/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	JRV
 *
 *	$Log:	man_mast.c,v $
 * Revision 2.2  92/04/29  15:08:38  trillium
 * Ohio Release
 * 
 *      Function:	- "main" for man_mast, the manager process
 *                        for the mandelbrot demo.  Some explanation
 *                        of the terminology use is in order first:
 *                        'grid' means the whole region of the complex
 *                        plane which is to be examined.  'region'
 *                        refers to a chunk of the grid which is farmed
 *                        off to the worker processes to be handled.
 *
 *      Returns:        - errno if fatal error terminated prog.
 */

#include <stdio.h>
#include <portable.h>
#include <typical.h>
#include <net.h>
#include "mandel.h"

/*
 * external variables
 */
extern int		errno;


/*
 * external functions
 */
extern double 		ttime();
void			terror();


/*
 * global variables
 */
char			*me;			/* tool name */


main(argc, argv)

int4			argc;
char			*argv[];

{
	struct nmsg	msg_h;
	float		grid_spec[4];		/* describes the whole grid */
	int4		reg_spec[4];		/* describes region by ul-lr */
	int4		grid_dim[2];		/* size of grid in pixels  */
	int4		region_dim[2];		/* size of region in pixels */
	char		*region;		/* holds the colors  */
	char		*reg_ptr;		/* a temp pointer */
	char		*answer;		/* collected results */
	int4		num_nodes;		/* count of itb nodes */
	int4		results = 0;	        /* count of results so far */
	int4		i, j;			/* loop counters */
	int4		myid;			/* master's node id */
	int4		region_size;       	/* size of region in bytes */
	int4		num_regions;		/* total number of regions */
	int4		more_regions = TRUE;	/* more to be assigned? */
	float		*floatdata;		/* temp areas to hold floats */
	double		start;			/* starting time */
	int4		mtvnode;		/* the node that will plot */

	me = argv[0];

	if (kinit(0)) {
		exit(errno);
	}

	num_nodes = getnitb();
	myid = getnodeid();
	get_grid_dimensions(grid_spec, grid_dim);
	get_region_dimensions(region_dim);

	printf("Which node should plot? (note, host=100) >>");
	fflush(stdout);
	scanf("%d", &mtvnode);

	if (rtvclr(mtvnode)) {
		terror(me);
		kexit(errno);
	}
	
	start = ttime();
/*
 *  Send translate and scale factors to each slave.
 */
	floatdata = (float *)msg_h.nh_data;
	floatdata[0] = (grid_spec[2]-grid_spec[0])/grid_dim[0];
	floatdata[1] = (grid_spec[1]-grid_spec[3])/grid_dim[1];
	floatdata[2] = grid_spec[0];
	floatdata[3] = grid_spec[3];
	msg_h.nh_event = EVENT_REQUEST;
	msg_h.nh_type = MESSAGE_WORK;
	msg_h.nh_flags = DFLTDATA;
	msg_h.nh_length = 0;
	for(i=1; i<num_nodes; i++) {
		msg_h.nh_node = i;
		if (nsend(&msg_h)) {
			terror(me);
			kexit(errno);
		}
	}
/*
 * Allocate space for grid and region.
 */
	answer = (char *) malloc(grid_dim[0]*grid_dim[1]);
	if (answer==NULL){
		terror(me);
		kexit(errno);
	}
	region_size = region_dim[0]*region_dim[1];
	region = (char *) malloc(region_size);
	if (answer==NULL){
		free(region);
		terror(me);
		kexit(errno);
	}
	num_regions = get_num_regions(grid_dim, region_dim);
/*
 *  Send a region to each node except the master.
 *  The worker should compute the values for that region
 *  and return an array of colors.
 */
	get_first_region(reg_spec);

	msg_h.nh_event = EVENT_REQUEST;
	msg_h.nh_type = MESSAGE_WORK;
	msg_h.nh_length = 0;
	msg_h.nh_flags = DINTMSG;
	msg_h.nh_data[0] = myid;
	for(i=1; ( (i<num_nodes) && (more_regions) ); i++) {
		msg_h.nh_node = i;
		msg_h.nh_data[1] = reg_spec[0];  /* pixel coordinates ul,lr  */
		msg_h.nh_data[2] = reg_spec[1];
		msg_h.nh_data[3] = reg_spec[2];
		msg_h.nh_data[4] = reg_spec[3];
		if (nsend(&msg_h)) {
			terror(me);
			kexit(errno);
		}
		more_regions = get_next_region(reg_spec);
	}
/*
 *  Continue to send work requests and gather results until
 *  all vectors have been received.
 */
	while (TRUE) {
		msg_h.nh_flags = DINTMSG;
		msg_h.nh_msg = (char *) region;
		msg_h.nh_event = EVENT_RESULT;
		msg_h.nh_type = MESSAGE_WORK;
		msg_h.nh_length = region_size;
		if (nrecv(&msg_h)) {
			terror(me);
			kexit(errno);
		}
		
		if (rtvrect(mtvnode, 
			    msg_h.nh_data[1], 
			    msg_h.nh_data[2], 
			    msg_h.nh_data[3]-msg_h.nh_data[1] + 1, 
			    msg_h.nh_data[4]-msg_h.nh_data[2] + 1,
			    region)) {

			terror(me);
			kexit(errno);
		}

		if (++results == num_regions)
			break;

		if (more_regions) {
			msg_h.nh_node = msg_h.nh_data[0];
			msg_h.nh_data[0] = myid;
			msg_h.nh_data[1] = reg_spec[0]; /* pixel coord's */
			msg_h.nh_data[2] = reg_spec[1];
			msg_h.nh_data[3] = reg_spec[2];
			msg_h.nh_data[4] = reg_spec[3];
			msg_h.nh_type = MESSAGE_WORK;
			msg_h.nh_event = EVENT_REQUEST;
			msg_h.nh_length = 0;
			msg_h.nh_flags = DINTMSG;
			if(nsend(&msg_h)) {
				terror(me);
				kexit(errno);
			}
			more_regions = get_next_region(reg_spec);
		}
	} 
/*    
 * We have all the answers now, so kill the workers.
 */
	msg_h.nh_flags = DINTMSG;
	msg_h.nh_event = EVENT_REQUEST;
	msg_h.nh_type = MESSAGE_DIE;
	msg_h.nh_length = 0;
	for (i=1; i<num_nodes; i++) {
		msg_h.nh_node = i;
		if(nsend(&msg_h)) {
			terror(me);
			kexit(errno);
		}
	}

	printf("Running time is %f seconds.\n", ttime()-start);
	kexit(0);
}
