/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	JRV
 *
 *	$Log:	man_slave.c,v $
 * Revision 2.2  92/04/29  15:09:20  trillium
 * Ohio Release
 * 
 *	Function:	- "main" for man_slave, the worker process
 *                         for the mandelbrot demo.
 */

#include <stdio.h>
#include <portable.h>
#include <typical.h>
#include <net.h>
#include "mandel.h"

/*
 * external variables
 */
extern int		errno;


/*
 * local variables
 */
static char		*me;			/* tool name */


main(argc, argv)
int4			argc;
char			*argv[];
{
	struct nmsg	msg_h;
	int4		ulx, uly;		/* upper-left point */
	int4		lrx, lry;   		/* lower-right point */
	char		*region;		/* ptr to array of colors */
	char		*ptr;			/* temp pointer */
	int4		region_size;		/* in bytes */
	int4		i, j;			/* loop counters */
	int4		myid;			/* node id of this worker */
	float		scale_x; 		/* used to convert from */
	float		scale_y;		/*   pixels to colors   */
	float		translate_x;		/*                      */
	float		translate_y;		/*                      */
	float		*floatdata;		/* alias to nh_data */

	me = argv[0];

	if (kinit(0)) {
		exit(errno);
	}
	myid = getnodeid();
/*
 * Receive scaling and translating factors for grid.
 */
	msg_h.nh_event = EVENT_REQUEST;
	msg_h.nh_type = MESSAGE_WORK;
	msg_h.nh_length = 0;
	msg_h.nh_flags = DFLTDATA;
	if (nrecv(&msg_h)) {
		terror(me);
		kexit(errno);
	}

	floatdata = (float *)msg_h.nh_data;
	scale_x = floatdata[0];
	scale_y = floatdata[1];
	translate_x = floatdata[2];
	translate_y = floatdata[3];
/*
 * Now loop, getting requests from the master
 */
	while (TRUE) {
		msg_h.nh_event = EVENT_REQUEST;
		msg_h.nh_type = MESSAGE_WORK;
		msg_h.nh_flags = DINTMSG;
		msg_h.nh_length = 0;
		if (nrecv(&msg_h)) {
			terror(me);
			kexit(errno);
		}
/*
 * If we get a DIE message, we are all done, so just kexit.
 */
		if (msg_h.nh_type == MESSAGE_DIE) {
			kexit(0);
		}
/*
 * If we get here, we've gotten a message requesting us to do
 * some more work.  Compute the given region, send the result back to 
 * the master process, and then wait for a new request.
 */
		ulx = msg_h.nh_data[1];
		uly = msg_h.nh_data[2];
		lrx = msg_h.nh_data[3];
		lry = msg_h.nh_data[4];
		region_size = (lrx-ulx+1) * (lry-uly+1);
		region = (char *)malloc((unsigned)region_size);
		
		if (region == NULL) {
			terror(me);
			kexit(errno);
		}
		ptr = region;

		for(j=uly; j <= lry; j++) {
			for(i=ulx; i <= lrx; i++) {
				*ptr++ = xy2color(scale_x * i + translate_x,
						  scale_y * j + translate_y);

			}
		}
		msg_h.nh_msg = (char *)region;
		msg_h.nh_length = region_size;
		msg_h.nh_event = EVENT_RESULT;
		msg_h.nh_flags = DINTMSG;
		msg_h.nh_node = msg_h.nh_data[0];
		msg_h.nh_data[0] = myid;
		if (nsend(&msg_h)) {
			free(region);
			terror(me);
			kexit(errno);
		}
		free(region);
	}
}



