/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	JRV
 *
 *	$Log:	region.c,v $
 * Revision 2.2  92/04/29  15:09:26  trillium
 * Ohio Release
 * 
 *      Module:         - region.c 
 * 
 *      Functions:      - void get_region_dimensions();
 *			  void get_grid_dimensions();
 *			  int4 get_data_file();
 *                        void get_first_region();
 *                        void get_grid_dimensions();
 *                        int4 get_next_region();
 *			  int4 get_num_regions();
 *			  void write_answer();
 *
 *      Description:    - This module contains functions
 *                        which manipulate the little regions (parts
 *                        of the whole graph) which are sent to  the
 *                        slave processes.
 */

#include <fcntl.h>
#include <stdio.h>
#include <portable.h>
#include <typical.h>
#include <freq.h>
#include <net.h>

/*
 * useful defines
 */
#define	MAX_FNAME	128

/*
 * external variables
 */
extern int4		errno;
extern char		*me;

/*
 * local variables
 */
static int4		x_dim, y_dim;		/* dimensions of the regions */
static int4		ulx, uly;		/* current region coords */
static int4		lrx, lry; 		/*   in pixels           */
static int4		grid_dim[2]; 		/* save pixel dimensions  */

 
/*
 *	get_grid_dimensions
 *
 *	Function:	This gets from the user the area on the complex plane
 *			which the program will compute.
 *	
 *	Accepts:	address of coordinates in complex plane
 *			address of number of pixels in grid
 *
 *	Returns:	nothing, but fills in the above values
 */
void get_grid_dimensions(grid, dim)
float			*grid;
int4			*dim;
{
	printf("\nSpecify the region of the complex plane to be plotted\n");
	printf("by entering the upper-left and lower-right coordinates.\n");

	printf("Enter the upper-left x (real) value >> ");
	fflush(stdout);
	scanf("%f", &grid[0]);
	printf("Enter the upper-left y (imaginary) value >> ");
	fflush(stdout);
	scanf("%f", &grid[1]);

	printf("Enter the lower-right x (real) value >> ");
	fflush(stdout);
	scanf("%f", &grid[2]);
	printf("Enter the lower-right y (imaginary) value >> ");
	fflush(stdout);
	scanf("%f", &grid[3]);

	printf("\nHow many pixels to plot in the X direction? >> ");
	fflush(stdout);
	scanf("%d", &dim[0]);
	printf("How many pixels to plot in the Y direction? >> ");
	fflush(stdout);
	scanf("%d", &dim[1]);

	grid_dim[0] = dim[0];
	grid_dim[1] = dim[1];
}


/*
 *	get_region_dimensions
 *
 *	Function:	This is used to determine the size of each region
 *			which will be handed off to the worker processes.
 *
 *	Accepts:	region dimensions, which it modifies as input
 *			by the user
 */
void get_region_dimensions(region_dim)
int4			*region_dim;
{
	printf("\nEnter the width (in pixels) of the partitions >> ");
	fflush(stdout);
	scanf("%d", &x_dim);
	printf("Enter the height (in pixels) of the partitions >> ");
	fflush(stdout);
	scanf("%d", &y_dim);
	region_dim[0] = x_dim;
	region_dim[1] = y_dim;
}


/*
 *	get_num_regions
 *
 *	Function:	figures out how many regions there will be, given
 *			the grid size and region size
 *
 *	Accepts:	the grid and regions dimensions
 *
 *	Returns:	the number of regions
 */ 
int4 get_num_regions(grid_dim, region_dim)
int4			*grid_dim;
int4			*region_dim;
{
	int4		num_regions_x;
	int4		num_regions_y;
	float		tmp_float;

	num_regions_x = (grid_dim[0])/(region_dim[0]);
	tmp_float = (float)(grid_dim[0])/(float)(region_dim[0]);
	if ((float)num_regions_x != tmp_float) {
		num_regions_x++;
	}

	num_regions_y = (grid_dim[1])/(region_dim[1]);
	tmp_float = (float)(grid_dim[1])/(float)(region_dim[1]);
	if ((float)num_regions_y != tmp_float) {
		num_regions_y++;
	}
	return(num_regions_x * num_regions_y);
}


/*
 *	get_data_file
 *
 *	Function:	prompts the user for the filename to save results in
 *
 *	Returns:	a valid (FILE *) if ok, NULL if not
 */ 
int4 get_data_file()
{
	char		file_name[MAX_FNAME];

	printf("File name in which to store results? >> ");
	fflush(stdout);
	scanf("%s", file_name);
	return(open(file_name, O_CREAT | O_WRONLY | O_1WAY, 0600));
}


/*
 *	write_answer
 *
 *	Function:	writes the results to disk if user wants to
 *
 *	Accepts:	ptr to buffer with answer
 */ 
void write_answer(answer, grid_dim)
char			*answer;
int4			*grid_dim;
{
	char		doit[80];
	int4		fp;			/* the output file */
	char 		line[80];		/* holds output line */
	int4		i,j;			/* loop counters */

	printf("\nDo you wish to save the results to disk? (y/n) >> ");
	fflush(stdout);
	fflush(stdin);
	scanf("%s", doit);
	if ((doit[0] == 'y') || (doit[0] == 'Y')) {
		fp = get_data_file();
		if (fp==NULL) {
			terror(me);
			kexit(errno);
		}

		printf("Now writing results to disk");
		fflush(stdout);
		for (i=0; i<grid_dim[0]; i++) {
			for(j=0; j<grid_dim[1]; j++) {
				sprintf(line, "x: %d, y: %d, count: %d\n",i,j,
					(int)(answer[i+j*(grid_dim[1]-1)]));
				write(fp, line, strlen(line));
			}
			printf(".");
			fflush(stdout);
		}
		close(fp);
		printf("OK.\n");
	}
}


/*
 *	get_first_region
 *
 *	Function:	This returns in reg_spec[0..3] the x and y pixel
 *			coordinates of the first region in the grid.  It
 *			also initializes the various current values used
 *			later in calls to this module.
 *
 *	Accepts:	array of integers which are filled in with the 
 *			region coordinates
 */
void get_first_region(reg_spec)
int4			*reg_spec;
{
	ulx = 0;
        uly = 0;
	lrx = x_dim-1;
	lry = y_dim-1;
	reg_spec[0] = ulx;
	reg_spec[1] = uly;
	reg_spec[2] = lrx;
	reg_spec[3] = lry;
}


/*
 *	get_next_region
 *
 *	Function:	gives a new piece of the grid to the caller
 *
 *	Accepts:	array of integers which are filled in with the
 *			region coordinates
 *
 *	Returns:	TRUE if this is a valid region, FALSE if there are
 *			 no more
 */ 
int4 get_next_region(reg_spec)
int4			*reg_spec;
{
	ulx += x_dim;
	lrx += x_dim;

	if (ulx >=grid_dim[0]) {
		ulx = 0;
		lrx = x_dim-1;
		uly += y_dim;
		lry += y_dim;
	}

	if (lrx >= grid_dim[0]) {      /* Clip RHS of region against grid. */
		lrx = grid_dim[0]-1;
	}
	if (lry >= grid_dim[1]) {      /* Clip bottom against grid. */
		lry = grid_dim[1]-1;
	}

	reg_spec[0] = ulx;
	reg_spec[1] = uly;
	reg_spec[2] = lrx;
	reg_spec[3] = lry;
    
	if (uly >= grid_dim[1]) {
		return(FALSE);
	}
	else {
		return(TRUE);
	}
}






