c
c	OSU Research Computing					Trillium
c
c	Copyright 1991 The Ohio State University
c
c	RKM
c	
c	$Log:	dglbsolv.f,v $
c Revision 2.2  92/04/24  15:35:54  trillium
c Ohio Release
c 
c Revision 2.1  91/05/02  14:46:51  gdburns
c Ohio release
c 
c

	subroutine DGLBSOLV(u,y,z,x,ierror)
c
c	Parallel Tridiagonal Linear Solver
c
c	Function:	To solve a larger system of equations by communicating
c			with neighboring nodes.
c
c	Input:		common block envar containing -- 
c				 nitb = number of nodes in topology
c				 mynode = node-id.
c				 firstnd = first node in chain
c				 lastnod = last node in chain
c				 npart  = size of assigned partition.
c			
c			u,y,z --  local solution vectors. 
c
c	Output:		x -- actual global solution vector.
c			ierror -- error indicator.
c
c	Trollius Constants
c
	parameter(NHDSIZE=8)
c
c 	Machine/Process Constants
c
	parameter(INTSIZE=4)
	parameter(IDLSIZE=8)
	parameter(IDDBLMSG=8388608)
	parameter(NOTEVENT=-2147483647)

c
c	Data Constraints
c	Some day I would like the user send in work-space buffers....
c
	parameter(MAXSIZE=1000)

c
c	Define events with numbers.
c
	parameter(IEVNEX = 100)
c
c	Programming Constants
c
	double precision ONE,ZERO
	parameter(ONE=1d0)
	parameter(ZERO=0d0)

	integer nitb,mynode,firstnd,lastnod,
     #		ndata(NHDSIZE),npart,ierror,
     #		nevent,ntype,nlength,nflags,ndsize,
     #		ndlevnt1,ndlevnt2,ndlrht,ndllft

	double precision k11,k12,k13,k21,k22,k23,g1,g2,
     #		rval1,rval2,rval3,rvar1,rvar2,rvar3,r1,r2,xl,xn,
     #		u(1),y(1),z(1),x(1),
     #		inbuf(3),outbuf(3)
c 
c	Omnibus common block.
c
	common/envar/nitb,mynode,firstnd,lastnod,npart

c
c 	In this module, simultaneous linear equations are solved are
c	solved across several processors arranged as a 1-d mesh. The
c	solution method is similar to Block Cyclic Reduction used to
c	solve block triangular systems. See Golub and Van Loan's 
c	"Matrix Computation" for further details.
c
c	Processor Topology:
c	0 --- 1 --- 2 --- 3 --- 4  --- ... --- nitb-1
c
c	Nodes "0" and "nitb-1" have an equation each while nodes 1 through
c	nitb-2 have two equations each. Thus the total number of equations to 
c	be solved are 2 * (nitb-1) where nitb = number of compute nodes.
c	Each node receives constants from its left or right neighbours and 
c	after internal elimination the constants are send to right and left 
c	neighbours. The middle node does a solution which is then migrated
c	back to the other nodes. 
c
c	For more details see "A method to parallelise Tridiagonal Solvers"
c	by Silvia Muller presented at the Fifth Distributed Memory Conference, 
c	April 1990.
c
c	In all the equations listed below for the sake of explanations, the 
c	vector v is used. Elements of vector v correspond to
c
c		v(1) 	= x(2;1)
c		v(2) 	= x(1;npart)
c		v(3)	= x(3;1)
c		v(4)	= x(2;npart)
c		.	.	.	
c		.	.	.	
c		v(d-1)	= x(d;1)
c		v(d)	= x(d-1;npart)
c

c
c	Some message passing infrastructure.
c
	isize = IDLSIZE
	nevent = IEVNEX
	nflags = IDDBLMSG
	ndsize = 0

	if (mynode .eq. firstnd) then
c
c	This is node 0. Determine if configuration has more than one node,
c	by checking if a route to node mynode+1 (=1) exists.
c
	   ndlrht = NOTEVENT
	    if ( nitb .gt. 1) then
		call GROUTE(mynode+1,ndlrht,ierr)
		if (ierr .ne. 0) then
		   ierror = -1
		   return
		endif
	    endif

	if ( ndlrht .ne. NOTEVENT ) then
c
c	Yes there is a node to the right.
c

c
c	Equations involved:
c	nitb = 1 : Do nothing
c	nitb >= 2 : k11 * v(1) + k12 * v(2) = g1
c

c
c	Determine the number of itb nodes.
c
		if ( nitb .eq. 2) then
			k11	= z(npart)
			k12	= ONE 
			g1	= u(npart)	
c
c 	Receive elimination information from second node,
c	also the last node, since there are only two nodes in the 
c	configuration.
c
			ntype =  0
			nlength = 3 * isize
		
			call DRECV(nevent,ntype,nlength,nflags,
     #				ndata,ndsize,inbuf(1),ierr)
			if (ierr .ne. 0) then
				ierror = -1
				return
			endif

			k21 	= inbuf(1)
			k22 	= inbuf(2)
			g2 	= inbuf(3)

c
c	Reduction Operations:
c	There exist only two equations to be solved one in each processor.
c
c	k11 * v(1) + k12 * v(2) = g1 		( 1 )
c	k21 * v(1) + k22 * v(2) = g2		( 2 )
c
c		Receive inbuf(1)=k21, inbuf(2)=k22 and inbuf(3)=g2
c		Eliminate v(2) from first equation
c		Solve for v(1) = x(2;1)
c
			r1 = g1 * k22 - g2 * k12
			r2 = k11 * k22 - k21 * k12
			xn = r1 / r2
c
c		Send the computed value x(2;1) to right neighbour
c
			outbuf(1) = xn
			nlength = isize
			call DSEND(ndlrht,nevent,ntype,nlength,
     #				nflags,ndata,ndsize,outbuf(1),ierr)
			if (ierr .ne. 0) then
				ierror = -1
				return
			endif	
		else
c
c		Number of nodes > 2.
c
c		There is a difference in the computation when the number
c		of nodes is greater then 2. The right node has three
c		terms in its equations rather than two as was the case 
c		earlier. Node 1 is not the solution node anymore. Like
c		any other  node it sends the constants of its equation
c		k11, k12 and g1 to node 2.
c
			k11	   = z(npart)
			k12 	   = ONE
			g1  	   = u(npart)

			outbuf(1)  = k11
			outbuf(2)  = k12
			outbuf(3)  = g1

			ntype = 1
			nlength = 3 * isize
			call DSEND(ndlrht,nevent,ntype,nlength,
     #				nflags,ndata,ndsize,outbuf(1),ierr)
			if (ierr .ne. 0) then
				ierror = -1
				return
			endif	
c
c		Receive the value of v(2) or x(1:npart) in short from right 
c		neighbour node 1.
c

			ntype = 0
			nlength = isize
			call DRECV(nevent,ntype,nlength,nflags,ndata,
     #				ndsize, inbuf(1),ierr)
			if (ierr .ne. 0) then
				ierror = -1
				return
			endif
c
c		Determine xn = x(2;1). 
c
			xn = (g1 - k12 * inbuf(1))/ k11
		endif
		else
			xn = ZERO
		endif
	
c
c		Determine vector. Not the most optimal way, but I am
c		lazy.
c
		do 10 i = 1,npart
			x(i) = u(i) - xn *  z(i)
10		continue


	elseif ((mynode .eq. lastnod) .and. (firstnd .ne. lastnod)) then
c
c	This is the last node in the chain. Get the route info for the 
c	left neighbour.
c
		ndllft = NOTEVENT
		call GROUTE(mynode-1,ndllft,ierr)
		if (ierr .ne. 0) then
			ierror = -1
			return
		endif

c
c	Equations: 
c		d = 2(nitb-1)
c		v(d-1) + y(npart;1) * v(d) = u(npart;1)
c
c		k11 * v(d-1) + k12 * v(d) = g1
c	Not a solution node. Sole purpose is to send the constants of its
c	equation and obtain the value of v(d-1) from its left neighbour.
c
		k11 = ONE
		k12 = y(1)
		g1  = u(1) 

		outbuf(1) = k11
		outbuf(2) = k12
		outbuf(3) = g1
c
c	Send elements to node, lastnode-1 on the left.
c
		ntype = 2
		nlength = 3 * isize
		call DSEND(ndllft,nevent,ntype,nlength,nflags,
     #			ndata(1),ndsize,outbuf,ierr)

		if (ierr .ne. 0) then
			ierror = -1
			return
		endif
c
c	Receive solution from left, node lastnod-1 which is 
c	xn for node lastnode-1.
c
		ntype = 0
		nlength = isize
		call DRECV(nevent,ntype,nlength,nflags,
     #				ndata(1),ndsize,inbuf,ierr)

		if (ierr .ne. 0) then
			ierror = -1
			return
		endif  
c
c	use 	k11 * v(d-1) + k12 * v(d) = g1
c	solve for xl = v(d) = (g1 - k11 * v(d-1)) / k12
c
		xl = (g1 - k11 * inbuf(1) ) / k12
		
		do 20 i = 1,npart
			x(i) = u(i) - xl * y(i)
20		continue

	else
c
c	This is not a boundary node, let us call this node i, 0 < i< lastnod.
c

c
c	Get some routing information so that communcation could be carried
c	out with neighbours.
c
		call GROUTE(mynode+1,ndlrht,ierr)
		if (ierr .ne. 0) then
			ierror = -1
			return
		endif

		call GROUTE(mynode-1,ndllft,ierr)
		if (ierr .ne. 0) then
			ierror = -1
			return
		endif

c
c	Set values of the "k" and g constants.
c
		k11 = ONE
		k12 = y(1)
		k13 = z(1)
		g1  = u(1)
		k21 = y(npart)
		k22 = z(npart)
		k23 = ONE
		g2  = u(npart)

		midnod = (nitb-1) / 2

	if (mynode .ne. midnod) then
c
c	The case of the middle node is different.
c	If this node is not a middle node then all it does is pass 
c	its "k" and "g" constants to its left or right neighbours depending
c	on its position in the linear array.  It then receives the
c	solution value xl or xn.

c
c	Obtain elimination information from node (i-1) or node (i+1).
c
		ntype = 0
		nlength = 3 * isize
		call DRECV(nevent,ntype,nlength,nflags,ndata,ndsize,
     #			inbuf(1),ierr)
		if (ierr .ne. 0) then
			ierror = -1
			return
		endif

		if ( mynode .lt. midnod ) then

c
c	Reduction operations performed on node i, 0 < i < midnod :
c	k22 * v(p-1)+ v(p) * k23 		   		= g2 (i-1;npart)
c
c	k11 * v(p-1)+ k12 * v(p)   + k13 * v(p+1) 		= g1 (i;1)
c	              k21 * v(p)  + k22 * v(p+1) + k23 * v(p+2) = g2 (i;npart)
c
c	Receive inbuf(1)=k22, inbuf(2)=k23 and inbuf(3)=g2 from node i-1.
c	Eliminate v(p-1) from second equation.
c	Eliminate v(p) from third equation.
c	Send values of outbuf(1)=k22, outbuf(2)=k23 and outbuf(3)=g2 to 
c	node i+1.
c	Some bookkeeping: Message type = 1 ( downstream )
c			  ndlevnt1 = dl_out for going right
c			  ndlevnt2 = dl_out for going left
c
			k12 = k12 * inbuf(1) - k11 * inbuf(2)
			k13 = inbuf(1) * k13
			g1 = g1 * inbuf(1) - inbuf(3) * k11

			k22 = k12 * k22 - k21 * k13
			k23 = k23 * k12
			g2  = g2 * k12 - g1 * k21

			ndlevnt1 = ndlrht
			ndlevnt2 = ndllft
			ntype = 1

			outbuf(1) = k22
			outbuf(2) = k23
			outbuf(3) = g2
		else

c
c	Reduction operations performed on node i, midnod < i < lastnod :  
c	k11 * v(p-1) + k12 * v(p)   + k13 * v(p+1) = g1	(i;1)
c	k21 * v(p)   + k22 * v(p+1) + k23 * v(p+2) = g2	(i;npart)	
c
c		       k11 * v(p+1) + k12 * v(p+2) = g1 (i+1;1)
c
c	Receive inbuf(1)=k11, inbuf(2)=k12  and inbuf(3)=g1 from node i+1
c	Eliminate v(p+2) from equation 2 
c	Eliminate v(p+1) from equation 1
c	Send k11, k21 and g1 to node i-1.
c	Some bookkeeping: Message type = 2 ( upstream )
c			  ndlevnt1 = dl_out for going left
c			  ndlevnt2 = dl_out for going right
c
			k21 = k21 * inbuf(2)
			k22 = inbuf(2) * k22 - inbuf(1) * k23
			g2 = g2 * inbuf(2) - inbuf(3) * k23
			
			k11 = k11 * k22
			k12 = k12 * k22 - k21 * k13
			g1  = g1 * k22 - g2 * k13

			ndlevnt1 = ndllft
			ndlevnt2 = ndlrht
			ntype = 2

			outbuf(1) = k11
			outbuf(2) = k12
			outbuf(3) = g1
		endif

c
c	Send to the right or left neighbour the "k" constants and "g" value.
c
			nlength = 3 * isize
			call DSEND(ndlevnt1,nevent,ntype,nlength,
     #				nflags,ndata(1),ndsize,outbuf,ierr)
			if (ierr .ne. 0) then
			    ierror = -1
			    return
			endif
c
c	i < midnod
c	Receive vector components from neighbour.
c

			ntype = 0
			nlength = isize
			call DRECV(nevent,ntype,nlength,
     #				nflags,ndata(1),ndsize,inbuf,ierr)
			if (ierr .ne. 0) then
			    ierror = -1 
			    return
			endif

		if ( mynode .lt. midnod ) then
c
c	Received v(p+2) and now calculate v(p+1), then v(p). Dispatch v(p)
c	to node i-1.
c
			xn = (g2 - k23 * inbuf(1)) / k22
			xl = (g1 - xn * k13) / k12
			outbuf(1) = xl
		elseif( mynode .gt. midnod ) then
c
c	Received v(p-1) and now calculate v(p), then v(p+1). Dispatch v(p+1)
c	to node i+1.
c
			xl = (g1 - k11 * inbuf(1)) / k12
			xn = (g2 - k21 * xl) / k22
			outbuf(1) = xn
		endif
c
c 	Send values to left or right neigbour, the value of v(p) or xl.
c
			ntype  = 0
			nlength = isize
			call DSEND(ndlevnt2,nevent,ntype,nlength,
     #				nflags,ndata(1),ndsize,outbuf,ierr)
			if (ierr .ne. 0) then
			    ierror = -1
			    return
			endif
		else
c
c	MIDDLE NODE is the solution node. This where the BUCK STOPS.
c	Obtain elimination information from nodes (i-1) and (i+1)
c	This receipt is stateless and the sender ( left or right neighbour)
c	is determined from the type of message. Computations are done only
c	after both neighbours send their "k" and "g" values. Notice that
c	type is set to 0.
c
			ntype = 0
			nlength = 3 * isize
			call DRECV(nevent,ntype,nlength,nflags,
     #				ndata,ndsize,inbuf(1),ierr)
			if (ierr .ne. 0) then
				ierror = -1
				return
			endif

c
c	Is it from the right or the left neighbour. No checks are done if 
c	messages are received from the same direction twice. Dont think it 
c	is necessary, but shall do it anyway in the C version.
c
			if ( ntype .eq. 1 ) then
				rval1 = inbuf(1)
				rval2 = inbuf(2)
				rval3 = inbuf(3)
			elseif ( ntype .eq. 2) then
				rvar1 = inbuf(1)
				rvar2 = inbuf(2)
				rvar3 = inbuf(3)
			endif

c
c	Wait again.
c
			ntype = 0
			nlength = 3 * isize
			call DRECV(nevent,ntype,nlength,
     #				nflags,ndata(1),ndsize,inbuf,ierr)
			if (ierr .ne. 0) then
			    ierror = -1
			    return
			endif
 
			if ( ntype .eq. 1 ) then
				rval1 = inbuf(1)
				rval2 = inbuf(2)
				rval3 = inbuf(3)
			elseif ( ntype .eq. 2) then
				rvar1 = inbuf(1)
				rvar2 = inbuf(2)
				rvar3 = inbuf(3)
			endif

c
c	Do elimination again.
c	k22 * v(p-1) + k23 * v(p) 		  	        = g2 (i-1;npart)
c
c	k11 * v(p-1) + k12 * v(p)   + k13 * v(p+1) 		 = g1	(i;1)
c	               k21 * v(p)   + k22 * v(p+1) + k23 * v(p+2)= g2 (i;npart)
c
c		                      k11 * v(p+1) + k12 * v(p+2) = g1 (i+1;1)
c
c	Receive k22, k23 and g2 from node i-1
c	Receive k11, k12 and g1 from node i+1
c	Eliminate v(p-1) from equation 2
c	Eliminate v(p) from equation 3
c	Eliminate v(p+2) from equation 3
c
			k12 = k12 * rval1 - k11 * rval2
			k13 = rval1 * k13
			g1 = g1 * rval1 - rval3 * k11

			k22 = k12 * k22 - k21 * k13
			k23 = k23 * k12
			g2  = g2 * k12 - g1 * k21

c
c	Solve for xl=v(p) and xn=v(p+1)
c
			r1 = g2 * rvar2 - rvar3 * k23
			r2 = k22 * rvar2 - rvar1 * k23
			xn = r1 / r2
			xl = (g1 - k13 * xn) / k12

c
c	Send xl to left neighbour and xn to right neighbour.
c
			ntype = 0
			outbuf(1) = xn
			nlength = isize
			call DSEND(ndlrht,nevent,ntype,nlength,
     #				nflags,ndata(1),ndsize,outbuf,ierr)

			if (ierr .ne. 0) then
			    ierror = -1
			    return
			endif
		
			outbuf(1) = xl
			nlength = isize
			call DSEND(ndllft,nevent,ntype,nlength,
     #				nflags,ndata(1),ndsize,outbuf,ierr)

			if (ierr .ne. 0) then
			    ierror = -1
			    return
			endif
		endif
c
c	Determine your own vector.
c
			do 30 i = 1,npart
				x(i) = u(i) - xl * y(i) - xn * z(i)
30			continue

	endif

	ierror = 0
c
c	Whew that was quite a code.!!!
c

	return
	end
