c
c	OSU Research Computing					Trillium
c
c	Copyright 1991 The Ohio State University
c
c	RKM
c	
c	$Log:	tdgtsl.f,v $
c Revision 2.2  92/04/24  15:35:59  trillium
c Ohio Release
c 
c Revision 2.1  91/05/02  14:46:53  gdburns
c Ohio release
c 
c

c
c	Parallel Tridiagonal Linear Solver
c
c	Function:	to solve a tridiagonal system of equations on
c			a node.
c	Input:		n 	- order of matrix	
c			c,d,e 	- diagonals of tridiagonal matrix
c					c , e 	-- arrays of dimension n-1
c					d	-- array of dimension n
c			b 	- right hand side -- array of dimension n
c			ireq	- type of request
c	Output: 	common block envar
c				nitb	-- # of itb nodes
c				mynode	-- node id executing this module
c				ifrstnd	-- first node-id in chain
c				lastnod	-- last node-id in chain
c				npart	-- size assigned to mynode 
c			b 	- the solution vector
c			info	- error indicator
c				  info = 0 if there was neither a singularity
c				 	nor a Trollius problem.
c				  info = k if d(k) = 0.
c				  info = -1 if there was a trollius problem.
c

c
c	Locally defined routines	-- none
c

c
c	Externally defined routines
c		IGNITB(),IGNDID()	-- Trollius goodies
c		dlclsolv		-- local solver
c		dglbsolv		-- global solver
c		dgtsl			-- linpack tridiagonal solver
c
	
	subroutine TDGTSL(n,c,d,e,b,ireq,info)
c
c	Data constraints. At the moment we are handling upto 1000x1000 matrices.
c	This would change in the future.
c
	parameter(IFIRST=1)
	parameter(MAXSIZE=1000)

	integer n,ireq,info,ierr,ifrstnd,ising,lastnod,mynode,nitb
     #		,npart

	double precision c(1),d(1),e(1),b(1),u(MAXSIZE),y(MAXSIZE),
     #		z(MAXSIZE)

	common/envar/nitb,mynode,ifrstnd,lastnod,npart


	ierr = 0
	nitb = IGNITB() 
	mynode = IGNDID()
	ifrstnd = 0
	lastnod = nitb - 1
	npart = n
c
c	Check for spurious node numbers.
c
	if ((mynode .lt. ifrstnd) .or. (mynode .gt. lastnod)) then
	    info = -1
            return
	endif

c
c	Solve locally first. Notice that the only kind of errors
c	are the singularity kind. It is strictly local.
c
	call DLCLSOLV(c,d,e,b,u,y,z,ising)
	if( ising .ne. 0)  then
	   info = ising
	   return
	endif

c
c	Do a global solve now.
c
	call DGLBSOLV(u,y,z,b,ierr)
	if( ierr .eq. -1) then
	    info = -1
	    return
	endif

	info = 0
	return
	end
