c
c	OSU Research Computing					Trillium
c
c	Copyright 1991 The Ohio State University
c
c	RKM
c	
c	$Log:	zglbsolv.f,v $
c Revision 2.2  92/04/24  15:36:02  trillium
c Ohio Release
c 
c Revision 2.1  91/05/02  14:46:54  gdburns
c Ohio release
c 
c

	subroutine ZGLBSOLV(u,y,z,x,ierror)
*
*	Parallel Tridiagonal Linear Solver
*
*	Function:	To solve a larger system of equations by communicating
*			with neighboring nodes.
*
*	Input:		common block envar containing -- 
*				 nitb = number of nodes in topology
*				 mynode = node-id.
*				 ifrstnd = first node in chain
*				 lastnod = last node in chain
*				 npart  = size of assigned partition.
*			
*			u,y,z --  local solution vectors. 
*
*	Output:		x -- actual global solution vector.
*			ierror -- error indicator.
*
*	Trollius Constants
*
	parameter(NHDSIZE=8)
*
* 	Machine/Process Constants
*
	parameter(INTSIZE=4)
	parameter(IZLSIZE=16)
	parameter(IFLTMSG=4194304)
	parameter(NOTEVENT=-2147483647)

*
*	Data Constraints
*
	parameter(MAXSIZE=1000)

*
*	Define events with numbers.
*
	parameter(IEVNEX = 100)


	integer ndata(NHDSIZE),nitb,mynode,npart,ierror

	complex*16 k11,k12,k21,k22,u(1),y(1),z(1),x(1),inbuf(3)
     #	,outbuf(3),k13,k23,g1,g2,rval1,rval2,rval3,rvar1,rvar2,
     #	rvar3,r1,r2,xl,xn

	common/envar/nitb,mynode,ifrstnd,lastnod,npart

*
* 	In this module, simultaneous linear equations are solved. The
*	solution method is similar to cyclic reduction (?).
*	
*
*	In all the equations listed below, the vector v is used. Elements
*	of vector v corresponds to
*
*		v(1) 	= x(2;1)
*		v(2) 	= x(1;npart)
*		v(3)	= x(3;1)
*		v(4)	= x(2;npart)
*		.	.	.	
*		.	.	.	
*		v(d-1)	= x(d;1)
*		v(d)	= x(d-1;npart)
*

	nevent = IEVNEX
	nflags = IFLTMSG
	ndsize = 0

	if ((mynode .eq. ifrstnd)) then
*
*	This is node 0. Check if the configuration has more than one node,
*	by checking if a route to node mynode+1 (=1) exists.
*
	   ndlrht = NOTEVENT
	    if ( nitb .gt. 1) then
		call GROUTE(mynode+1,ndlrht,ierr)
		if (ierr .ne. 0) then
		   call TERROR("solver(glbsolv)")
		   ierror = -1
		   return
		endif
	    endif

	if ( ndlrht .ne. NOTEVENT ) then
*
*	Yes there is a node to the right.
*

*
*	Equations involved:
*       z(1;npart) * v(1) + v(2) = u(1;npart) 		(1)
*
*	Reductions Done : on node 0.
*	v(1) + ( 1.0 /z(1;npart) ) * v(2) = u(1;npart) /z(1;npart)
*
*	nitb = 1 : Do nothing
*	nitb >= 2 : k11 * v(1) + v(2) = g1
*

*
*	Determine the number of itb nodes.
*
		if ( nitb .eq. 2) then
			k11	= z(npart)
			k12	= (1d0,0d0) 
			g1	= u(npart)	
*
*	 	Receive elimination information from second node, i.e., 
*		also the last node.
*
		ntype =  0
		nlength = 3 * IZLSIZE
		
		call DRECV(nevent,ntype,nlength,nflags,ndata,ndsize,
     #			inbuf(1),ierr)
			if (ierr .ne. 0) then
				ierror = -1
				return
			endif

		k21 	= inbuf(1)
		k22 	= inbuf(2)
		g2 	= inbuf(3)

		xn = (g1 * k22 - g2 * k12) / (k11 * k22 - k21 * k12)
			
		outbuf(1) = xn
		nlength = IZLSIZE
		call DSEND(ndlrht,nevent,ntype,nlength,
     #			nflags,ndata,ndsize,outbuf(1),ierr)
		if (ierr .ne. 0) then
			ierror = -1
			return
		endif	

		else
*
*	Number of nodes > 2
*
*		Send the necessary information to node 2, the right neighbor.
*
			k11	   = z(npart)
			k12 	   = (1d0,0d0)
			g1  	   = u(npart)

			outbuf(1)  = k11
			outbuf(2)  = k12
			outbuf(3)  = g1

			ntype = 1
			nlength = 3 * IZLSIZE
			call DSEND(ndlrht,nevent,ntype,nlength,
     #				nflags,ndata,ndsize,outbuf(1),ierr)
			if (ierr .ne. 0) then
				ierror = -1
				return
			endif	
*
*	Receive the value of v(2) or x(npart) in short from right 
*	neighbour node 1.
*

			ntype = 0
			nlength = IZLSIZE
			call DRECV(nevent,ntype,nlength,nflags,ndata,
     #				ndsize, inbuf(1),ierr)
			if (ierr .ne. 0) then
				ierror = -1
				return
			endif
			xn = (g1 - k12 * inbuf(1))/ k11
		endif
		else
			xn = (0d0,0d0)
		endif
	
*
*	Determine vector.
*
		do 10 i = 1,npart
			x(i) = u(i) - xn *  z(i)
10		continue

****************************************************************************

	elseif ((mynode .eq. lastnod) .and. (ifrstnd .ne. lastnod)) then
*
*	This is the last node in the chain. Get the route info for the 
*	left neighbour.
*
		ndllft = NOTEVENT
		call GROUTE(mynode-1,ndllft,ierr)
		if (ierr .ne. 0) then
			ierror = -1
			return
		endif

*
*	Equations involved:
*	d = 2(nitb-1)
*		v(d-1) + y(npart;1) * v(d) = u(npart;1)
*
*		k11 * v(d-1) + k12 * v(d) = g1
*
		k11 = (1d0,0d0) 
		k12 = y(1)
		g1  = u(1) 

		outbuf(1) = k11
		outbuf(2) = k12
		outbuf(3) = g1
*
*	Send elements to node, lastnode-1 on the left.
*
		ntype = 2
		nlength = 3 * IZLSIZE
		call DSEND(ndllft,nevent,ntype,nlength,nflags,
     #			ndata(1),ndsize,outbuf,ierr)

		if (ierr .ne. 0) then
			call TERROR("psgtsl(glbsolv)")
			ierror = -1
			return
		endif
*
*	Receive solution from left, node lastnod-1 which is 
*	xn for node lastnode-1.	 For this node it has v(d) in hand.
*
		ntype = 0
		nlength = IZLSIZE
		call DRECV(nevent,ntype,nlength,nflags,
     #				ndata(1),ndsize,inbuf,ierr)

		if (ierr .ne. 0) then
			call TERROR("psgtsl(glbsolv)")
			ierror = -1
			return
		endif  
*
*	use 	k11 * v(d-1) + k12 * v(d) = g1
*	solve for xl = v(d) = (g1 - k11 * v(d-1)) / k12
*
		xl = (g1 - k11 * inbuf(1) ) / k12
		
		do 20 i = 1,npart
			x(i) = u(i) - xl * y(i)
20		continue

	else
*
*	This is not a boundary node, let us call this node i, 0 < i< lastnod.
*
* Equation on node i-1:
* i = 2 ; p=2
* v(1) * y(i-1;npart) + v(2) * z(i-1;npart) + v(3) =u(i-1;npart)
* i > 2
* p = 2(i-2)
* v(p-1) * y(i-1;npart) + v(p) * z(i-1;npart) + v(p+1) =u(i-1;npart)
*	
* Equations on node i:
* p = 2(i-1)
* v(p-1) + v(p) * y(i;1) + v(p+1) * z(i;1) = u(i;1)
* v(p) * y(i;npart) + v(p+1) * z(i;npart) + v(p+2) =u(i;npart)
*
* Equation on node i+1:
* i != lastnode-1 ; p = 2(i-1)
* v(p) + v(p+1) * y(i+1;1) + v(p+2) * z(i+2;1) 	=u(i+1;1)	
* i = lastnode-1; p =2(lastnode-1)
* v(p) + v(p+1) * y(i+1;1) 			= u(i+1;1)
*

		midnod = (nitb-1) / 2
*
*	Get some routing information.
*
		call GROUTE(mynode+1,ndlrht,ierr)
		if (ierr .ne. 0) then
			call TERROR("solver(glbsolv)")
			ierror = -1
			return
		endif

		call GROUTE(mynode-1,ndllft,ierr)
		if (ierr .ne. 0) then
			call TERROR("solver(glbsolv)")
			ierror = -1
			return
		endif

*
*	Reduction operations performed on node i, 0 < i < midnod :
*	v(p-1)+ v(p)   * k11(i) + v(p+1) * k12(i) = g1 		(i;1)
*	v(p)  + v(p+1) * k21(i) + v(p+2) * k21(i) = g2		(i;npart)
*
*	Reduction operations performed on node i, midnod < i < lastnod :  
*	v(p-1) * k11(i) + v(p) *  k12(i) +   v(p+1) = g1(i)	(i;1)
*	v(p)   * k21(i) + v(p+1) *  k22(i) + v(p+2) = g2(i)	(i;npart)	
*	

		k11 = (1d0,0d0)
		k12 = y(1)
		k13 = z(1)
		g1  = u(1)
		k21 = y(npart)
		k22 = z(npart)
		k23 = (1d0,0d0)
		g2  = u(npart)

*
*	Obtain elimination information from node (i-1)/(i+1)
*
	if (mynode .ne. midnod) then
		ntype = 0
		nlength = 3 * IZLSIZE
		call DRECV(nevent,ntype,nlength,nflags,ndata,ndsize,
     #			inbuf(1),ierr)
		if (ierr .ne. 0) then
			call TERROR("solver(glbsolv)")
			ierror = -1
			return
		endif

		if ( mynode .lt. midnod ) then

			k12 = k12 * inbuf(1) - k11 * inbuf(2)
			k13 = inbuf(1) * k13
			g1 = g1 * inbuf(1) - inbuf(3) * k11

			k22 = k12 * k22 - k21 * k13
			k23 = k23 * k12
			g2  = g2 * k12 - g1 * k21

			ndlevnt1 = ndlrht
			ndlevnt2 = ndllft
			ntype = 1

			outbuf(1) = k22
			outbuf(2) = k23
			outbuf(3) = g2
		else

			k21 = k21 * inbuf(2)
			k22 = inbuf(2) * k22 - inbuf(1) * k23
			g2 = g2 * inbuf(2) - inbuf(3) * k23
			
			k11 = k11 * k22
			k12 = k12 * k22 - k21 * k13
			g1  = g1 * k22 - g1 * k13

			ndlevnt1 = ndllft
			ndlevnt2 = ndlrht
			ntype = 2

			outbuf(1) = k11
			outbuf(2) = k12
			outbuf(3) = g1
		endif

*
*	Send to the right/left neighbour, i+1
*
			nlength = 3 * IZLSIZE
			call DSEND(ndlevnt1,nevent,ntype,nlength,
     #				nflags,ndata(1),ndsize,outbuf,ierr)
			if (ierr .ne. 0) then
			    call TERROR("solver(glbsolv)")
			    ierror = -1
			    return
			endif
*
*	i < midnod
*	Receive values from neighbour i+1, value of v(p+2).
*	Calculate v(p+1), then v(p).
*

			ntype = 0
			nlength = IZLSIZE
			call DRECV(nevent,ntype,nlength,
     #				nflags,ndata(1),ndsize,inbuf,ierr)
			if (ierr .ne. 0) then
			    call TERROR("solver(glbsolv)") 
			    ierror = -1 
			    return
			endif

		if ( mynode .lt. midnod ) then
			xn = (g2 - k23 * inbuf(1)) / k22
			xl = (g1 - xn * k13) / k12
			outbuf(1) = xl
		elseif( mynode .gt. midnod ) then
			xl = (g1 - k11 * inbuf(1)) / k12
			xn = (g2 - k21 * xl) / k22
			outbuf(1) = xn
		endif
*
* 	Send values to left/right neigbour, the value of v(p) or xl.
*
			ntype  = 0
			nlength = IZLSIZE
			call DSEND(ndlevnt2,nevent,ntype,nlength,
     #				nflags,ndata(1),ndsize,outbuf,ierr)
			if (ierr .ne. 0) then
			    call TERROR("solver(glbsolv)")
			    ierror = -1
			    return
			endif
		else
*
*	MIDDLE NODE
*	Obtain elimination information from node (i-1)/(i+1)
*
			ntype = 0
			nlength = 3 * IZLSIZE
			call DRECV(nevent,ntype,nlength,nflags,
     #				ndata,ndsize,inbuf(1),ierr)
			if (ierr .ne. 0) then
				call TERROR("solver(glbsolv)")
				ierror = -1
				return
			endif

			if ( ntype .eq. 1 ) then
				rval1 = inbuf(1)
				rval2 = inbuf(2)
				rval3 = inbuf(3)
			elseif ( ntype .eq. 2) then
				rvar1 = inbuf(1)
				rvar2 = inbuf(2)
				rvar3 = inbuf(3)
			endif

			ntype = 0
			nlength = 3 * IZLSIZE
			call DRECV(nevent,ntype,nlength,
     #				nflags,ndata(1),ndsize,inbuf,ierr)
			if (ierr .ne. 0) then
			    call TERROR("solver(glbsolv)")
			    ierror = -1
			    return
			endif
 
			if ( ntype .eq. 1 ) then
				rval1 = inbuf(1)
				rval2 = inbuf(2)
				rval3 = inbuf(3)
			elseif ( ntype .eq. 2) then
				rvar1 = inbuf(1)
				rvar2 = inbuf(2)
				rvar3 = inbuf(3)
			endif

			k12 = k12 * rval1 - k11 * rval2
			k13 = rval1 * k13
			g1 = g1 * rval1 - rval3 * k11

			k22 = k12 * k22 - k21 * k13
			k23 = k23 * k12
			g2  = g2 * k12 - g1 * k21

			r1 = g2 * rvar2 - rvar3 * k23
			r2 = k22 * rvar2 - rvar1 * k23
			xn = r1 / r2
			xl = (g1 - k13 * xn) / k12

			ntype = 0
			outbuf(1) = xn
			nlength = IZLSIZE
			call DSEND(ndlrht,nevent,ntype,nlength,
     #				nflags,ndata(1),ndsize,outbuf,ierr)

			if (ierr .ne. 0) then
			    call TERROR("solver(glbsolv)")
			    ierror = -1
			    return
			endif
		
			outbuf(1) = xl
			nlength = IZLSIZE
			call DSEND(ndllft,nevent,ntype,nlength,
     #				nflags,ndata(1),ndsize,outbuf,ierr)

			if (ierr .ne. 0) then
			    call TERROR("solver(glbsolv)")
			    ierror = -1
			    return
			endif
		endif
			do 30 i = 1,npart
				x(i) = u(i) - xl * y(i) - xn * z(i)
30			continue

	endif

	ierror = 0

	return
	end
