c
c	OSU Research Computing					Trillium
c
c	Copyright 1991 The Ohio State University
c
c	RKM
c	
c	$Log:	zlclsolv.f,v $
c Revision 2.2  92/04/24  15:36:04  trillium
c Ohio Release
c 
c Revision 2.1  91/05/02  14:46:55  gdburns
c Ohio release
c 
c

c
c	Parallel Tridiagonal Linear Solver
c
c	Function:	Called by routine tzgtsl().
c			To solve three system of equations locally with 
c			no communication.
c	
c	Input:		common block envar
c				nitb	-- # of itb nodes
c				mynode	-- node id executing this module
c				ifrstnd	-- first node-id in chain
c				lastnod	-- last node-id in chain
c				npart	-- size assigned to mynode 
c			c,d,e 	- diagonals of tridiagonal matrix
c					c , e 	-- arrays of dimension n-1
c					d	-- array of dimension n
c			b 	-- right hand side -- array of dimension n
c
c	Output:		u,y,z 	-- local solution vectors of dimension n-1.
c			info	-- error indicator
c				info > 0 -- singularity occured
c				info = 0 -- peachy-pie situation.
c

c
c	Externally defined routines
c		zgtsl	-- Linpack solver
c
	subroutine ZLCLSOLV(c,d,e,b,u,y,z,info)

	parameter(MAXSIZE=1000)

	integer ierr,info,ifrstnd,lastnod,mynode,nitb,npart

	complex*16 c(1),d(1),e(1),b(1),c1(MAXSIZE),d1(MAXSIZE),
     #		e1(MAXSIZE),b1(MAXSIZE),u(1),y(1),z(1)

	common/envar/nitb,mynode,ifrstnd,lastnod,npart

c
c 	It is required to solve three systems of linear equations.
c	Let A = [ c,d,e], then the three systems to solve are
c	A u = h, 	where h = [b(1),...,b(npart)] 	(S1)
c	A y = f, 	where f = [c(1),...,0]		(S2)
c	A z = g 	where h = [0,...,e(npart)]	(S3)
c
	
c
c	Initialize
c
	do 10 i = 1,npart
		c1(i) = c(i)
		d1(i) = d(i)
		e1(i) = e(i)
		b1(i) = b(i)
		u(i) = (0d0,0d0)
		y(i) = (0d0,0d0)
		z(i) = (0d0,0d0)
10	continue

c
c	Solve system S1 on all nodes.
c
	call ZGTSL(npart,c1,d1,e1,b1,ierr)
	if ( ierr .ne. 0 ) then
		info = ierr
		return
	endif

c
c	Zgtsl destroys the input vectors, so c1,d1,e1 have to be restored every 
c	time zgtsl is called.
c

	do 30 i = 1,npart
		u(i)  = b1(i)
		c1(i) =c(i)
		d1(i) =d(i)
		e1(i) =e(i)
		b1(i) = (0d0,0d0)
30	continue

c
c	Solve second system S2 on all nodes except the first one.
c
	if ( mynode .ne. ifrstnd) then
		b1(1) = c(1)

		call ZGTSL(npart,c1,d1,e1,b1,ierr)
		if ( ierr .ne. 0 ) then
			info = ierr
			return
		endif

		do 40 i = 1,npart
			y(i) = b1(i)
			c1(i) =c(i)
			d1(i) =d(i)
			e1(i) =e(i)
			b1(i) = (0d0,0d0)
40		continue
	endif
c
c	Solve third system S3 on all nodes except the last one.
c
	if ( mynode .ne. lastnod) then
		b1(npart) = e(npart)

		call ZGTSL(npart,c1,d1,e1,b1,ierr)
		if ( ierr .ne. 0 ) then
			info = ierr
			return
		endif

		do 50 i = 1,npart
			z(i) = b1(i)
50		continue
	endif

	info = 0

	return

	end
