c
c       OSU Research Computing					Trillium
c
c       Copyright 1991 The Ohio State University
c
c       VAR
c
c	$Log:	wavcrs.f,v $
c Revision 2.2  92/04/24  15:43:05  trillium
c Ohio Release
c 
c
c       Function:       Solve the one-dimensional wave equation problem.
c
c       Version:        F3 : CUBIX+CROS program in FORTRAN
c
c       Reference:      Solving Problems on Concurrent Processors - Vol I
c                       Fox, Johnson et al
c
	program wavcrs

c	Global declarations and define constants.

	integer INITIA, UPDATE, QUIT
	integer INFUNIT, OUTFUNIT

	parameter (MAXSIZE = 500)
	parameter (INITIA = 1, UPDATE = 2, QUIT = 3)

c	Select standard input/output or file input/output.
c	IOSTREAM = 0 for I/O from STDIN/STDOUT,
c		 = 1 for I/O from FILEIN/FILEOUT

	parameter (IOSTREAM = 1)
	parameter (ISTDIN = 5, ISTDOUT = 6)
	parameter (INFUNIT = 7, OUTFUNIT = 8)

	double precision values(0:MAXSIZE+1), oldval(0:MAXSIZE+1),
     #  newval(0:MAXSIZE+1)

	character*6 ifname, ofname

c	Common blocks.

	common /extcom/ myproc, nproc
	common /nptcom/ nmin, nleft, ntot
	common /valcom/ values, oldval, newval
	common /files/ ifname, ofname
	common /iodev/ nindev, noutdev
	common /cros/ klchan, krchan

c	Local variables.

	double precision delt

        call KINIT(0,nerr)
        if (nerr.ne.0) then
                print *, 'node(main): Err in kinit',nerr
                call EXIT(-1)
        endif

	call USINGL(ISTDIN, nerr)
        if (nerr.ne.0) then
                print *, 'node(main): Err in usingl : ',nerr
                call KEXIT(-1)
        endif

	call USINGL(ISTDOUT, nerr)
        if (nerr.ne.0) then
                print *, 'node(main): Err in usingl : ',nerr
                call KEXIT(-1)
        endif

	if (IOSTREAM .eq. 0) then
		nindev = ISTDIN
		noutdev = ISTDOUT
	else
		nindev = INFUNIT
		noutdev = OUTFUNIT

ccccccc BUG-1: Reading any input in singular mode expects excess
ccccccc	       carriage returns, one for each processor.
ccccccc	       Hence parameter input is from file instead of stdin.

c		print *,'Enter input file name (exactly 6 char) : '
c		read(*,10) ifname
c10		format(a6)

c		print *,'Enter output file name (exactly 6 char) : '
c		read(*,11) ofname
c11		format(a6)

c	Caution: Make sure that the file resides in the current
c	working directory which can be set by command "tcd".

		ifname = 'INFILE'
 		ofname = 'OUTFIL'

		print *, 'Input file : ',ifname

		open (nindev, file = ifname, status = 'old',
     *		form = 'formatted', err = 901)

		call USINGL(nindev, nerr)
        	if (nerr.ne.0) then
                  print *, 'node(main): Err in usingl : ',nerr
                  call KEXIT(-1)
        	endif

ccccccc	BUG-2: Opening new file with status = 'new' causes problems.

c		open (noutdev, file = ofname, status = 'new',
c    *		form = 'formatted', err = 902)

ccccccc	BUG-3: UMULTI() for file output causes processors to 
ccccccc        to overwrite each others data in the file.
ccccccc	       Hence data points are output on stdout 
ccccccc        instead of file.

 		open (noutdev, file = ofname, 
     *		form = 'formatted', err = 902)
 		print *, 'Opened outfile : ', ofname

 		call USINGL(noutdev, nerr)
         	if (nerr.ne.0) then
                  print *, 'node(main): Err in usingl : ',nerr
                  call KEXIT(-1)
         	endif
	endif

        write(*,9)
 9      format(/'This is a 1-D wave equation solver -'/)

	call gparam

	call decompose

	last = 0
 20	call getask (last, ntask)
 	if (ntask.ne.QUIT) then
                if (ntask.eq.INITIA) then
			last = INITIA
                        call initze (jlen, delt)
                elseif (ntask.eq.UPDATE) then
                        call dostep (jlen, delt)
                        call outres
                endif
                goto 20 
        endif

	call KEXIT(0)

        stop

 901	nerr = IERRNO()
	print *, 'Err in opening infile ',ifname,' : ',nerr
	stop
 902	nerr = IERRNO()
	print *, 'Err in opening outfile ',ofname,' : ',nerr

        end

c       --------------------------------------------------------------------

        subroutine gparam
 
	common /extcom/ myproc, nproc
	common /nptcom/ nmin, nleft, ntot
	common /valcom/ values, oldval, newval
	common /files/ ifname, ofname
	common /iodev/ nindev, noutdev
	common /cros/ klchan, krchan

        nproc = IGNITB()
 
        myproc = IGNDID()
        if ((myproc .lt. 0) .or. (myproc .gt. nproc-1)) then
                print *,'Err: Node ids must be 0 thru (nproc-1)'
                call KEXIT(0)
        endif
c       print *,'nproc = ', nproc, ' myproc = ',myproc

        return
        end
 
c       --------------------------------------------------------------------

        subroutine decompose
 
	common /extcom/ myproc, nproc
	common /nptcom/ nmin, nleft, ntot
	common /valcom/ values, oldval, newval
	common /files/ ifname, ofname
	common /iodev/ nindev, noutdev
	common /cros/ klchan, krchan

c	Mapping topology onto cartesian grid.

	nerr = KGRDIN (1, nproc)
        if (nerr.eq.-1) then
                print *, 'node(decompose): Err in kgrdin',nerr
                call EXIT(-1)
        endif

c	Determine left and right channels.

	klchan = KGRDCH (myproc, 0, -1)
        if (klchan.eq.-1) then
                print *, 'node(decompose): Err in kgrdch',nerr
                call EXIT(-1)
        endif

	krchan = KGRDCH (myproc, 0, 1)
        if (krchan.eq.-1) then
                print *, 'node(decompose): Err in kgrdch',nerr
                call EXIT(-1)
        endif

	if (myproc .eq. 0) then 
		klchan = 0
	endif
	if (myproc .eq. (nproc-1)) then 
		krchan = 0
	endif

	return
	end

c       --------------------------------------------------------------------

        subroutine getask (last, ntask)
 
	integer INITIA, UPDATE, QUIT

	common /extcom/ myproc, nproc
	common /nptcom/ nmin, nleft, ntot
	common /valcom/ values, oldval, newval
	common /files/ ifname, ofname
	common /iodev/ nindev, noutdev
	common /cros/ klchan, krchan

        parameter (INITIA=1, UPDATE=2, QUIT=3)

	character*1 resp

 10	write(*,100)
 100    format(/'Select one of the following:'/
     #  ' 1 ... to initialize a new problem'/
     #  ' 2 ... to compute points and print results'/
     #  ' 3 ... to terminate program'/)

	read (nindev, 110) resp
 110	format(a)

	if (resp.eq.'1') then
		ntask = INITIA
	elseif (resp.eq.'2') then
		ntask = UPDATE
	elseif (resp.eq.'3') then
		ntask = QUIT
	else
		write(*,120)
 120		format(/'*** Invalid input - please try again ***'/)
		goto 10
	endif

	print *,'option = ',ntask
	if ((last.eq.0).and.(ntask.eq.UPDATE)) then
		print *,'Problem not initialized'
		goto 10
	endif

	return
	end

c	--------------------------------------------------------------------

        subroutine initze (jlen, delt)
        integer jlen
        double precision delt

	common /extcom/ myproc, nproc
	common /nptcom/ nmin, nleft, ntot
	common /valcom/ values, oldval, newval
	common /files/ ifname, ofname
	common /iodev/ nindev, noutdev
	common /cros/ klchan, krchan

        parameter (MAXSIZE = 500)

        double precision values(0:MAXSIZE+1), oldval(0:MAXSIZE+1),
     #  newval(0:MAXSIZE+1)

        double precision pi, fac, x
        parameter (pi=3.14159265)

 10     write(*,100)
 100    format(/' Enter total number of points :'/)

	read (nindev, 110) ntot
 110	format(i5)

 	print *, 'ntot = ',ntot
c 	write(noutdev, 120) ntot
c120	format('ntot = ', i5)

        if (ntot.eq.0) then
                write(*,130)
 130    	format(/' *** Zero points specified ***'/)
                call KEXIT(0)
        endif

        nmin = ntot / nproc
        nleft = mod (ntot, nproc)
        if (nmin.ge.MAXSIZE) then
                write(*,140)
 140    	format(/' *** Too many points specified ***'/)
                goto 10
        endif
 
        write(*,150)
 150    format(/' Enter time advance parameter :'/)

        read(nindev, 160) delt
 160	format(f10.2)

 	write(*, 165) delt
 165	format('delt = ',f8.3)

c       	write(noutdev, 170) delt
c170	format('delt = ',f8.3)
 
        fac = 2.0 * pi
        k = 0
 
        do 180 i = 0, nproc-1
                npts = nmin
                if (i .lt. nleft) npts = nmin + 1
 
                if (myproc .eq. i) then
                        jlen = npts
c			print *,'initial data in ',myproc
                        do 190 j = 1, npts
                                x = dble(k) / dble(ntot+1)
                                values(j) = dsin(fac*x)
                                k = k+1
c                               write(*,185) values(j)
c185                            format(f30.5)
 190                     continue
                else
                        k = k + npts
                endif
 180     continue
 
        do 200 i = 1, jlen
 200             oldval(i) = values(i)
 
        return
        end
 
c       --------------------------------------------------------------------
 
        subroutine dostep(jlen, del)
        integer jlen
        double precision del
 
        integer dblsize
        parameter (dblsize = 8)
        parameter (MAXSIZE = 500)
 
        double precision values(0:MAXSIZE+1), oldval(0:MAXSIZE+1),
     #  newval(0:MAXSIZE+1)

	common /extcom/ myproc, nproc
	common /nptcom/ nmin, nleft, ntot
	common /valcom/ values, oldval, newval
	common /files/ ifname, ofname
	common /iodev/ nindev, noutdev
	common /cros/ klchan, krchan

        double precision c, delx, pfac

        write(*,100)
 100    format(/' Enter number of time steps :'/)

        read(nindev, 110) nstep
 110	format(i5)

 	print *, 'nstep = ', nstep
c 	write(noutdev, 120) nstep
c120	format('nstep = ', i4) 
 
        c = 1.0
        delx = 1.0
        pfac = (c*del/delx)**2

        do 130 i = 1, nstep

            nerr = KCSHIF (values(0), klchan, dblsize, 
     #			values(jlen), krchan, dblsize)
            if (nerr.eq.-1) then
                print *, 'node(', myproc, ')(dostep):Err in kcshif-1',
     #                  nerr
                call KEXIT(-1)
            endif
 
            nerr = KCSHIF (values(jlen+1), krchan, dblsize, 
     #			values(1), klchan, dblsize)
            if (nerr.eq.-1) then
                print *, 'node(', myproc, ')(dostep):Err in kcshif-2',
     #                  nerr
                call KEXIT(-1)
            endif
 
            do 140 j = 1, jlen
                if ((myproc .eq. 0) .and. (j .eq. 1)) then
                    newval(j) = 0.0
                elseif ((myproc.eq.(nproc-1)).and.(j.eq.jlen)) then
                    newval(j) = 0.0
                else
                    newval(j) = 2.0*values(j) - oldval(j)
     #                  + ( values(j+1) - 2.0*values(j) + values(j-1))
     #                  * pfac
                endif
 140         continue
 
            do 150 j = 1, jlen
		oldval(j) = values(j)
 150            values(j) = newval(j)
 
 130     continue
 
        return
        end

c       --------------------------------------------------------------------

        subroutine outres

        parameter (MAXSIZE = 500)
	parameter (ISTDOUT = 6)

        double precision values(0:MAXSIZE+1), oldval(0:MAXSIZE+1),
     #  newval(0:MAXSIZE+1)

	common /extcom/ myproc, nproc
	common /nptcom/ nmin, nleft, ntot
	common /valcom/ values, oldval, newval
	common /files/ ifname, ofname
	common /iodev/ nindev, noutdev
	common /cros/ klchan, krchan

        if (myproc .lt. nleft) then
            npts = nmin+1
        else
            npts = nmin
        endif

         write(*, 100)
 100    format(/'Data points on the wave :'/)

        write(noutdev, 105)
 105    format(/'Data points on the wave :'/)

  	call UMULTI(ISTDOUT, nerr)
  	if (nerr .ne. 0) then
  		print *, 'Err in umulti: ', nerr
  		call KEXIT(-1)
  	endif

  	call UMULTI(noutdev, nerr)
  	if (nerr .ne. 0) then
  		print *, 'Err in umulti: ', nerr
  		call KEXIT(-1)
  	endif

c	With CUBIX, it is ensured that the data points will be output
c	in the order of the processor numbers.

  	write(*, 110) (values(j), j=1,npts)
 110    format(f30.5)

  	write(noutdev, 120) (values(j), j=1,npts)
 120    format(f30.5)

 	call USINGL(ISTDOUT, nerr)
 	if (nerr .ne. 0) then
 		print *, 'Err in usingl: ', nerr
 		call KEXIT(-1)
 	endif

	call TCLOSE(nindev, nerr)
c 	call TCLOSE(noutdev, nerr)
 
        return
        end
 
 
c       --------------------------------------------------------------------

