c
c	OSU Research Computing, architects and contractors	Trillium
c
c	VAR
c
c	$Log:	wavfcp.f,v $
c Revision 2.2  92/04/24  15:43:16  trillium
c Ohio Release
c 
c Revision 2.1  91/05/02  14:00:23  gdburns
c initial version
c 
c
c	Function:	Solve the one-dimensional wave equation problem.
c
c	Version:	F1 : Host-node programs in FORTRAN
c
c	Reference: 	Solving Problems on Concurrent Processors - Vol I
c			Fox, Johnson et al
c
c	Host Program
 
	program wavfcp

c	Global declarations and define constants.

	integer intsize, dblsize
	integer INITIA, UPDATE, QUIT

	parameter (MAXSIZE = 500)
	parameter (IEVTASK=11, IEVNTOT=12, IEVDELT=13, IEVNSTEP=14,
     #		   IEVCHECK=15, IEVRES=16, IEVOUT=100)
c	Do not use events from (IEVOUT+1) thru (IEVOUT+nproc)
	parameter (INITIA=1, UPDATE=2, QUIT=3)
	parameter (intsize = 4, dblsize = 8)

c	Common blocks.

	common /extcom/ myproc, nproc
	common /nptcom/ nmin, nleft, ntot
	common /msgcom/ nevent, ntype, nlength, nflags,	
     #				ndata, ndsize

	write(*,9)
 9	format(/'This is a 1-D wave equation solver -'/)

	call KINIT(0,nerr)
	if (nerr.ne.0) then
                print *, 'host(main): Err in kinit',nerr
                call EXIT(-1)
        endif

	call setvar

	ntype = 0
	nlength = 0
	nflags = 0
	ndata = 0
	ndsize = 0

c	Get task options and perform corresponding task.

	last = 0
 10	call getask (last, ntask)
	if (ntask.ne.QUIT) then
		if (ntask.eq.INITIA) then
			last = INITIA
			call initcp
		elseif (ntask.eq.UPDATE) then
			call dostep
			call outres
		endif
		goto 10
	endif

	call KEXIT(0)

	stop
	end

c	--------------------------------------------------------------------

	subroutine setvar

c	Find number of processors from bnail file and verify their
c	nodeids.

	common /extcom/ myproc, nproc
	parameter (NOTEVENT = -2147483647)

	nproc = IGNITB()

c	print *,'nproc = ', nproc

c	Check if compute nodeids are 0,1,2,...,(nproc-1)

	nevent = 0
	do 10 i = 0, nproc-1
	   call GROUTE(i, nevent, nerr)
	   if (nerr .eq. NOTEVENT) then
	     print *,'*** Node',i,' does not exist - verify bnail file'
	     call KEXIT(-1)
	   endif
 10	continue

	return
	end

c	--------------------------------------------------------------------

	subroutine getask (last, ntask)

c	Get the task option selected and send this to all compute nodes.

	integer INITIA, UPDATE, QUIT

        common /extcom/ myproc, nproc
        common /msgcom/ nevent, ntype, nlength, nflags,
     #                  ndata, ndsize
 
	parameter (IEVTASK=11)
        parameter (INITIA=1, UPDATE=2, QUIT=3)
	parameter (intsize=4)

	character*1 resp

c	Read task option.

 10	write(*,100)
	read(*,200) resp
	if (resp.eq.'1') then
		ntask = INITIA
	elseif (resp.eq.'2') then
		ntask = UPDATE
	elseif (resp.eq.'3') then
		ntask = QUIT
	else
		write(*,101)
		goto 10
	endif

	if ((last.eq.0).and.(ntask.eq.UPDATE)) then
		print *,'Problem not initialized'
		goto 10
	endif

	nlength = intsize

	do 300 i = 0, nproc-1
	    	nevent = IEVTASK
c		print *,'sending task to node i'
		itask = IRW(ntask)
		call NSEND(i, nevent, ntype, nlength, nflags,
     #			   ndata, ndsize, itask, nerr)
        	if (nerr.ne.0) then
        	    print *, 'host(getask): Err in nsend(ntask)',
     #			nerr
		    call TERROR('host')
        	    call KEXIT(-1)
        	endif
 300	continue

	return

 100	format(/'Select one of the following:'/
     #  ' 1 ... to initialize a new problem'/
     #  ' 2 ... to compute points and print results'/
     #  ' 3 ... to terminate program'/)

 101	format(/'*** Invalid input - please try again ***'/)
 200	format(a)

	end

c	--------------------------------------------------------------------

	subroutine initcp

c	Get parameters ntot and delt. Send these to all nodes.

	parameter (MAXSIZE = 500)

	common /extcom/ myproc, nproc

	common /nptcom/ nmin, nleft, ntot
        common /msgcom/ nevent, ntype, nlength, nflags,
     #                  ndata, ndsize

	integer dblsize
	double precision delt, ddelt
	parameter (dblsize=8, intsize=4)
	parameter (IEVNTOT=12, IEVDELT=13)
	double precision FRD


 10	write(*,100)
	read(*,*) ntot
	if (ntot.eq.0) then
		write(*,103)
		Call KEXIT(-1)
	endif
	nmin = ntot / nproc
	nleft = mod (ntot, nproc)
	if (nmin.ge.MAXSIZE) then
		write(*,101)
		goto 10
	endif

	write(*,102)
	read(*,*) delt

        nevent = IEVNTOT
	nlength = intsize

        do 300 i = 0, nproc-1
c	    print *,'sending npts = ',ntot,' to node ',i
	    itot = IRW(ntot)
            call NSEND(i, nevent, ntype, nlength, nflags,
     #                 ndata, ndsize, itot, nerr)
            if (nerr.ne.0) then
            	print *, 'host(initcp): Err in nsend(ntot)',nerr
            	call KEXIT(-1)
            endif
 300	continue
 
	nevent = IEVDELT
	nlength = dblsize

        do 400 i = 0, nproc-1
c	    print *,'sending delt = ',delt,' to node ',i
	    ddelt = FRD(delt)
            call NSEND(i, nevent, ntype, nlength, nflags,
     #                 ndata, ndsize, ddelt, nerr)
            if (nerr.ne.0) then
                print *, 'host(initcp): Err in nsend(delt)',nerr
                call KEXIT(-1)
            endif
 400    continue

	return

 100	format(/' Enter total number of points :'/)
 101	format(/' *** Too many points specified ***'/)
 102	format(/' Enter time advance parameter :'/)
 103	format(/' *** Zero points specified ***'/)

	end

c	--------------------------------------------------------------------

	subroutine dostep

c	Get parameter nstep, send this to all nodes.

	common /extcom/ myproc, nproc

        common /msgcom/ nevent, ntype, nlength, nflags,
     #                  ndata, ndsize

	integer check
	parameter (intsize = 4)
	parameter (IEVNSTEP = 14, IEVCHECK = 15)

	write(*,100)
	read(*,*) nstep

        nevent = IEVNSTEP
        nlength = intsize

        do 200 i = 0, nproc-1
c	    print *,'sending nstep = ',nstep,' to node ',i
	    istep = IRW(nstep)
            call NSEND(i, nevent, ntype, nlength, nflags,
     #                 ndata, ndsize, istep, nerr)
            if (nerr.ne.0) then
                print *, 'host(dostep): Err in nsend(nstep)',nerr
                call KEXIT(-1)
            endif
 200    continue

        nevent = IEVCHECK
        nlength = intsize

c	Receive check message from each of 'nproc' node processes.
 
        do 300 i = 0, nproc-1
c	    print *,'waiting for done msg from node ',i
            call NRECV(nevent, ntype, nlength, nflags, ndata,
     #                 ndsize, check, nerr)
            if (nerr.ne.0) then
                print *, 'host(dostep): Err in nrecv(check)',nerr
                call KEXIT(-1)
            endif
	    check = IRW(check)
 300    continue
 
	return

 100	format(/' Enter number of time steps :'/)

	end

c	--------------------------------------------------------------------

	subroutine outres

c	Get results from nodes in order and print them.

	parameter (MAXSIZE = 500)

	common /extcom/ myproc, nproc

        common /nptcom/ nmin, nleft, ntot 
        common /msgcom/ nevent, ntype, nlength, nflags,
     #                  ndata, ndsize

	double precision values(0:MAXSIZE+1)
	integer dblsize
	parameter (intsize = 4, dblsize = 8)
	parameter (IEVTASK = 11, IEVRES = 16, IEVOUT=100)
	double precision FRD

c	Send the request message and receive data from the 'nproc' 
c	processes. 

	write(*,9)
 9	format(/'Data points on the wave :'/)
	do 10  i = 0, nproc-1

c	    Send request message.
c	    To have this application workable on both the multi-node
c	    Trollius and the Stand-alone version, the event (see below)
c	    is assigned the value (IEVOUT+i) because :
c	    Stand-alone Trollius does not distinguish processes waiting
c	    on the same event. In order that the Host node receive and
c	    print the final data from the pseudo-compute nodes in
c	    ascending order, we need to specify distinct event numbers
c	    for each process. 

	    nevent = IEVOUT+i
	    nlength = 0
	    ntype = 0
c	    print *,'sending getres msg to node ',i
	    itask = IRW(ntask)
	    call NSEND(i, nevent, ntype, nlength, nflags,
     #                     ndata, ndsize, itask, nerr)
	    if (nerr.ne.0) then
                print *, 'host(getask): Err in nsend(ntask)',nerr
                call KEXIT(-1)
            endif

c	Receive data.

	    nevent = IEVRES
	    nlength = (nmin+1)*dblsize

	    ntype = 0
	    npts = nmin
	    if (i .lt. nleft) npts = nmin+1

            call NRECV(nevent, ntype, nlength, nflags, ndata,
     #                 ndsize, values(1), nerr)
            if (nerr.ne.0) then
                print *, 'host(outres): Err in nrecv(data)',nerr
                call KEXIT(-1)
            endif

c	Convert data from transputer to local.
	    do 20 j = 1, npts
		values(j) = FRD(values(j))
 20	    continue

	    write(*,100) (values(j), j=1,npts)
	    write(*,102) 
 10	continue

	return

 100	format(f30.5)
 102	format(/)
	end

c	--------------------------------------------------------------------


