c
c       OSU Research Computing, architects and contractors      Trillium
c
c       VAR
c
c       $Log:	wavnod.f,v $
c Revision 2.2  92/04/24  15:43:18  trillium
c Ohio Release
c 
c Revision 2.1  91/05/02  14:00:29  gdburns
c initial version
c 
c
c       Function:       Solve the one-dimensional wave equation problem.
c
c       Version:        F1 : Host-node programs in FORTRAN
c
c       Reference:      Solving Problems on Concurrent Processors - Vol I
c                       Fox, Johnson et al
c
c       Node Program

	program wavnod

c	Global declarations and define constants.

	integer INITIA, UPDATE, QUIT
	integer intsize, dblsize

	parameter (MAXSIZE = 500)
	parameter (INITIA=1, UPDATE=2, QUIT=3)
        parameter (IEVTASK=11, IEVNTOT=12, IEVDELT=13, IEVNSTEP=14,
     #             IEVCHECK=15, IEVRES=16, IEVOUT=100)
        parameter (intsize = 4, dblsize = 8)


	double precision values(0:MAXSIZE+1), oldval(0:MAXSIZE+1),
     #  newval(0:MAXSIZE+1)

c	Common blocks.

	common /extcom/ myproc, nproc
	common /nptcom/ nmin, nleft, ntot
	common /valcom/ values, oldval, newval
        common /msgcom/ nnode, nevent, ntype, nlength, nflags,
     #                  ndata, ndsize

c	Local functions.

	double precision delt
	integer mlen

        call KINIT(0,nerr)
        if (nerr.ne.0) then
                print *, 'node(main): Err in kinit',nerr
                call KEXIT(-1)
        endif

	call gparam

        ntype = 0
        nlength = 0
        nflags = 0
        ndsize = 0
 
 10	call getask (ntask)
 	if (ntask.ne.QUIT) then
                if (ntask.eq.INITIA) then
                        call initze (mlen, delt)
                elseif (ntask.eq.UPDATE) then
                        call dostep (mlen, delt)
                        call outres
                endif
                goto 10
        endif

	call KEXIT(0)

        stop
        end

c       --------------------------------------------------------------------

        subroutine gparam

c	Get nproc, myproc.
 
        common /extcom/ myproc, nproc
 
        nproc = IGNITB()
 
        myproc = IGNDID()
        if ((myproc .lt. 0) .or. (myproc .gt. nproc-1)) then
                print *,'*** Node ids must be 0 thru (nproc-1) ***'
                print *,'Terminate and rerun "wavnod"'
        endif
c       print *,'nproc = ', nproc, ' myproc = ',myproc
        return
        end
 
c       --------------------------------------------------------------------

        subroutine getask (ntask)

c	Get task option from host.
 
        common /extcom/ myproc, nproc
        common /msgcom/ nnode, nevent, ntype, nlength, nflags,
     #                  ndata, ndsize
 
        parameter (intsize = 4)
        parameter (IEVTASK = 11)
 
        nevent = IEVTASK
        nlength = intsize
        ntype = 0
	ndsize = 0
 
c       print *,'node ',myproc,' waiting for task'
        call NRECV(nevent, ntype, nlength, nflags, ndata,
     #             ndsize, ntask, nerr)
        if (nerr.ne.0) then
            print *,'node(',myproc,')(getask): Err in nrecv(task)',
     #          nerr
            call TERROR('node')
            call KEXIT(-1)
        endif

        return
        end
 
c       --------------------------------------------------------------------
 
        subroutine initze (jlen, delt)

c	Receive npts, delt from host.

        integer jlen
        double precision delt

        common /extcom/ myproc, nproc

        common /msgcom/ nnode, nevent, ntype, nlength, nflags,
     #                  ndata, ndsize

        parameter (MAXSIZE = 500)
        parameter (IEVNTOT = 12, IEVDELT = 13)

        double precision values(0:MAXSIZE+1), oldval(0:MAXSIZE+1),
     #  newval(0:MAXSIZE+1)
        common /valcom/ values, oldval, newval
        common /nptcom/ nmin, nleft, ntot
 
        integer dblsize
        double precision pi, fac, x
        parameter (pi=3.14159265, intsize=5, dblsize=8)
 
        nevent = IEVNTOT
        nlength = intsize
        ntype = 0
	ndsize = 0
 
        call NRECV(nevent, ntype, nlength, nflags, ndata,
     #             ndsize, ntot, nerr)
        if (nerr.ne.0) then
            print *, 'node(', myproc, ')(initze): Err in nrecv(ntot)',
     #          nerr
            call KEXIT(-1)
        endif

c       print *,'npts = ',ntot
 
        nmin = ntot / nproc
        nleft = mod (ntot, nproc)
 
        fac = 2.0 * pi
        k = 0
 
        do 20 i = 0, nproc-1
                npts = nmin
                if (i .lt. nleft) npts = nmin + 1
 
                if (myproc .eq. i) then
                        jlen = npts
c                       print *,'initial data in ',myproc
                        do 30 j = 1, npts
                                x = dble(k) / dble(ntot+1)
                                values(j) = dsin(fac*x)
                                k = k+1
c                               write(*,31) values(j)
c31                             format(f30.5)
 30                     continue
                else
                        k = k + npts
                endif
 20     continue
 
        nevent = IEVDELT
        nlength = dblsize
        ntype = 0
	ndsize = 0
 
        call NRECV(nevent, ntype, nlength, nflags, ndata,
     #             ndsize, delt, nerr)
        if (nerr.ne.0) then
            print *, 'node(', myproc, ')(initze): Err in nrecv(delt)',
     #          nerr
            call KEXIT(-1)
        endif

        do 40 i = 1, jlen
 40             oldval(i) = values(i)
 
        return
        end
 
c       --------------------------------------------------------------------
 
        subroutine dostep(ilen, del)

c	Update nodal points for nstep iterations.

        integer ilen
        double precision del
 
        common /extcom/ myproc, nproc
        common /msgcom/ nnode, nevent, ntype, nlength, nflags,
     #                  ndata(8), ndsize
 
        parameter (MAXSIZE = 500)
        parameter (IEVNSTEP = 14, IEVCHECK = 15)
 
        double precision values(0:MAXSIZE+1), oldval(0:MAXSIZE+1),
     #  newval(0:MAXSIZE+1)
        common /valcom/ values, oldval, newval
 
        double precision c, delx, pfac
        integer dblsize, check
        parameter (intsize=4, dblsize = 8)
 
        nevent = IEVNSTEP
        nlength = intsize
        ntype = 0
	ndsize = 0
 
        call NRECV(nevent, ntype, nlength, nflags, ndata,
     #             ndsize, nstep, nerr)
        if (nerr.ne.0) then
            print *, 'node(', myproc, ')(dostep): Err in nrecv(nstep)',
     #                  nerr
            call KEXIT(-1)
        endif
c       print *,'nstep = ',nstep
 
        c = 1.0
        delx = 1.0
        pfac = (c*del/delx)**2
 
        do 20 i = 1, nstep
 
            call kcshif (values(0), (myproc-1), values(ilen),
     #                  (myproc+1), nerr)
            if (nerr.ne.0) then
                print *, 'node(', myproc, ')(dostep): Err in kcshif-1',
     #                  nerr
                call KEXIT(-1)
            endif
 
            call kcshif (values(ilen+1), (myproc+1), values(1),
     #                  (myproc-1), nerr)
            if (nerr.ne.0) then
                print *, 'node(', myproc, ')(dostep): Err in kcshif-2',
     #                  nerr
                call KEXIT(-1)
            endif
 
            do 30 j = 1,ilen
                if ((myproc .eq. 0) .and. (j .eq. 1)) then
                    newval(j) = 0.0
                elseif ((myproc .eq. nproc-1) .and. (j .eq. ilen)) then
                    newval(j) = 0.0
                else
                    newval(j) = 2.0*values(j) - oldval(j)
     #                  + ( values(j+1) - 2.0*values(j) + values(j-1) )
     #                  * pfac
                endif
 30         continue
 
            do 40 j = 1,ilen
                oldval(j) = values(j)
 40             values(j) = newval(j)
 
 20     continue
 
        check = 0

        nevent = IEVCHECK
        nlength = intsize
        ntype = 0
	ndsize = 0

        nhost = IGORGN()
        call NSEND(nhost, nevent, ntype, nlength, nflags, 
     #                  ndata, ndsize, check, nerr)
        if (nerr.ne.0) then
            print *, 'node(', myproc, ')(dostep): Err in nsend(check)',
     #                  nerr
            call KEXIT(-1)
        endif

        return
        end

c       --------------------------------------------------------------------

        subroutine kcshif (inval, inindex, outval, outindex, nerr)

c	Exchange boundary nodal values with neighbour nodes.
 
        double precision inval, outval
        integer inindex, outindex, nerr
        integer dblsize
        parameter (dblsize = 8)
 
        integer nrerr, nwerr
        common /extcom/ myproc, nproc
        common /msgcom/ nnode, nevent, ntype, nlength, nflags,
     #                  ndata, ndsize

	nerr = 0
	nrerr = 0
	nwerr = 0
 
        ntype = 0
        nlength = dblsize
	ndsize = 0
 
        if ( mod(myproc,2) .eq. 0 ) then
                                 
c       Even node index.
 
c           Send to outindex only if outindex in (0,nproc-1).
 
            if ((outindex .ge. 0) .and. (outindex .le. nproc-1)) then
                nevent = myproc * 65536 + outindex
c               print *,'node',myproc, ' --> ', outindex,
c    #                  ' on',nevent,' :',outval

                call NSEND(outindex, nevent, ntype, nlength, nflags,
     #                  ndata, ndsize, outval, nwerr)
            endif
 
c           Receive from inindex only if inindex in (0,nproc-1).
 
            if ((inindex .ge. 0) .and. (inindex .le. nproc-1)) then
                nevent = inindex * 65536 + myproc
c               print *,'node',myproc, ' <-- ',inindex, ' on',
c    #                  nevent
                call NRECV(nevent, ntype, nlength, nflags, 
     #                     ndata, ndsize, inval, nrerr)
            endif
 
        else
 
c       Odd node index.
        
c           Receive from inindex only if inindex in (0,nproc-1).

            if ((inindex .ge. 0) .and. (inindex .le. nproc-1)) then
                nevent = inindex * 65536 + myproc
c               print *,'node',myproc, ' <-- ', inindex, ' on',
c    #                  nevent
                call NRECV(nevent, ntype, nlength, nflags, 
     #                     ndata, ndsize, inval, nrerr)
            endif
 
c           Send to outindex only if outindex in (1,nproc).
 
            if ((outindex .ge. 0) .and. (outindex .le. nproc-1)) then
                nevent = myproc * 65536 + outindex
c               print *,'node',myproc, ' --> ', outindex,
c    #                  ' on',nevent,' :',outval

                call NSEND(outindex, nevent, ntype, nlength, nflags,
     #                   ndata, ndsize, outval, nwerr)
            endif
 
        endif     
 
        if (nwerr .ne. 0) then
            nerr = nwerr
        else
            nerr = nrerr
        endif
 
        return
        end
 
 
c       --------------------------------------------------------------------
 
        subroutine outres

c	Send results to host.

        parameter (MAXSIZE = 500)
         
        common /extcom/ myproc, nproc
        common /msgcom/ nnode, nevent, ntype, nlength, nflags,
     #                  ndata, ndsize
 
        double precision values(0:MAXSIZE+1), oldval(0:MAXSIZE+1),
     #  newval(0:MAXSIZE+1)
        common /valcom/ values, oldval, newval
        common /nptcom/ nmin, nleft, ntot
 
        integer dblsize
        parameter (dblsize = 8)
        parameter (IEVRES = 16, IEVOUT = 100)
 
        if (myproc .lt. nleft) then
            npts = nmin+1
        else
            npts = nmin
        endif
 
c       Get OUT msg from control processor.
 
        nevent = IEVOUT+myproc
        nlength = 0
        ntype = 0
	ndsize = 0
 
        call NRECV(nevent, ntype, nlength, nflags, ndata,
     #             ndsize, ntask, nerr)
        if (nerr.ne.0) then
            print *,'node(',myproc,')(outres): Err in nrecv ',nerr
            call TERROR('node')
            call KEXIT(-1)
        endif
 
c       Send data to control processor.
 
        nevent = IEVRES
        nlength = npts*dblsize
        ntype = 0
        nhost = IGORGN()
c       print *,'node ',myproc, ' sending results'

        call NSEND (nhost, nevent, ntype, nlength, nflags,
     #              ndata, ndsize, values(1), nerr)
        if (nerr.ne.0) then
            print *, 'node(', myproc, ')(outres): Err in nsend(data)',
     #                  nerr
            call KEXIT(-1)
        endif
 
        return
        end
 
c       --------------------------------------------------------------------

