/*
 *      OSU Research Computing					Trillium
 *      Copyright 1992 The Ohio State University
 *      RBD
 *
 *      $Log:	mtvmachine.c,v $
 * Revision 2.2  92/04/24  16:12:09  trillium
 * Ohio Release
 * 
 *	Function:	- machine dependent module for mtvd
 *			- ITB stub version
 *			- G300 controller version
 */

#include <stdio.h>
#include <memory.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>
#include <tvreq.h>
#include <mtvmachine.h>

/*
 * external variables
 */
extern int		errno;			/* global error code */

/*
 * static variables
 */
static int4		g3_back_porch;		/* back porch timing */
static int4		g3_broad_pulse;		/* vertical sync pulse */
static int4		g3_display;		/* visible display period */
static int4		g3_half_sync;		/* 1/2 line sync pulse */
static int4		g3_linetime;		/* total horiz. line period */
static int4		g3_meminit;		/* memory init timing */
static int4		g3_pll_factor;		/* PLL multiplication factor */
static int4		g3_short_disp;		/* short horiz. line period */
static int4		g3_vblank;		/* vertical blanking period */
static int4		g3_vdisplay;		/* vertical display */
static int4		g3_vsync;		/* vertical sync period */
static int4		g3_xfer_delay;		/* data transfer timing */
static int4		xsize;			/* horizontal screen size */
static int4		ysize;			/* vertical screen size */
static char		*me;			/* program name */

static char *usage = "usage: %s [-h] [-b bkp] [-d dsp] [-k vb] [-l lt] [-m mi] [-p pll] [-s sdsp] [-v vs] [-x td] [-B bp] [-H hs] [-V vdsp]\n";

/*
 * external functions
 */
extern void		exit();
extern void		followed();
extern void		validopts();
extern int		do_args();
extern int		opt_taken();

/*
 * static functions
 */
static void		g3_help();
static void		g3_initclut();
static void		g3_initcrtc();
static void		g3_plotchar();
static void		g3_reset();
static void		g3_setpalette();
static int4		g3_col_normalize();

/*
 *	tv_init
 *
 *	Function:	- initializes mtvd graphics interface
 *			- ITB G300 version
 *			- exits on some conditions
 *
 *	Accepts:	- argc, argv from command line
 *
 *	Returns:	- 0 or ERROR
 */
int
tv_init(argc, argv)

int			argc;
char			*argv[];

{
	me = argv[0];
	validopts("h");
	followed("bdklmpsvxBHV");
/*
 * Parse the command line.
 */
	if ((do_args(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {
		fprintf(stderr, usage, me);
		exit(1);
	}

	if (opt_taken('h')) {
		g3_help();
		exit(0);
	}
/*
 * Assign the controller parameters.
 */
	g3_back_porch = opt_taken('b') ? stoi4(getparam('b')) : BPORCH;
	g3_display = opt_taken('d') ? stoi4(getparam('d')) : DISPLAY;
	g3_pll_factor = opt_taken('p') ? stoi4(getparam('p')) : PLLFACTOR;
	g3_vblank = opt_taken('k') ? stoi4(getparam('k')) : VBLANK;
	g3_linetime = opt_taken('l') ? stoi4(getparam('l')) : LINETIME;
	g3_meminit = opt_taken('m') ? stoi4(getparam('m')) : MEMINIT;
	g3_short_disp = opt_taken('s') ? stoi4(getparam('s')) : SHORTDISP;
	g3_vsync = opt_taken('v') ? stoi4(getparam('v')) : VSYNC;
	g3_xfer_delay = opt_taken('x') ? stoi4(getparam('x')) : XFERDELAY;
	g3_broad_pulse = opt_taken('B') ? stoi4(getparam('B')) : BPULSE;
	g3_half_sync = opt_taken('H') ? stoi4(getparam('H')) : HALFSYNC;
	g3_vdisplay = opt_taken('V') ? stoi4(getparam('V')) : VDISPLAY;
/*
 * Assign the screen size in pixels.
 */
	xsize = g3_display * 4;
	ysize = g3_vdisplay / 2;
/*
 * Initialize the CLUT to default mode 0.  Initialize the CRT controller.
 */
	g3_initclut();
	g3_initcrtc();
/*
 * Clear the screen.
 */
	(void) tv_clr();

	return(0);
}

/*
 *	g3_initclut
 *
 *	Function:	- initialize the CLUT of the G300 controller
 *			- set the CLUT to "IBM standard" mode
 */
static void
g3_initclut()

{
	int4		i;		/* favorite counter */
	int4		red;
	int4		green;
	int4		blue;
	int4		bias_value = 9;
	int4		colour_index[4];

	G3_MASK = 0xFF;
	colour_index[0] = 0;
	colour_index[1] = 18;
	colour_index[2] = 36;
	colour_index[3] = 54;
	g3_setpalette(0, 0, 0, 0);

	for (i = 1; i < 256; i++) {
		red = colour_index[i & 0x3];
		green = colour_index[(i >> 2) & 0x3];
		blue = colour_index[(i >> 4) & 0x3];
		if (i >= 192) {
			blue += bias_value;
		} else if (i >= 128) {
			green += bias_value;
		} else if (i >= 64) {
			red += bias_value;
		}
		g3_setpalette(i, red, green, blue);
	}
}

/*
 *	g3_setpalette
 *
 *	Function:	- set a palette entry in the colour
 *			  lookup table (CLUT)
 *
 *	Accepts:	- colour (CLUT index)
 *			- red colour value
 *			- green colour value
 *			- blue colour value
 */
static void
g3_setpalette(colour, red, green, blue)

int4			colour,
			red,
			green,
			blue;

{
	int4		real_colour;		/* real 24 bit colour */
/*
 * Normalize the colour levels.
 */
	red = g3_col_normalize(red);
	green = g3_col_normalize(green);
	blue = g3_col_normalize(blue);
/*
 * Compute the real colour code and store it in the CLUT.
 */
	real_colour = ((blue << 16) | (green << 8) | red);
	G3_REGISTER[colour] = real_colour;
}

/*
 *	g3_col_normalize
 *
 *	Function:	- normalize the RGB colour value
 *			- normalized value fits into a byte
 *
 *	Accepts:	- RGB colour value
 *
 *	Returns:	- normalized RGB colour value
 */
static int4
g3_col_normalize(colour)

int4			colour;

{
	if (colour < 0) {
		return(INT4_NIL);
	} else if (colour > 63) {
		return((int4) 255);
	} else {
		return(4 * colour);
	}
}

/*
 *	g3_initcrtc
 *
 *	Function:	- initialize the CRT controller section of
 *			  the G300 graphic controller chip
 *			- transputer version
 */
static void
g3_initcrtc()

{
/*
 * Reset the graphics controller.
 */
	g3_reset();
/*
 * Set the Phase Locked Loop factor.
 */
	G3_CNTL = 0;
	G3_BOOT = g3_pll_factor;
	tin(ldtimer() + RESETDELAY);
/*
 * Program the VTG registers.
 */
	G3_HALFSYNC = g3_half_sync;
	G3_BACKPORCH = g3_back_porch;
	G3_DISPLAY = g3_display;
	G3_SHORTDSP = g3_short_disp;
	G3_BPULSE = g3_broad_pulse;
	G3_VSYNC = g3_vsync;
	G3_VBLANK = g3_vblank;
	G3_VDISPLAY = g3_vdisplay;
	G3_LINETIME = g3_linetime;
	G3_LINESTART = 0;
	G3_MEMINIT = g3_meminit;
	G3_XFERDELAY = g3_xfer_delay;
	G3_MASK = 255;
	G3_TOPSCR = 0;
/*
 * Turn the VTG on.
 */
	G3_CNTL = 0x0001;
}

/*
 *	g3_reset
 *
 *	Function:	- reset the G300 graphic controller chip
 *			- transputer version
 */
static void
g3_reset()

{
/*
 * Wait for the reset delay then set the reset line.
 */
	tin(ldtimer() + RESETDELAY);
	G3_RESET = 1;
/*
 * Wait for the reset delay then clear the reset line.
 */
	tin(ldtimer() + RESETDELAY);
	G3_RESET = 0;
}

/*
 *	tv_clr
 *
 *	Function:	- clear the entire screen to black
 *			- G300 version
 *
 *	Returns:	- 0 or ERROR
 */
int
tv_clr()

{
	int4		nwords;		/* # words per screen line */
	int4		nchars;		/* # remaining line characters */
	int4		i;		/* favorite counter */
	char		*p;		/* favorite pointer */
	int4		*pp;		/* another pointer */

	nwords = xsize / 4;
	nchars = xsize % 4;
/*
 * Blank the first screen line.
 */
	for (i = 0, pp = (int4 *) G3_SCRBASE; i < nwords; i++, pp++) {
		*pp = INT4_NIL;
	}
	for (i = 0, p = (char *) pp; i < nchars; i++, p++) {
		*p = (char) 0;
	}
/*
 * Blank all remaining lines by copying from the first line.
 */
	p = (char *) G3_SCRBASE + xsize;
	for (i = 1; i < ysize; i++, p += xsize) {
		(void) memcpy(p, (char *) G3_SCRBASE, xsize);
	}

	return(0);
}

/*
 *	tv_rect
 *
 *	Function:	- plot a multicoloured rectangle
 *			- handles boundary clipping
 *			- G300 version
 *
 *	Accepts:	- upper left corner x-ordinate
 *			- upper left corner y-ordinate
 *			- rectangle horizontal size (width)
 *			- rectangle vertical size (height)
 *			- array of pixel colours
 *
 *	Returns:	- 0 or ERROR
 */
int
tv_rect(x1, y1, xlen, ylen, colours)

int4			x1,
			y1,
			xlen,
			ylen;
char			*colours;

{
	int4		i, j;		/* favorite indices */
	int4		xmax, ymax;	/* upper rectangle bounds */
	char		*p;		/* favorite pointer */

	xmax = x1 + xlen - 1;
	ymax = y1 + ylen - 1;

	if ((x1 >= 0) && (y1 >= 0) && (xmax < xsize) && (ymax < ysize)) {
/*
 * If the full rectangle is within the screen, plot it with no pixel checks.
 */
		p = (char *) G3_SCRBASE + (y1 * xsize) + x1;
		for (j = y1; j <= ymax; j++) {
			for (i = x1; i <= xmax; i++) {
				*(p++) = *(colours++);
			}
			p += xsize - xlen;
		}
	}
	else if ((x1 >= xsize) || (y1 >= ysize) || (xmax < 0) || (ymax < 0)) {
/*
 * Do nothing if the full rectangle is outside the screen.
 */
	}
	else {
/*
 * The rectangle has a portion outside the screen, check each pixel.
 */
		for (j = y1; j <= ymax; j++) {
			for (i = x1; i <= xmax; i++) {
				if ((i >= 0) && (j >= 0) &&
					(i < xsize) && (j < ysize)) {

					p = (char *) G3_SCRBASE +
						(j * xsize) + i;
					*p = *colours;
				}
				colours++;
			}
		}
	}

	return(0);
}

/*
 *	tv_snake
 *
 *	Function:	- plot a multicoloured snake-like region
 *			- handles boundary clipping
 *			- G300 version
 *
 *	Accepts:	- starting point ordinate
 *			- snake length
 *			- array of pixel colours
 *
 *	Returns:	- 0 or ERROR
 */
int
tv_snake(x, len, colours)

int4			x,
			len;
char			*colours;

{
	int4		i;		/* favorite counter */
	char		*p;		/* favorite pointer */
/*
 * Do nothing if the entire snake is outside the screen.
 */
	if (x >= (xsize * ysize)) {
		return(0);
	}
/*
 * Clip the head of the snake if it is outside the screen.
 */
	if (x < 0) {
		colours -= x;
		len += x;
		x = 0;
	}
/*
 * Clip the tail of the snake if it is outside the screen.
 */
	if ((x + len) > (xsize * ysize)) {
		len = (xsize * ysize) - x;
	}
/*
 * Plot the snake.
 */
	p = (char *) G3_SCRBASE + x;
	for (i = 0; i < len; i++) {
		*(p++) = *(colours++);
	}

	return(0);
}

/*
 *	tv_text
 *
 *	Function:	- plot a text string
 *			- handles boundary clipping
 *			- handles only horizontal strings
 *			- no wrap around at edge of screen
 *			- G300 version
 *
 *	Accepts:	- horizontal text ordinate of first letter
 *			- vertical text ordinate of first letter
 *			- string colour
 *			- string
 *
 *	Returns:	- 0 or ERROR
 */
int
tv_text(x, y, colour, string)

int4			x,
			y,
			colour;
char			*string;

{
	int4		len;		/* string length */

	len = strlen(string);
/*
 * If the string is outside the screen, do nothing.
 */
	if ((x >= xsize/XCHAR) || (y >= ysize/YCHAR) ||
		(x + len <= 0) || (y < 0)) {

		return(0);
	}
/*
 * Clip the head of the string if needed.
 */
	if (x < 0) {
		string -= x;
		len += x;
		x = 0;
	}
/*
 * Clip the tail of the string if needed.
 */
	if (x + len > xsize/XCHAR) {
		len = xsize/XCHAR - x;
		string[len] = '\0';
	}
/*
 * Plot the string.
 */
	for ( ; *string; string++, x++) {
		g3_plotchar(x, y, colour, *string);
	}

	return(0);
}

/*
 *	tv_state
 *
 *	Function:	- get mtvd state information
 *
 *	Accepts:	- ptr to state structure
 *
 *	Returns:	- 0 or ERROR
 */
int
tv_state(pstate)

struct tvstate	*pstate;

{
	pstate->tvs_xlen = xsize;
	pstate->tvs_ylen = ysize;
	pstate->tvs_ncolour = 256;
	pstate->tvs_xtext = xsize/XCHAR;
	pstate->tvs_ytext = ysize/YCHAR;

	return(0);
}

/*
 *	g3_plotchar
 *
 *	Function:	- plot a character on the screen
 *			- G300 version
 *
 *	Accepts:	- horizontal character position
 *			- vertical character position
 *			- colour
 *			- character code
 */
static void
g3_plotchar(xc, yc, col, char_code)

int4			xc,
			yc,
			col;
char			char_code;

{
	int4		xpixel, ypixel;		/* char. pixel coordinates */
	int4		i;			/* favorite counter */
	unsigned char	row;			/* char. row pixel map */
	unsigned char	*pcm;			/* ptr character map */
	char		*p;			/* favorite pointer */
/*
 * Do nothing if the character is not valid.
 */
	i = (int4) char_code;
	if ((i < 0) || (i > NCHAR)) {
		return;
	}
/*
 * Loop over the rows of the character to be plotted.
 */
	pcm = charset[i].cm_rows;
	ypixel = yc * YCHAR;
	for (i = 0; i < YCHAR; i++, pcm++, ypixel++) {
		xpixel = xc * XCHAR;
		for (row = *pcm; row != 0; row <<= 1, xpixel++) {
			if (row & 0x80) {
				p = (char *) G3_SCRBASE + xpixel +
					(ypixel * xsize);
				*p = (char) (col & 0xFF);
			}
		}
	}
}

/*
 *	g3_help
 *
 *	Function:	- print the help menu
 */
static void 
g3_help()

{
	printf("\nSynopsis:\t%s [options]\n", me);
	printf("\nDescription:\tITB graphics daemon.\n");
	printf("\nOptions:\t-h                      Print this message.\n");
	printf("\t\t-b <bkp>                Back porch timing.\n");
	printf("\t\t-d <dsp>                Display line timing.\n");
	printf("\t\t-k <vb>                 Vertical blank timing.\n");
	printf("\t\t-l <lt>                 Horizontal line timing.\n");
	printf("\t\t-m <mi>                 Memory initialization timing.\n");
	printf("\t\t-p <pll>                PLL pixel clock factor.\n");
	printf("\t\t-s <sdsp>               Short display timing.\n");
	printf("\t\t-v <vs>                 Vertical sync timing.\n");
	printf("\t\t-x <td>                 Data transfer delay.\n");
	printf("\t\t-B <bp>                 Broad pulse timing.\n");
	printf("\t\t-H <hs>                 Half-sync timing.\n");
	printf("\t\t-V <vdsp>               Vertical display.\n");
}
