/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	mtvmachine.h,v $
 * Revision 2.2  92/04/24  16:12:18  trillium
 * Ohio Release
 * 
 *	Function:	- contants and structures associated with mtvd
 *			- ITB, INMOS G300 version
 */

#ifndef _MTVMACHINE
#define _MTVMACHINE

/*
 * register locations (base and offsets)
 */
#define G3_REGBASE		0x40000000
#define G3_REGISTER		((int *) G3_REGBASE)
#define G3_BOOT			G3_REGISTER[0x1A0]
#define G3_TOPSCR		G3_REGISTER[0x180]
#define G3_CNTL			G3_REGISTER[0x160]
#define G3_MASK			G3_REGISTER[0x140]
#define G3_BYTECNT		G3_REGISTER[0x140]
#define G3_HALFSYNC		G3_REGISTER[0x121]
#define G3_BACKPORCH		G3_REGISTER[0x122]
#define G3_DISPLAY		G3_REGISTER[0x123]
#define G3_SHORTDSP		G3_REGISTER[0x124]
#define G3_BPULSE		G3_REGISTER[0x125]
#define G3_VSYNC		G3_REGISTER[0x126]
#define G3_VBLANK		G3_REGISTER[0x127]
#define G3_VDISPLAY		G3_REGISTER[0x128]
#define G3_LINETIME		G3_REGISTER[0x129]
#define G3_LINESTART		G3_REGISTER[0x12A]
#define G3_MEMINIT		G3_REGISTER[0x12B]
#define G3_XFERDELAY		G3_REGISTER[0x12C]
#define G3_HINC			G3_REGISTER[0x12D]
#define G3_VINC			G3_REGISTER[0x12E]
#define G3_TBINC		G3_REGISTER[0x12F]

#define G3_RESET		(*((int *) 0x000000F0))
#define G3_SCRBASE		0xC0000000
#define G3_SCREEN		((char *) G3_SCRBASE)

/*
 * constants (set for Hitachi HM-4219/4119 monitor and 5MHz system clock)
 */
#define BPORCH			54		/* back porch */
#define BPULSE			205		/* broad pulse */
#define DISPLAY			320		/* screen width (words) */
#define HALFSYNC		21		/* 1/2 horiz. sync pulse */
#define LINETIME		422		/* total horiz. line */
#define MEMINIT			238		/* MemInit register */
#define PLLFACTOR		21		/* PLL factor for 5MHz clock */
#define SHORTDISP		109		/* short line timing */
#define VBLANK			56		/* vertical blank */
#define VDISPLAY		2048		/* screen height (1/2 lines) */
#define VSYNC			6		/* vertical sync */
#define XFERDELAY		18		/* data transfer timing */
#define RESETDELAY		20		/* (low-pri ticks) 1.28 ms */

/*
 * character set
 */
#define XCHAR			8		/* # horizontal pixels/char */
#define YCHAR			8		/* # vertical pixels/char */
#define NCHAR			128		/* # characters in the set */

struct charmap {
	unsigned char	cm_rows[YCHAR];		/* rows of 8 pixels */
};

static struct charmap	charset[NCHAR] = {	/* numbers in hexadecimal */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII   0 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII   1 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII   2 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII   3 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII   4 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII   5 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII   6 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII   7 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII   8 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII   9 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  10 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  11 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  12 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  13 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  14 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  15 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  16 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  17 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  18 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  19 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  20 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  21 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  22 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  23 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  24 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  25 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  26 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  27 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  28 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  29 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  30 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  31 ' ' */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  32 ' ' */
	{ 0x18, 0x18, 0x18, 0x18, 0x18, 0x00, 0x18, 0x00 }, /* ASCII  33  !  */
	{ 0x6C, 0x6C, 0x6C, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  34  "  */
	{ 0x6C, 0x6C, 0xFE, 0x6C, 0xFE, 0x6C, 0x6C, 0x00 }, /* ASCII  35  #  */
	{ 0x18, 0x3E, 0x58, 0x3C, 0x1A, 0x7C, 0x18, 0x00 }, /* ASCII  36  $  */
	{ 0x00, 0xC6, 0xCC, 0x18, 0x30, 0x66, 0xC6, 0x00 }, /* ASCII  37  %  */
	{ 0x38, 0x6C, 0x38, 0x76, 0xDC, 0xCC, 0x76, 0x00 }, /* ASCII  38  &  */
	{ 0x18, 0x18, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  39  '  */
	{ 0x0C, 0x18, 0x30, 0x30, 0x30, 0x18, 0x0C, 0x00 }, /* ASCII  40  (  */
	{ 0x30, 0x18, 0x0C, 0x0C, 0x0C, 0x18, 0x30, 0x00 }, /* ASCII  41  )  */
	{ 0x00, 0x66, 0x3C, 0xFF, 0x3C, 0x66, 0x00, 0x00 }, /* ASCII  42  *  */
	{ 0x00, 0x18, 0x18, 0x7E, 0x18, 0x18, 0x00, 0x00 }, /* ASCII  43  +  */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x30 }, /* ASCII  44  ,  */
	{ 0x00, 0x00, 0x00, 0x7E, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  45  -  */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00 }, /* ASCII  46  .  */
	{ 0x06, 0x0C, 0x18, 0x30, 0x60, 0xC0, 0x80, 0x00 }, /* ASCII  47  /  */
	{ 0x7C, 0xC6, 0xCE, 0xD6, 0xE6, 0xC6, 0x7C, 0x00 }, /* ASCII  48  0  */
	{ 0x18, 0x38, 0x18, 0x18, 0x18, 0x18, 0x7E, 0x00 }, /* ASCII  49  1  */
	{ 0x3C, 0x66, 0x06, 0x3C, 0x60, 0x66, 0x7E, 0x00 }, /* ASCII  50  2  */
	{ 0x3C, 0x46, 0x06, 0x1C, 0x06, 0x66, 0x3C, 0x00 }, /* ASCII  51  3  */
	{ 0x18, 0x38, 0x58, 0x98, 0xFE, 0x18, 0x3C, 0x00 }, /* ASCII  52  4  */
	{ 0x7E, 0x62, 0x60, 0x3C, 0x06, 0x66, 0x3C, 0x00 }, /* ASCII  53  5  */
	{ 0x3C, 0x66, 0x60, 0x7C, 0x66, 0x66, 0x3C, 0x00 }, /* ASCII  54  6  */
	{ 0x7E, 0x46, 0x06, 0x0C, 0x18, 0x18, 0x18, 0x00 }, /* ASCII  55  7  */
	{ 0x3C, 0x66, 0x66, 0x3C, 0x66, 0x66, 0x3C, 0x00 }, /* ASCII  56  8  */
	{ 0x3C, 0x66, 0x66, 0x3E, 0x06, 0x66, 0x3C, 0x00 }, /* ASCII  57  9  */
	{ 0x00, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00 }, /* ASCII  58  :  */
	{ 0x00, 0x00, 0x18, 0x18, 0x00, 0x18, 0x18, 0x30 }, /* ASCII  59  ;  */
	{ 0x0C, 0x18, 0x30, 0x60, 0x30, 0x18, 0x0C, 0x00 }, /* ASCII  60  <  */
	{ 0x00, 0x00, 0x7E, 0x00, 0x00, 0x7E, 0x00, 0x00 }, /* ASCII  61  =  */
	{ 0x60, 0x30, 0x18, 0x0C, 0x18, 0x30, 0x60, 0x00 }, /* ASCII  62  >  */
	{ 0x3C, 0x66, 0x06, 0x0C, 0x18, 0x00, 0x18, 0x00 }, /* ASCII  63  ?  */
	{ 0x7C, 0xC6, 0xDE, 0xDE, 0xDE, 0xC0, 0x7C, 0x00 }, /* ASCII  64  @  */
	{ 0x18, 0x3C, 0x66, 0x66, 0x7E, 0x66, 0x66, 0x00 }, /* ASCII  65  A  */
	{ 0xFC, 0x66, 0x66, 0x7C, 0x66, 0x66, 0xFC, 0x00 }, /* ASCII  66  B  */
	{ 0x3C, 0x66, 0xC0, 0xC0, 0xC0, 0x66, 0x3C, 0x00 }, /* ASCII  67  C  */
	{ 0xF8, 0x6C, 0x66, 0x66, 0x66, 0x6C, 0xF8, 0x00 }, /* ASCII  68  D  */
	{ 0xFE, 0x62, 0x68, 0x78, 0x68, 0x62, 0xFE, 0x00 }, /* ASCII  69  E  */
	{ 0xFE, 0x62, 0x68, 0x78, 0x68, 0x60, 0xF0, 0x00 }, /* ASCII  70  F  */
	{ 0x3C, 0x66, 0xC0, 0xC0, 0xCE, 0xC6, 0x7E, 0x00 }, /* ASCII  71  G  */
	{ 0x66, 0x66, 0x66, 0x7E, 0x66, 0x66, 0x66, 0x00 }, /* ASCII  72  H  */
	{ 0x7E, 0x18, 0x18, 0x18, 0x18, 0x18, 0x7E, 0x00 }, /* ASCII  73  I  */
	{ 0x1E, 0x0C, 0x0C, 0x0C, 0xCC, 0xCC, 0x78, 0x00 }, /* ASCII  74  J  */
	{ 0xE6, 0x66, 0x6C, 0x78, 0x6C, 0x66, 0xE6, 0x00 }, /* ASCII  75  K  */
	{ 0xF0, 0x60, 0x60, 0x60, 0x62, 0x66, 0xFE, 0x00 }, /* ASCII  76  L  */
	{ 0xC6, 0xEE, 0xFE, 0xFE, 0xD6, 0xC6, 0xC6, 0x00 }, /* ASCII  77  M  */
	{ 0xC6, 0xE6, 0xF6, 0xDE, 0xCE, 0xC6, 0xC6, 0x00 }, /* ASCII  78  N  */
	{ 0x38, 0x6C, 0xC6, 0xC6, 0xC6, 0x6C, 0x38, 0x00 }, /* ASCII  79  O  */
	{ 0xFC, 0x66, 0x66, 0x78, 0x60, 0x60, 0xF0, 0x00 }, /* ASCII  80  P  */
	{ 0x38, 0x66, 0xC6, 0xC6, 0xDA, 0xCC, 0x76, 0x00 }, /* ASCII  81  Q  */
	{ 0xFC, 0x66, 0x66, 0x7C, 0x6C, 0x66, 0xE2, 0x00 }, /* ASCII  82  R  */
	{ 0x3C, 0x66, 0x60, 0x3C, 0x06, 0x66, 0x3C, 0x00 }, /* ASCII  83  S  */
	{ 0x7E, 0x5A, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00 }, /* ASCII  84  T  */
	{ 0x66, 0x66, 0x66, 0x66, 0x66, 0x66, 0x3C, 0x00 }, /* ASCII  85  U  */
	{ 0x66, 0x66, 0x66, 0x66, 0x66, 0x3C, 0x18, 0x00 }, /* ASCII  86  V  */
	{ 0xC6, 0xC6, 0xC6, 0xD6, 0xFE, 0xEE, 0xC6, 0x00 }, /* ASCII  87  W  */
	{ 0xC6, 0x6C, 0x38, 0x38, 0x6C, 0xC6, 0xC6, 0x00 }, /* ASCII  88  X  */
	{ 0x66, 0x66, 0x66, 0x3C, 0x18, 0x18, 0x3C, 0x00 }, /* ASCII  89  Y  */
	{ 0xFE, 0xC6, 0x8C, 0x18, 0x32, 0x66, 0xFE, 0x00 }, /* ASCII  90  Z  */
	{ 0x3C, 0x30, 0x30, 0x30, 0x30, 0x30, 0x3C, 0x00 }, /* ASCII  91  [  */
	{ 0xC0, 0x60, 0x30, 0x18, 0x0C, 0x06, 0x02, 0x00 }, /* ASCII  92  \  */
	{ 0x3C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x3C, 0x00 }, /* ASCII  93  ]  */
	{ 0x18, 0x3C, 0x66, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  94  ^  */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF }, /* ASCII  95  _  */
	{ 0x00, 0x30, 0x18, 0x0C, 0x00, 0x00, 0x00, 0x00 }, /* ASCII  96  `  */
	{ 0x00, 0x00, 0x78, 0x0C, 0x7C, 0xCC, 0x76, 0x00 }, /* ASCII  97  a  */
	{ 0xE0, 0xE0, 0x7C, 0x66, 0x66, 0x66, 0xBC, 0x00 }, /* ASCII  98  b  */
	{ 0x00, 0x00, 0x3C, 0x66, 0x60, 0x66, 0x3C, 0x00 }, /* ASCII  99  c  */
	{ 0x1C, 0x0C, 0x7C, 0xCC, 0xCC, 0xCC, 0x76, 0x00 }, /* ASCII 100  d  */
	{ 0x00, 0x00, 0x3C, 0x66, 0x7E, 0x60, 0x3C, 0x00 }, /* ASCII 101  e  */
	{ 0x1C, 0x36, 0x30, 0x78, 0x30, 0x30, 0x78, 0x00 }, /* ASCII 102  f  */
	{ 0x00, 0x00, 0x3E, 0x66, 0x66, 0x3E, 0x06, 0x7C }, /* ASCII 103  g  */
	{ 0xE0, 0x60, 0x6C, 0x76, 0x66, 0x66, 0xE6, 0x00 }, /* ASCII 104  h  */
	{ 0x18, 0x00, 0x38, 0x18, 0x18, 0x18, 0x3C, 0x00 }, /* ASCII 105  i  */
	{ 0x06, 0x00, 0x0E, 0x06, 0x06, 0x66, 0x66, 0x3C }, /* ASCII 106  j  */
	{ 0xE0, 0x60, 0x66, 0x6C, 0x78, 0x6E, 0xE6, 0x00 }, /* ASCII 107  k  */
	{ 0x38, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00 }, /* ASCII 108  l  */
	{ 0x00, 0x00, 0x6C, 0xFE, 0xD6, 0xD6, 0xC6, 0x00 }, /* ASCII 109  m  */
	{ 0x00, 0x00, 0xD8, 0x66, 0x66, 0x66, 0x66, 0x00 }, /* ASCII 110  n  */
	{ 0x00, 0x00, 0x3C, 0x66, 0x66, 0x66, 0x3C, 0x00 }, /* ASCII 111  o  */
	{ 0x00, 0x00, 0xDC, 0x66, 0x66, 0x7C, 0x60, 0xF0 }, /* ASCII 112  p  */
	{ 0x00, 0x00, 0x76, 0xCC, 0xCC, 0x7C, 0x0C, 0x1E }, /* ASCII 113  q  */
	{ 0x00, 0x00, 0x6C, 0x36, 0x30, 0x30, 0x78, 0x00 }, /* ASCII 114  r  */
	{ 0x00, 0x00, 0x3C, 0x60, 0x3C, 0x06, 0x7C, 0x00 }, /* ASCII 115  s  */
	{ 0x30, 0x30, 0x7C, 0x30, 0x30, 0x36, 0x1C, 0x00 }, /* ASCII 116  t  */
	{ 0x00, 0x00, 0x66, 0x66, 0x66, 0x66, 0x3E, 0x00 }, /* ASCII 117  u  */
	{ 0x00, 0x00, 0x66, 0x66, 0x66, 0x3C, 0x18, 0x00 }, /* ASCII 118  v  */
	{ 0x00, 0x00, 0xC6, 0xD6, 0xD6, 0xFE, 0x6C, 0x00 }, /* ASCII 119  w  */
	{ 0x00, 0x00, 0xC6, 0x6C, 0x38, 0x6C, 0xC6, 0x00 }, /* ASCII 120  x  */
	{ 0x00, 0x00, 0x66, 0x66, 0x66, 0x3E, 0x06, 0x7C }, /* ASCII 121  y  */
	{ 0x00, 0x00, 0x7E, 0x4C, 0x18, 0x30, 0x7E, 0x00 }, /* ASCII 122  z  */
	{ 0x0E, 0x18, 0x18, 0x70, 0x18, 0x18, 0x0E, 0x00 }, /* ASCII 123  {  */
	{ 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x00 }, /* ASCII 124  |  */
	{ 0x70, 0x18, 0x18, 0x0E, 0x18, 0x18, 0x70, 0x00 }, /* ASCII 125  }  */
	{ 0x76, 0xD8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* ASCII 126  ~  */
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }  /* ASCII 127 ' ' */
};

#endif	/* _MTVMACHINE */
