.TH BRENDALIB l  "March 14, 1990" "Release 2.1" "TROLLIUS PROGRAMMING TOOLS"
.SH NAME
brendalib
\- the Brenda library functions interface 
.SH SYNOPSIS
.nf
.ft B
subroutine PUT  (tag, buffer, size, flags, snode, ierror)
subroutine GRAB (tag, buffer, size, flags, snode, ierror)
subroutine LOOK (tag, buffer, size, flags, snode, ierror)
subroutine TRYGRAB (tag, buffer, size, flags, snode, ierror)
subroutine TRYLOOK (tag, buffer, size, flags, snode, ierror)
subroutine BUMP (tag, oldvalue, increment, flags, snode, ierror)
subroutine FREEZE ()
subroutine THAW ()

integer*4 tag, size, oldvalue, increment, snode
integer flags, ierror, buffer(*)
.ft R
.fi
.SH DESCRIPTION
The \fIbrenda\fR functions provide a form of virtual shared memory over a
network of Trollius nodes. 
Objects are tagged and thrown into a pool. 
The first process to request an object with that tag picks the object out of
the pool. 
Since objects cannot be modified while in place in the pool,
there are no synchronization problems. 
The location of actual storage, if necessary, and the details of the
message passing scheme behind the brenda services are hidden from the caller.
The overhead, for small objects, is about three times that of direct
message passing. 
For objects that are large relative to the maximum message size the
transmission costs dominate and the overhead is negligible. 
Brenda avoids double-buffering as much as possible.
.PP
The PUT subroutine puts an object into the common
.I "tuple space"
(TS). 
The object is simply \fIsize\fR bytes from the array \fIbuffer\fR. 
Although buffer is declared as an integer array, any
non-character array type can be used. 
A copy of the object is made, so when PUT returns the caller may
overwrite the data.
The \fIsize\fR must be greater than or equal to 4 bytes (the size of 
integer*4).
.PP
The GRAB function removes a matching object from TS and returns
up to \fIsize\fR bytes of it into the provided buffer \fIbuffer\fR.
The function LOOK finds a matching object in TS and returns
a copy, but does \fInot\fR remove it from TS.
.PP
If no matching object is found in TS, GRAB and LOOK block
until one appears. 
The functions TRYGRAB and TRYLOOK are similar to GRAB and LOOK, but
do not block. 
If no matching object is found they return an error code.
.PP
The BUMP function is an optimized operation that combines reading
the old value of an integer object and adding a given constant to the
stored object. 
For the BUMP operation, the affected object must already exist and
must be of the size of a integer*4 (4 bytes). 
Note that in this case the size is implied, and the amount to increment
by takes the place of the \fIsize\fR parameter.
.PP
The order of bytes inside the object is normally not touched.
In the special case where \fIsize\fR is equal to the size of a integer*4 
(4 bytes), the byte ordering is automatically converted to fit each machine.  
This was done to enable the BUMP function.  
So beware of the case where an object, larger than one integer,
is put by one kind of processor and grabbed by another.
.PP
An object is tagged by an arbitrary 32-bit non-zero integer value.
The value of the tag is up to the user, but the tag specified in
a PUT must not be zero, and the tag given to GRAB or
LOOK must match the original tag for the object to be retrieved.
The tag "ANYTAG" (0) matches any other tag.
.PP
The \fIflags\fR parameter is normally 0.
The following values may be added to this parameter:
.TP 16
B_SETTLED
means the object is to be stored in the node specified by the parameter
\fIsnode\fR. 
The GRAB and LOOK calls must also specify B_SETTLED and must know what
\fIsnode\fR to specify or the object that was PUT as B_SETTLED will
not be found. 
The flag B_CAL means means that the object is to be stored in the
local node, and is equivalent to it being B_SETTLED in the node GETNODEID(l).
The default is B_NOMADIC (0), which means the object may be stored anywhere,
and the actual storage location is a function of the tag. 
The usage of B_SETTLED is recommended as an optimization when the object
is to be used by one consuming process which runs on a known node.
.TP
B_GBAL
means that copies of the object are to be sent to \fIall\fR the
participating storage nodes, using a broadcast mechanism.
This is an optimization that is useful when it is known that
the object will be LOOK-ed at by processes on all or most of the nodes. 
The LOOK, too, must specify the
B_GBAL flag (or, equivalently, the B_LOCAL flag) for the optimization
to yield any gains by avoiding long-range communication.
A GRAB of a B_GBAL object is not permitted, except in the special
case where the GRAB is done by the same
process that originally created that object via a B_GBAL PUT.
In this special case a request to remove the object is broadcast
to all the storage nodes.
This restriction of a B_GBAL GRAB to one process eliminates
the need for a global query, but does not guarantee that a last-minute
LOOK on a remote node will fail to find the object, even after
copies of the object have been removed from some other nodes.
.TP
B_VERIFY
means that a return message will be sent from the Brenda server
during the processing of a PUT request, slowing things down somewhat
but checking that there is space for storing the object. 
If this flag is not specified, a call can return \fIOK\fR while the storage
node is out of memory. 
B_VERIFY is relevant only when PUT-ing objects that are smaller than MAXNMSGLEN.
The default is B_BLIND (0).
It is recommended that VERIFY be used while debugging. 
B_VERIFY can also provide a form of flow control for processes that put a long
stream of objects, since each PUT call will not return until the
brenda server on the storage node has responded to this specific
request. 
Otherwise the storage node may be flooded with messages,
possibly causing a deadlock.
.PP
The \fIsnode\fR parameter specifies which node the object should be stored on
if it is declared as B_SETTLED in the flags parameter, or if the object's
tag is ANYTAG (0). 
Otherwise this parameter is ignored.
.SH "ADDITIONAL FUNCTIONS"
The following functions are used in debugging, and are not part of
the shared-memory scheme proper.
.PP
The two functions, FREEZE and THAW, may be used as an aid in debugging. 
The FREEZE call requests that
\fBall\fR the Brenda servers stop serving normal requests.
(It takes time to propogate the FREEZE request to all
the servers, and they do not all freeze at the exact same time.)
The FREEZE action causes
the current status of the tuple space to be preserved. 
After any one process calls FREEZE, any process that requests
a Brenda service blocks indefinitely, waiting for the reply from
a frozen server. 
Processes that do not call Brenda do not block.
Even the process that called FREEZE does not block until
its next normal Brenda call. 
After the freeze takes effect, additional FREEZE requests are ignored. 
To resume normal operation, the THAW function must be called.
The \fIbrenda\fR shell command also has options
to freeze and thaw the servers, but putting a call to FREEZE
inside the program being debugged has the advantage of knowing
where the program stops.
While frozen, the Brenda servers will accept a standard Brenda request
if the B_TORCH flag bit is set.  
.SH ERROR CODES
.TP 21
[EWOULDBLOCK]
TRYGRAB or TRYLOOK did not find the requested object.
.TP
[EMEMORY]
Necessary memory allocation failed on a remote node.
.TP
[EBADREQUEST]
The request type is unknown or a parameter is invalid.
.TP
[EMSGLEN]
The size parameter is too large or too small.
.TP
[EADDRESS]
The node address specified is invalid.
.TP
[EGARBAGE]
Memory has been overwritten with garbage somewhere.
.TP
[ENOEVENTS]
The Trollius kernel ran out of message descriptor space.
.TP
[EIMPOSSIBLE]
The wrong kind of message arrived (this should not happen).
.SH OTHER CONSIDERATIONS
To use these functions one must link to the \fIlibbrenda\fR library.
One must also load the brenda server,
\fIbrendad\fR, into each node on which objects may be stored, but not
into nodes that only send brenda requests but do not store objects.
In the current implementation, all compute nodes are storage nodes.
When loading \fIbrendad\fR, 1K of stack space is enough.
.SH ACKNOWLEDGEMENTS
The concepts behind Brenda were derived from the Linda project at Yale.
See:
.I "Linda on Hypercube Multicomputers"
by Nicholas Carriero and David Gelernter, pp45--56 in: M.T. Heath, editor,
.I "Hypercube Multiprocessors 1986"
---Proceedings of the First Conference on Hypercube Multiprocessors,
SIAM, Philadelphia. or:
.I "Linda and Friends"
by Ahuja, Carriero and Gelernter, IEEE \fICOMPUTER\fR, August, 1986, pp26--34.
.SH SEE ALSO
brenda(l)
