.TH BRENDALIB l  "March 12, 1990" "Release 2.0" "TROLLIUS PROGRAMMING TOOLS"
.SH NAME
brendalib
\- the Brenda library functions interface  (Trollius)
.SH SYNOPSIS
.nf
.ft B
#include <terror.h>
#include "brenda.h"

int	put (long tag, char *buffer, long size, int flags, long snode);
int	grab (long tag, char *buffer, long size, int flags, long snode);
int	look (long tag, char *buffer, long size, int flags, long snode);
int	trygrab (long tag, char *buffer, long size, int flags, long snode);
int	trylook (long tag, char *buffer, long size, int flags, long snode);
int	bump (long tag, long *oldvalue, long increment, int flags, long snode);

int	brenda (int request, long tag, char *buffer, long size, int flags, long snode);

int	freeze ();
int	thaw ();

int	_brenda (struct brenda_pb *argblk);
.ft R
.fi
.SH DESCRIPTION
The \fIbrenda\fR functions provide a form of virtual shared memory over a
network of Trollius nodes.  Objects are tagged and thrown into a pool.  The
first process to request an object with that tag picks the object out of
the pool.  Since objects cannot be modified while in place in the pool,
there are no synchronization problems.  The location of actual storage, if
necessary, and the details of the message passing scheme behind the brenda
services are hidden from the caller.  The overhead, for small objects, is
about three times that of direct message passing.  For objects that are
large relative to the maximum message size the transmission costs dominate
and the overhead is negligible.  Brenda avoids double-buffering as much as
possible.
.PP
The \fIput\fR function puts an object into the common
.I "tuple space"
(TS).  The object is simply the \fIsize\fR bytes starting
at the address \fIbuffer\fR.  A copy of the object is
made, so when \fIput\fR returns the caller may overwrite the data.
The \fIsize\fR must be greater than or equal to 4 bytes \fI(sizeof(long))\fR.
.PP
The \fIgrab\fR function removes a matching object from TS and returns
up to \fIsize\fR bytes of it into the provided byffer \fIbuffer\fR.
The function \fIlook\fR finds a matching object in TS and returns
a copy, but does \fInot\fR remove it from TS.
.PP
If no matching object is found in TS, \fIgrab\fR and \fIlook\fR block
until one appears.  The functions \fItrygrab\fR and \fItrylook\fR are
similar to \fIgrab\fR and \fIlook\fR, but do not block.  If no matching
object is found they return an error code.
.PP
The \fIbump\fR function is an optimized operation that combines reading
the old value of an integer object and adding a given constant to the
stored object.  The operation is logically equivalent to the combination:
.br
.nf
.ft I
	long tmp;
	cc = grab (tag, &oldvalue, sizeof(long), flags, snode);
	tmp = oldvalue + incr;
	(void) put (tag, &tmp, sizeof(long), flags, snode);
	return (cc);
.ft R
.fi
For the \fIbump\fR operation the affected object must already exist and
must be of the size of a \fIlong int\fR (4 bytes).  Note that in this case
the size is implied, and the amount to increment by takes the place of the
\fIsize\fR parameter.
.PP
The order of bytes inside the object is normally not touched.
In the special case where \fIsize\fR is equal to \fIsizeof(long)\fR,
the byte ordering inside the int is automatically converted
to fit each machine.  This was done to enable the \fIbump\fR function.  
So beware of the case where an object, larger than one integer,
is put by one kind of processor and grabbed by another.
.PP
An object is tagged by an arbitrary 32-bit non-zero integer value.
The value of the tag is up to the user, but the tag specified in
a \fIput\fR must not be zero, and the tag given to \fIgrab\fR or
\fIlook\fR must match the original tag for the object to be retrieved.
The tag "ANYTAG" (0) matches any other tag.
.PP
The \fIflags\fR parameter is normally 0.
The following values may be OR-ed into this parameter:
.TP 16
B_SETTLED
means the object is to be stored in the node specified by the parameter
\fIsnode\fR.  The \fIgrab\fR and \fIlook\fR calls must also specify
B_SETTLED and must know what \fIsnode\fR to specify or the object that was
\fIput\fR as B_SETTLED will not be found.  The flag B_LOCAL means
means that the object is to be stored in the local node, and is
equivalent to it being B_SETTLED in the node \fIgetnodeid\fR.
The default is B_NOMADIC (0),
which means the object may be stored anywhere, and the actual storage
location is a function of the tag.  The usage of B_SETTLED is recommended
as an optimization when the object is to be used by one consuming process
which runs on a known node.
.TP
B_GLOBAL
means that copies of the object are to be sent to \fIall\fR the
participating storage nodes, using a broadcast mechanism.
This is an optimization that is useful when it is known that
the object will be \fIlook\fR-ed at by processes on all
or most of the nodes.  The \fIlook\fR, too, must specify the
B_GLOBAL flag (or, equivalently, the B_LOCAL flag) for the optimization
to yield any gains by avoiding long-range communication.
A \fIgrab\fR of a B_GLOBAL object is not permitted, except in the special
case where the \fIgrab\fR is done by the same
process that originally created that object via a B_GLOBAL \fIput\fR.
In this special case a request to remove the object is broadcast
to all the storage nodes.
This restriction of a B_GLOBAL \fIgrab\fR to one process eliminates
the need for a global query, but does not guarantee that a last-minute
\fIlook\fR on a remote node will fail to find the object, even after
copies of the object have been removed from some other nodes.
.TP
B_VERIFY
means that a return message will be sent from the Brenda server
during the processing of a \fIput\fR request, slowing things down somewhat
but checking that there is space for storing the object.  If this
flag is not specified, a call can return \fIOK\fR while the storage
node is out of memory.  B_VERIFY is relevant only when \fIput\fR-ing
objects that are smaller than MAXNMSGLEN.  The default is B_BLIND (0).
It is recommended that VERIFY be used while debugging.  B_VERIFY can
also provide a form of flow control for processes that put a long
stream of objects, since each \fIput\fR call will not return until the
brenda server on the storage node has responded to this specific
request.  Otherwise the storage node may be flooded with messages,
possibly causing a deadlock.
.PP
The \fIsnode\fR parameter specifies which node the object should be stored on
if it is declared as B_SETTLED in the flags parameter, or if the object's
tag is ANYTAG (0).  Otherwise this parameter is ignored.
.PP
All of the functions described above (\fIput\fR, \fIgrab\fR, etc.)
are really macros using the common function \fIbrenda\fR.
An additional parameter, the request code, is added
as the leading parameter.  The request code is one of the following:
B_PUT, B_GRAB, B_LOOK, B_TRYGRAB, B_TRYLOOK or B_BUMP.
.SH "ADDITIONAL FUNCTIONS"
The following functions are used in debugging, and are not part of
the shared-memory scheme proper.
.PP
The two functions, \fIfreeze\fR and \fIthaw\fR, may be used
as an aid in debugging.  The \fIfreeze\fR call requests that
\fBall\fR the Brenda servers stop serving normal requests.
(It takes time to propogate the \fIfreeze\fR request to all
the servers, and they do not all freeze at the exact same time.)
The \fIfreeze\fR action causes
the current status of the tuple space to be preserved.  After
any one process calls \fIfreeze\fR, any process that requests
a Brenda service blocks indefinitely, waiting for the reply from
a frozen server.  Processes that do not call Brenda do not block.
Even the process that called \fIfreeze\fR does not block until
its next normal Brenda call.  After the freeze takes effect,
additional \fIfreeze\fR requests
are ignored.  To resume normal operation, the \fIthaw\fR
function must be called.
The \fIbrenda\fR shell command also has options
to freeze and thaw the servers, but putting a call to \fIfreeze\fR
inside the program being debugged has the advantage of knowing
where the program stops.
While frozen, the Brenda servers will accept a standard Brenda request
if the B_TORCH flag bit is set.  They also accept the debugging requests
described below.  Normally, these out-of-line requests would come
from the \fIbrenda\fR shell command, which is designed for
this sort of debugging.
.PP
The following request types are
accessible via the function \fI_brenda\fR.
The \fI_brenda\fR function accepts a pointer to a structure of type
.I "struct brenda_pb"
that serves both for holding parameters describing the request and
for returning information from the Brenda server.
The structure has the following form:
.nf
	struct brenda_pb {
		long	request;
		long	tag;
		char	*buffer;
		long	size;
		long	flags;
		long	snode;
		long	data;
	};
.fi
The \fI_brenda\fR function modifies the tag and size fields
of the structure to reflect the actual tag and size of an object
that has been found.  It also sets the \fIsnode\fR field to the absolute
address of the node to which the request was directed.
The possible request types are as follows:
.TP 16
_OBJ_SUM
Requests a short report on how many objects of the
specified tag are stored, and the total bytes used.
The results of the query are returned in the
\fItag, snode, size\fR and \fIdata\fR fields of the structure.
(The \fIdata\fR field holds the number of found objects.)
.TP
_REQ_SUM
Requests a short report on how many requests for objects,
of a specific tag, are queued.
.TP
_OBJ_LST
Requests a long report on objects stored, including the
tag and size of each.  The \fIbuffer\fR field must point to
an array of \fIstruct obj_lst\fR.  The \fIsize\fR field
must be set to the size, in bytes, of the array.
The array is filled with object descriptors, each
including the tag and size of the object.
.TP
_REQ_LST
Requests a long report on requests queued.  Each pending request
is described by the node and event (PID) of the requesting process,
the request code, tag requested, and size specified in the request.
The \fIbuffer\fR field must point to an array of \fIstruct req_lst\fR.  
.PP
In the last two cases the \fIsize\fR field, upon calling, must hold the size,
in bytes, of the buffer allocated by the caller.  This buffer must fit in one
message, i.e., its size must not exceed MAXNMSGLEN.
Upon successful return, the \fIdata\fR field holds the number of objects or
requests found (possibly 0).  If the space allocated for the long reports
is insufficient, the value returned in the \fIsize\fR field will be
larger than the number of slots in the \fIbuffer\fR, but the buffer
boundary will not be overwritten.
If the \fItag\fR is ANYTAG (0), all tags are included in the reports.
Otherwise, only objects or requests of the specified \fItag\fR are reported.
The storage node to query is implied by the tag and the hashing scheme,
except for the ANYTAG case (where the tag is 0, which implies SETTLED),
and the case where SETTLED is specified explicitly.  In the latter cases the
address of the storage node must be specified in the \fIsnode\fR
field.
.TP 16
_CLEAR_
Requests that all objects and requests of the specified tag be
deleted from the lists kept by the Brenda server on the implied node.
No data is sent nor received in this request -- the \fIbuffer\fR and
\fIsize\fR fields are ignored.
In the special case where ANYTAG (a zero) is given in the
\fItag\fR field, the server at the node specified in the \fIsnode\fR
field clears all its tables, returns all dynamically allocated
memory to the system, and starts afresh.  This call returns, in the
\fIsize\fR field of the header, the number of objects and/or requests deleted
(0 upon a full reset).
.SH RETURN VALUES
The Brenda functions return 0 if all went well.  In addition,
fields in the header passed to \fI_brenda\fR
are modified, as described above.
Upon error -1 is returned, and the global variable \fIerrno\fR is set to
one of the error codes below.
.SH ERROR CODES
.TP 21
[EWOULDBLOCK]
\fItrygrab\fR or \fItrylook\fR did not find the requested object.
.TP
[EMEMORY]
Necessary memory allocation failed on a remote node.
.TP
[EBADREQUEST]
The request type is unknown or a parameter is invalid.
.TP
[EMSGLEN]
The size parameter is too large or too small.
.TP
[EADDRESS]
The node address specified is invalid.
.TP
[EGARBAGE]
Memory has been overwritten with garbage somewhere.
.TP
[ENOEVENTS]
The Trollius kernel ran out of message descriptor space.
.TP
[EIMPOSSIBLE]
The wrong kind of message arrived (this should not happen).
.SH OTHER CONSIDERATIONS
To use these functions one must link to the \fIlibbrenda\fR library.
One must also load the brenda server,
\fIbrendad\fR, into each node on which objects may be stored, but not
into nodes that only send brenda requests but do not store objects.
In the current implementation, all compute nodes are storage nodes.
When loading \fIbrendad\fR, 1K of stack space is enough.
.SH ACKNOWLEDGEMENTS
The concepts behind Brenda were derived from the Linda project at Yale.  See:
.I "Linda on Hypercube Multicomputers"
by Nicholas Carriero and David Gelernter, pp45--56 in: M.T. Heath, editor,
.I "Hypercube Multiprocessors 1986"
---Proceedings of the First Conference on Hypercube Multiprocessors,
SIAM, Philadelphia. or:
.I "Linda and Friends"
by Ahuja, Carriero and Gelernter, IEEE \fICOMPUTER\fR, August, 1986, pp26--34.
.SH SEE ALSO
brenda(l)

