.TH CROS l "April 10, 1990" "Release 2.0" "TROLLIUS PROGRAMMING TOOLS"
.SH NAME
CrOS \- CrOS III Compatible Library
.SH DESCRIPTION
The CrOS III compatible functions make programming typical domain-decomposition
applications easier than with straight message-passing.  Thus, users of
Trollius are encouraged to consider the use of these library functions.
One CrOS function will sometimes replace a whole sequence of nsend(l)/nrecv(l)
calls, and remove the necessity of devising group communication algorithms
afresh in each application.
This manual page is oriented towards C programmers, but FORTRAN versions
of the functions do exist (see FORTRAN PROGRAMMING).
.PP
All of the functions documented at the end of the book
"Solving Problems on Concurrent Processors" by Fox et al. are
implemented, except for
.I rdstat, cubeldle
and 
.I cubeldve.
.PP
When porting programs designed for the original CrOS, remember that
programs that assume a hypercube topology will have to be modified.
Programs that use Parallel I/O (CUBIX) calls will need minor
modifications to use the Parallel I/O inherent in Trollius instead.
.SH CONFIGURING TROLLIUS TO USE THE LIBRARY
The CrOS III compatible Trollius library assumes a grid topology,
of any dimension (one, two, etc) up to a limit (currently four).
The physical machine must be booted to
support the desired grid, hopefully with direct links.
If direct links are missing, the grid will still be supported
through message forwarding, but at much reduced performance.
And some features will not work at all, such as the "welding"
of the grid connections into virtual circuits.
.PP
The n nodes must be labeled 0 ... n-1 so as to fit the grid order,
in row-major format.  For example, a 2-rows-by-3-columns 2-d mesh has 
node identifiers:
.PP
.nf
.if t			(0)	(1)	(2)	<--- second index
.if n			(0)	(1)	(2)	<------------ second index
		(0)	 0	 1	 2
		(1)	 3	 4	 5
.if t		 ^--------------------------------------------- first index
.if n		 ^------------------------------- first index
.fi
.PP
so that node 3 is at grid position (1,0).
Additionally, node 0 should be directly connected to the host.
.PP
Runtime errors will result if the node IDs are not 0..n-1, but no
errors will result if the physical connections do not fully support
the logical grid.  Message forwarding will be implicitly invoked.
.SH BEHAVIOR OF CROS III FUNCTIONS
See the manual pages at the end of the book by Fox et al. for detailed
descriptions of the library functions.  The rest of this section
only describes differences between the actual behavior of the
functions in this library and the descriptions in the book.
.PP
All programs using this library must #include <cros.h>.
The first CrOS call a program makes must be 
.I gridinit, 
in the standard CrOS III syntax (except when
.I gridsplit
is used, in which case it is invoked first).  The parameters given to 
.I gridinit 
must match the grid that was actually configured (see above).  After that, 
any legal CrOS calls are acceptable in any legal order.
.I Gridsplit
is used to split the processors of the grid between the specified grid
dimensions.  The number of processors in each dimension is returned in the 
dimension array.  The function returns a 0 when successfull and a non-zero
value otherwise. Note that the processors must have been numbered according
to the labeling described above, for the number of dimensions used.
.PP
Channel IDs (mask bit positions) are not the same as they
would be on a hypercube.  Instead, looking up along
grid axis K is channel 2*K+1, down is 2*K.  Note that the
sender of a message and the receiver of the same message use
the same physical channel but do 
.B not
use the same channel masks!
In any case, the function 
.I gridchan 
should always be
consulted for the correct channel mask rather than trying
to guess it.
.PP
Wrap-around at the edges of the grid is supported by default,
but may involve message-forwarding, and thus be inefficient,
in the case where the wrap-around links do not exist as normal
links in the hardware.  If an element in the array of grid sizes
passed to 
.I gridinit 
is negative, it is taken to mean that no
wrap-around is to be done on that axis, so relevant channel masks
will be set to 0.  Wrap-around on anything but real physical links
does not work under virtual circuits.
.PP
The sending functions return once outgoing data is sent out, although
outgoing data may not have been received yet at its destination.
Trollius currently does not provide for an acknowledgment when messages
are forwarded, and the use of separate acknowledgement messages is
not deemed worth the overhead.  Note, though, that in the case of
virtual circuits between all grid neighbors, the sender and receiver
do synchronize.
.PP
No error is returned from any receiving functions upon reception of
less data than was sent.  There is no 
.I errcnt 
variable since we
cannot put anything useful in it.
.PP
The host processor ID is whatever its Trollius ID is, not necessarily 1024.
.PP
Other minor, function-specific, differences between the implemented CrOS
functions and their description in the book follow.
.PP
.I Cwrite 
and 
.I vwrite 
return 0 if the 
.I mask 
parameter is 0.
.I Cread 
and 
.I vread 
return 0 if the 
.I inmask 
parameter is 0.
.PP
.I Cread 
and 
.I vread 
cannot operate unless given an actual buffer.
If given NULLPTR, an error is returned.
.PP
.I Cread 
and 
.I vread 
do 
.B not 
forward the tail of the message if they
did not have a large enough buffer to receive it.
.PP
Both 
.I vread 
and 
.I vwrite 
use malloc(l) to allocate a big block to
temporarily hold the whole vector in gathered form.  If this
malloc(l) fails the 
.I vread 
or 
.I vwrite 
call fails.
.PP
.I Vshift 
requires two calls to malloc(l), for input and output buffers for
the gathered forms of the vectors.
.PP
.I Broadcast 
.B does 
pick up and forward the message directed to
the processor even if given a NULL buffer pointer, although the
data is discarded.  Unlike the original CrOS, it does
not return the number of bytes that the CP sent.  Instead, it
returns the number of bytes received.
.PP
.I Combine 
allows a NULL buffer if 
.I nitems 
is 0.
Such a data-less combine is useful as a synchronization barrier:
.ti +5
combine (NULLPTR, (int (*)()) 0, 0, 0);
.PP
.I Concat 
requires that the 
.I inbuf 
parameter be valid (not NULL),
and it is assumed that the supplied 
.I inbuf 
has space for at least (
.I n 
* 
.I binsz
) bytes, where 
.I n 
is the number of nodes in the grid.  A NULL 
.I inbuf 
is an error.
.PP
.I Concat 
is rather inefficient if 
.I nbytes 
is much smaller than 
.I binsz,
since 
.I binsz 
is the amount actually passed in messages.
.PP
.I Bcastelt 
does not return the number of bytes that the CP sent.
Instead, it returns the number of bytes received.
.PP
The functions 
.I fdumpcp 
and 
.I floadcp 
use malloc(l) to get a buffer for 
.B one 
processor's data.
.PP
.I Floadcp 
and 
.I mloadcp 
return an error if the number of nodes in
.I bufmap[0] 
does not 
.B match 
the actual number of processors, whether
it is higher or lower.
.PP
.I Cparam 
is not void (as is seems to be in the book).
Instead it returns 0 (OK) or -1 (ERROR).
.SH ADDITIONAL FUNCTIONS
A number of utility functions have been added to the standard CrOS III
basic set.  Most of these additional functions are not intended
for use by application programs.  Their use is possible, but is not normally
of any advantage, and they are not portable.  Here is a list of the more
useful ones:
.PP
int	cweld (void);
.PP
This requests that all the grid connections be set up as virtual
circuits.  This reduces communications setup times tremendously.
But while the grid is welded, no nonlocal Trollius system calls
are allowed.  This includes I/O to the host, except from node 0, which can
always talk to the host, since the link from it to the host
is not part of the grid.  The 
.I loadelt 
and 
.I dumpelt
functions use direct host I/O from all nodes and therefore are better not
used while the grid is welded, although an automatic temporary
unwelding is done if they are called.  
.I Cweld 
should not be used
unless each and every grid connection (including wrap-around if allowed)
exists as an independent physical link in the hardware.  Call 
.I cweld
from all the grid nodes at once.  Do not call on the host.
.PP
int	cunweld (void);
.PP
This call undoes what 
.I cweld 
does.  It should be used before non-CrOS host I/O, such as 
printf(3s), is attempted.  Call 
.I cunweld
from all the grid nodes at once.  Do not call on the host.
.PP
int	griddims (int *num);
.PP
This call fills the given array 
.I num
with the length of the grid
along each of the dimensions. It is used by some of the library
functions, and may be used by user-written libraries, to supplement
the information returned by 
.I cparam.  
This is the same information
given to 
.I gridinit 
as its second argument.
.PP
In addition to the functions, there are two global variables used:
.PP
extern int	errno;
.PP
This standard global holds the code identifying the most recent error.
.PP
extern int	cwelded;
.PP
This global variable knows whether virtual circuits are in effect or not.
Holds HOLD (nonzero, defined in <net.h>) or 0.  This variable may be looked
at, but not altered, by user programs.
.SH DEBUGGING PROGRAMS
Virtual circuits should 
.B not 
be used until the application is fully debugged.
With virtual circuits off, one can insert a printf(3s) anywhere in the code,
in any node, at any time.  Additionally, while the program is running (or
crashing), one can query the system on any node from commands on the host,
such as state(l), mstate(l), and bfstate(l).  
In order to interpret the reports from state(l) and bfstate(l), 
one must know what Trollius events and types are used
for CrOS messages.  The messages use events that identify the direction
of travel in the grid.  The event encodes both the axis traveled on,
the direction of travel along that axis, and whether the message is
wrapping around the end of the grid.
.PP
Here is a table of events used in one and two dimensional grids:
.PP
.nf
.if n .ta 10n 20n 30n 40n 50n
.if t .ta 1.0i 2.0i 3.0i 4.0i 5.0i
Axis	Going	Going 	Over-	Under-
Index	Up	Down	Wrap	Wrap

0	3	1	19	17

1	7	5	23	21

.fi
Two more events are used for communication to and from the control processor,
and are defined in the header file <cros.h>: CTOHOST (34) and CFROMHOST (38).
.PP
The message type encodes the CrOS function involved.  Here is the list
of types (defined in the header file <cros.h>):
.PP
.if n .ta 12n 24n 36n
.if t .ta 1.25i 2.5i 3.75i
.nf
Symbol	Value	Functions

CRDWR	0x001	\fIcwrite/cread/vwrite/vread/cshift\fR
CDUMP	0x004	\fIdumpelt/fdumpcp/mdumpcp\fR
CLOAD	0x008	\fIloadelt/floadcp/mloadcp\fR
CBCAST	0x010	\fIbroadcast\fR (part of \fIbcastelt\fR, too)
CCOMBINE	0x020	\fIcombine\fR (part of \fIcweld\fR and \fIcombelt\fR, too)
CCONCAT	0x040	\fIconcat\fR: main data
CCONCATS	0x080	\fIconcat\fR: size array
CBCASTCP	0x100	\fIbcastelt/bcastcp\fR
CCOMBCP	0x200	\fIcombelt/combcp\fR: normal case
CCOMBCPE	0x400	\fIcombelt/combcp\fR: error case
CWELD	0x800	\fIcweld/cunweld\fR
.fi
.SH FORTRAN PROGRAMMING
The FORTRAN functions mentioned in the book by Fox et al. are also
implemented, but at this point for grid nodes only, i.e., not for
the host node.  Their names are in upper case, for example, KCREAD, KVSHIF,
etc.  Note that when calling KCPARA, the second argument must point
to an array of 6 integers, not 5.  Functions added to those in the
book include:
.TP 14
KCWELD
same as \fIcweld\fR above.
.TP
KCUNWD
same as \fIcunweld\fR above.
.TP
KCWDED
returns the value of the "cwelded" variable.
.TP
KFULMK
same as \fIfullmask\fR above.
.TP
KGRDDI
same as \fIgriddims\fR above.
.SH ERRORS
Finally, in the case when CrOS functions return errors, make sure to
inspect the 
.I errno 
global variable.  It may hold a standard UNIX or
Trollius error code (defined in <errno.h> or <terror.h>) or a CrOS-specific
error code (defined in <cros.h>), which may be one of the following:
.TP 17
[ENUMPROC]
asked \fIgriddims\fR for 0 or too many processors.
.TP
[EBADMASK]
channel mask has too many or few bits set.
.TP
[EBADCHAN]
channel mask includes invalid set bit(s).
.TP
[EBADBUF]
null data buffer.
.TP
[EOVERLAP]
input and output data buffers overlap.
.TP
[ESIZEDIFF]
\fIinsize\fR != \fIoutsize\fR in self-\fIvshift\fR.
.TP
[ENUMPROCS]
\fIbufmap[0]\fR does not fit actual \fInproc\fR.
.TP
[ENOGRID]
\fIgridinit\fR not called.
.TP
[EGRIDDIM]
grid dimension does not match.
.TP
[EDATASIZE]
data block too large for buffer or system.
.TP
[EFAKEGRID]
welding logical links that are not physical.
.SH SEE ALSO
"Solving Problems on Concurrent Processors", volume 1, by G. Fox,
M. Johnson, G. Lyzenga, S. Otto, J. Salmon and D. Walker.
Prentice Hall, 1988, ISBN 0-13-8230226
.sp
The Trollius CrOS III Compatible Library.  Programming Tools Series 2/1.
Cornell University Theory Center and The Ohio State University Office of 
Research Computing.
