.TH RTV l "March 25, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rtvclr, rtvrect, rtvsnake, rtvtext, rtvstate \- Control remote graphics screens.
.SH SYNOPSIS
.nf
.ft B
#include <tvreq.h>

.if n int	rtvclr (int nodeid);
.if n int	rtvrect (int nodeid, int x1, int y1, int xlen, int ylen,
.if n		char *colours);
.if n int	rtvsnake (int nodeid, int x1, int len, char *colours);
.if n int	rtvtext (int nodeid, int x1, int y1, int colour,
.if n		char *text);
.if n int	rtvstate (int nodeid, struct tvstate *ptvstate);
.if t int	rtvclr (int nodeid);
.if t int	rtvrect (int nodeid, int x1, int y1, int xlen, int ylen, char *colours);
.if t int	rtvsnake (int nodeid, int x1, int len, char *colours);
.if t int	rtvtext (int nodeid, int x1, int y1, int colour, char *text);
.if t int	rtvstate (int nodeid, struct tvstate *ptvstate);
 
.fi
.SH DESCRIPTION
Some processor boards in a multicomputer system might be equipped with
dedicated graphics frame buffers.
In addition, typically the host node is a workstation running a windowing
environment capable of supporting high resolution graphics.
This allows the user to plot the results of the parallel computation as it
happens, providing immediate visual feedback (and nice demos !).
Such video capabilities are managed by graphics daemons, mtvd(l) on ITB
nodes, and xmtvd(l) on OTB nodes.
.PP
The functions described here give the user a primitive interface to these
graphical capabilities.
In all of these functions, the
.I nodeid
parameter refers to the nodeid of the node that is running the graphics daemon.
Before these functions can be used intelligently, some basic information
must be learned about the graphics frame buffer.
The 
.I rtvstate
function returns key information about the frame buffer by filling a tvstate
structure passed by reference.
Tvstate is defined in <tvreq.h> as :
.PP
.nf
struct tvstate {
.if n	int	tvs_xlen;		/* screen pixel width */
.if n	int	tvs_ylen;		/* screen pixel height */
.if t	int	tvs_xlen;	/* screen pixel width */
.if t	int	tvs_ylen;	/* screen pixel height */
	int	tvs_ncolour;	/* number of colours */
	int	tvs_xtext;	/* screen character width */
	int	tvs_ytext;	/* screen character height */
};
.fi
.PP
The
.I rtvclr
function clears the frame buffer by setting all pixels to the background
colour (black).
.PP
The
.I rtvrect
function plots a rectangular region of pixels anywhere inside the frame buffer.
.I X1
and
.I y1
are the pixel coordinates of the upper left corner of the rectangle.
.I Xlen
and
.I ylen
are the width and height of the rectangle, i.e. the dimensions of the rectangle
along the X and Y axes.
.I Colours
is an array of (xlen * ylen) bytes, one per pixel.
Each pixel is an index into the colour map whose size is given in the
information returned by the
.I rtvstate
function.
.PP
The
.I rtvsnake
function is similar to 
.I rtvrect.
It plots a snake-like region in the frame buffer. 
This region is defined by a starting pixel whose single ordinate
.I x1
is an integer ranging from 0 for the upper left pixel in the screen, to
(tvs_xlen * tvs_ylen) \- 1, for the lower right pixel in the screen.
The length of the snake in pixels is given by
.I len.
.I Colours
is the array of bytes representing the colour of each pixel, as used in the
.I rtvrect
function.
The region plot starts at the 
.I x1
pixel and proceeds along the horizontal rows of the frame buffer, from left
to right, for the given length of the snake.
.PP
The
.I rtvtext
function views the screen in characters, not in pixels, but otherwise
operates much like
.I rtvsnake.
At the starting character coordinates
.I x1
and 
.I y1, rtvtext
plots the null-terminated ASCII character array (string)
.I text
across the horizontal character row of the screen.
.I Colour
is the single colour used to plot all characters in the string.
.SH "RETURN VALUE"
Upon successful completion, all functions return 0.
If an error occurs, these functions return
\-1, and set the global error variable 
.I errno
to the error code.
