.TH TRADE l "March 30, 1992" "Release 2.2" "TROLLIUS UTILITIES"
.SH NAME
trade  \- Prepares ttrace data for viewing with ParaGraph.
.SH SYNTAX
.B trade
[-hV] [<infile>] [-o <outfile>]
.br
.SH OPTIONS
.TP 14
.B \-h
Print useful information on this tool.
.TP
.B \-V
Format and print the ttrace data in ASCII with no conversion.
.TP
.B <infile>
Redirect standard input from \fIinfile\fR.
.TP
.B \-o <outfile>
Redirect standard output to \fIoutfile\fR.
.SH DESCRIPTION
.PP
.I trade
is a Trollius utility which converts the output file written by ttrace(l) into
a file which can be read by ParaGraph and used to visualize the execution of
a Trollius application.  This allows the application programmer to detect
critical paths and bottlenecks in communication, appraise load-balancing and
processor utilization, and other nifty things.
.PP
Trace data is generated by using loadgo(l) with the -t option to start the
application.  The data is collected from the nodes into a file using the
ttrace command.  Finer control of tracing is provided by the functions
ktrstop(l) and ktrstart(l), which allow changing of trace identifiers (TIDs) 
and selective tracing of regions in the application.
.PP
If 
.I trade
is invoked with no infile, the input is read from standard input.  Likewise,
if no outfile is given, the output is written to standard output.
.PP
Three passes are made over the data.  The first pass performs three tasks:
trstart/trstop trace records with no trace events between them are removed,
the time stamp of the earliest meaningful start (i.e., not removed by the 
above 
process) is found, and a list is built which maps the TIDs to their nodes.
This 'cleaned' input file is written to a temporary file in /tmp as each record
is examined.  This out-of-core approach allows for processing of trace files
whose size is limited only by disk storage capacity.
.PP
The second pass reads the cleaned input, one record at a time, and generates
the appropriate trace records in PICL format.  PICL (Portable Instrumented
Communication Library) is the format expected by ParaGraph.  These records
are written to another file in /tmp.
.PP
The correspondence between Trollius trace events and PICL events is as follows:
For each Trollius trstart record, one trace_start PICL record is generated.
For each Trollius trend	record, one trace_exit and two compstats PICL records
are generated.  For each Trollius trsend, a PICL send record is generated, but
the to_node field is filled in with the TID rather than the node (this is
taken from the map created in pass one).  For each Trollius trrecv, either one
PICL recv_blocking, one compstats, one recv_waking and another compstats are 
generated or else one PICL recv is generated, followed by two compstats.  
.PP
The recv record is generated if no clock ticks occur between
the entry into nrecv and the return.  The recv_blocking/recv_waking pair is
generated if some time elapses.  The idea here is to show the difference
between an nrecv where the message is already available, and an nrecv which
blocks.  Usually the nrecv takes enough time to execute to give
the appearance of blocking for a short time.  'Small' times here should be
interpreted as non-blocking receives.
.PP
When the second pass has finished, 
.I trade
makes a third pass to sort the temporary file containing the PICL records, 
placing the output into outfile, or else writing it to standard ouput if 
outfile was not specified.  This
obviates the need for running the shell script supplied with ParaGraph to
order the records.
.SH EXAMPLES
.TP 4
.I "trade mbozo.dat -o mbozo.trf"
Reads in the file called mbozo.dat, which must have been created with ttrace,
converts it to ParaGraph form, and writes the sorted output into mbozo.trf.
.TP
.I "trade mbozo.dat -V"
Reads in the file called mbozo.dat, which must have been created with ttrace,
and formats it in ASCII.  This prints only the ttrace data and performs no
conversion to ParaGraph form.
.SH DIAGNOSTICS
.PP
If some part of the tracefile supplied is unintelligible to
.I trade,
the program will print the following error message and halt:
.TP 14
"trade: unknown trace record: %d, aborting."
.PP
If a program changes its TID more than once, the user is warned with:
.TP 14
"trade: warning, TID %d has been reset to TID %d, continuing."
.PP
In this case, execution continues, but the resulting tracefile may not 
be meaningful.
.SH BUGS
.PP
.I trade
assumes that the clocks on all ITB nodes are synchronized.  For this to
be the case, the boot schema used to boot the machine must contain only 
one link to the host.  If this is not done, the clocks on all ITB nodes
will not be synchronized, and the view from ParaGraph will
be meaningless; it will appear that many sends occurred at time zero.  As
a result, Trollius code running on the host should not be included in the
trace, since currently the host's clock is not synchronized with the ITB
nodes.
.SH FILES
.TP 4
/tmp/trclean.??????
the file which has redundant start/stop pairs removed
.PP
.TP 4
/tmp/trtmpout.??????
the file which holds the unsorted output
.PP
Both of these files are removed before 
.I trade
exits, but abnormal termination could cause them to be left behind.
.SH "SEE ALSO"
loadgo(l), ttrace(l), ktrstart(l)
