.TH WAVE l  "May 2, 1991" "Release 2.1" "TROLLIUS APPLICATIONS"
.SH NAME
\Solve the one-dimensional wave equation problem using Trollius.
.SH DESCRIPTION
The time-dependent motion of a simple vibrating string is described 
by a hyperbolic partial differential equation. The solution is the 
vibration amplitude A(x,t) expressed as a function of position and time.
The problem becomes fully posed with the addition of boundary 
conditions on the spatial domain, and initial position and velocity 
distributions. We solve the problem by a numerical method which 
lends itself to a natural concurrent decomposition. The problem 
decomposition is based upon the assignment of spatial groups of nodal 
points to processors. 
.PP 
The basic algorithm is :
.TP 12
1. Identify the linear topology.
.TP
2. Load the node program on each compute node.
.TP
3. Run the host program.
.TP
4. Set problem domain and subdomain size.
.TP
5. Decide on the parameter values.
.TP
6. Initialize starting values for vibration amplitude.
.TP
7. Exchange end nodal values with neighbouring nodes.
.TP
8. Perform update on all nodal points.
.TP
9. Repeat for specified number of iterations.
.PP
Given the data domain size, speedup or efficiency can be achieved by 
distributing the data domain over a large number of processors. 
However communication inefficiency dictates assignment of 
large numbers of nodal points to each processor. The subdomain 
size should be a balance between the two conflicting goals. 
Load Balancing is assured by assigning equal numbers of nodal points 
to each processor since the problem is homogeneous.
.PP
The processor configuration is a linear array of N processors connected 
to a control processor. In Trollius environment, this topology must 
be specified in a boot schema or the 
.I bnail 
file. The host node will function as the control processor and 
it can be assigned any nodeid. The N compute 
nodes must be assigned nodeid's 0 through N-1. The bnail file must 
include bidirectional paths (of length greater than or equal to 1) 
between node i and i+1 (for exchanging nodal values) and paths 
such that all compute nodes are reachable from the host node 
(for broadcasting parameters) and vice versa (for sending final 
data values back from the compute nodes to the host). 
.PP
The input to the wave program consists of the bnail file, the 
corresponding route file, and parameters - domain size (npts),
time step size (delt), and number of time steps (nstep). The output 
is the vibration amplitudes at each of the points of the domain.
.PP
More sophisticated versions of the wave program include the 
.I cubix 
code which eliminates the control processor, the 
.I cros 
code which replaces the lengthy communication codes by special 
crystalline communication routines, and finally the cubixcros version 
which combines the features of both cubix and cros versions.
.SH SEE ALSO 
nail(l), cubix(l), cros(l)

