/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	brenda.c,v $
 * Revision 2.2  92/04/24  16:16:45  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:04:19  gdburns
 * Ohio release
 * 
 * Revision 2.0.1.3  91/04/25  17:29:28  gdburns
 * "max" is a bad name for a variable.
 * 
 * Revision 2.0.1.2  90/10/15  13:25:07  raghu
 * Changed the format of debug/error messages to conform to standards.
 * 
 * Revision 2.0.1.1  90/10/12  15:59:20  raghu
 * Changed brenda to pick up brendad from default directory if not available
 * in current directory.
 * 
 * Revision 2.0  90/09/26  12:36:43  gdburns
 * Homo sapien
 * 
 *
 *	Function:	- command-line program to get brenda status information
 *			- also allows loading, dooming, freezing and thawing the
 *			  brenda servers, clearing the object space, manually
 *			  putting and grabbing, etc.
 */

#include <sys/file.h>		/* file I/O constants */
#include <sys/types.h>		/* for types in struct stat definition */
#include <sys/stat.h>		/* for struct stat definition */
#include <stdio.h>

#include "brenda.h"

#include <typical.h>
#include <t_types.h>		/* type conversion routines */
#include <terror.h>		/* error codes */
#include <kio.h>		/* for rloadgo() RTF_* flags */
#include <preq.h>		/* for rstate() call parameters */
#include <ksignal.h>		/* for SIGUDIE */
#undef NULL
#include <bufreq.h>		/* for rbuffer req types */
#undef NULL
#define NULL 0

/*
 * External functions.
 */

extern char *findpath();
extern char *getparam();	 /* this was in do_args.h: */
extern int parse_nodes();	 /* these were in parse_nodes.h: */
extern char *mnemonic();
extern void getnode();
extern void mnusage();

#define MAXLINES	64	/* max lines in report per snode */

extern int errno;		/* Global error variable */

long	tag, size, node, mynode;
int	verbose, nbase, nnodes;
long	ibuf [MAXNMSGLEN / sizeof(long)];
long	*nodes;
char	*filename;
int 	n_brenda; 		/* # of brendad pathname prefixes */
char 	**p_brenda;		/* brendad pathname prefix 	*/
char 	*f_brenda;		/* brenda daemon */

/*
 *
 *	  main
 *
 *	  Accepts  : - argument string which do_args will process
 *
 */

main (argc, argv)
int    argc;    
char   *argv[];
{
	long	mnode;
	int	i, index, pindex, flags;
	void	usage(), do_param(), req_report(), obj_report();
	void	manipulate(), install(), clear(), doom();
	void	Freeze(), Thaw();

#ifdef DEBUG
printf ("brenda: started\n");
#endif

	/* process command-line arguments */

	validopts ("hxvrtspglbICDfnzu");
	exclusive ("rpglbICDzu");
	followed  ("tspglbfn");
	if (do_args (&argc, argv) < 0) {
		if (errno == EUSAGE)
			usage();
		else
			puts ("brenda(main): do_args returned unknown error");
			exit(errno);
	}

	if (opt_taken('h')) {
		help();
		exit(0);
	}

	verbose = opt_taken ('v');

	if (opt_taken ('x'))
		nbase = 16;
	else
		nbase = 10;

	filename = NULL;
	if (opt_taken ('f'))
		filename = getparam ('f');
	if (opt_taken ('S'))
		filename = getparam ('S');
	if (opt_taken ('R'))
		filename = getparam ('R');
	if (opt_taken ('A'))
		filename = getparam ('A');
	if ((   opt_taken('f') || opt_taken('S')
	     || opt_taken('R') || opt_taken('A'))
	&&  filename == NULL) {
			puts ("brenda(main): need a file name");
			usage();
			exit (errno);
	}

	i = 0;
	do_param ('s', &i);
	size = (long) i;

	i = 0;
	do_param ('t', &i);
	do_param ('p', &i);
	do_param ('g', &i);
	do_param ('l', &i);
	do_param ('b', &i);
	tag = (long) i;

	i = NOTNODEID;
	flags = 0;
	do_param ('n', &i);
	node = i;

	if (parse_nodes (&argc, argv) < 0 || argc > 1) {
		mnusage();
		exit (errno);
	}

	/* attach to kernel */

	if (kinit(0) == ERROR) {
		terror("brenda(main)");
		exit(errno);
	}

#ifdef DEBUG
printf ("brenda(main): kinit() OK\n");
#endif

	/* get the list of storage nodes */

	nnodes = getngrid();		/* use only GRID nodes */
	if (nnodes < 1) {		/* host only */
		nnodes = 1;
		mynode = getnodeid();
		nodes = &mynode;
	} else {
		nodes = (long *) malloc (nnodes * sizeof(long));
		if (nodes == NULL) {
			puts("brenda(main):malloc failed.");
			my_kexit(errno);
		}
	}
	if (getgnodes (nodes, nnodes) != 0) {
		terror("brenda(main)");
		my_kexit(errno);
	}

	/* parse and check node IDs given on command line */

	getnode (&pindex, &mnode, &flags);	/* mnemonics given? */
	index = (-1);
	while (index != pindex) {
#ifdef DEBUG
printf ("brenda(main): mnemonics given: node %s\n", mnemonic(mnode));
#endif
		for (i=0; i<nnodes; i++)
			if (mnode == nodes[i])  break;
		if (i == nnodes)	{
			printf ("brenda(main): node %s not a server node\n",
					mnemonic(mnode));
			my_kexit (0);
		}
		getnode (&index, &mnode, &flags);
	}

#ifdef DEBUG
printf("brenda(main): verbose=%c, tag=%d, size=%d, node=%s, filename=%s\n",
verbose? 'y' : 'n', tag, size,
(node==NOTNODEID? "(none)" : mnemonic(node)),
(filename? filename : "(none)"));
#endif

	/* do the chosen op */

	switch (opt_chosen('r')) {

		case 'r':
			req_report();
			break;
		case 'p':
			manipulate (B_PUT);
			break;
		case 'g':
			manipulate (B_TRYGRAB);
			break;
		case 'l':
			manipulate (B_TRYLOOK);
			break;
		case 'b':
			manipulate (B_BUMP);
			break;
		case 'I':
			install();
			break;
		case 'C':
			clear();
			break;
		case 'D':
			doom();
			break;
		case 'z':
			Freeze();
			break;
		case 'u':
			Thaw();
			break;
		default:
			obj_report();
			break;
	}

	my_kexit(0);
}

void
do_param (c, p)
	int c, *p;
{
	void usage();

	if (opt_taken (c) && intparam (c, p) != 0) {
		usage();
		exit (errno);
	}
}

void
usage()
{
puts ("usage: brenda [-h] [-v] [-x] [-r] [-t <tag>] [-{p|g|l|b} <tag>]");
puts ("   [-f <filename>] [-s <size>] [{-n<node#>|<node(s)>}] [-{I|z|u|C|D}]");
}

help()
{
    printf("\nSynopsis:\tbrenda [options] <nodes>\n");
    printf("\nDescription:\tMonitor and control a Brenda daemon.\n");
    printf("\nOptions:        -h\t\tthis help message\n");
    printf("\t\t-v\t\tturn on verbose mode\n");
    printf("\t\t-x\t\tdisplay tags and PIDs in hexadecimal\n");
    printf("\t\t-r\t\treport requests rather than objects\n");
    printf("\t\t-t <tag>\tuse this tag\n");
    printf("\t\t-p <tag>\tput (create) an object\n");
    printf("\t\t-g <tag>\tget (read and destroy) an object\n");
    printf("\t\t-l <tag>\tlook at an object (non-destructive get)\n");
    printf("\t\t-b <tag>\tbump (look and increment) an integer object\n");
    printf("\t\t-s <size>\tsize of object to create or read, in bytes,\n");
    printf("\t\t\t\t  or amount to bump an integer object\n");
    printf("\t\t-f <filename>\tfile to write object obtained by -g or -l,\n");
    printf("\t\t\t\t  or file to read object created by -p\n");
    printf("\t\t-n <node>\tspecify node explicitly\n");
    printf("\t\t-u\t\tunfreeze (thaw) the Brenda servers\n");
    printf("\t\t-z\t\tfreeze the Brenda servers\n");
    printf("\t\t-C\t\tclear all objects and requests\n");
    printf("\t\t-D\t\t-C, and doom the Brenda servers\n");
    printf("\t\t-I\t\tinstall the brenda servers\n");
    printf("\t\t<none>\t\treport stored Brenda objects\n");
    mnusage();
    printf("\nExample:        brenda n0-7\n");
    printf("\t\t\t\"Report stored objects on nodes 0 thru 7.\"\n\n");
}

/*
 *	Report objects in one node.
 */
void
one_oreport (snode)
	long snode;
{
	int i, count;
	unsigned int lsize;
	struct obj_lst *list;
	struct brenda_pb header;
	char *format;

	if (! verbose) {
		header.request = _OBJ_SUM;
		header.tag     = tag;
		header.flags   = B_SETTLED;
		header.snode   = snode;
		if (_brenda (&header) != B_OK)
			printf ("brenda(one_oreport): unable to query node %s\n"
				,mnemonic(snode));
		if (header.data == 0)
			return;		/* silent if no objects found */
		if (tag == ANYTAG) {
			printf ("    (all)     %8d     %8d      %s\n",
				header.data, header.size, mnemonic(snode));
		} else {
			if (nbase == 16)
				format = "0x%08x    %8d     %8d     %s\n";
			else
				format = "%9d     %8d     %8d      %s\n";
			printf (format, tag, header.data,
					header.size, mnemonic(snode));
		}
		return;
	}

	list = (struct obj_lst *) ibuf;
	count = MAXNMSGLEN / sizeof(struct obj_lst);
	if (count > MAXLINES)  count = MAXLINES;
	lsize = count * sizeof(struct obj_lst);
	header.request = _OBJ_LST;
	header.tag     = tag;
	header.buffer  = (char *) list;
	header.size    = lsize;
	header.flags   = B_SETTLED;
	header.snode   = snode;
	if (_brenda (&header) != B_OK)
		printf ("brenda(one_oreport): unable to query node %s\n", 
			mnemonic(snode));
	for (i=0; i<header.data && i<count; i++) {
		if (nbase == 16)
			format = "0x%08x    %8d       %s\n";
		else
			format = "%9d     %8d       %s\n";
		printf (format, list[i].tag, list[i].size, mnemonic(snode));
	}
	if (count < header.data)
		printf ("etc...\n");
}

/*
 *  Report about objects
 */
void
obj_report()
{
	long	mnode;
	int	i, index, pindex, flags;

	if (verbose)
		printf ("     TAG          SIZE      SNODE\n");
	else
		printf ("     TAG          COUNT        SIZE     SNODE\n");

	if (node == NOTNODEID) {
		getnode (&pindex, &mnode, &flags);	/* mnemonics given? */
		index = (-1);
		while (index != pindex) {		/* yes */
			one_oreport (mnode);
			getnode (&index, &mnode, &flags);
		}
		if (pindex < 0) {
			if (tag != ANYTAG)
				one_oreport (hashnode(tag));
			else
				for (i=0; i<nnodes; i++)
					one_oreport (nodes[i]);
		}

	} else {				/* node ID specified */
			one_oreport (node);
	}
}

/*
 *	Report requests in one node.
 */
void
one_rreport (snode)
	long snode;
{
	int i, count;
	unsigned int lsize;
	struct req_lst *list;
	struct brenda_pb header;
	char *format;

	list = (struct req_lst *) ibuf;
	count = MAXNMSGLEN / sizeof(struct req_lst);
	if (count > MAXLINES)  count = MAXLINES;
	lsize = count * sizeof(struct req_lst);
	header.request = _REQ_LST;
	header.tag     = tag;
	header.buffer  = (char *) list;
	header.size    = lsize;
	header.flags   = B_SETTLED;
	header.snode   = snode;
	if (_brenda (&header) != B_OK)
		printf ("brenda(one_rreport): unable to query node %s\n", 
			mnemonic(snode));
	for (i=0; i<header.data && i<count; i++) {
		if (nbase == 16)
			format = "0x%08x   %6d        %s";
		else
			format = "%9d     %6d       %s";
		printf (format, list[i].tag, list[i].size,
			mnemonic(list[i].node));
		if (nbase == 16)
			format = "      0x%08x      %s\n";
		else
			format = "    %10d        %s\n";
		printf (format, list[i].event, mnemonic(snode));
	}
	if (count < header.data)
		printf ("etc...\n");
}

/*
 *  Report about requests
 */
void
req_report()
{
	long	mnode;
	int	i, index, pindex, flags;

	printf ("     TAG       RSIZE      CNODE       CPID        SNODE\n");

	if (node == NOTNODEID) {

		getnode (&pindex, &mnode, &flags);	/* mnemonics given? */
		index = (-1);
		while (index != pindex) {		/* yes */
			one_rreport (mnode);
			getnode (&index, &mnode, &flags);
		}
		if (pindex < 0) {
			if (tag != ANYTAG)
				one_rreport (hashnode(tag));
			else
				for (i=0; i<nnodes; i++)
					one_rreport (nodes[i]);
		}

	} else {				/* node ID specified */
			one_rreport (node);
	}
}

/*
 *  Display a block of bytes in hex.
 */
void
hex (buf, nbytes)
	char *buf;
	int nbytes;
{
	register int  c, i, j;
	int  offset, ceil;
	char string[20];

	if (nbytes == sizeof(long)) {
		printf("  Type 	= integer \n");
		printf("  Value	= %ld\n", *((long *) buf));
		return;
	}

	offset = 0;
	string[16] = '\0';
	ceil = ((nbytes + 15) & (~15));
	for (i=0; i<ceil; i++) {
		j = (i & 15);
		if (j == 0)
			printf ("%06x: ", offset);
		if (i < nbytes) {
			c = (0xFF & (int) buf[i]);
			printf (" %02x", c);
			if (c >= ' ' && c <= '~')	/* assumes ASCII */
				string[j] = c;
			else
				string[j] = '.';
		} else {
			printf ("   ");
			string[j] = '\0';
		}
		if (j == 15) {
			printf ("   %s\n", string);
			offset += 16;
		}
	}		
	if (j != 15)
		printf ("\n", offset);
}

/*
 *  Try and put, grab or look at an object.
 *  If grab or look and not found, complain.
 *  If verbose, display object contents, in hex.
 */
void
manipulate (opcode)
	int opcode;
{
	int  i, ii, r, s, flags, malloced, fd;
	long bsize, nodeid;
	char *cp;
	struct obj_lst list;
	struct brenda_pb header;
	struct stat statbuf;

	if (node == NOTNODEID) {
		getnode (&i, &nodeid, &flags);
		if (i >= 0) {
			node = nodeid;
			getnode (&ii, &nodeid, &flags);
			if (ii != i) {
			    printf("brenda(manipulate):no multicasts.\n");
			    return;
			}
		}
	}
	if (tag == ANYTAG && node == NOTNODEID) {
		printf("brenda(manipulate):untagged object needs node spec\n");
		return;
	}
	flags = (node == NOTNODEID ?  B_NORMAL :  B_SETTLED);
	if (size == 0 || opcode == B_BUMP) {
		if (opcode == B_PUT) {
			if (filename) {
			    s = stat (filename, &statbuf);
			    if (s) {
			    printf("brenda(manipulate):cannot access file %s\n",
				filename);
				return;
			    }
			    size = statbuf.st_size;
			    if (size < sizeof(long)) {
			        printf("brenda(manipulate):file '%s' short\n",
					filename);
				return;
			    }
			} else {
			    size = sizeof(long);	/* put an int */
			}
		} else /* != B_PUT */ {
			header.request = _OBJ_LST;
			header.tag     = tag;
			header.buffer  = (char *) &list;
			header.size    = sizeof (list);
			header.flags   = flags;
			header.snode   = node;
			if (_brenda (&header) != B_OK) {
			printf ("brenda(manipulate):unable to query node %s\n", 
				mnemonic(node));
				return;
			}
			if (header.data == 0) {
				printf("brenda(manipulate):object not found\n");
				return;
			}
			/*   (This is not quite right since this object
			 *   might be grabbed by somebody else before
			 *   we do the grab or look below.)
			 */
			if (opcode == B_BUMP) {
			    if (list.size != sizeof(long)) {
			    printf("brenda(manipulate):cannot bump non-ints\n");
			        return;
			    } else {
				if (size == 0)	    /* increment not given: */
				        size = 1;	/* increment by 1 */
			    }
			} else /* != B_BUMP */ {
				size = list.size;
			}
		}
	}
	if (opcode == B_BUMP)
		bsize = sizeof(long);
	else
		bsize = size;
	if (bsize > MAXNMSGLEN) {
		cp = malloc ((unsigned int) bsize);
		if (cp == NULL) {
			puts("brenda(manipulate):out of memory");
			return;
		}
		malloced = 1;
	} else {
		cp = (char *) ibuf;
		malloced = 0;
	}
	if (opcode == B_PUT) {
		if (filename) {
			fd = open (filename, O_RDONLY, 0);
			if (fd < 0) {
			printf ("brenda(manipulate):unable to open file %s\n", 
				filename);
				if (malloced)  free (cp);
				return;
			}
			s = read (fd, cp, (int) bsize);
			if (s < bsize) {
				if (s >= sizeof(long))
					bsize = s;
				else {
					printf("brenda(manipulate):error reading file %s\n", filename);
				    if (malloced)  free (cp);
				    return;
				}
			}
			close (fd);
		} else {
			for (i=0; i<bsize; i++)
				cp[i] = 0;
		}
	}
	if (opcode == B_BUMP)
		s = trygrab (tag, cp, bsize, flags, node);
	else
		s = brenda (opcode, tag, cp, bsize, flags, node);
	if (s == B_OK && opcode != B_PUT && (bsize == sizeof(long) || verbose))
		hex (cp, (int) bsize);
	while (opcode != B_PUT && filename != NULL) {
		fd = open (filename, O_WRONLY | O_TRUNC | O_CREAT, 0644);
		if (fd < 0) {
			printf ("brenda(manipulate): unable to create file %s\n"				, filename);
			break;
		}
		if (write (fd, cp, (int) bsize) < bsize)
			printf ("brenda(manipulate): error writing file %s\n",
				filename);
		close (fd);
		break;
	}
	if (malloced)
		free (cp);
	if (opcode == B_BUMP) {
		ii = *((int *) cp) + size;
		r = put (tag, (char *) &ii, (long) sizeof(long), flags, node);
		if (r != B_OK) {
		printf("brenda(manipulate):'put' part of 'bump' failed\n");
			return;
		}
		printf("\n");
		printf("  Bumped value by %d. \n",size);
		printf("  New value = %d\n", ii);
	}
	if (s != B_OK) {
		if (opcode == B_PUT)
			cp = "put";
		else if (opcode == B_TRYGRAB)
			cp = "trygrab";
		else if (opcode == B_TRYLOOK)
			cp = "trylook";
		else
			cp = "bump";
		if (opcode != B_PUT && errno == EWOULDBLOCK)
			printf ("brenda(manipulate): object not found\n");
		else
			printf("brenda(manipulate): %s() returned error %d\n", 
				cp, errno);
	}
}

/*
 *  Freeze and thaw the Brenda servers.
 */
void
Freeze()
{
	if (freeze() != B_OK)
		printf ("brenda(Freeze): unable to freeze\n");
}
void
Thaw()
{
	if (thaw() != B_OK)
		printf ("brenda(Thaw): unable to thaw\n");
}

/*
 *  Clear object space objects in one node.
 */
void
one_clear (snode)
	long snode;
{
	struct brenda_pb header;

	header.request = _CLEAR_;
	header.tag     = tag;
	header.flags   = B_SETTLED;
	header.snode   = snode;
	if (_brenda (&header) != B_OK)
		printf ("brenda(one_clear): unable to clear node %s\n", 
				mnemonic(snode));
	if (rbuffer (snode, BFSWEEP, 0) < 0)
		terror("brenda(one_clear)");
}

/*
 *	Install the brenda servers.
 *	Reset existing servers, doom redundant ones.
 */
void
install()
{
	int i, s, t, pid;
	long nodeid;
	struct {
		char *p[2];
		char name[8];
	} argvst;
	struct pstate state;
	char *fullpath; 	/* brenda's fullpath */

/*
 * Initialize varibales for searching.
 */
	n_brenda = 0;
	p_brenda = (char **) 0;
	f_brenda = DEFBRENDAD;
/*
 *	Create a list of paths to be searched. The current directory is
 *	first searched and then the default directory DEFPBRENDAD.
 */

	addarg( &n_brenda, &p_brenda, "");
	addarg( &n_brenda, &p_brenda, ".");
	addarg( &n_brenda, &p_brenda, DEFPBRENDAD);

	fullpath = findpath(f_brenda,p_brenda,S_IEXEC);
	if ( !(fullpath) ) {	
		return;
	} 
	
	strcpy (argvst.name, DEFBRENDAD);

	getnode (&i, &nodeid, &t);
	if (node != NOTNODEID || i >= 0) {
	    printf ("brenda(install): Install does not take node specifiers\n");
	    return;
	}

	if (tag != ANYTAG) {
		printf ("brenda(install): Install does not take a tag\n");
		return;
	}

#ifdef BCAST
	if (verbose)
		printf ("doom (old brendad's)...\n");

	s = rdoom (HOST2GRID, SELECT_BLKD, BRENDA, SIGUDIE);
	if (s < 0) {
		printf ("brenda(install): rdoom() failed\n");
		return;
	}

	if (verbose)
		printf ("sweep...\n");
	s = rbuffer (HOST2GRID, BFSWEEP, 0);
	if (s < 0) {
		printf ("brenda(install): rbuffer(BFSWEEP) failed\n");
		return;
	}
	if (verbose)
		printf ("buffer -b%d...\n", BNBUFS);
	s = rbuffer (HOST2GRID, BFWORKER, BNBUFS);
	if (s < 0) {
		printf ("brenda(install): rbuffer(BFWORKER) failed\n");
		return;
	}

	if (verbose)
		printf("brenda: loading %s into %d nodes...\n",fullpath,nnodes);
	s = rloadgo (HOST, HOST2GRID, BDSTACK, RTF_NOWAIT,
		fullpath, 1, &argvst, sizeof(argvst), &pid);
	if (s < 0) {
	    printf ("brenda(install): rloadgo() failed\n");
	    return;
	}
#else
	for (i=0; i<nnodes; i++) {
		if (verbose)
			printf ("rstate...\n");
		s = rstate (nodes[i],SELECT_BLKD,BRENDA,&state,sizeof(state));
		if (s < 0) {
		    printf ("brenda(install): rstate() call to node %s failed\n"
			,mnemonic(nodes[i]));
			return;
		}
		if (s > 1) {
			if (verbose) 
		    	    printf("brenda:dooming brendad copies in node %s\n"
				,mnemonic(nodes[i]));
			s = rdoom (nodes[i], SELECT_BLKD, BRENDA, SIGUDIE);
			if (s < 0) {
			printf("brenda(install):rdoom() to node %s failed\n",
						mnemonic(nodes[i]));
				return;
			}
			s = 0;		/* no brendad's left now */
		}
		if (verbose)
			printf ("sweep %s...\n", mnemonic(nodes[i]));
		if (rbuffer (nodes[i], BFSWEEP, 0) < 0) {
			printf ("brenda: rbuffer(BFSWEEP) failed\n");
			return;
		}
		if (verbose)
			printf ("buffer -b%d %s...\n",
				BNBUFS, mnemonic(nodes[i]));
		if (rbuffer (nodes[i], BFWORKER, BNBUFS) < 0) {
			printf ("brenda(install): rbuffer(BFWORKER) failed\n");
			return;
		}
		if (s > 0) {
			if (verbose) 
			    printf("brenda:clearing brendad in node %s..\n"
				, mnemonic(nodes[i]));
			one_clear (nodes[i]);
		} else {
			if (verbose)
			    printf ("brenda: loading brendad into node %s...\n",
					mnemonic(nodes[i]));
			s = rloadgo (HOST, nodes[i], BDSTACK, RTF_NOWAIT,
				fullpath, 1, &argvst, sizeof(argvst), &pid);
			if (s < 0) {
			    printf ("brenda: rloadgo() into node %s failed\n",
					mnemonic(nodes[i]));
			    return;
			}
		}
	}
#endif	BCAST
}

/*
 *  Clear all or part of object space
 */
void
clear()
{
	long	mnode;
	int	i, index, pindex, flags;

	if (tag == ANYTAG && node == NOTNODEID)
		if(verbose)
		    printf ("brenda: Clearing object space...\n");
	else
		if (verbose)
		    printf ("brenda: Clearing part of object space...\n");

	if (node == NOTNODEID)	{

		getnode (&pindex, &mnode, &flags);	/* mnemonics given? */
		index = (-1);
		while (index != pindex) {		/* yes */
			one_clear (mnode);
			getnode (&index, &mnode, &flags);
		}
		if (pindex < 0) {
			if (tag != ANYTAG)
				one_clear (hashnode(tag));
			else
				for (i=0; i<nnodes; i++)
					one_clear (nodes[i]);
		}
		
	} else {				/* node ID specified */
		one_clear (node);
	}
}

/*
 *   Doom the servers.
 */
void
doom()
{
	int	i, s;
	long	n;
	struct pstate state;

	getnode (&i, &n, &s);
	if (node != NOTNODEID || i >= 0) {
		printf ("brenda(doom): Doom does not take node specifiers\n");
		return;
	}

#ifdef BCAST
	if (verbose)
	    printf ("brenda: Dooming the brendad servers...\n");
	s = rdoom (HOST2GRID, SELECT_BLKD, BRENDA, SIGUDIE);
	if (s < 0) {
		printf ("brenda(doom): rdoom() failed\n");
		return;
	}
#else
	for (i=0; i<nnodes; i++) {
		n = nodes[i];
/*		s = rstate (n, SELECT_BLKD, BRENDA, &state, sizeof(state));
		if (s < 0) {
			printf ("brenda(doom): rstate call to node %s failed\n",
					mnemonic(n));
			return;
		}
*/ s = 1;
		if (s > 0) {
			printf ("brenda(doom): dooming brendad in node %s...\n",
					mnemonic(n));
			s = rdoom (n, SELECT_BLKD, BRENDA, SIGUDIE);
			if (s < 0) {
				printf("brenda(doom):rdoom to node %s failed\n",
					 mnemonic(n));
				return;
			}
		}
	}
#endif
}

/*
 *  This is necessary due to a current Trollius bug:
 */
my_kexit (code)
	int code;
{
	fflush (stdout);
	fflush (stderr);
	kexit (code);
}
