/*
 *      OSU Research Computing                            	Trillium
 *      Copyright 1992 The Ohio State University
 *      GDB
 *
 *      $Log:	lestate.c,v $
 * Revision 2.2  92/04/24  16:17:23  trillium
 * Ohio Release
 * 
 *	Function:	- monitors LEDs on remote Trollius nodes
 *	Accepts:	- lestate <nodeids>
 */

#include <ledreq.h>
#include <net.h>
#include <stdio.h>
#include <terror.h>
#include <typical.h>
#include <shared.h>

/*
 * external variables
 */
extern int      errno;          	/* global error variable */

/*
 * external functions
 */
extern int	rlestate();		/* get LED status */

/*
 * local functions
 */
static void	help();

/*
 * syntax for lestate command
 */
static char *usage = "lestate [-h] <nodeids>";

int
main(argc, argv)

int   			argc;
char 			*argv[];

{
	int		n_index;		/* index of node entry */
	int		nodeid;			/* value of node entry */
	int		n_flags;		/* extra node entry info */
	struct lestate	state;			/* LED status */
/*
 * Parse the command line.
 */
	validopts("h");
 
	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {
 
		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n\n", usage);
		} else {
			terror("lestate (args)");
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Initialize and attach to Trollius.
 */
	if (kinit(0)) {
		terror("lestate (kinit)");
		exit(errno);
	}
/*
 * Get and display status for every target node.
 */
	getnode(&n_index, &nodeid, &n_flags);
 
	if (n_index < 0) {
		fprintf(stderr, "usage: %s\n\n", usage);
		kexit(EUSAGE);
	}

	printf("NODE	#LED	SHADOW		CC\n");

	do {
		if (rlestate(nodeid, &state) < 0) {
			terror("lestate (rlestate)");
			kexit(errno);
		}

		printf("%-7.7s", mnemonic(nodeid));
		printf(" %-7d 0x%-8x",
			state.les_nled, state.les_shadow);

		if (state.les_cc == NOTNODEID) {
			printf("\tnone\n");
		} else {
			printf("\t%s\n", mnemonic(state.les_cc));
		}

		getnode(&n_index, &nodeid, &n_flags);

	} while (n_index);

	kexit(0);
}

/*
 *	help
 *
 *	Function:	- Print helpful information on the lestate command.
 */

static void
help()
{
	printf("\nSynopsis:\tlestate <options> <nodeids>\n");
        printf("\nDescription:\tMonitor remote LEDs.\n");
        printf("\nOptions:");
	printf("\t-h\t\tthis help message\n");
	mnusage();
	printf("\t\th	(local host node)\n");
	printf("\nExample:	lestate n7\n");
	printf("\t\t\t\"Print LED status of node 7.\"\n\n");
}
