/*
 *      OSU Research Computing                            	Trillium
 *      Copyright 1992 The Ohio State University
 *      RBD
 *
 *      $Log:	mtv.c,v $
 * Revision 2.2  92/04/24  16:18:18  trillium
 * Ohio Release
 * 
 *	Function:	- controls graphics daemon on remote Trollius nodes
 */

#include <stdio.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>
#include <shared.h>
#include <tvreq.h>

/*
 * external variables
 */
extern int      	errno;		/* global error variable */

/*
 * external functions
 */
extern int		rtvclr();	/* clear remote graphics frame */

/*
 * local variables
 */
static char		*me;		/* tool name */

/*
 * local functions
 */
static void		help();

/*
 * syntax for mtv command
 */
static char *usage = "usage: %s [-h] <nodeids>\n";

int
main(argc, argv)

int			argc;
char			*argv[];

{
	int		n_index;		/* index of node entry */
	int4		nodeid;			/* value of node entry */
	int		n_flags;		/* extra node entry info */
/*
 * Parse the command line.
 */
	me = argv[0];
	validopts("h");
 
	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {
 
		if (errno == EUSAGE) {
			fprintf(stderr, usage, me);
		} else {
			terror(me);
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Initialize and attach to Trollius.
 */
	if (kinit(0)) {
		terror(me);
		exit(errno);
	}
/*
 * Perform the mtvd operation for all target nodes.
 */
	getnode(&n_index, &nodeid, &n_flags);
 
	if (n_index < 0) {
		fprintf(stderr, usage, me);
		kexit(EUSAGE);
	}

	do {
		if (rtvclr(nodeid)) {
			terror(me);
			kexit(errno);
		}

		getnode(&n_index, &nodeid, &n_flags);

	} while (n_index);

	kexit(0);
}

/*
 *	help
 *
 *	Function:	- Print helpful information on the mtv command.
 */
static void
help()

{
	printf("\nSynopsis:\t%s <options> <nodeids>\n", me);
        printf("\nDescription:\tControl remote graphics daemons.\n");
        printf("\nOptions:");
	printf("\t-h\t\tthis help message\n");
	mnusage();
	printf("\t\th	(local host node)\n");
	printf("\t\tC	(all compute nodes)\n");
	printf("\t\tH	(all host nodes)\n");
	printf("\t\tN	(all nodes)\n");
	printf("\nExample:	mtv n100\n");
	printf("\t\t\t\"Clear the graphics frame on node 100.\"\n\n");
}
