/*
 *      OSU Research Computing                            	Trillium
 *      Copyright 1992 The Ohio State University
 *      GDB
 *
 *      $Log:	tled.c,v $
 * Revision 2.2  92/04/24  16:19:42  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/24  14:20:01  gdburns
 * major rewrite
 * 
 * Revision 2.1  91/03/20  02:52:37  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:26:23  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:53:48  trillium
 * Homo erectus
 *
 *	Function:	- controls LEDs on remote Trollius nodes
 *	Accepts:	- see the usage variable
 *	Examples:	- tled -s 5 n50-52 
 *			  Set (turn on) first and third LEDs on nodes 50,51,52
 *			- tled -S 15 n5
 *			  Set the shadow register on node 5 to value 15.
 */

#include <ledreq.h>
#include <stdio.h>
#include <terror.h>
#include <typical.h>
#include <shared.h>

/*
 * external variables
 */
extern int      errno;          	/* global error variable */

/*
 * external functions
 */
extern int	rledcc();		/* LED copy & forward */
extern int	rledop();		/* basic LED operations */

/*
 * local functions
 */
static void	help();
static int	ledop();

/*
 * syntax for tled command
 */
static char *usage =
	"tled [-h] [-scn <bits>] [-S <shadow>] [-f <nodeid>] <nodeids>";

int
main(argc, argv)

int			argc;
char			*argv[];

{
	int		n_index;		/* index of node entry */
	int		nodeid;			/* value of node entry */
	int		n_flags;		/* extra node entry info */
	int		ccnodeid;		/* carbon copy nodeid */
/*
 * Parse the command line.
 */
	validopts("h");
	followed("cfnsS");
 
	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {
 
		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n\n", usage);
		} else {
			terror("tled (args)");
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Initialize and attach to Trollius.
 */
	if (kinit(0)) {
		terror("tled (kinit)");
		exit(errno);
	}
/*
 * Perform all LED operations.
 */
	getnode(&n_index, &nodeid, &n_flags);
 
	if (n_index < 0) {
		fprintf(stderr, "usage: %s\n\n", usage);
		kexit(EUSAGE);
	}

	do {
		if (ledop(nodeid, (int) 's', LEDSET) < 0) {
			terror("tled (op)");
			kexit(errno);
		}

		if (ledop(nodeid, (int) 'c', LEDCLR) < 0) {
			terror("tled (op)");
			kexit(errno);
		}

		if (ledop(nodeid, (int) 'n', LEDNOT) < 0) {
			terror("tled (op)");
			kexit(errno);
		}

		if (ledop(nodeid, (int) 'S', LEDALL) < 0) {
			terror("tled (op)");
			kexit(errno);
		}

		if (opt_taken('f')) {

			if (intparam('f', &ccnodeid) < 0) {
				terror("tled (cc)");
				kexit(errno);
			}

			if (rledcc(nodeid, ccnodeid) < 0) {
				terror("tled (cc)");
				kexit(errno);
			}
		}

		getnode(&n_index, &nodeid, &n_flags);

	} while (n_index);

	kexit(0);
}

/*
 *	ledop
 *
 *	Function:	- remote LED operation
 *
 *	Accepts:	- remote nodeid
 *			- tled option
 *			- LED opcode
 *
 *	Returns:	- 0 or ERROR
 */

static int
ledop(nodeid, option, opcode)

int4			nodeid;
int			option;
int4			opcode;

{
	int4		bits;		/* bit pattern argument to opcode */

	if (opt_taken(option)) {

		if (intparam(option, &bits) < 0) {
			return(ERROR);
		}

		if (rledop(nodeid, opcode, bits) < 0) {
			return(ERROR);
		}
	}

	return(0);
}

/*
 *	help
 *
 *	Function:	- Print helpful information on the tled command.
 */

static void
help()
{
	printf("\nSynopsis:\ttled <options> <nodeids>\n");
        printf("\nDescription:\tControl remote LEDs.\n");
        printf("\nOptions:");
	printf("\t-h\t\tthis help message\n");
	printf("\t\t-c <bits>\tclear LEDs enabled in <bits>\n");
	printf("\t\t-f <nodeid>\tcarbon copy operations to <nodeid>\n");
	printf("\t\t-n <bits>\tnegate LEDs enabled in <bits>\n");
	printf("\t\t-s <bits>\tset LEDs enabled in <bits>\n");
	printf("\t\t-S <shadow>\tset shadow register to <shadow>\n");
	mnusage();
	printf("\t\th	(local host node)\n");
	printf("\t\tC	(all compute nodes)\n");
	printf("\nExample:	tled -s 0xff n7\n");
	printf("\t\t\t\"Turn on LEDs 0 thru 7 on node 7.\"\n\n");
}
