/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	trstate.c,v $
 * Revision 2.2  92/04/24  16:26:00  trillium
 * Ohio Release
 * 
 *	Function:	- command to print Trillium traced client info
 *	Accepts:	- see the "usage" variable
 *	Examples:	- trstate n10
 *			- print status about monitored processes on node 10
 */

#include <portable.h>
#include <stdio.h>
#include <terror.h>
#include <trreq.h>
#include <typical.h>

/*
 * external variables
 */
extern int 	errno;		/* global error variable */

/*
 * external functions
 */
extern void	validopts();	/* set the valid options */
extern void	terror();
extern void	intparam();
extern void	followed();
extern char	*mnemonic();
extern char	*malloc();
extern int	do_args();	/* parse the valid options */
extern int	getnode();	/* get next node entry */
extern int	opt_taken();	/* Was a particular option selected? */
extern int	parse_nodes();	/* parse the node identifiers */
extern int4	rtrstate();	/* get remote traced status */

/*
 * local functions
 */
static void	header();
static void	help();
static void	do_trstate();

/*
 * local variables
 */
static char *_copyright =
	"Copyright 1991 The Ohio State University $Revision: 2.2 $";

static struct trstate	*trtable;		/* traced state descs */
static int4		trsmax;			/* max entries */

/*
 * syntax for trstate command
 */
static char *usage = "trstate [-h] [-B <maxmsgs>] <nodes>";

main(argc, argv)

int   			argc;    
char   			*argv[];

{
	int4		n_index;		/* index of node entry */
	int4		nodeid;			/* value of node entry */
	int4		n_flags;		/* extra node entry info */
/*
 * Parse the command line.
 */
	validopts("hB");
	followed("B");

	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n", usage);
		} else {
			terror("trstate (arguments)");
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Allocate traced state table.
 */
	if (opt_taken('B')) {
		 intparam('B', &trsmax);
	} else {
		trsmax = TRSMAX;
	}

	trtable = (struct trstate *) 
		malloc((unsigned) (trsmax * sizeof(struct trstate)));

	if (! trtable) {
		terror("trstate (malloc)");
	}
/*
 * Attach to kernel.
 */
	if (kinit(0) < 0) {
		terror("trstate (kinit)");
		exit(errno);
	}
/*
 * Loop through all specified nodes.
 */
	(void) getnode(&n_index, &nodeid, &n_flags);

	if (n_index >= 0) {
/*
 * Print status header.
 */
		header();

		do {
			if (n_flags) {
				fprintf(stderr, "usage: %s\n", usage);
				kexit(errno);
			}

			do_trstate(nodeid);
			(void) getnode(&n_index, &nodeid, &n_flags);
		} while (n_index);
	} else {
		fprintf(stderr, "usage: %s\n", usage);
		kexit(errno);
	}

	kexit(0);
}

/*
 *	do_trstate
 *
 *	Function:	- find and print the trace state of a remote node
 *			- may exit program
 *
 *	Accepts:	- node identifier
 */
static void
do_trstate(nodeid)

int4			nodeid;

{
	int4		nstate;		/* # of return status descs */
	int4		limit;		/* # of blocks to print */
	int4		i;		/* favourite index */
	int4		dettot;		/* detached trace length total */

	if ((nstate = rtrstate(nodeid, trtable, trsmax, &dettot)) < 0) {
		terror("trstate");
		kexit(errno);
	}

	limit = (nstate < trsmax) ? nstate : trsmax;

	for (i = 0; i < limit; ++i) {
/*
 * Look for the spacer descriptor.
 */
		printf("%-9.9s", mnemonic(nodeid));
		printf(" %-11d", trtable[i].trs_pid);
		printf(" %-11d", trtable[i].trs_length);

		if (trtable[i].trs_sig) {
			printf(" yes\n");
		} else {
			printf(" no\n");
		}
	}

	printf("%-9.9s", mnemonic(nodeid));
	printf(" %-11s", "detached");
	printf(" %-11d\n", dettot);

	if (nstate > trsmax) {
		printf("...and more traced processes...\n");
	}
}

/*
 *	  header
 *
 *	  Function:	- prints the trstate table header
 */
static void
header()
{
	printf("NODE      PID         LENGTH      SIGNAL?\n");
}

/*
 *	  help
 *
 *	  Function:	- prints helpful information on the trstate command
 */
static void
help()
{
	printf("\nSynopsis:	trstate [options] <nodes>\n");
	printf("\nDescription:	Print information on traced processes.\n");
	printf("\nOptions:	-h	this help message\n");
	printf("\t\t-B <#>	set maximum # of messages to print\n");
	mnusage();
	printf("\t\th\t(local host node)\n");
	printf("\nExample:	trstate n3-9,11\n");
	printf("\t\t\t\"Get traced status from nodes 3 thru 9 & 11.\"\n\n");
}
