/*
 *      OSU Research Computing                            	Trillium
 *      Copyright 1992 The Ohio State University
 *      RBD
 *
 *      $Log:	tvstate.c,v $
 * Revision 2.2  92/04/24  16:29:31  trillium
 * Ohio Release
 * 
 *	Function:	- returns status of graphics daemons on remote nodes
 *	Accepts:	- tvstate <nodeids>
 */

#include <stdio.h>
#include <portable.h>
#include <net.h>
#include <terror.h>
#include <typical.h>
#include <shared.h>
#include <tvreq.h>

/*
 * external variables
 */
extern int      	errno;         	/* global error variable */

/*
 * external functions
 */
extern int		rtvstate();	/* get mtvd status */

/*
 * local variables
 */
static char		*me;		/* tool name */

/*
 * local functions
 */
static void		help();

/*
 * syntax for tvstate command
 */
static char *usage = "usage: %s [-h] <nodeids>\n";

int
main(argc, argv)

int   			argc;
char 			*argv[];

{
	int		n_index;		/* index of node entry */
	int4		nodeid;			/* value of node entry */
	int		n_flags;		/* extra node entry info */
	struct tvstate	state;			/* mtvd status */
/*
 * Parse the command line.
 */
	me = argv[0];
	validopts("h");
 
	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {
 
		if (errno == EUSAGE) {
			fprintf(stderr, usage, me);
		} else {
			terror(me);
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Initialize and attach to Trollius.
 */
	if (kinit(0)) {
		terror(me);
		exit(errno);
	}
/*
 * Get and display status for every target node.
 */
	getnode(&n_index, &nodeid, &n_flags);
 
	if (n_index < 0) {
		fprintf(stderr, usage, me);
		kexit(EUSAGE);
	}

	printf("NODE	XLEN	YLEN	NCOL.	XTEXT	YTEXT\n");

	do {
		if (rtvstate(nodeid, &state)) {
			terror(me);
			kexit(errno);
		}

		printf("%-7.7s", mnemonic(nodeid));
		printf(" %-7d %-7d %-7d %-7d %-7d\n",
			state.tvs_xlen, state.tvs_ylen, state.tvs_ncolour,
			state.tvs_xtext, state.tvs_ytext);

		getnode(&n_index, &nodeid, &n_flags);

	} while (n_index);

	kexit(0);
}

/*
 *	help
 *
 *	Function:	- Print helpful information on the tvstate command.
 */
static void
help()

{
	printf("\nSynopsis:\t%s <options> <nodeids>\n", me);
        printf("\nDescription:\tGet status of remote graphics daemons.\n");
        printf("\nOptions:");
	printf("\t-h\t\tthis help message\n");
	mnusage();
	printf("\t\th	(local host node)\n");
	printf("\nExample:	tvstate n100\n");
	printf("\t\t\t\"Print status of graphic daemon on node 100.\"\n\n");
}
