/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	rtrfget.c,v $
 * Revision 2.2  92/04/24  16:32:50  trillium
 * Ohio Release
 * 
 *	Function:	- unloads trace data from remote trace daemon
 *			- receives trace data from traced length is 0
 *			- writes trace data into file
 *
 *	Accepts:	- target node ID
 *			- file descriptor
 *
 *	Returns:	- total length of received trace data
 */

#include <trreq.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* error number */
/*
 * external functions
 */
char			*malloc();
/*
 * local functions
 */
static int4		do_rtr();

int4
rtrfget(nodeid, fd)

int4			nodeid;
int			fd;

{
	return(do_rtr(nodeid, fd, TRQGET));
}

int4
rtrfforget(nodeid, fd)

int4			nodeid;
int			fd;

{
	return(do_rtr(nodeid, fd, TRQFORGET));
}

static int4
do_rtr(nodeid, fd, req)

int4			nodeid;
int			fd;
int4			req;

{
	struct nmsg	nhead;		/* network message desc. */
	struct trreq  	*request;	/* traced request */
	struct trreply	*reply;		/* traced reply */
	int4		trtotal;	/* total # replied descs */
	int4		mask;		/* favourite index */

	trtotal = 0;
	request = (struct trreq *) nhead.nh_data;
	reply = (struct trreply *) nhead.nh_data;

	request->trq_src_node = (nodeid == LOCAL) ? nodeid : getnodeid();
	request->trq_src_event = -getpid();
	request->trq_req = req;

	nhead.nh_node = nodeid;
	nhead.nh_event = EVTRACED;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();
	nhead.nh_type = 0;
	nhead.nh_length = MAXNMSGLEN;

	nhead.nh_msg = malloc((unsigned) MAXNMSGLEN);

	if (nhead.nh_msg == (char *) 0) {
		return(ERROR);
	}

	if (nrecv(&nhead)) {
		(void) ksigsetmask(mask);
		return(ERROR);
	}

	while ((nhead.nh_length > 0) || (reply->trr_errno)) {

		if (write(fd, nhead.nh_msg, nhead.nh_length) < 0) {
			(void) ksigsetmask(mask);
			return(ERROR);
		}

		trtotal += nhead.nh_length;
		nhead.nh_length = MAXNMSGLEN;

		if (nrecv(&nhead)) {
			(void) ksigsetmask(mask);
			return(ERROR);
		}
	}

	(void) ksigsetmask(mask);

	if (reply->trr_errno) {
		errno = reply->trr_errno;
		return(ERROR);
	} else {
		return(trtotal);
	}
}
