/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	rtrstate.c,v $
 * Revision 2.2  92/04/24  16:32:51  trillium
 * Ohio Release
 * 
 *	Function:	- gather state from a remote traced
 *			- receives trace status from traced until
 *			  reply code is EEOF
 *
 *	Accepts:	- target node ID
 *			- ptr to trace status table
 *			- # of entries in table
 *			- ptr to detached trace length total
 *
 *	Returns:	- number of replied descriptors or ERROR
 */

#include <trreq.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <terror.h>
#include <typical.h>

#define NMSGSTATE (MAXNMSGLEN / sizeof(struct trstate))

/*
 * external variables
 */
extern int		errno;		/* error number */

int4
rtrstate(nodeid, tbldesc, maxstate, dettot)

int4			nodeid;		/* target node ID */
int4			maxstate;	/* # state descs in table */
struct trstate		tbldesc[];	/* ptr to desc table */
int4			*dettot;	/* detached trace length total */

{
	struct nmsg	nhead;		/* network message desc. */
	struct trreq  	*request;	/* traced request */
	struct trreply	*reply;		/* traced reply */
	int4		nstate;		/* minimum # descs */
	int4		totstate;	/* total # replied descs */
	int4		mask;		/* favourite index */

	totstate = 0;
	request = (struct trreq *) nhead.nh_data;
	reply = (struct trreply *) nhead.nh_data;

	request->trq_src_node = (nodeid == LOCAL) ? nodeid : getnodeid();
	request->trq_src_event = -getpid();
	request->trq_req = TRQSTATE;

	nhead.nh_node = nodeid;
	nhead.nh_event = EVTRACED;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();
	nhead.nh_type = 0;
	nhead.nh_msg = (char *) tbldesc;

	do {
		nstate = (maxstate > NMSGSTATE) ? NMSGSTATE : maxstate;
		nhead.nh_length = nstate * sizeof(struct trstate);

		if (nrecv(&nhead)) {
			return(ERROR);
		}

		if (! reply->trr_errno || (reply->trr_errno == EEOF)) {
			totstate += reply->trr_nstate;

			nstate = (nstate > reply->trr_nstate)
				? reply->trr_nstate : nstate;
			maxstate -= nstate;
			tbldesc += nstate;
		}

	} while (! reply->trr_errno);

	(void) ksigsetmask(mask);

	if (reply->trr_errno == EEOF) {
		*dettot = reply->trr_length;
		return(totstate);
	} else {
		errno = reply->trr_errno;
		return(ERROR);
	}
}
