/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	rtvrect.c,v $
 * Revision 2.2  92/04/24  16:32:53  trillium
 * Ohio Release
 * 
 *	Function:	- plot a rectangular region of pixels
 *
 *	Accepts: 	- destination node
 *			- horizontal ordinate of first pixel
 *			- vertical ordinate of first pixel
 *			- rectangle width (horizontal)
 *			- rectangle height (vertical)
 *			- array of pixel colours
 *
 *	Returns:	- 0 or ERROR
 */

#include <portable.h>
#include <terror.h>
#include <net.h>
#include <events.h>
#include <typical.h>
#include <tvreq.h>

/*
 * external variables
 */
extern int		errno;


int
rtvrect(node, x1, y1, xlen, ylen, colours)

int4			node,
			x1,
			y1,
			xlen,
			ylen;
char			*colours;

{
	struct nmsg	nhead;		/* message header */
	struct tvreq	*request;	/* ptr to mtvd request */
	int4		yrows;
/*
 * Check the trivial cases.
 */
	if ((xlen == INT4_NIL) || (ylen == INT4_NIL)) {
		return(0);
	}
	if ((xlen < INT4_NIL) || (ylen < INT4_NIL)) {
		errno = EINVAL;
		return(ERROR);
	}
/*
 * Set up the request message.
 */
 	request = (struct tvreq *) nhead.nh_data;
	request->tvq_req = TVQRECT;

	nhead.nh_node	= node;
	nhead.nh_event	= EVMTVD;
	nhead.nh_type	= 0;
	nhead.nh_flags	= DINTDATA;
/*
 * Packetize the rectangle by cutting it in slices along the y-axis.
 * Send each slice to be plotted.  The assumption made here is that
 * no rectangle will have an xlen (width) larger than MAXNMSGLEN,
 * otherwise slicing along the x-axis would be needed too.
 */
	yrows = (int4) MAXNMSGLEN / xlen;
	request->tvq_x1		= x1;
	request->tvq_y1		= y1;
	request->tvq_xlen	= xlen;
	nhead.nh_msg		= colours;

	while (ylen > 0) {
		request->tvq_ylen = (yrows > ylen) ? ylen : yrows;
		nhead.nh_length = xlen * request->tvq_ylen;

		if (nsend(&nhead)) {
			return(ERROR);
		}

		request->tvq_y1 += request->tvq_ylen;
		nhead.nh_msg += nhead.nh_length;
		ylen -= request->tvq_ylen;
	}

	return(0);
}
