/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	CROS.c,v $
 * Revision 2.2  92/04/24  16:38:12  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:17:47  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:36:51  gdburns
 * Homo sapien
 * 
 *
 *	FORTRAN wrappers for CrOS library.
 *
 *	Function names and calling syntax are
 *	as described in the book by Fox et al.
 *
 *	As it stands, this file is for ITB only.
 *	Control-processor functions skipped for now,
 *	but skeletons for those wrappers are here.
 */

#include <cros.h>

void
KCREAD (retval, buf, inmask, outmask, nbytes)
	int *retval, *inmask, *outmask, *nbytes;
	char *buf;
{
	*retval = cread (buf, *inmask, *outmask, *nbytes);
}

void
KCWRIT (retval, buf, mask, nbytes)
	int *retval, *mask, *nbytes;
	char *buf;
{
	*retval = cwrite (buf, *mask, *nbytes);
}

void
KVREAD (retval, buf, inmask, outmask, size, offset, nitems)
	int *retval, *inmask, *outmask, *size, *offset, *nitems;
	char *buf;
{
	*retval = vread (buf, *inmask, *outmask, *size, *offset, *nitems);
}

void
KVWRIT (retval, buf, mask, size, offset, nitems)
	int *retval, *mask, *size, *offset, *nitems;
	char *buf;
{
	*retval = vwrite (buf, *mask, *size, *offset, *nitems);
}

void
KBROAD (retval, buf, src, mask, nbytes)
	int *retval, *src, *mask, *nbytes;
	char *buf;
{
	*retval = broadcast (buf, *src, *mask, *nbytes);
}

void
KCOMBI (retval, buf, func, size, nitems)
	int *retval, (*func)(), *size, *nitems;
	char *buf;
{
	*retval = combine (buf, func, *size, *nitems);
}

void
KCONCA (retval, outbuf, nbytes, inbuf, binsz, szarray)
	int *retval, *nbytes, *binsz, *szarray;
	char *outbuf, *inbuf;
{
	*retval = concat (outbuf, *nbytes, inbuf, *binsz, szarray);
}

void
KCSHIF (retval, inbuf, inmask, inbytes, outbuf, outmask, outbytes)
	int *retval, *inmask, *inbytes, *outmask, *outbytes;
	char *inbuf, *outbuf;
{
	*retval = cshift (inbuf, *inmask, *inbytes, outbuf, *outmask, *outbytes);
}

void
KVSHIF (retval, inbuf, inmask, insize, inoffset, initems,
		outbuf, outmask, outsize, outoffset, outitems)
	int *retval, *inmask, *insize, *inoffset, *initems,
			*outmask, *outsize, *outoffset, *outitems;
	char *inbuf, *outbuf;
{
	*retval = vshift (inbuf, *inmask, *insize, *inoffset, *initems,
			  outbuf, *outmask, *outsize, *outoffset, *outitems);
}

/*
void	KBCSTC (retval, buf, nbytes)
int	bcastcp  (char *buf, int nbytes);
*/

void
KBCSTE (retval, buf, nbytes)
	int *retval, *nbytes;
	char *buf;
{
	*retval = bcastelt (buf, *nbytes);
}

/*
void	KCOMBC ()
int	combcp  (char *buf, int size, int nitems);
*/

void
KCOMBE (retval, buf, func, size, nitems)
	int *retval, (*func)(), *size, *nitems;
	char *buf;
{
	*retval = combelt (buf, func, *size, *nitems);
}

void
KDUMPE (retval, buf, nbytes)
	int *retval, *nbytes;
	char *buf;
{
	*retval = dumpelt (buf, *nbytes);
}

/*
void	KFCREA ()
int	fcread  (int fd, unsigned int inmask,  int nbytes);

void	KFCWRI ()
int	fcwrite (int fd, unsigned int outmask, int nbytes);

void	KFDUMP ()
int	fdumpcp (int fd, int nbytes, int *bufmap);

void	KFLOAD ()
int	floadcp (int fd, int *bufmap);
*/

void
KLOADE (retval, buf, nbytes)
	int *retval, *nbytes;
	char *buf;
{
	*retval = loadelt (buf, *nbytes);
}

/*
void	KMDUMP ()
int	mdumpcp (char *buf, int nbytes, int *bufmap);

void	KMLOAD ()
int	mloadcp (char *buf, int *bufmap);

void	KCUBEL ()
int	cubeld (int doc, char *nodeprog);
*/

void
KGRDIN (retval, dim, num)
	int *retval, *dim, *num;
{
	*retval = gridinit (*dim, num);
}

void
KGRDSP (retval, proc, dim, num)
	int *retval, *proc, *dim, *num;
{
	*retval = gridsplit (*proc, *dim, num);
}

void
KGRDCO (retval, proc, coord)
	int *retval, *proc, *coord;
{
	*retval = gridcoord (*proc, coord);
}

void
KGRDPR (retval, coord)
	int *retval, *coord;
{
	*retval = gridproc (coord);
}

void
KGRDCH (retval, proc, dir, sign)
	int *retval, *proc, *dir, *sign;
{
	*retval = gridchan (*proc, *dir, *sign);
}

/* warning: we require env[] to have >>> 6 <<< fields */
void
KCPARA (retval, env)
	int *retval, *env;
{
	*retval = cparam ((struct cubenv *) env);
}

/*** the rest are functions we added: ***/

void
KCWELD (retval)
	int *retval;
{
	*retval = cweld();
}

void
KCUNWD (retval)
	int *retval;
{
	*retval = cunweld();
}

void
KCWDED (retval)
	int *retval;
{
	*retval = cwelded;
}

void
KFULMK (retval)
	int *retval;
{
	*retval = (int) fullmask();
}

void
KGRDDI (retval, num)
	int *retval, *num;
{
	*retval = griddims (num);
}
