/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	bcastelt.c,v $
 * Revision 2.2  92/04/24  16:38:16  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:17:54  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:36:53  gdburns
 * Homo sapien
 * 
 *
 *	CrOS library implementation: bcastelt().
 *
 *	Bcastelt() does not use a Trollius multicast, to allow
 *	usage while the grid is welded into virtual circuits.
 *	Instead, node 0 gets it from the host, then calls the
 *	CrOS broadcast() function.
 *
 *	Differences between this implementation and the CrOS
 *	man pages in the book by Fox et al:
 *
 *	- bcastelt() does not return an error if the incoming data
 *		is larger than the specified nbytes.
 *	- bcastelt() does not return the number of bytes that the CP sent.
 *		Instead, it returns the number of bytes received.
 *	- no check is made here for data transmission errors.
 */


#include <net.h>
#include <cros.h>

PUBLIC int
/*
bcastelt (char *buffer, int nbytes)
*/
bcastelt (buffer, nbytes)
	char *buffer;
	int nbytes;
{
	struct nmsg nhead;

	if (buffer == NULL && nbytes != 0) {
		errno = EBADBUF;
		return (-1);
	}

	if (getnodeid() == 0) {		/* directly connected to host */
		nhead.nh_event  = CFROMHOST;
		nhead.nh_type   = CBCASTCP;
		nhead.nh_flags  = 0;
		nhead.nh_msg    = buffer;
		nhead.nh_length = nbytes;
		if (nrecv (&nhead) != OK)
			return (-1);

		nbytes = (int) nhead.nh_length;
	}

	return (broadcast (buffer, 0, fullmask(), nbytes));
}
