/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	combelt.c,v $
 * Revision 2.2  92/04/24  16:38:18  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:17:57  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:36:54  gdburns
 * Homo sapien
 * 
 *
 *	CrOS library implementation: combelt().
 *
 *	Combelt() does not use a Trollius multireel, to allow
 *	usage while the grid is welded into virtual circuits.
 *	Instead, all nodes call the CrOS combine() function,
 *	then node 0 sends the result to the host.
 *
 *	Differences between this implementation and the CrOS
 *	man pages in the book by Fox et al:
 *
 *	- combelt() does not return an error if the incoming data
 *		is larger than the specified nitems.
 *	- no check is made here for data transmission errors.
 */


#include <net.h>
#include <cros.h>

PUBLIC int
/*
combelt (char *buf, int (*func)(), int size, int nitems)
*/
combelt (buffer, func, size, nitems)
	char *buffer;
	int (*func)();
	int size, nitems;
{
	struct nmsg nhead;
	struct cubenv env;

	nitems = combine (buffer, func, size, nitems);

	if (getnodeid() == 0) {		/* directly connected to host */

		if (cparam (&env) < 0)
			return (-1);

		if (sendnbr (env.cpmask, &nhead) < 0)
			return (-1);

		/* no VC to host even if grid welded: */
		nhead.nh_flags = 0;

		if (nitems < 0) {
			nhead.nh_type   = CCOMBCPE;
			nhead.nh_length = 0;
		} else {
			nhead.nh_type   = CCOMBCP;
			nhead.nh_length = size * nitems;
		}

		nhead.nh_msg = buffer;
		if (nsend (&nhead) != OK)
			return (-1);
	}

	return (nitems);
}
