/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	csend.c,v $
 * Revision 2.2  92/04/24  16:38:24  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:18:03  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:36:57  gdburns
 * Homo sapien
 * 
 *
 *	Dsend() and drecv() wrappers for the CrOS library
 *
 *	The features added to dsend():
 *	- uses psend() if HOLD mode (ITB only).
 *		Nonportable (?) assumption: psend() can handle large messages.
 *	- packetizing of large messages.
 *		Nh_data[0] passes the bytes remaining to be sent.
 */

#include <net.h>

int
csend (nhead)
	struct nmsg *nhead;
{
	int	length, oldlen, chunk, retval;
	char	*oldbuf;

#ifdef itb
	if (nhead->nh_flags == HOLD)
		return (psend(nhead));
#endif

	length = nhead->nh_length;

	if (length <= MAXNMSGLEN) {		/* make easy case go fast */
		nhead->nh_data[0] = 0;		/* for the receiver */
		return (dsend (nhead));
	}

	oldbuf = nhead->nh_msg;
	oldlen = length;
	retval = 0;

	while (length > 0) {
		chunk = (length > MAXNMSGLEN ? MAXNMSGLEN : length);
		nhead->nh_length = chunk;
		length -= chunk;
		nhead->nh_data[0] = length;	/* for the receiver */
		if (dsend (nhead) != 0) {
			retval = -1;
			break;
		}
		nhead->nh_msg += chunk;
	}

	nhead->nh_msg = oldbuf;
	nhead->nh_length = oldlen;

	return (retval);
}

int
crecv (nhead)
	struct nmsg *nhead;
{
	int	length, total, chunk, retval;
	char	*oldbuf;

#ifdef itb
	if (nhead->nh_flags == HOLD)
		return (precv(nhead));
#endif

	length = nhead->nh_length;
	if (length > MAXNMSGLEN)
		nhead->nh_length = MAXNMSGLEN;
	if (drecv (nhead) != 0)
		return (-1);
	if (nhead->nh_data[0] == 0)		/* sender is done */
		return (0);

	chunk = nhead->nh_length;
	total = chunk;
	oldbuf = nhead->nh_msg;
	retval = 0;

	while (nhead->nh_data[0] > 0) {
		length -= chunk;
		nhead->nh_msg += chunk;
		chunk = (length > MAXNMSGLEN ? MAXNMSGLEN : length);
		nhead->nh_length = chunk;
		if (drecv (nhead) != 0) {
			retval = -1;
			break;
		}
		total += nhead->nh_length;
	}

	nhead->nh_msg = oldbuf;
	nhead->nh_length = total;

	return (retval);
}
