/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	cubeld.c,v $
 * Revision 2.2  92/04/24  16:38:26  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:18:07  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:36:58  gdburns
 * Homo sapien
 * 
 *
 *	CrOS library implementation: cubeld(), cubeldl(),cubeldv().
 *
 *	Note: this file to be compiled for OTB usage only.
 *
 *	We cannot implement cubeldle() and cubeldve() because Trollius
 *	(at this time) does not support environment variables ITB.
 *
 *	To implement cubeldl() correctly we need varargs(), but since
 *	we cannnot count on having it everywhere yet, we kludge it.
 *	The kludge accepts up to 10 command-line arguments.  The total
 *	length of the strings is also limited to 256 bytes.
 *	Someday we should fix it up.
 *
 *	The CrOS cubeld*() functions do not support setting the stack
 *	size and location for the created processes.  Here we added
 *	functions that have the additional parameters stacksize, flags.
 *	Flags can take the value RTF_ONCHIP (from <kio.h>) for on-chip
 *	stack.  (An RTF_NOWAIT is automatically OR-ed in.)  These new
 *	functions have the same names as the old, with 'x' appended.
 *
 *	Differences between this implementation and the CrOS
 *	man pages in the book by Fox et al:
 *
 *	- gridinit() must be called before cubeld(), and the
 *		doc parameters must match.
 *	- cubeld*() returns once the data is sent out, although
 *		the receivers may not have received it yet.
 *	- no check is made here for data transmission errors.
 *	- returns 0 if no error.
 */


#include <net.h>
#include <kio.h>
#include <rreq.h>

#include <cros.h>

typedef char *string;


/*
 *  Launch a process on the simulated compute nodes.
 *  Only used in "stand alone" mode.
 */
PRIVATE int
launch (nnodes, nodeprog, argv)
	int	nnodes;
	char	*nodeprog;
	char	**argv;
{
	int	node, pid;

	for (node=0; node<nnodes; node++) {

		/* tell SA router to wait for our later rrattach: */
		if (rrhold() == ERROR)
			return (-1);

		/* Fork a child process: */
		if ((pid = fork()) == -1)
			return (-1);

		if (pid != 0) { 

			/* 
			 *  Parent process:
			 *  send ATTACH (n_pid, n_nodeid) message to sa_rtr.
			 */
			if (rrattach (pid, node) == ERROR)
				return (-1);

		} else { 

			/*
			 * Child process.
			 */
			errno = 0;
			execvp (nodeprog, argv);
			if (errno)
				perror(nodeprog);
			exit(errno);
	      	}
	}

	return (0);
}


#define STACKSIZE	((long) 8192)
#define MAXARGCHARS	256

PRIVATE char argst[MAXARGCHARS];

PRIVATE int
loadgo (castid, nodeprog, stacksize, flags, argv)
	long castid;
	string nodeprog;
	long stacksize;
	int flags;
	string *argv;
{
	int i, argc, len, size;
	long pid;
	char *cp;

	flags |= RTF_NOWAIT;	/* no reply from kenyad upon kexit */

	for (argc=0; argv[argc]!=NULL; argc++)
		/* just count them */ ;

	size = ((argc+1) * sizeof (char *));
	cp = argst + size;
	if (argc == 0) {
		*cp = 0;
		size += 1;
	} else for (i=0; i<argc; i++) {
		len = strlen(argv[i]) + 1;
		if ((size + len) > MAXARGCHARS) {
			errno = EDATASIZE;
			return (-1);
		}
		strcpy (cp, argv[i]);
		cp   += len;
		size += len;
	}

	if (rloadgo (ORIGIN, castid, stacksize, flags, nodeprog,
				argc, argst, size, &pid) < 0)
		return (-1);

	return (0);
}


PUBLIC int
/*
cubeldvx (int doc, char *nodeprog, long stacksize, int flags, char **argv)
*/
cubeldvx (doc, nodeprog, stacksize, flags, argv)
	int doc;	/* not used */
	long stacksize;
	int flags;
	string nodeprog;
	string *argv;
{
	struct cubenv env;

	if (cparam (&env) != 0)
		return (-1);

	if (env.cubeid == NOTNODEID)			/* standalone */
		return (launch (env.nproc, nodeprog, argv));
	else
		return (loadgo (env.cubeid, nodeprog, stacksize, flags, argv));
}


PUBLIC int
/*
cubeldlx (int doc, char *nodeprog, long stacksize, int flags, char *arg0, ...)
*/
cubeldlx (doc, nodeprog, stacksize, flags,
		arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9)
	long stacksize;
	int doc, flags;
	string nodeprog;
	string arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9;
{
	string args[11];

	args[0] = arg0;
	args[1] = arg1;
	args[2] = arg2;
	args[3] = arg3;
	args[4] = arg4;
	args[5] = arg5;
	args[6] = arg6;
	args[7] = arg7;
	args[8] = arg8;
	args[9] = arg9;
	args[10] = NULLPTR;
	return (cubeldvx (doc, nodeprog, stacksize, flags, args));
}


PUBLIC int
/*
cubeldx (int doc, char *nodeprog, long stacksize, int flags)
*/
cubeldx (doc, nodeprog, stacksize, flags)
	int doc, flags;
	string nodeprog;
	long stacksize;
{
	string args[1];

	args[0] = NULLPTR;
	return (cubeldv (doc, nodeprog, stacksize, flags, args));
}


PUBLIC int
/*
cubeldv (int doc, char *nodeprog, char **argv)
*/
cubeldv (doc, nodeprog, argv)
	int doc;
	string nodeprog;
	string *argv;
{
	return (cubeldvx (doc, nodeprog, STACKSIZE, 0, argv));
}

PUBLIC int
/*
cubeldl (int doc, char *nodeprog, char *arg0, ...)
*/
cubeldl (doc, nodeprog,
		arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9)
	int doc;
	string nodeprog;
	string arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9;
{
	string args[11];

	args[0] = arg0;
	args[1] = arg1;
	args[2] = arg2;
	args[3] = arg3;
	args[4] = arg4;
	args[5] = arg5;
	args[6] = arg6;
	args[7] = arg7;
	args[8] = arg8;
	args[9] = arg9;
	args[10] = NULLPTR;
	return (cubeldvx (doc, nodeprog, STACKSIZE, 0, args));
}


PUBLIC int
/*
cubeld (int doc, char *nodeprog)
*/
cubeld (doc, nodeprog)
	int doc;
	string nodeprog;
{
	string args[1];

	args[0] = NULLPTR;
	return (cubeldvx (doc, nodeprog, STACKSIZE, 0, args));
}
