/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	dumpelt.c,v $
 * Revision 2.2  92/04/24  16:38:30  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:18:12  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:37:00  gdburns
 * Homo sapien
 * 
 *
 *	CrOS library implementation: dumpelt(), loadelt().
 *
 *	Differences between this implementation and the CrOS
 *	man pages in the book by Fox et al:
 *
 *	- dumpelt returns once the data has been sent out,
 *		but it may not have been read by the CP yet.
 *	- no check is made on the validity of the buffer parameter.
 *	- no check is made here for data transmission errors.
 */


#include <net.h>
#include <cros.h>


PUBLIC int
/*
dumpelt (char *buf, int nbytes)
*/
dumpelt (buf, nbytes)
	char *buf;
	int nbytes;
{
	struct nmsg nhead;
	int welded = FALSE;

	if (cwelded) {
		welded = TRUE;
		if (cunweld() != OK)
			return (-1);
	}

	nhead.nh_node   = ORIGIN;
	nhead.nh_event  = (CTOHOST + (getnodeid() << 16));
	nhead.nh_type   = CDUMP;
	nhead.nh_flags  = 0;
	nhead.nh_msg    = buf;
	nhead.nh_length = nbytes;
	if (tsend (&nhead) != OK)
		return (-1);

	if (welded) {
		if (cweld() != OK)
			return (-1);
	}

	return (nbytes);
}


PUBLIC int
/*
loadelt (char *buf, int nbytes)
*/
loadelt (buf, nbytes)
	char *buf;
	int nbytes;
{
	struct nmsg nhead;
	int welded = FALSE;

	if (cwelded) {
		welded = TRUE;
		if (cunweld() != OK)
			return (-1);
	}

	nhead.nh_event  = (CFROMHOST + (getnodeid() << 16));
	nhead.nh_type   = CLOAD;
	nhead.nh_flags  = 0;
	nhead.nh_msg    = buf;
	nhead.nh_length = nbytes;
	if (nrecv (&nhead) != OK)
		return (-1);

	if (welded) {
		if (cweld() != OK)
			return (-1);
	}

	return (nbytes);
}
