/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	fdumpcp.c,v $
 * Revision 2.2  92/04/24  16:38:32  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:18:15  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:37:02  gdburns
 * Homo sapien
 * 
 *
 *	CrOS library implementation: fdumpcp, floadcp.
 *
 *	Note: this file to be compiled for OTB usage only.
 *
 *	These functions use malloc() to get a buffer for ONE processor's
 *	data, and fdumpcp() uses trecv to avoid message-buffer overflow.
 *
 *	Differences between this implementation and the CrOS
 *	man pages in the book by Fox et al:
 *
 *	- fdumpcp() does not return an error if the incoming data
 *		is larger than the specified buffer.
 *	- floadcp() returns an error if the number of nodes in bufmap[0]
 *		does not MATCH the actual number of processors, whether
 *		it is higher or lower.
 *	- no check is made here for data transmission errors.
 */


#include <net.h>
#include <cros.h>

extern char *malloc();
extern int read();
extern int write();


PUBLIC int
/*
fdumpcp (int fd, int nbytes, int *bufmap)
*/
fdumpcp (fd, nbytes, bufmap)
	int fd;
	int nbytes;
	int *bufmap;
{
	char *buffer;
	int i, n, size;
	struct cubenv env;
	struct nmsg nhead;

	buffer = malloc ((unsigned) nbytes);
	if (buffer == NULL)
		return (-1);

	if (cparam (&env) != 0)
		return (-1);
	n = env.nproc;

	for (i=0; i<n; i++) {
		nhead.nh_node   = i;
		nhead.nh_event 	= (CTOHOST + (i<<16));
		nhead.nh_type   = CDUMP;
		nhead.nh_flags  = 0;
		nhead.nh_msg    = buffer;
		nhead.nh_length = nbytes;
		if (trecv (&nhead) != 0)
			return (-1);
		size = nhead.nh_length;
		bufmap[i+1] = size;
		if (write (fd, buffer, size) != size) {
			free (buffer);
			return (-1);
		}
	}
	bufmap[0] = n;

	free (buffer);

	return (n);
}


PUBLIC int
/*
floadcp (int fd, int *bufmap)
*/
floadcp (fd, bufmap)
	int fd;
	int *bufmap;
{
	int i, n, size, maxbytes;
	char *buffer;
	struct cubenv env;
	struct nmsg nhead;

	if (cparam (&env) != 0)
		return (-1);
	n = env.nproc;
	if (bufmap[0] != n) {
		errno = ENUMPROCS;
		return (-1);
	}

	n = bufmap[0];
	maxbytes = 0;
	for (i=0; i<n; i++) {
		if (bufmap[i] > maxbytes)
			maxbytes = bufmap[i];
	}
	buffer = malloc ((unsigned) maxbytes);
	if (buffer == NULL)
		return (-1);

	for (i=0; i<n; i++) {
		size = bufmap[i+1];
		if (read (fd, buffer, size) != size) {
			free (buffer);
			return (-1);
		}
		nhead.nh_node   = i;
		nhead.nh_event 	= (CFROMHOST + (i<<16));
		nhead.nh_type   = CLOAD;
		nhead.nh_flags  = 0;
		nhead.nh_msg    = buffer;
		nhead.nh_length = size;
		if (nsend (&nhead) != 0)
			return (-1);
	}

	free (buffer);

	return (n);
}
