
/*
 *	Office of Research Computing
 *
 *	Copyright 1990 The Ohio State University
 *
 *	RBD
 *
 *	$Log:	gridsplit.c,v $
 * Revision 2.2  92/04/24  16:38:37  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:18:20  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:37:03  gdburns
 * Homo sapien
 *  
 *
 *	Function:	- split the grid of processors
 *
 *	Accepts:	- the total number of processors
 *			- the number of dimensions
 *			- ptr to the dimension array
 *
 *	Returns:	- 0 or ERROR
 */


#include <stdio.h>
#include <portable.h>
#include <net.h>				/* network constants */
#include <typical.h>				/* common constants */

/*
 * external functions
 */
extern int4		getnodeid();
extern int4		getntype();
extern int		getnodes();
extern void		*malloc();


int
gridsplit(nproc, ndim, dim)

int4		nproc,
		ndim,
		*dim;

{
	int4		jonestype, jonesmask, mynode;
	int4		*jones, tmp;
	int		i, j, index, njones;

/*
 * Initialize.
 */
	if (ndim < 1) return(ERROR);
	jonestype = (NT_JONES | NT_ITB) & ~NT_WASTE;
	jonesmask = NT_JONES | NT_ITB | NT_WASTE;
	mynode = getnodeid();
	for (i=0; i < ndim; i++) dim[i] = 1;
/*
 * Get the neighbouring nodes.
 */
	if ((njones = getntype(jonestype, jonesmask)) == 0) return(0);
	if ((jones = (int4 *)malloc(njones*sizeof(int4))) == NULL) {
		return(ERROR);
	}
	if (getnodes(jones, njones, jonestype, jonesmask)) {
		free(jones);
		return(ERROR);
	}
/*
 * Compute the initial node distances.  Zero out illegal nodeids.
 */
	for (i=0; i < njones; i++) {
		jones[i] = (jones[i] == NOTNODEID) ? 0 : abs(jones[i]-mynode);
	}
/*
 * Sort the node distances in increasing order.  An O(N^2) algorithm
 * is good enough since (a) this is an initialization call, (b) the
 * number of processor links, N, is small.
 */
	for (i=0; i < njones-1; i++) {
		tmp = jones[i];
		index = i;
		for (j=i+1; j < njones; j++) {
			if (tmp > jones[j]) {
				tmp = jones[j];
				index = j;
			}
		}
		if (index != i) {
			jones[index] = jones[i];
			jones[i] = tmp;
		}
	}
/*
 * Remove duplicate distances and adjust the array length.
 */
	for (i=1, j=0; i < njones; i++) {
		if (jones[i] != jones[j]) jones[++j] = jones[i];
	}
	njones = j+1;
/*
 * Compute the number of processors in each dimension as the ratio
 * of two consecutive node distances.
 */
	for (i=0; i < njones-1; i++) {
		jones[i] = jones[i+1] / jones[i];
	}
	jones[njones-1] = nproc / jones[njones-1];

	if (ndim >= njones) {
/*
 * The user requested enough dimensions to handle the hardware.
 * Simply copy the hardware dimensions as is.
 */
		for (i=0; i < njones; i++) dim[i] = jones[i];
	}
	else {
/*
 * The user requested less dimensions than the hardware has.
 * Compress all other dimensions to fit into the last one.
 */
		for (i=0; i < ndim; i++) dim[i] = jones[i];
		for (j=i--; j < njones; j++) dim[i] *= jones[j];
	}

	return(0);
}
