/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	mdumpcp.c,v $
 * Revision 2.2  92/04/24  16:38:38  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:18:23  gdburns
 * Ohio release
 * 
 * Revision 2.0  90/09/26  12:37:04  gdburns
 * Homo sapien
 * 
 *
 *	CrOS library implementation: mdumpcp, mloadcp.
 *
 *	Note: this file to be compiled for OTB usage only.
 *
 *	Mdumpcp uses trecv to avoid message-buffer overflow.
 *
 *	Differences between this implementation and the CrOS
 *	man pages in the book by Fox et al:
 *
 *	- mdumpcp() does not return an error if the incoming data
 *		is larger than the specified buffer.
 *	- mloadcp() returns an error if the number of nodes in bufmap[0]
 *		does not MATCH the actual number of processors, whether
 *		it is higher or lower.
 *	- no check is made here for data transmission errors.
 */


#include <net.h>
#include <cros.h>


PUBLIC int
/*
mdumpcp (char *buf, int nbytes, int *bufmap)
*/
mdumpcp (buf, nbytes, bufmap)
	char *buf;
	int nbytes;
	int *bufmap;
{
	int i, n;
	struct cubenv env;
	struct nmsg nhead;

	if (cparam (&env) != 0)
		return (-1);
	n = env.nproc;

	bufmap[0] = n;

	for (i=0; i<n; i++) {
		nhead.nh_node   = i;
		nhead.nh_event 	= (CTOHOST + (i<<16));
		nhead.nh_type  	= CDUMP;
		nhead.nh_flags 	= 0;
		nhead.nh_msg    = buf;
		nhead.nh_length = nbytes;
		if (trecv (&nhead) != 0)
			return (-1);
		bufmap[i+1] = nhead.nh_length;
		buf += nhead.nh_length;
	}

	return (n);
}


PUBLIC int
/*
mloadcp (char *buf, int *bufmap)
*/
mloadcp (buf, bufmap)
	char *buf;
	int *bufmap;
{
	int i, n;
	struct cubenv env;
	struct nmsg nhead;

	if (cparam (&env) != 0)
		return (-1);
	n = env.nproc;
	if (bufmap[0] != n) {
		errno = ENUMPROCS;
		return (-1);
	}

	for (i=0; i<n; i++) {
		nhead.nh_node   = i;
		nhead.nh_event 	= (CFROMHOST + (i<<16));
		nhead.nh_type  	= CLOAD;
		nhead.nh_flags 	= 0;
		nhead.nh_msg    = buf;
		nhead.nh_length = bufmap[i+1];
		if (nsend (&nhead) != 0)
			return (-1);
		buf += bufmap[i+1];
	}

	return (n);
}
