/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	VSD
 *
 *	$Log:	filedr.c,v $
 * Revision 2.2  92/04/24  16:42:15  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/24  14:54:07  gdburns
 * Change ASYNC to NOCBX.
 * Make zillions of cosmetic changes.
 * 
 * Revision 2.1  91/05/02  14:38:58  gdburns
 * Ohio release
 * 
 * Revision 2.0.1.1  90/12/13  17:56:58  gdburns
 * Added terror.h to get EUSAGE and fixed a few linting problems.
 * 
 * Revision 2.0  90/09/26  12:36:47  gdburns
 * Homo sapien
 *
 *	Function:	- File System Diagnostic Tool (File Doctor)
 *			  Tests the file system function calls -
 *			  Open, Close, Read, Write, Create.
 *			  The basic functions 'read' and 'write' only
 *			  are provided - 'create' is 'read/write' with
 *			  zero number of bytes, all reads and writes
 *			  are bracketed with 'open' and 'close' of
 *			  the specified file, except when a '-c'
 *			  option is specified, the file is left unclosed.
 *			  In addition, timings for reads and writes are
 *			  provided.
 *
 *	Usage: 		  filedr [-chtvN] -ro/wo/rw/wr [-f filename]
 *			  [-g flags] [-m modes] [-l length]
 *
 *	Examples:	  filedr -ct -rw -g ps -l 30
 *			  filedr -ro -f junk -m 664
 *			  filedr -vN -wr -f junk -l 1000
 */


#include <errno.h>			/* for ENOENT */
#include <fcntl.h>			/* for O_CREAT, O_APPEND, etc */
#include <terror.h>			/* for EUSAGE */
#include <typical.h>			/* for ERROR, SUCCESS */
#include <freq.h>			/* for O_NOCBX, O_SINGL, O_MULTI */
#include <stdio.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */

/*
 * external functions
 */
extern void		terror();
extern void		validopts();	/* valid options */
extern void		twoletter();	/* two-letter options */
extern void		followed();	/* options with parameters */
extern int		get_2nd_let();	/* for two-letter options */
extern int		do_args();	/* parse command line arguments */
extern int		opt_taken();	/* check if option selected */
extern int		intparam();	/* get integer opt parameter */
extern char		*getparam();	/* get string opt parameter */
extern int		kinit();	/* attach to kernel */
extern double		ttime();	/* get the current time */
extern char		*malloc();	/* allocate memory */
extern int		free();		/* free allocated memory */


/*
 * local functions
 */
static void		help();		/* print help */
static int		otod();		/* converts octal string to dec */
static int		init_flags();	/* init flags based on options */
static int 		do_open();	/* open file */
static int 		do_read();	/* read data from file */
static int		do_write();	/* write data into file */
static int		do_close();	/* close file */

/*
 * local variables
 */
static char		*_copyright =
"Copyright 1992 The Ohio State University $Revision: 2.2 $";
static char		*usage =
"filedr [-chtvN] -ro/wo/rw/wr [-f filename] [-g flags] [-m modes] [-l length]";

static int		fl_verbose;	/* verbose flag */
static int		fl_close;	/* close file */
static int		fl_time;	/* timing results */
static int		fl_pretend;	/* pretend flag */
static int		fl_stdfile;	/* stdin, stdout, stderr */
static int		fl_read;	/* file read flag */
static int		fl_write;	/* file write flag */
static int		fl_rdwr;	/* do read then write */
static int		fl_wrrd;	/* do write then read */

static int		flags;		/* open file with flags	*/
static int		modes;		/* open file with modes */
static int		numbytes;	/* num of bytes to rd/wr */
static int		handle;		/* file handle/udesc */
static char		*filename;	/* file name */
static char		*buffer;	/* read-write buffer */

/*
 * main program
 */

main(argc, argv)

int			argc;
char			*argv[];

{
/*
 * Parse command line.
 */
	validopts("chtvN");
	followed("fgml");
	twoletter("row");
	twoletter("wor");

	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {
		fprintf(stderr, "usage: %s\n", usage);
		kexit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		kexit(0);
	}
/*
 * Attach to kernel.
 */
	if (kinit(0) < 0) {
		terror("filedr (kinit)");
		kexit(0);
	}
/*
 * Initialize flags based on selected options.
 */
	if (init_flags() < 0) {
		terror("filedr(init_flags)");
		kexit(ERROR);
	}
/*
 * Perform specified file operations.
 */
	if (!fl_stdfile) {
	    if (do_open() == ERROR) {
		fprintf(stderr,"Error in opening file\n");
		kexit(ERROR);
	    }
	}
	if (numbytes > 0) {
		if (fl_read || fl_rdwr) {
			if (do_read() == ERROR) kexit(ERROR);
			if (fl_rdwr) {
/*
 * Close and open file again.
 */
			    if ((!fl_stdfile) && fl_close) {
	    			if (do_close() == ERROR) {
	fprintf(stderr,"Error in closing file\n");
					kexit(ERROR);
	    			}
			    }
			    if (!fl_stdfile && fl_close) {
				if (do_open() == ERROR) {
	fprintf(stderr,"Error in opening file\n"); 
					kexit(ERROR);
            			}
        		    }    
			    if (do_write() == ERROR) kexit(ERROR);
			}
		} 

		if (fl_write || fl_wrrd) {
			if (do_write() == ERROR) kexit(ERROR);
			if (fl_wrrd) {
/* 
 * Close and open file again.
 */
                            if ((!fl_stdfile) && fl_close) {
                                if (do_close() == ERROR) {
        fprintf(stderr,"Error in closing file\n");
                                        kexit(ERROR);
                                }
                            }
                            if (!fl_stdfile && fl_close) { 
                                if (do_open() == ERROR) {
        fprintf(stderr,"Error in opening file\n"); 
                                        kexit(ERROR);
                                }
                            }     
			    if (do_read() == ERROR) kexit(ERROR);
			}
		}
	}
	else if (fl_verbose) printf("Number of bytes read/written = 0\n");

	if ((!fl_stdfile) && fl_close) {
	    if (do_close() == ERROR) {
		fprintf(stderr,"Error in closing file\n");
		kexit(ERROR);
	    }
	}
}

/*
 * do_open()
 *
 * Function:	Open file, if it doesn't exist then create
 */

static int
do_open()

{
	double	timeval;

	if (fl_verbose) printf("Opening file : %s\n", filename);
	if (!fl_pretend) {
	    timeval = ttime();
	    if ((handle = topen(filename, flags, modes)) == ERROR) {
	    	if (errno == ENOENT) {
/*
 * File does not exist, so create it.
 */
			if (fl_verbose) printf("File not found - creating it\n");
			flags = flags | O_CREAT;
			timeval = ttime();
			handle = topen(filename, flags, modes);
	    	}
	    	else {
/*
 * Some other error.
 */
                	return(ERROR); 
            	}
	    }
	    timeval = ttime() - timeval;
	}
	else timeval = 0; 
	if (fl_time) {
		printf("Time taken for opening : %f\n",
			timeval);
	}
	return(SUCCESS);
} 


/*
 * do_close()
 *
 * Function:	Close file
 */

static int
do_close()

{
	double	timeval;

	if (fl_verbose) printf("Closing file : %s\n", filename);
	if (!fl_pretend) {
		timeval = ttime();
		(void) tclose(handle);
		timeval = ttime() - timeval;
	}
	else timeval = 0;
	if (fl_time) {
                printf("Time taken for closing : %f\n", timeval);
        }
	return(SUCCESS);
}


/*
 * do_read()
 *
 * Function:	Read data from file
 */

static int
do_read()

{
	double	timeval;
	int 	i, ret_bytes;
	
/*
 * Read numbytes amount of data.
 */
	if ((buffer = malloc ((unsigned) numbytes)) == NULL) {
		fprintf(stderr,"Error in malloc space for buffer\n");
		return(ERROR);
	}
	for (i=0; i<numbytes; i++) {
		buffer[i] = ' ';
	}

	if (!fl_pretend) {
	    timeval = ttime();
	    if ((ret_bytes = tread(handle,buffer,numbytes)) == ERROR) {
	      fprintf(stderr,"Error in reading file\n");
	      return(ERROR);
	    }
	    timeval = ttime() - timeval;
	}
	else {
	    timeval = 0;
	    ret_bytes = 0;
	}

	if (fl_verbose) printf("Number of bytes read : %d\n", ret_bytes);
        if (fl_time) {
                printf("Time taken for reading : %f\n", timeval);
        }
	free(buffer);
	return(SUCCESS);
}


/*
 * do_write()
 */

static int
do_write()

{
	double	timeval;
	int 	i, ret_bytes;

        if ((buffer = malloc ((unsigned) numbytes)) == NULL) {
                fprintf(stderr,"Error in malloc space for buffer\n");
                return(ERROR);
        }

	for (i=0; i<numbytes; i++) {
		buffer[i] = (char) (i % 256);
	}

	if (!fl_pretend) {
	    timeval = ttime();
	    if ((ret_bytes = twrite (handle,buffer,numbytes)) == ERROR) {
	      fprintf(stderr,"Error %d in writing file\n", errno);
              return(ERROR);
            }
	    timeval = ttime() - timeval;
	}
	else {
	    timeval = 0;
	    ret_bytes = 0;
	}

	if (fl_verbose) printf("Number of bytes written : %d\n", ret_bytes);

	if (fl_time) {
        	printf("Time taken for writing : %f\n", timeval);
        }
	free(buffer);
	return(SUCCESS);
}
	


/*
 * Print help on the filedr command. 
 */

static void
help()

{
  printf("\nSynopsis:\tfiledr [-chtvN] -ro/wo/rw/wr [-f filename]\n");
  printf("\t\t           [-g flags] [-m modes] [-l numbytes]\n");
  printf("\nDescription:\tTest and time file system commands\n");
  printf("\nOptions:\t-h                this help message\n");
  printf("\t\t-c                do not close specified file\n");
  printf("\t\t-t                time file operations\n");
  printf("\t\t-v                verbose mode\n");
  printf("\t\t-N                pretend mode\n");
  printf("\t\t-ro               read-only operation on file (default:ro)\n");
  printf("\t\t-wo               write-only operation on file\n");
  printf("\t\t-rw               read then write on file\n");
  printf("\t\t-wr               write then read on file\n");
  printf("\t\t-f filename       work on this file (default:testfile)\n");
  printf("\t\t-g flags          open file with flags - \n");
  printf("\t\t                    flags is a string of characters\n");
  printf("\t\t                    from the set r,w,u,p,c,t,x,e,a,s,m for\n");
  printf("\t\t                    read, write, rdwr(update), append,\n");
  printf("\t\t                    create, trunc, excl, reuse,\n");
  printf("\t\t                    async, single, multiple modes\n");
  printf("\t\t-m modes          set read/write/execute permissions on file\n");
  printf("\t\t                    (modes in octal only)\n");
  printf("\t\t-l numbytes       read/write numbytes of data from/to file\n\n");
  return;
}

/*
 * Initialize flags based on the options selected in the command line.
 * Set default values for unspecified parameters.
 * Returns SUCCESS or ERROR.
 */

static int
init_flags()

{

	extern char	*strchr();
	extern int	strncmp();

	char 	*str_flags;
	char 	*str_modes;
	int 	other_chr;
/*
 * Check for switches.
 */
	if (opt_taken('c')) 
		fl_close = FALSE;
	else
		fl_close = TRUE;

	if (opt_taken('t')) 
		fl_time = TRUE;
	else
		fl_time = FALSE;

	if (opt_taken('v')) 
		fl_verbose = TRUE;
	else
		fl_verbose = FALSE;

	if (opt_taken('N'))
		fl_pretend = TRUE;
	else
		fl_pretend = FALSE;

/*
 * Check for parameters.
 */
	fl_stdfile = FALSE;
	fl_read = FALSE;
	fl_write = FALSE;
	fl_rdwr = FALSE;
	fl_wrrd = FALSE;
	flags = 0;

	if ((other_chr = get_2nd_let('r')) == 'o') {
		fl_read = TRUE;
		flags = flags | O_RDONLY;
	}
	else if (other_chr == 'w') {
		fl_rdwr = TRUE;
		flags = flags | O_RDWR;
	}
	else if ((other_chr = get_2nd_let('w')) == 'o') {
		fl_write = TRUE;
		flags = flags | O_WRONLY;
	}
	else if (other_chr == 'r') {
		fl_wrrd = TRUE;
		flags = flags | O_RDWR;
	}
	else {
                fprintf(stderr, "usage: %s\n", usage);
                kexit(ERROR);
	}

	if (opt_taken('f')) {
		filename = getparam('f');
		if (strncmp(filename,"stdin",5) == 0) {
			handle = 0;
			fl_stdfile = TRUE;
		}
		else if (strncmp(filename,"stdout",6) == 0) {
			handle = 1;
			fl_stdfile = TRUE;
		}
		else if (strncmp(filename,"stderr",6) == 0) {
			handle = 2;
			fl_stdfile = TRUE;
		}
		else handle = -1;
	}
	else {
		filename = "testfile";
		printf("Operating on default file 'testfile'\n");
	}

	if (opt_taken('g')) {
		str_flags = getparam('g');
		if (strchr(str_flags,'r') != NULL) flags = flags | O_RDONLY;
		if (strchr(str_flags,'w') != NULL) flags = flags | O_WRONLY;
		if (strchr(str_flags,'u') != NULL) flags = flags | O_RDWR;
		if (strchr(str_flags,'p') != NULL) flags = flags | O_APPEND;
		if (strchr(str_flags,'c') != NULL) flags = flags | O_CREAT;
		if (strchr(str_flags,'t') != NULL) flags = flags | O_TRUNC;
		if (strchr(str_flags,'x') != NULL) flags = flags | O_EXCL;
		if (strchr(str_flags,'e') != NULL) flags = flags | O_REUSE;
		if (strchr(str_flags,'a') != NULL) flags = flags | O_NOCBX;
		if (strchr(str_flags,'s') != NULL) flags = flags | O_SINGL;
		if (strchr(str_flags,'m') != NULL) flags = flags | O_MULTI;
	}
	
	if (opt_taken('m')) {
		str_modes = getparam('m');
		if ((modes = otod(str_modes)) == -1) {
			fprintf(stderr, "Error : modes must be in octal\n");
			return(ERROR);
		}
	}
	else {
		modes = 0644;
	}

	if (opt_taken('l')) {
		if (intparam('l', &numbytes) != SUCCESS) {
			return(ERROR);
		}
	}
	else {
		numbytes = 0;
	}

	return(SUCCESS);
}


/*
 * otod()
 *
 * Function:	Converts a string of octal digits into decimal number.
 *		The string has to be unsigned and may begin with a zero.
 */


static int 
otod(s)
char *s;

{
    int  dec, n, mult;
    char c, *p;

    if (s[0] == '0') {
	s++;
    }
    if ((n = strlen(s)) > 1) {

	    p  = s + n - 1;

	    dec = 0;
	    mult = 1;
	    while (p+1 != s) {
		c = *p;
		if ((c >= '0') && (c <= '7')) {
		    n = c - '0';
		}
		else return(-1);
		dec = dec + (n * mult);
		mult *= 8;
		p = p-1;
	    }
	    return(dec);
    }

    return(atoi(s));
}

