/*
 *      OSU Research Computing					Trillium
 *      Copyright 1992 The Ohio State University
 *      GDB
 *
 *      $Log:	ledd.c,v $
 * Revision 2.2  92/04/24  16:48:41  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/24  15:08:14  gdburns
 * complete rewrite
 * 
 *	Function:	- control and status of Light Emitting Diodes
 *			- real or imaginary
 */

#include <events.h>
#include <ledreq.h>
#include <net.h>
#include <shared.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;

/*
 * local variables
 */
static int4		shadow;		/* shadow register for LEDs */
static int4		cc;		/* cc requests to this nodeid */

/*
 * external functions
 */
extern void		leinit();	/* initialize LED */
extern void		ledisplay();	/* toggle LEDs */
extern int4		lecount();	/* How many LEDs do we have? */

main(argc, argv)

int			argc;
char			*argv[];

{
	struct nmsg	nhead;
	struct lereq	*request;
	struct lestate	*plestate;

	cc = NOTNODEID;
/*
 * Attach to kernel.
 */
	if (kinit(0) < 0) {
		panic(errno);
	}
/*
 * Initialize LEDs.
 */
	leinit(argc, argv);

	request = (struct lereq *) nhead.nh_data;
	plestate = (struct lestate *) nhead.nh_data;

	nhead.nh_length = 0;
	nhead.nh_event = EVLEDD;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;

	for (;;) {
		nhead.nh_length = 0;
		nhead.nh_type = 0;
/*
 * Receive ledd request.
 */
		if (nrecv(&nhead) < 0) {
			panic(errno);
		}

		switch(request->leq_req) {
/*
 * LED manipulation
 */
			case LEQOP:

			switch(request->leq_op) {

				case LEDNOT:
				shadow ^= request->leq_mask;
				break;

				case LEDCLR:
				shadow &= ~request->leq_mask;
				break;

				case LEDSET:
				shadow |= request->leq_mask;
				break;

				case LEDALL:
				shadow = request->leq_mask;
				break;
			}

			ledisplay(request->leq_src_node, shadow);

			if (cc != NOTNODEID) {
				rledop(cc, LEDALL, shadow);
			}

			break;
/*
 * LED status
 */
			case LEQSTATE:
			nhead.nh_node = request->leq_src_node;
			nhead.nh_event = request->leq_src_event;
			nhead.nh_type = 0;
			nhead.nh_length = 0;
			plestate->les_nled = lecount();
			plestate->les_shadow = shadow;
			plestate->les_cc = cc;

			if (nsend(&nhead) < 0) {
				panic(errno);
			}

			nhead.nh_event = EVLEDD;
			break;

			case LEQCC:
			cc = request->leq_cc;
			break;
		}
	}
}
