/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	mtvd.c,v $
 * Revision 2.2  92/04/24  16:49:29  trillium
 * Ohio Release
 * 
 *	Function:	- graphic daemon
 *			- plots rectangles and "snakes"
 */

#include <portable.h>
#include <typical.h>
#include <t_types.h>
#include <events.h>
#include <net.h>
#include <shared.h>
#include <tvreq.h>

/*
 * external variables
 */
extern int		errno;

/*
 * external functions
 */
extern int		terror();
extern int		tv_clr();
extern int		tv_init();
extern int		tv_rect();
extern int		tv_snake();
extern int		tv_state();
extern int		tv_text();
extern void		exit();

/*
 * local functions
 */
static void		do_clr();
static void		do_rect();
static void		do_snake();
static void		do_state();
static void		do_text();
static void		recv_req();
static void		sendreply();

/*
 * local variables
 */
static struct nmsg	nhead;			/* message header */
static char		buf[MAXNMSGLEN];	/* message body */
static struct tvreq	*request;		/* ptr mtvd request */
static struct tvreply	*reply;			/* ptr mtvd reply */

static char *_copyright = "Copyright 1992 The Ohio State University";

static void		(*(do_tvreq[MAXTVREQ]))() = {
				do_clr,
				do_rect,
				do_snake,
				do_text,
				do_state
			};

/*
 * Main code starts here.
 */
int
main(argc, argv)

int			argc;
char			*argv[];

{
/*
 * Initialize the graphics screen.
 */
	if (tv_init(argc, argv)) {
		exit(errno);
	}
/*
 * Attach to the kernel.
 */
	if (kinit(0)) {
		terror("mtvd (kinit)");
		exit(errno);
	}
/*
 * Set up the request and reply message descriptors.
 */
	nhead.nh_msg = buf;
	request = (struct tvreq *) nhead.nh_data;
	reply = (struct tvreply *) nhead.nh_data;
 
	do {
		recv_req();
/*
 * Process the request.  Individual request functions handle any replies.
 */
 		if ((request->tvq_req < MAXTVREQ) &&
			(request->tvq_req >= 0)) {

			(*(do_tvreq[request->tvq_req]))();
		}

	} while (1);
}

/*
 *	do_snake
 *
 *	Function:	- plot a snake-like region of pixels
 */
static void
do_snake()
{
	(void) tv_snake(request->tvq_x1, request->tvq_len, nhead.nh_msg);
}

/*
 *	do_rect
 *
 *	Function:	- plot a rectangular region of pixels
 */
static void
do_rect()
{
	(void) tv_rect(request->tvq_x1, request->tvq_y1, 
		request->tvq_xlen, request->tvq_ylen, nhead.nh_msg);
}

/*
 *	do_clr
 *
 *	Function:	- clear the screen 
 */
static void
do_clr()
{
	(void) tv_clr();
}

/*
 *	do_state
 *
 *	Function:	- get status of the graphic frame buffer
 *			- send reply to client
 */
static void
do_state()
{
	struct tvstate	state;
	int4		client_node;
	int4		client_event;

	client_node = request->tvq_src_node;
	client_event = request->tvq_src_event;
	if (tv_state(&state)) {
		reply->tvr_ret = ERROR;
		reply->tvr_errno = errno;
	}
	else {
		reply->tvr_ret		= 0;
		reply->tvr_xlen		= state.tvs_xlen;
		reply->tvr_ylen		= state.tvs_ylen;
		reply->tvr_ncolour	= state.tvs_ncolour;
		reply->tvr_xtext	= state.tvs_xtext;
		reply->tvr_ytext	= state.tvs_ytext;
	}

	sendreply(client_node, client_event);
}

/*
 *	do_text
 *
 *	Function:	- plot the text string on the graphic frame
 */
static void
do_text()
{
	(void) tv_text(request->tvq_x1, request->tvq_y1,
		request->tvq_colour, nhead.nh_msg);
}

/*
 *	sendreply
 *
 *	Function:	- send reply back to the client
 *			- reply information is already in nh_data
 *			- abort if send error
 *
 *	Accepts:	- client nodeid
 *			- client event
 */
static void
sendreply(nodeid, event)

int4			nodeid;
int4			event;

{
	nhead.nh_node	= nodeid;
	nhead.nh_event	= event;
	nhead.nh_flags	= DINTDATA;
	nhead.nh_length	= 0;

	if (nsend(&nhead)) {
		terror("mtvd (sendreply)");
		kexit(errno);
	}
}

/*
 *	recv_req
 *
 *	Function:	- receive a request from a client
 *			- abort if receive error
 */
static void
recv_req()
{
	nhead.nh_event	= EVMTVD;
	nhead.nh_type	= 0;
	nhead.nh_flags	= DINTDATA;
	nhead.nh_length	= MAXNMSGLEN;

	if (nrecv(&nhead)) {
		terror("mtvd (recv_req)");
		kexit(errno);
	}
}
