/*
 *      OSU Research Computing					Trillium
 *      Copyright 1991 The Ohio State University
 *      GDB
 *
 *      $Log:	traced.c,v $
 * Revision 2.2  92/04/24  16:50:23  trillium
 * Ohio Release
 * 
 *	Function:	- trace daemon
 *			- manages local trace data
 *			- transport mechanism
 *			- prefix code is tr
 */

#include <events.h>
#include <net.h>
#include <terror.h>
#include <trreq.h>
#include <shared.h>

/*
 * external functions
 */
extern void		trbinit();		/* initialize trace list */
extern void		trpinit();		/* initialize client list */
extern void		trqattach();		/* attach client to traced */
extern void		trqdetach();		/* detach client from traced */
extern void		trqstore();		/* store trace data */
extern void		trqget();		/* unload trace data */
extern void		trqstate();		/* get traced status */
extern void		trqsweep();		/* sweep out current traces */
extern void		trqsig();		/* signal clients */
extern char		*memcpy();
/*
 * local functions
 */
static void		recv_req();		/* receive traced request */
/*
 * global functions
 */
void			send_reply();		/* send traced reply */
/*
 * global variables
 */
char			buf[MAXNMSGLEN];	/* I/O buffer */
/*
 * external variables
 */
extern int		errno;			/* global error variable */

main()

{
	struct trreq	request;		/* traced request */
/*
 * Attach to kernel.
 */
	if (kinit(0) < 0) {
		panic(errno);
	}
/*
 * Attach to kenyad.
 */
	if ((rpattach(LOCAL) < 0) && (errno != ESERVICE)) {
		panic(errno);
	}
/*
 * Initialize trace and client modules.
 */
	trpinit();
	trbinit();
/*
 * main request processing loop
 */
	for (;;) {
		recv_req(&request);

		if (request.trq_req == TRQATTACH) {
			trqattach(&request);
		} else if (request.trq_req == TRQDETACH) {
			trqdetach(&request);
		} else if (request.trq_req == TRQSTORE) {
			trqstore(&request);
		} else if (request.trq_req == TRQGET) {
			trqget(&request);
		} else if (request.trq_req == TRQFORGET) {
			trqget(&request);
		} else if (request.trq_req == TRQSTATE) {
			trqstate(&request);
		} else if (request.trq_req == TRQSWEEP) {
			trqsweep();
		} else if (request.trq_req == TRQSIG) {
			trqsig();
		}
	}
}

/*
 *	recv_req
 *
 *	Function:	- receives an traced request from a client
 *
 *	Accepts:	- ptr to traced request structure
 */

static void
recv_req(request)

struct trreq		*request;

{
	struct nmsg		nhead;		/* request descriptor */

	nhead.nh_event = EVTRACED;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = MAXNMSGLEN;
	nhead.nh_msg = buf;

	if (nrecv(&nhead) < 0) {
		panic(errno);
	}

	(void) memcpy((char *) request, (char *) nhead.nh_data,
		sizeof(struct trreq));
}

/*
 *	send_reply
 *
 *	Function:	- sends a traced reply to a client
 *
 *	Accepts:	- ptr to traced request
 *			- ptr to traced reply
 *			- length of reply message
 */

void
send_reply(request, reply, len)

struct trreq		*request;
struct trreply		*reply;
int4			len;

{
	struct nmsg		nhead;		/* reply descriptor */

	(void) memcpy((char *) nhead.nh_data, (char *) reply,
		sizeof(struct trreply));
	
	nhead.nh_node = request->trq_src_node;
	nhead.nh_event = request->trq_src_event;
	nhead.nh_type = 0;
	nhead.nh_flags = DINTMSG;
	nhead.nh_length = len;
	nhead.nh_msg = buf;

	if (nsend(&nhead) < 0) {
		panic(errno);
	}
}
