/*
 *      OSU Research Computing					Trillium
 *      Copyright 1991 The Ohio State University
 *      GDB
 *
 *      $Log:	trbuf.c,v $
 * Revision 2.2  92/04/24  16:50:30  trillium
 * Ohio Release
 * 
 *	Function:	- functions that manipulate the trace buffer
 *			  data structure
 */

#include <all_list.h>
#include <net.h>
#include <trreq.h>
#include <typical.h>
#include <shared.h>

/*
 * external variables
 */
extern int		errno;			/* global error variable */
extern char		buf[];			/* traced message I/O */

/*
 * local structures
 */
struct trbuf {
	char		*trb_buf;
	int4		trb_pid;
	int4		trb_length;
};
#define TRBNULL ((struct trbuf *) 0)

/*
 * local variables
 */
static LIST		*trb_list;

/*
 * external functions
 */
extern char		*malloc();
extern char		*memcpy();
extern void		send_reply();
extern void		trpzero();

/*
 * local functions
 */
static int4		trbcompare();

/*
 *	trbinit
 *
 *	Function:	- initializes trace buffer list
 */
void
trbinit()

{
	trb_list = al_init(sizeof(struct trbuf), trbcompare);

	if (trb_list == (LIST *) 0) {
		panic(errno);
	}
}

/*
 *	trqstore
 *
 *	Function:	- stores trace data from a client
 *			- no reply message
 *
 *	Accepts:	- client pid
 */
void
trqstore(request)

struct trreq		*request;

{
	struct trbuf	new;

	if (request->trq_length <= 0) {
		return;
	}
/*
 * We are not going to store the traces if the client is not attached.
 */
	if (trpinc(request->trq_pid, request->trq_length) < 0) {
		return;
	}
/*
 * Allocate space for the traces.
 */
	new.trb_buf = malloc((unsigned) request->trq_length);

	if (new.trb_buf == (char *) 0) {
		return;
	}

	new.trb_pid = request->trq_pid;
	new.trb_length = request->trq_length;
	(void) memcpy(new.trb_buf, buf, request->trq_length);

	if (al_insert(trb_list, &new) == 0) {
		panic(errno);
	}
}

/*
 *	trqget
 *
 *	Function:	- unloads trace data
 *			- if TRQGET, then maintain trace data
 *			- if TRQFORGET, then free trace data
 *			- sends all trace buffers to client
 *			- ends by sending a zero length trace buffer
 *
 *	Accepts:	- ptr to traced request
 */
void
trqget(request)

struct trreq		*request;

{
	struct trbuf	*curr;
	struct trbuf	*prev;
	struct trreply	reply;

	reply.trr_errno = 0;
	curr = (struct trbuf *) al_top(trb_list);

	while (curr != TRBNULL) {
/*
 * Ship traces to client.
 */
		(void) memcpy(buf, curr->trb_buf, curr->trb_length);
		send_reply(request, &reply, curr->trb_length);
		prev = curr;
		curr = (struct trbuf *) al_next(trb_list, curr);
/*
 * Free, if requested, the trace data.
 */
		if (request->trq_req == TRQFORGET) {
			(void) free(prev->trb_buf);
			(void) al_delete(trb_list, prev);
		}
	}

	send_reply(request, &reply, 0);

	if (request->trq_req == TRQFORGET) {
		trpzero();
	}
}

/*
 *	trqsweep
 *
 *	Function:	- sweeps out trace data
 *			- no reply message
 */
void
trqsweep()

{
	struct trbuf	*curr;
	struct trbuf	*prev;

	curr = (struct trbuf *) al_top(trb_list);

	while (curr != TRBNULL) {
		prev = curr;
		curr = (struct trbuf *) al_next(trb_list, curr);
		(void) free(prev->trb_buf);
		(void) al_delete(trb_list, prev);
	}

	trpzero();
}

/*
 *	trbcompare
 *
 *	Function:	- compares two trb list entries by pid
 *
 *	Accepts:	- ptr to list entry A
 *			- ptr to list entry B
 *
 *	Returns:	- -1 if A < B
 *			- 0 if A == B
 *			- 1 if A > B
 */
static int4
trbcompare(a, b)

struct trbuf		*a;
struct trbuf		*b;

{
	if (a->trb_pid < b->trb_pid) {
		return(-1);
	} else if (a->trb_pid == b->trb_pid) {
		return(0);
	} else {
		return(1);
	}
}
