/*
 *      OSU Research Computing                                  Trillium
 *      Copyright 1992 The Ohio State University
 *      JRV
 *
 *      $Log:	trade.c,v $
 * Revision 2.2  92/04/24  16:51:43  trillium
 * Ohio Release
 * 
 *
 *      Function:       - converts trace output to PagaGraph input
 */

#include <stdio.h>
#include <string.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>
#include <ktrace.h>
#include <trade.h>


/*
 * global variables
 */
char			*me;			/* tool name */
int4			num_nodes=0;		/* count of traced nodes  */
int4			idle_time[MAX_NODES];	/* cumm. for traced nodes */
int4			tid_map[MAX_NODES];	/* tid_map[node] = tid    */


/*
 * static variables
 */
char			*usage =
		 	     "usage: %s [-h] [-V] [<infile>] [-o <outfile>]\n";
char			*bad_record = 
			      "%s: unknown trace record: %d, aborting.\n";
char			*tmpFileName      = "/tmp/trtmpout.XXXXXX";
char			*cleanedFileName  = "/tmp/trclean.XXXXXX";

/*
 * external variables
 */
extern int		errno;


/* 
 * external functions
 */
extern void		fix_time();
extern uint4		get_min_start();
extern void		generate_PG_RECV();
extern void		generate_PG_RECV_WAKING();
extern void		generate_PG_RECV_BLOCKING();
extern void		generate_PG_SEND();
extern void		generate_PG_TRACE_EXIT();
extern void		generate_PG_CLOSE();
extern void		generate_PG_TRACE_START();
extern void		generate_PG_COMPSTATS();


/*
 * local functions
 */
void			help_trade();
void			sort_and_output();
void			generate_recs();
void			clean_and_exit();

/*
 * Main program.
 */
int
main(argc, argv)

int                     argc;
char                    *argv[];

{
	int4		w;            		/* tmp word for reading in   */
	char		outputFileName[256];    /* name of output file       */
	FILE		*fpInput;		/* input file                */
	FILE		*fpCleaned; 		/* redund start/stop removed */
	FILE		*fpTmp;			/* tmp output file           */
	uint4		time_offset;		/* timestamp of first start  */
	int4		i;			/* my favorite index         */
	int4		fl_V = 0; 		/* 'raw output' flag         */
	int4		fl_o = 0;		/* 'to output file' flag     */

/*
 * Assign the tool name.
 */
        if ((me = strrchr(argv[0], STRDIR)) == (char *) NULL) {
                me = argv[0];
        }
        else {
                me++;
        }
/*
 * Parse the command line.
 */
        validopts("hV");
        followed("o");

        if ((do_args(&argc, argv) == ERROR) ||
                (errno = ((argc == 1)||(argc == 2)) ? 0 : EUSAGE)) {
                fprintf(stderr, usage, me);
                exit(1);
        }

        if (opt_taken('h')) {
                help_trade();
                exit(0);
        }

        if (opt_taken('V')) { 
		fl_V = 1; 
	}

	if (opt_taken('o')) { 
		fl_o = 1;
		strcpy(outputFileName, getparam('o'));
	}

	if (argc == 2) {
                if ( (fpInput = fopen(argv[1], "rb")) == NULL) {
			perror(me);
			exit(errno);
		}
	}
	else /* it must be 1 */ {
		fpInput = stdin;
	}

	mktemp(cleanedFileName);
	if ((fpCleaned = fopen(cleanedFileName, "wb")) == 0) {
		perror(me);
		exit(errno);
	}

/*
 * Pass through the input once to find minimum start time and to clean out
 * redundant start/stop events; also build map of nodes/tids.
 */

	for (i=0; i<MAX_NODES; i++) {
		tid_map[i] = i;
	}

	time_offset = get_min_start(tid_map,
				    fpInput,
				    fpCleaned, 
				    outputFileName, 
				    fl_V, 
				    fl_o);
	fclose(fpCleaned);

	if ((fpCleaned = fopen(cleanedFileName,"rb")) == NULL) {
		perror(me);
		clean_and_exit(errno);
	}

/*
 * If the user selected 'raw output', just exit now.
 */
	if (fl_V) {
		clean_and_exit(0);
	}

	if ((fpTmp = fopen(mktemp(tmpFileName), "w"))  == NULL) {
		perror(me);
		clean_and_exit(errno);
	}

/*
 * Now, make a second pass to generate the trace records
 */
	generate_recs(tid_map, fpCleaned, fpTmp, time_offset);

/*
 * Output results and clean up.
 */
	fclose(fpTmp);
	fclose(fpCleaned);

	sort_and_output(tmpFileName, outputFileName, fl_o);

	clean_and_exit(0);

} /* end main */


/*
 *	generate_recs
 *
 *	Function:	- generates PG trace records, unsorted
 *			- this is the second pass throught the input
 *
 *	Accepts:	- input file, binary ttrace file
 *			- tmp output file
 *			- earliest start-record's timestamp
 *
 */
void
generate_recs(tid_map, fpInput, fpTmp, time_offset)

int4			*tid_map;
FILE			*fpInput;
FILE			*fpTmp;
uint4			time_offset;

{
	struct trtr	tr_head;
	struct trstart	start;
	struct trsend	send;
	struct trrecv	recv;
	uint4		count=1;

        while (count != 0) {
		count = fread(&tr_head, sizeof(struct trtr), 1, fpInput);
		
		if (count != 1) break;

 		fix_time(&tr_head, time_offset);

		switch (tr_head.trt_type) {

		case TRSTART:
			fread(&start.tr1_node,
			      sizeof(struct trstart) - sizeof(struct trtr), 
			      1, 
			      fpInput);

			idle_time[tr_head.trt_tid] = 0;
			generate_PG_TRACE_START(tr_head, start, fpTmp);
			break;
			
		case TREND:		
			generate_PG_COMPSTATS(tr_head, fpTmp);
			generate_PG_TRACE_EXIT(tr_head, fpTmp);
			break;
			
		case TRSEND:
			fread(&send.tr5_tonode, 
			      sizeof(struct trsend) - sizeof(struct trtr), 
			      1, 
			      fpInput);

			generate_PG_SEND(tr_head, send, fpTmp, tid_map);
			generate_PG_COMPSTATS(tr_head, fpTmp);
			idle_time[tr_head.trt_tid] += send.tr5_total;
			generate_PG_COMPSTATS(tr_head, fpTmp);
			break;
			
		case TRRECV:
			fread(&recv.tr6_frtid,  
			      sizeof(struct trrecv) - sizeof(struct trtr), 
			      1,
			      fpInput);

			if (recv.tr6_total < 2) {
				generate_PG_RECV(tr_head, recv, fpTmp);
				generate_PG_COMPSTATS(tr_head, fpTmp);
				generate_PG_COMPSTATS(tr_head, fpTmp);
			}
			else {
				generate_PG_RECV_BLOCKING(tr_head, recv,fpTmp);
				generate_PG_COMPSTATS(tr_head, fpTmp);
/*
 * Stop the next call from colliding with this one by spacing them by one tick.
 */
				recv.tr6_total -= 1;
				idle_time[tr_head.trt_tid] += recv.tr6_total;
				tr_head.trt_stamp += recv.tr6_total;
				generate_PG_RECV_WAKING(tr_head, recv, fpTmp);
				generate_PG_COMPSTATS(tr_head, fpTmp);
			}
			break;

		default:
			fprintf(stderr, bad_record, me, tr_head.trt_type);
			clean_and_exit(0);

		} /* end-switch */
	} /* end-while */
} /* end generate_recs */


/*
 *      help_trade
 *
 *      Function:       - print the help menu
 *
 */
void
help_trade()

{
        printf("\nSynopsis:\t%s [options]\n", me);
        printf("\nDescription:\tConvert ttrace output to ParaGraph input.\n");
        printf("\nOptions:\t-h                      Print this message.\n");
        printf("\t\t-V                      Print ASCII version of ttrace file.\n");
        printf("\t\t<infile>		Get input from <infile>.\n");
        printf("\t\t-o <outfile>            Send output to <outfile>.\n\n");
}


/*
 *	sort_and_output
 *
 *	Function:	- sort the temporary output file with fork/exec/sort
 *			- output to stdout or to the file specified by '-o'
 *
 *	Accepts:	- name of file with unsorted PG records
 *			- name of file to send output
 *			- flag which says if '-o' was specified
 */
void
sort_and_output(tmpFileName, outputFileName, fl_o)

char			*tmpFileName;
char			*outputFileName;
int4			fl_o;

{
	int4		i;
	int4		kid_pid;
	char		*kid_arg[9];
	char		outarg[80];

/*
 *  this is (nearly) how sort was called from the shellscript supplied with PG:
 *       sort +1n -2 +2n -3 +0n -1 -oOutputfile $1
 */
	kid_arg[0] = "sort";
	kid_arg[1] = "+1n";
	kid_arg[2] = "-2";
	kid_arg[3] = "+2n";
	kid_arg[4] = "+0n";
	kid_arg[5] = "-1";

	i = 6;

	if (fl_o) {
		sprintf(outarg, "-o%s", outputFileName);
		kid_arg[i++] = outarg;
	}

	kid_arg[i++] = tmpFileName;
	kid_arg[i] = (char *) 0;

	if ((kid_pid = fork()) == 0) {		 /* I'm the dude */
		if (execvp("sort", kid_arg) == -1) {
			perror(errno);
			exit(errno);
		}
	}
	else if (kid_pid == -1) {
		perror(errno);
		clean_and_exit(errno);
	}	

	while (wait(0) == -1);			/* wait, 'cause I'm mom */
}


/*
 *	clean_and_exit
 *
 *	Function:	- removes tmp files and exits
 *			- never returns to caller
 *
 *	Accepts:	- the errno, 0 if none
 *
 */
void
clean_and_exit(err)

int4		err;

{
	unlink(tmpFileName);
	unlink(cleanedFileName);
	exit(err);
}






