/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	JRV
 *
 *	$Log:	trio.c,v $
 * Revision 2.2  92/04/24  16:52:01  trillium
 * Ohio Release
 * 
 *	Function:	- various useful routines for trade
 *			- mostly to handle i/o
 */

#include <stdio.h>
#include <errno.h>
#include <portable.h>
#include <typical.h>
#include <all_list.h>
#include <ktrace.h>
#include <trade.h>

/*
 * local functions
 */
uint4		get_min_start();
void		check_issue_start();
int4		same_tid();
void		generate_PG_RECV();
void		generate_PG_RECV_WAKING();
void		generate_PG_RECV_BLOCKING();
void		generate_PG_SEND();
void		generate_PG_TRACE_EXIT();
void		generate_PG_CLOSE();
void		generate_PG_TRACE_START();
void		generate_PG_COMPSTATS();
int4		sec_from_stamp();
int4		usec_from_stamp();
void		fix_time();

/*
 * global data
 */
char 		*tid_reset =
		 "%s: warning, TID %d has been reset to TID %d, continuing.\n";

/*
 *  external variables
 */
extern char		*me;			/* tool name */
extern char		*bad_record;		/* msg for unexpected record */
extern int4		idle_time[MAX_NODES];	/* cumm. idle time for nodes */

/*
 *  external functions
 */
void			clean_and_exit();


/*
 *	get_min_start
 *
 *	Function:	- find earliest start
 *			- clean out unneeded start/stop pairs
 *			- build map of tids to nodes for mapping sends
 *
 *	Accepts:	- array of nodes/tids
 *			- the buffered input file
 *			- the 'cleaned' version of above
 *			- the name of the final output
 *			- 'raw output' flag
 *			- 'redirect ouput' flag
 *
 *	Returns:	- the lowest start's timestamp
 *
 */
uint4
get_min_start(tid_map, fpInput, fpCleaned, outputFileName, fl_V, fl_o)

int4			*tid_map;
FILE			*fpInput;
FILE			*fpCleaned;
char			*outputFileName;
int4			fl_V;
int4			fl_o;

{
	struct trtr	tr_head;
	struct trstart	start;
	struct trstart	*item;
	struct trsend	send;
	struct trrecv	recv;
	int4		count;
	uint4		time_offset;
	FILE 		*fpOutput;
	LIST		*startlist;


	if (fl_V) {
		if (fl_o == 0) {
			fpOutput = stdout;
		}
		else if ((fpOutput = fopen(outputFileName, "w")) == NULL) {
			perror(errno);
			clean_and_exit(errno);
		}
	}

	startlist = al_init(sizeof(struct trstart), same_tid);

	time_offset = UINT4_MAX;
	count = 1;

        while (count != 0) {
		count = fread(&tr_head, sizeof(struct trtr), 1, fpInput);

		if (count != 1) break;

		switch (tr_head.trt_type) {

		case TRSTART:
			fread(&start.tr1_node,
			      sizeof(struct trstart) - sizeof(struct trtr), 
			      1, 
			      fpInput);

			if (fl_V) {
				fprintf(fpOutput, "%d %d %d %d %d\n",
					tr_head.trt_type,
					tr_head.trt_stamp,
					tr_head.trt_tid,
					start.tr1_node,
					start.tr1_pid);
			}
			else { 
				if ((tid_map[start.tr1_node]!=start.tr1_node)&&
				    (tid_map[start.tr1_node]!=
				      			   tr_head.trt_tid) ) {

					fprintf(stderr, 
						tid_reset, 
						me, 
						tid_map[start.tr1_node],
						tr_head.trt_tid);
				}

				tid_map[start.tr1_node] = tr_head.trt_tid;
				memcpy(&start, &tr_head, sizeof(struct trtr));
				al_insert(startlist, &start);
			}

			break;
			
		case TREND:		

			if (fl_V) {
				fprintf(fpOutput, "%d %d %d\n",
					tr_head.trt_type,
					tr_head.trt_stamp,
					tr_head.trt_tid);
			}
			else {
/* 
 * Find the start that matches this end; if it's still
 * in the list, discard both.  Otherwise, the start has
 * already been generated, so just put out the 'end'.
 */				
				item = (struct trstart *) al_find(startlist,
								  &tr_head);

				if (item == (struct trstart *) NULL) {
					fwrite(&tr_head, 
					       sizeof(struct trtr),
					       1,
					       fpCleaned);
				}
				else {
					al_delete(startlist, item);
				}
			}

			break;
			
		case TRSEND:

			fread(&send.tr5_tonode, 
			      sizeof(struct trsend) - sizeof(struct trtr), 
			      1, 
			      fpInput);

			if (fl_V) {
			      fprintf(fpOutput,"%d %d %d %d %d %d %d %d %d\n",
					tr_head.trt_type,
					tr_head.trt_stamp,
					tr_head.trt_tid,
					send.tr5_tonode,
					send.tr5_event,
					send.tr5_length,
					send.tr5_total,
					send.tr5_blocked,
					send.tr5_transmit);
			}
			else {
				check_issue_start(fpCleaned, 
						  startlist,
						  &tr_head,
						  &time_offset);
				
				fwrite(&tr_head, 
				       sizeof(struct trtr),
				       1,
				       fpCleaned);

				fwrite(&send.tr5_tonode,
				       sizeof(struct trsend) - 
				       		sizeof(struct trtr),
				       1,
				       fpCleaned);
			}

			break;
			
		case TRRECV:

			fread(&recv.tr6_frtid,  
			      sizeof(struct trrecv) - sizeof(struct trtr), 
			      1,
			      fpInput);

			if (fl_V) {
			      fprintf(fpOutput,"%d %d %d %d %d %d %d %d %d\n",
					tr_head.trt_type,
					tr_head.trt_stamp,
					tr_head.trt_tid,
					recv.tr6_frtid,
					recv.tr6_event,
					recv.tr6_length,
					recv.tr6_total,
					recv.tr6_blocked,
					recv.tr6_transmit);
			}
			else {
				check_issue_start(fpCleaned, 
						  startlist,
						  &tr_head,
						  &time_offset);
				
				fwrite(&tr_head, 
				       sizeof(struct trtr),
				       1,
				       fpCleaned);

				fwrite(&recv.tr6_frtid,
				       sizeof(struct trrecv) - 
				       		sizeof(struct trtr),
				       1,
				       fpCleaned);
			}
				      
			break;

		default:
			fprintf(stderr, bad_record , me, tr_head.trt_type);
			clean_and_exit(0);
		}
	}

	return(time_offset);
}


/*
 *	check_issue_start
 *
 *	Function:	- updates the min start time
 *			- generates start if needed, discards it otherwise
 *
 *	Accepts:	- FILE * to version of input w/unneeded starts stripped
 *			- list of pending starts
 *			- the start in question
 *			- current min start time
 *
 */
void
check_issue_start(fpCleaned, startlist, tr_head, time_offset)

FILE			*fpCleaned;
LIST			*startlist;
struct trtr		*tr_head;
uint4			*time_offset;

{
	struct trstart	*item;

	item = (struct trstart *) al_find(startlist, tr_head);

	if (item == (struct trstart *)NULL) {
		return;
	}
	else {
		if ((*item).tr1_head.trt_stamp < *time_offset) {
			*time_offset = 	(*item).tr1_head.trt_stamp;
		}

		fwrite(item,
		       sizeof(struct trstart),
		       1,
		       fpCleaned);

		al_delete(startlist, item);
	}
}


/*
 *	same_tid(start, listitem)
 *
 *	Function:	- comparison function for all_list
 *
 *	Accepts:	- a traced start event record
 *			- the list of starts
 *
 *	Returns:	- 0 if the tids are the same,  -1 otherwise
 *
 */
int4
same_tid(start, listitem)

struct trstart 		*start;
struct trstart 		*listitem;

{
	if ((*start).tr1_head.trt_tid == (*listitem).tr1_head.trt_tid) {
		return(0);
	}
	else return(-1);
}


/*
 *      
 *
 *      Function:       - generate_PG_RECV
 *
 *      Accepts:        - trace event header
 *                      - body for trrecv event
 *			- temp output file
 */
void
generate_PG_RECV(tr_head, recv, fpTmp)

struct trtr		tr_head;
struct trrecv		recv;
FILE			*fpTmp;			/* tmp output file,
							 sorted later  */
{
	fprintf(fpTmp,"%d %d %d %d %d %d %d\n",
	       PG_RECV,
	       sec_from_stamp(tr_head.trt_stamp),
	       usec_from_stamp(tr_head.trt_stamp),
	       tr_head.trt_tid,
	       recv.tr6_frtid,
	       recv.tr6_event,
	       recv.tr6_length);
}

/*
 *      
 *
 *      Function:       - generate_PG_RECV_WAKING
 *
 *      Accepts:        - trace event header
 *                      - body for trrecv event
 *			- temp output file
 */
void
generate_PG_RECV_WAKING(tr_head, recv, fpTmp)

struct trtr		tr_head;
struct trrecv		recv;
FILE			*fpTmp;			/* tmp output file,
							 sorted later     */
{
	fprintf(fpTmp,"%d %d %d %d %d %d %d\n",
	       PG_RECV_WAKING,
	       sec_from_stamp(tr_head.trt_stamp),
	       usec_from_stamp(tr_head.trt_stamp),
	       tr_head.trt_tid,
	       recv.tr6_frtid,
	       recv.tr6_event,
	       recv.tr6_length);
}

/*
 *      
 *
 *      Function:       - generate_PG_RECV_BLOCKING
 *
 *      Accepts:        - trace event header
 *                      - body for trrecv event
 *			- temp output file
 */
void
generate_PG_RECV_BLOCKING(tr_head, recv, fpTmp)

struct trtr		tr_head;
struct trrecv		recv;
FILE			*fpTmp;			/* tmp output file, 
							sorted later     */
{
	fprintf(fpTmp,"%d %d %d %d %d\n",
	       PG_RECV_BLOCKING,
	       sec_from_stamp(tr_head.trt_stamp),
	       usec_from_stamp(tr_head.trt_stamp),
	       tr_head.trt_tid,
	       recv.tr6_event);
}


/*
 *      
 *
 *      Function:       - generate_PG_SEND
 *
 *      Accepts:        - trace event header
 *                      - body for trsend event
 *			- temp output file
 */
void
generate_PG_SEND(tr_head, send, fpTmp, tid_map)

struct trtr		tr_head;
struct trsend		send;
FILE			*fpTmp;			/* tmp output file */
int4			*tid_map;

{
	fprintf(fpTmp,"%d %d %d %d %d %d %d\n",
	       PG_SEND,
	       sec_from_stamp(tr_head.trt_stamp),
	       usec_from_stamp(tr_head.trt_stamp),
	       tr_head.trt_tid,
	       tid_map[send.tr5_tonode],
	       send.tr5_event,
	       send.tr5_length);
}


/*
 *      
 *
 *      Function:       - generate_PG_TRACE_EXIT
 *
 *      Accepts:        - trace event header
 *			- temp output file
 */
void
generate_PG_TRACE_EXIT(tr_head, fpTmp)

struct trtr		tr_head;
FILE			*fpTmp;			/* tmp output file, 
							sorted later     */
{
	fprintf(fpTmp,"%d %d %d %d %d\n",
	       PG_TRACE_EXIT,
	       sec_from_stamp(tr_head.trt_stamp),
	       usec_from_stamp(tr_head.trt_stamp),
	       tr_head.trt_tid,
	       UNUSED);
}

/*
 *      
 *
 *      Function:       - generate_PG_CLOSE
 *
 *      Accepts:        - trace event header
 *			- temp output file
 */
void
generate_PG_CLOSE(tr_head, fpTmp)

struct trtr		tr_head;
FILE			*fpTmp;			/* tmp output file, 
							sorted later     */

{
	fprintf(fpTmp,"%d %d %d %d\n",
	       PG_CLOSE,
	       sec_from_stamp(tr_head.trt_stamp),
	       usec_from_stamp(tr_head.trt_stamp),
	       tr_head.trt_tid);
}


/*
 *      
 *
 *      Function:       - generate_PG_TRACE_START
 *
 *      Accepts:        - trace event header
 *                      - event body for trstart record
 *			- temp output file
 */
void
generate_PG_TRACE_START(tr_head, start, fpTmp)

struct trtr		tr_head;
struct trstart		start;
FILE			*fpTmp;			/* tmp output file,
							 sorted later     */

{
	fprintf(fpTmp,"%d %d %d %d %d %d %d\n",
	       PG_TRACE_START,
	       sec_from_stamp(tr_head.trt_stamp),
	       usec_from_stamp(tr_head.trt_stamp),
	       tr_head.trt_tid,
	       UNUSED,
	       UNUSED,
	       UNUSED);
}

/*
 *      
 *
 *      Function:       - generate_PG_COMPSTATS
 *
 *      Accepts:        - trace event header
 *			- temp output file
 */
void
generate_PG_COMPSTATS(tr_head, fpTmp)

struct trtr		tr_head;
FILE			*fpTmp;			/* tmp output file,
							 sorted later     */
{
	fprintf(fpTmp,"%d %d %d %d %d %d\n",
	       PG_COMPSTATS,
	       sec_from_stamp(tr_head.trt_stamp),
	       usec_from_stamp(tr_head.trt_stamp),
	       tr_head.trt_tid,
	       sec_from_stamp(idle_time[tr_head.trt_tid]),
	       usec_from_stamp(idle_time[tr_head.trt_tid]));
}


/*
 *      
 *
 *      Function:       - sec_from_stamp
 *
 *      Accepts:        - a time stamp in units of clock ticks
 *
 *      Returns:	- number of whole seconds in time stamp
 */
int4
sec_from_stamp(stamp)

int4			stamp;

{
	return( (int4)(stamp*SEC_PER_TICK) );
}

/*
 *      
 *
 *      Function:       - return microseconds in remainder of stamp
 *
 *      Accepts:        - time stamp
 *
 *	Returns:	- remainder portion of fractions of a second
 *			  from the time stamp.
 */
int4
usec_from_stamp(stamp)

int4			stamp;

{
	return((int4)( ((stamp*SEC_PER_TICK) - 
		        (int4)(stamp*SEC_PER_TICK)) * 1000000 ));
}

/*
 *      
 *
 *      Function:       - adjust the timestamp on an event by the given offset
 *
 *      Accepts:        - a trace event header and the offset in clock ticks
 *
 */
void
fix_time(head, time_offset)

struct trtr		*head;
int4			time_offset;

{
	(* head).trt_stamp = (* head).trt_stamp - time_offset;
}

