/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	AKP
 *
 *	$Log:	connect.c,v $
 * Revision 1.1  88/06/05  01:23:37  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:50:32  trillium
 * Initial revision
 * 
 *
 *	Connect
 *
 *	Connect I/O from two pathnames together.  Primarily meant for
 *	RS-232 lines...
 *
 */

static char *_copyright = "Copyright 1987 Cornell University $Header: connect.c,v 1.1 88/06/05 01:23:37 trillium Exp $";

#include <stdio.h>
#include <sgtty.h>
#include <signal.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>

#define BUFSIZE		1024

extern int open();
extern int select();
extern int read();
extern int write();
extern int isatty();
extern int errno;
extern char *sys_errlist[];

int escape();
int quit();
int stop();

struct channel {
	int		 ch_inp;		/* input descriptor */
	int		 ch_out;		/* output descriptor */
	int		 ch_esc;		/* escape character */
	int		 ch_flo;		/* hardware flow control */
	int		 ch_sgl;		/* single byte access */
	struct sgttyb	 ch_tty;		/* current tty settings */
	struct sgttyb	 ch_sav;		/* original tty settings */
	char		*ch_nam;		/* file name */
};

static int baudrates[] = {
	   0,    50,    75,   110,
	 134,   150,   200,   300,
	 600,  1200,  1800,  2400,
	4800,  9600, 19200, 38400,
	-1
};

static char   iobuffer[BUFSIZE];
static char  *USAGE = "[-e] [-lf] [-le c] [-lb baud] [-rf] [-re c] [-rb baud] [local] remote";
static char **av;

static int dflag;
static int eflag;
static int Lb, Rb;
static int sending = 0;
static int savedes;

static struct channel local = {
	 0,				/* standard input */
	 1,				/* standard output */
	-1,				/* no default? */
	 0,				/* default is no flow control */
	 0,				/* default is multi-byte access */
	{B9600, B9600, 0, 0, 0},	/* tty stuff will get set correctly */
	{B9600, B9600, 0, 0, 0},	/* ditto */
	(char *) 0
};

static struct channel remote = {
	-1,				/* none */
	 1,				/* standard output */
	-1,				/* none */
	 0,				/* default is no flow control */
	 0,				/* default is multi-byte access */
	{B9600, B9600, 0, 0, 0},	/* ditto */
	{B9600, B9600, 0, 0, 0},	/* ditto */
	(char *) 0
};

main(argc, argv)
  int   argc;
  char *argv[];
{
    int i;

    Lb = -1;  Rb = 9600;
    dflag = 0;
    eflag = 0;

    av = argv;
    i  = 1;
    while (i < argc) {
	if (argv[i][0] == '-') {
	    switch(argv[i][1]) {
	        case 'd':
		    dflag++;
		    break;

		case 'e':
		    eflag++;
		    break;

		case 'l':
		case 'r':
		    if ((argv[i][2] == 'e') && ((i + 1) < argc)) {
			if (argv[i][1] == 'l') {
			    local.ch_esc = argv[++i][0];
			} else {
			    remote.ch_esc = argv[++i][0];
			}
		    } else if ((argv[i][2] == 'b') && ((i + 1) < argc)) {
			if (argv[i][1] == 'l') {
			    Lb = atoi(argv[++i]);
			} else if (argv[i][1] == 'r') {
			    Rb = atoi(argv[++i]);
			} else {
			    usage(1);
			}
		    } else if ((argv[i][2] == 'f') && ((i + 1) < argc)) {
			if (argv[i][1] == 'l') {
			    local.ch_flo++;
			} else if (argv[i][1] == 'r') {
			    remote.ch_flo++;
			} else {
			    usage(1);
			}
		    } else if (argv[i][2] == '1') {
			if (argv[i][1] == 'l') {
			    local.ch_sgl++;
			} else if (argv[i][1] == 'r') {
			    remote.ch_sgl++;
			} else {
			    usage(1);
			}
		    } else {
			usage(1);
		    }
		    break;

		default:
		    usage(2);
		    break;

	    } /* switch */
	} else {
	    if (local.ch_nam && remote.ch_nam) {
		usage(3);
	    }
	    if (remote.ch_nam) {
		local.ch_nam  = remote.ch_nam;
		remote.ch_nam = argv[i];
	    } else {
		remote.ch_nam = argv[i];
	    }
	}
	i++;
    }
    if (remote.ch_nam == (char *) 0) {
	usage(4);
    }

    openchannel(&local,   local.ch_nam, remote.ch_nam);
/*
 *  openchannel(&remote, remote.ch_nam, local.ch_nam);
 *
 * This is admittedly a hack.  Most drivers support multiple opens, but
 * some don't.  So I am hacking at the original symetric design to get
 * around the problem.  Note that openchannel now uses O_RDWR.  (GDB)
 */
    remote.ch_inp = local.ch_out;
    remote.ch_out = local.ch_inp;

    if (dflag) {
	debugchannel(&local);
	debugchannel(&remote);
	exit(0);
    }

    signal(SIGHUP,  quit);
    signal(SIGTERM, quit);
    signal(SIGTTIN, stop);
    signal(SIGINT,  SIG_IGN);
    signal(SIGTTOU, SIG_IGN);
    if (isatty(local.ch_inp)) {
	setmodes(&local, Lb);
    }
    if (isatty(remote.ch_inp)) {
	eflag = 0;
	setmodes(&remote, Rb);
    }

    couple(&local, &remote);

    quit(0);	/* never reached...but just in case... */
}


disconnect(chan, msg)
  struct channel *chan;
  char *msg;
{
    ioctl(chan->ch_inp, TIOCSETP, &chan->ch_sav);
    printf("\r\n%s\r\n", (msg) ? msg : "[disconnect]");
}


connect(chan, msg)
  struct channel *chan;
  char *msg;
{
    ioctl((sending ? savedes : chan->ch_inp), TIOCSETP, &chan->ch_tty);
    printf("\r\n%s\r\n", (msg) ? msg : "[connect]");
}


setmodes(chan, bf)
  struct channel *chan;
  int bf;
{
    int baud;

    ioctl(chan->ch_inp, TIOCGETP, &chan->ch_sav);
    ioctl(chan->ch_inp, TIOCGETP, &chan->ch_tty);

    if (!eflag) {
	chan->ch_tty.sg_flags |= RAW;
	chan->ch_tty.sg_flags &= ~TANDEM;
	chan->ch_tty.sg_flags &= ~CRMOD;
	chan->ch_tty.sg_flags &= ~ECHO;
	chan->ch_tty.sg_flags &= ~CBREAK;
    }

    if (bf != -1) {
	baud = findbaud(bf);
	if (baud > EXTB) {
	    fprintf(stderr, "%s: illegal baud rate\n", av[0]);
	    usage(5);
	}
	chan->ch_tty.sg_ispeed = chan->ch_tty.sg_ospeed = baud;
    }
    ioctl(chan->ch_inp, TIOCSETP, &chan->ch_tty);
    if (chan->ch_flo) {
#ifdef gould
	if (ioctl(chan->ch_inp, TIOCRTSFLOW, 0) < 0) {
	    fprintf(stderr, "%s: ioctl error: errno = %d\n", av[0], errno);
	    perror("TIOCRTSFLOW");
	}
#endif
    }
}


debugchannel(chan)
  struct channel *chan;
{
    fprintf(stderr, "ch_nam='%s'\n",
	(chan->ch_nam) ? chan->ch_nam : "**stdin**"
    );
    fprintf(stderr, "ch_inp=%d\n", chan->ch_inp);
    fprintf(stderr, "ch_out=%d\n", chan->ch_out);
    fprintf(stderr, "ch_esc=%d\n", chan->ch_esc);
    fprintf(stderr, "ch_flo=%s\n", chan->ch_flo ? "On" : "Off");
    fprintf(stderr, "\n");
}


openchannel(chan, inp, out)
  struct channel *chan;
  char *inp, *out;
{
    if (inp) {
	chan->ch_inp = open(inp, O_RDWR | O_NDELAY | O_EXCL, 0600);
	if (chan->ch_inp < 0) {
	    fprintf(stderr, "%s: openchannel: open: %s: %s\n",
		av[0], inp, sys_errlist[errno]
	    );
	    exit(1);
	}
    }
    if (out) {
	chan->ch_out = open(out, O_RDWR | O_NDELAY | O_EXCL, 0600);
	if (chan->ch_out < 0) {
	    fprintf(stderr, "%s: openchannel: open: %s: %s\n",
		av[0], out, sys_errlist[errno]
	    );
	    exit(1);
	}
    }
}


escape(chan, buf, pos, size)
  struct channel *chan;
  char *buf;
  int  pos, size;
{
    char filename[256];
    char buffer[1024];
    char next;
    int  n;

    if ((pos+1) == size) {
	read(chan->ch_inp, &next, 1);
    } else {
	next = buf[pos+1];
    }
    if (next == chan->ch_esc) {
	return(0);
    }
    switch(next) {
	case 't':	/* type a file... */
	    ioctl(chan->ch_inp, TIOCSETP, &chan->ch_sav);
	    printf("Type: "); fflush(stdout);
	    scanf("%s", filename);
	    n = glob(filename, buffer, sizeof(buffer));
	    if (n > 0) {
		swapdesc(buffer, chan);
		sprintf(filename, "[<CR> to send '%s']", buffer);
		connect(chan, filename);
	    } else {
		sprintf(buffer, "[no match: '%s']", filename);
		connect(chan, buffer);
	    }
	    break;

	case '>':	/* old. */
	    ioctl(chan->ch_inp, TIOCSETP, &chan->ch_sav);
	    printf("Send: "); fflush(stdout);
	    scanf("%s", filename);
	    n = glob(filename, buffer, sizeof(buffer));
	    if (n > 0) {
		download(buffer, chan, 0, 1);
		connect(chan, 0);
	    } else {
		sprintf(buffer, "[no match: '%s']", filename);
		connect(chan, buffer);
	    }
	    break;

	case 'l':	/* send a trillium load module to "load" */
	    ioctl(chan->ch_inp, TIOCSETP, &chan->ch_sav);
	    printf("Load: "); fflush(stdout);
	    scanf("%s", filename);
	    n = glob(filename, buffer, sizeof(buffer));
	    if (n > 0) {
		download(buffer, chan, 1, BUFSIZE);
		connect(chan, 0);
	    } else {
		sprintf(buffer, "[no match: '%s']", filename);
		connect(chan, buffer);
	    }
	    break;

	case '\032':	/* ^Z */
	    stop();
	    break;
	    
	case '.':
	    quit(0);
	    break;
	    
	case '?':
	    help();
	    break;

	default:
	    write(chan->ch_out, &buf[pos], 1);
	    write(chan->ch_out, &next, 1);
	    return(1);
	    break;
    }
    return(1);
}


help()
{
    disconnect(&local.ch_inp, "[help]");
    printf("connected from '%s' to '%s'\n",
	(local.ch_nam ? local.ch_nam : "**stdin**"),
	(remote.ch_nam ? remote.ch_nam : "**stdout**")
    );
    printf("local escape: ");
    escapechar(local.ch_esc);
    printf("local baud: %d\n", baudrates[local.ch_tty.sg_ispeed]);
    printf("remote escape: ");
    escapechar(remote.ch_esc);
    printf("remote baud: %d\n", baudrates[remote.ch_tty.sg_ispeed]);
    printf("\n\nValid escapes:\n");
    printf("\tt       \"type\" a file to remote.\n");
    printf("\t<       Send a file.\n");
    printf("\tl       Send the length and contents of a file.\n");
    printf("\tCTRL-Z  Suspend.\n");
    printf("\t?       This message.\n");
    printf("\t.       Quit.\n");
    sleep(1);
    connect(&local.ch_inp, 0);
    sleep(1);
}


escapechar(c)
  char c;
{
    if (c >= 0) {
	printf("%3d  ", c);
	if (c < ' ') {
	    printf("<CTRL-'%c'>\n", c+64);
	} else {
	    printf("'%c'\n", c);
	}
    } else {
	printf("NONE\n");
    }
}


stop()
{
    ioctl(local.ch_inp, TIOCSETP, &local.ch_sav);
    ioctl(remote.ch_inp, TIOCSETP, &remote.ch_sav);
    printf("\n[stopped]\n");
    kill(getpid(), SIGTSTP);
    printf("\n[reconnect]\n");
    ioctl(local.ch_inp, TIOCSETP, &local.ch_tty);
    ioctl(remote.ch_inp, TIOCSETP, &remote.ch_tty);
}


quit(n)
  int n;
{
    ioctl(local.ch_inp, TIOCSETP, &local.ch_sav);
    ioctl(remote.ch_inp, TIOCSETP, &remote.ch_sav);
    exit(n);
}


couple(lcl, rem)
  struct channel *lcl, *rem;
{
    int nfds, nfound;
    int readfds, lcl_mask, rem_mask, mask;

    nfds = lcl->ch_inp;
    lcl_mask = (1 << lcl->ch_inp);
    rem_mask = (1 << rem->ch_inp);
    mask = lcl_mask | rem_mask;
    for (;;) {
	readfds = mask;
	nfound = select(20, &readfds, 0, 0, 0);
	if (nfound > 0) {
	    if (lcl_mask & readfds) {
		transfer(lcl, iobuffer);
	    }
	    if (rem_mask & readfds) {
	        transfer(rem, iobuffer);
	    }
	} else {
	    fprintf(stderr, "%s: couple: select: %s\n",
		av[0], sys_errlist[errno]
	    );
	}
    }
}


transfer(chan, buf)
  struct channel *chan;
  register char  *buf;
{
    register int i, ibytes, escch;
    int  obytes, flush, unwritten, bsize;

    bsize = ((chan->ch_sgl || sending) ? 1 : BUFSIZE);
    if ((ibytes = read(chan->ch_inp, buf, bsize)) > 0) {
	unwritten = 1;
	if ((escch = chan->ch_esc) >= 0) {
	    for (i = 0; i < ibytes; i++) { /* scan the buf for escapes */
		if (buf[i] == escch) {
		    obytes = write(chan->ch_out, buf, i);
		    if (obytes >= 0) {
			flush = escape(chan, buf, i, ibytes);
			if (flush) {
			    i = ibytes;
			} else if (write(chan->ch_out, &buf[i], ibytes-i) < 0){
			    perror("transfer: write (1)");
			}
		    } else {
			perror("transfer: write (2)");
		    }
		    unwritten = 0;
		}
	    }
	} else { /* no escape character so just write it out */
	    if (write(chan->ch_out, buf, ibytes) < 0) {
		perror("transfer: write(3)");
	    }
	    unwritten = 0;
	}
	if (unwritten) {
	    if (write(chan->ch_out, buf, ibytes) < 0) {
		quit(0);
	    }
	}
    } else {
	if (sending) {
	    sending = 0;
	    close(chan->ch_inp);
	    chan->ch_inp = savedes;
	    connect(chan, 0);
	} else {
	    quit(0);
	}
    }

}


findbaud(n)
  int n;
{
    int i;

    i = 0;
    while ((baudrates[i] != n) && (baudrates[i] != -1)) {
	    i++;
    }
    return(i);
}


usage(n)
  int n;
{
    if (n)
	fprintf(stderr, "usage: %s %s\n", av[0], USAGE);
    exit(n);
}


download(path, chan, sendlen, bfsize)
  char *path;
  struct channel *chan;
  int sendlen, bfsize;
{
    struct stat stats;
    char        xbuf[BUFSIZE];
    int         fd, nbytes;

    if (sendlen) {
	if (stat(path, &stats) < 0) {
	    fprintf(stderr, "%s: download: stat: %s: %s\n",
		av[0], path, sys_errlist[errno]
	    );
	    return;
	} else {
	    sendlength(chan->ch_out, stats.st_size);
	}
    }
    fd = open(path, O_RDONLY, 0600);
    if (fd >= 0) {
	printf("[%s]\n", path);  fflush(stdout);
	do {
	    nbytes = read(fd, xbuf, bfsize);
	    write(chan->ch_out, xbuf, nbytes);
	    if (sendlen) {
		printf("."); fflush(stdout);
	    }
	} while (nbytes > 0);
	close(fd);
    } else {
	fprintf(stderr, "%s: download: open: %s: %s\n",
	    av[0], path, sys_errlist[errno]
	);
    }
}

sendlength(d, n)
  int   d;
  off_t n;
{
#ifdef vax
    /* Vax and Transputer byte ordering are the same */
    write(d, &n, sizeof(n));
#endif vax

#ifdef gould
    /* Gould/Sun byte ordering needs to be reversed */
    char b;
    b = (n)       & 0xFF;  write(d, &b, 1);
    b = (n >>  8) & 0xFF;  write(d, &b, 1);
    b = (n >> 16) & 0xFF;  write(d, &b, 1);
    b = (n >> 24) & 0xFF;  write(d, &b, 1);
#endif gould
}


swapdesc(f, chan)
  char *f;
  struct channel *chan;
{
    int d;

    d = open(f, O_RDONLY | O_NDELAY, 0600);
    if (d >= 0) {
	savedes = chan->ch_inp;
	sending = 1;
	chan->ch_inp = d;
    } else {
	fprintf(stderr, "%s: swapdesc: open: %s: %s\n",
	    av[0], f, sys_errlist[errno]
	);
    }
}
