/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: fmod64.c,v 2.2 92/04/21 10:54:59 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	File:	fmod64.c
 *
 *	Implements T800/T414 64 bit floating point "fmod(x,y)".
 *
 *	Based on INMOS supplied code.  Modified by Logical Systems for correct
 *	ANSI "C" setting of "errno" in ERANGE/EDOM situations.
 */

#include	<inline.h>
#include	<math.h>
#include	<stddef.h>

#pragma	asm_expand			/* Recognize EDOM/ERANGE in "asm" */

double
fmod(double x,double y)
	{
	#pragma	asm
#ifdef	T800
;
;	T800 64 bit version
;
	.ext	errno
	ldlp	[y]			;Point to "y"
	ldlp	[x]			;Point to "x"
	fpldnldb
	fpnotfinite
	stl	[x]			;Hold flag in old "x" parameter slot
	fpldnldb
	fpnotfinite
	cj	@L2			;Jump if "y" not NaN or infinite
L0:	fpadd				;Add to propagate NaN's
L1:	.ldc	EDOM			;Domain error (input was NaN probably)
	_asm_eval1(&errno);
	stnl	0			;Set errno to EDOM
	ret				;Return NaN as result
	.align
unstable
	.dw	0x00000000
	.dw	0x7FF00001		;"unstable" NaN
L2:	ldl	[x]			;Was "x" Nan or infinite?
	cj	@L3			;Nope
	.ldc	0
	cj	@L0			;Jump since "x" was NaN or infinite
L3:	fpdiv
	fpnotfinite
	cj	@L4			;Jump if result looks OK
	.ldc	unstable		;Result will be "unstable" NaN
	fpldnldb
	.ldc	0
	cj	@L1			;Jump if result is NaN
L4:	fpdup
	fpurz
	fpint
	fpsub
	ldlp	[y]
	fpldnlmuldb
	ret
#else	/* T800 */
;
;	T414 64 bit version
;
	.ext	errno
	.ext	_FPstate
	.ext	_r64fei
	.ext	_r64op
	.ext	_r64tei
	_asm_ajw(-5)
	.ldc	0x7FF00000		;NaN/Inf dividing line
	ldl	[x]+1
	gt
	eqc	0
	cj	@L5			;Jump if not NaN/Inf
;
;	"x" is Inf/NaN, figure out if "y" is also and if so which is larger.
;	Then use information to determine whether to return "x" or "y" in
;	conformance with NaN propagation rules.
;
	ldl	[x]+1			;Which is larger?
	ldl	[y]+1
	gt
	cj	@L1			;Jump if "x" <= "y"
L0:	ldlp	[x]			;Return "x" (NaN/Inf)
	.ldc	_FPstate
	.ldc	8
	move
	j	@L3
L1:	ldl	[x]+1
	ldl	[y]+1
	diff
	cj	@L4			;Jump if MSW's equal
L2:	ldlp	[y]			;Return "y" (NaN/Inf)
	.ldc	_FPstate
	.ldc	8
	move
L3:	.ldc	EDOM			;Domain error
	_asm_eval1(&errno);
	stnl	0			;Signal "domain" error
	j	@L7
L4:	ldl	[x]			;High order MSW's equal, check lows
	ldl	[y]
	gt
	eqc	0
	cj	L0			;Jump if "x" > "y"
	ldl	[x]
	ldl	[y]
	diff
	cj	@L2			;Jump if "x" <= "y"
	j	@L0			;Else return "x"
	.align
unstable:
	.dw	0x00000000		;"unstable" NaN
	.dw	0x7FF00001
;
;	We know "x" wasn't a NaN/Inf, now check "y".
;
L5:	.ldc	0x7FF00000		;NaN/Inf dividing line
	ldl	[y]+1
	gt
	cj	@L2			;Jump if NaN/Inf
;
;	Now we do the actual function work!
;
	ldlp	[y]
	stl	0
	.ldc	3			;We need a divide
	ldlp	[x]
	ldlp	1			;Result
	call	@_r64op			;"x"/"y"
;
;	Check for bogus result.
;
	.ldc	0x7FF00000		;NaN/Inf dividing line
	ldl	2
	gt
	eqc	0
	cj	@L6			;Jump if ! NaN/Inf
	.ldc	unstable		;Return "unstable" NaN
	.ldc	_FPstate
	.ldc	8
	move
	j	@L3			;Signal EDOM
;
;	Truncate result and subtract difference with original to get
;	remainder.
;
L6:	ldlp	1
	.ldc	0			;Truncate result
	ldlp	3
	call	@_r64tei		;Convert to 64 bit integer
	ldlp	3			;Temp storage
	.ldc	0			;Truncate result
	ldlp	[x]			;Re-use "x" slot
	call	@_r64fei		;Convert back to double
	ldlp	[x]			;Truncated version of quotient
	stl	0
	.ldc	1			;We need a subtract
	ldlp	1			;Original quotient
	ldlp	3			;Result
	call	@_r64op			;"x"/"y" - trunc("x"/"y")
	ldlp	3			;Fractional component
	stl	0
	.ldc	2			;We need a multiply
	ldlp	[y]
	.ldc	_FPstate		;Result
	call	@_r64op
L7:	_asm_ajw(5)
	ret
#endif	/* T800 */
	#pragma	endasm
	}
