/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: sqrt32.c,v 2.2 92/04/21 10:55:04 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	File:	sqrt32.c
 *
 *	Implements T800/T414 32 bit floating point "sqrtf(x)/sqrt(x)".
 *
 *	Based on INMOS supplied code.  Conversion by Scott Lloyd of CSA.
 *	Addition of ERRNO handling and changed calling sequence by Kirk Bailey
 *	of Logical Systems.
 */

#include	<inline.h>
#include	<math.h>
#include	<stddef.h>

#pragma	asm_expand			/* Recognize EDOM/ERANGE in "asm" */

#undef	sqrt
#undef	sqrtf

float
sqrtf(float x)
	{
	#pragma	asm
#ifdef	DOUBLE32
	.pub	sqrt
sqrt:
#endif	DOUBLE32

#ifdef	T800
;
;	T800 32 bit version
;
	.ext	errno
	fpldzerosn			;Used later to check argument sign
	ldlp	[x]
	fpldnlsn
	fpnan				;Input NaN?
	eqc	0
	cj	@L2			;Jump if so (pass on through)
	fpnotfinite			;Infinity?
	cj	@L1			;Jump if not
	.ldc	unstable
	fpldnlsn			;Load "unstable" NaN
	.ldc	EDOM			;Domain error
	_asm_eval1(&errno);
	stnl	0			;Signal "domain" error
	ret
L1
	fpgt				;Negative argument?
	cj	@L3			;Jump if non-negative argument
	.ldc	undefined
	fpldnlsn			;Load "undefined" NaN
L2	.ldc	EDOM			;Domain error
	_asm_eval1(&errno);
	stnl	0
	ret
	.align
unstable
	.dw	0x7F800008		;"unstable" NaN
undefined
	.dw	0x7F800010		;"undefined" NaN
L3	ldlp	[x]
	fpldnlsn			;Get fresh copy of argument
	fpusqrtfirst
	fpusqrtstep
	fpusqrtstep
	fpusqrtlast
#else	/* T800 */
;
;	T414 32 bit version
;
	.ext	?SN32TOFAREG
	.ext	errno
	_asm_ajw(-10);
	.ldc	L600
	stl	8
	ldl	[x]			;Operand
	ldl	8
	ldnl	0
	and				;Mask off sign
	stl	7
	.ldc	0
	ldl	7
	unpacksn
	stl	5			;Fractional part
	stl	6			;Exponent part
	stl	4			;Operand "type"
	ldl	4
	eqc	3			;NaN?
	eqc	0
	cj	@L601			;Jump if so
	ldl	4
	eqc	0			;Zero?
	cj	@L602			;Jump if not
	ldl	[x]
	stl	9
	j	@L617
L601:	ldl	[x]			;Pass NaN on through
	stl	9
	j	@L602A			;Signal EDOM
L602:	ldl	[x]
	ldl	7
	diff				;Negative?
	cj	@L603			;Jump if not
	ldl	8
	ldnl	1			;"undefined" NaN
	stl	9
L602A:	.ldc	EDOM			;Domain error
	_asm_eval1(&errno);
	stnl	0			;Signal "domain" error
	j	@L617
L603:	ldl	4
	eqc	2			;Infinite?
	cj	@L604			;Jump if not
	ldl	8
	ldnl	6			;"unstable" NaN
	stl	9
	j	@L602A			;Signal "domain" error
;
;	INMOS BUG!!!
;
;	The following statement was originally a "ldl 6", but it appears to be
;	checking for denormalized numbers based on the code which follows.  To
;	do this the operand "type" information in local offset 4 needs to be
;	checked to determine the value returned in "C" from the "unpacksn".
;	Alternatively, this could be a bug in Scott's disassembly but I'm
;	betting against INMOS (not that denormalized #'s are used very often
;	anyway, but...)
;
;	Kirk Bailey, 2/17/88
;
L604:	ldl	4
	eqc	1			;Denormalized?
	cj	@L605			;Jump if not
	ldl	5
	.ldc	0
	norm				;Normalize fraction
	stl	1			;Toss LSW
	stl	5			;Normalized fraction
	stl	2			;Exponent modifier
	ldl	6			;Old exponent
	ldl	2
	sub
	stl	6			;Updated exponent
L605:	ldl	6
	.ldc	1
	and
;	eqc	1			;Not needed, tough for OCCAM to detect
	cj	@L606			;Jump if even exponent (base 2)
	ldl	5
	.ldc	1
	shr
	stl	5
	.ldc	0
	ldl	8
	ldnl	2
	ldl	5
	lmul
	stl	2
	stl	3
	.ldc	0
	ldl	8
	ldnl	3
	ldl	3
	lsum
	stl	3
	stl	2
	j	@L607
	.align
L600:
	.dw	0x7fffffff		;Mask for sign bit
	.dw	0x7F800010		;NaN if - ("undefined", was 0x7f840000)
	.dw	0xd413ccd0
	.dw	0x4c1e24cd
	.dw	0x93000000
	.dw	0x6cffffff
	.dw	0x7F800008		;NaN for +INF ("unstable")
L606:	.ldc	0
	ldl	8
	ldnl	4
	ldl	5
	lmul
	stl	2
	stl	3
	.ldc	0
	ldl	8
	ldnl	5
	ldl	3
	lsum
	stl	3
	stl	2
L607:	ldl	5
	.ldc	1
	shr
	stl	2
	ldl	2
	.ldc	0
	ldl	3
	ldiv
	stl	1
	stl	0
	.ldc	0
	ldl	3
	.ldc	1
	shr
	ldl	1
	rev
	lsum
	stl	3
	stl	0
	ldl	2
	.ldc	0
	ldl	3
	ldiv
	stl	1
	stl	0
	.ldc	0
	ldl	3
	.ldc	1
	shr
	ldl	1
	rev
	lsum
	stl	3
	stl	0
	ldl	3
	.ldc	-64
	and
	stl	3
	.ldc	0
	ldl	3
	ldl	3
	lmul
	stl	2
	stl	1
	ldl	1
	ldl	5
	diff
	eqc	0
	cj	@L608
	ldl	2
	eqc	0
	cj	@L608
	j	@L613
L608:	.ldc	0
	ldl	1
	ldl	5
	diff
	gt
	eqc	0
	cj	@L611
	ldl	3
	.ldc	32
	diff
	stl	3
	ldl	3
	mint
	and
	eqc	0
	cj	@L610
	ldl	3
	.ldc	1
	shl
	stl	3
	ldl	6
	adc	-2
	stl	6
L610:	j	@L613
L611:	.ldc	0
	ldl	1
	ldl	5
	diff
	gt
	cj	@L612
	ldl	3
	adc	32
	stl	3
	j	@L613
L612:	stopp
L613:	ldl	3
	.ldc	0
	or
	cj	@L616
	ldl	6
	adc	127
	.ldc	1
	shr
	stl	0
	ldl	3
	.ldc	0
	norm
	postnormsn
	roundsn
L616:	stl	9
L617:
	ldl	9
	call	@?SN32TOFAREG		;Store 32 bit fp in A into FAreg
	_asm_ajw(10);
#endif	/* T800 */
	#pragma	endasm
	}
