/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: sqrt64.c,v 2.2 92/04/21 10:55:05 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	File:	sqrt64.c
 *
 *	Implements T800/T414 64 bit floating point "sqrt(x)".
 *
 *	Based on INMOS supplied code.  Conversion by Scott Lloyd of CSA.
 *	Addition of ERRNO handling and changed calling sequence by Kirk Bailey
 *	of Logical Systems.
 */

#include	<inline.h>
#include	<math.h>
#include	<stddef.h>

#pragma	asm_expand			/* Recognize EDOM/ERANGE in "asm" */

#undef	sqrt

double
sqrt(double x)
	{
	#pragma	asm
#ifdef	T800
;
;	T800 64 bit version
;
	.ext	errno
	fpldzerodb			;Used later to check argument sign
	ldlp	[x]
	fpldnldb
	fpnan				;Input NaN?
	eqc	0
	cj	@L2			;Jump if so (pass on through)
	fpnotfinite			;Infinity?
	cj	@L1			;Jump if not
	.ldc	unstable
	fpldnldb			;Load "unstable" NaN
	.ldc	EDOM			;Domain error
	_asm_eval1(&errno);
	stnl	0			;Signal "domain" error
	ret
L1
	fpgt				;Negative argument?
	cj	@L3			;Jump if non-negative argument
	.ldc	undefined
	fpldnldb			;Load "undefined" NaN
L2	.ldc	EDOM			;Domain error
	_asm_eval1(&errno);
	stnl	0
	ret
	.align
unstable
	.dw	0x00000000		;"unstable" NaN
	.dw	0x7FF00001
undefined
	.dw	0x00000000		;"undefined" NaN
	.dw	0x7FF00002
L3	ldlp	[x]
	fpldnldb			;Get fresh copy of argument
	fpusqrtfirst
	fpusqrtstep
	fpusqrtstep
	fpusqrtstep
	fpusqrtstep
	fpusqrtstep
	fpusqrtlast
#else	/* T800 */
;
;	T414 64 bit version.  A first cut at removing OCCAM generated redundant
;	code eliminated 25-40% for the small regions worked over.  If time ever
;	permits a careful assembly language re-implementation looks EXTREMELY
;	worthwhile.
;
	.ext	_FPstate
	.ext	errno
	_asm_ajw(-23);
	.ldc	_FPstate		;Addr of "soft" FAreg
	stl	22
	.ldc	data
	stl	21
	ldl	21
	ldlp	17
	.ldc	8
	move
	ldl	[x]			;Load low order word of number
	stl	19
	ldl	[x]+1			;Load high order word of number
	ldl	18
	and
	stl	20
	ldl	21
	ldnlp	2
	ldlp	17
	.ldc	8
	move
	.ldc	0
	stl	16
	ldlp	17
	ldlp	19
	ldlp	16
	call	@P1			;Set 16 to 1 if NaN
	ldl	19
	.ldc	0
	diff
	eqc	0
	cj	@L32			;Jump if argument != 0
	ldl	20
	.ldc	0
	diff
	eqc	0
L32:	eqc	0
	cj	@L33			;Jump if argument == 0
	ldl	16
	cj	@L34			;Jump if !NaN and !0
	.ldc	EDOM			;"domain" error code
	_asm_eval1(&errno);
	stnl	0			;Signal "domain" error
L33:	ldlp	[x]			;Return argument (0 or NaN)
	ldl	22
	.ldc	8
	move
	j	@L57
	.align
data:	.dw	0xffffffff
	.dw	0x7fffffff
	.dw	0x00000000
	.dw	0x7ff00000
;
;	The following floating point constant represents the value of the
;	"undefined" NaN.  The original INMOS code used a value for this which
;	is inconsistent with the latest INMOS Compiler Writers' Guide.  The
;	value has been changed to reflect the one used in the "Guide".
;
	.dw	0x00000000		;Was 0x00000000
	.dw	0x7ff00002		;Was 0x7FF08000
	.dw	0xe7799211
	.dw	0xd413cccf
	.dw	0x824c9a4e
	.dw	0x4c1e24cc
	.dw	0x00000000
	.dw	0x93000000
	.dw	0xffffffff
	.dw	0x6cffffff
	.dw	0x00000000		;"unstable" NaN (low word)
	.dw	0x7ff00001		;"unstable" NaN (high word)
L34:	ldl	[x]+1			;Load high order word of number
	ldl	20
	diff
	cj	@L36			;Jump if # positive
	.ldc	EDOM			;"domain" error code
	_asm_eval1(&errno);
	stnl	0			;Signal "domain" error
	ldl	21			;Negative argument, return NaN
	ldnlp	4
	ldl	22
	.ldc	8
	move
	j	@L57
L36:	ldl	21
	ldnlp	2
	ldlp	17
	.ldc	8
	move				;Get a copy of 0x7FF00000,0x00000000
	ldl	[x]			;Load low order word of number
	ldl	17
	diff
	eqc	0
	cj	@L37			;Jump if not +/- INF
	ldl	[x]+1			;Load high order word of number
	ldl	18
	diff
	eqc	0			;TRUE if argument is +/- INF
L37:	cj	@L38			;Jump if arg != +/- INF
	.ldc	EDOM			;"domain" error code
	_asm_eval1(&errno);
	stnl	0			;Signal "domain" error
	ldl	21			;Return "unstable" NaN
	ldnlp	14
	ldl	22
	.ldc	8
	move
	j	@L57
L38:	.ldc	0
	stl	10
	.ldc	0
	stl	11
	.ldc	12
	stl	1
	ldlp	19
	stl	0
	ldlp	10
	ldlp	13
	ldlp	17
	call	@P4
	ldl	17
	stl	12
	ldl	12
	eqc	0
	cj	@L39
	.ldc	0
	stl	8
	.ldc	0
	stl	9
	ldlp	8
	stl	1
	ldlp	13
	stl	0
	ldlp	10
	ldlp	13
	ldlp	17
	call	@P6
	ldl	17
	stl	11
	.ldc	0
	ldl	11
	sub
	stl	12
	j	@L40
L39:	ldl	14
	ldl	13
	.ldc	1
	lshr
	stl	10
	stl	11
	.ldc	0
	stl	8
	mint
	stl	9
	ldl	10
	ldl	8
	or
	stl	13
	ldl	11
	ldl	9
	or
	stl	14
L40:	ldl	12
	.ldc	1
	and
	eqc	1
	cj	@L41
	ldl	14
	ldl	13
	.ldc	1
	lshr
	stl	13
	stl	14
	ldl	21
	ldnlp	6
	ldlp	8
	.ldc	8
	move
	.ldc	0
	stl	6
	.ldc	0
	stl	7
	ldlp	6
	stl	1
	ldlp	8
	stl	0
	ldlp	13
	ldlp	10
	ldlp	15
	call	@P8
	ldl	21
	ldnlp	8
	ldlp	8
	.ldc	8
	move
	.ldc	0
	stl	6
	.ldc	0
	stl	7
	ldlp	6
	stl	1
	ldlp	8
	stl	0
	ldlp	15
	ldlp	15
	ldlp	10
	call	@P7
	j	@L42
L41:	ldl	21
	ldnlp	10
	ldlp	8
	.ldc	8
	move
	.ldc	0
	stl	6
	.ldc	0
	stl	7
	ldlp	6
	stl	1
	ldlp	8
	stl	0
	ldlp	13
	ldlp	10
	ldlp	15
	call	@P8
	ldl	21
	ldnlp	12
	ldlp	8
	.ldc	8
	move
	.ldc	0
	stl	6
	.ldc	0
	stl	7
	ldlp	6
	stl	1
	ldlp	8
	stl	0
	ldlp	15
	ldlp	15
	ldlp	10
	call	@P7
L42:	ldl	14
	ldl	13
	.ldc	1
	lshr
	stl	10
	stl	11
	.ldc	0
	stl	4
	.ldc	0
	stl	5
	ldlp	21
	stl	2
	ldlp	15
	stl	1
	ldlp	4
	stl	0
	ldlp	10
	ldlp	6
	ldlp	8
	call	@P2
	ldl	16
	ldl	15
	.ldc	1
	lshr
	stl	4
	stl	5
	.ldc	0
	stl	2
	.ldc	0
	stl	3
	ldlp	2
	stl	1
	ldlp	8
	stl	0
	ldlp	4
	ldlp	15
	ldlp	6
	call	@P7
	.ldc	0
	stl	4
	.ldc	0
	stl	5
	ldlp	21
	stl	2
	ldlp	15
	stl	1
	ldlp	4
	stl	0
	ldlp	10
	ldlp	6
	ldlp	8
	call	@P2
	ldl	16
	ldl	15
	.ldc	1
	lshr
	stl	4
	stl	5
	.ldc	0
	stl	2
	.ldc	0
	stl	3
	ldlp	2
	stl	1
	ldlp	8
	stl	0
	ldlp	4
	ldlp	15
	ldlp	6
	call	@P7
	.ldc	0
	stl	4
	.ldc	0
	stl	5
	ldlp	21
	stl	2
	ldlp	15
	stl	1
	ldlp	4
	stl	0
	ldlp	10
	ldlp	6
	ldlp	8
	call	@P2
	ldl	16
	ldl	15
	.ldc	1
	lshr
	stl	4
	stl	5
	.ldc	0
	stl	2
	.ldc	0
	stl	3
	ldlp	2
	stl	1
	ldlp	8
	stl	0
	ldlp	4
	ldlp	15
	ldlp	6
	call	@P7
	.ldc	-512
	stl	6
	.ldc	-1
	stl	7
	ldl	15
	ldl	6
	and
	stl	15
	ldl	16
	ldl	7
	and
	stl	16
	.ldc	0
	stl	6
	.ldc	0
	stl	7
	ldlp	6
	stl	1
	ldlp	15
	stl	0
	ldlp	15
	ldlp	10
	ldlp	8
	call	@P8
	.ldc	0
	stl	7
	ldl	8
	ldl	13
	diff
	eqc	0
	cj	@L43
	ldl	9
	ldl	14
	diff
	eqc	0
L43:	stl	7
	.ldc	0
	stl	5
	.ldc	0
	stl	6
	.ldc	0
	stl	4
	ldl	10
	ldl	5
	diff
	eqc	0
	cj	@L44
	ldl	11
	ldl	6
	diff
	eqc	0
L44:	stl	4
	ldl	7
	cj	@L45
	ldl	4
	cj	@L45
	j	@L50
L45:	.ldc	0
	ldl	8
	ldl	13
	ldiff
	stl	6
	ldl	9
	ldl	14
	ldiff
	stl	7
	.ldc	0
	stl	4
	.ldc	0
	stl	5
	.ldc	0
	stl	3
	ldlp	6
	ldlp	4
	ldlp	3
	call	@P1
	ldl	3
	eqc	0
	cj	@L49
	.ldc	256
	stl	6
	.ldc	0
	stl	7
	.ldc	0
	ldl	15
	ldl	6
	ldiff
	stl	15
	ldl	16
	ldl	7
	ldiff
	stl	16
	.ldc	0
	stl	6
	mint
	stl	7
	ldl	15
	ldl	6
	and
	stl	6
	ldl	16
	ldl	7
	and
	stl	7
	.ldc	0
	stl	4
	.ldc	0
	stl	5
	.ldc	0
	stl	3
	ldl	6
	ldl	4
	diff
	eqc	0
	cj	@L47
	ldl	7
	ldl	5
	diff
	eqc	0
L47:	stl	3
	ldl	3
	cj	@L48
	ldl	16
	ldl	15
	.ldc	1
	lshl
	stl	15
	stl	16
	ldl	12
	adc	-2
	stl	12
L48:	j	@L50
L49:	.ldc	256
	stl	6
	.ldc	0
	stl	7
	.ldc	0
	ldl	6
	ldl	15
	lsum
	stl	15
	ldl	7
	ldl	16
	ladd
	stl	16
L50:	ldl	12
	adc	1023
	.ldc	1
	shr
	xdble
	stl	17
	stl	18
	.ldc	2047
	stl	10
	.ldc	0
	stl	11
	ldl	15
	ldl	10
	and
	stl	13
	ldl	16
	ldl	11
	and
	stl	14
	ldl	16
	ldl	15
	.ldc	1
	lshl
	stl	8
	stl	9
	.ldc	12
	stl	1
	ldlp	8
	stl	0
	ldlp	17
	ldlp	15
	ldlp	10
	call	@P5
	.ldc	1024
	stl	10
	.ldc	0
	stl	11
	ldl	13
	ldl	10
	and
	stl	10
	ldl	14
	ldl	11
	and
	stl	11
	.ldc	0
	stl	8
	.ldc	0
	stl	9
	.ldc	0
	stl	7
	ldl	10
	ldl	8
	diff
	eqc	0
	cj	@L52
	ldl	11
	ldl	9
	diff
	eqc	0
L52:	stl	7
	ldl	7
	cj	@L53
	j	@L55
L53:	.ldc	1023
	stl	10
	.ldc	0
	stl	11
	ldl	13
	ldl	10
	and
	stl	10
	ldl	14
	ldl	11
	and
	stl	11
	.ldc	1
	stl	8
	.ldc	0
	stl	9
	ldl	15
	ldl	8
	and
	stl	8
	ldl	16
	ldl	9
	and
	stl	9
	ldl	10
	ldl	8
	or
	stl	10
	ldl	11
	ldl	9
	or
	stl	11
	.ldc	0
	stl	8
	.ldc	0
	stl	9
	.ldc	0
	stl	7
	ldl	10
	ldl	8
	diff
	eqc	0
	cj	@L54
	ldl	11
	ldl	9
	diff
	eqc	0
L54:	stl	7
	ldl	7
	eqc	0
	cj	@L55
	.ldc	1
	stl	10
	.ldc	0
	stl	11
	.ldc	0
	ldl	10
	ldl	15
	lsum
	stl	15
	ldl	11
	ldl	16
	ladd
	stl	16
L55:	ldlp	15
	ldl	22
	.ldc	8
	move
L57:	_asm_ajw(23);
	ret

P1:	ldl	2
	ldnl	1
	ldl	3
	ldnl	1
	diff
	eqc	0
	cj	@L60
	ldl	2
	ldnl	0
	mint
	xor
	ldl	3
	ldnl	0
	mint
	xor
	gt
	ldl	1
	sb
	j	@L61
L60:	ldl	2
	ldnl	1
	ldl	3
	ldnl	1
	gt
	ldl	1
	sb
L61:	ret

P2:	ajw	-11
	ldl	16
	ldnl	1
	eqc	0
	cj	@L30
	ldl	16
	ldnl	0
	stl	10
	ldl	14
	ldnl	0
	ldl	15
	ldnl	1
	ldl	10
	ldiv
	ldl	12
	stnl	1
	stl	9
	ldl	9
	ldl	15
	ldnl	0
	ldl	10
	ldiv
	ldl	12
	stnl	0
	ldl	13
	stnl	0
	.ldc	0
	ldl	13
	stnl	1
	j	@L31
L30:	ldl	16
	ldnl	1
	ldl	16
	ldnl	0
	norm
	stl	5
	stl	6
	stl	4
	ldl	14
	ldnl	1
	ldl	14
	ldnl	0
	ldl	4
	lshl
	stl	3
	stl	10
	ldl	14
	ldnl	0
	ldl	15
	ldnl	1
	ldl	4
	lshl
	stl	3
	stl	9
	ldl	15
	ldnl	1
	ldl	15
	ldnl	0
	ldl	4
	lshl
	stl	7
	stl	8
	.ldc	2
	stl	2
	ldlp	5
	stl	1
	ldlp	8
	stl	0
	ldlp	9
	ldlp	10
	ldl	12
	ldnlp	1
	call	@P3
	.ldc	2
	stl	2
	ldlp	5
	stl	1
	ldlp	7
	stl	0
	ldlp	8
	ldlp	9
	ldl	12
	call	@P3
	ldl	8
	ldl	7
	ldl	4
	lshr
	ldl	13
	stnl	0
	ldl	13
	stnl	1
L31:	ajw	11
	ret

P3:	ajw	-7
	ldl	12
	ldnl	1
	ldl	9
	ldnl	0
	diff
	eqc	0
	cj	@L25
	.ldc	-1
	stl	6
	ldl	9
	ldnl	0
	ldl	10
	ldnl	0
	sum
	stl	5
	j	@L26
L25:	ldl	9
	ldnl	0
	ldl	10
	ldnl	0
	ldl	12
	ldnl	1
	ldiv
	stl	6
	stl	5
L26:	ldl	6
	cj	@L28
	.ldc	0
	ldl	12
	ldnl	0
	ldl	6
	lmul
	stl	2
	stl	3
	.ldc	0
	ldl	2
	ldl	11
	ldnl	0
	ldiff
	stl	0
	stl	1
	ldl	1
	ldl	3
	ldl	5
	ldiff
	stl	0
	stl	1
	ldl	1
	eqc	0
	cj	@L27
	ldl	6
	.ldc	1
	diff
	stl	6
L27:	ldl	3
	ldl	12
	ldnl	1
	ldl	6
	lmul
	stl	3
	stl	4
	.ldc	0
	ldl	11
	ldnl	0
	ldl	2
	ldiff
	ldl	11
	stnl	0
	stl	1
	ldl	1
	ldl	10
	ldnl	0
	ldl	3
	ldiff
	ldl	10
	stnl	0
	stl	1
	ldl	1
	ldl	9
	ldnl	0
	ldl	4
	ldiff
	ldl	9
	stnl	0
	stl	1
	ldl	1
	cj	@L28
	ldl	6
	.ldc	1
	diff
	stl	6
	.ldc	0
	ldl	12
	ldnl	0
	ldl	11
	ldnl	0
	lsum
	ldl	11
	stnl	0
	stl	1
	ldl	1
	ldl	12
	ldnl	1
	ldl	10
	ldnl	0
	lsum
	ldl	10
	stnl	0
	stl	1
	ldl	1
	.ldc	0
	ldl	9
	ldnl	0
	lsum
	ldl	9
	stnl	0
	stl	1
L28:	ldl	6
	ldl	8
	stnl	0
	ajw	7
	ret

P4:	ajw	-3
	ldl	8
	stl	2
	.ldc	32
	ldl	2
	gt
	cj	@L1
	ldl	6
	ldnl	1
	ldl	6
	ldnl	0
	ldl	2
	lshl
	stl	1
	ldl	4
	stnl	1
	ldl	6
	ldnl	0
	ldl	7
	ldnl	1
	ldl	2
	lshl
	stl	1
	ldl	4
	stnl	0
	ldl	7
	ldnl	1
	ldl	7
	ldnl	0
	ldl	2
	lshl
	ldl	5
	stnl	0
	ldl	5
	stnl	1
	j	@L5
L1:	.ldc	64
	ldl	2
	gt
	cj	@L2
	ldl	2
	adc	-32
	stl	2
	ldl	6
	ldnl	0
	ldl	7
	ldnl	1
	ldl	2
	lshl
	stl	1
	ldl	4
	stnl	1
	ldl	7
	ldnl	1
	ldl	7
	ldnl	0
	ldl	2
	lshl
	ldl	5
	stnl	1
	ldl	4
	stnl	0
	.ldc	0
	ldl	5
	stnl	0
	j	@L5
L2:	.ldc	96
	ldl	2
	gt
	cj	@L3
	ldl	2
	adc	-64
	stl	2
	ldl	7
	ldnl	1
	ldl	7
	ldnl	0
	ldl	2
	lshl
	ldl	4
	stnl	0
	ldl	4
	stnl	1
	.ldc	0
	stl	0
	.ldc	0
	stl	1
	ldlp	0
	ldl	5
	.ldc	8
	move
	j	@L5
L3:	.ldc	128
	ldl	2
	gt
	cj	@L4
	ldl	2
	adc	-96
	stl	2
	ldl	7
	ldnl	0
	ldl	2
	shl
	ldl	4
	stnl	1
	.ldc	0
	ldl	4
	stnl	0
	.ldc	0
	stl	0
	.ldc	0
	stl	1
	ldlp	0
	ldl	5
	.ldc	8
	move
	j	@L5
L4:	.ldc	0
	stl	0
	.ldc	0
	stl	1
	ldlp	0
	ldl	4
	.ldc	8
	move
	.ldc	0
	stl	0
	.ldc	0
	stl	1
	ldlp	0
	ldl	5
	.ldc	8
	move
L5:	ajw	3
	ret

P5:	ajw	-3
	ldl	8
	stl	2
	.ldc	32
	ldl	2
	gt
	cj	@L9
	ldl	7
	ldnl	1
	ldl	7
	ldnl	0
	ldl	2
	lshr
	ldl	5
	stnl	0
	stl	1
	ldl	6
	ldnl	0
	ldl	7
	ldnl	1
	ldl	2
	lshr
	ldl	5
	stnl	1
	stl	1
	ldl	6
	ldnl	1
	ldl	6
	ldnl	0
	ldl	2
	lshr
	ldl	4
	stnl	0
	ldl	4
	stnl	1
	j	@L13
L9:	.ldc	64
	ldl	2
	gt
	cj	@L10
	ldl	2
	adc	-32
	stl	2
	ldl	6
	ldnl	0
	ldl	7
	ldnl	1
	ldl	2
	lshr
	ldl	5
	stnl	0
	stl	1
	ldl	6
	ldnl	1
	ldl	6
	ldnl	0
	ldl	2
	lshr
	ldl	5
	stnl	1
	ldl	4
	stnl	0
	.ldc	0
	ldl	4
	stnl	1
	j	@L13
L10:	.ldc	96
	ldl	2
	gt
	cj	@L11
	ldl	2
	adc	-64
	stl	2
	ldl	6
	ldnl	1
	ldl	6
	ldnl	0
	ldl	2
	lshr
	ldl	5
	stnl	0
	ldl	5
	stnl	1
	.ldc	0
	stl	0
	.ldc	0
	stl	1
	ldlp	0
	ldl	4
	.ldc	8
	move
	j	@L13
L11:	.ldc	128
	ldl	2
	gt
	cj	@L12
	ldl	2
	adc	-96
	stl	2
	ldl	6
	ldnl	1
	ldl	2
	shr
	ldl	5
	stnl	0
	.ldc	0
	ldl	5
	stnl	1
	.ldc	0
	stl	0
	.ldc	0
	stl	1
	ldlp	0
	ldl	4
	.ldc	8
	move
	j	@L13
L12:	.ldc	0
	stl	0
	.ldc	0
	stl	1
	ldlp	0
	ldl	4
	.ldc	8
	move
	.ldc	0
	stl	0
	.ldc	0
	stl	1
	ldlp	0
	ldl	5
	.ldc	8
	move
L13:	ajw	3
	ret

P6:	ajw	-11
	ldl	15
	ldnl	0
	stl	10
	ldl	15
	ldnl	1
	stl	9
	ldl	16
	ldnl	0
	stl	8
	ldl	16
	ldnl	1
	stl	7
	ldl	13
	stl	6
	ldl	13
	ldnlp	1
	stl	5
	ldl	14
	stl	4
	ldl	14
	ldnlp	1
	stl	3
	ldl	9
	cj	@L17
	ldl	9
	ldl	10
	norm
	stl	1
	ldl	5
	stnl	0
	stl	2
	ldl	10
	ldl	7
	ldl	2
	lshl
	stl	1
	ldl	6
	stnl	0
	ldl	7
	ldl	8
	ldl	2
	lshl
	ldl	4
	stnl	0
	ldl	3
	stnl	0
	j	@L21
L17:	ldl	10
	cj	@L18
	ldl	10
	ldl	7
	norm
	stl	1
	ldl	5
	stnl	0
	stl	2
	ldl	7
	ldl	8
	ldl	2
	lshl
	ldl	3
	stnl	0
	ldl	6
	stnl	0
	.ldc	0
	ldl	4
	stnl	0
	ldl	2
	adc	32
	stl	2
	j	@L21
L18:	ldl	7
	cj	@L19
	ldl	7
	ldl	8
	norm
	stl	1
	ldl	5
	stnl	0
	stl	2
	ldl	1
	ldl	6
	stnl	0
	.ldc	0
	stl	0
	.ldc	0
	stl	1
	ldlp	0
	ldl	14
	.ldc	8
	move
	ldl	2
	adc	64
	stl	2
	j	@L21
L19:	ldl	8
	cj	@L20
	ldl	8
	.ldc	0
	norm
	stl	1
	ldl	5
	stnl	0
	stl	2
	ldl	1
	ldl	6
	stnl	0
	.ldc	0
	stl	0
	.ldc	0
	stl	1
	ldlp	0
	ldl	14
	.ldc	8
	move
	ldl	2
	adc	96
	stl	2
	j	@L21
L20:	.ldc	0
	stl	0
	.ldc	0
	stl	1
	ldlp	0
	ldl	13
	.ldc	8
	move
	.ldc	0
	stl	0
	.ldc	0
	stl	1
	ldlp	0
	ldl	14
	.ldc	8
	move
	.ldc	128
	stl	2
L21:	ldl	2
	xdble
	ldl	12
	stnl	0
	ldl	12
	stnl	1
	ajw	11
	ret

P7:	ajw	-2
	ldl	7
	ldnl	0
	stl	0
	ldl	0
	ldl	6
	ldnl	0
	ldl	5
	ldnl	0
	lsum
	ldl	4
	stnl	0
	stl	1
	ldl	1
	ldl	6
	ldnl	1
	ldl	5
	ldnl	1
	lsum
	ldl	4
	stnl	1
	stl	1
	ldl	1
	xdble
	ldl	3
	stnl	0
	ldl	3
	stnl	1
	ajw	2
	ret

P8:	ajw	-7
	.ldc	0
	ldl	11
	ldnl	0
	ldl	10
	ldnl	0
	lmul
	stl	6
	stl	0
	ldl	0
	ldl	11
	ldnl	0
	ldl	10
	ldnl	1
	lmul
	stl	5
	stl	2
	ldl	5
	ldl	11
	ldnl	1
	ldl	10
	ldnl	0
	lmul
	stl	4
	stl	0
	ldl	0
	ldl	11
	ldnl	1
	ldl	10
	ldnl	1
	lmul
	stl	3
	stl	1
	.ldc	0
	ldl	12
	ldnl	0
	ldl	6
	lsum
	ldl	9
	stnl	0
	stl	0
	ldl	0
	ldl	12
	ldnl	1
	ldl	4
	lsum
	ldl	9
	stnl	1
	stl	0
	ldl	0
	ldl	2
	ldl	3
	lsum
	ldl	8
	stnl	0
	stl	0
	ldl	0
	.ldc	0
	ldl	1
	lsum
	ldl	8
	stnl	1
	stl	0
	ajw	7
#endif	/* T800 */
	#pragma	endasm
	}
