/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: switchp.c,v 2.2 92/04/21 10:59:50 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	File:	switchp.c
 *
 *	Implements T800/T414 "ProcToHigh()" and "ProcToLow()".
 *
 *	Copyright 1987, 1988, 1989 by Logical Systems.
 */

#include	<inline.h>
#include	<conc.h>		/* Concurrency includes */

/*
 *	ProcToLow
 *
 *	Change from either low or high priority to low priority.  Returns the
 *	previous priority.
 */
 
int
ProcToLow(void)
	{
	#pragma	asm
	ldpri				;Get current priority
	cj	@l1			;Jump if high prority
	.ldc	1			;Indicate was already low
	ret
l1
	ldlp	0			;Compute WS for runp instruction
	.ldc	1			;Will be low priority
	or
	runp				;Schedule low priority process
	stopp				;Stop high priority process
	.ldc	0			;Was high priority
	#pragma	endasm
	}

/*
 *	ProcToHigh
 *
 *	Change from either low or high priority to high priority.  Returns the
 *	previous priority.  Partially based on ideas from Todd Allen of
 *	CitiCorp/TTI.
 */

/*
 *	The following data structure is used as a small "static" stack for the
 *	residual low priority process generated by the following transistion.
 *	Actually, only "low_pri_stack0" is used (to hold the PC stored by
 *	"stopp"), but this makes the code wordlength independent.
 */
static	int	low_pri_stack0;
static	int	low_pri_stack1;
 
int
ProcToHigh(void)
	{
	#pragma	asm
	ldpri				;Get current priority
	eqc	0
	cj	@h1			;Jump if low priority
	.ldc	0			;Indicate already high priority
	ret
;
;	The meat of the matter, swap stacks so that when the low priority
;	task is resumed (all high priority tasks done), it has somewhere to
;	kill itself.
;
h1
	.ldc	h2			;PC addr for high priority task
	stl	-1			;Store as appropriate for later "runp"
	_asm_eval1(&low_pri_stack1);	;"static" WS for low priority tailings
	gajw				;Swap workspaces
	runp				;Start high priority process
	stopp		 		;Executed by low priority tailings only
h2
	.ldc	1			;Indicate was low priority
	#pragma	endasm
	}
