/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: pv.c,v 2.2 92/04/21 11:01:01 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	pv.c
 *
 *	Semaphore primitives for the Transputer.
 *
 *	Based on the Transputer concurrency function library designed by
 *	Jeffrey Mock of PIXAR (Version 2).  This implementation by Scott Lloyd
 *	of CSA.  Semaphore routines modified to return to original priority by
 *	Bob Eifrig.  Additional routines added by Logical Systems.
 *
 *	Special thanks to Jim Brown (Metron, Inc.), who revised the code to
 *	allow semaphores to be used as "counting semaphores" in addition
 *	to the previously supported "binary semaphores".
 *
 *	Copyright (C) 1987, 1988, 1989 by Computer System Architects and
 *	Logical Systems.
 */

#include	<inline.h>
#include	<conc.h>

/*
 *	High priority "_SemP".  Switches to high priority during operation to
 *	allow use with mixed high and low priority tasks.
 */
int
_HSemP(Semaphore *sem)
	{
	int	priority;
	int	rval;

	priority = ProcToHigh();
	rval = sem->use = (sem->use ? _SemP(sem) : 1);
	if(priority == 1)
		(void) ProcToLow();
	return (rval);
	}

/*
 *	High priority "_SemV".  Switches to high priority during operation to
 *	allow use with mixed high and low priority tasks.
 */
int
_HSemV(Semaphore *sem)
	{
	int	priority;
	int	rval;

	priority = ProcToHigh();
	rval = sem->use = ((sem->use == 1) ? 0 : _SemV(sem));
	if(priority == 1)
		(void) ProcToLow();
	return (rval);
	}

int
_SemP(Semaphore *sem)
	{
	int	pri;			/* priority (must be nwsp[0]) */
	;
	#pragma	asm
	LDL	[sem]			;sem->use++
	LDNL	0
	ADC	1
	LDL	[sem]
	STNL	0

;	LDL	[sem]			;if(sem->use == 1) goto SP3
;	LDNL	0
;	EQC	1
;	EQC	0
;	CJ	@SP3

	LDL	[sem]			;if(sem->use <= 1) goto SP3
	LDNL	0
	LDC	1
	GT
	CJ	@SP3

	LDPRI				;pri = ProcGetPriority()
	STL	[pri]

	LDL	[sem]			;if(sem->use != 2) goto SP1
	LDNL	0
	EQC	2
	CJ	@SP1

	LDLP	[pri]			;sem->head = nwsp
	LDL	[sem]
	STNL	1
	.LDC	0
	CJ	@SP2
SP1
	LDLP	[pri]			;sem->tail[-2] = nwsp
	LDL	[sem]
	LDNL	2
	STNL	-2
SP2
	LDLP	[pri]			;sem->tail = nwsp
	LDL	[sem]
	STNL	2
	STOPP				;Deschedule process
SP3
	LDL	[sem]			;Return sem->use
	LDNL	0
	#pragma endasm
	}

int
_SemV(Semaphore *sem)
	{
	int	wsp;			/* ptr to ws of process to run */
	;
	#pragma asm
	LDL	[sem]			;sem->use--
	LDNL	0
	ADC	-1
	LDL	[sem]
	STNL	0

;	LDL	[sem]			;if(sem->use == 0) goto SV1
;	LDNL	0
;	CJ	@SV1

	LDL	[sem]			;if(sem->use <= 0) goto SV1
	LDNL	0
	.LDC	0
	GT
	CJ	@SV1

	LDL	[sem]			;wsp = sem->head
	LDNL	1
	STL	[wsp]

	LDL	[wsp]			;sem->head = wsp[-2]
	LDNL	-2
	LDL	[sem]
	STNL	1

	LDL	[wsp]			;ProcReady(wsp | was[0])
	LDNL	0
	LDL	[wsp]
	OR
	RUNP
SV1
	LDL	[sem]			;Return sem->use
	LDNL	0
	#pragma endasm
	}
