/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: talloc.c,v 2.2 92/04/21 11:01:02 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	talloc.c
 *
 *	Transputer primitives which use dynamic memory.
 *
 *	Based on the Transputer concurrency function library designed by
 *	Jeffrey Mock of PIXAR (Version 2).  This implementation by Scott Lloyd
 *	of CSA (with some additions by Kirk Bailey of Logical Systems).
 *
 *	Copyright (C) Computer System Architects and Logical Systems, 1989.
 */

#include	<conc.h>
#include	<string.h>
#include	<stdlib.h>

#ifdef	T212
#define	DEFAULTWSSIZE	4096		/* Default workspace size */
#else	/* !T212 */
#define	DEFAULTWSSIZE	65536		/* Default workspace size */
#endif	/* T212 */
#define	MINWORK		5		/* Minimum workspace size (words) */
#define RESERVE		5		/* WS reserve for I/O etc. (words) */

Channel	*
ChanAlloc(void)
	{
	Channel			*c;

	if((c = (Channel *) malloc(sizeof(Channel))) != NULL)
		*c = NOPROCESS;
	return(c);
	}

void
ChanFree(Channel *ptr)
	{
	free(ptr);
	}

Process	*
ProcAlloc(void (*func)(),int sp,int nparam, ...)	/* p1, p2, ..., pN */
	{
	Process			*p;
	int			*wsp;
	int			i;
	int			size;
	int			minsize;
	int			work;

	if((p = (Process *) malloc(sizeof(Process))) == NULL)
		return(NULL);
	size = (sp == 0) ? (DEFAULTWSSIZE / sizeof(int)) :
		(sp + (sizeof(int) - 1)) / sizeof(int);
	work = nparam + 2;
	if(work < MINWORK)
		work = MINWORK;
	minsize = work + RESERVE;
	if(size < minsize)
		size = minsize;
	if((p->ws = malloc(size * sizeof(int))) != NULL)
		{
	        wsp = (int *) (p->ws) + size - work;	/* Workspace ptr */
	        wsp[1] = (int) p;	/* Process descriptor */
		p->pdesc = wsp;
		p->func = func;
		p->wssize = size * sizeof(int);
		p->nparam = nparam;
		memcpy((wsp+2),(((char *) (&nparam)) + sizeof(int)),
			(size_t) (nparam * sizeof(int)));
	        return(p);
		}
	else
		{
		free(p);
		return(NULL);
		}
	}

void
ProcFree(Process *ptr)
	{
	free(ptr->ws);			/* Free associated workspace */
	free(ptr);			/* Free process description block */
	}

Semaphore	*
SemAlloc(void)
	{
	Semaphore		*s;

	if((s = (Semaphore *) malloc(sizeof(Semaphore))) != NULL)
		{
		s->use = 0;
		s->head = (int *) NOPROCESS;
		s->tail = (int *) NOPROCESS;
		}
	return (s);
	}

void
SemFree(Semaphore *ptr)
	{
	if(! ptr->use)
		free(ptr);
	}
