/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: pow64.c,v 2.2 92/04/21 11:01:55 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	File:	pow64.c
 *
 *	Implements T800/T414 64 bit floating point "pow(x,y)".
 *
 *	Based on INMOS supplied code.  Modified by Logical Systems for correct
 *	operation with negative values of "x" and integral values of "y" per
 *	the ANSI "C" specification.  Also modified to correctly set "errno"
 *	for the various ERANGE/EDOM conditions.  This code is more than a
 *	little obscure, good luck!  (After all, "rewriting" the output of
 *	compiler generated machine language is what "real" programmers LOVE
 *	doing when they run out of hard jobs...).
 */

#include	<inline.h>
#include	<math.h>
#include	<stddef.h>

#pragma	asm_expand			/* Recognize EDOM/ERANGE in "asm" */

double
pow(double x,double y)
	{
	#pragma	asm
#ifdef	T800
;
;	T800 64 bit version
;
	.ext	errno
	.ext	_scale
	.ext	_unpk
	_asm_ajw(-31)
	.ldc	A3
	stl	29
	.ldc	A1
	stl	30
	ldlp	[x]
	fpldnldb
	fpnan
	stl	28
	ldlp	[x]
	fpldnldb
	fpnotfinite
	stl	27
	ldlp	[y]
	fpldnldb
	fpnan
	stl	26
	ldlp	[y]
	fpldnldb
	fpnotfinite
	stl	25
	ldl	28
	eqc	0
	cj	@L4
	ldl	26
	cj	@L5
L4:	ldlp	[x]
	ldlp	[y]
	fpldnldb
	fpldnladddb
	ldlp	23
	fpstnldb
	j	@L5B
	.align
A1:	.dw	0x00000000
	.dw	0x3ff00000
	.dw	0xa2a490da
	.dw	0x3feea4af
	.dw	0xdcfba487
	.dw	0x3fed5818
	.dw	0xdd85529c
	.dw	0x3fec199b
	.dw	0x995ad3ad
	.dw	0x3feae89f
	.dw	0x82a3f090
	.dw	0x3fe9c491
	.dw	0x422aa0db
	.dw	0x3fe8ace5
	.dw	0x73eb0187
	.dw	0x3fe7a114
	.dw	0x667f3bcd
	.dw	0x3fe6a09e
	.dw	0xdd485429
	.dw	0x3fe5ab07
	.dw	0xd5362a27
	.dw	0x3fe4bfda
	.dw	0x4c123422
	.dw	0x3fe3dea6
	.dw	0x0a31b715
	.dw	0x3fe306fe
	.dw	0x6e756238
	.dw	0x3fe2387a
	.dw	0x3c7d517b
	.dw	0x3fe172b8
	.dw	0x6cf9890f
	.dw	0x3fe0b558
	.dw	0x00000000
	.dw	0x3fe00000

A2:	.dw	0x179c2893
	.dw	0xbc8e9c23
	.dw	0x895048dd
	.dw	0x3c711065
	.dw	0xb071f2be
	.dw	0x3c6c7c46
	.dw	0xee04992f
	.dw	0xbc741577
	.dw	0x054647ad
	.dw	0x3c86324c
	.dw	0x11f09ebc
	.dw	0x3c7ada09
	.dw	0xb6c70573
	.dw	0x3c89b07e
	.dw	0x4adc610b
	.dw	0x3c88a62e
;
;	The first two floating point constants in the following array are the
;	"undefined" and "unstable" NaN constant values.  The original INMOS
;	code used values for these which are inconsistent with the latest INMOS
;	Compiler Writers' Guide.  The values have been changed to reflect the
;	ones used in the "Guide".
;
A3:	.dw	0x00000000		;Was 0x00000010
	.dw	0x7ff00002		;Was 0x7FF00000
	.dw	0x00000000		;Was 0x00000008
	.dw	0x7ff00001		;Was 0x7FF00000
	.dw	0x00000000		;1.0
	.dw	0x3ff00000
	.dw	0x00000000
	.dw	0x7ff00000
	.dw	0x00000000
	.dw	0xbff00000
	.dw	0xdb4afc29
	.dw	0x3f3c78fd
	.dw	0x2e278dac
	.dw	0x3f624924
	.dw	0x999e080f
	.dw	0x3f899999
	.dw	0x5555554d
	.dw	0x3fb55555
	.dw	0x94ae0bf9
	.dw	0x3fdc551d
	.dw	0x00000000
	.dw	0x40300000
	.dw	0x44868000
	.dw	0x00003fef
	.dw	0x00000000
	.dw	0x3fb00000
	.dw	0xe392cc80
	.dw	0x3eef4edd
	.dw	0xe0384c74
	.dw	0x3f242f7a
	.dw	0x18d7cda0
	.dw	0x3f55d87e
	.dw	0x6e131d99
	.dw	0x3f83b2ab
	.dw	0xd703026d
	.dw	0x3fac6b08
	.dw	0xff82c4cf
	.dw	0x3fcebfbd
	.dw	0xfefa39f0
	.dw	0x3fe62e42
L5:	.ldc	0
	stl	28			;Once held "BaseIsNan", now sign flag
	fpldzerodb
	ldlp	[x]
	fpldnldb
	fpgt
	cj	@L6
;
;	INMOS just called it a domain error if the base < 0.  We need to
;	see if the exponent value was integral before deciding whether to
;	call it a domain error or not.  If the value was integral we compute
;	what the sign of the resulting expression will be and store it in
;	local variable offset 28 (which has been initialized to "positive").
;
	ldlp	[y]			;Point to exponent
	fpldnldb
	fpdup				;Need three copies of exponent
	fpdup
	fpint				;See if integral
	fpeq
	eqc	0
	cj	@L5C			;Jump if so (not domain error time)
L5A:	ldl	29			;Indicate domain error
	dup
	ldnl	0
	stl	23			;Low order word of "undefined" NaN
	ldnl	1
	stl	24			;High order word of "undefined" NaN
L5B:	.ldc	EDOM			;Domain error
	_asm_eval1(&errno);
	stnl	0			;Set errno to EDOM
	j	@L44
L5C:	ldlp	28			;Compute sign of result
	fpstnli32			;Convert to integer (bottom 32 bits)
	ldl	28			;Get integer value (bottom 32 bits)
	.ldc	1
	and
	stl	28			;Save positive/negative status in flag
;
;	We have computed the sign of the result, now negate the "base".
;
	ldlp	[x]			;Point to base
	dup				;A copy of the address
	fpldnldb			;Get base value
	fpuabs				;Make base positive
	fpstnldb			;Update base value
L6:	ldlp	[x]
	fpldnldb
	fpldzerodb
	fpeq
	cj	@L11
	ldlp	[y]
	fpldnldb
	fpldzerodb
	fpgt
	eqc	0
	cj	@L7
	j	@L5A			;Return NaN and set "errno" to EDOM
L7:	ldl	25
	cj	@L8
	ldl	29
	ldnlp	2
	ldlp	23
	.ldc	8
	move
	j	@L5B			;Return NaN and set "errno" to EDOM
L8:	ldlp	[x]
	ldlp	23
	.ldc	8
	move
	j	@L44
L11:	ldl	29
	ldnlp	4
	fpldnldb
	ldlp	[x]
	fpldnldb
	fpgt
	cj	@L14
	ldl	25
	cj	@L14
	ldlp	[y]
	fpldnldb
	fpldzerodb
	fpgt
	cj	@L24A			;Return NaN and set "errno" to ERANGE
	fpldzerodb
	ldlp	23
	fpstnldb
	j	@L44
L14:	ldlp	[x]
	fpldnldb
	ldl	29
	ldnlp	4
	fpldnldb
	fpeq
	cj	@L18
	ldl	25
	cj	@L16
	ldl	29
	ldnlp	2
	ldlp	23
	.ldc	8
	move
	j	@L24B			;Return NaN and set "errno" to ERANGE
L16:	ldl	29
	ldnlp	4
	ldlp	23
	.ldc	8
	move
	j	@L44
L18:	ldl	27
	cj	@L23
	ldl	29
	ldnlp	4
	fpldnldb
	ldlp	[y]
	fpldnldb
	fpgt
	cj	@L24A			;Return NaN and set "errno" to ERANGE
	ldlp	[y]
	fpldnldb
	ldl	29
	ldnlp	8
	fpldnldb
	fpgt
	eqc	0
	cj	@L5A
	fpldzerodb
	ldlp	23
	fpstnldb
	j	@L44
L23:	ldlp	[y]
	fpldnldb
	ldl	29
	ldnlp	4
	fpldnldb
	fpeq
	cj	@L24
	ldlp	[x]
	ldlp	23
	.ldc	8
	move
	j	@L43A
L24:	ldl	25
	cj	@L27
	ldlp	[y]
	fpldnldb
	fpldzerodb
	fpgt
	cj	@L25
L24A:	ldl	29			;Indicate range error
	ldnlp	6
	ldlp	23
	.ldc	8
	move
L24B:	.ldc	ERANGE			;Range error
	_asm_eval1(&errno);
	stnl	0			;Set "errno" to ERANGE
	j	@L44
L25:	fpldzerodb
	ldlp	23
	fpstnldb
	j	@L44
L27:	ldlp	[x]
	ldlp	18
	ldlp	21
	call	@_unpk
	ldlp	18
	fpldnldb
	fpudivby2
	ldlp	18
	fpstnldb
	ldl	21
	adc	1
	stl	21
	.ldc	0
	stl	22
	ldlp	18
	fpldnldb
	ldl	30
	ldnlp	16
	fpldnldb
	fpgt
	eqc	0
	cj	@L28
	.ldc	8
	stl	22
L28:	ldlp	18
	fpldnldb
	ldl	22
	ldl	30
	ldnlp	8
	fpldnldbi
	fpgt
	eqc	0
	cj	@L29
	ldl	22
	adc	4
	stl	22
L29:	ldlp	18
	fpldnldb
	ldl	22
	ldl	30
	ldnlp	4
	fpldnldbi
	fpgt
	eqc	0
	cj	@L30
	ldl	22
	adc	2
	stl	22
L30:	ldl	22
	ldl	30
	ldnlp	2
	wsubdb
	ldlp	8
	.ldc	8
	move
	ldl	22
	.ldc	2
	div
	.ldc	A2
	wsubdb
	ldlp	6
	.ldc	8
	move
	ldlp	8
	ldlp	18
	fpldnldb
	fpldnldb
	fpsub
	ldlp	6
	fpldnldb
	fpsub
	ldlp	18
	ldlp	8
	fpldnldb
	fpldnladddb
	fpdiv
	ldlp	2
	fpstnldb
	ldlp	2
	fpldnldb
	fpumulby2
	ldlp	4
	fpstnldb
	ldlp	4
	ldlp	4
	fpldnldb
	fpldnlmuldb
	ldlp	2
	fpstnldb
	ldl	29
	ldnlp	10
	ldlp	2
	fpldnldb
	fpldnlmuldb
	ldl	29
	ldnlp	12
	fpldnladddb
	ldlp	2
	fpldnlmuldb
	ldl	29
	ldnlp	14
	fpldnladddb
	ldlp	2
	fpldnlmuldb
	ldl	29
	ldnlp	16
	fpldnladddb
	ldlp	2
	fpldnlmuldb
	ldlp	4
	fpldnlmuldb
	ldlp	0
	fpstnldb
	ldl	29
	ldnlp	18
	ldlp	0
	fpldnldb
	fpldnlmuldb
	ldlp	0
	fpldnladddb
	ldlp	4
	ldl	29
	ldnlp	18
	fpldnldb
	fpldnlmuldb
	fpadd
	ldlp	4
	fpldnladddb
	ldlp	14
	fpstnldb
	ldl	21
	.ldc	16
	mul
	ldl	22
	adc	1
	sub
	stl	21
	ldlp	21
	fpi32tor64
	ldl	29
	ldnlp	20
	fpldnldb
	fpdiv
	ldlp	16
	fpstnldb
	ldlp	16
	fpldnldb
	fpur64tor32
	ldlp	14
	fpldnldb
	fpur64tor32
	fpadd
	ldlp	9
	fpstnlsn
	ldlp	9
	fpldnlsn
	fpldzerosn
	fpeq
	cj	@L31
	ldl	29
	ldnlp	6
	ldlp	12
	.ldc	8
	move
	j	@L32
L31:	ldlp	9
	ldl	29
	ldnlp	22
	fpldnlsn
	fpldnlsn
	fpdiv
	fpur32tor64
	ldlp	10
	fpstnldb
	ldlp	10
	fpldnldb
	fpuabs
	ldlp	12
	fpstnldb
L32:	ldlp	[y]
	fpldnldb
	fpuabs
	ldlp	18
	fpstnldb
	ldlp	18
	fpldnldb
	ldlp	12
	fpldnldb
	fpgt
	cj	@L39
	ldlp	[y]
	fpldnldb
	fpldzerodb
	fpgt
	cj	@L35
	ldlp	10
	fpldnldb
	fpldzerodb
	fpgt
	eqc	0
	cj	@L24A			;Set "errno" to ERANGE
	j	@L36
L35:	ldlp	10
	fpldnldb
	fpldzerodb
	fpgt
	cj	@L24A			;Set "errno" to ERANGE
L36:	fpldzerodb
	ldlp	23
	fpstnldb
	j	@L44
L39:	ldlp	[y]
	ldlp	6
	call	@reduce
	ldlp	6
	ldlp	[y]
	fpldnldb
	fpldnldb
	fpsub
	ldlp	4
	fpstnldb
	ldlp	16
	ldlp	4
	fpldnldb
	fpldnlmuldb
	ldlp	14
	ldlp	[y]
	fpldnldb
	fpldnlmuldb
	fpadd
	ldlp	12
	fpstnldb
	ldlp	12
	ldlp	8
	call	@reduce
	ldlp	8
	ldlp	12
	fpldnldb
	fpldnldb
	fpsub
	ldlp	10
	fpstnldb
	ldlp	16
	ldlp	6
	fpldnldb
	fpldnlmuldb
	ldlp	8
	fpldnladddb
	ldlp	12
	fpstnldb
	ldlp	12
	ldlp	8
	call	@reduce
	ldlp	8
	ldlp	12
	fpldnldb
	fpldnldb
	fpsub
	ldlp	10
	fpldnladddb
	ldlp	10
	fpstnldb
	ldlp	10
	ldlp	12
	call	@reduce
	ldlp	12
	ldlp	10
	fpldnldb
	fpldnldb
	fpsub
	ldlp	10
	fpstnldb
	ldlp	8
	ldlp	12
	fpldnldb
	fpldnladddb
	ldl	29
	ldnlp	20
	fpldnlmuldb
	fpurz
	fprtoi32
	ldlp	20
	fpstnli32
	ldl	20
	ldl	29
	ldnl	23
	gt
	eqc	0
	cj	@L24A			;Set "errno" to ERANGE
	ldlp	10
	fpldnldb
	fpldzerodb
	fpgt
	cj	@L41
	ldl	20
	adc	1
	stl	20
	ldl	29
	ldnlp	24
	ldlp	10
	fpldnldb
	fpldnldb
	fpsub
	ldlp	10
	fpstnldb
L41:	.ldc	0
	ldl	20
	gt
	cj	@L42
	ldl	20
	.ldc	16
	div
	stl	21
	j	@L43
L42:	ldl	20
	.ldc	16
	div
	adc	1
	stl	21
L43:	ldl	21
	.ldc	16
	mul
	ldl	20
	sub
	stl	22
	ldl	22
	ldl	30
	wsubdb
	ldlp	6
	.ldc	8
	move
	ldlp	10
	ldl	29
	ldnlp	26
	fpldnldb
	fpldnlmuldb
	ldl	29
	ldnlp	28
	fpldnladddb
	ldlp	10
	fpldnlmuldb
	ldl	29
	ldnlp	30
	fpldnladddb
	ldlp	10
	fpldnlmuldb
	ldl	29
	ldnlp	32
	fpldnladddb
	ldlp	10
	fpldnlmuldb
	ldl	29
	ldnlp	34
	fpldnladddb
	ldlp	10
	fpldnlmuldb
	ldl	29
	ldnlp	36
	fpldnladddb
	ldlp	10
	fpldnlmuldb
	ldl	29
	ldnlp	38
	fpldnladddb
	ldlp	10
	fpldnlmuldb
	ldlp	6
	fpldnlmuldb
	ldlp	6
	fpldnladddb
	ldlp	8
	fpstnldb
	ldl	21
	ldlp	8
	ldlp	23
	call	@_scale
L43A:	ldl	28			;See if we have to negate the result
	cj	@L44			;Jump if not (no negative base stuff)
	fpldzerodb			;Compute -result
	ldlp	23
	fpldnldb
	fpsub
	.ldc	0
	cj	@L45
L44:	ldlp	23
	fpldnldb
L45:	_asm_ajw(31)
	ret
;
;	void reduce(r, x)
;	double *r;
;	double *x;
;
	.align
RDATA:	.dw	0x00000000
	.dw	0x43a00000
	.dw	0x00000000
	.dw	0xc3a00000
	.dw	0x00000000
	.dw	0x40300000
reduce:	ajw	-6
	.ldc	RDATA
	stl	9
	ldl	8
	fpldnldb
	ldl	9
	fpldnldb
	fpgt
	eqc	0
	cj	@L1
	ldl	9
	ldnlp	2
	fpldnldb
	ldl	8
	fpldnldb
	fpgt
	cj	@L2
L1:	ldl	8
	ldl	7
	.ldc	8
	move
	j	@L3
L2:	ldl	9
	ldnlp	4
	ldl	8
	fpldnldb
	fpldnlmuldb
	ldlp	2
	fpstnldb
	ldlp	2
	.ldc	1
	ldlp	0
	call	@dtoi64
	ldlp	0
	.ldc	1
	ldlp	4
	call	@i64tod
	ldl	9
	ldnlp	4
	ldlp	4
	fpldnldb
	fpldnldb
	fpdiv
	ldl	7
	fpstnldb
L3:	ajw	6
	ret
;
;	void dtoi64(i, m, d)
;	int i[2];
;	int m;
;	double *d;
;
dtoi64:	ldl	3
	fpldnldb
	ldl	2
	eqc	0
	cj	@L54
	fpurz
L54:	fpint
	fpuchki64
	ldl	1
	fpdup
	dup
	fpstnli32
	ldnlp	1
	fpuexpdec32
	fpstnli32
	ret
;
;	void i64tod(d, m, i)
;	double *d;
;	int m;
;	int i[2];
;
i64tod:	ldl	3
	dup
	fpb32tor64
	fpdup
	ldnlp	1
	fpi32tor64
	fpuexpinc32
	ldl	2
	eqc	0
	cj	@L55
	fpurz
L55:	fpadd
	ldl	1
	fpstnldb
	ret
#else	/* T800 */
;
;	T414 64 bit version (from D700D).  NO attempt is made here to optimize
;	the OCCAM generated code (hope to get to this someday...)  This code
;	has had the same modifications mentioned for the above T800 code and
;	has also had some helper function calls replaced with inline code, etc.
;	In general a fair bit of difficulty for such a small amount of code!
;
	.ext	errno
	.ext	_fdiv64
	.ext	_fix64
	.ext	_fmul64
	.ext	_FPstate
	.ext	_iexp64
	.ext	_refl64
	.ext	_r64op
	.ext	_r64fei
	.ext	_r64fmi
	.ext	_r64tei
	.ext	_r64toi
	.ext	_r64f32
	.ext	_r64t32
	.ext	_shr64
	_asm_ajw(-37)
	.ldc	0
	stl	36			;Initialize sign toggle to "NO"
	.ldc	A2
	stl	35
	ldl	[x]			;Compute |x|
	stl	31
	ldl	[x]+1
	.ldc	2147483647		;~MOSTNEG
	and
	stl	32
	ldl	[y]			;Compute |y|
	stl	29
	ldl	[y]+1
	.ldc	2147483647		;~MOSTNEG
	and
	stl	30
	ldl	32			;x is NaN or Inf?
	.ldc	2146435072		;DRealInf
	gt
	eqc	0
	cj	@L12
	ldl	32
	.ldc	2146435072		;DRealInf
	diff
	eqc	0
	cj	@L11
	ldl	31
	mint
	xor
	mint
	gt
L11:	eqc	0
L12:	eqc	0
	cj	@L17			;Jump if x is !NaN and !Inf
	ldl	32			;x > y?
	ldl	30
	gt
	eqc	0
	cj	@L14
	ldl	32
	ldl	30
	diff
	eqc	0
	cj	@L13
	ldl	31
	mint
	xor
	ldl	29
	mint
	xor
	gt
L13:	eqc	0
L14:	eqc	0
	cj	@L15
	ldlp	[x]			;x > y, return x NaN/Inf
	ldlp	33
	.ldc	8
	move
	.ldc	EDOM			;Domain error
	_asm_eval1(&errno);
	stnl	0			;Set errno to EDOM
	j	@L16
	.align
A1:	.dw	0x00000000
	.dw	0x40000000
	.dw	0x4921b30b
	.dw	0x3d495f45
	.dw	0xf7490e4c
	.dw	0x3ab031b9
	.dw	0x0aa53844
	.dw	0x383337bb
	.dw	0xb5a75abd
	.dw	0x35d13f32
	.dw	0x47e12039
	.dw	0x33892305
	.dw	0x5541b6b7
	.dw	0x3159ca84
	.dw	0xd6030db0
	.dw	0x2f4228e7
	.dw	0xfe779921
	.dw	0x2d413ccc
	.dw	0x90a852b2
	.dw	0x2b560fba
	.dw	0x6c544e3b
	.dw	0x297fb5aa
	.dw	0x2468446b
	.dw	0x27bd4c98
	.dw	0x636e2a5c
	.dw	0x260dfc14
	.dw	0xeac470ce
	.dw	0x2470f4dc
	.dw	0xfaa2f5ba
	.dw	0x22e57078
	.dw	0xf3121ec5
	.dw	0x216ab0d9
	.dw	0x00000000
	.dw	0x20000000

A2:	.dw	0x00000000		;"undefined" NaN
	.dw	0x7ff00002
	.dw	0x00000000		;"unstable" NaN
	.dw	0x7ff00001
	.dw	0x00000000		;DRealInf
	.dw	0x7ff00000
	.dw	0x00000000		;1.0
	.dw	0x3ff00000
	.dw	0x4cab5145
	.dw	0x00000012
	.dw	0x2b82fe17
	.dw	0x17154765
	.dw	0x7fffffff
	.dw	0x44868000
	.dw	0xd50d31b3
	.dw	0x0004f0c8
	.dw	0x00000000
	.dw	0x00000000
L15:	ldlp	[y]			;y >= x, return y NaN/Inf
	ldlp	33
	.ldc	8
	move
	.ldc	EDOM			;Domain error
	_asm_eval1(&errno);
	stnl	0			;Set errno to EDOM
L16:	j	@L102
L17:	ldl	30			;y is NaN/Inf?
	.ldc	2146435072		;DRealInf
	gt
	eqc	0
	cj	@L19
	ldl	30
	.ldc	2146435072		;DRealInf
	diff
	eqc	0
	cj	@L18
	ldl	29
	mint
	xor
	mint
	gt
L18:	eqc	0
L19:	eqc	0
	cj	@L20			;Jump if y is !NaN and !Inf
	ldlp	[y]			;Return y since is NaN/Inf
	ldlp	33
	.ldc	8
	move
	.ldc	EDOM			;Domain error
	_asm_eval1(&errno);
	stnl	0			;Set errno to EDOM
	j	@L102
L20:	.ldc	0			;x < 0?
	ldl	[x]+1
	gt
	cj	@L21			;Jump if not
;
;	Here we depart in a major way with the original code since we wish to
;	check for a negative base with a integral exponent and produce the
;	correct result.  We do this by seeing if the exponent is integral and
;	then computing the corresponding sign of the result, then storing the
;	fact that the sign of the result needs to be flipped later in local
;	offset #36.  We also make the base value be positive so that the rest
;	of the function will work correctly...
;
	ldlp	[y]
	.ldc	0			;Truncate result
	call	@_r64toi		;Convert to integer
	stl	0			;Store for a bit
	ldl	0
	ldlp	33			;Temp storage
	call	@_r64fmi		;Convert back to double
	ldl	[y]			;Compare for equality
	ldl	33
	diff
	eqc	0
	cj	@L20A			;Jump if not
	ldl	[y]+1
	ldl	34
	diff
	cj	@L20B			;Jump if equal (ie. integral)
L20A:	ldl	35			;If x < 0, return "undefined" NaN
	ldlp	33
	.ldc	8
	move
	.ldc	EDOM			;Domain error
	_asm_eval1(&errno);
	stnl	0			;Set errno to EDOM
	j	@L102
L20B:	ldl	[x]+1			;Make base be positive
	mint
	xor
	stl	[x]+1
	ldl	0			;Integer equivalent of exponent
	.ldc	1
	and
	cj	@L21			;Exponent is even, result will be ok
	mint				;Result sign will need toggling
	stl	36
L21:	ldl	29			;y == Inf?
	ldl	30
	.ldc	2146435072		;DRealInf
	diff
	or
	eqc	0
	cj	@L35			;Jump if y != Inf
	ldl	31			;x == 0.0?
	ldl	32
	or
	cj	@L22			;Jump if so
	ldl	[x]			;x == 1.0?
	ldl	[x]+1
	.ldc	1072693248		;1.0
	diff
	or
	eqc	0
	cj	@L23			;Jump if not
L22:	ldl	35			;Return "unstable" NaN
	ldnlp	2
	ldlp	33
	.ldc	8
	move
	.ldc	ERANGE			;Range error
	_asm_eval1(&errno);
	stnl	0			;Set errno to ERANGE
	j	@L32
L23:	ldl	[y]			;y = +Inf?
	ldl	[y]+1
	.ldc	2146435072		;DRealInf
	diff
	or
	eqc	0
	cj	@L28			;Jump if y != +Inf
	.ldc	1072693248		;1.0
	ldl	[x]+1			;x <= 1.0?
	gt
	eqc	0
	cj	@L25			;Jump if not
	.ldc	1072693248		;1.0
	ldl	[x]+1
	diff
	eqc	0
	cj	@L24			;Jump if x != 1.0
	mint
	ldl	[x]
	mint
	xor
	gt
L24:	eqc	0
L25:	eqc	0
	cj	@L26			;Jump if x < 1.0
	.ldc	0			;Otherwise return 0.0
	stl	33
	.ldc	0
	stl	34
	j	@L27
L26:	ldl	35			;Return DRealInf
	ldnlp	4
	ldlp	33
	.ldc	8
	move
	.ldc	ERANGE			;Range error
	_asm_eval1(&errno);
	stnl	0			;Set errno to ERANGE
L27:	j	@L32
L28:	.ldc	1072693248		;1.0
	ldl	[x]+1			;x <= 1.0?
	gt
	eqc	0
	cj	@L30			;Jump if not (return DRealInf)
	.ldc	1072693248		;1.0
	ldl	[x]+1
	diff
	eqc	0
	cj	@L29			;Jump if x != 1.0
	mint
	ldl	[x]
	mint
	xor
	gt
L29:	eqc	0
L30:	eqc	0
	cj	@L31
	ldl	35			;Return DRealInf
	ldnlp	4
	ldlp	33
	.ldc	8
	move
	.ldc	ERANGE			;Range error
	_asm_eval1(&errno);
	stnl	0			;Set errno to ERANGE
	j	@L32
L31:	.ldc	0			;Return 0.0
	stl	33
	.ldc	0
	stl	34
L32:	j	@L102
L35:	ldl	31			;x == 0.0?
	ldl	32
	or
	eqc	0
	cj	@L40			;Jump if x != 0.0
	ldl	[y]+1			;y <= 0.0?
	.ldc	0
	gt
	eqc	0
	cj	@L37			;Jump if not
	ldl	[y]+1
	eqc	0
	cj	@L36
	ldl	[y]
	mint
	xor
	mint
	gt
L36:	eqc	0
L37:	cj	@L38
	ldl	35			;Return "undefined" NaN
	ldlp	33
	.ldc	8
	move
	.ldc	EDOM			;Domain error
	_asm_eval1(&errno);
	stnl	0			;Set errno to EDOM
	j	@L39
L38:	ldlp	[x]			;Return x
	ldlp	33
	.ldc	8
	move
L39:	j	@L102
L40:	ldl	[x]			;x == 1.0?
	ldl	[x]+1
	.ldc	1072693248		;1.0
	diff
	or
	eqc	0
	cj	@L41			;Jump if x != 1.0
	ldl	35			;Else, return 1.0
	ldnlp	6
	ldlp	33
	.ldc	8
	move
	j	@L102
L41:	ldl	[x]			;x == DRealInf?
	ldl	[x]+1
	.ldc	2146435072		;DRealInf
	diff
	or
	eqc	0
	cj	@L50			;Jump if x != DRealInf
	.ldc	1072693248		;1.0
	ldl	30			;y >= 1.0?
	gt
	eqc	0
	cj	@L43			;Jump if not
	.ldc	1072693248		;1.0
	ldl	30
	diff
	eqc	0
	cj	@L42
	mint
	ldl	29
	mint
	xor
	gt
L42:	eqc	0
L43:	cj	@L48			;Jump if y < 1.0
	ldl	[y]+1			;0.0 >= y?
	.ldc	0
	gt
	eqc	0
	cj	@L45			;Jump if y > 0
	ldl	[y]+1
	eqc	0
	cj	@L44
	ldl	[y]
	mint
	xor
	mint
	gt
L44:	eqc	0
L45:	eqc	0
	cj	@L46
	ldl	35			;Return DRealInf
	ldnlp	4
	ldlp	33
	.ldc	8
	move
	.ldc	ERANGE			;Range error
	_asm_eval1(&errno);
	stnl	0			;Set errno to ERANGE
	j	@L47
L46:	.ldc	0			;Return 0.0
	stl	33
	.ldc	0
	stl	34
L47:	j	@L49
L48:	ldl	35			;Return "undefined" NaN
	ldlp	33
	.ldc	8
	move
	.ldc	ERANGE			;Range error
	_asm_eval1(&errno);
	stnl	0			;Set errno to ERANGE
L49:	j	@L102
L50:	ldl	[y]			;y = 1.0?
	ldl	[y]+1
	.ldc	1072693248		;1.0
	diff
	or
	eqc	0
	cj	@L51			;Jump if not
	ldlp	[x]			;Return x
	ldlp	33
	.ldc	8
	move
	j	@L101A
L51:	ldl	32
	ldl	31
	.ldc	52
	lshr
	stl	15
	stl	16
	ldl	15
	stl	27
	ldl	31
	stl	25
	ldl	32
	.ldc	1048575
	and
	stl	26
	ldl	27
	eqc	0
	cj	@L52
	ldlp	25
	stl	16
	ldlp	31
	stl	15
	ldl	15
	ldnl	1
	ldl	15
	ldnl	0
	norm
	rev
	stl	14
	rev
	stl	13
	ldl	16
	stnl	0
	ldl	14
	ldl	16
	stnl	1
	ldl	13
	stl	27
	.ldc	12
	ldl	27
	diff
	stl	27
	ldl	16
	ldnl	1
	ldl	16
	ldnl	0
	.ldc	2
	lshr
	ldl	16
	stnl	0
	ldl	16
	stnl	1
	j	@L53
L52:	ldl	25
	stl	13
	ldl	26
	.ldc	1048576
	or
	stl	14
	ldl	14
	ldl	13
	.ldc	9
	lshl
	stl	15
	stl	16
	ldl	15
	stl	25
	ldl	16
	stl	26
L53:	ldl	27
	adc	-1022
	stl	27
	ldl	26
	.ldc	759250124
	gt
	eqc	0
	cj	@L55
	ldl	26
	.ldc	759250124
	diff
	eqc	0
	cj	@L54
	ldl	25
	mint
	xor
	.ldc	2121767201
	gt
L54:	eqc	0
L55:	cj	@L56
	.ldc	8
	stl	28
	j	@L57
L56:	.ldc	0
	stl	28
L57:	ldl	28
	.ldc	2
	prod
	.ldc	A1
	wsub
	ldnlp	8
	ldlp	15
	.ldc	8
	move
	ldl	26
	ldl	16
	gt
	eqc	0
	cj	@L59
	ldl	26
	ldl	16
	diff
	eqc	0
	cj	@L58
	ldl	25
	mint
	xor
	ldl	15
	mint
	xor
	gt
L58:	eqc	0
L59:	cj	@L60
	ldl	28
	adc	4
	stl	28
L60:	ldl	28
	.ldc	2
	prod
	.ldc	A1
	wsub
	ldnlp	4
	ldlp	15
	.ldc	8
	move
	ldl	26
	ldl	16
	gt
	eqc	0
	cj	@L62
	ldl	26
	ldl	16
	diff
	eqc	0
	cj	@L61
	ldl	25
	mint
	xor
	ldl	15
	mint
	xor
	gt
L61:	eqc	0
L62:	cj	@L63
	ldl	28
	adc	3
	stl	28
	j	@L64
L63:	ldl	28
	adc	1
	stl	28
L64:	ldl	28
	.ldc	2
	prod
	.ldc	A1
	wsub
	ldlp	15
	.ldc	8
	move
	.ldc	0
	ldl	25
	ldl	15
	ldiff
	stl	9
	ldl	26
	ldl	16
	lsub
	stl	10
	ldl	10
	ldl	9
	.ldc	4
	lshl
	stl	7
	stl	8
	ldlp	7
	stl	6
	.ldc	0
	ldl	25
	ldl	15
	lsum
	stl	9
	ldl	26
	ldl	16
	ladd
	stl	10
	ldlp	9
	ldl	6
	ldlp	13
	call	@_fdiv64
	ldlp	13
	ldlp	13
	ldlp	11
	call	@_fmul64
	ldl	35
	ldnlp	8
	ldlp	11
	ldlp	9
	call	@_fmul64
	.ldc	0
	ldl	9
	.ldc	-1722569274
	lsum
	stl	5
	ldl	10
	.ldc	6553
	ladd
	stl	6
	ldlp	5
	ldlp	11
	ldlp	7
	call	@_fmul64
	.ldc	0
	ldl	7
	.ldc	-1431542164
	lsum
	stl	5
	ldl	8
	.ldc	2796202
	ladd
	stl	6
	ldlp	5
	ldlp	11
	ldlp	9
	call	@_fmul64
	ldlp	13
	ldlp	9
	ldlp	7
	call	@_fmul64
	.ldc	0
	ldl	13
	ldl	7
	lsum
	stl	5
	ldl	14
	ldl	8
	ladd
	stl	6
	ldlp	5
	ldl	35
	ldnlp	10
	ldlp	9
	call	@_fmul64
	ldlp	9
	ldlp	21
	call	@_refl64
	ldl	27
	.ldc	16
	prod
	ldl	28
	diff
	stl	27
	ldl	27
	eqc	0
	cj	@L65
	.ldc	0
	stl	19
	.ldc	0
	stl	20
	j	@L66
L65:	ldl	27
	ldlp	15
	call	@_r64fmi
	.ldc	-4
	ldlp	15
	ldlp	19
	call	@_iexp64
L66:	ldlp	19
	.ldc	1
	call	@_r64t32
	stl	15
	ldlp	21
	.ldc	1
	call	@_r64t32
	.ldc	0
	ldl	15
	call	@P2
	stl	16
	ldl	16
	stl	15
	ldl	15
	ldl	35
	ldnl	12
	and
	eqc	0
	cj	@L67
	ldl	35
	ldnlp	4
	ldlp	17
	.ldc	8
	move
	j	@L68
L67:	ldl	16
	.ldc	3
	ldl	35
	ldnl	13
	call	@P2
	ldlp	17
	call	@_r64f32
L68:	ldlp	17
	stl	16
	ldl	16
	ldnl	0
	stl	14
	ldl	16
	ldnl	1
	.ldc	2147483647
	and
	stl	15
	ldl	30
	ldl	15
	gt
	eqc	0
	cj	@L70
	ldl	30
	ldl	15
	diff
	eqc	0
	cj	@L69
	ldl	29
	mint
	xor
	ldl	14
	mint
	xor
	gt
L69:	eqc	0
L70:	eqc	0
	cj	@L83
	ldl	[y]+1
	.ldc	0
	gt
	eqc	0
	cj	@L72
	ldl	[y]+1
	eqc	0
	cj	@L71
	ldl	[y]
	mint
	xor
	mint
	gt
L71:	eqc	0
L72:	eqc	0
	cj	@L77
	ldl	16
	ldnl	1
	.ldc	0
	gt
	eqc	0
	cj	@L74
	ldl	16
	ldnl	1
	eqc	0
	cj	@L73
	ldl	16
	ldnl	0
	mint
	xor
	mint
	gt
L73:	eqc	0
L74:	eqc	0
	cj	@L75
	ldl	35
	ldnlp	4
	ldlp	33
	.ldc	8
	move
	j	@L76
L75:	.ldc	0
	stl	33
	.ldc	0
	stl	34
L76:	j	@L81
L77:	ldl	16
	ldnl	1
	.ldc	0
	gt
	eqc	0
	cj	@L79
	ldl	16
	ldnl	1
	eqc	0
	cj	@L78
	ldl	16
	ldnl	0
	mint
	xor
	mint
	gt
L78:	eqc	0
L79:	eqc	0
	cj	@L80
	.ldc	0
	stl	33
	.ldc	0
	stl	34
	j	@L81
L80:	ldl	35
	ldnlp	4
	ldlp	33
	.ldc	8
	move
	.ldc	ERANGE			;Range error
	_asm_eval1(&errno);
	stnl	0			;Set errno to ERANGE
L81:	j	@L102
L83:	ldlp	[y]
	ldlp	11
	call	@P1
	ldlp	11
	stl	0
	.ldc	1
	ldlp	[y]
	ldlp	9
	call	@_r64op
	ldlp	9
	stl	0
	.ldc	2
	ldlp	19
	ldlp	7
	call	@_r64op
	ldlp	[y]
	stl	0
	.ldc	2
	ldlp	21
	ldlp	5
	call	@_r64op
	ldlp	5
	stl	0
	.ldc	0
	ldlp	7
	ldlp	13
	call	@_r64op
	ldlp	13
	ldlp	15
	call	@P1
	ldlp	15
	stl	0
	.ldc	1
	ldlp	13
	ldlp	17
	call	@_r64op
	ldlp	11
	stl	0
	.ldc	2
	ldlp	19
	ldlp	7
	call	@_r64op
	ldlp	7
	stl	0
	.ldc	0
	ldlp	15
	ldlp	13
	call	@_r64op
	ldlp	13
	ldlp	15
	call	@P1
	ldlp	15
	stl	0
	.ldc	1
	ldlp	13
	ldlp	9
	call	@_r64op
	ldlp	9
	stl	0
	.ldc	0
	ldlp	17
	ldlp	11
	call	@_r64op
	ldl	11
	stl	17
	ldl	12
	stl	18
	ldlp	17
	ldlp	13
	call	@P1
	ldlp	13
	stl	0
	.ldc	0
	ldlp	15
	ldlp	11
	call	@_r64op
	ldl	11
	stl	15
	ldl	12
	stl	16
	ldlp	15
	stl	12
	ldl	12
	ldnl	0
	ldl	12
	ldnl	1
	or
	eqc	0
	cj	@L84
	.ldc	0
	stl	23
	.ldc	0
	stl	24
	j	@L85
L84:	.ldc	4
	ldlp	15
	ldlp	11
	call	@_iexp64
	ldlp	11
	.ldc	0
	ldlp	23
	call	@_r64tei
L85:	ldlp	13
	stl	0
	.ldc	1
	ldlp	17
	ldlp	11
	call	@_r64op
	ldl	11
	stl	17
	ldl	12
	stl	18
	ldl	24
	.ldc	0
	gt
	eqc	0
	cj	@L87
	ldl	24
	eqc	0
	cj	@L86
	ldl	23
	mint
	xor
	.ldc	-2147467280
	gt
L86:	eqc	0
L87:	eqc	0
	cj	@L88
	ldl	35
	ldnlp	4
	ldlp	33
	.ldc	8
	move
	.ldc	ERANGE			;Range error
	_asm_eval1(&errno);
	stnl	0			;Set errno to ERANGE
	j	@L102
L88:	ldlp	17
	ldlp	15
	call	@_fix64
	ldl	16
	.ldc	0
	gt
	eqc	0
	cj	@L90
	ldl	16
	eqc	0
	cj	@L89
	ldl	15
	mint
	xor
	mint
	gt
L89:	eqc	0
L90:	eqc	0
	cj	@L91
	.ldc	0
	ldl	23
	.ldc	1
	lsum
	stl	13
	ldl	24
	.ldc	0
	ladd
	stl	14
	ldl	13
	stl	23
	ldl	14
	stl	24
	ldl	15
	stl	13
	.ldc	0
	ldl	16
	.ldc	134217728
	lsub
	stl	14
	ldl	13
	stl	15
	ldl	14
	stl	16
L91:	ldl	23
	stl	14
	.ldc	0
	ldl	14
	gt
	cj	@L92
	ldl	14
	.ldc	16
	div
	stl	27
	j	@L93
L92:	ldl	14
	.ldc	16
	div
	adc	1
	stl	27
L93:	ldl	27
	.ldc	16
	prod
	ldl	14
	diff
	stl	28
	ldl	27
	adc	1
	stl	27
	ldl	28
	.ldc	2
	prod
	.ldc	A1
	wsub
	ldlp	13
	.ldc	8
	move
	ldl	35
	ldnlp	14
	ldlp	15
	ldlp	11
	call	@_fmul64
	.ldc	0
	ldl	11
	.ldc	-1403106907
	lsum
	stl	7
	ldl	12
	.ldc	2862769
	ladd
	stl	8
	ldlp	7
	ldlp	15
	ldlp	9
	call	@_fmul64
	.ldc	0
	ldl	9
	.ldc	2004863291
	lsum
	stl	7
	ldl	10
	.ldc	20654750
	ladd
	stl	8
	ldlp	7
	ldlp	15
	ldlp	11
	call	@_fmul64
	.ldc	0
	ldl	11
	.ldc	1026430217
	lsum
	stl	7
	ldl	12
	.ldc	119194165
	ladd
	stl	8
	ldlp	7
	ldlp	15
	ldlp	9
	call	@_fmul64
	.ldc	0
	ldl	9
	.ldc	-2122409752
	lsum
	stl	7
	ldl	10
	.ldc	515882495
	ladd
	stl	8
	ldlp	7
	ldlp	15
	ldlp	11
	call	@_fmul64
	.ldc	0
	ldl	11
	.ldc	-387535150
	lsum
	stl	7
	ldl	12
	.ldc	1488522235
	ladd
	stl	8
	ldlp	7
	ldlp	15
	ldlp	9
	call	@_fmul64
	ldlp	13
	ldlp	9
	ldlp	11
	call	@_fmul64
	.ldc	0
	ldl	11
	ldl	13
	lsum
	stl	9
	ldl	12
	ldl	14
	ladd
	stl	10
	ldlp	9
	ldlp	17
	call	@_refl64
	ldlp	17
	stl	14
	ldlp	33
	stl	11
	ldl	14
	ldnl	1
	ldl	14
	ldnl	0
	.ldc	52
	lshr
	stl	9
	stl	10
	ldl	27
	xdble
	stl	7
	stl	8
	.ldc	0
	ldl	9
	ldl	7
	lsum
	stl	12
	ldl	10
	ldl	8
	ladd
	stl	13
	ldl	13
	.ldc	0
	gt
	eqc	0
	cj	@L95
	ldl	13
	eqc	0
	cj	@L94
	ldl	12
	mint
	xor
	mint
	gt
L94:	eqc	0
L95:	cj	@L101
	.ldc	1
	ldl	12
	diff
	stl	10
	ldl	35
	ldnlp	16
	stl	0
	ldl	10
	stl	1
	ldl	14
	ldnl	0
	stl	4
	ldl	14
	ldnl	1
	.ldc	1048575
	and
	stl	5
	ldl	4
	stl	2
	ldl	5
	.ldc	1048576
	or
	stl	3
	ldlp	2
	ldlp	8
	ldlp	6
	call	@_shr64
	ldlp	6
	ldl	14
	.ldc	8
	move
	.ldc	0
	ldl	9
	gt
	eqc	0
	cj	@L96
	ldl	14
	ldl	11
	.ldc	8
	move
	j	@L99
L96:	ldl	8
	ldl	9
	mint
	diff
	or
	eqc	0
	cj	@L97
	ldl	14
	ldnl	0
	.ldc	1
	and
	stl	6
	.ldc	0
	stl	7
	ldl	6
	.ldc	1
	diff
	ldl	7
	or
	eqc	0
	cj	@L98
L97:	.ldc	0
	ldl	14
	ldnl	0
	.ldc	1
	lsum
	ldl	11
	stnl	0
	ldl	14
	ldnl	1
	.ldc	0
	ladd
	ldl	11
	stnl	1
	j	@L99
L98:	ldl	14
	ldl	11
	.ldc	8
	move
L99:	j	@L101A
L101:	ldl	14
	ldnl	0
	stl	9
	ldl	14
	ldnl	1
	.ldc	1048575
	and
	stl	10
	ldl	13
	ldl	12
	.ldc	52
	lshl
	stl	7
	stl	8
	ldl	9
	ldl	7
	or
	ldl	11
	stnl	0
	ldl	10
	ldl	8
	or
	ldl	11
	stnl	1
L101A:	
	ldl	36			;Should result be complemented?
	cj	@L102			;Jump if not
	ldl	34			;Toggle sign bit
	ldl	36
	xor
	stl	34
L102:	ldlp	33
	.ldc	_FPstate		;Addr of "soft" FAreg
	.ldc	8
	move
	_asm_ajw(37)
	ret

P1:	ajw	-12
	ldl	14
	ldlp	8
	.ldc	8
	move
	ldl	8
	stl	6
	ldl	9
	.ldc	2147483647
	and
	stl	7
	ldl	7
	.ldc	1125122048
	gt
	eqc	0
	cj	@L2
	ldl	7
	.ldc	1125122048
	diff
	eqc	0
	cj	@L1
	ldl	6
	mint
	xor
	mint
	gt
L1:	eqc	0
L2:	eqc	0
	cj	@L3
	ldl	8
	stl	4
	ldl	9
	stl	5
	ldl	4
	stl	10
	ldl	5
	stl	11
	j	@L8
L3:	.ldc	1067450368
	ldl	7
	gt
	eqc	0
	cj	@L5
	.ldc	1067450368
	ldl	7
	diff
	eqc	0
	cj	@L4
	mint
	ldl	6
	mint
	xor
	gt
L4:	eqc	0
L5:	eqc	0
	cj	@L6
	.ldc	0
	stl	10
	.ldc	0
	stl	11
	j	@L8
L6:	ldl	8
	stl	4
	ldl	9
	stl	5
	.ldc	4
	ldlp	4
	ldlp	2
	call	@_iexp64
	ldlp	2
	.ldc	1
	ldlp	0
	call	@_r64tei
	ldlp	0
	.ldc	1
	ldlp	10
	call	@_r64fei
	ldlp	10
	stl	5
	ldl	5
	ldnl	0
	ldl	5
	ldnl	1
	or
	eqc	0
	cj	@L7
	j	@L8
L7:	.ldc	-4
	ldlp	10
	ldlp	4
	call	@_iexp64
	ldl	4
	stl	10
	ldl	5
	stl	11
L8:	ldlp	10
	ldl	13
	.ldc	8
	move
	ajw	12
	ret

	.align
data:	.dw	N7,	N7,	N7,	N2
	.dw	N7,	N7,	N13,	N6
	.dw	N7,	N4,	N6,	N6
	.dw	N5,	N6,	N6,	N6
	.dw	N7,	N7,	N7,	N2
	.dw	N7,	N7,	N13,	N6
	.dw	N7,	N4,	N6,	N6
	.dw	N5,	N6,	N6,	N6
	.dw	N7,	N12,	N8,	N8
	.dw	N7,	N9,	N1,	N8
	.dw	N7,	N9,	N9,	N12
	.dw	N5,	N6,	N6,	N6
	.dw	N7,	N9,	N9,	N10
	.dw	N7,	N9,	N3,	N8
	.dw	N7,	N11,	N8,	N8
	.dw	N5,	N6,	N6,	N6
P2:	ajw	-9
	ldl	11
	ldl	12
	unpacksn
	stl	6
	stl	5
	ldl	10
	unpacksn
	stl	7
	stl	0
	.ldc	data
	wsub
	ldnl	0
	gcall
N1:	ldl	0
	ldl	5
	add
	adc	-126
	stl	0
	.ldc	0
	ldl	7
	ldl	6
	lmul
	norm
	postnormsn
	roundsn
	ldl	10
	ldl	12
	xor
	mint
	and
	or
	ajw	9
	ret
N2:	ldl	11
	eqc	1
	cj	@LL1
	mint
LL1:	ldl	12
	xor
	ldl	10
	and
	mint
	and
	ajw	9
	ret
N3:	ldl	7
	.ldc	0
	norm
	stl	3
	stl	7
	stl	2
	ldl	6
	.ldc	0
	norm
	stl	3
	stl	6
	adc	127
	ldl	0
	ldl	5
	sub
	add
	ldl	2
	sub
	stl	0
	ldl	7
	.ldc	1
	shr
	.ldc	0
	ldl	6
	ldiv
	rev
	norm
	postnormsn
	roundsn
	ldl	10
	ldl	12
	xor
	mint
	and
	or
	ajw	9
	ret
N4:	ldl	11
	eqc	1
	cj	@LL2
	mint
LL2:	ldl	12
	xor
	ldl	10
	xor
	mint
	and
	cj	@LL5
	.ldc	524288
	stl	1
	j	@LL10
N5:	ldl	10
	ldl	6
	ldl	7
	gt
	cj	@LL7
N6:
LL3:	ldl	12
	ldl	11
	eqc	1
	cj	@LL4
	mint
	xor
	ajw	9
	ret
LL4:	cj	@LL7
N7:
LL5:	ldl	10
	.ldc	0
LL7:	rev
	ldl	11
	.ldc	1
	gt
	cj	@LL9
	mint
	not
	and
	ldl	10
	ldl	12
	xor
	mint
	and
LL9:	or
	ajw	9
	ret
N8:	ldl	10
	ldl	12
	xor
	mint
	and
	ajw	9
	ret
N9:	ldl	10
	ldl	12
	xor
	mint
	and
	stl	1
	j	@LL10
N10:	.ldc	4194304
	stl	1
	j	@LL10
N11:	.ldc	2097152
	stl	1
	j	@LL10
N12:	.ldc	1048576
	stl	1
LL10:	ldl	1
	.ldc	2139095040
	or
	ajw	9
	ret
N13:	ldl	0
	ldl	5
	diff
	stl	4
	ldl	4
	.ldc	0
	gt
	cj	@LL14
	.ldc	26
	ldl	4
	gt
	cj	@LL5
	ldl	6
	.ldc	0
	ldl	4
	lshr
	stl	3
	stl	6
	j	@LL15
LL14:	ldl	4
	.ldc	-26
	gt
	cj	@LL3
	ldl	7
	.ldc	0
	ldl	4
	sub
	.ldc	0
	rev
	lshr
	stl	3
	stl	7
	ldl	5
	stl	0
LL15:	ldl	11
	eqc	1
	cj	@LL16
	mint
LL16:	ldl	12
	xor
	ldl	10
	xor
	mint
	and
	cj	@LL20
	.ldc	0
	.ldc	0
	ldl	3
	ldiff
	stl	3
	stl	2
	ldl	2
	ldl	7
	ldl	6
	ldiff
	stl	4
	cj	@LL18
	ldl	2
	ldl	6
	ldl	7
	ldiff
	ldl	3
	norm
	postnormsn
	roundsn
	ldl	11
	eqc	1
	cj	@LL17
	mint
LL17:	ldl	12
	xor
	mint
	and
	or
	ajw	9
	ret
LL18:	ldl	4
	cj	@LL19
	ldl	4
	ldl	3
	norm
	postnormsn
	roundsn
	ldl	10
	mint
	and
	or
LL19:	ajw	9
	ret
LL20:	ldl	6
	ldl	7
	lsum
	stl	7
	cj	@LL21
	ldl	0
	adc	1
	stl	0
	ldl	7
	ldl	3
	.ldc	1
	lshr
	stl	3
	mint
	or
	stl	7
	j	@LL22
LL21:	ldl	0
	eqc	1
	cj	@LL22
	ldl	7
	mint
	lsum
	rev
	stl	0
LL22:	ldl	0
	ldl	7
	ldl	3
	roundsn
	ldl	10
	mint
	and
	or
	ajw	9
	ret
#endif	/* T800 */
	#pragma	endasm
	}
