/*
 *	OSU Research Computing					Trillium
 *	(public domain software)
 *
 *	$Log:	fopen.c,v $
 * Revision 2.2  92/04/21  11:03:02  trillium
 * Ohio Release
 * 
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 * This file is originally from dLibs 1.2 by Dale Schumacher et al.
 * It was modified for Trollius by MB, 88/12.
 * MB: added call to atkexit() 891012
 * It was modified again for portability by GDB, 91/08.
 */

#include <fcntl.h>
#include <stdio.h>

#define isiodev(d) ((d)<3)	/* instead of isatty(d) */
#define isstdout(d) ((d)==1)
#define isstderr(d) ((d)==2)

/*
 * external variables
 */
extern int		errno;
extern int		_ioexit;

/*
 * external functions
 */
extern void		fcloseall();

/*
 * INTERNAL FUNCTION.  Attempt to open <filename> in the given
 * <mode> and attach it to the stream <fp>
 */
static FILE *
_fopen(filename, mode, fp)
	char *filename;
	char *mode;
	FILE *fp;
{
	int h, i, iomode = 0, f = 0;

	if (! _ioexit) {
		if (atkexit (fcloseall, (char *) 0) != 0)
			return (NULL);
		_ioexit = 1;
	}

	/*
	 * The CRLF --> \n translation left in fgetc/fputc in case we ever go
	 * back to MS-DOS.  To compensate we default to binary mode here.
	 */
	f |= _IOBIN;

	while(*mode) {
		switch(*mode++) {
			case 'r':
				f |= _IOREAD;
				break;
			case 'w':
				f |= _IOWRT;
				iomode |= (O_CREAT | O_TRUNC);
				break;
			case 'a':
				f |= _IOWRT;
				iomode |= (O_CREAT | O_APPEND);
				break;
			case '+':
				f |= (_IOREAD | _IOWRT);
				break;
			case 'b':
				f |= _IOBIN;
				break;
			case 't':
				f &= ~_IOBIN;
				break;
			default:
				return(NULL);
		}
	}

	if((i = (f & (_IOREAD | _IOWRT))) == 0)
		return(NULL);
	else if(i == _IOREAD)
		iomode |= O_RDONLY;
	else if(i == _IOWRT)
		iomode |= O_WRONLY;
	else
		iomode |= O_RDWR;
	h = open(filename, iomode, 0644);
	if(h < 0) {
		errno = h;
		return(NULL);		/* file open/create error */
	}
	if(isiodev(h)) {
		f |= _IODEV;
		if (isstderr(h))
			f |= _IONBF;
		else if (isstdout(h))
			f |= _IOLBF;
		else
			f |= _IOFBF;
	} else
		f |= _IOFBF;
	fp->_file = h;			/* file handle */
	fp->_flag = f;			/* file status flags */
	fp->_base = NULL;		/* base of file buffer */
	fp->_ptr = NULL;		/* current buffer pointer */
	fp->_bsiz = 0;			/* buffer size */
	fp->_cnt = 0;			/* # of bytes in buffer */
	return(fp);
}

FILE *
fopen(filename, mode)
	char *filename, *mode;
{
	int i;
	FILE *fp = NULL;

	for(i=0; (!fp && (i < _NFILE)); ++i)
		if(!(_iob[i]._flag & (_IOREAD | _IOWRT)))   /* empty slot? */
			fp = &_iob[i];
	if(fp)
		return(_fopen(filename, mode, fp));
	return(NULL);
}

FILE *
freopen(filename, mode, fp)
	char *filename, *mode;
	FILE *fp;
{
	if(fclose(fp))
		return(NULL);
	return(_fopen(filename, mode, fp));
}
