/*
 *	OSU Research Computing					Trillium
 *	(public domain software)
 *
 *	$Log:	stdio.h,v $
 * Revision 2.2  92/04/21  11:07:08  trillium
 * Ohio Release
 * 
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 * stdio.h - standard I/O include file
 */

/*
 * This file is originally from dLibs 1.2 by Dale Schumacher et al.
 * It was modified for Trollius by MB, 88/12.
 * It was modified again for portability by GDB, 91/08.
 */

#ifndef _STDIO
#define	_STDIO

/*
 * constants
 */
#define	_NFILE		(16)		/* maximum number of open streams */
#define	OPEN_MAX	_NFILE		/* ANSI equivalent (replaces _NFILE) */
#define	BUFSIZ		(1024)		/* default buffer size */
#define	EOF		(-1)		/* end-of-file indicator */
#define	EOS		('\0')		/* end-of-string indicator */
#define	NULL		(0)		/* null FILE stream ptr */

/*
 * lseek() origins
 */
#ifndef SEEK_SET
#define	SEEK_SET	0		/* from beginning of file */
#define	SEEK_CUR	1		/* from current location */
#define	SEEK_END	2		/* from end of file */
#endif

/*
 * FILE structure flags
 */
#define	_IOREAD		0x0001		/* file may be read from */
#define	_IOWRT		0x0002		/* file may be written to */
#define	_IOBIN		0x0004		/* file is in "binary" mode */
#define	_IODEV		0x0008		/* file is a character device */
#define	_IORW		0x0080		/* last i/o was 0:read/1:write */
#define	_IOFBF		0x0100		/* i/o is fully buffered */
#define	_IOLBF		0x0100		/* i/o is line buffered */
#define	_IONBF		0x0400		/* i/o is not buffered */
#define	_IOMYBUF	0x0800		/* standard buffer */
#define	_IOEOF		0x1000		/* EOF has been reached */
#define	_IOERR		0x4000		/* an error has occured */

/*
 * FILE structure
 */
typedef	struct
{
	int		_cnt;		/* # of bytes in buffer */
	unsigned char	*_ptr;		/* current buffer pointer */
	unsigned char	*_base;		/* base of file buffer */
	unsigned int	_flag;		/* file status flags */
	int		_file;		/* file handle */
	int		_bsiz;		/* buffer size */
	unsigned char	_ch;		/* tiny buffer, for "unbuffered" i/o */
} FILE;

/*
 * external functions
 */
extern FILE	*fopen(), *fdopen(), *freopen();
extern long	ftell();
extern void	rewind(), setbuf(), setbuffer();
extern char	*fgets(), *gets();
extern int	fclose(), fflush(), fgetc(), fputc(), fputs(), fseek(),
		fungetc(), puts(), setlinebuf(), setvbuf();

/*
 * standard streams
 */
#define stdin	(&_iob[0])
#define stdout	(&_iob[1])
#define stderr	(&_iob[2])
#define stdprn	(&_iob[3])
#define stdaux	(&_iob[4])

/*
 * external variables
 */
extern FILE	_iob[];

/*
 * stream macros
 */
#define clearerr(fp)	((void) ((fp)->_flag &= ~(_IOERR|_IOEOF)))
#define feof(fp)	((fp)->_flag & _IOEOF)
#define ferror(fp)	((fp)->_flag & _IOERR)
#define fileno(fp)	((fp)->_file)

/*
 * aliases
 */
#define	getc			fgetc
#define	putc			fputc
#define	getchar()		fgetc(stdin)
#define	putchar(c)		fputc((c),stdout)
#define	ungetc			fungetc
#define	ungetchar(c)		fungetc((c),stdin)

#endif _STDIO
