/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: tasm1.c,v 2.2 92/04/21 11:10:13 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/************************************************************************/
/*									*/
/*	File:	tasm1.c							*/
/*									*/
/*	Transputer Assembler (TASM), source file 1.  This file contains	*/
/*	the top level code for the assembler (command line parsing and	*/
/*	general sequencing of the assembly process).			*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1989 by Logical Systems.		*/
/*									*/
/*				09/01/86				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard I/O include file */
#include	<stdlib.h>		/* Standard library include file */
#include	<ctype.h>		/* Character type macros */
#include	<string.h>		/* String functions */
#include	"taldef.h"		/* Assembler/Linker definitions */
#include	"tasmdef.h"		/* Assembler definitions */
#include	"tasmtyp.h"		/* Assembler data declarations */
#include	"tasmext.h"		/* Assembler externals */

/************************************************************************/
/*									*/
/*			Global Data Definitions				*/
/*									*/
/************************************************************************/

/*
 *	I/O definitions.
 */

char	ifnbuf[FNSIZE + 1] =	{ 0 };	/* Input filename buffer */
char	lifnbuf[FNSIZE + 1] =	{ 0 };	/* Last filename in a "#line" stmt */
STREAM	in_fp		=	NULL;	/* Input file pointer */
char	ofnbuf[FNSIZE + 1] =	{ 0 };	/* Output filename buffer */
STREAM	out_fp		=	NULL;	/* Output file pointer */
char	listbuf[FNSIZE + 1] =	{ 0 };	/* Listing filename buffer */
STREAM	list_fp		=	NULL;	/* Listing file pointer */
char	temp1buf[FNSIZE + 1] =	{ 0 };	/* Temp file #1 filename buffer */
STREAM	temp1_fp	=	NULL;	/* Temp file #1 file pointer */
char	temp2buf[FNSIZE + 1] =	{ 0 };	/* Temp file #2 filename buffer */
STREAM	temp2_fp	=	NULL;	/* Temp file #2 file pointer */

STREAM	read_temp	=	NULL;	/* Temp file being read */
STREAM	write_temp	=	NULL;	/* Temp file being written */

/*
 *	General definitions.
 */
UCHAR	acc_dta[ACCDATASIZE] =	{ 0 };	/* T_DATA pack accumulator */
int	acc_dtasize	=	0;	/* T_DATA pack accumulated size */
SLONG	acc_storesize	=	0;	/* T_STORAGE accumulated size */
SLONG	aligned[NUMMODULES] =	{ 0L };	/* Loc. counter for last ".align" */
BOOL	eject		=	FALSE;	/* No initial formfeed needed */
BOOL	compress	=	FALSE;	/* TRUE if to pack T_DATA/T_STORAGE */
SLONG	con_expr	=	0;	/* Expression value from operand */
UINT	cpu_type	=	ALL;	/* CPU specific opcodes bitmask */
BOOL	emulate		=	FALSE;	/* TRUE if instruction emulation ON */
UINT	errors		=	0;	/* Total # of errors encountered */
BOOL	expr_err	=	0;	/* TRUE if an expression error found */
BOOL	had_amper	=	0;	/* TRUE if operand had "@" prefix */
BOOL	had_diff	=	0;	/* TRUE if operand was "rel-rel" */
BOOL	had_dollar	=	0;	/* TRUE if operand had "$" prefix */
BOOL	had_expr	=	0;	/* TRUE if operand contained expr */
BOOL	had_symbol	=	0;	/* TRUE if operand contained symbol */
UINT	max_i_size	=	MAX_I_SIZE_32B;	/* Max. instr. size in bytes */
SLONG	max_loc[NUMMODULES] =	{ 0L };	/* Maximum location counter */
SLONG	min_loc[NUMMODULES] =	{ 0L };	/* Minimum location counter */
SLONG	max_loc_adj[NUMMODULES] = { 0L };	/* "max_loc" adj for fwd ref */
SLONG	min_loc_adj[NUMMODULES] = { 0L };	/* "min_loc" adj for fwd ref */
UINT	module		=	0;	/* Current module # */
UINT	number		=	1;	/* Current input line # */
int	optimize	=	OHIGH;	/* High optimization level */
char	parse_buf[MAXLNSIZE + 1] = {0};	/* Parsing buffer */
STRING	parse_ptr	=	NULL;	/* Parsing buffer pointer */
BOOL	rel_only	=	FALSE;	/* TRUE if making relocatable code */
BOOL	retf_active	=	FALSE;	/* TRUE if holding a ".retf" pseudo */
UINT	retf_number	=	0;	/* Line # ".retf" appeared on */
SLONG	retf_value	=	0;	/* Value of ".retf" operand field */

#if	(H_OPSYS == MPW) || defined(__TURBOC__)
struct	symtab	*st	=	NULL;	/* Symbol table */
#else	/* !MPW && !__TURBOC__ */
struct	symtab	st[SYMTABSIZE] = { 0 };	/* Symbol table */
#endif	/* MPW || __TURBOC__ */

#ifdef	__TURBOC__
extern	unsigned	_stklen = 8192;
#endif	/* __TURBOC__ */

UINT	st_num1		=	0;	/* ST entry # for operand symbols */
UINT	st_num2		=	0;	/* ST entry # for operand symbols */
BOOL	translator	=	FALSE;	/* TRUE if source is translated */
BOOL	verbose		=	VERBOSE;/* Default: make output verbose */
UINT	word_size	=	WORD_SIZE_32B;	/* CPU word size in bytes */

/************************************************************************/
/*									*/
/*	Routine:	main						*/
/*									*/
/*	Opens the input and output files, deciphers any command line	*/
/*	switches, allocates free memory and starts the assembling.	*/
/*									*/
/*	Parameters:							*/
/*		"argc" An integer count of the command line arguments.	*/
/*		"argv" A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#endif	/* MSC_OPT */

int
main(argc,argv)
	int			argc;
	char			**argv;
	{
	extern	STRING		strchr();
	extern	STRING		strrchr();

	STRING			cptr;
	register	UINT	i;
	BOOL			list;
	STRING			ptr;
	BOOL			file_rename;
/*
 *	Check validity of input filename and open it.
 */
	if(argc < 2)
		usagerr();		/* Needs at least an input file */
	argv++;				/* Skip program name */
	makefn(*argv++,ifnbuf,TASM_INPUT_EXT);	/* Make full name */
#if	CRLF
	if((in_fp = (STREAM) fopen(ifnbuf,"rt")) == NULL)
#else
	if((in_fp = (STREAM) fopen(ifnbuf,"r")) == NULL)
#endif
		{
		fatal("Unable to open input file: ",ifnbuf);
		}
	strcpy(lifnbuf,ifnbuf);		/* Default #line filename template */
/*
 *	Given the input filename, create a default output filename in
 *	case the user leaves it to us.
 */
	strcpy(ofnbuf,ifnbuf);		/* Create default output filename */
	strcpy(strrchr(ofnbuf,'.'),TASM_OUTPUT_EXT);
	file_rename = FALSE;		/* And assume it will be used */
/*
 *	See if the user has specified a directory pathname for the temp files.
 *	If so, use it, otherwise put the temp files in the directory specified
 *	by the TEMP_PATH environment variable or the current directory if the
 *	variable doesn't exist.
 */
	if((*argv != NULL) && (**argv != '-'))	/* If explicit path given */
		strcpy(temp1buf,*argv++);	/* Create temp file names */
	else
		{
		temp1buf[0] = '\0';	/* No distinct temp directory name */
		if((cptr = getenv(TEMP_PATH)) != NULL)
			{
			if(strlen(cptr) < FNSIZE)
				strcpy(temp1buf,cptr);
			}
		}
/*
 *	Ensure temp directory has trailing delimiter and prune directory name
 *	from input file.
 */
#if	H_OPSYS == MPW
	if(((i = strlen(temp1buf)) > 0) && (temp1buf[i - 1] != ':'))
		strcat(temp1buf,":");
	if((cptr = strrchr(ifnbuf,':')) == NULL)
#else	/* !MPW */
	if(((i = strlen(temp1buf)) > 0) && (temp1buf[i - 1] != '/') &&
		(temp1buf[i - 1] != '\\'))
		{
		strcat(temp1buf,"/");
		}
	if(((cptr = strrchr(ifnbuf,'/')) == NULL) &&
		((cptr = strrchr(ifnbuf,'\\')) == NULL))
#endif	/* MPW */
		{
		cptr = &ifnbuf[-1];
		}
/*
 *	Add input filename to any existing temp directory path.
 */
	if((strlen(temp1buf) + strlen(++cptr)) < FNSIZE)
		strcat(temp1buf,cptr);
	else
		strcpy(temp1buf,cptr);
/*
 *	Substitute temp file suffixes.
 */
	strcpy(temp2buf,temp1buf);
	strcpy(strrchr(temp1buf,'.'),TEMP_NAME1);
	strcpy(strrchr(temp2buf,'.'),TEMP_NAME2);
/*
 *	Open temp files and allocate explicit buffers for them if needed.
 */
#if	CRLF
	if((temp1_fp = (STREAM) fopen(temp1buf,"w+b")) == NULL)
#else
	if((temp1_fp = (STREAM) fopen(temp1buf,"w+")) == NULL)
#endif
		{
		fatal("Unable to open temporary file: ",temp1buf);
		}
#if	IOBUFSIZE
	if((ptr = (STRING) malloc(IOBUFSIZE)) == NULL)
		fatal("Insufficient stream buffer memory for file: ",temp1buf);
	if(setvbuf(temp1_fp,ptr,_IOFBF,IOBUFSIZE) != 0)
		fatal("Error setting stream buffer for file: ",temp1buf);
#endif
#if	CRLF
	if((temp2_fp = (STREAM) fopen(temp2buf,"w+b")) == NULL)
#else
	if((temp2_fp = (STREAM) fopen(temp2buf,"w+")) == NULL)
#endif
		{
		fatal("Unable to open temporary file: ",temp2buf);
		}
#if	IOBUFSIZE
	if((ptr = (STRING) malloc(IOBUFSIZE)) == NULL)
		fatal("Insufficient stream buffer memory for file: ",temp2buf);
	if(setvbuf(temp2_fp,ptr,_IOFBF,IOBUFSIZE) != 0)
		fatal("Error setting stream buffer for file: ",temp2buf);
#endif
/*
 *	Set which temp file is to be initially read and which is to be written.
 */
	write_temp = temp1_fp;
	read_temp = temp2_fp;
/*
 *	Process any command line switches.
 */
	list = FALSE;			/* No listing file by default */
	while((ptr = *argv++) != NULL)	/* Process rest of command line */
		{
		if(*ptr == '-')
			{
			while(*++ptr != '\0')
				{
				switch((int) *ptr)
					{
/* Remove debug info from output */	case 'C':
					case 'c':
						lifnbuf[0] = '\0';
						compress = TRUE;
						break;

/* Generate a listing file */		case 'L':
					case 'l':
						if(translator)
							{
		fatal("Unable to generate non-assembly language listing","");
							}
						list = TRUE;
						break;

/* Rename output file */		case 'O':
					case 'o':
						file_rename = TRUE;
						break;

/* Optimization level flag */		case 'Q':
					case 'q':
						switch((int) *next_arg(&argv,
							&ptr))
							{
							case '0':
								optimize =
									OLOW;
								break;

							case '1':
								optimize =
									OMED;
								break;

							case '2':
								optimize =
									OHIGH;
								break;

							default:
								usagerr();
							}
						break;

/* Source was translated */		case 'T':
					case 't':
						if(list)
							{
		fatal("Unable to generate non-assembly language listing","");
							}
						translator = TRUE;
						break;

/* Turn OFF verbose */			case 'V':
					case 'v':
						verbose = !verbose;
						break;

					default:
						usagerr();
					}
				}
			}
		else if(file_rename)		/* New output filename? */
			{
/* Dflt nme, new drv? */if (ptr[1] == ':' && (ptr[2] == '\0'))
				{
				strcpy(ofnbuf,ptr);
				strcat(ofnbuf,ifnbuf[1]==':'
					? ifnbuf + 2 : ifnbuf);
				strcpy(strrchr(ofnbuf,'.'),
					TASM_OUTPUT_EXT);
				}
/* New filename */	else
/* Expand filename */		makefn(ptr,ofnbuf,TASM_OUTPUT_EXT);
			file_rename = FALSE;
			}
		else
			usagerr();
		}
/*
 *	Make sure input and output files are distinct.
 */
	if(! strcmp(ifnbuf,ofnbuf))
		fatal("Output file name same as input","");
/*
 *	Open output file.
 */
#if	CRLF
	if((out_fp = (STREAM) fopen(ofnbuf,"wb")) == NULL)
#else
	if((out_fp = (STREAM) fopen(ofnbuf,"w")) == NULL)
#endif
		{
		fatal("Unable to open output file: ",ofnbuf);
		}
/*
 *	Open listing file if requested.
 */
	if(list)
		{
		strcpy(listbuf,ofnbuf);
		strcpy(strrchr(listbuf,'.'),
		TASM_LIST_EXT);
#if	CRLF
		if((list_fp = (STREAM) fopen(listbuf,"wt")) == NULL)
#else
		if((list_fp = (STREAM) fopen(listbuf,"w")) == NULL)
#endif
			{
			fatal("Unable to open listing file: ",listbuf);
			}
		}
/*
 *	Initialize/get symbol table memory.
 */
#if	(H_OPSYS == MPW) || defined(__TURBOC__)
	if((st = (struct symtab *) calloc(SYMTABSIZE, sizeof(struct symtab)))
		== NULL)
		{
		fatal("Insufficient memory for symbol table","");
		}
#else	/* !MPW && !__TURBOC__ */
	for(ptr = (STRING) st, i = 0;
		i < (SYMTABSIZE * sizeof(struct symtab)); i++)
		{
		*ptr++ = '\0';		/* Zero symbol table out */
		}
#endif	/* MPW || __TURBOC__ */
/*
 *	Print sign-on messages if appropriate.
 */
	if(verbose)
		banner();
/*
 *	Verify that the various portability items are configured OK.
 */

/*lint -e506 lint isn't very found of constant valued booleans */

#ifdef	__TURBOC__
#pragma	warn	-rch
#endif	/* __TURBOC__ */

	if(sizeof(SLONG) < 4)
		fatal("The size of SLONG is not correctly configured","");

#ifdef	__TURBOC__
#pragma	warn	.rch
#endif	/* __TURBOC__ */

/*lint +e506 lint isn't very fond of constant valued booleans */

/*
 *	Everything is ready to go, now is the time to:
 */
	assemble();
	return (0);
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#endif	/* MSC_OPT */

/************************************************************************/
/*									*/
/*	Routine:	assemble					*/
/*									*/
/*	Handle the sequencing of the assembly process.			*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
assemble()
	{
#if	DEBUG
	putchar('[');			/* Bracket pass indicators */
#endif
/*
 *	First we process the input file into the internal file format used
 *	with the temp files.  This operation corresponds to the traditional
 *	first pass of a two pass assembler.
 */
	pass0();			/* Process whatever "parse" found */
/*
 *	Make a pass through the temp file and complete the error detection
 *	process by spotting undefined symbols.  This operation corresponds to
 *	the second pass of a traditional two pass assembler.  This pass also
 *	takes a first stab at figuring out how large instructions which refer
 *	to PC relative immediate data should be.
 */
	swap_temp();			/* Switch temp files around */
	if(optimize == OLOW)		/* Should we assume worst case? */
		pass1(TRUE);		/* Yes, all references are 8 bytes */
	else
		{			/* No, make them optimum length */
		pass1(FALSE);		/* Finish error detection */
/*
 *	Now we do the iterative "growing" of PC relative immediate data until
 *	enough prefix instructions are generated to handle the data.  Note that
 *	references to external data (or local references which are affected by
 *	external references), force the final binding to be performed by TLNK.
 *	These operations are not performed if the assembler found any errors.
 */
		if(! errors)
			{
			FOREVER
				{
				swap_temp();	/* Switch in&out temp files */
				if(pass1(FALSE))
					break;	/* Quit if no better */
				}
			if(optimize == OMED)
				{
				swap_temp();	/* Switch in&out temp files */
				pass2();	/* Grow boundary cases */
				}
			swap_temp();		/* Swap input & output files */
			pass1(TRUE);		/* Last pass, get loc. defs */
			}
		}
/*
 *	If a listing file was requested, now is the time to do it.
 */
	if(list_fp != NULL)
		{
		swap_temp();		/* Switch input & output temp files */
		list_pass();		/* Generate a listing */
		swap_temp();		/* Restore order of temp files */
		}
/*
 *	Well we've done as much as we can for now, translate the final temp
 *	file into the file format expected by TLNK and write it to the output
 *	file (assuming no errors were found).
 */
	if(! errors)
		{
		swap_temp();		/* Switch input & output temp files */
		pass3();		/* Write out relocatable file */
		}
#if	DEBUG
	putchar(']');			/* Bracket pass indicators */
#endif
	if(list_fp != NULL)
		{
		if(errors)
			fprintf(list_fp,"\n%u errors detected\n",errors);
		else
			fprintf(list_fp,"\nNo errors detected\n");
		}
	if(errors)
		{
		printf("\n%u errors detected\n",errors);
		terminate(ERRORS);		/* All done, found errors */
		}
	if(verbose)
		printf("\nNo errors detected\n");
	terminate(NOERRORS);	/* All done */
	}

/************************************************************************/
/*									*/
/*	Routine:	banner						*/
/*									*/
/*	Display sign-on message to user.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
banner()
	{
	printf("%s, %s, %s\n",ASM_NAME,VERSION,COPYRIGHT);
	}

/************************************************************************/
/*									*/
/*	Routine:	fatal						*/
/*									*/
/*	Print out a "FATAL" error message and quit.			*/
/*									*/
/*	Parameters:							*/
/*		"sptr1"	A pointer to the first of the error message.	*/
/*		"sptr2" A pointer to the rest of the error message.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
fatal(sptr1,sptr2)
	STRING			sptr1;
	STRING			sptr2;
	{
	fprintf(STDERR,"FATAL: %s%s\n",sptr1,sptr2);	/* Print message */
	terminate(ERRORS);		/* Bye... */
	}

/************************************************************************/
/*									*/
/*	Routine:	fatali						*/
/*									*/
/*	Print out a internal "FATAL" error message and quit.		*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to the error message.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
fatali(sptr)
	STRING			sptr;
	{
	fatal("TASM internal error #",sptr);
	}

/************************************************************************/
/*									*/
/*	Routine:	get_1b,get_2b,get_4b				*/
/*									*/
/*	Read 1, 2 or 4 bytes from the current "read" temp file.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The 1, 2 or 4 byte value read.			*/
/*									*/
/************************************************************************/

int
get_1b()
	{
	int			c;

	if((c = fgetc(read_temp)) == EOF)
		{
		if(ferror(read_temp) || feof(read_temp))
			{
			fatal("Error reading temp file: ",
				(read_temp == temp1_fp) ? temp1buf : temp2buf);
			}
		}
	return (c & 0xFF);
	}

int
get_2b()
	{
	int			rval;

	rval = get_1b();
	rval |= (get_1b() << 8);
	return (rval & 0xFFFF);
	}

SLONG
get_4b()
	{
	SLONG			rval;

	rval = (SLONG) get_1b();
	rval |= (((SLONG) get_1b()) << 8);
	rval |= (((SLONG) get_1b()) << 16);
	rval |= (((SLONG) get_1b()) << 24);
	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	lookup						*/
/*									*/
/*	Check if a symbol is in the symbol table, if not add it.	*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to the start of the symbol.		*/
/*									*/
/*	Returns:	A pointer to the existing/resulting symbol	*/
/*			table entry.					*/
/*									*/
/************************************************************************/

UINT
lookup(sptr)
	STRING			sptr;
	{
	register	UINT	i;
	register	UINT	num;
	register struct symtab	*ptr;
	register	STRING	tptr;

	tptr = sptr;			/* Point to symbol */
	num = 0;			/* Zero hash sum */
	while(*tptr)
		num = (num * 16) + *tptr++; /* Compute hash */
	num &= SYMTABMASK;		/* Mask to size of table */
/*
 *	We now have the symbol table entry number to start scanning at,
 *	but we need to convert that to the correct pointer, therefore:
 */
	ptr = st + num;
/*
 *	Scan table for exact match or empty slot.
 */
	for(i = 0; i < SYMTABSIZE; i++ , num++ , ptr++)
		{
		if(ptr >= &st[SYMTABSIZE])	/* Wrap around? */
			{
			num = 0;	/* If so */
			ptr = st;
			}
		if(ptr->id.name != NULL)	/* Entry filled? */
			{
			if(! strncmp(sptr,ptr->id.name,SYMNMESIZE))
				{
				return (num);	/* Exact match */
				}
			continue;	/* Filled, try next */
			}
		i = strlen(sptr) + 1;	/* Figure out how much space we need */
		if((tptr = (STRING)
			malloc((i < SYMNMESIZE) ? i : SYMNMESIZE)) == NULL)
			{
			fatal("Insufficient symbol table string memory","");
			}
		string_copy(tptr,sptr,SYMNMESIZE);
		ptr->id.name = tptr;	/* Remember string location */
		return (num);
		}
	fatal("Symbol table full","");
	return (0);			/* Make lint happy */
	}

/************************************************************************/
/*									*/
/*	Routine:	makefn						*/
/*									*/
/*	Expand a skeleton filename by adding a default extension if	*/
/*	none is supplied.						*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to the skeleton filename.		*/
/*		"bptr"	A pointer to the buffer the result goes in.	*/
/*		"ext"	A pointer to the default extension to use.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
makefn(sptr,bptr,ext)
	STRING			sptr;
	STRING			bptr;
	STRING			ext;
	{
	extern	STRING		strchr();
	extern	STRING		strrchr();

	STRING			s;

	if(strlen(sptr) > FNSIZE)
		usagerr();
	strcpy(bptr,sptr);
	if((s = strrchr(bptr,'.')) != NULL)
		{
		if((strchr(s,'/') == NULL) && (strchr(s,'\\') == NULL))
			return;
		}
	if(strlen(bptr) > (FNSIZE - 4))
		usagerr();
	strcat(bptr,ext);
	}

/************************************************************************/
/*									*/
/*	Routine:	next_arg					*/
/*									*/
/*	Fetch the next command line argument for the current flag.	*/
/*									*/
/*	Parameters:							*/
/*		"argv"	A pointer to the "main" "argv" parameter.	*/
/*		"ptr"	A pointer to the current parsing pointer.	*/
/*									*/
/*	Returns:	A pointer to the next argument string.		*/
/*									*/
/************************************************************************/

STRING
next_arg(argv,ptr)
	char			***argv;
	char			**ptr;
	{
	STRING			s;

	s = *ptr;
	if(*++s == '\0')
		{
		if((s = *(*argv)++) == NULL)
			fatal("Missing argument for flag: ",*ptr);
		}
	else
		*ptr = s + (strlen(s) - 1);

	return (s);
	}

/************************************************************************/
/*									*/
/*	Routine:	out_write					*/
/*									*/
/*	Write a specified # of characters to the output file.		*/
/*									*/
/*	Parameters:							*/
/*		"start"	Starting address of characters to write.	*/
/*		"size"	The number to write.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
out_write(start,size)
	STRING			start;
	int			size;
	{
	if(fwrite(start,1,(UINT) size,out_fp) != size)
		fatal("Error writing output file: ",ofnbuf);
	}

/************************************************************************/
/*									*/
/*	Routine:	out_1b,out_2b,out_4b				*/
/*									*/
/*	Write 1,2 or 4 bytes to the output file.			*/
/*									*/
/*	Parameters:							*/
/*		"value"	The 1 or 2 byte value to write.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
out_1b(value)
	int			value;
	{
	if(putc(value,out_fp) == EOF)
		{
		if(ferror(out_fp))
			fatal("Error writing output file: ",ofnbuf);
		}
	}

VOID
out_2b(value)
	int			value;
	{
	out_1b(value);
	out_1b(value >> 8);
	}

VOID
out_4b(value)
	SLONG			value;
	{
	out_1b((int) value);
	out_1b((int) (value >>= 8));
	out_1b((int) (value >>= 8));
	out_1b((int) (value >> 8));
	}

/************************************************************************/
/*									*/
/*	Routine:	put_1b,put_2b,put_4b				*/
/*									*/
/*	Write 1,2 or 4 bytes to the current "write" temp file.		*/
/*									*/
/*	Parameters:							*/
/*		"value"	The 1,2 or 4 byte value to write.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
put_1b(value)
	int			value;
	{
	if(putc(value,write_temp) == EOF)
		{
		if(ferror(write_temp))
			{
			fatal("Error writing temp file: ",
			(write_temp == temp1_fp) ? temp1buf : temp2buf);
			}
		}
	}

VOID
put_2b(value)
	int			value;
	{
	put_1b(value);
	put_1b(value >> 8);
	}

VOID
put_4b(value)
	SLONG			value;
	{
	put_1b((int) value);
	put_1b((int) (value >>= 8));
	put_1b((int) (value >>= 8));
	put_1b((int) (value >> 8));
	}

/************************************************************************/
/*									*/
/*	Routine:	string_copy					*/
/*									*/
/*	Copy at most "n" characters from one string to another.  No	*/
/*	more characters are copied past the '\0'.  This is functionally	*/
/*	identical to "strncpy" without the zero padding and return	*/
/*	value.								*/
/*									*/
/*	Parameters:							*/
/*		"dest"	Points to the string destination.		*/
/*		"src"	Points to the string source.			*/
/*		"n"	The maximum # of characters to copy.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
string_copy(dest,src,n)
	STRING			dest;
	STRING			src;
	register	UINT	n;
	{
	while(n--)
		{
		if((*dest++ = *src++) == '\0')
			return;
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	swap_temp					*/
/*									*/
/*	Make the current temp input file be the current temp output	*/
/*	file and vice-versa.  The files are also rewound.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
swap_temp()
	{
	if(write_temp == temp1_fp)
		{
		write_temp = temp2_fp;
		read_temp = temp1_fp;
		}
	else
		{
		write_temp = temp1_fp;
		read_temp = temp2_fp;
		}
	rewind(write_temp);
	rewind(read_temp);
	}

/************************************************************************/
/*									*/
/*	Routine:	terminate					*/
/*									*/
/*	Erase temp files and return to the operating system.  If errors	*/
/*	were detected remove the output file.				*/
/*									*/
/*	Parameters:							*/
/*		"rval"	The desired "exit" value to return.		*/
/*									*/
/*	Returns:	None, and takes its time about it too.		*/
/*									*/
/************************************************************************/

VOID
terminate(rval)
	int			rval;
	{
/*
 *	Remove output file if errors encountered.
 */
	if((rval == ERRORS) && (out_fp != NULL))
		{
		if(fclose(out_fp) == EOF)
			{
	fprintf(STDERR,"WARNING: Unable to close output file: %s\n",ofnbuf);
			}
		else if(unlink(ofnbuf) == -1)
			{
	fprintf(STDERR,"WARNING: Unable to remove output file: %s\n",ofnbuf);
			}
		}
/*
 *	Close and remove temp files unless debugging is active.
 */
#if	DEBUG
	exit(rval);
#else
	if(temp1_fp)
		{
		if(fclose(temp1_fp) == EOF)
			{
	fprintf(STDERR,"WARNING: Unable to close temp file: %s\n",temp1buf);
			}
		else if(unlink(temp1buf) == -1)
			{
	fprintf(STDERR,"WARNING: Unable to remove temp file: %s\n",temp1buf);
			}
		}
	if(temp2_fp)
		{
		if(fclose(temp2_fp) == EOF)
			{
	fprintf(STDERR,"WARNING: Unable to close temp file: %s\n",temp2buf);
			}
		else if(unlink(temp2buf) == -1)
			{
	fprintf(STDERR,"WARNING: Unable to remove temp file: %s\n",temp2buf);
			}
		}
	exit(rval);			/* We are history */
#endif
	}

/************************************************************************/
/*									*/
/*	Routine:	usagerr						*/
/*									*/
/*	Report incorrect assembler invocation to user with a suggestion	*/
/*	of how it should be done.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
usagerr()
	{
fprintf(STDERR,"FATAL: General usage is 'tasm in_file [temp_directory] [options] [out_file]'\n");
printf("Options syntax:\n\n");
printf("Compress output file:		-c\n");
printf("Generate a listing file:	-l\n");
printf("Specify output filename:	-o\n");
printf("Specify no prefix optimization:	-q0\n");
printf("Specify symbol optimization:	-q1\n");
printf("Specify full optimization:	-q2 (default)\n");
printf("Input file is translation:	-t\n");
printf("Disable verbose output:		-v\n");
	terminate(ERRORS);
	}
