/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: tasm2.c,v 2.2 92/04/21 11:10:14 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/************************************************************************/
/*									*/
/*	File:	tasm2.c							*/
/*									*/
/*	Transputer Assembler (TASM), source file 2.  This file contains	*/
/*	the code which reads the assembly language file and transforms	*/
/*	it into the internal file format which is written on the temp	*/
/*	files.								*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1989 by Logical Systems.		*/
/*									*/
/*				09/01/86				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<ctype.h>		/* Character classification macros */
#include	<string.h>		/* String functions */
#include	"taldef.h"		/* Assembler/Linker definitions */
#include	"tasmdef.h"		/* Assembler definitions */
#include	"tasmtyp.h"		/* Assembler data declarations */
#include	"tasmext.h"		/* Assembler externals */

/************************************************************************/
/*									*/
/*			Mnemonic State Definitions			*/
/*									*/
/************************************************************************/

/*
 *	For "state" variable in "nextc".
 */
#define	BOL		0		/* Beginning-of-line */
#define	COMMENT		1		/* Within comment */
#define	LABEL		2		/* Within label */
#define	BOO		3		/* Beginning-of-opcode */
#define	OPCODE		4		/* Within opcode/pseudo-op */
#define	BOOP		5		/* Beginning-of-operand field */
#define	OPERAND		6		/* Within operand */
#define	NUMBER		7		/* Update line # next time */

/************************************************************************/
/*									*/
/*			Local Data Definitions				*/
/*									*/
/************************************************************************/

/*
 *	Instruction length table.
 */

static	struct	opcode	op_info[] =
	{
		{	"#LINE"	,	ALL,		UNUSED,	pp_line	},
		{	".ALIGN",	ALL,		UNUSED,	align	},
		{	".ALL"	,	ALL,		UNUSED,	all	},
		{	".DB"	,	ALL,		UNUSED,	db	},
		{	".DBNZ"	,	ALL,		UNUSED,	dbnz	},
		{	".DS"	,	ALL,		UNUSED,	ds	},
		{	".DW"	,	ALL,		UNUSED,	dw	},
		{	".EMULATE",	ALL,		UNUSED,	do_em	},
		{	".END"	,	ALL,		UNUSED,	the_end	},
		{	".EXT"	,	ALL,		UNUSED,	do_extern },
		{	".LDC"	,	ALL,		DOTLDC,	op_i	},
		{	".MOD"	,	ALL,		UNUSED,	do_module },
		{	".NOEMULATE",	ALL,		UNUSED,	do_noem	},
		{	".NOREL",	ALL,		UNUSED, do_norel},
		{	".PUB"	,	ALL,		UNUSED,	do_public },
		{	".REAL32",	ALL,		FALSE,	do_float },
		{	".REAL64",	ALL,		TRUE,	do_float },
		{	".REL"	,	ALL,		UNUSED,	do_rel	},
		{	".RETF"	,	ALL,		UNUSED,	do_retf	},
		{	".SET"	,	ALL,		UNUSED,	do_set	},
		{	".SYM"	,	ALL,		UNUSED,	do_sym	},
		{	".T212"	,	ALL,		UNUSED,	t212	},
		{	".T414"	,	ALL,		UNUSED,	t414	},
		{	".T800"	,	ALL,		UNUSED,	t800	},
		{	".VAL"	,	ALL,		UNUSED,	do_val	},
		{	"?U32TOR32",	FALSE,		UNUSED,	noti	},
		{	"ADC"	,	ALL,		ADC,	op_i	},
		{	"ADD"	,	ALL,		0x05,	opr_i	},
		{	"AJW"	,	ALL,		AJW,	op_i	},
		{	"ALT"	,	ALL,		0x43,	opr_i	},
		{	"ALTEND",	ALL,		0x45,	opr_i	},
		{	"ALTWT"	,	ALL,		0x44,	opr_i	},
		{	"AND"	,	ALL,		0x46,	opr_i	},
		{	"BCNT"	,	ALL,		0x34,	opr_i	},
		{	"BITCNT",	ALL,		0x76,	opr_i	},
		{	"BITREVNBITS",	ALL,		0x78,	opr_i	},
		{	"BITREVWORD",	ALL,		0x77,	opr_i	},
		{	"BREAK"	,	ALL,		0xB1,	opr_i	},
		{	"BSUB"	,	ALL,		0x02,	opr_i	},
		{	"CALL"	,	ALL,		CALL,	op_i	},
		{	"CCNT1"	,	ALL,		0x4D,	opr_i	},
		{	"CFLERR",	T414,		0x73,	opr_i	},
		{	"CJ"	,	ALL,		CJ,	op_i	},
		{	"CLRHALTERR",	ALL,		0x57,	opr_i	},
		{	"CLRJ0BREAK",	ALL,		0xB2,	opr_i	},
		{	"CRCBYTE",	ALL,		0x75,	opr_i	},
		{	"CRCWORD",	ALL,		0x74,	opr_i	},
		{	"CSNGL"	,	ALL,		0x4C,	opr_i	},
		{	"CSUB0"	,	ALL,		0x13,	opr_i	},
		{	"CWORD"	,	ALL,		0x56,	opr_i	},
		{	"DIFF"	,	ALL,		0x04,	opr_i	},
		{	"DISC"	,	ALL,		0x2F,	opr_i	},
		{	"DISS"	,	ALL,		0x30,	opr_i	},
		{	"DIST"	,	ALL,		0x2E,	opr_i	},
		{	"DIV"	,	ALL,		0x2C,	opr_i	},
		{	"DUP",		ALL,		0x5A,	opr_i	},
		{	"ENBC"	,	ALL,		0x48,	opr_i	},
		{	"ENBS"	,	ALL,		0x49,	opr_i	},
		{	"ENBT"	,	ALL,		0x47,	opr_i	},
		{	"ENDP"	,	ALL,		0x03,	opr_i	},
		{	"EQC"	,	ALL,		EQC,	op_i	},
		{	"FMUL"	,	T414|T800,	0x72,	opr_i	},
		{	"FPADD"	,	T800,		0x87,	opr_i	},
		{	"FPB32TOR64",	T800,		0x9A,	opr_i	},
		{	"FPCHKERR",	T800,		0x83,	opr_i	},
		{	"FPDIV"	,	T800,		0x8C,	opr_i	},
		{	"FPDUP"	,	T800,		0xA3,	opr_i	},
		{	"FPENTRY",	T800,		0xAB,	opr_i	},
		{	"FPEQ"	,	T800,		0x95,	opr_i	},
		{	"FPGT"	,	T800,		0x94,	opr_i	},
		{	"FPI32TOR32",	T800,		0x96,	opr_i	},
		{	"FPI32TOR64",	T800,		0x98,	opr_i	},
		{	"FPINT"	,	T800,		0xA1,	opr_i	},
		{	"FPLDNLADDDB",	T800,		0xA6,	opr_i	},
		{	"FPLDNLADDSN",	T800,		0xAA,	opr_i	},
		{	"FPLDNLDB",	T800,		0x8A,	opr_i	},
		{	"FPLDNLDBI",	T800,		0x82,	opr_i	},
		{	"FPLDNLMULDB",	T800,		0xA8,	opr_i	},
		{	"FPLDNLMULSN",	T800,		0xAC,	opr_i	},
		{	"FPLDNLSN",	T800,		0x8E,	opr_i	},
		{	"FPLDNLSNI",	T800,		0x86,	opr_i	},
		{	"FPLDZERODB",	T800,		0xA0,	opr_i	},
		{	"FPLDZEROSN",	T800,		0x9F,	opr_i	},
		{	"FPMUL"	,	T800,		0x8B,	opr_i	},
		{	"FPNAN"	,	T800,		0x91,	opr_i	},
		{	"FPNOTFINITE",	T800,		0x93,	opr_i	},
		{	"FPORDERED",	T800,		0x92,	opr_i	},
		{	"FPREMFIRST",	T800,		0x8F,	opr_i	},
		{	"FPREMSTEP",	T800,		0x90,	opr_i	},
		{	"FPREV"	,	T800,		0xA4,	opr_i	},
		{	"FPRTOI32",	T800,		0x9D,	opr_i	},
		{	"FPSTNLDB",	T800,		0x84,	opr_i	},
		{	"FPSTNLI32",	T800,		0x9E,	opr_i	},
		{	"FPSTNLSN",	T800,		0x88,	opr_i	},
		{	"FPSUB"	,	T800,		0x89,	opr_i	},
		{	"FPTESTERR",	T414|T800,	0x9C,	opr_i	},
		{	"FPUABS",	T800,		0x0B,	fpu_i	},
		{	"FPUCHKI32",	T800,		0x0E,	fpu_i	},
		{	"FPUCHKI64",	T800,		0x0F,	fpu_i	},
		{	"FPUCLRERR",	T800,		0x9C,	fpu_i	},
		{	"FPUDIVBY2",	T800,		0x11,	fpu_i	},
		{	"FPUEXPDEC32",	T800,		0x09,	fpu_i	},
		{	"FPUEXPINC32",	T800,		0x0A,	fpu_i	},
		{	"FPUMULBY2",	T800,		0x12,	fpu_i	},
		{	"FPUNOROUND",	T800,		0x0D,	fpu_i	},
		{	"FPUR32TOR64",	T800,		0x07,	fpu_i	},
		{	"FPUR64TOR32",	T800,		0x08,	fpu_i	},
		{	"FPURM"	,	T800,		0x05,	fpu_i	},
		{	"FPURN"	,	T800,		0x22,	fpu_i	},
		{	"FPURP"	,	T800,		0x04,	fpu_i	},
		{	"FPURZ"	,	T800,		0x06,	fpu_i	},
		{	"FPUSETERR",	T800,		0x23,	fpu_i	},
		{	"FPUSQRTFIRST",	T800,		0x01,	fpu_i	},
		{	"FPUSQRTLAST",	T800,		0x03,	fpu_i	},
		{	"FPUSQRTSTEP",	T800,		0x02,	fpu_i	},
		{	"GAJW"	,	ALL,		0x3C,	opr_i	},
		{	"GCALL"	,	ALL,		0x06,	opr_i	},
		{	"GT"	,	ALL,		0x09,	opr_i	},
		{	"IN"	,	ALL,		0x07,	opr_i	},
		{	"J"	,	ALL,		J,	op_i	},
		{	"LADD"	,	ALL,		0x16,	opr_i	},
		{	"LB"	,	ALL,		0x01,	opr_i	},
		{	"LDC"	,	ALL,		LDC,	op_i	},
		{	"LDDEVID",	ALL,		0x17C,	opr_i	},
		{	"LDIFF"	,	ALL,		0x4F,	opr_i	},
		{	"LDINF"	,	T414,		0x71,	opr_i	},
		{	"LDIV"	,	ALL,		0x1A,	opr_i	},
		{	"LDL"	,	ALL,		LDL,	op_i	},
		{	"LDLP"	,	ALL,		LDLP,	op_i	},
		{	"LDMEMSTARTVAL",ALL,		0x7E,	opr_i	},
		{	"LDNL"	,	ALL,		LDNL,	op_i	},
		{	"LDNLP"	,	ALL,		LDNLP,	op_i	},
		{	"LDPI"	,	ALL,		0x1B,	opr_i	},
		{	"LDPRI"	,	ALL,		0x1E,	opr_i	},
		{	"LDTIMER",	ALL,		0x22,	opr_i	},
		{	"LEND"	,	ALL,		0x21,	opr_i	},
		{	"LMUL"	,	ALL,		0x31,	opr_i	},
		{	"LSHL"	,	ALL,		0x36,	opr_i	},
		{	"LSHR"	,	ALL,		0x35,	opr_i	},
		{	"LSUB"	,	ALL,		0x38,	opr_i	},
		{	"LSUM"	,	ALL,		0x37,	opr_i	},
		{	"MINT"	,	ALL,		MINT,	opr_i	},
		{	"MOVE"	,	ALL,		0x4A,	opr_i	},
		{	"MOVE2DALL",	T414|T800,	0x5C,	opr_i	},
		{	"MOVE2DINIT",	T414|T800,	0x5B,	opr_i	},
		{	"MOVE2DNONZERO",T414|T800,	0x5D,	opr_i	},
		{	"MOVE2DZERO",	T414|T800,	0x5E,	opr_i	},
		{	"MUL"	,	ALL,		0x53,	opr_i	},
		{	"NFIX"	,	ALL,		NFIX,	pfix_i	},
		{	"NORM"	,	ALL,		0x19,	opr_i	},
		{	"NOT"	,	ALL,		0x32,	opr_i	},
		{	"OPR"	,	ALL,		OPR,	op_i	},
		{	"OR"	,	ALL,		0x4B,	opr_i	},
		{	"OUT"	,	ALL,		0x0B,	opr_i	},
		{	"OUTBYTE",	ALL,		0x0E,	opr_i	},
		{	"OUTWORD",	ALL,		0x0F,	opr_i	},
		{	"PFIX"	,	ALL,		PFIX,	pfix_i	},
		{	"POP"	,	ALL,		0x79,	opr_i	},
		{	"POSTNORMSN",	T414,		0x6C,	opr_i	},
		{	"PROD"	,	ALL,		0x08,	opr_i	},
		{	"REM"	,	ALL,		0x1F,	opr_i	},
		{	"RESETCH",	ALL,		0x12,	opr_i	},
		{	"RET"	,	ALL,		RET,	opr_i	},
		{	"REV"	,	ALL,		0x00,	opr_i	},
		{	"ROT"	,	ALL,		0x79,	opr_i	},
		{	"ROUNDSN",	T414,		0x6D,	opr_i	},
		{	"RUNP"	,	ALL,		0x39,	opr_i	},
		{	"SAVEH"	,	ALL,		0x3E,	opr_i	},
		{	"SAVEL"	,	ALL,		0x3D,	opr_i	},
		{	"SB"	,	ALL,		0x3B,	opr_i	},
		{	"SETERR",	ALL,		0x10,	opr_i	},
		{	"SETHALTERR",	ALL,		0x58,	opr_i	},
		{	"SETJ0BREAK",	ALL,		0xB3,	opr_i	},
		{	"SHL"	,	ALL,		0x41,	opr_i	},
		{	"SHR"	,	ALL,		0x40,	opr_i	},
		{	"START"	,	ALL,		0x1FF,	opr_i	},
		{	"STARTP",	ALL,		0x0D,	opr_i	},
		{	"STHB"	,	ALL,		0x50,	opr_i	},
		{	"STHF"	,	ALL,		0x18,	opr_i	},
		{	"STL"	,	ALL,		STL,	op_i	},
		{	"STLB"	,	ALL,		0x17,	opr_i	},
		{	"STLF"	,	ALL,		0x1C,	opr_i	},
		{	"STNL"	,	ALL,		STNL,	op_i	},
		{	"STOPERR",	ALL,		0x55,	opr_i	},
		{	"STOPP"	,	ALL,		0x15,	opr_i	},
		{	"STTIMER",	ALL,		0x54,	opr_i	},
		{	"SUB"	,	ALL,		0x0C,	opr_i	},
		{	"SUM"	,	ALL,		0x52,	opr_i	},
		{	"TALT"	,	ALL,		0x4E,	opr_i	},
		{	"TALTWT",	ALL,		0x51,	opr_i	},
		{	"TESTERR",	ALL,		0x29,	opr_i	},
		{	"TESTHALTERR",	ALL,		0x59,	opr_i	},
		{	"TESTHARDCHAN",	ALL,		0x2D,	opr_i	},
		{	"TESTJ0BREAK",	ALL,		0xB4,	opr_i	},
		{	"TESTPRANAL",	ALL,		0x2A,	opr_i	},
		{	"TIMERDISABLEH",ALL,		0x7A,	opr_i	},
		{	"TIMERDISABLEL",ALL,		0x7B,	opr_i	},
		{	"TIMERENABLEH",	ALL,		0x7C,	opr_i	},
		{	"TIMERENABLEL",	ALL,		0x7D,	opr_i	},
		{	"TIN"	,	ALL,		0x2B,	opr_i	},
		{	"UNPACKSN",	T414,		0x63,	opr_i	},
		{	"WCNT"	,	ALL,		0x3F,	opr_i	},
		{	"WSUB"	,	ALL,		0x0A,	opr_i	},
		{	"WSUBDB",	ALL,		0x81,	opr_i	},
		{	"XDBLE"	,	ALL,		0x1D,	opr_i	},
		{	"XOR"	,	ALL,		0x33,	opr_i	},
		{	"XWORD"	,	ALL,		0x3A,	opr_i	},
	};

/************************************************************************/
/*									*/
/*	Routine:	align						*/
/*									*/
/*	Handle the ".align" pseudo-op.  Forces the next data to be word	*/
/*	aligned by generating a T_ALIGN record (a word is defined to be	*/
/*	32 bits for now).						*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
align(parameter)
	SLONG			parameter;
	{
/*
 *	First make sure that some code or data has occurred since the last
 *	".align" (redundant ".align"s mess up the prefix minimization code).
 */
	if(aligned[module] == min_loc[module])
		return;
/*
 *	Otherwise, remember that we have one now and add to temp file.
 */
	aligned[module] = min_loc[module];	/* Remember that we had one */
	put_1b((int) T_ALIGN);		/* Nope, adjust us */
	put_2b((int) number);		/* Line # */
	loc_pad();			/* Update location counters */
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	all						*/
/*									*/
/*	Handle the ".all" pseudo-op.  This selects the set of opcodes	*/
/*	common to all the 32 bit transputers supported by TASM.		*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
all(parameter)
	SLONG			parameter;
	{
	cpu_type = ALL;			/* Select common legal opcodes */
	max_i_size = MAX_I_SIZE_32B;	/* Maximum instruction length */
	word_size = WORD_SIZE_32B;	/* Word size in bytes */
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	build_i						*/
/*									*/
/*	Build an instruction and output it to the temp file given a	*/
/*	function (instruction) code and a 32 bit constant.		*/
/*									*/
/*	Parameters:							*/
/*		"length"	The length of the instruction to build.	*/
/*		"function"	The second 4 bits of this hold the	*/
/*				basic instruction opcode bits.		*/
/*		"constant"	The 32 bit constant value.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
build_i(length,function,constant)
	UINT			length;
	UINT			function;
	SLONG			constant;
	{
	register	UINT	i;
	UCHAR			preamble[7];
/*
 *	First see if we are doing the "optimizing" trick for ".LDC" opcodes.
 */
	if((function == DOTLDC) && (constant <= -2146435074L))
		{
		build_i((UINT) MINT_SIZE,(UINT) OPR,(SLONG) MINT);
		length -= MINT_SIZE;
		constant &= 0x7FFFFFFFL;
		if(constant != 0L)
			{
			build_i((UINT) length,(UINT) ADC,(SLONG) constant);
			return;
			}
		else
			{
			while(length--)
				build_i((UINT) 1,(UINT) NOP,(SLONG) 0L);
			return;
			}
		}

	function &= 0xF0;
/*
 *	Otherwise, generate the instruction in the normal fashion.
 */
	if((length <= 0) || (length > 8))
		fatali("4");
	if(i_size(function,constant) > length)
		fatali("5");
	if(length > 1)
		{
/*
 *	Well we know we need prefix instructions, figure out whether the first
 *	one should be a negative prefix.
 */
		if(constant < 0)
			preamble[0] = NFIX;
		else
			preamble[0] = PFIX;
/*
 *	Now determine the basic function/instruction byte which goes at the end
 *	of the string of prefixes.
 */
		function = function | ((UINT) (constant & 0xF));
/*
 *	Figure out the remainder of the necessary prefixes.
 */
		for(i = length - 2; i > 0; i--)
			preamble[i] = PFIX | ((constant >>= 4) & 0xF);
		if(preamble[0] == NFIX)
			{
/*lint -e502 lint doesn't like complementing signed integral types */
			preamble[0] |= (( ~(constant >> 4)) & 0xF);
/*lint +e502 lint doesn't like complementing signed integral types */
			}
		else
			preamble[0] |= ((constant >> 4) & 0xF);
		}
	else				/* If no prefix */
		function = function | ((UINT) (constant & 0xF));
/*
 *	Time to write the instruction encoding out to the temp file.
 */
	for(i = 0; i < (length - 1); i++)
		put_1b((int) preamble[i]);
	put_1b((int) function);
	}

/************************************************************************/
/*									*/
/*	Routine:	data						*/
/*									*/
/*	Generate T_DATA record for an opcode which can be completely	*/
/*	defined when first encountered (no forward references).		*/
/*									*/
/*	Parameters:							*/
/*		"line_num"	The line number to put in the record.	*/
/*		"opcode"	The opcode value to use.		*/
/*		"parameter"	The value to encode into the opcode.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
data(line_num,opcode,parameter)
	UINT			line_num;
	UINT			opcode;
	SLONG			parameter;
	{
	UINT			length;

	put_1b((int) T_DATA);
	put_2b((int) line_num);		/* Line # */
	length = i_size(opcode,parameter);	/* Figure out how long it is */
	put_2b((int) length);		/* Instruction length */
	build_i(length,opcode,parameter);	/* Build instruction */
/*
 *	Update minimum and maximum location counters.
 */
	min_loc[module] += length;
	max_loc[module] += length;
	}

/************************************************************************/
/*									*/
/*	Routine:	db						*/
/*									*/
/*	Handle define bytes (".db"), pseudo-op.				*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
db(parameter)
	SLONG			parameter;
	{
	UINT			count;
	STRING			temp;
/*
 *	First we pass over line and determine how much storage we will be
 *	allocating.
 */
	temp = parse_ptr;		/* Remember start of expressions */
	if(match("\""))			/* Was string literal */
		{
		if(str_size(TRUE))
			return;		/* String format error? */
		count = (UINT) con_expr;	/* String length */
		}
	else				/* Was constant expression */
		{
		if(operand(word_size))
			return;		/* Expression error */
		if(had_symbol || had_diff || had_dollar || had_amper ||
			(! had_expr))
			{
			post_error(ERR_EXPR,NULL);
			return;
			}
		count = 1;		/* # of expressions */
		}
	while(match(","))
		{
		if(match("\""))		/* Was string literal */
			{
			str_size(TRUE);	/* Compute string length */
			count = count + ((UINT) con_expr);	/* Str len */
			}
		else			/* Was constant expression */
			{
			if(operand(word_size))
				return;	/* Expression error */
			if(had_symbol || had_diff || had_dollar || had_amper ||
				(! had_expr))
				{
				post_error(ERR_EXPR,NULL);
				return;
				}
			count++;
			}
		}
/*
 *	We now know the amount of storage that is being defined.  Generate the
 *	header part of the temp file record.
 */
	put_1b((int) T_DATA);
	put_2b((int) number);		/* Input line # */
	put_2b((int) count);		/* # of bytes to reserve */
/*
 *	Now go back through and re-evaluate the expressions, this time put the
 *	results in the temp file record.
 */
	parse_ptr = temp;		/* Restore parsing buffer pointer */
	if(match("\""))
		str_out(TRUE);		/* Output string constant */
	else
		{
		operand(word_size);		/* Parse expression */
		put_1b((int) con_expr);	/* Or byte value */
		}
	while(match(","))
		{
		if(match("\""))
			str_out(TRUE);	/* Output string constant */
		else
			{
			operand(word_size);	/* Parse expression */
			put_1b((int) con_expr);	/* Or byte value */
			}
		}
/*
 *	Update minimum and maximum location counters.
 */
	min_loc[module] += count;
	max_loc[module] += count;
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	dbnz						*/
/*									*/
/*	Same as ".db" but text strings are not null byte terminated.	*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
dbnz(parameter)
	SLONG			parameter;
	{
	UINT			count;
	STRING			temp;
/*
 *	First we pass over line and determine how much storage we will be
 *	allocating.
 */
	temp = parse_ptr;		/* Remember start of expressions */
	if(match("\""))			/* Was string literal */
		{
		if(str_size(FALSE))
			return;		/* String format error? */
		count = ((UINT) con_expr);	/* String length */
		}
	else				/* Was constant expression */
		{
		if(operand(word_size))
			return;		/* Expression error */
		if(had_symbol || had_diff || had_dollar || had_amper ||
			(! had_expr))
			{
			post_error(ERR_EXPR,NULL);
			return;
			}
		count = 1;		/* # of expressions */
		}
	while(match(","))
		{
		if(match("\""))		/* Was string literal */
			{
			str_size(FALSE);	/* Compute string length */
			count = count + ((UINT) con_expr);	/* Str len */
			}
		else			/* Was constant expression */
			{
			if(operand(word_size))
				return;	/* Expression error */
			if(had_symbol || had_diff || had_dollar || had_amper ||
				(! had_expr))
				{
				post_error(ERR_EXPR,NULL);
				return;
				}
			count++;
			}
		}
/*
 *	We now know the amount of storage that is being defined.  Generate the
 *	header part of the temp file record.
 */
	put_1b((int) T_DATA);
	put_2b((int) number);		/* Input line # */
	put_2b((int) count);		/* # of bytes to reserve */
/*
 *	Now go back through and re-evaluate the expressions, this time put the
 *	results in the temp file record.
 */
	parse_ptr = temp;		/* Restore parsing buffer pointer */
	if(match("\""))
		str_out(FALSE);		/* Output string constant */
	else
		{
		operand(word_size);		/* Parse expression */
		put_1b((int) con_expr);	/* Or byte value */
		}
	while(match(","))
		{
		if(match("\""))
			str_out(FALSE);	/* Output string constant */
		else
			{
			operand(word_size);	/* Parse expression */
			put_1b((int) con_expr);	/* Or byte value */
			}
		}
/*
 *	Update minimum and maximum location counters.
 */
	min_loc[module] += count;
	max_loc[module] += count;
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	do_em						*/
/*									*/
/*	Handle the ".emulate" pseudo-op.				*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
do_em(parameter)
	SLONG			parameter;
	{
	emulate = TRUE;
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	do_extern					*/
/*									*/
/*	Handle the ".ext" pseudo-op.					*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
do_extern(parameter)
	SLONG			parameter;
	{
	if(note_extern())
		return;
	while(match(","))
		{
		if(note_extern())
			return;
		}
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	do_float					*/
/*									*/
/*	Handle the ".real32" and ".real64" pseudo-ops.			*/
/*									*/
/*	Parameters:							*/
/*		"parameter"-	TRUE if 64 bit IEEE wanted, FALSE if 32.*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
do_float(parameter)
	SLONG			parameter;
	{
	UINT			count;
	UINT			length;
	SLONG			result[2];
	STRING			temp;
/*
 *	First we pass over line and make sure no errors are lurking.
 */
	temp = parse_ptr;		/* Remember start of expressions */
	for(count = 1; ; count++)
		{
		if(atoieee(result,(BOOL) parameter))
			{
			post_error(ERR_FLOAT,NULL);
			return;
			}
		if(! match(","))
			break;		/* If no more items on line */
		}
/*
 *	Now we pass over floating point constants and generate T_DATA record.
 */
	parse_ptr = temp;		/* Reset to point at start of expr's */
	put_1b((int) T_DATA);	/* Record type */
	put_2b((int) number);	/* Line # */
	if(parameter)		/* Float size? */
		put_2b((int) (length = (count << 3)));	/* 64 bit */
	else
		put_2b((int) (length = (count << 2)));	/* 32 bit */
	while(count--)		/* For each floating point constant */
		{
		atoieee(result,(BOOL) parameter);
		if(parameter)
			{
			put_4b((SLONG) result[0]);	/* 64 bit */
			put_4b((SLONG) result[1]);
			}
		else
			put_4b((SLONG) result[0]);	/* 32 bit */
		if(count != 0)
			match(",");		/* Slurp comma */
		}
	min_loc[module] += length;
	max_loc[module] += length;
	}

/************************************************************************/
/*									*/
/*	Routine:	do_module					*/
/*									*/
/*	Change the current module # (".mod" pseudo-op).			*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
do_module(parameter)
	SLONG			parameter;
	{
	if(operand(word_size))
		return;			/* If expression error */
/*
 *	Make sure module # was constant expression and  within legal range.
 */
	if(had_symbol || had_diff || had_amper || had_dollar || (! had_expr) ||
		(con_expr < 0) || (con_expr >= NUMMODULES))
		{
		post_error(ERR_EXPR,NULL);
		return;
		}
/*
 *	Otherwise, everything looks OK, generate appropriate temp file
 *	record and update active module #.
 */
	put_1b((int) T_MODULE);
	put_2b((int) number);		/* Input line # */
	put_1b((int) con_expr);		/* Module # */
	module = ((UINT) con_expr);
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	do_noem						*/
/*									*/
/*	Handle the ".noemulate" pseudo-op.				*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
do_noem(parameter)
	SLONG			parameter;
	{
	emulate = FALSE;
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	do_norel					*/
/*									*/
/*	Handle the ".norel" pseudo-op.					*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
do_norel(parameter)
	SLONG			parameter;
	{
	rel_only = FALSE;
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	do_public					*/
/*									*/
/*	Handle the ".pub" pseudo-op.					*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
do_public(parameter)
	SLONG			parameter;
	{
	if(note_public())
		return;
	while(match(","))
		{
		if(note_public())
			return;
		}
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	do_rel						*/
/*									*/
/*	Handle the ".rel" pseudo-op.					*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
do_rel(parameter)
	SLONG			parameter;
	{
	rel_only = TRUE;
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	do_retf						*/
/*									*/
/*	Called by "pass0" when a ".retf" pseudo-op is encountered.	*/
/*	Sets the "retf_active" flag and remembers the pertinent data	*/
/*	for use when the ".retf" gets turned into a "ajw", "ret" pair.	*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
do_retf(parameter)
	SLONG			parameter;
	{
	if(operand(word_size))
		return;			/* If error in operand field */
	if(had_symbol || had_diff || had_dollar || had_amper || (! had_expr))
		{
		post_error(ERR_EXPR,NULL);
		return;
		}
	retf_active = TRUE;		/* Indicate we had a ".retf" */
	retf_number = number;		/* Current line # */
	retf_value = con_expr;		/* Amount to "ajw" */
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	do_set						*/
/*									*/
/*	Handle the constant equate pseudo-op.				*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
do_set(parameter)
	SLONG			parameter;
	{
	UINT			st_index;
	STRING			temp;
/*
 *	First parse symbol to be "set".
 */
	skip_space();			/* Skip blanks */
	temp = parse_ptr;		/* Remember symbol start */

	if(operand(word_size))
		return;		/* Parse symbol */
	if((! had_symbol) || had_diff || had_dollar || had_amper || had_expr)
		{
		post_error(ERR_EXPR,NULL);	/* Illegal symbol */
		return;
		}
/*
 *	Check for multiple definition errors.
 */
	if(st[st_index = st_num1].type & (S_DEFINED | S_EXTERN))
		{
		*parse_ptr = '\0';	/* Terminate it for error message */
		post_error(ERR_DUPL,temp);
		return;
		}
/*
 *	Now parse constant expression to set it to.
 */
	if(! match(","))
		{
		post_error(ERR_EXPR,NULL);	/* Need comma */
		return;
		}
	if(operand(word_size))
		return;		/* If expression error */
	if(had_symbol || had_diff || had_dollar || had_amper || (! had_expr))
		{
		post_error(ERR_EXPR,NULL);	/* Illegal expression */
		return;
		}
/*
 *	Looks OK, update symbol table status.  We indicate that the symbol is
 *	defined and is a constant value which is known at assembly time.
 */
	st[st_index].type |= (S_DEFINED | S_CONSTANT);
	st[st_index].min_value = st[st_index].max_value = con_expr;
/*
 *	Generate the T_SET record.
 */
	put_1b((int) T_SET);		/* Record type */
	put_2b((int) number);		/* Line # */
	put_2b((int) st_index);		/* Symbol # */
	put_4b((SLONG) con_expr);	/* Equated value of symbol */
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	do_sym						*/
/*									*/
/*	Handle the debug symbol information pseudo-op.			*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
do_sym(parameter)
	SLONG			parameter;
	{
	char			buf[SYMNMESIZE + 1];
	UINT			buflen;
	int			c;
	UINT			count;
	BOOL			lhads;
	register STRING		s;
	SLONG			value_e;
	UINT			value_s;
/*
 *	First parse debugging symbol name.
 */
	skip_space();			/* Skip blanks */
	if(symbol(*parse_ptr))
		{
		for(s = parse_ptr + 1; (symbol(*s) || isdigit(*s)); s++)
			;
		c = *s;			/* Remember the original terminator */
		*s = '\0';		/* Replace with string terminator */
		string_copy(buf,parse_ptr,SYMNMESIZE);
		buf[SYMNMESIZE] = '\0';
		buflen = strlen(buf);	/* Record string length */
		*s = c;			/* Replace with original terminator */
		parse_ptr = s;
		}
	else
		{
		post_error(ERR_EXPR,NULL);	/* Illegal symbol */
		return;
		}
/*
 *	Now parse assigned expression field.
 */
	if(! match(","))
		{
		post_error(ERR_EXPR,NULL);	/* Need comma */
		return;
		}
	if(operand(4))			/* Parse expression */
		return;
	if(had_diff || had_dollar || had_amper)
		{
		post_error(ERR_EXPR,NULL);	/* Illegal expression */
		return;
		}
	if(had_expr)
		value_e	= con_expr;
	else
		value_e = 0L;

	lhads = had_symbol;		/* Remember for later */
	if(lhads)
		value_s	= st_num1;
/*
 *	Now parse additional fields.  Do this twice, the first time to see if
 *	we have any errors and to see how many expressions are present, the
 *	second time to build the appropriate data records.
 */
	s = parse_ptr;			/* Remember start of expressions */
	for(count = 0; match(","); count++)
		{
		if(operand(4))
			return;		/* Expression error */
		if(had_symbol || had_diff || had_dollar || had_amper ||
			(! had_expr))
			{
			post_error(ERR_EXPR,NULL);
			return;
			}
		}
	parse_ptr = s;			/* Restore parsing pointer */
/*
 *	We now know the amount of storage that is being defined.  Generate the
 *	header part of the temp file record.
 */
	if(lhads)
		put_1b((int) T_DEBUGSYM_DATA);
	else
		put_1b((int) T_DEBUG_DATA);
	put_2b((int) number);		/* Input line # */
	put_4b((SLONG) value_e);	/* Constant expr/address offset */
	put_2b((int) (1 + buflen + (count * 4)));	/* Length of rest */
	if(lhads)
		put_2b((int) value_s);	/* Symbol table entry # */
	put_1b((int) buflen);		/* Debugging symbol name length */
	s = buf;
	while(*s)			/* Copy symbol name into record */
		put_1b((int) *s++);
/*
 *	Now go back through and re-evaluate the expressions, this time put the
 *	results in the temp file record.
 */
	while(match(","))
		{
		operand(4);		/* Parse expression */
		put_4b((SLONG) con_expr);	/* Add data field */
		}
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	do_val						*/
/*									*/
/*	Handle the ".val" pseudo-op.					*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
do_val(parameter)
	SLONG			parameter;
	{
	int			c;
	register	STRING	s;
	UINT			st_index;
	STRING			temp;
/*
 *	First parse symbol to be ".val"ed.
 */
	skip_space();			/* Skip blanks */
	temp = parse_ptr;		/* Remember symbol start */

	if(symbol(*parse_ptr))
		{
		for(s = parse_ptr+1; (symbol(*s)||isdigit(*s)); s++)
			;
		c = *s;			/* Remember the original terminator */
		*s = '\0';		/* Replace with string terminator */
		st_index = lookup(parse_ptr);	/* Get ST entry # */
		*s = c;			/* Replace with original terminator */
		parse_ptr = s;
		}
	else
		{
		post_error(ERR_EXPR,NULL);	/* Illegal symbol */
		return;
		}
/*
 *	Check for multiple definition errors.
 */
	if((! (st[st_index].type & S_VALUE)) &&
		(st[st_index].type & (S_DEFINED | S_EXTERN | S_PUBLIC)))
		{
		*parse_ptr = '\0';	/* Terminate it for error message */
		post_error(ERR_DUPL,temp);
		return;
		}
/*
 *	Now parse constant expression to set it to.
 */
	if(! match(","))
		{
		post_error(ERR_EXPR,NULL);	/* Need comma */
		return;
		}
	if(operand(word_size))
		return;			/* If expression error */
	if(had_symbol || had_diff || had_dollar || had_amper || (! had_expr))
		{
		post_error(ERR_EXPR,NULL);	/* Illegal expression */
		return;
		}
/*
 *	Looks OK, update symbol table status.  We indicate that the symbol is
 *	defined and has a new (or original), constant value which is known at
 *	assembly time.
 */
	st[st_index].type |= (S_DEFINED | S_VALUE);
	st[st_index].min_value = st[st_index].max_value = con_expr;
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	ds						*/
/*									*/
/*	Reserve a specified amount of storage (".ds").			*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
ds(parameter)
	SLONG			parameter;
	{
	if(operand(word_size))
		return;			/* If expression error */
	if(had_symbol || had_diff || had_dollar || had_amper || (! had_expr))
		{
		post_error(ERR_EXPR,NULL);	/* Illegal expression */
		return;
		}
/*
 *	Everything looks OK, generate appropriate temp file record.
 */
	put_1b((int) T_STORAGE);
	put_2b((int) number);		/* Input line # */
	put_4b((SLONG) con_expr);	/* # of bytes to reserve */
/*
 *	Update minimum and maximum location counters.
 */
	min_loc[module] += con_expr;
	max_loc[module] += con_expr;
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	dw						*/
/*									*/
/*	Handle define word (".dw", 2/4 bytes), pseudo-op.		*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#endif	/* MSC_OPT */

VOID
dw(parameter)
	SLONG			parameter;
	{
	UINT			count;
	UINT			length;
	STRING			temp;
/*
 *	First we pass over line and make sure no errors are lurking.
 */
	temp = parse_ptr;		/* Remember start of expressions */
	FOREVER
		{
		if(operand(word_size))
			return;		/* Bad expression */
		if(had_dollar)
			{
			post_error(ERR_EXPR,NULL);
			return;
			}
		if(! match(","))
			break;		/* If no more items on line */
		}
/*
 *	Now we pass over expressions and generate T_DATA records where we can
 *	and the appropriate relocatable records elsewhere.  Note that the
 *	following code combines all the pure data expressions which are
 *	adjacent into one larger T_DATA record to minimize the output record
 *	size.
 */
	parse_ptr = temp;		/* Reset to point at start of expr's */
	count = 0;			/* # of pure data exprs in row */
	FOREVER
		{
		operand(word_size);		/* Parse expression */
		if(! (had_diff || had_amper || had_symbol))
			{
			count++;	/* If pure data just count */
			if(! match(","))
				break;	/* If time to quit */
			}
		else			/* Otherwise, symbol was involved */
			{
			if(count)	/* First put out pure data if needed */
				{
				length = count * word_size;
				put_1b((int) T_DATA);	/* Record type */
				put_2b((int) number);	/* Line # */
				put_2b((int) length);	/* length */
				parse_ptr = temp;	/* Expression start */
				while(count)	/* For each "data" expr */
					{
					operand(word_size);	/* Parse expr */
					if(word_size == 2)
						put_2b((int) con_expr);
					else
						put_4b((SLONG) con_expr);
					match(",");	/* Slurp comma */
					count--;
					}
				min_loc[module] += length;
				max_loc[module] += length;
				operand(word_size);	/* Re-evaluate symbol expr */
				}
/*
 *	Now we need to figure out what type of relocatable data record to write
 *	out.
 */
			if(had_diff)		/* "rel-rel"? */
				{
				put_1b((int) T_RELREL_DATA);
				put_2b((int) number);	/* Input line # */
				put_2b((int) st_num1);	/* Left operand */
				put_2b((int) st_num2);	/* Right operand */
				if(had_expr)
					put_4b((SLONG) con_expr);
				else
					put_4b((SLONG) 0L);
				}
			else if(had_amper)	/* Did it start with a '@'? */
				{
				if(had_symbol)	/* Generate T_RELSYM_DATA */
					{
					put_1b((int) T_RELSYM_DATA);
					put_2b((int) number);	/* Line # */
					put_2b((int) st_num1);	/* ST entry# */
					if(had_expr)
						put_4b((SLONG) con_expr);
					else
						put_4b((SLONG) 0L);
					}
				else		/* Generate T_REL_DATA */
					{
					put_1b((int) T_REL_DATA);
					put_2b((int) number);	/* Line # */
					put_4b((SLONG) con_expr);
					}
				}
			else
				{
				put_1b((int) T_ADDR_DATA);
				put_2b((int) number);	/* Current line # */
				put_2b((int) st_num1);	/* ST entry # */
				if(had_expr)
					put_4b((SLONG) con_expr);
				else
					put_4b((SLONG) 0L);	/* No offset */
				}
			min_loc[module] += word_size;	/* Update loc cnts */
			max_loc[module] += word_size;
			if(! match(","))
				break;	/* If time to quit */
			temp = parse_ptr;	/* Update past symbol expr */
			}
		}
	if(count)			/* If we have some data expr's left */
		{
		length = count * word_size;
		put_1b((int) T_DATA);	/* Record type */
		put_2b((int) number);	/* Line # */
		put_2b((int) length);	/* length */
		parse_ptr = temp;	/* Expression start */
		while(count)		/* For each "data" expr */
			{
			operand(word_size);	/* Parse expr */
			if(word_size == 2)
				put_2b((int) con_expr);
			else
				put_4b((SLONG) con_expr);
			match(",");	/* Slurp comma */
			count--;
			}
		min_loc[module] += length;
		max_loc[module] += length;
		}
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#endif	/* MSC_OPT */

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	filename					*/
/*									*/
/*	Write a filename identification record to the temp file.	*/
/*									*/
/*	Parameters:							*/
/*		"new_num"	The starting line # in the new file.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
filename(new_num)
	UINT			new_num;
	{
	STRING			sptr;

	put_1b((int) T_FILENAME);	/* Filename type record */
	if(translator)			/* If original source code not asm */
		{
		put_2b((int) number);	/* Last line # before #line */
		put_2b((int) new_num);	/* Starting line # for current file */
		number = new_num;	/* Update line # counter */
		}
	else
		{
		put_2b((int) number - 1);	/* Last line # before #line */
		put_2b((int) new_num);	/* Starting line # for current file */
		number = new_num - 1;	/* Update line # counter */
		}
	put_1b((int) strlen(lifnbuf));	/* Filename length */
	sptr = lifnbuf;
	while(*sptr)			/* Put filename in record */
		put_1b((int) *sptr++);
	}

/************************************************************************/
/*									*/
/*	Routine:	flush_retf					*/
/*									*/
/*	Called by "pass0" when a label, EOF, or non-".retf" pseudo-op	*/
/*	is encountered.  If the value of "retf_active" is TRUE, this	*/
/*	routine spits out the required "ajw"/"ret" instructions as	*/
/*	needed.								*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
flush_retf()
	{
	if(! retf_active)
		return;			/* No ".retf" buffered up */
	retf_active = FALSE;		/* Or won't be for much longer */
	if(retf_value)
		data(retf_number,(UINT) AJW,retf_value);
	data(retf_number,(UINT) OPR, (SLONG) RET);
	}

/************************************************************************/
/*									*/
/*	Routine:	fpu_i						*/
/*									*/
/*	Handles the floating point "sequence" sub-instructions which	*/
/*	consist of a "ldc #", followed by a "fpentry".			*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	The constant to use with the "ldc".	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
fpu_i(parameter)
	SLONG			parameter;
	{
	UINT			length;
/*
 *	Generate data record with appropriate stuff.
 */
	put_1b((int) T_DATA);
	put_2b((int) number);		/* Input line # */
	length = i_size(LDC,parameter);	/* Figure length */
	put_2b((int) (length + FPENTRYSIZE));	/* Instruction length */
	build_i((UINT) length,(UINT) LDC,parameter);	/* Build "ldc" */
	build_i((UINT) FPENTRYSIZE,(UINT) OPR,(SLONG) FPENTRY);	/* "fpentry" */
/*
 *	Update minimum and maximum location counters.
 */
	min_loc[module] += (length + FPENTRYSIZE);
	max_loc[module] += (length + FPENTRYSIZE);
	}

/************************************************************************/
/*									*/
/*	Routine:	i_size						*/
/*									*/
/*	Determine the length of an instruction with immediate data.	*/
/*									*/
/*	Parameters:							*/
/*		"function" -	The basic instruction opcode.		*/
/*		"imm_data" -	The immediate data to scrutinize.	*/
/*									*/
/*	Returns:	The instruction length in bytes.		*/
/*									*/
/************************************************************************/

UINT
i_size(function,imm_data)
	UINT			function;
	register	SLONG	imm_data;
	{
	if((function == DOTLDC) && (imm_data <= -2146435074L))
		{
		imm_data &= 0x7FFFFFFFL;
		if(imm_data != 0L)
			return (MINT_SIZE + i_size(ADC,imm_data));
		else
			return (MINT_SIZE);
		}
	if(imm_data >= 0)
		{
		if(imm_data > 65535L)
			{
			if(imm_data > 16777215L)
				{
				if(imm_data > 268435455L)
					return (8);
				else
					return (7);
				}
			else
				{
				if(imm_data > 1048575L)
					return (6);
				else
					return (5);
				}
			}
		else
			{
			if(imm_data > 255)
				{
				if(imm_data > 4095)
					return (4);
				else
					return (3);
				}
			else
				{
				if(imm_data > 15)
					return (2);
				else
					return (1);
				}
			}
		}
	else
		{
		if(imm_data < -65536L)
			{
			if(imm_data < -16777216L)
				{
				if(imm_data < -268435456L)
					return (8);
				else
					return (7);
				}
			else
				{
				if(imm_data < -1048576L)
					return (6);
				else
					return (5);
				}
			}
		else
			{
			if(imm_data < -256)
				{
				if(imm_data < -4096)
					return (4);
				else
					return (3);
				}
			else	
				return (2);
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	nc1						*/
/*									*/
/*	A helping function for "nextc".  Reads input characters and	*/
/*	filters out NULL's.  Note that the file input library routines	*/
/*	already convert carriage-return, linefeed pairs into linefeeds	*/
/*	on input (for operating systems which require it).		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The next character read.			*/
/*									*/
/************************************************************************/

int
nc1()
	{
	int			c;

	FOREVER				/* Or at least a long time! */
		{
		if((c = getc(in_fp)) == '\0')
			continue;	/* Try again */
		if(c == '\f')
			continue;
		else if(c == EOF)
			{
			if(ferror(in_fp))
				fatal("Error reading input file: ",ifnbuf);
			}
		return (c);
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	nextc						*/
/*									*/
/*	Get the next character from the input file.  This routine makes	*/
/*	the assembly language input into a "standard" form:		*/
/*									*/
/*	1. Input lines which contain labels are split into a line	*/
/*	containing the label and one containing the rest of the		*/
/*	statement.							*/
/*	2. If EOF is detected within a line, a line terminator (NULL)	*/
/*	is prepended to make it appear on its own line.			*/
/*	3. One tab is present before a opcode/pseudo-op.		*/
/*	4. Opcodes/pseudo-ops are translated to upper case.		*/
/*	5. One tab is present between the opcode/pseudo-op and the	*/
/*	operand field (if an operand field exists).			*/
/*	6. If both a symbol and a constant expression are present in	*/
/*	the operand field they are seperated by a tab.			*/
/*	7. The character manipulations described in "nc1" also apply.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The next virtual character.			*/
/*									*/
/************************************************************************/

int
nextc()
    {
    static	UINT		state = BOL;	/* Start @ beginning-of-line */

    register 			int c;

    FOREVER
	{
	switch(state)			/* What happened last time? */
	    {
	    case BOL:			/* We are starting a new line */
		switch(c = nc1())	/* Get first character */
		    {
		    case '\n':
			if(translator)
				break;	/* Leave alone if src ! asm */
			number++;	/* Increment input line # */
			break;		/* Skip blank lines */

		    case ';':
			state = COMMENT;/* Whole line comment */
			break;		/* Skip comments */

		    case '#':		/* So preprocesor "#line" works */
		    case ' ':
		    case '\t':
			state = BOO;	/* Must be an opcode/pseudo-op */
			break;

		    case EOF:
			return (c);	/* We've found the end of the road */

		    default:
			state = LABEL;	/* Must be a label */
			return (c);
		    }
		break;

	    case COMMENT:		/* We are within a comment */
		state = BOL;
		while(c != '\n')	/* Skip comments */
		    {
		    if((c = nc1()) == EOF)
			return (c);
		    }
		if(translator)
		    break;		/* Leave # alone if src ! asm */
		number++;		/* Increment input line # */
		break;

	    case LABEL:			/* Doing a label now */
		switch(c = nc1())
		    {
		    case '\n':		/* The end of it by definition */
			state = NUMBER;	/* Increment input line # */
			return ('\0');

		    case ' ':
		    case '\t':
			(VOID) ungetc(c,in_fp);	/* Put separator back */
			state = BOL;
			return ('\0');

		    case ':':		/* Toss colons in labels */
			break;

		    case ';':
			state = COMMENT;/* Rest of line is comment */
			return ('\0');

		    case EOF:
			state = NUMBER;
			return ('\0');	/* Put on own line */

		    default:
			return (c);
		    }
		break;

	    case BOO:			/* We are trying to find the opcode */
		while((c == ' ') || (c == '\t'))
			c = nc1();
		switch(c)
		    {
		    case '\n':		/* Must have been a blank line */
			state = BOL;
			if(translator)
			    break;	/* Leave alone if src ! asm */
			number++;	/* Increment input line # */
			break;

		    case ';':
			state = COMMENT;/* Rest of line is comment */
			break;		/* Skip comments */

		    case EOF:		/* Funny place for it but hey? */
			state = BOL;
			return (c);

		    default:
			(VOID) ungetc(c,in_fp);	/* Put start of opcode back */
			state = OPCODE;
			return ('\t');
		    }
		break;

	    case OPCODE:		/* Within a opcode/pseudo-op */
		switch(c = nc1())
		    {
		    case '\n':
			state = NUMBER;	/* Increment input line # */
			return ('\0');

		    case ' ':
		    case '\t':		/* End of opcode/pseudo-op */
			state = BOOP;
			return ('\t');

		    case ';':
			state = COMMENT;/* Rest of line is comment */
			return ('\0');

		    case EOF:
			state = NUMBER;
			return ('\0');

		    default:
			return (islower(c) ? toupper(c) : c);
		    }

	    case BOOP:			/* Find beginning of operand field */
		c = nc1();
		while((c == ' ') || (c == '\t'))
			c = nc1();
		switch(c)
		    {
		    case '\n':
			state = NUMBER;	/* Increment input line # */
			return ('\0');

		    case ';':
			state = COMMENT;/* Rest of line is comment */
			return ('\0');

		    case EOF:
			state = NUMBER;
			return ('\0');

		    default:
			state = OPERAND;
			return (c);
		    }

	    case OPERAND:		/* Within operand field */
		switch(c = nc1())
		    {
		    case '\n':
			state = NUMBER;	/* Increment input line # */
			return ('\0');

		    case EOF:
			state = NUMBER;
			return ('\0');

		    default:
			return (c);
		    }

	    case NUMBER:		/* Update input line # */
		state = BOL;
		if(translator)
		    break;		/* Leave # alone if src ! asm */
		number++;
		break;
	    }
	}
    }

/************************************************************************/
/*									*/
/*	Routine:	note_extern					*/
/*									*/
/*	Update the symbol table with an external reference definition.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if an error occurred, FALSE otherwise.	*/
/*									*/
/************************************************************************/

BOOL
note_extern()
	{
	UINT			index;
	STRING			sptr;
	int			temp;

	sptr = parse_ptr;		/* Remember start of symbol */
	if(! symbol(*parse_ptr++))	/* Make sure it was a symbol! */
		{
		post_error(ERR_EXPR,NULL);	/* Otherwise nogoodnik */
		return (ERRORS);
		}
	while(*parse_ptr && (symbol(*parse_ptr) || isdigit(*parse_ptr)))
		parse_ptr++;		/* Find end of symbol */
	temp = *parse_ptr;		/* Record char past symbol end */
	*parse_ptr = '\0';		/* Replace with termination for now */
	index = lookup(sptr);		/* Find ST index for symbol */
	*parse_ptr = temp;		/* Replace original char */
	if(st[index].type & (S_DEFINED | S_PUBLIC))
		{
		post_error(ERR_DUPL,sptr);	/* Multiple definition error */
		return (ERRORS);
		}
	st[index].type |= S_EXTERN;	/* Remember its an external ref. */
	return (NOERRORS);
	}

/************************************************************************/
/*									*/
/*	Routine:	note_public					*/
/*									*/
/*	Update the symbol table with an public symbol definition.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if an error occurred, FALSE otherwise.	*/
/*									*/
/************************************************************************/

BOOL
note_public()
	{
	UINT			index;
	STRING			sptr;
	int			temp;

	sptr = parse_ptr;		/* Remember start of symbol */
	if(! symbol(*parse_ptr++))	/* Make sure it was a symbol! */
		{
		post_error(ERR_EXPR,NULL);	/* Otherwise nogoodnik */
		return (ERRORS);
		}
	while(*parse_ptr && (symbol(*parse_ptr) || isdigit(*parse_ptr)))
		parse_ptr++;		/* Find end of symbol */
	temp = *parse_ptr;		/* Record char past symbol end */
	*parse_ptr = '\0';		/* Replace with termination for now */
	index = lookup(sptr);		/* Find ST index for symbol */
	*parse_ptr = temp;		/* Replace original char */
	if(st[index].type & (S_EXTERN | S_VALUE))
		{
		post_error(ERR_DUPL,sptr);	/* Multiple definition error */
		return (ERRORS);
		}
	st[index].type |= S_PUBLIC;	/* Remember its a public definition */
	return (NOERRORS);
	}

/************************************************************************/
/*									*/
/*	Routine:	noti						*/
/*									*/
/*	Called in response to opcodes/pseudo-ops which are not yet	*/
/*	implemented.							*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
noti(parameter)
	SLONG			parameter;
	{
	post_error(ERR_NOTI,NULL);
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	operand						*/
/*									*/
/*	Parse the operand field of an instruction or pseudo-op.		*/
/*									*/
/*	If the field contains a leading "$" (word size scaling factor),	*/
/*	which cannot be fully evaluated (the remaining expression isn't	*/
/*	constant), "had_dollar" is set TRUE.  If the field contains a	*/
/*	"rel-rel" expression, "had_diff" is set TRUE, and the ST	*/
/*	identifiers are stored in "st_num1" and "st_num2".  Otherwise,	*/
/*	if the field had a ampersand, "had_amper" is set to TRUE.  If a	*/
/*	symbol is also present, "st_num1" is set to contain the symbol	*/
/*	table index returned by "lookup", and "had_symbol" is set to	*/
/*	TRUE.  Finally, if the field contained a constant expression	*/
/*	optionally following any of the other forms, set "had_expr" to	*/
/*	TRUE and return the value in "con_expr".			*/
/*									*/
/*	Parameters:							*/
/*		"num_size"	The # of bytes any associated numeric	*/
/*				expression must fit into.		*/
/*									*/
/*	Returns:	NOERRORS (FALSE), if the operand was OK, ERRORS	*/
/*			(TRUE), otherwise (the error will be "posted"	*/
/*			prior to "operand" returning).			*/
/*									*/
/************************************************************************/

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#endif	/* MSC_OPT */

BOOL
operand(num_size)
	UINT			num_size;
	{
	int			c;
	BOOL			local_amper;
	register	STRING	s;
	STRING			s0;
	STRING			s1;
	STRING			s2;
/*
 *	Do we have a leading '$'?
 */
	if(match("$"))
		{
		had_dollar = TRUE;
		skip_space();
		}
	else
		had_dollar = FALSE;
/*
 *	Do we have a '@'?
 */
	s0 = parse_ptr;			/* Remember start of operand field */
	if(match("@"))
		{
		local_amper = had_amper = TRUE;	/* We have a '@'! */
		skip_space();		/* Skip blanks */
		}
	else
		local_amper = had_amper = FALSE;	/* No '@' */
/*
 *	Time to check and see if we have a symbol.
 */
	s1 = parse_ptr;			/* Remember where first symbol began */
	if(symbol(*parse_ptr))
		{
		had_symbol = TRUE;	/* We've got a symbol! */
		for(s = parse_ptr+1; (symbol(*s)||isdigit(*s)); s++)
			;
		c = *s;			/* Remember the original terminator */
		*s = '\0';		/* Replace with string terminator */
		st_num1 = lookup(parse_ptr);	/* Get ST entry # */
		*s = c;			/* Replace with original terminator */
		parse_ptr = s;
		skip_space();		/* Skip blanks */
		}
	else
		had_symbol = FALSE;	/* No symbol */
/*
 *	Look ahead and see if we had another symbol with a leading "-", thus
 *	forming a "sym-sym" or "@sym-@sym" expression.  Note that in a
 *	"@sym-sym" expression, the second symbol is assumed to have a
 *	"constant" value and be part of the trailing constant expression.
 */
	had_diff = FALSE;		/* Assume FALSE initially */
	s2 = parse_ptr;			/* Remember where we were */
	FOREVER				/* To allow nested "break" */
		{
		if(had_symbol && match("-"))
			{
			if(! (match("@") || symbol(*parse_ptr++)))
				{
				parse_ptr = s2;	/* Reset parse pointer */
				break;
				}
			else		/* Was it legal rel-rel? */
				{
				parse_ptr--;	/* Backup past whatever was matched */
/* "@sym-@sym"? */		if(match("@"))
					{
					skip_space();	/* Skip whitespace */
					if((! symbol(*parse_ptr)) ||
						(! had_amper))
						{
						post_error(ERR_EXPR,NULL);
						return (ERRORS);
						}
					}
/* "sym-sym"? */		else
					{
/* "@sym-sym" */			if(had_amper)
						{
/* Assume it is a "constant" symbol */		parse_ptr = s2;
						break;
						}
					}
				had_diff = TRUE;	/* Assume so */
				had_amper = had_symbol = FALSE;
				for(s = parse_ptr + 1;
					(symbol(*s) || isdigit(*s)); s++)
					;
				c = *s;		/* Remember old terminator */
				*s = '\0';	/* Replace with '\0' */
				st_num2 = lookup(parse_ptr);	/* Get ST # */
				*s = c;		/* Put old terminator back */
				parse_ptr = s;
				skip_space();	/* Skip blanks */
				}
			}
		break;			/* Only one trip per customer */
		}
/*
 *	Okay, the above "sample" parse of the operand field may be too "strong"
 *	if some of the symbols involved are really constants.  Go through and
 *	figure out which are constants and note the appropriate simplifications
 *	which are possible (and back the parse pointer up so that the trailing
 *	constant expression evaluation will process the constant symbols).
 *
 *	First check for symbols which cannot appear anywhere but in constant
 *	expressions.  These symbols are the result of ".val" pseudo-ops.
 */
	if(had_diff || had_symbol)
		{
		if(st[st_num1].type & S_VALUE)
			{
			had_diff = had_symbol = FALSE;
			parse_ptr = s0;	/* Start over in operand field */
			}
		else if(had_diff && (st[st_num2].type & S_VALUE))
			{
			had_diff = FALSE;
			had_symbol = TRUE;
			parse_ptr = s2;	/* Keep first symbol */
			}
		}
/*
 *	Now check for symbols which are ".set" constants which CAN appear
 *	in general symbolic expressions.  However, we may be able to reduce
 *	the power of the relocatable expression...
 */
	if(had_diff && (! local_amper) && (st[st_num2].type & S_CONSTANT))
		{
		had_diff = FALSE;	/* Don't need this much power */
		if(st[st_num1].type & S_CONSTANT)
			parse_ptr = s1;
		else
			{
			parse_ptr = s2;	/* Keep first symbol */
			had_symbol = TRUE;
			}
		}
	else if(had_symbol && (st[st_num1].type & S_CONSTANT))
		{
		had_symbol = FALSE;	/* Don't need any symbols... */
		parse_ptr = s1;		/* Re-parse all except optional "@" */
		}
/*
 *	See if we have a trailing constant expression.
 */
	if((*parse_ptr == '\0') || (*parse_ptr == ',') || (*parse_ptr == ';'))
		{
		had_expr = FALSE;	/* Nope */
/*
 *	No '$' or '@' allowed alone.
 */
		if((had_dollar && (! (had_diff || had_symbol))) ||
			(had_amper && (! had_symbol)))
			{
			post_error(ERR_EXPR,NULL);
			return (ERRORS);
			}
		return (NOERRORS);	/* Looked OK */
		}
	else				/* We have an expression */
		{
		had_expr = TRUE;
		expr_err = FALSE;	/* Assume no expr errors */
		con_expr = eval();	/* Evaluate expression */
		if(had_dollar && (! (had_amper || had_diff || had_symbol)))
			{
			had_dollar = FALSE;
			if((con_expr % (SLONG) word_size) != 0L)
				{
				post_error(ERR_EXPR,NULL);
				return (ERRORS);
				}
			con_expr /= (SLONG) word_size;
			}
		if(expr_err || ((num_size == 2) &&
			((con_expr < (SLONG) -32768L) ||
				(con_expr > (SLONG) 65535L))))
			{
			post_error(ERR_EXPR,NULL);
			return (ERRORS);
			}
		return (NOERRORS);	/* Looking good */
		}
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#endif	/* MSC_OPT */

/************************************************************************/
/*									*/
/*	Routine:	opr_i						*/
/*									*/
/*	Handles the "operate" sub-instructions which have no operand	*/
/*	fields.								*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	The extended opcode encoding to use.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
opr_i(parameter)
	SLONG			parameter;
	{
/*
 *	Generate data record with appropriate stuff.
 */
	data((UINT) number,(UINT) OPR,(SLONG) parameter);
	}

/************************************************************************/
/*									*/
/*	Routine:	op_i						*/
/*									*/
/*	Handles instructions which reference immediate data (and thus	*/
/*	have operand fields), not including the explicit prefixing	*/
/*	instructions (implicit prefixing instructions are generated	*/
/*	where required).						*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	The basic function (opcode) to use.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
op_i(parameter)
	SLONG			parameter;
	{
/*
 *	First parse the operand to figure out whether it was a "rel-rel"
 *	symbolic reference (an absolute on anything except a Transputer...),
 *	or some other form of operand.
 */
	if(operand(word_size))
		return;			/* Return if expression error */
	if(had_diff)			/* "rel-rel"? */
		{
		if(had_dollar)
			put_1b((int) T_WRELREL_OP);
		else
			put_1b((int) T_RELREL_OP);
		put_2b((int) number);	/* Input line # */
		if(optimize == OLOW)
			put_1b((int) max_i_size);
		else
			put_1b((int) 1);	/* Minimum instr. length */
		put_2b((int) st_num1);	/* ST entry #1 (left operand) */
		put_2b((int) st_num2);	/* ST entry #2 (right operand) */
		if(had_expr)
			put_4b((SLONG) con_expr);	/* Offset */
		else
			put_4b((SLONG) 0L);
		put_1b((int) parameter);	/* Basic opcode */
/*
 *	Update minimum and maximum location counters, since we have not yet
 *	tied down the relative reference, the instruction could be anywhere
 *	from 1 to 8 bytes long (currently).
 */
		if(optimize == OLOW)
			min_loc[module] += max_i_size;
		else
			min_loc[module]++;
		max_loc[module] += max_i_size;
		}
	else if(had_dollar)		/* Not allowed for now */
		{
		post_error(ERR_EXPR,NULL);
		return;
		}
/*
 *	It wasn't a "rel-rel" record (with or without the '$' word size
 *	modifier), figure out what type of operand field we had and do
 *	whatever is required...
 */
	else if(had_amper)			/* Did it start with a '@'? */
		{
		if(had_symbol)		/* Generate T_RELSYM_OP record */
			{
			if(had_expr)
				{
				rel_sym((UINT) parameter,(UINT) st_num1,
					(SLONG) con_expr);
				}
			else
				{
				rel_sym((UINT) parameter,(UINT) st_num1,
					(SLONG) 0L);
				}
			}
		else			/* Generate T_REL_OP record */
			{
			put_1b((int) T_REL_OP);
			put_2b((int) number);	/* Input line # */
			if(optimize == OLOW)
				put_1b((int) max_i_size);
			else
				put_1b((int) 1);	/* Min. instr. len. */
			put_4b((SLONG) con_expr);	/* Offset */
			put_1b((int) parameter);	/* Basic opcode */
/*
 *	Update minimum and maximum location counters, since we have not yet
 *	tied down the relative reference the instruction could be anywhere
 *	from 1 to 8 bytes long (currently).
 */
			if(optimize == OLOW)
				min_loc[module] += max_i_size;
			else
				min_loc[module]++;
			max_loc[module] += max_i_size;
			}
		}
	else
		{
		if(had_symbol)		/* Generate T_ADDR_OP record */
			{
			put_1b((int) T_ADDR_OP);
			put_2b((int) number);	/* Input line # */
/*
 *	Set mininum and maximum location counters and instruction lengths
 *	as appropriate.
 */
			if(optimize == OLOW)
				{
				if(rel_only && (parameter == DOTLDC))
					{
					put_1b((int) (max_i_size + LDPI_SIZE));
					min_loc[module] +=
						(max_i_size + LDPI_SIZE);
					}
				else
					{
					put_1b((int) max_i_size);
					min_loc[module] += max_i_size;
					}
				}
			else
				{
				if(rel_only && (parameter == DOTLDC))
					{
					put_1b((int) (1 + LDPI_SIZE));
					min_loc[module] += (1 + LDPI_SIZE);
					}
				else
					{
					put_1b((int) 1);
					min_loc[module]++;
					}
				}
			if(rel_only && (parameter == DOTLDC))
				max_loc[module] += (max_i_size + LDPI_SIZE);
			else
				max_loc[module] += max_i_size;

			put_2b((int) st_num1);	/* ST entry # */
			if(had_expr)
				put_4b((SLONG) con_expr);	/* Offset */
			else
				put_4b((SLONG) 0L);
			if(rel_only && (parameter == DOTLDC))
				put_1b((int) DOTLDCR);
			else
				put_1b((int) parameter);	/* Opcode */
			}
		else			/* Generate T_DATA record */
			{
			if(! had_expr)
				{
				post_error(ERR_EXPR,NULL);
				return;
				}
			data((UINT) number,(UINT) parameter,(SLONG) con_expr);
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	parse						*/
/*									*/
/*	Read a line of code from the input into the parse buffer.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	FALSE if something was available from the input	*/
/*			file, TRUE if a hard EOF.			*/
/*									*/
/************************************************************************/

BOOL
parse()
	{
	register	int	c;
	register	UINT	count;
	register	STRING	cptr;

	cptr = parse_buf;		/* Place to put statement */
	count = 1;			/* Line length counter */
	while((c = nextc()) != EOF)	/* Read while we've got data */
		{
		if(c != '\0')
			{
			*cptr++ = c;
			if(count++ > (MAXLNSIZE - 1))
				fatal("Line too long in input file: ",ifnbuf);
			}
		else
			{
			*cptr = c;	/* Null terminate */
			return (FALSE);	/* We got something */
			}
		}
	if(ferror(in_fp))		/* Check if error or EOF */
		fatal("Error reading input file: ",ifnbuf);
	return (TRUE);			/* Hard EOF */
	}

/************************************************************************/
/*									*/
/*	Routine:	pass0						*/
/*									*/
/*	Convert the input source file into the internal file format	*/
/*	used by later passes.  Fill the ST with symbol definitions and	*/
/*	min/max location values.  This operation corresponds to the	*/
/*	traditional first pass of a two pass assembler.			*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
pass0()
	{
	register	int	lower;
	register	int	middle;
	register	int	result;
	register	int	upper;

#if	DEBUG
	putchar('0');			/* Indicate doing pass 0 */
#endif
	filename(1);			/* Initial file being read (line 1) */
	if(! translator)
		number++;		/* Counteract initial "filename" */
	FOREVER
		{
		if(parse())
			break;		/* Read next line (until EOF) */
/*
 *	First figure out if it is a label, or an opcode.
 */
		if(parse_buf[0] == '\t')
			{			/* Its an opcode/pseudo-op */
			for(parse_ptr = &parse_buf[1]; *parse_ptr; parse_ptr++)
				{
				if(*parse_ptr == '\t')
					{
/*	End of opcode */		*parse_ptr++ = '\0';
					break;
					}
				}
/*
 *	Recognize opcode/pseudo-op using a binary search of the opcode table
 *	and call the appropriate table routine to handle it.
 */
			lower = 0;		/* Lower bound of key */
			upper = (sizeof(op_info) / sizeof(struct opcode)) - 1;
			while(lower <= upper)	/* While table to search */
				{
/*	Key to compare */	middle = (lower + upper) >> 1;
				result = strcmp(&parse_buf[1],
						op_info[middle].name);
				if(result > 0)
					{
/*	Look in top half of table */	lower = middle + 1;
					continue;
					}
				else if(result < 0)
					{
/*	Check lower half of table */	upper = middle - 1;
					continue;
					}
				else			/* We have found it */
					{
/*	Legal for this CPU? */		if(cpu_type &
						op_info[middle].supported)
						{
/*	See if we have a ".retf" waiting */	if(retf_active)
							{
/*	Yep, output if we have a pseudo op */		if(parse_buf[1] == '.')
								{
/*	Unless it is another ".retf" */	if(strcmp(&parse_buf[1],".RETF") != 0)
						flush_retf();
					else
						break;
								}
/*	Toss normal statements */			else
								break;
							}
/*	Process statement */			(*op_info[middle].decode)
						(op_info[middle].parameter);
						}
/*	Nope, emulate it? */		else if(emulate)
						{
/*	Yes	*/				parse_buf[0] = '?';
						middle = lookup(parse_buf);
/*	Generate "call" */			rel_sym((UINT) CALL,
							(UINT) middle,
							(SLONG) 0L);
/*	Make ".extern" if needed */		if(st[middle].type &
							(S_DEFINED | S_PUBLIC))
								break;
						st[middle].type |= S_EXTERN;
						}
/*	Nope, give opcode error */	else
						{
						post_error(ERR_CODE,
							&parse_buf[1]);
						}
					break;
					}
				}
/*	Bad name */	if(lower > upper)
				post_error(ERR_CODE,&parse_buf[1]);
			}
/*
 *	If it wasn't an opcode/pseudo-op it must have been a label.  Add it
 *	to the symbol table with the appropriate status.
 */
		else
			{
			flush_retf();	/* Flush any ".retf" we are holding */
			middle = lookup(parse_buf);	/* Get/make ST entry */
/*
 *	Check for multiple definition errors.
 */
			if(st[middle].type & (S_DEFINED | S_EXTERN))
				post_error(ERR_DUPL,parse_buf);
/*
 *	Looks OK, update symbol table status.
 */
			else
				{
/*	Indicate defined */	st[middle].type |= S_DEFINED;
/*	Module # */		st[middle].module = module;
/*	Min. location */	st[middle].min_value = min_loc[module];
/*	Max. location */	st[middle].max_value = max_loc[module];

/*
 *	And generate a "definition" record for the temp file.
 */
				put_1b((int) T_DEF);
				put_2b((int) number);	/* Input line # */
				put_2b((int) middle);	/* ST entry # */
				}
			}
		}
	flush_retf();			/* Output possible trailing ".retf" */
	post_eof();			/* Generate EOF record */
	}

/************************************************************************/
/*									*/
/*	Routine:	pfix_i						*/
/*									*/
/*	Handles positive and negative prefix instructions.  These	*/
/*	instructions are not usually explicitly written but are usually	*/
/*	implied implicitly by instructions whose immediate data		*/
/*	occupies more than 4 bits.  When used explicitly these		*/
/*	instructions are limited to a 0-15 value constant expression	*/
/*	with no symbolic references allowed.				*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	The basic function (opcode) to use.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
pfix_i(parameter)
	SLONG			parameter;
	{
/*
 *	First parse the prefix expression.
 */
	if(operand(word_size))
		return;			/* Bad expression */
/*
 *	Make sure prefix is within legal range.
 */
	if(had_symbol || had_diff || had_dollar || had_amper || (! had_expr) ||
		(con_expr < 0) || (con_expr > 15))
		{
		post_error(ERR_EXPR,NULL);
		return;
		}
/*
 *	Looks OK, generate the instruction.
 */
	put_1b((int) T_DATA);
	put_2b((int) number);		/* Input line # */
	put_2b((int) 1);		/* Instruction length */
	put_1b((int) (parameter | con_expr));
/*
 *	Update minimum and maximum location counters.
 */
	min_loc[module]++;
	max_loc[module]++;
	}

/************************************************************************/
/*									*/
/*	Routine:	post_eof					*/
/*									*/
/*	Write a EOF record to the temp file.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
post_eof()
	{
	put_1b((int) T_EOF);		/* Generate EOF record */
	put_2b((int) number);		/* Remember input line # */
	}

/************************************************************************/
/*									*/
/*	Routine:	post_error					*/
/*									*/
/*	Called to record syntax errors during the conversion of the	*/
/*	input file into the file format used with the temp files.	*/
/*									*/
/*	Parameters:							*/
/*		"error_code"	The exact problem.			*/
/*		"sptr"		An optional string pointer.  NULL if	*/
/*				not used.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
post_error(error_code,sptr)
	UINT			error_code;
	STRING			sptr;
	{
	STRING			str;

	errors++;			/* Update total # of errors */
	put_1b((int) T_RESERVED);	/* Indicate this is an error record */
	put_2b((int) number);		/* Remember input line # */
	switch(error_code)
		{
		case	ERR_DUPL:
				str = "Duplicate symbol definition";
				put_1b((int) 0);	/* Assume 0 length */
				break;

		case	ERR_EXPR:
				str = "Expression field error";
				if(optimize == OLOW)
					{
/* Assume max length? */		put_1b((int) max_i_size);
					min_loc[module] += max_i_size;
					max_loc[module] += max_i_size;
					}
				else
					{
/* Assume length 1? */			put_1b((int) 1);
					min_loc[module]++;
					max_loc[module]++;
					}
				break;

		case	ERR_FLOAT:
				str = "Floating point constant error";
				put_1b((int) 0);	/* Assume 0 length */
				break;

		case	ERR_NOTI:
				str = "Not implemented (pseudo-op)";
				put_1b((int) 0);	/* Assume 0 length */
				break;

		case	ERR_CODE:
				str = "Opcode/pseudo-op unknown";
				put_1b((int) 0);	/* Assume 0 length */
				break;

		case	ERR_UDEF:
				str = "Undefined symbol";
				if(optimize == OLOW)
					{
/* Assume max length? */		put_1b((int) max_i_size);
					min_loc[module] += max_i_size;
					max_loc[module] += max_i_size;
					}
				else
					{
/* Assume length 1? */			put_1b((int) 1);
					min_loc[module]++;
					max_loc[module]++;
					}
				break;
		}
	put_1b((int) error_code);	/* Put error code in record */
	if(sptr == NULL)
		fprintf(STDERR,"<%s> @ %u: %s\n",lifnbuf,number,str);
	else
		fprintf(STDERR,"<%s> @ %u: %s: %s\n",lifnbuf,number,str,sptr);
	}

/************************************************************************/
/*									*/
/*	Routine:	pp_line						*/
/*									*/
/*	Handles the preprocessor "#line" directive and causes the file	*/
/*	name and line number info to be automatically updated so that	*/
/*	the listing matches the input even with preprocessor handled	*/
/*	include files.  The information is also written to the output	*/
/*	file which is used by the linker so that the information is	*/
/*	available for a future source level debugger.			*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
pp_line(parameter)
	SLONG			parameter;
	{
	UINT			temp;

	if(operand(word_size))
		return;
	if(had_symbol || had_diff || had_dollar || had_amper || (! had_expr))
		{
		post_error(ERR_EXPR,NULL);	/* Illegal symbol */
		return;
		}
	if(match("\""))			/* New filename also? */
		{
		if(! compress)
			{
			for(temp = 0; temp < FNSIZE; temp++)
				{
				if((*parse_ptr == '\0')||(*parse_ptr == '\"'))
					break;
				lifnbuf[temp] = *parse_ptr++;
				}
			lifnbuf[temp] = '\0';
			filename((UINT) con_expr);	/* Note change */
			}
		}
	else				/* Record new source line # */
		{
		if(translator)
			number = ((UINT) con_expr);
		else
			number = ((UINT) con_expr) - 1;
		}
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	rel_sym						*/
/*									*/
/*	Handle T_RELSYM_OP record generation.				*/
/*									*/
/*	Parameters:							*/
/*		"opcode"	The basic instruction opcode.		*/
/*		"sym_index"	The symbol table entry # to use.	*/
/*		"offset"	The constant offset to use.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
rel_sym(opcode,sym_index,offset)
	UINT			opcode;
	UINT			sym_index;
	SLONG			offset;
	{
	put_1b((int) T_RELSYM_OP);
	put_2b((int) number);		/* Input line # */
	if(optimize == OLOW)
		put_1b((int) max_i_size);
	else
		put_1b((int) 1);	/* Minimum instruction length */
	put_2b((int) sym_index);		/* ST entry # */
	put_4b((SLONG) offset);		/* Offset */
	put_1b((int) opcode);		/* Basic opcode */
/*
 *	Update minimum and maximum location counters, since we have not yet
 *	tied down the relative reference the instruction could be anywhere
 *	from 1 to 8 bytes long (currently).
 */
	if(optimize == OLOW)
		min_loc[module] += max_i_size;
	else
		min_loc[module]++;
	max_loc[module] += max_i_size;
	}

/************************************************************************/
/*									*/
/*	Routine:	skip_space					*/
/*									*/
/*	Skip whitespace in the parsing buffer.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
skip_space()
	{
	while(isspace(*parse_ptr))
		parse_ptr++;
	}

/************************************************************************/
/*									*/
/*	Routine:	str_out						*/
/*									*/
/*	Write the data in a "C" style string constant to the output	*/
/*	temp file.  This routine assumes that the string is correctly	*/
/*	formed (checked by a previous call to "str_size").		*/
/*									*/
/*	Parameters:	"terminator" -	TRUE if a '\0' byte terminator	*/
/*					is desired, FALSE otherwise.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
str_out(terminator)
	BOOL			terminator;
	{
	register	int	c;
	register	UINT	count;
	register	int	value;

	while((c = *parse_ptr++) != '\"')
		{
		if(c == '\\')
			{
			switch(c = *parse_ptr++)
				{
				case 'a':	c = '\a'; break;
				case 'b':	c = '\b'; break;
				case 'f':	c = '\f'; break;
#if	H_OPSYS == MPW
				case '\012':	c = '\012'; break;
#else	/* H_OPSYS == MPW */
				case 'n':	c = '\n'; break;
#endif	/* H_OPSYS == MPW */
				case 'r':	c = '\r'; break;
				case 't':	c = '\t'; break;
				case 'v':	c = '\v'; break;
				case '\\':	c = '\\'; break;
				case '\'':	c = '\''; break;
				case '\"':	c = '\"'; break;
				case '?':	c = '?'; break;

/*	Octal escape */		case '0': case '1': case '2': case '3' :
				case '4': case '5': case '6': case '7' :
					for(count = 2, value = c - '0';
						(isoct(c = *parse_ptr) &&
						(count <= 3)); count++,
							parse_ptr++)
						{
						value = (value << 3) + c - '0';
						}
					c = value;
					break;

/*	Hex escape */		case 'x':
					value = (int) hexbin(*parse_ptr++);
					for(count = 2;
						(ishex(c = *parse_ptr) &&
						(count <= 3)); count++,
							parse_ptr++)
						{
						value = (value << 4) +
							((int) hexbin(c));
						}
					c = value;
					break;
				}
			}
		put_1b((int) c);	/* Write a string byte */
		}
	if(terminator)
		put_1b((int) '\0');	/* Zero terminate */
	}

/************************************************************************/
/*									*/
/*	Routine:	str_size					*/
/*									*/
/*	Determine how many bytes it will take to store a "C" string and	*/
/*	put the length in "con_expr".					*/
/*									*/
/*	Parameters:	"terminator" -	TRUE if a '\0' byte terminator	*/
/*					is desired, FALSE otherwise.	*/
/*									*/
/*	Returns:	TRUE if any error was encountered and "posted",	*/
/*			FALSE otherwise.				*/
/*									*/
/************************************************************************/

BOOL
str_size(terminator)
	BOOL			terminator;
	{
	register	int	c;
	register	UINT	count;
	register	UINT	length;

	if(terminator)
		length = 1;
	else
		length = 0;
	for(; (c = *parse_ptr++) != '\"'; length++)
		{
		if(c == '\\')
			{
			switch(*parse_ptr++)
				{
				case 'a':	/* Is it a character escape? */
				case 'b':
				case 'f':
				case 'n':
				case 'r':
				case 't':
				case 'v':
				case '\\':
				case '\'':
				case '\"':
				case '?':
					break;

/*	Octal escape */		case '0': case '1': case '2': case '3' :
				case '4': case '5': case '6': case '7' :
					for(count = 2;
						(isoct(*parse_ptr) &&
						(count <= 3)); count++,
							parse_ptr++)
						;
					break;

/*	Hex escape */		case 'x':
					if(! ishex(*parse_ptr++))
						{
						post_error(ERR_EXPR,NULL);
						return (ERRORS);
						}
					for(count = 2;
						(ishex(c = *parse_ptr) &&
						(count <= 3)); count++,
							parse_ptr++)
						;
					break;

				default:
					post_error(ERR_EXPR,NULL);
					return (ERRORS);
				}
			}
		else
			{
			if(c == '\0')
				{
				post_error(ERR_EXPR,NULL);
				return (ERRORS);
				}
			}
		}
	con_expr = length;		/* String length */
	return	(NOERRORS);		/* Everything OK */
	}

/************************************************************************/
/*									*/
/*	Routine:	symbol						*/
/*									*/
/*	Determine if the parameter could be the start of a symbol ("_",	*/
/*	"?" or an alpha).						*/
/*									*/
/*	Parameters:							*/
/*		"c"	The character to test.				*/
/*									*/
/*	Returns:	TRUE if so, FALSE otherwise.			*/
/*									*/
/************************************************************************/

BOOL
symbol(c)
	char			c;
	{
	if(isalpha(c) || (c == '_') || (c == '?'))
		return (TRUE);
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	t212						*/
/*									*/
/*	Handle the ".t212" pseudo-op.  This selects the legal opcodes	*/
/*	for the T212/T222/T225 Transputer processors.			*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
t212(parameter)
	SLONG			parameter;
	{
	cpu_type = T212 | ALL;		/* Select T212/T222/T225 opcodes */
	max_i_size = MAX_I_SIZE_16B;	/* Maximum instruction length */
	word_size = WORD_SIZE_16B;	/* Word size in bytes */
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	t414						*/
/*									*/
/*	Handle the ".t414" pseudo-op.  This selects the legal opcodes	*/
/*	for the T414/T425 Transputer processors.			*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
t414(parameter)
	SLONG			parameter;
	{
	cpu_type = T414 | ALL;		/* Select T414/T425 legal opcodes */
	max_i_size = MAX_I_SIZE_32B;	/* Maximum instruction length */
	word_size = WORD_SIZE_32B;	/* Word size in bytes */
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	t800						*/
/*									*/
/*	Handle the ".t800" pseudo-op.  This selects the legal opcodes	*/
/*	for the T800/T801/T805 Transputer processors.			*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
t800(parameter)
	SLONG			parameter;
	{
	cpu_type = T800 | ALL;		/* Select T800/T801/T805 opcodes */
	max_i_size = MAX_I_SIZE_32B;	/* Maximum instruction length */
	word_size = WORD_SIZE_32B;	/* Word size in bytes */
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/************************************************************************/
/*									*/
/*	Routine:	the_end						*/
/*									*/
/*	Handle the ".end" pseudo-op.  Anything after this in the input	*/
/*	file is ignored.						*/
/*									*/
/*	Parameters:							*/
/*		"parameter"	A dummy parameter.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

VOID
the_end(parameter)
	SLONG			parameter;
	{
	while(getc(in_fp) != EOF)
		;	/* Suck it dry */
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */
