/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: tcx1.c,v 2.2 92/04/21 11:11:05 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	TCX Compiler (1 of 11)
 *
 *	Copyright 1987-1991 by Quality Computer Systems and Logical Systems.
 *	All rights reserved.
 *
 *	Published by:	Logical Systems
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	03/17/87
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<string.h>
#include	"tcx.h"
#define	MAIN				/* Actually generate storage */
#include	"tcxdecl.h"
#include	"tcxext.h"

#ifdef	__TURBOC__
extern	unsigned	_stklen = 32767;
#endif	/* __TURBOC__ */

/*
 *	Local function declarations.
 */
PRIVATE	VOID			inittable();
PRIVATE	VOID			parse();
PRIVATE	VOID			emit_globals();
PRIVATE	VOID			summary();
PRIVATE	struct	id_node		*make_glb();
PRIVATE	struct	id_node		*make_loc();
PRIVATE	struct	id_node		*make_tag();
PRIVATE	INT32			initvar();
PRIVATE	INT32			initvhelp();
PRIVATE	INT32			initaggreg();
PRIVATE	INT32			addbitf();
PRIVATE	INT32			initbitf();
PRIVATE	INT32			skip_ds();
PRIVATE	INT32			initstr();
PRIVATE	INT32			initchar();
PRIVATE	INT32			initword();
PRIVATE	VOID			add_init();

/*
 *	Top level compiler operational outline.
 */
PUBLIC	int
main(argc,argv)
	int			argc;
	char			*argv[];
	{
	getoptions(argc,argv);		/* Get user options		*/
	maininit();			/* Main level initialization	*/
	p_init();			/* Initialize peephole optimizer*/
	getspace();			/* Get table and buffer space	*/
	getinfil();			/* Open first/only input file	*/
	getoutfil();			/* Open output/temp files	*/
	inittable();			/* Initialize compiler tables	*/
	kill_line();			/* Clear parsing line		*/
	parse();			/* Parse program		*/
	flushout();			/* Flush diversions		*/
	p_flush();			/* Flush peephole optimizer	*/
	closeout();			/* Close output/temp files	*/
	p_summary();			/* Print peephole opt. summary	*/
	summary();
	if((Errcnt != 0) && Errdel)
		deleteout();		/* Remove output file		*/
	exit((Errcnt == 0) ? GOODEXIT: BADEXIT);
	return (0);
	}

/*
 *	Initialize compiler tables.
 */
PRIVATE	VOID
inittable()
	{
	inittypes();			/* Initialize type table */
	}

/*
 *	Process all input text.
 */
PRIVATE	VOID
parse()
	{
	struct scope_info	old_scope;
	int			sc;	/* This used to be "static" -- why? */
	struct id_node		*type;

	enter_scope(&old_scope);
/*
 *	If needed, add the function entry/exit trace functions to the list
 *	of external variables.
 */
	if(Tracein)
		(VOID) add_loc(Triname,Ifunctype,SC_EXTERN,NS_VAR,
			(INT32) (FUNC_NODE | REFERENCED_NODE));
	if(Traceout)
		(VOID) add_loc(Troname,Ifunctype,SC_EXTERN,NS_VAR,
			(INT32) (FUNC_NODE | REFERENCED_NODE));
/*
 *	Parse the file.
 */
	blanks();
	while(! Eof)
		{
		if(isdecl(&sc,&type,Inttype))
			{		/* Is it a declaration? */
			switch ((int) sc)
				{	/* Check storage class */
				case SC_REG:
				case SC_AUTO:
					scerr();	/* Bad storage class */

				case SC_NONE:	/* Use default */
					sc = SC_GLOBAL;
					break;

				case SC_STATIC:
					sc = SC_ST_GLB;
					break;
				}
			declglb(sc,type);
			}
		else			/* Must be a function definition */
			declglb(SC_GLOBAL,Inttype);
		blanks();
		}

	if(Ncmp)			/* Anything left hanging? */
		error("Unbalanced braces");

/*
 *	Scan for any undefined externals at this GLOBAL level.  We MUST do
 *	this before the final exit_scope() so that the symbols transfered are
 *	still around for emit_globals() to see.
 */
	scan_ripup(xfer_externs);

	/* Emit all file-level externs for things not defined */
	scan_global(emit_globals);
	scan_literals(emit_literals);

	exit_scope(&old_scope);

	del_literals();		/* For completeness, delete the literal list */
	}

/*
 *	Emit external declarations.
 */
PRIVATE VOID
emit_globals(sym)
	struct	id_node		*sym;
	{
	struct	id_node		*p;

	if(((p = lookup_node(sym->name, NS_OTHER)) == NULL) ||
	        (p->info.storage_class == SC_EXTERN))
		{
		instruction(TOP_EXT,ABS,sym->name,NULL,(INT32) 0,NULL);
		}
	}

/*
 *	Report statistics and errors.
 */
PRIVATE	VOID
summary()
	{
	char			s[20];

	if(Verbose || (Errcnt != 0))
		{
		if(Errcnt == 0)
			fprintf(STDERR,"No errors found\n");
		else
			{
			sprintf(s,"%d",(int) Errcnt);
			fprintf(STDERR,"%s error%s found\n",s,
				((Errcnt == 1) ? "" : "s"));
			}
		}
	}

/*
 *	Check for declarations.
 *
 *	Used by:
 *		parse (externals)
 *		decltag (struct/union members)
 *		procarg (arguments)
 */
PUBLIC	BOOL
isdecl(sc,type,deftype)
	int			*sc;
	struct	id_node		**type;
	struct	id_node		*deftype;
	{
	register int		holdsc;
	register struct	id_node	*holdtype;
	register int		rsc;
	register struct	id_node	*rtype;

	holdsc = 0;
	holdtype = NULL;
	do
		{
		if((rsc = getsc()) != 0)
			{		/* Storage class? */
			if(holdsc == 0)	/* First one? */
				holdsc = rsc;	/* Remember it */
			else
				scerr();
			}

		if(holdtype && (istypedef() != NULL))
			break;		/* Local redefinition */
					/* of typedef name */

		if((rtype = gettype()) != NULL)
			{
			if(holdtype == NULL)
			holdtype = rtype;
			else
				typerr();
			}
		} while (rsc || rtype);	/* Loop if either found */

	*sc = holdsc;
	if(holdsc && (!holdtype))	/* If storage class but */
		*type = deftype;	/* no type, use default */
	else
		*type = holdtype;

	return (holdsc || holdtype);
	}

/*
 *	Declare a global variable.
 */
PUBLIC	VOID
declglb(sc,type)
	int			sc;
	struct	id_node		*type;
	{
	BOOL			func_def;
	char			name[NAMEMAX+1];

	if(! matchc(';'))
		{
		do
			{
			(VOID) declvar(name,type,sc,make_glb,&func_def);

			if(! func_def)
				dvrt_out(dvrt_get());	/* Write div. out */

			} while((! func_def) && matchc(','));

		if(! func_def)
			ns();
		}
	}

PRIVATE struct id_node *
make_glb(name,t,sc,label)
	char			*name;
	struct	id_node		*t;
	int			sc;
	LABEL			label;
	{
	struct	id_node		*sym;

	if((t->type & FUNC_NODE) && (sc != SC_TYPEDEF))
		return (procfunc(name,t,sc));

	if((sym = add_loc(name,t,sc,NS_VAR,t->type)) != NULL)
		{
		/* Set surrogate symbol # */
		sym->address.surrogate = label;
		mark_scope_used();
		lineline(Lineno);
		}
	else
		{
		sym = lookup_node(name,NS_VAR);	/* Find the symbol */

		if(sc == SC_EXTERN)
			{
			if(! same_type(t,sym->type_ptr.base_type,
				/* ptr_eq_array = */ FALSE))
				{
				errname("Extern type redefinition",name);
				}
			}
		else if(sc == SC_TYPEDEF)
			{
			if(! same_type(t,sym->type_ptr.base_type,
				/* ptr_eq_array = */ FALSE))
				{
				errname("Typedef redefinition",name);
				}
			}
		else if((sym->info.storage_class == SC_EXTERN) &&
			same_type(t,sym->type_ptr.base_type,
				/* ptr_eq_array = */ FALSE))
			{
			/* Just change storage class to not external */
			sym->info.storage_class = sc;

			/* Set surrogate symbol # */
			sym->address.surrogate = label;
			mark_scope_used();
			lineline(Lineno);
			}
		else
			multidef(name);
		}

	return (sym);
	}

/*
 *	Declare local variables.
 */
PUBLIC	VOID
declloc(sc,type)
	register int		sc;
	struct	id_node		*type;
	{
	BOOL			func_def;
	char			name[NAMEMAX+1];

	if(! matchc(';'))
		{
		do
			{
			(VOID) declvar(name,type,sc,make_loc,&func_def);
			} while(matchc(','));

		ns();			/* Check for semicolon */
		}
	}

/*
 *	Do work to create local var.
 */
PRIVATE	struct	id_node	*
make_loc(name,t,sc,label)
	char			*name;
	struct	id_node		*t;
	int			sc;
	LABEL			label;
	{
	int			locsc;
	struct	id_node		*sym;

	if((t->type & FUNC_NODE) && (sc != SC_ST_EXT))
		locsc = SC_EXTERN;
	else
		locsc = sc;

	if((sym = add_loc(name,t,locsc,NS_VAR,t->type)) != NULL)
		{
		sym->address.surrogate = label;	/* Set surrogate symbol # */
		if(locsc == SC_STATIC)
			mark_scope_used();
/*
 *	I think this is correct -- get size of this node to be the size of
 *	the base type.
 */
		sym->val.integral = t->val.integral;
		}
	else
		{
		sym = lookup_node(name,NS_VAR);	/* Find the symbol */

		if((locsc == SC_EXTERN) &&
			(sym->info.storage_class == SC_EXTERN))
			{
			if(! same_type(t,sym->type_ptr.base_type,
				/* ptr_eq_array = */ FALSE))
				{
				errname("Extern type redefinition",name);
				}
			}
		else if(locsc == SC_TYPEDEF)
			{
			if(! same_type(t,sym->type_ptr.base_type,
				/* ptr_eq_array = */ FALSE))
				{
				errname("Typedef redefinition",name);
				}
			}
		else
			multidef(name);
		}
	return (sym);
	}

/*
 *	Get required array size.
 */
PUBLIC	INT32
arraysize()
	{
	struct	op_node		*cexp;
	register INT32		size;

	if(matchc(']'))
		return (-1);	/* Null size */

	cexp = constexp();

	if(! ISCONSTANT(cexp))
		{
		cexperr();		/* Not a constant expression */
		size = 1L;
		}
	else
		{
		size = cexp->right.ptr.terminal->val.integral;
		if(size < 0L)
			{
			error("Negative array size illegal");
			size = -size;
			}
		else if(size == 0L)
			{
			error("Zero array size illegal");
			size = 1L;
			}
		}
	needpunc(']');
	free_op_node(cexp);
	return (size);
	}

/*
 *	Declare (or find) a structure/union tag.
 */

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#endif	/* MSC_OPT */

PUBLIC	struct id_node *
decltag(typecode)
	INT32			typecode;	/* Type fld frm id_node.type */
	{
	int			bitposition;
	struct	op_node		*exp;
	BOOL			func_def;
	register int		hastag;
	register struct id_node	*mp;
	register struct id_node	*mt;
	register INT32		offset;
	register INT32		size;
	register struct id_node	*type;
	int			width;

	int			curr_id_space;
	struct	id_node		*memlist;
	int			memsc;
	struct	id_node		*memtype;
	char			name[NAMEMAX+1];

	type = NULL;
	hastag = FALSE;

	if(! nextc('{'))
		{			/* Should be tag name */
		if(! symname(name))
			illname();
		else
			{
			type = find_tag(name);	/* Tag declared? */
			hastag = TRUE;
			}
		}
	if(type == NULL)
		{			/* First appearance? */
		if(! nextc('{'))	/* If so must be definition */
			{
			if(! hastag)
				error("Structure/union definition needed");
			}
		type = add_loc(hastag ? name : NULL,(struct id_node *) NULL,
			SC_NONE,NS_TAG,typecode);
		}

	if(matchc('{'))
		{			/* If there was a tag, we have it */
		if(type->val.integral != 0)
			errname("Redefinition of structure/union tag",name);
		else
			{
			/* Assign name space */
			curr_id_space = type->name_space =
				new_name_space(TRUE);
			}

		offset = size = bitposition = 0;
		memlist = NULL;		/* No member list yet */

		while(isdecl(&memsc,&memtype,Inttype))
			{
			if(memsc != SC_NONE)
				error("Member can't have storage class");
			do
				{
				blanks();
				if(ch() == ':')
				    {
				    mt = memtype;	/* Blank ':' tag */
				    name[0] = '\0';	/* And a blank name */
				    }
				else
				    {
				    mt = declvar(name,memtype,SC_NONE,make_tag,
						 &func_def);
				    }
/*
 *	We have already parsed the storage class (Should be NONE),
 *	the type (limit it to unsigned int and signed int) and the
 *	name. Check for a ':' and if found, parse the bit-field
 *	width.
 */
				if(matchc(':'))
					{
			if((mt->type &
				(TYPE_MASK & ~SIGNED_NODE & ~UNSIGNED_NODE))
					!= INT_NODE)
						{
				error("Bit-fields must be type \"int\"");
						mt = Uinttype;
						}

					exp = constexp();
					width = (int) exp->right.ptr.terminal->
						val.integral;

					if(! ISCONSTANT(exp))
						{
				error("Bit-field must be constant integral");
						width = -1;
						}
					else if((width < 0) ||
						(width > Sizeof_int * 8))
						{
					error("Bit-field size out of range");
						width = -1;
						}
					else if(width == Sizeof_int * 8)
						width = -1;
					else
						{
						mt = add_type(
							(BITFIELD_NODE |
							INT_NODE |
				    (mt->type & (SIGNED_NODE|UNSIGNED_NODE))),
							    NULL,(INT32) 0);

						mt->val.bitfield.fieldwidth =
							(char) width;
						}
					free_op_node(exp);
					}
				else
					width = -1;	/* Not a bitfield */

				if((mt == type) || (mt->type & FUNC_NODE))
					{
					errname("Can't be a member",name);
					continue;
					}
				else if((name[0] != '\0') &&
					(lookup_node(name,curr_id_space)))
					{
					redefmem(name);
					}
				else if((mp =
					add_loc((name[0]!='\0') ? name : NULL,
						mt,NS_MEMBER,curr_id_space,
						(INT32) MEMBER_NODE)) != NULL)
					{
					if(width >= 0)
						{
						/* Doing a bitfield def */
						offset = (offset +
							Sizeof_int - 1) &
/*lint -e502 Doesn't like signed complement */
							~(Sizeof_int - 1);
/*lint +e502 Doesn't like signed complement */

						if(((width == 0) &&
							(bitposition != 0)) ||
							((bitposition + width)
							> (Sizeof_int * 8)))
							{
/*
 *	Overflowing bitfield to next boundary.
 */
							bitposition = 0;
							offset += Sizeof_int;
							}
						mt->val.bitfield.bitoffset =
							(char) bitposition;
						bitposition += width;
						}
					else
						{
/* Round up to char */				offset += (bitposition + 7) /
							8;
						bitposition = 0;
						}
/*
 *	Word align everything except chars & char arrays.
 */
					if(! ((mt->type & CHAR_NODE) ||
						((mt->type & ARRAY_NODE) &&
						(mt->type_ptr.base_type->type &
							CHAR_NODE))))
						{
						offset = (offset + Sizeof_int
/*lint -e502 Doesn't like signed complement */
							- 1) & ~(Sizeof_int-1);
/*lint +e502 Doesn't like signed complement */
						}
					mp->address.offset = offset;
					mp->info.next_member = NULL;

					if(memlist == NULL)
						{
						type->type_ptr.member_head =
							memlist = mp;
						}
					else
						{
						memlist->info.next_member = mp;
						memlist = mp;
						}
					}
				if(typecode & STRUCT_NODE)
					{
					if(! (mt->type & BITFIELD_NODE))
						offset += mt->val.integral;
					}
/* Union */			else if(mt->val.integral > size)
					{
					size = mt->val.integral;
					offset = bitposition = 0;
					}
				} while(matchc(','));
			ns();
			}
		needpunc('}');
		/* Fill in structure/union size padded through last word */
		if(typecode & STRUCT_NODE)
			size = offset + (bitposition + 7) / 8;

		if((type->val.integral = (size + Sizeof_int - 1) &
/*lint -e502 Doesn't like signed complement */
			~(Sizeof_int - 1)) == 0)
/*lint +e502 Doesn't like signed complement */
			{
			error("No members specified");
			}
		}
	return (type);
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#endif	/* MSC_OPT */

/*
 *	Placeholder routine for "declvar" call.
 */
PRIVATE struct	id_node	*
make_tag()
	{
	return (NULL);
	}

/*
 *	Declare (or find) an enum tag.
 */
PUBLIC	struct	id_node	*
enumtag()
	{
	register struct id_node	*em;
	register struct id_node	*type;
	register int		hastag;
	register INT32		enumnext;
	struct	op_node		*cexp;

	char			name[NAMEMAX+1];

	type = NULL;
	hastag = FALSE;

	if(! nextc('{'))
		{			/* Should be tag name */
		if(! symname(name))
			illname();
		else
			{
			type = find_tag(name);	/* Tag declared? */
			hastag = TRUE;
			}
		}
	if(type == NULL)
		{			/* First appearance */
		type = add_loc(hastag ? name : NULL,(struct id_node *) NULL,
			SC_NONE,NS_TAG,(INT32) ENUM_NODE);
		type->val.integral = Sizeof_int;
		}
	else if((type->type & ENUM_NODE) == 0)
		errname("Redefinition of tag",name);

	if(matchc('{'))
		{			/* If there was a tag, we have it */
		enumnext = 0;		/* Next enum value */

		do
			{
			if(! symname(name))
				{
				illname();
				break;
				}
			else if(matchc('='))
				{
				/* Shift gears -- user supplying an expr */
				cexp = constexp();	/* Get expression */

				if(ISCONSTANT(cexp))
					{
					enumnext = cexp->right.ptr.terminal->
						val.integral;
					}
				else
					error("Invalid enum constant");

				free_op_node(cexp);	/* Don't need expr */
				}

			if((em = add_loc(name,Inttype,SC_CONSTANT,NS_VAR,
				(INT32) INT_NODE | SIGNED_NODE)) != NULL) 
				{
				em->val.integral = enumnext++;	/* Give val */
				}
			else
				errname("Enum member already defined",name);

			} while(matchc(','));

		needpunc('}');
		}
	return (type);
	}

/*
 *	Declare a variable.
 */
PUBLIC	struct	id_node	*
declvar(name,type,sc,make_var,func_def)
	char			*name;
	register struct id_node	*type;
	register int		sc;
	struct	id_node		*(*make_var)();
	BOOL			*func_def;
	{
	struct	id_node		*sym;
	struct	id_node		*ts;
	register int		do_init;
	register int		putds;
	LABEL			label;

	ts = NULL;			/* Initialize activation info */
	*name = '\0';			/* No name to start with */
	dodecl(&ts,name,TRUE);		/* Analyze the declarator */
	if(*name == '\0')
		illname();		/* Needed a name */
	type = loadtype(&ts,type,sc);
	if((type->type & VOID_NODE) && ((type->type_ptr.base_type == NULL) ||
		(type->type_ptr.base_type->type != FUNC_NODE)))
		{
		error("Can't have void variable");
		}
	do_init = matchc('=');		/* Check for initialization */
	putds = FALSE;

	if(type->type & FUNC_NODE)
		{
#ifdef NON_TPUTER
		align();		/* Some machines need alignment */
#endif		/* NON_TPUTER */
		if(do_init)
			error("Can't initialize function");
		if(sc == SC_STATIC)	/* I.E. local static func decl */
			sc = SC_ST_EXT;
		sym = (*make_var)(name,type,sc,(LABEL) 0);
		}
	else if((sc == SC_GLOBAL) || (sc == SC_ST_GLB))
		{
		if(do_init)
			segment(SEG_INIT);
		else
			segment(SEG_UNINIT);
		chk_align(type);
		sym = (*make_var)(name,type,sc,(LABEL) 0);	/* Make sym */
		glblabel(name,sc);
		putds = TRUE;
		}
	else if(sc == SC_STATIC)
		{
		dvrt_stack();		/* Save current diversion */
		dvrt_append(Static_dvrt);	/* Pick up all old statics */
		if(do_init)
			segment(SEG_INIT);
		else
			segment(SEG_UNINIT);
		chk_align(type);
		printlabel(label = getlabel());
		sym = (*make_var)(name,type,sc,label);	/* Create symbol */
		putds = TRUE;
		}
	else
		{
		if(do_init && ((sc == SC_AUTO) || (sc == SC_REG)) &&
			(type->type & (SU_NODE | ARRAY_NODE)))
			{
			set_literal();
			chk_align(type);
			printlabel(label = getlabel());
			}
		else
			label = 0;
		sym = (*make_var)(name,type,sc,label);	/* Make var */
		if(do_init && (sc != SC_AUTO) && (sc != SC_REG))
			{
			error("Cannot initialize this storage class");
			segment(SEG_INIT);
			}
		}
	if(do_init)
		(VOID) initvar(sym,type);
	else if(putds)			/* Default initialization */
		(VOID) definit(type->val.integral);

	if((type->type & ARRAY_NODE) && (type->val.integral <= 0) &&
		(sc != SC_EXTERN) && (sc != SC_ARG))
		{
		sizerr();
		type->val.integral = Sizeof_int;
		}

	if(sc == SC_STATIC)
		Static_dvrt = dvrt_unstack();	/* Return to norm diversion */
	else if(do_init && ((sc == SC_AUTO) || (sc == SC_REG)) &&
		(type->type & (SU_NODE | ARRAY_NODE)))
		{
		set_code();
		}
	if(sym == NULL)			/* Was it a function definition? */
		*func_def = TRUE;
	else 
		*func_def = FALSE;
	return (type);
	}

/*
 *	Do initialization for one variable.
 */
PRIVATE	INT32
initvar(sym,tp)
	struct	id_node		*sym;
	register struct	id_node	*tp;
	{
#ifdef	DEBUG
	char			buf[80];
#endif	/* DEBUG */
	struct	op_node		*exp;	/* Holds data re: constant expr	*/
	INT32			init;
	struct	id_node		*source;

	if((sym->type & (SU_NODE | ARRAY_NODE)) &&
		((sym->info.storage_class == SC_AUTO) ||
			(sym->info.storage_class == SC_REG)))
		{
/*
 *	First check for a auto/reg struct/union initializer which is a
 *	value returned by a function call or assignment.
 */
		blanks();
		if((ch() != '{') && (sym->type & SU_NODE))
			{
			exp = constexp();
			if(! same_type(exp->node_type,tp,/*ptr_eq_array*/
				FALSE))
				{
				error("Illegal aggregate initializer");
				}
			}
		else
/*
 *	Otherwise get a normal "piecewise" struct/union/array initializer.
 */
			{
			init = initvhelp(sym,tp,/* in_aggregate = */ FALSE);
			source = add_type(sym->type,sym->type_ptr.base_type,
				sym->val.integral);
			source->info.storage_class = SC_LITERAL;
			source->address.surrogate = sym->address.surrogate;
			exp = terminal(source);
			}
		add_init(sym,exp);
		}
	else
		init = initvhelp(sym,tp,/* in_aggregate = */ FALSE);
#ifdef	DEBUG
	if(Debug && ((sym->info.storage_class == SC_STATIC)	||
		(sym->info.storage_class == SC_ST_GLB)		||
		(sym->info.storage_class == SC_GLOBAL)))
		{
		sprintf(buf,"; Initialized %ld bytes",(long) init);
		outcomment(buf);
		}
#endif
	return (init);
	}

PRIVATE INT32
initvhelp(sym,tp,in_aggregate)
	struct	id_node		*sym;
	register struct	id_node	*tp;
	BOOL			in_aggregate;
	{
	BOOL			match_brace;
	struct	id_node		*memb_ptr;
	INT32			thisinit;

	lineline(Lineno);

	match_brace = matchc('{');
	if(tp->type & UNION_NODE)
		{
/*
 *	Initialize as first union member.  Protect ourselves in case union
 *	doesn't have any members.
 */
		if((memb_ptr = tp->type_ptr.member_head) != NULL)
			{
			thisinit = initvhelp(sym,memb_ptr->type_ptr.base_type,
				/* in_aggregate = */ TRUE);
			if(thisinit < tp->val.integral)
				{
				thisinit += skip_ds(sym,
					tp->val.integral-thisinit);
				}
			}
		}
	else if(tp->type & (STRUCT_NODE | ARRAY_NODE))
		thisinit = initaggreg(sym,tp);
	else if(tp->type & CHAR_NODE)
		thisinit = initchar(sym,in_aggregate);
	else
		thisinit = initword(sym,tp,in_aggregate);

	if(match_brace)
		{
		(VOID) matchc(',');
		if(! matchc('}'))
			{
			error("Too many initializers");
/*
 *	Try to get back in sync.
 */
			while((! endst()) && (ch() != ';'))
				{
				if(gch() == '}')
					break;
				}
			}
		}
	return (thisinit);
	}

/*
 *	Initialize an aggregate.
 */

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#endif	/* MSC_OPT */

PRIVATE	INT32
initaggreg(sym,tp)
	struct	id_node		*sym;
	register struct	id_node	*tp;
	{
	BOOL			bitfactive;	/* TRUE if init. bitfield */
	INT32			bitfinit;	/* Init. bitfield accum. */
	register struct	id_node	*currmem;
	INT32			initloc;
	BOOL			more;
	char			name[NAMEMAX+1];
	BOOL			resize;
	register	INT32	size;
	struct	id_node		*type;
	register	INT32	typecode;

	static	int		level = 0;	/* Must be aggreg if level>1 */

	++level;
	size = tp->val.integral;	/* Size of this item */
	resize = (size == -1);		/* TRUE if unknown size */
	typecode = tp->type;

	if(typecode & STRUCT_NODE)
		{
		currmem = tp->type_ptr.member_head;
		bitfactive = FALSE;	/* Assume first member !bitfield */
		bitfinit = 0L;		/* Accumulator for bitfields */
		}
	else if(typecode & ARRAY_NODE)
		type = tp->type_ptr.base_type;
	else if(level > 1)
		error("Only aggregates can be initialized this way");

	initloc = 0;
	do
		{
		if(typecode & STRUCT_NODE)
			{
			if(currmem == NULL)
				{
/*
 *	Error recovery -- Someone else will report the error.
 */
				initloc +=
					initword(sym,tp,
						/* in_aggregate = */ TRUE);
				}
			else
				{
/*
 *	If we have a unnamed bitfield member skip it since it can't be
 *	initialized.
 */
				while((currmem != NULL) &&
					(currmem->type_ptr.base_type->type &
						BITFIELD_NODE) &&
					(currmem->name == NULL))
					{
					currmem = currmem->info.next_member;
					}
				if(currmem == NULL)
					{
					if(bitfactive)
						initloc += initbitf(bitfinit);
					break;
					}
/*
 *	If we've previously encountered a bitfield make sure that the next
 *	member is in the same "int" (ie. another bitfield), otherwise write
 *	out the vector of bits accumulated so far.
 */
				if(bitfactive &&
					(currmem->address.offset != initloc))
					{
					bitfactive = FALSE;
					initloc += initbitf(bitfinit);
					}
/*
 *	Skip any space remaining from partial initialization of a previous
 *	member.
 */
				initloc +=
					skip_ds(sym,currmem->address.offset -
						initloc);
/*
 *	If we have a bitfield initializer add it in the accumulator.
 */
				if((type = currmem->type_ptr.base_type)->type &
					BITFIELD_NODE)
					{
					if(! bitfactive)
						{
						bitfactive = TRUE;
						bitfinit = 0L;
						}
					bitfinit = addbitf(type,bitfinit);
					}
				else
/*
 *	Do non-bitfield structure initializer.
 */
					{
					initloc += initvhelp(sym,type,
						/* in_aggregate = */ TRUE);
					}
				currmem = currmem->info.next_member;
				}
			if(currmem == NULL)
				{
				if(bitfactive)
					initloc += initbitf(bitfinit);
				break;
				}
			}
		else
/*
 *	Must be of type ARRAY_NODE.
 */
			{
			if((type->type & CHAR_NODE) && matchc('\"'))
				{
				initloc += initstr(tp);
				break;
				}
			else
				{
				initloc += initvhelp(sym,type,
						/* in_aggregate = */ TRUE);
				}
			
			if((! resize) && (initloc >= size))
				break;
			}
		more = matchc(',');	/* If TRUE, may be more inits */
		blanks();

		if((ch() == '}') || (ch() == ';') || ((chksym(name) != NULL) &&
			(strcmp(name,"sizeof") != 0) && chkeyword()))
			{
			more = FALSE;	/* Nothing more here */
			}
		else if(! more)
			{
			if(isdigit(ch()) || issymf(ch()))
				{
				needpunc(',');	/* More inits, so give error */
				more = TRUE;	/* Pretend there is more */
				}
			else
				{
				error("Garbage characters in initialization");
/*
 *	Try to get back in sync.
 */
				while((ch() != ',') && (! endst()) &&
					(ch() != '}'))
					{
					(VOID) gch();
					}
				}
			}
		} while(more);		/* Loop if more to examine */
/*
 *	END OF LOOP
 */
	if(resize)			/* Fix size in type table */
		sym->val.integral = tp->val.integral = initloc;
	else if(size > initloc)		/* Default init if any left */
		{
		if((sym->info.storage_class == SC_STATIC)	||
			(sym->info.storage_class == SC_GLOBAL)	||
			(sym->info.storage_class == SC_ST_GLB)	||
			(sym->info.storage_class == SC_AUTO)	||
			(sym->info.storage_class == SC_REG))
			{
			(VOID) definit(size - initloc);
			}
		initloc = size;
		}
	--level;

	return(initloc);
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#endif	/* MSC_OPT */

/*
 *	Add a bitfield initialization to "bitfinit" (the accumulator for an
 *	initialized bitvector).
 */
PRIVATE INT32
addbitf(tp,bitfinit)
	struct	id_node		*tp;
	INT32			bitfinit;
	{
	struct	op_node		*exp;	/* Holds data re: constant expr	*/
	INT32			bitfsmask;
	INT32			bitfumask;
	INT32			val;	/* Holds result of constant expr */

	exp = constexp();
	if((! ISCONSTANT(exp)) || (EXPRTYPE(exp) & FP_NODE))
		error("Bitfield initializer must be an integral constant");
	else
		{
		val = exp->right.ptr.terminal->val.integral;
/*lint -e502 Doesn't like signed complement */
		bitfumask = ~((1L << tp->val.bitfield.fieldwidth) - 1L);
		bitfsmask = ~((1L << (tp->val.bitfield.fieldwidth - 1)) - 1L);
/*lint +e502 Doesn't like signed complement */
		if(tp->type & UNSIGNED_NODE)
			{
			if((bitfumask & val) != 0L)
				{
			error("Bitfield initializer too large for field");
				val = 0L;
				}
			}
		else
			{
			if((val & bitfsmask) !=
				((val >= 0L) ? 0L : (-1L & bitfsmask)))
				{
		error("Bitfield initializer not representable in field width");
				val = 0L;
				}
			}
/*lint -e502 Doesn't like signed complement */
		bitfinit |= ((val & (~bitfumask)) <<
			tp->val.bitfield.bitoffset);
/*lint +e502 Doesn't like signed complement */
		}
	free_op_node(exp);

	return (bitfinit);
	}

/*
 *	Write the accumulator for a bitfield vector ("bitfinit") to the
 *	output stream.
 */
PRIVATE INT32
initbitf(bitfinit)
	INT32			bitfinit;
	{
	instruction(TOP_DW,ABS,NULL,NULL,bitfinit,NULL);

	return (Sizeof_int);
	}

/*
 *	If the symbol is a static or static-global or auto/reg  aggregate
 *	variable, generate a skip (e.g. via .DS pseudo-op) the appropriate
 *	number of bytes.
 */
PRIVATE	INT32
skip_ds(sym,distance)
	struct	id_node		*sym;
	INT32			distance;
	{
	if((sym->info.storage_class == SC_STATIC)	||
		(sym->info.storage_class == SC_ST_GLB)	||
		(sym->info.storage_class == SC_GLOBAL)	||
		(sym->info.storage_class == SC_AUTO)	||
		(sym->info.storage_class == SC_REG))
		{
		return (definit(distance));
		}
	else
		return ((INT32) 0);
	}

/*
 *	Initialize an array of characters with a string.
 */
PRIVATE	INT32
initstr(tp)
	struct	id_node		*tp;
	{
	char			buf[80];
	register	int	c;
	char			*cur;
	char			*end;
	register	INT32	size;
	register	INT32	str_size;
	register	int	warned;

	size = tp->val.integral;
	warned = FALSE;

	cur = buf;
	end = buf + (sizeof(buf) - 1);

	for(str_size = 0; (c = getstring()) != -1; )
		{
		if(str_size == size)
			{
			if(! warned)
				{
				error("String is bigger than array");
				warned = TRUE;
				}
			}
		else
			{
			if((str_size % MAXDBCH) != 0)
				cur = safecat(cur,",",end);
			cur = safedec(cur,(INT32) c,end);

			if(((++str_size % MAXDBCH) == 0) &&
				((size == -1) || (str_size < size)))
				{
				*cur = '\0';
				outline(TOP_DB,buf,NULL);
				cur = buf;
				end = buf + (sizeof(buf) - 1);
				}
			}
		}
	if((size == -1) || (str_size < size))
		{
		if((str_size % MAXDBCH) != 0)
			cur = safecat(cur,",",end);
		cur = safecat(cur,"0",end);	/* End of string */
		str_size++;
		}
	*cur = '\0';
	outline(TOP_DB,buf,NULL);

	return (str_size);
	}

/*
 *	Initialize a character scalar.
 */
PRIVATE	INT32
initchar(sym,in_aggregate)
	struct	id_node		*sym;
	BOOL			in_aggregate;
	{
	struct	op_node		*exp;
	register	int	val;

	exp = constexp();

	if(ISCONSTANT(exp))
		{
		if(sym->type & SIGNED_NODE)
			{
			if((exp->right.ptr.terminal->val.integral < -128) ||
				(exp->right.ptr.terminal->val.integral > 127))
				{
				warning(1,"Signed char initializer error");
				}
			}
		else
			{
			if(exp->right.ptr.terminal->val.uintegral > 255)
				warning(1,"Unsigned char initializer error");
			}
		}

	if((! in_aggregate) &&
		((sym->info.storage_class == SC_AUTO)	||
	        (sym->info.storage_class == SC_REG)))
		{
		add_init(sym,exp);
		}
	else
		{
		if(ISCONSTANT(exp))
			{
			val = (int) (exp->right.ptr.terminal->val.integral &
				0xFF);
			}
		else if(isptrclass(exp))
			{
			error("Char cannot hold address");
			val = 0;
			}
		else
			{
			initerr();
			val = 0;
			}
		instruction(TOP_DB,ABS,NULL,NULL,(INT32) val,NULL);
		free_op_node(exp);
		}

	return ((INT32) SIZEOF_CHAR);
	}

/*
 *	Initialize one or more words.
 *
 *	The parameter "tp" specifies the size of the type to be
 *	initialized (e.g. INT_NODE) using the "type" field bits in id_node.
 */
PRIVATE INT32
initword(sym,tp,in_aggregate)
	struct	id_node		*sym;
	struct	id_node		*tp;
	BOOL			in_aggregate;
	{
	BOOL			addrflag;
	char			buf[NAMEMAX+1];
	struct	op_node		*exp;	/* Holds data re: constant expr	*/
	struct	id_node		*esym;
	double			fval;
	struct	op_node		*op;
	INT32			val;	/* Holds result of constant expr */

	exp = constexp();

	if((! in_aggregate) &&
		((sym->info.storage_class == SC_AUTO)	||
		(sym->info.storage_class == SC_REG)))
		{
		add_init(sym,exp);
		}
	else
		{
		if(ISCONSTANT(exp))
			{
			if(EXPRTYPE(exp) & FP_NODE)
				fval = exp->right.ptr.terminal->val.fvalue;
			else
				val = exp->right.ptr.terminal->val.integral;
			esym = NULL;
			}
		else if(is_simple(exp))
			{
			esym = constaddr(exp,&val,&addrflag)->
				right.ptr.terminal;
			if(! addrflag)
				error("Not a pointer, but assumed so");
			}
		else if((exp->operator == OP_CONVERT) && isptr(exp) &&
			is_simple(RIGHT(exp)))
			{
			esym = constaddr(RIGHT(exp),&val,&addrflag)->
				right.ptr.terminal;
			}
/*
 *	Note that in the future we should arrange that "binaryop" make OP_SUB
 *	into OP_ADD of -X when X is a numeric constant.  We should also go
 *	through and replace the calls to "is_simple" with something that will
 *	make TCX more portable (cases where more can be done in the way of
 *	addressing modes for normal instructions than can be done in external
 *	initialization...)
 */
		else if(((exp->operator == OP_ADD) ||
				(exp->operator == OP_SUB)) &&
			isptr(exp) &&
			(is_simple(op = LEFT(exp)) ||
				((LEFT(exp)->operator == OP_CONVERT) &&
					is_simple(op = RIGHT(LEFT(exp))))) &&
			ISRIGHTCONSTANT(exp))
			{
			esym = constaddr(op,&val,&addrflag)->
				right.ptr.terminal;
			if(exp->operator == OP_ADD)
				val += RIGHTCONSTANT(exp);
			else
				val -= RIGHTCONSTANT(exp);
			}
		else
			{
			initerr();
			fval = val = 0;
			esym = NULL;
			}
		if(esym != NULL)
			esym->type |= REFERENCED_NODE;	/* It was referenced */

		if(tp->type & FP_NODE)
			{
			if(EXPRTYPE(exp) & FP_NODE)
				sprintf(buf,"%.20g",fval);
			else
				sprintf(buf,"%ld",val);

			if(tp->val.integral > SIZEOF_FLOAT)
				{
				instruction(TOP_REAL64,ABS,buf,NULL,(INT32) 0,
					NULL);
				}
			else
				{
				instruction(TOP_REAL32,ABS,buf,NULL,(INT32) 0,
					NULL);
				}
			}
		else
			{
			if(EXPRTYPE(exp) & FP_NODE)
				val = f_to_l(fval,TRUE);

			instruction(TOP_DW,ABS,
				(esym != NULL) ? real_name(esym,buf) : NULL,
				NULL,val,
#ifdef	DEBUG
				(Debug && (esym != NULL) &&
					strcmp(real_name(esym,buf),
						esym->name)) ?
							esym->name : NULL
#else
				NULL
#endif	/* DEBUG */
				);
			}

		if((esym != NULL ) && (esym->diversion != NULL))
			{
			dvrt_append(esym->diversion);
			esym->diversion = NULL;
			}
		free_op_node(exp);
		}
	return ((INT32) tp->val.integral);
	}

/*
 *	The only alignment considerations are that words be word aligned.
 *	Characters (and arrays of characters) do not have to be aligned.
 *	Structures should probably be word aligned.  Thus, the rule is
 *	that if the base type (after skipping all array nodes) is a
 *	CHAR_NODE, then no alignment is necessary, otherwise align to
 *	a word boundary.
 */
VOID
chk_align(type)
	struct	id_node		*type;
	{
	while(type->type & ARRAY_NODE)
		type = type->type_ptr.base_type;
/*
 *	Always generate .ALIGN if not a CHAR type node.
 */
	if(! (type->type & CHAR_NODE))
		align();
	}

/*
 *	Add an expression to the auto initialization list.
 */
PRIVATE	VOID
add_init(sym,exp)
	struct	id_node		*sym;
	struct	op_node		*exp;
	{
	struct	op_node		*assignment;

	assignment = binaryop(OP_ASSIGN,lterminal(sym),exp);

	if(Initlist != NULL)
		Initlist = binaryop(OP_SEQ,Initlist,assignment);
	else
		Initlist = assignment;
	}

/* End of TCX1.C */
