/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: tcx5.c,v 2.2 92/04/21 11:11:12 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	TCX Compiler (5 of 11)
 *
 *	Copyright 1987-1991 by Quality Computer Systems and Logical Systems.
 *	All rights reserved.
 *
 *	Published by:	Logical Systems
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	03/17/87
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<string.h>
#include	<math.h>
#include	"tcx.h"
#include	"tcxdecl.h"
#include	"tcxext.h"

/*
 *	Local function declarations.
 */
PRIVATE	INT32			chktype();
PRIVATE	int			chksc();
PRIVATE	VOID			dodecl1();
PRIVATE	VOID			dodecl2();
PRIVATE	VOID			stack_type();
PRIVATE	INT32			c_hex();
PRIVATE	INT32			c_octal();
PRIVATE	INT32			c_decimal();
PRIVATE	INT32			charconst();
PRIVATE	int			escape();

/*
 *	Local data declarations.
 */
PRIVATE	TEMP	tempcount = 0;		/* Max number of temps per func */

/*
 *	This section identifies tokens in declarations and manages all of the
 *	compiler tables.
 */

/*
 *	Check next token to see if it is a keyword.
 */
PUBLIC BOOL
chkeyword()
	{
	char			name[NAMEMAX + 1];

	return ((chksym(name) != NULL) &&
		((lookup_node(name,NS_CKEYWORD) != NULL) ||
		(lookup_node(name,NS_CTYPES) != NULL) ||
		(lookup_node(name,NS_CSTORAGE) != NULL)));
	}

/*
 *	Check for a valid C variable type.
 *
 *	The symbol table has a name space (NS_CTYPES) for type symbols.
 *	The "type" of these symbols is not important, and is uninitialized.
 *	The integral cell contains the type mask, as that is where it is easy
 *	to put "user defined" stuff when manipulating the symbol table
 *	entry.  Later, it might make some sense to put the type information
 *	in the type cell.	DCR	4 Jun 87
 */
PRIVATE	INT32
chktype()
	{
	struct	id_node		*item;
	char			name[NAMEMAX + 1];

	if(chksym(name) == NULL)
		return ((INT32) 0);

	if((item = lookup_node(name,NS_CTYPES)) != NULL)
		return (item->val.integral);
	else
		return ((INT32) 0);
	}

/*
 *	Check for a valid C variable storage class.
 */
PRIVATE	int
chksc()
	{
	struct	id_node		*item;
	char			name[NAMEMAX + 1];

	if(chksym(name) == NULL)
		return (0);

	if((item = lookup_node(name,NS_CSTORAGE)) != NULL)
		return ((int) item->val.integral);
	else
		return (SC_NONE);
	}

/*
 *	Get a valid C variable type if present.
 */
PUBLIC	struct	id_node	*
gettype()
	{
	register INT32		newtype;
	struct	id_node		*rval;
	register INT32		type;

	type = 0;			/* Start with no type */
	for(;;)
		{			/* Process all type specifiers */
		newtype = chktype();
		if(newtype & (VOID_NODE | FLOAT_NODE | ENUM_NODE |
			STRUCT_NODE | UNION_NODE | SIGNED_NODE |
				UNSIGNED_NODE))
			{
			if(type == 0)
				type = newtype;
			else
				{
				type = newtype;
				typerr();
				}
			}
		else if(newtype & CHAR_NODE)
			{
			if(type == 0)
				{
				type = (Cpu_schar ? (CHAR_NODE | SIGNED_NODE)
					: (CHAR_NODE | UNSIGNED_NODE));
				}
			else if(type & ~(UNSIGNED_NODE | SIGNED_NODE))
				{
				type = CHAR_NODE | SIGNED_NODE;
				typerr();
				}
			else
				type |= CHAR_NODE;
			}
		else if(newtype & SHORT_NODE)
			{
			if(type == 0)
				type = SHORT_NODE | SIGNED_NODE;
			else if(type & ~(UNSIGNED_NODE | SIGNED_NODE))
				{
				type = SHORT_NODE | SIGNED_NODE;
				typerr();
				}
			else
				type |= SHORT_NODE;
			}
		else if(newtype & INT_NODE)
			{
			if(type == 0)
				type = INT_NODE | SIGNED_NODE;
			else if(type & (SHORT_NODE | LONG_NODE))
				;
			else if(type & ~(UNSIGNED_NODE | SIGNED_NODE))
				{
				type = INT_NODE | SIGNED_NODE;
				typerr();
				}
			else
				type |= INT_NODE;
			}
		else if(newtype & LONG_NODE)
			{
			if(type == 0)
				type = LONG_NODE | SIGNED_NODE;
			else if(type & ~(UNSIGNED_NODE | SIGNED_NODE))
				{
				type = LONG_NODE | SIGNED_NODE;
				typerr();
				}
			else
				type |= LONG_NODE;
			}
		else if(newtype & DOUBLE_NODE)
			{
			if(type == 0)
				type = newtype;
			else if((type & (LONG_NODE | SIGNED_NODE |
				UNSIGNED_NODE)) == (LONG_NODE | SIGNED_NODE))
				{
				type = LDOUBLE_NODE;
				}
			else
				{
				type = newtype;
				typerr();
				}
			}
		else
			{
			if(type == 0)	/* Must be a typedef... */
				{
				/* Skip type name if so */
				if((rval = istypedef()) != NULL)
					skip();
				return (rval);
				}
			if(type & VOID_NODE)
				return (Voidtype);
			if(type & FLOAT_NODE)
				return (Floattype);
			if(type & DOUBLE_NODE)
				return (Doubletype);
			if(type & LDOUBLE_NODE)
				return (Ldoubletype);
			if(type & CHAR_NODE)
				{
				if(type & UNSIGNED_NODE)
					return (Uchartype);
				else
				 	return (Chartype);
				}
			if(type & SHORT_NODE)
				{
				if(type & UNSIGNED_NODE)
					return (Ushorttype);
				else
					return (Shorttype);
				}
			if(type & INT_NODE)
				{
				if(type & UNSIGNED_NODE)
					return (Uinttype);
				else
					return (Inttype);
				}
			if(type & LONG_NODE)
				{
				if(type & UNSIGNED_NODE)
					return (Ulongtype);
				else
					return (Longtype);
				}
			if(type & (UNION_NODE | STRUCT_NODE))
				return (decltag(type));
			if(type & ENUM_NODE)
				return (enumtag());
			if(type & SIGNED_NODE)
				return (Inttype);
			if(type & UNSIGNED_NODE)
				return (Uinttype);

			fatal("Compiler error: gettype");
			}
		skip(); 		/* Skip the type keyword */
		}
/*lint -e533 		Function doesn't return anything at this point */
	}
/*lint +e533 */


/*
 *	See if next token is a typedef name.
 */
PUBLIC	struct	id_node	*
istypedef()
	{
	char			name[NAMEMAX + 1];
	register struct	id_node	*sym;

	if(! chksym(name) || (sym = lookup_node(name,NS_TYPEDEF)) == NULL)
		return (NULL);

	if(sym->info.storage_class == SC_TYPEDEF)
		return (sym->type_ptr.base_type);

	return (NULL);
	}

/*
 *	Get a valid C variable storage class if present.
 */
PUBLIC	int
getsc()
	{
	int			sc;

	if((sc = chksc()) != 0)
		skip(); 		/* Skip over the keyword */

	return (sc);
	}

/*
 *	Parse an abstract declarator.
 */
PUBLIC	struct	id_node	*
abstdecl()
	{
	struct	id_node		*stack;
	struct	id_node		*type;

	if((type = gettype()) != NULL)
		{
		stack = NULL;		/* Initialize stack */
		dodecl(&stack,NULL,FALSE);	/* Analyze the declarator */
		type = loadtype(&stack,type,SC_NONE);
		if((type->type & ARRAY_NODE) && (type->val.integral == -1))
			sizerr();
		return (type);
		}
	return (NULL);
	}

/*
 *	Load a local type entry into the symbol table.
 */
PUBLIC	struct	id_node	*
loadtype(stack,type,sc)
	register struct id_node	**stack;
	register struct id_node	*type;
	int			sc;
	{
	register INT32		size;
	register struct	id_node	*temp;

	for(temp = *stack; temp != NULL; temp = temp->level_ptr)
		{
		if(temp->type & PTR_NODE)
			type = ptrto(type);
		else if(temp->type & FUNC_NODE)
			{
			if(type->type & (ARRAY_NODE | FUNC_NODE))
				{
		error("Function can't return an array or another function");
				type = Inttype;
				}
			type = add_type((INT32) FUNC_NODE,type,(INT32) 0);
			type->proto_list = temp->proto_list;
			temp->proto_list = NULL; /* Don't want deleted! */
			}
		else if(temp->type & ARRAY_NODE)
			{
			if(type->type & FUNC_NODE)
				{
				error("No arrays of functions");
				type = ptrto(type);
				}
			if(temp->val.integral != -1)
				size = temp->val.integral * type->val.integral;
			else
				{
				/* 1st dimension missing */
				if(temp->level_ptr == NULL)
					size = -1;
				else if(((temp->level_ptr)->level_ptr == NULL)
						&&
					((temp->level_ptr)->type & PTR_NODE))
					/* Pointer-to-array */
					{
					size = -1;
					}
				else
					{
					sizerr();
					size = type->val.integral;
					}
				}
			type = add_type((INT32) ARRAY_NODE,type,size);
			}
		}
	free_level(*stack);		/* Free up "stack" */
/*
 *	The following code tests to see if we have a struct/union tag which
 *	never got declared (has no members).  Additionally note that
 *	this test only works at the top level of a declaration and will not
 *	catch things like a pointer to an undefined struct.  These cases are
 *	handled by similar code located in "delete_level".  This separation
 *	is necessary since you have to allow pointer-to-struct forward
 *	references (as well as extern/typedef forward references...)
 */
	if((type->type & SU_NODE) && (type->type_ptr.member_head == NULL) &&
		(sc != SC_TYPEDEF) && (sc != SC_EXTERN))
		{
		errname("Undefined structure/union tag",type->name);
		}
	return (type);
	}

/*
 *	Process a declarator.
 */
PUBLIC	VOID
dodecl(stack,name,name_rqd)
	struct	id_node			**stack;
	char				*name;
	BOOL				name_rqd;
	{
	if(matchc('*'))
		{
		dodecl(stack,name,name_rqd);
		if(name_rqd && (name != NULL) && (*name == '\0'))
			illname();
/*
 *	Note: stack_type does not need to stack the SIGNED_NODE info, where
 *	pointers are signed, since ptrto() will build the SIGNED_NODE bit into
 *	the type if necessary.
 */
		stack_type(stack, (INT32) PTR_NODE, (INT32) Sizeof_ptr, NULL);
		}
	dodecl1(stack,name,name_rqd);
	}

/*
 *	Helping function for dodecl.
 */
PRIVATE	VOID
dodecl1(stack,name,name_rqd)
	struct	id_node		**stack;
	char			*name;
	BOOL			name_rqd;
	{
	dodecl2(stack,name,name_rqd);
	for(;;)
		{
		if(matchc('('))
			{
			if(name_rqd && (name != NULL) && (*name == '\0'))
				illname();
			stack_type(stack,(INT32) FUNC_NODE,(INT32) 1,
				procarg());
			dodecl2(stack,name,name_rqd);
			}
		else if(matchc('['))
			{
			if(name_rqd && (name != NULL) && (*name == '\0'))
				illname();
			stack_type(stack,(INT32) ARRAY_NODE,arraysize(),NULL);
			dodecl2(stack,name,name_rqd);
			}
		else
			break;
		}
	}

/*
 *	Helping function for dodecl.
 */
PRIVATE	VOID
dodecl2(stack,name,name_rqd)
	struct	id_node		**stack;
	char			*name;
	BOOL			name_rqd;
	{
	if(matchc('('))
		{
		if(chkeyword() || nextc(')'))
			putback('(');	/* Function with/without proto */
		else
			{
			dodecl(stack,name,name_rqd);
			needpunc(')');
			}
		}
/*
 *	"name == NULL" means an abstract declarator.  "stack == NULL" means
 *	we haven't already parsed the inner most nesting (ie. we could have a
 *	name around here...)
 */
	else if((name != NULL) && (*stack == NULL))
		{
		if((*name != '\0') && symname(name))
			illname();	/* One name per customer please */
		else
			(VOID) symname(name);
		}
	}

/*
 *	Stacking routine for parsing type expresions.
 */
PRIVATE	VOID
stack_type(stack,type,size,proto_list)
	struct	id_node		**stack;
	INT32			type;
	INT32			size;
	struct	prototype	*proto_list;
	{
	struct	id_node		*temp;

	temp = get_node();		/* Get "id_node" to add to stack */
	temp->level_ptr = *stack;	/* Place on top of stack */
	*stack = temp;			/* Let stack know about it */
	temp->type = type;		/* Set "type" of stack node */
	temp->val.integral = size;	/* Set size of item */
	temp->proto_list = proto_list;	/* Add possible prototype list */
	}

/*
 *	Test for legal symbol name.
 */
PUBLIC	BOOL
symname(name)
	char			*name;
	{
	if(! chksym(name))
		return (FALSE);
	else if(chkeyword())
		{
		skip();
		keyerr(name);		/* Illegal use of keyword */
		return (FALSE);
		}
	else
		{			/* Valid symbol name */
		skip();
		return (TRUE);
		}
	}

/*
 *	Check for a valid 'C' symbol.
 */
PUBLIC	char	*
chksym(name)
	register char		*name;
	{
	register int		i;
	register char		*p;

	blanks();
	if(! issymf(ch()))
		p = NULL;		/* Name must start with an alpha */
	else
		{
		for(i = 0, p = Lptr; issym((int) *p); ++p)
			{
			if(i++ < NAMEMAX)
				*name++ = *p;
			}
		}
	*name = '\0';
	return (p);			/* Return loc of char after name */
	}

/*
 *	Return next available internal label number.
 */
PUBLIC	LABEL
getlabel()
	{
	static	LABEL		nextlabel = 1;

	return (++nextlabel);
	}

/*
 *	Return next available "double" internal temp number.
 */
PUBLIC	TEMP
get_dtemp()
	{
	if((tempcount += 2) > Tempspace)
		Tempspace = tempcount;

	return (tempcount-2);		/* Return temp number from 0 */
	}

/*
 *	Return next available internal temp number.
 */
PUBLIC	TEMP
get_temp()
	{
	if(++tempcount > Tempspace)
		Tempspace = tempcount;

	return (tempcount-1);		/* Return temp number from 0 */
	}

/*
 *	Release "double" temp.
 */

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

PUBLIC	VOID
free_dtemp(temp)
	TEMP			temp;
	{
	tempcount -= 2;
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/*
 *	Release temp.
 */

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

PUBLIC	VOID
free_temp(temp)
	TEMP			temp;
	{
	--tempcount;
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/*
 *	Add an entry to the switch/loop continue/break label queue.
 *
 *	The stack of loop continue and loop exit labels is really stored on
 *	the stack (in auto variables) of the parsing routines.  Each level,
 *	(e.g. dofor) contains a "swq" variable, which is linked to the previous
 *	such block.  The "addswq" and "delswq" routines just link and unlink
 *	these blocks.
 */
PUBLIC	VOID
addswq(p)
	struct	swq		*p;
	{
	p->swq_prev = Swqueue;
	Swqueue = p;
	}

/*
 *	Delete an entry from the switch/loop queue.
 */
PUBLIC	VOID
delswq()
	{
	if(Swqueue == NULL)
		error("No active switches or loops to delete");
	else
		Swqueue = Swqueue->swq_prev;
	}

/*
 *	Add an entry to the switch stack
 *
 *	The stack of switch statement blocks is really stored on the stack
 *	(in auto variables) of the parsing routine (doswitch).  Each level
 *	of doswitch has its own switch header block.  We only need to get
 *	to the most recent block, so the symbol "Casestack" points to the
 *	current header.  Each switch block is linked to the previous block
 *	to allow for backing out.
 */
PUBLIC	VOID
addswheader(p)
	struct	swheader	*p;
	{
	p->sh_prev = Casestack;
	Casestack = p;
	}

/*
 *	Remove a switch level from the stack
 */
PUBLIC	VOID
delswheader()
	{
	struct	switem		*handle;
	struct	switem		*scan;

	if(Casestack == NULL)
		{
		error("No active switches or loops to delete");
		return;
		}
/*
 *	Free up all of the case labels.
 */
/*lint -e530 "handle" is initialized in the loop */
	for(scan = Casestack->sh_caselist; scan != NULL; scan = handle)
		{
		handle = scan->sw_next;
		free((char *) scan);
		}
/*lint +e530 */

	Casestack = Casestack->sh_prev;
	}

/*
 *	Add an entry to the current switch statement
 *
 *	The descriptor to this item cannot be allocated on the stack
 *	of the "docase" routine, as the descriptor for case labels outlives
 *	the "docase" statement.  So, case entries are built from heap space.
 */
PUBLIC	VOID
addcase(val,label)
	INT32			val;
	LABEL			label;
	{
	struct	switem		*newitem;
	struct	switem		**scan;

	if(Casestack == NULL)
		{
		error("No active switch statement");
		return;
		}
	for(scan = &Casestack->sh_caselist; *scan != NULL;
		scan = &(*scan)->sw_next)
		{
		if(val == (*scan)->sw_value)
			{
			error("Duplicate switch case");
			return;
			}
		}
	if((newitem = (struct switem *) malloc(sizeof(struct switem))) == NULL)
		{
		mem_err();
		return;
		}
	*scan = newitem;
	newitem->sw_next = NULL;
	newitem->sw_value = val;
	newitem->sw_label = label;
	}

/*
 *	Evaluate a constant-like expression into a op_node structure.
 *
 *	Should this routine be in TCX4.C?  It seems a lot like "expression".
 *	DCR		17 May 87.
 */
PUBLIC	struct	op_node	*
constexp()
	{
	return (hier3());
	}

/*
 *	If next token is a constant build an op_node with its value.
 */
PUBLIC	BOOL
constant(op)
	struct	op_node		*op;
	{
	struct	id_node		*base_type;
	register int		c;
	int			is_char;
	int			is_float;
	int			is_hex;
	int			is_longf;
	int			is_singlef;
	int			leading_zero;
	int			long_no;
	char			num[NAMEMAX + 1];	/* OK for now */
	register char		*p;
	register int		sign;
	char			*state;
	register INT32		type;
	int			unsigned_no;
	union	value		val;

	blanks();			/* Skip leading white space */
	leading_zero=is_hex=is_char=is_float=is_singlef=is_longf=sign=FALSE;

	switch((int) (c = ch()))
		{
		case '\'':
			is_char = TRUE;
			(VOID) gch();	/* Eat up leading apostrophe */
			break;

		case '+':
		case '-':
			if(! isdigit(nch()) && (nch() != '.'))
				return (FALSE);
			sign = gch();
			break;

		case '.':
			if(! isdigit(nch()))
				error("Invalid float constant");
			break;

		case '0':
			if(match("0x") || match("0X"))
				{
				is_hex = TRUE;
				if(! isxdigit(ch()))
					error("Invalid hex constant");
				}
			else
				leading_zero = TRUE;
			break;

		default:
			if(! isdigit(c))
				return (FALSE);
			break;
		}

	if(! is_char)
		{
		p = num;		/* Build a string containing number */

		if(sign)
			*p++ = sign;

		if(is_hex)
			state = "0123456789abcdefABCDEF";
		else
			state = "0123456789.Ee";

		while((ch() != '\0') && (strchr(state, ch()) != NULL))
			{
			c = *p++ = gch();	/* Accept another character */

			if(! is_hex)
				{
				if(c == '.')
					{
					leading_zero = FALSE;
					state = "0123456789Ee";
					is_float = TRUE;
					}
				else if((c == 'E') || (c == 'e'))
					{
					leading_zero = FALSE;
					state = "0123456789+-";
					is_float = TRUE;
					}
				else if((c == '+') || (c == '-'))
					state = "0123456789";
				}
			}

		if(is_float)
			{
			if((c = (isupper(c = ch()) ? tolower(c) : c)) == 'f')
				{
				is_singlef = TRUE;
				(VOID) gch();		/* Toss 'f' */
				}
			else if(c == 'l')
				{
				is_longf = TRUE;
				(VOID) gch();		/* Toss 'l' */
				}
			}
		else
			{
			for(long_no = unsigned_no = c = FALSE; !c; )
				{
				if((! long_no) &&
					((ch() == 'l') || (ch() == 'L')))
					{
					long_no = TRUE;
					*p++ = gch();
					}
				else
				if((! unsigned_no) &&
					((ch() == 'u') || (ch() == 'U')))
					{
					unsigned_no = TRUE;
					*p++ = gch();
					}
				else
					c = TRUE;
				}
			}

		*p = '\0';
		}

	if(is_float)
		{
		val.fvalue = atof(num);
		if(is_singlef)
			type = FLOAT_NODE;
		else if(is_longf)
			type = LDOUBLE_NODE;
		else
			type = DOUBLE_NODE;
		}
	else if(is_hex)
		type = c_hex(&val, num);	/* Hex? */
	else if(is_char)
		type = charconst(&val);		/* Char? */
	else if(leading_zero)
		type = c_octal(&val, num);	/* Octal? */
	else
		type = c_decimal(&val, num);	/* Decimal? */

	if(type == 0)
		{
		base_type = Inttype;
		error("Compiler error: constant");
		}
	else
		base_type = basetype(type);

	op->right.ptr.terminal = add_type(type,base_type,val.integral);
	op->right.ptr.terminal->info.storage_class = SC_CONSTANT;

	if(type & FP_NODE)
		op->right.ptr.terminal->val.fvalue = val.fvalue;

	op->node_type = base_type;

	return (TRUE);
	}

/*
 *	Check for hex constant.
 */
PRIVATE	INT32
c_hex(val,s)
	union	value		*val;
	char			*s;
	{
	register	int	d;
	INT32			type;
	UINT32			v;

	for(v = 0, type = (SIGNED_NODE | INT_NODE); isxdigit(*s);)
		{
		d = *s++;
		if(isdigit(d))
			d -= '0';
		else if(isupper(d))
			d = d - 'A' + 10;
		else
			d = d - 'a' + 10;
		v = (v << 4) | d;
		if(Sizeof_int == WORD_SIZE_16B)
			{
			if(v > (UINT32) MAX_INT_16B)
				{
				if(v > (UINT32) MAX_UINT_16B)
					{
					if(v > (UINT32) MAX_LONG_16B)
						type=(UNSIGNED_NODE|LONG_NODE);
					else
						type = (SIGNED_NODE|LONG_NODE);
					}
				else
					type = (UNSIGNED_NODE | INT_NODE);
				}
			}
		else			/* Must be 32 bit machine */
			{
			if(v > (UINT32) MAX_INT_32B)
				{
				if(v > (UINT32) MAX_UINT_32B)
					{
					if(v > (UINT32) MAX_LONG_32B)
						type=(UNSIGNED_NODE|LONG_NODE);
					else
						type = (SIGNED_NODE|LONG_NODE);
					}
				else
					type = (UNSIGNED_NODE | INT_NODE);
				}
			}
		}
	while((*s != '\0') && (strchr("UuLl",(int) *s) != NULL))
		{
		if((*s == 'L') || (*s == 'l'))
			{
			if(type & UNSIGNED_NODE)
				type = (LONG_NODE | UNSIGNED_NODE);
			else
				type = (LONG_NODE | SIGNED_NODE);
			}
		else if((*s == 'U') || (*s == 'u'))
			{
			if(type & LONG_NODE)
				type = (UNSIGNED_NODE | LONG_NODE);
			else
				type = (UNSIGNED_NODE | INT_NODE);
			}
		++s;
		}
	val->integral = (INT32) v;
	return (type);
	}

/*
 *	Check for octal constant.
 */
PRIVATE	INT32
c_octal(val,s)
	union	value		*val;
	char			*s;
	{
	UINT32			v;
	INT32			type;

	for(v = 0,type = (SIGNED_NODE | INT_NODE); (*s >= '0') && (*s <= '7');)
		{
		v = (v << 3) | (*s++ - '0');
		if(Sizeof_int == WORD_SIZE_16B)
			{
			if(v > (UINT32) MAX_INT_16B)
				{
				if(v > (UINT32) MAX_UINT_16B)
					{
					if(v > (UINT32) MAX_LONG_16B)
						type=(UNSIGNED_NODE|LONG_NODE);
					else
						type = (SIGNED_NODE|LONG_NODE);
					}
				else
					type = (UNSIGNED_NODE | INT_NODE);
				}
			}
		else			/* Must be 32 bit machine */
			{
			if(v > (UINT32) MAX_INT_32B)
				{
				if(v > (UINT32) MAX_UINT_32B)
					{
					if(v > (UINT32) MAX_LONG_32B)
						type=(UNSIGNED_NODE|LONG_NODE);
					else
						type = (SIGNED_NODE|LONG_NODE);
					}
				else
					type = (UNSIGNED_NODE | INT_NODE);
				}
			}
		}
	while((*s != '\0') && (strchr("UuLl",(int) *s) != NULL))
		{
		if((*s == 'L') || (*s == 'l'))
			{
			if(type & UNSIGNED_NODE)
				type = (LONG_NODE | UNSIGNED_NODE);
			else
				type = (LONG_NODE | SIGNED_NODE);
			}
		else if((*s == 'U') || (*s == 'u'))
			{
			if(type & LONG_NODE)
				type = (UNSIGNED_NODE | LONG_NODE);
			else
				type = (UNSIGNED_NODE | INT_NODE);
			}
		++s;
		}
	val->integral = (INT32) v;
	return (type);
	}

/*
 *	Check for decimal constant.
 */
PRIVATE	INT32
c_decimal(val,s)
	union	value		*val;
	char			*s;
	{
	register	int	c;
	UINT32			d;
	int			negative;
	INT32			type;
	UINT32			v;

	type = 0;

	c = *s;
	negative = (c == '-');
	if((c == '+') || (c == '-'))
		++s;			/* Eat up the leading sign */
	v = 0;
	while(isdigit(*s))
		{			/* Build negative to handle largest */
		d = *s++ - '0';		/* negative number in 2's complement */
		v = v * 10 + d;

		if(Sizeof_int == WORD_SIZE_16B)
			{
			if(v > (UINT32) MAX_INT_16B)
				{
				type = LONG_NODE;
				if(v > (UINT32) MAX_LONG_16B)
					type |= UNSIGNED_NODE;
				else
					type |= SIGNED_NODE;
				}
			}
		else			/* Must be 32 bit machine */
			{
			if(v > (UINT32) MAX_INT_32B)
				{
				type = LONG_NODE;
				if(v > (UINT32) MAX_LONG_32B)
					type |= UNSIGNED_NODE;
				else
					type |= SIGNED_NODE;
				}
			}
		}
	while((*s != '\0') && (strchr("UuLl",(int) *s)))
		{
		if((*s == 'L') || (*s == 'l'))
			{
			if(type & UNSIGNED_NODE)
				type = (LONG_NODE | UNSIGNED_NODE);
			else
				type = (LONG_NODE | SIGNED_NODE);
			}
		else if((*s == 'U') || (*s == 'u'))
			{
			if(type & LONG_NODE)
				type = (UNSIGNED_NODE | LONG_NODE);
			else
				type = (UNSIGNED_NODE | INT_NODE);
			}
		++s;
		}
	if(negative)
		v = - (INT32) v;
	val->integral = v;

	return (type ? type : (INT_NODE | SIGNED_NODE));
	}

/*
 *	Check for character constant.
 *
 *	Later - put in code to check for overflow when building a constant
 *		so that the compiler can issue a warning to the user.
 */
PRIVATE	INT32
charconst(val)
	union	value		*val;
	{
	register	int	c;
	INT32			type;
	register	UINT32	v;

	for(v = type = 0; ((c = gch()) != '\'') && (c != 0);)
		{
		if(c == '\\')
			c = escape();

		v = (v << 8) + (c & 0xFF);

		if(Sizeof_int == WORD_SIZE_16B)
			{
			if(v > (UINT32) MAX_INT_16B)
				{
				type = LONG_NODE;
				if(v > (UINT32) MAX_LONG_16B)
					type |= UNSIGNED_NODE;
				else
					type |= SIGNED_NODE;
				}
			}
		else			/* Must be 32 bit machine */
			{
			if(v > (UINT32) MAX_INT_32B)
				{
				type = LONG_NODE;
				if(v > (UINT32) MAX_LONG_32B)
					type |= UNSIGNED_NODE;
				else
					type |= SIGNED_NODE;
				}
			}
		}
	val->integral = v;

	return (type ? type : (INT_NODE | SIGNED_NODE));
	}

/*
 *	Check for string, build an operand, and add it to the literal pool.
 *
 *	The parameter "op" points to an op_node than can be filled in with
 *	the pertinant info if the item is a string.
 */
PUBLIC	BOOL
string(op)
	struct	op_node		*op;
	{
	char			buf[80];
	register	int	c;
	char			*cur;
	char			*end;
	LABEL			label;
	int			perline;
	int			stringlen;

	if(! matchc('\"'))
		return (FALSE);

	dvrt_stack();
	set_literal();		/* Define string to be in literal area */

	label = getlabel();
	printlabel(label);

	stringlen = 1;		/* Starts at 1 because strings have 0 at end */
	perline = 0;

	while((c = getstring()) != -1)
		{
		if(perline == 0)
			{
			cur = buf;
			end = buf + (sizeof(buf) - 1);
			}
		else
			cur = safecat(cur,",",end);

		cur = safedec(cur,(INT32) c,end);
		stringlen++;

		if(++perline == MAXDBCH)
			{
			*cur = '\0';
			outline(TOP_DB,buf,NULL);
			perline = 0;
			}
		}

	if(perline == 0)
		{
		cur = buf;
		end = buf + (sizeof(buf) - 1);
		}
	else
		cur = safecat(cur,",",end);

	cur = safedec(cur,(INT32) 0,end);
	*cur = '\0';
	outline(TOP_DB,buf,NULL);
/*
 *	Need to create node that is a type ARRAY, with value of size of the
 *	string and base type of Chartype (or Uchartype).
 */
	op->right.ptr.terminal = add_type((INT32) 0,
		add_type(ARRAY_NODE | LITERAL_NODE,
			(Cpu_schar ? Chartype : Uchartype),(INT32) stringlen),
		(INT32) stringlen);

	op->node_type = op->right.ptr.terminal->type_ptr.base_type;
	op->right.ptr.terminal->info.storage_class = SC_LITERAL;
	op->right.ptr.terminal->address.surrogate = label;
	set_code();			/* Back to code space */
	op->right.ptr.terminal->diversion = dvrt_unstack();

	return (TRUE);
	}

/*
 *	Get the next character in a literal string.
 */
PUBLIC	int
getstring()
	{
	int			c;

	switch((int) (c = gch()))
		{
		case '\"':
			if(matchc('\"'))
				return (getstring());
			c = -1;
			break;

		case '\0':
			c = -1;
			break;

		case '\\':
			c = escape();
			break;
		}
	return (c);
	}

/*
 *	Evaluate an escape character constant.
 */
PRIVATE	int
escape()
	{
	register int		c;
	register int		code;
	register int		count;

	switch((int) (c = gch()))
		{
		case '\'':
		case '\"':
		case '\\':
		case '\?':
			return (c);

		case 'a':		/* Alarm */
			return ('\a');
/*
 *	Too bad if your compiler doesn't work.  DCR 24 May 87
 */
		case 'b':		/* Backspace */
			return ('\b');

		case 'f':		/* Formfeed */
			return ('\f');

		case 'n':		/* Newline */
			return ('\n');

		case 'r':		/* Carriage return */
			return ('\r');

		case 't':		/* Horizontal tab */
			return ('\t');

		case 'v':		/* Vertical tab */
			return ('\v');

		case 'x':		/* Hex sequence - max 2 hex digits */
			count = 0;
			code = 0;
			while(isxdigit(ch()) && (++count < 3))
				{
				code = code * 16;
				if(isdigit(c = gch()))
					code += c - '0';
				else if(isupper(c))
					code += c - 'A' + 10;
				else
					code += c - 'a' + 10;
				}
			return (code);

		default:
			if(isdigit(c))
				{	/* Octal seq. - max 3 octal digits */
				count = 0;
				code = c - '0';	/* Read up to 2 MORE digits */
				while(isdigit(ch()) && (++count < 3))
					code = code * 8 + gch() - '0';
				return (code);
				}
			else
				warning(1,"Invalid escape character");
		}
	return (c);
	}

/*
 *	Convert "type" to "ptr-to-type".
 */
PUBLIC	struct	id_node	*
ptrto(type)
	struct	id_node		*type;
	{
	return (add_type((INT32) (PTR_NODE | SIGNED_NODE),
		type, (INT32) Sizeof_ptr));
	}

/*
 *	Check op to see if it is a pointer.
 */
PUBLIC	BOOL
isptr(op)
	struct	op_node		*op;
	{
	return ((op->node_type->type & PTR_NODE) != 0);
	}

/*
 *	Check op to see if it is a array node.
 */
PUBLIC	BOOL
isarray(op)
	struct	op_node		*op;
	{
	return ((op->node_type->type & ARRAY_NODE) != 0);
	}

/*
 *	Check to see if it is a pointer or array node.
 */
PUBLIC	BOOL
isptrclass(op)
	struct	op_node		*op;
	{
	return ((op != NULL) &&
		((EXPRTYPE(op) & PTYPE_NODE) != 0));
	}

/*
 *	Modify function call structure return pointer to new expression.
 */
PUBLIC	VOID
modify_rslink(call_expr,link_expr)
	struct	op_node		*call_expr;
	struct	op_node		*link_expr;
	{
	struct	op_node		*op;

	if((link_expr == NULL) || (call_expr == NULL) ||
		(call_expr->operator != OP_CALL))
		{
		fatal("Compiler error: modify_rslink");
		}
	if(((op = LEFT(call_expr)) == NULL) || (op->operator != OP_PARAM) ||
		(RIGHT(op) == NULL))
		{
	error("Unable to find implied function parameter during optimization");
		return;
		}
	free_op_node(RIGHT(op));
	RIGHT(op) = link_expr;
	}

/* End of TCX5.C */
