/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: tcx6a.c,v 2.2 92/04/21 11:11:13 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	TCX Compiler (6a of 11)	Expression Syntax Tree Walker
 *
 *	Copyright 1987-1991 by Logical Systems.  All rights reserved.
 *
 *	Published by:	Logical Systems
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	03/17/87
 */

#include	<stdio.h>
#include	<ctype.h>
#include	<string.h>
#include	"tcx.h"
#include	"tcxdecl.h"
#include	"tcxext.h"

/*
 *	Local function declarations.
 */
#ifdef	DEBUG

PRIVATE	char			*expr_dhelp();
PRIVATE	char			*tree_print();
PRIVATE	char			*outtype();

#endif	/* DEBUG */
/*
 *	Expression "depth" usage notes:
 *
 *	Since the Transputer has only a 3 element deep stack we wish to make
 *	the most of it for expression evaluation.  To do this we keep track of
 *	the current maximum amount of stack which is used by any routine.  This
 *	record lets us make intelligent decisions about "temp" allocation, etc.
 *	The rules for routines which return "depth" values are:
 *
 *	1.	The depth returned is equal to the maximum amount used by the
 *		routine and all routines which it calls.  The maximum amount
 *		used should includes all input and output runtime "parameters"
 *		the routine uses.
 *	2.	The depth returned is never greater than 3.
 */
PUBLIC	VOID
expr_out(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	BOOL			old_sinfo;

	if(expr != NULL)
		{
		if((Errcnt == 0) || (! Errdel))
			{
			old_sinfo = Source_info;
			Source_info = FALSE;	/* Turn off #line in expr */
#ifdef	DEBUG
			if(Debug)
				{
				char		buf[LINESIZE];
				char		*cur;
				char		*end;
	
				end = buf + (sizeof(buf) - 1);	/* End ptr */
				cur = safecat(buf,";",end);
				cur = expr_dhelp(cur,expr,end);
				cur = safecat(cur,"\n;",end);
				cur = tree_print(cur,expr,0,end);
				*cur = '\0';
				outcomment(buf);
				}
#endif	/* DEBUG */
			(VOID) expr_code(NULL,expr,value_flag);
			no_oper(TOP_BETWEEN);	/* Indicate between exprs */
			}
		free_op_node(expr);

		Newfile = FALSE;	/* Diversion stuff turned it on */
		Source_info = old_sinfo;	/* Restore old status */
		}
	}

/*
 *	Generate code by recursively processing the expression tree.
 */
PUBLIC	unsigned	int
expr_code(totype,expr,vf)
	struct	id_node		*totype;
	struct	op_node		*expr;
	int			vf;
	{
	unsigned	int	depth;

	if(is_simple(expr))
		{
		depth = do_load((totype != NULL) ? totype : expr->node_type,
			expr,vf);
		}
	else
		{
		switch((unsigned int) (expr->operator))
			{
			case OP_MEMBER:
				depth = do_member(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_INDIRECT:
				depth = do_indirect(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_ADDRESS:
				depth = do_address(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_UMIN:
				depth = do_umin(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_UPLUS:
				depth = do_uplus(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_LNOT:
				depth = do_lnot(expr,vf);
				break;
			case OP_BNOT:
				depth = do_bnot(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_POSTADD:
				depth = do_postop(expr,vf & ~(RELAX_BOOL),1);
				break;
			case OP_POSTSUB:
				depth = do_postop(expr,vf & ~(RELAX_BOOL),-1);
				break;
			case OP_MUL:
				depth = do_mul(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_DIV:
				depth = do_div(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_MOD:
				depth = do_mod(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_ADD:
				depth = do_add(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_SUB:
				depth = do_sub(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_RSHIFT:
				depth = do_rshift(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_LSHIFT:
				depth = do_lshift(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_AMUL:
				depth = do_amul(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_ADIV:
				depth = do_adiv(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_AMOD:
				depth = do_amod(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_AADD:
				depth = do_aadd(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_ASUB:
				depth = do_asub(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_ARSHIFT:
				depth = do_arshift(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_ALSHIFT:
				depth = do_alshift(expr,vf & ~(RELAX_BOOL));
				break;
/* Rel ops changed 8-Jun-90: added "& ~RELAX_BOOL."  G. Oliver */
			case OP_LT:
				depth = do_lt(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_GT:
				depth = do_gt(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_LE:
				depth = do_le(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_GE:
				depth = do_ge(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_EQ:
				depth = do_ecomp(expr,vf,FALSE);
				break;
			case OP_NE:
				depth = do_ecomp(expr,vf,TRUE);
				break;
			case OP_BAND:
				depth = do_band(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_BXOR:
				depth = do_bxor(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_BOR:
				depth = do_bor(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_ABAND:
				depth = do_aband(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_ABXOR:
				depth = do_abxor(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_ABOR:
				depth = do_abor(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_LAND:
				depth = do_land(expr,vf);
				break;
			case OP_LOR:
				depth = do_lor(expr,vf);
				break;
			case OP_TERN:
				depth = do_tern(expr,vf);
				break;
			case OP_SEQ:
				depth = do_seq(expr,vf);
				break;
			case OP_CALL:
			case OP_ICALL:
				depth = do_call(expr);
				break;
			case OP_ASSIGN:
				depth = do_assign(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_CONVERT:
				depth = do_convert(expr,vf & ~(RELAX_BOOL));
				break;
			case OP_BOOL:
				depth = do_bool(expr, vf);
				break;

			default:
				fatal("Compiler error: expr_code");
			}
		if(vf & NEED_VALUE)
			{
			if((totype != NULL) && (! (vf & EVAL_ADDR)))
				{
/*
 *	A type to convert to, and not evaluating for address...
 */
				depth = depth_max(depth,cnvt_type(totype,
					expr->node_type));
				}
			}
		}
	return (depth);
	}

#ifdef	DEBUG

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#endif	/* MSC_OPT */

/*
 *	Format operator for debugging purposes.
 */
PRIVATE	char	*
expr_dop(cur,expr,end)
	char			*cur;
	struct	op_node		*expr;
	char			*end;
	{
	switch((unsigned int) (expr->operator))
		{
		case	OP_OFFSET:	cur = safecat(cur,"@",end); break;
		case	OP_MEMBER:	cur = safecat(cur,".",end); break;
		case	OP_INDIRECT:	cur = safecat(cur,"*",end); break;
		case	OP_ADDRESS:	cur = safecat(cur,"&",end); break;
		case	OP_UMIN:	cur = safecat(cur,"-",end); break;
		case	OP_UPLUS:	cur = safecat(cur,"+",end); break;
		case	OP_LNOT:	cur = safecat(cur,"!",end); break;
		case	OP_BNOT:	cur = safecat(cur,"~",end); break;
		case	OP_POSTADD:	cur = safecat(cur,"++",end); break;
		case	OP_POSTSUB:	cur = safecat(cur,"--",end); break;
		case	OP_MUL:		cur = safecat(cur,"*",end); break;
		case	OP_DIV:		cur = safecat(cur,"/",end); break;
		case	OP_MOD:		cur = safecat(cur,"%",end); break;
		case	OP_ADD:		cur = safecat(cur,"+",end); break;
		case	OP_SUB:		cur = safecat(cur,"-",end); break;
		case	OP_RSHIFT:	cur = safecat(cur,">>",end); break;
		case	OP_LSHIFT:	cur = safecat(cur,"<<",end); break;
		case	OP_AMUL:	cur = safecat(cur,"*=",end); break;
		case	OP_ADIV:	cur = safecat(cur,"/=",end); break;
		case	OP_AMOD:	cur = safecat(cur,"%=",end); break;
		case	OP_AADD:	cur = safecat(cur,"+=",end); break;
		case	OP_ASUB:	cur = safecat(cur,"-=",end); break;
		case	OP_ARSHIFT:	cur = safecat(cur,">>=",end); break;
		case	OP_ALSHIFT:	cur = safecat(cur,"<<=",end); break;
		case	OP_LT:		cur = safecat(cur,"<",end); break;
		case	OP_GT:		cur = safecat(cur,">",end); break;
		case	OP_LE:		cur = safecat(cur,"<=",end); break;
		case	OP_GE:		cur = safecat(cur,">=",end); break;
		case	OP_EQ:		cur = safecat(cur,"==",end); break;
		case	OP_NE:		cur = safecat(cur,"!=",end); break;
		case	OP_BAND:	cur = safecat(cur,"&",end); break;
		case	OP_BXOR:	cur = safecat(cur,"^",end); break;
		case	OP_BOR:		cur = safecat(cur,"|",end); break;
		case	OP_ABAND:	cur = safecat(cur,"&=",end); break;
		case	OP_ABXOR:	cur = safecat(cur,"^=",end); break;
		case	OP_ABOR:	cur = safecat(cur,"|=",end); break;
		case	OP_LAND:	cur = safecat(cur,"&&",end); break;
		case	OP_LOR:		cur = safecat(cur,"||",end); break;
		case	OP_TERN:	cur = safecat(cur,"?",end); break;
		case	OP_TSEL:	cur = safecat(cur,":",end); break;
		case	OP_SEQ:		cur = safecat(cur,",",end); break;
		case	OP_PARAM:	cur = safecat(cur,",",end); break;
		case	OP_ASSIGN:	cur = safecat(cur,"=",end); break;
		case	OP_CONVERT:	cur = safecat(cur,"(cast)",end); break;
		case	OP_BOOL:	cur = safecat(cur,"!!",end); break;

		default:		cur = safecat(cur,"*bad*",end); break;
		}
	cur = outtype(cur,expr->node_type->type,end);
	return (cur);
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#endif	/* MSC_OPT */

/*
 *	Format scalar type for debugging purposes.
 */
PRIVATE	char	*
outtype(cur,type,end)
	char			*cur;
	INT32			type;
	char			*end;
	{
	cur = safecat(cur,"<",end);

	if(type & UNSIGNED_NODE)
		cur = safecat(cur,"U",end);
	else
	if(type & SIGNED_NODE)
		cur = safecat(cur,"S",end);

	if(type & CHAR_NODE)
		cur = safecat(cur,"C",end);
	else
	if(type & SHORT_NODE)
		cur = safecat(cur,"S",end);
	else
	if(type & INT_NODE)
		cur = safecat(cur,"I",end);
	else
	if(type & LONG_NODE)
		cur = safecat(cur,"L",end);
	else
	if(type & FLOAT_NODE)
		cur = safecat(cur,"F",end);
	else
	if(type & DOUBLE_NODE)
		cur = safecat(cur,"D",end);
	else
	if(type & LDOUBLE_NODE)
		cur = safecat(cur,"LD",end);
	else
	if(type & PTR_NODE)
		cur = safecat(cur,"*",end);
	else
		cur = safecat(cur,"?",end);

	cur = safecat(cur,">",end);
	return (cur);
	}

/*
 *	Format an expression for debugging purposes.
 */
PRIVATE	char	*
expr_dhelp(cur,expr,end)
	char			*cur;
	struct	op_node		*expr;
	char			*end;
	{
	char			buf[NAMEMAX + 1];

	if(expr == NULL)
		return (cur);

	if(ISTERMINAL(expr))
		{
		if(expr->right.ptr.terminal->info.storage_class ==
			SC_CONSTANT)
			{
			if(expr->right.ptr.terminal->type & FP_NODE)
				{
				cur = safefloat(cur,expr->right.ptr.terminal->
					val.fvalue,end);
				}
			else
				{
				cur = safedec(cur,(INT32) expr->
					right.ptr.terminal->val.integral,end);
				}
			}
		else
			{
			cur = safecat(cur,real_name(expr->right.ptr.terminal,
				buf),end);
			}
		cur = outtype(cur,expr->right.ptr.terminal->type,end);
		return (cur);
		}

	cur = safecat(cur,"{",end);
	switch((int) (expr->operator))
		{
		case	OP_CALL:
			cur = expr_dhelp(cur,RIGHT(expr),end);
			cur = safecat(cur,"(",end);
			cur = expr_dhelp(cur,LEFT(expr),end);
			cur = safecat(cur,")",end);
			break;

		case	OP_ICALL:
			cur = expr_dhelp(cur,RIGHT(expr),end);
			cur = safecat(cur,"$(",end);
			cur = expr_dhelp(cur,LEFT(expr),end);
			cur = safecat(cur,")$",end);
			break;

		case	OP_PARAM:
			cur = expr_dhelp(cur,RIGHT(expr),end);
			if (LEFT(expr) != NULL)
				{
				cur = safecat(cur,",",end);
				cur = expr_dhelp(cur,LEFT(expr),end);
				}
			break;

		default:
			cur = expr_dhelp(cur,LEFT(expr),end);
			cur = expr_dop(cur,expr,end);
			cur = expr_dhelp(cur,RIGHT(expr),end);
			break;
		}
	return (safecat(cur,"}",end));
	}

/*
 *	Format the expression tree more graphically for debug purposes.
 */
PRIVATE	char	*
tree_print(cur,expr,level,end)
	char			*cur;
	struct	op_node		*expr;
	int			level;
	char			*end;
	{
	char			buf[NAMEMAX + 1];
	int			count;

	if(expr == NULL)
		return (cur);

	if(ISTERMINAL(expr))
		{
		if(expr->right.ptr.terminal->info.storage_class ==
			SC_CONSTANT)
			{
			if(expr->right.ptr.terminal->type & FP_NODE)
				{
				cur = safefloat(cur,expr->right.ptr.terminal->
					val.fvalue,end);
				}
			else
				{
				cur = safedec(cur,(INT32) expr->
					right.ptr.terminal->val.integral,end);
				}
			}
		else
			{
			cur = safecat(cur,real_name(expr->right.ptr.terminal,
				buf),end);
			}
		cur = outtype(cur,expr->right.ptr.terminal->type,end);
		cur = safecat(cur,"\n;",end);
		return (cur);
		}

	switch((int) (expr->operator))
		{
		case	OP_CALL:
			cur = safecat(cur,"call()",end);
			break;

		case	OP_ICALL:
			cur = safecat(cur,"icall()",end);
			break;

		case	OP_PARAM:
			cur = safecat(cur,",",end);
			break;
		default:
			cur = expr_dop(cur,expr,end);
			break;
		}
	cur = safecat(cur,"\t",end);
	cur = tree_print(cur,RIGHT(expr),level + 1,end);

	for(count = 0; count < level; count++)
		cur = safecat(cur,"\t",end);

	cur = tree_print(cur,LEFT(expr),level,end);
	return (cur);
	}
#endif	/* DEBUG */

/*
 *	Load operands for a binary operator.
 */
PUBLIC	unsigned	int
bin_op(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	deep_lop;
	unsigned	int	deep_rop;
	struct	dvrt		*dvrt_lop;
	struct	dvrt		*dvrt_rop;
	struct	id_node		*exprtype;
	VOID			(*free_tmp)();
	TEMP			(*get_tmp)();
	struct	op_node		*left;
	struct	op_node		*right;
	unsigned	int	tdeep_inc;
	unsigned	int	tdeep_lop;
	unsigned	int	tdeep_max;
	unsigned	int	tdeep_rop;
	TEMP			temp;
	unsigned	int	total_deep;

	exprtype = op_types(expr);

	if(expr->operator == OP_ASSIGN)
		{
		/* Do in reverse for assignment operator */
		dvrt_stack();
		deep_rop = expr_code(exprtype,LEFT(expr),value_flag);
		dvrt_rop = dvrt_unstack();
		right = LEFT(expr);

		dvrt_stack();
		deep_lop = expr_code(exprtype,RIGHT(expr),value_flag);
		dvrt_lop = dvrt_unstack();
		left = RIGHT(expr);
		}
	else
		{
		dvrt_stack();
		deep_lop = expr_code(exprtype,LEFT(expr),value_flag);
		dvrt_lop = dvrt_unstack();
		left = LEFT(expr);

		dvrt_stack();
		deep_rop = expr_code(exprtype,RIGHT(expr),value_flag);
		dvrt_rop = dvrt_unstack();
		right = RIGHT(expr);
		}

	/* Choose the depth cells to examine */
	if(exprtype->type & FP_NODE)
		{
		if(exprtype->val.integral > Sizeof_int)
			{
			free_tmp = free_dtemp;
			get_tmp = get_dtemp;
			}
		else
			{
			free_tmp = free_temp;
			get_tmp = get_temp;
			}
		tdeep_lop = deep_lop & FLT_STACK;
		tdeep_rop = deep_rop & FLT_STACK;
		tdeep_inc = FLT_DEEP(1);
		tdeep_max = MAX_FLT_STACK;
		}
	else
		{
		free_tmp = free_temp;
		get_tmp = get_temp;
		tdeep_lop = deep_lop & INT_STACK;
		tdeep_rop = deep_rop & INT_STACK;
		tdeep_inc = INT_DEEP(1);
		tdeep_max = MAX_INT_STACK;
		}
/*
 *	The values loaded are not needed by a higher level, so emit any
 *	code (generated by side-effects) and leave.
 */
	if(! (value_flag & NEED_VALUE))
		{
		dvrt_append(dvrt_lop);
		dvrt_append(dvrt_rop);
		total_deep = depth_max(deep_lop,deep_rop);
		}
	else
		{
/*
 *	Always load diversion which uses the most stack space first.
 */
		if(tdeep_lop < tdeep_rop)
			{
			/* Load right first */
			dvrt_append(dvrt_rop);
			dvrt_append(dvrt_lop);
			if(! (expr->operator & OP_COMM))
				rev(exprtype);

			total_deep = depth_max(deep_lop + tdeep_inc,deep_rop);
			}
		else if((tdeep_rop < tdeep_lop) || (tdeep_lop < tdeep_max))
			{
			/* Load left first */
			dvrt_append(dvrt_lop);
			dvrt_append(dvrt_rop);
			total_deep = depth_max(deep_lop,deep_rop + tdeep_inc);
			}
		else
			{
/*
 *	Well, both diversions use the full three elements.  Load in the
 *	preferred order for non-commutative operators using a "temp".
 */

/*
 *	You may not believe the following:  We know that we are going to
 *	overflow the stack, so we must allocate a temp that is not used
 *	in either of the expressions.  The expressions have already been
 *	evaluated, but may be using the wrong temp, since we need to
 *	essentially "insert" a temp between the code already generated.
 *	Instead of anything fancy, we simply throw away the diversion,
 *	allocate a temp and then re-evaluate the expression.  The second
 *	time of evaluation will use a higher temp, safely out of range
 *	of the temp we use for intermediate storage.
 */
			temp = (*get_tmp)();
			dvrt_free(dvrt_rop);	/* Flush the original */
			dvrt_free(dvrt_lop);	/* And the other side */
			(VOID) expr_code(exprtype,right,value_flag);
			store_temp(expr->node_type,temp);
			(VOID) expr_code(exprtype,left,value_flag);
			load_temp(expr->node_type,temp);
			(*free_tmp)(temp);
			total_deep = depth_max(depth_max(deep_lop,deep_rop),
				tdeep_max);
			}
		}
	return (total_deep);
	}

/*
 *	This routine evaluates the operands in the "best" way to preserve
 *	minimum stack depth.  A temp is passed into this routine that is
 *	used to hold the address of the target (left-side) Lvalue.
 *	The values returned are both "values" and in the proper order
 *	for non-commutative operators.
 *
 *	The caller uses the temp to refetch the address for storing after
 *	the operation is performed.
 *
 *	One other optimization would be to note when the left-operand
 *	(the one evaluated to an address) is actually a pointer (almost
 *	simple) and not use a temp in those cases.  How to do this??
 */
PUBLIC	unsigned	int
assign_op(expr,addrtemp)
	struct	op_node		*expr;
	TEMP 			addrtemp;
	{
	unsigned	int	deep_lop;
	unsigned	int	deep_rop;
	struct	dvrt		*dvrt_lop;
	struct	dvrt		*dvrt_rop;
	unsigned	int	final_deep;
	unsigned	int	tdeep_lop;
	unsigned	int	tdeep_max;
	unsigned	int	tdeep_rop;
	unsigned	int	total_deep;
	struct	id_node		*totype;

	totype = op_types(expr);

	dvrt_stack();
	deep_rop = expr_code(totype,RIGHT(expr),NEED_VALUE);
	dvrt_rop = dvrt_unstack();

	dvrt_stack();
	deep_lop = expr_code(NULL,LEFT(expr),NEED_VALUE | EVAL_ADDR);
	dvrt_lop = dvrt_unstack();

	/* Choose the depth cells to examine */
	if(totype->type & FP_NODE)
		{
		tdeep_lop = depth_max(FLT_DEEP(1),deep_lop & FLT_STACK);
		tdeep_rop = deep_rop & FLT_STACK;
		tdeep_max = MAX_FLT_STACK;
		final_deep = FLT_DEEP(2);
		}
	else
		{
		tdeep_lop = depth_max(INT_DEEP(1),deep_lop & INT_STACK);
		tdeep_rop = deep_rop & INT_STACK;
		tdeep_max = MAX_INT_STACK;
		final_deep = INT_DEEP(2);
		}
/*
 *	Always load diversion which uses the most stack space first.
 */
	if(tdeep_lop < tdeep_rop)
		{
		dvrt_append(dvrt_rop);
		dvrt_append(dvrt_lop);
		if(CPU_DUP)
			no_oper(TOP_DUP);
		store_temp(Inttype,addrtemp);
		if(CPU_NO_DUP)
			load_temp(Inttype,addrtemp);
/*
 *	Compute the depth used by the left + right operands loaded at the same
 *	time.  Note that one may occupy the FLOATING stack and the other WILL
 *	(lop) occupy the INTEGER stack (an address).
 */
		total_deep = depth_max(deep_rop,(unsigned int) (deep_lop +
			((totype->type & FP_NODE) ?
				FLT_DEEP(1) : INT_DEEP(1))));
		if(CPU_DUP)
			{
			total_deep = depth_max(total_deep,(unsigned int)
				((totype->type & FP_NODE) ?
					INT_DEEP(2) : INT_DEEP(3)));
			}
		total_deep = depth_max(total_deep,
			load_indirect(LEFT(expr),totype,(INT32) 0,NEED_VALUE));

		if(! (expr->operator & OP_COMM))
			rev(totype);
		}
	else if((tdeep_rop < tdeep_lop) || (tdeep_lop < tdeep_max))
		{
		dvrt_append(dvrt_lop);
		if(CPU_DUP)
			no_oper(TOP_DUP);
		store_temp(Inttype,addrtemp);
		if(CPU_NO_DUP)
			{
			load_temp(Inttype,addrtemp);
			total_deep = deep_lop;
			}
		else
			total_deep = depth_max(deep_lop,INT_DEEP(2));

		total_deep = depth_max(total_deep,load_indirect(LEFT(expr),
			totype,(INT32) 0,NEED_VALUE));
		dvrt_append(dvrt_rop);
/*
 *	Total depth includes the one added from evaluating the address of
 *	the left operand.
 */
		total_deep = depth_max(total_deep, (unsigned int) (deep_rop +
			((totype->type & FP_NODE) ?
				FLT_DEEP(1) : INT_DEEP(1))));
		}
	else
		{
/*
 *	Well, both diversions use the full three elements.
 */
		dvrt_append(dvrt_lop);
		store_temp(Inttype,addrtemp);
		dvrt_append(dvrt_rop);
		load_temp(Inttype,addrtemp);
		total_deep = depth_max(deep_lop,deep_rop);
		total_deep = depth_max(total_deep,load_indirect(LEFT(expr),
			totype,(INT32) 0,NEED_VALUE));
		if(! (expr->operator & OP_COMM))
			rev(totype);
		}

	return (depth_max(final_deep,total_deep));
	}

/*
 *	Issue instructions necessary to load an lvalue into the "accumulator".
 */

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#pragma	optimize("g",off)		/* Disable global common subs */
#pragma	optimize("l",off)		/* Disable loop optimizations */
#endif	/* MSC_OPT */

PUBLIC	unsigned	int
do_load(totype,expr,value_flag)
	struct	id_node		*totype;
	struct	op_node		*expr;
	int			value_flag;
	{
	BOOL			addrflag;
	char			buffer1[NAMEMAX + 1];
	BOOL			converted;
	unsigned	int	depth;
	INT32			offset;
	INT32			value;
	struct	op_node		*var;

	converted = FALSE;

	depth = INT_DEEP(1);		/* Default to 1 integer on stack */

	if(value_flag & NEED_VALUE)
		{
		if(isarray(expr))
			{
/*
 *	If this node is still an array, we only need the address.
 */
			value_flag |= EVAL_ADDR;
			}

		if(! ISTERMINAL(expr))
			{
			var = constaddr(expr,&offset,&addrflag);
			if(addrflag)
				value_flag |= EVAL_ADDR;
			}
		else
			{
			offset = 0;
			var = expr;
			}

		if(var->right.ptr.terminal->diversion != NULL)
			dvrt_copy(var->right.ptr.terminal->diversion);

		switch((int) (var->right.ptr.terminal->info.storage_class))
			{
			case	SC_AUTO:
			case	SC_REG:
			case	SC_ARG:
		if(((value_flag & EVAL_ADDR)			||
			(EXPRTYPE(expr) & (FP_NODE | CHAR_NODE))||
			(totype->type & FP_NODE)) &&
			(! (EXPRTYPE(expr) & BITFIELD_NODE)))
			{
			instruction(TOP_LDLP,ABS,WORKSPACE,NULL,
				-(word(-var->right.ptr.terminal->
					address.offset) - word(offset)),
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
				);

			if((offset %= Sizeof_int) != 0)
				abs_oper(TOP_ADC,(INT32) offset);

			if(! (value_flag & EVAL_ADDR))
				{
				if(EXPRTYPE(expr) & FP_NODE)
					{
					fp_oper(expr->node_type,TOP_BAD,
						TOP_FPLDNLSN,TOP_FPLDNLDB);
					depth = INT_DEEP(1) | FLT_DEEP(1);
					}
				else if(EXPRTYPE(expr) &
					(INT_NODE | SHORT_NODE | LONG_NODE))
					{
					if(EXPRTYPE(expr) & SIGNED_NODE)
						{
						fp_oper(totype,TOP_BAD,
							TOP_FPI32TOR32,
							TOP_FPI32TOR64);
						}
					else
						{
						if(CPU_FPU)
							{
						no_oper(TOP_FPB32TOR64);
						if(totype->val.integral <=
							Sizeof_int)
								{
						no_oper(TOP_FPUR64TOR32);
								}
							}
						else
							{
						fp_oper(totype,TOP_BAD,
							TOP_U32TOR32,
							TOP_FPB32TOR64);
							}
						}
					depth = INT_DEEP(1) | FLT_DEEP(1);
					converted = TRUE;
					}
				else
					{
					no_oper(TOP_LB);
					if(EXPRTYPE(expr) & SIGNED_NODE)
						{
						/* Extend the sign */
						loadconstant((INT32) 0x80);
						no_oper(TOP_XWORD);
						depth = INT_DEEP(2);
						}
					}
				}
			}
		else
			{
			instruction(TOP_LDL, ABS, WORKSPACE, NULL,
				-(word(-var->right.ptr.terminal->
					address.offset) - word(offset)),
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
				);
			}
		break;

			case	SC_CONSTANT:
		if((value_flag & EVAL_ADDR) || (! (ISCONSTANT(expr))))
			{
			loadconstant((INT32) var->right.ptr.terminal->
				val.integral + offset);
			if(! (value_flag & EVAL_ADDR))
				{
				if(EXPRTYPE(expr) & FP_NODE)
					{
					fp_oper(expr->node_type,TOP_BAD,
						TOP_FPLDNLSN,TOP_FPLDNLDB);
					depth = INT_DEEP(1) | FLT_DEEP(1);
					}
				else if((totype->type & FP_NODE) &&
					(EXPRTYPE(expr) & (INT_NODE |
						SHORT_NODE | LONG_NODE)))
					{
					if(EXPRTYPE(expr) & SIGNED_NODE)
						{
						fp_oper(totype,TOP_BAD,
							TOP_FPI32TOR32,
							TOP_FPI32TOR64);
						}
					else
						{
						if(CPU_FPU)
							{
						no_oper(TOP_FPB32TOR64);
						if(totype->val.integral <=
							Sizeof_int)
								{
						no_oper(TOP_FPUR64TOR32);
								}
							}
						else
							{
						fp_oper(totype,TOP_BAD,
							TOP_U32TOR32,
							TOP_FPB32TOR64);
							}
						}
					depth = INT_DEEP(1) | FLT_DEEP(1);
					converted = TRUE;
					}
 				else if(EXPRTYPE(expr) & CHAR_NODE)
					{
					no_oper(TOP_LB);
					if(EXPRTYPE(expr) & SIGNED_NODE)
						{
						loadconstant((INT32) 0x80);
						no_oper(TOP_XWORD);
						depth = INT_DEEP(2);
						}
					}
				else
					abs_oper(TOP_LDNL,(INT32) 0);
				}
			}
		else
			{
			if(totype->type & FP_NODE)
				{
				depth = fp_val(totype,var);
				converted = TRUE;
				}
			else
				{
				if(var->node_type->type & FP_NODE)
					{
/*
 *	Floating constant that is ref'd as integral...
 */
					value = f_to_l(var->
						right.ptr.terminal->val.fvalue,
							TRUE);
					converted = TRUE;
					}
				else
					{
					value = var->right.ptr.terminal->
						val.integral;
					}

				if(totype->type & CHAR_NODE)
					{
					if(totype->type & SIGNED_NODE)
						{
						if(value & 0x80)
							value |= 0xFFFFFF80L;
						else
							value &= 0x7F;
						}
					else
						value &= 0xFF;
					}
				loadconstant((INT32) value);
				}
			}
		break;

			case	SC_EXTERN:
			case	SC_ST_EXT:
		var->right.ptr.terminal->type |= REFERENCED_NODE;
		/* FALL THROUGH */

			case	SC_GLOBAL:
			case	SC_ST_GLB:
			case	SC_STATIC:
			case	SC_LITERAL:
		if(Static_link &&
			(! (var->right.ptr.terminal->type & FUNC_NODE)) &&
				(var->right.ptr.terminal->info.storage_class !=
					SC_LITERAL))
			{
/*
 *	Compute global data address using static link approach.
 */
			instruction(TOP_LDL,ABS,WORKSPACE,NULL,Slboff,
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
				);
			instruction(TOP_ADC,ABS,
				real_name(var->right.ptr.terminal,buffer1),
					SLBNAME,offset,
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
				);
			}
		else
			{
/*
 *	Use assembler pseudo-op to allow for optimized constant loading using
 *	the same rules at link time as loadconstant().
 */
			instruction(TOP_LDC,ABS,
				real_name(var->right.ptr.terminal,buffer1),
					NULL,offset,
#ifdef	DEBUG
				((Debug && strcmp(real_name(var->
					right.ptr.terminal,buffer1),
					var->right.ptr.terminal->name)) ?
					var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
				);
			}

		if(! (value_flag & EVAL_ADDR))
			{
			if(EXPRTYPE(expr) & BITFIELD_NODE)
				abs_oper(TOP_LDNL,(INT32) 0);
			else if(EXPRTYPE(expr) & FP_NODE)
				{
				fp_oper(expr->node_type,TOP_BAD,TOP_FPLDNLSN,
					TOP_FPLDNLDB);
				depth = INT_DEEP(1) | FLT_DEEP(1);
				}
			else if((totype->type & FP_NODE) &&
				(EXPRTYPE(expr) & (INT_NODE | SHORT_NODE |
					LONG_NODE)))
				{
				if(EXPRTYPE(expr) & SIGNED_NODE)
					{
					fp_oper(totype,TOP_BAD,TOP_FPI32TOR32,
						TOP_FPI32TOR64);
					}
				else
					{
					if(CPU_FPU)
						{
						no_oper(TOP_FPB32TOR64);
						if(totype->val.integral <=
							Sizeof_int)
							{
						no_oper(TOP_FPUR64TOR32);
							}
						}
					else
						{
						fp_oper(totype,TOP_BAD,
							TOP_U32TOR32,
							TOP_FPB32TOR64);
						}
					}
				depth = INT_DEEP(1) | FLT_DEEP(1);
				converted = TRUE;
				}
			else if(EXPRTYPE(expr) & CHAR_NODE)
				{
				no_oper(TOP_LB);
				if(EXPRTYPE(expr) & SIGNED_NODE)
					{
					loadconstant((INT32) 0x80);
					no_oper(TOP_XWORD);
					depth = INT_DEEP(2);
					}
				}
			else
				abs_oper(TOP_LDNL,(INT32) 0);
			}
		break;

			case	SC_RSTEMP:
		value = var->right.ptr.terminal->address.tempsize;
		if(value > Structspace)
			Structspace = value;
		instruction(TOP_LDLP,ABS,STRUCTBASE,
			((value != 0) ? NULL : var->right.ptr.terminal->name),
				(INT32) 0,
#ifdef	DEBUG
			(Debug ? "struct/union function return ptr" : NULL)
#else
			NULL
#endif	/* DEBUG */
			);
		break;
			}
		/* Do some fix-up if it was a bitfield that was loaded */
		if(EXPRTYPE(expr) & BITFIELD_NODE)
			depth = depth_max(depth,bit_field(expr,value_flag));

		if(! converted && (! (value_flag & EVAL_ADDR)))
			{
			depth = depth_max(depth,cnvt_type(totype,
				expr->node_type));
			}
		}
	else
		depth = INT_DEEP(0);

	return (depth);
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#pragma	optimize("g",on)		/* Enable global common subs */
#pragma	optimize("l",on)		/* Enable loop optimizing */
#endif	/* MSC_OPT */

/*
 *	Store the top stack element somewhere.
 */

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#endif	/* MSC_OPT */

PUBLIC	unsigned	int
do_store(expr)
	struct	op_node		*expr;
	{
	BOOL			addrflag;
	char			buffer1[NAMEMAX + 1];
	unsigned	int	depth;
	INT32			offset;
	struct	op_node		*var;

	depth = INT_DEEP(1);

	if(! ISTERMINAL(expr))
		{
		if(! expr->lvalue)
			fatal("Compiler error: do_store: not Lvalue");
		else
			var = constaddr(expr,&offset,&addrflag);
		}
	else
		{
		offset = 0;
		var = expr;
		}

	if(var->right.ptr.terminal->diversion != NULL)
		{
		/* Divert literal portion of terminal into literal stream */
		dvrt_copy(var->right.ptr.terminal->diversion);
		}

	if(EXPRTYPE(expr) & BITFIELD_NODE)
		{
		/* Mask and optionally shift the value to stored in bitfield */
		loadconstant((INT32) (1L <<
			expr->node_type->val.bitfield.fieldwidth) - 1);

		no_oper(TOP_AND);

		if(expr->node_type->val.bitfield.bitoffset != 0)
			{
			loadconstant((INT32) expr->node_type->
				val.bitfield.bitoffset);
			no_oper(TOP_SHL);
			}

		depth = INT_DEEP(2);
		}

	switch((int) (var->right.ptr.terminal->info.storage_class))
		{
		case	SC_AUTO:
		case	SC_REG:
		case	SC_ARG:
			if(EXPRTYPE(expr) & BITFIELD_NODE)
				{
				/* Local offset.  Load, mask, or in, store */
				instruction(TOP_LDL,ABS,WORKSPACE,NULL,
					-(word(-var->right.ptr.terminal->
						address.offset)-word(offset)),
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
					);
/*
 *	Load a shifted mask to remove the changing part of target.
 */
				loadconstant((INT32)
/*lint -e502 Doesn't like signed shift values */
		~(((1L << expr->node_type->val.bitfield.fieldwidth) - 1)
			<< expr->node_type->val.bitfield.bitoffset));
/*lint +e502 Doesn't like signed shift values */

				no_oper(TOP_AND);
				no_oper(TOP_OR);
				instruction(TOP_STL,ABS,WORKSPACE,NULL,
					-(word(-var->right.ptr.terminal->
						address.offset)-word(offset)),
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
					);
				depth = INT_DEEP(3);
				}
			else if(EXPRTYPE(expr) & (CHAR_NODE | FP_NODE))
				{
				instruction(TOP_LDLP,ABS,WORKSPACE,NULL,
					-(word(-var->right.ptr.terminal->
						address.offset)-word(offset)),
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
					);

				if(EXPRTYPE(expr) & FP_NODE)
					{
					fp_oper(expr->node_type,TOP_BAD,
						TOP_FPSTNLSN,TOP_FPSTNLDB);
					depth = FLT_DEEP(1) | INT_DEEP(1);
					}
				else
					{
					if((offset %= Sizeof_int) != 0)
						{
						abs_oper(TOP_ADC,
							(INT32) offset);
						}

					no_oper(TOP_SB);
					depth = INT_DEEP(2);
					}
				}
			else
				{
				instruction(TOP_STL,ABS,WORKSPACE,NULL,
					-(word(-var->right.ptr.terminal->
						address.offset)-word(offset)),
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
					);
				}
			break;

		case	SC_CONSTANT:
			if(EXPRTYPE(expr) & BITFIELD_NODE)
				{
/*
 *	A constant address to bitfield.  Load, mask, or in, store.
 */
				loadconstant(var->right.ptr.terminal->
					val.integral + offset);
				abs_oper(TOP_LDNL,(INT32) 0);
/*
 *	Load mask to clear part of target being updated.
 */
/*lint -e502 Doesn't like signed shift values */
				loadconstant((INT32)
		~(((1L << expr->node_type->val.bitfield.fieldwidth) - 1)
			<< expr->node_type->val.bitfield.bitoffset));
/*lint +e502 Doesn't like signed shift values */

				no_oper(TOP_AND);
				no_oper(TOP_OR);
				loadconstant(var->right.ptr.terminal->
					val.integral + offset);
				abs_oper(TOP_STNL,(INT32) 0);
				depth = INT_DEEP(3);
				}
			else
				{
				/* A constant address of something */
				loadconstant(var->right.ptr.terminal->
					val.integral + offset);

				if(expr->node_type->val.integral ==
					SIZEOF_CHAR)
					{
					no_oper(TOP_SB);
					}
				else
					abs_oper(TOP_STNL,(INT32) 0);
				depth = INT_DEEP(2);
				}
			break;

		case	SC_EXTERN:
		case	SC_ST_EXT:
			var->right.ptr.terminal->type |= REFERENCED_NODE;
			/* FALL THROUGH */

		case	SC_GLOBAL:
		case	SC_ST_GLB:
		case	SC_STATIC:
			if(Static_link)
				{
/*
 *	Compute global data address using static link approach.
 */
				instruction(TOP_LDL,ABS,WORKSPACE,NULL,Slboff,
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
					);
				instruction(TOP_ADC,ABS,
					real_name(var->right.ptr.terminal,
						buffer1),SLBNAME,offset,
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
					);
				}
			else
				{
/*
 *	Use assembler pseudo-op to allow for optimized constant loading using
 *	the same rules at link time as loadconstant().
 */
				instruction(TOP_LDC,ABS,
					real_name(var->right.ptr.terminal,
						buffer1),NULL,offset,
#ifdef	DEBUG
				((Debug && strcmp(real_name(var->
					right.ptr.terminal,buffer1),
					var->right.ptr.terminal->name)) ?
					var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
					);
				}
			if(EXPRTYPE(expr) & BITFIELD_NODE)
				{
				abs_oper(TOP_LDNL,(INT32) 0);
/*
 *	Load mask to clear part of target being updated.
 */
/*lint -e502 Doesn't like signed shift values */
				loadconstant((INT32)
		~(((1L << expr->node_type->val.bitfield.fieldwidth) - 1)
			<< expr->node_type->val.bitfield.bitoffset));
/*lint +e502 Doesn't like signed shift values */

				no_oper(TOP_AND);
				no_oper(TOP_OR);

				/* Now do the store */
				if(Static_link)
					{
/*
 *	Compute global data address using static link approach.
 */
					instruction(TOP_LDL,ABS,WORKSPACE,NULL,
						Slboff,
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
						);
					instruction(TOP_ADC,ABS,real_name(var->
						right.ptr.terminal,buffer1),
							SLBNAME,offset,
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
						);
					}
				else
					{
/*
 *	Use assembler pseudo-op to allow for optimized constant loading using
 *	the same rules at link time as loadconstant().
 */
					instruction(TOP_LDC,ABS,real_name(var->
						right.ptr.terminal,buffer1),
							NULL,offset,
#ifdef	DEBUG
				((Debug && strcmp(real_name(var->
					right.ptr.terminal,buffer1),
					var->right.ptr.terminal->name)) ?
					var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
						);
					}
				abs_oper(TOP_STNL,(INT32) 0);
				depth = INT_DEEP(3);
				}
			else
				{
				if(EXPRTYPE(expr) & FP_NODE)
					{
					fp_oper(expr->node_type,TOP_BAD,
						TOP_FPSTNLSN,TOP_FPSTNLDB);
					depth = FLT_DEEP(1) | INT_DEEP(1);
					}
				else
					{
					if(EXPRTYPE(expr) & CHAR_NODE)
						no_oper(TOP_SB);
					else
						abs_oper(TOP_STNL,(INT32) 0);

					depth = INT_DEEP(2);
					}
				}
			break;
		}
	return (depth);
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#endif	/* MSC_OPT */

/*
 *	Assumes the top of stack contains the address (byte or word)
 *	and the next-on-stack contains the value to be stored.
 *
 *	The entry parameter points to the op_node for the base of
 *	the operand to be stored, so the type may be inferred.
 */
PUBLIC	unsigned	int
store_indirect(node,fromtype)
	struct	op_node		*node;
	struct	id_node		*fromtype;
	{
	unsigned	int	depth;
	TEMP			temp;

	if(fromtype == NULL)
		fromtype = node->node_type;

	if(EXPRTYPE(node) & BITFIELD_NODE)
		{
		temp = get_temp();
		store_temp(Inttype,temp);	/* Store the address */
/*
 *	Mask and (optionally) shift the value to be stored in bitfield.
 */
		loadconstant((INT32)
			(1L << node->node_type->val.bitfield.fieldwidth) - 1);
		no_oper(TOP_AND);

		if(node->node_type->val.bitfield.bitoffset != 0)
			{
			loadconstant((INT32) node->node_type->
				val.bitfield.bitoffset);
			no_oper(TOP_SHL);
			}

		load_temp(Inttype,temp);	/* Reload address */
		abs_oper(TOP_LDNL,(INT32) 0);	/* Get contents */

		/* Mask out contents of old value to be replaced */
/*lint -e502 Doesn't like signed shift values */
		loadconstant((INT32)
		~(((1L << node->node_type->val.bitfield.fieldwidth) - 1)
			<< node->node_type->val.bitfield.bitoffset));
/*lint +e502 Doesn't like signed shift values */

		no_oper(TOP_AND);
		no_oper(TOP_OR);
		load_temp(Inttype,temp);	/* Get address again */

		/* The final store will be accomplished below */
		free_temp(temp);
		depth = INT_DEEP(3);	/* Three total used */
		}
	else if(fromtype->type & FP_NODE)
		{
		if(EXPRTYPE(node) & FP_NODE)
			{
			if(fromtype->val.integral !=
				node->node_type->val.integral)
				{
				fp_oper(fromtype,TOP_BAD,TOP_FPUR32TOR64,
					TOP_FPUR64TOR32);
				}
			fp_oper(node->node_type,TOP_BAD,TOP_FPSTNLSN,
				TOP_FPSTNLDB);
			return (FLT_DEEP(1) | INT_DEEP(1));
			}
		else if(EXPRTYPE(node) & (SHORT_NODE | INT_NODE | LONG_NODE))
			{
			if(CPU_FPU && (EXPRTYPE(node) & SIGNED_NODE))
				{
				if(! Round_mode)
					no_oper(TOP_FPURZ);
				no_oper(TOP_FPINT);
				depth = FLT_DEEP(1) | INT_DEEP(2);
				}
			else
				depth = FLT_DEEP(1) | INT_DEEP(1);
			no_oper(TOP_FPSTNLI32);
			return (depth);
			}
		else
			{
/*
 *	Convert to an integer then allow normal store to happen.
 */
			depth = cnvt_type(node->node_type,fromtype);
			rev(Inttype);	/* Leave address on top of stack */
			}
		}

	if(EXPRTYPE(node) & CHAR_NODE)
		{
		/* Char type -- store just a byte */
		no_oper(TOP_SB);
		}
	else
		{
		/* Other (INT or POINTER) type. */
		abs_oper(TOP_STNL,(INT32) 0);
		}

	depth = depth_max(depth,INT_DEEP(2));

	return (depth);
	}

/*
 *	Assumes the top of stack contains the address (byte or word.)
 *
 *	The entry parameter points to the op_node for the base of
 *	the operand to be stored, so the type may be inferred.
 */
PUBLIC	unsigned	int
load_indirect(node,totype,offset,value_flag)
	struct	op_node		*node;
	struct	id_node		*totype;
	INT32			offset;
	int			value_flag;
	{
	BOOL			converted;
	unsigned	int	depth;

	if(totype == NULL)
		totype = node->node_type;

	converted = FALSE;

	if((EXPRTYPE(node) & FP_NODE) && (totype->type & FP_NODE))
		{
		if(offset != 0)
			abs_oper(TOP_ADC,(INT32) offset);

		fp_oper(node->node_type,TOP_BAD,TOP_FPLDNLSN,TOP_FPLDNLDB);
		if(totype->val.integral == node->node_type->val.integral)
			converted = TRUE;

		depth = INT_DEEP(1) | FLT_DEEP(1);
		}
	else if((EXPRTYPE(node) & (SHORT_NODE | INT_NODE | LONG_NODE)) &&
		(totype->type & FP_NODE))
		{
		if(offset != 0)
			abs_oper(TOP_ADC,(INT32) offset);

		if(EXPRTYPE(node) & UNSIGNED_NODE)
			{
			if(CPU_FPU)
				{
				no_oper(TOP_FPB32TOR64);
				if(totype->val.integral <= Sizeof_int)
					no_oper(TOP_FPUR64TOR32);
				}
			else
				{
				fp_oper(totype,TOP_BAD,TOP_U32TOR32,
					TOP_FPB32TOR64);
				}
			}
		else
			fp_oper(totype,TOP_BAD,TOP_FPI32TOR32,TOP_FPI32TOR64);

		depth = INT_DEEP(1) | FLT_DEEP(1);
		converted = TRUE;
		}
	else
		{
		depth = INT_DEEP(1);

		if(EXPRTYPE(node) & CHAR_NODE)
			{
			/* Char type.  Load it and see if to sign extend */
			if(offset != 0)
				abs_oper(TOP_ADC,(INT32) offset);

			no_oper(TOP_LB);

			if(node->node_type->type & SIGNED_NODE)
				{
				/* Need to sign extend */
				loadconstant((INT32) 0x80);
				no_oper(TOP_XWORD);
				depth = INT_DEEP(2);
				}
			}
		else
			{
			/* All other types - INT or POINTER come here */
			abs_oper(TOP_LDNL,(INT32) word(offset));
			}
		}

	/* Do some fix-up if it was a bitfield that was loaded */
	if(EXPRTYPE(node) & BITFIELD_NODE)
		depth = depth_max(depth,bit_field(node,value_flag));

	if(! converted)
		depth = depth_max(depth,cnvt_type(totype,node->node_type));

	return (depth);
	}

/*
 *	Create a string containing the "real name" of a symbol.  If the
 *	symbol is a private variable (static global or static local) then
 *	the name will be the generated number of the label assigned to the
 *	symbol.  Otherwise, use the name given by the programmer.
 */
PUBLIC	char	*
real_name(sym,buf)
	struct	id_node		*sym;
	char			*buf;
	{
	switch((int) (sym->info.storage_class))
		{
		case	SC_LITERAL:	/* For initializers */
		case	SC_STATIC:	/* Static local */
			(VOID) strlabel(sym->address.surrogate,buf);
			break;

		default:
			/* All others use the original name */
			sprintf(buf,"%s%c",sym->name,ASM_UNIQ);
			break;
		}
	return (buf);
	}

/*
 *	 Print specified number as label.
 */
PUBLIC	VOID
printlabel(label)
	LABEL			label;
	{
	char			str[MAXLABEL + 1];

	if(label > 0)
		outline(TOP_LABEL,strlabel(label,str),NULL);
	}

/*
 *	Format specified label as a string.
 */
PUBLIC	char	*
strlabel(inlabel,outstring)
	LABEL			inlabel;
	char			*outstring;
	{
	sprintf(outstring,"?%ld",(long) inlabel);
	return (outstring);
	}

/*
 *	Calculate a maximum stack depth
 *
 *	The depth value is a composite number with four fields:
 *		int_expr_stack_usage	: 3;
 *		int_temps_used		: 5;
 *		float_expr_stack_usage	: 3;
 *		float_temps_used	: 5;
 *
 *	return
 *		min(max(left, right), max_size)
 *	for each field.
 */
PUBLIC	unsigned	int
depth_max(left,right)
	unsigned	int	left;
	unsigned	int	right;
	{
	return (depth_extract(left,right,INT_STACK,MAX_INT_STACK) |
		depth_extract(left,right,INT_TEMPS,MAX_INT_TEMPS) |
		depth_extract(left,right,FLT_STACK,MAX_FLT_STACK) |
		depth_extract(left,right,FLT_TEMPS,MAX_FLT_TEMPS));
	}

PUBLIC	unsigned	int
depth_extract(left,right,mask,max_val)
	unsigned	int	left;
	unsigned	int	right;
	unsigned	int	mask;
	unsigned	int	max_val;
	{
	unsigned	int	temp;

	left &= mask;
	right &= mask;
	temp = ((left > right) ? left : right);
	return ((temp < max_val) ? temp : max_val);
	}

/*
 *	Short form for "EQC 0" instruction.
 */
PUBLIC	VOID
eqc()
	{
	abs_oper(TOP_EQC, (INT32) 0);
	}

/*
 *	Reverse A and B stack registers.
 */
PUBLIC	VOID
rev(exprtype)
	struct	id_node		*exprtype;
	{
	fp_oper(exprtype,TOP_REV,TOP_FPREV,TOP_FPREV);
	}

/*
 *	Store into a temp
 */
PUBLIC	VOID
store_temp(exprtype,temp)
	struct	id_node		*exprtype;
	TEMP			temp;
	{
	if(exprtype->type & FP_NODE)
		{
		instruction(TOP_LDLP,ABS,TEMPBASE,NULL,(INT32) temp,NULL);
		fp_oper(exprtype,TOP_BAD,TOP_FPSTNLSN,TOP_FPSTNLDB);
		}
	else
		instruction(TOP_STL,ABS,TEMPBASE,NULL,(INT32) temp,NULL);
	}

/*
 *	Load from a temp
 */
PUBLIC	VOID
load_temp(exprtype,temp)
	struct	id_node		*exprtype;
	TEMP			temp;
	{
	if(exprtype->type & FP_NODE)
		{
		instruction(TOP_LDLP,ABS,TEMPBASE,NULL,(INT32) temp,NULL);
		fp_oper(exprtype,TOP_BAD,TOP_FPLDNLSN,TOP_FPLDNLDB);
		}
	else
		instruction(TOP_LDL,ABS,TEMPBASE,NULL,(INT32) temp,NULL);
	}

/*
 *	Compute the address of a temp
 */
PUBLIC	VOID
addr_temp(temp)
	TEMP			temp;
	{
	instruction(TOP_LDLP,ABS,TEMPBASE,NULL,(INT32) temp,NULL);
	}

/*
 *	Test for a "simple" operand -- one that can be loaded/stored
 *	easily.  For the transputer, this means an expression that:
 *
 *		1) Is a terminal (simple variable)
 *		2) Subscripted array of constant-only subscripts
 *		3) Indirect to a (2) constant-only subscripted array
 *		4) Indirect to a simple
 *		5) Member or Offset based upon a simple.
 */
PUBLIC	BOOL
is_simple(expr)
	struct	op_node		*expr;
	{
	if(ISTERMINAL(expr))
		return (TRUE);
	else
		return (constnode(expr));
	}

PUBLIC	BOOL
constnode(expr)
	struct	op_node		*expr;
	{
	if(ISTERMINAL(expr))
		{
		if(expr->node_type->type & SU_NODE)
			return (TRUE);
		else
			return (!expr->lvalue);	/* FALSE/lvalue in the expr */
		}
	else if((expr->operator == OP_ADDRESS) && is_simple(RIGHT(expr)))
		return (TRUE);
	else if((expr->operator == OP_INDIRECT) && const2node(RIGHT(expr)))
		return (TRUE);
	else if(((expr->operator == OP_MEMBER) ||
		(expr->operator == OP_OFFSET)) && constnode(LEFT(expr)))
		{
		return (TRUE);
		}
	else
		return (const2node(expr));
	}

PUBLIC	BOOL
const2node(expr)
	struct	op_node		*expr;
	{
	if(ISTERMINAL(expr) &&
		(isarray(expr)				||
		(expr->node_type->type & SU_NODE)	||
		ISCONSTANT(expr)))
		{
		return (TRUE);
		}
	else
		return ((expr->operator == (int)OP_ADD)	&&
			ISRIGHTCONSTANT(expr)		&&
			!isptr(expr)			&&
			isarray(expr)			&&
			constnode(LEFT(expr)));
	}

/*
 *	Calculate constant address expressions yielding one address and
 *	a single offset.  *addr is set to TRUE if to eval for address,
 *	not value.
 */
PUBLIC	struct	op_node	*
constaddr(expr,offset,addr)
	struct	op_node		*expr;
	INT32			*offset;
	BOOL			*addr;
	{
	struct	op_node		*retv;
	INT32			value;

	*offset = 0;

	if(ISTERMINAL(expr))
		{
		/* Remember if this is an address or not */
		*addr = (isarray(expr) ||
			(ISCONSTANT(expr) && (EXPRTYPE(expr) & PTR_NODE)));

		retv = expr;
		}
	else if(expr->operator == OP_ADDRESS)
		{
		if(ISTERMINAL(RIGHT(expr)))
			retv = RIGHT(expr);
		else
			retv = constaddr(RIGHT(expr),offset,addr);
		*addr = TRUE;
		}
	else if(expr->operator == OP_INDIRECT)
		{
		retv = constaddr(RIGHT(expr),offset,addr);
		*addr = (isarray(expr) || (EXPRTYPE(expr) & SU_NODE));
		}
	else if((expr->operator == OP_MEMBER) || (expr->operator == OP_OFFSET))
		{
		retv = constaddr(LEFT(expr),offset,addr);
		*offset += RIGHT(expr)->right.ptr.terminal->val.integral;
		*addr = (isarray(expr) || (EXPRTYPE(expr) & SU_NODE));
		}
	else
		{
		retv = constaddr(LEFT(expr),offset,addr);

		/* Now calculate subscript portion at this level */
		if(ISRIGHTCONSTANT(expr))
			{
			value = RIGHTCONSTANT(expr);

			if(expr->operator == (int) OP_ADD)
				*offset += value;
			else
				*offset -= value;

			*addr = (isarray(expr) || (EXPRTYPE(expr) & SU_NODE));
			}
		else
			fatal("Compiler error: constaddr: not constant");
		}
	return (retv);
	}

/*
 *	Generate non-relative instruction.
 */
PUBLIC	VOID
abs_oper(opcode,value)
	int			opcode;
	INT32			value;
	{
	instruction(opcode,ABS,NULL,NULL,value,NULL);
	}

/*
 *	Generate instructions without operand fields.
 */
PUBLIC VOID
no_oper(opcode)
	int			opcode;
	{
	instruction(opcode,NONE,NULL,NULL,(INT32) 0,NULL);
	}
/*
 *	Generate an instruction to load a integer or floating point zero.
 */
PUBLIC	unsigned	int
ld_zero(type)
	struct	id_node		*type;
	{
	if(type->type & FP_NODE)
		{
		if(type->val.integral > SIZEOF_FLOAT)
			no_oper(TOP_FPLDZERODB);
		else
			no_oper(TOP_FPLDZEROSN);
		return (FLT_DEEP(1));
		}
	loadconstant((INT32) 0);
	return (INT_DEEP(1));
	}

/*
 *	Generate a single or double precision floating point instruction.
 */
PUBLIC	VOID
fp_oper(type,intins,sfpins,lfpins)
	struct	id_node		*type;
	int			intins;
	int			sfpins;
	int			lfpins;
	{
	if(type->type & FP_NODE)
		{
		if(type->val.integral > SIZEOF_FLOAT)
			no_oper(lfpins);
		else
			no_oper(sfpins);
		}
	else
		no_oper(intins);
	}

/*
 *	Load the address of the op_node term (converted to be a floating
 *	point constant).
 */
PUBLIC	unsigned	int
fp_oplit(type,term)
	struct	id_node		*type;
	struct	op_node		*term;
	{
	char			buffer[NAMEMAX + 1];
	char			buffer1[MAXLABEL + 1];
	struct	id_node		*sym;
	struct	id_node		*var;

	var = term->right.ptr.terminal;

	if(var->type & FP_NODE)
		sprintf(buffer,"%.20g",var->val.fvalue);
	else
		{
		if(var->type & SIGNED_NODE)
			sprintf(buffer,"%ld",var->val.integral);
		else
			sprintf(buffer,"%lu",var->val.uintegral);
		}

	if(type->val.integral > SIZEOF_FLOAT)
		sym =  add_lit(buffer,type,NS_FP64LITS);
	else
		sym =  add_lit(buffer,type,NS_FP32LITS);

	if(sym != NULL)
		sym->address.surrogate = getlabel();
	else
		{
		if(type->val.integral > SIZEOF_FLOAT)
			sym = lookup_node(buffer,NS_FP64LITS);
		else
			sym = lookup_node(buffer,NS_FP32LITS);
		}
	sym->type |= REFERENCED_NODE;
	instruction(TOP_LDC,ABS,strlabel(sym->address.surrogate,buffer1),NULL,
		(INT32) 0,
#ifdef	DEBUG
		Debug ? var->name : NULL);
#else
		NULL);
#endif	/* DEBUG */
	return (INT_DEEP(1));
	}

/*
 *	Generate a literal floating point value and load it.
 */
PUBLIC	unsigned	int
fp_val(type,term)
	struct	id_node		*type;
	struct	op_node		*term;
	{
	unsigned int		depth;

	if(((term->right.ptr.terminal->type & FP_NODE) &&
		(term->right.ptr.terminal->val.fvalue != 0.0)) ||
			(!(term->right.ptr.terminal->type & FP_NODE) &&
		(term->right.ptr.terminal->val.integral != 0)))
		{
		depth = depth_max(fp_oplit(type,term),FLT_DEEP(1));
		fp_oper(type,TOP_BAD,TOP_FPLDNLSN,TOP_FPLDNLDB);
		}
	else
		depth = ld_zero(type);

	return (depth);
	}

/*
 *	Convert type of current expression to new type.
 */
PUBLIC	unsigned	int
cnvt_type(totype,fromtype)
	struct	id_node		*totype;
	struct	id_node		*fromtype;
	{
	unsigned	int	depth;
	TEMP			temp;

	if((totype->type & FP_NODE) && (fromtype->type & INTEGRAL_NODE))
		{
/*
 *	Converting an integral to a float.
 */
		temp = get_temp();
		store_temp(Inttype,temp);
		instruction(TOP_LDLP,ABS,TEMPBASE,NULL,(INT32) temp,NULL);
		if(fromtype->type & UNSIGNED_NODE)
			{
/*
 *	Unsigned to floating.
 */
			if(CPU_FPU)
				{
				no_oper(TOP_FPB32TOR64);
				if(totype->val.integral <= Sizeof_int)
					no_oper(TOP_FPUR64TOR32);
				}
			else
				{
				fp_oper(totype,TOP_BAD,TOP_U32TOR32,
					TOP_FPB32TOR64);
				}
			}
		else
			fp_oper(totype,TOP_BAD,TOP_FPI32TOR32,TOP_FPI32TOR64);

		depth = INT_DEEP(1) | FLT_DEEP(1);
		free_temp(temp);
		}
	else if((totype->type & INTEGRAL_NODE) && (fromtype->type & FP_NODE))
		{
		temp = get_temp();
		if(CPU_FPU && (totype->type & SIGNED_NODE))
			{
			if(! Round_mode)
				no_oper(TOP_FPURZ);
			no_oper(TOP_FPINT);
			}
		instruction(TOP_LDLP,ABS,TEMPBASE,NULL,(INT32) temp,NULL);
		no_oper(TOP_FPSTNLI32);
		load_temp(Inttype,temp);
		depth = INT_DEEP(1) | FLT_DEEP(1);
		free_temp(temp);
		}
	else if((totype->type & FP_NODE) && (fromtype->type & FP_NODE) &&
		(totype->val.integral != fromtype->val.integral))
		{
		fp_oper(fromtype,TOP_BAD,TOP_FPUR32TOR64,TOP_FPUR64TOR32);
		depth = INT_DEEP(1) | FLT_DEEP(1);
		}
	else
		depth = 0;

	return (depth);
	}

/*
 *	Return the largest type of the two operands for the given subexpression
 */
PUBLIC	struct	id_node	*
op_types(expr)
	struct	op_node		*expr;
	{
	return (greatest_type(LEFT(expr),RIGHT(expr)));
	}

/* End of tcx6a.c */
