/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: tcx6b.c,v 2.2 92/04/21 11:11:15 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	TCX Compiler (6b of 11)	Expression Syntax Tree Walker
 *
 *	Copyright 1987-1991 by Logical Systems.  All rights reserved.
 *
 *	Published by:	Logical Systems
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	03/17/87
 */

#include	<stdio.h>
#include	<ctype.h>
#include	"tcx.h"
#include	"tcxdecl.h"
#include	"tcxext.h"

/*
 *	Local function declarations.
 */
PRIVATE	VOID			const_sign();

/*
 *	"x.x".
 */
PUBLIC	unsigned	int
do_member(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;
	INT32			value;

	depth = expr_code(NULL,LEFT(expr),value_flag | EVAL_ADDR);

	if(value_flag & NEED_VALUE)
		{
		value = RIGHTCONSTANT(expr);

		if(isarray(expr) || (value_flag & EVAL_ADDR))
			{
			if(value != 0)
				abs_oper(TOP_ADC,(INT32) value);
			}
		else if(expr->node_type->type & FP_NODE)
			{
			depth = depth_max(depth,load_indirect(expr,NULL,value,
				value_flag));
			}
		else if(expr->node_type->val.integral == Sizeof_int)
			{
			abs_oper(TOP_LDNL,(INT32) word(value));

			if(expr->node_type->type & BITFIELD_NODE)
				{
				depth = depth_max(depth,bit_field(expr,
					value_flag));
				}
			}
		else
			{
			depth = depth_max(depth,load_indirect(expr,NULL,value,
				value_flag));
			/* load_indirect() fixes it own bit_field loads */
			}
		}
	return (depth);
	}

/*
 *	Fix up operand after being loaded from bitfield.
 */
PUBLIC	unsigned	int
bit_field(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	/* Do some fix-up if it was a bitfield that was loaded */
	if(value_flag & RELAX_BOOL)
		{
/*
 *	We know the value will be used as a boolean -- don't bother with
 *	shifting before the mask.  Just mask it by constant.
 */
		loadconstant((INT32) ((1L << expr->node_type->
			val.bitfield.fieldwidth) - 1) <<
				expr->node_type->val.bitfield.bitoffset);

		no_oper(TOP_AND);
		}
	else
		{
		if(expr->node_type->val.bitfield.bitoffset != 0)
			{
			loadconstant((INT32) expr->node_type->
				val.bitfield.bitoffset);
			no_oper(TOP_SHR);
			}

		loadconstant((INT32) (1L <<
			expr->node_type->val.bitfield.fieldwidth) - 1);

		no_oper(TOP_AND);

		if(expr->node_type->type & SIGNED_NODE)
			{
			/* Sign extend */
			loadconstant((INT32) (1L << (expr->node_type->
				val.bitfield.fieldwidth - 1)));

			no_oper(TOP_XWORD);
			}
		}
	return (INT_DEEP(2));
	}

/*
 *	"*x".
 */
PUBLIC	unsigned	int
do_indirect(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;

/*
 *	Look ahead and determine if this indirection operator is an array
 *	reference:
 *
 *	1) Not evaluating for address
 *	2) Need final value
 *	3) Next operator is +
 *	4) Left operand is an array
 *	5) Right operand is a constant
 *	6) Current node is not an array (fully instantiated lvalue)
 */
	if((! (value_flag & EVAL_ADDR)) 		&&
		(value_flag & NEED_VALUE)		&&
		(RIGHT(expr)->operator == (int) OP_ADD)	&&
		isptrclass(LEFT(RIGHT(expr)))		&&
		ISRIGHTCONSTANT(RIGHT(expr))		&&
		(! isarray(expr)))
		{
		/* Evaluate the left side of the sum for address */
		depth = expr_code(NULL,LEFT(RIGHT(expr)),
			value_flag & ~EVAL_ADDR);

		/* Now load-indirect using the right-side constant offset */
		depth = depth_max(depth,load_indirect(expr,NULL,
			RIGHTCONSTANT(RIGHT(expr)),value_flag));
		}
	else
		{
		depth = expr_code(NULL,RIGHT(expr),value_flag & ~EVAL_ADDR);
/*
 *	Generate the dereferencing load if not an array, struct, or union
 *	and the value is needed (not address).
 */
		if((! isarray(expr))					&&
			(!(EXPRTYPE(expr) & (STRUCT_NODE | UNION_NODE)))&&
			(value_flag & NEED_VALUE)			&&
			(! (value_flag & EVAL_ADDR)))
			{
			depth = depth_max(depth,load_indirect(expr,NULL,
				(INT32) 0,value_flag));
			}
		}

	return (depth);
	}

/*
 *	"&x".
 */
PUBLIC	unsigned	int
do_address(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	return (expr_code(NULL,RIGHT(expr),value_flag | EVAL_ADDR));
	}

/*
 *	"-x".
 */
PUBLIC	unsigned	int
do_umin(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;

	depth = expr_code(NULL,RIGHT(expr),value_flag);

	if(value_flag & NEED_VALUE)
		{
		if(EXPRTYPE(expr) & FP_NODE)
			{
/*
 *	Subtract current number from zero to achieve negation.
 */
			(VOID) ld_zero(expr->node_type);
			no_oper(TOP_FPREV);
			no_oper(TOP_FPSUB);
			depth = depth_max(depth,FLT_DEEP(2));
			}
		else
			{
			no_oper(TOP_NOT);
			abs_oper(TOP_ADC,(INT32) 1);
			}
		}
	return (depth);
	}

/*
 *	"+x".
 */
PUBLIC	unsigned	int
do_uplus(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	return (expr_code(NULL,RIGHT(expr),value_flag));
	}

/*
 *	"!".
 */
PUBLIC	unsigned	int
do_lnot(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;

	depth = expr_code(expr->node_type,RIGHT(expr),value_flag);

	if(value_flag & NEED_VALUE)
		{
		if(EXPRTYPE(expr) & FP_NODE)
			{
			(VOID) ld_zero(expr->node_type);
			no_oper(TOP_FPEQ);
			depth = depth_max(depth,INT_DEEP(1) | FLT_DEEP(2));
			}
		eqc();
		}

	return (depth);
	}

/*
 *	"~".
 */
PUBLIC	unsigned	int
do_bnot(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned int		depth;

	depth = expr_code(expr->node_type,RIGHT(expr),value_flag);

	if(value_flag & NEED_VALUE)
		{
		no_oper(TOP_NOT);
		}
	return (depth);
	}

/*
 *	Currently, the implementation of op++ and op-- are defined as *dyadic*
 *	operators, whose left node is the Lvalue to be "incremented" and right
 *	node is the *constant* to add to it.  It would be fairly easy to define
 *	an operator that allowed general expressions on the right side, but
 *	since C doesn't have these, we won't implement here.  A test is made
 *	just in case, though.
 *
 *	Also, we KNOW the constant is not ZERO, so no test is made.
 */
PUBLIC	unsigned	int
do_postop(expr,value_flag,sign)
	struct	op_node		*expr;
	int			value_flag;
	int			sign;
	{
	unsigned	int	depth;
	TEMP			temp;
	BOOL			used_temp;

	if(! ISRIGHTCONSTANT(expr))
		fatal("Compiler error: do_postop");
	else if(value_flag & NEED_VALUE)
		{
		if(is_simple(LEFT(expr)))
			{
			/* If a simple, do by loading twice and incrementing */
			depth = do_load(LEFT(expr)->node_type,LEFT(expr),
				NEED_VALUE);

			if(LEFTTYPE(expr) & FP_NODE)
				no_oper(TOP_FPDUP);
			else if(CPU_DUP)
				{
				no_oper(TOP_DUP);
				depth = depth_max(depth,INT_DEEP(2));
				}
			else
				{
				depth = depth_max(depth,
					do_load(LEFT(expr)->node_type,
						LEFT(expr),NEED_VALUE) +
							INT_DEEP(1));
				}

			if(LEFTTYPE(expr) & FP_NODE)
				{
				/* Make it negative if needed */
				RIGHT(expr)->right.ptr.terminal->
					val.integral *= sign;
				depth = depth_max(depth,fp_oplit(LEFT(expr)->
					node_type,RIGHT(expr)) + FLT_DEEP(3));
				fp_oper(LEFT(expr)->node_type,TOP_BAD,
					TOP_FPLDNLADDSN,TOP_FPLDNLADDDB);
				/* Clean up negation if it was performed */
				RIGHT(expr)->right.ptr.terminal->
					val.integral *= sign;
				depth = depth_max(depth,do_store(LEFT(expr)));
				}
			else
				{
/*
 *	Figure out if "do_store" will require the entire integer stack.  If so,
 *	we need to use a temp.
 */
				dvrt_stack();
				if((do_store(LEFT(expr)) & INT_STACK) ==
					INT_DEEP(3))
					{
					used_temp = TRUE;
					}
				else
					used_temp = FALSE;
				dvrt_free(dvrt_unstack());

				if(used_temp)
					{
					temp = get_temp();
					store_temp(Inttype,temp);
					}

				abs_oper(TOP_ADC,
					(INT32) (sign * RIGHTCONSTANT(expr)));

				if(used_temp)
					{
					depth = depth_max(depth,
						do_store(LEFT(expr)));
					load_temp(Inttype,temp);
					free_temp(temp);
					}
				else
					{
					depth = depth_max(depth,
						do_store(LEFT(expr)) +
							INT_DEEP(1));
					}
				}
			}
		else
			{
			depth = expr_code(NULL,LEFT(expr),
				NEED_VALUE | EVAL_ADDR);

			temp = get_temp();
			store_temp(Inttype,temp);
			load_temp(Inttype,temp);
			depth = depth_max(depth,load_indirect(LEFT(expr),NULL,
				(INT32) 0,value_flag));

			if(LEFTTYPE(expr) & FP_NODE)
				no_oper(TOP_FPDUP);
			else if(CPU_DUP)
				{
				no_oper(TOP_DUP);
				depth = depth_max(depth,INT_DEEP(2));
				}
			else
				{
				load_temp(Inttype,temp);
				depth = depth_max(depth,
					load_indirect(LEFT(expr),NULL,
						(INT32) 0,value_flag) +
							INT_DEEP(2));
				}

			if(LEFTTYPE(expr) & FP_NODE)
				{
				/* Make it negative if needed */
				RIGHT(expr)->right.ptr.terminal->
					val.integral *= sign;
				depth = depth_max(depth,fp_oplit(LEFT(expr)->
					node_type,RIGHT(expr)) + FLT_DEEP(3));
				fp_oper(LEFT(expr)->node_type,TOP_BAD,
					TOP_FPLDNLADDSN,TOP_FPLDNLADDDB);
				/* Clean up negation if it was performed */
				RIGHT(expr)->right.ptr.terminal->
					val.integral *= sign;
				load_temp(Inttype, temp);
				depth = depth_max(depth,
					store_indirect(LEFT(expr),NULL));

				}
			else
				{
				abs_oper(TOP_ADC,
					(INT32) sign * RIGHTCONSTANT(expr));
				load_temp(Inttype,temp);
				depth = depth_max(depth,
					store_indirect(LEFT(expr),NULL) +
						INT_DEEP(1));
				}
			free_temp(temp);
			}
		}
	else
		{
/*
 *	It's simple: do it as a += or -=.  The final results won't be reloaded.
 */
		if(sign >= 0)
			depth = do_aadd(expr,value_flag);
		else
			depth = do_asub(expr,value_flag);
		}
	return (depth);
	}

/*
 *	"x*x".
 */
PUBLIC	unsigned int
do_mul(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;
	double			fvalue;
	BOOL			isintconst;
	struct	id_node		*optype;
	INT32			value;

	if(! (value_flag & NEED_VALUE))
		return (bin_op(expr,value_flag));

	optype = op_types(expr);	/* Find greatest type of operands */

	if(ISRIGHTCONSTANT(expr))
		{
		isintconst = FALSE;	/* Neither is constant */
		if(RIGHTTYPE(expr) & FP_NODE)
			{
			fvalue = RIGHTFCONSTANT(expr);
			value = f_to_l(fvalue,FALSE);

			if(((double) value) == fvalue)
				isintconst = TRUE;
			}
		else
			{
			/* Expression type is integer */
			isintconst = TRUE;
			value = RIGHTCONSTANT(expr);
			}
		if(isintconst && (value == 0))
			{
			depth = expr_code(optype,LEFT(expr),NO_VALUE);
			depth = depth_max(depth,ld_zero(optype));
			}
		else if(isintconst && (value == 1))
			depth = expr_code(optype,LEFT(expr),value_flag);
		else
			{
			if(optype->type & FP_NODE)
				{
				if(isintconst && (value == 2))
					{
					depth = expr_code(optype,LEFT(expr),
						value_flag);
					no_oper(TOP_FPUMULBY2);
					depth = depth_max(depth,INT_DEEP(1));
					}
				else
					{
					depth = expr_code(optype,LEFT(expr),
						value_flag);
					depth = depth_max(depth,
						fp_oplit(optype,RIGHT(expr)) +
							FLT_DEEP(2));
					fp_oper(optype,TOP_BAD,TOP_FPLDNLMULSN,
						TOP_FPLDNLMULDB);
					}
				}
			else
				{
				if(isintconst && (value == Sizeof_int))
					{
					depth = expr_code(optype,LEFT(expr),
						value_flag);
					no_oper(TOP_BCNT);
					}
				else
					{
/*
 *	An optimization is performed here to load the constant last, so the
 *	multiply is (typically) faster.  The Transputer multiply is faster when
 *	the smaller operand is in A, and the presumption is made that constants
 *	are typically of small magnitude.
 */
					depth = expr_code(optype,LEFT(expr),
						value_flag);
					loadconstant(value);
					no_oper(TOP_PROD);
					depth = depth_max(depth,INT_DEEP(2));
					}
				}
			}
		}
	else if(optype->type & FP_NODE)
		{
		if((RIGHTTYPE(expr) & FP_NODE) && (RIGHT(expr)->
			node_type->val.integral == optype->val.integral) &&
				is_simple(RIGHT(expr)))
			{
			depth = expr_code(optype,LEFT(expr),value_flag);

			depth = depth_max(depth,expr_code(optype,RIGHT(expr),
				NEED_VALUE | EVAL_ADDR) + FLT_DEEP(1));
			fp_oper(optype,TOP_BAD,TOP_FPLDNLMULSN,
				TOP_FPLDNLMULDB);
			depth = depth_max(depth,FLT_DEEP(2));
			}
		else if((LEFTTYPE(expr) & FP_NODE) && (LEFT(expr)->
			node_type->val.integral == optype->val.integral) &&
				is_simple(LEFT(expr)))
			{
			depth = expr_code(optype,RIGHT(expr),value_flag);

			depth = depth_max(depth,expr_code(optype,LEFT(expr),
				NEED_VALUE | EVAL_ADDR) + FLT_DEEP(1));
			fp_oper(optype,TOP_BAD,TOP_FPLDNLMULSN,
				TOP_FPLDNLMULDB);
			depth = depth_max(depth,FLT_DEEP(2));
			}
		else
			{
			depth = bin_op(expr,value_flag);
			no_oper(TOP_FPMUL);
			}
		}
	else
		{
		/* Non-constant integer expression */
		depth = bin_op(expr,value_flag);
		no_oper(TOP_PROD);
		}
	return (depth);
	}

/*
 *	Use double precision signed division to handle single precision
 *	unsigned.
 */
PUBLIC	unsigned	int
do_udiv(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	deep_lop;
	unsigned	int	deep_rop;
	struct	dvrt		*dvrt_lop;
	struct	dvrt		*dvrt_rop;
	unsigned	int	ideep_lop;
	unsigned	int	ideep_rop;
	struct	id_node		*optype;
	TEMP			temp;

	optype = op_types(expr);	/* Find greatest type of operands */

	dvrt_stack();
	deep_lop = expr_code(optype,LEFT(expr),value_flag);
	dvrt_lop = dvrt_unstack();

	dvrt_stack();
	deep_rop = expr_code(optype,RIGHT(expr),value_flag);
	dvrt_rop = dvrt_unstack();
/*
 *	Given the left and right depths, figure out how to set up the stack
 *	to minimize the use of "rev" instructions and "temp" variables.
 */
	ideep_lop = deep_lop & INT_STACK;
	ideep_rop = deep_rop & INT_STACK;

	if(ideep_rop == INT_DEEP(1))
		{
		if(ideep_lop < INT_DEEP(3))
			{
			loadconstant((INT32) 0);
			dvrt_append(dvrt_lop);
			dvrt_append(dvrt_rop);
			}
		else
			{
			dvrt_append(dvrt_lop);
			loadconstant((INT32) 0);
			rev(Inttype);
			dvrt_append(dvrt_rop);
			}
		}
	else if((ideep_rop == INT_DEEP(2)) && (ideep_lop == INT_DEEP(1)))
		{
		loadconstant((INT32) 0);
		dvrt_append(dvrt_rop);
		dvrt_append(dvrt_lop);
		rev(Inttype);
		}
	else if((ideep_rop == INT_DEEP(3)) && (ideep_lop == INT_DEEP(1)))
		{
		dvrt_append(dvrt_rop);
		loadconstant((INT32) 0);
		rev(Inttype);
		dvrt_append(dvrt_lop);
		rev(Inttype);
		}
/*
 *	Well, no way to avoid using a "temp".  Do the code so as to minimize
 *	the number of "rev" instructions needed.
 */
	else
		{
		/* For details about expr_code reevaluation, see bin_op */
		temp = get_temp();
		dvrt_free(dvrt_rop);	/* Flush the original */
		dvrt_free(dvrt_lop);	/* And the other side */
		(VOID) expr_code(optype,RIGHT(expr),value_flag);

		store_temp(Inttype,temp);
		load_temp(Inttype,temp);
		free_temp(temp);

		if(ideep_lop == INT_DEEP(2))
			{
			loadconstant((INT32) 0);
			(VOID) expr_code(optype,LEFT(expr),value_flag);
			}
		else
			{
			(VOID) expr_code(optype,LEFT(expr),value_flag);
			loadconstant((INT32) 0);
			rev(Inttype);
			}
		load_temp(Inttype,temp);
		free_temp(temp);
		}
	no_oper(TOP_LDIV);

	return (depth_max(depth_max(deep_lop,deep_rop),INT_DEEP(3)));
	}

/*
 *	Signed division.
 */
PUBLIC	unsigned	int
do_div(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;
	double			fvalue;
	BOOL			isintconst;
	struct	op_node		*op;
	struct	id_node		*optype;
	INT32			shift;
	INT32			value;

	if(! (value_flag & NEED_VALUE))
		return (bin_op(expr,value_flag));

	optype = op_types(expr);	/* Find greatest type of operands */

	if(ISRIGHTCONSTANT(expr) || ISLEFTCONSTANT(expr))
		{
		/* Constant expression */
		isintconst = FALSE;

		if(ISRIGHTCONSTANT(expr))
			{
			if((RIGHTTYPE(expr) & FP_NODE))
				{
				fvalue = RIGHTFCONSTANT(expr);

				value = f_to_l(fvalue,FALSE);

				if(((double) value) == fvalue)
					isintconst = TRUE;
				}
			else
				{
				/* Expression type is integer */
				isintconst = TRUE;

				value = RIGHTCONSTANT(expr);
				fvalue = l_to_f(RIGHT(expr));
				}

			if(isintconst && (value == 0))
				{
				error("Division by zero");
				depth = 0;
				}
			else if(isintconst && (value == 1))
				{
				/* Absorb it -- division by one is ignored */
				depth = expr_code(optype,LEFT(expr),
					value_flag);
				}
			else
				{
				if(optype->type & FP_NODE)
					{
					if(isintconst && (value == 2))
						{
						depth = expr_code(optype,
							LEFT(expr),value_flag);
						no_oper(TOP_FPUDIVBY2);
						depth = depth_max(depth,
							INT_DEEP(1));
						}
					else
						{
						depth = expr_code(optype,
							LEFT(expr),value_flag);
						op = make_fconst(1.0 / fvalue,
							optype);
						depth = depth_max(depth,
							fp_oplit(optype,op) +
								FLT_DEEP(2));
						fp_oper(optype,TOP_BAD,
							TOP_FPLDNLMULSN,
							TOP_FPLDNLMULDB);
						free_op_node(op);
						}
					}
				else
					{
					if((optype->type & UNSIGNED_NODE) &&
						isintconst &&
						((shift=bitlog2(value)) >= 0))
						{
/*
 *	Unsigned division by a power of two -- shift it.
 */
						depth = expr_code(optype,
							LEFT(expr),value_flag);

						loadconstant(shift);
						no_oper(TOP_SHR);
						depth = depth_max(depth,
							INT_DEEP(2));
						}
					else if(optype->type & UNSIGNED_NODE)
						{
						depth = do_udiv(expr,
							value_flag);
						}
					else		/* Signed division */
						{
						depth = bin_op(expr,
							value_flag);
						no_oper(TOP_DIV);
						}
					}
				}
			}
		else
			{
			/* Left side constant */
			if((LEFTTYPE(expr) & FP_NODE))
				{
				fvalue = LEFTFCONSTANT(expr);

				value = f_to_l(fvalue,FALSE);

				if(((double) value) == fvalue)
					isintconst = TRUE;
				}
			else
				{
				/* Expression type is integer */
				isintconst = TRUE;

				value = LEFTCONSTANT(expr);
				fvalue = l_to_f(LEFT(expr));
				}

			if(isintconst && (value == 0))
				{
				/* Ignore value and return a zero */
				(VOID) expr_code(optype,RIGHT(expr),NO_VALUE);
				depth = ld_zero(optype);
				}
			else
				{
				if(optype->type & FP_NODE)
					{
					depth = bin_op(expr,value_flag);
					no_oper(TOP_FPDIV);
					}
				else
					{
					if(optype->type & UNSIGNED_NODE)
						{
						depth = do_udiv(expr,
							value_flag);
						}
					else		/* Signed division */
						{
						depth = bin_op(expr,
							value_flag);
						no_oper(TOP_DIV);
						}
					}
				}
			}
		}
	else if(optype->type & FP_NODE)
		{
		depth = bin_op(expr,value_flag);
		no_oper(TOP_FPDIV);
		}
	else
		{
		/* Non constant integer expression */
		if(optype->type & UNSIGNED_NODE)
			depth = do_udiv(expr,value_flag);
		else			/* Signed division */
			{
			depth = bin_op(expr,value_flag);
			no_oper(TOP_DIV);
			}
		}
	return (depth);
	}

/*
 *	Convert value to a bit number, iff one bit is set.
 */
PUBLIC	INT32
bitlog2(value)
	INT32			value;
	{
	int 			bit;
	UINT32			val;

	for(bit = 0, val = value; (val!=0)&&((val&1) == 0); val /= 2, ++bit)
		;			/* Nothing */

	if((val & ~1) != 0)
		return ((INT32) -1);
	else
		return ((INT32) bit);
	}

/*
 *	"%".
 */
PUBLIC	unsigned	int
do_mod(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned int		depth;
	struct id_node		*optype;
	INT32			value;

	if(! (value_flag & NEED_VALUE))
		return (bin_op(expr,value_flag));

	optype = op_types(expr);	/* Find greatest type of operands */

	if(ISRIGHTCONSTANT(expr))
		{
		value = RIGHTCONSTANT(expr);
		if(value == 0)
			{
			error("Division by zero");
			loadconstant((INT32) 0);
			return (INT_DEEP(1));
			}
		else if(value == 1)
			{
			depth = expr_code(optype,LEFT(expr),NO_VALUE);
			loadconstant((INT32) 0);
			depth = depth_max(depth,INT_DEEP(1));
			return (depth);
			}
		}
	else if(ISLEFTCONSTANT(expr))
		{
		value = LEFTCONSTANT(expr);
		if(value == 0)
			{
			depth = expr_code(optype,RIGHT(expr),NO_VALUE);
			loadconstant((INT32) 0);
			depth = depth_max(depth,INT_DEEP(1));
			return (depth);
			}
		}
/*
 *	We have to actually do a remainder.  The question is, is it unsigned?
 */
	if(optype->type & UNSIGNED_NODE)	/* Unsigned remainder? */
		{
		depth = do_udiv(expr,value_flag);
		rev(Inttype);		/* Put remainder on top of stack */
		}
	else				/* Signed remainder */
		{
		depth = bin_op(expr,value_flag);
		no_oper(TOP_REM);
		}
	return (depth);
	}

/*
 *	"+".
 */
PUBLIC	unsigned	int
do_add(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;
	double			fvalue;
	struct	op_node		*op;
	struct	id_node		*optype;
	INT32			value;

	if(! (value_flag & NEED_VALUE))
		return (bin_op(expr,value_flag));

	optype = op_types(expr);	/* Find greatest type of operands */

	if(ISRIGHTCONSTANT(expr))
		{
		depth = expr_code(optype,LEFT(expr),value_flag);

		if(optype->type & FP_NODE)
			{
			if(RIGHTTYPE(expr) & FP_NODE)
				fvalue = RIGHTFCONSTANT(expr);
			else
				fvalue = l_to_f(RIGHT(expr));

			if(fvalue != 0.0)
				{
				op = make_fconst(fvalue,optype);
				depth = depth_max(depth,fp_oplit(optype,op) +
					FLT_DEEP(2));
				fp_oper(optype,TOP_BAD,TOP_FPLDNLADDSN,
					TOP_FPLDNLADDDB);
				free_op_node(op);
				}
			}
		else
			{
			if(RIGHTTYPE(expr) & FP_NODE)
				value = f_to_l(RIGHTFCONSTANT(expr),TRUE);
			else
				value = RIGHTCONSTANT(expr);

			if(value != 0)
				abs_oper(TOP_ADC,(INT32) value);
			}
		}
	else if((RIGHTTYPE(expr) & FP_NODE) && (RIGHT(expr)->node_type->
		val.integral == optype->val.integral) &&
			is_simple(RIGHT(expr)))
		{
		depth = expr_code(optype,LEFT(expr),value_flag);

		depth = depth_max(depth,expr_code(optype,RIGHT(expr),
			NEED_VALUE | EVAL_ADDR) + FLT_DEEP(1));
		fp_oper(optype,TOP_BAD,TOP_FPLDNLADDSN,TOP_FPLDNLADDDB);
		depth = depth_max(depth,FLT_DEEP(2));
		}
	else if((LEFTTYPE(expr) & FP_NODE) && (LEFT(expr)->node_type->
		val.integral == optype->val.integral) &&
			is_simple(LEFT(expr)))
		{
		depth = expr_code(optype,RIGHT(expr),value_flag);

		depth = depth_max(depth,expr_code(optype,LEFT(expr),
			NEED_VALUE | EVAL_ADDR) + FLT_DEEP(1));
		fp_oper(optype,TOP_BAD,TOP_FPLDNLADDSN,TOP_FPLDNLADDDB);
		depth = depth_max(depth,FLT_DEEP(2));
		}
	else
		{
		depth = bin_op(expr,value_flag);
		fp_oper(optype,TOP_ADD,TOP_FPADD,TOP_FPADD);
		}
	return (depth);
	}

/*
 *	"-".
 */
PUBLIC	unsigned	int
do_sub(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	struct	dvrt		*code;
	unsigned	int	depth;
	double			fvalue;
	struct	op_node		*op;
	struct	id_node		*optype;
	INT32			value;

	if(! (value_flag & NEED_VALUE))
		return (bin_op(expr,value_flag));

	optype = op_types(expr);	/* Find greatest type of operands */

	if(ISLEFTCONSTANT(expr))
		{
		dvrt_stack();
		depth = expr_code(optype,RIGHT(expr),value_flag);
		code = dvrt_unstack();

		if(optype->type & FP_NODE)
			{
			if(LEFTTYPE(expr) & FP_NODE)
				fvalue = LEFTFCONSTANT(expr);
			else
				fvalue = l_to_f(LEFT(expr));

			op = make_fconst(fvalue,optype);

			if((depth & FLT_STACK) == FLT_DEEP(3))
				{
				dvrt_append(code);
				depth = depth_max(depth,fp_val(optype,op) +
					FLT_DEEP(1));
				no_oper(TOP_FPREV);
				}
			else
				{
				depth = depth_max(depth + FLT_DEEP(1),
					fp_val(optype,op));
				dvrt_append(code);
				}
			no_oper(TOP_FPSUB);
			free_op_node(op);
			}
		else
			{
			/* Left constant is an integer */
			value = LEFTCONSTANT(expr);

			if(value == 0)
				{
/*
 *	0 - expr -> just negate to avoid extra stack usage.
 */
				dvrt_append(code);
				no_oper(TOP_NOT);
				abs_oper(TOP_ADC,(INT32) 1);
				}
			else
				{
				if((depth & INT_STACK) < INT_DEEP(3))
					{
					loadconstant(value);
					dvrt_append(code);
					depth = depth_max(depth,
						depth + INT_DEEP(1));
					}
				else
					{
					dvrt_append(code);
					loadconstant(value);
					rev(Inttype);
					depth = depth_max(depth,INT_DEEP(2));
					}
				no_oper(TOP_SUB);
				}
			}
		}
	else if(ISRIGHTCONSTANT(expr))
		{
		depth = expr_code(optype,LEFT(expr),value_flag);

		if(optype->type & FP_NODE)
			{
			if(RIGHTTYPE(expr) & FP_NODE)
				fvalue = RIGHTFCONSTANT(expr);
			else
				fvalue = l_to_f(RIGHT(expr));

			if(fvalue != 0.0)
				{
				op = make_fconst(-fvalue,optype);
/*
 *	Generate load and add of negative constant.
 */
				depth = depth_max(depth,fp_oplit(optype,op) +
					FLT_DEEP(2));
				fp_oper(optype,TOP_BAD,TOP_FPLDNLADDSN,
					TOP_FPLDNLADDDB);
				free_op_node(op);
				}
			}
		else
			{
			if(RIGHTTYPE(expr) & FP_NODE)
				value = f_to_l(RIGHTFCONSTANT(expr),TRUE);
			else
				value = RIGHTCONSTANT(expr);

			if(value != 0)
				abs_oper(TOP_ADC,(INT32) -value);
			}
		}
	else
		{
		depth = bin_op(expr,value_flag);
		fp_oper(optype,TOP_SUB,TOP_FPSUB,TOP_FPSUB);
		}
	return (depth);
	}

/*
 *	"<<".
 */
PUBLIC	unsigned	int
do_lshift(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;

	if(! (value_flag & NEED_VALUE))
		return (bin_op(expr,value_flag));

	if(ISLEFTCONSTANT(expr) && (LEFTCONSTANT(expr) == 0))
		{
		depth = expr_code(expr->node_type,RIGHT(expr),NO_VALUE);

		loadconstant((INT32) 0);
		depth = depth_max(depth,INT_DEEP(1));
		}
	else if(ISRIGHTCONSTANT(expr) && (RIGHTCONSTANT(expr) == 0))
		depth = expr_code(expr->node_type,LEFT(expr),value_flag);
	else
		{
		depth = bin_op(expr,value_flag);

		no_oper(TOP_SHL);
		}

	return (depth);
	}

/*
 *	">>".
 */
PUBLIC	unsigned	int
do_rshift(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	struct	dvrt		*code;
	unsigned	int	depth;
	TEMP			temp;
	INT32			value;

	if(! (value_flag & NEED_VALUE))
		return (bin_op(expr,value_flag));

	if(ISLEFTCONSTANT(expr))
		{
		value = LEFTCONSTANT(expr);
		if(value == 0)
			{
			depth = expr_code(expr->node_type,RIGHT(expr),
				NO_VALUE);

			loadconstant((INT32) 0);
			depth = depth_max(depth,INT_DEEP(1));
			}
		else
			{
			dvrt_stack();
			depth = expr_code(expr->node_type,RIGHT(expr),
				value_flag);
			code = dvrt_unstack();
/*
 *	Only do signed right shifts if parameter has been set.
 */
			if(Cpu_srshift && (LEFTTYPE(expr) & SIGNED_NODE))
				{
				switch((int) (depth & INT_STACK))
					{
					case	INT_DEEP(1):
						const_sign(LEFT(expr));
						loadconstant(value);
						dvrt_append(code);
						break;

					case	INT_DEEP(2):
						const_sign(LEFT(expr));
						dvrt_append(code);
						loadconstant(value);
						rev(Inttype);
						break;

					case	INT_DEEP(3):
						dvrt_append(code);
						const_sign(LEFT(expr));
						rev(Inttype);
						loadconstant(value);
						rev(Inttype);
						break;

					default:
				fatal("Compiler error: do_rshift");
					}
				no_oper(TOP_LSHR);
				depth = depth_max(depth,INT_DEEP(3));
				}
			else
				{
				/* Unsigned shift -- just do the simple case */
				if(depth < INT_DEEP(3))
					{
					loadconstant(value);
					dvrt_append(code);
					depth = depth_max(depth,depth+
						INT_DEEP(1));
					}
				else
					{
					dvrt_append(code);
					loadconstant(value);
					rev(Inttype);
					depth = depth_max(depth,INT_DEEP(2));
					}
				no_oper(TOP_SHR);
				}
			}
		}
	else if(ISRIGHTCONSTANT(expr))
		{
		value = RIGHTCONSTANT(expr);
		if(value == 0)
			{
			depth = expr_code(expr->node_type,LEFT(expr),
				value_flag);
			}
		else if((value > (8 * Sizeof_int)) && ((! Cpu_srshift) ||
			(LEFTTYPE(expr) & UNSIGNED_NODE)))
			{
			/* Evaluate for side-effects only */
			depth = expr_code(expr->node_type,LEFT(expr),NO_VALUE);

			loadconstant((INT32) 0);
			depth = depth_max(depth,INT_DEEP(1));
			}
		else if(Cpu_srshift && (LEFTTYPE(expr) & SIGNED_NODE))
			{
			depth = expr_code(expr->node_type,LEFT(expr),
				value_flag);
			no_oper(TOP_XDBLE);
			loadconstant(value);
			no_oper(TOP_LSHR);
			depth = depth_max(depth,INT_DEEP(3));
			}
		else
			{
			/* Unsigned shift right */

			depth = expr_code(expr->node_type,LEFT(expr),
				value_flag);
			loadconstant(value);
			no_oper(TOP_SHR);
			depth = depth_max(depth,INT_DEEP(2));
			}
		}   
	else
		{
/*
 *	Only do signed right shifts if parameter has been set.
 */
		if(Cpu_srshift && (LEFTTYPE(expr) & SIGNED_NODE))
			{
			dvrt_stack();
			depth = expr_code(expr->node_type,RIGHT(expr),
				value_flag);
			code = dvrt_unstack();

			if((depth & INT_STACK) > INT_DEEP(1))
				{
				temp = get_temp();
				dvrt_free(code);
				depth = expr_code(expr->node_type,RIGHT(expr),
					value_flag);
				store_temp(Inttype,temp);
				depth = depth_max(depth,expr_code(expr->
					node_type,LEFT(expr),value_flag));
				no_oper(TOP_XDBLE);
				load_temp(Inttype,temp);
				}
			else
				{
				depth = depth_max(depth,expr_code(expr->
					node_type,LEFT(expr),value_flag));
				no_oper(TOP_XDBLE);
				dvrt_append(code);
				}
			no_oper(TOP_LSHR);
			depth = depth_max(depth,INT_DEEP(3));
			}
		else
			{
			depth = bin_op(expr,value_flag);

			no_oper(TOP_SHR);
			}
		}
	return (depth);
	}

/*
 *	Load a sign for a constant of the type specified by op_node.
 */
PRIVATE	VOID
const_sign(expr)
	struct	op_node		*expr;
	{
	if((expr->right.ptr.terminal->type & SIGNED_NODE) &&
		(expr->right.ptr.terminal->val.integral < 0))
		{
		loadconstant((INT32) -1);
		}
	else
		loadconstant((INT32) 0);
	}

/* End of tcx6b.c */
