/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: tcx6d.c,v 2.2 92/04/21 11:11:17 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	TCX Compiler (6d of 11)	Expression Syntax Tree Walker
 *
 *	Copyright 1987-1991 by Logical Systems.  All rights reserved.
 *
 *	Published by:	Logical Systems
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	03/17/87
 */

#include	<stdio.h>
#include	<ctype.h>
#include	<string.h>
#include	"tcx.h"
#include	"tcxdecl.h"
#include	"tcxext.h"

/*
 *	Local function declarations.
 */
PRIVATE	unsigned	int	aband_cr1();
PRIVATE	unsigned	int	aband_cr2();
PRIVATE	unsigned	int	aband_g();
PRIVATE	unsigned	int	abxor_cr1();
PRIVATE	unsigned	int	abxor_cr2();
PRIVATE	unsigned	int	abxor_g();
PRIVATE	unsigned	int	abor_cr1();
PRIVATE	unsigned	int	abor_cr2();
PRIVATE	unsigned	int	abor_g();
PRIVATE	unsigned	int	log_help();
PRIVATE	VOID			do_call1();
PRIVATE	VOID			do_call2();
PRIVATE	unsigned	int	eval_1arg();
PRIVATE	unsigned	int	eval_2arg();
PRIVATE	unsigned	int	eval_3arg();

/*
 *	"&".
 */
PUBLIC	unsigned	int
do_band(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;
	struct	id_node		*optype;
	INT32			value;

	if(! (value_flag & NEED_VALUE))
		return (bin_op(expr,value_flag));

	optype = op_types(expr);	/* Find greatest type of operands */

	if(ISRIGHTCONSTANT(expr))
		{
		value = RIGHTCONSTANT(expr);
		if(value == 0)
			{
			depth = expr_code(optype,LEFT(expr),NO_VALUE);
			loadconstant((INT32) 0);
			depth = depth_max(depth,INT_DEEP(1));
			}
		else if(((Sizeof_int == WORD_SIZE_16B) ?
				((value & 0xFFFF) == 0xFFFF) :
					(value == 0xFFFFFFFFL)) ||
			((value == 0x000000FFL) && 
				(LEFTTYPE(expr) & CHAR_NODE) &&
					(LEFTTYPE(expr) & UNSIGNED_NODE)))
			{
			depth = expr_code(optype,LEFT(expr),value_flag);
			}
		else
			{
			depth = bin_op(expr,value_flag);
			no_oper(TOP_AND);
			}
		}
	else
		{
		depth = bin_op(expr,value_flag);
		no_oper(TOP_AND);
		}
	return (depth);
	}

/*
 *	"^".
 */
PUBLIC	unsigned	int
do_bxor(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;
	struct	id_node		*optype;
	INT32			value;

	if(! (value_flag & NEED_VALUE))
		return (bin_op(expr,value_flag));

	optype = op_types(expr);	/* Find greatest type of operands */

	if(ISRIGHTCONSTANT(expr))
		{
		value = RIGHTCONSTANT(expr);
		if(value == 0)
			depth = expr_code(optype,LEFT(expr),value_flag);
		else if((Sizeof_int == WORD_SIZE_16B) ?
			((value & 0xFFFF) == 0xFFFF) :
				(value == 0xFFFFFFFFL))
			{
			depth = expr_code(optype,LEFT(expr),value_flag);
			no_oper(TOP_NOT);
			}
		else
			{
			depth = bin_op(expr,value_flag);
			no_oper(TOP_XOR);
			}
		}
	else
		{
		depth = bin_op(expr,value_flag);
		no_oper(TOP_XOR);
		}
	return (depth);
	}

/*
 *	"|".
 */
PUBLIC	unsigned	int
do_bor(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;
	struct	id_node		*optype;
	INT32			value;

	if(! (value_flag & NEED_VALUE))
		return (bin_op(expr,value_flag));

	optype = op_types(expr);	/* Find greatest type of operands */

	if(ISRIGHTCONSTANT(expr))
		{
		value = RIGHTCONSTANT(expr);
		if(value == 0)
			depth = expr_code(optype,LEFT(expr),value_flag);
		else if((Sizeof_int == WORD_SIZE_16B) ?
				((value & 0xFFFF) == 0xFFFF) :
					(value == 0xFFFFFFFFL))
			{
			depth = expr_code(optype,LEFT(expr),NO_VALUE);
			loadconstant((INT32) -1);
			depth = depth_max(depth,INT_DEEP(1));
			}
		else
			{
			depth = bin_op(expr,value_flag);
			no_oper(TOP_OR);
			}
		}
	else
		{
		depth = bin_op(expr,value_flag);
		no_oper(TOP_OR);
		}
	return (depth);
	}

/*
 *	"&=".
 */
PUBLIC	unsigned	int
do_aband(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	return (do_aop(expr,value_flag,do_band,aband_cr1,aband_cr2,NULL,
		aband_g));
	}

PRIVATE	unsigned	int
aband_cr1(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned int		depth;

	if(RIGHTCONSTANT(expr) == 0)
/*
 *	"a &= 0" is equivalent to "a = 0"
 */
		{
		depth = do_assign(expr,value_flag);
		}
	else if(((Sizeof_int == WORD_SIZE_16B) ?
		((RIGHTCONSTANT(expr) & 0xFFFF) == 0xFFFF) :
			(RIGHTCONSTANT(expr) == 0xFFFFFFFFL)) ||
		((LEFTTYPE(expr) & CHAR_NODE) &&
			(LEFTTYPE(expr) &  UNSIGNED_NODE) &&
				(RIGHTCONSTANT(expr) == 0x000000FFL)))
		{
		depth = expr_code(expr->node_type,LEFT(expr),value_flag);
		}
	else
		depth = UNDEF_DEEP;

	return (depth);
	}

PRIVATE	unsigned	int
aband_cr2(expr)
	struct	op_node		*expr;
	{
	loadconstant(RIGHTCONSTANT(expr));
	no_oper(TOP_AND);

	return(INT_DEEP(2));
	}

PRIVATE	unsigned	int
aband_g(expr,temp)
	struct	op_node		*expr;
	TEMP			temp;
	{
	unsigned	int	depth;

	depth = assign_op(expr,temp);

	no_oper(TOP_AND);

	return (depth);
	}

/*
 *	"^=".
 */
PUBLIC	unsigned	int
do_abxor(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	return (do_aop(expr,value_flag,do_bxor,abxor_cr1,abxor_cr2,NULL,
		abxor_g));
	}

PRIVATE	unsigned	int
abxor_cr1(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;

	if(RIGHTCONSTANT(expr) == 0)
		depth = expr_code(expr->node_type,LEFT(expr),value_flag);
	else
		depth = UNDEF_DEEP;

	return (depth);
	}

PRIVATE	unsigned	int
abxor_cr2(expr)
	struct	op_node		*expr;
	{
	unsigned	int	depth;

	if((Sizeof_int == WORD_SIZE_16B) ?
		((RIGHTCONSTANT(expr) & 0xFFFF) == 0xFFFF) :
			(RIGHTCONSTANT(expr) == 0xFFFFFFFFL))
		{
		no_oper(TOP_NOT);
		depth = INT_DEEP(1);
		}
	else
		{
		loadconstant(RIGHTCONSTANT(expr));
		no_oper(TOP_XOR);
		depth = INT_DEEP(2);
		}
	return (depth);
	}

PRIVATE	unsigned	int
abxor_g(expr,temp)
	struct	op_node		*expr;
	TEMP			temp;
	{
	unsigned	int	depth;

	depth = assign_op(expr,temp);

	no_oper(TOP_XOR);

	return (depth);
	}

/*
 *	"|=".
 */
PUBLIC	unsigned	int
do_abor(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	return (do_aop(expr,value_flag,do_bor,abor_cr1,abor_cr2,NULL,abor_g));
	}

PRIVATE	unsigned	int
abor_cr1(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;

	if(RIGHTCONSTANT(expr) == 0)
		depth = expr_code(expr->node_type,LEFT(expr),value_flag);
	else if((Sizeof_int == WORD_SIZE_16B) ?
		((RIGHTCONSTANT(expr) & 0xFFFF) == 0xFFFF) :
			(RIGHTCONSTANT(expr) == 0xFFFFFFFFL))
/*
 *	"a |= 0xFFFFFFFF" is equivalent to "a = 0xFFFFFFFF"
 */
		{
		depth = do_assign(expr,value_flag);
		}
	else
		depth = UNDEF_DEEP;

	return (depth);
	}

PRIVATE unsigned	int
abor_cr2(expr)
	struct	op_node		*expr;
	{
	loadconstant(RIGHTCONSTANT(expr));
	no_oper(TOP_OR);

	return(INT_DEEP(2));
	}

PRIVATE	unsigned	int
abor_g(expr,temp)
	struct	op_node		*expr;
	TEMP			temp;
	{
	unsigned	int	depth;

	depth = assign_op(expr,temp);

	no_oper(TOP_OR);

	return (depth);
	}

/*
 *	A unary "operator" which converts its operand into a canonical
 *	boolean value (either TRUE or FALSE).
 */
PUBLIC	unsigned	int
do_bool(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;
	
	depth = expr_code(NULL,RIGHT(expr),value_flag | RELAX_BOOL);

	if(value_flag & NEED_VALUE)
		{
		if(RIGHTTYPE(expr) & FP_NODE)
			{
			(VOID) ld_zero(RIGHT(expr)->node_type);
			no_oper(TOP_FPEQ);
			eqc();
			depth = depth_max(depth,INT_DEEP(1) | FLT_DEEP(2));
			}
		else
			{
			eqc();
			eqc();
			depth = depth_max(depth,INT_DEEP(1));
			}
		}
	return (depth);
	}

/*
 *	"&&".
 *
 *	Since all constant operands are converted to canonical True/False
 *	values during parsing we need only check against them.
 */
PUBLIC	unsigned	int
do_land(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;
	LABEL			endlab;

	endlab = getlabel();

	depth = log_help(OP_LAND,LEFT(expr),endlab);

	if(ISRIGHTCONSTANT(expr))
		{
		if(value_flag & NEED_VALUE)
			{
			if(RIGHTCONSTANT(expr) == TRUE)
				loadconstant((INT32) TRUE);
			else
				loadconstant((INT32) FALSE);
			depth = depth_max(depth,INT_DEEP(1));
			}
		}
	else
		{
		depth = depth_max(depth,expr_code(NULL,RIGHT(expr),
			value_flag | RELAX_BOOL));

		if(value_flag & NEED_VALUE)
			{
			if(RIGHTTYPE(expr) & FP_NODE)
				{
				(VOID) ld_zero(RIGHT(expr)->node_type);
				no_oper(TOP_FPEQ);
				eqc();
				depth = depth_max(depth,
					INT_DEEP(1) | FLT_DEEP(2));
				}
			else if(! (value_flag & RELAX_BOOL))
				{
				eqc();
				eqc();
				}
			}
		}

	printlabel(endlab);

	return (depth);
	}

/*
 *	"||".
 *
 *	Since all constant operands are converted to canonical True/False
 *	values during parsing we need only check against them.
 */
PUBLIC	unsigned	int
do_lor(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;
	LABEL			endlab;

	endlab = getlabel();

	depth = log_help(OP_LOR,LEFT(expr),endlab);

	if(ISRIGHTCONSTANT(expr))
		{
		if(value_flag & NEED_VALUE)
			{
			if(RIGHTCONSTANT(expr) == TRUE)
				loadconstant((INT32) FALSE);
			else
				loadconstant((INT32) TRUE);
			depth = depth_max(depth,INT_DEEP(1));
			}
		}
	else
		{
		depth = depth_max(depth,expr_code(NULL,RIGHT(expr),
			value_flag | RELAX_BOOL));

		if(value_flag & NEED_VALUE)
			{
			if(RIGHTTYPE(expr) & FP_NODE)
				{
				(VOID) ld_zero(RIGHT(expr)->node_type);
				no_oper(TOP_FPEQ);
				depth = depth_max(depth,
					INT_DEEP(1) | FLT_DEEP(2));
				}
			else
				eqc();
			}
		}

	printlabel(endlab);

	if(value_flag & NEED_VALUE)
		eqc();

	return (depth);
	}

/*
 *	Helping function for comparison operators.
 */
PRIVATE	unsigned	int
log_help(op,expr,label)
	int			op;
	struct	op_node		*expr;
	LABEL			label;
	{
	char			buf[MAXLABEL + 1];
	unsigned	int	depth;

	if(expr->operator == op)
		{
		depth = log_help(op,LEFT(expr),label);
		expr = RIGHT(expr);
		}

	depth = depth_max(depth,expr_code(NULL,expr,NEED_VALUE | RELAX_BOOL));

	if(EXPRTYPE(expr) & FP_NODE)
		{
		/* Convert floating value to integer value */
		(VOID) ld_zero(expr->node_type);
		no_oper(TOP_FPEQ);
		if(op == OP_LAND)
			eqc();
		depth = depth_max(depth,INT_DEEP(1) | FLT_DEEP(2));
		}
	else if(op == OP_LOR)
		eqc();

	instruction(TOP_CJ,REL,strlabel(label,buf),NULL,(INT32) 0,NULL);

	return (depth);
	}

/*
 *	"?:".
 */
PUBLIC	unsigned	int
do_tern(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	char			buf[MAXLABEL + 1];
	unsigned	int	depthf;
	unsigned	int	deptht;
	LABEL			elabel;
	LABEL			flabel;

	flabel = getlabel();		/* False jump label */
	elabel = getlabel();		/* End jump label */

	deptht = expr_code(NULL,LEFT(expr),NEED_VALUE | RELAX_BOOL);

	if(LEFTTYPE(expr) & FP_NODE)
		{
		(VOID) ld_zero(LEFT(expr)->node_type);
		no_oper(TOP_FPEQ);
		eqc();
		deptht = depth_max(deptht,INT_DEEP(1) | FLT_DEEP(2));
		}
	instruction(TOP_CJ,REL,strlabel(flabel,buf),NULL,(INT32) 0,NULL);

	deptht = depth_max(deptht,
		expr_code(expr->node_type,LEFT(RIGHT(expr)),value_flag));

	loadconstant((INT32) 0);

	instruction(TOP_CJ,REL,strlabel(elabel,buf),NULL,(INT32) 0,NULL);
	deptht = depth_max(deptht,INT_DEEP(2));

	printlabel(flabel);
/*
 *	At this point there is one ZERO on the stack in A.
 */
	depthf = expr_code(expr->node_type,RIGHT(RIGHT(expr)),value_flag);

	if((depthf & INT_STACK) >= INT_DEEP(3))
		{
/*
 *	Zero from CJ was lost in stack overflow.
 */
		loadconstant((INT32) 0);
		}

    printlabel(elabel);
/*
 *	Delete the ZERO from both branches.  Note that for the FALSE branch,
 *	the ZERO might be the TOP of the stack.  But who cares.
 */
	no_oper(TOP_ADD);
	return (depth_max(deptht,depthf + INT_DEEP(1)));
	}

/*
 *	",".
 */
PUBLIC	unsigned	int
do_seq(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned	int	depth;
/*
 *	Evaluate all left-hand nodes and forget value.
 */
	depth = expr_code(expr->node_type,LEFT(expr),NO_VALUE);
/*
 *	Evaluate the right-hand node and keep value if top-most.
 */
	return (depth_max(depth,expr_code(expr->node_type,RIGHT(expr),
		value_flag)));
	}

/*
 *	Function calling.
 */
PUBLIC	unsigned int
do_call(expr)
	struct	op_node		*expr;
	{
	struct	arg_info	args[3];
	INT32			argsize;
	struct	op_node		*current;
	register int		numargs;
/*
 *	Initialize info table for first three parameters.
 */
	for(numargs = 0; numargs < 3; numargs++)
		{
		args[numargs].code = NULL;
		args[numargs].depth = 0;
		args[numargs].expr = NULL;
		args[numargs].node = NULL;
		args[numargs].size = 0;
		}
/*
 *	Divert code for up to the first three parameters.  Keep track of the
 *	first three expressions, the code each generates, the depth of the
 *	various stacks which evaluating the parameter trashes and the size
 *	of each parameter.
 */
	for(current = LEFT(expr), numargs = 0;
		((current != NULL) && (numargs < 3));
			current = LEFT(current),numargs++)
		{
		/* Save node handle */
		args[numargs].node = current;

		/* Save expression handle */
		args[numargs].expr = RIGHT(current);

		/* Compute & save parameter size */
		if(RIGHTTYPE(current) & (FP_NODE | SU_NODE))
			{
/*
 *	Test for a parameter which is function call returning struct/union.
 *	Arbitrarily indicate that it is > 3 words in size so that the
 *	"do_call2" function call code generation scheme is used if the
 *	returned struct is one of the first three parameters.  This is done
 *	since the parameter function call can be optimized to copy the
 *	struct being returned into the "final" spot it will occupy for this
 *	function call without a second copy operation.
 */
			if((RIGHTTYPE(current) & SU_NODE) &&
				(RIGHT(current)->operator == OP_CALL))
				{
				args[numargs].size = 3 + 1;
				}
			else
				{
				args[numargs].size =
					word((RIGHT(current)->node_type->
						val.integral +
					(Sizeof_int - 1)));
				}
			}
		else
			args[numargs].size = 1;
#ifdef	DEBUGGING
	printf("Exprtype=%lX, Size = %ld\n",EXPRTYPE(args[numargs].expr),
		args[numargs].size);
#endif	/* DEBUGGING */
		/* Compute and save parameter evaluation code */
		if(args[numargs].size <= 3)
			{
			/* Save expression code */
			dvrt_stack();
			if(expr->operator == OP_ICALL)
				{
				args[numargs].depth = expr_code(NULL,
					RIGHT(current),NEED_VALUE);
				}
			else
				{
				switch((int) args[numargs].size)
				{
				case	1:
					args[numargs].depth =
						eval_1arg(RIGHT(current));
					break;
				case	2:
					args[numargs].depth =
						eval_2arg(RIGHT(current));
					break;
				case	3:
					args[numargs].depth =
						eval_3arg(RIGHT(current));
					break;
				default:
					fatal("Compiler error: do_call");
				}
				}
			args[numargs].code = dvrt_unstack();
			}
		else
			args[numargs].code = NULL;
		}
/*
 *	If a intrinsic function is being called call the dispatch routine.
 */
	if(expr->operator == OP_ICALL)
		return (do_icall(expr,numargs,args));
/*
 *	Based on the size of the arguments and the amount of the integer
 *	stack each requires for evaluation, determine the optimum code
 *	sequence for this function call.
 */
	argsize = 0;			/* Additional param stack frame size */
	if((numargs == 0) ||
		((numargs == 1) && (args[0].size == 1)) ||
		((numargs == 2) && (args[0].size == 1) &&
			(args[1].size == 1)) ||
		((numargs == 3) && (args[0].size == 1) &&
			(args[1].size == 1) && (args[2].size == 1)))
		{
		if(current != NULL)	/* Handle additional parameters */
			argsize = do_args(current,(INT32) 0,FALSE);
		(VOID) do_call3(numargs,args,TRUE);
		do_call1(expr);
		}
	else if((numargs >= 2) &&
		((args[0].size == 1) &&
			(args[1].size == 2) &&
			(args[0].depth <= INT_DEEP(1))) ||
		((args[0].size == 2) &&
			(args[1].size == 1) &&
			(args[0].depth <= INT_DEEP(2))))
		{
		if(numargs > 2)		/* Handle additional parameters */
			{
			dvrt_free(args[2].code);
			argsize = do_args(args[2].node,(INT32) 0,FALSE);
			}
		dvrt_append(args[1].code);
		dvrt_append(args[0].code);
		do_call1(expr);
		}
	else if((args[0].size == 3) || ((numargs == 1) && (args[0].size == 2)))
		{
		if(numargs > 1)		/* Handle additional parameters */
			{
			dvrt_free(args[1].code);
			dvrt_free(args[2].code);
			argsize = do_args(args[1].node,(INT32) 0,FALSE);
			}
		dvrt_append(args[0].code);
		do_call1(expr);
		}
	else
		{
		dvrt_free(args[0].code);
		dvrt_free(args[1].code);
		dvrt_free(args[2].code);
		/* Note ajw base bumped up by 1 to allow for return address */
		argsize = do_args(args[0].node,(INT32) 1,TRUE);
		do_call2(expr);
/*
 *	Special cleanup code required since workspace values are now out of
 *	alignment with the "?ws" and "?tb" values.
 */
		assign(WORKSPACE,WORKSPACE,FALSE,(INT32) argsize);
		assign(TEMPBASE,TEMPBASE,FALSE,(INT32) argsize);
		assign(STRUCTBASE,STRUCTBASE,FALSE,(INT32) argsize);
		if((argsize -= 4) > 0)
			abs_oper(TOP_AJW,(INT32) argsize);
		return (INT_DEEP(3) | FLT_DEEP(3));
		}
/*
 *	Re-adjust workspace registers after the call if required.
 */
	if(argsize > 0)
		{
		assign(WORKSPACE,WORKSPACE,FALSE,(INT32) argsize);
		assign(TEMPBASE,TEMPBASE,FALSE,(INT32) argsize);
		assign(STRUCTBASE,STRUCTBASE,FALSE,(INT32) argsize);
		abs_oper(TOP_AJW,(INT32) argsize);
		}
/*
 *	Return maximum stack usage since call trashes stack.
 */
	return (INT_DEEP(3) | FLT_DEEP(3));
	}

/*
 *	Generate transfer-of-control code assuming parameters are ready.  This
 *	particular routine uses the "CALL" instruction to generate the transfer
 *	for cases were no objects cross the boundary between the first three
 *	words and the rest.
 */
PRIVATE	VOID
do_call1(expr)
	struct	op_node		*expr;
	{
	char			buf1[MAXLABEL + 1];
	char			buf2[MAXLABEL + 1];
	LABEL			label1;
	LABEL			label2;

	if(! ISTERMINAL(RIGHT(expr)))	/* Indirect call */
		{
		instruction(TOP_CALL,REL,strlabel(label1 = getlabel(),buf1),
			NULL,(INT32) 0,NULL);
/*
 *	Adjust offset variables to account for three new cells plus the return
 *	address.
 */
		assign(WORKSPACE,WORKSPACE,TRUE,(INT32) 4);
		assign(TEMPBASE,TEMPBASE,TRUE,(INT32) 4);
		assign(STRUCTBASE,STRUCTBASE,TRUE,(INT32) 4);

		printlabel(label1);

		instruction(TOP_ADC,ABS,strlabel(label2 = getlabel(),buf2),
			buf1,(INT32) 0,NULL);
/*
 *	Change return address generated by the above CALL.
 */
		abs_oper(TOP_STL,(INT32) 0);
/*
 *	Evaluate the address of the function to call.
 */
		(VOID) expr_code(NULL,RIGHT(expr),NEED_VALUE | EVAL_ADDR);

		no_oper(TOP_GCALL);
		printlabel(label2);

		assign(WORKSPACE,WORKSPACE,FALSE,(INT32) 4);
		assign(TEMPBASE,TEMPBASE,FALSE,(INT32) 4);
		assign(STRUCTBASE,STRUCTBASE,FALSE,(INT32) 4);
		}
	else
		{
		RIGHT(expr)->right.ptr.terminal->type |= REFERENCED_NODE;

		instruction(TOP_CALL,REL,RIGHT(expr)->right.ptr.terminal->name,
			NULL,(INT32) 0,NULL);
		}
	}

/*
 *	Generate transfer-of-control code assuming parameters are ready.  This
 *	particular routine manually builds the call stack frame for cases where
 *	objects cross the boundary between the first three words and the rest.
 */
PRIVATE	VOID
do_call2(expr)
	struct	op_node	*expr;
	{
	char			buf1[MAXLABEL + 1];
	char			buf2[MAXLABEL + 1];
	LABEL			label1;
	LABEL			label2;

	instruction(TOP_LDC,ABS,strlabel(label2 = getlabel(),buf2),
		strlabel(label1 = getlabel(),buf1),(INT32) 0,NULL);
	no_oper(TOP_LDPI);
	printlabel(label1);
	abs_oper(TOP_STL,(INT32) 0);

	if(! ISTERMINAL(RIGHT(expr)))	/* Indirect call */
		{
/*
 *	Evaluate the address of the function to call.
 */
		(VOID) expr_code(NULL,RIGHT(expr),NEED_VALUE | EVAL_ADDR);
		no_oper(TOP_GCALL);
		}
	else
		{
		RIGHT(expr)->right.ptr.terminal->type |= REFERENCED_NODE;
		instruction(TOP_J,REL,RIGHT(expr)->right.ptr.terminal->name,
			NULL,(INT32) 0,NULL);
		}
	printlabel(label2);
	}

/*
 *	Handle special case of up to the first three parameters being 1
 *	word in size.  Also used by the intrinsic routines which have integral
 *	"arguments" to load them in desired order.
 */
PUBLIC	unsigned int
do_call3(numargs,args,real_func)
	int			numargs;
	struct	arg_info	args[];
	BOOL			real_func;
	{
	TEMP			temp1;
	TEMP			temp2;
/*
 *	Given the number of parameters, generate code to place up to the first
 *	three in registers.
 */
	if(numargs == 1)
		{
		dvrt_append(args[0].code);
		return (args[0].depth);
		}
	else if(numargs == 2)
		{
		if((args[0].depth & INT_STACK) < INT_DEEP(3))
			{
			dvrt_append(args[1].code);
			dvrt_append(args[0].code);
			return (depth_max(args[1].depth,args[0].depth +
				INT_DEEP(1)));
			}
		else if((args[1].depth & INT_STACK) <
			(args[2].depth & INT_STACK))
			{
			dvrt_append(args[0].code);
			dvrt_append(args[1].code);
			rev(Inttype);
			return (depth_max(args[0].depth,args[1].depth +
				INT_DEEP(1)));
			}
		else
			{
			temp1 = get_temp();
			dvrt_free(args[0].code);
			if(real_func)
				(VOID) eval_1arg(args[0].expr);
			else
				(VOID) expr_code(NULL,args[0].expr,NEED_VALUE);
			store_temp(Inttype, temp1);
			dvrt_free(args[1].code);
			if(real_func)
				(VOID) eval_1arg(args[1].expr);
			else
				(VOID) expr_code(NULL,args[1].expr,NEED_VALUE);
			load_temp(Inttype, temp1);
			free_temp(temp1);
			/* Note that integer depth of one MUST be 3 */
			return (depth_max(args[0].depth,args[1].depth));
			}
		}
	else if(numargs >= 3)
		{
/*
 *	This case is a little complicated:  We have three parameters to stack,
 *	each of which can have 3 possible depths.  These 27 possibilities have
 *	been reduced to 9 distinct code generation sequences.  The following
 *	table shows which depth values translate into which cases:
 *
 *		DEPTH		CASE
 *	arg#0	arg#1	arg#2
 *	1	1	1	A
 *	1	1	2	C
 *	1	1	3	D
 *	1	2	1	A
 *	1	2	2	E
 *	1	2	3	E
 *	1	3	1	F
 *	1	3	2	G
 *	1	3	3	G
 *	2	1	1	A
 *	2	1	2	C
 *	2	1	3	D
 *	2	2	1	A
 *	2	2	2	E
 *	2	2	3	E
 *	2	3	1	F
 *	2	3	2	G
 *	2	3	3	G
 *	3	1	1	A
 *	3	1	2	C
 *	3	1	3	E
 *	3	2	1	A
 *	3	2	2	E
 *	3	2	3	E
 *	3	3	1	H
 *	3	3	2	I
 *	3	3	3	B
 */
/* CASE A */	if(((args[0].depth & INT_STACK) == INT_DEEP(1)) &&
		    ((args[1].depth & INT_STACK) < INT_DEEP(3)))
			{
			dvrt_append(args[2].code);
			dvrt_append(args[1].code);
			dvrt_append(args[0].code);
			}
/* CASE B */	else if(((args[0].depth & INT_STACK) == INT_DEEP(3)) &&
			 ((args[1].depth & INT_STACK) == INT_DEEP(3)) &&
			 ((args[2].depth & INT_STACK) == INT_DEEP(3)))
			{
			dvrt_append(args[0].code);
			temp1 = get_temp();
			store_temp(Inttype, temp1);
			dvrt_free(args[1].code);
			if(real_func)
				(VOID) eval_1arg(args[1].expr);
			else
				(VOID) expr_code(NULL,args[1].expr,NEED_VALUE);
			temp2 = get_temp();
			store_temp(Inttype, temp2);
			dvrt_free(args[2].code);
			if(real_func)
				(VOID) eval_1arg(args[2].expr);
			else
				(VOID) expr_code(NULL,args[2].expr,NEED_VALUE);
			load_temp(Inttype, temp2);
			load_temp(Inttype, temp1);
			free_temp(temp1);
			free_temp(temp2);
			}
/* CASE C */	else if(((args[0].depth & INT_STACK) == INT_DEEP(2)) &&
			 ((args[1].depth & INT_STACK) == INT_DEEP(1)))
			{
			dvrt_append(args[2].code);
			dvrt_append(args[0].code);
			dvrt_append(args[1].code);
			rev(Inttype);
			}
/* CASE D */	else if(((args[0].depth & INT_STACK) == INT_DEEP(3)) &&
			 ((args[1].depth & INT_STACK) == INT_DEEP(1)) &&
			 ((args[2].depth & INT_STACK) < INT_DEEP(3)))
			{
			dvrt_append(args[0].code);
			dvrt_append(args[2].code);
			rev(Inttype);
			dvrt_append(args[1].code);
			rev(Inttype);
			}
/* CASE E */	else if((((args[0].depth & INT_STACK) > INT_DEEP(1)) &&
			  ((args[1].depth & INT_STACK) == INT_DEEP(2))) ||
			 (((args[0].depth & INT_STACK) == INT_DEEP(3)) &&
			  ((args[1].depth & INT_STACK) == INT_DEEP(1)) &&
			  ((args[2].depth & INT_STACK) == INT_DEEP(3))))
			{
			dvrt_append(args[0].code);
			temp1 = get_temp();
			store_temp(Inttype, temp1);
			dvrt_free(args[2].code);
			dvrt_free(args[1].code);
			if(real_func)
				{
				(VOID) eval_1arg(args[2].expr);
				(VOID) eval_1arg(args[1].expr);
				}
			else
				{
				(VOID) expr_code(NULL,args[2].expr,NEED_VALUE);
				(VOID) expr_code(NULL,args[1].expr,NEED_VALUE);
				}
			load_temp(Inttype, temp1);
			free_temp(temp1);
			}
/* CASE F */	else if(((args[0].depth & INT_STACK) == INT_DEEP(1)) &&
			 ((args[1].depth & INT_STACK) == INT_DEEP(3)) &&
			 ((args[2].depth & INT_STACK) < INT_DEEP(3)))
			{
			dvrt_append(args[1].code);
			dvrt_append(args[2].code);
			rev(Inttype);
			dvrt_append(args[0].code);
			}
/* CASE G */	else if(((args[0].depth & INT_STACK) > INT_DEEP(1)) && 
			 ((args[1].depth & INT_STACK) == INT_DEEP(3)) &&
			 ((args[2].depth & INT_STACK) < INT_DEEP(3)))
			{
			dvrt_append(args[0].code);
			temp1 = get_temp();
			store_temp(Inttype, temp1);
			dvrt_free(args[1].code);
			dvrt_free(args[2].code);
			if(real_func)
				{
				(VOID) eval_1arg(args[1].expr);
				(VOID) eval_1arg(args[2].expr);
				}
			else
				{
				(VOID) expr_code(NULL,args[1].expr,NEED_VALUE);
				(VOID) expr_code(NULL,args[2].expr,NEED_VALUE);
				}
			rev(Inttype);
			load_temp(Inttype, temp1);
			free_temp(temp1);
			}
/* CASE H */	else if(((args[0].depth & INT_STACK) == INT_DEEP(1)) &&
			 ((args[1].depth & INT_STACK) == INT_DEEP(3)) &&
			 ((args[2].depth & INT_STACK) == INT_DEEP(3)))
			{
			dvrt_append(args[1].code);
			temp1 = get_temp();
			store_temp(Inttype, temp1);
			dvrt_free(args[2].code);
			dvrt_free(args[0].code);
			if(real_func)
				(VOID) eval_1arg(args[2].expr);
			else
				(VOID) expr_code(NULL,args[2].expr,NEED_VALUE);
			load_temp(Inttype, temp1);
			if(real_func)
				(VOID) eval_1arg(args[0].expr);
			else
				(VOID) expr_code(NULL,args[0].expr,NEED_VALUE);
			free_temp(temp1);
			}
/* CASE I */	else if(((args[0].depth & INT_STACK) == INT_DEEP(2)) &&
			 ((args[1].depth & INT_STACK) == INT_DEEP(3)) &&
			 ((args[2].depth & INT_STACK) == INT_DEEP(3)))
			{
			dvrt_append(args[1].code);
			temp1 = get_temp();
			store_temp(Inttype, temp1);
			dvrt_free(args[2].code);
			dvrt_free(args[0].code);
			if(real_func)
				{
				(VOID) eval_1arg(args[2].expr);
				(VOID) eval_1arg(args[0].expr);
				}
			else
				{
				(VOID) expr_code(NULL,args[2].expr,NEED_VALUE);
				(VOID) expr_code(NULL,args[0].expr,NEED_VALUE);
				}
			load_temp(Inttype, temp1);
			rev(Inttype);
			free_temp(temp1);
			}
		else
			fatal("Compiler error: do_call3");
		}
	return (depth_max(depth_max(depth_max(args[0].depth,args[1].depth),
		args[2].depth),INT_DEEP(3)));
	}

PUBLIC	INT32
do_args(param,cargsize,min_ajw)
	struct	op_node		*param;
	INT32			cargsize;
	BOOL			min_ajw;
	{
	struct op_node		*op;
	INT32			size;
	struct id_node		*sym;
	INT32			targsize;
/*
 *	Compute size of parameter.
 */
	if(EXPRTYPE(RIGHT(param)) & (FP_NODE | SU_NODE))
		{
		size = word((RIGHT(param)->node_type->val.integral +
				(Sizeof_int - 1)));
		}
	else
		size = 1;
/*
 *	Go recursive to handle additional parameters.
 */
	if(LEFT(param) != NULL)
		targsize = do_args(LEFT(param),cargsize + size,min_ajw);
	else
		{
		targsize = cargsize + size;
		if(min_ajw && (targsize < 4))
			targsize = 4;
		assign(WORKSPACE,WORKSPACE,TRUE,(INT32) targsize);
		assign(TEMPBASE,TEMPBASE,TRUE,(INT32) targsize);
		assign(STRUCTBASE,STRUCTBASE,TRUE,(INT32) targsize);
		abs_oper(TOP_AJW,(INT32) -targsize);
		}
/*
 *	Generate the code for evaluating this parameter.  First handle special
 *	cases where parameters are easily loaded onto the integer stack.
 */
	if(((RIGHTTYPE(param) & FP_NODE) ||
		((RIGHTTYPE(param) & SU_NODE) && (size <= 2))) &&
			(is_simple(RIGHT(param)) || (RIGHT(param)->lvalue &&
				((RIGHT(param))->operator != OP_CONVERT))))
		{
		switch((int) size)
			{
			case	1:
				(VOID) eval_1arg(RIGHT(param));
				abs_oper(TOP_STL,(INT32) cargsize);
				break;
			case	2:
				(VOID) eval_2arg(RIGHT(param));
				abs_oper(TOP_STL,(INT32) cargsize);
				abs_oper(TOP_STL,(INT32) cargsize + 1);
				break;
			default:
				fatal("Compiler error: do_args");
			}
		}
	else if(RIGHTTYPE(param) & SU_NODE)	/* struct/union */
		{
		if((RIGHT(param)->operator == OP_CALL) &&
			(ISTERMINAL(op = RIGHT(LEFT(RIGHT(param))))))
			{
			sym = op->right.ptr.terminal;
			assign(sym->name,STRUCTBASE,FALSE,(INT32) cargsize);
			(VOID) expr_code(NULL,RIGHT(param),NEED_VALUE);
			}
		else
			{
			(VOID) expr_code(NULL,RIGHT(param),
				NEED_VALUE | EVAL_ADDR);
			abs_oper(TOP_LDLP,(INT32) cargsize);
			loadconstant((INT32)
				(RIGHT(param)->node_type->val.integral));
			no_oper(TOP_MOVE);
			}
		}
	else				/* FP or int expression */
		{
		(VOID) expr_code(NULL,RIGHT(param),NEED_VALUE);
		if(RIGHTTYPE(param) & FP_NODE)
			{
			abs_oper(TOP_LDLP,(INT32) cargsize);
			if(size == 1)
				no_oper(TOP_FPSTNLSN);
			else
				no_oper(TOP_FPSTNLDB);
			}
		else
			abs_oper(TOP_STL,(INT32) cargsize);
		}
	return (targsize);
	}

/*
 *	Evaluates a single word floating point, integral or struct/union
 *	argument onto the integer stack.
 */
PRIVATE	unsigned	int
eval_1arg(expr)
	struct	op_node		*expr;
	{
	BOOL			addrflag;
	char			buffer1[MAXLABEL + 1];
	unsigned	int	depth;
	INT32			offset;
	TEMP			temp;
	INT32			value;
	struct	op_node		*var;

	if(EXPRTYPE(expr) & (SU_NODE | FP_NODE))
		{
		if(is_simple(expr))
			{
			depth = INT_DEEP(1);	/* All 1 deep */

			var = constaddr(expr,&offset,&addrflag);
			if(var->right.ptr.terminal->diversion != NULL)
				dvrt_copy(var->right.ptr.terminal->diversion);

			switch((int) (var->right.ptr.terminal->
				info.storage_class))
				{
				case	SC_AUTO:
				case	SC_REG:
				case	SC_ARG:
		instruction(TOP_LDL,ABS,WORKSPACE,NULL,
			-(word(-var->right.ptr.terminal->address.offset)
				- word(offset)),
#ifdef	DEBUG
			(Debug ? var->right.ptr.terminal->name : NULL)
#else
			NULL
#endif	/* DEBUG */
			);
		break;

				case	SC_CONSTANT:
		if(var->right.ptr.terminal->val.fvalue == 0.0)
			loadconstant((INT32) 0);	/* == (float) 0 */
		else
			{
/*
 *	It would generate cleaner code if the compiler knew how to generate the
 *	actual bit pattern for a floating constant and issue a loadconstant()
 *	here instead of bypassing this to the assembler.
 */
			(VOID) fp_oplit(expr->node_type,var);
			abs_oper(TOP_LDNL,(INT32) 0);
			}
		break;

				case	SC_EXTERN:
				case	SC_ST_EXT:
		var->right.ptr.terminal->type |= REFERENCED_NODE;
		/* Fall into normal static var reference */

				case	SC_GLOBAL:
				case	SC_ST_GLB:
				case	SC_STATIC:
				case	SC_LITERAL:
		if(Static_link &&
			(! (var->right.ptr.terminal->type & FUNC_NODE)) &&
				(var->right.ptr.terminal->info.storage_class !=
					SC_LITERAL))
			{
/*
 *	Compute global data address using static link approach.
 */
			instruction(TOP_LDL,ABS,WORKSPACE,NULL,Slboff,
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
				);
			instruction(TOP_ADC,ABS,real_name(var->
				right.ptr.terminal,buffer1),SLBNAME,offset,
#ifdef	DEBUG
				(Debug ? var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
				);
			}
		else
			{
/*
 *	Use assembler pseudo-op to allow for optimized constant loading using
 *	the same rules at link time as loadconstant().
 */
			instruction(TOP_LDC,ABS,real_name(var->
				right.ptr.terminal,buffer1),NULL,offset,
#ifdef	DEBUG
				((Debug && strcmp(real_name(var->
					right.ptr.terminal,buffer1),
					var->right.ptr.terminal->name)) ?
					var->right.ptr.terminal->name : NULL)
#else
				NULL
#endif	/* DEBUG */
				);
			}

		abs_oper(TOP_LDNL, (INT32) 0);
		break;

				case	SC_RSTEMP:
		value = var->right.ptr.terminal->address.tempsize;
		if(value > Structspace)
			Structspace = value;
		instruction(TOP_LDLP,ABS,STRUCTBASE,
			((value != 0) ? NULL : var->right.ptr.terminal->name),
				(INT32) 0,
#ifdef	DEBUG
			(Debug ? "struct/union function return ptr" : NULL)
#else
			NULL
#endif	/* DEBUG */
			);
		    break;
				}
			}
		else if(((EXPRTYPE(expr) & SU_NODE) || expr->lvalue) &&
			(expr->operator != OP_CONVERT))
			{
/*
 *	Evaluate for address and then load it into integer stack.
 */
			depth = expr_code(NULL,expr,EVAL_ADDR | NEED_VALUE);
			abs_oper(TOP_LDNL,(INT32) 0);
			}
		else
			{
			/* A FP expr which is not a simple or lvalue */
			depth = expr_code(NULL,expr,NEED_VALUE);
			temp = get_temp();
			store_temp(expr->node_type,temp);
			load_temp(Inttype,temp);
			free_temp(temp);
			}
		}
	else
		{
/*
 *	Handle integral cases + arrays/functions (treated as "pointers-to").
 */
		depth = expr_code(NULL,expr,NEED_VALUE);
		}

	return (depth & INT_STACK);	/* Only return integer component */
	}

/*
 *	Evaluates a double word floating point or struct/union argument onto
 *	the integer stack.
 */
PRIVATE	unsigned	int
eval_2arg(expr)
	struct	op_node		*expr;
	{
	BOOL			addrflag;
	char			buffer1[MAXLABEL + 1];
	unsigned	int	depth;
	INT32			offset;
	TEMP			temp;
	struct	op_node		*var;

	if(is_simple(expr))
		{
		depth = INT_DEEP(2);	/* All 2 deep */

		var = constaddr(expr,&offset,&addrflag);
		if(var->right.ptr.terminal->diversion != NULL)
			dvrt_copy(var->right.ptr.terminal->diversion);

		switch((int) (var->right.ptr.terminal->info.storage_class))
			{
			case	SC_AUTO:
			case	SC_REG:
			case	SC_ARG:
		instruction(TOP_LDL,ABS,WORKSPACE,NULL,
			-(word(-var->right.ptr.terminal->address.offset)
				- word(offset) - 1),
#ifdef	DEBUG
			(Debug ? var->right.ptr.terminal->name : NULL)
#else
			NULL
#endif	/* DEBUG */
			);
		instruction(TOP_LDL,ABS,WORKSPACE,NULL,
			-(word(-var->right.ptr.terminal->address.offset)
				- word(offset)),
#ifdef	DEBUG
			(Debug ? var->right.ptr.terminal->name : NULL)
#else
			NULL
#endif	/* DEBUG */
			);
		break;

			case	SC_CONSTANT:
		if(var->right.ptr.terminal->val.fvalue == 0.0)
			{
			loadconstant((INT32) 0);	/* == (float) 0 */
			loadconstant((INT32) 0);
			}
		else
			{
/*
 *	It would generate cleaner code if the compiler knew how to generate the
 *	actual bit pattern for a floating constant and issue a loadconstant()
 *	here instead of bypassing this to the assembler.
 */
			if(CPU_DUP)
				{
				(VOID) fp_oplit(expr->node_type,var);
				no_oper(TOP_DUP);
				abs_oper(TOP_LDNL,(INT32) 1);
				rev(Inttype);
				abs_oper(TOP_LDNL,(INT32) 0);
				}
			else
				{
				(VOID) fp_oplit(expr->node_type,var);
				temp = get_temp();
				store_temp(Inttype,temp);
				load_temp(Inttype,temp);
				abs_oper(TOP_LDNL,(INT32) 1);
				load_temp(Inttype,temp);
				abs_oper(TOP_LDNL,(INT32) 0);
				free_temp(temp);
				}
			}
		break;

			case	SC_EXTERN:
			case	SC_ST_EXT:
		var->right.ptr.terminal->type |= REFERENCED_NODE;
		/* Fall into normal static var reference */

			case	SC_GLOBAL:
			case	SC_ST_GLB:
			case	SC_STATIC:
		if(Static_link)
			{
/*
 *	Compute global data address using static link approach.
 */
			instruction(TOP_LDL,ABS,WORKSPACE,NULL,Slboff,
#ifdef	DEBUG
			(Debug ? var->right.ptr.terminal->name : NULL)
#else
			NULL
#endif	/* DEBUG */
				);
			instruction(TOP_ADC,ABS,real_name(var->
				right.ptr.terminal,buffer1),SLBNAME,offset,
#ifdef	DEBUG
			(Debug ? var->right.ptr.terminal->name : NULL)
#else
			NULL
#endif	/* DEBUG */
				);
			}
		else
			{
/*
 *	Use assembler pseudo-op to allow for optimized constant loading using
 *	the same rules at link time as loadconstant().
 */
			instruction(TOP_LDC,ABS,real_name(var->
				right.ptr.terminal,buffer1),NULL,offset,
#ifdef	DEBUG
			((Debug && strcmp(real_name(var->
				right.ptr.terminal,buffer1),
				var->right.ptr.terminal->name)) ?
				var->right.ptr.terminal->name : NULL)
#else
			NULL
#endif	/* DEBUG */
				);
			}

		if(CPU_DUP)
			{
			no_oper(TOP_DUP);
			abs_oper(TOP_LDNL,(INT32) 1);
			rev(Inttype);
			abs_oper(TOP_LDNL,(INT32) 0);
			}
		else
			{
			temp = get_temp();
			store_temp(Inttype,temp);
			load_temp(Inttype,temp);
			abs_oper(TOP_LDNL,(INT32) 1);
			load_temp(Inttype,temp);
			abs_oper(TOP_LDNL,(INT32) 0);
			free_temp(temp);
			}
		break;
			}
		}
	else if(((EXPRTYPE(expr) & SU_NODE) || expr->lvalue) &&
		(expr->operator != OP_CONVERT))
		{
/*
 *	Evaluate for address and then load it into integer stack.
 */
		depth = depth_max(expr_code(NULL,expr,EVAL_ADDR | NEED_VALUE),
			INT_DEEP(2));
		if(CPU_DUP)
			{
			no_oper(TOP_DUP);
			abs_oper(TOP_LDNL,(INT32) 1);
			rev(Inttype);
			abs_oper(TOP_LDNL,(INT32) 0);
			}
		else
			{
			temp = get_temp();
			store_temp(Inttype,temp);
			load_temp(Inttype,temp);
			abs_oper(TOP_LDNL,(INT32) 1);
			load_temp(Inttype,temp);
			abs_oper(TOP_LDNL,(INT32) 0);
			free_temp(temp);
			}
		}
	else
		{
/*
 *	Complex double precision fp expression as parameter.  Not a simple and
 *	not lvalue. Evaluate and copy to int stack.
 */
		depth = depth_max(expr_code(NULL,expr,NEED_VALUE),INT_DEEP(2));
		temp = get_dtemp();
		store_temp(expr->node_type,temp);
		load_temp(Inttype,temp + 1);
		load_temp(Inttype,temp);
		free_dtemp(temp);
		}

	return (depth & INT_STACK);	/* Only return int component */
	}

/*
 *	Evaluates a 3 word struct/union argument onto the integer stack.
 */
PRIVATE	unsigned	int
eval_3arg(expr)
	struct	op_node		*expr;
	{
	BOOL			addrflag;
	char			buffer1[MAXLABEL + 1];
	INT32			offset;
	TEMP			temp;
	struct	op_node		*var;

	if(is_simple(expr))
		{
		var = constaddr(expr,&offset,&addrflag);
		if(var->right.ptr.terminal->diversion != NULL)
			dvrt_copy(var->right.ptr.terminal->diversion);

		switch((int) (var->right.ptr.terminal->info.storage_class))
			{
			case	SC_AUTO:
			case	SC_REG:
			case	SC_ARG:
		instruction(TOP_LDL,ABS,WORKSPACE,NULL,
			-(word(-var->right.ptr.terminal->address.offset)
				- word(offset) - 2),
#ifdef	DEBUG
			(Debug ? var->right.ptr.terminal->name : NULL)
#else
			NULL
#endif	/* DEBUG */
			);
		instruction(TOP_LDL,ABS,WORKSPACE,NULL,
			-(word(-var->right.ptr.terminal->address.offset)
				- word(offset) - 1),
#ifdef	DEBUG
			(Debug ? var->right.ptr.terminal->name : NULL)
#else
			NULL
#endif	/* DEBUG */
			);
		instruction(TOP_LDL,ABS,WORKSPACE,NULL,
			-(word(-var->right.ptr.terminal->address.offset)
				- word(offset)),
#ifdef	DEBUG
			(Debug ? var->right.ptr.terminal->name : NULL)
#else
			NULL
#endif	/* DEBUG */
			);
		break;

			case	SC_EXTERN:
			case	SC_ST_EXT:
		var->right.ptr.terminal->type |= REFERENCED_NODE;
		/* Fall into normal static var reference */

			case	SC_GLOBAL:
			case	SC_ST_GLB:
			case	SC_STATIC:
		if(Static_link)
			{
/*
 *	Compute global data address using static link approach.
 */
			instruction(TOP_LDL,ABS,WORKSPACE,NULL,Slboff,
#ifdef	DEBUG
			(Debug ? var->right.ptr.terminal->name : NULL)
#else
			NULL
#endif	/* DEBUG */
				);
			instruction(TOP_ADC,ABS,real_name(var->
				right.ptr.terminal,buffer1),SLBNAME,offset,
#ifdef	DEBUG
			(Debug ? var->right.ptr.terminal->name : NULL)
#else
			NULL
#endif	/* DEBUG */
				);
			}
		else
			{
/*
 *	Use assembler pseudo-op to allow for optimized constant loading using
 *	the same rules at link time as loadconstant().
 */
			instruction(TOP_LDC,ABS,real_name(var->
				right.ptr.terminal,buffer1),NULL,offset,
#ifdef	DEBUG
			((Debug && strcmp(real_name(var->
				right.ptr.terminal,buffer1),
				var->right.ptr.terminal->name)) ?
				var->right.ptr.terminal->name : NULL)
#else
			NULL
#endif	/* DEBUG */
				);
			}

		if(CPU_DUP)
			{
			no_oper(TOP_DUP);
			abs_oper(TOP_LDNL,(INT32) 2);
			rev(Inttype);
			no_oper(TOP_DUP);
			abs_oper(TOP_LDNL,(INT32) 1);
			rev(Inttype);
			abs_oper(TOP_LDNL,(INT32) 0);
			}
		else
			{
			temp = get_temp();
			store_temp(Inttype,temp);
			load_temp(Inttype,temp);
			abs_oper(TOP_LDNL,(INT32) 2);
			load_temp(Inttype,temp);
			abs_oper(TOP_LDNL,(INT32) 1);
			load_temp(Inttype,temp);
			abs_oper(TOP_LDNL,(INT32) 0);
			free_temp(temp);
			}
		break;
			}
		}
	else
		{
/*
 *	Evaluate for address and then load it into integer stack.
 */
		(VOID) expr_code(NULL,expr,EVAL_ADDR | NEED_VALUE);
		if(CPU_DUP)
			{
			no_oper(TOP_DUP);
			abs_oper(TOP_LDNL,(INT32) 2);
			rev(Inttype);
			no_oper(TOP_DUP);
			abs_oper(TOP_LDNL,(INT32) 1);
			rev(Inttype);
			abs_oper(TOP_LDNL,(INT32) 0);
			}
		else
			{
			temp = get_temp();
			store_temp(Inttype,temp);
			load_temp(Inttype,temp);
			abs_oper(TOP_LDNL,(INT32) 2);
			load_temp(Inttype,temp);
			abs_oper(TOP_LDNL,(INT32) 1);
			load_temp(Inttype,temp);
			abs_oper(TOP_LDNL,(INT32) 0);
			free_temp(temp);
			}
		}

	return (INT_DEEP(3));		/* Uses all of integer stack */
	}

/*
 *	Do assignment operator.
 */

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#pragma	optimize("g",off)		/* Disable global common subs */
#pragma	optimize("l",off)		/* Disable loop optimizations */
#endif	/* MSC_OPT */

PUBLIC	unsigned int
do_assign(expr,vf)
	struct	op_node		*expr;
	int			vf;
	{
	unsigned	int	depth;
	BOOL			dupped;
	double			fvalue;
	unsigned	int	ldeep;
	struct	dvrt		*left;
	struct	id_node		*oldleft;
	BOOL			quick_right;
	unsigned	int	rdeep;
	struct	dvrt		*right;
	unsigned	int	sdepth;
	TEMP			temp;
	INT32			value;
/*
 *	Determine if the value is needed and it will be shorter to just reload
 *	via ".LDC" (or "FPLDZEROSN"/"FPLDZERODB"), instead of whatever DUP or
 *	"load_indirect" or "do_load" scheme which would otherwise be used.
 */
	quick_right = FALSE;
	if((vf & NEED_VALUE) && ISRIGHTCONSTANT(expr))
		{
		if(EXPRTYPE(expr) & FP_NODE)
			{
			if(RIGHTTYPE(expr) & FP_NODE)
				fvalue = RIGHTFCONSTANT(expr);
			else
				fvalue = l_to_f(RIGHT(expr));
			if(fvalue == 0.0)
				quick_right = TRUE;
			}
		else
			{
			if(RIGHTTYPE(expr) & FP_NODE)
				value = f_to_l(RIGHTFCONSTANT(expr),TRUE);
			else
				value = RIGHTCONSTANT(expr);

			if(Sizeof_int == WORD_SIZE_16B)
				{
				if((value == 0x8000) ||
					((value >= -256) && (value <= 255)))
					{
					quick_right = TRUE;
					}
				}
			else
				{
				if((value == 0x80000000L) ||
					((value >= -256) && (value <= 255)))
					{
					quick_right = TRUE;
					}
				}
			}
		}
/*
 *	If both sides are structures of the same type, then build the
 *	appropriate code to copy a structure across the assignment.
 */
	if((expr->operator == OP_ASSIGN) &&
		(LEFTTYPE(expr) & (ARRAY_NODE | SU_NODE)))
		{
		if(RIGHT(expr)->operator == OP_CALL)
			{
/*
 *	Handle case where function struct return value has already been
 *	stored into a destination struct whose assignment was more tightly
 *	bound.  Now we need to figure out the address of THAT structure so
 *	that we can use it as the right side of THIS assignment...
 */
			(VOID) expr_code(NULL,RIGHT(expr),
				NEED_VALUE | EVAL_ADDR);
			(VOID) expr_code(NULL,RIGHT(LEFT(RIGHT(expr))),
				NEED_VALUE | EVAL_ADDR);
			(VOID) expr_code(NULL,LEFT(expr),
				NEED_VALUE | EVAL_ADDR);
			}
		else
			(VOID) bin_op(expr,NEED_VALUE | EVAL_ADDR);
		if((vf & NEED_VALUE) && (! (is_simple(LEFT(expr)))))
			{
			temp = get_temp();
			if(CPU_DUP)
				{
				no_oper(TOP_DUP);
				store_temp(Inttype,temp);
				}
			else
				{
				store_temp(Inttype,temp);
				load_temp(Inttype,temp);
				}
			}
		loadconstant(RIGHT(expr)->node_type->val.integral);
		no_oper(TOP_MOVE);
		/* Remember that "move" can be timesliced */
		depth = INT_DEEP(3) | FLT_DEEP(3);
		if(vf & NEED_VALUE)
			{
			if(is_simple(LEFT(expr)))
				{
				(VOID) expr_code(NULL,LEFT(expr),
					NEED_VALUE | EVAL_ADDR);
				}
			else
				{
				load_temp(Inttype,temp);
				free_temp(temp);
				}
			}
		}
/*
 *	Is it a simple floating point store that can be done entirely on the
 *	integer stack?
 */
	else if((expr->operator == OP_ASSIGN) &&
		((LEFTTYPE(expr) & FP_NODE) &&
			is_simple(LEFT(expr)) &&
				(LEFT(expr)->node_type->val.integral <=
					Sizeof_int) &&
						is_simple(RIGHT(expr)) &&
		(RIGHT(expr)->node_type->val.integral <= Sizeof_int) &&
		((RIGHTTYPE(expr) & FP_NODE) || ISRIGHTCONSTANT(expr))))
		{
/*
 *	If a simple 32 bit floating point assignment, use integer stack to
 *	perform the copy, and then reload to the fp stack if value needed.
 */
		/* Change type to fool do_store */
		oldleft = LEFT(expr)->node_type;
		LEFT(expr)->node_type = Inttype;

		if(ISRIGHTCONSTANT(expr))
			{
			depth = fp_oplit(RIGHT(expr)->node_type,RIGHT(expr));
			abs_oper(TOP_LDNL,(INT32) 0);
			depth = depth_max(depth,INT_DEEP(1));
			}
		else
			depth = eval_1arg(RIGHT(expr));

		depth = depth_max(depth,do_store(LEFT(expr)));
		LEFT(expr)->node_type = oldleft;

		/* Reload if needed */
		depth = depth_max(depth,do_load(expr->node_type,LEFT(expr),
			vf));
		}
/*
 *	Otherwise, is the left easily accessable?
 */
	else if(is_simple(LEFT(expr)))
		{
		depth = expr_code(expr->node_type,RIGHT(expr),NEED_VALUE);
		sdepth = 0;		/* Integer stack used below store */
		if((vf & NEED_VALUE) && (! quick_right))
			{
			dvrt_stack();
			if((do_store(LEFT(expr)) & INT_STACK) == INT_DEEP(3))
				dupped = FALSE;
			else
				dupped = TRUE;
			dvrt_free(dvrt_unstack());
			if(dupped)
				{
				if(LEFTTYPE(expr) & FP_NODE)
					{
					dupped = TRUE;
					no_oper(TOP_FPDUP);
					depth = depth_max(depth,FLT_DEEP(2));
					}
				else if(CPU_DUP &&
		    ((! (LEFTTYPE(expr) & CHAR_NODE))||ISRIGHTCONSTANT(expr)))
					{
					dupped = TRUE;
					no_oper(TOP_DUP);
					sdepth = INT_DEEP(1);
					}
				else
					dupped = FALSE;
				}
			}
		depth = depth_max(do_store(LEFT(expr)) + sdepth,depth);

		if(vf & NEED_VALUE)
			{
			if(quick_right)
				{
				depth = depth_max(do_load(expr->node_type,
					RIGHT(expr),vf),depth);
				}
			else if(! dupped)
				{
				depth = depth_max(do_load(expr->node_type,
					LEFT(expr),vf),depth);
				}
			}
		}
/*
 *	Otherwise we need to determine the optimum load order for the
 *	left and right operands to minimize the use of temps and rev's.
 */
	else
		{
		dvrt_stack();
		ldeep = expr_code(NULL,LEFT(expr),NEED_VALUE | EVAL_ADDR);
		left = dvrt_unstack();

		dvrt_stack();
		rdeep = expr_code(expr->node_type,RIGHT(expr),NEED_VALUE);
		right = dvrt_unstack();

		if(((! (vf & NEED_VALUE)) || quick_right) &&
			(((EXPRTYPE(expr) & FP_NODE) &&
				((ldeep & FLT_STACK) < FLT_DEEP(3))) ||
			((! (EXPRTYPE(expr) & FP_NODE)) &&
				((ldeep & INT_STACK) < INT_DEEP(3)))))
			{
			dvrt_append(right);
			dvrt_append(left);
			depth = depth_max(rdeep,ldeep +
		((EXPRTYPE(expr) & FP_NODE) ? FLT_DEEP(1) : INT_DEEP(1)));
			depth = depth_max(depth,store_indirect(expr,NULL));
			if(quick_right)
				{
				depth = depth_max(depth,do_load(expr->
					node_type,RIGHT(expr),vf));
				}
			}
		else if(((CPU_DUP &&
			((! (LEFTTYPE(expr) & CHAR_NODE)) ||
				ISRIGHTCONSTANT(expr))) ||
					(EXPRTYPE(expr) & FP_NODE)) &&
			(((EXPRTYPE(expr) & FP_NODE) &&
				((ldeep & FLT_STACK) <= FLT_DEEP(1))) ||
			((! (EXPRTYPE(expr) & FP_NODE)) &&
				((ldeep & INT_STACK) <= INT_DEEP(1)))))
			{
			dvrt_append(right);
			fp_oper(expr->node_type,TOP_DUP,TOP_FPDUP,TOP_FPDUP);
			dvrt_append(left);
			depth = depth_max(rdeep,ldeep +
		((EXPRTYPE(expr) & FP_NODE) ? FLT_DEEP(2) : INT_DEEP(2)));
			depth = depth_max(depth,store_indirect(expr,NULL));
			}
		else if(((! (vf & NEED_VALUE)) || quick_right) &&
			((rdeep & INT_STACK) < INT_DEEP(3)))
			{
			dvrt_append(left);
			dvrt_append(right);
			if(! (EXPRTYPE(expr) & FP_NODE))
				rev(Inttype);
			depth = depth_max(ldeep,rdeep + INT_DEEP(1));
			depth = depth_max(depth,store_indirect(expr,NULL));
			if(quick_right)
				{
				depth = depth_max(depth,do_load(expr->
					node_type,RIGHT(expr),vf));
				}
			}
		else
			{
			dvrt_free(right);	/* To avoid temp overlap */
			dvrt_append(left);
			temp = get_temp();
			store_temp(Inttype,temp);	/* Save address */
			(VOID) expr_code(expr->node_type,RIGHT(expr),
				NEED_VALUE);
			depth = depth_max(ldeep,rdeep);
			load_temp(Inttype,temp);	/* Reload address */
			depth = depth_max(depth,
				(unsigned int) ((EXPRTYPE(expr) & FP_NODE) ?
					INT_DEEP(1) : INT_DEEP(2)));
			depth = depth_max(depth,store_indirect(expr,NULL));

			if(vf & NEED_VALUE)
				{
				if(quick_right)
					{
					depth = depth_max(depth,do_load(expr->
						node_type,RIGHT(expr),vf));
					}
				else
					{
					/* Reload address */
					load_temp(Inttype,temp);
					depth = depth_max(depth,
						load_indirect(LEFT(expr),NULL,
							(INT32) 0,
								NEED_VALUE));
					}
				}
			free_temp(temp);
			}
		}
	return (depth);
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#pragma	optimize("g",on)		/* Enable global common subs */
#pragma	optimize("l",on)		/* Enable loop optimizing */
#endif	/* MSC_OPT */

PUBLIC	unsigned int
do_convert(expr,value_flag)
	struct	op_node		*expr;
	int			value_flag;
	{
	unsigned int		depth;

	if(! (value_flag & NEED_VALUE))
		return (expr_code(expr->node_type,RIGHT(expr),NO_VALUE));
/*
 *	For now the tests are simple:  If the type of the cast node
 *	is (char) and the type of the old node is NOT char, then we
 *	must convert.  Note that we make expr_code do any conversions
 *	we don't want to deal with.
 */
	if((RIGHT(expr)->node_type != Chartype) &&
		(RIGHT(expr)->node_type != Uchartype))
		{
		if(expr->node_type == Chartype)
			{
			/* Converting non-char to char */
			depth = expr_code(expr->node_type,RIGHT(expr),
					value_flag);
			loadconstant((INT32) 0xFF);
			no_oper(TOP_AND);
			loadconstant((INT32) 0x80);
			no_oper(TOP_XWORD);
			depth = depth_max(depth,INT_DEEP(2));
			}
		else if(expr->node_type == Uchartype)
			{
			/* Converting non-char to unsigned char */
			depth = expr_code(expr->node_type,RIGHT(expr),
					value_flag);
			loadconstant((INT32) 0xFF);
			no_oper(TOP_AND);
			depth = depth_max(depth,INT_DEEP(2));
			}
		else
			{
			depth = expr_code(expr->node_type,RIGHT(expr),
					value_flag);
			}
		}
	else
		{
/*
 *	We are casting from some kind of char.  If we are casting
 *	to a char, then change type of lower node and just evaluate.
 */
		if(expr->node_type->type & CHAR_NODE)
			RIGHT(expr)->node_type = expr->node_type;

		depth = expr_code(expr->node_type,RIGHT(expr),value_flag);
		}

	return (depth);
	}

/* End of tcx6d.c */
