/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: tcx7.c,v 2.2 92/04/21 11:11:19 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	TCX Compiler (7 of 11)	Peephole optimizer
 *
 *	Copyright 1988-1991 by Logical Systems.
 *
 *	All rights reserved.
 *
 *	Published by:	Logical Systems
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	6/10/88
 *	Dave Regan & Kirk Bailey
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<string.h>
#include	"tcx.h"
#include	"tcxdecl.h"
#include	"tcxext.h"

/*
 *	Local function declarations.
 */
PRIVATE	char			*p_constant();
PRIVATE	BOOL			p_line();
PRIVATE	VOID			p_optimize();
PRIVATE	int			p_replace();
PRIVATE	VOID			p_write();
PRIVATE	BOOL			m_blank();
PRIVATE	BOOL			m_cvar1();
PRIVATE	BOOL			m_cvar2();
PRIVATE	BOOL			m_jvar1();
PRIVATE	BOOL			m_nvar1();
PRIVATE	BOOL			m_nvar2();
PRIVATE	BOOL			m_nvar3();
PRIVATE	BOOL			m_svar1();
PRIVATE	BOOL			m_var1();
PRIVATE	BOOL			m_var3();
PRIVATE	BOOL			m_var4();
PRIVATE	BOOL			m_var5();
PRIVATE	BOOL			m_varx();
PRIVATE	BOOL			m_zero();
PRIVATE	BOOL			r_blank();
PRIVATE	BOOL			r_case1();
PRIVATE	BOOL			r_case2();
PRIVATE	BOOL			r_case3();
PRIVATE	BOOL			r_case4();
PRIVATE	BOOL			r_case5();
PRIVATE	BOOL			r_var1();
PRIVATE	BOOL			r_var3();
PRIVATE	BOOL			r_var4();
PRIVATE	BOOL			r_var5();

/*
 *	This file is a largely self-contained module with a limited number of
 *	entry points (functions and global variables).
 *
 *	p_init()	Start the optimizer
 *	p_out()		Read a diversion buffer into the optimizer
 *	p_flush()	Clear out the peephole window
 *	p_summary()	Print a summary of the rules fired
 *
 *	Debug		Allow debugging information
 */

/*
 *	Things which could be improved:
 *
 *	When dealing with constants and math, if it is possible to do the math
 *	internally, do so.  At least remove redundant "+" signs.
 *
 *	Allow multiple things to be in the operand string.  E.g. both constant
 *	characters as well as numeric constants.
 */

/*
 *	Possible new rules and why they cannot yet be used.  This section is
 *	included to collect good thoughts, but cannot currently be implemented
 *	as it would require some extension to the basic optimizer.
 *
 *	1.	.LDC	X	==>	.LDC	X
 *		.LDC	Y		DUP
 *					ADC	Y - X
 *
 *	Useful whex X and Y are close and the instruction length for .LDC Y
 *	is long.  However, this requires some math.
 *
 *	2.	LDLP	X	==>	LDLP	X
 *		LDLP	X		DUP
 *
 *	Useful when X requires at least one prefix byte, but not if X requires
 *	no prefixes.  Will require keeping track of the current values of "?ws"
 *	as well as math capabilities.
 */

/*
 *	Obvious rules not included, and why they can't work (in general)
 *	This section is here to prevent people from going down promising paths,
 *	which have unfortunately been unsuccessfully traveled at least once
 *	before.
 */

/* NOTHING IN THIS SECTION CURRENTLY */

/************************************************************************/
/*									*/
/*   Definitions/structures and variables used only by the optimizer.	*/
/*									*/
/************************************************************************/

#define	WINDOW_SIZE	5		/* # lines in peephole window	*/
#define	WINDOW_VARS	5		/* # of match registers		*/

struct	peepreg				/* A template matching register	*/
	{
	BOOL	in_use;			/* TRUE if register is in use	*/
	char	*contents;		/* Register contents		*/
	};

struct	peepbuf				/* A line in the peephole buf	*/
	{
	int	opcode;			/* Opcode token for this line	*/
	char	*operand;		/* Operand string for this line */
	char	*comment;		/* Comment string for this line	*/
	};

struct	subp				/* A line in a optimize pattern	*/
	{
	int	opcode;			/* Opcode */
	BOOL	(*operand)();		/* Predicate func for oper. fld.*/
	};

struct	pattern				/* Full pattern (set of lines)	*/
	{
	struct	subp	*src;		/* Points to array of src lines	*/
	struct	subp	*dst;		/* Points to array of dst lines */
	int	count;			/* # of times rule gets fired	*/
	int	cpu_type;		/* Allowed CPU type(s) for rule	*/
	};

PRIVATE	int	FP_rcnt		=	0;	/* FP conc. opt. count	*/
PRIVATE	struct	peepbuf	Lines[WINDOW_SIZE]={0};	/* Peephole buffers	*/
PRIVATE	int		P_lcur	=	0;	/* Line # @ head of buf	*/
PRIVATE	int		P_lnum	=	0;	/* # of buffers in use	*/
PRIVATE	struct	peepreg	Vars[WINDOW_VARS] = {0};/* Peephole registers	*/

/************************************************************************/
/*									*/
/*	The reduction tables.  The general rule is that a replacement	*/
/*	pattern must NOT be longer than the pattern which was matched.	*/
/*									*/
/************************************************************************/

/*
 *	LDLP <s1>\nADC <s2>\nLDNL <s3>	==>  LDL <s1>+(<s2>/"wordsize")+<s3>
 */
PRIVATE	struct	subp	src0[]	=
	{
		{	TOP_LDLP,	m_nvar1	},
		{	TOP_ADC,	m_nvar2	},
		{	TOP_LDNL,	m_nvar3	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst0[]	=
	{
		{	TOP_LDL,	r_case3	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	ADC <s1>\nLDNL <s2>	==>	LDNL $<s1>+(<s2>*"wordsize")
 */
PRIVATE	struct	subp	src1[]	=
	{
		{	TOP_ADC,	m_nvar1	},
		{	TOP_LDNL,	m_nvar2	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst1[]	=
	{
		{	TOP_LDNL,	r_case1	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	EQC <s1>\nEQC 0\nEQC 0	==>	EQC <s1>
 */
PRIVATE	struct	subp	src2[]	=
	{
		{	TOP_EQC,	m_var1	},
		{	TOP_EQC,	m_zero	},
		{	TOP_EQC,	m_zero	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst2[]	=
	{
		{	TOP_EQC,	r_var1	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	ADC <s1>\nADC <s2>	==>	ADC <s1>+<s2>
 */
PRIVATE	struct	subp	src3[]	=
	{
		{	TOP_ADC,	m_nvar1	},
		{	TOP_ADC,	m_nvar2	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst3[]	=
	{
		{	TOP_ADC,	r_case2	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	ADC <s1>\nSTNL <s2>	==>	STNL $<s1>+(<s2>*"wordsize")
 */
PRIVATE	struct	subp	src4[]	=
	{
		{	TOP_ADC,	m_nvar1	},
		{	TOP_STNL,	m_nvar2	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst4[]	=
	{
		{	TOP_STNL,	r_case1	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	AJW <s1>\n.VAL <s3>\n.VAL <s4>\n.VAL <s5>\nAJW <s2>	==>
 *	.VAL <s3>\n.VAL <s4>\n.VAL <s5>\n.AJW <s1>+<s2>
 */
PRIVATE	struct	subp	src5[]	=
	{
		{	TOP_AJW,	m_nvar1	},
		{	TOP_VAL,	m_var3	},
		{	TOP_VAL,	m_var4	},
		{	TOP_VAL,	m_var5	},
		{	TOP_AJW,	m_nvar2	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst5[]	=
	{
		{	TOP_VAL,	r_var3	},
		{	TOP_VAL,	r_var4	},
		{	TOP_VAL,	r_var5	},
		{	TOP_AJW,	r_case2	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	AJW <s1>\n.RETF <s2>	==>	.RETF <s1>+<s2>
 */
PRIVATE	struct	subp	src6[]	=
	{
		{	TOP_AJW,	m_nvar1	},
		{	TOP_RETF,	m_nvar2	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst6[]	=
	{
		{	TOP_RETF,	r_case2	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	STL <s1>\n<BETWEEN>\nLDL <s1>	==>	DUP\nSTL <s1>
 */
PRIVATE	struct	subp	t8src7[]	=
	{
		{	TOP_STL,	m_var1	},
		{	TOP_BETWEEN,	m_blank	},
		{	TOP_LDL,	m_var1	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	t8dst7[]	=
	{
		{	TOP_DUP,	r_blank	},
		{	TOP_STL,	r_var1	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	LDC <s1>\nLDC <s1>	==>	LDC <s1>\nDUP
 *	Only useful for <s1> a constant requiring at least one prefix byte.
 *	The assumption is made that if the constant value is symbolic it is
 *	EXTREMELY likely to require at least one prefix byte.
 */
PRIVATE	struct	subp	t8src8[]	=
	{
		{	TOP_LDC,	m_svar1	},
		{	TOP_LDC,	m_svar1	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	t8dst8[]	=
	{
		{	TOP_LDC,	r_var1	},
		{	TOP_DUP,	r_blank	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	LDC <s1>\nDUP\nLDC <s1>	==>	LDC <s1>\nDUP\nDUP
 *	Only useful for <s1> a constant requiring at least one prefix byte.
 *	The assumption is made that if the constant value is symbolic it is
 *	EXTREMELY likely to require at least one prefix byte.
 */
PRIVATE	struct	subp	t8src9[]	=
	{
		{	TOP_LDC,	m_svar1	},
		{	TOP_DUP,	m_blank	},
		{	TOP_LDC,	m_svar1	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	t8dst9[]	=
	{
		{	TOP_LDC,	r_var1	},
		{	TOP_DUP,	r_blank	},
		{	TOP_DUP,	r_blank	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	LDLP <s1>\nLDNL <s2>	==>	LDL <s1> + <s2>
 */
PRIVATE	struct	subp	src10[]	=
	{
		{	TOP_LDLP,	m_nvar1	},
		{	TOP_LDNL,	m_nvar2	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst10[]	=
	{
		{	TOP_LDL,	r_case2	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	LDL <s1>\nLDL <s1>	==>	LDL <s1>\nDUP
 */
PRIVATE	struct	subp	t8src11[]	=
	{
		{	TOP_LDL,	m_var1	},
		{	TOP_LDL,	m_var1	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	t8dst11[]	=
	{
		{	TOP_LDL,	r_var1	},
		{	TOP_DUP,	r_blank	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	J @<l1>\n<l1>		==>	<l1>
 */
PRIVATE	struct	subp	src12[]	=
	{
		{	TOP_J,		m_jvar1	},
		{	TOP_LABEL,	m_var1	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst12[]	=
	{
		{	TOP_LABEL,	r_var1	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	FPLDNLDB\nFPADD		==>	FPLDNLADDDB
 */
PRIVATE	struct	subp	src13[]	=
	{
		{	TOP_FPLDNLDB,	m_blank	},
		{	TOP_FPADD,	m_blank	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst13[]	=
	{
		{	TOP_FPLDNLADDDB,r_blank	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	FPLDNLSN\nFPADD		==>	FPLDNLADDSN
 */
PRIVATE	struct	subp	src14[]	=
	{
		{	TOP_FPLDNLSN,	m_blank	},
		{	TOP_FPADD,	m_blank	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst14[]	=
	{
		{	TOP_FPLDNLADDSN,r_blank	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	FPLDNLDB\nFPMUL		==>	FPLDNLMULDB
 */
PRIVATE	struct	subp	src15[]	=
	{
		{	TOP_FPLDNLDB,	m_blank	},
		{	TOP_FPMUL,	m_blank	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst15[]	=
	{
		{	TOP_FPLDNLMULDB,r_blank	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	FPLDNLSN\nFPMUL		==>	FPLDNLMULSN
 */
PRIVATE	struct	subp	src16[]	=
	{
		{	TOP_FPLDNLSN,	m_blank	},
		{	TOP_FPMUL,	m_blank	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst16[]	=
	{
		{	TOP_FPLDNLMULSN,r_blank	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	LDC <s1>\nLDC <s2>\nAND	==>	LDC <s1> & <s2>
 *
 *	Generated by bitfield "store side" code.
 */
PRIVATE	struct	subp	src17[]	=
	{
		{	TOP_LDC,	m_cvar1	},
		{	TOP_LDC,	m_cvar2	},
		{	TOP_AND,	m_blank	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst17[]	=
	{
		{	TOP_LDC,	r_case4	},
		{	0,		(BOOL (*)())	0	}
	};

/*
 *	LDC <s1>\nLDC <s2>\nSHL	==>	LDC <s1> << <s2>
 *
 *	Generated by bitfield "store side" code.
 */
PRIVATE	struct	subp	src18[]	=
	{
		{	TOP_LDC,	m_cvar1	},
		{	TOP_LDC,	m_cvar2	},
		{	TOP_SHL,	m_blank	},
		{	0,		(BOOL (*)())	0	}
	};
PRIVATE	struct	subp	dst18[]	=
	{
		{	TOP_LDC,	r_case5	},
		{	0,		(BOOL (*)())	0	}
	};

PRIVATE	struct	pattern	Opts[]	=
	{
		{	src0,	dst0,	0,	(int) HAS_ANY	},
		{	src1,	dst1,	0,	(int) HAS_ANY	},
		{	src2,	dst2,	0,	(int) HAS_ANY	},
		{	src3,	dst3,	0,	(int) HAS_ANY	},
		{	src4,	dst4,	0,	(int) HAS_ANY	},
		{	src5,	dst5,	0,	(int) HAS_ANY	},
		{	src6,	dst6,	0,	(int) HAS_ANY	},
		{	t8src7,	t8dst7,	0,	(int) HAS_DUP	},
		{	t8src8,	t8dst8,	0,	(int) HAS_DUP	},
		{	t8src9,	t8dst9,	0,	(int) HAS_DUP	},
		{	src10,	dst10,	0,	(int) HAS_ANY	},
		{	t8src11,t8dst11,0,	(int) HAS_DUP	},
		{	src12,	dst12,	0,	(int) HAS_ANY	},
		{	src13,	dst13,	0,	(int) HAS_ANY	},
		{	src14,	dst14,	0,	(int) HAS_ANY	},
		{	src15,	dst15,	0,	(int) HAS_ANY	},
		{	src16,	dst16,	0,	(int) HAS_ANY	},
		{	src17,	dst17,	0,	(int) HAS_ANY	},
		{	src18,	dst18,	0,	(int) HAS_ANY	},
		{	NULL,	NULL,	0,	0	}
	};

/************************************************************************/
/*									*/
/*		Main entry points for the optimizer.			*/
/*									*/
/************************************************************************/

/*
 *	This routine clears out the peep-hole window.  This means that any text
 *	to this point cannot be combined with anything after.
 *
 *	The other typical use for this routine is to clear the window upon end
 *	of input file to flush everything out before closing the output file.
 */
PUBLIC	VOID
p_flush()
	{
	while(P_lnum > 0)
		p_optimize();
	}

/*
 *	Set up the peep-hole optimizer.
 */
PUBLIC	VOID
p_init()
	{
	register	int	i;
	register	int	j;
	register	int	k;
/*
 *	If the CPU for which code is being generated doesn't support floating
 *	point directly there is no point in doing the FP concurrency
 *	optimization stuff.
 */
	if(CPU_NO_FPU)
		FP_reorder = FALSE;
/*
 *	Go through the "Opcodes" table and adjust the appropriate instruction
 *	timings if we have a 16 bit Transputer.
 */
	if(Cpu_type & (T212 | T2EX))
		{
		for(i = 0; Opcodes[i].name != NULL; i++)
			{
			if(Opcodes[i].info & I_WORDTIMING)
				{
				Opcodes[i].info =
/* Preserve flags */			(Opcodes[i].info & I_BITMASK) |
/* New estimate for timing */		((Opcodes[i].info & (~I_BITMASK)) / 2);
				}
			}
		}
/*
 *	Allocate storage for peephole window buffers.
 */
	for(i = 0; i < WINDOW_SIZE; i++)
		{
		if((Lines[i].operand = malloc(LINESIZE)) == NULL)
			mem_err();
		if((Lines[i].comment = malloc(NAMEMAX + 1)) == NULL)
			mem_err();
		}
/*
 *	Allocate storage for peephole window registers.
 */
	for(i = 0; i < WINDOW_VARS; i++)
		{
		if((Vars[i].contents = malloc(LINESIZE)) == NULL)
			mem_err();
		}
/*
 *	Initialize the "Opcodes" table in regard to which opcodes can start a
 *	optimization pattern.  Also verify that the patterns match the
 *	configuration of the rest of the system.
 */
	for(i = 0; Opts[i].src != NULL; i++)
		{
		Opcodes[Opts[i].src[0].opcode].info |= I_PEEPABLE;
		for(j = 0; Opts[i].src[j].operand != NULL; j++)
			;
		for(k = 0; Opts[i].dst[k].operand != NULL; k++)
			;
		if((j > WINDOW_SIZE) || (k > WINDOW_SIZE))
			fatal("Compiler error: p_init#1");
		if(j < k)
			fatal("Compiler error: p_init#2");
		}
	}

/*
 *	Run the specified code through the peephole optimizer and write the
 *	result to the output file.
 */
PUBLIC	VOID
p_out(buf,count)
	char			*buf;
	int			count;
	{
/*
 *	Fill the peephole buffer.
 */
	for(;;)
		{
		while(P_lnum < WINDOW_SIZE)
			{
			if(p_line(((P_lcur + P_lnum) % WINDOW_SIZE),&buf,
				&count))
				{
				return;
				}
			P_lnum++;	/* Another line in the buffer */
			}
/*
 *	Perform the optimization.
 */
		p_optimize();
		}
	}

/*
 *	Print a summary of statistics.
 */
PUBLIC	VOID
p_summary()
	{
	register	int	i;

	if(! Debug)
		return;
/*
 *	Tell user which optimizations were triggered.
 */
	for(i = 0; Opts[i].src != NULL; i++)
		{
		fprintf(STDERR,"Optimization #%d fired %d times\n",i,
			Opts[i].count);
		}
	fprintf(STDERR,"%d FPU concurrency reorderings completed\n",FP_rcnt);
	}

/************************************************************************/
/*									*/
/*	Utility routines used only from within this file.		*/
/*									*/
/************************************************************************/

/*
 *	Parses a decimal constant string and returns a pointer to the
 *	the terminating string position.  The value is returned via a parameter
 *	pointer.
 */
PRIVATE	char	*
p_constant(str,value)
	char			*str;
	INT32			*value;
	{
	register	int	c;
	BOOL			negative;
	INT32			result;

	c = *str;
	negative = (c == '-');
	if((c == '+') || (c == '-'))
		++str;			/* Eat up the leading sign */
	result = 0;
	while(isdigit(*str))
		result = (result * 10) + (*str++ - '0');
	*value = negative ? (-result) : result;
	return (str);
	}

/*
 *	Given the text from a diversion, convert it into separate lines and
 *	return one per call.  Returns TRUE if more data is needed to complete a
 *	line.
 */

#define	PL_OPCODE	0		/* State value for finding opcode  */
#define	PL_OPERAND	1		/* State value for finding operand */
#define	PL_COMMENT	2		/* State value for finding comment */

PRIVATE	BOOL
p_line(line,buf,count)
	int			line;
	char			**buf;
	int			*count;
	{
	static			state	=	PL_OPCODE;
	static	char		*bufptr	=	NULL;

	switch((int) state)
		{
		case	PL_OPCODE:
				if(*count <= 0)
					return (TRUE);	/* Need more */
				Lines[line].opcode = *(*buf)++;
				(*count)--;
				state = PL_OPERAND;
				bufptr = Lines[line].operand;
				/* FALL THROUGH */
		case	PL_OPERAND:
				while(*count > 0)
					{
					(*count)--;
					if((*bufptr++ = *(*buf)++) == '\0')
						{
						state = PL_COMMENT;
						bufptr = Lines[line].comment;
						break;
						}
					}
				if(state != PL_COMMENT)
					return (TRUE);	/* Need more */
		case	PL_COMMENT:
				while(*count > 0)
					{
					(*count)--;
					if((*bufptr++ = *(*buf)++) == '\0')
						{
						state = PL_OPCODE;
						return (FALSE);
						}
					}
				return (TRUE);	/* Need more */
		}
	return (FALSE);			/* Needed only for lint... */
	}

/*
 *	Given a set of one or more lines in the optimizing window, perform
 *	optimizations until at least one less line is in the window. If it
 *	isn't possible to optimize then simply write a line to the output
 *	file.
 */
PRIVATE	VOID
p_optimize()
	{
	register	int	i;
	register	int	j;
	int			k;
	register	int	ltmp1;
	register		ltmp2;
	int			misc;
	char			*smisc;

	for(;;)
		{
/*
 *	Assuming peephole optimization is enabled do so.  Also verify that
 *	opcode at start of pattern belongs to a legal pattern!
 */
		if(Peepopt && (Opcodes[Lines[P_lcur].opcode].info &I_PEEPABLE))
			{
			for(i = 0; Opts[i].src != NULL; i++)
				{
				if(! (Cpu_type & Opts[i].cpu_type))
					continue;
				for(j = 0; j < WINDOW_VARS; j++)
					Vars[j].in_use = FALSE;
				for(j = 0; Opts[i].src[j].operand != NULL; j++)
					{
/*
 *	Test for enough lines in the window.
 */
					if(j >= P_lnum)
						break;
					ltmp1 = (P_lcur + j) % WINDOW_SIZE;
/*
 *	Test for matching opcode.
 */
					if(Opts[i].src[j].opcode !=
						Lines[ltmp1].opcode)
						{
						break;
						}
/*
 *	Test operand predicate.
 */
					if(! (*Opts[i].src[j].operand)(ltmp1))
						break;
					}
				if(Opts[i].src[j].operand != NULL)
					continue;	/* Chk next pattern */
/*
 *	We have a pattern match, replace the pattern.  If the new pattern has
 *	fewer lines we just return, otherwise we see if the replacement
 *	pattern could itself be part of a pattern and rescan the pattern list,
 *	or quit scanning the list if the "new" first opcode isn't part of any
 *	possible pattern.
 */
				if(p_replace(i,j) != 0)
					return;	/* Replacement was shorter */
				if(Opcodes[Lines[P_lcur].opcode].info &
					I_PEEPABLE)
					{
					i = 0;	/* Reset pattern counter */
					}
				else
					break;
				}
			}
/*
 *	If we are doing FPU concurrency reorganization, here is where the
 *	magic happens!  The basic rules in deciding whether to do this:
 *
 *	1. The first instruction in the buffer must have the "I_FPOP1ROLE"
 *	bit set in the "Opcodes" "info" field.
 *
 *	2. There must be at least three instructions in the peephole buffer.
 *
 *	3. The second instruction in the buffer must have the "I_FPOP2ROLE"
 *	bit set in the "Opcodes" "info" field.
 *
 *	4. The third instruction in the buffer must have the "I_OP3ROLE" bit
 *	set in the "Opcodes" "info" field.
 *
 *	5. The estimated execution time for the first instruction must exceed
 *	the second instruction OR the estimated execution time of the first
 *	instruction must be longer than the third instruction.
 *
 *	6. We have to be allowed to perform this optimization.
 *
 *	Given that we decide to perform the rearrangements, we then need to
 *	find a contiguous sequence of "I_OP3ROLE" instructions whose total
 *	estimated execution time is maximized while remaining less than the
 *	time taken by the first instruction.
 */
		if(((i = Opcodes[Lines[P_lcur].opcode].info)
			& I_FPOP1ROLE) &&
		(P_lnum >= 3) &&
		((j=Opcodes[Lines[ltmp1=(P_lcur+1)%WINDOW_SIZE].opcode].info)
			& I_FPOP2ROLE) &&
		((Opcodes[Lines[ltmp2=((ltmp1+1)%WINDOW_SIZE)].opcode].info)
			& I_OP3ROLE) &&
		(((i &= (~I_BITMASK)) > (j & (~I_BITMASK))) || (i > (int)
			(Opcodes[Lines[ltmp2].opcode].info&(~I_BITMASK)))) &&
		FP_reorder)
			{
			FP_rcnt++;	/* Keep track of # of reorderings */
/*
 *	Now we run through and figure out how many instructions (in addition to
 *	the third), we will be moving between the first two.  The first
 *	instruction can successfully "shadow" the # of machine cycles initially
 *	held in variable "i".
 */
			i -= (Opcodes[Lines[ltmp2].opcode].info&(~I_BITMASK));
			for(k = 3; k <= P_lnum; k++)
				{
/*
 *	Swap what was originally the second instruction with the current
 *	instruction (which has been determined to be of type "I_OP3ROLE").
 */
				misc = Lines[ltmp1].opcode;
				Lines[ltmp1].opcode = Lines[ltmp2].opcode;
				Lines[ltmp2].opcode = misc;
				smisc = Lines[ltmp1].operand;
				Lines[ltmp1].operand = Lines[ltmp2].operand;
				Lines[ltmp2].operand = smisc;
				smisc = Lines[ltmp1].comment;
				Lines[ltmp1].comment = Lines[ltmp2].comment;
				Lines[ltmp2].comment = smisc;
/*
 *	Okay, now figure out what the next instruction is, and whether it is
 *	of the appropriate type and execution time.
 */
				ltmp2 = ((ltmp1 = ltmp2) + 1) % WINDOW_SIZE;
				if(!((j = Opcodes[Lines[ltmp2].opcode].info) &
						I_OP3ROLE))
					break;
				if((i -= (j & (~I_BITMASK))) <= 0)
					break;
				}
			}
/*
 *	Time to write the first line out if no (more) optimizations are
 *	possible.
 */
		p_write();
		P_lcur = ((P_lcur + 1) % WINDOW_SIZE);
		P_lnum--;		/* One less line */
		return;			/* Wrote our one required line */
		}
	}

/*
 *	Substitute a replacement pattern for a pattern being matched and update
 *	the "P_lnum" and "P_lcur" global variables.  Passed the pattern # which
 *	was matched and the # of lines in that pattern.  Returns the # of lines
 *	which the replacement removed from the peephole window.
 */
PRIVATE	int
p_replace(pnum,lnum)
	int			pnum;
	int			lnum;
	{
	register	int	j;
	register	int	ltemp;
	register	int	rval;
/*
 *	Increment "fired" count in rule.
 */
	Opts[pnum].count++;		/* We've got a live one */
/*
 *	Figure out how many lines in replacement pattern.
 */
	for(j = 0; Opts[pnum].dst[j].operand != NULL; j++)
		;
/*
 *	Compute return value, new "first" line in window and # of lines
 *	remaining in window.
 */
	rval = lnum - j;		/* # of lines removed */
	P_lcur = (P_lcur + rval) % WINDOW_SIZE;	/* New "first" line # */
	P_lnum -= rval;			/* New # of lines remaining total */
/*
 *	Substitute replacement string into window.
 */
	for(j = 0; Opts[pnum].dst[j].operand != NULL; j++)
		{
		ltemp = (P_lcur + j) % WINDOW_SIZE;
		Lines[ltemp].opcode = Opts[pnum].dst[j].opcode;	/* Opcode */
		(VOID) (*Opts[pnum].dst[j].operand)(ltemp);	/* Operand */
		}
	return (rval);			/* All done */
	}

/*
 *	Write the first line in the peephole window to the output file.
 */
PRIVATE	VOID
p_write()
	{
	static	char		err[]	=	"Error writing to output file";
	register	int	opcode;

	if((opcode = Lines[P_lcur].opcode) <= TOP_SPECIAL_MAX)
		{
/*
 *	Special processing of one form or another required on output.
 */
		if(opcode == TOP_BETWEEN)
			{
			}		/* A placeholder in output stream */
		else
		if(opcode == TOP_COMMENT)
			{
			if(fputs(Lines[P_lcur].comment,Coutput) == EOF)
				fatal(err);
			fputc('\n',Coutput);
			}
		else			/* Must be TOP_LABEL */
			{
			if(fputs(Lines[P_lcur].operand,Coutput) == EOF)
				fatal(err);
			fputc('\n',Coutput);
			}
		}
	else
		{
/*
 *	Normal operand processing.
 */
		fputc('\t',Coutput);
		if(fputs(Opcodes[opcode].name,Coutput) == EOF)
			fatal(err);
		if(Lines[P_lcur].operand[0] != '\0')
			{
			fputc('\t',Coutput);
			if(fputs(Lines[P_lcur].operand,Coutput) == EOF)
				fatal(err);
			}
		if(Lines[P_lcur].comment[0] != '\0')
			{
			fputc('\t',Coutput);
			fputc(';',Coutput);
			if(fputs(Lines[P_lcur].comment,Coutput) == EOF)
				fatal(err);
			}
		fputc('\n',Coutput);
		}
	}

/************************************************************************/
/*									*/
/*	Pattern matching predicates and replacement routines.		*/
/*									*/
/************************************************************************/

/*
 *	Return TRUE if the operand of the specified line is "blank".
 */
PRIVATE	BOOL
m_blank(line)
	int			line;
	{
	return (Lines[line].operand[0] == '\0');
	}

/*
 *	Returns TRUE the first time this function is called for variable #1,
 *	if the variable is a numeric constant.  It returns TRUE on subsequent
 *	calls if the value being matched is the same as present on the
 *	previous calls.
 */
PRIVATE	BOOL
m_cvar1(line)
	int			line;
	{
	INT32			value;

	if(*p_constant(Lines[line].operand,&value) != '\0')
		return (FALSE);

	return (m_varx(line,0));
	}

/*
 *	Returns TRUE the first time this function is called for variable #2,
 *	if the variable is a numeric constant.  It returns TRUE on subsequent
 *	calls if the value being matched is the same as present on the
 *	previous calls.
 */
PRIVATE	BOOL
m_cvar2(line)
	int			line;
	{
	INT32			value;

	if(*p_constant(Lines[line].operand,&value) != '\0')
		return (FALSE);

	return (m_varx(line,1));
	}

/*
 *	Returns TRUE the first time this function is called for variable #1,
 *	and on subsequent calls if the operand contents being matched are
 *	exactly the same as present on the previous calls.  Note that this
 *	predicate EXPLICITLY tests for the presence of the "@" assembly
 *	language "relative" operand as would be used in a "j" or "cj"
 *	instruction.  Note that the "@" is NOT copied to the save register!
 */
PRIVATE	BOOL
m_jvar1(line)
	int			line;
	{
	if(Lines[line].operand[0] != '@')	/* Not relative format */
		return (FALSE);
	if(Vars[0].in_use)		/* Already something in variable */
		{
		return (! strcmp(Vars[0].contents,&Lines[line].operand[1]));
		}
	Vars[0].in_use = TRUE;
	strcpy(Vars[0].contents,&Lines[line].operand[1]);
	return (TRUE);			/* Always TRUE first time through */
	}

/*
 *	Returns TRUE the first time this function is called for variable #1,
 *	UNLESS the variable starts with a "$" (wordsize) modifier.  It returns
 *	TRUE on subsequent calls if the operand contents being matched are
 *	exactly the same as present on the previous calls.
 */
PRIVATE	BOOL
m_nvar1(line)
	int			line;
	{
	if(*Lines[line].operand == '$')
		return (FALSE);

	return (m_varx(line,0));
	}

/*
 *	Returns TRUE the first time this function is called for variable #2,
 *	UNLESS the variable starts with a "$" (wordsize) modifier.  It returns
 *	TRUE on subsequent calls if the operand contents being matched are
 *	exactly the same as present on the previous calls.
 */
PRIVATE	BOOL
m_nvar2(line)
	int			line;
	{
	if(*Lines[line].operand == '$')
		return (FALSE);

	return (m_varx(line,1));
	}

/*
 *	Returns TRUE the first time this function is called for variable #3,
 *	UNLESS the variable starts with a "$" (wordsize) modifier.  It returns
 *	TRUE on subsequent calls if the operand contents being matched are
 *	exactly the same as present on the previous calls.
 */
PRIVATE	BOOL
m_nvar3(line)
	int			line;
	{
	if(*Lines[line].operand == '$')
		return (FALSE);

	return (m_varx(line,2));
	}

/*
 *	Returns TRUE the first time this function is called for variable #1 if
 *	the operand is likely to require at least one prefix byte to represent
 *	(ie. not a constant value between 0 and 15).  Return TRUE on
 *	subsequent calls if the operand contents being matched are exactly the
 *	same as present on the previous calls.
 */
PRIVATE	BOOL
m_svar1(line)
	int			line;
	{
	register	int	c;
	INT32			value;

	if(Vars[0].in_use)		/* Already something in variable */
		{
		return (! strcmp(Vars[0].contents,Lines[line].operand));
		}
	c = Lines[line].operand[0];	/* Get first character */
	if((c == '-') || (c == '+') || isdigit(c))	/* Numeric constant? */
		{
		if((*p_constant(Lines[line].operand,&value) == '\0') &&
			((value >= 0L) && (value <= 15L)))
				return (FALSE);	/* Constant too small */
		}
	Vars[0].in_use = TRUE;
	strcpy(Vars[0].contents,Lines[line].operand);
	return (TRUE);			/* Always TRUE first time through */
	}

/*
 *	Returns TRUE the first time this function is called for variable #1,
 *	and on subsequent calls if the operand contents being matched are
 *	exactly the same as present on the previous calls.
 */
PRIVATE	BOOL
m_var1(line)
	int			line;
	{
	return (m_varx(line,0));
	}

/*
 *	Returns TRUE the first time this function is called for variable #3,
 *	and on subsequent calls if the operand contents being matched are
 *	exactly the same as present on the previous calls.
 */
PRIVATE	BOOL
m_var3(line)
	int			line;
	{
	return (m_varx(line,2));
	}

/*
 *	Returns TRUE the first time this function is called for variable #4,
 *	and on subsequent calls if the operand contents being matched are
 *	exactly the same as present on the previous calls.
 */
PRIVATE	BOOL
m_var4(line)
	int			line;
	{
	return (m_varx(line,3));
	}

/*
 *	Returns TRUE the first time this function is called for variable #5,
 *	and on subsequent calls if the operand contents being matched are
 *	exactly the same as present on the previous calls.
 */
PRIVATE	BOOL
m_var5(line)
	int			line;
	{
	return (m_varx(line,4));
	}

/*
 *	Helper function for "m_var1()", "m_var2()", ...
 */
PRIVATE	BOOL
m_varx(line,vnum)
	int			line;
	int			vnum;
	{
	if(Vars[vnum].in_use)		/* Already something in variable */
		return (! strcmp(Vars[vnum].contents,Lines[line].operand));
	Vars[vnum].in_use = TRUE;
	strcpy(Vars[vnum].contents,Lines[line].operand);
	return (TRUE);			/* Always TRUE first time through */
	}

/*
 *	Returns TRUE if the operand of the specified line is "0".
 */
PRIVATE	BOOL
m_zero(line)
	int			line;
	{
	return ((Lines[line].operand[0] == '0') &&
		(Lines[line].operand[1] == '\0'));
	}

/*
 *	"blanks" out the specified operand field.
 */
PRIVATE	BOOL
r_blank(line)
	int			line;
	{
	Lines[line].operand[0] = '\0';
	return (FALSE);			/* Ignored anyway */
	}

/*
 *	Replace the specified operand field with the contents of variable #1
 *	and variable #2 as: $v1+((v2)*"wordsize")
 */
PRIVATE	BOOL
r_case1(line)
	int			line;
	{
	register	char	*str;
	INT32			value1;
	INT32			value2;
/*
 *	See if both operands are constants, if so do reduction to a single
 *	constant so that the resulting operand expression may take part in
 *	further peephole template matching.
 */
	if((*p_constant(Vars[0].contents,&value1) == '\0') &&
		(*p_constant(Vars[1].contents,&value2) == '\0'))
		{
		sprintf(Lines[line].operand,"%ld",value1/Sizeof_int + value2);
		}
	else
		{
		str = Lines[line].operand;
		strcpy(str,"$");
		strcat(str,Vars[0].contents);
		strcat(str,"+((");
		strcat(str,Vars[1].contents);
		if(Sizeof_int == 2)
			strcat(str,")*2)");
		else
			strcat(str,")*4)");
		}
	return (FALSE);			/* Ignored anyway */
	}

/*
 *	Replace the specified operand field with the contents of variable #1
 *	and variable #2 as: v1+v2
 */
PRIVATE	BOOL
r_case2(line)
	int			line;
	{
	register	char	*str;
	INT32			value1;
	INT32			value2;
/*
 *	See if both operands are constants, if so do reduction to a single
 *	constant so that the resulting operand expression may take part in
 *	further peephole template matching.
 */
	if((*p_constant(Vars[0].contents,&value1) == '\0') &&
		(*p_constant(Vars[1].contents,&value2) == '\0'))
		{
		sprintf(Lines[line].operand,"%ld",value1 + value2);
		}
	else
		{
		str = Lines[line].operand;
		strcpy(str,Vars[0].contents);
		strcat(str,"+");
		strcat(str,Vars[1].contents);
		}
	return (FALSE);			/* Ignored anyway */
	}

/*
 *	Replace the specified operand field with the contents of variable #1,
 *	variable #2 and variable #3 as: v1+((v2)/"wordsize")+v3
 */
PRIVATE	BOOL
r_case3(line)
	int			line;
	{
	register	char	*str;

	str = Lines[line].operand;

	strcpy(str,Vars[0].contents);
	strcat(str,"+((");
	strcat(str,Vars[1].contents);
	if(Sizeof_int == 2)
		strcat(str,")/2)+");
	else
		strcat(str,")/4)+");
	strcat(str,Vars[2].contents);

	return (FALSE);			/* Ignored anyway */
	}

/*
 *	Replace the specified operand field with the contents of variable #1
 *	and variable #2 as the value of: (v1)&(v2)
 */
PRIVATE	BOOL
r_case4(line)
	int			line;
	{
	INT32			value1;
	INT32			value2;
/*
 *	Since both operands are constants, do reduction to a single constant so
 *	that the resulting operand expression may take part in further peephole
 *	template matching.
 */
	if((*p_constant(Vars[0].contents,&value1) == '\0') &&
		(*p_constant(Vars[1].contents,&value2) == '\0'))
		{
		sprintf(Lines[line].operand,"%ld",value1 & value2);
		}
	else
		fatal("Compiler error: r_case4");

	return (FALSE);			/* Ignored anyway */
	}

/*
 *	Replace the specified operand field with the contents of variable #1
 *	and variable #2 as the value of: (v1)<<(v2)
 */
PRIVATE	BOOL
r_case5(line)
	int			line;
	{
	INT32			value1;
	INT32			value2;
/*
 *	Since both operands are constants, do reduction to a single constant so
 *	that the resulting operand expression may take part in further peephole
 *	template matching.
 */
	if((*p_constant(Vars[0].contents,&value1) == '\0') &&
		(*p_constant(Vars[1].contents,&value2) == '\0'))
		{
		sprintf(Lines[line].operand,"%ld",value1 << value2);
		}
	else
		fatal("Compiler error: r_case5");

	return (FALSE);			/* Ignored anyway */
	}

/*
 *	Replaces the specified operand field with the contents of variable #1.
 */
PRIVATE	BOOL
r_var1(line)
	int			line;
	{
	strcpy(Lines[line].operand,Vars[0].contents);
	return (FALSE);			/* Ignored anyway */
	}

/*
 *	Replaces the specified operand field with the contents of variable #3.
 */
PRIVATE	BOOL
r_var3(line)
	int			line;
	{
	strcpy(Lines[line].operand,Vars[2].contents);
	return (FALSE);			/* Ignored anyway */
	}

/*
 *	Replaces the specified operand field with the contents of variable #4.
 */
PRIVATE	BOOL
r_var4(line)
	int			line;
	{
	strcpy(Lines[line].operand,Vars[3].contents);
	return (FALSE);			/* Ignored anyway */
	}

/*
 *	Replaces the specified operand field with the contents of variable #5.
 */
PRIVATE	BOOL
r_var5(line)
	int			line;
	{
	strcpy(Lines[line].operand,Vars[4].contents);
	return (FALSE);			/* Ignored anyway */
	}

/* End of TCX7.C */
