/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: tcx9.c,v 2.2 92/04/21 11:11:21 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	TCX Compiler (9 of 11)
 *
 *	Copyright 1987-1991 by Logical Systems.
 *	All rights reserved.
 *
 *	Published by:	Logical Systems
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	03/17/87
 *
 *	This file implements text diversions.  As code is generated, it
 *	is stuffed into a diversion, not written to the file.  These diversions
 *	can then be reordered, deleted, or combined with other diversions
 *	before being passed to a higher level of software, which can also
 *	munge the diversion.  When a function (or other zero level item)
 *	is fully compiled, the diversion is written to the disk.
 *
 *	When the user of a diversion has a pointer to a diversion, he has
 *	a pointer to the last block in the linked list of diversion blocks.
 *	The global variable "current" likewise points to the last of a list.
 *	This makes it easy to append new diversions onto the diversion.  To
 *	get to the start of the diversion, this code simply uses the "first"
 *	pointer.
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<string.h>
#include	"tcx.h"
#include	"tcxdecl.h"
#include	"tcxext.h"

/*
 *	Local function declarations.
 */
PRIVATE	struct	dvrt		*dvrt_init();
PRIVATE	VOID			fill_cache();
PRIVATE	VOID			write_cache();

/*
 *	Local data definitions.
 */
#ifdef	SEEK_SET
#define FBEGIN	SEEK_SET
#else
#define FBEGIN	0
#endif

PRIVATE char	cache[DVRTBLKSIZE];	/* Buffer for current dvrt	*/
PRIVATE unsigned	cache_blk = 0;	/* Current block # in cache	*/
PRIVATE BOOL	cache_dirty = FALSE;	/* TRUE/cache needs writing	*/
PRIVATE	struct	dvrt	*free_dvrt_list = NULL;	/* Free list/old dvrts	*/
PRIVATE	struct	dvrt	*current = NULL;/* The current diversion	*/
PRIVATE	struct	dvrt	*stack_top = NULL;	/* Top of dvrt stack	*/
PRIVATE	unsigned	max_blk_num = 0;/* Largest block # assigned	*/

#ifdef	DEBUG
PUBLIC	int		Alloc_dvrt = 0;	/* Number of allocated dvrts	*/
PUBLIC	long		Dvrttotal = 0;	/* Total bytes sent to dvrt	*/
PUBLIC	int		Dvrtcount = 0;	/* # of diversion written	*/
PUBLIC	int		Nfree_dvrt = 0;	/* Number currently free	*/
#endif	/* DEBUG */

/*
 *	Append a diversion onto the current diversion.
 *
 *	Given a diversion list pointer, this function manipulates the list
 *	pointers to append the given diversion to the "current" diversion.
 */
PUBLIC	VOID
dvrt_append(diversion)
	struct	dvrt		*diversion;
	{
	Newfile = TRUE;		/* Force #line in diversion */
	if(diversion != NULL)
		{
		if(current != NULL)
			{
/* Append to current */	diversion->first->prev = current;
			current->next = diversion->first;
/* Set head pointer */	diversion->first = current->first;
			}
		current = diversion;	/* Set tail pointer */
		}
	}

/*
 *	Copy a diversion into the current diversion.
 *
 *	Same syntax as dvrt_append, but duplicates the information that is
 *	appended.
 */
PUBLIC	VOID
dvrt_copy(diversion)
	struct	dvrt		*diversion;
	{
	struct	dvrt		*dp;
	struct	dvrt		*new;

	for(dp = diversion->first; dp != NULL; dp = dp->next)
		{
		new = dvrt_init();	/* Get a free diversion */
		new->count = dp->count;	/* Length */
		new->literal = dp->literal;	/* Literal status flag */
		if(Inter_file)
			{
			if(dp->text.blk_num != cache_blk)
				fill_cache(dp);
			else if(cache_dirty)
				write_cache(dp->text.blk_num);

			write_cache(new->text.blk_num);
			}
		else		/* We are not using an intermediate file */
			{
			memcpy(new->text.buf,dp->text.buf,
				(unsigned) dp->count);	/* Copy data */
			}
		dvrt_append(new);	/* Append to current diversion */
		}
	}

/*
 *	Put the specified diversion list buffers back in the free list.
 *
 *	All buffers which the user gets handles to via "dvrt_unstack" and
 *	"dvrt_get" should be disposed of in this fashion.  Note that this
 *	function processes entire diversion lists (THE INTERNAL POINTERS HAD
 *	BETTER NOT BEEN MUCKED WITH BY WHOEVER GOT THE DIVERSION HANDLE).
 */
PUBLIC	VOID
dvrt_free(diversion)
	struct	dvrt		*diversion;
	{
	if(diversion == NULL)
		return;			/* Empty list */
#ifdef	DEBUG
		{
		struct	dvrt	*dp;

		for(++Nfree_dvrt,dp = diversion->first; dp != diversion;
			dp = dp->next)
			{
			++Nfree_dvrt;	/* One more free */
			}
		}
#endif	/* DEBUG */
	diversion->next = free_dvrt_list;	/* Link into list */
	free_dvrt_list = diversion->first;
	}

/*
 *	Return a handle to what was the current diversion (before
 *	"dvrt_get" was called).
 *
 *	A new current diversion is in effect after the previous current
 *	diversion is returned.
 */
PUBLIC	struct	dvrt	*
dvrt_get()
	{
	struct	dvrt		*rval;

	Newfile = TRUE;			/* Force #line in diversion */
	rval = current;
	current = NULL;

	return (rval);
	}

/*
 *	Get the memory for, and initialize, a diversion buffer.
 *
 *	First tries to get the memory from the diversion free list.  Failing
 *	that "malloc" is used to obtain a block of diversions.  The contents
 *	of the buffer are initialized as if it will be the only diversion
 *	buffer in the list (higher level routines may modify this assumption
 *	explicitly if desired).
 *
 *	If the global variable Inter_file is true then the an intermediate
 *	file is utilized.
 */
PRIVATE	struct	dvrt	*
dvrt_init()
	{
	struct	dvrt		*rval;
	int			count;

	if(free_dvrt_list == NULL)
		{
		unsigned	size;

		if(Inter_file)
			size = DVRT_ALLOC * sizeof(struct dvrt);
		else
			size = sizeof(struct dvrt) - 2 + DVRTBLKSIZE;

		if((rval = (struct dvrt *) malloc(size)) == NULL)
			mem_err();
#ifdef	DEBUG
		Alloc_dvrt += (Inter_file ? DVRT_ALLOC : 1);	/* Update # */
#endif	/* DEBUG */
/*
 *	Place the newly created diversion(s) on the free list and assign disk
 *	block number(s).
 */
		if(Inter_file)
			{
			if(fseek(Interfp,(long) max_blk_num * DVRTBLKSIZE,
				FBEGIN) != 0)
				{
				fatal("Can't seek in intermediate file");
				}
			for(count = 0; count < DVRT_ALLOC; count++)
				{
				if(fwrite(cache,DVRTBLKSIZE,1,Interfp)
					!= 1)
					{
				fatal("Can't write to intermediate file");
					}
/* Seq. assign blk #'s */	rval->text.blk_num = max_blk_num++;
				rval->next = free_dvrt_list;
				free_dvrt_list = rval++;
				}
#ifdef	DEBUG
			Nfree_dvrt += DVRT_ALLOC;
#endif	/* DEBUG */
			}
		else			/* Place one dvrt on the free list */
			{
			rval->next = free_dvrt_list;
			free_dvrt_list = rval;
#ifdef	DEBUG
			++Nfree_dvrt;	/* One more free */
#endif	/* DEBUG */
			}
		}	/* End of if (free_dvrt_list == NULL) */
/*
 * 	Remove a diversion from the free list.
 */
	rval = free_dvrt_list;		/* Head of free list */
	free_dvrt_list = rval->next;	/* Next item on list */
#ifdef	DEBUG
	--Nfree_dvrt;			/* One less free */
#endif	/* DEBUG */
/*
 *	Initialize contents of buffer as if it is the only one in a diversion
 *	list.
 */
	rval->first = rval;		/* Is both tail and head of list */
	rval->prev = rval->next = NULL;	/* Zap list pointers */
	rval->count = 0;		/* Nothing in text buffer yet */
	rval->literal = FALSE;		/* Does not contain literal code */

	return (rval);
	}

/*
 *	Write the specified diversion buffer to the peephole optimizer.
 */
PUBLIC	VOID
dvrt_out(diversion)
	struct	dvrt		*diversion;
	{
	register	int	count;
	struct	dvrt		*literal;	/* Place for literal dvrts */
	struct	dvrt		*next;
	register struct dvrt	*temp;

	if(diversion == NULL)
		return;			/* If nothing to write */

	literal = NULL;
/*
 *	kludge to keep from adding a flag to diversion element:
 */
	diversion->first->prev = diversion;	/* Set to NON NULL */

/*lint -e530 next really is ok */
	for(temp = diversion->first; temp != NULL; temp = next)
		{
		if(temp->count == -1)
			fatal("loop in diversion list");

		next = temp->next;	/* Save ptr to next block */

		if((count = temp->count) != 0)	/* If characters in buffer */
			{
			if(temp->literal)
				{
/*
 *	Bust this block loose from diversion list and hand off to the literal
 *	pool.
 */
				temp->literal = FALSE;	/* No longer literal */

				if(temp == diversion->first)
					{
					/* First block in list */
					if(next != NULL)
						diversion->first = next;
					else
/* Totally empty now */				diversion = NULL;
					}
				else if(temp == diversion)
					{
/*
 *	Last block in list -- need to change "diversion".
 */
					diversion = diversion->prev;
					}
				else
					{
/*
 *	Somewhere in the middle -- just unhook.
 */
					temp->prev->next = next;
					next->prev = temp->prev;
					}

/* Add to current diversion */	temp->prev = literal;
				temp->next = NULL;

				if(literal != NULL)
					{
/* Point to new */			literal->next = temp;
/* Set head pointer */			temp->first = literal->first;
					}
				else
					temp->first = temp;

				literal = temp;	/* Set tail pointer */
				}
			else		/* Not a literal */
				{
#ifdef	DEBUG
/* Bytes written in dvrt */	Dvrttotal += temp->count;
/* Add one more */		Dvrtcount++;
#endif	/* DEBUG */
				if(Inter_file)
					{
					if(temp->text.blk_num != cache_blk)
						fill_cache(temp);
					p_out(cache,count);
					}
				else
/* Run through peephole opt. */		p_out(temp->text.buf,count);

/* All done with this */	temp->count = -1;
				}
			}
/*lint +e530 */
		}
	dvrt_free(diversion);	/* Recycle diversion blocks in list */
	dvrt_out(literal);	/* Now do the literals */
	}

/*
 *	Stack the current diversion and create a new current diversion.
 *
 *	The "first" pointer of the latest diversion buffer in the list is used
 *	as a list pointer to create a list (stack) of diversion lists.  The
 *	misuse of the "first" pointer is hidden within "dvrt_stack" and
 *	"dvrt_unstack" and thus may be ignored by everyone else.
 */
PUBLIC	VOID
dvrt_stack()
	{
	Newfile = TRUE;			/* Force #line in diversion */
	if(current == NULL)
		current = dvrt_init();

	if(current->prev != NULL)
		current->prev->first = current->first;

	current->first = stack_top;	/* Link into stack */
	stack_top = current;
	current = NULL;			/* No "current" diversion for now */
	}

/*
 *	Return a pointer to the current diversion and restore the previous one.
 *
 *	The return value is the current diversion list.  The most recent victim
 *	of a "dvrt_stack" operation becomes the "current" diversion again.
 */
PUBLIC	struct	dvrt	*
dvrt_unstack()
	{
	struct	dvrt		*rval;

	Newfile = TRUE;			/* Force #line in diversion */
	if(stack_top == NULL)
		fatal("Diversion stack underflow");
	rval = current;
	current = stack_top;
	stack_top = current->first;	/* Update top of stack */
/*
 *	Restore correct contents of "first" pointer.
 */
	if(current->prev == NULL)
		current->first = current;
	else
		current->first = current->prev->first;

	return (rval);
	}

/*
 *	Write a character into the current diversion buffer.
 *
 *	If the current buffer is completely full, gets a new buffer and links
 *	it into the current diversion list.
 */
PUBLIC	VOID
dvrt_write(c)
	int			c;
	{
	struct	dvrt		*temp;
/*
 *	First see if we need to get a new buffer (or any buffer).
 */
	if((current == NULL) || (current->count >= DVRTBLKSIZE) ||
		(c == LIT_STREAM) || (c == CODE_STREAM))
		{
		temp = dvrt_init();	/* Get another buffer */

		if(c == LIT_STREAM)
			temp->literal = TRUE;

		if(current != NULL)
			{
			if((c != CODE_STREAM) && (c != LIT_STREAM))
				{
/*
 *	We know that temp->literal starts out as FALSE, so only set if the
 *	char is regular data.
 */
				temp->literal = current->literal;
				}
			temp->first = current->first;	/* Copy head pointer */
			temp->prev = current;	/* Link into list */
			current->next = temp;
			}

		current = temp;		/* Update tail pointer */

		if((c != LIT_STREAM) && (c != CODE_STREAM))
			{
			if(Inter_file)
				{
				if(current->text.blk_num != cache_blk)
					fill_cache(current);
				cache[current->count++] = c;
				cache_dirty = TRUE;
				}
			else	/* We are not using an intermediate file */
				current->text.buf[current->count++] = c;
			}
		}
	else
		{
/*
 *	This store is done in the ELSE clause to keep the "normal" operation
 *	of dvrt_write() more efficient.  Fewer tests mean faster operation.
 *	(Or a better optimizing compiler.)
 */

		if(Inter_file)
			{
			if(current->text.blk_num != cache_blk)
				fill_cache(current);
			cache[current->count++] = c;
			cache_dirty = TRUE;
			}
		else
			current->text.buf[current->count++] = c;
		}
	}

#ifdef	DVRT_DEBUG
/*
 *	Print out the text in the diversion to the screen, as well as
 *	diversion addresses.  This is used to debug twisted diversions.
 */
PUBLIC	VOID
dvrt_check(diversion)
	struct	dvrt		*diversion;
	{
	int			i;

	if(diversion == NULL)
		printf("Diversion is empty\n");

	for(diversion = diversion->first; diversion != NULL;
		diversion = diversion->next)
		{
		printf("Diversion %08p: \"",diversion);
		for(i = 0; i < diversion->count; i++)
			{
			if(Inter_file)
				{
				if(diversion->text.blk_num != cache_blk)
					fill_cache(diversion);
				putchar(cache[i]);
				}
			else
				putchar(diversion->text.buf[i]);
			}
		printf("\"\n");
		if(Inter_file)
			printf("(blk_num: %u)\n",diversion->text.blk_num);
		}
	}
#endif	/* DVRT_DEBUG */

/*
 *	Emit a CODE_STREAM byte into current diversion.
 */
PUBLIC	VOID
set_code()
	{
	dvrt_write(CODE_STREAM);
	}

/*
 *	Emit a LIT_STREAM byte into current diversion.
 */
PUBLIC	VOID
set_literal()
	{
	dvrt_write(LIT_STREAM);
	}

/*
 *	Read the contents of the disk block into the cache and set the global
 *	variable "cache_blk" accordingly.
 *
 *	Sets cache_dirty = FALSE.  Should never be called if
 *	Inter_file = FALSE.
 */
PRIVATE VOID
fill_cache(dvrt)
	struct	dvrt		*dvrt;
	{
/*
 *	See if anything to be written before we read in the "new" stuff.
 */
	if(cache_dirty)
		write_cache(cache_blk);
/*
 *	Read "new" diversion buffer contents back in.
 */
	if(dvrt->count > 0)
		{
		if(fseek(Interfp,(long) dvrt->text.blk_num * DVRTBLKSIZE,
			FBEGIN) != 0)
			{
			fatal("Can't seek in intermediate file");
			}
		if(fread(cache,1,(unsigned) dvrt->count,Interfp) !=
			dvrt->count)
			{
			fatal("Can't read from intermediate file");
			}
		}
	cache_blk = dvrt->text.blk_num;
	cache_dirty = FALSE;
	}

/*
 *	Write the contents of the cache to a disk block.
 *
 *	Sets cache_dirty = FALSE, cache_blk = blk_num.  Should never be called
 *	if Inter_file = FALSE.
 */
PRIVATE	VOID
write_cache(blk_num)
	unsigned		blk_num;
	{
	if(fseek(Interfp,(long) blk_num * DVRTBLKSIZE,FBEGIN) != 0)
		fatal("Can't seek in intermediate file");
	if(fwrite(cache,DVRTBLKSIZE,1,Interfp) != 1)
		fatal("Can't write to intermediate file");

	cache_blk = blk_num;
	cache_dirty = FALSE;
	}

/* End of TCX9.C */
