/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: tcxenv.h,v 2.2 92/04/21 20:06:25 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/*
 *	TCX Compiler Environment Header File
 *
 *	Copyright (c) 1987 by Quality Computer Systems and Logical Systems.
 *	All rights reserved.
 *
 *	Published by:	Logical Systems
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	03/17/87
 *
 *	This file defines the host and target environments for
 *	the TCX compiler and library.  It is included in the
 *	other source files to enable environment-dependent code
 *	to be selected where necessary.  The definitions of the
 *	names currently used are:
 *
 *		H_OPSYS	- the host operating system that TCX will
 *			  run under
 *
 *	The host environments are:
 *
 *		H_OPSYS
 *		-------
 *		MS_DOS	- TCX will run under MSDOS
 *		VMS	- TCX will run under VAX/VMS
 *		UNIX	- TCX will run under Unix
 *		BSD	- TCX will run under "new" BSD Unix variants
 *		MPW	- TCX will run under MPW on the Mac
 *		XENIX	- TCX will run under Microsoft XENIX on PC's
 *
 *	The target environment is always assumed to be a Inmos Transputer
 *	processor.
 *
 */

#define	MS_DOS		1
#define	VMS		2
#define	UNIX		3
#define	BSD		4
#define	MPW		5
#define	XENIX		6

/*
 *	Define host environment.
 */

#define	H_OPSYS		BSD

/*
 *	Some versions of BSD are missing functions
 */

#if	H_OPSYS == BSD

#ifndef OSU
#define	strchr			index
#define	strrchr			rindex
#endif
#define	memcpy(dst, src, len)	bcopy(src, dst, len)

#endif	/* BSD */

/*
 *	The parameters defined here tailor the TCX compiler to the
 *	characteristics of the host environment where it will run.  This
 *	includes initial values for global variables to fit the size of the
 *	environment and features which may be included in TCX.
 */

#define	TEMP_NAME	".1"		/* Filename extension for temp file */
#define	TEMP_PATH	"TMP"		/* Env. var. for temp directory */

/*
 *	CRLF defines whether the system needs to convert CR,LF pairs into
 *	linefeeds for text files (needed on many micro's).  The effect of this
 *	switch (if TRUE), is to make "fopen" calls contain a trailing "b" for
 *	binary files.  For most UNIX installations this	switch should be FALSE.
 */

#if	((H_OPSYS == MS_DOS) || (H_OPSYS == MPW))
#define	CRLF		1
#else
#define	CRLF		0
#endif	/* MS_DOS || MPW */

#if	H_OPSYS == MS_DOS

typedef	long		INT32;		/* 32 bit integer		*/
typedef unsigned long	UINT32;		/* 32 bit unsigned integer	*/
typedef	int		BOOL;		/* Boolean variable		*/

#define FILNAMSIZE	64		/* Size of largest pathname	*/
#endif	/* MS_DOS */

#if	((H_OPSYS == VMS) || (H_OPSYS == UNIX) || (H_OPSYS == MPW) || (H_OPSYS == BSD) || (H_OPSYS == XENIX))

typedef	long		INT32;		/* 32 bit integer		*/
typedef unsigned long	UINT32;		/* 32 bit unsigned integer	*/
typedef	int		BOOL;		/* Boolean variable		*/

#define FILNAMSIZE	128		/* Size of largest pathname */
#endif	/* VMS || UNIX || MPW || BSD || XENIX */

/*
 *	The parameters defined here are used to generate output acceptable to
 *	the target assembler.
 */

#define ASM_FEXT	"tal"		/* Lower case file extension	   */
#define ASM_UNIQ	'\0'		/* Prefix/Suffix to make assembler */
					/* names unique.		   */
#define	WORKSPACE	"?ws"		/* Symbol for val. of workspace size */
#define	RSLINK		"?rslink"	/* Struct return link symbol	   */
#define	STRUCTBASE	"?sb"		/* Symbol for offset from WS to	*/
					/* struct function return temp space */
#define	TEMPBASE	"?tb"		/* Symbol for offset from WS to temp */

#define	SLBMODULE	250		/* First of 2 "static link" modules */
#define	SLBNAME		"?slb"		/* Global "static link" base name */

#ifdef	SCOPE_INFO
#define	SCOPE_START	"?ss"		/* Debugging scope start symbol */
#define	SCOPE_END	"?se"		/* Debugging scope end symbol */
#endif	/* SCOPE_INFO */

/* End TCXENV.H */
