/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: tlnk3.c,v 2.2 92/04/21 11:13:00 trillium Rel $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/************************************************************************/
/*									*/
/*	File:	tlnk3.c							*/
/*									*/
/*	Transputer Linker (TLNK), source file 3.  This file contains	*/
/*	the code which resolves all the unbound references.  This code	*/
/*	performs a function which is essentially equivalent to the	*/
/*	assembly process which TASM started, but on all the references	*/
/*	which were inter-module or external and thus out of the scope	*/
/*	of TASM.							*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1986-1989 by Logical Systems.		*/
/*									*/
/*				09/01/86				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<ctype.h>		/* Character type functions */
#include	<string.h>		/* String functions */
#include	"taldef.h"		/* Assembler/Linker definitions */
#include	"tlnkdef.h"		/* Linker definitions */
#include	"tlnktyp.h"		/* Linker data declarations */
#include	"tlnkext.h"		/* Linker externals */

/************************************************************************/
/*									*/
/*	Routine:	addr_pad					*/
/*									*/
/*	Updates the minimum and maximum module addresses as a result of	*/
/*	a T_ALIGN record.  If the minimum and maximum addresses PRIOR	*/
/*	to the record are the same, convert the record into a T_DATA	*/
/*	record of the appropriate length (1/1-3 bytes of zero).		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if the record was converted.		*/
/*									*/
/************************************************************************/

BOOL
addr_pad()
	{
	int			temp;

	if(min_addr == max_addr)
		{
		if((temp = (int) (((SLONG) word_size -
			(min_addr & ((SLONG) (word_size - 1)))) &
				((SLONG) (word_size - 1)))) != 0)
			{
			min_addr += temp;	/* Update minimum address */
			max_addr += temp;	/* Update maximum address */
			PUT_1B((int) T_DATA);	/* Convert record over */
			copy_temp(2);		/* Line # */
			put_2b((int) temp);	/* # of zero's to add */
			while(temp-- > 0)
				PUT_1B((int) 0);
			}
		else
			get_2b();	/* Toss line # */
		return (TRUE);
		}
	else
		{
		PUT_1B((int) T_ALIGN);
		copy_temp(2);		/* Line # */
		max_addr += (SLONG) (word_size - 1);
		}
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	adj_max_addr					*/
/*									*/
/*	This routine is a helping function for "pass1".  It figures out	*/
/*	if the specified symbol definition is a forward reference from	*/
/*	the current location address.  If not, the "max_addr" for the	*/
/*	symbol is returned.  If so, the "max_addr" is adjusted with the	*/
/*	current value of "max_addr_adj" to compensate for the change	*/
/*	which has occurred in the maximum address counter since the	*/
/*	last time the symbol definition address was updated (the	*/
/*	previous pass of "pass1").					*/
/*									*/
/*	Parameters:							*/
/*		"index"	A symbol table index value.			*/
/*									*/
/*	Returns:	The correct "max_addr" to use for the symbol.	*/
/*									*/
/************************************************************************/

SLONG
adj_max_addr(index)
	UINT			index;
	{
	SLONG			max_temp;
	SLONG			min_temp;
	BOOL			our_region;
	UINT			temp_index;
/*
 *	Figure out if it is a "forward" reference (and in same code/data
 *	region).  To start with we compute the equivalent of "get_max(index)",
 *	"get_min(index)" and "region == get_region(index)" and place the
 *	results in "max_temp","min_temp" and "our_region" respectively.  We do
 *	this INLINE to eliminate the normal function call overhead within this
 *	heavily used function.
 */
	if((temp_index = locals[index].gst_num) < MAX_GLOBALS)
		{
		if((temp_index = globals[temp_index].lst_num) < MAX_SYMBOLS)
			{
			max_temp = locals[temp_index].max_addr;
			min_temp = locals[temp_index].min_addr;
			our_region = (region == locals[temp_index].region);
			}
		else
			{
			max_temp = 0;	/* Undefined symbol */
			min_temp = 0;
			our_region = TRUE;
			}
		}
	else
		{
		max_temp = locals[index].max_addr;
		min_temp = locals[index].min_addr;
		our_region = (region == locals[index].region);
		}
/*
 *	Now we check for a forward reference...
 */
	if(((max_temp - min_temp) > (max_addr - min_addr)) && our_region)
		return(max_temp - max_addr_adj);
	return (max_temp);	/* Nope */
	}

/************************************************************************/
/*									*/
/*	Routine:	adj_min_addr					*/
/*									*/
/*	This routine is a helping function for "pass1" and "pass2".  It	*/
/*	figures out if the specified symbol definition is a forward	*/
/*	reference from the current location address.  If not, the	*/
/*	"min_addr" for the symbol is returned.  If so, the "min_addr"	*/
/*	is adjusted with the current value of "min_addr_adj" to		*/
/*	compensate for the change which has occurred in the minimum	*/
/*	address counter since the last time the symbol definition	*/
/*	address was updated (the previous pass of "pass1").		*/
/*									*/
/*	Parameters:							*/
/*		"index"	A symbol table index value.			*/
/*									*/
/*	Returns:	The correct "min_addr" to use for the symbol.	*/
/*									*/
/************************************************************************/

SLONG
adj_min_addr(index)
	UINT			index;
	{
	SLONG			max_temp;
	SLONG			min_temp;
	BOOL			our_region;
	UINT			temp_index;
/*
 *	Figure out if it is a "forward" reference (and in same code/data
 *	region).  To start with we compute the equivalent of "get_max(index)",
 *	"get_min(index)" and "region == get_region(index)" and place the
 *	results in "max_temp","min_temp" and "our_region" respectively.  We do
 *	this INLINE to eliminate the normal function call overhead within this
 *	heavily used function.
 */
	if((temp_index = locals[index].gst_num) < MAX_GLOBALS)
		{
		if((temp_index = globals[temp_index].lst_num) < MAX_SYMBOLS)
			{
			max_temp = locals[temp_index].max_addr;
			min_temp = locals[temp_index].min_addr;
			our_region = (region == locals[temp_index].region);
			}
		else
			{
			max_temp = 0;	/* Undefined symbol */
			min_temp = 0;
			our_region = TRUE;
			}
		}
	else
		{
		max_temp = locals[index].max_addr;
		min_temp = locals[index].min_addr;
		our_region = (region == locals[index].region);
		}
/*
 *	Now we check for a forward reference...
 */
	if(((max_temp - min_temp) > (max_addr - min_addr)) && our_region)
		return(min_temp + min_addr_adj);
	return (min_temp);	/* Nope */
	}

/************************************************************************/
/*									*/
/*	Routine:	bind_constant					*/
/*									*/
/*	Takes a PC relative constant which represents the offset from	*/
/*	the current address counter and figures out how long the	*/
/*	corresponding instruction needs to be taking into account that	*/
/*	the Transputer increments the PC to point just beyond the	*/
/*	instruction BEFORE it processes the PC offset, thus the length	*/
/*	of the instruction affects the needed offset AND the offset in	*/
/*	turn affects the needed prefix string length for the		*/
/*	instruction!							*/
/*									*/
/*	Parameters:							*/
/*		"function" -	The desired opcode.			*/
/*		"value"	-	The desired PC offset from the start of	*/
/*				the instruction/prefix string.		*/
/*									*/
/*	Returns:	The instruction length needed.			*/
/*									*/
/************************************************************************/

UINT
bind_constant(function,value)
	UINT			function;
	SLONG			value;
	{
	UINT			length1;
	UINT			length2;
/*
 *	First we calculate the needed instruction length ignoring the effect
 *	of the offset provided by the instruction length itself.
 */
	length1 = i_size((UINT) function,(SLONG) value);
/*
 *	Now we take a first cut at the instruction length including the self
 *	referential aspects.
 */
	length2 = i_size((UINT) function,(SLONG) (value - length1));
/*
 *	Finally, we do the calculation again to ensure that we aren't
 *	oscillating across a instruction length straddle point.  The two
 *	iterations of the length calculation are sufficent since we can be
 *	straddling at most one boundary.
 */
	length1 = i_size((UINT) function,(SLONG) (value - length2));
/*
 *	Now take the "max" of the above two calculated lengths to ensure that
 *	we cover the worst case.
 */
	if(length1 > length2)
		return (length1);
	else
		return (length2);
	}

/************************************************************************/
/*									*/
/*	Routine:	build_i						*/
/*									*/
/*	Build an instruction and output it to the temp file given a	*/
/*	function (instruction) code and a 32 bit constant.		*/
/*									*/
/*	Parameters:							*/
/*		"length"	The length of the instruction to build.	*/
/*		"function"	The second 4 bits of this hold the	*/
/*				basic instruction opcode bits.		*/
/*		"constant"	The 32 bit constant value.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
build_i(length,function,constant)
	UINT			length;
	UINT			function;
	SLONG			constant;
	{
	register	UINT	i;
	UCHAR			preamble[7];
/*
 *	Bulletproofing checks.
 */
	if((length <= 0) || (length > max_i_size))
		fatali("4");
	if(i_size(function,constant) > length)
		fatali("5");
/*
 *	First see if we are doing the "optimizing" trick for ".LDC" opcodes.
 */
	if((function == DOTLDC) && (constant <= -2146435074L))
		{
		build_i((UINT) MINT_SIZE,(UINT) OPR,(SLONG) MINT);
		length -= MINT_SIZE;
		constant &= 0x7FFFFFFFL;
		if(constant != (SLONG) 0)
			{
			build_i((UINT) length,(UINT) ADC,(SLONG) constant);
			return;
			}
		else
			{
			while(length--)
				build_i((UINT) 1,(UINT) NOP,(SLONG) 0);
			return;
			}
		}

	function &= 0xF0;
/*
 *	Otherwise, generate the instruction in the normal fashion.
 */
	if(length > 1)
		{
/*
 *	Well we know we need prefix instructions, figure out whether the first
 *	one should be a negative prefix.
 */
		if(constant < 0)
			preamble[0] = NFIX;
		else
			preamble[0] = PFIX;
/*
 *	Now determine the basic function/instruction byte which goes at the end
 *	of the string of prefixes.
 */
		function = function | ((UINT) (constant & 0xF));
/*
 *	Figure out the remainder of the necessary prefixes.
 */
		for(i = length - 2; i > 0; i--)
			preamble[i] = PFIX | ((constant >>= 4) & 0xF);
		if(preamble[0] == NFIX)
			{
/*lint -e502 lint doesn't like complementing signed integral types */
			preamble[0] |= (( ~(constant >> 4)) & 0xF);
/*lint +e502 lint doesn't like complementing signed integral types */
			}
		else
			preamble[0] |= ((constant >> 4) & 0xF);
		}
	else				/* If no prefix */
		function = function | ((UINT) (constant & 0xF));
/*
 *	Time to write the instruction encoding out to the temp file.
 */
	for(i = 0; i < (length - 1); i++)
		PUT_1B((int) preamble[i]);
	PUT_1B((int) function);
	}

/************************************************************************/
/*									*/
/*	Routine:	debug_addr					*/
/*									*/
/*	A helping function for "pass1" which handles the		*/
/*	"T_DEBUGSYM_DATA" records.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
debug_addr()
	{
	SLONG			big_temp;
	register	UINT	i;
	register	UINT	length;

	number = get_2b();		/* Input line # */
	big_temp = get_4b();		/* Offset value */
	length = get_2b();		/* Data section length */
	i = get_2b();			/* ST entry # */
/*
 *	See if we know enough to convert the reference into a T_DEBUG_DATA
 *	record (is the symbol address completely known?).
 */
	if(adj_min_addr(i) == adj_max_addr(i))
		{
		PUT_1B((int) T_DEBUG_DATA);	/* We've bagged another one */
		put_2b((int) number);	/* Line # */
		put_4b((SLONG) (adj_min_addr(i) + big_temp)); /* Address */
		put_2b((int) length);	/* Data section length */
		}
/*
 *	Otherwise we were unable to convert it at this time.
 */
	else
		{
		PUT_1B((int) T_DEBUGSYM_DATA);	/* Record type */
		put_2b((int) number);	/* Line # */
		put_4b((SLONG) big_temp);	/* Address offset */
		put_2b((int) length);	/* Data section length */
		put_2b((int) i);	/* ST index # */
		}
/*
 *	Copy rest of data on through.
 */
	copy_temp(length);
	}

/************************************************************************/
/*									*/
/*	Routine:	do_addr						*/
/*									*/
/*	A helping function for "pass1" which handles the "T_ADDR_DATA"	*/
/*	records.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
do_addr()
	{
	register	UINT	i;

	number = get_2b();		/* Input line # */
	i = get_2b();			/* ST entry # */
/*
 *	See if we can bind the reference (is the symbol address completely
 *	known and we aren't using the "static link" global data model?).
 */
	if((! static_link) && (adj_min_addr(i) == adj_max_addr(i)))
		{
		PUT_1B((int) T_DATA);	/* We've bagged another one */
		put_2b((int) number);	/* Line # */
		put_2b((int) word_size);
		if(word_size == 2)
/* Address */		put_2b((int) (adj_min_addr(i) + get_4b()));
		else
/* Address */		put_4b((SLONG) (adj_min_addr(i) + get_4b()));
		}
/*
 *	Otherwise we were unable to convert it at this time.
 */
	else
		{
		PUT_1B((int) T_ADDR_DATA);	/* Record type */
		put_2b((int) number);	/* Line # */
		put_2b((int) i);	/* ST index # */
		copy_temp(4);		/* Copy reference offset */
		}
/*
 *	Update current addresses.
 */
	min_addr += (SLONG) word_size;
	max_addr += (SLONG) word_size;
	}

/************************************************************************/
/*									*/
/*	Routine:	do_diff						*/
/*									*/
/*	A helping function for "pass1" which handles the		*/
/*	"T_RELREL_DATA" records.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
do_diff()
	{
	register	UINT	left;
	SLONG			lmax_a;
	SLONG			lmin_a;
	register	UINT	right;
	SLONG			rmax_a;
	SLONG			rmin_a;

	number = get_2b();		/* Input line # */
	left = get_2b();		/* "left" ST entry # */
	right = get_2b();		/* "right" ST entry # */
/*
 *	See if we know enough to convert the reference into a normal data
 *	record.
 */
	lmin_a = adj_min_addr(left);
	lmax_a = adj_max_addr(left);
	rmin_a = adj_min_addr(right);
	rmax_a = adj_max_addr(right);
	if((((lmax_a - rmax_a) == (lmin_a - rmin_a)) &&
		(get_region(left) == get_region(right))) ||
			((lmax_a == lmin_a) && (rmax_a == rmin_a)))
		{
		PUT_1B((int) T_DATA);	/* We've bagged another one */
		put_2b((int) number);	/* Line # */
		put_2b((int) word_size);
		if(word_size == 2)
			put_2b((int) (get_4b() + lmax_a - rmax_a));
		else
			put_4b((SLONG) (get_4b() + lmax_a - rmax_a));
		}
/*
 *	Otherwise we were unable to convert it at this time.
 */
	else
		{
		PUT_1B((int) T_RELREL_DATA);	/* Record type */
		put_2b((int) number);	/* Line # */
		put_2b((int) left);	/* "left" ST index # */
		put_2b((int) right);	/* "right" ST index # */
		copy_temp(4);		/* Copy reference offset */
		}
/*
 *	Update current addresses.
 */
	min_addr += (SLONG) word_size;
	max_addr += (SLONG) word_size;
	}

/************************************************************************/
/*									*/
/*	Routine:	do_i_addr					*/
/*									*/
/*	A helping function for "pass1" which handles the "T_ADDR_OP"	*/
/*	records.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if the reference was converted into an	*/
/*			normal data record.  FALSE otherwise.		*/
/*									*/
/************************************************************************/

BOOL
do_i_addr()
	{
	SLONG			con;
	UINT			function;
	register	UINT	i;
	UINT			ldpi_sze;
	UINT			min_length;

	number = get_2b();		/* Input line # */
	min_length = GET_1B();		/* Get minimum instruction length */
	i = get_2b();			/* ST entry # */
/*
 *	See if we can bind the reference (is the symbol address completely
 *	known?).
 */
	if((adj_min_addr(i) == adj_max_addr(i)) &&
		(min_addr == max_addr))
		{
		PUT_1B((int) T_DATA);	/* We've bagged another one */
		put_2b((int) number);	/* Line # */
		con = get_4b() + adj_min_addr(i);
		function = GET_1B();
		i = ldpi_size((UINT) function,(SLONG) con,min_addr,&ldpi_sze);
		if((i > min_length) &&
			(((function == DOTLDCR) && (min_length > 3)) ||
			((function != DOTLDCR) && (min_length != 1))))
			{
			fatali("6");
			}
		if(i < min_length)
			i = min_length;	/* Minimum */
		put_2b((int) i);	/* Instruction size */
		ldpi_build(i,(UINT) function,(SLONG) con,ldpi_sze);
		min_addr += i;		/* Update addresses */
		max_addr += i;
		return (TRUE);		/* We've performed a conversion */
		}
/*
 *	Otherwise copy reference to the temp file.
 */
	PUT_1B((int) T_ADDR_OP);	/* Record type */
	put_2b((int) number);		/* Line # */
	PUT_1B((int) min_length);	/* Minimum length */
	put_2b((int) i);		/* ST index # */
	put_4b((SLONG) (con = get_4b()));	/* Copy offset value */
	PUT_1B((int) (function = GET_1B()));	/* Copy basic opcode */
/*
 *	Update current addresses.
 */
	ldpi_addr((UINT) function,(SLONG) (adj_min_addr(i) + con),
		(SLONG) (adj_max_addr(i) + con),min_length);
	return (FALSE);
	}

/************************************************************************/
/*									*/
/*	Routine:	do_i_diff					*/
/*									*/
/*	A helping function for "pass1" which handles the "T_RELREL_OP"	*/
/*	records.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if the reference was converted into an	*/
/*			normal data record.  FALSE otherwise.		*/
/*									*/
/************************************************************************/

BOOL
do_i_diff()
	{
	SLONG			con;
	UINT			function;
	register	UINT	left;
	UINT			length;
	SLONG			lmax_a;
	SLONG			lmin_a;
	UINT			min_length;
	register	UINT	right;
	SLONG			rmax_a;
	SLONG			rmin_a;

	number = get_2b();		/* Input line # */
	min_length = GET_1B();		/* Get minimum instruction length */
	left = get_2b();		/* "left" ST entry # */
	right = get_2b();		/* "right" ST entry # */
/*
 *	See if we now know enough to convert the reference to a normal data
 *	record.
 */
	lmin_a = adj_min_addr(left);
	lmax_a = adj_max_addr(left);
	rmin_a = adj_min_addr(right);
	rmax_a = adj_max_addr(right);
	if((((lmax_a - rmax_a) == (lmin_a - rmin_a)) &&
		(get_region(left) == get_region(right))) ||
			((lmax_a == lmin_a) && (rmax_a == rmin_a)))
		{
		PUT_1B((int) T_DATA);	/* Absolute data record type */
		put_2b((int) number);	/* Line # */
		con = get_4b() + lmax_a - rmax_a;
		function = GET_1B();
		length = i_size((UINT) function,(SLONG) con);
		if((min_length != 1) && (length > min_length))
			fatali("7");
		if(length < min_length)
			length = min_length;	/* Minimum */
		put_2b((int) length);	/* Instr. size */
		build_i(length,(UINT) function,(SLONG) con);
		min_addr += length;
		max_addr += length;
		return (TRUE);		/* We've performed a conversion */
		}
/*
 *	We can't finish the binding, determine the best and worst case for
 *	instruction length and update the address counters appropriately.
 *	We also need to copy the reference to the temp file so that it may get
 *	converted in some future pass.
 */
	PUT_1B((int) T_RELREL_OP);	/* Record type */
	put_2b((int) number);		/* Line # */
	PUT_1B((int) min_length);	/* Minimum allowable length */
	put_2b((int) left);		/* "left" ST index # */
	put_2b((int) right);		/* "right" ST index # */
	put_4b((SLONG) (con = get_4b()));	/* Copy offset */
	PUT_1B((int) (function = GET_1B()));	/* Copy basic opcode */
/*
 *	Figure out how to update the location counters.
 */
	update_addr((UINT) function,
		(SLONG) (con + lmax_a - rmax_a),
		(SLONG) (con + lmin_a - rmin_a),
			min_length,FALSE);
	return (FALSE);			/* No conversion performed */
	}

/************************************************************************/
/*									*/
/*	Routine:	do_i_wdiff					*/
/*									*/
/*	A helping function for "pass1" which handles the "T_WRELREL_OP"	*/
/*	records.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if the reference was converted into an	*/
/*			normal data record.  FALSE otherwise.		*/
/*									*/
/************************************************************************/

BOOL
do_i_wdiff()
	{
	SLONG			con;
	UINT			function;
	register	UINT	left;
	UINT			length;
	SLONG			lmax_a;
	SLONG			lmin_a;
	UINT			min_length;
	register	UINT	right;
	SLONG			rmax_a;
	SLONG			rmin_a;

	number = get_2b();		/* Input line # */
	min_length = GET_1B();		/* Get minimum instruction length */
	left = get_2b();		/* "left" ST entry # */
	right = get_2b();		/* "right" ST entry # */
/*
 *	See if we now know enough to convert the reference to a normal data
 *	record.
 */
	lmin_a = adj_min_addr(left);
	lmax_a = adj_max_addr(left);
	rmin_a = adj_min_addr(right);
	rmax_a = adj_max_addr(right);
	if((((lmax_a - rmax_a) == (lmin_a - rmin_a)) &&
		(get_region(left) == get_region(right))) ||
			((lmax_a == lmin_a) && (rmax_a == rmin_a)))
		{
		PUT_1B((int) T_DATA);	/* Absolute data record type */
		put_2b((int) number);	/* Line # */
		con = (get_4b() + lmax_a - rmax_a) / (SLONG) word_size;
		function = GET_1B();
		length = i_size((UINT) function,(SLONG) con);
		if((min_length != 1) && (length > min_length))
			fatali("7");
		if(length < min_length)
			length = min_length;	/* Minimum */
		put_2b((int) length);	/* Instr. size */
		build_i(length,(UINT) function,(SLONG) con);
		min_addr += length;
		max_addr += length;
		return (TRUE);		/* We've performed a conversion */
		}
/*
 *	We can't finish the binding, determine the best and worst case for
 *	instruction length and update the address counters appropriately.
 *	We also need to copy the reference to the temp file so that it may get
 *	converted in some future pass.
 */
	PUT_1B((int) T_WRELREL_OP);	/* Record type */
	put_2b((int) number);		/* Line # */
	PUT_1B((int) min_length);	/* Minimum allowable length */
	put_2b((int) left);		/* "left" ST index # */
	put_2b((int) right);		/* "right" ST index # */
	put_4b((SLONG) (con = get_4b()));	/* Copy offset */
	PUT_1B((int) (function = GET_1B()));	/* Copy basic opcode */
/*
 *	Figure out how to update the location counters.
 */
	update_addr((UINT) function,
		(SLONG) (con + lmax_a - rmax_a) / (SLONG) word_size,
		(SLONG) (con + lmin_a - rmin_a) / (SLONG) word_size,
			min_length,FALSE);
	return (FALSE);			/* No conversion performed */
	}

/************************************************************************/
/*									*/
/*	Routine:	dor_ref						*/
/*									*/
/*	A helping function for "pass1" which handles the "T_REL_DATA"	*/
/*	records.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
dor_ref()
	{
	number = get_2b();		/* Input line # */
/*
 *	See if we know enough to convert the reference into a normal data
 *	record.
 */
	if(max_addr == min_addr)
		{
		PUT_1B((int) T_DATA);	/* We've bagged another one */
		put_2b((int) number);	/* Line # */
		put_2b((int) word_size);
		if(word_size == 2)
			put_2b((int) (get_4b() - min_addr));
		else
			put_4b((SLONG) (get_4b() - min_addr));
		}
/*
 *	Otherwise we were unable to convert it at this time.
 */
	else
		{
		PUT_1B((int) T_REL_DATA);	/* Record type */
		put_2b((int) number);	/* Line # */
		copy_temp(4);		/* Copy address */
		}
/*
 *	Update current addresses.
 */
	min_addr += (SLONG) word_size;
	max_addr += (SLONG) word_size;
	}

/************************************************************************/
/*									*/
/*	Routine:	dor_i_ref					*/
/*									*/
/*	A helping function for "pass1" which handles the non-symbolic	*/
/*	PC relative instructions (T_REL_OP) and tries to completely	*/
/*	bind the addresses involved so that the reference can be	*/
/*	converted into a normal data reference.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if the reference was converted into an	*/
/*			normal data record.  FALSE otherwise.		*/
/*									*/
/************************************************************************/

BOOL
dor_i_ref()
	{
	SLONG			con;
	UINT			function;
	register	UINT	i;
	UINT			min_length;

	number = get_2b();		/* Input line # */
	min_length = GET_1B();		/* Get minimum instruction length */
/*
 *	See if we now know enough to convert the reference to a normal data
 *	record.
 */
	if(min_addr == max_addr)
		{
		PUT_1B((int) T_DATA);	/* Absolute data record type */
		put_2b((int) number);	/* Line # */
		con = get_4b() - min_addr;
		function = GET_1B();
		i = bind_constant((UINT) function,(SLONG) con);
		if((min_length != 1) && (i > min_length))
			fatali("8");
		if(i < min_length)
			i = min_length;	/* Minimum */
		put_2b((int) i);	/* Instr. size */
		build_i(i,(UINT) function,(SLONG) (con - i));
		min_addr += i;
		max_addr += i;
		return (TRUE);		/* We've performed a conversion */
		}
/*
 *	We can't finish the binding, determine the best and worst case for
 *	instruction length and update the address counters appropriately.
 *	We also need to copy the reference to the temp file so that it may get
 *	converted in some future pass.
 */
	PUT_1B((int) T_REL_OP);	/* Record type */
	put_2b((int) number);		/* Line # */
	PUT_1B((int) min_length);	/* Minimum allowable length */
	put_4b((SLONG) (con = get_4b()));	/* Copy address */
	PUT_1B((int) (function = GET_1B()));	/* Copy basic opcode */
/*
 *	Figure out how to update the location counters.
 */
	update_addr((UINT) function,
		(SLONG) (con - min_addr),(SLONG) (con - max_addr),
			min_length,TRUE);
	return (FALSE);			/* No conversion performed */
	}

/************************************************************************/
/*									*/
/*	Routine:	dors_ref					*/
/*									*/
/*	A helping function for "pass1" which handles the		*/
/*	"T_RELSYM_DATA" records.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
dors_ref()
	{
	register	UINT	i;
	SLONG			max_a;
	SLONG			min_a;

	number = get_2b();		/* Input line # */
	i = get_2b();			/* ST entry # */
/*
 *	See if we know enough to convert the reference into a normal data
 *	record.
 */
	min_a = adj_min_addr(i);
	max_a = adj_max_addr(i);
	if((((max_a - max_addr) == (min_a - min_addr)) &&
		(get_region(i) == region)) ||
			((max_a == min_a) && (max_addr == min_addr)))
		{
		PUT_1B((int) T_DATA);	/* We've bagged another one */
		put_2b((int) number);	/* Line # */
		put_2b((int) word_size);
		if(word_size == 2)
			put_2b((int) (get_4b() + max_a - max_addr));
		else
			put_4b((SLONG) (get_4b() + max_a - max_addr));
		}
/*
 *	Otherwise we were unable to convert it at this time.
 */
	else
		{
		PUT_1B((int) T_RELSYM_DATA);	/* Record type */
		put_2b((int) number);	/* Line # */
		put_2b((int) i);	/* ST index # */
		copy_temp(4);		/* Copy reference offset */
		}
/*
 *	Update current addresses.
 */
	min_addr += (SLONG) word_size;
	max_addr += (SLONG) word_size;
	}

/************************************************************************/
/*									*/
/*	Routine:	dors_i_ref					*/
/*									*/
/*	A helping function for "pass1" which handles the symbolic PC	*/
/*	relative instructions (T_RELSYM_OP), and tries to completely	*/
/*	bind the addresses involved so that the reference can be	*/
/*	converted into a normal data reference.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if the reference was converted into an	*/
/*			normal data record.  FALSE otherwise.		*/
/*									*/
/************************************************************************/

BOOL
dors_i_ref()
	{
	SLONG			con;
	UINT			function;
	register	UINT	i;
	SLONG			max_a;
	SLONG			min_a;
	UINT			min_length;

	number = get_2b();		/* Input line # */
	min_length = GET_1B();		/* Get minimum instruction length */
	i = get_2b();			/* ST entry # */
/*
 *	See if we now know enough to convert the reference to a normal data
 *	record.
 */
	min_a = adj_min_addr(i);
	max_a = adj_max_addr(i);
	if((((max_a - max_addr) == (min_a - min_addr)) &&
		(get_region(i) == region)) ||
			((max_a == min_a) && (max_addr == min_addr)))
		{
		PUT_1B((int) T_DATA);	/* Absolute data record type */
		put_2b((int) number);	/* Line # */
		con = get_4b() + max_a - max_addr;
		function = GET_1B();
		i = bind_constant((UINT) function,(SLONG) con);
		if((min_length != 1) && (i > min_length))
			fatali("3");
		if(i < min_length)
			i = min_length;	/* Minimum */
		put_2b((int) i);	/* Instr. size */
		build_i(i,(UINT) function,(SLONG) (con - i));
		min_addr += i;
		max_addr += i;
		return (TRUE);		/* We've performed a conversion */
		}
/*
 *	We can't finish the binding, determine the best and worst case for
 *	instruction length and update the address counters appropriately.
 *	We also need to copy the reference to the temp file so that it may get
 *	converted in some future pass.
 */
	PUT_1B((int) T_RELSYM_OP);	/* Record type */
	put_2b((int) number);		/* Line # */
	PUT_1B((int) min_length);	/* Minimum allowable length */
	put_2b((int) i);		/* ST index # */
	put_4b((SLONG) (con = get_4b()));	/* Copy offset */
	PUT_1B((int) (function = GET_1B()));	/* Copy basic opcode */
/*
 *	Figure out how to update the location counters.
 */
	update_addr((UINT) function,(SLONG) (con + max_a - max_addr),
		(SLONG) (con + min_a - min_addr),min_length,TRUE);
	return (FALSE);			/* No conversion performed */
	}

/************************************************************************/
/*									*/
/*	Routine:	get_max						*/
/*									*/
/*	Return the "maximum" address associated with the local symbol	*/
/*	table entry parameter.						*/
/*									*/
/*	Parameters:							*/
/*		"index"	The symbol table entry #.			*/
/*									*/
/*	Returns:	The "maximum" symbol address.			*/
/*									*/
/************************************************************************/

SLONG
get_max(index)
	UINT			index;
	{
	UINT			temp_index;

	if((temp_index = locals[index].gst_num) < MAX_GLOBALS)
		{
		if((index = globals[temp_index].lst_num) < MAX_SYMBOLS)
			return (locals[index].max_addr);
		else
			return ((SLONG) 0);	/* Undefined symbol */
		}
	return (locals[index].max_addr);
	}

/************************************************************************/
/*									*/
/*	Routine:	get_min						*/
/*									*/
/*	Return the "minimum" address associated with the local symbol	*/
/*	table entry parameter.						*/
/*									*/
/*	Parameters:							*/
/*		"index"	The symbol table entry #.			*/
/*									*/
/*	Returns:	The "minimum" symbol address.			*/
/*									*/
/************************************************************************/

SLONG
get_min(index)
	UINT			index;
	{
	UINT			temp_index;

	if((temp_index = locals[index].gst_num) < MAX_GLOBALS)
		{
		if((index = globals[temp_index].lst_num) < MAX_SYMBOLS)
			return (locals[index].min_addr);
		else
			return ((SLONG) 0);	/* Undefined symbol */
		}
	return (locals[index].min_addr);
	}

/************************************************************************/
/*									*/
/*	Routine:	get_region					*/
/*									*/
/*	Return the region tag associated with the local symbol table	*/
/*	entry parameter.						*/
/*									*/
/*	Parameters:							*/
/*		"index"	The symbol table entry #.			*/
/*									*/
/*	Returns:	The symbol definition region tag.		*/
/*									*/
/************************************************************************/

UINT
get_region(index)
	UINT			index;
	{
	UINT			temp_index;

	if((temp_index = locals[index].gst_num) < MAX_GLOBALS)
		{
		if((index = globals[temp_index].lst_num) < MAX_SYMBOLS)
			return (locals[index].region);
		else
			return (region);	/* Undefined symbol, fake it */
		}
	return (locals[index].region);
	}

/************************************************************************/
/*									*/
/*	Routine:	i_size						*/
/*									*/
/*	Determine the length of an instruction with immediate data.	*/
/*									*/
/*	Parameters:							*/
/*		"function" -	The basic instruction opcode.		*/
/*		"data"	-	The immediate data to scrutinize.	*/
/*									*/
/*	Returns:	The instruction length in bytes.		*/
/*									*/
/************************************************************************/

UINT
i_size(function,data)
	UINT			function;
	register	SLONG	data;
	{
	static	char	*errmsg = "Address out of 16 bit addressing range";

	if(data >= (SLONG) 0)
		{
		if(data > (SLONG) 65535L)
			{
			if(data > (SLONG) 16777215L)
				{
				if(data > (SLONG) 268435455L)
					{
					if(word_size == 2)
						fatal(errmsg,"");
					return (8);
					}
				else
					{
					if(word_size == 2)
						fatal(errmsg,"");
					return (7);
					}
				}
			else
				{
				if(data > (SLONG) 1048575L)
					{
					if(word_size == 2)
						fatal(errmsg,"");
					return (6);
					}
				else
					{
					if(word_size == 2)
						fatal(errmsg,"");
					return (5);
					}
				}
			}
		else
			{
			if(data > (SLONG) 255)
				{
				if(data > (SLONG) 4095)
					return (4);
				else
					return (3);
				}
			else
				{
				if(data > (SLONG) 15)
					return (2);
				else
					return (1);
				}
			}
		}
	else
		{
		if(data < (SLONG) -65536L)
			{
			if(data < (SLONG) -16777216L)
				{
				if(data < (SLONG) -268435456L)
					{
					if((function == DOTLDC) &&
						(data <= (SLONG) -2146435074L))
						{
						data &= 0x7FFFFFFFL;
						if(data != (SLONG) 0)
							{
							return (MINT_SIZE +
								 i_size(ADC,
									data));
							}
						else
							return (MINT_SIZE);
						}
					if(word_size == 2)
						fatal(errmsg,"");
					return (8);
					}
				else
					{
					if(word_size == 2)
						fatal(errmsg,"");
					return (7);
					}
				}
			else
				{
				if(data < (SLONG) -1048576L)
					{
					if(word_size == 2)
						fatal(errmsg,"");
					return (6);
					}
				else
					{
					if(word_size == 2)
						fatal(errmsg,"");
					return (5);
					}
				}
			}
		else
			{
			if(data < (SLONG) -256)
				{
				if(data < (SLONG) -4096)
					return (4);
				else
					return (3);
				}
			else
				return (2);
			}
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	ldpi_addr					*/
/*									*/
/*	Takes two possible immediate values for instructions which have	*/
/*	immediate data, computes the length of an instruction resulting	*/
/*	from each and updates the "min_addr" and "max_addr" address	*/
/*	counters appropriately.  This function is a helper for		*/
/*	"do_i_addr" which handles the T_ADDR_OP record.			*/
/*									*/
/*	Parameters:							*/
/*		"function" -	The opcode being processed.		*/
/*		"min"	-	A SLONG which is used to compute the	*/
/*				first possible instruction length based	*/
/*				on the minimum symbol address.		*/
/*		"max"	-	A SLONG which is used to compute the	*/
/*				second possible instruction length	*/
/*				based on the maximum symbol address.	*/
/*		"min_length"	The minimum instruction length to	*/
/*				generate.				*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
ldpi_addr(function,min,max,min_length)
	UINT			function;
	SLONG			min;
	SLONG			max;
	UINT			min_length;
	{
	UINT			ldpi_sze1;
	UINT			ldpi_sze2;
	UINT			length1;
	UINT			length2;

	if((function != DOTLDCR) && (function != DOTLDC))
		{
		update_addr(function,min,max,min_length,FALSE);
		return;
		}

	length1 = ldpi_size(function,min,min_addr,&ldpi_sze1);
	length2 = ldpi_size(function,max,max_addr,&ldpi_sze2);
/*
 *	First see if the function specifies the "ldc"/"ldpi" code sequence, if
 *	so compute the min and max address increments as appropriate.  If not
 *	we set the min address increment to "min_length" since we can't easily
 *	compute the REAL address minimum.  In any event the address maximum is
 *	computed in the normal fashion (at least as long as "min_length" also).
 */
	if(function == DOTLDCR)
		{
		if(((min - min_addr - length1) ^
			(max - max_addr - length2)) < 0)
			{
			if(min_length > 3)
				min_addr += min_length;
			else
				min_addr += 3;
			}
		else
			{
			if(length1 > length2)
				{
				if(length2 < min_length)
					min_addr += min_length;
				else
					min_addr += length2;
				}
			else
				{
				if(length1 < min_length)
					min_addr += min_length;
				else
					min_addr += length1;
				}
			}
		}
	else
		min_addr += min_length;
/*
 *	Update maximum address as appropriate.
 */
	if(length1 > length2)
		{
		if(length1 < min_length)
			max_addr += min_length;
		else
			max_addr += length1;
		}
	else
		{
		if(length2 < min_length)
			max_addr += min_length;
		else
			max_addr += length2;
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	ldpi_build					*/
/*									*/
/*	Build an instruction and output it to the temp file given a	*/
/*	function (instruction) code and a 32 bit constant.  This	*/
/*	particular routine is a helper function for "do_i_addr" for	*/
/*	handling the T_ADDR_OP records.					*/
/*									*/
/*	Parameters:							*/
/*		"length"	The length of the instruction to build.	*/
/*		"function"	The second 4 bits of this hold the	*/
/*				basic instruction opcode bits.		*/
/*		"constant"	The 32 bit constant value.		*/
/*		"ldpi_sze"	Total size of "ldc"/"ldpi" instructions	*/
/*				to generate (zero if these instructions	*/
/*				are not to be used).			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
ldpi_build(length,function,constant,ldpi_sze)
	UINT			length;
	UINT			function;
	SLONG			constant;
	UINT			ldpi_sze;
	{
/*
 *	Bulletproofing check.
 */
	if((length <= 0) ||
		((ldpi_sze > 0) && (length > (max_i_size + LDPI_SIZE))) ||
		((ldpi_sze == 0) && (length > max_i_size)))
		{
		fatali("9");
		}
/*
 *	First see if we are doing the "ldc"/"ldpi" "optimizing" trick for
 *	".LDC" opcodes.
 */
	if(ldpi_sze > 0)
		{
		if(length < ldpi_sze)	/* Not enough room to generate */
			fatali("10");
		length -= ldpi_sze;
		ldpi_sze -= LDPI_SIZE;
		build_i((UINT) ldpi_sze,(UINT) DOTLDC,
			(SLONG) (constant - min_addr - (SLONG) LDPI_SIZE -
				(SLONG) ldpi_sze));
		build_i((UINT) LDPI_SIZE,(UINT) OPR,(SLONG) LDPI);
		while(length--)
			build_i((UINT) 1,(UINT) NOP,(SLONG) 0);
		}
/*
 *	Otherwise just build the instruction in the normal fashion.
 */
	else
		build_i(length,function,constant);
	}

/************************************************************************/
/*									*/
/*	Routine:	ldpi_length					*/
/*									*/
/*	Takes two possible immediate values for instructions which have	*/
/*	immediate data, computes the length of an instruction resulting	*/
/*	from each and updates the "min_addr" and "max_addr" address	*/
/*	counters appropriately.  The worst case of the two instruction	*/
/*	lengths is returned.  This routine is called from "pass2" to	*/
/*	"grow" the minimum instruction lengths for T_ADDR_OP records	*/
/*	where required.  Note that a running total of the amount we	*/
/*	have adjusted the "min_addr" address counter is kept to ensure	*/
/*	that the definitions of forward referenced symbols are adjusted	*/
/*	(via "adj_min_addr"), enough to prevent the "min_addr" -	*/
/*	"min_addr" term of the definition/reference difference from	*/
/*	changing sign.							*/
/*									*/
/*	Parameters:							*/
/*		"function" -	The instruction's basic opcode.		*/
/*		"min"	-	A SLONG which is used to compute the	*/
/*				minimum possible instruction length.	*/
/*		"max"	-	A SLONG which is used to compute the	*/
/*				maximum possible instruction length.	*/
/*		"minimum" -	The minimum instruction length to	*/
/*				generate.				*/
/*									*/
/*	Returns:	The "new" minimum instruction length.		*/
/*									*/
/************************************************************************/

UINT
ldpi_length(function,min,max,minimum)
	UINT			function;
	SLONG			min;
	SLONG			max;
	UINT			minimum;
	{
	UINT			ldpi_sze1;
	UINT			ldpi_sze2;
	UINT			length1;
	UINT			length2;
	UINT			max_length;
	UINT			min_length;

	if((function != DOTLDCR) && (function != DOTLDC))
		return (update_length(function,min,max,minimum,FALSE));

	length1 = ldpi_size(function,min,min_addr,&ldpi_sze1);
	length2 = ldpi_size(function,max,max_addr,&ldpi_sze2);

	if(length1 > length2)
		{
		max_length = length1;
		min_length = length2;
		}
	else
		{
		max_length = length2;
		min_length = length1;
		}
	if(max_length < minimum)
		max_length = minimum;
	if(min_length < minimum)
		min_length = minimum;
	if((minimum < max_length) &&
		(((function == DOTLDCR) && (minimum > 3)) ||
		((function != DOTLDCR) && (minimum != 1))))
		{
		fatali("11");
		}
	if(function == DOTLDCR)
		{
/*
 *	See if "ldc" data sign is still unknown (only for "ldc" before "ldpi").
 */
		if(((min - min_addr - length1) ^
			(max - max_addr - length2)) < 0)
			{
			if(minimum > 3)
				min_addr_adj += (max_length - minimum);
			else
				min_addr_adj += (max_length - 3);
			}
		else
			min_addr_adj += (max_length - min_length);
		}
	else
		min_addr_adj += (max_length - minimum);
/*
 *	Update addresses as appropriate.
 */
	min_addr += max_length;
	max_addr += max_length;

	return (max_length);
	}

/************************************************************************/
/*									*/
/*	Routine:	ldpi_size					*/
/*									*/
/*	Determine the length of an instruction with immediate data.	*/
/*	Used by "do_i_addr as a helper for T_ADDR_OP records.		*/
/*									*/
/*	Parameters:							*/
/*		"function" -	The basic instruction opcode.		*/
/*		"data"	-	The immediate data to scrutinize.	*/
/*		"cur_addr" -	Current address used for "ldc"/"ldpi"	*/
/*				offsetting.				*/
/*		"ldpi_sze" -	A pointer to a UINT type to be set to	*/
/*				the non-zero length of the "ldc"/"ldpi"	*/
/*				instruction length if that is the best	*/
/*				way.					*/
/*									*/
/*	Returns:	The instruction length in bytes.		*/
/*									*/
/************************************************************************/

UINT
ldpi_size(function,data,cur_addr,ldpi_sze)
	UINT			function;
	SLONG			data;
	SLONG			cur_addr;
	UINT			*ldpi_sze;
	{
	register	UINT	ldpi_tlen;
	UINT			rval;

	*ldpi_sze = 0;			/* Initialize to "ldc"/"ldpi" !used */
	if(function != DOTLDCR)
		{
		rval = i_size(function,data);
		if(function != DOTLDC)
			return (rval);
		}
	else
		rval = 100;		/* Infinity for instruction length */
/*
 *	The opcode was either a "DOTLDC" or "DOTLDCR".  In either case we need
 *	to evaluate the "ldc"/"ldpi" instruction as an alternative.
 */
	for(ldpi_tlen = 1; i_size(DOTLDC,
		(SLONG) (data - cur_addr - (SLONG) LDPI_SIZE -
			(SLONG) ldpi_tlen)) > ldpi_tlen; ldpi_tlen++)
		;
	ldpi_tlen += LDPI_SIZE;
	if(ldpi_tlen < rval)
		*ldpi_sze = rval = ldpi_tlen;

	return (rval);
	}

/************************************************************************/
/*									*/
/*	Routine:	pass0						*/
/*									*/
/*	Makes a pass through the temp files recording the minimum and	*/
/*	maximum addresses for the symbols.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#pragma	optimize("g",off)		/* Disable global common subs */
#pragma	optimize("l",off)		/* Disable loop optimizations */
#endif	/* MSC_OPT */

VOID
pass0()
	{
	SLONG			big_temp;
	BOOL			notdone;
	register	UINT	temp;

#if	DEBUG
	putchar('0');			/* Indicate doing pass 0 */
#endif
/*
 *	Initialize for this pass.
 */
	min_addr = max_addr = load_addr;	/* Default load address */
	region = 0;			/* Initialize code/data region tag */
/*
 *	Loop through all the records in all the module temp files.  Do
 *	whatever is needed for each different record type.
 */
	for(module = 0; module < NUMMODULES; module++)
		{
		if(mod_fp[module] != NULL)	/* If something in module */
			{
			swap_temp();	/* Swap module & extra temp file */
			notdone = TRUE;	/* Until we need to find next module */
			while(notdone)
				{
				switch(temp = GET_1B())
					{
/* Normal data definition */		case T_DATA:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					put_2b((int) (temp = get_2b()));
					copy_temp(temp);
					min_addr += temp;
					max_addr += temp;
					break;

/* Symbolic PC rel. instruction */	case T_RELSYM_OP:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					PUT_1B((int) (temp = GET_1B()));
					copy_temp(7);	/* Rest of record */
					min_addr += temp;
					max_addr += max_i_size;
					break;

/* Label definition */			case T_DEF:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					put_2b((int) (temp = get_2b()));
					put_min(temp,min_addr);
					put_max(temp,max_addr);
					put_region(temp);
					break;

/* Word alignment pseudo-op */		case T_ALIGN:
					addr_pad();	/* Update addr. cnts */
					break;

/* Reserve space pseudo-op */		case T_STORAGE:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					put_4b((SLONG) (big_temp = get_4b()));
					min_addr += big_temp;
					max_addr += big_temp;
					break;

/* "rel-rel" reference instruction */	case T_RELREL_OP:
/* Scaled "rel-rel" ref. instr. */	case T_WRELREL_OP:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					PUT_1B((int) (temp = GET_1B()));
					copy_temp(9);	/* Syms/off./opcode */
					min_addr += temp;
					max_addr += max_i_size;
					break;

/* Address ref. in data */		case T_ADDR_DATA:
					PUT_1B((int) temp);
					copy_temp(8);	/* Ln#, ST #, offset */
					min_addr += (SLONG) word_size;
					max_addr += (SLONG) word_size;
					break;

/* Address reference in instruction */	case T_ADDR_OP:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					PUT_1B((int) (temp = GET_1B()));
					min_addr += temp;
					copy_temp(6);	/* Rest of record */
					PUT_1B((int) (temp = GET_1B()));
					if(temp == DOTLDCR)
						{
						max_addr +=
							(max_i_size +
								LDPI_SIZE);
						}
					else
						max_addr += max_i_size;
					break;

/* "rel-rel" reference in data */	case T_RELREL_DATA:
					PUT_1B((int) temp);
					copy_temp(10);	/* ln#, syms, offset */
					min_addr += (SLONG) word_size;
					max_addr += (SLONG) word_size;
					break;

/* Non-symbolic PC relative instr. */	case T_REL_OP:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					PUT_1B((int) (temp = GET_1B()));
					copy_temp(5);	/* Rest of record */
					min_addr += temp;
					max_addr += max_i_size;
					break;

/* Filename record */			case T_FILENAME:
					PUT_1B((int) temp);
					copy_temp(4);	/* Old & new ln #'s */
					PUT_1B((int) (temp = GET_1B()));
					copy_temp(temp);
					break;

/* Symbolic PC relative ref. in data */	case T_RELSYM_DATA:
					PUT_1B((int) temp);
					copy_temp(8);	/* ln#, sym, offset */
					min_addr += (SLONG) word_size;
					max_addr += (SLONG) word_size;
					break;

/* PC relative address ref. in data */	case T_REL_DATA:
					PUT_1B((int) temp);
					copy_temp(6);	/* ln#, address */
					min_addr += (SLONG) word_size;
					max_addr += (SLONG) word_size;
					break;

/* Non-symbolic debug data record */	case T_DEBUG_DATA:
					PUT_1B((int) temp);
					copy_temp(6);	/* Line #, value */
					put_2b((int) (temp = get_2b()));
					copy_temp(temp);	/* Data */
					break;

/* Symbolic debug data record */	case T_DEBUGSYM_DATA:
					PUT_1B((int) temp);
					copy_temp(6);	/* Line#, offset */
					put_2b((int) (temp = get_2b()));
					copy_temp(temp + 2);	/* ST#, data */
					break;

/* Set load address instruction */	case T_LOAD:
					region++;
					PUT_1B((int) temp);
					put_4b((SLONG) (min_addr = get_4b()));
					max_addr = min_addr;
					break;

/* Found EOF record */			case T_EOF:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					notdone = FALSE;
					break;

					default:
					fatal("Corrupted temporary file","");
					}
				}
			}
		}
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#pragma	optimize("g",on)		/* Enable global common subs */
#pragma	optimize("l",on)		/* Enable loop optimizing */
#endif	/* MSC_OPT */

/************************************************************************/
/*									*/
/*	Routine:	pass1						*/
/*									*/
/*	Makes a pass through the temp files converting whatever PC	*/
/*	relative references which can be bound.  It updates the symbol	*/
/*	table addresses as the actual instruction lengths are		*/
/*	determined more accurately.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if no additional reference records could	*/
/*			be converted in this pass, otherwise FALSE (so	*/
/*			that "pass1" will be called again to iterate	*/
/*			towards the final result).			*/
/*									*/
/************************************************************************/

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#pragma	optimize("g",off)		/* Disable global common subs */
#pragma	optimize("l",off)		/* Disable loop optimizations */
#endif	/* MSC_OPT */

BOOL
pass1()
	{
	SLONG			big_temp;
	BOOL			min_size;
	BOOL			notdone;
	register	UINT	temp;

#if	DEBUG
	putchar('1');			/* Indicate doing pass 1 */
#endif
/*
 *	Initialize for this pass.
 */
	min_addr = max_addr = load_addr;	/* Default load address */
	max_addr_adj = min_addr_adj = 0;	/* Fwd reference min/max pad */
	region = 0;			/* Initialize code/data region tag */
	min_size = TRUE;		/* FALSE if we reduce something */
/*
 *	Loop through all the records in all the module temp files.  Do
 *	whatever is needed for each different record type.
 */
	for(module = 0; module < NUMMODULES; module++)
		{
		if(mod_fp[module] != NULL)	/* If something in module */
			{
			swap_temp();	/* Swap module & extra temp file */
			notdone = TRUE;	/* Until we need to find next module */
			while(notdone)
				{
				switch(temp = GET_1B())
					{
/* Normal data definition */		case T_DATA:
					PUT_1B((int) temp);
/* Line # */				PUT_1B(GET_1B()); PUT_1B(GET_1B());
					put_2b((int) (temp = get_2b()));
					copy_temp(temp);	/* Real data */
					min_addr += temp;
					max_addr += temp;
					break;

/* Symbolic PC rel. instruction */	case T_RELSYM_OP:
					if(dors_i_ref())
						min_size = FALSE;
					break;

/* Label definition */			case T_DEF:
					PUT_1B((int) temp);
/* Line # */				PUT_1B(GET_1B()); PUT_1B(GET_1B());
					put_2b((int) (temp = get_2b()));
					if((get_min(temp) != min_addr)
						|| (get_max(temp) != max_addr))
						{
						min_size = FALSE;
/* Adjust fwd reference fudge factors */	min_addr_adj = min_addr -
							get_min(temp);
						max_addr_adj = get_max(temp) -
							max_addr;
						put_min(temp,min_addr);
						put_max(temp,max_addr);
						}
					break;

/* Word alignment pseudo-op */		case T_ALIGN:
					if(addr_pad())
						min_size = FALSE;
					break;

/* Reserve space pseudo-op */		case T_STORAGE:
					PUT_1B((int) temp);
/* Line # */				PUT_1B(GET_1B()); PUT_1B(GET_1B());
					put_4b((SLONG) (big_temp = get_4b()));
					min_addr += big_temp;
					max_addr += big_temp;
					break;

/* "rel-rel" reference instruction */	case T_RELREL_OP:
					if(do_i_diff())
						min_size = FALSE;
					break;

/* Scaled "rel-rel" ref instruction */	case T_WRELREL_OP:
					if(do_i_wdiff())
						min_size = FALSE;
					break;

/* Address reference in data */		case T_ADDR_DATA:
					do_addr();
					break;

/* Address reference in instruction */	case T_ADDR_OP:
					if(do_i_addr())
						min_size = FALSE;
					break;

/* "rel-rel" reference in data */	case T_RELREL_DATA:
					do_diff();
					break;

/* Non-symbolic PC relative instr. */	case T_REL_OP:
					if(dor_i_ref())
						min_size = FALSE;
					break;

/* Filename record */			case T_FILENAME:
					PUT_1B((int) temp);
					copy_temp(4);	/* Old & new ln #'s */
					PUT_1B((int) (temp = GET_1B()));
					copy_temp(temp);
					break;

/* Symbolic PC relative ref. in data */	case T_RELSYM_DATA:
					dors_ref();
					break;

/* PC relative address ref. in data */	case T_REL_DATA:
					dor_ref();
					break;

/* Non-symbolic debug data record */	case T_DEBUG_DATA:
					PUT_1B((int) temp);
					copy_temp(6);	/* Line #, value */
					put_2b((int) (temp = get_2b()));
					copy_temp(temp);	/* Data */
					break;

/* Symbolic debug data record */	case T_DEBUGSYM_DATA:
					debug_addr();
					break;

/* Set load address instruction */	case T_LOAD:
					region++;
					PUT_1B((int) temp);
					put_4b((SLONG) (big_temp = get_4b()));
					min_addr = big_temp;
					max_addr = big_temp;
					max_addr_adj = min_addr_adj = 0;
					break;

/* Found EOF record */			case T_EOF:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					notdone = FALSE;
					break;

					default:
					fatal("Corrupted temporary file","");
					}
				}
			}
		}
	return (min_size);
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#pragma	optimize("g",on)		/* Enable global common subs */
#pragma	optimize("l",on)		/* Enable loop optimizing */
#endif	/* MSC_OPT */

/************************************************************************/
/*									*/
/*	Routine:	pass2						*/
/*									*/
/*	Makes a pass through the temp files increasing the minimum size	*/
/*	of all variable length instructions which aren't currently	*/
/*	listed as being of maximum instruction length.  This routine is	*/
/*	called once during the linkage process to resolve mutually	*/
/*	dependent variable length instructions by increasing the size	*/
/*	of ALL the boundary cases (which means all the cases left when	*/
/*	"pass2" is called since the others were taken care of by	*/
/*	iterating "pass1").  In the best of all possible worlds this	*/
/*	process would only increase the size of one instruction at a	*/
/*	time and examine the effects, however this could potentially	*/
/*	cause the temp files to be processed once for each unbound	*/
/*	instruction (or even worse if the optimum instruction to be	*/
/*	grown is chosen).  In the interest of reasonable execution	*/
/*	speed all instructions are handled together with the		*/
/*	understanding that a minor amount of unnecessary code might	*/
/*	result (prefix instructions execute rapidly).			*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#pragma	optimize("g",off)		/* Disable global common subs */
#pragma	optimize("l",off)		/* Disable loop optimizations */
#endif	/* MSC_OPT */

VOID
pass2()
	{
	SLONG			big_temp;
	register	UINT	function;
	UINT			lsym;
	register	UINT	min_length;
	BOOL			notdone;
	UINT			rsym;
	register	UINT	temp;

#if	DEBUG
	putchar('2');			/* Indicate doing pass 2 */
#endif
/*
 *	Initialize for this pass.
 */
	min_addr = max_addr = load_addr;	/* Default load address */
	max_addr_adj = min_addr_adj = 0;	/* Fwd reference minimum pad */
	region = 0;			/* Initialize code/data region tag */
/*
 *	Loop through all the records in all the module temp files.  Do
 *	whatever is needed for each different record type.
 */
	for(module = 0; module < NUMMODULES; module++)
		{
		if(mod_fp[module] != NULL)	/* If something in module */
			{
			swap_temp();	/* Swap module & extra temp file */
			notdone = TRUE;	/* Until we need to find next module */
			while(notdone)
				{
				switch(temp = GET_1B())
					{
/* Normal data definition */		case T_DATA:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					put_2b((int) (temp = get_2b()));
					copy_temp(temp);
					min_addr += temp;
					max_addr += temp;
					break;

/* Symbolic PC rel. instruction */	case T_RELSYM_OP:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					min_length = GET_1B();
					temp = get_2b();	/* Symbol # */
					big_temp = get_4b();	/* Offset */
					function = GET_1B();
					min_length = update_length(
						(UINT) function,
						(SLONG) (big_temp +
							adj_max_addr(temp) -
							max_addr),
						(SLONG) (big_temp +
							adj_min_addr(temp) -
							min_addr),
						min_length,TRUE);
					PUT_1B((int) min_length);
					put_2b((int) temp);
					put_4b((SLONG) big_temp);
					PUT_1B((int) function);	/* Opcode */
					break;

/* Label definition */			case T_DEF:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					put_2b((int) (temp = get_2b()));
					if((get_min(temp) != min_addr)
						|| (get_max(temp) !=
							max_addr))
						{
/* Adjust fwd reference fudge factors */	min_addr_adj = min_addr -
							get_min(temp);
						max_addr_adj = get_max(temp) -
							max_addr;
						put_min(temp,min_addr);
						put_max(temp,max_addr);
						}
					break;

/* Word alignment pseudo-op */		case T_ALIGN:
					addr_pad();	/* Update addr. cnts */
					break;

/* Reserve space pseudo-op */		case T_STORAGE:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					put_4b((SLONG) (big_temp = get_4b()));
					min_addr += big_temp;
					max_addr += big_temp;
					break;

/* "rel-rel" reference instruction */	case T_RELREL_OP:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					min_length = GET_1B();
					lsym = get_2b();	/* Left sym. */
					rsym = get_2b();	/* Right sym */
					big_temp = get_4b();	/* Offset */
					function = GET_1B();
					min_length = update_length(
						(UINT) function,
						(SLONG) (big_temp +
							adj_max_addr(lsym) -
							adj_max_addr(rsym)),
						(SLONG) (big_temp +
							adj_min_addr(lsym) -
							adj_min_addr(rsym)),
						min_length,FALSE);
					PUT_1B((int) min_length);
					put_2b((int) lsym);	/* Left sym. */
					put_2b((int) rsym);	/* Right sym */
					put_4b((SLONG) big_temp);
					PUT_1B((int) function);	/* Opcode */
					break;

/* Scaled "rel-rel" ref instruction */	case T_WRELREL_OP:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					min_length = GET_1B();
					lsym = get_2b();	/* Left sym. */
					rsym = get_2b();	/* Right sym */
					big_temp = get_4b();	/* Offset */
					function = GET_1B();
					min_length = update_length(
						(UINT) function,
						(SLONG) (big_temp +
							adj_max_addr(lsym) -
							adj_max_addr(rsym)) /
							(SLONG) word_size,
						(SLONG) (big_temp +
							adj_min_addr(lsym) -
							adj_min_addr(rsym)) /
							(SLONG) word_size,
						min_length,FALSE);
					PUT_1B((int) min_length);
					put_2b((int) lsym);	/* Left sym. */
					put_2b((int) rsym);	/* Right sym */
					put_4b((SLONG) big_temp);
					PUT_1B((int) function);	/* Opcode */
					break;

/* Address reference in data */		case T_ADDR_DATA:
					PUT_1B((int) temp);
					copy_temp(8);	/* Ln#, ST #, offset */
					min_addr += (SLONG) word_size;
					max_addr += (SLONG) word_size;
					break;

/* Address reference in instruction */	case T_ADDR_OP:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					min_length = GET_1B();
					temp = get_2b();	/* Symbol # */
					big_temp = get_4b();	/* Offset */
					function = GET_1B();
					min_length = ldpi_length(
						(UINT) function,
						(SLONG) (adj_min_addr(temp)+
							big_temp),
						(SLONG) (adj_max_addr(temp)+
							big_temp),min_length);
					PUT_1B((int) min_length);
					put_2b((int) temp);
					put_4b((SLONG) big_temp);
					PUT_1B((int) function);	/* Opcode */
					break;

/* "rel-rel" reference in data */	case T_RELREL_DATA:
					PUT_1B((int) temp);
					copy_temp(10);	/* Ln#, syms, offset */
					min_addr += (SLONG) word_size;
					max_addr += (SLONG) word_size;
					break;

/* Non-symbolic PC relative instr. */	case T_REL_OP:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					min_length = GET_1B();
					big_temp = get_4b();	/* Offset */
					function = GET_1B();
					min_length = update_length(
						(UINT) function,
						(SLONG) (big_temp - min_addr),
						(SLONG) (big_temp - max_addr),
						min_length,TRUE);
					PUT_1B((int) min_length);
					put_4b((SLONG) big_temp);
					PUT_1B((int) function);	/* Opcode */
					break;

/* Filename record */			case T_FILENAME:
					PUT_1B((int) temp);
					copy_temp(4);	/* Old & new ln #'s */
					PUT_1B((int) (temp = GET_1B()));
					copy_temp(temp);
					break;

/* Symbolic PC relative ref. in data */	case T_RELSYM_DATA:
					PUT_1B((int) temp);
					copy_temp(8);	/* Ln#, ST #, offset */
					min_addr += (SLONG) word_size;
					max_addr += (SLONG) word_size;
					break;

/* PC relative address ref. in data */	case T_REL_DATA:
					PUT_1B((int) temp);
					copy_temp(6);	/* Ln#, ST #, offset */
					min_addr += (SLONG) word_size;
					max_addr += (SLONG) word_size;
					break;

/* Non-symbolic debug data record */	case T_DEBUG_DATA:
					PUT_1B((int) temp);
					copy_temp(6);	/* Line #, value */
					put_2b((int) (temp = get_2b()));
					copy_temp(temp);	/* Data */
					break;

/* Symbolic debug data record */	case T_DEBUGSYM_DATA:
					PUT_1B((int) temp);
					copy_temp(6);	/* Line#, offset */
					put_2b((int) (temp = get_2b()));
					copy_temp(temp + 2);	/* ST#, data */
					break;

/* Set load address instruction */	case T_LOAD:
					region++;
					PUT_1B((int) temp);
					put_4b((SLONG) (big_temp = get_4b()));
					min_addr = big_temp;
					max_addr = big_temp;
					max_addr_adj = min_addr_adj = 0;
					break;

/* Found EOF record */			case T_EOF:
					PUT_1B((int) temp);
					copy_temp(2);	/* Line # */
					notdone = FALSE;
					break;

					default:
					fatal("Corrupted temporary file","");
					}
				}
			}
		}
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#pragma	optimize("g",on)		/* Enable global common subs */
#pragma	optimize("l",on)		/* Enable loop optimizing */
#endif	/* MSC_OPT */

/************************************************************************/
/*									*/
/*	Routine:	put_max						*/
/*									*/
/*	Update the "maximum" address associated with the local symbol	*/
/*	table entry parameter.						*/
/*									*/
/*	Parameters:							*/
/*		"index"	The local sybol table entry # to update.	*/
/*		"value"	The new value to update the address with.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
put_max(index,value)
	UINT			index;
	SLONG			value;
	{
	UINT			temp_index;

	if((temp_index = locals[index].gst_num) < MAX_GLOBALS)
		{
		if((index = globals[temp_index].lst_num) < MAX_SYMBOLS)
			{
			locals[index].max_addr = value;
			return;
			}
		else
			return;		/* Undefined, shouldn't happen */
		}
	locals[index].max_addr = value;
	}

/************************************************************************/
/*									*/
/*	Routine:	put_min						*/
/*									*/
/*	Update the "minimum" address associated with the local symbol	*/
/*	table entry parameter.						*/
/*									*/
/*	Parameters:							*/
/*		"index"	The local sybol table entry # to update.	*/
/*		"value"	The new value to update the address with.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
put_min(index,value)
	UINT			index;
	SLONG			value;
	{
	UINT			temp_index;

	if((temp_index = locals[index].gst_num) < MAX_GLOBALS)
		{
		if((index = globals[temp_index].lst_num) < MAX_SYMBOLS)
			{
			locals[index].min_addr = value;
			return;
			}
		else
			return;		/* Undefined, shouldn't happen */
		}
	locals[index].min_addr = value;
	}

/************************************************************************/
/*									*/
/*	Routine:	put_region					*/
/*									*/
/*	Update the region tag information associated with the local	*/
/*	symbol table entry parameter.  A helper routine for "pass0()".	*/
/*									*/
/*	Parameters:							*/
/*		"index"	The local sybol table entry # to update.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
put_region(index)
	UINT			index;
	{
	UINT			temp_index;

	if((temp_index = locals[index].gst_num) < MAX_GLOBALS)
		{
		if((index = globals[temp_index].lst_num) < MAX_SYMBOLS)
			{
			locals[index].region = region;
			return;
			}
		else	
			fatali("12");	/* Undefined, shouldn't happen */
		}
	locals[index].region = region;
	}

/************************************************************************/
/*									*/
/*	Routine:	update_addr					*/
/*									*/
/*	Takes two possible immediate values for instructions which have	*/
/*	immediate data, computes the length of an instruction resulting	*/
/*	from each and update the "min_addr" and "max_addr" address	*/
/*	counters appropriately.						*/
/*									*/
/*	Parameters:							*/
/*		"function" -	The opcode being processed.		*/
/*		"data1"	-	A SLONG which is used to compute the	*/
/*				first possible instruction length.	*/
/*		"data2"	-	A SLONG which is used to compute the	*/
/*				second possible instruction length.	*/
/*		"min_length"	The minimum instruction length to	*/
/*				generate.				*/
/*		"rel" -		TRUE if the immediate data is relative	*/
/*				to the PC of the instruction being	*/
/*				generated (i.e. does the instruction	*/
/*				length effect the data).		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
update_addr(function,data1,data2,min_length,rel)
	UINT			function;
	SLONG			data1;
	SLONG			data2;
	UINT			min_length;
	BOOL			rel;
	{
	UINT			length1;
	UINT			length2;

	if(rel)
		{
		length1 = bind_constant((UINT) function,(SLONG) data1);
		length2 = bind_constant((UINT) function,(SLONG) data2);
/*
 *	See if possibly either a forward or backward reference.
 */
		if(((data1 - (SLONG) length1) ^ (data2 - (SLONG) length2)) < 0)
			{
			min_addr += min_length;
			if(length1 > length2)
				{
				if(length1 < min_length)
					max_addr += min_length;
				else
					max_addr += length1;
				}
			else
				{
				if(length2 < min_length)
					max_addr += min_length;
				else
					max_addr += length2;
				}
			return;
			}
		}
	else
		{
		length1 = i_size((UINT) function,(SLONG) data1);
		length2 = i_size((UINT) function,(SLONG) data2);
/*
 *	See if instruction immediate data could be either positive or negative.
 */
		if((data1 ^ data2) < 0)
			{
			min_addr += min_length;
			if(length1 > length2)
				{
				if(length1 < min_length)
					max_addr += min_length;
				else
					max_addr += length1;
				}
			else
				{
				if(length2 < min_length)
					max_addr += min_length;
				else
					max_addr += length2;
				}
			return;
			}
		}
/*
 *	Otherwise we know the direction of the reference, or the sign of the
 *	immediate data.
 */
	if(length1 > length2)
		{
		if(length2 < min_length)
			min_addr += min_length;
		else
			min_addr += length2;
		if(length1 < min_length)
			max_addr += min_length;
		else
			max_addr += length1;
		}
	else
		{
		if(length1 < min_length)
			min_addr += min_length;
		else
			min_addr += length1;
		if(length2 < min_length)
			max_addr += min_length;
		else
			max_addr += length2;
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	update_length					*/
/*									*/
/*	Takes two possible immediate values for instructions which have	*/
/*	immediate data, computes the length of an instruction resulting	*/
/*	from each and updates the "min_addr" and "max_addr" address	*/
/*	counters appropriately.  The worst case of the two instruction	*/
/*	lengths is returned (which will be 1 greater than the best	*/
/*	case).  This routine is called from "pass2" to "grow" the	*/
/*	minimum instruction lengths where required.  Note that a	*/
/*	running total of the amount we have adjusted the "min_addr"	*/
/*	address counter is kept to ensure that the definitions of	*/
/*	forward referenced symbols are adjusted (via			*/
/*	"adj_min_addr"), enough to prevent the "min_addr-min_addr"	*/
/*	term of the definition/reference difference from changing sign.	*/
/*									*/
/*	Parameters:							*/
/*		"function" -	The instruction's basic opcode.		*/
/*		"data1"	-	A SLONG which is used to compute the	*/
/*				first possible instruction length.	*/
/*		"data2"	-	A SLONG which is used to compute the	*/
/*				second possible instruction length.	*/
/*		"min_length" -	The minimum instruction length to	*/
/*				generate.				*/
/*		"rel" -		TRUE if the immediate data is relative	*/
/*				to the PC of the instruction being	*/
/*				generated (i.e. does the instruction	*/
/*				length effect the data).		*/
/*									*/
/*	Returns:	The "new" minimum instruction length.		*/
/*									*/
/************************************************************************/

UINT
update_length(function,data1,data2,min_length,rel)
	UINT			function;
	SLONG			data1;
	SLONG			data2;
	UINT			min_length;
	BOOL			rel;
	{
	UINT			length1;
	UINT			length2;

	if(rel)
		{
		length1 = bind_constant((UINT) function,(SLONG) data1);
		length2 = bind_constant((UINT) function,(SLONG) data2);
/*
 *	See if possibly either a forward or backward reference.
 */
		if(((data1 - (SLONG) length1) ^ (data2 - (SLONG) length2)) < 0)
			{
			if((length1 == length2) && (min_length < length1))
				min_addr_adj++;
			}
		}
	else
		{
		length1 = i_size((UINT) function,(SLONG) data1);
		length2 = i_size((UINT) function,(SLONG) data2);
/*
 *	See if instruction immediate data could be either positive or negative.
 */
		if((data1 ^ data2) < 0)
			{
			if((length1 == length2) && (min_length < length1))
				min_addr_adj++;
			}
		}
	if(length1 > length2)
		{
		if(length1 > min_length)
			{
			if(min_length != 1)
				fatali("1");
			min_length = length1;
			min_addr_adj++;		/* Adj for fwd definitions */
			}
		min_addr += min_length;
		max_addr += min_length;
		}
	else
		{
		if(length2 > min_length)
			{
			if(min_length != 1)
				fatali("2");
			min_length = length2;
			if(length1 < length2)
				min_addr_adj++;
			}
		min_addr += min_length;
		max_addr += min_length;
		}
	return (min_length);
	}
