/*
 *	OSU Research Computing					Trillium
 *	(third party software provided under license)
 *
 *	$Header: tcc.c,v 2.2.1.1 92/05/18 16:29:22 raja Exp $
 *	We are just putting our stamp on this for housekeeping reasons.
 */

/************************************************************************/
/*									*/
/*	File:	tcc.c							*/
/*									*/
/*	Transputer Front End (TCC).					*/
/*									*/
/*	Written by Anthony Clem of Computer Specific Consulting.	*/
/*	Modified by Kirk Bailey of Logical Systems.			*/
/*									*/
/*		Copyright (c) 1991 by Logical Systems.			*/
/*									*/
/*				10/16/91				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard I/O include file	*/
#include	<ctype.h>		/* Character type functions	*/
#include	<string.h>		/* String functions		*/
#include	<time.h>		/* Time related functions	*/
#include	"taldef.h"		/* special definitions		*/

#if	H_OPSYS == MS_DOS
#include	<stdlib.h>		/* Std library include file	*/
#include	<process.h>		/* for spawning processes	*/
#define STRDIR	"\\"
#else	/* !MS_DOS */
#include	<sys/wait.h>
#define STRDIR	"/"
#endif	/* MS_DOS */

#ifdef OSU
#if	H_OPSYS == BSD
#include	<stdlib.h>
#include	<sys/wait.h>
#endif	/* BSD */
#endif	/* OSU */

/************************************************************************/
/*									*/
/*			Global Data Definitions				*/
/*									*/
/************************************************************************/

#define	TCC_ENV_VARIABLE	"TCC"	/* Name of environment variable	*/

/*
 *	There is one of these structures for each item to be passed to any of
 *	of the tools TCC exec's, as well as for each input file TCC is to
 *	process.
 */
struct	args
	{
	STRING		data;		/* Points to the data item	*/
	struct	args	*next;		/* Next element in the list	*/
	};

/*
 *	This structure is a combination of queue overhead information, together
 *	with "argc"/"argv" to be used when calling the associated tool.
 */
struct	targs
	{
	int		argc;		/* Argument count for cmd line	*/
	char		**argv;		/* Pointer to pointer to args	*/
	struct	args	*head;		/* Argument queue head pointer	*/
	struct	args	*tail;		/* Argument queue tail pointer	*/
	};

/*
 *	This structure holds the required information about a temporary file.
 */
struct	tfile
	{
	STREAM		fp;		/* Stream pointer for open file	*/
	char		name[FNSIZE+1];	/* File name			*/
	};

STRING		address	= NULL;		/* Specify initial PC		*/
BOOL		debug	= FALSE;	/* TRUE/enable debug info	*/
struct	targs	eargs;			/* Environment variable args	*/
STRING		entry	= NULL;		/* Program entrypoint info	*/
struct	targs	files;			/* List of input files		*/
struct	tfile	file1;			/* TCX input temp file		*/
struct	tfile	file2;			/* TCX output/TASM input file	*/
struct	targs	lflags;			/* TLNK flags			*/
STRING		lib_default = "t4lib";	/* Default library if not specd	*/
struct	targs	libs;			/* List of library files	*/
STRING		map	= NULL;		/* Map output filename pointer	*/
BOOL		no_link	= FALSE;	/* TRUE/no linking required	*/
STRING		output	= NULL;		/* Output filename pointer	*/
struct	targs	pp;			/* Arguments for PP		*/
BOOL		pp_only	= FALSE;	/* TRUE/only preprocess input	*/
struct	targs	tasm;			/* Arguments for TASM		*/
char		tempdir[FNSIZE + 1];	/* Directory for "temp" files	*/
UINT		temptime;		/* Holds "temp" file time seed	*/
struct	targs	tcx;			/* Arguments for TCX		*/
BOOL		tcx_only = FALSE;	/* TRUE/only compile input	*/
struct	targs	tlnk;			/* Arguments for TLNK		*/
int		verbose = VERBOSE;	/* TRUE/be verbose to user	*/
STRING		workspace = NULL;	/* Specify initial workspace	*/

#ifdef OSU
int		pretend = FALSE;	/* Do not execute commands	*/
int		startup = TRUE;		/* Link Trollius startup code	*/
struct	targs	tcvt;			/* Arguments for tcvt		*/
struct	targs	libdir;			/* Lib. directories to search	*/
struct	args	*plib;			/* ptr for lib. dir. access	*/
STRING		tlibenv = NULL;		/* memory for TLIB env variable */
STRING		incdir = NULL;		/* LSTT include directory	*/
STRING		incdir2 = NULL;		/* Trollius include directory	*/
STRING		pscratch;		/* used with the scratch buffer */
char		scratch[256];		/* Scratch buffer space		*/
#endif

/************************************************************************/
/*									*/
/*			Function Type Declarations			*/
/*									*/
/************************************************************************/

VOID	add_arg();
VOID	argify();
VOID	fatal();
VOID	name_temp();
STRING	next_arg();
VOID	parse_cpu();
char	**parse_env();
VOID	process_files();
VOID	run_pp();
VOID	run_tasm();
VOID	run_tcx();
VOID	run_tlnk();
VOID	run_tool();
STRING	swap_ext();
VOID	terminate();
VOID	usagerr();

/************************************************************************/
/*									*/
/*	Routine:	main						*/
/*									*/
/*	deciphers command line, and runs appropriate utilities for	*/
/*	given files (including flags).					*/
/*									*/
/*	Parameters:							*/
/*		argc -	An integer count of the command line arguments.	*/
/*		argv -	A pointer to an array of pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

int
main(argc,argv)
	int			argc;
	char			**argv;
	{
	BOOL			cpu_type	= FALSE;
	char			**eargv;
	BOOL			had_v		= FALSE;
	int			i;
	STRING			ptr;
	char			***targv;
	STRING			tptr;
/*
 *	Initialize argument structures for the various tools.
 */
	eargs.argc = files.argc = pp.argc = lflags.argc = libs.argc =
		tasm.argc = tcx.argc = tlnk.argc = 0;
	eargs.argv = files.argv = pp.argv = lflags.argv = libs.argv =
		tasm.argv = tcx.argv = tlnk.argv = NULL;
	eargs.head = files.head = pp.head = lflags.head = libs.head =
		tasm.head = tcx.head = tlnk.head = NULL;
	eargs.tail = files.tail = pp.tail = lflags.tail = libs.tail =
		tasm.tail = tcx.tail = tlnk.tail = NULL;
#ifdef OSU
	tcvt.argc = libdir.argc = 0;
	tcvt.argv = libdir.argv = NULL;
	tcvt.head = libdir.head = NULL;
	tcvt.tail = libdir.tail = NULL;
#endif
/*
 *	Initialize the temporary file data structures.
 */
	file1.fp = file2.fp = NULL;
	file1.name[0] = file2.name[0] = '\0';
/*
 *	Process possible arguments associated with the TCC environment
 *	variable.
 */
	eargv = parse_env();
/*
 *	Process environment and command line switches/arguments.
 */
	if(argc < 2)
		usagerr();		/* Needs at least one file argument */
	++argv;				/* Skip TCC program name */

	while((**(targv = &eargv) != NULL) || (**(targv = &argv) != NULL))
		{
		ptr = *(*targv)++;
		if(*ptr == '-')
			{
			while(*++ptr != '\0')
				{
				switch((int) *ptr)
					{
#ifndef OSU
/* Specify program load address */	case 'A':
					case 'a':
						if(address != NULL)
							{
	fatal("Only one load address is permitted","");
							}
						address = next_arg(targv,&ptr);
						break;

/* Do not link */			case 'C':
					case 'c':
						if((output != NULL) ||
							pp_only || tcx_only)
							{
	fatal("-C flag may not be used with -O, -P, or -S flags","");
							}
#else
					case 'c':
						if(pp_only || tcx_only)
							{
	fatal("-c flag may not be used with -o, -E, or -S flags","");
							}
#endif	/* OSU */
						if(no_link)
							{
#ifdef OSU
	fatal("Duplicate -c flag","");
#else
	fatal("Duplicate -C flag","");
#endif
							}
						no_link = TRUE;
						break;

/* Define macro in PP */		case 'D':
#ifndef OSU
					case 'd':
						add_arg(&pp,"-d");
#else
						add_arg(&pp,"-D");
#endif
						add_arg(&pp,next_arg(targv,
							&ptr));
						break;

#ifndef OSU
/* Specify program entrypoint */	case 'E':
					case 'e':
						if(entry != NULL)
							{
	fatal("Only one program entry point is permitted","");
							}
						entry = next_arg(targv,&ptr);
						break;

					case 'G':
					case 'g':
#else
					case 'd':
#endif	/* OSU */
						if(debug)
							{
	fatal("Duplicate -G flag","");
							}
						debug = TRUE;
						break;

/* Alternate path for include */	case 'I':
#ifndef OSU
					case 'i':
						add_arg(&pp,"-i");
#else
						add_arg(&pp,"-I");
#endif
						add_arg(&pp,next_arg(targv,
							&ptr));
						break;

#ifdef OSU
					case 'L':
						add_arg(&libdir,
							next_arg(targv, &ptr));
						break;

					case 'N':
						pretend = TRUE;
						break;

					case '0':
						startup = FALSE;
						break;

					case 'l':
						add_arg(&libs,next_arg(targv,
							&ptr));
						break;

					case 'O':
/*
 * Optimization turned on by default.
 */
						break;
#else	/* OSU */
/* Link using this library */		case 'L':
					case 'l':
						add_arg(&libs,next_arg(targv,
							&ptr));
						break;
/* Linker map output file name */	case 'M':
					case 'm':
						if(map != NULL)
							{
	fatal("Only one TLNK map file is permitted","");
							}
						map = next_arg(targv,&ptr);
						break;

/* TLNK output file name */		case 'O':
#endif	/* OSU */
					case 'o':
#ifdef OSU
						if(pp_only || tcx_only)
							{
	fatal("-o flag may not be used with -c, -E, or -S flags","");
							}
#else
						if(no_link || pp_only ||
							tcx_only)
							{
	fatal("-O flag may not be used with -C, -P, or -S flags","");
							}
#endif
						if(output != NULL)
							{
	fatal("Only one output file is permitted","");
							}
						output = next_arg(targv,&ptr);
						break;

#ifdef OSU
					case 'E':
#else
/* Preprocess source files only */	case 'P':
					case 'p':
#endif
						if(no_link ||
							(output != NULL) ||
								tcx_only)
							{
#ifdef OSU
	fatal("-E flag may not be used with -c, -o, or -S flags","");
#else
	fatal("-P flag may not be used with -C, -O, or -S flags","");
#endif
							}
						if(pp_only)
							{
#ifdef OSU
	fatal("Duplicate -E flag","");
#else
	fatal("Duplicate -P flag","");
#endif
							}
						pp_only = TRUE;
						break;

/* Halt after compiling files */	case 'S':
#ifndef OSU
					case 's':
#endif
						if(tcx_only)
							{
	fatal("Duplicate -S flag","");
							}
						tcx_only = TRUE;
						break;

#ifndef OSU
					case 'T':
					case 't':
						if(cpu_type)
							{
	fatal("Duplicate -T flag","");
							}
						cpu_type = TRUE;
						parse_cpu(targv,&ptr);
						break;

/* Undefine macro in PP */		case 'U':
					case 'u':
#else
					case 'U':
#endif
						add_arg(&pp,"-u");
						add_arg(&pp,next_arg(targv,
							&ptr));
						break;

#ifndef OSU
/* Make everything verbose */		case 'V':
#endif
					case 'v':
						if(had_v)
							{
#ifdef OSU
	fatal("Duplicate -v flag","");
#else
	fatal("Duplicate -V flag","");
#endif
							}
						had_v = TRUE;
						verbose = !verbose;
#ifndef OSU
						add_arg(&pp,"-v");
						add_arg(&tcx,"-v");
						add_arg(&tasm,"-v");
#endif
						break;

#ifndef OSU
/* Specify program workspace pointer */	case 'W':
					case 'w':
						if(workspace != NULL)
							{
	fatal("Only one initial workspace is permitted","");
							}
						workspace = next_arg(targv,
							&ptr);
						break;
#endif	/* OSU */

					default:
						usagerr();
					}
				}
			}
		else  if(*ptr == '+')
			{
			switch((int) *++ptr)
				{
#ifndef OSU
/* Pass flags to PP */		case 'P':
#endif
				case 'p':
					add_arg(&pp,next_arg(targv,&ptr));
					break;

#ifndef OSU
/* Pass flags to TCX */		case 'C':
#endif
				case 'c':
					add_arg(&tcx,next_arg(targv,&ptr));
					break;

#ifndef OSU
/* Pass flags to TASM */	case 'A':
#endif
				case 'a':
					add_arg(&tasm,next_arg(targv,&ptr));
					break;

#ifndef OSU
/* Pass flags to TLNK */	case 'L':
#endif
				case 'l':
					add_arg(&lflags,next_arg(targv,&ptr));
					break;

				default:
					fatal("Invalid '+' flag: ",ptr);
				}
			}
		else
			{
/*
 *	Add input files to the list to be processed.
 */
			add_arg(&files,ptr);
			}
		}
/*
 *	Add the debug information flags if needed.
 */
	if(! debug)
		{
		add_arg(&tcx,"-c");
		add_arg(&tasm,"-c");
		add_arg(&lflags,"c");
		}
#ifdef OSU
/*
 * Add some Trollius specific flags for the tools.
 */
/*
 * Add the Trollius include directory.  It is either
 * $(TROLLIUSHOME)/h or the default INCDIR2.
 */
	if ((pscratch = getenv("TROLLIUSHOME")) != NULL)
		{
		(void) strcpy(scratch, pscratch);
		(void) strcat(scratch, STRDIR);
		(void) strcat(scratch, "h");
		if ((incdir2 = malloc(strlen(scratch) + 1)) == NULL)
			fatal("Insufficient memory for Trollius include", "");
		(void) strcpy(incdir2, scratch);
		add_arg(&pp, "-I");
		add_arg(&pp, incdir2);
		}
	else
		{
		add_arg(&pp, "-I");
		add_arg(&pp, INCDIR2);
		}
/*
 * Add the LSTT include directory.  It is either $(LSTTHOME)/h/lstt or INCDIR.
 */
	if ((pscratch = getenv("LSTTHOME")) != NULL)
		{
		(void) strcpy(scratch, pscratch);
		(void) strcat(scratch, STRDIR);
		(void) strcat(scratch, "h");
		(void) strcat(scratch, STRDIR);
		(void) strcat(scratch, "lstt");
		if ((incdir = malloc(strlen(scratch) + 1)) == NULL)
			fatal("Insufficient memory for LSTT include", "");
		(void) strcpy(incdir, scratch);
		add_arg(&pp, "-I");
		add_arg(&pp, incdir);
		}
	else
		{
		add_arg(&pp, "-I");
		add_arg(&pp, INCDIR);
		}
/*
 * Add the compiler flags.
 */
	add_arg(&tcx, "-s");
	add_arg(&tcx, "-c");
	add_arg(&tcx, "-p8");
	add_arg(&tcx, "-w0");
/*
 * Add the assemble flags.
 */ 
	add_arg(&tasm,"-c");
/*
 * Add the linker flags.
 */
	add_arg(&lflags,"cs");

	if (address == NULL)
		address = LOADPT;
/*
 *	Add the Trollius default libraries at the end.
 */
	if (startup)
		{
		add_arg(&libs, STDLIB3);
		add_arg(&libs, STDLIB2);
		add_arg(&libs, STDLIB);
		}
/*
 *	Modify the TLIB environment variable.
 */
	strcpy(scratch, "TLIB=");
	for (plib = libdir.head; plib; plib = plib->next)
		{
		if (plib->data)
			{
			strcat(scratch, plib->data);
			strcat(scratch, ";");
			}
		}
	if ((pscratch = getenv("TROLLIUSHOME")) != NULL)
		{
		strcat(scratch, pscratch);
		strcat(scratch, STRDIR);
		strcat(scratch, "itb");
		strcat(scratch, STRDIR);
		strcat(scratch, "lib");
		strcat(scratch, STRDIR);
		strcat(scratch, "lstt");
		strcat(scratch, ";");
		}
	else
		{
		strcat(scratch, LIBDIR2);
		strcat(scratch, ";");
		}
	if ((pscratch = getenv("LSTTHOME")) != NULL)
		{
		strcat(scratch, pscratch);
		strcat(scratch, STRDIR);
		strcat(scratch, "itb");
		strcat(scratch, STRDIR);
		strcat(scratch, "lib");
		strcat(scratch, STRDIR);
		strcat(scratch, "lstt");
		strcat(scratch, ";");
		}
	else
		{
		strcat(scratch, LIBDIR);
		strcat(scratch, ";");
		}
		
	if ((tlibenv = malloc(strlen(scratch) + 1)) == NULL)
		fatal("Insufficient memory for TLIB", "");
	(void) strcpy(tlibenv, scratch);
	putenv(tlibenv);

#else	/* OSU */
/*
 *	Add the default library to the library list (at the end).
 */
	add_arg(&libs,lib_default);
#endif	/* OSU */
/*
 *	Go through and create skeleton "argc"/"argv" structures associated
 *	with the tool command lines which will be issued.
 */
	argify(&pp);
	pp.argv[0] = "pp";

	argify(&tcx);
	tcx.argv[0] = "tcx";

	argify(&tasm);
	tasm.argv[0] = "tasm";

	argify(&tlnk);
	tlnk.argv[0] = "tlnk";
/*
 *	Determine suitable names for the required temporary files.
 */
	if(! pp_only)
		{
/*
 *	Get a time stamp to use in conjuring temp file names.
 */
		temptime = (UINT) time(NULL);
/*
 *	See if the user has specified a special "temp" directory.  If so, make
 *	sure it is the correct format.
 */
		tempdir[0] = '\0';	/* No distinct temp directory name */
#ifdef OSU
		if( ((tptr = getenv(TEMP_PATH)) != NULL) ||
			((tptr = TEMP) != NULL) )
#else
		if((tptr = getenv(TEMP_PATH)) != NULL)
#endif	/* OSU */
			{
			if(strlen(tptr) < FNSIZE)
				{
				strcpy(tempdir,tptr);
/*
 *	Ensure temp directory has trailing delimiter.
 */
				if(((i = strlen(tempdir)) > 0) &&
#if	H_OPSYS == MPW
					(tempdir[i - 1] != ':'))
					{
					strcat(tempdir,":");
					}
#else	/* !MPW */
#if	H_OPSYS == MS_DOS
					(tempdir[i - 1] != '/') &&
					(tempdir[i - 1] != '\\'))
					{
					strcat(tempdir,"\\");
					}
#else	/* !MS_DOS */
					(tempdir[i - 1] != '/'))
					{
					strcat(tempdir,"/");
					}
#endif	/* MS_DOS */
#endif	/* MPW */
				}
			}
		name_temp(&file1);
		if(! tcx_only)
			name_temp(&file2);
		}
/*
 *	Okay, we've parsed the whole mess, time to get to it!
 */
#ifndef OSU
	if(verbose)
		printf("%s, %s\n%s\n\n",TCC_NAME,VERSION,COPYRIGHT);
#endif

	process_files();

	terminate(NOERRORS);		/* Head for the barn */
	return (0);
	}

/************************************************************************/
/*									*/
/*	Routine:	add_arg						*/
/*									*/
/*	Add the specified string to the argument list for the specified	*/
/*	tool.								*/
/*									*/
/*	Parameters:							*/
/*		tool -	A pointer to the arg structure for the tool.	*/
/*		s -	A pointer to the argument to add.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
add_arg(tool,s)
	struct	targs		*tool;
	STRING			s;
	{
	struct	args		*item;

	if((item = (struct args *) malloc(sizeof(struct args))) == NULL)
		fatal("Insufficient argument memory","");

	item->data = s;
	item->next = NULL;
/*
 *	Add the newly created argument structure to the queue.
 */
	if(tool->head == NULL)
		tool->head = tool->tail = item;
	else
		{
		(tool->tail)->next = item;
		tool->tail = item;
		}
	(tool->argc)++;
	}

/************************************************************************/
/*									*/
/*	Routine:	argify						*/
/*									*/
/*	Given a list of parameters to be passed to a tool, build the	*/
/*	"argv" vector for the "exec"/"spawn" call.  Leave the first	*/
/*	and last two slots empty for later customization.		*/
/*									*/
/*	Parameters:							*/
/*		tool -	A pointer to the arg structure for the tool.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
argify(tool)
	struct	targs		*tool;
	{
	int			i;
	struct	args		*ptr;
/*
 *	"malloc" the storage for the "argv" array.  We need two extra slots at
 *	the beginning, two extra at the end, and one for the terminating NULL
 *	pointer.
 */
	if((tool->argv = (char **) malloc(sizeof(char *) *
		(tool->argc + (2 + 2 + 1)))) == NULL)
		{
		fatal("Insufficient 'argv' memory","");
		}
/*
 *	Transfer the argument pointers from the list structure into the
 *	new "argv" structure.
 */
	for(i = 2, ptr = tool->head; ptr != NULL; i++, ptr = ptr->next)
		tool->argv[i] = ptr->data;
/*
 *	Skip the last two slots (actually, initialize them to NULL also), then
 *	write a NULL into the termination slot.
 */
	tool->argv[i] = NULL;
	tool->argv[i + 1] = NULL;
	tool->argv[i + 2] = NULL;
	tool->argc += (2 + 2);
	}

/************************************************************************/
/*									*/
/*	Routine:	fatal						*/
/*									*/
/*	Print out a "FATAL" error message and quit.			*/
/*									*/
/*	Parameters:							*/
/*		sptr1 -	A pointer to the first of the error message.	*/
/*		sptr2 -	A pointer to the rest of the error message.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
fatal(sptr1,sptr2)
	STRING			sptr1;
	STRING			sptr2;
	{
	fprintf(STDERR,"FATAL: %s%s\n",sptr1,sptr2);	/* Print message */
	terminate(ERRORS);		/* Bye... */
	}

/************************************************************************/
/*									*/
/*	Routine:	name_temp					*/
/*									*/
/*	Create a suitable temporary file name and record the results in	*/
/*	the specified data structure.					*/
/*									*/
/*	Parameters:							*/
/*		info -	A pointer to the temp file info structure.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
name_temp(info)
	struct	tfile		*info;
	{
	extern	STRING		strchr();

	static	UINT		instance = 0;

	if((strlen(tempdir) + 9) > FNSIZE)
		fatal("Temporary filename too long","");

	strcpy(info->name,tempdir);
#ifdef	OSU
	if(++instance > 26)
		fatal("Too many temp files","");
	sprintf(strchr(info->name,'\0'),"T%c%u.TCC", instance + 'A' - 1,
		getpid());
#else
	if(instance++ > 32767)
		fatal("Too many temp files","");
	sprintf(strchr(info->name,'\0'),"T%u.TCC",(temptime + instance) &
		0x7FFF);
#endif
	}

/************************************************************************/
/*									*/
/*	Routine:	next_arg					*/
/*									*/
/*	Fetch the next command line argument for the current flag.	*/
/*									*/
/*	Parameters:							*/
/*		argv -	A pointer to the "main" "argv" parameter.	*/
/*		ptr -	A pointer to the current parsing pointer.	*/
/*									*/
/*	Returns:	A pointer to the next argument string.		*/
/*									*/
/************************************************************************/

STRING
next_arg(argv,ptr)
	char			***argv;
	char			**ptr;
	{
	STRING			s;

	s = *ptr;
	if(*++s == '\0')
		{
		if((s = *(*argv)++) == NULL)
			fatal("Missing argument for flag: ",*ptr);
		}
	else
		*ptr = s + (strlen(s) - 1);

	return (s);
	}

/************************************************************************/
/*									*/
/*	Routine:	parse_cpu					*/
/*									*/
/*	Parse the Transputer type.					*/
/*									*/
/*	Parameters:							*/
/*		argv -	A pointer to the "main" "argv" parameter.	*/
/*		ptr -	A pointer to the current parsing pointer.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
parse_cpu(argv,ptr)
	char			***argv;
	char			**ptr;
	{
	static	STRING		err		= "Illegal Transputer type";
	BOOL			double32	= FALSE;
	STRING			s;

	s = next_arg(argv,ptr);
	if((*s == 'F') || (*s == 'f'))
		{
		double32 = TRUE;
		add_arg(&pp,"-dDOUBLE32");
		add_arg(&tcx,"-f2");
		s++;
		}
	switch((int) *s)
		{
		case '0':
			add_arg(&tcx,"-p0");
			break;

		case '2':
			add_arg(&pp,"-dT212");
			if(double32)
				lib_default = "t232lib";
			else
				lib_default = "t2lib";
			if(*++s == '\0')
				{
				add_arg(&tcx,"-p2");
				--s;
				}
			else if(*s == '5')
				add_arg(&tcx,"-p25");
			else
				fatal(err,"");
			break;

		case '4':
			add_arg(&pp,"-dT414");
			if(double32)
				lib_default = "t432lib";
			else
				lib_default = "t4lib";
			if(*++s == '\0')
				{
				add_arg(&tcx,"-p4");
				--s;
				}
			else if(*s == '5')
				add_arg(&tcx,"-p45");
			else
				fatal(err,"");
			break;

		case '8':
			add_arg(&pp,"-dT800");
			if(double32)
				lib_default = "t832lib";
			else
				lib_default = "t8lib";
			if(*++s == '\0')
				{
				add_arg(&tcx,"-p8");
				--s;
				}
			else if(*s == '5')
				add_arg(&tcx,"-p85");
			else
				fatal(err,"");
			break;

		default:
			fatal(err,"");
		}
	}

/************************************************************************/
/*									*/
/*	Routine:	parse_env					*/
/*									*/
/*	Parse any arguments associated with the TCC environment		*/
/*	variable into the "eargs" structure and return a pointer to the	*/
/*	resulting "pseudo argv" array.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	A pointer to the resulting "argv" array.	*/
/*									*/
/************************************************************************/

char	**
parse_env()
	{
	STRING			ptr;
	STRING			s;
/*
 *	See if we have an environment variable by the right name, if so add
 *	the contents of any string associated with it to the "eargs"
 *	structure.
 */
	if((s = getenv(TCC_ENV_VARIABLE)) != NULL)
		{
/*
 *	"malloc" storage for a copy of the environment variable.
 */
		if((ptr = (STRING) malloc(strlen(s) + 1)) == NULL)
			fatal("Insufficient environment copy memory","");
		strcpy(ptr,s);
/*
 *	Go through the copy of the environment string we've made and replace
 *	all occurances of whitespace with '\0' so that we can treat the non-
 *	whitespace items as '\0' terminated strings.  Some real crude command
 *	line processing we have here...
 */
		while(*ptr != '\0')
			{
			for(s = ptr; (! ((*s == '\0') || isspace(*s))); s++)
				;
			if(*s != '\0')
				*s++ = '\0';
			if(s != ptr)
				add_arg(&eargs,ptr);
			ptr = s;
			}
		}
	argify(&eargs);
	return (eargs.argv + 2);	/* Skip first two empty slots */
	}

/************************************************************************/
/*									*/
/*	Routine:	process_files					*/
/*									*/
/*	Go through the input file list, one file at a time, and perform	*/
/*	all required processing.  We need to pay attention to the flags	*/
/*	the user may have supplied to modify our operation.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
process_files()
	{
	extern	STRING		strrchr();

	char			buf[FNSIZE + 1];
	STRING			bptr;
	struct	args		*ptr;
	STRING			sptr;
	BOOL			was_c;
/*
 *	First we do whatever PP/TCX/TASM work has been requested for each file.
 */
	for(ptr = files.head; ptr != NULL; ptr = ptr->next)
		{
		if((sptr = strrchr(ptr->data,'.')) != NULL)
			{
/*
 *	Make a lower case copy of any extension the file name has.
 */
			for(bptr = buf; *sptr != '\0'; bptr++, sptr++)
				{
				if(isupper(*sptr))
					*bptr = tolower(*sptr);
				else
					*bptr = *sptr;
				}
			*bptr = '\0';
/*
 *	Look at the file extension to figure out what to do.
 */		
			was_c = (strcmp(buf,PP_CINPUT_EXT) == 0);
			if(was_c || (strcmp(buf,PP_AINPUT_EXT) == 0))
				{
				run_pp(ptr->data,was_c);
				if(! pp_only)
					{
#ifdef	OSU
/*
 * To emulate the behaviour of Unix "cc", strip the full paths out
 * of the file names.  This will force the objects and binaries to
 * be created in the local directory not that of the sources.
 */
#if	H_OPSYS == MS_DOS
					bptr = strrchr(ptr->data, '\\');
#else
					bptr = strrchr(ptr->data, '/');
#endif
					if (bptr != NULL)
						ptr->data = bptr + 1;
#endif
					if(was_c)
						run_tcx(ptr->data);
					if(! tcx_only)
						{
						run_tasm(ptr->data,FALSE,
							was_c);
						}
					}
				}
			else if((strcmp(buf,TASM_INPUT_EXT) == 0) &&
				(! pp_only) && (! tcx_only))
				{
				run_tasm(ptr->data,TRUE,was_c);
				}
			}
		}
/*
 *	If we are linking, now is the time to do it.
 */
	if(no_link || pp_only || tcx_only)
		return;

	run_tlnk();
	}

/************************************************************************/
/*									*/
/*	Routine:	run_pp						*/
/*									*/
/*	Execute the preprocessor module, passing a source file and the	*/
/*	specified flags.						*/
/*									*/
/*	Parameters:							*/
/*		s -	A pointer to the name of the input file.	*/
/*		was_c -	TRUE if the source file is a ".c" file.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
run_pp(s,was_c)
	STRING			s;
	BOOL			was_c;
	{
/*
 *	Customize the input and output file names and run PP.
 */
	pp.argv[1] = s;
	if(pp_only || (tcx_only && (! was_c)))
		pp.argv[pp.argc - 2] = NULL;
	else
		{
		pp.argv[pp.argc - 2] = "-o";
		if(was_c)
			pp.argv[pp.argc - 1] = file1.name;
		else
			pp.argv[pp.argc - 1] = file2.name;
		}

	run_tool(&pp);
	}

/************************************************************************/
/*									*/
/*	Routine:	run_tasm					*/
/*									*/
/*	Execute the assembler module, passing a source file and the	*/
/*	specified flags.						*/
/*									*/
/*	Parameters:							*/
/*		s -	A pointer to the name of the input file.	*/
/*		was_a -	TRUE if the source file is a ".tal" file.	*/
/*		was_c -	TRUE if the source file is a ".c" file.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
run_tasm(s,was_a,was_c)
	STRING			s;
	BOOL			was_a;
	BOOL			was_c;
	{
	char			buf[FNSIZE + 1];
/*
 *	Customize the input and output file names and run TASM.
 */
	if(was_a)
		{
		tasm.argv[1] = s;
		tasm.argv[tasm.argc - 2] = NULL;
		}
	else
		{
		tasm.argv[1] = file2.name;
		if(was_c)
			tasm.argv[tasm.argc - 2] = "-to";
		else
			tasm.argv[tasm.argc - 2] = "-o";
#ifdef OSU
		if (no_link && (output != NULL))
			tasm.argv[tasm.argc - 1] = output;
		else
			tasm.argv[tasm.argc - 1] =
				swap_ext(buf,s,TASM_OUTPUT_EXT);
#else
		tasm.argv[tasm.argc - 1] = swap_ext(buf,s,TASM_OUTPUT_EXT);
#endif
		}

	run_tool(&tasm);
	}

/************************************************************************/
/*									*/
/*	Routine:	run_tcx						*/
/*									*/
/*	Execute the compiler module, passing a source file and the	*/
/*	specified flags.						*/
/*									*/
/*	Parameters:							*/
/*		s -	A pointer to the name of the input file.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
run_tcx(s)
	STRING			s;
	{
	char			buf[FNSIZE + 1];
/*
 *	Customize the input and output file names and run TCX.
 */
	tcx.argv[1] = file1.name;
	tcx.argv[tcx.argc - 2] = "-o";
	if(tcx_only)
		tcx.argv[tcx.argc - 1] = swap_ext(buf,s,TCX_OUTPUT_EXT);
	else
		tcx.argv[tcx.argc - 1] = file2.name;

	run_tool(&tcx);
	}

/************************************************************************/
/*									*/
/*	Routine:	run_tlnk					*/
/*									*/
/*	Execute the linker module using a command file we build here.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
run_tlnk()
	{
	char			buf[FNSIZE + 1];
	struct	args		*ptr;
/*
 *	Open the temp file to be used for the command file.
 */
#if	CRLF
	if((file1.fp = (STREAM) fopen(file1.name,"wt")) == NULL)
#else
	if((file1.fp = (STREAM) fopen(file1.name,"w")) == NULL)
#endif
		{
		fatal("Unable to open TLNK command file: ",file1.name);
		}
/*
 *	Write the various fields in the command file.
 */
#ifdef OSU
	if (debug)
		{
		printf("TLNK command file: %s\n", file1.name);
		printf("TLIB=%s\n", getenv("TLIB"));
		}
#endif
/*
 *	First, write any flags to the command file.
 */
	if(lflags.head != NULL)
		{
		fprintf(file1.fp,"FLAG\t");
		for(ptr = lflags.head; ptr != NULL; ptr = ptr->next)
			fprintf(file1.fp,"%s ",ptr->data);
		fputc('\n',file1.fp);
		}
#ifdef OSU
	if ((lflags.head != NULL) && (debug))
		{
		printf("FLAG\t");
		for(ptr = lflags.head; ptr != NULL; ptr = ptr->next)
			printf("%s ",ptr->data);
		printf("\n");
		}
/*
 * Add the TEMP directory.  Hack the last character (the directory
 * separator) out of it temporarily.
 */
	if (tempdir[0] != '\0')
		{
			char	last, *plast;

			plast = tempdir + strlen(tempdir) -1;
			last = *plast;
			*plast = '\0';
			fprintf(file1.fp, "TEMP\t%s\n", tempdir);
			if (debug)
				printf("TEMP\t%s\n", tempdir);
			*plast = last;
		}
#endif
/*
 *	Write the map/listing file to the command file.
 */
	if(map != NULL)
		fprintf(file1.fp,"LIST\t%s\n",map);
#ifdef OSU
	if ((map != NULL) && (debug))
		printf("LIST\t%s\n",map);
#endif
/*
 *	Write the list of input files to the command file.
 */
#ifndef OSU
	for(ptr = files.head; ptr != NULL; ptr = ptr->next)
		fprintf(file1.fp,"INPUT\t%s\n",swap_ext(buf,ptr->data,""));
#else
	if (startup)
		{
		if (getenv("TROLLIUSHOME") != NULL)
			{
			(void) strcpy(scratch, getenv("TROLLIUSHOME"));
			(void) strcat(scratch, STRDIR);
			(void) strcat(scratch, "itb");
			(void) strcat(scratch, STRDIR);
			(void) strcat(scratch, "lib");
			(void) strcat(scratch, STRDIR);
			(void) strcat(scratch, "lstt");
			}
		else
			{
			(void) strcpy(scratch, LIBDIR2);
			}
		(void) strcat(scratch, STRDIR);
		(void) strcat(scratch, STARTUP);
		fprintf(file1.fp,"INPUT\t%s\n", scratch);
		}

	for(ptr = files.head; ptr != NULL; ptr = ptr->next)
		fprintf(file1.fp,"INPUT\t%s\n",swap_ext(buf,ptr->data,
			TLNK_INPUT_EXT));

	if (debug)
		{
		if (startup)
			printf("INPUT\t%s\n", scratch);
		for(ptr = files.head; ptr != NULL; ptr = ptr->next)
			printf("INPUT\t%s\n",swap_ext(buf,ptr->data,
				TLNK_INPUT_EXT));
		}
#endif
/*
 *	Write the entrypoint to the command file.
 */
	if(entry != NULL)
		fprintf(file1.fp,"ENTRY\t%s\n",entry);
#ifdef OSU
	if (debug)
		{
		if(entry != NULL)
			printf("ENTRY\t%s\n",entry);
		}
#endif
/*
 *	Write the list of library files to the command file.
 */
	for(ptr = libs.head; ptr != NULL; ptr = ptr->next)
		fprintf(file1.fp,"LIB\tlib%s\n",ptr->data);
#ifdef OSU
	if (debug)
		{
		for(ptr = libs.head; ptr != NULL; ptr = ptr->next)
			printf("LIB\tlib%s\n",ptr->data);
		}
#endif
/*
 *	Write the output pathname to the command file.
 */
#ifndef OSU
	if(output != NULL)
		fprintf(file1.fp,"OUTPUT\t%s\n",output);
#else
	fprintf(file1.fp,"OUTPUT\t%s\n", file2.name);

	if (debug)
		printf("OUTPUT\t%s\n", file2.name);
#endif
/*
 *	Write the initial load address to the command file.
 */
	if(address != NULL)
		fprintf(file1.fp,"LOAD\t%s\n",address);
#ifdef OSU
	if ((address != NULL) && (debug))
		printf("LOAD\t%s\n",address);
#endif
/*
 *	Write the initial workspace to the command file.
 */
	if(workspace != NULL)
		fprintf(file1.fp,"STACK\t%s\n",workspace);
#ifdef OSU
	if ((workspace != NULL) && (debug))
		printf("STACK\t%s\n",workspace);
#endif
/*
 *	Close the file and prepare to run TLNK.
 */
	if(ferror(file1.fp))
		fatal("Error writing TLNK command file: ",file1.name);
	if(fclose(file1.fp) == EOF)
		fatal("Unable to close TLNK command file: ",file1.name);
/*
 *	Customize the command line and run TLNK.
 */
	tlnk.argv[1] = file1.name;
	tlnk.argv[2] = NULL;

	run_tool(&tlnk);
#ifdef OSU
/*
 *	Run the tcvt tool to the linker output.
 */
	if (output)
		add_arg(&tcvt,output);
	else
		add_arg(&tcvt,"a.out");
	add_arg(&tcvt, file2.name);
	argify(&tcvt);
	tcvt.argv[0] = "tcvt";
	tcvt.argv[1] = "-o";
	run_tool(&tcvt);
#endif
	}

/************************************************************************/
/*									*/
/*	Routine:	run_tool					*/
/*									*/
/*	Run the specified tool and display the command line if needed.	*/
/*									*/
/*	Parameters:							*/
/*		tool -	A pointer to the arg structure for the tool.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
run_tool(tool)
	struct	targs		*tool;
	{
	int			i;
	int			retrn;
/*
 *	Show the user the command line if requested.
 */
	if(verbose)
		{
		for(i = 0; tool->argv[i] != NULL; i++)
			{
			if(i == 0)
				printf("%s",tool->argv[i]);
			else
				printf(" %s",tool->argv[i]);
			}
		putchar('\n');
		}
#ifdef OSU
	if (pretend)
/*
 * Don't run the tool, just return.
 */
		{
		return;
		}
#endif	/* OSU */
/*
 *	Run the tool.
 */
#if	H_OPSYS == MS_DOS
	retrn = spawnvp(P_WAIT,tool->argv[0],tool->argv);
	if(retrn != 0)
		terminate(retrn);
#else	/* !MS_DOS */
	retrn = fork();
	if(retrn)
		{
		wait(&retrn);
		if(retrn !=0)
			terminate(WEXITSTATUS(retrn));
		}
	else
#ifndef OSU
		execvp(tool->argv[0],tool->argv);
#else
		{
		execvp(tool->argv[0],tool->argv);
		perror("tcc");
		exit(2);
		}
#endif
#endif	/* MS_DOS */
	}

/************************************************************************/
/*									*/
/*	Routine:	swap_ext					*/
/*									*/
/*	Take a "skeleton" filename and replace any file extension it	*/
/*	might have with the specified extension and return the result.	*/
/*									*/
/*	Parameters:							*/
/*		bptr -	A pointer to the buffer the result goes in.	*/
/*		sptr -	A pointer to the skeleton filename.		*/
/*		ext -	A pointer to the default extension to use.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

STRING
swap_ext(bptr,sptr,ext)
	STRING			bptr;
	STRING			sptr;
	STRING			ext;
	{
	extern	STRING		strrchr();

	STRING			s;

	if(! (((s = strrchr(strcpy(bptr,sptr),'.')) == NULL) ||
#if	H_OPSYS == MPW
		(strrchr(s,':') != NULL)))
#else	/* !MPW */
		((strrchr(s,'/') != NULL) ||
			(strrchr(s,'\\') != NULL))))
#endif	/* MPW */
		{
/*
 *	We had a filename extension, replace it.
 */
		*s = '\0';
		}
	if((strlen(bptr) + strlen(ext)) > FNSIZE)
		fatal("Filename too long","");

	return (strcat(bptr,ext));
	}

/************************************************************************/
/*									*/
/*	Routine:	terminate					*/
/*									*/
/*	Remove any temp files which were created and exit with the	*/
/*	specified exit code.						*/
/*									*/
/*	Parameters:							*/
/*		rval -	The desired "exit" value to return.		*/
/*									*/
/*	Returns:	None, and takes its time about it too.		*/
/*									*/
/************************************************************************/

VOID
terminate(rval)
	int			rval;
	{
	if(file1.name[0] != '\0')
		(VOID) remove(file1.name);

	if(file2.name[0] != '\0')
		(VOID) remove(file2.name);
#ifdef	OSU
	if (tlibenv)
		(void) free(tlibenv);
	if (incdir)
		(void) free(incdir);
	if (incdir2)
		(void) free(incdir2);
#endif

	exit(rval);			/* We are history */
	}

/************************************************************************/
/*									*/
/*	Routine:	usagerr						*/
/*									*/
/*	Report incorrect TCC invocation to user with a suggestion of	*/
/*	how it should be done.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
usagerr()
	{
#ifdef OSU
	fprintf(STDERR,"FATAL: Usage is 'tcc [pathname.ext]* [options]*'\n");
	printf("\tDo not link:                          '-c'\n");
	printf("\tGenerate debug information:           '-d'\n");
	printf("\tDefine macro:                         '-D macro[=value]'\n");
	printf("\tPreprocess only:                      '-E'\n");
	printf("\tSpecify include file path:            '-I pathname'\n");
	printf("\tSpecify link library:                 '-l library'\n");
	printf("\tSpecify link library path:            '-L pathname'\n");
	printf("\tPretend mode:                         '-N'\n");
	printf("\tSpecify linked output file:           '-o pathname'\n");
	printf("\tTurn optimization on:                 '-O'\n");
	printf("\tPreprocess & compile only:            '-S'\n");
	printf("\tUn-define macro:                      '-U macro'\n");
	printf("\tToggle verbose status:                '-v'\n");
	printf("\tDo not link the start up module       '-0'\n");
	printf("\tPass explicit argument to TASM:       '+a argument'\n");
	printf("\tPass explicit argument to TCX:        '+c argument'\n");
	printf("\tPass explicit argument to TLNK:       '+l argument'\n");
	printf("\tPass explicit argument to PP:         '+p argument'\n");
#else
	fprintf(STDERR,"FATAL: Usage is 'tcc [pathname.ext]* [options]*'\n");
	printf("\tSpecify initial load address:         '-a address'\n");
	printf("\tDo not link:                          '-c'\n");
	printf("\tDefine macro:                         '-d macro[=value]'\n");
	printf("\tSpecify entrypoint address or symbol: '-e entrypoint'\n");
	printf("\tGenerate debug information:           '-g'\n");
	printf("\tSpecify include file path:            '-i pathname'\n");
	printf("\tSpecify link library:                 '-l pathname'\n");
	printf("\tSpecify linkage map output file:      '-m pathname'\n");
	printf("\tSpecify linked output file:           '-o pathname'\n");
	printf("\tPreprocess only:                      '-p'\n");
	printf("\tPreprocess & compile only:            '-s'\n");
	printf("\tSet FP/CPU type T2/T25/T4/T45/T8/T85: '-t [f]type'\n");
	printf("\tUn-define macro:                      '-u macro'\n");
	printf("\tToggle verbose status:                '-v'\n");
	printf("\tSpecify initial workspace address:    '-w address'\n");
	printf("\tPass explicit argument to TASM:       '+a argument'\n");
	printf("\tPass explicit argument to TCX:        '+c argument'\n");
	printf("\tPass explicit argument to TLNK:       '+l argument'\n");
	printf("\tPass explicit argument to PP:         '+p argument'\n");
	printf("PROCESSING FLOWCHART BASED ON PATHNAME EXTENSION:\n");
	printf("\t[*.c]->PP->(*.pp)->TCX->[*.tal]->TASM->[*.trl]->TLNK->(*.tld)\n");
	printf("\t[*.pal]->PP->(*.pp)--------------/   (Legal input types in []'s)\n");
#endif	/* OSU */
	terminate(ERRORS);
	}
