/*
 *	OSU Research Computing & Cornell Theory Center		Trillium
 *
 *	Copyright 1990 Cornell University & The Ohio State University
 *
 *	MB
 *
 *	$Log:	tcvt.c,v $
 * Revision 2.2  92/04/21  11:22:20  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  13:23:37  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:36:23  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:59:51  trillium
 * Homo erectus
 * 
 *
 *	Convert Logical Systems .tld files to t.out format.
 *
 *	To be of any use, the .tld file must be compiled "-s".
 *	This program heavily uses code borrowed from both the
 *	Logical Systems "ld-one.c" and the Trollius "patch.c".
 *
 *	Copyright 1988 Logical Systems and Cornell University.
 */

/*
 *  notes:
 *
 *  Assumes that bss records appear after all text and initialized data.
 *  
 *  For the relocations to work, loadpoint must be much below datapoint
 *  so that all code offsets are negative.  If we choose 0 as the standard
 *  phantom data segment start address then we should choose, say, minint
 *  for text segment start.
 *  
 *  There is supposed to be a T_LOAD record for the start of the
 *  data segment, separating it from the text segment.
 */

#include	<stdio.h>
#include	<sys/file.h>		/* for open() constants */
#include	<strings.h>		/* String functions */
#include	<t.out.h>		/* t.out format definitions */
#include	"taldef.h"		/* Assembler/Linker definitions */

#define	WORDSIZE 4



/************************************************************************/
/*									*/
/*			Global Data Definitions				*/
/*									*/
/************************************************************************/

/*
 *  
 *	Structure used for list of records.
 */

struct tal_rec {
	int		type;		/* type of record */
	UINT		size;		/* size of data records */
	SLONG		pc;		/* reloc rec pos in data segment */
	SLONG		offset;		/* reloc rec contents */
	char		*data;		/* where data rec's data is kept */
	struct tal_rec	*next;		/* ptr to next record in list */
};

struct tal_rec *listhead = NULL;	/* head of list of records	*/
struct tal_rec *datahead = NULL;	/* first data record in list	*/
struct tal_rec *bsshead  = NULL;	/* first bss record in list	*/

/*
 *	I/O definitions.
 */

char inbuf[FNSIZE + 1];			/* input filename buffer */
char *outfile = "t.out";		/* name of output file */
STREAM	in_fp = NULL;			/* Input file pointer */
STREAM	verbout = stdout;		/* "verbose" output stream */
int  outfd;				/* output file descriptor */

/*
 *	General definitions.
 */

struct exec header;			/* output executable file header */

STRING	cpu_name;			/* Processor name (see "cpu_type") */
UINT	cpu_type;			/* Processor type */

int  verbose = 0;			/* whether to report main steps */
int  Verbose = 0;			/* even more verbose */

SLONG  tcodept = -1;			/* load point as TLNK thinks */
SLONG  tdatapt = -1;			/* start of data segment */

extern int errno;


/************************************************************************/
/*									*/
/*			Function Type Declarations			*/
/*									*/
/************************************************************************/

VOID	convert();
VOID	toutput();
VOID	read_block();
VOID	exit();
VOID	fatal();
VOID	main();
VOID	makefn();
VOID	usagerr();

extern char *malloc();

/*
 *	The following functions may be called only for side effects.
 */
/*lint +fvr */

int	in_1b();
int	in_2b();
SLONG	in_4b();

/*lint -fvr */

/************************************************************************/
/*									*/
/*	Routine:	main						*/
/*									*/
/*	Reads the ".tld" file, converts it into "t.out" format,		*/
/*	and writes it out.						*/
/*									*/
/*	Command line parameters:					*/
/*		[-v] [-V]		-- verbosity			*/
/*		[-o <outfile>]		-- default "t.out"		*/
/*		file to convert						*/
/*									*/
/************************************************************************/

VOID
main(argc, argv)
	int	argc;
	char	*argv[];
{
	/* get output file name if specified */
	while (argc > 1 && argv[1][0] == '-') {
		if (argv[1][1] == 'o') {
			if (argv[1][2]) {
				outfile = &argv[1][2];
				++argv;
				--argc;
			} else {
				if (argc < 4)  usagerr();
				outfile = argv[2];
				argv += 2;
				argc -= 2;
			}
		} else if (argv[1][1] == 'v') {
			verbose = 1;
			++argv;
			--argc;
		} else if (argv[1][1] == 'V') {
			verbose = 1;
			Verbose = 1;
			++argv;
			--argc;
		} else {
			usagerr();
		}
	}

	if(argc != 2) usagerr();	/* Must have an input file */

	if (verbose)
		fprintf (verbout, "\n.tld-->t.out converter, v2.0\n\n");

/*
 *	Open file to be converted.
 */
	makefn(argv[1],inbuf,TLNK_OUTPUT_EXT);	/* Make complete filename */
#if	CRLF
	if((in_fp = (STREAM) fopen(inbuf,"rb")) == NULL)
#else
	if((in_fp = (STREAM) fopen(inbuf,"r")) == NULL)
#endif
	{
		fatal("Unable to open input file: ",inbuf);
	}

/*
 *	Read file, convert, and write to output file.
 */
	convert();

	fclose(in_fp);

	if (verbose)
		fprintf (verbout, "Program conversion finished\n\n");

	exit(NOERRORS);
}


/************************************************************************/
/*									*/
/*	Routine:	read_block					*/
/*									*/
/*	Read a specified # of bytes from the input file			*/
/*	to a buffer.							*/
/*									*/
/*	Parameters:							*/
/*		"buffer" The address of the buffer			*/
/*		"count"	The number of bytes to copy.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
read_block (buffer, count)
	register char	*buffer;
	register UINT	count;
{
	while(count--) {
		*buffer++ = in_1b();
	}
}

/************************************************************************/
/*									*/
/*	Helper routine for convert(): add a node to the list		*/
/*									*/
/************************************************************************/

struct tal_rec *
addone (last, type, size)
	struct tal_rec *last;
	int type;
	UINT size;
{
	struct tal_rec *new;
	char *buf;
	unsigned int i;

	i = sizeof (struct tal_rec);
	new = (struct tal_rec *) malloc (i);
	if (size)
		buf = malloc (size);
	else
		buf = NULL;
	if (new==NULL || (size && (buf==NULL)))
		fatal ("Out of memory!", "");
	if (listhead == NULL)
		listhead = new;
	else
		last->next = new;
	new->type = type;
	new->data = buf;
	new->next = NULL;
	return (new);
}

/************************************************************************/
/*									*/
/*	Routine:	Convert						*/
/*									*/
/*	Convert .tld format to t.out format				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
convert()
{
	register UINT temp;
	struct tal_rec *record = NULL;

/*
 *	Get file type and CPU type.
 */
	if(in_1b() != T_LD_FILE)
		fatal("Specified input file isn't in .tld format","");
	cpu_type = in_1b();		/* Get load CPU type */
	switch(cpu_type) {

	case	TYPE_UNKNOWN:
		cpu_name = "UNKNOWN";
		fatal("Cannot handle unknown Transputer CPU type","");
			break;

	case	TYPE_414:
		cpu_name = "T414B";
		header.a_magic = TO_MAGIC_T4_X;
		break;

	case	TYPE_800:
		cpu_name = "T800";
		header.a_magic = TO_MAGIC_T8_X;
		break;

	default:
		fatal("Unrecognizable Transputer CPU type in input file", "");
	}

/*
 *	Initialize t.out header:
 */
	header.a_text   = 0;
	header.a_data   = 0;
	header.a_bss    = 0;
	header.a_syms   = 0;
	header.a_entry  = 0;
	header.a_trsize = 0;
	header.a_drsize = 0;

/*
 *	Loop through all the records in load file.  Write those that are
 *	needed to the output file, toss the rest.
 */
	FOREVER {
		switch ((int) in_1b())
		{
		case T_ENTRY:		/* Set entry point address */
			header.a_entry = in_4b();
			if (verbose) {
				fprintf (verbout,
				"Entry point: 0x%x\n", header.a_entry);
			}
			bsshead = NULL;
			break;

		case T_LOAD:		/* Set load address */
			if (tcodept == -1) {
				tcodept = in_4b();
				if (verbose) {
				    fprintf (verbout,
				    "Code load point: 0x%x\n", tcodept);
				}
			} else if (tdatapt == -1) {
				tdatapt = in_4b();
				if (verbose) {
				    fprintf (verbout,
				    "Data load point: 0x%x\n", tdatapt);
				}
			} else {
				fatal ("Third load point?!", "");
			}
			datahead = NULL;
			bsshead = NULL;
			break;

		case T_DATA:		/* Normal code or data record */
			in_2b();		/* Toss line # */
			temp = in_2b();		/* Get length */
			record = addone (record, T_DATA, temp);
			read_block (record->data, temp);
			record->size = temp;
			if (Verbose) {
				fprintf (verbout,
				"%d bytes code\n", temp);
			}
			if (datahead == NULL)
				datahead = record;
			bsshead = NULL;
			break;

		case T_ADDR_DATA:	/* data relocation record */
			in_2b();		/* Toss line # */
			in_2b();		/* Toss symbol # */
			record = addone (record, T_ADDR_DATA, (UINT) 0);
			record->offset = in_4b();
			record->size = WORDSIZE;
			if (Verbose) {
				fprintf (verbout,
					"data reloc at offset %d\n",
					record->offset);
			}
			if (datahead == NULL)
				datahead = record;
			bsshead = NULL;
			break;

		case T_STORAGE:		/* Uninitialized-data record */
			in_2b();		/* Toss line # */
			record = addone (record, T_STORAGE, (UINT) 0);
			record->size = (UINT) in_4b();
			if (Verbose) {
				fprintf (verbout,
				"%d bytes storage\n", record->size);
			}
			if (datahead == NULL)
				datahead = record;
			if (bsshead == NULL)
				bsshead = record;
			break;

		case T_FILENAME:	/* Filename record */
			in_4b();		/* Toss old & new ln #'s */
			temp = in_1b();		/* Get length */
			while(temp--) in_1b();	/* Toss filename */
			break;

		case T_STACK:		/* Set stack pointer address */
			in_4b();		/* toss stack address */
			break;

		case T_DEBUG_DATA:
			in_2b();		/* Toss line # */
			in_4b();		/* Toss value */
			temp = in_2b();		/* Get length */
			while(temp--) in_1b();	/* Toss data */
			break;

		case T_DEBUGSYM_DATA:
			in_2b();		/* Toss line # */
			in_4b();		/* Toss value */
			temp = in_2b();		/* Get length */
			in_2b();		/* Toss symbol # */
			while(temp--) in_1b();	/* Toss data */
			break;

		case T_EOF:		/* Found EOF record */
			toutput (datahead, bsshead);
			return;

		default:
			fatal ("Bad record format on load file: ", inbuf);
		}
	}
}


/************************************************************************/
/*									*/
/*	Routine:	toutput						*/
/*									*/
/*	Output the fields of the t.out format				*/
/*									*/
/************************************************************************/

VOID
toutput()
{
	int  i, n, size;
	SLONG  pc, offset;
	char *zerobuf, *drbuf;
	struct tal_rec *record;
	struct to_relative *relocbuf, *relocp;
	SLONG zeroword = 0;

	/* allocate space large enough for
	   largest non-BSS T_STORAGE segment
	   and initialize it to all zeros: */

	size = WORDSIZE;
	record = listhead;
	while (record) {
		if (record == bsshead)
			break;
		if (record->type == T_STORAGE) {
			if (size < record->size)
				size = record->size;
		}
		record = record->next;
	}
	if (size > WORDSIZE) {
		zerobuf = malloc (size);
		if (zerobuf == NULL)
			fatal ("Out of memory!", "");
		for (i=0; i<size; i++)
			zerobuf[i] = 0;
	} else {
		zerobuf = (char *) &zeroword;
	}

	/* make all non-bss record->data point to SOMETHING	*/
	/* -- this affects T_STORAGE and T_ADDR_DATA records	*/

	record = listhead;
	while (record) {
		if (record == bsshead)
			break;
		if (record->type==T_STORAGE || record->type==T_ADDR_DATA)
			record->data = zerobuf;
		record = record->next;
	}

	/* count the relocation records,
	   and compute their location too: */

	n = 0;
	pc = 0;
	record = datahead;
	while (record) {
		if (record == bsshead)
			break;
		if (record->type == T_ADDR_DATA) {
			++n;
			record->pc = pc;
		}
		pc += record->size;
		record = record->next;
	}
	header.a_data = pc;

	/* if any, get space for and construct relocation segment: */

	if (n > 0) {
		if (tdatapt == -1)
			fatal ("data relocation without datapt?!", "");
		size = n * sizeof(struct to_relative);
		header.a_drsize = size;
		drbuf = malloc (size);
		if (drbuf == NULL)
			fatal ("Out of memory!", "");
		relocbuf = (struct to_relative *) drbuf;
		relocp = relocbuf;
		record = datahead;
		while (record) {
			if (record == bsshead)
				break;
			if (record->type == T_ADDR_DATA) {
				offset = record->offset;
				if (offset < 0) {	/* offset into code */
					offset += tdatapt - tcodept;
					relocp->s1_text = 1;
				} else {		/* offset into data */
					relocp->s1_text = 0;
				}
				relocp->type = 1;
				relocp->s1_external = 0;
				relocp->s2_external = 0;
				relocp->pc = record->pc / WORDSIZE;
				relocp->cnst = offset;
				relocp->s1.index = -1;
				relocp->s2.index = -1;
				++relocp;
			}
			record = record->next;
		}
	}

	/* complete t.out header */
	header.a_entry -= tcodept;
	record = listhead;
	while (record) {
		if (record==datahead || record==bsshead)
			break;
		header.a_text += record->size;
		record = record->next;
	}
	record = bsshead;
	while (record) {
		header.a_bss += record->size;
		record = record->next;
	}

	/* open output file */
	outfd = open (outfile,
		(O_WRONLY | O_CREAT | O_TRUNC), 0775);
	if (outfd < 0)
		fatal("Unable to open output file: ",
		outfile);

	/* write out header */
	if (verbose)
		fprintf (verbout, "Writing out header\n");
	ewrite (outfd, (char *) &header, (int) sizeof(header));

	/* write out code segment */
	if (verbose) {
		fprintf (verbout,
			"Writing code segment, %ld bytes\n", header.a_text);
	}
	record = listhead;
	while (record) {
		if (record==datahead || record==bsshead)
			break;
		if (Verbose)
			fprintf (verbout, "\t%u bytes\n", record->size);
		ewrite (outfd, record->data, (int) record->size);
		record = record->next;
	}

	/* write data segment */
	if (verbose) {
		fprintf (verbout,
			"Writing data segment, %ld bytes\n", header.a_data);
	}
	record = datahead;
	while (record) {
		if (record == bsshead)
			break;
		if (Verbose)
			fprintf (verbout, "\t%ud bytes\n", record->size);
		ewrite (outfd, record->data, (int) record->size);
		record = record->next;
	}

	/* write out data relocation table, if any */
	if (header.a_drsize != 0) {
		if (verbose) {
			fprintf (verbout,
				"Writing data reloc table, %ld bytes\n",
					header.a_drsize);
		}
		ewrite (outfd, relocbuf, (int) header.a_drsize);
	}

	/* we don't have any symbol table */

	/* we don't have any "strings" */
	ewrite (outfd, (char *) &zeroword, sizeof(zeroword));

	/* close output file */
	if (close (outfd) < 0)
		fatal("Unable to close output file: ", outfile);
}


/************************************************************************/
/*									*/
/*	Routines:	in_1b,in_2b,in_4b				*/
/*									*/
/*	Read 1, 2 or 4 bytes from the download input file.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The 1, 2 or 4 byte value read.			*/
/*									*/
/************************************************************************/

int
in_1b()
{
	register int	c;

	if((c = fgetc(in_fp)) == EOF) {
		if(ferror(in_fp) || feof(in_fp)) {
			fatal("Error reading input file: ", inbuf);
		}
	}
	return (c & 0xFF);
}

int
in_2b()
{
	int	rval;

	rval = in_1b();
	rval |= (in_1b() << 8);
	return (rval & 0xFFFF);
}

SLONG
in_4b()
{
	SLONG	rval;

	rval = (SLONG) in_1b();
	rval |= (((SLONG) in_1b()) << 8);
	rval |= (((SLONG) in_1b()) << 16);
	rval |= (((SLONG) in_1b()) << 24);
	return (rval);
}

/************************************************************************/
/*									*/
/*	Routine:	makefn						*/
/*									*/
/*	Expand a skeleton filename by adding a default extension if	*/
/*	none is supplied.						*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to the skeleton filename.		*/
/*		"bptr"	A pointer to the buffer the result goes in.	*/
/*		"ext"	A pointer to the default extension to use.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
makefn (sptr,bptr,ext)
	STRING	sptr,bptr,ext;
{
	extern	char	*strchr(),*strrchr();

	STRING	s;

	if(strlen(sptr) > FNSIZE) usagerr();
	strcpy(bptr,sptr);
	if((s = strrchr(bptr,'.')) != NULL) {
		if((strchr(s,'/') == NULL) && (strchr(s,'\\') == NULL)) return;
	}
	if(strlen(bptr) > (FNSIZE - 4)) usagerr();
	strcat(bptr,ext);
}

/*************************************************************************
	ewrite

	FUNCTION:	- error proof write operation
			- makes the write system call and checks for all errors
	ACCEPTS:	- file descriptor open for write
			- pointer to data
			- size of data
	RETURNS:	- may call error or perror and never return
**************************************************************************/


ewrite(fd, buffer, bufsize)

int		fd,			/* read file descriptor */
		bufsize;		/* size of data buffer */
char		*buffer;		/* pointer to data */

{
	if (write(fd, buffer, bufsize) != bufsize)
		fatal ("Error writing to output file: ", outfile);
}

/************************************************************************/
/*									*/
/*	Routine:	usagerr						*/
/*									*/
/*	Report incorrect loader invocation to user with a suggestion	*/
/*	of how it should be done.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
usagerr()
{
	fatal ("Usage: 'convert [-v] [-V] [-o outfile] infile'","");
}

/************************************************************************/
/*									*/
/*	Routine:	fatal						*/
/*									*/
/*	Print out a "FATAL" error message and quit.			*/
/*									*/
/*	Parameters:							*/
/*		"sptr1"	A pointer to the first of the error message.	*/
/*		"sptr2" A pointer to the rest of the error message.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID
fatal(sptr1,sptr2)
	STRING	sptr1,sptr2;
{
	fprintf (verbout, "FATAL: %s%s\n",sptr1,sptr2);
	exit(ERRORS);
}
