/*
 *	OSU Research Computing & Cornell Theory Center		Trillium
 *
 *	Copyright 1990 Cornell University & The Ohio State University
 *
 *	$Log:	trar.c,v $
 * Revision 2.2  92/04/21  11:28:01  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  13:25:09  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:36:25  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:59:54  trillium
 * Homo erectus
 * 
 *
 *	MB
 *
 *	ar-like wrapper for LSTT tlib.
 *
 *	Usage: trar options libname objfiles...
 *
 *	For now, no "posname" allowed.  Output: an invocation of:
 *		tlib libname -[options] objfiles...
 *	where the ar options are translated to tlib options thus:
 *
 *	AR option	TLIB option	function
 *
 *	d		d		delete files from archive
 *	r		u		replace named files
 *	u		(n.a.)		replace only if newer
 *	q		a		quickly append files at end
 *	t		(no option)	print a table of contents
 *	x		c		extract named files (or all)
 *	v		v		be verbose
 *	c		(n.a.)		silently create new archive
 *		translated to:
 *	(n.a.)		b		create archive, order files
 *
 *	Other options not supported for now.
 *	The 'c' ar option was chosen to invoke the 'b' tlib option,
 *	The actual operation is similar to that of ar followed by ranlib.
 */

#include <stdio.h>
#include <ctype.h>
#include <signal.h>

#ifndef TLIB
#define TLIB	"tlib"
#endif

main (argc, argv)
	int argc;
	char **argv;
{
	int c, i, j, t, status, dobuild;
	char *progname, *libname, *f, *aropts;
	char tlibopts[4];

	if (argc < 3) {
		fprintf (stderr, "Usage: trar options libname [objfiles...]\n");
		exit(1);
	}

	progname = argv[0];
	aropts   = argv[1];
	libname  = argv[2];

	f = TLIB;

	argv[0] = f;
	argv[1] = libname;
	argv[2] = tlibopts;

	tlibopts[0] = '-';
	dobuild = 0;
	j = 1;
	for (i=0; (c=aropts[i])!=0; i++) {
		switch (c) {
		case 'd':
			if (j == 1)
				tlibopts[j++] = 'd';
			break;
		case 'r':
			if (j == 1)
				tlibopts[j++] = 'u';
			break;
		case 'q':
			if (j == 1)
				tlibopts[j++] = 'a';
			break;
		case 't':
			/* no tlib options needed */
			break;
		case 'x':
			if (j == 1)
				tlibopts[j++] = 'c';
			break;
		case 'v':
			tlibopts[j++] = 'v';
			break;
		case 'c':
			dobuild = 1;
			break;
		default:
			fprintf (stderr,
			"trar: '%c' option not supported\n", c);
			exit(1);
			break;
		}
	}
	if (dobuild)
		tlibopts[1] = 'b';
	if (j == 1)
		j = 0;
	tlibopts[j] = 0;

#ifdef DEBUG
	fprintf (stderr, "%s", f);
	for (i=1; argv[i] != 0; i++)
		fprintf (stderr, " %s", argv[i]);
	fprintf (stderr, "\n");
	exit(0);
#endif

	execvp(f, argv);
	/* should not reach this point -- unless execv() fails: */
	fprintf (stderr, "%s: can't find %s\n", progname, f);
	exit(1);
}
