/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	NET.h,v $
 * Revision 2.2  92/04/30  14:44:42  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/01/19  20:23:32  gdburns
 * facelift and sync with new features
 * 
 * Revision 2.1  91/03/20  11:34:35  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:28:18  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/07/03  15:37:57  gdburns
 * Gave NREEL the correct value.
 * 
 * Revision 2.0  90/02/20  11:55:05  trillium
 * Homo erectus
 *
 *	Function:	- F77 constants and structures associated with the
 *			  network subsystem
 *			- based on Trollius 2.0 Copyright 1992
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#ifndef _NETF
#define _NETF

/*
 * network constants
 */
#define	LOCAL		(-2147483647)	/* on the local node */
#define	MAXNMSGLEN	(4096)		/* maximum network frame size */
#define	NOTEVENT	(-2147483647)	/* not a valid message event */
#define	NOTNODEID	(-2147483648)	/* not a valid node ID */
#define	NOTNODETYPE	(-2147483648)	/* not a valid node type */
#define	ORIGIN		(IGORGN())	/* should be user's node */

/*
 * node types
 */
#define NTOTB		(0)		/* OTB, not-ITB */
#define NTITB		(1)		/* ITB, not-OTB */
#define NTCAST		(2)		/* a multicast of nodes */
#define NTWASTE		(4)		/* not fully used/wasted */
#define NTDISK		(8)		/* node has a disk */
#define NTTUBE		(16)		/* node has a video unit */
#define NTJONES		(32)		/* neighbour node */
#define NTBOOT		(64)		/* child node in boot tree */
#define NTALL		(127)		/* match any bit-flag */

/*
 * predefined cast identifiers (implies a nodeid set and a nodeid source)
 */
#define HOST2ALL	(-2147483646)	/* origin to all nodes */
#define HOST2ITB	(-2147483645)	/* origin to all ITB nodes */
#define HOST2OTB	(-2147483643)	/* origin to all OTB nodes */

/*
 * network message header
 */
#define NHDSIZE		8		/* # of data words in header */

#define	KHOLD		(1)		/* set virtual circuit */
#define NOBUF		(16)		/* don't do buffering at all */
#define DFORWARD	(256)		/* datalink store and forward */
#define NREEL		(1024)		/* multireel */
#define DRAWDATA	(65536)		/* don't touch data pouch */
#define DINTDATA	(131072)	/* data pouch is integers */
#define DFLTDATA	(262144)	/* data pouch is floats */
#define DDBLDATA	(524288)	/* data pouch is doubles */
#define DRAWMSG		(1048576)	/* don't touch message */
#define DINTMSG		(2097152)	/* message is integers */
#define DFLTMSG		(4194304)	/* message is floats */
#define DDBLMSG		(8388608)	/* message is doubles */

/*
 * multireel combining functions
 */
#define CBN_ERR		(-1)		/* error codes in first int only */
#define CBN_ADD		(-2)		/* add ints */
#define CBN_MAX		(-3)		/* maximize ints */
#define CBN_MIN		(-4)		/* minimize ints */
#define CBN_XOR		(-5)		/* exclusive OR ints */
#define CBN_OR		(-6)		/* OR ints */
#define CBN_AND		(-7)		/* AND ints */
#define CBN_CAT		(-8)		/* concatenate messages */

#endif
