/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD
 *
 *	$Log:	all_hash.h,v $
 * Revision 2.2  92/04/30  14:44:47  trillium
 * Ohio Release
 * 
 * Revision 2.1  92/01/19  22:00:20  raja
 * initial revision
 * 
 *	Function:	- generic hash table templates and constants
 *			- for both the static and dynamic versions
 */


#ifndef _ALLHASH
#define _ALLHASH

#include <portable.h>

/*
 * flags defining the modes of operation
 */
#define AHFIXED		((int4) 0)	/* no LRU counters, fixed hashing */
#define AHLRU		((int4) 1)	/* LRU counters are used */
#define AHNOINIT	((int4) 2)	/* hash table not initialized */

/*
 * type definitions
 */
typedef struct ah_desc	HASH;
typedef struct ah_desc	SHASH;

/*
 * function prototypes (all_hash.c, all_shash.c)
 */
extern HASH		*ah_init();		/* create dynamic hash table */
extern SHASH		*ahs_init();		/* create static hash table */
extern void		ah_free();		/* free dynamic hash table */
extern void		*ah_find();		/* read an element */
extern void		*ah_next();		/* read next element */
extern int		ah_delete();		/* remove an element */
extern int		ah_expand();		/* enlarge the hash table */
extern int		ah_insert();		/* insert an element */
extern int		ah_kick();		/* insert an element by force */

/*
 * function macros
 */
#define ah_count(ahd)		((ahd)->ah_nelem)
#define ah_size(ahd)		((ahd)->ah_maxnelem)
#define ah_top(ahd)		(ah_next(ahd, (void *) 0))

#define ahs_find(ahd, x)	ah_find(ahd, x)
#define ahs_delete(ahd, x)	ah_delete(ahd, x)
#define ahs_insert(ahd, x)	ah_insert(ahd, x)
#define ahs_kick(ahd, x)	ah_kick(ahd, x)
#define ahs_count(ahd)		ah_count(ahd)
#define ahs_size(ahd)		ah_size(ahd)
#define ahs_top(ahd)		ah_top(ahd)
#define ahs_next(ahd, x)	ah_next(ahd, x)

/* 
 * structure definitions
 */
struct ah_desc {			/* hash table descriptor */
	int4		ah_maxnelem;	/* maximum nbr. elements */
	int4		ah_nelem;	/* current nbr. elements */
	int4		ah_elemsize;	/* size of element */
	int4		ah_nullkey;	/* null hash key value */
	int4		ah_mode;	/* mode of operation */
	int4		*ah_lru;	/* table of LRU counters */
	void		*ah_table;	/* ptr to the hash table */
};

#endif
