/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	bootreq.h,v $
 * Revision 2.2  92/04/30  14:44:54  trillium
 * Ohio Release
 * 
 * Revision 2.1  92/01/19  20:28:40  gdburns
 * initial revision
 * 
 *	Function:	- boot daemon interface templates and constants
 */

#ifndef _BOOTREQ
#define _BOOTREQ

#include <portable.h>

/*
 * bootd request codes
 */
#define BTQBOOT		0		/* boot neighbour ITB node */
#define BTQADD		1		/* add programs, data */
#define BTQFLUSH	2		/* delete all programs */
#define BTQSTAGE	3		/* stage/load programs */
#define BTQSPREAD	4		/* recursive booting */
#define BTQSTATE	5		/* bootd status */
#define BTQNUM		6		/* # of bootd requests (services) */

/*
 * bootd objects
 */
#define BTO_STAGE	0x01		/* stage program table */
#define BTO_SPAWN	0x02		/* spawn program table */
#define BTO_STRAP	0x04		/* bootstrap program */
#define BTO_ROUTE	0x08		/* route table */
#define BTO_FROZEN	0x10		/* frozen system image */

/*
 * option flags
 */
#define BTSHARE		0x01		/* share text segment */
#define BTONCHIPSTK	0x02		/* place stack segment in ONCHIP */

/*
 * bootd request
 */
struct bootreq {
	int4		btq_src_node;	/* reply to this node */
	int4		btq_src_event;	/* reply to this event */
	int4		btq_request;
	int4		btq_parm1;
#define btq_object	btq_parm1
	int4		btq_parm2;
#define btq_ilink	btq_parm2
#define btq_node	btq_parm2
#define btq_tag		btq_parm2
#define btq_maxprog	btq_parm2
	int4		btq_parm3;
#define btq_flags	btq_parm3
#define btq_size	btq_parm3
	int4		btq_parm4;
#define btq_stksize	btq_parm4
	int4		btq_parm5;
#define btq_priority	btq_parm5
};

/*
 * bootd reply
 */
struct bootreply {
	int4		btr_reply;	/* error code */
	int4		btr_nprog;	/* # of program state descs */
};

/*
 * bootd program status
 */
#define BTMAXNAME	20
#define BTSMAX		40

struct bootstate {
	void		*bts_load;	/* load module addr, includes text */
	void		*bts_data;	/* data segment address */
	int4		bts_stksize;	/* stack size */
	int4		bts_flags;	/* miscellaneous flags */
	int4		bts_priority;	/* process priority */
	char		bts_name[BTMAXNAME];
};

#endif
