/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation & The Ohio State University
 *
 *	MB
 *
 *	$Log:	brenda.h,v $
 * Revision 2.2  92/04/24  16:44:08  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:44:02  gdburns
 * Ohio release
 * 
 * Revision 2.0.1.2  91/04/25  17:30:27  gdburns
 * Getnodeid() is declared in net.h.
 * 
 * Revision 2.0.1.1  90/10/12  16:07:39  raghu
 * Defaults DEFBRENDAD (brenda daemon) and DEFPBRENDAD (path) defined.
 * 
 * Revision 2.0  90/09/26  12:36:50  gdburns
 * Homo sapien
 * 
 *
 *	Function:	- constants for BRENDA implementation via Trollius 2.0.
 */

#ifndef BRENDA_H


/****** configuration stuff ******/

/* no. of slots in the hash table internal to the Brenda deamon: */
/* (this should be prime relative to the number of nodes) */
#define B_SLOTS		59

/* how many list nodes to malloc() at once: */
#define B_NBLOCK	63

/* how large a nodeid table to statically allocate in brendlib: */
/* (a larger table is rmalloc()ed if necessary.) */
#define NODETABLESIZE	18

/* #define FREEALL if the rfreeall() is available on all storage nodes */
/* The rfreeall() function returns rmalloc()-ed memory to the system ITB */
#define FREEALL

/* #define BCAST if the storage nodes are all the compute nodes, HOST2ITB
   is defined in net.h, and broadcast doom & load are supported and desired. */
#define BCAST

/* the max number of Trollius buffer workers, set by brenda -I */
#define BNBUFS 16

/* this is the stack size given to brendad upon brenda -I */
#define BDSTACK 2048


/****** C extensions... ******/

#define PUBLIC		/*public*/
#define PRIVATE		static

#ifndef NULL
#define NULL 0
#endif
#define NULLP ((char *) 0)


/*
 * Default file names and search path prefixes.
 */

#ifndef DEFPBRENDAD
#define	DEFPBRENDAD	"/usr/local/trillium/itb/bin"
#endif
#define	DEFBRENDAD	"brendad"
	 
/****** Trollius stuff ******/

#ifndef MAXNMSGLEN
#include <net.h>
#endif

/* this will go away: */
#ifndef HOST2GRID
#define HOST2GRID	HOST2ITB
#endif

#define getngrid()      getntype (NT_ITB & ~NT_WASTE, NT_ITB | NT_WASTE)
#define getgnodes(p,n)  getnodes (p, n, NT_ITB & ~NT_WASTE, NT_ITB | NT_WASTE)


/****** hashing scheme ******/

/* random-number-generation constants: */
#define TAG_M		139968
#define TAG_A		3877

/* randomize a given tag to some positive number: */
#define RANDTAG(x)	(((TAG_A * (x)) % TAG_M) & 0x7FFFFFFF)

/* calc the hash slot inside the storage node given a tag: */
#define HASHSLOT(x)	(RANDTAG(x) % B_SLOTS)


/****** types and constants ******/

/** Event the brenda server waits for **/
#define BRENDA	-23

/** User-specified object flags **/

#define B_NOMADIC	0		/* default */
#define B_SETTLED	0x01
#define B_VERIFY	0x02
#define B_BLIND		0		/* default */
#define B_LOCAL		0x10
#define B_GLOBAL	0x20
#define B_NORMAL	0		/* default */
#define B_TORCH		0x80		/* get through to frozen server */

/** Requests to the brenda server **/

#define B_PUT		0x01
#define B_GRAB		0x02
#define B_LOOK		0x03
#define B_TRYGRAB	0x04
#define B_TRYLOOK	0x05
#define B_BUMP		0x06

/* freeze/thaw request codes */

#define _FREEZE_	0x0C
#define _THAW_		0x0D

/* Debugging request codes */

#define _OBJ_SUM	0x11
#define _REQ_SUM	0x12
#define _OBJ_LST	0x13
#define _REQ_LST	0x14
#define _CLEAR_		0x15

/* Internal request codes */

#define _ACK_		0x21
#define _LPUT_		0x31
#define _GPUT_		0x32
#define _LGPUT_		0x33
#define _GGRAB_		0x41
#define _FOUND_		0x51
#define _LFOUND_	0x52

#define _ERROR_		0xFF

/* nh_types */

#define _REQ_		0x0100		/* required bit in client request */
#define _REPLY_		0x0200		/* required bit in reply to client */
#define _DATA_		0x0400		/* required bit in data message */
#define _FROZEN_	0x0800		/* required bit in frozen requests */

#define ANYTAG		((long) 0)

/* Codes that may be returned to the user */

#define B_OK		0
#define B_ERROR		(-1)

/* Brenda function call (i.e. parameter block) structure */

struct brenda_pb {
	long	request;
	long	tag;
	char	*buffer;
	long	size;
	long	flags;
	long	snode;
	long	data;
};

/*
 *  Brenda request and reply structure.
 *  This is placed in the nh_data field of the nmsg header.
 */
struct brenda_req {
	long	node;
	long	event;
	long	request;
	long	tag;
	long	size;
	long	data;
	long	data2;		/* not used for now */
	char	*buf;		/* only used locally */
};

/* Brenda report structures */

struct obj_lst {
	long	tag;
	long	size;
};

struct req_lst {
	long	node;		/* client node */
	long	event;		/* client event */
	long	request;
	long	tag;
	long	size;
};


/****** variable and function declarations ******/

extern int errno;
extern char *malloc();
extern char *realloc();
#ifdef FREEVOID
extern void free();	/* #define FREEVOID if your free() is void */
#else
extern int  free();	/* else we check the return value for error! */
#endif
extern bcopy();		/* type of bcopy() ? */
#ifdef FREEALL
extern void rfreeall();		/* return rmalloc()ed memory to system */
extern void _reset_malloc();	/* tell malloc() to start over */
#endif
extern int getntype();
extern int getnodes();

#ifdef __STDC__
int brenda (int req, long tag, char *buf, long size, int flags, long snode);
int freeze (void);
int thaw (void);
int _brenda (struct brenda_pb *argblk);
long hashnode (long tag);
#else
extern int brenda();
extern int freeze(), thaw();
extern int _brenda();
extern long hashnode();
#endif

/* wrappers for standard brenda calls */

#define  put(tag,b,s,f,n)	brenda(B_PUT,tag,b,s,f,n)
#define  grab(tag,b,s,f,n)	brenda(B_GRAB,tag,b,s,f,n)
#define  look(tag,b,s,f,n)	brenda(B_LOOK,tag,b,s,f,n)
#define  trygrab(tag,b,s,f,n)	brenda(B_TRYGRAB,tag,b,s,f,n)
#define  trylook(tag,b,s,f,n)	brenda(B_TRYLOOK,tag,b,s,f,n)
#define  bump(tag,b,incr,f,n)	brenda(B_BUMP,tag,b,incr,f,n)

#endif BRENDA_H
