/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	bufreq.h,v $
 * Revision 2.2  92/04/30  14:45:01  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/27  12:22:31  gdburns
 * Add buffer flow control of user messages.
 * 
 * Revision 2.1.1.1  92/01/27  17:51:02  gdburns
 * Adjust memory and worker limits.
 * 
 * Revision 2.1  91/03/20  11:34:43  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/02/04  22:50:01  gdburns
 * Bumped up default maximum workers.
 * 
 * Revision 2.0.1.2  90/07/13  19:28:28  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/05/06  15:36:34  raja
 * Added the BFPROBE request number.  Parametrized the structure definitions.
 * 
 * Revision 2.0  90/02/20  11:55:09  trillium
 * Homo erectus
 *
 *	Function:	- constants associated with the buffer server
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#ifndef _BUFREQ
#define _BUFREQ

#include <net.h>

/*
 * bufferd request codes
 */
#define	BFWORKER	1
#define BFSTATE		2
#define BFSWEEP		3
#define	BFSPACE		4
#define	BFREADY		5
#define	BFPROBE		6
#define	BFUSRSPACE	7

/*
 * general bufferd constants
 */
#define BFMAX		32
#define BSMAX		(BFMAX * 4)
#define BFSTACK		(0x300)
#define BFDEFSPACE	0x40000
#define BFDEFUSR	0x30000

/*
 * message filters
 */
#define BFALL		0			/* all messages */
#define BFUSR		1			/* user messages */
#define BFSYS		2			/* system messages */
#define BFREQ		4			/* bufferd messages */
#define BFDLI		8			/* datalink input messages */

#ifdef itb
#define BFDEFWORKERS	32
#else
#define BFDEFWORKERS	8
#endif

#define NULL		((struct buffer *) 0)

/*
 * buffer descriptor
 */
struct buffer {
	struct nmsg	bf_head;
	struct buffer	*bf_next;
};

/*
 * bufferd request
 */
struct bufreq {
	int		bfq_src_node;
	int		bfq_src_event;
	int		bfq_req;
	int		bfq_parm3;
#define bfq_maxspace	bfq_parm3
#define bfq_maxworkers	bfq_parm3
#define bfq_maxusr	bfq_parm3
#define bfq_dl_event	bfq_parm3
	int		bfq_parm4;
#define bfq_type	bfq_parm4
	int		bfq_parm5;
#define bfq_event	bfq_parm5
};

/*
 * bufferd reply
 */
struct bufreply {
	int		bfr_reply;
	int		bfr_param1;
#define	bfr_maxworkers	bfr_param1
#define	bfr_type	bfr_param1
	int		bfr_param2;
#define	bfr_maxspace	bfr_param2
#define	bfr_length	bfr_param2
	int		bfr_param3;
#define	bfr_nworkers	bfr_param3
#define	bfr_flags	bfr_param3
	int		bfr_param4;
#define	bfr_nspace	bfr_param4
	int		bfr_param5;
#define	bfr_nfull	bfr_param5
	int		bfr_param6;
#define	bfr_maxusr	bfr_param6
};

/*
 * buffer state
 */
struct bfstate {
	int		bfs_node;	/* destination node */
	int		bfs_event;	/* destination event */
	int		bfs_type;	/* destination type */
	int		bfs_length;	/* message length */
};

/*
 * buffer parameters
 */
struct bfparms {
	int		bfp_maxworkers;
	int		bfp_maxspace;
	int		bfp_nworkers;
	int		bfp_nspace;
	int		bfp_nfull;
	int		bfp_maxusr;
};

#endif _BUFREQ
