/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB, RBD
 *
 *	$Log:	castreq.h,v $
 * Revision 2.2  92/04/30  14:45:03  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:34:44  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  90/11/12  17:30:25  gdburns
 * Switch the link mask to unsigned.
 * 
 * Revision 2.0.1.2  90/07/13  19:27:48  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/07/04  14:21:54  gdburns
 * Revised for new castd registration code.
 * 
 * Revision 2.0  90/02/20  11:54:44  trillium
 * Homo erectus
 *
 *	Function:	- constants associated with the broadcast server
 */

#ifndef _CASTREQ
#define _CASTREQ

#include <portable.h>
#include <net.h>
#include <all_list.h>

/*
 * castd constants
 */
#define MAXCAST		((int4) 29)		/* size of cast table */
#define MAXCLINK	((int4) 4)		/* maximum number of links */
#define MAXSPACE	((int4) 0x40000)	/* maximum message space */

/*
 * castd message filters
 */
#define CFALL		((int4) 0);		/* all messages */
#define CFSYS		((int4) 4);		/* system messages only */

/*
 * castd null pointers
 */
#define CNULL		((struct cast *) 0)
#define CQNULL		((struct castreq *) 0)
#define CSNULL		((struct cstate *) 0)
#define CS2NULL		((struct cstate2 *) 0)
#define CRNULL		((struct castreel *) 0)
#define CPNULL		((struct castpkt *) 0)

/*
 * castd request
 */
struct castreq {
	int4		cq_src_node;		/* source node ID */
	int4		cq_root_node;		/* spanning root node ID */
	int4		cq_src_event;		/* source process event */
	int4		cq_req;			/* castd request */
#define CADD		((int4) 1)		/* add a cast */
#define CREM		((int4) 2)		/* remove a cast */
#define CACK		((int4) 3)		/* parent acknowledgement */
#define CASK		((int4) 4)		/* inquire # of branches */
#define CERR		((int4) 5) 		/* indicate error to parent */
#define CSTAT		((int4) 6) 		/* get castd status */
#define CSTAT2		((int4) 7) 		/* get castd secondary status */
	int4		cq_castid;		/* cast identifier */
	int4		cq_combine;		/* reel combination selector */
	int4		cq_maxstat;		/* max state entries */
};

/*
 * castd reply
 */
struct castreply {
	int4		cr_reply;		/* status code */
	int4		cr_result;		/* some useful stuff */
#define	cr_nbranch	cr_result		/* number of branches */
};

/*
 * multicast entry
 */
struct cast {
	int4		c_castid;		/* cast identifier */
	int4		c_root_node;		/* root node of spanning tree */
	int4		c_parent_node;		/* real parent of node */
	int4		c_combine;		/* combining service for reel */
	int4		c_nbranch;		/* out branches at this node */
	uint4		c_links;		/* in/out link mask */
#define	CLOCAL		0x80000000
	LIST		*c_reels;		/* list of reels taking place */
	int4		c_reg_event;		/* registration request event */
	int4 		c_reg_nack;		/* registration acks. left */
	int4		c_reg_retcode;		/* registration return code */
};

struct castreel {
	int4		cr_event;		/* message event */
	int4		cr_type;		/* message type */
	LIST		*cr_pkt;		/* list message packets */
};

struct castpkt {
	int4		cp_pktnum;		/* packet number */
	int4		cp_nreel;		/* # of reels remaining */
	int4		cp_nh_flags;		/* packet flags */
	int4		cp_nh_data[NHDSIZE];	/* packet data pouch */
	int4		cp_length;		/* packet length */
	char		*cp_msg;		/* body of reeled message */
};

struct cstate {
	int4		cs_castid;		/* cast identifier */
	int4		cs_root_node;		/* root node of spanning tree */
	int4		cs_parent_node;		/* real parent of node */
	int4		cs_combine;		/* combining service for reel */
	int4		cs_nbranch;		/* out branches at this node */
	int4		cs_nreels;		/* number of active reels */
};

struct cstate2 {
	int4		cs2_castid;		/* cast identifier */
	int4		cs2_event;		/* reeling event */
	int4		cs2_type;		/* event type */
	int4		cs2_npkt;		/* # of packets */
};

/*
 * combining functions
 */
#define CBN_ERR		((int4) -1)
#define CBN_ADD		((int4) -2)
#define CBN_MAX		((int4) -3)
#define CBN_MIN		((int4) -4)
#define CBN_XOR		((int4) -5)
#define CBN_OR		((int4) -6)
#define CBN_AND		((int4) -7)

#endif
