/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1990 Cornell Research Foundation and The Ohio State University
 *
 *	MB
 *
 *	$Log:	cros.h,v $
 * Revision 2.2  92/04/24  16:44:12  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/05/02  14:44:09  gdburns
 * Ohio release
 * 
 *	
 *	MB, last modified 900312
 *
 *	Header file for all the CrOS library code.
 *
 */

#ifndef CROS_H

#define CROS_H

#include <net.h>

/* bounds */

/*
 *  Max grid dimensions.
 *  Must allow more than 16 simultaneous virtual circuits
 *  before MAXDIMS is increased beyond 4!
 */
#define MAXDIMS		4

					
/* common symbols */

#undef NULL
#undef OK
#undef ERROR
#undef TRUE
#undef FALSE

#define NULL	0
#define NULLPTR ((char *) 0)
#define OK	0
#define ERROR	(-1)
#define TRUE	1
#define FALSE	0

#define PUBLIC	/* not static */
#define PRIVATE	static


/* events */

#define	NOTEVENT	0x80000001

#define CTOHOST		(4*MAXDIMS+16+2)
#define CFROMHOST	(4*MAXDIMS+16+6)


/* message types */

#define CRDWR		0x001
#define CDUMP		0x004
#define CLOAD		0x008
#define CBCAST		0x010
#define CCOMBINE	0x020
#define CCONCAT		0x040
#define CCONCATS	0x080
#define CBCASTCP	0x100
#define CCOMBCP		0x200
#define CCOMBCPE	0x400
#define CWELD		0x800


/* error codes */

#define ENUMPROC	601	/* asked griddim for 0 or too many processors */
#define EBADMASK	602	/* channel mask has too many or few bits set */
#define EBADCHAN	603	/* channel mask includes invalid set bit(s) */
#define EBADBUF		604	/* null data buffer */
#define EOVERLAP	605	/* input and output data buffers overlap */
#define ESIZEDIFF	606	/* insize != outsize in self-vshift */
#define ENUMPROCS	607	/* bufmap[0] does not fit actual nproc */
#define ENOGRID		608	/* gridinit() not called */
#define EGRIDDIM	609	/* grid dimension does not match */
#define EDATASIZE	610	/* data block too large for buffer or system */
#define EFAKEGRID	611	/* welding logical links that are not physical*/


/* macros to twiddle channel mask bits */

#define rightbit(x)	((x) & ~((x)-1))		/* isolate LS set bit */
#define leftbits(x)	((x) & ((x)-1))			/* clear LS set bit */
#define ismultibit(x)	(leftbits(x) != 0)		/* > one bit set */
/*
#define isonebit(x)	(leftbits(x) == 0)
#define isevenbit(x)	(((x) & 0x5555) != 0)
#define isoddbit(x)	(((x) & 0xAAAA) != 0)
*/


/*
 *  Types ands structures:
 */

struct cubenv {
	int	doc;
	int	procnum;
	int	nproc;
	int	cpmask;
	int	cubemask;
	long	cubeid;		/* added: host-->cube multicast ID */
};


/* external variables */

extern int errno;
extern int cwelded;


/* function prototypes */

#ifdef ANSI_C		/* s.b. __STDC__ but the Apollo lint chokes! */

int	cread (char *buf, unsigned int inmask,outmask, int nbytes);
int	cwrite (char *buf, unsigned int mask, int nbytes);
int	vread (char *buf, unsigned int inmask,outmask, int size,offset,nitems);
int	vwrite (char *buf, unsigned int mask, int size, offset, nitems);

int	broadcast (char *buf, int src, unsigned int mask, int nbytes);
int	combine (char *buf, int (*func)(), int size, int nitems);
int	concat (char *outbuf, int nbytes, char *inbuf, int binsz, int *szarray);
int	cshift (char *inbuf,  unsigned int inmask,  int inbytes,
		char *outbuf, unsigned int outmask, int outbytes);
int	vshift (char *inbuf, unsigned int inmask, int insize, inoffset, initems,
	  char *outbuf, unsigned int outmask, int outsize, outoffset, outitems);

int	bcastcp  (char *buf, int nbytes);
int	bcastelt (char *buf, int nbytes);
int	combcp  (char *buf, int size, int nitems);
int	combelt (char *buf, int (*func)(), int size, int nitems);
int	dumpelt (char *buf, int nbytes);
int	fcread  (int fd, unsigned int inmask,  int nbytes);
int	fcwrite (int fd, unsigned int outmask, int nbytes);
int	fdumpcp (int fd, int nbytes, int *bufmap);
int	floadcp (int fd, int *bufmap);
int	loadelt (char *buf, int nbytes);
int	mdumpcp (char *buf, int nbytes, int *bufmap);
int	mloadcp (char *buf, int *bufmap);

int	cubeld (int doc, char *nodeprog);
int	cubeldl (int doc, char *nodeprog, char *argv0, ...);
int	cubeldv (int doc, char *nodeprog, char **argv);

int	gridinit (int dim, int *num);
int	gridsplit (int proc, int dim, int *num);
int	gridcoord (int proc, int *coord);
int	gridproc (int *coord);
int	gridchan (int proc, int dir, int sign);
int	cparam (struct cubenv *env);

unsigned int fullmask (void);
int	griddims (int *num);
int	sendnbr (unsigned int mask, struct nmsg *nheader);
int	recvnbr (unsigned int mask, struct nmsg *nheader);
int	isoddnode (unsigned int mask);

int	cweld (void);
int	cunweld (void);

#else	/* not ANSI C */

extern int cread();
extern int cwrite();
extern int vread();
extern int vwrite();
extern int broadcast();
extern int combine();
extern int concat();
extern int cshift();
extern int vshift();
extern int bcastcp();
extern int bcastelt();
extern int combcp();
extern int combelt();
extern int dumpelt();
extern int fcread();
extern int fcwrite();
extern int fdumpcp();
extern int floadcp();
extern int loadelt();
extern int mdumpcp();
extern int mloadcp();
extern int cubeld();
extern int cubeldl();
extern int cubeldv();
extern int gridinit();
extern int gridsplit();
extern int gridcoord();
extern int gridproc();
extern int gridchan();
extern int cparam();
extern unsigned int fullmask();
extern int griddims();
extern int sendnbr();
extern int recvnbr();
extern int isoddnode();
extern int cweld();
extern int cunweld();

#endif	/* __STDC__ */

extern int csend();
extern int crecv();

#endif	CROS_H
