/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	kio.h,v $
 * Revision 2.2  92/04/30  14:45:31  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/01/10  17:38:42  gdburns
 * Add TID, the process's trace identifier.
 * 
 * Revision 2.1.1.2  91/12/16  15:00:11  gdburns
 * Add RTF_MALLOC switch.
 * 
 * Revision 2.1.1.1  91/10/03  15:14:25  gdburns
 * Add runtime support for traced.
 * 
 * Revision 2.1  91/03/20  11:34:50  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/01/14  15:34:53  gdburns
 * Pick_aux* are now local parameters in ksend/krecv.
 * 
 * Revision 2.0.1.2  90/10/18  23:34:16  gdburns
 * Added RTF_DEBUG, a hook for debuggers.
 * 
 * Revision 2.0.1.1  90/07/13  19:28:24  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:55:10  trillium
 * Homo erectus
 * 
 *	Function:	- define templates and constants for the
 *			  Kernel Input Output block
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 */

#ifndef _KIO
#define _KIO

#include <kreq.h>
#include <ksignal.h>
#include <ktrace.h>
#include <rreq.h>

#define MAXKHOLD	32		/* max # of virtual circuits */
#define MAXRCACHE	29		/* max # of cached routes */
#define HOLDBITS	0xf		/* mask virtual circuit index bits */
#define	TNEXIT		8		/* maximum number of exit routines */

#undef  NULL
#define NULL		((struct khold_t *) 0)

/*
 * Trillium Runtime Flags
 */
#define RTF_KERNEL	0x0001		/* attached to kernel? */
#define RTF_KENYA	0x0002		/* attached to kenya? */
#define RTF_NODEID	0x0004		/* ki_nodeid is valid? */
#define RTF_SYSGEN	0x0008		/* system process? */
#define RTF_NOWAIT	0x0010		/* background process? */
#define RTF_ONCHIP	0x0020		/* on chip stack */
#define RTF_DEBUG	0x0040		/* for debuggers */
#define RTF_I860	0x0080		/* single i860 process */
#define RTF_TRACE	0x0100		/* traced process? */
#define RTF_TRSWITCH	0x0200		/* trace on/off switch */
#define RTF_TRON	0x0300		/* comb. of TRACE & TRSWITCH */
#define RTF_MALLOC	0x0400		/* runtime malloc active */

/*
 * virtual circuit descriptor
 */
struct khold_t {
	int		kh_event;	/* KHOLD event number */
	int		kh_peer_port;	/* peer's primary communication port */
	int		kh_peer_aux;	/* peer's auxiliary comm. port */
	int		kh_aux;		/* my auxiliary communication port */
	int		kh_auxd;	/* my auxiliary comm. port descriptor */
	int		kh_flags;	/* need to remember KREVPORT */
};

/*
 * Kernel Interface
 */
struct kio_t {				/* Kernel Input Output template */
	int		ki_client;	/* read replies from kernel */
	int		ki_port;	/* primary communciation port */
#define ki_break	ki_port		/* signal end of KHOLD, ITB */
#define ki_udpdata	ki_port		/* data transmission, UDP */
#define ki_msgdata	ki_port		/* data transmission, msgop(2) */
	int		ki_aux;		/* auxiliary communication port */
#define ki_itbdata	ki_aux		/* data transmission, ITB */
#define ki_sync		ki_aux		/* synchronize data transmission, UDP */
#define ki_sender	ki_aux		/* sender's PID, msgop(2) */
	int		ki_peer_port;	/* peer's ki_port */
#define ki_peer_break	ki_peer_port
#define ki_peer_udpdata	ki_peer_port
#define ki_peer_msgdata	ki_peer_port
	int		ki_peer_aux;	/* peer's ki_aux */
#define ki_peer_itbdata	ki_peer_aux
#define ki_peer_sync	ki_peer_aux
#define ki_peer_sender	ki_peer_aux
	int		ki_portd;	/* primary comm. descriptor */
	int		ki_auxd;	/* auxiliary comm. descriptor */
	int		ki_clientd;	/* kernel reply comm. descriptor */
/*	int		ki_pick_aux;*/	/* my chosen auxiliary port */
/*	int		ki_pick_auxd;*/	/* my chosen auxiliary descriptor */
	int		ki_kernel;	/* the kernel's request port */
	int		ki_priority;	/* my kernel process priority */
	int		ki_rtf;		/* Run Time Flags */
	int		ki_slush;	/* for KCLOBBER - bad engineering */
	int		ki_signal;	/* pending signals */
	int		ki_sigmask;	/* signal masks */
	int		ki_sigretry;	/* flags to retry after EINTR */
	int		ki_index;	/* kernel's way of identifying me */
	int		ki_nodeid;	/* my node identifier */
	int		ki_origin;	/* origin node identifier */
	int		ki_stdin;	/* standard input descriptor */
	int		ki_stdout;	/* standard output descriptor */
	int		ki_stderr;	/* standard error descriptor */
/*
 * KHOLD (local virtual circuit) Table
 */
	struct khold_t	ki_khold[MAXKHOLD];
/*
 * Signal Handling Table
 */
	int		((*(ki_sig_handlers[TNSIG]))());
	int		((*(ki_sig_default[TNSIG]))());
/*
 * at exit table
 */
	int		((*(ki_exit[TNEXIT]))());
	char		*ki_arg[TNEXIT];
/*
 * route cache
 */
	struct route	ki_route[MAXRCACHE];
/*
 * trace support
 */
	struct trhead	*ki_trhead;	/* trace buffer headers */
	int		ki_trcurr;	/* current trace buffer index */
	int		ki_tid;		/* trace process identifier */
};

#endif
