/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	preq.h,v $
 * Revision 2.2  92/04/30  14:46:12  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:35:06  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:28:05  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:53  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:10:56  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:46:21  trillium
 * Initial revision
 * 
 *
 *	Function:	- define request and reply structures for the
 *			  process control server (aka kenya)
 */


#ifndef _PREQ

#define _PREQ

#include "kreq.h"
#include "t.out.h"

#define PCREATE		1			/* create a process in place */
#define PDOOM		2			/* doom a process */
#define PSTATE		3			/* find state of a process */
#define PSPAWN		4			/* spawn an existing process */
#define PDETACH		5			/* detach an existing process */
#define PATTACH		6			/* attach an existing process */

#define PREQS		6			/* number of kenya requests */
#define PNORMAL		1			/* normalize request to 0 */
#ifndef NOTPROCESS
#define NOTPROCESS	0x80000000
#endif
#define PSMAXNAME	64
#define PSTKSIZE	0x2000			/* default stack size */

/*
 * process request selection flags
 */
#define SELECT_PID	0x01
#define SELECT_INDEX	0x02
#define SELECT_BLKD	0x04
#define SELECT_LOADPT	0x08
#define SELECT_ALL	0x10
#define SELECT_SYS	0x20
#define SELECT_APPL	0x40

/*
 * kenyad request
 */
struct preq {
	int		pq_src_node;
	int		pq_src_event;
	int		pq_req;
	int		pq_parm1;
#define pq_sel_method	pq_parm1		/* process selection method */
#define pq_datapt	pq_parm1		/* address of data segment */
	int		pq_parm2;
#define pq_sel_value	pq_parm2		/* process selection value */
#define pq_loadpt	pq_parm2		/* load module flat tag */
	int		pq_parm3;
#define pq_flags	pq_parm3		/* kenyad option flags */
#define pq_status	pq_parm3		/* exit status */
#define pq_signal	pq_parm3		/* deliver these signals */
	int		pq_parm4;
#define pq_stksize	pq_parm4		/* stack size requirement */
	int		pq_parm5;
#define pq_argc		pq_parm5		/* number of runtime args */
	int		pq_parm6;
#define pq_argv		pq_parm6		/* argument structure tag */
	int		pq_parm7;
#define pq_nodeid	pq_parm7		/* spawn target nodeid */
#define pq_pid		pq_parm7		/* detach this process ID */
	int		pq_parm8;
#define pq_stack	pq_parm8		/* stack address */
#define pq_stdin	pq_parm8		/* std input desc. */
	int		pq_parm9;
#define pq_stdout	pq_parm9		/* std input desc. */
	int		pq_parm10;
#define pq_stderr	pq_parm10		/* std input desc. */
};

/*
 * kenyad reply
 */
struct preply {
	int		pr_reply;
	int		pr_result1;
#define pr_pid		pr_result1		/* kenyad process ID */
	int		pr_result2;
#define pr_index	pr_result2		/* created process index */
#define pr_nodeid	pr_result2		/* kenyad node ID */
};

/*
 * kenayd process descriptor
 */
struct pprocess {
	int		p_pid;
	int		p_stksize;
	char		*p_stack;
	char		*p_static;
	struct exec	*p_loadpt;
	int		p_argc;
	char		**p_argv;
	int		p_flags;
	int		p_nodeid;		/* parent node ID */
	int		p_event;		/* parent's reply event */
};

/*
 * kenyad reply for the PSTATE request
 */
struct pstate {
	int		ps_reply;
	int		ps_index;
	struct pprocess	ps_kenya;
	KS_STATE	ps_kernel;
	char		ps_name[PSMAXNAME];
};

#define PSTATESIZE (sizeof(struct pstate) - PSMAXNAME)

#endif
