/*
 *	OSU Research Computing & Cornell Theory Center		Trillium
 *
 *	(third party software provided under license)
 *
 *	$Log:	t.out.h,v $
 * Revision 2.2  92/04/30  14:46:20  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:35:10  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/03/18  13:16:44  gdburns
 * Use a special nlist structure for LSTT, which disagrees
 * on structure packing with ULS and Sun.
 * 
 * Revision 2.0.1.2  90/09/23  00:36:23  gdburns
 * Added N_TDB.
 * 
 * Revision 2.0.1.1  90/07/13  19:28:30  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:55:13  trillium
 * Homo erectus
 * 
 *
 *	We're just putting our stamp on this for housekeeping reasons.
 */

#ifndef _TOUT

#define _TOUT

/*	
 *	This is the transputer equivalent of a.out.h, altered for use
 *	with the Unidot Language System (ULS).  For usage information
 *	with LSTT, see below.  For further information on t.out.h, refer
 *	to the Unix man page on a.out.  The differences between a.out
 *	and this format are listed below.
 *
 *	We have two kinds of magic numbers.  TO_MAGIC is produced by
 *	the assembler and the output of tld.  It indicates a normal
 *	object file where all unknown prefix values are left as
 *	8 byte things.  In fact, in files with the magic number
 *	TO_MAGIC the unknown prefix values aren't even filled in,
 *	they're left as zero bytes.  A reference to the prefix is 
 *	stored in the relocation entries (discussed in a moment).
 *	TO_MAGIC_X is produced as the output of tbn, the final stage
 *	after linking.  In this file, all of the text relocations
 *	have been removed from the file and the prefixes are
 *	optimized.  The upshot is that TO_MAGIC files can be linked
 *	and relinked and so forth, but TO_MAGIC_X files cannot, 
 *	they can only be executed.  ULS tools do not pass the specific
 *	transputer type to tas (though they know it, from ulsmch), so
 *	a generic magic number must be used.
 *
 *	The values of symbols and relocation entries might be in
 *	bytes or words, depending are where they occur.  This is
 *	done to match the transputer where sometimes offsets might
 *	be bytes, sometimes words.
 *
 *	All of the items in struct exec are in bytes, size of data segment,
 *	text segment, etc.
 *
 *	When a symbol (struct nlist) is a text symbol, its value (n_value)
 *	is in bytes from the base of the text segment.
 *
 *	When a symbol is a data or bss symbol, its value is
 *	in words from the base of the data segment.  
 *
 *	The pc portion of a text relocation entry (struct to_relative) is
 *	in bytes from the base of the text segment.
 *
 *	The pc portion of a data relocation entry (struct to_relative) is
 *	in words from the base of the data segment.
 *
 *	tas and tld produce output where the text segment is based at
 *	zero and the data segment is also based at zero.  bss values
 *	are placed above the data segment.  Common areas are converted
 *	by tld into a bss entry of the same name.  As far as tas and tld know,
 *	there are separate address and data spaces.  After jumps
 *	are optimized by tld, the text relocation entries are removed
 *	since they aren't useful anymore.  Text is relative and
 *	can be loaded anywhere.
 *
 *	Data relocations are saved by tbn since the data segment needs
 *	to be fixed up at load time.  see load.c in the host library
 *	to see how to fix up the data segment before loading.  In the
 *	end, the text segment can be loaded anywhere as it comes out
 *	of tld.  The data segment needs to have values in it fixed
 *	since it can contain absolute addresses.  This information is
 *	contained in the data relocation entries.
 *
 *	Relocation entries are of the form: s1 - s2 + constant
 *	This allows the following kinds of cases at assembly time:
 *
 *		x:	.byte	0	; just to make it complicated
 *			ldl	1
 *			ldl	2
 *			ldlp	3
 *			ldc	y-x+1	; <- a complicated text relocation
 *			lend
 *		y:
 *
 *	If one of the symbols isn't needed in the relocation entry,
 *	it's index is set to -1 (0 is a valid symbol index).
 *
 *	LSTT only uses t.out.h in a final conversion step.  The tcvt
 *	program converts typical LSTT .tld format into t.out format.
 *	There is no string or symbol table at present, and relocations
 *	conform to s1 - s2 + constant, though s1 and s2 are always 0.
 *	LSTT specifies the transputer type, hence specific
 *	magic numbers are used.
 */

/*
 * T212 compatible magic numbers
 */
#define TO_MAGIC_T2	0x54323132	/* "T212" */
#define TO_MAGIC_T2_X	0x54324558	/* "T2EX" */

/*
 * T414 compatible magic numbers
 */
#define TO_MAGIC_T4	0x54343134	/* "T414" */
#define TO_MAGIC_T4_X	0x54344558	/* "T4EX" */

/*
 * T800 magic compatible numbers
 */
#define TO_MAGIC_T8	0x54383030	/* "T800" */
#define TO_MAGIC_T8_X	0x54384558	/* "T8EX" */

/*
 * generic transputer magic numbers
 */
#define TO_MAGIC_TN	0x544e3030	/* "TN00" */
#define TO_MAGIC_TN_X	0x544e4558	/* "TNEX" */

#define TO_MAGIC	TO_MAGIC_TN	/* transputer ld format */
#define TO_MAGIC_X	TO_MAGIC_TN_X	/* executable, jumps optimized */

/*
 * header
 */
struct exec {
	unsigned long	a_magic;	/* magic number */
	unsigned long	a_text;		/* size of text segment */
	unsigned long	a_data;		/* size of initialized data */
	unsigned long	a_bss;		/* size of uninitialized data */
	unsigned long	a_syms;		/* size of symbol table */
	unsigned long	a_entry;	/* entry point */
	unsigned long	a_trsize;	/* size of text relocation */
	unsigned long	a_drsize;	/* size of data relocation */
};

/*
 * symbol record
 */
#ifdef lstt
struct	nlist {
	union {
		char	*n_name;	/* for use when in-core */
		long	n_strx;		/* index into file string table */
	} n_un;
	unsigned char	n_type;		/* type flag (N_TEXT,..)  */
	char		n_other;	/* used for internal memory placement */
	char		n_desc;		/* see <stab.h> */
	unsigned long	n_value;	/* value of symbol (or sdb offset) */
};
#else
struct	nlist {
	union {
		char	*n_name;	/* for use when in-core */
		long	n_strx;		/* index into file string table */
	} n_un;
	unsigned char	n_type;		/* type flag (N_TEXT,..)  */
	char		n_other;	/* used for internal memory placement */
	short		n_desc;		/* see <stab.h> */
	unsigned long	n_value;	/* value of symbol (or sdb offset) */
};
#endif

/*
 * bits used in n_other
 */
#define	N_REG	1

#define	n_hash	n_desc

#define	N_UNDF	0x0		/* undefined */
#define	N_ABS	0x2		/* absolute */
#define	N_TEXT	0x4		/* text */
#define	N_DATA	0x6		/* data */
#define	N_BSS	0x8		/* bss */
#define	N_TDB	0xa		/* tdb */
#define	N_COMM	0x12		/* common (internal to ld) */
#define	N_FN	0x1f		/* file name symbol */

#define	N_EXT	01		/* external bit, or'ed in */
#define	N_TYPE	0x1e		/* mask for all the type bits */

/*
 * Dbx entries have some of the N_STAB bits set.
 * These are given in <stab.h>
 */

#define	N_STAB	0xe0		/* if any of these bits set, a dbx symbol */

/*
 * relocation record
 */
struct	to_relative {
	unsigned	pc,		/* location to be fixed */
			type,		/* 0 - prefix, 1 - word */
			s1_external,	/* for ld's use */
			s2_external,
			s1_text;	/* short cut for downloader */
	int		cnst;
	union {
	  int		index;		/* symbol table index */
	  struct nlist	*p;
	} s1;
	union {
	  int		index;		/* symbol table index */
	  struct nlist	*p;  		/* for use in core */
	} s2;
};

/*
 * offsets for x as a structure
 */
#define	N_TXTOFF(x) \
	(sizeof (struct exec))

#define	N_SYMOFF(x) \
	(N_TXTOFF(x)+(x).a_text+(x).a_data+(x).a_trsize+(x).a_drsize)

#define	N_STROFF(x) \
	(N_SYMOFF(x) + (x).a_syms)

#define	N_BADMAG(x) \
	((x).a_magic != TO_MAGIC)

/*
 * offsets for x as a structure pointer
 */
#define P_TXTOFF(x) \
	(sizeof (struct exec))

#define P_RELOFF(x) \
	(P_TXTOFF(x) + (x)->a_text)

#define P_SYMOFF(x) \
	(P_RELOFF(x) + (x)->a_trsize + (x)->.a_drsize)

#define P_DATOFF(x) \
	(P_SYMOFF(x) + (x)->a_syms)

#define P_BSSOFF(x) \
	(P_DATOFF(x) + (x)->a_data)

#endif
