/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	btfrozen.c,v $
 * Revision 2.2  92/04/30  10:19:43  trillium
 * Ohio Release
 * 
 *	Function:	- contains all routines that operate on the
 *			  frozen file
 */

#include <bootreq.h>
#include <net.h>
#include <shared.h>
#include <typical.h>

#undef NULL
#define NULL		((char *) 0)

/*
 * global functions
 */
void			btinitfrozen();
int			btaddfrozen();
int			btflushfrozen();
int			btstagefrozen();
int			btthaw();

/*
 * external variables
 */
extern int		errno;			/* global error variable */
extern struct bootreq	*request;		/* bootd request */

/*
 * external functions
 */
extern void		btmosescat();
extern int		rbtaddfrozen();		/* add frozen object to bootd */

/*
 * local variables
 */
static char		*pfrozen;		/* ptr to frozen file */
static int4		sfrozen;		/* size of frozen file */

/*
 *	btinitfrozen
 *
 *	Function:	- initializes frozen file ptr
 */

void
btinitfrozen()

{
	pfrozen = NULL;
	sfrozen = 0;
}

/*
 *	btaddfrozen
 *
 *	Function:	- identifies a frozen file and sets frozen file ptr
 *	Returns:	- 0 or ERROR
 */

int
btaddfrozen()

{
	if ((pfrozen = (char *) rforget(LOCAL, request->btq_tag)) == NULL) {
		return(ERROR);
	}

	sfrozen = request->btq_size;
	return(0);
}

/*
 *	btflushfrozen
 *
 *	Function:	- frees frozen table
 *	Returns:	- 0 or ERROR
 */

int
btflushfrozen()

{
	int		r;		/* return value */

	r = rfree(LOCAL, pfrozen);
	btinitfrozen();
	return(r);
}

/*
 *	btthaw
 *
 *	Function:	- boots frozen file over link
 *	Accepts:	- output link
 *	Returns:	- 0
 */

int
btthaw(ilink)

int4			ilink;

{
	btmosescat(ilink, pfrozen, sfrozen);
	return(0);
}

/*
 *	btstagefrozen
 *
 *	Function:	- ships frozen file to remote bootd
 *	Accepts:	- destination node ID
 *	Returns:	- 0 or ERROR
 */

int
btstagefrozen(nodeid)

int4			nodeid;

{
	if (pfrozen == NULL) {
		return(0);
	}
/*
 * Ship the frozen file over to the newly booted node.
 */
	if (rflat(nodeid, pfrozen, sfrozen, sfrozen, getpid()) < 0) {
		return(ERROR);
	}

	if (rbtaddfrozen(nodeid, getpid(), sfrozen) < 0) {
		return(ERROR);
	}

	return(0);
}
