/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	btroute.c,v $
 * Revision 2.2  92/04/30  10:19:44  trillium
 * Ohio Release
 * 
 *	Function:	- contains all routines that operate on the
 *			  route file
 */

#include <all_hash.h>
#include <bootreq.h>
#include <net.h>
#include <rreq.h>
#include <rtio.h>
#include <shared.h>
#include <terror.h>
#include <typical.h>

#undef NULL
#define NULL		((char *) 0)

/*
 * global functions
 */
void			btinitroute();
int			btaddroute();
int			btflushroute();
int			btrout();
int			btstageroute();

/*
 * external variables
 */
extern int		errno;			/* global error variable */
extern struct bootreq	*request;		/* bootd request */

/*
 * external functions
 */
extern int		rbtadd();		/* add object to bootd */

/*
 * local variables
 */
static char		*proute;		/* ptr to route file */
static int4		sroute;			/* size of route file */

/*
 *	btinitroute
 *
 *	Function:	- initializes route file ptr
 */

void
btinitroute()

{
	proute = NULL;
	sroute = 0;
}

/*
 *	btaddroute
 *
 *	Function:	- identifies a route file and sets route file ptr
 *	Returns:	- 0 or ERROR
 */

int
btaddroute()

{
	struct rtiohead	*phead;

	if ((proute = (char *) rforget(LOCAL, request->btq_tag)) == NULL) {
		return(ERROR);
	}

	phead = (struct rtiohead *) proute;

	if (phead->rh_magicno != RTIO_MAGIC) {
		errno = EMAGIC;
		return(ERROR);
	}

	getsmart();
	sroute = sizeof(struct rtiohead) + (phead->rh_nentries *
		sizeof(struct rtiodir)) + (phead->rh_nentries *
		getnall() * sizeof(struct route));
	return(0);
}

/*
 *	btflushroute
 *
 *	Function:	- frees route table
 *	Returns:	- 0 or ERROR
 */

int
btflushroute()

{
	int		r;		/* return value */

	r = rfree(LOCAL, proute);
	btinitroute();
	return(r);
}

/*
 *	btrout
 *
 *	Function:	- sends network information and route table to
 *			  router on given node
 *			- net info and route table taken from locally
 *			  staged route file
 *			- does nothing if route file is not present
 *
 *	Accepts:	- destination node ID
 *	Returns:	- 0 or ERROR
 */

int
btrout(nodeid)

int			nodeid;

{
	struct rtiohead *phead;		/* route file header ptr */
	struct rtiodir	*pindex;	/* index table ptr */
	struct rtiodir	*p1index;	/* index table element ptr */
	struct route    *ptroute;	/* route table ptr */
	struct route    *p1route;	/* route table element ptr */
	struct net_info	net;		/* basic network info */
	SHASH		ihd;		/* index table descriptor */
	SHASH		rhd;		/* route table descriptor */

	if (proute == NULL) {
		return(0);
	}
/*
 * Set pointers to the route file header and index table.
 */
	phead = (struct rtiohead *) proute;
	pindex = (struct rtiodir *) (proute + sizeof(struct rtiohead));
/*
 * Initialize the static hash descriptor for the index table.
 */
	if (ahs_init(phead->rh_nentries, (int4) sizeof(struct rtiodir),
		(int4) NOTNODEID, (int4) (AHFIXED | AHNOINIT),
		(void *) pindex, (int4 *) 0, &ihd) == (SHASH *) 0) {

		return(ERROR);
	}
/*
 * Find the index element for the given node.
 */
	p1index = (struct rtiodir *) ahs_find(&ihd, nodeid);

	if (p1index == (struct rtiodir *) 0) {
		errno = EBADNODE;
		return(ERROR);
	}
/*
 * Set pointer to the given node's route table.
 */
	ptroute = (struct route *) (proute + p1index->rd_rtoffset);
/*
 * Initialize the static hash descriptor for the individual route table.
 */
	if (ahs_init(phead->rh_nentries, (int4) sizeof(struct route),
		(int4) NOTNODEID, (int4) (AHFIXED | AHNOINIT),
		(void *) ptroute, (int4 *) 0, &rhd) == (SHASH *) 0) {

		return(ERROR);
	}
/*
 * Find the given node's own route entry in its route table.
 */
	p1route = (struct route *) ahs_find(&rhd, nodeid);

	if (p1route == (struct route *) 0) {
		panic(EIMPOSSIBLE);
	}
/*
 * Fill in network information for the given node.
 */
	net.ni_nodetype = p1route->r_nodetype;
	net.ni_origin = phead->rh_rootnode;
	net.ni_nodeid = nodeid;
/*
 * Ship the basic network information to the destination router.
 */
	if (rsetnet(nodeid, &net)) {
		return(ERROR);
	}
/*
 * Ship the route table to the destination router.
 */
	if (rroute(nodeid, ptroute, phead->rh_nentries)) {
		return(ERROR);
	}

	return(0);
}

/*
 *	btstageroute
 *
 *	Function:	- ships route file to remote bootd
 *	Accepts:	- destination node ID
 *	Returns:	- 0 or ERROR
 */

int
btstageroute(nodeid)

int4			nodeid;

{
	if (proute == NULL) {
		return(0);
	}
/*
 * Ship the route file over to the newly booted node.
 */
	if (rflat(nodeid, proute, sroute, sroute, getpid()) < 0) {
		return(ERROR);
	}

	if (rbtadd(nodeid, BTO_ROUTE, getpid()) < 0) {
		return(ERROR);
	}

	return(0);
}
