/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	btstrap.c,v $
 * Revision 2.2  92/04/30  10:19:45  trillium
 * Ohio Release
 * 
 *	Function:	- all routines to manage the bootstrap loader
 */

#include <bootd.h>
#include <bootreq.h>
#include <net.h>
#include <shared.h>
#include <terror.h>
#include <typical.h>

/*
 * global functions
 */
void			btinitstrap();		/* initialize bootstrap */
int			btaddstrap();		/* add bootstrap */
int			btexport();		/* load & execute bootstrap */
int			btflushstrap();		/* free bootstrap */
int			btstagestrap();		/* stage bootstrap remotely */

/*
 * external functions
 */
extern int		rbtadd();		/* add object to bootd */
extern int4		touttextsize();		/* text segment size */
extern void		out();			/* output to link */
extern void		*touttextptr();		/* ptr to text segment */

/*
 * external variables
 */
extern int		errno;			/* global error number */
extern struct bootreq	*request;		/* bootd request */

/*
 * local variables
 */
static struct btprog	strap;			/* bootstrap prog desc. */

/*
 *	btinitstrap
 *
 *	Function:	- initializes bootstrap module
 */

void
btinitstrap()

{
	strap.btp_load = (void *) 0;
}

/*
 *	btaddstrap
 *
 *	Function:	- set pointer to bootstrap code
 *	Returns:	- 0 or ERROR
 */

int
btaddstrap()

{
	strap.btp_load = (void *) rforget(LOCAL, request->btq_tag);

	if (strap.btp_load == (void *) 0) {
		return(ERROR);
	}

	return(0);
}

/*
 *	btflushstrap
 *
 *	Functions:	- frees memory of bootstrap load module
 *	Returns:	- 0 or ERROR
 */

int
btflushstrap()

{
	int		r;		/* return value */

	if (strap.btp_load) {
		r = rfree(LOCAL, (char *) strap.btp_load);
		btinitstrap();
		return(r);
	} else {
		errno = ENOBOOTSTRAP;
		return(ERROR);
	}
}

/*
 *	btexport
 *
 *	Function:	- loads bootstrap into a transputer
 *			- loads the text and data segments of the strap
 *			  program through a link to a transputer
 *			  ready to boot from a link
 *			- first byte through the channel is the length
 *			  of the following load module
 *			- after load, the booting transputer automatically
 *			  transfers control to the first loaded byte
 *			- obviously, transputer specific routine
 *
 *	Accepts:	- link number
 *	Returns:	- 0 or ERROR (sets errno)
 */

int
btexport(ilink)

int				ilink;

{
	unsigned char		length;		/* text and data length */
	void			*alink;		/* link address */

	if (strap.btp_load) {
/*
 * This is transputer specific code.
 * Derive link address from link index.
 */
		alink = (void *) (0x80000000 + (ilink * 0x4));
		length = (char) touttextsize(strap.btp_load);
		out(alink, (char *) &length, sizeof(char));
		out(alink, (char *) touttextptr(strap.btp_load),
			(int) ((unsigned int) length));
/*
 * end of transputer specific code
 */
		return(0);
	} else {
		errno = ENOBOOTSTRAP;
		return(ERROR);
	}
}

/*
 *	btstagestrap
 *
 *	Function:	- sends a bootstrap to neighbour bootd
 *	Accepts:	- neighbour nodeid
 *	Returns:	- 0 or ERROR
 */

int
btstagestrap(nodeid)

int4			nodeid;

{
	int4		strapsize;

	if (strap.btp_load == (void *) 0) {
		errno = ENOBOOTSTRAP;
		return(ERROR);
	}

	strapsize = touthtxtsize(strap.btp_load);
/*
 * Load the bootstrap to the remote node.
 */
	if (rflat(nodeid, (char *) strap.btp_load, strapsize,
			strapsize, getpid()) < 0) {
		return(ERROR);
	}

	return(rbtadd(nodeid, BTO_STRAP, getpid()));
}
