/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	dlo_ortho.c,v $
 * Revision 2.2  92/04/30  10:23:05  trillium
 * Ohio Release
 * 
 *	Function:	- datalink output program for transputer link
 *			- orthodox version
 */

#include <dlreq.h>
#include <net.h>
#include <kio.h>
#include <ksignal.h>
#include <pcb.h>
#include <portable.h>
#include <shared.h>
#include <terror.h>
#include <typical.h>
#ifdef itb
#include <c.h>
#endif

/*
 * external variables
 */
extern int		errno;		/* global error code */
extern struct kio_t	_kio;		/* Kernel I/O */
extern struct pcb_t	*_pcb;		/* Process Control Block */

/*
 * local variables
 */
static char		*myname = "dlo_ortho";
static int		filter;		/* message filter */

/*
 * local functions
 */
static void		valve();	/* positive event valve */


void
dlo_ortho(outlink)

int			outlink;	/* output channel address */

{
	int		ihold;		/* index of our dl event in VC tbl */
	int		mask;		/* signal mask */
	int		r;		/* return value from drecv() */
	struct nmsg	nhead;		/* network message descriptor */
	char		nmsg[MAXNMSGLEN];

/*
 * Connect to the kernel.
 */
	if (kinit(0) < 0) {
		panic(errno);
	}
/*
 * Become a high priority process.
 */
	setpri(0);
/*
 * Connect to dld.
 */
	if (ldlattach(outlink) < 0) {
		panic(errno);
	}
/*
 * Catch SIGA and SIGB.
 */
	if ((ksignal(SIGA, valve) == 0) ||
		(ksignal(SIGB, valve) == 0)) {

		panic(errno);
	}

	mask = ksigretry(0);
	ksigsetretry(mask & ~(sigmask(SIGA) | sigmask(SIGB)));
	filter = DLALL;
/*
 * Fuss with our arguments.
 */
	_pcb->pc_argv = &myname;
	_pcb->pc_argc = 1;
/*
 * Connect to kenyad.
 */
	if (rpattach(LOCAL) < 0) {
		panic(errno);
	}

	ihold = (outlink & HOLDBITS) + ((outlink & INT4_SIGN)
		? (HOLDBITS + 1) : 0);

	for (;;) {
/*
 * Receive network frame.
 */
		nhead.nh_event = outlink;
		nhead.nh_type = filter;
		nhead.nh_flags = DFORWARD;
		nhead.nh_length = MAXNMSGLEN;
		nhead.nh_msg = nmsg;

		while (((r = drecv(&nhead)) < 0) && (errno == EINTR)) {
			nhead.nh_type = filter;
		}

		if (r < 0) {
			panic(errno);
		}
/*
 * Pull a very sneaky manouver to maintain a virtual circuit if
 * so indicated by the header.
 */
		if (nhead.nh_flags & KHOLD) {
			_kio.ki_khold[ihold].kh_event = outlink;
		}

		out(outlink, (char *) &nhead, sizeof(struct nmsg));

		if (nhead.nh_length > 0) {
			out(outlink, nmsg, nhead.nh_length);
		}
	}
}

/*
 *	valve
 *
 *	Function:	- turns positive event value on or off by
 *			  altering the message type
 *	Accepts:	- signal number
 */

static void
valve(sig)

int4			sig;

{
	filter = (sig == SIGA) ? DLALL : DLSYS;
}
