/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	clientio.c,v $
 * Revision 2.2  92/04/30  10:28:25  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/26  16:40:15  gdburns
 * Support buffer flow control.
 * 
 * Revision 2.1.1.1  92/02/23  19:10:10  gdburns
 * Add #ifdef on c.h to facilitate make depend.
 * 
 * Revision 2.1  91/03/20  03:01:39  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.5  91/01/14  15:32:38  gdburns
 * Changed pick_aux* from static kio to passed parameters.
 * 
 * Revision 2.0.1.4  90/11/29  19:52:37  gdburns
 * Converted to new filed.
 * 
 * Revision 2.0.1.3  90/11/16  16:00:41  raja
 * Added siesta list handling to enable datalinks to run at hi-pri.
 * 
 * Revision 2.0.1.2  90/07/13  19:26:48  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/07/02  16:46:13  gdburns
 * Added _kbail() for machine dependent process termination.
 * 
 * Revision 2.0  90/02/20  11:54:07  trillium
 * Homo erectus
 *
 *	Function:	- interface between kernel runtime request library
 *			  and native environment
 *			- special T414/T800 datalink version
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#ifdef itb
#include <c.h>
#endif
#include <freq.h>
#include <kernel.h>
#include <kio.h>
#include <kmsg.h>
#include <typical.h>
#include <terror.h>
#include <transputer/transputer.h>
#include <machine/maconfig.h>

/*
 * external variables
 */
extern int		errno;			/* global error variable */
extern struct kio_t	_kio;			/* Kernel I/O block */

/*
 * external functions
 */
extern void		_follow_signal();	/* check and follow signals */
extern void		take_siesta();		/* stop the process */

/*
 * global functions
 */
int			_init_kio();		/* initialize kernel I/O */
int			_make_kreq();		/* make a kernel request */
int			_send_kmsg();		/* send message to peer */
int			_recv_kmsg();		/* receive message from peer */
void			_kbail();		/* terminate process */

/*
 * global variables
 */
int			channel;		/* datalink's I/O port */
struct fclient		_ufd[FUMAX] = {0};	/* user file desc. */

/*
 *	_init_kio
 *
 *	Function:	- initialize kernel interface
 *	Returns:	- error indicator
 */

int
_init_kio()

{
	_kio.ki_portd = (int) NOTPROCESS;
	_kio.ki_clientd = (int) NOTPROCESS;
	_kio.ki_port = (int) &_kio.ki_portd;
	_kio.ki_client = (int) &_kio.ki_clientd;
	_kio.ki_itbdata = channel;

	return(0);
}

/*
 *	_createport
 *
 *	Function:	- get the auxiliary port address
 *			- initializes descriptor
 *	Accepts:	- ptr to a communication descriptor
 *	Returns:	- ptr to the same descriptor
 */

int
_createport(chan)

int			*chan;

{
	*chan = NOTPROCESS;
	return((int) chan);
}

/*
 *	_make_kreq
 *
 *	Function:	- make a kernel request
 *			- sends the request structure to the kernel
 *			- receives the reply structure from the kernel
 *
 *	Accepts:	- ptr to kernel request structure
 *			- ptr to kernel reply structure
 *
 *	Returns:	- error indicator
 *			- sets errno accordingly if a kernel error
 *			  is returned
 */

int
_make_kreq(req, reply)

struct kreq		*req;
struct kreply		*reply;

{
	int		reply_code;	/* kernel reply code */

/*
 * Loop through possible signal interruptions.
 */
	do {
/*
 * Send the request to the kernel.  If the channel is already
 * taken, deschedule and take a small siesta.  The kernel will
 * wake you up when ready for you to continue.
 * Based on an idea by Aarne Rantala at VTT.
 */
		if ( ((*(int *) KERNEL_CHAN) > ONCHIP_START) ||
			((*(int *) FRONT_SIESTA) != NOTPROCESS) ) {

			do {
				take_siesta();
			} while ( (*(int *) KERNEL_CHAN) > ONCHIP_START);
		}

		out(KERNEL_CHAN, (char *) req, sizeof(struct kreq));
/*
 * Receive the reply from the kernel;
 */
		in(_kio.ki_client, (char *) reply, sizeof(struct kreply));

		reply_code = reply->kr_reply;

		if (reply->kr_signal) {
			_kio.ki_signal |= reply->kr_signal;

			if (reply_code == EINTR) {
				_follow_signal();
			}
		}

	} while ((reply_code == EINTR) &&
		(reply->kr_signal & _kio.ki_sigretry));

	if (reply_code != SUCCESS) {
		errno = reply_code;
		return(ERROR);
	}

	return(0);
}

/*
 *	_recv_kmsg
 *
 *	Function:	- receive a message from another process
 *	Accepts:	- a pointer to a kernel message header
 *			- my chosen aux port
 *			- my chosen aux port descriptor
 *	Returns:	- an error indicator
 */

int
_recv_kmsg(msgptr, pick_aux, pick_auxd)

struct kmsg		*msgptr;
int			pick_aux;
int			pick_auxd;

{
	char		break_hole;
	int		data_port;

	if (msgptr->k_flags & KDLI) {
/*
 * Block on break port until the destination process releases us.
 */
/*
 *		out(_kio.ki_peer_break, &break_hole, sizeof(break_hole));
 */
 		out(_kio.ki_break, &break_hole, sizeof(break_hole));
	} else {
		data_port = (msgptr->k_flags & KREVPORT) ?
			_kio.ki_peer_itbdata : pick_aux;

		if (msgptr->k_length) {
			in(data_port, msgptr->k_msg, msgptr->k_length);
		}

		if ((msgptr->k_flags & KBREAK) &&
			(data_port <= Link3In)) {
/*
 *		 	out(_kio.ki_peer_break, &break_hole,
 *				sizeof(break_hole));
 *		 	out(_kio.ki_break, &break_hole,
 *				sizeof(break_hole));
 */
 		 	out(_kio.ki_peer_break, &break_hole,
 				sizeof(break_hole));
		}
	}

	return(0);
}

/*
 *	_send_kmsg
 *
 *	Function:	- send a message to another process
 *	Accepts:	- a pointer to a kernel message header
 *			- my chosen aux port
 *			- my chosen aux port descriptor
 *	Returns:	- an error indicator
 */

int
_send_kmsg(msgptr, pick_aux, pick_auxd)

struct kmsg		*msgptr;	/* pointer to message header */
int			pick_aux;
int			pick_auxd;

{
	char		break_hole;

	if (msgptr->k_length) {
		out(_kio.ki_peer_itbdata, msgptr->k_msg, msgptr->k_length);
	}
/*
 * Do we need to break a virtual circuit with a sync message?
 */
	if ((msgptr->k_flags & KBREAK) && (_kio.ki_peer_itbdata <= Link3Out)) {

		in(_kio.ki_peer_break, &break_hole, sizeof(break_hole));
	}

	return(0);
}

/*
 *	_kcleanup
 *
 *	Function:	- a stub for this implementation
 */

int
_kcleanup()

{
	return(0);
}

/*
 *	_kbail
 *
 *	Function:	- stop transputer process
 *			- does not return
 */

void
_kbail()

{
	stopp(0);
}
