/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	dl_mom.c,v $
 * Revision 2.2  92/04/30  10:28:28  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/26  16:40:34  gdburns
 * Support buffer flow control.
 * 
 * Revision 2.1  91/03/20  03:01:40  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/11/16  16:09:27  raja
 * Change in the configuration of on-chip RAM.
 * 
 * Revision 2.0.1.1  90/07/13  19:26:44  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:05  trillium
 * Homo erectus
 *
 *	Function:	- creates datalink process pairs on transputer links
 *			- command line determines which links are used
 *			- usage: dl_t4 outlinkaddr outlinkaddr ...
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#include <dlreq.h>
#include <kernel.h>
#include <net.h>
#include <machine/maconfig.h>
#include <pcb.h>

#define DLOSTACK	0x400	
#define DLISTACK	DLOSTACK + MAXNMSGLEN

/*
 * external variables
 */
extern int		errno;			/* error number */
extern int4		ppid[MAXNLINK];		/* dlo child pids */
extern struct pcb_t	*_pcb;

/*
 * external functions
 */
extern char		*kpar_Fe();		/* heavyweight process */
extern void		dld();			/* datalink daemon */
extern void		dli_t4();		/* datalink input program */
extern void		dlo_t4();		/* datalink output program */

/*
 * local variables
 */
static char		*myname = "dld";

main(argc, argv)

int			argc;
char			*argv[];

{
	int		inlink;			/* input link address */
	int		outlink;		/* output link address */
	int		pid;			/* child process ID */
	int		i;			/* favourite index */
	char		*memstart;		/* memory alloc'd to child */

	for (i = 0; i < MAXNLINK; ++i) {
		ppid[i] = NOTPROCESS;
	}

	if (kinit(0) < 0) {
		panic(errno);
	}

	--argc;
	++argv;
/*
 * Loop through command line pairs.
 */
	while (argc) {
/*
 * Open the device.
 */
		outlink = stoi(argv[0]);
		inlink = outlink + 0x10;
/*
 * Create the datalink output process.
 */
		memstart = kpar_Fe(dlo_t4, DLOSTACK, &pid, 1, outlink);

		if (memstart == (char *) 0) {
			panic(errno);
		}

		ppid[dl_link(outlink)] = pid;
/*
 * Create the datalink input process.
 */
		memstart = kpar_Fe(dli_t4, DLISTACK, &pid, 1, inlink);

		if (memstart == (char *) 0) {
			panic(errno);
		}

		--argc;
		++argv;
	}
/*
 * Become datalink daemon.
 */
	_pcb->pc_argv = &myname;
	_pcb->pc_argc = 1;
	dld();
	return(0);
}
