/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	dli_t4.c,v $
 * Revision 2.2  92/04/30  10:28:29  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/26  16:40:37  gdburns
 * Support buffer flow control.
 * 
 * Revision 2.1.1.1  92/02/23  19:10:59  gdburns
 * Add #ifdef on c.h to facilitate make depend.
 * 
 * Revision 2.1  91/03/20  03:01:41  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  90/11/16  16:17:34  raja
 * Changed to high priority.  All local messages are routed to LOCAL.
 * 
 * Revision 2.0.1.2  90/07/14  16:02:40  gdburns
 * Added more robustness.
 * 
 * Revision 2.0.1.1  90/07/13  19:26:46  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:06  trillium
 * Homo erectus
 *
 *	Function:	- datalink input program for transputer links
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#ifdef itb
#include <c.h>
#endif
#include <net.h>
#include <typical.h>
#include <kmsg.h>
#include <kio.h>
#include <pcb.h>
#include <events.h>
#include <terror.h>

/*
 * external functions
 */
extern int		getnodeid();
extern int		kinit();
extern int		rsetroute();
extern void		_follow_signal();

/*
 * external variables
 */
extern int		errno;
extern struct kio_t	_kio;
extern struct pcb_t	*_pcb;
extern int		channel;

/*
 * local variables
 */
static char		*myname = "dli_t4";

void
dli_t4(inlink)

int			inlink;			/* input channel address */

{
	int		dl_partner;
	int		mynodeid;
	int		ks_req;
	char		break_hole;
	struct kmsg	dl_khead;
	struct netframe	dl_inframe;

/*
 * Initialize.
 */
	channel = inlink;
	dl_partner = channel - 0x10;

	dl_khead.k_length = sizeof(dl_inframe.nf_header);
	dl_khead.k_flags = KDLI;
	dl_khead.k_msg = (char *) &dl_inframe;
/*
 * Connect to the kernel.
 */
	if (kinit(0) < 0) {
		panic(errno);
	}
/*
 * Become a high priority process.
 */
	setpri(0);
/*
 * Fuss with our argv pointer.
 */
	_pcb->pc_argv = &myname;
	_pcb->pc_argc = 1;
/*
 * Connect to kenyad.
 */
	if (rpattach(LOCAL)) {
		panic(errno);
	}
/*
 * Input frame header.
 */
	in(channel, ((char *) &dl_inframe.nf_header), sizeof(struct nmsg));
/*
 * Inform the router of activity on the link.
 * NOTNODEID will cause no damage if route table is already in place.
 * Otherwise it will set the dumb route and all outbound messages will be
 * sent to my dlo partner.
 */
	(void) rsetroute(LOCAL, NOTNODEID, dl_partner, dl_partner);

	while ((mynodeid = getnodeid()) == NOTNODEID) {
/*
 * Preamble:  Do something simple and stupid until I learn my node ID.
 */
/*
 * Store and forward.
 */
	    if (dl_inframe.nf_header.nh_length) {
		in(channel, dl_inframe.nf_data, dl_inframe.nf_header.nh_length);
	    }

	    dl_inframe.nf_header.nh_flags &= ~BYBUF;
	    dl_inframe.nf_header.nh_flags |= DDLI;
	    dl_inframe.nf_header.nh_msg = dl_inframe.nf_data;
	    dl_inframe.nf_header.nh_node = LOCAL;
	    (void) nsend(&dl_inframe.nf_header);
/*
 * Input ack character if not in a virtual circuit.
 */
	    if (! (dl_inframe.nf_header.nh_flags & KHOLD)) {
		in(channel, &break_hole, sizeof(break_hole));
	    }
/*
 * Input frame header.
 */
	    in(channel, ((char *) &dl_inframe.nf_header), sizeof(struct nmsg));
	}
/*
 * End of Preamble
 */
/*
 * Input and forward datalink frames.
 */
	for (;;) {
	    dl_inframe.nf_header.nh_flags &= ~BYBUF;
/*
 * Is this frame bound for this node?
 */
	    if ((dl_inframe.nf_header.nh_node == mynodeid) ||
		(dl_inframe.nf_header.nh_node == LOCAL)) {
/*
 * Attempt to use the optimized protocol.
 * Set port reversal bit in the network header.  This reverses the normal
 * practice of using the receiver's port to communicate.  Instead, the
 * destination process will now begin using the sender's (us) port which
 * is conveniently the address of the input link channel.
 */
		dl_inframe.nf_header.nh_flags |= KREVPORT;
		dl_khead.k_event = dl_inframe.nf_header.nh_event;
		dl_khead.k_type = dl_inframe.nf_header.nh_type;
/*
 * Forward to destination event or buffer process.
 * Though we are not using a virtual circuit on this end, we do not
 * want to follow any possible signal (like SIGBACK) until after
 * the break message.  This is indicated by KDLI in k_flags.
 */
		ks_req = (dl_inframe.nf_header.nh_flags & KHOLD) ?
			KSEND : KTRY_SEND;

		if (do_ksend(&dl_khead, ks_req)) {
/*
 * We are assuming EWOULDBLOCK here and now we will not use the optimized
 * protocol.  Read in the message body for store and forward.
 */
		    if (dl_inframe.nf_header.nh_length) {
			in(channel, dl_inframe.nf_data,
			    dl_inframe.nf_header.nh_length);
		    }

		    dl_inframe.nf_header.nh_node = LOCAL;
		    dl_inframe.nf_header.nh_msg = dl_inframe.nf_data;
		    dl_inframe.nf_header.nh_flags &= ~KREVPORT;
		    dl_inframe.nf_header.nh_flags |= DDLI;

		    (void) nsend(&dl_inframe.nf_header);
		}

		else {
/*
 * Block on peer's break port (remember KREVPORT!) until destination process
 * releases us.
 */
 		    in(_kio.ki_break, &break_hole, sizeof(break_hole));

		    if (_kio.ki_signal) {
			_follow_signal();
		    }
		}
/*
 * Input ack character.
 */
		in(channel, &break_hole, sizeof(break_hole));
	    }

	    else {
/*
 * Store and forward.
 */
		if (dl_inframe.nf_header.nh_length) {
		    in(channel, dl_inframe.nf_data,
			dl_inframe.nf_header.nh_length);
		}

		dl_inframe.nf_header.nh_msg = dl_inframe.nf_data;
		dl_inframe.nf_header.nh_flags |= DDLI;

		(void) nsend(&dl_inframe.nf_header);
/*
 * Input ack character if not continuing a virtual circuit.
 */
		if (! (dl_inframe.nf_header.nh_flags & KHOLD)) {
		    in(channel, &break_hole, sizeof(break_hole));
		}
	    }
/*
 * Input frame header.
 */
	    in(channel, ((char *) &dl_inframe.nf_header), sizeof(struct nmsg));
	}
}
