/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	dlo_t4.c,v $
 * Revision 2.2  92/04/30  10:28:30  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/26  16:40:40  gdburns
 * Support buffer flow control.
 * 
 * Revision 2.1.1.1  92/02/23  19:11:02  gdburns
 * Add #ifdef on c.h to facilitate make depend.
 * 
 * Revision 2.1  91/03/20  03:01:43  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  90/11/16  16:21:02  raja
 * Changed to high priority.
 * 
 * Revision 2.0.1.2  90/07/14  16:02:56  gdburns
 * Added more robustness.
 * 
 * Revision 2.0.1.1  90/07/13  19:26:47  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:06  trillium
 * Homo erectus
 * 
 *	Function:	- datalink output program
 *			- dlo_t4 has special version of clientio.c
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#ifdef itb
#include <c.h>
#endif
#include <dlreq.h>
#include <net.h>
#include <kio.h>
#include <pcb.h>
#include <typical.h>
#include <kmsg.h>
#include <ksignal.h>
#include <terror.h>

/*
 * external functions
 */
extern int		kinit();

/*
 * external variables
 */
extern int		errno;
extern struct kio_t	_kio;
extern struct pcb_t	*_pcb;
extern int		channel;

/*
 * local variables
 */
static char		*myname = "dlo_t4";
static int		filter;		/* message type filter */

/*
 * local functions
 */
static void		valve();	/* positive event valve */

void
dlo_t4(outlink)

int			outlink;	/* output channel address */

{
	struct kmsg	dl_khead;	/* kernel message descriptor */
	char		ack;		/* acknowledgement byte */
	int		mask;		/* signal mask */

/*
 * Initialize.
 */
	dl_khead.k_event = channel = outlink;
	dl_khead.k_flags = KDLI;
	dl_khead.k_length = sizeof(struct nmsg);
/*
 * Connect to the kernel.
 */
	if (kinit(0) < 0) {
		panic(errno);
	}
/*
 * Catch SIGA and SIGB.
 */
	if ((ksignal(SIGA, valve) == 0) ||
		(ksignal(SIGB, valve) == 0)) {

		panic(errno);
	}

	mask = ksigretry(0);
	ksigsetretry(mask & ~(sigmask(SIGA) | sigmask(SIGB)));
	filter = DLALL;
/*
 * Become a high priority process.
 */
	setpri(0);
/*
 * Connect to kenyad.
 */
	_pcb->pc_argv = &myname;
	_pcb->pc_argc = 1;

	if (rpattach(LOCAL)) {
	    panic(errno);
	}

	for (;;) {
		dl_khead.k_type = filter;
/*
 * Synchronize on dlo event for the receipt of the header.
 */
		while (krecv(&dl_khead) < 0) {
			dl_khead.k_type = filter;
		}
/*
 * Send out an ack byte.
 */
		out(outlink, &ack, sizeof(ack));
	}

	/* POC */
}

/*
 *	valve
 *
 *	Function:	- turns positive event value on or off by
 *			  altering the message type
 *	Accepts:	- signal number
 */

static void
valve(sig)

int4			sig;

{
	filter = (sig == SIGA) ? DLALL : DLSYS;
}
