/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	flatd.c,v $
 * Revision 2.2  92/04/30  10:30:37  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  03:02:03  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:27:17  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/05/16  15:44:18  gdburns
 * Added extra reply message to client to ensure that the data
 * is fully loaded before the client can start announcing the tag value.
 * 
 * Revision 2.0  90/02/20  11:54:27  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:01:50  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0.1.2  88/06/01  16:56:00  gdburns
 * Whoops!  Forgot to restore correct flatreq.h reference.
 * 
 * Revision 1.0.1.1  88/06/01  16:41:33  gdburns
 * Rewrote FL_LOAD request to spawn a separate process which receives
 * data messages in any order.
 * Added FL_DONE command to clean up after worker process completes.
 * 
 * Revision 1.0  87/12/19  15:43:11  trillium
 * Initial revision
 * 
 *
 *	Function:	- flatd server
 *			- provides symbolic access to memory
 *			- ``flattens'' memory in a distributed system in
 *			  order to support broadcasting operations
 *			- accepts requests to load memory (and mark it
 *			  with a given tag), inquire the address associated
 *			  with a given tag (and optionally forget the tag)
 */


#include <net.h>
#include <events.h>
#include <terror.h>
#include <kreq.h>
#include <flatreq.h>
#include <kio.h>
#include <shared.h>


struct flat_request	*request;		/* request from client */
struct flat_reply	*reply;			/* reply to client */

struct nmsg		head_req,		/* request desc. from client */
			head_reply,		/* reply desc. to client */
			head_data;		/* data desc. from client */

struct flat_region	fltable[FLMAX];		/* flat address table */

int			suck_it_in();

extern int		errno;

extern struct kio_t _kio;

main()

{
	int		i;			/* favourite index variable */


	if (kinit(0)) {
		panic(errno);
	}

	if (rpattach(LOCAL)) {
		panic(errno);
	}
/*
 * Initialize table.
 */
	for (i = 0; i < FLMAX; ++i) {
		fltable[i].fl_addr = EMPTY;
	}

	request = (struct flat_request *) head_req.nh_data;
	reply = (struct flat_reply *) head_reply.nh_data;
/*
 * Initialize message descriptors.
 */
	head_req.nh_event = EVFLATD;
	head_req.nh_type = 0;
	head_req.nh_flags = 0;
	head_req.nh_msg = (char *) 0;
	head_req.nh_length = 0;

	head_reply.nh_flags = NREEL;
	head_reply.nh_type = 0;
	head_reply.nh_msg = (char *) 0;
	head_reply.nh_length = 0;

	head_data.nh_flags = 0;

	for (;;) {

		if (nrecv(&head_req)) {
			panic(errno);
		}

		head_reply.nh_node = request->fl_src_node;
		head_reply.nh_event = request->fl_src_event;

		if (request->fl_req == FL_LOAD) {
			do_load();
		}

		else if ((request->fl_req == FL_GET) ||
			(request->fl_req == FL_FORGET)) {

			do_get();
		}

		else if (request->fl_req == FL_DONE) {
			do_done();
		}

		else {
			reply->fl_reply = EBADREQUEST;

			if (nsend(&head_reply)) {
				panic(errno);
			}
		}
	}
}


/*
 *	do_load
 *
 *	Function:	- load a region of memory and mark it with a tag
 *			- locally mallocs requested amount of memory
 */

do_load()

{
	int		ld_length,		/* length of load operation */
			ma_length,		/* length to malloc */
			i,			/* favourite index variable */
			pid;			/* new process's ID */
	char		*memblock;		/* memory allocated for stack
						   and static data */


	ld_length = request->fl_ld_length;
	ma_length = request->fl_ma_length;

	for (i = 0; (i < FLMAX) && (fltable[i].fl_addr != EMPTY); ++i);

	if (i >= FLMAX) {
		reply->fl_reply = ENOFLDESCRIPTORS;
	}

	else if ((fltable[i].fl_addr = (char *) rmalloc(LOCAL, ma_length, 0, 0))
			== (char *) 0) {

		reply->fl_reply = errno;
		fltable[i].fl_addr = EMPTY;
	}

	else {
		reply->fl_reply = 0;
		fltable[i].fl_tag = request->fl_tag;

		if (ld_length) {

			/* Create a parallel process to load the data */
			memblock = (char *) kpar_Fe(suck_it_in,
						FLSTACK,
						&pid,
						4,
						fltable[i].fl_addr,
						ld_length,
						request->fl_src_event,
						request->fl_src_node);

			if (memblock == (char *) 0) {
				reply->fl_reply = errno;
			}
		}
	}

	if (nsend(&head_reply)) {
		panic(errno);
	}
}


/*
 *	do_done
 *
 *	Function:	- wait for a worker process to complete
 *			  and then free its memory
 */

do_done()

{
	KS_STATE	done;		/* kernel state of worker process */


	while (kstate(request->fl_pid, &done) == 0) {
		tin(ldtimer() + 1);
	}

	if (errno != ENOTPROCESS) {
		panic(errno);
	}

	if (rfree(LOCAL, (char *) request->fl_dseg)) {
		panic(errno);
	}
}


/*
 *	do_get
 *
 *	Function:	- find and reply the address associated with a
 *			  given tag
 *			- FL_FORGET also cleans out the fltable entry
 *			  associated with the tag
 */

do_get()

{
	int		i;			/* favourite index variable */


	for (i = 0; (i < FLMAX)
			&& ((fltable[i].fl_tag != request->fl_tag)
			|| (fltable[i].fl_addr == EMPTY)); ++i);

	if (i >= FLMAX) {
		reply->fl_reply = EBADTAG;
	}

	else {
		reply->fl_reply = 0;
		reply->fl_address = fltable[i].fl_addr;
	}

	if (nsend(&head_reply)) {
		panic(errno);
	}

	if (request->fl_req == FL_FORGET) {
		fltable[i].fl_addr = EMPTY;
	}
}


/*
 *	suck_it_in
 *
 *	Function:	- parallel process to load data into allocated space
 *			- attempts to die after loading data using a method
 *			  that could potentially be fatal - someday we
 *			  need to allow Trillium processes a way to clean
 *		 	  up after themselves that is safe.
 *
 *	Accepts:	- length
 *			- address of message
 *			- receiving event
 */

suck_it_in(addr, length, event, nodeid)

char			*addr;		/* address of data */
int			length;		/* length of data */
int			event;		/* client event */
int			nodeid;		/* client node ID */

{
	struct flat_request	*donereq;	/* send done notice to
						   flatd administrator */
	struct flat_reply	*donereply;	/* also tell client */
	struct nmsg		donemsg;	/* message header for donereq */
	int			numpkts;	/* # of packets to receive */
	char			buf[MAXNMSGLEN];


	if (kinit(0)) {
		panic(errno);
	}

	donereq = (struct flat_request *) donemsg.nh_data;
	donereply = (struct flat_reply *) donemsg.nh_data;

	head_data.nh_flags = 0;
	head_data.nh_msg = buf;
	head_data.nh_length = MAXNMSGLEN;
	head_data.nh_event = event;

	numpkts = (length + (MAXNMSGLEN - 1)) / MAXNMSGLEN;

	while (numpkts--) {
		head_data.nh_type = 0;
		head_data.nh_length = MAXNMSGLEN;

		if (nrecv(&head_data)) {
			panic(errno);
		}

		bcopy(buf, addr + head_data.nh_type, head_data.nh_length);
	}
/*
 * Inform flatd client that the data has been fully loaded.
 */
	donereply->fl_reply = 0;
	donemsg.nh_node = nodeid;
	donemsg.nh_event = event;
	donemsg.nh_type = 0;
	donemsg.nh_flags = NREEL;
	donemsg.nh_length = 0;
	donemsg.nh_msg = (char *) 0;

	if (nsend(&donemsg)) {
		panic(errno);
	}
/*
 * Inform flatd administrator of impending death so it can clean up.
 */
	donereq->fl_req = FL_DONE;
	donereq->fl_pid = getpid();
	donereq->fl_dseg = __slink();

	donemsg.nh_node = LOCAL;
	donemsg.nh_event = EVFLATD;

	if (nsend(&donemsg)) {
		panic(errno);
	}

	kexit(0);
}
