/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	kenyad.c,v $
 * Revision 2.2  92/04/30  10:31:19  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  03:02:19  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/03/19  23:54:49  gdburns
 * Be more careful with NOBUF.
 * 
 * Revision 2.0.1.2  90/11/29  19:45:40  gdburns
 * Shared.h is dangerous and should be listed last.
 * 
 * Revision 2.0.1.1  90/07/13  19:26:58  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:11  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:02:06  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0.1.2  88/06/01  17:23:36  gdburns
 * Removed the redundant (with nreq) stub function.
 * 
 * Revision 1.0.1.1  88/06/01  16:18:40  gdburns
 * Removed redundant (with nreq) set_order function.
 * 
 * Revision 1.0  87/12/19  15:43:22  trillium
 * Initial revision
 * 
 *
 *	Function:	- itb process control server
 *			- higher level control of processes,
 *			  particularly application/user processes
 */

#include "kenya.h"
#include <kio.h>
#include <events.h>
#include <typical.h>
#include <terror.h>
#include <shared.h>

/*
 * external variables
 */
extern int		errno;			/* global error variable */

/*
 * external functions
 */
extern void		do_pcreate(),		/* declare kenya requests */
			do_pdoom(),
			do_pstate(),
			do_pspawn(),
			do_pdetach(),
			do_pattach();

/*
 * local variables
 */
static char *_copyright = "Copyright 1987 Cornell University $Revision: 2.2 $";

static void		(*do_req[PREQS])() = {	/* request routines */
				do_pcreate,
				do_pdoom,
				do_pstate,
				do_pspawn,
				do_pdetach,
				do_pattach
			};
/*
 * global variables
 */
struct pprocess		ptable[PMAX];		/* the process table */
struct preq		request;		/* request from client */
struct preply		reply;			/* reply to client */


main()

{
	struct pprocess		*p;		/* favourite pointer name */
	int			i,		/* favourite index name */
				pcode;		/* normalized request code */


	if (kinit(0)) {
		panic(errno);
	}
/*
 * Initialize process table.
 */
	for (p = ptable, i = 0; i < PMAX; ++i, ++p) {
		p->p_loadpt = (struct exec *) 0;
		p->p_static = (char *) 0;
		p->p_stack = (char *) 0;
		p->p_flags = 0;
		p->p_argc = 0;
		p->p_argv = (char **) 0;
		p->p_pid = NOTPROCESS;
	}
/*
 * main request processing loop
 */
	for (;;) {

		if (recv_req()) {
			panic(errno);
		}

		pcode = request.pq_req - PNORMAL;
			
		if ((pcode >= 0) && (pcode < PREQS)) {
			(*do_req[pcode])();
		}

		else {
			reply.pr_reply = EBADREQUEST;

			if (send_reply((char *) &reply, sizeof(reply))) {
				panic(errno);
			}
		}
	}
}


/*
 *	recv_req
 *
 *	Function:	- receive a kenya request from a client
 *
 *	Returns:	- 0 or ERROR
 */

int
recv_req()

{
	struct nmsg		nhead;		/* request descriptor */


	nhead.nh_event = EVKENYAD;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = sizeof(request);
	nhead.nh_msg = (char *) &request;

	return(nrecv(&nhead));
}

	
/*
 *	send_reply
 *
 *	Function:	- send a kenya reply to a client
 *
 *	Accepts:	- ptr to kenyad reply
 *			- length of kenyad reply
 *			- buffering flag
 *
 *	Returns:	- 0 or ERROR
 */

int
send_reply(preply, length, flags)

char				*preply;
int				length;
int				flags;

{
	struct nmsg		header;		/* reply descriptor */


	header.nh_node = request.pq_src_node;
	header.nh_event = request.pq_src_event;
	header.nh_type = 0;
	header.nh_flags = flags | NREEL;
	header.nh_length = length;
	header.nh_msg = preply;

	return(nsend(&header));
}


/*
 *	search_ptable
 *
 *	Function:	- find a process's pid in the table of process
 *			  descriptors
 *			- simple linear search is used
 *
 *	Accepts:	- process ID of client to be found
 *
 *	Returns:	- an index of a process descriptor or ERROR
 */

int
search_ptable(search_id)

int				search_id;	/* pid to be found */

{
	int			i;		/* favourite index variable */


	for (i = 0; (ptable[i].p_pid != search_id) && (i < PMAX); ++i);

	return((i < PMAX) ? i : ERROR);
}


/*
 *	match
 *
 *	Function:	- does a given process descriptor match the
 *			  request pattern?
 *			- abstracts the vagueries of process specification
 *			  options
 *
 *	Accepts:	- pointer to a process descriptor
 *			- matching flags
 *			- matching value (meaning varies with flags)
 *
 *	Returns:	- TRUE or FALSE or ERROR
 */

int
match(pmatch, mflags, mvalue)

struct pprocess		*pmatch;		/* does this process match? */
int			mflags,			/* matching flags */
			mvalue;			/* matching value */

{
	int		matched;		/* do we have a match? */
	KS_STATE	ks_match;		/* kernel state for pmatch */


	if (pmatch->p_pid == NOTPROCESS) {
		return(FALSE);
	}

	switch (mflags) {

	case SELECT_ALL:
		matched = TRUE;
		break;

	case SELECT_SYS:
		matched = (pmatch->p_flags & RTF_SYSGEN) ? TRUE : FALSE;
		break;

	case SELECT_APPL:
		matched = (pmatch->p_flags & RTF_SYSGEN) ? FALSE : TRUE;
		break;

	case SELECT_PID:
		matched = (pmatch->p_pid == mvalue) ? TRUE : FALSE;
		break;

	case SELECT_INDEX:
		--mvalue;
		matched = ((ptable + mvalue) == pmatch) ? TRUE : FALSE;
		break;

	case SELECT_BLKD:
/*
 * Get state info from the kernel.
 */
		if (kstate(pmatch->p_pid, &ks_match)) {
			matched = ERROR;
		}

		else {
			matched = (((ks_match.ks_state == SSEND_BLKD) ||
					(ks_match.ks_state == SRECV_BLKD)) &&
					(ks_match.ks_event == mvalue)) ?
					TRUE : FALSE;
		}

		break;

	case SELECT_LOADPT:
		matched = (pmatch->p_loadpt == (struct exec *) (int *) mvalue)
				? TRUE : FALSE;
		break;

	default:
		break;
	}

	return(matched);
}
