/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	pcreate.c,v $
 * Revision 2.2  92/04/30  10:31:27  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/26  17:09:11  gdburns
 * Add hooks for i860 executables.
 * 
 * Revision 2.1.1.1  92/01/02  23:14:33  gdburns
 * Use tout package instead of bootd for load module relocation.
 * 
 * Revision 2.1  91/03/20  03:02:21  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/03/19  23:55:07  gdburns
 * Be more careful with NOBUF.
 * 
 * Revision 2.0.1.2  90/11/29  19:46:34  gdburns
 * Shared.h is dangerous and should be listed last.
 * 
 * Revision 2.0.1.1  90/07/13  19:27:01  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:12  trillium
 * Homo erectus
 *
 *	Function:	- implements the PCREATE request
 *			- finds pointer to static data
 *			- zeros out the bss segment
 *			- calls create()
 *			- passes off i860 load module to i860d
 *
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#include <events.h>
#include <i8req.h>
#include <mng.h>
#include <net.h>
#include <typical.h>
#include <terror.h>
#include <pcb.h>
#include "kenya.h"
#include <kio.h>
#include <shared.h>

/*
 * external variables
 */
extern int		errno;
extern struct pprocess	ptable[];		/* kenya process table */
extern struct preq	request;		/* kenya request */
extern struct preply	reply;			/* kenya reply */
extern struct pcb_t	*_pcb;

/*
 * external functions
 */
extern void		patch_argv();		/* relocate argument pointers */
extern void		toutpatch();		/* patch load module */
extern void		*toutrelptr();		/* ptr to relocation table */
extern void		*touttextptr();		/* ptr to text segment */
extern int		rw();			/* reverse word */

/*
 * local functions
 */
static int		ri8create();

void
do_pcreate()

{
	int		*bss;		/* bss pointer */
	int		c;		/* I'm trying this out as a counter. */
	int		i;		/* favourite index */
	struct exec	*loadpt;	/* address of load module */
	char		*static_link;	/* static data ptr */
	struct mng	*pmngload;	/* mng file header ptr */

	reply.pr_reply = 0;
/*
 * Find the address of the load module.
 */
	loadpt = (struct exec *) rforget(LOCAL, request.pq_loadpt);
	pmngload = (struct mng *) loadpt;

	if (loadpt == (struct exec *) 0) {
		reply.pr_reply = errno;
	}
/*
 * Check for a valid load module.
 */
	else if ((loadpt->a_magic != TO_MAGIC_TN_X) &&
		(loadpt->a_magic != TO_MAGIC_T4_X) &&
		(loadpt->a_magic != TO_MAGIC_T8_X) &&
		! mng_magic(pmngload)) {

		reply.pr_reply = EMAGIC;
	}
/*
 * Convert argv from tag to pointer.
 */
	if (request.pq_argc) {

		if ((request.pq_argv = (int) rforget(LOCAL, request.pq_argv))
			== 0) {

			reply.pr_reply = errno;
		}
	} else {
		request.pq_argv = 0;
	}

	if (reply.pr_reply) {

		if (loadpt) (void) rfree(LOCAL, (char *) loadpt);
		if (request.pq_argv)
			(void) rfree(LOCAL, (char *) request.pq_argv);

		if (send_reply((char *) &reply, sizeof(reply), 0)) {
			panic(errno);
		}

		return;
	}
/*
 * If we have an MNG file we take a hard turn to the left and do something
 * totally different.
 */
	if (mng_magic(pmngload)) {
/*
 * Check for the existance of an i860 process.  We only allow
 * one at a time.
 */
		for (i = 0; (i < PMAX) && ((ptable[i].p_pid == NOTPROCESS) ||
			!(ptable[i].p_flags & RTF_I860)); ++i);

		if (i < PMAX) {
			reply.pr_reply = EI860PROC;

			if (loadpt) (void) rfree(LOCAL, (char *) loadpt);
			if (request.pq_argv)
				(void) rfree(LOCAL, (char *) request.pq_argv);
		} else {
/*
 * Create the process by making request to i860d.
 */
			if (ri8create(pmngload) < 0) {
				reply.pr_reply = errno;
			}
		}
	} else {
/*
 * Calculate the address of external data.
 */
		static_link = ((char *) loadpt)
				+ sizeof(struct exec)
				+ loadpt->a_text
				+ loadpt->a_syms
				+ loadpt->a_drsize;
/*
 * Zero out the bss segment.
 */
		bss = (int *) (static_link + loadpt->a_data + loadpt->a_bss);
		bss--;
		c = loadpt->a_bss / sizeof(int);

		for (; c; --c, --bss) {
			*bss = 0;
		}

		if (create(loadpt, static_link) < 0) {
			reply.pr_reply = errno;
		}
	}
/*
 * Reply to client.
 */
	if (send_reply((char *) &reply, sizeof(reply), 0)) {
		panic(errno);
	}
}

/*
 *	create
 *
 *	Function:	- creates a new process for a given load module
 *			  already residing in memory
 *			- fills in reply.pr_pid with new process ID
 *			- fills in reply.pr_index with new process index
 *			- argv string is assumed to be contiguous and
 *			  argv[argc] must be a null pointer
 *
 *			  --------------------
 *			  |    0 pointer     |  argv + 0
 *			  --------------------
 *			  |    1 pointer     |  argv + 1
 *			  --------------------
 *				   :	
 *			  --------------------
 *			  | argc - 1 pointer |  argv + (argc - 1)
 *			  --------------------
 *			  |        0         |
 *			  --------------------
 *			  :     0 string     :  argv[0]
 *			  --------------------
 *			  :     1 string     :  argv[1]
 *			  --------------------
 *				   :	
 *			  --------------------
 *			  : argc - 1 string  :  argv[argc - 1]
 *			  --------------------
 *
 *	Accepts:	- ptr to load module
 *			- ptr to private static data
 *
 *	Returns:	- 0 or ERROR
 */

int
create(loadpt, static_link)

struct exec		*loadpt;	/* load module ptr */
char			*static_link;	/* static data ptr */

{
	struct pprocess	*pcreated;	/* newborn process pointer */
	int		icreated;	/* index of newborn process */
	struct kproc_t	newproc;	/* new process info for kcreate() */

/*
 * Find a slot in the process table.
 */
	icreated = search_ptable((int) NOTPROCESS);

	if (icreated == ERROR) {
		errno = ENOPDESCRIPTORS;
		return(ERROR);
	}

	pcreated = ptable + icreated;
	pcreated->p_static = static_link;
	pcreated->p_argc = request.pq_argc;
	pcreated->p_argv = (char **) request.pq_argv;
/*
 * Set stack size.
 */
	newproc.kp_stksize = pcreated->p_stksize = (request.pq_stksize) ?
				request.pq_stksize : PSTKSIZE;
/*
 * Perform all necessary address relocation.
 */
	toutpatch(loadpt, toutrelptr(loadpt), static_link,
		touttextptr(loadpt), static_link);
	patch_argv(pcreated->p_argc, pcreated->p_argv);

	newproc.kp_header = pcreated->p_loadpt = loadpt;
	newproc.kp_slink = static_link;
	newproc.kp_rtf = pcreated->p_flags = request.pq_flags | RTF_KENYA;
	_pcb->pc_stdin = request.pq_stdin;
	_pcb->pc_stdout = request.pq_stdout;
	_pcb->pc_stderr = request.pq_stderr;
/*
 * Create the process, notice that the top of stack argument is the process ID.
 */
	if ((pcreated->p_pid = kcreate(&newproc, pcreated->p_argv)) == ERROR) {
		pcreated->p_pid = NOTPROCESS;
		return(ERROR);
	}

	pcreated->p_stack = newproc.kp_stack;
	pcreated->p_nodeid = request.pq_src_node;
	pcreated->p_event = request.pq_src_event;

	reply.pr_pid = pcreated->p_pid;
	reply.pr_index = icreated + 1;

	return(0);
}

/*
 *	ri8create
 *
 *	Function:	- passes the responsibility of creating
 *			  the process to i860d
 *
 *	Accepts:	- ptr to load module
 */

static int
ri8create(pmngload)

struct mng		*pmngload;

{
	struct pprocess	*pcreated;	/* newborn process pointer */
	int		icreated;	/* index of newborn process */
	struct nmsg	nhead;		/* message descriptor to/from i860d */
	struct i8req	i8request;	/* i860d request */
	struct i8reply	i8reply;	/* i860d reply */
/*
 * Find a slot in the process table.
 */
	icreated = search_ptable((int) NOTPROCESS);

	if (icreated < 0) {
		errno = ENOPDESCRIPTORS;
		return(ERROR);
	}

	i8request.i8q_req = I8QCREATE;
	i8request.i8q_src_node = LOCAL;
	i8request.i8q_src_event = -getpid();
	i8request.i8q_pload = pmngload;
	i8request.i8q_pargv = (char **) request.pq_argv;
	i8request.i8q_argc = request.pq_argc;
	i8request.i8q_rtf = request.pq_flags | RTF_KENYA | RTF_I860;
	i8request.i8q_stksize = request.pq_stksize;
	i8request.i8q_stdin = request.pq_stdin;
	i8request.i8q_stdout = request.pq_stdout;
	i8request.i8q_stderr = request.pq_stderr;

	nhead.nh_node = LOCAL;
	nhead.nh_event = EVI860D;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = sizeof(struct i8req);
	nhead.nh_msg = (char *) &i8request;

	if (nsend(&nhead) < 0) {
		panic(errno);
	}

	nhead.nh_event = -getpid();
	nhead.nh_length = sizeof(struct i8reply);
	nhead.nh_msg = (char *) &i8reply;

	if (nrecv(&nhead) < 0) {
		panic(errno);
	}

	if (i8reply.i8r_reply != 0) {
		errno = i8reply.i8r_reply;
		return(ERROR);
	}

	pcreated = ptable + icreated;
	pcreated->p_pid = i8reply.i8r_pid;
	pcreated->p_argc = request.pq_argc;
	pcreated->p_stack = (char *) i8reply.i8r_pstack;
	pcreated->p_static = (char *) i8reply.i8r_pstatic;
	pcreated->p_nodeid = request.pq_src_node;
	pcreated->p_event = request.pq_src_event;
	pcreated->p_loadpt = (struct exec *) i8reply.i8r_pload;
	pcreated->p_argv = i8reply.i8r_pargv;
	pcreated->p_flags = request.pq_flags | RTF_KENYA | RTF_I860;

	reply.pr_pid = pcreated->p_pid;
	reply.pr_index = icreated + 1;

	return(0);
}
