/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	pdetach.c,v $
 * Revision 2.2  92/04/30  10:31:30  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/03/05  20:00:56  gdburns
 * Kludge termination event so as not to clash with rloadgo().
 * 
 * Revision 2.1.1.1  91/06/05  13:27:38  gdburns
 * Use unmodified parent event (-pid).
 * 
 * Revision 2.1  91/03/20  03:02:22  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/03/19  23:55:27  gdburns
 * Be more careful with NOBUF.
 * 
 * Revision 2.0.1.2  90/11/29  19:46:39  gdburns
 * Shared.h is dangerous and should be listed last.
 * 
 * Revision 2.0.1.1  90/07/13  19:26:51  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:09  trillium
 * Homo erectus
 *	Function:	- implement the PDETACH request
 *			- only allow detach by pid
 *			- wait for the process to die
 *			- until we add a way to identify the client
 *			  we have to reply first and THEN carry out the request
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#include <typical.h>
#include <terror.h>
#include "kenya.h"
#include <kio.h>
#include <shared.h>

/*
 * external variables
 */
extern struct pprocess	ptable[];		/* kenya process table */
extern struct preq	request;		/* kenya request */
extern struct preply	reply;			/* kenya reply */
extern int		errno;

/*
 * external functions
 */
extern int		send_reply();
extern int		search_ptable();

/*
 * local functions
 */
static int		wait_die();
static int		free_pprocess();


void
do_pdetach()

{
	int		idetach;		/* descriptor index of target */
	struct pprocess	*pdetach;		/* descriptor ptr of target */

/*
 * Find the process descriptor for the given identifier.
 */
	if ((idetach = search_ptable(request.pq_pid)) == ERROR) {
		reply.pr_reply = ENOTPROCESS;
	}

	else {
		reply.pr_reply = 0;
	}
/*
 * Send reply of detach operation.
 */
	if (send_reply((char *) &reply, sizeof(reply), NOBUF)) {
		panic(errno);
	}

	if (! reply.pr_reply) {

		pdetach = ptable + idetach;

		if (! (pdetach->p_flags & RTF_NOWAIT)) {
/*
 * Reply to the parent process.
 */
			request.pq_src_node = pdetach->p_nodeid;
			request.pq_src_event = pdetach->p_event & 0x40000000;
			reply.pr_reply = request.pq_status;
			reply.pr_nodeid = getnodeid();
			reply.pr_pid = pdetach->p_pid;

			if (send_reply((char *) &reply, sizeof(reply), 0)) {
				panic(errno);
			}
		}
/*
 * Wait for that sucker to die.  Pdetach is a dangerous request which should
 * only be used by processes to detach themselves immediately before detaching
 * from the kernel or by guru programmers who want to clean up dead kenya
 * descriptors.
 */
		if (wait_die(request.pq_pid)) {
			panic(errno);
		}
/*
 * Detach target process.
 */
		pdetach->p_pid = NOTPROCESS;

		if (free_pprocess(pdetach)) {
/*
 * panic(errno);
 * Let's not get too excited if we can't free some space.
 */
		}
	}
}


/*
 *	wait_die
 *
 *	Function:	- wait for a process to die
 *			- pdetach assumes that the process is proceding
 *			  directly to death
 *
 *	Accepts:	- pid
 */

static int
wait_die(pid)

int			pid;		/* target process ID */

{
	KS_STATE	done;		/* kernel state of target process */


	while (kstate(pid, &done) == 0) {
		tin(ldtimer() + 1000);
	}
	
	if (errno == ENOTPROCESS) {
		return(0);
	}

	else {
		return(ERROR);
	}
}


/*
 *	free_pprocess
 *
 *	Function:	- free a process descriptor
 *			- marks process descriptor as empty
 *			- frees memory allocated to text, stack and argv
 *
 *	Accepts:	- ptr to a process descriptor to be freed
 *
 *	Returns:	- 0 or ERROR
 */

static int
free_pprocess(pfree)

struct pprocess		*pfree;		/* free this descriptor */

{
	int		found = FALSE,	/* found load module flag */
			i;		/* favourite index variable */


/*
 * Check to see if nobody else is using the text segment.
 */
	for (i = 0; (i < PMAX) && !found; ++i) {

	    if ((ptable[i].p_pid != NOTPROCESS) && ((ptable + i) != pfree)
			&& (ptable[i].p_loadpt == pfree->p_loadpt)) {

		found = TRUE;
	    }
	}

	if (!found && (rfree(LOCAL, (char *) pfree->p_loadpt))) {
	    return(ERROR);
	}

	(void) rfree(LOCAL, pfree->p_stack);
	(void) rfree(LOCAL, pfree->p_static);

	if (pfree->p_argc) {
	    found = FALSE;
/*
 * Check to see if nobody else is using the argv structure.
 */
	    for (i = 0; (i < PMAX) && !found; ++i) {

		if ((ptable[i].p_pid != NOTPROCESS) && ((ptable + i) != pfree)
				&& (ptable[i].p_argv == pfree->p_argv)) {

		    found = TRUE;
		}
	    }

	    if (!found && (rfree(LOCAL, (char *) pfree->p_argv))) {
		return(ERROR);
	    }
	}

	return(0);
}


/*
 *	do_pattach
 *
 *	Function:	- attach an existing process to kenya
 *			- process will become part of all kenya services
 */

void
do_pattach()

{
	struct exec	*tout;
	struct pprocess	*pattached;	/* new process pointer */
	int		iattached;	/* index of new process */


	tout = (struct exec *) request.pq_loadpt;
/*
 * Is the process already attached?
 */
	iattached = search_ptable(request.pq_pid);

	if (iattached != ERROR) {
		reply.pr_reply = 0;
	}
/*
 * Find a slot in the process table.
 */
	else if ((iattached = search_ptable((int) NOTPROCESS)) == ERROR) {
		reply.pr_reply = ENOPDESCRIPTORS;
	}
/*
 * Check for valid load module header.
 */
	else if ((tout->a_magic != TO_MAGIC_TN_X) &&
		(tout->a_magic != TO_MAGIC_T4_X) &&
		(tout->a_magic != TO_MAGIC_T8_X)) {

		reply.pr_reply = EMAGIC;
	}

	else {
/*
 * Fill in new kenya descriptor.
 */
		pattached = ptable + iattached;
		pattached->p_pid = request.pq_pid;
		pattached->p_stksize = request.pq_stksize;
		pattached->p_stack = (char *) request.pq_stack;
		pattached->p_static = (char *) request.pq_datapt;
		pattached->p_loadpt = tout;
		pattached->p_argc = request.pq_argc;
		pattached->p_argv = (char **) request.pq_argv;
		pattached->p_flags = (request.pq_flags | RTF_KENYA |
					RTF_NOWAIT);
		reply.pr_reply = 0;
	}
/*
 * Reply to client.
 */
	if (send_reply((char *) &reply, sizeof(reply), NOBUF)) {
		panic(errno);
	}
}
