/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	pdoom.c,v $
 * Revision 2.2  92/04/30  10:31:31  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  03:02:24  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/03/19  23:55:30  gdburns
 * Be more careful with NOBUF.
 * 
 * Revision 2.0.1.2  90/11/29  19:46:42  gdburns
 * Shared.h is dangerous and should be listed last.
 * 
 * Revision 2.0.1.1  90/07/13  19:27:02  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:13  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:02:21  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0.1.1  88/02/29  17:16:52  gdburns
 * Doom would not work with processes not attached to kenyad.
 * 
 * Revision 1.0  87/12/19  15:43:27  trillium
 * Initial revision
 * 
 *
 *	Function:	- implement the PDOOM request
 *			- check for a match on each used process descriptor
 *			- dooms each matched process using kdoom()
 *			- if no process is matched and a pid is supplied,
 *			  attempts to doom process with the kernel only
 */

#include <typical.h>
#include <terror.h>
#include <preq.h>
#include "kenya.h"
#include <ksignal.h>
#include <shared.h>

/*
 * external variables
 */
extern struct pprocess	ptable[];		/* kenya process table */
extern struct preq	request;		/* kenya request */
extern struct preply	reply;			/* kenya reply */
extern struct pstate	reply_state;		/* kenya state reply */
extern int		errno;

/*
 * external functions
 */
extern int		free_pprocess();
extern int		send_reply();


void
do_pdoom()

{
	int		i,			/* favourite index variable */
			matched,		/* result of match function */
			found_match = FALSE;	/* did anything match? */


	reply.pr_reply = 0;

	for (i = 0; (i < PMAX) && !reply.pr_reply; ++i) {

	    matched = match(ptable + i, request.pq_sel_method,
				request.pq_sel_value);

	    if (matched) {
		found_match = TRUE;

		if (matched == ERROR) {
		    reply.pr_reply = errno;
		}
/*
 * Doom process in kernel.
 */
		else if (kdoom(ptable[i].p_pid, request.pq_signal)) {
		    reply.pr_reply = errno;
		}
	    }
	}

	if (!found_match && (request.pq_sel_method & SELECT_PID)) {
/*
 * No matching process was found.  Try to doom a process only
 * if a pid is supplied (because the pid is the only key the kernel will take).
 */
		if (kdoom(request.pq_sel_value, request.pq_signal)) {
			reply.pr_reply = errno;
		}
	}
/*
 * Send reply of doom operation back to client.
 */
	if (send_reply((char *) &reply, sizeof(reply), 0)) {
		panic(errno);
	}
}
