/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	pspawn.c,v $
 * Revision 2.2  92/04/30  10:31:33  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/01/02  23:14:56  gdburns
 * Use tout package instead of bootd for load module relocation.
 * 
 * Revision 2.1  91/03/20  03:02:25  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/03/19  23:55:34  gdburns
 * Be more careful with NOBUF.
 * 
 * Revision 2.0.1.2  90/11/29  19:46:44  gdburns
 * Shared.h is dangerous and should be listed last.
 * 
 * Revision 2.0.1.1  90/07/13  19:27:03  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:14  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:02:24  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:43:28  trillium
 * Initial revision
 * 
 *
 *	Function:	- implement the PSPAWN request
 *			- check for a match on each used process descriptor
 *			- spawns each matched process
 *			- if local, copy data segment and create process locally
 *			- if remote, load the program and external data to
 *			  the requested node, then use rcreate
 *			- I am not too proud of this stretch of code.  (GDB)
 */

#include <typical.h>
#include <terror.h>
#include <preq.h>
#include "kenya.h"
#include <kio.h>
#include <shared.h>

/*
 * external variables
 */
extern struct pprocess	ptable[];		/* kenya process table */
extern struct preq	request;		/* kenya request */
extern struct preply	reply;			/* kenya reply */
extern int		errno;

/*
 * external functions
 */
extern char		*strncpy();
extern int		match();
extern int		send_reply();
extern int		create();
extern int		strlen();
extern void		patch_argv();
extern void		toutpatch();
extern void		*toutrelptr();
extern void		*touttextptr();

/*
 * local functions
 */
static int		lspawn();
static int		spawn();


void
do_pspawn()

{
	int		i;			/* favourite index variable */
	int		matched;		/* result of match function */
	int		argn;			/* size of argument list */
	int		found_match;		/* did we match anything? */
	char		*p;			/* favourite ptr variable */
	char		**argv;			/* runtime arguments */
	struct pprocess	*pspawn;		/* target kenyad descriptor */


	found_match = FALSE;
	reply.pr_reply = 0;
/*
 * Establish an argv pointer.
 */
	if (request.pq_argc) {
/*
 * Arguments are supplied by the caller.
 */
	    if (! (argv = (char **) rforget(LOCAL, request.pq_argv))) {
		reply.pr_reply = errno;
	    }
	}
/*
 * Spawn every selected process.
 */
	for (i = 0; (i < PMAX) && !reply.pr_reply; ++i) {

	    matched = match(ptable + i, request.pq_sel_method,
				request.pq_sel_value);

	    if (matched == ERROR) {
		reply.pr_reply == errno;
	    }

	    else if (matched) {
		pspawn = ptable + i;

		if (request.pq_argc) {
/*
 * Fill in the correct argv[0].
 */
		    patch_argv(request.pq_argc, argv);
		    strncpy(argv[0], pspawn->p_argv[0], strlen(argv[0])); 
		}

		else {
		    argv = pspawn->p_argv;
		    request.pq_argc = pspawn->p_argc;
		}
/*
 * Find the size of the arg structure.
 */
		p = argv[request.pq_argc - 1];
		argn = (p - (char *) argv) + strlen(p) + 1;
/*
 * Is it a local spawn?
 */
		if ((request.pq_nodeid == LOCAL) ||
			(request.pq_nodeid == getnodeid())) {
/*
 * Make a copy of the argment structure.
 */
		    request.pq_argv = (int) rmalloc(LOCAL, argn, 0, 0);

		    if (! request.pq_argv) {
			reply.pr_reply == errno;
		    }

		    else {
			bcopy((char *) argv, (char *) request.pq_argv, argn);

			if (lspawn(pspawn)) {
			    reply.pr_reply = errno;
			}
		    }
		}

		else {
/*
 * It's a remote spawn.
 */
		    if (spawn(request.pq_nodeid, pspawn, argv, argn)) {
			reply.pr_reply = errno;
		    }
		}

		found_match = TRUE;
	    }
	}

	if ((argv != pspawn->p_argv) && rfree(LOCAL, (char *) argv)) {
		reply.pr_reply = errno;
	}

	if (! found_match) {
		reply.pr_reply = ENOTPROCESS;
	}
/*
 * Send reply of spawn operation.
 */
	if (send_reply((char *) &reply, sizeof(reply), 0)) {
	    panic(errno);
	}
}

/*
 *	lspawn
 *
 *	Function:	- spawn a process locally
 *			- runs new process from the same text segment
 *			- copies static data segment
 *			- calls create()
 *
 *	Accepts:	- ptr to kenya process descriptor to be spawned
 *
 *	Returns:	- 0 or ERROR
 */

static int
lspawn(pspawn)

struct pprocess		*pspawn;

{
	int		static_size;		/* size of static data seg */
	char		*new_static;		/* new static data ptr */
	struct exec	*loadpt;		/* load module ptr */


	loadpt = pspawn->p_loadpt;
/* 
 * Calculate static data segment size.
 */
	static_size = loadpt->a_data + loadpt->a_bss;
/*
 * Malloc space for the data segment.
 */
	if (! (new_static = (char *) rmalloc(LOCAL, static_size, 0, 0))) {
		return(ERROR);
	}
/*
 * Copy the data segment.
 */
	bcopy(pspawn->p_static, new_static, static_size);

	request.pq_flags = pspawn->p_flags | RTF_NOWAIT;
	request.pq_stksize = pspawn->p_stksize;
/*
 * Patch the new data segment.
 */
	toutpatch(loadpt, toutrelptr(loadpt), new_static,
		touttextptr(loadpt), new_static);

	return(create(loadpt, new_static));
}


/*
 *	spawn
 *
 *	Function:	- spawn a process to a remote node
 *			- does not handle local spawns
 *			- remotely loads the load module to the
 *			  destination node
 *			- call rcreate with kenyad's current request
 *			  parameters
 *
 *	Accepts:	- destination node ID
 *			- ptr to process descriptor to be spawned
 *			- ptr to argument list
 *			- size of argument list
 *
 *	Returns:	- 0 or ERROR
 */

static int
spawn(nodeid, pspawn, argv, argn)

int			nodeid;		/* spawn to this node */
struct pprocess		*pspawn;	/* spawn this process */
char			**argv;
int			argn;

{
	struct exec	*loadpt;	/* local load module ptr */
	int		prgm_size,	/* size of program */
			mypid;		/* kenya's pid */


	mypid = getpid();
	loadpt = pspawn->p_loadpt;
/*
 * Calculate sizeof program (load module).
 */
	prgm_size = sizeof(struct exec) +
			loadpt->a_text +
			loadpt->a_data +
			loadpt->a_bss +
			loadpt->a_drsize +
			loadpt->a_syms;
/*
 ***BUG***
 * Copy the right data segment, not necessarily the one attached
 * to the load module.
 */
	if (rflat(nodeid, (char *) loadpt, prgm_size, prgm_size, mypid)) {
		return(ERROR);
	}
/*
 * Load arguments to the remote node.
 */
	request.pq_argv = mypid;

	if (rflat(nodeid, (char *) argv, argn, argn, mypid)) {
/*
 * Clean up load module on remote node.
 */
	    (void) rfree(nodeid, (char *) rforget(nodeid, mypid));
	    return(ERROR);
	}

	reply.pr_index = rcreate(nodeid, mypid, pspawn->p_flags | RTF_NOWAIT,
				pspawn->p_stksize, &reply.pr_pid,
				request.pq_argc, request.pq_argv);

	return((reply.pr_index == ERROR) ? ERROR: 0);
}
