/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	kernelio.c,v $
 * Revision 2.2  92/04/30  10:38:54  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/04/26  17:13:43  gdburns
 * Remove stray debug stmts.
 * 
 * Revision 2.1.1.2  92/04/08  20:43:36  gdburns
 * Add stub for kkillrewrite().
 * 
 * Revision 2.1.1.1  92/02/23  19:13:47  gdburns
 * Add #ifdef to c.h.
 * 
 * Revision 2.1  91/03/20  03:02:42  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/11/15  16:19:15  raja
 * Added siesta list handling to enable multiple hi-pri processes to
 * communicate with the kernel.
 * 
 * Revision 2.0.1.1  90/07/13  19:26:42  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:03  trillium
 * Homo erectus
 *
 *	Function:	- machine dependent portion of kernel
 *			- receives kernel requests and sends replys
 *			- ITB specific code
 */

#ifdef itb
#include <c.h>
#endif
#include <portable.h>
#include <kernel.h>
#include <kreq.h>
#include <machine/maconfig.h>

extern struct kreq	request;		/* kernel request */
extern struct kreply	reply;			/* kernel reply */

/*
 * external functions
 */
extern void		end_siesta();		/* wake up a process */

/*
 *	init_request
 *
 *	Function:	- initialize communication structures for
 *			  sending and receiving kernel requests
 *			- initialize the siesta list for processes
 *			  colliding on the kernel channel
 */

int
init_request()

{
	(*(int *) KERNEL_CHAN) = NOTPROCESS;
	(*(int *) FRONT_SIESTA) = NOTPROCESS;
	(*(int *) BACK_SIESTA) = NOTPROCESS;
}

/*
 *	_kinit
 *	_kbail
 *	kkillwrite
 *	kkillrewrite
 */

int _kinit() { return(0); }
void _kbail() { stopp(); }
int kkillwrite() { return(0); }
int kkillrewrite() { return(0); }

/*
 *	recv_req
 *
 *	Function:	- receive the next kernel request
 *			- if the kernel channel is free activate
 *			  a process from the siesta list
 */

int
recv_req()

{
	if ( (*(int *) KERNEL_CHAN) == NOTPROCESS) {
/*
 * The kernel channel is free, run the top process on the siesta list.
 * Based on an idea by Aarne Rantala at VTT.
 */
		end_siesta();
	}

	in(KERNEL_CHAN, (char *) &request, sizeof(request));
}

/*
 *	send_reply
 *
 *	Function:	- send reply back to the client
 *			- message contains result of kernel request
 *
 *	Accepts:	- client's kernel communication port
 */

int
send_reply(client_port)

int			client_port;		/* client's data channel */

{
	out(client_port, (char *) &reply, sizeof(reply));
}
