/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	malist.c,v $
 * Revision 2.2.1.1  93/05/24  15:42:50  gdburns
 * BYSIZE test should be <=.
 *
 * Revision 2.2  92/04/30  10:39:45  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/11/10  23:36:04  gdburns
 * Minor overhaul fixes several problems.
 * 
 * Revision 2.1  91/03/20  03:03:13  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:26:39  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:01  trillium
 * Homo erectus
 *
 *	Function:	- list management routines for mallocd
 */

#include <typical.h>
#include <mareq.h>

/*
 * external variables
 */
extern struct memblock	*empty;			/* top of empty desc. list */
extern int		nempty;			/* # empty descriptors */

/*
 *	delete
 *
 *	Function:	- delete a memory block from a list
 *
 *	Accepts:	- ptr to top of list
 *			- search key
 *			- search criteria
 *
 *	Returns:	- ptr to the deleted memory block
 */

struct memblock
*delete(top, key, search)

struct memblock		**top;
int			key;
int			search;

{
	int		found;			/* found the correct block? */
	struct memblock	*curr;			/* current pointer */
	struct memblock	*prev;			/* previous pointer */
	int		align;			/* equally divides start */
	char		*start;			/* aligned starting addr */

	found = FALSE;
	prev = curr = *top;

	while (curr && !found) {

		switch (search) {

			case BYPTR:
			found = (curr == (struct memblock *) key);
			break;

			case BYSIZE:
			align = ((int *) key)[1];
			start = (char *) (((unsigned int) curr->ma_start +
				(align - 1)) / align * align);
			found = (start + ((int *) key)[0]) <=
				(curr->ma_start + curr->ma_size);
			break;

			case BYSTART:
			found = (curr->ma_start == (char *) key);
			break;

			case BYPID:
			found = (curr->ma_pid == key);
			break;

			case BYLAST:
			found = (curr->ma_next == MANULL);
			break;
		}

		if (! found) {
			prev = curr;
			curr = curr->ma_next;
		}
	}

	if (found) {

		if (curr == *top) {
			*top = curr->ma_next;
		} else {
			prev->ma_next = curr->ma_next;
		}

		return(curr);
	} else {
		return(MANULL);
	}
}

/*
 *	insert
 *
 *	Function:	- insert a memory block into a list
 *			- performs optional block compaction
 *			- may call itself recursively once
 *
 *	Accepts:	- ptr to top of list
 *			- ptr to the new block
 *			- compaction flag
 */

void
insert(top, new, compact)

struct memblock		**top;
struct memblock		*new;
int			compact;

{
	struct memblock	*curr;			/* current pointer */
	struct memblock	*prev;			/* previous pointer */

	prev = curr = *top;

	while (curr && ! (curr->ma_start >= new->ma_start)) {
		prev = curr;
		curr = curr->ma_next;
	}

	new->ma_next = curr;

	if (curr == *top) {
		*top = new;
	} else {
		prev->ma_next = new;
	}
/*
 * Check for block compaction.
 */
	if (compact) {
/*
 * backwards compaction
 */
		if ((new != *top) && ((prev->ma_start + prev->ma_size) >=
			new->ma_start)) {

			new->ma_size += prev->ma_size;
			new->ma_start = prev->ma_start;
			prev = delete(top, (int) prev, BYPTR);
			prev->ma_start = (char *) 0;
			insert(&empty, prev, NOCOMPACT);
			++nempty;
		}
/*
 * forwards compaction
 */
		if (curr && ((new->ma_start + new->ma_size) >=
			curr->ma_start)) {

			new->ma_size += curr->ma_size;
			curr = delete(top, (int) curr, BYPTR);
			curr->ma_start = (char *) 0;
			insert(&empty, curr, NOCOMPACT);
			++nempty;
		}
	}
}
