/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	mallocd.c,v $
 * Revision 2.2  92/04/30  10:39:46  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  91/11/10  23:36:07  gdburns
 * Minor overhaul fixes several problems.
 * 
 * Revision 2.1.1.1  91/10/09  22:29:04  gdburns
 * Allow descriptor blocks to be reported by mstate.
 * 
 * Revision 2.1  91/03/20  03:03:15  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/09/13  11:20:22  gdburns
 * Removed machine dependencies into separate routines and header files.
 * 
 * Revision 2.0.1.1  90/07/13  19:26:37  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:00  trillium
 * Homo erectus
 * 
 *	Function:	- itb memory allocater
 *			- can allocate memory from multiple "regions"
 *			- support deallocation and simple compaction
 *			- prefix is ma
 *
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#include <net.h>
#include <events.h>
#include <kernel.h>
#include <priority.h>
#include <typical.h>
#include <terror.h>
#include <machine/maconfig.h>
#include <mareq.h>
#include <shared.h>

/*
 * external variables
 */
extern int		errno;			/* error number */

/*
 * external functions
 */
extern struct memblock	*getdesc();
extern struct memblock	*delete();
extern void		freedesc();
extern void		initdesc();
extern void		insert();
extern int		region_start();
extern int		region_size();

/*
 * global variables
 */
struct memblock		*memalloc[NREGIONS];	/* tops of allocated lists */
struct memblock		*memfree[NREGIONS];	/* tops of free lists */
struct memblock		*empty;			/* top of empty desc. list */
struct memblock		*tbldesc;		/* top of desc. table list */
struct memblock		memdesc[MAMAX];		/* memory block descriptors */
int			nempty;			/* # of unused descriptors */

/*
 * local functions
 */
static char		*maqalloc();
static int		maqfree();
static int		maqfreeall();
static void		maqstate();
static void		init_mallocd();

/*
 * local variables
 */
static struct mareq	*request;	/* malloc request */
static struct mareply	*reply;		/* malloc reply */

main()

{
	struct nmsg	nhead;		/* network message desc. */

	if (kinit(PRMALLOCD)) {
		panic(errno);
	}

	if (rpattach(LOCAL)) {
		panic(errno);
	}

	request = (struct mareq *) nhead.nh_data;
	reply = (struct mareply *) nhead.nh_data;
/*
 * Initialize memory block data structures.
 */
	init_mallocd();

	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;
/*
 * main request processing loop
 */
	for (;;) {
		nhead.nh_event = EVMALLOCD;

		if (nrecv(&nhead)) {
			panic(errno);
		}

		nhead.nh_node = request->maq_src_node;
		nhead.nh_event = request->maq_src_event;

		switch(request->maq_request) {

			case MAQALLOC:
			reply->mar_address = maqalloc();
			reply->mar_reply = (!reply->mar_address && errno) ?
				errno : 0;
			break;

			case MAQFREE:
			reply->mar_reply = (maqfree()) ? errno : 0;
			break;

			case MAQFREEALL:
			reply->mar_reply = (maqfreeall()) ? errno : 0;
			break;

			case MAQSTATE:
			maqstate();
			break;

			default:
			reply->mar_reply = EBADREQUEST;
			break;
		}
/*
 * Reply to the client for simple requests.
 */
		if (request->maq_request != MAQSTATE) {
			nhead.nh_type = 0;
			nhead.nh_length = 0;
			nhead.nh_flags = BYBUF;

			if (nsend(&nhead)) {
				panic(errno);
			}
		}
	}
}

/*
 *	init_mallocd
 *
 *	Function:	- initialize all data structures for mallocd
 */

static void
init_mallocd()

{
	int			i;		/* favourite index variable */

	nempty = 0;
/*
 * Initialize the descriptor table.
 */
	tbldesc = MANULL;
	initdesc(memdesc, 0);
/*
 * Initialize the allocated lists.
 * Initialize the free lists.
 * There are at least NREGIONS + 1 descriptors in the static descriptor table.
 * ie: NREGIONS + 1 < MAMAX
 */
	for (i = 0; i < NREGIONS; ++i) {
		memalloc[i] = MANULL;
		memfree[i] = delete(&empty, (int) empty, BYPTR);
		memfree[i]->ma_pid = NOTPROCESS;
		memfree[i]->ma_start = (char *) region_start(i + 1);
		memfree[i]->ma_size = region_size(i + 1);
		memfree[i]->ma_next = MANULL;
		--nempty;
	}
}

/*
 *	maqalloc
 *
 *	Function:	- allocates the requested amount of memory aligned to
 *			  the requested boundary in the requested region
 *			- searches the free lists for a free memory block
 *			  that can contain a free block of the requested size
 *			- if found, the free block is chopped as follows:
 *
 *			  -------------------------
 *			  | after free space >=0  |
 *			  -------------------------
 *			  | newly allocated space |
 *			  -------------------------	aligned start address
 *			  | before free space >=0 |
 *			  -------------------------	original free address
 *
 *	Returns:	- the address of the allocated memory or 0
 *			- sets errno accordingly
 */

static char
*maqalloc()

{
	int		upper;			/* upper region boundary */
	int		lower;			/* lower region boundary */
	int		i;			/* favourite index variable */
	int		*pint;			/* index pointer */
	int		align;			/* alignment quantity */
	int		key[2];			/* BYSIZE parameters */
	char		*start;			/* start of alloc mem block */
	struct memblock *fit;			/* large enough mem block */
	struct memblock *before;		/* free space before */
	struct memblock *after;			/* free space after */

	if (request->maq_amount <= 0) {
		errno = EINVAL;
		return((char *) 0);
	}
/*
 * Set the alignment quantity.  Zero means the word size.
 * Alignment must be word size divisible.  Round it up.
 */
	align = (request->maq_align == 0) ? sizeof(int) : request->maq_align;
	align = (align + (sizeof(int) - 1)) / sizeof(int) * sizeof(int);
/*
 * Round up amount to a full word.
 */
	request->maq_amount = (request->maq_amount + ((sizeof(int) - 1)))
		/ sizeof(int) * sizeof(int);
/*
 * Set region search bounds.
 */
	if (request->maq_region == ANYBANK) {
		lower = 0;
		upper = NMAIN - 1;
	} else {
		lower = request->maq_region - 1;
		upper = lower;
	}
/*
 * Malloc may require two new descriptors.
 */
	before = getdesc();
	after = getdesc();

	if ((before == MANULL) || (after == MANULL)) {

		if (before != MANULL) {
			insert(&empty, before, NOCOMPACT);
		}

		if (after != MANULL) {
			insert(&empty, after, NOCOMPACT);
		}

		return((char *) 0);
	}
/*
 * Look for some memory to allocate.
 */
	key[0] = request->maq_amount;
	key[1] = align;

	for (i = lower; (i <= upper) &&
		(! (fit = delete(&memfree[i], (int) key, BYSIZE))); ++i) {
	}

	if (! fit) {
		errno = ENOMEM;
		return((char *) 0);
	}

	start = (char *) (((unsigned int) fit->ma_start + (align - 1)) /
		align * align);
	before->ma_size = start - fit->ma_start;

	if (before->ma_size > 0) {
		before->ma_start = fit->ma_start;
		before->ma_pid = NOTPROCESS;
		insert(&memfree[i], before, NOCOMPACT);
		--nempty;
	} else {
		before->ma_start = (char *) 0;
		insert(&empty, before, NOCOMPACT);
	}

	after->ma_size = (fit->ma_start + fit->ma_size) -
		(start + request->maq_amount);

	if (after->ma_size > 0) {
		after->ma_start = start + request->maq_amount;
		after->ma_pid = NOTPROCESS;
		insert(&memfree[i], after, NOCOMPACT);
		--nempty;
	} else {
		after->ma_start = (char *) 0;
		insert(&empty, after, NOCOMPACT);
	}

	fit->ma_start = start;
	fit->ma_size = request->maq_amount;
	fit->ma_node = request->maq_node;
	fit->ma_pid = request->maq_pid;

	insert(&memalloc[i], fit, NOCOMPACT);
	errno = 0;
/*
 * Initialize allocated memory to zeros.  We need an efficient bzero() here.
 */
	pint = (int *) fit->ma_start;
	i = fit->ma_size / 4;

	while (i) {
		*pint++ = 0;
		--i;
	}

	return(fit->ma_start);
}

/*
 *	maqfree
 *
 *	Function:	- free a block of allocated memory
 *			- search the allocated list of every region until
 *			  the block is found
 *
 *	Returns:	- 0 or ERROR
 */

static int
maqfree()

{
	int		i;		/* favourite index variable */
	struct memblock	*old;		/* allocated block */

	for (i = 0; (i < NREGIONS) &&
		(! (old = delete(&memalloc[i], (int) request->maq_address,
					BYSTART))); ++i) {
	}

	if (! old) {
		errno = ENOTMALLOCED;
		return(ERROR);
	}

	old->ma_pid = NOTPROCESS;
	insert(&memfree[i], old, COMPACT);
/*
 * Can we release a descriptor block back to the free memory list?
 */
	if (nempty >= (MAMAX * 1.5)) {
		freedesc();
	}

	return(0);
}

/*
 *	maqfreeall
 *
 *	Function:	- free all blocks allocated to a process ID
 *			- search the allocated list of every region until
 *			  all blocks are found
 *
 *	Returns:	- 0 or ERROR if no blocks were allocated to PID
 */

static int
maqfreeall()

{
	int		i;		/* favourite index variable */
	int		found;		/* found an allocated block? */
	struct memblock	*old;		/* allocated block */

	found = FALSE;

	for (i = 0; (i < NREGIONS); ++i) {

		while (old = delete(&memalloc[i], request->maq_pid, BYPID)) {
			found = TRUE;
			old->ma_pid = NOTPROCESS;
			insert(&memfree[i], old, COMPACT);
/*
 * Can we release a descriptor block back to the free memory list?
 */
			if (nempty >= (MAMAX * 1.5)) {
				freedesc();
			}
		}
	}

	if (found) {
		return(0);
	} else {
		errno = ENOTMALLOCED;
		return(ERROR);
	}
}

/*
 *	maqstate
 *
 *	Function:	- sends memory block descriptor to client
 *			- does not exceed block limit given by client
 */

static void
maqstate()

{
	struct nmsg	nhead;		/* network message descriptor */
	struct memblock	*tbl;		/* current mem block desc. table */
	struct mareply	*sreply;	/* state reply */
	int		iregion;	/* save region index */

	nhead.nh_node = request->maq_src_node;
	nhead.nh_event = request->maq_src_event;
	nhead.nh_type = 0;
	nhead.nh_flags = BYBUF | DINTMSG;
/*
 * Send used memory block descriptors until none are left or the
 * client supplied limit is reached.
 */
	sreply = (struct mareply *) nhead.nh_data;
	sreply->mar_reply = 0;
	sreply->mar_nblks = 0;
/*
 * Count the total number of descriptors.
 */
	for (tbl = tbldesc; tbl; tbl = tbl->ma_next) {
		sreply->mar_nblks += MAMAX;
	}

	for (tbl = tbldesc; tbl && request->maq_maxblks; tbl = tbl->ma_next) {

		nhead.nh_length = ((MAMAX > request->maq_maxblks) ?
			request->maq_maxblks : MAMAX) * sizeof(struct memblock);
		nhead.nh_msg = tbl->ma_start;
		iregion = tbl->ma_iregion;
		tbl->ma_iregion = MAMAX * sizeof(struct memblock);

		if (nsend(&nhead)) {
			panic(errno);
		}

		tbl->ma_iregion = iregion;
		request->maq_maxblks -= nhead.nh_length /
			sizeof(struct memblock);
	}
}
