/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	clientio.c,v $
 * Revision 2.2  92/04/30  11:46:18  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/02/23  20:28:46  gdburns
 * Add #ifdef to c.h.
 * 
 * Revision 2.1  91/03/20  13:10:39  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.4  91/01/14  15:33:27  gdburns
 * Changed pick_aux* from static kio to passed parameters.
 * 
 * Revision 2.0.1.3  90/11/15  16:57:42  raja
 * Added siesta list handling to enable multiple hi-pri processes to
 * co-exist with the kernel.
 * 
 * Revision 2.0.1.2  90/07/13  19:32:31  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/07/02  16:43:12  gdburns
 * Added _kbail() for machine dependent process termination.
 * 
 * Revision 2.0  90/02/20  11:57:45  trillium
 * Homo erectus
 *
 *	Function:	- interface between kernel runtime request library
 *			  and native environment
 *			- T414/T800 version
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#ifdef itb
#include <c.h>
#endif
#include <kernel.h>
#include <kio.h>
#include <kmsg.h>
#include <typical.h>
#include <terror.h>
#include <transputer/transputer.h>
#include <machine/maconfig.h>

/*
 * external variables
 */
extern int		errno;			/* global error variable */
extern struct kio_t	_kio;			/* Kernel I/O block */

/*
 * external functions
 */
extern void		_follow_signal();	/* check and follow signals */
extern void		take_siesta();		/* stop the process */

/*
 * global functions
 */
int			_make_kreq();		/* make a kernel request */
int			_send_kmsg();		/* send message to peer */
int			_recv_kmsg();		/* receive message from peer */
void			_init_kio();		/* initialize kernel I/O */
void			_kbail();		/* terminate self */

/*
 * local functions
 */
static void		clobber();		/* when send len > recv len */

/*
 *	_init_kio
 *
 *	Function:	- initializes kernel interface
 */

void
_init_kio()

{
	_kio.ki_portd = NOTPROCESS;
	_kio.ki_auxd = NOTPROCESS;
	_kio.ki_clientd = NOTPROCESS;
	_kio.ki_port = (int) &_kio.ki_portd;
	_kio.ki_aux = (int) &_kio.ki_auxd;
	_kio.ki_client = (int) &_kio.ki_clientd;
}


/*
 *	_createport
 *
 *	Function:	- create a transputer communication port
 *			- real simple - initialize the channel and
 *			  return a pointer to it
 *
 *	Accepts:	- pointer to a channel
 *
 *	Returns:	- the same pointer
 */

int
_createport(pchan)

int			*pchan;

{
	*pchan = NOTPROCESS;
	return((int) pchan);
}


/*
 *	_make_kreq
 *
 *	Function:	- make a kernel request
 *			- sends the request structure to the kernel
 *			- receives the reply structure from the kernel
 *
 *	Accepts:	- ptr to kernel request structure
 *			- ptr to kernel reply structure
 *
 *	Returns:	- error indicator
 *			- sets errno accordingly if a kernel error
 *			  is returned
 */

int
_make_kreq(req, reply)

struct kreq			*req;
struct kreply			*reply;

{
	int			reply_code;	/* kernel reply code */

/*
 * Loop through possible signal interruptions.
 */
	do {
/*
 * Send the request to the kernel.  If the channel is already
 * taken, deschedule and take a small siesta.  The kernel will
 * wake you up when ready for you to continue.
 * Based on an idea by Aarne Rantala at VTT.
 */
		if ( ((*(int *) KERNEL_CHAN) > ONCHIP_START) ||
			((*(int *) FRONT_SIESTA) != NOTPROCESS) ) {

			do {
				take_siesta();
			} while ( (*(int *) KERNEL_CHAN) > ONCHIP_START);
		}

		out(KERNEL_CHAN, (char *) req, sizeof(struct kreq));
/*
 * Receive the reply from the kernel;
 */
		in(_kio.ki_client, (char *) reply, sizeof(struct kreply));

		reply_code = reply->kr_reply;

		if (reply->kr_signal) {
			_kio.ki_signal |= reply->kr_signal;

			if (reply_code == EINTR) {
				_follow_signal();
			}
		}

	} while ((reply_code == EINTR) &&
		(reply->kr_signal & _kio.ki_sigretry));

	if (reply_code != SUCCESS) {
		errno = reply_code;
		return(ERROR);
	}

	return(0);
}


/*
 *	_recv_kmsg
 *
 *	Function:	- receive a message from another process
 *
 *	Accepts:	- a pointer to a kernel message header
 *			- my chosen aux port
 *			- my chosen aux port descriptor
 *
 *	Returns:	- an error indicator
 */


int
_recv_kmsg(msgptr, pick_aux, pick_auxd)

struct kmsg			*msgptr;	/* pointer to message header */
int				pick_aux;
int				pick_auxd;

{
	int			flags;
	int			data_port;
	int			break_port;
	char			break_hole;


	flags = msgptr->k_flags;
/*
 * Do we reverse ports?
 */
	if (flags & KREVPORT) {
		data_port = _kio.ki_peer_itbdata;
		break_port = _kio.ki_peer_break;
/*
 *		break_port = _kio.ki_break;
 */
	}

	else {
		data_port = pick_aux;
		break_port = _kio.ki_peer_break;
	}
/*
 * Is some idiot sending us more data than we can handle?
 */
	if (flags & KCLOBBER) {
		clobber(data_port, msgptr);
	}

	else {
		if (msgptr->k_length) {
			in(data_port, msgptr->k_msg, msgptr->k_length);
		}
	}
/*
 * Do we need to break a virtual circuit with a sync message?
 */
	if ((flags & KBREAK) && (data_port <= Link3In)) {
		out(break_port, &break_hole, sizeof(break_hole));
	}

	return(0);
}


/*
 *	_send_kmsg
 *
 *	Function:	- send a message to another process
 *
 *	Accepts:	- a pointer to a kernel message header
 *			- my chosen aux port
 *			- my chosen aux port descriptor
 *
 *	Returns:	- an error indicator
 */


int
_send_kmsg(msgptr, pick_aux, pick_auxd)

struct kmsg			*msgptr;	/* pointer to message header */
int				pick_aux;
int				pick_auxd;

{
	char			break_hole;


	if (msgptr->k_length) {
		out(_kio.ki_peer_itbdata, msgptr->k_msg, msgptr->k_length);
	}
/*
 * Do we need to break a virtual circuit with a sync message?
 */
	if ((msgptr->k_flags & KBREAK) && (_kio.ki_peer_itbdata <= Link3Out)) {
/* 
 *		in(_kio.ki_break, &break_hole, sizeof(break_hole));
 */
 		in(_kio.ki_peer_break, &break_hole, sizeof(break_hole));
	}

	return(0);
}


/*
 *	clobber
 *
 *	Function:	- Since the t800/t414 don't like unequal length
 *			  messages, we have to go out of our way to ensure
 *			  that it never happens.
 *			- To add to our woes, we must handle the soft versus
 *			  hard cases differently.
 *
 *	Accepts:	- communication port address
 *			- message pointer
 */

static void
clobber(port, msgptr)

int			port;
struct kmsg		*msgptr;

{
	char		slush_hole;

	if (port <= Link3In) {
/*
 * hard channel
 * Read what we need and then trash the rest.
 */
		if (msgptr->k_length > 0) {
			in(port, msgptr->k_msg, msgptr->k_length);
		}

		while(_kio.ki_slush > 0) {
			in(port, &slush_hole, sizeof(slush_hole));
			--_kio.ki_slush;
		}
	}

	else {
/*
 * soft channel
 * Busy wait for the peer to arive.  They are synchronized, so this
 * won't take forever.
 */
		while(*((int *) port) == MININT);

		if (msgptr->k_length > 0) {
			in(port, msgptr->k_msg, msgptr->k_length);
		}

		else {
			in(port, &slush_hole, sizeof(slush_hole));
		}
	}
}

/*
 *	_kcleanup
 *
 *	Function:	- a stub for this implementation
 */

int
_kcleanup()

{
	return(0);
}

/*
 *	_kbail
 *
 *	Function:	- stop transputer process
 *			- does not return
 */

void
_kbail()

{
	stopp();
}
