/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	kcreate.c,v $
 * Revision 2.2  92/04/30  11:46:22  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/01/07  22:26:09  gdburns
 * Declare rmalloc().
 * 
 * Revision 2.1  91/03/20  13:10:42  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/09/13  14:54:29  gdburns
 * Include machine dependent memory region header file.
 * 
 * Revision 2.0.1.1  90/07/13  19:32:30  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:57:41  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  02:00:32  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  16:21:19  trillium
 * Initial revision
 * 
 *
 *	Function:	- create a new process
 *			- initializes the workspace and schedules the process
 *			
 *			- A new process must have three pointers:
 *			  an instruction pointer, a workspace pointer
 *			  and a static link which points to the external
 *			  data space.  On dispatch, the new process's
 *			  workspace is laid out as below.
 *
 *			  |-----------------|
 *			  | PCB ptr         |  WP + 2
 *			  |-----------------|
 *			  | static link ptr |  WP + 1
 *			  |-----------------|
 *			  | return address  |  WP + 0
 *			  |-----------------|
 *			  | instruction ptr |  WP - 1
 *			  |-----------------|
 *
 *			  As the entry point will usually be a runtime
 *			  startoff routine that does not return, the
 *			  return address need not be initialized.
 *
 *			- argv string must be null terminated in order
 *			  to calculate argc
 *
 *	Accepts:	- ptr to a new process structure
 *			- argv to be passed to main
 */

#include <kio.h>
#include <machine/maconfig.h>
#include <mareq.h>
#include <net.h>
#include <pcb.h>
#include <t.out.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;
extern struct pcb_t	*_pcb;

/*
 * external functions
 */
extern int		kexit();
extern char		*rmalloc();


int
kcreate(newproc, argv)

struct kproc_t		*newproc;
char			*argv[];

{
	struct crt0_t	*crt;			/* initial stack layout */
	int		argc;			/* argument count */
	int		region;			/* stack memory region */

/*
 * Check for valid load module.
 */
	if ((newproc->kp_header->a_magic != TO_MAGIC_TN_X) &&
		(newproc->kp_header->a_magic != TO_MAGIC_T4_X) &&
		(newproc->kp_header->a_magic != TO_MAGIC_T8_X)) {
		errno = EMAGIC;
		return(ERROR);
	}

	newproc->kp_stksize += sizeof(struct crt0_t);
/*
 * Allocate stack space.
 */
	region = (newproc->kp_rtf & RTF_ONCHIP) ? ONCHIP : ANYBANK;

	if (! (newproc->kp_stack = rmalloc(LOCAL, newproc->kp_stksize,
						region, 0))) {
		return(ERROR);
	}
/*
 * Count arguments.
 */
	if (argv == (char **) 0) {
		argc = 0;
	}

	else {
		for (argc = 0; argv && argv[argc]; ++argc);
	}

	crt = (struct crt0_t *) (newproc->kp_stack + (newproc->kp_stksize -
					sizeof(struct crt0_t)));

	crt->t0_ip = (char *) newproc->kp_header + sizeof(struct exec) +
			newproc->kp_header->a_entry;
	crt->t0_ret = (char *) (int) kexit;
/*
 * Set the static link.
 */
	crt->t0_slink = (newproc->kp_slink) ? newproc->kp_slink :
			(char *) newproc->kp_header +
			newproc->kp_header->a_text +
			newproc->kp_header->a_drsize +
			newproc->kp_header->a_syms +
			sizeof(struct exec);

	crt->t0_ppcb = &crt->t0_pcb;
	crt->t0_pcb.pc_pid = ((int) newproc->kp_stack + newproc->kp_stksize + 1)
				& ~MININT;
	crt->t0_pcb.pc_header = newproc->kp_header;
	crt->t0_pcb.pc_nodeid = getnodeid();
	crt->t0_pcb.pc_argc = argc;
	crt->t0_pcb.pc_argv = argv;
	crt->t0_pcb.pc_rtf = newproc->kp_rtf;
	crt->t0_pcb.pc_stack = newproc->kp_stack;
	crt->t0_pcb.pc_stksize = newproc->kp_stksize;
	crt->t0_pcb.pc_stdin = _pcb->pc_stdin;
	crt->t0_pcb.pc_stdout = _pcb->pc_stdout;
	crt->t0_pcb.pc_stderr = _pcb->pc_stderr;
/*
 * Run process at low priority.
 */
	runp(((char *) &crt->t0_ret) + 1);
	
	return(crt->t0_pcb.pc_pid);
}
