/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	kpar_Fe.c,v $
 * Revision 2.2  92/04/30  11:46:25  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/02/23  20:28:24  gdburns
 * Remove c.h.
 * 
 * Revision 2.1.1.2  91/12/16  15:04:42  gdburns
 * Add RTF_TRACE carry-over from parent.
 * 
 * Revision 2.1.1.1  91/11/27  16:59:30  gdburns
 * Declare rmalloc().
 * 
 * Revision 2.1  91/03/20  13:10:43  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:32:33  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:57:45  trillium
 * Homo erectus
 *
 *	Function:	- creates an iron weight process from a function
 *			- a copy is made of the current data segment; thus,
 *			  the new process inherits the current value of all
 *			  static variables
 *			- a new stack is allocated
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 *
 *	Accepts:	- ptr to function to use as text in new process
 *			- required stack size for new process
 *			- ptr to new process ID (value returned)
 *			- number of parameters to pass to new process
 *
 *	Returns:	- ptr to malloc'd space (stack/static data)
 */

#include <t.out.h>
#include <pcb.h>
#include <kio.h>
#include <net.h>
#include <typical.h>

/*
 * external variables
 */
extern int			errno;
extern struct pcb_t		*_pcb;
extern struct kio_t		_kio;

/*
 * external functions
 */
extern int			kexit();
extern char			*rmalloc();


char *kpar_Fe(par_func, stksize, newpid, nparms)

int			(*par_func)(),	/* ptr to parallel function */
			stksize,	/* stack size required */
			nparms,		/* # of runtime parameters */
			*newpid;	/* value returned is new process ID */

{
	char		*wp,		/* workspace */
			*new_slink;	/* new static data ptr */
	int		(*bug)(),	/* gentp bug */
			totalsize,	/* size of data/bss, stack and PCB */
			datasize,	/* size of data and bss segments */
			rtf_save,	/* save the runtime flags */
			*parmptr,	/* points at stack parameters */
			i;		/* favourite index variable */
	struct pcb_t	*newpcb,	/* new process's PCB */
			*pcb_save;	/* save the PCB ptr */


/*
 * Allocate space for data segment and stack segment.
 */
	datasize = _pcb->pc_header->a_data + _pcb->pc_header->a_bss;
	totalsize = stksize + datasize + sizeof (struct pcb_t);

	if (((new_slink = rmalloc(LOCAL, totalsize, 0, 0)) == (char *) 0)) {
		return((char *) 0);
	}

	newpcb = (struct pcb_t *) (new_slink +
			(totalsize - sizeof(struct pcb_t)));
	newpcb->pc_pid = *newpid = ((int) new_slink + totalsize + 1) & ~MININT;
	newpcb->pc_header = _pcb->pc_header;
	newpcb->pc_nodeid = _pcb->pc_nodeid;
	newpcb->pc_stack = new_slink + datasize;
	newpcb->pc_stksize = stksize;
	newpcb->pc_argc = _pcb->pc_argc;
	newpcb->pc_argv = _pcb->pc_argv;
/*
 * Temporary solution for standard descriptors.  We need a RTF_CHILD flag.
 */
	newpcb->pc_stdin = 0;
	newpcb->pc_stdout = 1;
	newpcb->pc_stderr = 2;

	wp = new_slink + datasize + stksize;
/*
 * Set up and copy static data segment.
 */
	pcb_save = _pcb;
	rtf_save = _kio.ki_rtf;
	_pcb = newpcb;
	_kio.ki_rtf = rtf_save & (RTF_SYSGEN | RTF_TRACE);

	bcopy((char *) __slink(), new_slink, datasize);

	_pcb = pcb_save;
	_kio.ki_rtf = rtf_save;
/*
 * Drop the user supplied parameters onto the stack starting at wp - 1.
 */
	parmptr = (&nparms) + 1;

	for (i = nparms; i > 0; --i) {
		*(((int *) wp) - i) = *(parmptr++);
	}

	wp = (char *) (((int *) wp) - nparms);

	*(((int *) wp) - 1) = (int) new_slink;
	*(((int *) wp) - 2) = (int) kexit;
	bug = par_func;
	*(((int *) wp) - 3) = (int) bug;
/*
 * Make that baby low priority.
 */
	wp++;
/*
 * Cast to int to make lint happy.
 */
	runp((int) (((int *) wp) - 2));

	return(new_slink);
}
