/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	kpar_He.c,v $
 * Revision 2.2  92/04/30  11:46:26  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/02/23  20:28:29  gdburns
 * Remove c.h.
 * 
 * Revision 2.1  91/03/20  13:10:44  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:32:34  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:57:46  trillium
 * Homo erectus
 *
 *	Function:	- create a helium weight process from a function
 *			- the existing data segment is shared by the new
 *			  process
 *			- a stack is supplied by the caller
 *
 *			- A new process must have three pointers:
 *			  an instruction pointer, a workspace pointer
 *			  and a static link which points to the external
 *			  data space.  On dispatch, the new process's
 *			  workspace is be laid out as below.
 *
 *			  |-----------------|
 *			  | static link ptr |  WP + 1
 *			  |-----------------|
 *			  | return address  |  WP + 0
 *			  |-----------------|
 *			  | instruction ptr |  WP - 1
 *			  |-----------------|
 *
 *			  The workspace address passed to kpar_He points
 *			  to the first location *past* the valid workspace
 *
 *			- "nparms" word only parameters are expected to
 *			  follow "nparms" on the calling stack to kpar_He
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 *
 *	Accepts:	- ptr to function to use as text in new process
 *			- workspace (top of stack) pointer
 *			- number of parameters to pass to par_func
 */

/*
 * external functions
 */
extern int		_exit();


kpar_He(par_func, wp, nparms)

int			(*par_func)(),	/* ptr to parallel function */
			nparms;		/* # of runtime parameters */
char			*wp;		/* stack ptr */

{
	int		(*bug)(),	/* gentp bug */
			*parmptr,	/* points at stack parameters */
			i;		/* favourite index variable */


/*
 * Drop the user supplied parameters onto the stack starting at wp - 1.
 */
	parmptr = (&nparms) + 1;

	for (i = nparms; i > 0; --i) {
		*(((int *) wp) - i) = *(parmptr++);
	}

	wp = (char *) (((int *) wp) - nparms);

	*(((int *) wp) - 1) = (int) __slink();
	*(((int *) wp) - 2) = (int) _exit;
	
	bug = par_func;
	*(((int *) wp) - 3) = (int) bug;
/*
 * Make that baby low priority.
 */
	wp++;
/*
 * Cast to int to make lint happy.
 */
	runp((int) (((int *) wp) - 2));

	return(0);
}
