/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	siesta.lstt.s,v $
 * Revision 2.2  92/04/30  11:46:30  trillium
 * Ohio Release
 * 
| Revision 2.1  92/02/23  20:30:23  gdburns
| initial revision
| 
 *	Function:	- management of the siesta queue
 *			- used when collisions occur during kernel
 *			  kernel requests by high priority processes
 *			- enqueue a process and stop it
 *			- dequeue a process and run it
 *			- t800 (lstt) version
 */

#include <machine/maconfig.h>
#define NOTPROCESS	0x80000000

/*
 * defined global functions
 */
	.t800					; CPU type
	.rel					; relocatable code
	.pub		take_siesta		; stop a process
	.pub		end_siesta		; run a stopped process

/*
 *	take_siesta
 *
 *	Function:	- link process at the end of the siesta list
 *			- stop the process
 *			- returns when the kernel reschedules the process
 */
take_siesta
	.ldc		FRONT_SIESTA		; read the contents of the
	ldnl		0			;	front siesta list ptr
	eqc		NOTPROCESS		; check if the list is empty
	cj		@not_first		;
/*
 * First process on the list so link it at the front of the list.
 */
	ldlp		0			; get the current Wptr
	.ldc		FRONT_SIESTA		; store it in the front
	stnl		0			;	siesta list ptr
	.ldc		0			; unconditionally jump
	cj		@link_end		;	to link_end
not_first
/*
 * Link it to the last process on the list.
 */
	ldlp		0			; get the current Wptr
	.ldc		BACK_SIESTA		; get ptr to the WS of the last
	ldnl		0			;	process on the list
	stnl		-2			; link it to the new process
link_end
/*
 * Link it to the end of the list.
 */
	ldlp		0			; get the current Wptr
	.ldc		BACK_SIESTA		; store it int the back
	stnl		0			;	siesta list ptr
	.ldc		NOTPROCESS		; declare the new process as the
	ldlp		0			;	last process on the list
	stnl		-2			;	by clearing its link
/*
 * Stop the process just linked.
 */
	stopp					; take a siesta
	.retf		0			; return

/*
 *	end_siesta
 *
 *	Function:	- remove and run the top process on the siesta list
 */
end_siesta
/*
 * Check if the siesta queue is empty.
 */
	ajw		-1			; reserve a local word
	.ldc		FRONT_SIESTA		; read the contents of the
	ldnl		0			;	front siesta list ptr
	dup					; store it in the temporary
	stl		0			;	local word location
	eqc		NOTPROCESS		; check if the list is empty
	cj		@not_empty		;
/*
 * Siesta list is empty.
 */
	.retf		1			; return deallocating the word
not_empty
/*
 * Siesta list is not empty.  Set the front of the list to point
 * to the next process in the list.
 */
	ldl		0			; get the pointer to the next
	ldnl		-2			;	process in the list
	dup					;
	.ldc		FRONT_SIESTA		; store it in the front list
	stnl		0			;	pointer
	eqc		NOTPROCESS		; check if the siesta list
	cj		@run_process		;	is now empty
/*
 * That was the only process on the list.  Declare the list empty.
 */
	.ldc		NOTPROCESS		; store the NOTPROCESS
	.ldc		BACK_SIESTA		;	constant in the back
	stnl		0			;	list pointer
/*
 * Run the process.
 */
run_process
	.ldc		NOTPROCESS		; store the NOTPROCESS constant
	ldl		0			;	in the next process ptr
	stnl		-2			;	of the current process
	ldl		0			; run the current process
	runp					;
	.retf		1			; return
	.end
