/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	siesta.uls.s,v $
 * Revision 2.2  92/04/30  11:46:35  trillium
 * Ohio Release
 * 
| Revision 2.1  92/02/23  20:34:16  gdburns
| initial revision
| 
 *	Function:	- management of the siesta queue
 *			- used when collisions occur during kernel
 *			  kernel requests by high priority processes
 *			- enqueue a process and stop it
 *			- dequeue a process and run it
 *			- t414/t800 (uls) version
 */

#include <machine/maconfig.h>
#define NOTPROCESS	0x80000000


	EXPORT	_take_siesta, _end_siesta

/*
 *	take_siesta
 *
 *	Function:	- link process at the end of the siesta list
 *			- stop the process
 *			- returns when the kernel reschedules the process
 */
	CSEG
_take_siesta:

/*
 * Check if the siesta queue is empty.
 */
	ldc	FRONT_SIESTA
	ldnl	0
	eqc	NOTPROCESS
	cj	not_first
/*
 * First process on the list so link it at the front of the list.
 */
	ldlp	0
	ldc	FRONT_SIESTA
	stnl	0
	ldc	0
	cj	link_end
/*
 * Link it to the last process on the list.
 */
not_first:
	ldlp	0
	ldc	BACK_SIESTA
	ldnl	0
	stnl	-2
/*
 * Link it to the end of the list.
 */
link_end:
	ldlp	0
	ldc	BACK_SIESTA
	stnl	0
	ldc	NOTPROCESS
	ldlp	0
	stnl	-2
/*
 * Stop the process just linked.
 */
	stopp
	ret

/*
 *	end_siesta
 *
 *	Function:	- remove and run the top process on the siesta list
 */
	CSEG
_end_siesta:
	
/*
 * Check if the siesta queue is empty.
 */
	ajw	-1
	ldc	FRONT_SIESTA
	ldnl	0
	dup
	stl	0
	eqc	NOTPROCESS
	cj	not_empty
	ajw	1
	ret
/*
 * Siesta list is not empty.
 */
not_empty:
	ldl	0
	ldnl	-2
	dup
	ldc	FRONT_SIESTA
	stnl	0
	eqc	NOTPROCESS
	cj	run_process
/*
 * That was the only process on the list.  Declare the list empty.
 */
	ldc	NOTPROCESS
	ldc	BACK_SIESTA
	stnl	0
/*
 * Run the process.
 */
run_process:
	ldc	NOTPROCESS
	ldl	0
	stnl	-2
	ldl	0
	runp
	ajw	1
	ret
