/*
 *      Office of Research Computing				Trillium
 *      Copyright 1989 The Ohio State University
 *	RBD,VSD
 *
 *	$Log:	LTOT.c,v $
 * Revision 2.2  92/04/30  11:53:32  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  91/07/13  04:32:36  raja
 * Enhanced FLTOTF portability and a face-lift for the file.
 * 
 * Revision 2.1.1.1  91/07/10  18:17:36  raja
 * Fixed function referencing problem.  Fortran wrappers use the C macros
 * for byte ordering.
 * 
 * Revision 2.1  91/03/20  13:34:51  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:34:44  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/15  15:54:20  dixit-v
 * Changed function names.
 * 
 * Revision 2.0  90/02/20  11:58:56  trillium
 * Homo erectus
 * 
 *	Function:	- f77 otb interface to ltot, ltots, ltotd, ltotf
 */

#include <t_types.h>
#include <math.h>

/*
 *	iltot_
 *
 *	Function:	- f77 callable wrapper (ILTOT) for ltot
 *			- FORTRAN integer function
 *			- change word to Trollius byte ordering
 *
 *	FORTRAN:	- new_word = ILTOT(word)
 *
 *	Accepts:	- word (pass by reference)
 *
 *	Returns:	- word in Trollius byte ordering
 */
int
iltot_(iword)

int			*iword;

{
	return(ltot(*iword));
}

/*
 *	iltots_
 *
 *	Function:	- f77 callable wrapper (ILTOTS) for ttols
 *			- FORTRAN integer*2 function
 *			- change integer*2 to Trollius byte ordering
 *
 *	FORTRAN:	- new_short = ILTOTS(short)
 *
 *	Accepts:	- short (pass by reference)
 *
 *	Returns:	- short in Trollius byte ordering
 */
short
iltots_(sword)

short			*sword;

{
	return(ltots(*sword));
}

/*
 *	fltotd_
 *
 *	Function:	- f77 callable wrapper (FLTOTD) for ltotd
 *			- FORTRAN real*8 function
 *			- change double precision real to Trollius byte ordering
 *
 *	FORTRAN:	- new_double = FLTOTD(double)
 *
 *	Accepts:	- real*8 (pass by reference)
 *
 *	Returns:	- real*8 in Trollius byte ordering
 */
double
fltotd_(dword)

double			*dword;

{
	return(ltotd(*dword));
}

/*
 *	fltotf_
 *
 *	Function:	- f77 callable wrapper (FLTOTF) for ltotf
 *			- FORTRAN real function
 *			- change real to Trollius byte ordering
 *
 *	FORTRAN:	- new_real = FLTOTF(real)
 *
 *	Accepts:	- real (pass by reference)
 *
 *	Returns:	- real in Trollius byte ordering
 */
FLOATFUNCTIONTYPE
fltotf_(fword)

float			*fword;

{
	float		new_fword;
/*
 * Uses architecture dependent macros to allow returning a C float to a
 * FORTRAN real (see <math.h>).  This is needed because otherwise
 * C would upgrade the float to a double before passing it back.
 */
	new_fword = ltotf(*fword);
	RETURNFLOAT(new_fword);
}
