/*
 *      Office of Research Computing				Trillium
 *      Copyright 1989 The Ohio State University
 *	RBD,VSD
 *
 *	$Log:	TTOL.c,v $
 * Revision 2.2  92/04/30  11:53:37  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  91/07/13  04:29:57  raja
 * Enhanced FTTOLF portability and a face-lift for the file.
 * 
 * Revision 2.1.1.1  91/07/10  18:22:05  raja
 * Fixed function referencing problem.  Fortran wrappers use the C macros
 * for byte ordering.
 * 
 * Revision 2.1  91/03/20  13:34:56  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:34:45  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/15  16:06:54  dixit-v
 * Changed function names.
 * 
 * Revision 2.0  90/02/20  11:58:56  trillium
 * Homo erectus
 * 
 *	Function:	- f77 otb interface to ttol, ttols, ttold, ttolf
 */

#include <t_types.h>
#include <math.h>

/*
 *	ittol_
 *
 *	Function:	- f77 callable wrapper (ITTOL) for ttol
 *			- FORTRAN integer function
 *			- change word to local byte ordering
 *
 *	FORTRAN:	- new_word = ITTOL(word)
 *
 *	Accepts:	- word (pass by reference)
 *
 *	Returns:	- word in local byte ordering
 */
int
ittol_(iword)

int			*iword;

{
	return(ttol(*iword));
}

/*
 *	ittols_
 *
 *	Function:	- f77 callable wrapper (ITTOLS) for ttols
 *			- FORTRAN integer*2 function
 *			- change integer*2 to local byte ordering
 *
 *	FORTRAN:	- new_short = ITTOLS(short)
 *
 *	Accepts:	- short (pass by reference)
 *
 *	Returns:	- short in local byte ordering
 */
short
ittols_(sword)

short			*sword;

{
	return(ttols(*sword));
}

/*
 *	fttold_
 *
 *	Function:	- f77 callable wrapper (FTTOLD) for ttold
 *			- FORTRAN real*8 function
 *			- change double precision real to local byte ordering
 *
 *	FORTRAN:	- new_double = FTTOLD(double)
 *
 *	Accepts:	- real*8 (pass by reference)
 *
 *	Returns:	- real*8 in local byte ordering
 */
double
fttold_(dword)

double			*dword;

{
	return(ttold(*dword));
}

/*
 *	fttolf_
 *
 *	Function:	- f77 callable wrapper (FTTOLF) for ttolf
 *			- FORTRAN real function
 *			- change real to local byte ordering
 *
 *	FORTRAN:	- new_real = FTTOLF(real)
 *
 *	Accepts:	- real (pass by reference)
 *
 *	Returns:	- real in local byte ordering
 */
FLOATFUNCTIONTYPE
fttolf_(fword)

float			*fword;

{
	float		new_fword;
/*
 * Uses architecture dependent macros to allow returning a C float to a
 * FORTRAN real (see <math.h>).  This is needed because otherwise
 * C would upgrade the float to a double before passing it back.
 */
	new_fword = ttolf(*fword);
	RETURNFLOAT(new_fword);
}
