
/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	RBD
 *
 *	$Log:	ttime.c,v $
 * Revision 2.2  92/04/30  11:55:01  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  13:36:00  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:34:10  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:58:36  trillium
 * Homo erectus
 *  
 *
 *	Function:	- Returns the contents of a system timer
 *			  expressed in seconds.
 *			- Used primarily for recording elapsed 
 *			  time in the user code, in a portable
 *			  manner.
 *			- The resolution is not guaranteed, as it
 *			  depends on the underlying hardware,
 *			  but the units (seconds) are
 *			  guaranteed.
 *			- This code is for OTB and uses the 
 *			  Unix(tm) timing support.
 */

#include <sys/time.h>
#include <typical.h>

/*
 * External variables.
 */
extern int	errno;


/*
 * ttime() returns the contents of the system timer (a measure
 * of time since an arbitrary but fixed point in time) expressed
 * in seconds. 
 */

double
ttime()

{
/*
 * Local variables.
 */
	struct	timeval	now;
	double		sec;

/*
 * Get the system time.
 */
	if (gettimeofday(&now, (struct timezone *)0) != SUCCESS) {
		panic(errno);
	}

/*
 * Return time in seconds.
 */
	sec = (double)now.tv_sec + (double)now.tv_usec/1000000;
	return(sec);
}

