/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RKM/GDB
 *
 *	$Log:	clientio.msg.c,v $
 * Revision 2.2  92/04/30  11:55:42  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  09:29:42  gdburns
 * enhanced robustness and facial
 * 
 * Revision 2.1  91/03/20  15:11:25  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.5  91/01/14  15:34:13  gdburns
 * Changed pick_aux* from static kio to passed parameters.
 * 
 * Revision 2.0.1.4  90/11/29  19:55:49  gdburns
 * Converted to new filed.
 * 
 * Revision 2.0.1.3  90/10/18  15:29:09  gdburns
 * _kbail exits with caller supplied code.
 * 
 * Revision 2.0.1.2  90/07/13  19:34:05  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/14  16:42:57  gdburns
 * To terminate process, call _kbail() in system dependent clientio module.
 * 
 * Revision 2.0  90/02/20  11:58:34  trillium
 * Homo erectus
 * 
 *	Function:	- interface between kernel runtime request library
 *	  		  and native environment
 *			- OTB message queue version
 */

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#include <stdio.h>
#include <freq.h>
#include <kernel.h>
#include <kio.h>
#include <kmsg.h>
#include <net.h>
#include <terror.h>
#include <typical.h>

#define MAXKMSGLEN	2048			/* kernel message packet size */

/*
 * external variables
 */
extern int		errno;			/* global error variable */
extern struct kio_t	_kio;			/* Kernel I/O block */

/*
 * external functions
 */
extern void		_follow_signal();	/* check and follow signals */
extern key_t		ftok();			/* produce tokens */
extern char		*getenv();
extern char		*memcpy();

/*
 * global functions
 */
int			_create_port();		/* create comm. port */
int			_make_kreq();		/* make a kernel request */
int			_recv_kmsg();		/* receive message from peer */
int			_send_kmsg();		/* send message to peer */
void			_init_kio();		/* initialize kernel I/O */
void			_kcleanup();		/* remove kernel interface */
void			_kbail();		/* terminate self */

/*
 * local functions
 */
static void		_ktrap();		/* trap interrupts */

/*
 * local variables
 */
static key_t 		token;			/* unique token */

/*
 * global variables
 */
struct kio_t		_kio;			/* Kernel I/O block */
struct fclient		_ufd[FUMAX];		/* user file desc. */

/*
 *	_init_kio
 *
 *	Function:	- initializes kernel interface
 *			- learns the msqid associated with the kernel
 *			- creates a personal msqid
 */

void
_init_kio()

{
	int		fd_token;		/* token file desc. */
	char		f_token[32];		/* pathname to create token */
/*
 * Trap some common signals so that we can remove the message queue.
 */
 	(void) signal(SIGTERM, _ktrap);
 	(void) signal(SIGINT, _ktrap);
	(void) signal(SIGHUP, _ktrap);
	(void) signal(SIGFPE, _ktrap);
	(void) signal(SIGSEGV, _ktrap);
/*
 * Generate the token used by kernel to create its message queue.
 */
	sprintf(f_token, "/tmp/%s%dk", getenv("USER"), getuid());

	if ((token = ftok(f_token, KERNEL_ID)) < 0) {
		fprintf(stderr, "client[%d]: no running kernel\n", getpid());
		fprintf(stderr, "client[%d]: %s\n", getpid(), f_token);
		perror("client (_init_kio (ftok))");
		exit(errno);
	}		
/*
 * Learn kernel qid.
 */
	if ((_kio.ki_kernel = msgget(token, 0600)) < 0) {

		if (errno == ENOENT) {
			fprintf(stderr, "client[%d]: no running kernel\n",
				getpid());
			exit(ENOKERNEL);
		} else {
			fprintf(stderr, "client[%d]: can't find kernel",
				getpid());
			perror("client (_init_kio (msgget))");
			exit(errno);
		}
	}
/*
 * Create qid keyed by my process ID.
 */
	sprintf(f_token, "/tmp/%s%dc", getenv("USER"), getpid());

	if ((fd_token = open(f_token, O_CREAT, 0600)) < 0) {
		fprintf(stderr, "client[%d]: %s\n", getpid(), f_token);
		perror("client (_init_kio (open))");
		exit(errno);
	}

	if ((token = ftok(f_token, KERNEL_ID)) < 0) {
		fprintf(stderr, "client[%d]: %c %s\n", getpid(), KERNEL_ID,
			f_token);
		perror("client (_init_kio (ftok))");
		exit(errno);
	}		

	(void) close(fd_token);
/*
 * Create client qid.
 */
	if ((_kio.ki_port = msgget(token, IPC_CREAT | IPC_EXCL | 0600)) < 0) {
		fprintf(stderr, "client[%d]: can't create msg queue\n",
			getpid());
		perror("client (_init_kio (msgget))");
		exit(errno);
	}

	_kio.ki_client = _kio.ki_port;
/*
 * Set auxillary port to my message queue token.
 */
	_kio.ki_aux = (int) token;
/*
 * Set default stdio descriptors.
 */
	_kio.ki_stdin =  0;
	_kio.ki_stdout = 1;
	_kio.ki_stderr = 2;
}

/*
 *	_ktrap
 *
 *	Function:	- calls _kcleanup
 *			- re-installs default handler
 *			- delivers signal to ourselves
 *	Accepts:	- signal number
 */

static void
_ktrap(sig)

int			sig;

{
	_kcleanup();
	(void) signal(sig, SIG_DFL);
	(void) kill(getpid(), sig);
	exit(0);
}

/*
 *	_kcleanup
 *
 *	Function:	- cleans up kernel interface 
 *			- removes the client message queue
 */

void
_kcleanup()

{
	char		f_token[32];		/* pathname to create token */

/*
 * Remove the message queue.
 */
 	(void) msgctl(_kio.ki_port, IPC_RMID, (struct msqid_ds *) 0);
/*
 * Remove the client token creation file.
 */
	sprintf(f_token, "/tmp/%s%dc", getenv("USER"), getpid());
	(void) unlink(f_token);
}

/*
 *	_createport
 *
 *	Function:	- get filter for senders
 *			- We use the sender's message queue token to select
 *			  the sender in case multiple virtual circuits cause
 *			  multiple senders to use our message queue.
 *	Accepts:	- ptr to a communication port descriptor - meaningless
 *			  in this implementation
 *	Returns:	- our message queue token
 */

int
_createport(desc)

int			*desc;

{
	*desc = 0;
	return((int) token);
}

/*
 *	_make_kreq
 *
 *	Function:	- makes a kernel request
 *			- sends the request structure to the kernel
 *			- receives the reply structure from the kernel
 *	Accepts:	- ptr to kernel request structure
 *			- ptr to kernel reply structure
 *	Returns:	- 0 or ERROR
 *			- sets errno accordingly if a kernel error is returned
 */

int
_make_kreq(req, reply)

struct kreq		*req;
struct kreply		*reply;

{
	int		ret;		/* syscall return code */
	int		reply_code;	/* kernel reply code */

	struct {			/* request message */
		long		mq_type;
		struct kreq	mq_text;
	} msgreq;

	struct {			/* reply message */
		long		mr_type;
		struct kreply	mr_text;
	} msgreply;

/*
 * Check if attached to the kernel.
 */
	if (! ((_kio.ki_rtf & RTF_KERNEL) || (req->kq_request == KATTACH))) {
		errno = ENOTATTACHED;
		return(ERROR);
	}

	msgreq.mq_type = KERNEL_REPLY; 
/*
 * Loop through possible Trollius signal interruptions.
 */
	do {
		memcpy((char *) &msgreq.mq_text, (char *) req,
			sizeof(struct kreq));
/*
 * Send request.
 */
		while (((ret = msgsnd(_kio.ki_kernel, (struct msgbuf *) &msgreq,
			sizeof(struct kreq), 0)) < 0) && (errno == EINTR));

		if (ret < 0) {
			return(ret);
		}
/*
 * Receive reply.
 */
		while (((ret = msgrcv(_kio.ki_client,
			(struct msgbuf *) &msgreply,
			sizeof(struct kreply), (long) KERNEL_REPLY, 0)) < 0) &&
			(errno == EINTR)); 

		memcpy((char *) reply, (char *) &msgreply.mr_text,
			sizeof(struct kreply)); 

		if (ret < 0) {
			return(ret);
		} 

		reply_code = reply->kr_reply;

		if (reply->kr_signal) {
			_kio.ki_signal |= reply->kr_signal;
	    
			if (reply_code == EINTR) {
				_follow_signal();
			}
		}

	} while ((reply_code == EINTR) &&
		(reply->kr_signal & _kio.ki_sigretry));

	if (reply_code != 0) {
		errno = reply_code;
		return(ERROR);
	}

	return(0);
}

/*
 *	_recv_kmsg
 *
 *	Function:	- receives a message from another process
 *			- no synchronization on zero length messages
 *	Accepts:	- a pointer to a kernel message header
 *			- my chosen aux port
 *			- my chosen aux port descriptor
 *	Returns:	- an error indicator
 */

int
_recv_kmsg(pkhead, pick_aux, pick_auxd)

struct kmsg		*pkhead;	/* pointer to message header */
int			pick_aux;
int			pick_auxd;

{
	int		ret;		/* syscall return code */
	int		peerrem;	/* sender's idea of k_length */
	int		remaining;	/* bytes left to receive */
	int		slicelen;	/* bytes per slice */
	char		*sliceptr;	/* one slice of full message */

	struct {			/* data message */
		long	m_type;
		char	m_text[MAXKMSGLEN];
	} msg;

	if (!pkhead->k_length && (!(pkhead->k_flags & KCLOBBER))) {
		return(0);
	}

	remaining = pkhead->k_length;
	peerrem = (pkhead->k_flags & KCLOBBER) ? remaining + _kio.ki_slush :
		remaining;
	sliceptr = pkhead->k_msg;
/* 
 * Receive the peer's packetized messages.  Send a synchronization
 * message prior to each packet, inviting the peer to send a packet.
 */
	do {
		slicelen = (remaining > MAXKMSGLEN) ? MAXKMSGLEN : remaining;
		msg.m_type = (long) _kio.ki_sender;  
/*
 * Synchronize.
 */
		while (((ret = msgsnd(_kio.ki_peer_msgdata,   
			(struct msgbuf *) &msg, 1, 0)) < 0) &&
			(errno == EINTR)); 

		if (ret < 0) return(ret);
/*
 * Receive packet.
 */
		while (((ret = msgrcv(_kio.ki_msgdata, (struct msgbuf *) &msg,
			slicelen, (long) _kio.ki_peer_sender,
			MSG_NOERROR)) < 0) && (errno == EINTR));

		if (ret < 0) return(ret);

		if (slicelen > 0) {
			memcpy(sliceptr, msg.m_text, slicelen);
		}

		sliceptr += slicelen;
		remaining -= slicelen;
		peerrem -= MAXKMSGLEN;
	} while ((remaining > 0) || (peerrem > 0));

	return(0);
}

/*
 *	_send_kmsg
 *
 *	Function:	- sends a message to another process
 *			- zero length messages should be checked by the caller
 *	Accepts:	- a pointer to a kernel message header
 *			- my chosen aux port
 *			- my chosen aux port descriptor
 *	Returns:	- an error indicator
 */

int
_send_kmsg(pkhead, pick_aux, pick_auxd)

struct kmsg		*pkhead;	/* pointer to message header */
int			pick_aux;
int			pick_auxd;

{
	int		ret;		/* syscall return code */
	int		remaining;	/* bytes left to transmit */
	int		slicelen;	/* bytes per slice */
	char		*sliceptr;	/* one slice of full message */

	struct {			/* data message */
		long	m_type;
		char	m_text[MAXKMSGLEN];
	} msg;

	if (! pkhead->k_length) {
		return(0);
	}

	remaining = pkhead->k_length;
	sliceptr = pkhead->k_msg;
/*   
 * Packetize and send the message to the peer.  Recieve a synchronization
 * message prior to each packet, indicating the peer can recieve a packet.
 * This is required because there is no limit to the kernel message length.
 */
	do {
		slicelen = (remaining > MAXKMSGLEN) ? MAXKMSGLEN : remaining;
/*
 * Synchronize.
 */
		while (((ret = msgrcv(_kio.ki_msgdata, (struct msgbuf *) &msg,
			1, (long) _kio.ki_peer_sender, 0)) < 0) && 
			(errno == EINTR)); 

		if (ret < 0) return(ret);
/*
 * Send packet.
 */
		memcpy(msg.m_text, sliceptr, slicelen);
		msg.m_type = _kio.ki_sender;

		while (((ret = msgsnd(_kio.ki_peer_msgdata,
			(struct msgbuf *) &msg, slicelen, 0)) < 0)
			&& (errno == EINTR));

		if (ret < 0) return(ret);

		sliceptr += slicelen;
		remaining -= slicelen;
	} while (remaining > 0);

	return(0);
}

/*
 *	_kbail
 *
 *	Function:	- terminates the program
 *	Accepts:	- exit status
 */

void
_kbail(status)

int			status;

{
	exit(status);
}
