/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	clientio.udp.c,v $
 * Revision 2.2  92/04/30  11:55:44  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/04/27  09:30:21  gdburns
 * enhanced robustness and facial
 * 
 * Revision 2.1.1.2  91/10/27  21:32:34  gdburns
 * Exit functions are declared type void.
 * 
 * Revision 2.1.1.1  91/10/09  18:29:58  gdburns
 * Do not receive a zero length datagram.
 * 
 * Revision 2.1  91/03/20  15:11:28  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.6  91/01/14  15:34:37  gdburns
 * Changed pick_aux* from static kio to passed parameters.
 * 
 * Revision 2.0.1.5  90/11/29  19:55:57  gdburns
 * Converted to new filed.
 * 
 * Revision 2.0.1.4  90/10/18  15:29:31  gdburns
 * _kbail exits with caller supplied code.
 * 
 * Revision 2.0.1.3  90/07/13  19:34:02  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.2  90/06/14  16:52:12  gdburns
 * To terminate process, call _kbail() in system dependent clientio module.
 * 
 * Revision 2.0.1.1  90/06/01  15:52:32  gdburns
 * Upgraded the select() on KATTACH with more robust file descriptor bitmasks.
 * 
 * Revision 2.0  90/02/20  11:58:32  trillium
 * Homo erectus
 *
 *	Function:	- interface between kernel runtime request library
 *	  		  and native environment
 *			- OTB datagram version
 *
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#include <memory.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <netdb.h>

#include <freq.h>
#include <kernel.h>
#include <kio.h>
#include <kmsg.h>
#include <terror.h>
#include <typical.h>

#define NETADDR	struct sockaddr *

#if !defined(IPPORT_USERRESERVED)
#define IPPORT_USERRESERVED	5000
#endif

#define MAXKMSGLEN		4096

/*
 * local variables
 */
static struct sockaddr_in
			client_in;		/* client process address */

/*
 * external variables
 */
extern int		errno;			/* global error variable */
extern struct kio_t	_kio;			/* Kernel I/O block */

/*
 * external functions
 */
extern void		kexit();		/* Trillium stdio exit */
extern void		exit();			/* non-Trillium stdio exit */
extern void		_follow_signal();	/* check and follow signals */

/*
 * local functions
 */
static int		findport();		/* find an available UDP port */

/*
 * global functions
 */
int			_make_kreq();		/* make a kernel request */
int			_send_kmsg();		/* send message to peer */
int			_recv_kmsg();		/* receive message from peer */
int			_createport();		/* create comm. port */
void			_init_kio();		/* initialize kernel I/O */
void			_kbail();		/* terminate self */

/*
 * global variables
 */
struct kio_t		_kio;			/* Kernel I/O block */
struct fclient		_ufd[FUMAX];		/* user file desc. */

/*
 *	_init_kio
 *
 *	Function:	- initializes kernel interface
 *			- establishes UDP INET domain sockets to communicate
 *			  with the kernel and other Trillium processes
 *			- attempts to use descriptors at the top of the
 *			  descriptor table so as to stay away from normal
 *			  host node program activity
 */

void
_init_kio()

{
	char		myhostname[MAXHNAME];	/* space for hostname */
	struct hostent	*myhostent;		/* my host entry */
	u_long		ip_addr;		/* client's IP address */
	int		lsize;			/* size of INET address */

	lsize = sizeof(struct sockaddr_in);
	(void) strcpy(myhostname, "localhost");

	if ((myhostent = gethostbyname(myhostname)) == (struct hostent *) 0) {
		fprintf(stderr, "client[%d]: cannot get host entry for %s\n",
			getpid(), myhostname);
		perror("client (gethostbyname)");
		exit(1);
	}

	memcpy((caddr_t) &ip_addr, (caddr_t) myhostent->h_addr,
		myhostent->h_length);
/*
 * Set kernel port number.
 */
	_kio.ki_kernel = (int) htons((u_short) KERNEL_PORT + geteuid());
/*
 * Set up local address.
 */
	memset((caddr_t) &client_in, 0, lsize);
	memcpy((caddr_t) &client_in.sin_addr, (caddr_t) &ip_addr,
		sizeof(ip_addr));
	client_in.sin_family = AF_INET;
	client_in.sin_port = htons((u_short) 0);
/*
 * Create the primary communication descriptor.
 */
	_kio.ki_port = _createport(&_kio.ki_portd);
	_kio.ki_client = _kio.ki_port;

	if (_kio.ki_port < 0) {
		fprintf(stderr, "client[%d]: cannot get inet port\n",
			getpid());
		perror("client (_createport)");
		exit(errno);
	}
/*
 * Create the auxiliary communication descriptor.
 */
	_kio.ki_aux = _createport(&_kio.ki_auxd);

	if (_kio.ki_aux < 0) {
		fprintf(stderr, "client[%d]: cannot get aux inet port\n",
			getpid());
		perror("client (_createport)");
		exit(errno);
	}
/*
 * Set default stdio descriptors.
 */
	_kio.ki_stdin = 0;
	_kio.ki_stdout = 1;
	_kio.ki_stderr = 2;
}

/*
 *	_createport
 *
 *	Function:	- create a new socket and bind it to a UDP port
 *			  number by calling findport()
 *	Accepts:	- ptr to socket descriptor
 *	Returns:	- port number or ERROR
 */

int
_createport(newsock)

int			*newsock;

{
/*
 * Create the new socket.
 */
	if ((*newsock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		return(ERROR);
	}
/*
 * Bind the client's name to the socket.
 */
	if (findport(*newsock, &client_in) < 0) {
		return(ERROR);
	}

	return((int) (unsigned) client_in.sin_port);
}

/*
 *	findport
 *
 *	Function:	- find a scummy port and bind to it
 *			- bind(2) will indeed pick a port for us if sin_port
 *			  is initialized to 0, but it won't put the selected
 *			  port number into sin_port!!  we *really* need to
 *			  know that port number...
 *
 *	Accepts:	- a socket to be bound
 *			- pointer to a name to be bound with all fields set
 *			  except the port number
 *	Returns:	- 0 or ERROR
 */

static int
findport(s, sin)

int			s;		/* the socket */
struct sockaddr_in	*sin;		/* INET address */

{
	u_short		port;

	for (port = IPPORT_RESERVED; port < IPPORT_USERRESERVED; ++port) {
		sin->sin_port = htons(port);

		if (bind(s, (NETADDR) sin, sizeof(struct sockaddr_in)) >= 0) {
			return(0);
		}

		else if ((errno != EADDRINUSE) && (errno != EADDRNOTAVAIL)) {
			return(ERROR);
		}
	}

	return(ERROR);
}

/*
 *	_make_kreq
 *
 *	Function:	- makes a kernel request
 *			- sends the request structure to the kernel
 *			- receives the reply structure from the kernel
 *	Accepts:	- ptr to kernel request structure
 *			- ptr to kernel reply structure
 *	Returns:	- 0 or ERROR
 *			- sets errno accordingly if a kernel error
 *			  is returned
 */

int
_make_kreq(req, reply)

struct kreq		*req;
struct kreply		*reply;

{
	int		lsize;		/* size of INET address */
	int		ret;		/* syscall return code */
	int		reply_code;	/* kernel reply code */
	fd_set 		readfds;	/* descriptor bit selector */
	struct timeval 	timeout;	/* time out for KATTACH */


	if (! ((_kio.ki_rtf & RTF_KERNEL) || (req->kq_request == KATTACH))) {
		errno = ENOTATTACHED;
		return(ERROR);
	}

	client_in.sin_port = (u_short) _kio.ki_kernel;
/*
 * Loop through possible Trollius signal interruptions.
 */
	do {
/*
 * Send the request to the kernel.
 */
		while (((ret = sendto(_kio.ki_portd, (caddr_t) req,
			sizeof(struct kreq), 0, (NETADDR) &client_in,
			sizeof(struct sockaddr_in))
			) < 0) && (errno == EINTR));

		if (ret < 0) {
			return(ret);
		}
/*
 * We will do an error check for the presence of the kernel iff the
 * request is KATTACH and the process is uncontrolled.
 */
		if ((req->kq_request == KATTACH) && !req->kqa_priority) {
			FD_ZERO(&readfds);
			FD_SET(_kio.ki_portd, &readfds);
			timeout.tv_sec = 5;
			timeout.tv_usec = 0;
/*
 * Wait for a reply or a timeout.
 */
			if ((ret = select(getdtablesize(), &readfds,
				(fd_set *) 0, (fd_set *) 0,
				&timeout)) < 0) {

				return(ret);
			}

			if (ret == 0) {
				errno = ENOKERNEL;
				return(ERROR);
			}
		}
/*
 * Receive the reply from the kernel;
 */
		while (((ret = recvfrom(_kio.ki_portd, (caddr_t) reply,
			sizeof(struct kreply), 0, (NETADDR) &client_in,
			&lsize)) < 0) && (errno == EINTR));

		if (ret < 0) {
			return(ret);
		}

		reply_code = reply->kr_reply;

		if (reply->kr_signal) {
			_kio.ki_signal |= reply->kr_signal;

			if (reply_code == EINTR) {
				_follow_signal();
			}
		}

	} while ((reply_code == EINTR) &&
		(reply->kr_signal & _kio.ki_sigretry));

	if (reply_code != 0) {
		errno = reply_code;
		return(ERROR);
	}

	return(0);
}

/*
 *	_recv_kmsg
 *
 *	Function:	- receives a message from another process
 *	Accepts:	- a pointer to a kernel message header
 *			- my chosen aux port
 *			- my chosen aux port descriptor
 *	Returns:	- an error indicator
 */

int
_recv_kmsg(msgptr, pick_aux, pick_auxd)

struct kmsg		*msgptr;	/* pointer to message header */
int			pick_aux;
int			pick_auxd;

{
	int		lsize;		/* size of sender's addr */
	int		peerrem;	/* sender's idea of k_length */
	int		remaining;	/* bytes left to receive */
	int		ret;		/* syscall return code */
	int		slicelen;	/* bytes per slice */
	char		minmsg;		/* minimum message buffer */
	char		*sliceptr;	/* one slice of full message */

	if ((!msgptr->k_length) && (!(msgptr->k_flags & KCLOBBER))) {
		return(0);
	}

	remaining = msgptr->k_length;
	peerrem = (msgptr->k_flags & KCLOBBER) ? remaining + _kio.ki_slush :
		remaining;
	sliceptr = msgptr->k_msg;
/* 
 * Receive the peer's packetized messages.  Send a synchronization
 * message prior to each packet, inviting the peer to send a packet.
 */
	do {
		client_in.sin_port = (u_short) _kio.ki_peer_sync;
		slicelen = (remaining > MAXKMSGLEN) ? MAXKMSGLEN : remaining;
/*
 * Synchronize.
 */
		while (((ret = sendto(pick_auxd, &minmsg,
			sizeof(minmsg), 0, (NETADDR) &client_in,
			sizeof(struct sockaddr_in))) < 0) && (errno == EINTR));

		if (ret < 0) return(ret);
/*
 * Receive packet.
 */
		if (slicelen > 0) {

			while (((ret = recvfrom(_kio.ki_portd, sliceptr,
				slicelen, 0, (NETADDR) &client_in, &lsize)) < 0)
				&& (errno == EINTR));
		} else {

			while (((ret = recvfrom(_kio.ki_portd, &minmsg,
				sizeof(minmsg), 0, (NETADDR) &client_in,
				&lsize)) < 0) && (errno == EINTR));
		}

		if (ret < 0) return(ret);

		sliceptr += slicelen;
		remaining -= slicelen;
		peerrem -= MAXKMSGLEN;
	} while ((remaining > 0) || (peerrem > 0));

	return(0);
}

/*
 *	_send_kmsg
 *
 *	Function:	- sends a message to another process
 *	Accepts:	- a pointer to a kernel message header
 *			- my chosen aux port
 *			- my chosen aux port descriptor
 *	Returns:	- an error indicator
 */

int
_send_kmsg(msgptr, pick_aux, pick_auxd)

struct kmsg		*msgptr;	/* pointer to message header */
int			pick_aux;
int			pick_auxd;

{
	int		lsize;		/* size of sender's addr */
	int		remaining;	/* bytes left to transmit */
	int		ret;		/* syscall return code */
	int		slicelen;	/* bytes per slice */
	char		minmsg;		/* minimum message buffer */
	char		*sliceptr;	/* one slice of full message */

	if (! msgptr->k_length) {
		return(0);
	}

	remaining = msgptr->k_length;
	sliceptr = msgptr->k_msg;
/*   
 * Packetize and send the message to the peer.  Recieve a synchronization
 * message prior to each packet, indicating the peer can recieve a packet.
 * This is required because there is no limit to the kernel message length.
 */
	do {
		slicelen = (remaining > MAXKMSGLEN) ? MAXKMSGLEN : remaining;
/*
 * Synchronize.
 */
		while (((ret = recvfrom(pick_auxd, &minmsg, sizeof(minmsg), 0,
			(NETADDR) &client_in, &lsize)) < 0)
			&& (errno == EINTR));

		if (ret < 0) return(ret);
/*
 * Send packet.
 */
		client_in.sin_port = (u_short) _kio.ki_peer_udpdata;

		while (((ret = sendto(_kio.ki_portd, sliceptr,
			slicelen, 0, (NETADDR) &client_in,
			sizeof(struct sockaddr_in))) < 0) && (errno == EINTR));

		if (ret < 0) return(ret);

		sliceptr += slicelen;
		remaining -= slicelen;
	} while (remaining > 0);

	return(0);
}

/*
 *	_kcleanup
 *
 *	Function:	- just a stub for UDP
 */

int
_kcleanup()

{
	return(0);
}

/*
 *	_kbail
 *
 *	Function:	- terminates the process
 *	Accepts:	- exit status
 */

void
_kbail(status)

int			status;

{
	exit(status);
}
