/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	dlo_dev.c,v $
 * Revision 2.2  92/04/30  12:18:19  trillium
 * Ohio Release
 * 
 *	Function:	- Receive network frames on a fixed event and write
 *			  them to a device.
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University
 *			  and Cornell Research Foundation
 *
 *	Accepts:	- device file descriptor open for write
 *			- dlo event
 */

#include <dlreq.h>
#include <stdio.h>
#include <net.h>
#include <debug.h>
#include <kio.h>
#include <portable.h>
#include <t_types.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */
extern struct kio_t	_kio;

/*
 * local variables
 */
static int		filter;		/* message filter */

/*
 * local functions
 */
static void		valve();	/* positive event valve */

void
dlo_dev(fd, event)

int			fd;
int			event;

{
	int 		length;		/* total frame length */
	int		nbytes;		/* return value from write */
	int		i;		/* favourite counter variable */
	int		holding;	/* Are we holding a VC? */
	int		data_conv;	/* data conversion flags */
	int		mask;		/* signal mask */
	int		*pint;		/* integer pointer */
	float		*pflt;		/* float pointer */
	double		*pdbl;		/* double pointer */
	int		ihold;		/* index of our dl event in VC tbl */
	int		r;		/* drecv() return value */
	char		ack;		/* acknowledgement byte */
	struct netframe	outframe;	/* network frame (header and data) */

/*
 * Initialize and attach to the kernel.
 */
	if (kinit(0)) {
		terror("dlo_dev");
		exit(errno);
	}
/*
 * Connect to dld.
 */
	if (ldlattach(event) < 0) {
		panic(errno);
	}
/*
 * Catch SIGA and SIGB.
 */
	if ((ksignal(SIGA, valve) == 0) ||
		(ksignal(SIGB, valve) == 0)) {

		panic(errno);
	}

	mask = ksigretry(0);
	ksigsetretry(mask & ~(sigmask(SIGA) | sigmask(SIGB)));
	filter = DLALL;

	ihold = (event & HOLDBITS) + ((event & INT4_SIGN) ? (HOLDBITS + 1) : 0);
/*
 * Process network frames forever.
 */
	for(;;) {	
/*
 * Receive network frame.
 */
		outframe.nf_header.nh_event = event;
		outframe.nf_header.nh_type = filter;
		outframe.nf_header.nh_flags = DFORWARD;
		outframe.nf_header.nh_length = MAXNMSGLEN;
		outframe.nf_header.nh_msg = outframe.nf_data;

		while (((r = drecv(&outframe.nf_header)) < 0) &&
				(errno == EINTR)) {
			outframe.nf_header.nh_type = filter;
		}

		if (r < 0) {
			terror("dlo_dev");
			kexit(errno);
		}

	db(("dlo_dev: outgoing frame, node=%x event=%d type=%x length=%d\n",
			outframe.nf_header.nh_node,
			outframe.nf_header.nh_event,
			outframe.nf_header.nh_type,
			outframe.nf_header.nh_length));
/*
 * Pull a very sneaky manouver to maintain a virtual circuit if
 * so indicated by the header.
 */
		if (outframe.nf_header.nh_flags & KHOLD) {
			_kio.ki_khold[ihold].kh_event = event;
		}

		length = sizeof(struct nmsg) + outframe.nf_header.nh_length;
/*
 * Convert nh_data to Trillium byte ordering.
 */
		data_conv = outframe.nf_header.nh_flags & DMSKDATA;

		if (!data_conv || (data_conv == DINTDATA)) {
		    pint = outframe.nf_header.nh_data;

		    for (i = 0; i < NHDSIZE; ++i) {
			*pint = ltot(*pint);
			++pint;
	            }
		}

		else if (data_conv == DFLTDATA) {
		    pflt = (float *) outframe.nf_header.nh_data;

		    for (i = 0; i < ((NHDSIZE * sizeof(int)) / sizeof(float));
			++i) {

			*pflt = ltotf(*pflt);
			++pflt;
		    }
		}

		else if (data_conv == DDBLDATA) {
		    pdbl = (double *) outframe.nf_header.nh_data;

		    for (i = 0; i < ((NHDSIZE * sizeof(int)) / sizeof(double));
			++i) {

			*pdbl = ltotd(*pdbl);
			++pdbl;
		    }
		}
/*
 * Convert nh_msg to the local data representation.
 */
		data_conv = outframe.nf_header.nh_flags & DMSKMSG;

		if (data_conv == DINTMSG) {
		    pint = (int *) outframe.nf_header.nh_msg;

		    for (i = 0; i < (outframe.nf_header.nh_length /
			sizeof(int)); ++i) {

			*pint = ltot(*pint);
			++pint;
	            }
		}

		else if (data_conv == DFLTMSG) {
		    pflt = (float *) outframe.nf_header.nh_msg;

		    for (i = 0; i < (outframe.nf_header.nh_length /
			sizeof(float)); ++i) {

			*pflt = ltotf(*pflt);
			++pflt;
	            }
		}

		else if (data_conv == DDBLMSG) {
		    pdbl = (double *) outframe.nf_header.nh_msg;

		    for (i = 0; i < (outframe.nf_header.nh_length /
			sizeof(double)); ++i) {

			*pdbl = ltotd(*pdbl);
			++pdbl;
		    }
		}
/*
 * Convert header (sans nh_data and nh_msg) to Trillium byte ordering.
 */
		holding = outframe.nf_header.nh_flags & KHOLD;
		pint = (int *) &outframe.nf_header;

		for (i = 0; i < ((sizeof(struct nmsg) / sizeof(int)) -
		    (NHDSIZE + 1)); ++i) {

		    *pint = ltot(*pint);
		    ++pint;
		}
/*
 * Write the entire frame to the device.
 */
		nbytes = write(fd, (char *) &outframe, length);

		if (nbytes < 0) {
			terror("dlo_dev");
			kexit(errno);
		}

		else if (nbytes < length) {
			fprintf(stderr, "dlo_dev: write(fd, b, n) < n\n");
			kexit(errno);
		}
/*
 * Write acknowledgement byte to the device if we are not continuing
 * a virtual circuit.
 * This is needed for compatibility with dl_t4.
 */
/*
 *		if (!holding) {
 *
 *			if (write(fd, &ack, sizeof(ack)) == ERROR) {
 *				kexit(errno);
 *			}
 *		}
 */
	}   /* for */
}

/*
 *	valve
 *
 *	Function:	- turns positive event value on or off by
 *			  altering the message type
 *	Accepts:	- signal number
 */

static void
valve(sig)

int4			sig;

{
	filter = (sig == SIGA) ? DLALL : DLSYS;
}
