/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	tfish.c,v $
 * Revision 2.2  92/04/30  12:23:23  trillium
 * Ohio Release
 * 
 *	Function:	- "tfish"
 *			- Archipel Volvox version
 *			- reset ITB nodes associated with an OTB-ITB link
 *			- issues device driver reset control and
 *			  executes Voltop netreset command
 *			- so named from our original reset device, fish line.
 */

#include <fcntl.h>
#include <machine/machine.h>
#include <machine/volvox.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

/*
 * external functions
 */
extern int		do_args();
extern void		followed();
extern void		intparam();
extern void		validopts();

/*
 * local functions
 */
static void		help();

/*
 * local variables
 */
static char		*usage = "usage: tfish [-hr] [-d <delay>]  <device>\n";

/*
 * external variables
 */
extern int		errno;

main(argc, argv)

int			argc;
char			**argv;

{
	int		delay;		/* #secs to delay after reset */
	int		fd;		/* device file descriptor */
	int		reset;		/* the reset mask */
	int		pid;		/* child pid */
	char		*dreset;	/* device file name */
	union wait	status;		/* status after forking */
/*
 * Initialize option parser.
 */
	validopts("dhr");
	followed("d");

	if (do_args(&argc, argv) < 0) {
		(void) fprintf(stderr, usage);
		exit(errno);
	}

	if (opt_taken('h')) {
		help();
		exit(0);
	}

	if (argc > 2) {
		(void) fprintf(stderr, usage);
		exit(1);
	}

	if (argc > 1) {
		dreset = argv[1];
	} else {
		dreset = DEFFDEVICE;
	}
/*
 * Execute netreset command.
 */
	if (!opt_taken('r')) {
/*
 * Direct stdout to /dev/null to make netreset silent.
 */
		fd = open("/dev/null", O_WRONLY);

		if (fd < 0) {
			perror("tfish");
			exit(errno);
		}

		if (dup2(fd, 1) < 0) {
			perror("tfish");
			exit(errno);
		}
/*
 * Execute Voltop reset since the above only resets the device connected
 * node.
 */
		if ((pid = fork()) < 0) {
			perror("tfish");
			exit(errno);
		}
/*
 * child
 */
		else if (pid == 0) {
			execlp("netreset", "netreset", "-d", dreset,
				(char *) 0);
			exit(1);
		}
/*
 * parent
 */
		else {
			while ((wait(&status) != 0) && (! WIFEXITED(status)));

			if (status.w_retcode != 0) {
				(void) fprintf(stderr,
					"tfish: netreset exit status %d\n",
					(int) status.w_retcode);
				exit((int) status.w_retcode);
			}
		}
	}

	if ((fd = open(dreset, O_RDWR)) < 0) {
		(void) fprintf(stderr,
			"tfish: unable to reset %s (may be busy)\n", dreset);
		perror("tfish");
		exit(1);
	}

	reset = 0x10;

	if (ioctl(fd, VXV_RESET, &reset) < 0) {
		(void) fprintf(stderr,
			"tfish: unable to reset %s (may be busy)\n", dreset);
		exit(1);
	}

	if (close(fd) < 0) {
		perror("tfish");
		exit(errno);
	}
/*
 * Delay.
 */
	if (opt_taken('d')) {
		intparam('d', &delay);
		sleep((unsigned) delay);
	}
/*
 * Delete the command file.
 */
	exit(0);
}

/*
 *	help
 *
 *	Function:	- prints useful information on the tfish command
 */

static void
help()

{
	printf("\nSynopsis:	tfish [-h] [-d <delay>] [<filename>]\n");
	printf("\nDescription:	Reset itb node(s).\n");
	printf("\nOptions:	-h		this help message\n");
	printf("\t\t-d <delay>	pause after reseting\n");
	printf("\t\t-r		skip netreset\n");
	printf("\t\t<filename>	specific reset device\n");
}
