/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD/RKM
 *
 *	$Log:	solder.h,v $
 * Revision 2.2  92/04/30  12:57:55  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/27  10:15:36  gdburns
 * overhaul of machine independent code
 * 
 * Revision 2.1.1.1  91/04/01  18:19:11  raja
 * Upgraded to handle multiple device names.
 * 
 * Revision 2.1  91/03/11  14:42:25  gdburns
 * Ohio Release
 *
 *	Function:	- constants for the solder tool
 */

#ifndef _SOLDER
#define _SOLDER

#define	CBAD -1

/*
 * default options and devices
 */
#define DEFSOPT			"-cs"
#define	DEFDEVNAME		"/dev/bxiv0"

/*
 * Describe a board. There exist three main entities on a board, slots,
 * edges and connectors.
 */

/*
 * A slot in a board has some number of communication links. These links 
 * are connected to a link of a switch.
 */

struct slot_connections {
	int sl_switch;				/* which switch */
	int sl_sw_link;				/* which link */
} ;

/*
 * There are different kinds of edges ( by virtue of their connections ).
 * An edge is connected to a switch and to a connector.
 *	
 */

struct edge_connections {
	int ed_switch;				/* which switch */
	int ed_sw_link;				/* which link */
	int ed_connector;			/* which d_connector */
} ;

/*
 * Describe Inmos boards still at conceptual level. 
 * Herewith  im == Inmos.
 */

/*
 * B014 specific constants.
 */

#define	NUMBEROFSLOTS		8		/* slot/tram population */
#define NUMBEROFSLOTLINKS	4		/* tram links */
#define NUMBEROFEDGES		27		/* edge connector population */
#define NUMBEROFCONNECTORS	3		/* d_connector population */
#define NUMBEROFSWITCHES	2		/* switch population */
#define	NUMBEROFSWLINKS		32		/* links on a switch */
#define NUMBEROFBOARDS		1		/* single or multiple */
#define NUMBEROFDPINS		37		/* pins on d_connector */

typedef struct slot_connections 
			im_slots[NUMBEROFSLOTS][NUMBEROFSLOTLINKS] ;

typedef struct edge_connections im_edges[NUMBEROFEDGES] ;

typedef	char im_linkswitch[NUMBEROFSWITCHES][NUMBEROFSWLINKS] ;


/*
 * Hard coded slot and edge connections.
 */

im_slots		slots_b014 =  {
	{ { CBAD,CBAD }, {CBAD, CBAD }, { 1, 1  }, { 0, 1  } },	/* slot 0 */
	{ {  0,   2   }, { 1,    2   }, { 1, 3  }, { 0, 3  } },	/* slot 1 */
	{ {  0,   4   }, { 1,    4   }, { 1, 5  }, { 0, 5  } },	/* slot 2 */
	{ {  0,   6   }, { 1,    6   }, { 1, 7  }, { 0, 7  } },	/* slot 3 */
	{ {  0,   8   }, { 1,    8   }, { 1, 9  }, { 0, 9  } },	/* slot 4 */
	{ {  0,   10  }, { 1,    10  }, { 1, 11 }, { 0, 11 } },	/* slot 5 */
	{ {  0,   12  }, { 1,    12  }, { 1, 13 }, { 0, 13 } },	/* slot 6 */
	{ {  0,   14  }, { 1,    14  }, { 1, 15 }, { 0, 15 } }	/* slot 7 */
};

/*
 * There are four kinds of edges namely:
 *	[ P2 = back connector; P4, P5 = front connectors ]
 *
 *		edges 0-7 	source P2 		sink C004 0.
 *		edges 7-15 	source C004 1 		sink P4.
 *		edges 16-23 	source C004 1		sink P5.
 *		edges 24-26	source C004 0/1		sink none.
 */

im_edges 		edges_b014 = {
	{ 0, 16, 2 }, { 0, 17, 2 }, { 0, 18, 2 },	/* edge 0-2   */
	{ 0, 19, 2 }, { 0, 20, 2 }, { 0, 21, 2 }, 	/* edge 3-5   */
	{ 0, 22, 2 }, { 0, 23, 2 },			/* edge 6-7   */
	{ 1, 16, 4 }, { 1, 17, 4 }, { 1, 18, 4 }, 	/* edge 8-10  */
	{ 1, 19, 4 }, { 1, 20, 4 }, { 1, 21, 4 }, 	/* edge 11-13 */
	{ 1, 22, 4 }, { 1, 23, 4 },			/* edge 14-15 */
	{ 1, 24, 5 }, { 1, 25, 5 }, { 1, 26, 5 },	/* edge 16-18 */
	{ 1, 27, 5 }, { 1, 28, 5 }, { 1, 29, 5 }, 	/* edge 19-21 */
	{ 1, 30, 5 }, { 1, 31, 5 },			/* edge 22-23 */
	{ 0,  0, 0 }, { 0, 31, 0 }, { 1, 0, 0 }		/* edge 24-26 */
};

/*
 * D-connectors. Not used yet.
 */

typedef	char	im_d_connector[NUMBEROFDPINS];

/*
 * Struct to carry around information abou a physical node.
 */

struct	im_hwnodeinfo {
	int	hw_board;
	int	hw_slot;
};

#define	NULLHW	(struct im_hwnodeinfo *) 0

/*
 * Put it all together.
 */

struct im_board{
	struct	slot_connections	*bd_slots ;		/* slot ptr   */
	struct 	edge_connections 	*bd_edges ;		/* edge ptr   */
	im_linkswitch			 bd_switch ;		/* c004 array */
	im_d_connector			 d_p2, d_p4, d_p5;	/* dconnector */
};

#endif
