/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	tfish.c,v $
 * Revision 2.2  92/04/30  12:59:46  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/30  12:59:27  gdburns
 * Change name to tfish.
 * 
 * Revision 2.1.1.1  91/04/15  14:17:33  raja
 * Exit with a status code 0 upon successful termination.
 * 
 * Revision 2.1  91/03/11  14:43:43  gdburns
 * Ohio Release
 *
 *	Function:	- "tfish"
 *			- INMOS B014 version
 *			- resets ITB nodes associated with an OTB-ITB link
 *			- so named from our original reset device, fish line.
 */

#include <machine/machine.h>
#include <stdio.h>

/*
 * external functions
 */
extern int		OpenLink();
extern int		ResetLink();
extern int		do_args();
extern void		followed();
extern void		intparam();
extern void		validopts();

/*
 * local functions
 */
static void		help();

/*
 * local variables
 */
static char		*me;		/* name I was exec'd by */
static char		*usage = "usage: %s <device>\n";

/*
 * external variables
 */
extern int		errno;


main(argc, argv)

int			argc;
char			**argv;

{
	int		delay;		/* #secs to delay after reset */
	int		fd;		/* device file descriptor */
	char		*dreset;	/* device file name */

	me = argv[0];
/*
 * Initialize option parser.
 */
	validopts("dh");
	followed("d");

	if (do_args(&argc, argv) < 0) {
		(void) fprintf(stderr, usage, me);
		return(errno);
	}

	if (opt_taken('h')) {
		help();
		return(0);
	}

	if (argc > 2) {
		(void) fprintf(stderr, usage, me);
		return(1);
	}

	if (argc > 1) {
		dreset = argv[1];
	} else {
		dreset = DEFFDEVICE;
	}
/*
 * Open the device.
 */
	if ((fd = OpenLink(dreset)) < 0) {
		(void) fprintf(stderr,
			"%s: unable to open %s (may be busy)\n", me, dreset);
		return(1);
	}

	if (ResetLink(fd) < 0) {
		(void) fprintf(stderr, "%s: unable to reset %s\n", me, dreset);
		return(1);
	}
/*
 * Delay.
 */
	if (opt_taken('d')) {
		intparam('d', &delay);
		sleep((unsigned) delay);
	}

	return(0);
}

/*
 *	help
 *
 *	Function:	- prints useful information on the tfish command
 */

static void
help()

{
	printf("\nSynopsis:	tfish [-h] [-d <delay>] [<filename>]\n");
	printf("\nDescription:	Reset ITB node(s).\n");
	printf("\nOptions:	-h		this help message\n");
	printf("\t\t-d <delay>	pause after reseting\n");
	printf("\t\t<filename>	specific reset device\n");
}
