/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	dlo_meiko.c,v $
 * Revision 2.2  92/04/30  13:40:30  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  10:22:20  gdburns
 * buffer flow control enhancements
 * 
 * Revision 2.1  91/03/20  11:45:48  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/10/05  23:16:11  gdburns
 * upgraded to modern portable.h
 * 
 * Revision 2.0  90/10/04  16:40:15  gdburns
 * Meiko VCS
 * 
 *	Function:	- datalink output program for transputer link
 *			- Meiko VCS version
 */

#ifdef itb
#include <c.h>
#endif
#include <dlreq.h>
#include <net.h>
#include <kio.h>
#include <pcb.h>
#include <portable.h>
#include <shared.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;
extern struct kio_t	_kio;
extern struct pcb_t	*_pcb;

/*
 * local variables
 */
static char		*myname = "dlo_meiko";
static int		filter;		/* message type filter */

/*
 * local functions
 */
static void		valve();	/* positive event valve */

void
dlo_meiko(outlink)

int			outlink;	/* output channel address */

{
	int		ihold;		/* index of our dl event in VC tbl */
	struct nmsg	nhead;		/* network message descriptor */
	char		nmsg[MAXNMSGLEN];
	int		nw;		/* # of words (VCS protocol) */
	int		mask;		/* signal mask */

/*
 * Connect to the kernel.
 */
	if (kinit(0) < 0) {
		panic(errno);
	}
/*
 * Connect to dld.
 */
	if (ldlattach(outlink) < 0) {
		panic(errno);
	}
/*
 * Catch SIGA and SIGB.
 */
	if ((ksignal(SIGA, valve) == 0) ||
		(ksignal(SIGB, valve) == 0)) {

		panic(errno);
	}

	mask = ksigretry(0);
	ksigsetretry(mask & ~(sigmask(SIGA) | sigmask(SIGB)));
	filter = DLALL;
/*
 * Become a high priority process.
 */
	setpri(0);
/*
 * Connect to kenyad.
 */
	_pcb->pc_argv = &myname;
	_pcb->pc_argc = 1;

	if (rpattach(LOCAL) < 0) {
		panic(errno);
	}

	ihold = (outlink & HOLDBITS) + ((outlink & INT4_SIGN)
		? (HOLDBITS + 1) : 0);

	for (;;) {
/*
 * Receive network frame.
 */
		nhead.nh_event = outlink;
		nhead.nh_type = filter;
		nhead.nh_flags = DFORWARD;
		nhead.nh_length = MAXNMSGLEN;
		nhead.nh_msg = nmsg;

		while (drecv(&nhead)) {
			nhead.nh_type = filter;
		}
/*
 * Pull a very sneaky manouver to maintain a virtual circuit if
 * so indicated by the header.
 */
		if (nhead.nh_flags & KHOLD) {
			_kio.ki_khold[ihold].kh_event = outlink;
		}

		nw = sizeof(struct nmsg) / sizeof(int);
		out(outlink, (char *) &nw, sizeof(nw));
		out(outlink, (char *) &nhead, sizeof(struct nmsg));

		if (nhead.nh_length > 0) {
			nw = (nhead.nh_length + (sizeof(int) - 1))
				/ sizeof(int);
			out(outlink, (char *) &nw, sizeof(nw));
			out(outlink, nmsg, nw * sizeof(int));
		}
	}
}

/*
 *	valve
 *
 *	Function:	- turns positive event value on or off by
 *			  altering the message type
 *	Accepts:	- signal number
 */

static void
valve(sig)

int4			sig;

{
	filter = (sig == SIGA) ? DLALL : DLSYS;
}
