/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	dli_dev.c,v $
 * Revision 2.2  92/04/30  13:44:19  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  10:24:32  gdburns
 * buffer flow control enhancements
 * 
 * Revision 2.1  91/03/20  11:45:13  gdburns
 * Ohio Release
 * 
 * Revision 2.0  90/10/04  16:37:43  gdburns
 * Meiko VCS
 * 
 *	Function:	- Read network frames from a device and forward them
 *			  using nsend().
 *			- Meiko VCS version
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and
 *			  Cornell Research Foundation
 *
 *	Accepts:	- device file descriptor open for read
 */

#include <stdio.h>
#include <net.h>
#include <debug.h>
#include <t_types.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */

/*
 * external functions
 */
extern int		svcs_recv();	/* read from the device */

dli_dev(fd)

int			fd;

{
	int		i;		/* favourite counter variable */
	int		data_conv;	/* data conversion flags */
	int		len;		/* incoming length in words */
	int		nbytes;		/* total # of bytes read */
	int		*pint;		/* integer pointer */
	float		*pflt;		/* float pointer */
	double		*pdbl;		/* double pointer */
	struct nmsg	nhead;		/* network message descriptor */
	char		nmsg[MAXNMSGLEN];

/*
 * Initialize and attach to the kernel.
 */
	if (kinit(0) < 0) {
		terror("dli_dev");
		exit(errno);
	}
/*
 * Process network frames forever.
 */
	for(;;) {	
/*
 * Read network frame header from device.
 */
	    nbytes = svcs_recv(fd, &len, (char *) &nhead);

	    if (nbytes < 0) {
		terror("dli_dev");
		kexit(errno);
	    }
/*
 * Convert header (sans nh_data and nh_msg) to local byte ordering.
 */
	    pint = (int *) &nhead;

	    for (i = 0; i < ((sizeof(struct nmsg) / sizeof(int)) -
		(NHDSIZE + 1)); ++i) {

		*pint = ttol(*pint);
		++pint;
	    }

  db(("dli_dev: incoming frame: node=%x event=%d type=%x flags=%x length=%d\n",
		nhead.nh_node,
		nhead.nh_event,
		nhead.nh_type,
		nhead.nh_flags,
		nhead.nh_length));
/*
 * Convert nh_data to the local data representation.
 */
	    data_conv = nhead.nh_flags & DMSKDATA;

	    if (!data_conv || (data_conv == DINTDATA)) {
		pint = nhead.nh_data;

		for (i = 0; i < NHDSIZE; ++i) {
		    *pint = ttol(*pint);
		    ++pint;
	        }
	    }

	    else if (data_conv == DFLTDATA) {
		pflt = (float *) nhead.nh_data;

		for (i = 0; i < ((NHDSIZE * sizeof(int)) / sizeof(float));
		    ++i) {

		    *pflt = ttolf(*pflt);
		    ++pflt;
	        }
	    }

	    else if (data_conv == DDBLDATA) {
		pdbl = (double *) nhead.nh_data;

		for (i = 0; i < ((NHDSIZE * sizeof(int)) / sizeof(double));
		    ++i) {

		    *pdbl = ttold(*pdbl);
		    ++pdbl;
		}
	    }
/*
 * Read network frame data from device.
 */
	    if (nhead.nh_length > 0) {
		nbytes = svcs_recv(fd, &len, nmsg);

		if (nbytes < 0) {
		    terror("dli_dev");
		    kexit(errno);
		}
	    }

	    nhead.nh_msg = nmsg;
/*
 * Convert nh_msg to the local data representation.
 */
	    data_conv = nhead.nh_flags & DMSKMSG;

	    if (data_conv == DINTMSG) {
		pint = (int *) nhead.nh_msg;

		for (i = 0; i < (nhead.nh_length / sizeof(int));
		    ++i) {

		    *pint = ttol(*pint);
		    ++pint;
	        }
	    }

	    else if (data_conv == DFLTMSG) {
		pflt = (float *) nhead.nh_msg;

		for (i = 0; i < (nhead.nh_length / sizeof(float));
		    ++i) {

		    *pflt = ttolf(*pflt);
		    ++pflt;
	        }
	    }

	    else if (data_conv == DDBLMSG) {
		pdbl = (double *) nhead.nh_msg;

		for (i = 0; i < (nhead.nh_length / sizeof(double));
		    ++i) {

		    *pdbl = ttold(*pdbl);
		    ++pdbl;
		}
	    }
/*
 * Reset the BYBUF flag.
 */
	    nhead.nh_flags &= ~BYBUF;
	    nhead.nh_flags |= DDLI;
/*
 * Forward network message.
 */
	    if (nhead.nh_node == NOTNODEID) {
		nhead.nh_node = LOCAL;
	    }

	    if (nsend(&nhead)) {
		terror("dli_dev");
		kexit(errno);
	    }
	}   /* for */
}
