/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	dlo_dev.c,v $
 * Revision 2.2  92/04/30  13:44:21  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  10:24:34  gdburns
 * buffer flow control enhancements
 * 
 * Revision 2.1  91/03/20  11:45:16  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/10/05  23:14:31  gdburns
 * changed to modern portable.h
 * 
 * Revision 2.0  90/10/04  16:37:44  gdburns
 * Meiko VCS
 * 
 *	Function:	- Receive network frames on a fixed event and write
 *			  them to a device.
 *			- Meiko VCS version
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and
 *			  Cornell Research Foundation
 *
 *	Accepts:	- device file descriptor open for write
 *			- dlo event
 */

#include <stdio.h>
#include <net.h>
#include <dlreq.h>
#include <debug.h>
#include <kio.h>
#include <ksignal.h>
#include <portable.h>
#include <t_types.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */
extern struct kio_t	_kio;

/*
 * external functions
 */
extern int		svcs_send();	/* write to device */

/*
 * local functions
 */
static void		valve();	/* positive event valve */

/*
 * local variables
 */
static int		filter;		/* message filter */

void
dlo_dev(fd, event)

int			fd;		/* output device */
int			event;		/* datalink event */

{
	int		nbytes;		/* return value from write */
	int		i;		/* favourite counter variable */
	int		ihold;		/* index of our dl event in VC tbl */
	int		data_conv;	/* data conversion flags */
	int		mask;		/* signal retry mask */
	int		mlen;		/* # words in msg body */
	int		*pint;		/* integer pointer */
	int		r;		/* favourite return value */
	float		*pflt;		/* float pointer */
	double		*pdbl;		/* double pointer */
	struct nmsg	nhead;		/* network msg descriptor */
	char		nmsg[MAXNMSGLEN];
	
/*
 * Initialize and attach to the kernel.
 */
	if (kinit(0) < 0) {
		terror("dlo_dev");
		exit(errno);
	}
/*
 * Catch SIGA and SIGB.
 */
	if ((ksignal(SIGA, valve) == 0) ||
		(ksignal(SIGB, valve) == 0)) {

		terror("dlo_dev");
		kexit(errno);
	}

	mask = ksigretry(0);
	ksigsetretry(mask & ~(sigmask(SIGA) | sigmask(SIGB)));

	ihold = (event & HOLDBITS) + ((event & INT4_SIGN) ? (HOLDBITS + 1) : 0);
	filter = DLALL;
/*
 * Process network frames forever.
 */
	for(;;) {	
/*
 * Receive network frame.
 */
		nhead.nh_event = event;
		nhead.nh_type = filter;
		nhead.nh_flags = DFORWARD;
		nhead.nh_length = MAXNMSGLEN;
		nhead.nh_msg = nmsg;

		while (((r = drecv(&nhead)) < 0) && (errno == EINTR)) {
			nhead.nh_type = filter;
		}

		if (r < 0) {
			terror("dlo_dev");
			kexit(errno);
		}

	db(("dlo_dev: outgoing frame, node=%x event=%d type=%x length=%d\n",
			nhead.nh_node,
			nhead.nh_event,
			nhead.nh_type,
			nhead.nh_length));
/*
 * Pull a very sneaky manouver to maintain a virtual circuit if
 * so indicated by the header.
 */
		if (nhead.nh_flags & KHOLD) {
			_kio.ki_khold[ihold].kh_event = event;
		}
/*
 * Convert nh_data to Trillium byte ordering.
 */
		data_conv = nhead.nh_flags & DMSKDATA;

		if (!data_conv || (data_conv == DINTDATA)) {
		    pint = nhead.nh_data;

		    for (i = 0; i < NHDSIZE; ++i) {
			*pint = ltot(*pint);
			++pint;
	            }
		}

		else if (data_conv == DFLTDATA) {
		    pflt = (float *) nhead.nh_data;

		    for (i = 0; i < ((NHDSIZE * sizeof(int)) / sizeof(float));
			++i) {

			*pflt = ltotf(*pflt);
			++pflt;
		    }
		}

		else if (data_conv == DDBLDATA) {
		    pdbl = (double *) nhead.nh_data;

		    for (i = 0; i < ((NHDSIZE * sizeof(int)) / sizeof(double));
			++i) {

			*pdbl = ltotd(*pdbl);
			++pdbl;
		    }
		}
/*
 * Convert nh_msg to the local data representation.
 */
		data_conv = nhead.nh_flags & DMSKMSG;

		if (data_conv == DINTMSG) {
		    pint = (int *) nhead.nh_msg;

		    for (i = 0; i < (nhead.nh_length /
			sizeof(int)); ++i) {

			*pint = ltot(*pint);
			++pint;
	            }
		}

		else if (data_conv == DFLTMSG) {
		    pflt = (float *) nhead.nh_msg;

		    for (i = 0; i < (nhead.nh_length /
			sizeof(float)); ++i) {

			*pflt = ltotf(*pflt);
			++pflt;
	            }
		}

		else if (data_conv == DDBLMSG) {
		    pdbl = (double *) nhead.nh_msg;

		    for (i = 0; i < (nhead.nh_length /
			sizeof(double)); ++i) {

			*pdbl = ltotd(*pdbl);
			++pdbl;
		    }
		}
/*
 * Calculate body length by rounding up to nearest word.
 */
		mlen = (nhead.nh_length + (sizeof(int) - 1)) / sizeof(int);
/*
 * Convert header (sans nh_data and nh_msg) to Trillium byte ordering.
 */
		pint = (int *) &nhead;

		for (i = 0; i < ((sizeof(struct nmsg) / sizeof(int)) -
		    (NHDSIZE + 1)); ++i) {

		    *pint = ltot(*pint);
		    ++pint;
		}
/*
 * Write header to the device.
 */
		nbytes = svcs_send(fd, sizeof(struct nmsg) / sizeof(int),
			(char *) &nhead);

		if (nbytes < 0) {
			terror("dlo_dev");
			kexit(errno);
		}
/*
 * Write msg to the device.
 */
		if (mlen > 0) {
			nbytes = svcs_send(fd, mlen, nmsg);

			if (nbytes < 0) {
				terror("dlo_dev");
				kexit(errno);
			}
		}
	}   /* for */
}

/*
 *	valve
 *
 *	Function:	- turns positive event value on or off by
 *			  altering the message type
 *	Accepts:	- signal number
 */

static void
valve(sig)

int4			sig;

{
	filter = (sig == SIGA) ? DLALL : DLSYS;
}
