/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	link.c,v $
 * Revision 2.2  92/04/30  13:45:34  trillium
 * Ohio Release
 * 
 *	Function:	- link device interface
 *			- Meiko VCS version
 */

#include <sys/types.h>
#include <sys/uio.h>
#include <machine/machine.h>
#include <t_types.h>

#define MAXPKSIZE	4096

/*
 * external functions
 */
extern int		read();
extern int		write();

/*
 * global functions
 */
int			CloseLink();	/* close device */
int			OpenLink();	/* open device */
int			ReadLink();	/* read from device */
int			WriteLink();	/* write to device */

/*
 *	OpenLink
 *
 *	Function:	- opens link device
 *	Accepts:	- link name
 *	Returns:	- handle for open link device
 */

int
OpenLink(name)

char			*name;

{
	return(VCSFD);
}

/*
 *	ReadLink
 *
 *	Function:	- reads data from link device
 *	Accepts:	- link handle
 *			- buffer address
 *			- buffer length
 *			- timeout value (not used)
 *	Returns:	- #bytes written or -1 if error
 */

int
ReadLink(fd, buf, nbytes, timeout)

int			fd;
char			*buf;
int			nbytes;
int			timeout;

{
	int		hdr;
	int		rc;
	struct iovec	iov[2];

	iov[0].iov_base = (caddr_t) &hdr;
	iov[0].iov_len = sizeof(hdr);
	iov[1].iov_base = (caddr_t) buf;
	iov[1].iov_len = nbytes;
    
	rc = readv(fd, iov, 2);

	return (rc > 4 ? rc - 4: rc);
}

/*
 *	WriteLink
 *
 *	Function:	- writes data to link device
 *			- uses negative length protocol
 *			- due to device driver limitations, must packetize
 *	Accepts:	- link handle
 *			- buffer address
 *			- buffer length
 *			- timeout value (not used)
 *	Returns:	- #bytes written or -1 if error
 */

int
WriteLink(fd, buf, nbytes, timeout)

int			fd;
char			*buf;
int			nbytes;
int			timeout;

{
	int		hdr;
	int		rc;
	int		rctot;
	int		pkbytes;
	struct iovec	iov[2];

	rctot = 0;

	while (nbytes > 0) {
		pkbytes = (nbytes > MAXPKSIZE) ? MAXPKSIZE : nbytes;
		hdr = ltot(-pkbytes);
		iov[0].iov_base = (caddr_t) &hdr;
		iov[0].iov_len = sizeof(hdr);
		iov[1].iov_base = (caddr_t) buf;
		iov[1].iov_len = pkbytes;
    
		rc = writev(fd, iov, 2);

		if (rc < 0) {
			return(rc);
		} else if (rc > sizeof(hdr)) {
			rc -= sizeof(hdr);
			rctot += rc;
			nbytes -= rc;
			buf += rc;
		} else {
			return(rctot);
		}
	}

	return(rctot);
}

/*
 *	CloseLink
 *
 *	Function:	- stub
 *	Accepts:	- link handle
 *	Returns:	- 0
 */

int
CloseLink(fd)

int			fd;

{
	return(0);
}
