/*
 *	OSU Research Computing					Trillium
 *
 *	(third party software provided under license)
 *
 *	$Log:	mzos.h,v $
 * Revision 2.2  92/04/30  13:46:19  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:46:49  gdburns
 * Ohio Release
 * 
 * Revision 2.0  90/10/04  16:53:29  gdburns
 * Meiko VCS
 * 
 *
 *	We're just putting our stamp on this for housekeeping reasons.
 */

#ifndef mzos_h
#define mzos_h 1
/*
 * This defines the interface to the management of MK200 resources,
 * e.g. claiming processors and links and specifying the wiring.
 * However, this interface may change: better to use the abstract functions
 * in svcs.h.
 */
#include "bsdnet.h"

#define MAXDATA (9*1024)
#define HDRSIZE (sizeof(long))

typedef struct Msg {
    long   m_nwords;		/* size of message in words (little-endian) */
    char   m_data[MAXDATA];     /* contents of message */
} Msg;
 
#define MAXPROCS 256

#define T414 0x01
#define T800 0x02
#define T425 0x04
#define T801 0x08

#define T4 (T414 | T425)
#define T8 (T800 | T801)

/*
 * Memory requirements are in units of 256Kb, so 12Mb = 48 units.
 * Is this worthwhile ?  Neater just to have a 32-bit word.  But anyway,
 * I hope to hide details of the protocol behind a procedural interface.
 */
#define MEMUNITBITS 18
#define MEMUNIT (1 << MEMUNITBITS)
#define MEGABYTES ((1024*1024) / MEMUNIT)

#define ANY 0xff

typedef struct ProcessorSpec {
    unsigned char p_types;     /* set of acceptable processor types */
    unsigned char p_memunits;  /* memory size needed in MEMUNITs */
    unsigned char p_boardtype; /* MKxxx board type code, or 0xff for any */
    unsigned char p_module;    /* module, or 0xff for any */
    unsigned char p_board;     /* board,  or 0xff for any */
    unsigned char p_proc;      /* proc,   or 0xff for any */
    unsigned char p_pad0;      /* pad to 32-bit word boundary */
    unsigned char p_pad1;
} ProcessorSpec;

typedef struct WireSpec {
    unsigned char w_proc0;     /* allows up to 256 processors per domain */
    unsigned char w_link0;
    unsigned char w_proc1;
    unsigned char w_link1;
} WireSpec;

typedef struct DomainSpec {
    int d_pid;                  /* process id of requesting process */
    int d_gid;			/* group id of requesting process */
    int d_lifetime;             /* a guess at domain lifetime in seconds */
    int d_priority;             /* [-10..+10], -10 is high: use 0 for now */
    int d_wait;                 /* wait for resources, or give up ? */
    int d_nspecs;
    int d_nwires;
    ProcessorSpec *d_spec;
    WireSpec *d_wire;
} DomainSpec;

#define PS_MAGIC   0xdead
#define PS_DOMAINS 1            /* get list of processes using resources */
#define PS_ELMAP   2            /* get state of each processor */

typedef struct PsSpec {
    int ps_magic;		/* magic to say this is new tps */
    int ps_cmd;                 /* commands for getting more info */
    int ps_base;
    int ps_lim;                
} PsSpec;

typedef struct Processor {
    unsigned char p_type;       /* processor type */
    unsigned char p_memunits;   /* memory size */
    short         p_svbusid;    /* supervisor bus id of processor */
} Processor;

typedef struct Resources {
    int  r_p1;
    int  r_p2;
    int  r_p3;
    int  r_rc;
    char r_linkname[16]; /* null-terminated name of link device */
    int  r_board;
    int  r_link;
    int  r_nprocs;
    Processor r_proc[MAXPROCS];
} Resources;

typedef struct PsResults {
    int psr_nprocs;		 /* number of processes - that's all for now */
    struct PsDomain {
	int psd_pid;		 /* unix pid of the mzrun running here */
	int psd_state;		 /* monitor state */
	struct PsResources {
	    /* subset of Resources - don't want too much stuff 'cos otherwise
	     * it might not fit in the reply message!
	     */
	    int psp_board;
	    int psp_link;
	    int psp_nprocs;
	    Processor psp_proc0;	/* only processor 0 */
	} psd_resources; /* resources being used */
    } psr_data[1]; /* extensible structure - cast from a Msg */
} PsResults;

#define PS_NULLPID 0
#define PS_SYSPID  1

typedef struct PsElement {
    short e_pid;
    char  e_virtual;
    char  e_module;
    char  e_board;
    char  e_procno;
    char  e_boardtype;
    char  e_proctype;
    int   e_memsize;
} PsElement;

typedef struct PsElmap {
    int pse_nels;                 /* total # of svbus elements in machine */
    int pse_base;                 /* svbus id of first element in pse_el */
    int pse_lim;                  /* svbus id of first element after pse_el */
    PsElement pse_el[1];
} PsElmap;

typedef struct Request {
    int req_type;		  /* type of request */
    int req_pid;                  /* process id of requesting process */
    int req_gid;		  /* group id of caller */
    struct sockaddr req_sockaddr; /* Internet address it is listening on */
    char *req_addr;		  /* for MsgChan */
    union {
	DomainSpec req_domain;
	PsSpec req_ps;
    } req_data;
} Request;

typedef struct Reply {
   int rep_rc;			 /* return status */
   int rep_type;		 /* echo request type */
   union {
       Resources rep_resources;
       PsResults rep_ps;
   } rep_data;
} Reply;

/*
 * request types
 */
#define REQ_NULL   0		/* not requesting anything */
#define REQ_DOMAIN 1		/* requesting domain allocation */
#define REQ_PS     2		/* requesting ps info */

#endif

