/*
 *	OSU Research Computing					Trillium
 *
 *	(third party software provided under license)
 *
 *	$Log:	svcs.h,v $
 * Revision 2.2  92/04/30  13:46:23  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:46:55  gdburns
 * Ohio Release
 * 
 * Revision 2.0  90/10/04  16:53:30  gdburns
 * Meiko VCS
 * 
 *
 *	We're just putting our stamp on this for housekeeping reasons.
 */

#ifndef svcs_h
#define svcs_h 1
/*
 * @(#)svcs.h	1.4 17:54:27 5/2/89 
 *
 * Interfaces to Sun Virtual Computing Surface (SVCS) library functions
 * for using processors on a Meiko In-Sun Computing Surface.
 *
 * Copyright Meiko Ltd 1988.
 */
#ifndef mzos_h
#  include "mzos.h"
#endif
#ifndef svcserrs_h
#  include "svcserrs.h"
#endif

/*
 * Functions for constructing a specification of the required set of
 * processors (a domain), and grabbing a suitable set.
 */

extern DomainSpec *svcs_newspec();

extern int  svcs_procspec(/* DomainSpec *d, int proctypes, int minmembytes */);

extern void svcs_wirespec(/* DomainSpec *d, int proc0, link0, proc1, link1 */);

extern void svcs_freespec(/* DomainSpec *d */);

/*
 * Function for looking at an open file descriptor for a .ex4, .ex8, or
 * .cpr file and deciding what resources it needs. 
 * If the file is a valid .exe, give the domain spec and the exe header,
 * and leave the file pointer after the header, else exe header is NULL 
 * (a .cpr does not contain a domain specification) and the file pointer
 * is reset to start of file.  The header should be deallocated with free().
 */

extern int  svcs_filespec(/* int fd, DomainSpec **d, int **header */);

/*
 * Function for grabbing a set of processors by negotiating with the
 * daemon.  Hands back open vcsfd, and supfd and description of domain if
 * you want (pass in NULL if not interested).
 */

extern int  svcs_grab(/* DomainSpec *d, int *vcsfd, *supfd, Resources *r */);

/*
 * Functions for booting the domain with an Occam program (either in .cpr
 * format as produced by the Occam Programming System, or as raw boot data
 * extracted from a .cpr file by cprtoaout and built into the image), 
 * or a single-processor C or Fortran .ex8 image (linked with OPS-style 
 * libraries).  A .ex8 will be booted inside an appropriate harness,
 * which provides filesystem access and command line args through the
 * standard C or Fortran library.
 */
 
typedef struct BuiltIn {
    int  b_nbytes;  /* number of bytes of data following */
    char b_data[4]; /* the data - really b_data[b_nbytes] */
} BuiltIn;
 
extern int svcs_bootbuiltin(/* int vcsfd, BuiltIn *b */);

extern int svcs_bootexe(/* int vcsfd, int *header, int exefd,
                           char *cmdline, Resources *r */);

extern int svcs_bootcpr(/* int vcsfd, int cprfd */);

/*
 * Low-level functions for sending and receiving messages to the domain.
 * In each case nwords is the number of 32-bit words of following data
 * to send and receive.  This should be less than 9Kbytes (e.g. some header
 * information, and up to 8Kbytes of data).
 * 
 * These calls should be matched by processor 0 in the domain thus:
 *
 *    svcs_send(fd, nw, buf);   ==>  linksIn[0] ? nw; [buf FROM 0 FOR nw]
 *    
 *    svcs_recv(fd, &nw, buf);  <==  linksOut[0] ! nw; [buf FROM 0 FOR nw]
 */

extern int svcs_send(/* int vcsfd, int nwords, char *data */);

extern int svcs_recv(/* int vcsfd, int *nwords, char *data */);

/*
 * Equivalent functions for sending and receiving messages down Unix
 * pipes, in particular the supervisor connection to the mzmaster daemon.
 */
 
extern int svcs_pipesend(/* int pipefd, int nwords, char *data */);

extern int svcs_piperecv(/* int pipefd, int *nwords, char *data */);

/*
 * Silicon Graphics Irix does not support select() on user devices,
 * so you should call svcs_select() instead, which uses ioctl() calls
 * on any mz or mp file descriptors to simulate a full select().
 */
 
extern int svcs_select(/* int nfds, int *rx, *tx, *ex, struct timeval *tout */);

/*
 * Functions for converting data from Meiko to Host format (hence mtoh...)
 * and vice versa (htom...).
 * For 68020 or SPARC host and Meiko MK200 transputer board, this involves
 * bytesex conversion, hence htomxxx == mtohxxx.
 * We assume short is 16 bits, long 32 bits, float 32 bits, double 64 bits.
 * All these functions work either with dst == src, or with
 * non-overlapping dst and src arrays.  All data should be appropriately
 * aligned for the host machine before and after conversion.
 */

extern void svcs_mtohshorts(/* short *dst, short *src, int nshorts */);

extern void svcs_mtohlongs(/* long *dst, long *src, int nlongs */);

extern void svcs_mtohfloats(/* float *dst, float *src, int nfloats */);

extern void svcs_mtohdoubles(/* double *dst, double *src, int ndoubles */);

#define svcs_htomshorts  svcs_mtohshorts
#define svcs_htomlongs   svcs_mtohlongs
#define svcs_htomfloats  svcs_mtohfloats
#define svcs_htomdoubles svcs_mtohdoubles

/*
 * Useful typedefs for defining interfaces to filesys and other server code.
 * Typically a server will have an xx_init() proc which takes at least an
 * fd and an InstallerProc, and will call installerproc(fd, xx_handler) to
 * say that it wants to handle input on fd.  Then xx_handler will call the
 * recvproc passed in to xx_init() ... This structure is needed to cope
 * with the fact that the main select() loop (if any) may be in user code
 * or in the SunView notifier.
 */

typedef int SendProc(/* int fd, int nwords, char *data */);

typedef int RecvProc(/* int fd, int *nwords, char *data */); 
 
typedef void InputReadyProc(/* int *opaque, int fd */);

typedef void InstallerProc(/* int *opaque, InputReadyProc *handler, int fd */);

#endif
