/*
 *	OSU Research Computing					Trillium
 *
 *	Copyright 1990 The Ohio State University
 *
 *	AKP/GDB
 *
 *	$Log:	moses.lstt.s,v $
 * Revision 2.2  92/04/30  13:47:11  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:54:10  gdburns
 * Ohio Release
 * 
 * Revision 2.0  90/10/04  17:28:39  gdburns
 * Meiko VCS
 * 
 *
 *	Function:	- generic transputer implementation of the moses
 *			- It is installed as "moses.t4" for historical reasons.
 *			- This assembly language grammar is tailored to LSTT.
 *			- Meiko VCS version
 *			- based on Trollius 2.0 moses, Copyright 1990
 *			  The Ohio State University
 *			  & Cornell Research Foundation
 */

#include <transputer/transputer.h>
#include <moses.h>

/*
 * This is the way the stack looks after the initial register save call.
 */
#define retaddr		iReg		/* prescribed by call */
#define oldiReg		aReg		/* prescribed by call */
#define oldwReg		bReg		/* prescribed by call */
#define bootlink	cReg		/* prescribed by call */
#define cmnd 		4		/* moses command code */
#define arg0 		5		/* generic argument 0 */
#define arg1 		6		/* generic argument 1 */
#define arg2 		7		/* generic argument 2 */
#define lnkI 		8		/* moses input link */
#define lnkO 		9		/* moses output link */
#define minint		10		/* less instr than mint */
#define scratch		11		/* scratch pad temporary */

#define stksize		16		/* maximum stack depth */
#define timer0		36		/* #bytes past minint */
#define timer1		40		/* #bytes past minint */
#define iodiff		-16		/* #bytes between in and out links */


	.t800
	.pub	_start

_start:
	ajw	stksize
	call	@regsave
regsave:
	mint
	stl	minint
/*
 * Clear out the process queue.
 */
	ldl	minint
	sthf
	ldl	minint
	stlf
/*
 * Reset timers.
 */
	ldl	minint
	ldl	minint
	adc	timer0
	stnl	0
	ldl	minint
	ldl	minint
	adc	timer1
	stnl	0
	ldc	0
	sttimer
/*
 * Clear flags.
 */
	clrhalterr
	testerr
/*
 * Set input and output links.
 */
	ldl	bootlink
	stl	lnkI
	ldl	lnkI
	adc	iodiff
	stl	lnkO

	j	@forever
/*
 * Get one word from link in aReg and store it at address in bReg.
 */
getwrd:
	ldl	bReg
	ldl	aReg
	ldc	4
	in
	ret
/*
 * Get two words from link in aReg.  Store first word at address in cReg
 * and second word in address in bReg.
 */
get2wrds:
	ldl	cReg
	ldl	aReg
	call	@getwrd
	ldl	bReg
	ldl	aReg
	call	@getwrd
	ret
/*
 * M_IN
 *
 * Perform an in instruction.
 * First argument is buffer pointer.
 * Second argument is channel address.
 * Third argument is buffer length.
 */
_in:
	ldl	cmnd
	eqc	M_IN
	cj	@forever

	ldlp	arg0
	ldlp	arg1
	ldl	lnkI
	call	@get2wrds
	ldlp	arg2
	ldl	lnkI
	call	@getwrd
_doin:
	ldl	arg0
	ldl	arg1
	ldl	arg2
	in
	j	@forever
/*
 * M_OUT
 *
 * Perform an out instruction.
 * First argument is buffer pointer.
 * Second argument is channel address.
 * Third argument is buffer length.
 */
_out:
	ldl	cmnd
	eqc	M_OUT
	cj	@_in

	ldlp	arg0
	ldlp	arg1
	ldl	lnkI
	call	@get2wrds
	ldlp	arg2
	ldl	lnkI
	call	@getwrd
_doout:
	ldl	arg0
	ldl	arg1
	ldl	arg2
	out
	j	@forever
/*
 * M_WARM
 *
 * Perform a warm start by restarting moses.
 */
_warm:
	ldl	cmnd
	eqc	M_WARM
	cj	@_out

	j	@_start
/*
 * M_STTIMER
 *
 * Set the timers.
 */
_sttimer:
	ldl	cmnd
	eqc	M_STTIMER
	cj	@_warm

	ldlp	arg0
	ldl	lnkI
	call	@getwrd

	ldl	arg0
	sttimer
	j	@forever
/*
 * M_RUNP
 *
 * Perform a runp instruction.
 * First argument is iReg.
 * Second argument is wReg.
 */
_runp:
	ldl	cmnd
	eqc	M_RUNP
	cj	@_sttimer

	ldlp	arg0
	ldlp	arg1
	ldl	lnkI
	call	@get2wrds
	ldl	arg0
	ldl	arg1
	stnl	INACTIVE_IP
	ldl	arg1
	runp
	j	@forever
/*
 * M_CALL
 *
 * Perform a gcall instruction.
 * First argument is iReg.
 * Second argument is wReg.
 */
_call:
	ldl	cmnd
	eqc	M_CALL
	cj	@_runp

	ldlp	arg0
	ldlp	arg1
	ldl	lnkI
	call	@get2wrds
/*
 * Adjust my priority to what it needs to be when I call the new program.
 */
	ldc	__cont-here
	ldpi
here:
	stl	INACTIVE_IP
	ldlp	0
	ldl	arg1
	ldc	1
	and
	or
	runp
	stopp
__cont:
/*
 * Call the new program with given workspace (rounded down to word addr).
 */
	ldl	arg0
	ldl	arg1
	ldc	~3
	and
	gajw
	rev
	gcall
/*
 * M_FTCH
 *
 * Fetch data to the output link.
 * First argument is length.
 * Second argument is starting address.
 */
_ftch:
	ldl	cmnd
	eqc	M_FTCH
	cj	@_call

	ldlp	arg1
	ldlp	arg0
	ldl	lnkI
	call	@get2wrds
/*
 * Speak Meiko VCS protocol - written by DH.
 */
	ldl	lnkO
	ldl	arg1
	adc	3
	ldc	2
	shr
	outword

	ldl	arg0
	ldl	arg1
	adc	3
	ldc	3
	not
	and
	ldl	lnkO
	rev
	out

	j	@forever
/*
 * M_LOAD
 *
 * Load data from input link.
 * First argument is length.
 * Second argument is starting address.
 */
_load:
	ldl	cmnd
	eqc	M_LOAD
	cj	@_ftch

	ldl	lnkI
	stl	arg1
	ldlp	arg2
	ldlp	arg0
	ldl	lnkI
	call	@get2wrds
	j	@_doin
/*
 * M_READ
 *
 * Perform a ldnl 0 instruction.
 * First argument is word address.
 */
_read:
	ldl	cmnd
	eqc	M_READ
	cj	@_load

	ldlp	arg0
	ldl	lnkI
	call	@getwrd
/*
 * Speak Meiko VCS protocol.  First output # of words that follow.
 */
	ldl	lnkO
	ldc	1
	outword
	ldl	arg0
	ldl	lnkO
	ldc	4
	out
/*
 *	j	forever
 */
/*
 * Read in next command from input link.
 */
forever:
	ldc	0
	stl	cmnd

	ldlp	cmnd
	ldl	lnkI
	ldc	1
	in
/*
 * M_WRIT
 *
 * Perform a stnl 0 instruction.
 * First argument is address.
 * Second argument is data.
 */
_write:
	ldl	cmnd
	eqc	M_WRIT
	cj	@_read

	ldlp	arg0
	ldlp	arg1
	ldl	lnkI
	call	@get2wrds
	ldl	arg1
	ldl	arg0
	stnl	0
	j	@forever

END:
