/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	solder.c,v $
 * Revision 2.2  92/04/30  13:49:28  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  10:31:19  gdburns
 * overhaul machine independent code
 * 
 * Revision 2.1  91/03/20  11:48:41  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  91/03/19  14:35:37  raja
 * Overhaul of the topology layer tools.
 * 
 * Revision 2.0  90/10/04  17:23:12  gdburns
 * Meiko VCS
 * 
 *	Function:	- system specific topology reconfiguration tool
 *			- topology is specified using NaIL
 *			- Meiko VCS version
 */

#include <signal.h>
#include <stdio.h>
#include <portable.h>
#include <unistd.h>
#include <string.h>
#include <terror.h>
#include <typical.h>
#include <debug.h>
#include <nail.h>
#include <net.h>
#include <machine/machine.h>
#include <machine/svcs.h>

#define SSH		"ssh"

/*
 * global variables
 */
char			*me;		/* tool name */

/*
 * external variables
 */
extern int		errno;
extern FILE		*yyin;		/* lex input file */

/*
 * static variables
 */
static struct bschema	nail;		/* network description */
static int		fl_debug;	/* debug option flag */
static int		fl_pretend;	/* pretend option flag */
static int		fl_verbose;	/* verbose option flag */
static char		*f_bnail;	/* boot schema filename */
static char		*fullbnail;	/* full boot schema filename */
static char		**p_bnail;	/* schema pathname prefix */
static DomainSpec	*dspec;		/* ptr to Meiko domain spec. */

static char *usage = "usage: %s [-dhvNV] [-c <bnail>]\n";

/* 
 * external functions
 */
extern void		followed();
extern void		(*(signal()))();
extern void		validopts();
extern char		*findpath();
extern char		*getenv();
extern char		*getparam();
extern char		*strcpy();
extern char		*strrchr();
extern int		addarg();
extern int		cnfexec();
extern int		do_args();
extern int		find_host();
extern int		opt_taken();
extern int		parse_nail();
extern int		unlink();
extern int4		intparam();
extern int4		stoi();

/*
 * static functions
 */
static void		do_connect();
static void		help_solder();
static void		prt_defaults();
static void		set_defaults();
static int		check_connect();
static int		check_node();
static int		do_error();
static int		do_solder();

/*
 * main program
 */
int
main(argc, argv)  

int			argc;
char			*argv[];

{
/*
 * Assign the tool name.
 */
	if ((me = strrchr(argv[0], STRDIR)) == NULL) {
		me = argv[0];
	} else {
		me++;
	}
/*
 * Parse the command line.
 */
	validopts("dhvNV");
	followed("c");

	if ((do_args(&argc, argv) < 0) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		fprintf(stderr, usage, me);
		exit(errno);
	}

	if (opt_taken('h')) {
		help_solder();
		exit(0);
	}
/*
 * Set the mode of operation.
 */
	set_defaults();

	if (fl_debug) {
		prt_defaults();
	}
/* 
 * Locate the boot schema file.
 */
	if ((fullbnail = findpath(f_bnail, p_bnail, R_OK)) == NULL) {
		fprintf(stderr, "%s: cannot locate bnail file \"%s\"\n",
			me, f_bnail);
		exit(1);
	}
/*
 * Open the boot schema file.
 */
	DBUG("%s: opening the boot schema file: %s ...\n", me, f_bnail);
	if ((yyin = fopen(fullbnail, "r")) == NULL) {
		perror(me);
		exit(errno);
	}
/*
 * Parse the boot schema.
 */
	if (parse_nail(&nail) < 0) {
		fprintf(stderr, "%s: cannot parse the NaIL syntax\n", me);
		exit(errno);
	}
/*
 * Close the boot schema.
 */
	if (fclose(yyin) == EOF) {
		perror(me);
		exit(ERROR);
	}
/*
 * Print the boot schema out if needed.
 */
	if ((opt_taken('V')) || (fl_debug)) {
		pretty_bnail(&nail);

		if (opt_taken('V')) {
			exit(0);
		}
	}
/*
 * Ignore the INT signal.
 */
	signal(SIGINT, SIG_IGN);
/*  
 * Solder the network.
 */
	DBUG("%s: soldering ...\n", me);
	if (do_solder() < 0) {
		fprintf(stderr, "%s: unable to solder\n", me);
		exit(errno);
	}

	DBUG("%s: soldered OK\n", me);
	return(0);
}

/*
 *	help_solder
 *
 *	Function:	- prints the help menu
 */

static void 
help_solder()
{
	printf("\nSynopsis:\t%s [options]\n", me);
	printf("\nDescription:\tConfigure the topology of a multicomputer.\n");
	printf("\nOptions:\t-h\t\tPrint this message.\n");
	printf("\t\t-d\t\tPrint debug information.\n");
	printf("\t\t-v\t\tBe verbose.\n");
	printf("\t\t-N\t\tPretend to take action.\n");
	printf("\t\t-V\t\tPretty print the bnail file.\n");
	printf("\t\t-c <bnail>\tUse <bnail> as boot schema.\n");
}

/*
 *	set_defaults
 *
 *	Function:	- sets the values of key variables controlling
 *			  the operation of the tool
 */

static void 
set_defaults()

{
	int		n_bnail;

	n_bnail = 0;
	p_bnail = (char **) 0;
	addarg(&n_bnail, &p_bnail, "");
	addarg(&n_bnail, &p_bnail, "$TROLLIUSHOME/boot");
	addarg(&n_bnail, &p_bnail, DEFPBNAIL);
/*
 * Set the file names.
 */
	if (opt_taken('c')) {
		f_bnail = getparam('c');
	} else if (f_bnail = getenv("TROLLIUSBNAIL")) {
	} else {
		f_bnail = DEFFBNAIL;
	}
/*
 * Set the flags.
 */
	fl_verbose = opt_taken('v');
	fl_pretend = opt_taken('N');

	if (opt_taken('d')) { fl_debug = TRUE;
		fl_verbose = TRUE;
	} else {
		fl_debug = FALSE;
	}
}

/*
 *	prt_defaults
 *
 *	Function:	- prints the values of key variables
 *			- mostly used for debugging
 */

static void 
prt_defaults()

{
	printf("\nDefault Settings\n\n");
	printf("boot schema: %s\n", f_bnail);
	printf("debug flag: %d\n", fl_debug);
	printf("verbose flag: %d\n", fl_verbose);
	printf("pretend flag: %d\n", fl_pretend);
}

/*
 *	do_error
 *
 *	Function:	- reports error messages from do_solder()
 *			- closes and unlinks the command file
 *	Accepts:	- the error message
 *	Returns:	- ERROR
 */

static int
do_error(message)

char			*message;

{
	fprintf(stderr, "%s (do_solder)", me);

	if (message) {
		fprintf(stderr, ": %s\n", message);
	} else {
		fprintf(stderr, "\n");
	}

	return(ERROR);
}

/*
 *	do_solder
 *
 *	Function:	- reconfigures the system links
 *			- system depend code: Meiko VCS version
 *	Returns:	- 0 or ERROR
 */

static int
do_solder()

{
	int4		i;
	int4		sysid;
	int		fl_error;
	int		n_cmd;
	int		status;
	char		**v_cmd;
	struct t_node	*tnode;			/* ptr to tnode element */
	struct t_link	*tlink;			/* ptr to tlink element */
	struct link	*plink;			/* ptr to list of links */
	struct nspecs	*spec;			/* ptr to node specs */

	int		vcsfd, supfd;		/* Meiko VCS file desc. */
	Resources	resource;		/* Meiko resource struct */
/*
 * Locate the host node of the Trollius multicomputer.
 */
	if (find_host(&nail)) {
		return(do_error("cannot find host node"));
	}
/*
 * A constraint of the Meiko hardware is to have a node with a Meiko
 * nodeid 0 connected via its out going link 0 to the host node.
 * Check for this by verifying all the host incoming connections for
 * a node with a system nodeid of 0 and a link 0 connection to the host.
 */
	tnode = nail.bs_hostnode;
	spec = tnode->tn_specs;
	plink = spec->ns_lkin;
	fl_error = ERROR;

	for (i = 0; i < spec->ns_nin; i++, plink = plink->lk_next) {
		tlink = plink->lk_link;
/*
 * Get the system nodeid of the source node of this link.
 */
		sysid = stoi(((tlink->tl_srcnode)->tn_specs)->ns_sysnode);

		if (sysid == 0) {

			if (tlink->tl_srclink != 0) {
				return(do_error("Meiko node 0 misconnected"));
			} else {
				fl_error = 0;
				break;
			}
		}
	}

	if (fl_error == ERROR) {
		return(do_error("Meiko node 0 not found"));
	}
/*
 * Define the number of Meiko nodes needed and their type.
 * Type is not implemented here because Trollius doesn't use it.
 * The default node type is set to 1 Meg T8 nodes.  Make sure
 * it does not wait if it cannot grab the required network.
 */ 
	if ((dspec = svcs_newspec()) == (DomainSpec *) 0) {
		return(do_error("cannot grab a Meiko domain specification"));
	}

	dspec->d_wait = 0;

	for (i = 0; i < (ah_count(nail.bs_tnhd) - 1); i++) {
		svcs_procspec(dspec, T800, MEGABYTES);
	}
/*
 * Issue the link reconfiguration commands.   Some links should not
 * be reconfigured (system specific) as determined by the
 * check_connect() function.
 */
	tlink = nail.bs_tlink;

	for (i = 0; i < nail.bs_ntlink; i++, tlink = tlink->tl_next) {

		if (check_connect(tlink)) {
			do_connect(tlink);
		}
	}
/* 
 * Execute the system soldering command.
 */
	VERBOSE("soldering ...\n");
	if (! fl_pretend) {
		svcs_grab(dspec, &vcsfd, &supfd, &resource);

		if (vcsfd < 0) {
			return(ERROR);
		}

		if (dup2(vcsfd, VCSFD) < 0) {
			return(do_error("cannot dup VCS file desc."));
		}

		if (dup2(supfd, SUPFD) < 0) {
			return(do_error("cannot dup SUP. file desc."));
		}
	}
/*
 * Due to the design of the Meiko daemon (svcsd), every process that
 * wants to access the link device has to inherit the file descriptor
 * from solder.  We therefore fork a special shell to pass along 
 * the descriptor to any children the user wants to specify, in particular
 * the spread tool.
 */
	n_cmd = 0;
	v_cmd = (char **) 0;
	addarg(&n_cmd, &v_cmd, SSH);
	DBUG("forking the shell\n");

	status = cnfexec(v_cmd);

	if (status < 0) {
		fprintf(stderr, "%s: can't exec %s\n", me, v_cmd[0]);
		perror(me);
		return(ERROR);
	} else if (status > 0) {
		fprintf(stderr, "%s: %s exited with code %d\n",
			me, v_cmd[0], status);
		errno = status;
		return(ERROR);
	}

	return(0);
}

/*
 *	check_connect
 *
 *	Function:	- checks if link connection is allowed
 *			- system specific
 *			- Meiko VCS version
 *	Accepts:	- ptr to link description
 *	Returns:	- TRUE or FALSE
 */

static int
check_connect(link)

struct t_link		*link;

{
	struct nspecs	*src_spec;
	struct nspecs	*dest_spec;
/*
 * Check the node types being linked.  If any one is an otb node
 * the connection should be inhibited.
 */
	src_spec = (link->tl_srcnode)->tn_specs;
	dest_spec = (link->tl_destnode)->tn_specs;

	if (tisotb(src_spec->ns_type) || tisotb(dest_spec->ns_type)) {
		return(FALSE);
	}
/*
 * Check the syntax of the system specific node names.
 */
	if (! check_node(src_spec->ns_sysnode)) {
		fprintf(stderr, "%s: illegal system node syntax: %s\n", 
			me, src_spec->ns_sysnode);
		return(FALSE);
	}

	if (! check_node(dest_spec->ns_sysnode)) {
		fprintf(stderr, "%s: illegal system node syntax: %s\n", 
			me, dest_spec->ns_sysnode);
		return(FALSE);
	}
/*
 * Check if both nodes have the same host.  For now links are connected
 * only when they share the same host.  This restriction should be
 * removed later.
 */
	if (src_spec->ns_host && dest_spec->ns_host) {

		if (strcmp(src_spec->ns_host, dest_spec->ns_host)) {
			return(FALSE);
		}
	}

	else if (src_spec->ns_host || dest_spec->ns_host) {
		return(FALSE);
	}
/*
 * Check if the link is hardwired.  For this system, the hardwired link
 * is link 0 of Meiko system node 0.
 */
	if ((link->tl_srclink == (int4) 0) &&
		(stoi(((link->tl_srcnode)->tn_specs)->ns_sysnode) == 0)) {
		return(FALSE);
	}
/*
 * Check if the link is a duplicate.  A link is duplicate if the
 * source link number is greater than the destination link number or if
 * they are equal but the source nodeid is greater or equal to the
 * destination nodeid.  As nodeids are unique, this forms a complete
 * ordering among (node,link) pairs and the connection is allowed
 * iff (srcnode, srclink) < (destnode, destlink).
 */
	if (link->tl_srclink > link->tl_destlink) {
		return(FALSE);
	}

	else if (link->tl_srclink < link->tl_destlink) {
		return(TRUE);
	}

	else if ((link->tl_srcnode)->tn_nodeid <
		(link->tl_destnode)->tn_nodeid) {
		return(TRUE);
	}

	else {
		return(FALSE);
	}
}

/*
 *	do_connect
 *
 *	Function:	- issues the command to connect the link
 *			- system specific
 *			- Meiko VCS version
 *	Accepts:	- ptr to the link description
 */

static void
do_connect(link)

struct t_link		*link;

{
	if ( ((link->tl_srcnode)->tn_specs)->ns_sysnode == NULL) {
		fprintf(stderr, "%s (do_connect): node %d has no sysnode\n",
			me, (link->tl_srcnode)->tn_nodeid);
		return;
	}

	if ( ((link->tl_destnode)->tn_specs)->ns_sysnode == NULL) {
		fprintf(stderr, "%s (do_connect): node %d has no sysnode\n",
			me, (link->tl_destnode)->tn_nodeid);
		return;
	}

	if (! fl_pretend) {
		VERBOSE("wire %d.%d %d.%d\n", (link->tl_srcnode)->tn_nodeid,
			link->tl_srclink, (link->tl_destnode)->tn_nodeid,
			link->tl_destlink);
		svcs_wirespec(dspec, (link->tl_srcnode)->tn_nodeid,
			link->tl_srclink, (link->tl_destnode)->tn_nodeid,
			link->tl_destlink);
	}
}

/*
 *	check_node
 *
 *	Function:	- checks system specific node name syntax
 *			- Meiko VCS version
 *			- allowed syntax: all characters are digits
 *	Accepts:	- node name
 *	Returns:	- TRUE or FALSE
 */

static int 
check_node(name)

char			*name;

{
/*
 * Handle the trivial case.
 */
	if (name  == NULL) {
		return(FALSE);
	}
/*
 * Loop through all the characters checking if they
 * are decimal digits.
 */
	for ( ; *name; name++) {

		if (! isdigit((int) *name)) {
			return(FALSE);
		}
	}

	return(TRUE);
}
