/*
 *	OSU Research Computing				Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB/RBD
 *
 *	$Log:	ssh.c,v $
 * Revision 2.2  92/04/30  13:49:51  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:48:53  gdburns
 * Ohio Release
 * 
 * Revision 2.0  90/10/04  17:23:44  gdburns
 * Meiko VCS
 *
 *	Function:	- small shell for Meiko VCS Trollius
 *			- used to invoke spread, hboot, sboot, export, cboot,
 *			  and moses which all must inherit the link device fd
 */

#include <signal.h>
#include <stdio.h>
#include <terror.h>
#include <typical.h>
#include <sys/wait.h>
#include <ssh.h>

/*
 * external functions
 */
extern void		(*(signal()))();
extern void		validopts();
extern int		do_args();

/*
 * local functions
 */
static void		bail();
static void		cmdinvoke();
static void		escape();
static void		help();
static void		prompt();
static void		resetopts();
static void		usage();
static int		cmdtoken();

/*
 * local variables
 */
static char		line[MAXLINE];		/* input command line */

static char		*cmdnames[] = {		/* builtin command names */
				"exit",
				"help",
				"quit",
				"",
			};

static void		(*cmdfuncs[])() = {	/* builtin command functions */
				bail,
				help,
				bail,
			};

static char *invoke = "ssh [-dh]";

/*
 * external variables
 */
extern int		errno;

/*
 * global variables
 */
int			fdebug;			/* debug mode on/off */


main(argc, argv)

int			argc;
char			*argv[];

{
	int		ac;			/* # of command line args */
	char		*av[MAXARGV];		/* command line args */

	validopts("dh");

	if ((do_args(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n", invoke);
		}

		else {
			terror("ssh (arguments)");
		}

		return(errno);
	}

	if (opt_taken('h')) {
		usage();
		return(0);
	}

	fdebug = opt_taken('d');
	prompt();
	*line = '\0';

	for (;;) {
/*
 * Read a command line.
 */
		if (fgets(line, MAXLINE, stdin) == NULL) {
			exit(0);
		}

		else {

			if (*line == ESCAPE) {
				escape(line + 1);
			}

			else {
				ac = cmdtoken(line, av);

				if (ac > 0) {
					cmdinvoke(ac, av);
				}
			}

			prompt();
		}
	}
}

/*
 *	cmdtoken
 *
 *	Function:	- tokenize command line into argc/argv format
 *
 *	Accepts:	- ptr to command line
 *			- argv, a ptr to an array of token ptrs
 *
 *	Returns:	- number of token ptrs
 */

static int
cmdtoken(cmdline, av)

char			*cmdline;
char			**av;

{
	int		ac;			/* # command line args */
	char		*pcmd;			/* command line ptr */
	int		ftoken;			/* in middle of token? */


	ac = 0;
	ftoken = FALSE;
	pcmd = cmdline;

	for (pcmd = cmdline; (*pcmd != '\0') && (ac < (MAXARGV - 1)); ++pcmd) {

		if ((*pcmd == ' ') || (*pcmd == '\t') || (*pcmd == '\n') ||
			(*pcmd == '\0')) {

			if (ftoken) {
				*pcmd = '\0';
				ftoken = FALSE;
			}
		}

		else if (!ftoken) {
			av[ac] = pcmd;
			++ac;
			ftoken = TRUE;
		}
	}

	if (ac >= (MAXARGV - 1)) {
		fprintf(stderr, "PC-T: maximum number of arguments exceeded\n");
		ac = 0;
	}

	av[ac] = NULL;
	return(ac);
}

/*
 *	cmdinvoke
 *
 *	Function:	- invoke function or exec program
 *
 *	Accepts:	- number of arguments (passed to command function)
 *			- ptr to arguments (passed to command function)
 */

static void
cmdinvoke(ac, av)

int			ac;
char			**av;

{
	int		i;		/* favourite index */
    	union wait	status;		/* child exit status */


	if ((i = strsrch(cmdnames, av[0])) >= 0) {

		(*cmdfuncs[i])(ac, av);
	}

	else {
		i = fork();

		if (i == ERROR) {
			perror("ssh");
			exit(errno);
		}

		else if (i == 0) {
			(void) signal(SIGINT, SIG_DFL);
			execvp(av[0], av);
			perror(av[0]);
			exit(errno);
		}

		else {
			while ((wait(&status) != 0) && (! WIFEXITED(status)) &&
				( ! WIFSIGNALED(status)));
		}
	}
}

/*
 *	strsrch
 *
 *	Function:	- search for a key string in a table of strings
 *			- key string only need match to the length
 *			  of the table string
 *
 *	Accepts:	- ptr to table of strings
 *			- ptr to key string
 *
 *	Returns:	- table index of matched string
 */

static int
strsrch(strs, key)

char			*strs[];
char			*key;

{
	int		i;			/* favourite index */


	i = 0;

	while ((*strs[i] != '\0') &&
		(strncmp(strs[i], key, strlen(key)))) {

		++i;
	}

	if (*strs[i] == '\0') {
		i = ERROR;
	}

	return(i);
}


/*
 *	help
 *
 *	Function:	- prints helpful information on ssh commands
 */

static void
help()

{
	printf("\n\tssh commands:\n\n");
	printf("\tspread\t\tBoot Trollius onto a multicomputer.\n");
	printf("\thboot\t\tBoot programs onto the local host node.\n");
	printf("\tsboot\t\tBoot programs onto the first compute node.\n");
	printf("\texport\t\tExport bootstrap onto the first compute node.\n");
	printf("\tcboot\t\tBoot single program onto the first compute node.\n");
	printf("\tmoses\t\tInterface to moses bootstrap.\n");
	printf("\ttkill\t\tTerminate local host node programs.\n");
	printf("\texit\t\tExit ssh.\n");
	printf("\tquit\t\tExit ssh.\n");
	printf("\t!<cmd>\t\tEscape <cmd> to the local environment.\n\n");
}

/*
 *	  usage
 *
 *	  Function:	- prints helpful information ssh options
 */

static void
usage()
{
	printf("\nSynopsis:	ssh [options]\n");
	printf("\nOptions:	-h		this help message\n");
	printf("\nDescription:	small shell for Trollius boot commands\n\n");
}

/*
 *	prompt
 *
 *	Function:	- display the ssh prompt
 */

static void
prompt()

{
	fputs("T> ", stdout);
	fflush(stdout);
}

/*
 *	escape
 *
 *	Function:	- Feed a command to the native operating system.
 *
 *	Accepts:	- ptr to command line
 */

static void
escape(cmdline)

char			*cmdline;

{
	(void) system(cmdline);
}

/*
 *	bail
 *
 *	Function:	- Bail out of ssh.
 */

static void
bail()

{
	char		*av[2];			/* command line args */

	av[0] = "tkill";
	av[1] = NULL;
	cmdinvoke(1, av);
	exit(0);
}
