/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	tfish.c,v $
 * Revision 2.2  92/04/30  13:50:09  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/03/08  10:23:45  gdburns
 * Change name to tfish.
 * 
 * Revision 2.1.1.1  91/04/15  14:20:38  raja
 * Exit with a status code 0 upon successful termination.
 * 
 * Revision 2.1  91/03/20  11:48:03  gdburns
 * Ohio Release
 * 
 * Revision 2.0  90/10/04  17:15:06  gdburns
 * Meiko VCS
 *
 *	Function:	- "tfish"
 *			- Meiko VCS version
 *			- reset itb nodes associated with an otb-itb link
 *			- device name is accepted for compatibility
 *			  with other versions of tfish
 *			- so named from our original reset device, fish line.
 */

#include <stdio.h>
#include <typical.h>
#include <machine/machine.h>
#include <machine/svcs.h>
#include <machine/supercmd.h>
#include <sys/types.h>
#include <sys/stat.h>

/*
 * external functions
 */
extern int		do_args();
extern void		followed();
extern void		intparam();
extern void		validopts();

/*
 * local functions
 */
static void		help();
static int		writeReset();

/*
 * local variables
 */
static char		*me;		/* name I was exec'd by */
static char		*usage = "usage: %s <device> <bootstrap>\n";

/*
 * external variables
 */
extern int		errno;

main(argc, argv)
int			argc;
char			**argv;
{
	int		delay;		/* #secs to delay after reset */
	int		j;		/* simple counter */

	me = argv[0];
/*
 * Initialize option parser.
 */
	validopts("dh");
	followed("d");

	if (do_args(&argc, argv) < 0) {
		(void) fprintf(stderr, usage, me);
		exit(errno);
	}

	if (opt_taken('h')) {
		help();
		exit(0);
	}

	if (argc > 2) {
		(void) fprintf(stderr, usage, me);
		exit(1);
	}
/*
 * How do we issue a reset command to svcsd?
 * This does not quite work.
 *
 *	for (j = 0; writeReset(SUPFD, j, 1) >= 0; ++j);
 *	for (j = 0; writeReset(SUPFD, j, 0) >= 0; ++j);
 *	if (j == 0) (void) fprintf(stderr, "%s: can't reset, errno = %d\n", me,
 *		errno);
 */
/*
 * Delay.
 */
	if (opt_taken('d')) {
		intparam('d', &delay);
		sleep((unsigned) delay);
	}

	return(0);
}

/*
 *	help
 *
 *	Function:	- prints useful information on the tfish command
 */

static void
help()

{
	printf("\nSynopsis:	tfish [-h] [-d <delay>] [<filename>]\n");
	printf("\nDescription:	Reset ITB node(s).\n");
	printf("\nOptions:	-h		Print this help message.\n");
	printf("\t\t-d <delay>	Pause after reseting\n");
	printf("\t\t<filename>	This is not used for Meiko tfish.\n");
}

/*
 *	writeReset
 *
 *	Function:	- reset a process using svcsd and the supervisor bus
 *			- supplied by Titch at Meiko
 *
 *	Accepts:	- supervisor bus file descriptor
 *			- processor number (Meiko style)
 *			- reset signal level (high or low)
 */

static int
writeReset(supfd, procno, value)

int			supfd;
int			procno;
int			value;

{
	int		rc;
	int		nw;
	SuperMsg	m;
    
/*
 * Build up request message.
 */
	m.super_cmd = SUPER_SBWRITE;
	m.super_id = procno;
	m.super_p1 = FN_RESET;
	m.super_p2 = (value == 0) ? 0 : 0xff;
/*
 * Convert Host to Meiko (little-endian) bytesex.
 */
	svcs_htomlongs(&m, &m, SUPER_ENTRY);
/*
 * Send it down the supervisor socket to the svcs daemon.
 */
	rc = svcs_pipesend(supfd, SUPER_ENTRY, &m);

	if (rc < 0)  {
		return(rc);
	}
/*
 * Receive reply.  I can't quite remember whether we ought to
 * do a select() before this, but I don't think we need to ...
 */
	rc = svcs_piperecv(supfd, &nw, &m);
	if (rc < 0) {
		return(rc);
	}
	svcs_mtohlongs(&m, &m, SUPER_ENTRY);
/*
 * Check if it succeeded (could have been bad processor no).
 */
	if (m.super_id < 0) return(-1);
	return(0);   
}
