/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	dl_mom.c,v $
 * Revision 2.2  92/04/30  13:53:47  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  10:38:01  gdburns
 * buffer flow control enhancements
 * 
 * Revision 2.1  91/03/20  11:55:55  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:31:48  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:57:13  trillium
 * Homo erectus
 * 
 *	Function:	- creates datalink process pairs 
 *			- command line arguments determine datalink
 *			  configuration
 *			- does one centralized open for each datalink device
 *			- usage: dl_dev devfile event devfile event ...
 */

#include <stdio.h>
#include <dlreq.h>
#include <fcntl.h>
#include <kernel.h>
#include <net.h>

/*
 * external variables
 */
extern int		errno;			/* global error variable */
extern int4		ppid[MAXNLINK];		/* child dlo processes */

/*
 * external functions
 */
extern void		dld();			/* datalink daemon */
extern void		dlo_dev();		/* datalink output */
extern void		dli_dev();		/* datalink input */

/*
 * global variables
 */
char			*me;


main(argc, argv)

int			argc;
char			*argv[];

{
	int		event;			/* dlo event */
	int		fd;			/* device descriptor */
	int		i;			/* favourite index */
	int		pid;			/* dl child process ID */

	for (i = 0; i < MAXNLINK; ++i) {
		ppid[i] = NOTPROCESS;
	}

	me = argv[0];
	argv++;
	argc--;
/*
 * We must have an even number of options.
 */
	if ((argc % 2) != 0) {
		fprintf(stderr, "usage: dl_dev device event ...\n");
		exit(1);
	}
/*
 * Loop through command line pairs.
 */
	while (argc > 0) {
/*
 * Open the device.
 */
		if ((fd = open(argv[0], O_RDWR)) < 0) {
			fprintf(stderr, "%s: can't open %s\n", me, argv[0]);
			perror(me);
			exit(errno);
		}
/*
 * Fork the datalink input process.
 */
		if ((pid = fork()) < 0) {
			perror(me);
			exit(errno);
		} else if (pid == 0) {
			dli_dev(fd);
			exit(0);
		}
/*
 * Fork the datalink output process.
 */
		event = stoi(argv[1]);

		if ((pid = fork()) < 0) {
			perror(me);
			exit(errno);
		} else if (pid == 0) {
			dlo_dev(fd, event);
			exit(0);
		}

		ppid[dl_link(event)] = pid;
		argv += 2;
		argc -= 2;
	}

	dld();
	return(0);
}
