/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	dli_dev.c,v $
 * Revision 2.2  92/04/30  13:53:48  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  10:38:13  gdburns
 * buffer flow control enhancements
 * 
 * Revision 2.1  91/03/20  11:55:57  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.5  90/07/31  17:33:54  dixit-v
 * Fixed bug : Read nh_msg from device before performing byte conversions.
 * 
 * Revision 2.0.1.4  90/07/14  16:06:10  gdburns
 * Added more robustness.
 * 
 * Revision 2.0.1.3  90/07/13  19:31:51  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.2  90/06/06  14:37:46  gdburns
 * Corrected error message typo.
 * 
 * Revision 2.0.1.1  90/05/02  14:11:19  dixit-v
 * Fixed bug - changed D***DATA to D***MSG for nh_msg field.
 * 
 * Revision 2.0  90/02/20  11:57:15  trillium
 * Homo erectus
 * 
 *	Function:	- Read network frames from a device and forward them
 *			  using nsend().
 *	Accepts:	- device file descriptor open for read
 */

#include <stdio.h>
#include <net.h>
#include <debug.h>
#include <t_types.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */

void
dli_dev(fd)

int			fd;

{
	int		i;		/* favourite counter variable */
	int		data_conv;	/* data conversion flags */
	int		*pint;		/* integer pointer */
	float		*pflt;		/* float pointer */
	double		*pdbl;		/* double pointer */
	char		ack;		/* acknowledgement byte */
	struct netframe	inframe;	/* network frame (header and data) */

/*
 * Initialize and attach to the kernel.
 */
	if (kinit(0) < 0) {
		terror("dli_dev");
		exit(errno);
	}
/*
 * Process network frames forever.
 */
	for(;;) {	
/*
 * Read network frame header from device.
 */
	    if (mread(fd, (char *) &inframe.nf_header, sizeof(struct nmsg))) {
		kexit(errno);
	    }
/*
 * Convert header (sans nh_data and nh_msg) to local byte ordering.
 */
	    pint = (int *) &inframe.nf_header;

	    for (i = 0; i < ((sizeof(struct nmsg) / sizeof(int)) -
		(NHDSIZE + 1)); ++i) {

		*pint = ttol(*pint);
		++pint;
	    }

  db(("dli_dev: incoming frame: node=%x event=%d type=%x flags=%x length=%d\n",
		inframe.nf_header.nh_node,
		inframe.nf_header.nh_event,
		inframe.nf_header.nh_type,
		inframe.nf_header.nh_flags,
		inframe.nf_header.nh_length));
/*
 * Convert nh_data to the local data representation.
 */
	    data_conv = inframe.nf_header.nh_flags & DMSKDATA;

	    if (!data_conv || (data_conv == DINTDATA)) {
		pint = inframe.nf_header.nh_data;

		for (i = 0; i < NHDSIZE; ++i) {
		    *pint = ttol(*pint);
		    ++pint;
	        }
	    }

	    else if (data_conv == DFLTDATA) {
		pflt = (float *) inframe.nf_header.nh_data;

		for (i = 0; i < ((NHDSIZE * sizeof(int)) / sizeof(float));
		    ++i) {

		    *pflt = ttolf(*pflt);
		    ++pflt;
	        }
	    }

	    else if (data_conv == DDBLDATA) {
		pdbl = (double *) inframe.nf_header.nh_data;

		for (i = 0; i < ((NHDSIZE * sizeof(int)) / sizeof(double));
		    ++i) {

		    *pdbl = ttold(*pdbl);
		    ++pdbl;
		}
	    }
/*
 * Read network frame data from device.
 */
	    if (mread(fd, inframe.nf_data, inframe.nf_header.nh_length)) {
		kexit(errno);
	    }
	    inframe.nf_header.nh_msg = inframe.nf_data;
/*
 * Convert nh_msg to the local data representation.
 */

	    data_conv = inframe.nf_header.nh_flags & DMSKMSG;

	    if (data_conv == DINTMSG) {
		pint = (int *) inframe.nf_header.nh_msg;

		for (i = 0; i < (inframe.nf_header.nh_length / sizeof(int));
		    ++i) {

		    *pint = ttol(*pint);
		    ++pint;
	        }
	    }

	    else if (data_conv == DFLTMSG) {
		pflt = (float *) inframe.nf_header.nh_msg;

		for (i = 0; i < (inframe.nf_header.nh_length / sizeof(float));
		    ++i) {

		    *pflt = ttolf(*pflt);
		    ++pflt;
	        }
	    }

	    else if (data_conv == DDBLMSG) {
		pdbl = (double *) inframe.nf_header.nh_msg;

		for (i = 0; i < (inframe.nf_header.nh_length / sizeof(double));
		    ++i) {

		    *pdbl = ttold(*pdbl);
		    ++pdbl;
		}
	    }
/*
 * Reset the BYBUF flag.
 */
	    inframe.nf_header.nh_flags &= ~BYBUF;
	    inframe.nf_header.nh_flags |= DDLI;
/*
 * Forward network message.
 */
	    if (inframe.nf_header.nh_node == NOTNODEID) {
		inframe.nf_header.nh_node = LOCAL;
	    }

	    if (nsend(&inframe.nf_header)) {
/*
 *		terror("dli_dev");
 *		kexit(errno);
 */
	    }
/*
 * Read acknowledgement byte from device if not continuing a VC.
 * This is needed for compatibility with dl_t4.
 */
	    if (! (inframe.nf_header.nh_flags & KHOLD)) {

		if (mread(fd, &ack, sizeof(ack))) {
		    kexit(errno);
		}
	    }
	}   /* for */
}

/*
 *	mread
 *
 *	Function:	- execute multiple reads up to a given length
 *	Accepts:	- file descriptor
 *			- length 
 *			- buffer pointer
 *	Returns:	- 0 or ERROR
 */

static
mread(fd, q, length)

int			fd;
int			length;
char			*q;

{
	int		ret;		/* # of bytes actually read */

	while ((ret = read(fd, q, length)) < length) {

		if (ret == ERROR) {
			terror("dli_dev");
			return(ERROR);
		} else if (ret == 0) {
			fprintf(stderr, "dli_dev: premature EOF\n");
			return(ERROR);
		} else {
			length -= ret;
			q += ret;
		}
	}

	return(0);
}
