/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	dlo_dev.c,v $
 * Revision 2.2  92/04/30  13:53:49  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/27  10:38:15  gdburns
 * buffer flow control enhancements
 * 
 * Revision 2.1.1.1  91/10/09  15:48:25  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1  91/03/20  11:55:58  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.4  90/09/27  18:14:58  gdburns
 * Converted to modern portable.h.
 * 
 * Revision 2.0.1.3  90/07/14  16:06:18  gdburns
 * Added more robustness.
 * 
 * Revision 2.0.1.2  90/07/13  19:31:50  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/05/02  14:15:39  dixit-v
 * Fixed bug - changed D***DATA to D***MSG for nh_msg field.
 * 
 * Revision 2.0  90/02/20  11:57:14  trillium
 * Homo erectus
 * 
 *	Function:	- Receive network frames on a fixed event and write
 *			  them to a device.
 *	Accepts:	- device file descriptor open for write
 *			- dlo event
 */

#include <dlreq.h>
#include <stdio.h>
#include <net.h>
#include <debug.h>
#include <kio.h>
#include <ksignal.h>
#include <t_types.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */
extern struct kio_t	_kio;

/*
 * local functions
 */
static void		valve();	/* positive event valve */

/*
 * local variables
 */
static int		filter;		/* message filter */

void
dlo_dev(fd, event)

int			fd;
int			event;

{
	int 		length;		/* total frame length */
	int		nbytes;		/* return value from write */
	int		i;		/* favourite counter variable */
	int		r;		/* favourite return value */
	int		holding;	/* Are we holding a VC? */
	int		data_conv;	/* data conversion flags */
	int		mask;		/* signal retry mask */
	int		*pint;		/* integer pointer */
	float		*pflt;		/* float pointer */
	double		*pdbl;		/* double pointer */
	int		ihold;		/* index of our dl event in VC tbl */
	char		ack;		/* acknowledgement byte */
	struct netframe	outframe;	/* network frame (header and data) */

/*
 * Initialize and attach to the kernel.
 */
	if (kinit(0) < 0) {
		terror("dlo_dev");
		exit(errno);
	}
/*
 * Catch SIGA and SIGB.
 */
	if ((ksignal(SIGA, valve) == 0) ||
		(ksignal(SIGB, valve) == 0)) {

		terror("dlo_dev");
		kexit(errno);
	}

	mask = ksigretry(0);
	ksigsetretry(mask & ~(sigmask(SIGA) | sigmask(SIGB)));

	ihold = (event & HOLDBITS) + ((event & INT4_SIGN) ? (HOLDBITS + 1) : 0);
	filter = DLALL;
/*
 * Process network frames forever.
 */
	for(;;) {	
/*
 * Receive network frame.
 */
		outframe.nf_header.nh_event = event;
		outframe.nf_header.nh_type = filter;
		outframe.nf_header.nh_flags = DFORWARD;
		outframe.nf_header.nh_length = MAXNMSGLEN;
		outframe.nf_header.nh_msg = outframe.nf_data;

		while (((r = drecv(&outframe.nf_header)) < 0) &&
			(errno == EINTR)) {
			outframe.nf_header.nh_type = filter;
		}

		if (r < 0) {
			terror("dlo_dev");
			kexit(errno);
		}

	db(("dlo_dev: outgoing frame, node=%x event=%d type=%x length=%d\n",
			outframe.nf_header.nh_node,
			outframe.nf_header.nh_event,
			outframe.nf_header.nh_type,
			outframe.nf_header.nh_length));
/*
 * Pull a very sneaky manouver to maintain a virtual circuit if
 * so indicated by the header.
 */
		if (outframe.nf_header.nh_flags & KHOLD) {
			_kio.ki_khold[ihold].kh_event = event;
		}

		length = sizeof(struct nmsg) + outframe.nf_header.nh_length;
/*
 * Convert nh_data to Trillium byte ordering.
 */
		data_conv = outframe.nf_header.nh_flags & DMSKDATA;

		if (!data_conv || (data_conv == DINTDATA)) {
		    pint = outframe.nf_header.nh_data;

		    for (i = 0; i < NHDSIZE; ++i) {
			*pint = ltot(*pint);
			++pint;
	            }
		}

		else if (data_conv == DFLTDATA) {
		    pflt = (float *) outframe.nf_header.nh_data;

		    for (i = 0; i < ((NHDSIZE * sizeof(int)) / sizeof(float));
			++i) {

			*pflt = ltotf(*pflt);
			++pflt;
		    }
		}

		else if (data_conv == DDBLDATA) {
		    pdbl = (double *) outframe.nf_header.nh_data;

		    for (i = 0; i < ((NHDSIZE * sizeof(int)) / sizeof(double));
			++i) {

			*pdbl = ltotd(*pdbl);
			++pdbl;
		    }
		}
/*
 * Convert nh_msg to the local data representation.
 */
		data_conv = outframe.nf_header.nh_flags & DMSKMSG;

		if (data_conv == DINTMSG) {
		    pint = (int *) outframe.nf_header.nh_msg;

		    for (i = 0; i < (outframe.nf_header.nh_length /
			sizeof(int)); ++i) {

			*pint = ltot(*pint);
			++pint;
	            }
		}

		else if (data_conv == DFLTMSG) {
		    pflt = (float *) outframe.nf_header.nh_msg;

		    for (i = 0; i < (outframe.nf_header.nh_length /
			sizeof(float)); ++i) {

			*pflt = ltotf(*pflt);
			++pflt;
	            }
		}

		else if (data_conv == DDBLMSG) {
		    pdbl = (double *) outframe.nf_header.nh_msg;

		    for (i = 0; i < (outframe.nf_header.nh_length /
			sizeof(double)); ++i) {

			*pdbl = ltotd(*pdbl);
			++pdbl;
		    }
		}
/*
 * Convert header (sans nh_data and nh_msg) to Trillium byte ordering.
 */
		holding = outframe.nf_header.nh_flags & KHOLD;
		pint = (int *) &outframe.nf_header;

		for (i = 0; i < ((sizeof(struct nmsg) / sizeof(int)) -
		    (NHDSIZE + 1)); ++i) {

		    *pint = ltot(*pint);
		    ++pint;
		}
/*
 * Write the entire frame to the device.
 */
		nbytes = write(fd, (char *) &outframe, length);

		if (nbytes == ERROR) {
			terror("dlo_dev");
			kexit(errno);
		} else if (nbytes < length) {
			fprintf(stderr, "dlo_dev: write(fd, b, n) < n\n");
			kexit(errno);
		}
/*
 * Write acknowledgement byte to the device if we are not continuing
 * a virtual circuit.
 * This is needed for compatibility with dl_t4.
 */
		if (!holding) {

			if (write(fd, &ack, sizeof(ack)) < 0) {
				kexit(errno);
			}
		}
	}   /* for */
}

/*
 *	valve
 *
 *	Function:	- turns positive event value on or off by
 *			  altering the message type
 *	Accepts:	- signal number
 */

static void
valve(sig)

int4			sig;

{
	filter = (sig == SIGA) ? DLALL : DLSYS;
}
