/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	solder.c,v $
 * Revision 2.2  92/04/30  13:57:23  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/04/27  10:58:53  gdburns
 * Overhaul machine independent code.
 * 
 * Revision 2.1.1.2  92/01/31  18:22:06  gdburns
 * Look at the environment for the boot schema filename.
 * 
 * Revision 2.1.1.1  91/11/27  16:04:00  raja
 * Modified to use libtools and the new search module.
 * 
 * Revision 2.1  91/03/20  11:58:27  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/11/16  15:29:24  raja
 * Overhaul of the topology layer tools.
 * 
 * Revision 2.0.1.1  90/07/13  19:31:29  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:57:01  trillium
 * Homo erectus
 *
 *	Function:	- system specific topology reconfiguration tool
 *			- topology is specified using NaIL
 *			- NT1000 version
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 */

#include <stdio.h>
#include <portable.h>
#include <unistd.h>
#include <string.h>
#include <terror.h>
#include <typical.h>
#include <debug.h>
#include <nail.h>
#include <net.h>
#include <machine/machine.h>

/*
 * global variables
 */
char			*me;		/* tool name */

/*
 * external variables
 */
extern int		errno;
extern FILE		*yyin;		/* lex input file */

/*
 * static variables
 */
static struct bschema	nail;		/* network description */
static int		fl_debug;	/* debug option flag */
static int		fl_pretend;	/* pretend option flag */
static int		fl_verbose;	/* verbose option flag */
static char		*f_bnail;	/* boot schema filename */
static char		*fullbnail;	/* full boot schema filename */
static char		f_solder[24];	/* soldering command file */
static char		*t_solder;	/* link connect tool */
static char		**p_bnail;	/* schema pathname prefix */
static FILE		*fp_solder;	/* solder command file */

static char *usage = "usage: %s [-dhvNV] [-c <bnail>]\n";

/* 
 * external functions
 */
extern void		followed();
extern void		validopts();
extern char		*findpath();
extern char		*getenv();
extern char		*getparam();
extern char		*mktemp();
extern char		*sprintf();
extern char		*strcpy();
extern char		*strrchr();
extern int		addarg();
extern int		cnfexec();
extern int		do_args();
extern int		opt_taken();
extern int		parse_nail();
extern int		unlink();
extern int4		intparam();
extern int4		stoi();

/*
 * static functions
 */
static void		do_connect();
static void		help_solder();
static void		prt_defaults();
static void		set_defaults();
static int		check_connect();
static int		check_node();
static int		do_error();
static int		do_solder();

/*
 * main program
 */
int
main(argc, argv)  

int			argc;
char			*argv[];

{
/*
 * Assign the tool name.
 */
	if ((me = strrchr(argv[0], STRDIR)) == NULL) {
		me = argv[0];
	} else {
		me++;
	}
/*
 * Parse the command line.
 */
	validopts("dhvNV");
	followed("c");

	if ((do_args(&argc, argv) < 0) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		fprintf(stderr, usage, me);
		exit(errno);
	}

	if (opt_taken('h')) {
		help_solder();
		exit(0);
	}
/*
 * Set the mode of operation.
 */
	set_defaults();

	if (fl_debug) {
		prt_defaults();
	}
/* 
 * Locate the boot schema file.
 */
	if ((fullbnail = findpath(f_bnail, p_bnail, R_OK)) == NULL) {
		fprintf(stderr, "%s: cannot locate bnail file \"%s\"\n",
			me, f_bnail);
		exit(1);
	}
/*
 * Open the boot schema file.
 */
	DBUG("%s: opening the boot schema file: %s ...\n", me, f_bnail);
	if ((yyin = fopen(fullbnail, "r")) == NULL) {
		perror(me);
		exit(errno);
	}
/*
 * Parse the boot schema.
 */
	if (parse_nail(&nail) < 0) {
		fprintf(stderr, "%s: cannot parse the NaIL syntax\n", me);
		exit(errno);
	}
/*
 * Close the boot schema.
 */
	if (fclose(yyin) == EOF) {
		perror(me);
		exit(ERROR);
	}
/*
 * Print the boot schema out if needed.
 */
	if ((opt_taken('V')) || (fl_debug)) {
		pretty_bnail(&nail);

		if (opt_taken('V')) {
			exit(0);
		}
	}
/*  
 * Solder the network.
 */
	DBUG("%s: soldering ...\n", me);
	if (do_solder() < 0) {
		fprintf(stderr, "%s: unable to solder\n", me);
		exit(errno);
	}

	DBUG("%s: soldered OK\n", me);
	return(0);
}

/*
 *	help_solder
 *
 *	Function:	- prints the help menu
 */

static void 
help_solder()
{
	printf("\nSynopsis:\t%s [options]\n", me);
	printf("\nDescription:\tConfigure the topology of a multicomputer.\n");
	printf("\nOptions:\t-h\t\tPrint this message.\n");
	printf("\t\t-d\t\tPrint debug information.\n");
	printf("\t\t-v\t\tBe verbose.\n");
	printf("\t\t-N\t\tPretend to take action.\n");
	printf("\t\t-V\t\tPretty print the bnail file.\n");
	printf("\t\t-c <bnail>\tUse <bnail> as boot schema.\n");
}

/*
 *	set_defaults
 *
 *	Function:	- sets the values of key variables controlling
 *			  the operation of the tool
 */

static void 
set_defaults()

{
	int		n_bnail;

	n_bnail = 0;
	p_bnail = (char **) 0;
	addarg(&n_bnail, &p_bnail, "");
	addarg(&n_bnail, &p_bnail, "$TROLLIUSHOME/boot");
	addarg(&n_bnail, &p_bnail, DEFPBNAIL);
/*
 * Set the tool name.
 */
	t_solder = DEFTLCONN;
/*
 * Set the file names.
 */
	if (opt_taken('c')) {
		f_bnail = getparam('c');
	} else if (f_bnail = getenv("TROLLIUSBNAIL")) {
	} else {
		f_bnail = DEFFBNAIL;
	}
/*
 * Set the flags.
 */
	fl_verbose = opt_taken('v');
	fl_pretend = opt_taken('N');

	if (opt_taken('d')) {
		fl_debug = TRUE;
		fl_verbose = TRUE;
	} else {
		fl_debug = FALSE;
	}
}

/*
 *	prt_defaults
 *
 *	Function:	- prints the values of key variables
 *			- mostly used for debugging
 */

static void 
prt_defaults()

{
	printf("\nDefault Settings\n\n");
	printf("boot schema: %s\n", f_bnail);
	printf("debug flag: %d\n", fl_debug);
	printf("verbose flag: %d\n", fl_verbose);
	printf("pretend flag: %d\n", fl_pretend);
	printf("local solder command: %s\n", t_solder);
}

/*
 *	do_solder
 *
 *	Function:	- reconfigures the system links
 *			- system dependent code: NT1000 board version
 *			- generates nt_ctl commands
 *
 *	Returns:	- 0 or ERROR
 */

static int
do_solder()
{
	int4		i;
	int		n_cmd;
	int		min_site;
	int		tmp_site;
	int		min_exist;
	int		status;
	char		*sysnode;
	char		**v_cmd;
	struct t_node	*tnode;			/* ptr to tnode element */
	struct t_link	*tlink;			/* ptr to tlink element */
	HASH		*tnhd;			/* TNODE hash table desc. */
/* 
 * Create and open the nt_ctl temporary command file. 
 */
	(void) strcpy(f_solder, "/tmp/solderXXXXXX");
	(void) mktemp(f_solder);

	if ((fp_solder = fopen(f_solder, "w")) == NULL) {
		perror(me);
		return(ERROR);
	}
/*
 * Set the subsystem control registers.  Issue an "s m 0" command
 * where "m" is the smallest site number and issue "s # 5" commands
 * for all other site numbers "#".
 */
	min_site = MAXINT;
	min_exist = FALSE;
	tnhd = nail.bs_tnhd;
	tnode = (struct t_node *) ah_top(tnhd);

	while (tnode) {
		sysnode = (tnode->tn_specs)->ns_sysnode;
		if (sysnode) {
			if ((tmp_site = stoi(sysnode)) < min_site) {
				min_site = tmp_site;
				min_exist = TRUE;
			}
		}
		tnode = (struct t_node *) ah_next(tnhd, (void *) tnode);
	}
	if (! min_exist) {
		return(do_error("no minimum site number found"));
	}
/*
 * Issue the command for the minimum site number.
 */
	VERBOSE("s %d 0\n", min_site);
	fprintf(fp_solder, "s %d 0\n", min_site);
/*
 * Issue all the "s # 5" commands for the non-minimum site numbers.
 * This creates duplicate commands but they execute fast in this system. 
 */
	tnode = (struct t_node *) ah_top(tnhd);
	while (tnode) {
		sysnode = (tnode->tn_specs)->ns_sysnode;
		if (sysnode) {
			tmp_site = stoi(sysnode);
			if (tmp_site != min_site) {
				VERBOSE("s %d 5\n", tmp_site);
				fprintf(fp_solder, "s %d 5\n", tmp_site);
			}
		}
		tnode = (struct t_node *) ah_next(tnhd, (void *) tnode);
	}
/*
 * Issue the link reconfiguration commands.   Some links should not
 * be reconfigured (system specific) as determined by the
 * check_connect() function.
 */
	tlink = nail.bs_tlink;
	for (i = 0; i < nail.bs_ntlink; i++, tlink = tlink->tl_next) {
		if (check_connect(tlink)) {
			do_connect(tlink);
		}
	}
/*
 * Close the command file.
 */
	(void) fclose(fp_solder);
/* 
 * Execute the system soldering command.
 */
	n_cmd = 0;
	v_cmd = (char **) 0;
	addarg(&n_cmd, &v_cmd, t_solder);
	addarg(&n_cmd, &v_cmd, f_solder);

	if (fl_debug) {
		printfv(n_cmd, v_cmd);
	}

	VERBOSE("soldering ...\n");

	if (! fl_pretend) {
		status = cnfexec(v_cmd);

		if (status < 0) {
			fprintf(stderr, "%s: can't exec %s\n", me, v_cmd[0]);
			perror(me);
			exit(errno);
		} else if (status > 0) {
			fprintf(stderr, "%s: %s exited with code %d\n",
					me, v_cmd[0], status);
			exit(status);
		}
	}
/*
 * Delete the command file.
 */
	if (unlink(f_solder) < 0) {
		perror(me);
		exit(errno);
	}
 
	return(0);
}

/*
 *	check_connect
 *
 *	Function:	- checks if link connection is allowed
 *			- system specific, NT1000 version
 *	Accepts:	- ptr to link description
 *	Returns:	- TRUE or FALSE
 */

static int
check_connect(link)

struct t_link		*link;

{
	struct nspecs	*src_spec, *dest_spec;
/*
 * Check the node types being linked.  If any one is an otb node
 * the connection should be inhibited.
 */
	src_spec = (link->tl_srcnode)->tn_specs;
	dest_spec = (link->tl_destnode)->tn_specs;

	if (tisotb(src_spec->ns_type) || tisotb(dest_spec->ns_type)) {
		return(FALSE);
	}
/*
 * Check the syntax of the system specific node names.
 */
	if (! check_node(src_spec->ns_sysnode)) {
		fprintf(stderr, "%s: illegal system node syntax: %s\n", 
			me, src_spec->ns_sysnode);
		return(FALSE);
	}

	if (! check_node(dest_spec->ns_sysnode)) {
		fprintf(stderr, "%s: illegal system node syntax: %s\n", 
			me, dest_spec->ns_sysnode);
		return(FALSE);
	}
/*
 * Check if both nodes have the same host.  For now links are connected
 * only when they share the same host.  This restriction should be
 * removed later.
 */
	if (src_spec->ns_host && dest_spec->ns_host) {

		if (strcmp(src_spec->ns_host, dest_spec->ns_host)) {
			return(FALSE);
		}
	} else if (src_spec->ns_host || dest_spec->ns_host) {
		return(FALSE);
	}
/*
 * Check if the link is hardwired.  For this system, hardwired links
 * use the transputer links 1 and 2.
 */
	if ((link->tl_srclink == (int4) 1) ||
		(link->tl_srclink == (int4) 2) ||
		(link->tl_destlink == (int4) 1) ||
		(link->tl_destlink == (int4) 2)) {
		return(FALSE);
	}
/*
 * Check if the link is a duplicate.  A link is duplicate if the
 * source link number is greater than the destination link number or if
 * they are equal but the source nodeid is greater or equal to the
 * destination nodeid.  As nodeids are unique, this forms a complete
 * ordering among (node,link) pairs and the connection is allowed
 * iff (srcnode, srclink) < (destnode, destlink).
 */
	if (link->tl_srclink > link->tl_destlink) {
		return(FALSE);
	} else if (link->tl_srclink < link->tl_destlink) {
		return(TRUE);
	} else if ((link->tl_srcnode)->tn_nodeid <
		(link->tl_destnode)->tn_nodeid) {
		return(TRUE);
	} else {
		return(FALSE);
	}
}

/*
 *	do_connect
 *
 *	Function:	- issue the command to connect the link
 *			- system specific, NT1000 version
 *	Accepts:	- ptr to the link description
 */

static void
do_connect(link)

struct t_link		*link;

{
	if ( ((link->tl_srcnode)->tn_specs)->ns_sysnode == NULL) {
		fprintf(stderr, "%s (do_connect): node %d has no sysnode\n",
			me, (link->tl_srcnode)->tn_nodeid);
		return;
	}

	if ( ((link->tl_destnode)->tn_specs)->ns_sysnode == NULL) {
		fprintf(stderr, "%s (do_connect): node %d has no sysnode\n",
			me, (link->tl_destnode)->tn_nodeid);
		return;
	}

	VERBOSE("c %s.%d %s.%d\n", ((link->tl_srcnode)->tn_specs)->ns_sysnode,
		link->tl_srclink, ((link->tl_destnode)->tn_specs)->ns_sysnode,
		link->tl_destlink);

	fprintf(fp_solder, "c %s.%d %s.%d\n",
		((link->tl_srcnode)->tn_specs)->ns_sysnode, link->tl_srclink,
		((link->tl_destnode)->tn_specs)->ns_sysnode, link->tl_destlink);
}

/*
 *	check_node
 *
 *	Function:	- check system specific node name syntax
 *			- NT1000 version
 *			- allowed syntax: "site.module" or "site.e"
 *	Accepts:	- node name
 *	Returns:	- TRUE or FALSE
 */

static int 
check_node(name)

char			*name;

{
/*
 * Handle the trivial cases.
 */
	if (name  == NULL) {
		return(FALSE);
	}

	if (! isdigit((int) *name)) {
		return(FALSE); 
	}
/*
 * Skip the first integer.
 */
	for ( ; isdigit((int) *name); name++) { }
/*
 * The following character must be a '.' in a valide node name.
 */
	if (*(name++) != '.') {
		return(FALSE);
	}
/*
 * After the '.' character, the name must either end with an 'e'
 * or with another integer.
 */
	if (*name == 'e') {
		if (*(++name) == NULL) {
			return(TRUE);
		} else {
			return(FALSE);
		}
	}

	else if (isdigit((int) *name)) {
/* 
 * Scan for the second integer.
 */
		name++;

		for ( ; isdigit((int) *name); name++) { }

		if (*name == NULL) {
			return(TRUE);
		} else {
			return(FALSE);
		}
	}

	else {
		return(FALSE);
	}
}

/*
 *	do_error
 *
 *	Function:	- reports error messages from do_solder()
 *			- closes and unlinks the command file
 *	Accepts:	- the error message
 *	Returns:	- ERROR
 */

static int
do_error(message)

char			*message;

{
	fprintf(stderr, "%s (do_solder)", me);

	if (message) {
		fprintf(stderr, ": %s\n", message);
	} else {
		fprintf(stderr, "\n");
	}

	(void) fclose(fp_solder);
	(void) unlink(f_solder);
	return(ERROR);
}
