/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	tfish.c,v $
 * Revision 2.2  92/04/30  13:57:46  trillium
 * Ohio Release
 * 
 * Revision 2.1  92/02/23  22:57:52  gdburns
 * initial revision
 * 
 *
 *	Function:	- "fish"
 *			- nt1000 version
 *			- reset itb nodes associated with an otb-itb link
 *			- so named from our original reset device, fish line.
 *			- name changed to tfish so as not to conflict with
 *			  games/fish
 */

#include <fcntl.h>
#include <machine/machine.h>
#include <stdio.h>
#include <sys/ioctl.h>

#define SITE_RESET	_IO(n,0)	/* nt1000 specific */

/*
 * external functions
 */
extern int		do_args();
extern void		followed();
extern void		intparam();
extern void		validopts();

/*
 * local functions
 */
static void		help();

/*
 * local variables
 */
static char		*me;		/* name I was exec'd by */
static char		*usage = "usage: %s <device>\n";

/*
 * external variables
 */
extern int		errno;

main(argc, argv)

int			argc;
char			**argv;

{
	int		delay;		/* #secs to delay after reset */
	int		fd;		/* device file descriptor */
	char		*dreset;	/* device file name */

	me = argv[0];
/*
 * Initialize option parser.
 */
	validopts("dh");
	followed("d");

	if (do_args(&argc, argv) < 0) {
		(void) fprintf(stderr, usage, me);
		exit(errno);
	}

	if (opt_taken('h')) {
		help();
		exit(0);
	}

	if (argc > 2) {
		(void) fprintf(stderr, usage, me);
		exit(1);
	}

	dreset = (argc > 1) ? argv[1] : DEFFDEVICE;
/*
 * Open the device.
 */
	if ((fd = open(dreset, O_RDWR, 0)) < 0) {
		(void) fprintf(stderr, "%s: unable to open %s\n", me, dreset);
		perror(me);
		exit(errno);
	}
/*
 * Reset device.
 */
	if (ioctl(fd, SITE_RESET) < 0) {
		perror(me);
		exit(errno);
	}
/*
 * Delay.
 */
	if (opt_taken('d')) {
		intparam('d', &delay);
		sleep((unsigned) delay);
	}

	return(0);
}

/*
 *	help
 *
 *	Function:	- prints useful information on the tfish command
 */

static void
help()

{
	printf("\nSynopsis:	tfish [-h] [-d <delay>] [<filename>]\n");
	printf("\nDescription:	Reset itb node(s).\n");
	printf("\nOptions:	-h		this help message\n");
	printf("\t\t-d <delay>	pause after reseting\n");
	printf("\t\t<filename>	specific reset device\n");
}
