/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	btface.c,v $
 * Revision 2.2  92/04/30  13:59:19  trillium
 * Ohio Release
 * 
 *	Function:	- link interface for moses and reset
 *			- vmtm version
 */

#include <machine/machine.h>
#include <moses.h>
#include <portable.h>

/*
 * global functions
 */
void			btclock();	/* synchronize clocks */
void			btmosescall();	/* transfer control */
void			btmosescat();	/* cat moses commands */
void			btmosesload();	/* load bytes */
void			btmosesrunp();	/* spawn process */
void			btmoseswrite();	/* write one word */
int			btreset();	/* reset neighbour */

/*
 * external functions
 */
extern void		out();		/* transputer output channel */

/*
 *	btmosesload
 *
 *	Function:	- executes a moses load command across
 *			  the given link
 *	Accepts:	- link number
 *			- local buffer address
 *			- remote buffer address
 *			- buffer length
 */

void
btmosesload(ilink, lbuf, rbuf, len)

int4			ilink;
void			*lbuf;
void			*rbuf;
int4			len;

{
	char		cmd = M_LOAD;	/* moses load command */
	void		*alink;		/* link address */

	if (len == 0) {
		return;
	}
/*
 * Derive link address from link index.
 */
	alink = (void *) (0x80000000 + (ilink * 0x4));
/*
 * Output the moses command.
 */
	out(alink, &cmd, (int4) sizeof(char));
/*
 * Output length, then remote address, then local buffer.
 */
	out(alink, (char *) &len, (int4) sizeof(int4));
	out(alink, (char *) &rbuf, (int4) sizeof(char *));
	out(alink, (char *) lbuf, len);
}

/*
 *	btmoseswrite
 *
 *	Function:	- executes a moses write command across
 *			  the given link
 *	Accepts:	- link number
 *			- remote word address
 *			- data word
 */

void
btmoseswrite(ilink, raddr, dword)

int4			ilink;
int4			*raddr;
int4			dword;

{
	char		cmd = M_WRIT;	/* moses load command */
	void		*alink;		/* link address */
/*
 * Derive link address from link index.
 */
	alink = (void *) (0x80000000 + (ilink * 0x4));
/*
 * Output the moses command.
 */
	out(alink, &cmd, (int4) sizeof(char));
/*
 * Output remote address, then data word.
 */
	out(alink, (char *) &raddr, (int4) sizeof(char *));
	out(alink, (char *) &dword, (int4) sizeof(int4));
}

/*
 *	btmosesrunp
 *
 *	Function:	- executes a moses runp command across
 *			  the given link
 *	Accepts:	- link number
 *			- instruction pointer
 *			- stack pointer
 */

void
btmosesrunp(ilink, ip, wp)

int4			ilink;
void			*ip;
void			*wp;

{
	char		cmd = M_RUNP;	/* moses load command */
	void		*alink;		/* link address */
/*
 * Derive link address from link index.
 */
	alink = (void *) (0x80000000 + (ilink * 0x4));
/*
 * Output the moses command.
 */
	out(alink, &cmd, (int4) sizeof(char));
/*
 * Output instruction pointer, then stack pointer.
 */
	out(alink, (char *) &ip, (int4) sizeof(void *));
	out(alink, (char *) &wp, (int4) sizeof(void *));
}

/*
 *	btmosescall
 *
 *	Function:	- executes a moses call command across
 *			  the given link
 *	Accepts:	- link number
 *			- instruction pointer
 *			- stack pointer
 */

void
btmosescall(ilink, ip, wp)

int4			ilink;
void			*ip;
void			*wp;

{
	char		cmd = M_CALL;	/* moses load command */
	void		*alink;		/* link address */
/*
 * Derive link address from link index.
 */
	alink = (void *) (0x80000000 + (ilink * 0x4));
/*
 * Output the moses command.
 */
	out(alink, &cmd, (int4) sizeof(char));
/*
 * Output instruction pointer, then stack pointer.
 */
	out(alink, (char *) &ip, (int4) sizeof(void *));
	out(alink, (char *) &wp, (int4) sizeof(void *));
}

/*
 *	btmosescat
 *
 *	Function:	- executes a buffer of preset moses commands
 *	Accepts:	- link index
 *			- ptr to moses buffer
 *			- size of moses buffer
 */

void
btmosescat(ilink, pmoses, smoses)

int4			ilink;
char			*pmoses;
int4			smoses;

{
	void		*alink;		/* link address */
/*
 * Derive link address from link index.
 */
	alink = (void *) (0x80000000 + (ilink * 0x4));
	out(alink, pmoses, smoses);
}

/*
 *	btreset
 *
 *	Function:	- resets the neighbouring node connected
 *			  by the given link
 *	Accepts:	- link number
 */

int
btreset(ilink)

int4			ilink;

{
/*
 * Begin voodoo sequence to reset ITB neighbour on other side of the
 * given link.
 */
	*ARESET = 0;
	*ARESET = 1;
	*ARESET = 2;
	*ARESET = 3;

	*ARESET = 1 << ilink;
	tin(ldtimer() + 2);
	*ARESET = 0;

	return(0);
}

/*
 *	btclock
 *
 *	Function:	- Attempt to synchronize local and neighbouring
 *			  clocks by setting the neighbouring clock to the
 *			  local value.
 *	Accepts:	- link index
 */

void
btclock(ilink)

int4			ilink;

{
	int4		local;
	char		command;
	void		*alink;		/* link address */

/*
 * Derive link address from link index.
 */
	alink = (void *) (0x80000000 + (ilink * 0x4));
	command = M_STTIMER;
/*
 * Output the moses command.
 */
	out(alink, &command, sizeof(char));
/*
 * Read the value of the local clock and output it to the neighbour.
 */
	local = ldtimer();
	out(alink, (char *) &local, sizeof(int));
}
