/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	solder.c,v $
 * Revision 2.2  92/04/30  14:03:38  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  11:16:12  gdburns
 * Overhaul machine independent code.
 * 
 * Revision 2.1  91/04/24  13:18:11  gdburns
 * vmtm version
 *
 *	Function:	- system specific topology reconfiguration tool
 *			- topology is specified using NaIL
 *			- Parsytec VMTM version
 */

#include <debug.h>
#include <machine/machine.h>
#include <nail.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "solder.h"

/*
 * external variables
 */
extern int		errno;
extern FILE		*yyin;		/* lex input file */

/*
 * static variables
 */
static char		*me;		/* tool name */
static int		fl_debug;	/* debug option flag */
static int		fl_verbose;	/* verbose option flag */
static char		*f_bnail;	/* boot schema filename */
static char		*t_solder;	/* link connect tool */
static char		**p_tool;	/* tool pathname prefix */

static char *usage = "usage: %s [-dhvN] [-c <bschema>]\n";

/* 
 * external functions
 */
extern void		followed();
extern void		validopts();
extern char		*findpath();
extern char		*getenv();
extern char		*getparam();
extern char		*malloc();
extern int		addarg();
extern int		do_args();
extern int		opt_taken();

/*
 * static functions
 */
static void		help_solder();
static void		prt_defaults();
static void		set_defaults();

int
main(argc, argv)  

int			argc;
char			*argv[];

{
	char		*mtmsun;
	char		*vmtm;
	char		*env_mtmsun;
	char		*env_vmtm;
	char		*p;
	char		*path_mtmsun;
	char		*path_vmtm;
	char		**v_cmd;
	int		n_cmd;

/*
 * Assign the tool name.
 */
	if ((me = strrchr(argv[0], STRDIR)) == NULL) {
		me = argv[0];
	} else {
		me++;
	}
/*
 * Parse the command line.
 */
	validopts("dhvN");
	followed("c");

	if ((do_args(&argc, argv) < 0) || (argc > 1)) {
		fprintf(stderr, usage, me);
		return(1);
	}

	if (opt_taken('h')) {
		help_solder();
		return(0);
	}
/*
 * Set the mode of operation.
 */
	set_defaults();

	if (fl_debug) {
		prt_defaults();
	}
/*
 * Formulate the mtm and vmtm configuration tool names.
 */
	mtmsun = malloc((unsigned) (strlen(f_bnail) + strlen(DEFMTMSUN) + 1));

	if (mtmsun == NULL) {
		perror(me);
		return(errno);
	}

	(void) strcpy(mtmsun, DEFMTMSUN);
	p = strchr(f_bnail, '.');

	if (p != NULL) {
		(void) strcat(mtmsun, p);
		DBUG("mtmconf is %s ...\n", mtmsun);
	}

	vmtm = malloc((unsigned) (strlen(f_bnail) + strlen(DEFVMTM) + 1));

	if (vmtm == NULL) {
		perror(me);
		return(errno);
	}

	(void) strcpy(vmtm, DEFVMTM);
	p = strchr(f_bnail, '.');

	if (p != NULL) {
		(void) strcat(vmtm, p);
		DBUG("vmtmconf is %s ...\n", vmtm);
	}
/* 
 * Locate the configuration tools.
 * If they are found, assign the appropriate environment variables.
 */
	path_mtmsun = findpath(mtmsun, p_tool, R_OK);
	
	if (path_mtmsun != NULL) {
		DBUG("%s found to be %s ...\n", mtmsun, path_mtmsun);
		env_mtmsun = malloc((unsigned) (strlen(path_mtmsun) +
			strlen(ENV_MTMSUN) + 1));

		if (env_mtmsun == NULL) {
			perror(me);
			return(errno);
		}

		(void) strcpy(env_mtmsun, ENV_MTMSUN);
		(void) strcat(env_mtmsun, path_mtmsun);
		DBUG("setting mtmsun environment to %s ...\n", env_mtmsun);

		if (putenv(env_mtmsun) < 0) {
			perror(me);
			return(errno);
		}

		(void) free(env_mtmsun);
		(void) free(path_mtmsun);
	} else {
		VERBOSE("did not find %s, using default ...\n", mtmsun);
	}

	path_vmtm = findpath(vmtm, p_tool, R_OK);
	
	if (path_vmtm != NULL) {
		DBUG("%s found to be %s ...\n", vmtm, path_vmtm);
		env_vmtm = malloc((unsigned) (strlen(path_vmtm) +
			strlen(ENV_VMTM) + 1));

		if (env_vmtm == NULL) {
			perror(me);
			return(errno);
		}

		(void) strcpy(env_vmtm, ENV_VMTM);
		(void) strcat(env_vmtm, path_vmtm);
		DBUG("setting vmtm environment to %s ...\n", env_vmtm);

		if (putenv(env_vmtm) < 0) {
			perror(me);
			return(errno);
		}

		(void) free(env_vmtm);
		(void) free(path_vmtm);
	} else {
		VERBOSE("did not find %s, using default ...\n", vmtm);
	}

	(void) free(mtmsun);
	(void) free(vmtm);
/*
 * Execute the Parsytec configuration tool.
 */
	n_cmd = 0;
	v_cmd = (char **) 0;
	addarg(&n_cmd, &v_cmd, t_solder);
	addarg(&n_cmd, &v_cmd, "-c");

	if (fl_verbose) {
		addarg(&n_cmd, &v_cmd, "-v");
	}

	VERBOSE("soldering ...\n");

	if (opt_taken('N')) {
		return(0);
	}

	return(cnfexec(v_cmd));
}

/*
 *	help_solder
 *
 *	Function:	- prints the help menu
 */

static void 
help_solder()

{
	printf("\nSynopsis:\t%s [options]\n", me);
	printf("\nDescription:\tConfigure the topology of a multicomputer.\n");
	printf("\nOptions:\t-h                      Print this message.\n");
	printf("\t\t-d                      Turn on debugging.\n");
	printf("\t\t-v                      Be verbose.\n");
	printf("\t\t-N                      Only \"pretend\" to configure.\n");
	printf("\t\t-c nailfile             Use this boot schema.\n");
}

/*
 *	set_defaults
 *
 *	Function:	- sets the values of key variables controlling
 *			  the operation of the tool
 */

static void 
set_defaults()

{
	int		n_tool;
/*
 * Set the pathnames for the vendor's solder utility.
 */
	n_tool = 0;
	p_tool = (char **) 0;
	addarg(&n_tool, &p_tool, "");		/* absolute */
	addarg(&n_tool, &p_tool, ".");		/* cwd */
	addarg(&n_tool, &p_tool, DEFPTOOL);	/* default */
/*
 * Set the tool name.
 */
	t_solder = DEFTLCONN;
/*
 * Set the file names.
 */
	if (opt_taken('c')) {
		f_bnail = getparam('c');
	} else if (getenv("TROLLIUSBNAIL")) {
	} else {
		f_bnail = DEFFBNAIL;
	}
/*
 * Set the flags.
 */
	fl_debug = opt_taken('d');
	fl_verbose = opt_taken('v');
}

/*
 *	prt_defaults
 *
 *	Function:	- prints the values of key variables
 *			- mostly used for debugging
 */

static void 
prt_defaults()

{
	printf("boot schema: %s\n", f_bnail);
	printf("debug flag: %d\n", fl_debug);
	printf("verbose flag: %d\n", fl_verbose);
	printf("pretend flag: %d\n", opt_taken('N'));
	printf("t_solder = \"%s\"\n", t_solder);
}
