.TH BSEND l "January 19, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
BSEND, BRECV \- Basic Trollius message passing.
.SH SYNOPSIS
.nf
subroutine BSEND (bdestnode, bmsg, blength, bdatatype, ierror)
subroutine BRECV (bsrcnode, bmsg, blength, bdatatype, ierror);

integer bdestnode, bsrcnode, bmsg(*), blength, bdatatype, ierror
.fi
.SH DESCRIPTION
.PP
.I BSEND()
and
.I BRECV()
provide very simple message passing at the Trollius network level.
They allow the user to make point-to-point data transfers between two nodes.
These functions are meant to be used in pairs, each message sent by
.I BSEND()
being received by \fIBRECV()\fR.
.PP
.I Bdestnode
is used by
.I BSEND()
to identify the destination node running the intended receiving process, local
or remote.
.I Bsrcnode
is used by
.I BRECV()
to identify the source node running the process that is sending the message.
A receiver thus can directly specify its synchronizing sender.
The special node identifier LOCAL (see CONSTANTS(l)) can be used as it refers
to the local node.
Since the sender and receiver are only identified by the nodes on
which they run (source node and destination node), this limits the applicability
of these functions to the case of a single user process per node.
.PP
The message being transmitted takes the form of an array of data elements.
.I Bmsg
holds the address of that array, i.e. the address of the first byte of data
to be sent.
The data must be stored contiguously in memory.
.PP
.I Blength
holds the length of the message as a number of data elements.
For example, if the message to be sent is an array of 20 floating point numbers,
.I blength
should be set to 20.
If the sender and the receiver specify different lengths, the lesser amount
will be transferred.
.PP
The type of the data elements is given by
.I bdatatype.
The functions transmit the message between any two Trollius nodes.
On nodes of different types, data may have different representations.
For example, integers may be stored with the most significant byte
first in memory (big endian) or with the most significant byte last in
memory (little endian).
.I BSEND()
and
.I BRECV()
automatically modify the message data to suit the internal data
representation of each node.
The data types supported are (see CONSTANTS(l)):
.TP 14
DINTMSG
The message is an array of 32-bit integers.
.TP
DFLTMSG
The message is an array of single precision real numbers.
.TP
DDBLMSG
The message is an array of double precision real numbers.
.TP
DRAWMSG
The message representation will not be changed.
.PP
DRAWMSG is typically used when sending an array of bytes (e.g. an ASCII
text string) or when the message array contains a mixture of data types.
In this later case, the user will have to specify the size of the array
in bytes and perform any necessary changes in the data representation by
using the several byte-order translation functions available in Trollius
(see ITTOL(l)).
.SH "EXAMPLE USAGE"
.PP
The following is an example of a typical usage of
.I BSEND()
and
.I BRECV()
to transmit an array of single precision floating point numbers
between two nodes (n3 and n7).
.PP
The sender on node 3 sends the contents of the data array of length
5 to the receiver on node 7 by executing:
.nf
.ta 1i
.ft B

REAL*4	data(5)
INTEGER	ierror

CALL BSEND(7, data, 5, DFLTMSG, ierror)
.fi
.PP
The receiver on node 7, not knowing how many floating point numbers
are going to be sent, sets a maximum limit of 20.
.I BRECV()
modifies the value of the
.I blength
variable to indicates how many numbers were actually received.
For this reason,
.I BRECV()
should never be called with a constant length, always use a variable instead.
The code executed by the receiver is:
.nf
.ta 1i
.ft B

REAL*4	indata(20)
INTEGER	numdata, ierror

numdata = 20
CALL BRECV(3, indata, numdata, DFLTMSG, ierror)
.fi
.SH BLOCKING
.PP
A process calling
.I BRECV()
blocks until the message sent by the process calling
.I BSEND()
arrives.
A process calling
.I BSEND()
blocks until its message is picked by the Trollius message routing system.
The only thing that is guaranteed by a successful return from
.I BSEND()
is that the message has entirely left the calling process, thus the sender
is free to reuse the space of the message array.
In the example above, the user can modify the values in the
.I data
array after
.I BSEND()
returns.
.SH "RETURN VALUE"
.PP
Upon successful completion of \fIBSEND()\fR,
.I ierror
is set to 0, otherwise it is set to one of the error codes returned by NSEND(l).
Upon successful completion of \fIBRECV()\fR,
.I ierror
is set to 0 and
.I blength
is changed to reflect the number of data elements received (i.e. the lesser
amount between the length of the array sent and the length of the array
expected).
Otherwise,
.I ierror
is set to one of the error codes returned by NRECV(l).
.SH "SEE ALSO"
CONSTANTS(l), NSEND(l), ITTOL(l)
