.TH DSEND l "January 16, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
DSEND, DTSEND, DRECV, DTRECV  \- Trollius datalink sublevel message passing
.SH SYNOPSIS
.nf
subroutine DSEND (ndlevent, nevent, ntype, nlength, nflags, ndata,
 	ndsize, nmsg, ierror)
subroutine DTSEND (ndlevent, nevent, ntype, nlength, nflags, ndata,
 	ndsize, nmsg, ierror)

integer ndlevent, nevent, ntype, nlength, nflags, ndata(*), ndsize,
 	nmsg(*), ierror  

subroutine DRECV (nevent, ntype, nlength, nflags, ndata, ndsize, 
 	nmsg, ierror)
subroutine DTRECV (nevent, ntype, nlength, nflags, ndata, ndsize,
 	nmsg, ierror)

integer nevent, ntype, nlength, nflags, ndata(*), ndsize, nmsg(*), ierror  
.fi
.SH DESCRIPTION
.PP
These subroutines access the datalink sublevel message passing services of
the Trollius network level. 
The datalink sublevel subroutines do the actual work of network communication. 
Although they can be invoked directly, they are most often called by the
network and transport level subroutines. 
The datalink sublevel subroutines do not have built-in routing.
Instead, the user supplies information on the forwarding process.
.PP
.I DSEND()
and
.I DRECV()
are used to send messages to, or receive messages from,
nearest neighbour nodes (see "TYPICAL USAGE").
.I DTSEND()
and
.I DTRECV()
differ from
.I DSEND()
and
.I DRECV()
in that they will never cause the calling process to block.
The message is either immediately transferred, or an error is immediately
returned, indicating that the process would have blocked.
Both the sending process and the synchronizing receiving process
may use the non-blocking subroutines (unlike KSEND(l)).
.SH ARGUMENTS
.PP
The process that will receive a message sent by
.I DSEND()
is determined by the 
.I ndlevent
argument to
.I DSEND(),
and the
.I nevent
and
.I ntype
arguments to
.I DSEND()
and
.I DRECV().
.I Ndlevent
determines where the message will go directly after leaving the sender.
.I Nevent
and
.I ntype
determine which, of all the possible receivers, are appropriate
receivers for that specific message.
.PP
The
.I ndlevent
argument is used by
.I DSEND()
to a) synchronize the sender and a forwarding process
such as a link process, or b) synchronize the sender and a local receiver.
In the first case, the message will be passed if 
.I ndlevent
equals the published event number of the forwarding link process.
The link process event is returned by the integer function IDLEVT(l)
and the subroutine GROUTE(l).
In the second case, the message will be passed if
.I ndlevent
equals the receiver's
.I nevent 
argument.
This argument remains unchanged after calling
.I DSEND().
.PP
A receiver is matched to a sender by one or both of the
.I nevent
and
.I ntype
arguments.  
.I Nevent
is an arbitrary positive integer value.
In general, the values assigned by the sender and the receiver must be equal.
The exception is when the sender synchronizes with a local 
destination process, in which case the sender's
.I ndlevent
argument must match the receiver's
.I nevent
argument.
.I Nevent
remains unchanged after calling
.I DSEND()
or
.I DRECV().
.PP
.I Ntype
further filters messages that match on 
.I nevent.
A message will be passed only if the
.I ntype
arguments of the sender and receiver have at least one bit
set in an identical position.
A zero value for
.I ntype 
is a special case - that process will match any other process using the same
.I nevent.
.I Ntype
remains unchanged after calling
.I DSEND(),
but is set to the sender's
.I ntype
after calling
.I DRECV().
.PP
.I Nlength
holds the length (in bytes) of the message to be sent.
If the sender and the receiver specify different lengths,
the lesser amount will be transferred.
The maximum length of messages transferred with
.I DSEND()
and
.I DRECV()
is MAXNMSGLEN (defined in NET.h, currently set to 4096 bytes, see CONSTANTS(l)).
.I Nlength
remains unchanged after calling
.I DSEND(),
but is set to the minimum of the sender's and receiver's lengths after calling
.I DRECV().
.PP
.I Nflags
is normally set to 0, and remains unchanged after calling
.I DSEND()
or
.I DRECV().
When the NOBUF bit is set in
.I nflags,
buffers will not be used.
The KHOLD bit is set to create a virtual circuit (see "VIRTUAL CIRCUITS").
When the DFORWARD bit is set in the
.I nflags
argument to
.I DRECV(),
all common arguments will be changed to those of the sender except for
.I nmsg.
DFORWARD has no effect when used with
.I DSEND().
It is not typically used by application processes.
Flags used to ensure that the data representation is correct for the 
receiving node are discussed under "DATA REPRESENTATION".
Appropriate values for NOBUF, KHOLD, DFORWARD, and the data representation 
flags are listed in CONSTANTS(l).
.PP
.I Ndata
is a short message buffer, the length of which (in words) is specified by 
.I ndsize.
Although
.I ndata
is declared as an integer array, any type of numerical array can be used.
Character arrays may not be used.
.I Ndata
can be used instead of
.I nmsg 
(in which case
.I nlength
is set to 0), or for appending control information to the message body.
.I Ndata
is more efficient than
.I nmsg,
because the actual data is inserted into the message header, rather
than just the address.
After the message is passed, the information placed in 
.I ndata
by the sender will appear (possibly with conversion, see "DATA REPRESENTATION")
in the receiver's
.I ndata
array.
The sender's
.I ndata
array will not change.
.PP
.I Ndsize 
is the number of bytes in the
.I ndata
message buffer.
The maximum size is NHDSIZE
.B words,
which is currently set to 8 (see CONSTANTS(l)).
If different amounts are specified by
.I DSEND()
and
DRECV(), the smaller amount will be delivered.
This argument remains unchanged after calling
.I DSEND()
or
DRECV().
Thus, there is no way for the receiver to check whether the requested
amount of information arrived.
.PP
.I Nmsg
holds the actual message data.
Although
.I nmsg
is declared as an integer array, any type of numerical array can be used.
Character arrays may not be used.
The size of this array (in bytes) is set by the 
.I nlength 
argument.
.I Nmsg
remains unchanged after calling
.I DSEND().
It is filled with message data after calling
.I DRECV().
.PP
.I Ierror
is set to the error code if an error occurs.
Otherwise, it is set to 0.
.SH "TYPICAL USAGE"
.PP
.I DSEND()
and
.I DRECV()
are used for nearest neighbour message passing, when the user wishes to avoid
the expense of automatic message routing (as provided by NSEND(l)).
.PP
To synchronize the sender and the receiver, the sender must set
.I ndlevent
to the link process event connecting the destination node, and
.I nevent
to the value specified by the receiver.
The receiver must only set
.I nevent
to the same value as the sender.
.SH "DATA REPRESENTATION"
.PP
On nodes of different types, data may have different representations.
For example, integers may be stored with the most significant byte
first in memory (big endian) or with the most significant byte last in
memory (little endian).
All arguments, except
.I nmsg,
are automatically converted if passed to a node with
different data representation.
.I Ndata
is converted assuming that it holds 32-bit integers.
.PP
The
.I nflags
argument to
.I DSEND()
can be set to the following data representation flags
(they will have no effect when used with \fIDRECV()\fR).
Their values are listed on the CONSTANTS(l) man page.
Each flag assumes a data type, and will make the appropriate change in the
data representation of the given argument.
They will have no effect if data conversion is not needed.
.TP 14
DINTDATA 
.I Ndata
holds 32-bit integers.
.TP
DFLTDATA
.I Ndata
holds single precision real numbers.
.TP
DDBLDATA 
.I Ndata
holds double precision real numbers.
.TP
DRAWDATA
.I Ndata 
representation will not be changed.
.TP
DINTMSG
.I Nmsg
holds integers.
.TP
DFLTMSG
.I Nmsg
holds single precision real numbers.
.TP
DDBLMSG
.I Nmsg
holds double precision real numbers.
.TP
DRAWMSG
.I Nmsg
representation will not be changed.
.PP
Data representation can also be changed with the functions
ILTOT(l), ITTOL(l), etc.
.SH BUFFERS
.PP
If the sender does not immediately synchronize with
the forwarding process specified by 
.I ndlevent,
.I DSEND()
will secretly try to synchronize on a buffer event.
For this reason, the forwarding process identified by
.I ndlevent
should never be a buffer process.
A message forwarded to a buffer process will eventually be forwarded to
the original 
.I ndlevent.
This behavior can be disabled by setting the NOBUF bit in the
.I nflags
argument.
The user should be aware of the presence of buffer processes
even though no programming is required to utilize them.
.SH BLOCKING
.PP
A process calling
.I DRECV()
blocks until the message sent by the process calling
.I DSEND()
arrives.
A process calling
.I DSEND()
blocks only until its message is picked up by a) a local process calling
.I DRECV(),
b) a buffer process on the same node, or
c) a link process on the same node.
The only thing that is guaranteed by a successful return from
.I DSEND()
is that the message has entirely left the sender.
.I DSEND()
thus differs from KSEND(l), which blocks unless a receiver is
already waiting.
.PP
The loose blocking behavior of
.I DSEND()
introduces a fundamental danger of Trollius message passing - a sender can
transmit a message that may never be received due to programming
error or deadlock.
Messages are never dropped or timed out.
Some Trollius process will always be stuck with it, waiting for a synchronizing
.I DRECV()
that will never happen.
If that unfortunate process is a buffer, it can
be located by the user and swept clean (see sweep(l)).
However, if that process is a link proprietor,
the link is henceforth plugged and useless.
An intelligent buffering system does try to prevent this from
happening (see buffer(l)), but it remains possible.
.PP
On current transputer implementations when buffering is turned off and 
.I nlength
is not equal to 0, transputer in/out channels allow stronger synchronization
between nearest neighbour nodes than described above.
In this case, 
.I DSEND()
does not return until the synchronizing process has received the message. 
This behavior is identical to that described in KSEND(l).
.SH "VIRTUAL CIRCUITS"
.PP
The virtual circuit capability described in KSEND(l) is also available
for network messages.
A virtual circuit eliminates the need for interaction with the kernel
and other server processes on subsequent
.I DSEND()
and
.I DRECV()
calls, and thus increases the speed of message passing.
A network virtual circuit is a chain of local virtual circuits
beginning with the sender, following through link proprietor processes,
and ending with the destination receiver.
For the life of the circuit, the intervening links cannot be
used by other processes, including system processes.
Thus it is recommended that virtual circuits be used with extreme
caution and for nearest neighbour communication only.
.PP
A virtual circuit is created by setting the KHOLD bit in the
.I nflags
argument to both
.I DSEND()
and
.I DRECV().
Unlike the kernel level, there is protection against length mismatch for
network level virtual circuits.
The lesser amount specified in the
.I nlength
arguments of the sender and receiver will be transmitted.
Resetting the KHOLD flag of both message passing subroutines to 0
sends or receives the message along the virtual circuit, but
terminates the circuit after transmission.
Communicating parties must set and reset the KHOLD bit on the same message,
or both parties will block and the node may fail.
Use virtual circuits with care and only after debugging an application
without them.
Buffers are automatically disabled (NOBUF) when KHOLD is set.
.SH ERRORS
.TP 16
[EWOULDBLOCK]
.I DTSEND()
or
.I DTRECV()
failed because the message could not be immediately forwarded or received.
A call to
.I DSEND()
or
.I DRECV()
would have blocked.
.SH "SEE ALSO"
NSEND(l), KSEND(l), PSEND(l), TSEND(l)
