.TH GNET l "January 16, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
GNET, IGNDID, IGNDTP, IGORGN, IGNITB, IGNOTB, IGNALL,
IGNCST, IGNTP, GSMART
\- Get information on nodes and on the Trollius network.
.SH SYNOPSIS
.nf
subroutine GNET (node, nodetype, origin, ierror)
integer node, nodetype, origin, ierror

integer function IGNDID ()
integer function IGNDTP()
integer function IGORGN ()

integer function IGNITB ()
integer function IGNOTB ()
integer function IGNALL ()
integer function IGNCST ()
integer function IGNTP (nodetype, typemask)
integer nodetype, typemask

subroutine GSMART (ierror)
integer ierror
.fi
.SH DESCRIPTION
.PP
These functions return information obtained from the local route daemon
regarding the currently booted Trollius network.
.PP
.I GNET()
supplies the local node identifier
.I node,
the local node type
.I nodetype,
and the identifier 
.I origin
of the host node from which the Trollius multicomputer was booted. 
The node type is a combination of flags describing the capabilities of the
node.
These flags (see CONSTANTS(l) for current values) are:
.PP
.TP 12
NTOTB
OTB node
.TP
NTITB
ITB node
.TP
NTCAST
multicast definition
.TP
NTWASTE
node not available to the user
.TP
NTDISK
node has a disk
.TP
NTTUBE
node has a video display unit
.TP
NTJONES
node is a neighbour of the local node
.TP
NTBOOT
node is booted by the local node
.PP
.I Ierror
is set to the error code if an error occurs. 
Otherwise, it is set to 0.
.PP
.I IGNDID()
returns the local node identifier.
.I IGNDTP()
returns the local node type.
IGORGN()
returns the host node identifier.
.PP
.I IGNITB()
returns the number of ITB nodes that have no additional capabilities,
i.e., ITB nodes that are available to the user, and do not have an attached
disk or video display unit.
.I IGNOTB()
returns the total number of OTB nodes.
This will always be 1 for Trollius 2.2.
.I IGNALL()
returns the total number of nodes.
.I IGNCST()
returns the number of multicast definitions passing through the 
local node (see RCAST(l)). 
This includes all multicasts which are routed 
through the local node, not just those that are registered by the local node.
.I IGNTP()
is a general function that is used to determine the number of nodes whose
node types have certain bits set to certain values. 
This is a flexible tool which allows the user to obtain very specific
information about the capabilities of nodes in the system.
Use of
.I IGNTP()
will be described in detail below.
.PP
.I GSMART()
blocks the caller until the router has received its network and routing
information and can provide intelligent answers to 
.I GNET()
and GROUTE(l).
.SH NODE SELECTION()
.PP
You may need more detailed information on the number and types of nodes in the
system than just the number of ITB, OTB, or cast definitions. 
You may, for example, want to know the number of ITB nodes with attached
video displays. 
The
.I IGNTP()
function is used for this purpose.
.PP
To understand \fIIGNTP()\fR, you must realise that node types are interpreted
as bit fields, and that each node type has a different bit set. 
NTOTB is an exception: it has no bits set. 
A bitmask having all bits set, NTALL, is also available. 
The node types are thus:
.PP
.RS
.nf
.ta 1.3i 2.3i
Nodetype	Value	Bitmask

NTOTB	0	....00000
NTITB	1	....00001
NTCAST	2	....00010
NTWASTE	4	....00100
NTDISK	8	....01000
NTTUBE	16	...010000
NTALL	31	...011111
NTJONES	32	..0100000
NTBOOT	64	.01000000
.fi
.RE
.PP
The 
.I typemask 
argument to
.I IGNTP()
is used to specify which bits are of interest.
These are set to 1, all others are set to 0. 
The 
.I nodetype 
argument is used to specify what values these bits should take.  
.I IGNTP()
then compares the relevant bits (as specified by 
.I typemask) 
in the node type of each node in the system, to 
see if they have the appropriate values (as specified by 
.I nodetype).
.PP
A simple example is \fIIGNITB()\fR, which calls
.I IGNTP()
to learn the number of compute nodes that have no additional capabilities.
To do this,
.I IGNTP()
has to look at all bits. 
Thus, the 
.I typemask 
argument is NTALL. 
The values these bits can take is NTITB: all bits must equal 0 except
for the NTITB bit.
Thus,
.I IGNITB()
actually calls:
.PP
.RS
nnodes = IGNTP(NTITB, NTALL)
.RE
.PP
To learn the number of ITB nodes that have an attached video 
display, but no other capabilities, you would again want to look at all bits,
but you'd like two bits to be set: the NTITB bit and the NTTUBE bit. 
Adding these values will produce a bitmask with both bits set.
Thus, you would call:
.PP
.RS
nodetype = NTITB + NTTUBE
.br
nnodes = IGNTP(nodetype, NTALL)
.RE
.SH "RETURN VALUE"
.PP
Upon successful completion of
.I GNET()
or \fIGSMART()\fR, 0 is returned in \fIierror\fR.
Otherwise an error code is returned.
For all other functions, the values returned
upon successful completion are described above. 
If an error occurs, -1 is returned.
None of these functions have specific errors of their own,
but can incur any of the errors associated with NSEND(l)
or NRECV(l).
