.TH ILTOT l "January 18, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
FLTOTD, FLTOTF, FTTOLD, FTTOLF, ILTOT, ILTOTS, ITTOL, ITTOLS
\- Trollius network data representation translation
.SH SYNOPSIS
.nf
double precision function FLTOTD (dword)
double precision dword

single precision function FLTOTF (fword)
single precision fword

double precision function FTTOLD (dword)
double precision dword

single precision function FTTOLF (fword)
single precision fword

integer function ILTOT (iword)
integer iword

integer*2 function ILTOTS (sword)
integer*2 sword

integer function ITTOL (iword)
integer iword

integer*2 function ITTOLS (sword)
integer*2 sword
.fi
.SH DESCRIPTION
.PP
The representation and storage of basic data types may not be uniform
across the nodes in any given Trollius network. 
Often the OTB node uses a different byte order than the ITB nodes.
As a result, message and file data must frequently
be converted to the appropriate representation.
Conversion is not necessary if the processes involved are on the same node.
.PP
The functions convert the data representation between the local
format, regardless of what that is, and a standard (Trollius) format.
If no conversion is necessary, the functions will have no effect.
.PP
When sending messages between nodes with different data representation,
only the message data, i.e., what is sent in the 
.I nmsg
argument to NSEND(l), must be converted.
All other arguments, including \fIndata\fR (which is assumed to hold
word length integers), are automatically converted when the message is passed.
Data conversion functions should be used by both the 
sending and receiving processes. 
Data representation should be converted from local to Trollius
representation before sending, and converted from Trollius to local
representation after receiving.  
.PP
As an alternative to the byte order translation functions, the programmer 
can set various bits in the
.I nflags
argument of NSEND(l) to convert the
.I nmsg
data, or to convert
.I ndata
in other than the default manner. 
.PP
File data should be converted from
local to Trollius representation before writing, and from Trollius to local 
representation after reading.  
.PP
The conversion functions are:
.TP 10
ITTOL()
Converts a word length integer from Trollius to local representation.
.TP
ILTOT()
Converts a word length integer from local to Trollius representation.
.TP
ITTOLS()
Converts a half-word integer from Trollius to local representation.
.TP
ILTOTS()
Converts a half-word integer from local to Trollius representation.
.TP
FTTOLF()
Converts a single precision real from Trollius to local representation.
.TP
FLTOTF()
Converts a single precision real from local to Trollius representation.
.TP
FTTOLD()
Converts a double precision real from Trollius to local representation.
.TP
FLTOTD()
Converts a double precision real from local to Trollius representation.
.SH BUGS
.PP
These functions were designed for 32 bit machines and need to be redesigned
for 16 and 64 bit machines.
Outside of 32 bit machines they are guaranteed to cause confusion.
.SH "SEE ALSO"
NSEND(l), IRW(l)
