.TH KSEND l "January 18, 1992" "Release 2.2" "TROLLIUS LOCAL LIBRARY"
.SH NAME
KSEND, KTSEND, KRECV, KTRECV  \- Send and receive local node
Trollius messages.
.SH SYNOPSIS
.nf
subroutine KSEND (kevent, ktype, klength, kflags, kmsg, ierror)
subroutine KTSEND (kevent, ktype, klength, kflags, kmsg, ierror)
subroutine KRECV (kevent, ktype, klength, kflags, kmsg, ierror)
subroutine KTRECV (kevent, ktype, klength, kflags, kmsg, ierror)

integer kevent, ktype, klength, kflags, kmsg(*), ierror
.fi
.SH DESCRIPTION
These functions access Trollius local message passing services.
.I KSEND()
attempts to send a message to another process on the same node.
.I KRECV()
attempts to receive a message from another process on the same node.
The Trollius kernel arranges for the two processes to meet.
The process calling
.I KSEND()
always blocks unless a receiver is already waiting, blocked on
.I KRECV().
.PP
.I KTSEND()
and
.I KTRECV()
are similar to
.I KSEND()
and \fIKRECV()\fR, but they never block.
The message is transferred if a synchronizing process is waiting.
If not,
.I ierror
is set to EWOULDBLOCK.
Never use the non-blocking functions on both sides, or they will never meet.
.PP
The
.I kevent
argument serves to synchronize the message passing.
The values passed to
.I KSEND()
and
.I KRECV()
must be equal.
This argument remains unchanged after calling
.I KSEND()
or \fIKRECV()\fR.
.PP
.I Ktype
filters messages further.
A message will be passed only if the 
.I ktype 
arguments to
.I KSEND()
and
.I KRECV()
have at least one bit set in an identical position.
A zero value for
.I ktype 
is a special case - it will ``fit'' any other caller using the same
.I kevent.
This argument remains unchanged after calling \fIKSEND()\fR,
but is set to the sender's
.I ktype
after calling \fIKRECV()\fR.
.PP
.I Klength
holds the length (in bytes) of the message to be sent.
If the sender and the receiver specify different lengths,
the lesser amount will be transferred.
This argument is set to the minimum of the sender's and receiver's length
after calling
.I KSEND()
or \fIKRECV()\fR.
.PP
.I Kflags
is normally set to 0.
It can be set to KHOLD (see CONSTANTS(l)) to request that a virtual 
circuit be kept open for the next message on the same event.
.PP
.I Kmsg
holds the actual message data.
Although
.I kmsg
is declared as an integer array, any type of numerical array can be used.
Character arrays may not be used.  
The size of this array (in bytes) is set by the
.I klength
argument.
.I Kmsg
remains unchanged after calling \fIKSEND()\fR.
It is filled with the message data after calling \fIKRECV()\fR.
.PP
.I Ierror
is set to the error code if an error occurs.  
Otherwise, it is set to 0.
.SH "VIRTUAL CIRCUITS"
.PP
In Trollius, the kernel is not an executive with complete control of
the processor, but a server providing a standardized rendezvous service
on top of a lower layer of process management.
On Trollius ITB nodes, this lower layer can be implemented by hardware.
On Trollius OTB (host) nodes, this lower layer is implemented by the
native operating system.
.PP
Using the Trollius kernel is convenient, powerful, portable and expensive.
Therefore, virtual circuits have been provided to bypass the kernel
once an initial rendezvous has been made.
This is done by setting the KHOLD bit in the
.I kflags
argument to both
.I KSEND
and \fIKRECV()\fR.
KHOLD causes
.I KSEND()
and
.I KRECV()
to remember the rendezvous information learned from the kernel and
associate it with future uses of the same event, without consulting
the kernel again.
After the first message establishing the virtual circuit, there is no
protection against mismatched message lengths.  
If the sender and receiver specify different values in the
.I klength
argument, the program will almost certainly fail.
.PP
A virtual circuit is bi-directional, and there can be up to 16 simultaneous
virtual circuits for each process.
The restriction (for the sake of performance) is that all active
virtual circuit events must differ in the last four bits.
.PP
Using a virtual circuit event with the KHOLD bit reset (0) sends
or receives the message along the virtual circuit, but terminates
the circuit after transmission.
Communicating parties must set and reset a virtual circuit on the
same message, or both parties will block and the node may fail.
Use virtual circuits with care only after debugging an application without them.
.SH ERRORS
.TP 23
[EWOULDBLOCK]
.I KTSEND()
or
.I KTRECV()
was called and returned an error condition because the message
could not be immediately sent.
No process is blocked on a matching event and type.
At that time, a call to
.I KSEND()
or
.I KRECV()
would have blocked.
.TP
[ENOEVDESCRIPTORS]
The Trollius kernel ran out of internal message descriptors.
.SH "SEE ALSO"
NSEND(l), NRECV(l)
