.TH NSEND l "February 11, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
NSEND, NTSEND, NRECV, NTRECV  \- Trollius network sublevel message passing
.SH SYNOPSIS
.nf
subroutine NSEND (nnode, nevent, ntype, nlength, nflags, ndata, 
	ndsize, nmsg, ierror)
subroutine NTSEND (nnode, nevent, ntype, nlength, nflags, ndata, 
	ndsize, nmsg, ierror)

integer nnode, nevent, ntype, nlength, nflags, ndata(*), ndsize, 
	nmsg(*), ierror

subroutine NRECV (nevent, ntype, nlength, nflags, ndata, ndsize, 
	nmsg, ierror)
subroutine NTRECV (nevent, ntype, nlength, nflags, ndata, ndsize, 
	nmsg, ierror)

integer nevent, ntype, nlength, nflags, ndata(*), ndsize, nmsg(*), 
	ierror
.fi
.SH DESCRIPTION
.PP
There are four sublevels of message passing within the Trollius network
(inter-node) level.  
They are, in order of increasing functionality,
the physical, datalink, network, and transport sublevels.
The network sublevel functions are the most commonly used, most generic 
communication functions in Trollius.  
They provide routing services and invoke datalink sublevel functions.
.PP
.I NSEND()
and 
.I NRECV()
send messages to, and receive messages from, any process on any node.  
If no receiver, sender, or forwarding process is ready,
the process calling 
.I NSEND()
or
.I NRECV()
will block (more on this below).
.I NTSEND()
and
.I NTRECV()
differ from
.I NSEND()
and \fINRECV()\fR,
in that they will never cause the calling process to block.
The message is either immediately transferred,
or an error is immediately returned, indicating that the
process would have blocked.
Both the sender and the synchronizing receiver
may use the non-blocking functions (unlike KSEND(l)).
.SH ARGUMENTS
.PP
The
.I nnode
argument to
.I NSEND()
identifies the node running the intended receiver, local or remote.
The special node identifier LOCAL (see CONSTANTS(l)) refers to the local node 
and causes
.I NSEND()
to bypass its routing step.
.I Nnode
remains unchanged after calling \fINSEND()\fR.
Note that
.I NRECV()
does not directly specify its synchronizing sender.
Instead, the receiver on the correct node is synchronized to the sender by their
.I nevent
and
.I ntype
arguments.
.PP
A message can be sent to, or received from, a group of nodes.
The nodes must first be registered as a multicast destination with RCAST(l).  
The cast identifier can then be used as the
.I nnode
argument to \fINSEND()\fR.
See RCAST(l) for more information.
Two casts are predefined and do not need to be registered with RCAST(l).  
HOST2ALL identifies a multicast from the origin OTB node to all Trollius nodes.
HOST2ITB identifies a multicast from the origin OTB 
node to all ITB nodes.
See CONSTANTS(l) for exact values.
.PP
.I Nevent
is an arbitrary positive integer value, used
as a second level of synchronization after 
.I nnode.
The values passed to
.I NSEND()
and
.I NRECV()
must be equal.  
This argument remains unchanged after calling either subroutine.
.PP
The
.I ntype
argument further filters messages that match on 
.I nevent.
A message will be passed only if the 
.I ntype 
arguments passed to
.I NSEND()
and
.I NRECV()
have at least one bit set in an identical position.
A zero value for
.I ntype 
is a special case \- that process will match any other process using the same
.I nevent.
The
.I ntype
argument remains unchanged after calling \fINSEND()\fR,
but is set to the sender's
.I ntype
after calling \fINRECV()\fR.
.PP
.I Nlength
holds the length (in bytes) of the message to be sent.
If the sender and the receiver specify different lengths,
the lesser amount will be transferred.
The datalink functions called by
.I NSEND()
have a maximum message length of MAXNMSGLEN (see CONSTANTS(l)).
Messages exceeding this length will be implicitly broken down into
multiple smaller messages, which will be implicitly recombined by the receiver.
In this way, one call to
.I NSEND()
can introduce many messages into the network, a possible source of confusion
during debugging.
.I Nlength
remains unchanged after calling \fINSEND()\fR,
but is set to the minimum of the sender's and receiver's lengths after calling
\fINRECV()\fR.
.PP
.I Nflags
is normally set to 0, and remains unchanged after calling either subroutine.
When the NOBUF bit is set in \fInflags\fR, buffers will not be used.  
The KHOLD bit is set to create a virtual circuit (see "VIRTUAL CIRCUITS").
The NREEL bit of the sender is set for multireeling (see RCAST(l)).
Flags used to ensure that the data representation is correct for the receiver's
node are discussed under "DATA REPRESENTATION".  
For the precise values of NOBUF, KHOLD, NREEL, and the data representation
flags, see CONSTANTS(l).
.PP
.I Ndata
is a short message buffer, the length of which (in bytes) is specified by 
.I ndsize.
Although
.I ndata
is declared as an integer array, any type of numerical array can be used.
Character arrays may not be used.
.I Ndata
can be used instead of
.I nmsg 
(in which case
.I nlength
is set to 0),
or for appending control information to the message body.
.I Ndata
is more efficient than
.I nmsg,
because the actual data is inserted into the message header, rather
than just the address.
After the message is passed, the information placed in 
.I ndata
by the sender will appear (possibly with conversion) in the receiver's
.I ndata
array.  
The sender's
.I ndata
argument will not change.
.PP
.I Ndsize 
is the number of bytes in the
.I ndata
message buffer.  
The maximum size is NHDSIZE words, which is currently set to 8
(see CONSTANTS(l)).
If different amounts are specified by
.I NSEND()
and \fINRECV()\fR, the smaller amount will be transferred.
This argument remains unchanged after calling either subroutine.
Thus, there is no way for the receiver to check whether 
the requested amount of information arrived.
.PP
.I Nmsg
holds the actual message data.  
Although
.I nmsg
is declared as an integer array, any type of numerical array can be used.
Character arrays may not be used.
The size of this array (in bytes) is set by the 
.I nlength 
argument.
.I Nmsg
remains unchanged after calling \fINSEND()\fR.
It is filled with message data after calling \fINRECV()\fR. 
.PP
.I Ierror
is set to the error code if an error occurs, otherwise it is set to 0.
.SH "EXAMPLE USAGE"
.PP
The following is an example of a typical usage of the network communication 
subroutines for a message passed between two nodes with similar data 
representations, utilizing a minimum level of message filtering.
This is intended only as a summary of a simple case.  
Many variations can be constructed using the detailed information 
given in the above section.
.PP
The sender sets
.I nnode
to the identifier of the receiver's node,
.I nevent 
to an arbitrary integer,
.I ntype 
to 0,
.I nlength
to the number of bytes of data in the
.I nmsg
array,
.I nflags
to 0,
.I ndata
to 0,
.I ndsize
to 0, and
.I ierror
to an uninitialized integer variable.
The elements of the 
.I nmsg
array hold the data to be sent.
These arguments are passed to \fINSEND()\fR.
.PP
The receiver sets
.I nevent
to the same integer specified by the sender,
.I ntype
to 0,
.I nlength
to the amount of data to be received,
.I nflags
to 0,
.I ndata
to 0,
.I ndsize
to 0, and
.I ierror
to an uninitialized integer variable.
The elements of the
.I nmsg
array are unassigned.
The arguments are passed to \fINRECV()\fR.
After the message is passed, the receiver's 
.I nmsg
array holds information passed by the sender, and the receiver's
.I nlength
argument is set to the minimum of the 
.I nlength
arguments passed to \fINSEND()\fR and \fINRECV()\fR.
.I Ierror 
is set to 0 for both parties,
indicating that the subroutines incurred no errors.
.SH "DATA REPRESENTATION"
.PP
On nodes of different types, data may have different representations.
For example, integers may be stored with the most significant byte
first in memory (big endian) or with the most significant byte last in
memory (little endian).
All arguments, except
.I nmsg,
are automatically converted if passed to a node with
different data representation.
.I Ndata
is converted assuming that it holds 32-bit integers.
.PP
The
.I nflags
argument to
.I NSEND()
can be set to the following data representation flags
(they will have no effect when used with \fINRECV()\fR).
Their values are listed on the CONSTANTS(l) man page.
Each flag assumes a data type, and will make the appropriate change in the
data representation of the given argument.
They will have no effect if data conversion is not needed.
.TP 14
DINTDATA 
.I Ndata
holds 32-bit integers.
.TP
DFLTDATA
.I Ndata
holds single precision real numbers.
.TP
DDBLDATA 
.I Ndata
holds double precision real numbers.
.TP
DRAWDATA
.I Ndata 
representation will not be changed.
.TP
DINTMSG
.I Nmsg
holds integers.
.TP
DFLTMSG
.I Nmsg
holds single precision real numbers.
.TP
DDBLMSG
.I Nmsg
holds double precision real numbers.
.TP
DRAWMSG
.I Nmsg
representation will not be changed.
.PP
Data representation can also be changed with the functions
ILTOT(l), ITTOL(l), etc.
.SH BLOCKING
.PP
A process calling
.I NRECV()
blocks until the message sent by the process calling
.I NSEND()
arrives.
A process calling
.I NSEND()
blocks only until its message is picked up by a) a local process calling
.I NRECV(),
b) a buffer process on the same node, or
c) a link process on the same node.
The only thing that is guaranteed by a successful return from
.I NSEND()
is that the message has entirely left the sender.
.I NSEND()
thus differs from KSEND(l), which blocks unless a receiver is
already waiting.
.PP
The loose blocking behavior of
.I NSEND()
introduces a fundamental danger of Trollius message passing - a sender can
transmit a message that may never be received due to programming
error or deadlock.
Messages are never dropped or timed out.
Some Trollius process will always be stuck with it, waiting for a synchronizing
.I NRECV()
that will never happen.
If that unfortunate process is a buffer, it can
be located by the user and swept clean (see sweep(l)).
However, if that process is a link proprietor,
the link is henceforth plugged and useless.
An intelligent buffering system does try to prevent this from
happening (see buffer(l)), but it remains possible.
.PP
On current transputer implementations when buffering is turned off and 
.I nlength
is not equal to 0, transputer in/out channels allow stronger synchronization
between nearest neighbour nodes than described above.
In this case, 
.I NSEND()
does not return until the synchronizing process has received the message. 
This behavior is identical to that described in KSEND(l).
.SH "VIRTUAL CIRCUITS"
.PP
The virtual circuit capability described in KSEND(l) is also available
for network messages.
A virtual circuit eliminates the need for interaction with the kernel
and other server processes on subsequent
.I NSEND()
and
.I NRECV()
calls, and thus increases the speed of message passing.
A network virtual circuit is a chain of local virtual circuits
beginning with the sender, following through link proprietor processes,
and ending with the destination receiver.
For the life of the circuit, the intervening links cannot be
used by other processes, including system processes.
Thus it is recommended that virtual circuits be used with extreme
caution and for nearest neighbour communication only.
.PP
A virtual circuit is created by setting the KHOLD bit in the
.I nflags
argument to both
.I NSEND()
and
.I NRECV().
Unlike the kernel level, there is protection against length mismatch for
network level virtual circuits.
The lesser amount specified in the
.I nlength
arguments of the sender and receiver will be transmitted.
Resetting the KHOLD flag of both message passing subroutines to 0
sends or receives the message along the virtual circuit, but
terminates the circuit after transmission.
Communicating parties must set and reset the KHOLD bit on the same message,
or both parties will block and the node may fail.
Use virtual circuits with care and only after debugging an application
without them.
Buffers are automatically disabled (NOBUF) when KHOLD is set.
.SH TRACING
.PP
The network sublayer functions are capable of generating traces that
record the time of nature of their activity.
Tracing is enabled for a process by the -t option of loadgo(l).
Tracing is provided at the cost of one word of storage in the data pouch.
If trace generation is enabled, the
.I ndata[7]
field of the data pouch is not available to the program.
.SH ERRORS
.TP 16
[EWOULDBLOCK]
.I NTSEND()
or
.I NTRECV()
failed because the message could not be forwarded or received, 
respectively.  
A call to
.I NSEND()
or
.I NRECV()
would have blocked.
.TP
[EBADNODE]
The specified destination node does not exist.
.SH "SEE ALSO"
DSEND(l), KSEND(l), PSEND(l)
