.TH RCAST l "January 18, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
RCAST, RMELT, RCASK \- Register or remove a Trollius multicast destination.
.SH SYNOPSIS
.nf
subroutine RCAST (node, castid, combine, pcast, ncast, ierror)
integer node, castid, combine, pcast(*), ncast, ierror

subroutine RMELT (node, castid, ierror)
integer node, castid, ierror

subroutine RCASK (node, castid, numbr, ierror)
integer node, castid, numbr, ierror
.fi
.SH DESCRIPTION
.PP
Multicasting and multireeling allow messages to be sent to, or received
from, a group of nodes, using NSEND(l) and NRECV(l).
This is achieved by defining the group of nodes by a multicast identifier,
which can then be used as the destination node argument to NSEND(l). 
There can be only one participating process per node per multicast
at any one time.
The multicast definition also includes a root node, from which
messages will be sent or receiving to the multicast group.
The root node itself may also be a member of the multicast group.
The same multicast group with another root node requires another
multicast identifier.
.PP
Multicast identifiers can also be used as the destination node argument to many
remote functions, requesting service on more than one node in parallel.
.PP
.I RCAST()
must be used before general message passing to register the multicast
identifier with all the nodes needed to participate in a multicast or multireel.
.I RMELT()
is the inverse of RCAST.
It de-registers the multicast identifier from all participating nodes.
.I RCASK()
may be used in multireeling (see below).  
It reports the number of messages that are expected to arrive at a 
given multicast member on the way to the root node.
.PP
The
.I node
argument is the root node identifier of the multicast.
.I Castid
is the multicast identifier.
.I Combine
indicates the combining function to be used during multireeling.
.I Pcast
is an array of node identifiers, the members of the multicast group, and
.I ncast
is the number of members.
.I Ierror
is set upon return of the subroutine to the error code if an error occurs
or 0 if the operation was successful.
Since the caller of
.I RCAST
is typically running on the root node of the multicast,
.I Node
is usually set to LOCAL (see CONSTANTS(l)).
The value of
.I combine 
is unimportant if multireeling is not intended.
.I RMELT()
and
.I RCASK()
require the root
.I node,
the
.I castid,
and 
.I ierror.  
.I RCASK()
sets
.I numbr
to the expected number of multireel messages at the \fInode\fR.
.SH "MULTICAST REGISTRATION"
.PP
.I RCAST()
operates as follows.
A registration message is sent to the root node.
The root node decides which outgoing multicomputer links are used to send
messages to each member in the multicast group.
The multicast members are then partitioned according to outgoing link.
This information is remembered locally.
The registration message is then propagated across the links for which there
is a non-zero list of multicast members, to neighbouring nodes.
The neighbour nodes proceed through the steps performed by the
root node, subdividing their sub-group of multicast members, storing 
information, and propagating the registration message.
In this way,
.I RCAST()
proceeds through several recursive multicast registration requests.
.SH MULTICASTING
.PP
Assuming the multicast is registered, a process multicasts a message by
calling NSEND(l) with the destination node argument set to the multicast
identifier, \fIcastid\fR, instead of the usual single destination node.
The multicast nodes each receive a copy of the message with NRECV(l).
Messages are further synchronized on event and type, as usual.
.PP
Three multicasts originating from the origin OTB node are predefined,
and do not need to be registered with \fIRCAST\fR.
HOST2ALL identifies a multicast from the origin OTB node to all Trollius nodes;
HOST2ITB identifies a multicast from the origin OTB to all ITB nodes;
and HOST2OTB identifies a multicast from the origin OTB to all OTB nodes.
HOST2OTB is meaningless in version 2.2.
It will be utilized in version 2.3.
See CONSTANTS(l) for the precise values of these constants.
If you wish to multireel using a predefined multicast identifier, you must
multicast first to trigger the implicit registration.
.PP
Multicast messages are naturally more efficient than multiple messages
due to their multiple forwarding characteristic.
However, multicast messages incur an added overhead because they
must pass through an additional process, the cast daemon, on each node
of the journey.
.SH MULTIREELING
.PP
In a multireel, the multicast nodes reverse the operation of a 
multicast, i.e., they all send a message to the root node.
The root node calls NRECV(l).
Each of the multicast nodes call NSEND(l) with the
.I castid
as the destination node.
They must also include NREEL in the
.I nflags
argument to NSEND(l).
.I Nlength
is set to a number no greater than MAXNMSGLEN, currently set to
4096 bytes (see CONSTANTS(l)).  
.PP
Note that although the root node registers 
.I castid, 
the multicast nodes use
.I castid.  
This makes it imperative that the call to 
.I RCAST()
be completed before the multicast nodes call NSEND(l).
The recommended procedure is for the root node to register the 
.I castid,
and multicast an initial synchronizing message to the nodes before
starting the multireel.
In this way the multicast nodes cannot initiate the multireel until the
multicast is completed, which occurs after the multicast registration.
.PP
A multireel operates as follows.
When a node receives two messages from nodes of the multicast, and
each time it receives a subsequent message, combines them into a single message.
When all the multicast nodes (defined at that node, see the registration
description above) have been heard, the resulting combined message
is forwarded to the parent neighbour node (recalled from the original
multicast definition).
Eventually, a single message is forwarded to the root node.
.PP
The function used to combine message information is specified in the
.I combine
argument to \fIRCAST()\fR.
Several built-in combining functions that operate on integers (or, more
specifically, on word length items interpreted as integers) are identified by
the codes listed below (see CONSTANTS(l)).
These functions combine both the 
.I ndata
and 
.I nmsg
arguments to NSEND(l).  
All the functions combine all the elements of the 
.I ndata
and 
.I nmsg
arrays, except for CBN_ERR.
.PP
.TP 10
CBN_ERR
Combine error codes in the first integer only, and retain non-zero value.  
If both values are non-zero, either one may be retained.
.TP 
CBN_ADD
Combine by adding all corresponding integers.
.TP
CBN_MIN
Combine by taking the minimum.
.TP
CBN_MAX
Combine by taking the maximum.
.TP
CBN_AND
Combine by applying a bitwise AND.
.TP
CBN_OR
Combine by applying a bitwise OR.
.TP
CBN_XOR
Combine by applying a bitwise XOR.
.PP
If a multireel is to be used purely for synchronization, use CBN_ERR and set
.I nlength
to 0.
.PP
The user can choose to supply their own combining process.
When registering the multicast identifier with \fIRCAST\fR, the 
.I combine
argument should be set to the positive event used to synchronize messages
passed to and from the combining process.
The combining process should be written in the form of a server, and
should be dedicating strictly to the purpose of combining.
It should loop endlessly, first receiving two messages on the registered
event with NRECV(l), then combining the two messages, and then sending 
the combined message, on the same event, to the LOCAL node.
The
.I ndata
argument, the
.I nmsg
argument, or both may be combined, as desired.  
All messages must have the same length, and the length must not 
change as the messages are combined.
In order to ensure that the combining process runs on every node on the
multireel route that needs to do combining, the combining process should run
on every ITB node.
.PP
Another method is to receive and combine the reeled messages directly
in the user program.
.I Combine
should be set to the event that will be used to receive incoming multireel
messages.
All nodes in the multicast, and all nodes along the route to the destination,
should use
.I RCASK()
to determine the number of messages to be combined locally.
This would include messages received from branches of the multicast, and
from the local node if it is a multicast member.
The user program should then repeat the combining loop described above for
local combining processes
.B "one time less"
than the number returned by \fIRCASK()\fR.
The final, combined message should be sent out on the same event, with the
.I nnode
argument of NSEND(l) set to LOCAL.
This method of combining should not be used with the NOBUF bit of the
.I nflags
argument set.
.I RCASK()
must be called before sending the message to be combined, or deadlock
may result.
.SH ERRORS
.PP
As well as the errors listed below,
.I RCAST()
can incur any of the errors associated with NSEND(l) or NRECV(l).
.PP
.TP 16
[ECOLLIDE]
The cast identifier is already defined.
.TP
[EFULL]
A node cannot hold any more multicast definitions.
.TP
[EBADNODE]
One of the multicast member nodes does not exist.
.SH BUGS
.PP
Problems may arise if more than one process uses the same multicast identifier.
This should be avoided for Release 2.2.
In addition, a given process should never have two simultaneous 
multireels running under the same multicast identifier.  
Finally, simultaneous multireels on overlapping nodes, even if 
given different multicast identifiers, will not work.
To be safe, avoid using the same multicast identifier for multiple multireels,
and synchronize multireeling with the root node.  
This can be done by multicasting first.
.SH "SEE ALSO"
NSEND(l)
