.TH ALL_HASH l  "April 8, 1992" "Release 2.2" "TROLLIUS INTERNALS"
.SH NAME
.B all_hash, all_shash
\- general purpose hash table processing package
.SH SYNOPSIS
.nf
.ta .7i 1.2i
#include <all_hash.h>

HASH	*ah_init(int size, int elemsize, int nullkey, int mode);
int	ah_delete(HASH *ahd, int key);
int	ah_expand(HASH *ahd, int newsize);
int	ah_insert(HASH *ahd, void *elem);
int	ah_kick(HASH *ahd, void *elem);
int	ah_count(HASH *ahd);
int	ah_size(HASH *ahd);
void	*ah_find(HASH *ahd, int key);
void	*ah_top(HASH *ahd);
void	*ah_next(HASH *ahd, void *elem);
void	ah_free(HASH *ahd);

SHASH	*ahs_init(int size, int elemsize, int nullkey, int mode,
		void *htable, int *lru, SHASH *ahsd);
int	ahs_delete(SHASH *ahsd, int key);
int	ahs_insert(SHASH *ahsd, void *elem);
int	ahs_kick(SHASH *ahsd, void *elem);
int	ahs_count(SHASH *ahsd);
int	ahs_size(SHASH *ahsd);
void	*ahs_find(SHASH *ahsd, int key);
void	*ahs_top(SHASH *ahsd);
void	*ahs_next(SHASH *ahsd, void *elem);
.SH DESCRIPTION
.I All_hash
and
.I all_shash
are general purpose hash table processing packages.
They differ only in the way memory is allocated for a hash table.
.I All_hash,
the dynamic package, obtains memory from malloc(l)(3) whenever a new hash
table is created or its size expanded and returns memory with free(l)(3)
whenever a hash table is destroyed.
.I All_shash,
the static package, requires that the caller provide memory for the maximum
number of hash table entries when the table is first created.
Functions that operate on a dynamic hash table are named
.I ah_*
and functions that operate on a static hash table are named
.I ahs_*.
.PP
A hash table is created and initialized with the
.I ah_init()
or
.I ahs_init()
functions which both return a pointer to a hash table descriptor, typedef
.I HASH
or
.I SHASH
respectively.
The hash table descriptor pointer is used in all subsequent hash table
operation functions.
In the static function,
.I ahs_init(),
the caller supplies space not only for the maximum number of hash table
entries, but for the hash table descriptor and the LRU counters when
this mode of operation is used.
The LRU (Least Recently Used) counters keep track of the number of accesses
made to each hash table entry and are used when the AHLRU mode is chosen.
This enables
.I ah_kick()
to pick the least recently used entry as a good
candidate to be replaced by the new entry to be stored in the table.
If the AHFIXED mode is chosen, the LRU counters are not required.
In this mode, a new entry replaces the old entry at the first location it
is hashed to.
The
.I mode
parameter is a bit\-mapped flag, each flag controlling some characteristic
of the hash table.
.I Mode
values are constructed by ORing flags from the following list:
.TP 15
AHLRU
The least recently used entry is overwriten if the hash table is full when
.I ah_kick()
is called.
.TP
AHFIXED
The first hashed location is overwriten if the hash table is full when
.I ah_kick()
is called.
.TP
AHNOINIT
The hash table is not initialized.
Usually, if the data structures are
properly designed (data hiding), this mode of operation would not be required.
If used, the burden of properly initializing the hash table entries rests
on the user.
.PP
A dynamic hash table is freed with the
.I ah_free()
function.
A static hash table is simply forgotten, since the caller is responsible
for all the memory involved.
Allocating the space for a static hash table is straight forward.
The user needs to allocate two separate arrays of equal number of entries.
One,
.I htable,
is the actual hash table and each of its entries is a user-defined structure.
The second,
.I lru,
is only used if the AHLRU mode is chosen and consists of an array of integers
(32\-bit integers) each being used as a counter for an entry in the hash table.
If the mode of operation is set to AHFIXED, there is no need to allocate the
.I lru
array.
An example of how to allocate space for a static hash table for the AHLRU mode
of operation is given below:
.RS
.sp 1
.nf
struct myelement htable[31];
int lru[31];
SHASH ahsd;
#define ELEMSIZE sizeof(struct myelement)
ahs_init(31, ELEMSIZE, -1, AHLRU, htable, lru, &ahsd);
.fi
.sp 1
.RE
Thirty one elements of type
.I myelement
are allocated and named
.I htable
and thirty one 32\-bit counters are allocated and named
.I lru.
.PP
The following functions operate on dynamic hash tables:
.TP 15
ah_init()
Allocate and initialize a dynamic hash table.
A hash table descriptor pointer is returned, but 0 is returned if allocation
fails.
The caller supplies the total number of entries in the hash table, the
size of each element, the value in the key of an empty element, and the
mode of operation.
The size of the element must be at least the size of an integer (32 bits)
in order to be able to contain the hashing key.
.TP
ah_delete()
Delete an existing element from a hash table.
The element is specified by its key.
The function returns -1 and sets
.I errno
to EDELETE if the element could not be found in the given hash table.
.TP
ah_insert()
Insert a new element into a hash table.
The caller prepares and supplies a pointer to the new element.
The function copies the contents of the caller supplied element into
the appropriate space in the hash table.
The caller can reuse the element.
.I Ah_insert()
returns -1 and sets
.I errno
to EFULL if the hash table has no empty slots to store the element.
.TP
ah_free()
Free all allocated memory in a dynamic hash table including the hash
table descriptor.
The hash table is effectively blown away.
The hash table descriptor pointer is no longer valid.
.TP
ah_find()
Find an existing element in the hash table.
The caller provides the search key for the element.
A pointer to the found element is returned, or the null pointer
if the element is not found.
.TP
ah_top()
Find the first element in the hash table.
A pointer to the element is returned, or the null pointer if the
hash table is empty.
.TP
ah_next()
Find the next element in the hash table.
A pointer to the element is returned, or the null pointer if the
hash table is empty.
This function is typically used in conjunction with
.I ah_top()
in order to access all the elements in the hash table with no prior
knowledge of the keys in these elements.
.TP
ah_count()
A count of all elements in a given hash table is returned.
.TP
ah_size()
The size of the given hash table is returned.
.TP
ah_expand()
Expand the size of a dynamic hash table in order to accomodate more elements.
The caller provides the desired new hash table size.
The new size has to be larger than that of the current hash table.
The new hash table inherits the operation mode of the previous hash table
except for the AHNOINIT status which is turned off and the new hash table
is initialized.
If the AHLRU mode was set, the LRU counters are reset to zero thus giving
all entries equal chance to be kicked out once the expanded hash table
fills up again.
The function returns -1 is a new hash table could not be allocated.
.PP
The static hash table functions are very similar.
The differences are listed below.
.TP 15
ahs_init()
As explained above, this function requires the caller to allocate all the
memory used by the hash table descriptor.
.TP
ahs_free()
This function does not exist.
.TP
ahs_expand()
This function does not exist.
.SH "SEE ALSO"
all_list(l), all_queue(l)
