.TH ALL_LIST l  "April 8, 1992" "Release 2.2" "TROLLIUS INTERNALS"
.SH NAME
.B all_list, all_slist
\- general purpose list processing package
.SH SYNOPSIS
.nf
.ta .7i 1.2i
#include <all_list.h>

LIST	*al_init(int elemsize, int (*comp)());
int	al_delete(LIST *ald, char *old);
char	*al_insert(LIST *ald, char *unew);
char	*al_append(LIST *ald, char *unew);
int	al_free(LIST *ald);
char	*al_find(LIST *ald, char *key);
char	*al_next(LIST *ald, char *old);
char	*al_prev(LIST *ald, char *old);
int	al_count(LIST *ald, char *old);
char	*al_top(LIST *ald);
char	*al_bottom(LIST *ald);

#include <all_slist.h>

SLIST	*als_init(int elemsize, int (*comp)(), int nlist,
		char *plist, SLIST *pdesc);
int	als_delete(SLIST *ald, char *old);
char	*als_insert(SLIST *ald, char *unew);
char	*als_append(SLIST *ald, char *unew);
char	*als_find(SLIST *ald, char *key);
char	*als_next(SLIST *ald, char *old);
char	*als_prev(SLIST *ald, char *old);
int	als_count(SLIST *ald, char *old);
char	*als_top(LIST *ald);
char	*als_bottom(LIST *ald);
.SH DESCRIPTION
.I All_list
and
.I all_slist
are general purpose list processing packages.
They differ only in the way memory is allocated for a list.
.I All_list,
the dynamic package,
obtains memory from malloc(l)(3) whenever a new element
is inserted into a list and returns memory with free(l)(3)
whenever an element is deleted from a list.
.I All_slist,
the static package,
requires that the caller provide memory for the maximum number of
list elements when the list is first created.
Functions that operate on a dynamic list are named
.I al_*
and functions that operate on a static list are named
.I als_*.
.PP
A list is created and initialized with the
.I al_init()
or
.I als_init()
functions which both return a pointer to a list descriptor, typedef
.I LIST
or
.I SLIST
respectively.
The list descriptor pointer is used in all subsequent list operation
functions.
In the static function,
.I als_init(),
the caller supplies space not only for the maximum number of list
elements, but for the list descriptor also.
A dynamic list is freed with the
.I al_free()
function.
A static list is simply forgotten, since the caller is responsible
for all the memory involved.
Allocating the space for a static list is a bit tricky since
each element must contain the caller's structure and the list
management overhead structure.
An example of how to allocate space for a static list is given below:
.RS
.sp 1
.nf
char *mylist[25 * (sizeof(struct al_head) +
	sizeof(myelement))];
.fi
.sp 1
.RE
Twenty-five elements of type
.I myelement
are allocated and named
.I mylist.
Other than this space allocation, the caller never sees the
.I al_head
structure.
Element pointers always point to the caller's element type.
.PP
The following functions operate on dynamic lists:
.TP 15
al_init()
Allocate and initialize a list.
A list descriptor pointer is returned, but 0 is returned if allocation
failed.
The caller supplies a pointer to a function that will compare two
list elements, given pointers to each.
The comparison function must return 0 if the elements are equal,
-1 if the first element is less than the second, and
1 if the first element is greater than the second.
.TP
al_delete()
Delete an existing element from a list.
The function returns -1 and sets
.I errno
to EDELETE if the element could not be found in the given list.
.TP
al_insert()
Insert a new element into a list.
The caller prepares and supplies a pointer to the new element.
The function allocates its own element to actually insert into the
list and copies the contents of the caller supplied element.
The caller can reuse the element.
.I al_insert()
returns a pointer to internally allocated list element.
.TP
al_append()
This function is identical to
.I al_insert()
except that the new element is always placed at the end of the list.
.TP
al_free()
Free all elements in a list and free the list descriptor.
The list is effectively blown away.
The list descriptor pointer is no longer valid.
This function always returns 0.
.TP
al_find()
Find an existing list element.
The caller prepares a search key element, filling in only the
key fields needed by the compare function to identify an element.
A pointer to the found element is returned, or the null pointer
if the element is not found.
.TP
al_next()
The caller gives a pointer to an existing list element and a
pointer to the next element in the list is returned.
This pointer may be null.
.TP
al_prev()
The caller gives a pointer to an existing list element and a
pointer to the previous element in the list is returned.
This pointer may be null.
.TP
al_count()
A count of all elements in a given list is returned.
.TP
al_top()
Return a pointer to the first element in the list.
The is pointer may be null if the list is empty.
.TP
al_bottom()
Return a pointer to the last element in the list.
The is pointer may be null if the list is empty.
.PP
The static list functions are very similar.
The differences are listed below.
.TP 15
als_insert()
This function returns the null pointer and sets
.I errno
to EFULL if there are no free elements available.
.TP
als_append()
The same is true for this function.
.TP
als_free()
This function does not exist.
.SH "SEE ALSO"
all_hash(l), all_queue(l)
