.TH BUFFER l  "April 3, 1992" "Release 2.2" "TROLLIUS COMMANDS"
.SH NAME
buffer, sweep \- Control Trollius buffers.
.SH SYNTAX
.B buffer
[-hS] [-b <workers>] [-s <space>] [-u <usrspace>] <node(s)>
.sp
.B sweep
<node(s)>
.SH OPTIONS
.TP 15
.B \-b <workers>
Set maximum number of buffer worker processes to <workers>.
.TP
.B \-h
Print helpful information on this command.
.TP
.B \-s <space>
Set maximum size in bytes of the internal buffer pool to <space>.
.TP
.B \-u <usrspace>
Set maximum size in bytes of user space (messages with positive events)
in the internal buffer pool to <usrspace>.
.TP
.B \-S
Sweep (clean) out all buffers.
.SH DESCRIPTION
.PP
.I Buffer
controls buffering parameters on any Trollius node.
It must be called with an option:
.I buffer <node(s)>
by itself has no function.
.I Sweep
is used after an application program error or premature termination to
remove all messages held in buffers.
.PP
The node is specified on the command line.  The host node is identified
as \fIh\fR.
ITB nodes are identified as
.I n<list>,
where <list> can be a single node identifier or a list of node identifiers.
Node identifiers are established in the boot schema (see nail(l)), and can be
written in decimal or hexadecimal notation.
The
.I buffer
or
.I sweep
command can be broadcast to all ITB nodes, host nodes, or all ITB and
host nodes by specifying
.I <node(s)>
as \fIC\fR, \fIH\fR, or \fIN\fR, respectively.
.PP
The maximum number of buffer worker processes on a node can be set with the
.I -b <workers>
option; the default maximum is 32 ITB and 8 OTB.
Although the maximum number of workers can be set to zero, this is not
recommended, as it will prevent the buffer daemon from forwarding any messages.
.PP
The total space that can be consumed by the buffer daemon's internal buffer
pool is adjusted with
.I -s <space>
option, where
.I <space>
is the maximum number of bytes in the buffer pool; the default is 256 Kbytes.
.I Space
should not be less than MAXNMSGLEN (defined in <net.h>).
.PP
The
.I -u <usrspace>
option sets the buffer pool threshold size that will cause the buffer daemon
to stop accepting user messages (messages with positive events).
The buffer daemon will continue to accept user messages when the
pool size drops below <usrspace>.
By setting this threshold size at least MAXNMSGLEN above the total
pool size limit, this user message flow control feature will be
effectively disabled.
The default value is 128 Kbytes.
.PP
In the event of an application program error or premature termination
of an application process, unwanted messages often collect in the buffers.
The user will need to "sweep" the buffers clean before
running the application program again.
.I Buffer -S <node(s)>
will empty all the buffer workers and remove all messages from the
internal buffer pool on the given nodes.
.I Sweep <node(s)>
is equivalent to
.I buffer -S <node(s)>.
.SH "MESSAGE BUFFERING"
.PP
The purpose of Trollius buffering is to receive, store, and
forward messages that cannot immediately be received by other
processes.
This prevents a process that is sending a message from stopping execution
until a receiving process is ready.
If the five conditions listed below are true, a Trollius process
will send the message to the buffer daemon.
.sp
.RS 5
.TP 3
.B 1.
The NOBUF flag in the
.I nh_flags
field of the network message descriptor (see nsend(l)) is not set.
.TP
.B 2.
The KHOLD flag in the
.I nh_flags
field of the network message descriptor (see nsend(l)) is not set.
.TP
.B 3.
Sufficient space is available in the buffer daemon's internal buffer pool.
.TP
.B 4
The
.I nh_event
field of the network message descriptor is negative, or the
buffer daemon's postive event buffer pool size threshold is not exceeded.
.TP
.B 5.
The forwarding or destination process of the network
message is not synchronized (receive blocked) at the exact moment
the send is attempted.
This can happen anywhere along the route between sender and receiver.
.RE
.sp
.PP
The sending process will block only until the buffer
daemon receives the message.
.PP
The buffer daemon forwards the message to a subordinate buffer worker process.
Each buffer worker process holds one message, which it will forward to the
original forwarding or destination process.
All messages held by the buffer worker processes must be
distinct in their combination of forwarding event
.I (nh_dl_event)
and forwarding type (see nsend(l)).
The forwarding event equals the message event
.I (nh_event)
for a message that has arrived at the destination node.
.PP
If a message received by the buffer daemon is not distinct, it will be
stored in the buffer daemon's internal buffer pool until it becomes distinct.
Similarly, if the maximum number of worker processes exist and are full,
messages will be stored in the internal buffer pool until
a buffer worker becomes free, or until the maximum number of workers is
increased.
.PP
The buffer daemon will refuse to receive any message for buffering
if the current size of the internal buffer pool exceeds the upper size limit.
It will resume receiving messages when space is cleared
through forwarding messages to buffer workers.
.PP
The buffer daemon will refuse to receive any messages with a positive
destination event if the current size of the internal buffer pool
exceeds a threshold size value.
It will resume receiving messages with a postive event if the current
size of the internal buffer pool drops below the threshold size value.
.PP
For any source/destination pair, messages with the same event and type will
be delivered in the order in which they were sent.
This cannot be guaranteed for messages with different events or types.
Deadlock, a situation in which no messages can be delivered, can occur when more
forwarding event/type combinations exist than there are workers to hold them.
The message on which a process is blocked may therefore be trapped in the
internal buffer pool.
Also, deadlock may occur if the message on which a process
is blocked cannot reach the internal buffer pool because it is already full.
This could occur even if buffer workers are available.
.SH EXAMPLES
.TP 4
.I buffer -b 4 h
Use a maximum of four buffer workers on the host node.
.TP
.I sweep C
Clean out all buffers on all ITB nodes.
.SH BUGS
The maximum space limit does not include the space used by buffer workers.
.SH SEE ALSO
bfstate(l), rbuffer(l)
