.TH CBOOT l  "January 3, 1992" "Release 2.2" "TROLLIUS TOOLS"
.SH NAME
cboot \- Boot a single program onto a neighbour ITB node.  (Trollius)
.SH SYNTAX
.B cboot
[-hvL] [-b{cr}] [-i <program>] [-o <device>] [-m <fd>]
.br
[-p <idx>] [-w <topstk>] [-k <priority>] [-t <text>]
.br
-l <base> -s <stksize> [-- <args>]
.SH OPTIONS
.TP 14
.B \-h
Print a helpful message.
.TP
.B \-v
Be verbose.
.TP
.B \-L
Output is a link device as opposed to a plain file.
This option is required for stupid device drivers.
A plain file is appended, if it already exists.
.TP
.B \-br
Create a process, maintaining moses(l) control.
.TP
.B \-bc
Transfer control to program, losing moses(l) control.
.TP
.B \-i <program>
Specify the input file, which must be an executable ITB program.
Without this option, stdin is used.
There are no restrictions on program or data size, as with export(l).
.TP
.B \-o <device>
Specify the output file.
Without this option, stdout is used.
.TP
.B \-m <fd>
Write the high water mark, or the first available address after
the program load, to the given file descriptor.
The 4 byte word is written raw and it is assumed that the
file descriptor is one end of a pipe.
.TP
.B \-p <idx>
Load one entry of the spawn table on the target node (the
output file) whose index in the table is given by the integer
argument.
Update the spawn table size to be
.I <idx>
+ 1.
.TP
.B \-w <topstk>
Specify the first address above the stack.
This fixes the stack to a memory location and overrides the
stack's default placement after the data/bss segments.
.TP
.B \-s <stksize>
Set the stack size, a mandatory input for
.I cboot.
.TP
.B \-l <base>
Set the base address where program loading begins with the
first segment.
This is mandatory input for
.I cboot.
.TP
.B \-t <text>
Give a specific address for the header and text segments.
These would normally be the first segments loaded at the
base address.
Instead, the data/bss segments will be loaded first.
.TP
.B \-k <priority>
Specify a process priority at process creation time.
A suitable default is selected in the absence of this option.
.TP
.B \-\- <args>
These arguments are passed untouched to main() within the program.
argv[argc] is the null pointer.
The minimum argument segment is therefore a single null pointer.
.SH DESCRIPTION
.I Cboot
loads and optionally executes an input program on an output device
connected ITB node, which must be running moses(l).
The output may also be a plain file, in which case the moses
commands and data are simply collected, possibly for later
use by thaw(l).
.PP
There are three ways of specifying the input and output files.
.sp
.nf
	1% cboot <program> <device>
	2% cboot -i <program> -o <device>
	3% cboot < <program> > <device>
.fi
.LP
The first method is compatible with older versions of
.I cboot.
The second method is easier to remember while the third method permits
.I cboot
to be used in a pipeline.
.PP
The load procedure involves laying out the various segments
in memory starting from a given base address, then relocating
the input program's data segment to reflect those addresses
(in the absence of h/w address translation, this is required),
then loading the segments, then preparing the stack for
execution, then executing the program.
.SH "MEMORY LAYOUT"
.PP
Starting at the base address, which must be given on the
command line, and in the absence of the exceptions discussed
below, the program segments are laid out as follows:
.sp 1
.RS
.nf
header/text
static data/bss
stack
arguments
.fi
.RE
.LP
The first available word aligned address after the argument
segment is the new high water mark and is written to the
given file descriptor if the
.I -m
option is taken.
Sboot(l) uses this feature to boot several programs consecutively
in memory.
.PP
The stack size must be provided on the command line.
The stack may be located anywhere by giving the address immediately
above (following) it with the
.I -w
option.
When this option is present,
.I cboot
does not take the stack into account when laying out memory, though
it does initialize it as always.
In other words, the stack is missing from the list above and the
user should take care not to interfere with the natural layout
of the other segments (by placing the stack at a non-conflicting address).
.PP
The text segment may be shared from a previously loaded copy of the
program whose header/text address is given with the
.I -t
option.
.SH LOADING
.PP
Each segment is loaded into its mapped address.
Since all of the bss and most of the stack are uninitialized,
they are not loaded.
Also, if the
.I -t
option is taken, the header/text segment is not loaded.
The data and argument segments are always loaded.
.PP
Loading is accomplished by writing moses(l) commands to the output.
If the output is a device, indicated by the
.I -L
option, then moses(l) must be running on the ITB node connected to
the device.
Moses can be booted with the export(l) tool.
.PP
.I Cboot
will load one entry of a table that describes all booted programs
on an ITB node.
This table resides at a fixed memory address and the table entry
that is written is given by the
.I -p
option.
.I Cboot
updates the size of the table to be the given table entry index plus one.
This means that the
.I -p
options really only appends an entry onto the end of the table.
The first table entry is 0.
.SH EXECUTION
.PP
Execution can occur by either creating a process or by transferring
control away from moses with a jump instruction.
These options are selected by
.I -br
and
.I -bc
respectively.
If
.I cboot
is to be executed several times without resetting the node, the
.I -br
option must be used on all but the last invocation.
.PP
The default process priority can be overridden with the
.I -k
option.
.SH EXAMPLES
.TP 4
.I "cboot -l 0x1200 -s 0x400 < prog > cmds"
This is the simplest form of
.I cboot.
The moses(l) commands to load `prog' at address
0x1200 with a stack of 0x400 bytes are written into the plain
file, `cmds'.
.TP
.I "cboot -l 0x1200 -s 0x400 -L -br < prog > /dev/l0"
The program is loaded and executed on the ITB node connected
to `/dev/l0'.
.TP
.I "cboot -l 0x1200 -s 0x400 -L -br -w 0x1200 < prog > /dev/l0"
The stack is located immediately behind the base address, at address 0xC00.
.TP
.I "cboot -l 0x2800 -s 0x400 -L -bc -t 0x1200 < prog > /dev/l0"
The program is called and shares
the text segment with the previously loaded program at 0x1200.
.I Cboot
cannot be called again without reseting the node.
.SH DIAGNOSTICS
The verbose option,
.I -v,
will show the memory layout, the full pathname of the input file,
the final high water mark, and other useful information.
.SH SEE ALSO
export(l), moses(l), sboot(l), thaw(l)
