.TH CONF l  "February 24, 1992" "Release 2.2" "TROLLIUS TOOLS"
.SH NAME
conf \- Trollius node process schema
.SH DESCRIPTION
A process schema (once ambiguously called a configuration file), lists the
system processes that will constitute the Trollius Operating System on a
particular node.
Trollius developers will find process schemata very useful for
debugging and for generating custom systems.
.PP
Separate default configurations files are normally used
for booting the host node (OTB), device connected ITB nodes (neighbours of
the host), remote ITB nodes and Standalone Trollius.
In addition, the
.I \-c
option to hboot(l), sboot(l), and vboot(l) specifies the file name
of a custom process schema.
This capability is not available with spread(l).
.PP
Application programs can be booted with the operating system
by including the program name in the process schema.
.SH STRUCTURE AND SYNTAX
.PP
The process schema grammar defines three types of statements: comments,
processes, and options.
.PP
Comments begin with # and terminate with a newline.
.PP
Process statements consist of a filename, command line arguments,
and possibly options local to the process.
The command line arguments are passed to the process when it is started.
The process options are used by the booting tools before starting the process.
.PP
Filenames may include application programs and any of the system processes
listed below.
.PP
.nf
.ta 1.0i 1.75i
bootd	boots one ITB node from another ITB node.
bufferd	creates, kills, sweeps, and states buffers.
castd	sends and receives messages from groups of nodes.
dl_dev	spawn link I/O processes for OTB device drivers.
dl_t4	spawn link I/O processes for ITB transputer links.
echod	echoes messages; can be used to test nodes and links.
filed	OTB file server.
flatd	provides symbolic access to ITB memory.
i860d	controls an ITB i860 co-processor.
kenyad	controls and monitors ITB processes.
kernel	coordinates message passing.
ledd	controls LEDs.
loadd	loads executable files onto ITB nodes; runs OTB.
mallocd	allocates and frees ITB memory.
memd	remotely reads and writes ITB memory.
mtvd	controls a pixel map display.
router	maintains routing tables.
traced	collects and transports trace data.
shared	ITB shared library code block.
.fi
.PP
Processes are started in the order given in the process schema, and
for Trollius, the order is important.
On OTB nodes, the kernel must be first and dl_dev must be last.
On ITB nodes, shared must be first, followed by the kernel,
and dl_t4 must be last.
.PP
Process options are placed inside braces {} following the
command line arguments.
Process options consist of a keyword followed by an assigned
value in parentheses.
Options that are not explicitly given have default settings.
Options inside braces in process statements apply only to that process.
Options outside process statements apply to all processes that
appear later in the process schema, unless locally overridden.
All the options are listed below:
.PP
.TP 14
topology
a string of events and link names, global only, default null
.TP
stack
stack size, default 0x1400 bytes
.TP
delay
waiting time in seconds after process boot, OTB only, default 0
.TP
priority
transputer priority, transputers only, default 1
.TP
share
0/1 flag indicating text segment is shared with previous process with
identical filename, ITB only, default 0
.TP
onchip
0/1 flag indicating stack resides in onchip RAM, transputers only,
default 0
.PP
The topology option can be overridden from the command line using
hboot(l), sboot(l), vboot(l) or spread(l) with the
.I \-T
option.
.PP
.SH "FROZEN PROCESS SCHEMA"
.PP
A frozen process schema does not contain a description of processes
but rather an image of the complete binary stream that would have been
sent to an ITB node after parsing a process schema.
Booting with a frozen schema is much faster than the normal procedure
of parsing the process schema and invoking tools to handle each process
separately.
.PP
A default frozen process schema is usually created for each default
ITB process schema.
Spread(l), sboot(l), and vboot(l) have an option to use or not use
frozen schemata.
See the individual manual pages for more details.
An ITB schema is frozen with the \fI-Z\fR option of sboot(l).
.SH EXAMPLES
.PP
Following are example process schemata.
The first file is used for the OTB host node in Trollius.
.PP
.RS
.nf
.if t .ta 0.5i 1.0i 2.0i 3.0i
.if n .ta 8n 16n 24n 32n
##
## typical Trollius OTB process schemata
##

topology = /dev/nap0 0x80000000;

##
## The kernel is listed first.
##
kernel { delay(1); }

##
## daemons
##
router { }
loadd { }
echod { }
filed { }
castd { }
bufferd { }

##
## The datalink administrator with topology specific runtime
## parameters should be listed last.
##
dl_dev $topology { }
.fi
.RE
.LP
The second file is used for remote ITB nodes in Trollius.
.LP
.RS
.nf
##
## typical Trollius ITB process schemata
##

##
## The shared library and kernel are listed first.
##
shared { priority(0); }
kernel { priority(0); }

router { }
echod { }
mallocd { }
memd { }
flatd { }
bootd { }
kenyad { }
castd { }
bufferd { }
traced { }

##
## The datalink administrator should be listed last.
##
dl_t4 { }
.fi
.RE
.SH FILES
.TP 4
$TROLLIUSHOME/boot/conf.otb
default host (OTB) node process schema where $TROLLIUSHOME is the
installation directory
.TP
$TROLLIUSHOME/boot/conf.isb
default device connected ITB node process schema
.TP
$TROLLIUSHOME/boot/conf.itbz
default device connected ITB node frozen process schema
.TP
$TROLLIUSHOME/boot/conf.ivb
default remote ITB node process schema
.TP
$TROLLIUSHOME/boot/conf.ivbz
default remote ITB node frozen process schema
.TP
$TROLLIUSHOME/boot/conf.SA
Stand-alone Trollius process schema
.SH SEE ALSO
hboot(l), sboot(l), spread(l), vboot(l)
