.TH CUBIX l  "April 8, 1992" "Release 2.2" "TROLLIUS C LIBRARY"
.SH NAME
CUBIX \- overview of CUBIX-style parallel I/O under Trollius
.SH SYNOPSIS
.nf
.ft B
#include <freq.h>
.sp
int rforder(int fd, int sequence);
.sp
int rfchcx(int fd, int flags);
int async(int fd);
int singl(int fd);
int multi(int fd);
.fi
.SH DESCRIPTION
.LP
Trollius supports parallel input and output, based on CUBIX,
that closely represents sequential I/O but involves a transparent
synchronization among participating nodes.
Comprehensive literature on CUBIX is listed in the acknowledgements
section of this introduction and is recommended reading for a better
understanding of Trollius' CUBIX-based system.
This introduction will concentrate on how to use parallel I/O
under Trollius.
.LP
Parallel I/O provides for loosely synchronous communication among
a defined cast or group of nodes.
The only cast currently supported involves communication between the
host node on one hand and all ITB nodes on the other.
This restriction will be removed in future versions.
(The cast daemon is not actually involved in parallel I/O, though it
should be.)
One client process from each cast member is allowed to operate on any one
open file.
All clients participate in all I/O operation, and an I/O operation is not
processed until requests from all clients have made the request to
the remote file daemon.
.LP
Sequential I/O is also called asynchronous mode.
Parallel I/O has two flavours:  singular mode and multiple mode.
In singular mode, clients process identical data.
All client processes pass identical parameters to the file runtime function.
Singular mode is very useful for status messages, prompts, or global data I/O.
In multiple mode, clients process varying amounts of non-identical data.
.LP
Clients are serviced in the order that they appear in the open
descriptor's cast member list.
The usual order is from lowest to highest node identifier.
The order may be modified using the
.I rforder()
function where
.I fd
is the file descriptor handle and
.I sequence
is a sequence (ordering) number from 0 to the number of members minus one.
Note that each client process must call
.I rforder()
with its own distinct sequence number.
.LP
Singular and multiple modes are requested by the O_SINGL and O_MULTI
flags given to the open(2) function.
Sometimes it is desirable to change an open file descriptor from one
CUBIX mode to another and this is the purpose of the
.I rfchcx()
function and its accompanying wrappers.
.I Flags
indicate the new mode and is one of O_SINGL, O_MULTI or O_ASYNC.
None of the other original open(2) flags need to be repeated
and are unaffected.
.LP
Parallel I/O operations supported by Trollius are
open(2), close(2), read(2), write(2), and lseek(2).
These are called in the same manner as for sequential I/O except as
noted for open(2) above.
.LP
The file daemon will not allow multiple concurrent parallel opens on the
same file name.
Also, if a client fails to synchronize on any I/O operation, the
parallel file descriptor may be left open and in an intermediate
state of synchronization.
This will cause bad things to happen if the application is run
again without cleaning up the old file descriptor using tfile(l).
The state of synchronization on a parallel file descriptor can
.SH "RETURN VALUE"
Upon successful completion of the CUBIX specific functions,
the value 0 is returned.
If an error occurred, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
These error codes relate to the functions described here and
to the basic I/O operations which support CUBIX.
.TP 12
[ECXREQ]
The file operation is not the same as the operation other clients
are already attempting to synchronize.
.TP
[ECXMEMBER]
The caller is not a member of the CUBIX cast.
.TP
[ECXSINGL]
This is not really an error, but an indication that a write
operation is in singular mode, is redundant and need not continue.
.SH ACKNOWLEDGEMENTS
J. Salmon,
.I "CUBIX: Programming Hypercubes without Programming Hosts"
.B IN
M.T. Heath (ed.),
.I "Hypercube Multiprocessors, 1987",
SIAM, 1987
.PP
G. Fox et al.,
.I "Solving Problems on Concurrent Processors, vol. 1",
Prentice Hall, 1988.
.SH SEE ALSO
tfile(1), fstate(l)
