.TH DO_ARGS l  "April 8, 1992" "Release 2.2" "TROLLIUS INTERNALS"
.SH NAME
.B do_args
\- Process argc, argv command-line arguments.
.SH SYNOPSIS
.nf
.ft B
void	validopts (char *str);
void	exclusive (char *str);
void	followed  (char *str);
void	twoletter (char *str);
int	do_arg (int *argc, char **argv);
int	opt_taken (int let);
int	opt_chosen (int let);
int	get_2nd_let (int let);
char	*getparam (int let);
int	xtoi (char **cp, int *ip);
int	intparam (int let, int *ip);

int	argtailc;
char	**argtailv;
.ft R
.fi
.SH DESCRIPTION
.I Do_args
is a package of functions that make it easy for a C program to
process the command-line arguments
.I argc
and
.I argv,
in whatever way it chooses.
.PP
Option letters that are prefixed with a dash in the command line,
and parameters that follow selected option letters,
are processed directly by the
.I do_args
package.
Any additional arguments not prefixed by a dash, such
as file names, are left in the collapsed
.I argc, argv
argument list.
A "--" argument directs the
.I do_args
package to avoid processing any additional arguments to the right
of the "--".
These unprocessed arguments are stored using
.I argtailc
and
.I argtailv
(also in the standard
.I argc, argv
format).
The
.I do_args()
function also checks the command line for some errors, such as
missing parameters, or more than one mutually-exclusive option.
.PP
The main program calls
.I validopts()
to declare which option letters are valid,
.I exclusive()
to declare sets of option letters which are mutually exclusive, and
.I followed()
to declare option letters which are followed by a parameter.
Each of these functions accepts, as an argument, a null-terminated
character string that holds the list of applicable letters.
Option letters passed to the latter two functions are assumed to be valid
even if omitted from the call to
.I validopts().
If no option letters are mutually exclusive or followed by
parameters, the calls to
.I exclusive()
or
.I followed()
may be omitted.
.PP
The option "-#", where "#" is a decimal or hexadecimal
integer, should be listed as "0" in the string passed to
.I validopts().
It should
.I not
be included in the list for
.I followed().
Two letter options are permitted using the
.I twoletter()
function.
For example,
.I
twoletter("xabc")
or
.I
twoletter("x:abc")
permit the two letter options "xa", "xb" and "xc".
.PP
After declaring the valid options, the main program calls the function
.I do_args()
to process the command line given by the user.
The
.I argc
and
.I argv
parameters passed to main()
are passed to
.I do_args().
The
.I argc
parameter must be passed by reference.
The options and associated parameters processed by
.I do_args()
are stored in static variables that are queried by
the functions described below.
.SH OPTION QUERIES
.PP
The function
.I opt_taken()
checks whether a given option letter appeared on the command line.
If it did, a true (nonzero) value is returned.
Otherwise, false (0) is returned.
Since characters in C are handled as integers, the character specifying
the option can be passed as the integer \fIlet\fR parameter.
.PP
The function
.I opt_chosen()
returns the mutually exclusive option letter chosen from a set defined by
.I exclusive()
and identified by a member letter of the set passed as the argument to
.I opt_chosen().
If none was chosen, -1 is returned.
.PP
The parameter following a parameter-taking option can be retrieved
via the function
.I getparam().
This function is given the option letter.
A pointer to the corresponding parameter string is returned.
A null pointer is returned if the parameter is missing, if the option was
not taken, or if this option letter was not declared as one that is followed
by a parameter.
.PP
The numerical value of a command-line parameter that is always an integer
can be obtained by calling the function
.I intparam(),
passing two parameters: the option letter, and a pointer to a variable of type
.I int
that will receive the value.
The numerical parameter can be specified on the command line in decimal, or,
if preceded by the prefix
.I 0x,
in hexadecimal.
It may be either positive or negative.
A negative hexadecimal value must have the negative sign preceding the
.I 0x
prefix.
The function
.I intparam()
itself returns 0 if all went well, or -1 if the parameter is missing,
inappropriate, or badly formatted.
.PP
The function
.I xtoi(),
used by
.I intparam(),
converts a character string (pointed to by its first argument)
into an integer value (pointed to by the second argument).
The rules of conversion are as described for
.I intparam()
above.
The function
.I xtoi()
returns 0 if all went well, or -1 if the string is badly formatted.
.PP
Two letter options are queried with the
.I get_2nd_let()
function which accepts the first letter and returns the second letter
that was chosen.
If multiple second letter options were chosen, they are returned by
successive calls to
.I get_2nd_let()
and -1 is returned after none remain or if none existed.
.SH DIAGNOSTICS
In the case of errors, the functions described above return a value
indicating that an error has occurred.
In addition, the global variable
.I errno
is set to one of the following values, symbolically defined
in the include file
.I <terror.h>.
.TP 22
EUSAGE
the command line violates the rules as declared.
.TP
EBADASCIINUMB
a character string that is supposed to represent a number
has a format that is invalid.
.SH "SEE ALSO"
parse_nodes(l)
