.TH DSEND l "April 8, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
dsend, dtry_send, drecv, dtry_recv  \- Trollius datalink level message passing
.SH SYNOPSIS
.nf
.ft B
#include <net.h>
#include <events.h>
.PP
.ft B
int	dsend (struct nmsg *header);
int	dtry_send (struct nmsg *header);
int	drecv (struct nmsg *header);
int	dtry_recv (struct nmsg *header);
.fi
.SH DESCRIPTION
These functions access the datalink sublayer message-passing services of
the Trollius network level.
The datalink layer functions do the actual work of network communication.
Although they can be invoked directly, they are most often called by the
network and transport layer functions.
The datalink layer functions do not have built-in routing.
Instead, the user supplies information on the forwarding process.
.PP
.I Dsend()
and
.I drecv()
are typically used to send messages to, or receive messages from,
nearest-neighbor nodes (see "Typical usage").
.I Dtry_send()
and
.I dtry_recv()
differ from
.I dsend()
and
.I drecv()
in that they will never cause the calling process to block.
The message is either immediately transferred,
or an error is immediately returned, indicating that the
process would have blocked.
Both the sending process and the synchronizing receiving process
may use the non-blocking functions (unlike ksend(l)) unless
both parties are on the same node and NOBUF is used.
.SS THE NETWORK MESSAGE STRUCTURE
All of the functions accept a pointer to a message structure
which is an extension of the local-layer message structure
used by ksend(l) and krecv(l).
The network message structure is defined in <net.h> as:
.PP
.nf
.RS
.DT
struct nmsg {
	int	nh_dl_event,
		nh_node,
		nh_event,
		nh_type,
		nh_length,
		nh_flags,
		nh_data[NHDSIZE];
	char	*nh_msg;
};
.RE
.fi
.PP
The
.I nh_dl_event
field is used by
.I dsend()
to a) synchronize the sending process and a forwarding process
such as a link process, or b) synchronize the sending process
and a local receiving process.
In the first case, the message will be passed if
.I nh_dl_event
equals the published event number of the forwarding process.
This event number is returned by getroute(l), and by the
.I dl_event(x)
macro.
The argument to
.I dl_event
is the appropriate link number defined in the Boot Schema (see nail(l)).
Link process events and
.I dl_event
are defined in <events.h>.
In the second case, the message will be passed if
.I nh_dl_event
equals the value set in the receiving process'
.I nh_event
field.
The
.I nh_dl_event
field is not used by
.I drecv().
This field remains unchanged after calling
.I dsend()
or
.I drecv().
.PP
The
.I nh_node
field is used by
.I dsend()
to identify the remote node running the intended receiving process.
The
.I nh_node
field is not used by
.I drecv().
A receiving process thus cannot directly specify its synchronizing
sending process.  Instead, receiving processes are "matched" to
sending processes by one or both of the
.I nh_event
and
.I nh_type
fields.
This field remains unchanged after calling
.I dsend()
or
.I drecv().
.PP
.I Nh_event
is an arbitrary positive integer, used as a second level of
synchronization after
.I nh_node.
In general, the values assigned by the sending and the receiving
processes must be equal.
The exception is when the sending process synchronizes with a local
destination process,
in which case the sending process'
.I nh_dl_event
field must match the receiving process'
.I nh_event
field.
The
.I nh_event
field remains unchanged after calling
.I dsend()
or
.I drecv().
.PP
The
.I nh_type
field further filters messages that match on the
.I nh_event
field.
A message will be passed only if the
.I nh_type
fields of the sending and receiving processes have at least one bit
set in an identical position.
In other words, the bitwise logical AND of the type fields specified by
the two parties must not equal zero.
A zero value in the
.I nh_type
field is a special case -
that process will match any other process using the same
.I nh_event.
The
.I nh_type
field remains unchanged after calling
.I dsend(),
but is set to the sender's
.I nh_type
field after calling
.I drecv().
.PP
The
.I nh_length
field holds the length (in bytes) of the message to be sent.
If the sender and the receiver specify different lengths,
the lesser amount will be transferred.
The maximum length of messages transferred with
.I dsend()/drecv()
is MAXNMSGLEN (defined in <net.h> and currently set to 4096 bytes).
The
.I nh_length
field remains unchanged after calling
.I dsend(),
but is set to the minimum of the sender's and receiver's lengths after calling
.I drecv().
.PP
The
.I nh_flags
field is normally set to 0, and remains unchanged after calling
.I dsend()
or
.I drecv().
When the NOBUF bit is set in
.I nh_flags,
buffers will not be used.
The KHOLD bit is set to create a virtual circuit (see "Virtual Circuits").
When the DFORWARD bit is set in the receiving process'
.I nh_flags
field, all message structure fields will be changed to those of the
sending process except for
.I nh_msg.
DFORWARD has no effect when used with
.I dsend().
It is not typically used by user processes.
Flags used to assure that the data representation is correct for the
receiving node are discussed under "Data Representation".
NOBUF, KHOLD, DFORWARD, and the data representation flags are defined in
<net.h>.
.PP
The
.I nh_data
field is a convenient data pouch within the network message descriptor.
Its array size is NHDSIZE, which is defined in
<net.h> and is currently set to 8.
It can be used for sending short messages (in which case
.I nh_length
is set to 0)
or for appending control information to the message body.
After the message is passed, the information placed in
.I nh_data
by the sender will appear (possibly with conversion, see "Data Representation")
in the same field of the receiver.
The sender's
.I nh_data
field will not change.
.PP
The
.I nh_msg
field holds the address of the first byte of data to be sent.
The data must be stored contiguously in memory.
This field remains unchanged after calling
.I dsend()
or
.I drecv().
.SS DATA REPRESENTATION
On nodes of different types, data may have different representations.
For example, integers may be stored with the most significant byte
first in memory (big-endian) or with the most significant byte last in
memory (little-endian).
All fields in the network message structure, except
.I nh_msg,
are automatically converted if passed to a node with
different data representation.
In converting the \fInh_data\fR field, it is assumed to hold all integers.
.PP
The
.I nh_flags
field of the sender's message structure can be set to the following data
representation flags (they will have no effect when used with
.I drecv()
).
Each flag assumes a data type, and will make the appropriate change in the
data representation of the given field.
They will have no effect if data conversion is not needed.
.TP 14
DINTDATA
.I nh_data
holds 8 32-bit integers.
.TP
DFLTDATA
.I nh_data
holds 8 single precision real numbers.
.TP
DDBLDATA
.I nh_data
holds 4 double precision real numbers.
.TP
DRAWDATA
.I nh_data
representation will not be changed.
.TP
DINTMSG
.I nh_msg
points to integers.
.TP
DFLTMSG
.I nh_msg
points to single precision real numbers.
.TP
DDBLMSG
.I nh_msg
points to double precision real numbers.
.TP
DRAWMSG
.I nh_msg
representation will not be changed.
.PP
If
.I nh_data
or
.I nh_msg
contains a mixture of data types, the user will have to change the
representation using the functions ltot(l), ttol(l), etc.
.SH TYPICAL USAGE
The typical usage of
.I dsend()
and
.I drecv()
is for nearest neighbor message-passing, when the user wishes to avoid
the expense of automatic message routing (as provided by nsend(l)).
The datalink functions
are fully capable of local and multi-hop remote message-passing,
but will, in the latter case, need to invoke the router after the first "hop".
Multi-hop remote message-passing with
.I dsend()/drecv()
is most useful when the
user wants to avoid the established route between two nodes.
.PP
To synchronize the sending and receiving processes, the sending process must set
.I nh_node
to the destination neighbor's node ID,
.I nh_dl_event
to the link process event connecting the destination node, and
.I nh_event
to the value specified by the receiving process.
.SH FORWARDING DATALINK MESSAGES
If the sending process does not immediately synchronize with
the forwarding process specified by
.I nh_dl_event,
.I dsend()
will secretly try to synchronize on a buffer event.
For this reason, the forwarding process identified by
.I nh_dl_event
should never be a buffer process.
A message forwarded to a buffer process will eventually be forwarded
to the original
.I nh_dl_event.
If the sending process does not immediately synchronize
with the buffer process, the sending process will block on the original
.I nh_event.
This behavior can be disabled by setting the NOBUF bit in the
.I nh_flags
field.
The user should be aware of the presence of buffer processes
even though no programming is required to utilize them.
.SH BLOCKING
A process calling
.I drecv()
blocks until the message sent by the process calling
.I dsend()
entirely arrives.
A process calling
.I dsend()
blocks only until its message is picked up by a) a local
process calling
.I drecv(),
b) a buffer process on the same node, or
c) the multicast process on the same node, or
d) a link process on the same node.
The only thing that is guaranteed by a successful return from
.I dsend()
is that the message has entirely left the calling process.
.I Dsend()
thus differs from ksend(l), which blocks unless a receiver is
already waiting.
.PP
The loose blocking behavior of
.I dsend()
introduces a fundamental danger of Trollius message passing:  a sending
process can transmit a message that may never be received due to programming
error or deadlock.
This message will never be dropped or timed out.
Some Trollius process will always be stuck with it, waiting for a synchronizing
.I drecv()
that will never happen.
If that unfortunate process is a buffer, it can
be located by the user and swept clean (see sweep(l)).
However, if the process is a link proprietor,
the link is henceforth plugged and useless.
An intelligent buffering system does try to prevent this from
happening (see buffer(l)), but it remains theoretically possible.
.PP
Besides the legitimate buffer process, each datalink input proprietor
has the space to hold one message.
NOBUF does not affect this single link buffer.
.SH VIRTUAL CIRCUITS
The virtual circuit capability described
in ksend(l) is also available for network messages.
A virtual circuit eliminates the need for interaction with the kernel
and other server processes on subsequent
.I dsend()/drecv()
calls, and thus increases the speed of message-passing.
A network virtual circuit is a chain of local virtual circuits
beginning with the sending process, following through any number of link
proprietor processes, and ending with the destination receiving process.
For the life of the circuit, the intervening links can be used by
.I "no other processes,"
not even system processes.
Thus it is recommended that virtual circuits be used
.B "for nearest-neighbor message-passing only."
On transputers, a virtual circuit that cross an inter-node link
eliminates the afore-mentioned link buffer and causes a strong
synchronization between the sender and receiver on each side of the link.
.PP
A virtual circuit is created by setting the KHOLD bit in the
.I nh_flags
field of both the sending and receiving processes' message structures.
Unlike the kernel level, there is protection against length mismatch for
network level virtual circuits.
The lesser amount specified in the sender or receiver's
.I nh_length
field will always be transmitted.
Resetting the KHOLD flag of both message structures to 0
sends or receives the message along the virtual circuit, but
terminates the circuit after transmission.
Communicating parties must set and reset the KHOLD bit on the same message,
or both parties will block and the node may fail.
Use virtual circuits with care and only after debugging an application
without them.
Buffers are automatically disabled (NOBUF) when KHOLD is set.
.SH "RETURN VALUE"
.PP
Upon successful completion, 0 is returned.
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 23
[EWOULDBLOCK]
.I dtry_send()
or
.I dtry_recv()
failed because the message could not be forwarded or received, respectively.
A call to
.I dsend()
or
.I drecv()
would have blocked.
.TP
[ENOEVDESCRIPTORS]
The Trollius kernel ran out of message descriptor space.
.SH "SEE ALSO"
ksend(l), nsend(l), psend(l), tsend(l)
