.TH GETNET l "April 8, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
getnet, getnodeid, getnodetype, getorigin, getnitb, getnotb, getnall, getncast,
getntype, getsmart \- Get information on nodes and on the Trollius network.
.SH SYNOPSIS
.nf
.ft B
#include <net.h>

int	getnet (struct net_info *nodes);
int	getnodeid ();
int	getnodetype ();
int	getorigin ();
int	getnitb ();
int	getnotb ();
int	getnall ();
int	getncast ();
int	getntype (int nodetype, typemask);
int	getsmart ();
.fi
.SH DESCRIPTION
These functions return information obtained from the local route daemon
regarding the currently booted Trollius network.
.PP
.I Getnet()
accepts a pointer to a net_info structure, defined in <net.h>:
.nf
.ta .5i 1i 1.5i

	struct net_info {
		int	ni_nodeid,
			ni_nodetype,
			ni_origin;
	};

.fi
Upon return from
.I getnet(),
the fields in the structure hold, in order:
the local node identifier, the local node type and the identifier
of the host node from which the Trollius multicomputer was booted.
The node type is a combination of flags describing the capabilities of
the node.
These flags, defined in <net.h>, are:
.TP 12
NT_OTB
host node
.TP
NT_ITB
compute node
.TP
NT_CAST
multicast definition
.TP
NT_WASTE
node not available to the user
.TP
NT_DISK
node has a disk
.TP
NT_TUBE
node has a video display unit
.TP
NT_JONES
node is a neighbour of the local node
.TP
NT_BOOT
node is booted by the local node
.PP
.I Getnodeid()
returns the local node identifier.
.I Getnodetype()
returns the local node type.
.I Getorigin()
returns the host node identifier.
.PP
.I Getnitb()
returns the number of compute nodes that have no additional capabilities,
i.e., ITB nodes that are available to the user, and do not have an attached
disk or video display unit.
.I Getnotb()
returns the total number of host nodes.
This will always be 1 for Trollius Release 2.2.
.I Getnall()
returns the total number of nodes in the system.
.I Getncast()
returns the total number of defined multicast destinations passing
through the local node (see rcast(l)).
This includes all multicasts which are routed through the local node, not
just those that are registered by the local node.
.I Getntype()
is a general function that is used to determine the number of nodes whose node
types have certain bits set to certain values.
This is a flexible tool which allows the user to obtain very specific
information about the capabilities of nodes in the system.
Use of
.I getntype()
will be described in detail below.
.PP
.I Getsmart()
blocks the caller until the router has received its network and routing
information and can provide intelligent answers to
.I getnet()
and getroute(l).
.SS GETNTYPE
You may need more detailed information on the number and types of nodes in the
system than just the number of ITB, OTB, or cast definitions.
You may, for example, want to know the number of ITB nodes with attached video
displays.
The
.I getntype()
function is used for this purpose.
.PP
To understand
.I getntype(),
you must realize that node types are interpreted as
binary numbers, and that each node type has a different bit set.
NT_OTB is an exception: it has no bits set.
A bitmask having all bits set, NT_ALL, is also available.
Note that NT_ALL does not include NT_JONES nor NT_BOOT since these
node characteristics are not inherent to the nodeids, but depend on the node
from which the query is made.
The node types are thus:
.PP
.nf
.ta 1.3i 2.3i
Nodetype	Value	Bitmask

NT_OTB	0	....00000
NT_ITB	1	....00001
NT_CAST	2	....00010
NT_WASTE	4	....00100
NT_DISK	8	....01000
NT_TUBE	16	...010000
NT_ALL	31	...011111
NT_JONES	32	..0100000
NT_BOOT	64	.01000000
.fi
.PP
The
.I typemask
argument to
.I getntype()
is used to specify which bits are of interest.
These are set to 1, all others are set to 0.
The
.I nodetype
argument is used to specify what values these bits should take.
.I Getntype()
then compares the relevant bits (as specified by
.I typemask)
in the node type of each node in the system, to
see if they have the appropriate value (as specified by
.I nodetype).
.PP
A simple example is provided by
.I getnitb(),
which calls
.I getntype()
to learn the number of compute nodes that have no additional capabilities.
To do this,
.I getntype()
has to look at all bits.
Thus, the
.I typemask
argument is NT_ALL.
The values these bits can take is NT_ITB: all bits must equal 0 except
for the ITB bit.
Thus,
.I getnitb()
actually calls:
.PP
.ti +5
nnodes = getntype (NT_ITB, NT_ALL);
.PP
To learn the number of ITB nodes that have an attached video
display, but no other capabilities, you would again want to look at all bits,
but you'd like two bits to be set: the ITB bit and the TUBE bit.
A bitwise OR of these values will produce a bitmask with both bits set.
Thus, you would call:
.PP
.ti +5
nnodes = getntype (NT_ITB | NT_TUBE, NT_ALL);
.SH "RETURN VALUE
Upon successful completion of
.I getnet()
or
.I getsmart(),
0 is returned.
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.I Getnodeid()
and
.I getorigin()
both return node identifiers.
NOTNODEID (defined in <net.h>) is returned
if the local route daemon does not yet know any network information or if
an error occurred.
To distinguish between the two cases, errno is set to 0 if the route daemon
has no network information otherwise errno indicates the error.
Likewise,
.I getnodetype()
returns a valid node type or NOTNODETYPE (defined in <net.h>) in
the cases of error or lack of information in the local route daemon.
In the same manner, these two cases are distinguished by having errno set
to 0 if the route daemon lacks the network information otherwise errno
indicates the error.
The return values upon successful completion for the remaining functions are
described above; they return -1 if an error occurs.
None of these functions have specific errors of their own,
but can incur any of the errors associated with nsend(l).
.SH "SEE ALSO"
rsetnet(l)
