.TH GETNODES l "April 8, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
getnodes \- Get array of Trollius node identifiers.
.SH SYNOPSIS
.nf
.ft B
#include <net.h>

int	getnodes (int *pnodes, int nnodes, int nodetype, int typemask);
.fi
.SH DESCRIPTION
.I Getnodes()
returns, at the location pointed to by
.I pnodes,
an array of no more than
.I nnodes
node identifiers whose type match the given
.I nodetype
at the bit positions specified by
.I typemask.
Given sufficient space and a sufficiently large
.I nnodes,
all node identifiers having a node type field that satisfies the relation
imposed by the
.I nodetype
and
.I typemask
parameters will be returned.
.PP
The node type is a combination of flags describing the capabilities of the
node.
These flags, which are interpreted as binary numbers, are defined in
<net.h> as:
.PP
.nf
.ta 1.3i 2.3i
Nodetype	Value	Bitmask

NT_OTB	0	....00000
NT_ITB	1	....00001
NT_CAST	2	....00010
NT_WASTE	4	....00100
NT_DISK	8	....01000
NT_TUBE	16	...010000
NT_ALL	31	...011111
NT_JONES	32	..0100000
NT_BOOT	64	.01000000
.fi
.PP
A node type is said to satisfy the mentioned relation when it equals the given
.I nodetype
on all bit locations that are set to 1 in
.I typemask.
Specifically, there are three cases that arise when the user wants to
specify each bit of the desired node types:  a bit should be set to 0,
should be set to 1, or the bit is a don't-care.
For each bit that should be set to 0, the user sets the
.I nodetype
parameter bit to 0 and that of the
.I typemask
parameter to 1.
Likewise, for each bit that should be set to 1, the user sets the
.I nodetype
parameter bit to 1 and that of the
.I typemask
parameter to 1.
Finally,  for each don't-care bit, the user sets the
.I typemask
parameter bit to 0,  the
.I nodetype
parameter bit being irrelevant in this case.
For more discussion of the use of the
.I nodetype
and
.I typemask
parameters, read the getnet(l) manual page.
.PP
The number of nodes satisfying a given
.I nodetype
and
.I typemask
can be predicted with getntype(l) and the required amount of space
can be reserved.
No indication of the number of returned node identifiers is given by
.I getnodes().
.SH EXAMPLE
The following code finds, by calling getntype(l), the number of nodes
with their node type least significant bit set to 0.
Memory is then allocated for an array of this size, and the array is
filled by calling
.I getnodes().

.nf
#include <net.h>

int *pnodes, nnodes;

nnodes = getntype(0, 1);
if ((pnodes = (int *)malloc(nnodes*sizeof(int))) == NULL) {
        kexit(errno);
}
if (getnodes(pnodes, nnodes, 0, 1) != 0) {
        kexit(errno);
}
.fi

.SH "RETURN VALUE
Upon successful completion of
.I getnodes(),
0 is returned and
.I pnodes
points to an array of node identifiers.
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH "SEE ALSO"
getnet(l)
