.TH GETRENT l "April 8, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
getrent, getrentc, getroute, getroute2, getrtype \- Get Trollius routing
information.
.SH SYNOPSIS
.nf
.ft B
#include <net.h>
#include <rreq.h>
#include <events.h>

int	getrent (struct route *rent);
int	getrentc (struct route *rent);
int	getroute (int nodeid, int event);
int	getroute2 (int nodeid, int event);
int	getrtype (int nodeid);
.fi
.SH DESCRIPTION
These functions return information obtained from the local route daemon.
.I Getrent()
accepts a pointer to a route structure, defined in <rreq.h>:
.nf
.ta .5i 1i 1.5i

	struct route {
		int	r_nodeid,
			r_event,
			r_event2,
			r_nodetype;
	};

.fi
The
.I r_nodeid
field must be initialized, before calling
.I getrent(),
to the identifier of the node whose route table entry is required.
.I Getrent()
sends a request for routing information to the route daemon.
Upon return from
.I getrent(),
the fields in the structure hold, in order: the unmodified node identifier,
two events of link processes a message destined for that node would be
forwarded to, and the node type of that node.
The first event,
.I r_event,
is the one used during normal message passing.
The second event,
.I r_event2,
is used by the multicasting daemon and should be guaranteed to provide
a route with a minimum number of hops along the broadcast spanning tree
from that node.
The link process events are defined in <events.h>.
The node type is a combination of flags describing the capabilities of the node.
These flags are defined in <net.h>.
For an explanation of their usage, see getnet(l).
.PP
.I Getrentc()
has the exact functionality of
.I getrent(),
except that it stores the recently accessed route table entries
in a cache.
This speeds up later access to the same entries.
When invoked,
.I getrentc()
checks the cache for the route entry of the given node identifier.
If the entry is found, the requested information is returned.
Otherwise,
.I getrent()
is called, and returns a copy of the routing table entry from the route daemon.
This entry is then stored in the cache.
.PP
.I Getroute()
accepts a node identifier and an event, and returns the event
of the link process that a message destined for that node
should be forwarded to.
If the destination node is the local node,
.I getroute()
returns the supplied
.I event.
Otherwise
.I getrentc()
is called, and the
.I r_event
field of the route table entry for the given node identifier is returned.
.PP
.I Getroute2()
is similar to
.I getroute()
except that it returns the second event field in the route table
.I (r_event2).
.PP
.I Getrtype()
accepts a node identifier and returns the type of that node.
It invokes
.I getrentc(),
and returns the
.I r_nodetype
field from the route table entry for the specified
node.
.SH "RETURN VALUE
Upon successful completion of
.I getrent()
or
.I getrentc(),
0 is returned.
Otherwise, \-1 is returned.
Upon successful completion of
.I getroute()
or
.I getroute2(),
a forwarding event number is returned.
Otherwise, NOTEVENT (defined in <events.h>) is returned.
.I Getrtype()
returns a valid node type when successful.
NOTNODETYPE (defined in <net.h>) is returned if an error occurs,
or if the route daemon does not yet know any network information.
To distinguish between these two cases, errno is set to 0 if the route
daemon has no network information otherwise errno indicates the error.
When an error occurs, these functions set the global variable
.I errno
to the appropriate error code.
These functions have no specific errors of
their own, but can incur any of the errors associated with nsend(l).
.SH "SEE ALSO"
getnet(l), rsetroute(l), rroute(l)
