.TH KCREATE l "April 6, 1992" "Release 2.2" "TROLLIUS LOCAL LIBRARY"
.SH NAME
kcreate \- create a Trollius process from a load module
.SH ITB SYNOPSIS
.nf
.ft B
#include <pcb.h>
#include <t.out.h>

int	kcreate (struct kproc_t *newproc, char **argv);
.fi
.SH OTB SYNOPSIS
.nf
.ft B
int	kcreate (char *filename, char **argv);
.fi
.SH DESCRIPTION
This local layer function is used to create a process on the local
node from a program that is already in memory.
It is a system-level function.
Higher level functions are available that simplify process creation.
.PP
On a compute node (ITB), the program is described with a structure,
defined in <pcb.h>:
.nf

	struct kproc_t {
		struct exec	*kp_header;
		int		kp_rtf;
		int		kp_stksize;
		char		*kp_slink;
		char		*kp_stack;
	};

.fi
The
.I kp_header
field points to a header containing information about the loaded file,
including the number of bytes in the code, data, and bss segments.
It is assumed that the header, code, data, and bss segments are all
contiguous in memory.
.PP
The
.I kp_rtf
field contains runtime flags: 
RTF_SYSGEN indicates that the new process is a system process,
RTF_ONCHIP causes the stack to be allocated from on-chip RAM.
.PP
The
.I kp_stksize
field defines the stack size requirement in bytes.
.PP
The
.I kp_slink
field points to the data/bss segment.
If
.I kp_slink 
is set to 0, the data/bss segment found in the load module will be used.
Processes created with 
.I kcreate()
can share a code segment by using the same
.I kp_header,
but a different
.I kp_slink.
.PP
The 
.I kp_stack
field is set to the stack pointer upon return of the function.
.PP
The
.I argv
parameter points to an 
.I argv
structure consisting of (argc + 1) uninitialized pointers to strings,
followed by argc null-terminated strings.
The strings must be contiguous in memory.
This field is ignored if set to 0.
.PP
On a host node (OTB), the program is described with a file name,
which must contain a valid executable for the local host node.
The new process is created with fork(2) and execvp(3).
.PP
Both ITB and OTB, the new process inherits the parent's stdin, stdout and 
stderr Trollius file descriptors.
.PP
For most purposes, users will prefer to create Trollius processes using
a kpar(l) function, rloadgo(l), rspawn(l), or rcreate(l).
The kpar(l) functions allow a running process to be created on the local ITB
node.
The code for the new process is supplied as a function in the parent process'
code.
Rloadgo(l) loads an executable program on an ITB node or group of ITB nodes,
and creates a running process or processes.
Rspawn(l) duplicates a running process on an ITB node or group of ITB nodes.
Rcreate(l) starts a program running on the local ITB node that was loaded
with rload(l).
.SH "RETURN VALUE"
Upon successful completion of
.I kcreate(),
the new process identifier is returned.
If an error occurred, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 20
[ENOMEM]
There is insufficient memory to create a new process.
.SH "SEE ALSO"
kpar(l), rcreate(l), rload(l), rloadgo(l), rspawn(l)
