.TH KEXIT l "April 6, 1992" "Release 2.2" "TROLLIUS LOCAL LIBRARY"
.SH NAME
kexit, _kexit, kdetach, atkexit \- Detach from Trollius and exit.
.SH SYNOPSIS
.nf
.ft B
int	kexit (int errcode);
int	_kexit (int errcode);
int	kdetach (int pid);
int	atkexit (int (*func)(), char *arg);
.fi
.SH DESCRIPTION
A process that attaches to the Trollius kernel should detach itself
before quitting.
On an ITB node
.I kexit()
is implicitly called after the main program returns, though
and explicit call is recommended for portability with OTB nodes.
Also on an ITB node, exit(3) is equivalent to
.I kexit().
On an OTB node one must call
.I kexit()
explicitly at the end of every Trollius program, since the Trollius kernel
has no other way of knowing that a process has terminated.
The argument to
.I kexit()
is passed to the system when the process exits,
and is conventionally 0 for a normal exit, or some nonzero error code.
.I _Kexit()
may be called instead of
.I kexit()
when cleanup functions, including those for standard I/O, are not required.
This decreases the size of the executable file considerably.
.PP
A process can detach from the local Trollius kernel but keep running by calling
.I kdetach().
This only undoes what kattach(l) did.
.I Kdetach()
is called from
.I kexit().
On ITB nodes, rpdetach(l) is also called from
.I kexit().
The parameter passed to
.I kdetach()
is the process ID of the process to be detached.
A process ID of 0 is interpreted as a request to detach the calling
process, and is the normal usage of
.I kdetach().
Detaching peer Trollius processes is intended as an emergency clean up
measure and is not intended for normal usage.
.PP
Before detaching from Trollius,
.I kexit()
can automatically call any desired cleanup functions.
.I _Kexit()
cannot call cleanup functions and this is its only distinguishing feature.
A cleanup function is first registered by calling
.I atkexit().
The parameters passed to
.I atkexit()
are the address of the function to be called upon
.I kexit(),
and an argument to pass to the cleanup function
when it is called.
(Note that the function may ignore the argument, reverting to the
ANSI standard C library atexit(3) behavior, where no argument is passable.)
There is a limit on how many cleanup functions may be registered.
Cleanup functions are called in reverse order of their registration.
Some system modules register their own cleanup functions,
including memory allocation and file I/O.
.PP
To enhance portability of applications, there are wrappers in the
Trollius library such that, on compute nodes, a call to
exit(3) actually calls
.I kexit(),
and a call to
atexit(3)
actually calls
.I atkexit().
This is
.I not
true on OTB nodes.
.SH RETURN VALUE
.I Kexit()
and
.I _kexit()
are not supposed to return.
.I Kdetach()
and
.I atkexit()
return 0 upon successful completion.
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.I Kdetach()
fails if:
.TP 25
[ENOTPROCESS]
The process whose process identifier was passed to
.I kdetach()
is not attached to the kernel.
.PP
.I Atkexit()
fails if:
.TP 25
[ENOEXDESCRIPTORS]
There is no room left in the
.I kexit()
tables.
.SH BUGS
If too many user functions are registered with
.I atkexit()
before calling system functions that also call
.I atkexit(),
the latter calls to
.I atkexit()
may fail and the system functions may fail to clean up.
If the missing cleanup is file I/O, some output data
may be lost.
.SH SEE ALSO
kinit(l), rpdetach(l)
