.TH KINIT l "April 6, 1992" "Release 2.2" "TROLLIUS LOCAL LIBRARY"
.SH NAME
kinit, kattach \- Attach process to Trollius kernel.
.SH SYNOPSIS
.nf
.ft B
int	kinit (int priority);
int	kattach (int priority);
.fi
.SH DESCRIPTION
.I Kattach()
initializes communication between the calling process and the Trollius kernel.
.I Kinit()
initializes data structures necessary for interfacing with the Trollius
kernel, and calls
.I kattach().
.PP
Every OTB program that runs under Trollius must call
.I kinit()
before calling any Trollius functions.
ITB processes must call
.I kinit()
only if they are to run at a non-zero priority or are started with kpar_Fe(l).
An implicit call to
.I kinit(0)
is made before ITB programs start.
It is a good portability practise for ITB programs to call
.I kinit()
explicitly.
.PP
The argument passed to
.I kinit()
and
.I kattach()
is an arbitrary positive integer representing the process's scheduling
priority.
The maximum priority than can be assigned to a user process is 100000.
This ensures that user processes will be at a lower priority than the
buffer process.
Processes with priority zero (0) are called uncontrolled; they are
released by the kernel to be scheduled by the underlying hardware or
software whenever they are not blocked on a Trollius event.
All other non-zero priorities cause a process to be more carefully
controlled by the kernel.
Specifically, only the highest priority controlled process not blocked
on an event is released for scheduling by the underlying hardware or
software.
All other unblocked controlled processes are impeded by the kernel until
the highest priority controlled process blocks.
At that point, the next highest (now the highest) priority unblocked
controlled process is released by the kernel.
The Trollius kernel does not timeslice/round-robin controlled processes.
Processes with identical controlled priorities are ordered as they
registered using
.I kattach()/kinit()
with respect to time.
.PP
Controlled processes incur more overhead than uncontrolled processes.
If predictable process scheduling is not required, priority 0 should be used.
.PP
After a call to
.I kinit(), kattach()
may be called repeatedly without an intervening kdetach(l) to alter the
priority of the calling process.
.SH "RETURN VALUE"
Upon successful completion of
.I kinit()
or
.I kattach(),
0 is returned.
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.PP
.I Kinit()
may fail on OTB nodes due to errors arising from the native operating system.
.I Kattach(),
which is called from
.I kinit(),
fails if:
.TP 22
[ENOKDESCRIPTORS]
The Trollius kernel is out of process descriptors and is unable to attach
additional processes.
.SH SEE ALSO
kexit(l)
