.TH KPAR l "April 6, 1992" "Release 2.2" "TROLLIUS LOCAL LIBRARY"
.SH NAME
kpar_He, kpar_Na, kpar_Fe \- Create a lightweight process.
.SH SYNOPSIS
.ft B
int	kpar_He (void (*func)(), char *stkptr, int nparms, int parm1, 
int parm2,...);
.br
char	*kpar_Na (void (*func)(), int stksize, int nparms, int parm1, 
int parm2,...);
.br
char	*kpar_Fe (void (*func)(), int stksize, int *pid, int nparms, int parm1, 
int parm2,...);
.SH DESCRIPTION
These functions allow a process running on an ITB node
to create another, parallel, process on the same node.
The two processes share the processor's time.
When the function returns the process is destroyed.
.PP
The code for the parallel process is supplied as a function in
the parent process' code.
The code is used as given, without copying to another location.
This differentiates
.I kpar
from kcreate(l).
The function can accept arguments but cannot return a value.
Each argument is one word 
in size (T414/T800: 4 bytes in a word, int, long or pointer).
A double requires two words and counts as two arguments.
.PP
The three versions of the function to create lightweight processes
are denoted
.I "kpar_He(), kpar_Na(),"
and
.I kpar_Fe().
The three versions implement three different versatility / overhead
compromises.
.PP
.I Kpar_He(),
the lightest weight version of
.I kpar,
does nothing but set up the arguments for the new process on the stack
(workspace) and create the process.
.I Kpar_He()
accepts the address of the function to run as a process, the
.B top
address of an area to be used as the stack,
and the number of arguments to be passed
to the function, followed by the arguments themselves.
The stack address passed must point to memory
immediately above the memory block to be used.
The stack area must be allocated before calling
.I kpar_He(),
and may be reused after the process terminates.
For maximum speed, the stack can be allocated from ONCHIP RAM using rmalloc(l).
The child process created via
.I kpar_He() 
.B "must not"
attach to Trollius, i.e., must not call kinit(l) or kattach(l),
or any other Trollius services.
The child can communicate with the parent through the shared static data
segment and through other hardware services (such as the in() and out()
functions on transputer ITB nodes).
.PP
.I Kpar_Na(),
the medium weight version of
.I kpar,
is the same as
.I kpar_He()
except that it takes the desired stack size as an argument
and allocates the necessary memory using rmalloc(l).
The allocated stack area is not released when the child process terminates.
The parent process may reuse it (after the child is known
to have terminated) and should release it with rfree(l) when it is
no longer required.
.PP
The function code for both
.I kpar_He()
and
.I kpar_Na()
should be reentrant.
.PP
.I Kpar_Fe(),
the heavyweight version of
.I kpar,
creates a complete Trollius process, including its own static data segment.
The data segment starts out identical to the parent's data segment, but
it is a copy, so results cannot be returned through static variables.
The process created by
.I kpar_Fe()
may call kinit(l) and then use Trollius services, such as message passing.
.SH "RETURN VALUE"
Upon successful completion of
.I kpar_He(),
0 is returned.
If an error occurred, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.PP
.I Kpar_Na()
and
.I kpar_Fe()
return a pointer to the stack and/or data area they have allocated with
rmalloc(l).
This pointer can be used with rfree(l) to release the memory.
Upon error, a NULL pointer is returned and the global variable
.I errno
is set to indicate the error.
.PP
.I Kpar_Fe()
sets the 
.I pid
argument to the process ID of the created process.
.SH ERRORS
.TP 20
[ENOMEM]
There is insufficient memory to create the process.
.SH BUGS
The data segment of a process created with
.I kpar_Fe()
will not contain valid pointers to other data within its new segment.
These pointers will still reference the parent's data segment.
Text pointers and those returned by rmalloc(l) will point to the
desired code and data, respectively.
.SH "SEE ALSO"
kcreate(l)
