.TH KSEND l "April 6, 1992" "Release 2.2" "TROLLIUS LOCAL LIBRARY"
.SH NAME
ksend, ktry_send, krecv, ktry_recv  \- Send and receive local node
Trollius messages.
.SH SYNOPSIS
.nf
.ft B
#include <kmsg.h>

int	ksend (struct kmsg *header);
int	ktry_send (struct kmsg *header);
int	krecv (struct kmsg *header);
int	ktry_recv (struct kmsg *header);
.fi
.SH DESCRIPTION
These functions access Trollius local message passing services.
.I Ksend()
sends a message to another process on the same node.
.I Krecv()
receives a message from another process on the same node.
The Trollius kernel arranges for the two processes to meet.
The process calling
.I ksend()
always blocks unless a receiver is already waiting, blocked on
.I krecv().
The reverse is also true.
.PP
.I Ktry_send()
and
.I ktry_recv()
are similar to
.I ksend()
and
.I krecv()
, but they never block.
The message is transferred if a synchronizing process is waiting.
If not,
.I errno
is set to EWOULDBLOCK.
Never use the non-blocking functions on both sides, or they will never meet.
.PP
The parameter passed to these functions
is a pointer to a local message structure, the fields of which 
must be filled before the message is sent or received. 
The
.I kmsg
structure is defined in <kmsg.h> as:
.PP
.nf
.RS
.DT
struct kmsg {
	int	k_event,
		k_type,
		k_length,
		k_flags;
	char	*k_msg;	
};
.RE
.fi
.PP
The
.I k_event
field serves to synchronize the message passing.
The sender and the receiver must specify the same positive integer.
This field remains unchanged after calling
.I ksend()
or
.I krecv().
.PP
The
.I k_type
field filters messages further.
A message will be passed only if the 
.I k_type 
fields of the sending and receiving processes have at least one bit set
in an identical position.
In other words, the bitwise logical AND of the type fields specified by
the two parties must not equal zero.
A zero value in the 
.I k_type 
field is a special case -
it will ``fit'' any other caller using the same
.I k_event.
This field remains unchanged after calling
.I ksend(),
but is set to the sender's
.I k_type
field after calling
.I krecv().
.PP
The
.I k_length
field holds the length (in bytes) of the message to be sent.
If the sender and the receiver specify different lengths,
the lesser amount will be transferred.
This field is set to the minimum of the sender's and receiver's length
after calling
.I ksend()
or
.I krecv().
.PP
The
.I k_flags
field is normally set to 0.
It can be set to KHOLD (defined in <kmsg.h>) to request that a virtual 
circuit be kept open for the next message on the same event.
.PP
The
.I k_msg
field holds the address of the first byte of data to be sent.
The data must be stored contiguously in memory.
This field remains unchanged after calling
.I ksend()
or
.I krecv().
.SH VIRTUAL CIRCUITS
In Trollius, the kernel is not an executive with complete control of
the processor, but a server providing a standardized rendez-vous service
on top of a lower layer of process management.
On Trollius compute (ITB) nodes, this lower layer can be implemented
by hardware.
On Trollius host (OTB) nodes, this lower layer is implemented by the
native operating system.
.PP
Using the Trollius kernel is convenient, powerful, portable and expensive.
Therefore, virtual circuits have been provided to bypass the kernel
once an initial rendezvous has been made.
This is done by setting the
.I k_flags
field of both the sending and receiving processes'
local message descriptors to KHOLD.
Setting KHOLD causes
.I ksend()
or
.I krecv()
to remember the rendez-vous information learned from the kernel and
associate it with future uses of the same event, without consulting
the kernel again.
After the first message establishing the virtual circuit,
there is no protection against mismatched message lengths. 
If the sender and receiver specify different values in the
.I k_length
field, the program has every right to fail.
.PP
A virtual circuit is bi-directional, and there can be up to 16 simultaneous
virtual circuits for each process.
The restriction (for the sake of performance) is that all active
virtual circuit events must differ in the last four bits.
(In other words, event MOD 16 must be distinct.)
.PP
Using a virtual circuit event with the KHOLD flag reset (0) sends
or receives the message along the virtual circuit, but terminates
the circuit after transmission.
Communicating parties must set and reset a virtual circuit on the
same message, or both parties will block and the node may fail.
Use virtual circuits with care only after debugging an application without them.
.SH RETURN VALUE
Upon successful completion, 0 is returned.
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 20
[EWOULDBLOCK]
.I ktry_send()
or
.I ktry_recv()
was called and returned an error condition because the message
could not be immediately sent.
No process is blocked on a matching event and type.
At that time, a call to
.I ksend()
or
.I krecv()
would have blocked.
.TP
[ENOEVDESCRIPTORS]
The Trollius kernel ran out of message descriptor space.
.SH "SEE ALSO"
nsend(l), nrecv(l)
