.TH KSIGBLOCK l "March 16, 1992" "Release 2.2" "TROLLIUS LOCAL LIBRARY"
.SH NAME
ksigblock, ksigsetmask, ksigretry, ksigsetretry, kpause \- Manipulate Trollius signal handling policy.
.SH SYNOPSIS
.nf
#include <ksignal.h>

int	ksigblock (int mask);
int	ksigsetmask (int mask);
int	ksigretry (int mask);
int	ksigsetretry (int mask);
int	kpause ();
.fi
.SH DESCRIPTION
.PP
As part of the signal handling package, Trollius maintains two system
masks, a block mask and a retry mask.
Each contains a bit for every defined signal.
The block mask will prevent a signal from being handled if its
corresponding bit is set.
The retry mask indicates whether a system call to the kernel should be
retried after an given interrupt.
Signals are defined in <ksignal.h> and
listed on the ksignal(l) man page.
.PP
.I Ksigblock()
resets the block mask to 
block each signal whose corresponding bit in the mask parameter
is set, as well as those whose bits were set in the original block mask.
.I Ksigsetmask()
explictly sets the system block mask to the mask parameter.
All signals set in the mask parameter become blocked - all others
become unblocked.
Both routines return the previous value of the block mask.
.PP
All signals begin as unblocked, and privileged signals cannot be blocked.
A blocked signal is not ignored - it is impeded.
If the block should be removed for a signal that was pending, the
signal handler will immediately be called.
Several impeded signals will fuse together as one and not be
handled distinctly.
.PP
Since Trollius signals are only delivered during system calls to the
kernel, there is an excellent chance that a kernel system call will be
interrupted.
When this happens, the system call returns with the error EINTR.
If the bit corresponding to the delivered signal is set in the
retry mask, the system call will be automatically
retried after the interrupt handler returns.
.PP
.I Ksigretry()
resets the retry mask to 
retry each signal whose corresponding bit in the mask parameter 
is set, as well as those whose bits were set in the original retry mask.
.I Ksigsetmask()
explictly sets the system retry mask to the mask parameter.
All signals set in the mask parameter will be retried - all others
will not be retried.
Both routines return the previous value of the retry mask.
After calling kinit(l), all signals will be retried except SIGRELEASE,
which cannot be retried.
.PP
The macro \fIsigmask(signum)\fR, in <ksignal.h>, converts a signal number
into a mask with just the corresponding bit set.
.PP
.I Kpause()
causes the calling process to wait until any signal is delivered.
.SH "RETURN VALUE"
.PP
\fIKsigblock()\fR, \fIksigsetmask()\fR, \fIksigretry()\fR and
.I ksigsetretry()
all return the previous mask value.
.I Kpause()
always returns \-1 and the global variable
.I errno
is always set to EINTR.
.SH "SEE ALSO"
ksignal(l), kdoom(l)
