.TH KSTATE l "April 6, 1992" "Release 2.2" "TROLLIUS LOCAL LIBRARY"
.SH NAME
kstate \- Get kernel state of a process.
.SH SYNOPSIS
.nf
.ft B
#include <kreq.h>

int	kstate (int pid, struct kstate *pstate);
.fi
.SH DESCRIPTION
The function
.I kstate()
enables a Trollius process to inquire about the kernel state of
any process on the local OTB or ITB node, including itself.
The argument
.I pid
is the identifier of the process to be stated.
The special identifier 0 indicates the calling process.
.I Pstate
is a pointer to a structure where information from the kernel will be stored.
The structure is of type kstate, defined in <kreq.h>:
.nf

	struct kstate {
		int	ks_state;
		int	ks_priority;
		int	ks_event;
		int	ks_type;
		int	ks_signal;
	};

.fi
Upon return from
.I kstate(),
the fields in the structure hold:
the state of the process (see below),
the priority of the process as requested in the call to kinit(l) or kattach(l),
the event and type (if any) on which the process is blocked, and the hexadecimal
representation of all pending signals delivered by kdoom(l).
The process states, defined in <kreq.h>, are as follows:
.TP 15
SSYNC
ready, having called ksend(l) or krecv(l), often seen as SREADY.
.TP
SSTATE
ready, having called
.I kstate().
.TP
SERROR
ready, having encountered an error on any kernel request.
.TP
SATTACH
ready, having called kattach(l).
.TP
SREADY
ready, having called kdetach(l), kdoom(l), ksend(l) or krecv(l).
.TP
SSEND_BLKD
blocked sending a message to an event.
.TP
SRECV_BLKD
blocked receiving a message from an event.
.TP
SACTIVE
running - not under Trollius kernel control.
.SH "RETURN VALUE
Upon successful completion, 0 is returned.
Otherwise \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 20
[ENOTPROCESS]
The process whose process identifier was passed to
.I kstate()
is not attached to the kernel.
.SH "SEE ALSO"
rstate(l), state(l)
