.TH KTRSTART l "April 6, 1992" "Release 2.2" "TROLLIUS LOCAL LIBRARY"
.SH NAME
ktrstart, ktrstop  \- Start and stop Trollius application tracing.
.SH SYNOPSIS
.nf
#include <net.h>

void   ktrstart (int tid);
void   ktrstop (void);
.fi
.SH DESCRIPTION
.I Ktrstart()
turns on tracing and
.I ktrstop()
turns it off.
A Trollius application should be started using loadgo(l) with the -t option
if tracing is desired.
In this case, an implicit call to
.I ktrstart()
occurs at the call to kinit(l), and an implicit call to
.I ktrstop()
occurs at the call to kexit(l).
The implicit call to
.I ktrstart()
uses the nodeid for the trace identifier (tid).
To generate traces for a Trollius
application using the default tids, the programmer need never call
.I ktrstart()/ktrstop().
.PP
These functions are available to the programmer to provide flexibility in
two respects.
First, the programmer may wish to designate his/her own trace identifiers
rather than using the defaults.
This is accomplished by following the call to kinit(l) with a call to
.I ktrstop()
to cancel the implicit call to
.I ktrstart().
Then, a call to
.I ktrstart()
using the desired tid is made.
There are several restrictions to this.
Tids must start at zero and be contiguous.
The maximum number of tids currently supported is 512, numbered 0 through 511.
They must also be unique system-wide.
That is, no two nodes can have the same tid set.
Finally, an application can only change its tid from the default once;
traces with different tids should never be generated by the same process.
The exception to this is, of course, the trstart/trstop records generated by
calling
.I ktrstart()/ktrstop().
These trace records will be removed by trade(l) during processing for
viewing with ParaGraph.
.PP
The other way that these functions are used is to turn tracing on and off
to ensure that traces are generated only during certain phases of the
application's execution.
This can be used to considerably reduce the size of the trace files and
simplify viewing with ParaGraph.
Extreme care should be used with this feature, however.
All the restrictions on setting tids as described above still apply.
In addition, the programmer must ensure that matching sends and receives
on different nodes are either both traced or both not traced.
For example, if an nsend(l) is issued by a node which has tracing turned off
and the message is read with an nrecv(l) on a node which has tracing turned
on, ParaGraph will see an nrecv(l) complete for which it can find no matching
nsend(l).
This will cause spurious lines on the display, generally the nsend(l) will
appear to have occurred at time zero.
If an nsend(l) is issued by a node which has tracing turned on and the
message is received by a node which has tracing turned off, this nsend(l) will
be matched by ParaGraph to the first nrecv(l) occurring after tracing is
turned back on.
This will cause sends and receives to be erroneously matched-up in the display.
.PP
The trace structures which are generated by Trollius are defined in <ktrace.h>
as:
.PP
.nf
.RS
.DT

/*
 * trace structures
 */

struct trtr {
	int			trt_type;
	int			trt_stamp;
	int			trt_tid;
};

struct trstart {
	struct trtr	tr1_head;
	int			tr1_node;
	int			tr1_pid;
};

struct trend {
	struct trtr	tr2_head;
};

struct trsend {
	struct trtr	tr5_head;
	int			tr5_tonode;
	int			tr5_event;
	int			tr5_length;
	int			tr5_total;
	int			tr5_blocked;
	int			tr5_transmit;
};

struct trrecv {
	struct trtr	tr6_head;
	int			tr6_frtid;
	int			tr6_event;
	int			tr6_length;
	int			tr6_total;
	int			tr6_blocked;
	int			tr6_transmit;
};
.RE
.fi
.SH LIMITATIONS
.PP
In addition to the limitations imposed above, it is important to note that
if trace generation is enabled, the nh_data[7] field of the network message
descriptor is not available to the program.
This means that the programmer should not use this field in nsend(l) or
nrecv(l) in an application which is to be run with tracing.
.SH BUGS
If traced runs out of trace buffers, tracing is permanently disabled for that
node until ttrace(l) is used or until Trollius is restarted.
This means that the trace file recovered by ttrace(l) may have been truncated,
and some trace records generated by
.I ktrstart(), ktrstop(),
nsend(l) and nrecv(l) may be lost.
.SH "SEE ALSO"
loadgo(l), ttrace(l), trade(l), nsend(l)
