.TH LOADGO l  "April 7, 1992" "Release 2.2" "TROLLIUS COMMANDS"
.SH NAME
loadgo \- Load and execute files on Trollius ITB nodes.
.SH SYNTAX
.B loadgo
[-fhtvwy] [-i <infile>] [-o <outfile>] [-s <size>] <node(s)> <program(s)> [-- <args>]
.SH OPTIONS
.TP 14
.B \-f
Do not open Trollius standard file descriptors - use defaults.
.TP
.B \-h
Print useful information on this command.
.TP
.B \-i <infile>
Redirect standard input from \fIinfile\fR.
.TP
.B \-o <outfile>
Redirect standard output to \fIoutfile\fR.
.TP
.B \-s <size>
Set stack size to
.I size
bytes.
.TP
.B \-t
Turn on trace generation for the new processes.
.TP
.B \-v
Turn on verbose mode.
.TP
.B \-w
Wait for process completion before exiting \fIloadgo\fR.
.TP
.B \-y
Allocate stack from fast on-chip RAM.
Use in conjunction with \fI-s\fR.
.TP
.B -- <args>
Pass these runtime arguments to every new process.
This must always be the last argument to \fIloadgo\fR.
.SH DESCRIPTION
.PP
.I Loadgo
is a Trollius command that loads an executable program from the file
system of the local OTB (host) node onto the specified ITB node, and
creates a new running process with the given runtime arguments.
A single
.I loadgo
command may specify multiple nodes and multiple program files, but
only one set of runtime arguments may be specified for all new processes.
.I Loadgo
loads all the listed programs on all the listed ITB nodes.
To load different programs on different nodes, or to pass
different arguments to different programs, multiple
.I loadgo
invocations are necessary.
The files must be ITB executable load modules.
.PP
ITB nodes are identified as
.I n<list>,
where <list> can be a single node identifier or a list of node identifiers.
Node identifiers are established in the boot schema (see nail(l)), and
can be written in decimal and/or hexadecimal notation.
An example node list is n0-7,8,10,0xc,0x10-31.
A
.I loadgo
request can be broadcast to all ITB nodes by specifying the
.I C
multicast node identfier.
.PP
Normally,
.I loadgo
exits immediately, without waiting for the new processes to complete.
If the
.I \-w
option is used,
.I loadgo
will wait for all processes to complete, and will print a message if a
non-zero exit status, indicating an error, was obtained from any process.
.PP
The default stack size is PSTKSIZE, defined in <preq.h>,
which is set to 8 Kbytes.
A different stack size can be specified with the
.I \-s
option.
The
.I \-y
option places the stack in on-chip RAM.
Since there are only about 3 Kbytes of user-available on-chip RAM on the
T800, the
.I \-s
option must also be used to set the stack size.
.PP
Standard I/O for the created processes will appear at the terminal/window
from which
.I loadgo
was invoked.
Stdin and stdout can be redirected to files with the
.I \-i
and
.I \-o
options respectively.
Stderr cannot be redirected.
None of the three can be redirected to a pipe.
To avoid all interaction with the Trollius remote file daemon, use the
.I \-f
option and the defaults (same window/terminal that invoked spread(l)) will
be used.
.PP
The
.I \-t
option enables trace generation for the created processes.
The trace daemon must be running on the target nodes prior to
using this option.
.SH EXAMPLES
.TP 4
.I "loadgo n0-7 prog1"
Load and execute prog1 on nodes 0 through 7.
Do not wait for all processes to complete before exiting
.I loadgo.
.TP
.I "loadgo n8 -vwy -s 1024 prog1 prog2 -- -q
Load and execute prog1 and prog2 on node 8.
Report on each process as it is created.
Use an on-chip stack of one Kbyte and pass the -q option to each new process.
Wait for the processes to complete before exiting loadgo.
.SH DIAGNOSTICS
.PP
Due to lack of memory protection on the transputer, a defective loaded program
can overwrite and destroy the ITB node operating system.
Should
.I loadgo
fail to return, check the target node(s) with tping(l) and state(l).
If either of these commands fail to operate,
the node, and possibly the whole Trollius system, must be rebooted.
Even
.I loadgo
may hang if the node crashes before an acknowledgement
of a successful load and creation can be transferred off of the node.
.SH SEE ALSO
rloadgo(l), state(l), doom(l)
