.TH MALLOC l "April 7, 1992" "Release 2.2" "TROLLIUS LOCAL LIBRARY"
.SH NAME
malloc, free, calloc, realloc
\- high-level Trollius memory (de)allocation functions
.SH SYNOPSIS
.nf
.ft B
char	*malloc (unsigned int amount);
int	free (char *address);
char	*calloc (unsigned int nelem, unsigned int elsize);
char	*realloc (char *address, unsigned int amount);
.fi
.SH DESCRIPTION
The memory allocation functions
.I malloc()
and
.I free()
provide higher-level, portable access to rmalloc(l) and rfree(l)
(See the rmalloc(l) man page).
They are restricted to local allocation (inside the same node),
imply the region ANYBANK, and only guarantee word-alignment.
If you wish to allocate memory on a remote node or from a specific
region, or have special address alignment requirements, use rmalloc(l)
and rfree(l).
.PP
The first time
.I malloc()
is called in a program, it calls rmalloc(l) to obtain a chunk of memory
of a minimum size.
Subsequent calls to
.I malloc()
utilize this chunk from rmalloc(l) until it is used up, at which point
a new chunk is obtained from rmalloc(l).
The amount of memory that is obtained from rmalloc(l), if necessary,
by a call to
.I malloc()
is at least equal to the
.I amount
parameter passed to
.I malloc(),
plus a small overhead.
This amount is rounded up to a multiple of
ALLOCBLK, and has a minimum value of ALLOCMIN.
ALLOCBLK and ALLOCMIN are defined in <malloc.h>, and are currently
set to 1 Kbyte and 4 Kbyte, respectively.
Accessing chunks of memory obtained from rmalloc(l) by previous calls to
.I malloc()
does not involve communication with the malloc daemon (mallocd).
Thus,
.I malloc()
and
.I free()
are two orders of magnitude faster than rmalloc(l) and rfree(l)
when allocating and reallocating many small blocks of memory.
.PP
Blocks allocated by
.I malloc()
must be deallocated by
.I free(),
and the call to
.I free()
must be made from the same process that called
.I malloc().
The blocks that are freed are not returned to the system until all
other blocks residing within the same larger block (that was
implicitly obtained by an rmalloc(l) call) are also freed.
This fact can be used to advantage.
If an application repeatedly frees all of a malloc-ed chunk of memory,
performance will suffer due to redundant implicit calls to
rmalloc(l) and rfree(l).
On the other hand,
if a small block is always kept unfreed, the remainder of that block can
be reused, and unnecessary interaction with the system can be avoided.
In any case, all memory is returned to the system when the program exits.
.PP
The function
.I calloc()
is a variant of
.I malloc().
It is given the number of ``elements'' to be allocated,
and the size, in bytes, of each.
It simply allocates, using
.I malloc(),
an amount equal to the product of the two, and initializes the
allocated memory to all zeros.
.PP
The function
.I realloc()
allows one to enlarge or shrink a block previously obtained via
.I malloc().
The parameters passed to
.I realloc()
are the former address of the block, and the new size desired.
If shrunk, the part of the block starting at the lowest memory
address is retained.
If enlarged, additional memory space is appended at the high end of the block.
If necessary, the block is moved to a new location and the old data copied.
In either case,
.I realloc()
returns the address of the block.
.PP
Blocks returned from
.I calloc()
and
.I realloc()
are deallocated using
.I free().
One must be careful to pass pointers returned by
.I malloc(), calloc()
and
.I realloc()
only to
.I free()
or
.I realloc(),
and not to rfree(l).
Similarly, pointers returned by rmalloc(l) can be passed to rfree(l), but not to
.I free()
or
.I realloc().
Furthermore, one must not pass to
.I free()
a pointer to a block that was returned from
.I malloc()
in another process.
A thread created by a call to
.I kpar_Fe(l)
is considered a new process for this, as for most, purposes.
.SH "RETURN VALUE"
.I "Malloc(), calloc(),"
and
.I realloc()
return the address of the allocated block.
If the request cannot be satisfied, NULL (0) is returned.
Upon successful completion of
.I free(),
0 is returned.
Otherwise \-1 is returned.
In all cases of error, the global variable
.I errno
is set to one of the following error codes.
.SH ERRORS
.TP 24
[ENOMEM]
An allocation request cannot be satisfied because the amount requested
is too big, too small (0 bytes), or memory is exhausted or fragmented.
.TP
[ENOTMALLOCED]
The memory block was rejected by
.I free()
since it was never obtained by
.I malloc()
by this process, or because of an intervening call to rfreeall(l).
.TP
[EGARBAGE]
The internal lists kept by
.I malloc()
have been corrupted, probably due to the program (or another process)
writing to locations outside of the allocated blocks.
.SH "SEE ALSO"
rmalloc(l), mstate(l)
