.TH MAP l  "April 7, 1992" "Release 2.2" "TROLLIUS TOPOLOGY TOOLS"
.SH NAME
map \- Compile routing tables from a Trollius boot schema file.
.SH SYNTAX
.B map
[-dhsvwVW][-c <bnail>][-i <extra>][-p <route>]
.br
.SH OPTIONS
.TP 13
.B \-c <bnail>
Use
.I bnail
as the boot schema file.
.TP
.B \-d
Turn on debugging.
.TP
.B \-h
Print useful information on this tool.
.TP
.B \-i <extra>
Increase the size of the route table for each node by
.I extra
entries.
.TP
.B \-p <route>
Use
.I route
as the route file.
.TP
.B \-s
Report statistical information on the routes generated.
.TP
.B \-v
Turn on verbose mode.
.TP
.B \-w
Do not write the routes generated to file.
.TP
.B \-V
Format and print the boot schema.
.TP
.B \-W
Ignore the link weights when generating the second routing tables.
.SH DESCRIPTION
.I Map
is the topology tool that generates routing tables for a Trollius system from
the topology specified in a boot schema file (see nail(l)).
Unless overridden by the
.I \-p <route>
option, the route file containing the tables is named
`route' appended by the identical suffix (if any) of the boot schema filename.
A route file must be generated from the boot schema via
.I map
or rasm(l) before the system can be booted with spread(l).
.PP
Routing tables tell which link to pass a message to, intermediate to
reaching a given destination node.
Spread(l) calls trout(l) which reads the route file and
sends each node its specific routing information, to be used by the router
daemon during message-passing and multicasting operations.
.PP
A boot schema is a description of a bootable multicomputer system, specifying
all physical nodes and the links that connect them.
To this information may be added routing table entries or information on the
order in which unspecified routes are to be generated.
.I Map
is designed to be flexible, to match the variable amount of information in the
boot schema.
It compiles routing tables representing a fully connected topology.
It will incorporate the route table entries given in the boot schema and
generate all missing routes, thus providing full network connectivity.
.PP
.I Map
generates two separate routing tables for each node, the first one being
used for regular message passing, the second for multicasting and
multireeling operations.
In brief, while generating the first routing tables,
.I map
attempts to avoid congestion on a given link by assigning a "weight"
to a link that is part of a path.
A path is selected for having a lower weight than alternate paths.
After selecting a path,
.I map
increments the weights of the intermediate links along the path,
and uses the updated weights to determine the "lightest" path for the
subsequent pair of nodes.
Thus the order in which the pairs of nodes are considered is important and a
different order may result in a different set of valid routing tables.
The generation of the second routing tables follows a strict minimum-hop
spanning tree approach.
Link weights as defined in the boot schema represent the size of each hop.
For each node in the system,
.I map
generates a spanning tree minimizing the hop sizes, irrespective of
the order in which the nodes are considered.
Multicasting (multireeling) from (to) that node will follow the paths
formed by the branches of that tree.
A more detailed discussion of route generation by the
.I map
tool is given below.
.SH ROUTE GENERATION
The boot schema incorporates topology information in the following tuples
(see nail(l)):
.TP 14
TWEIGHT
Sets default link and path weights.
.TP
TNODE
Lists Trollius node identifiers and their types.
.TP
TLINK
Lists all unidirectional physical links.
.TP
TROUTE
Specifies one entry in one node's routing table.
.TP
TORDER
Gives priority to generating the optimal paths for the listed
non-neighboring nodes.
.PP
The boot schema must have the TNODE and TLINK tuples completely specified.
Information for TWEIGHT, TROUTE, and TORDER is optional.
In case of duplication or conflict of information between various
tuples, the first one to appear in the boot schema is chosen.
The subsequent tuples do not override the previous information.
.PP
The TWEIGHT tuple sets default link weights for links in four types
of paths:  to nearest neighbors, to the host node, between nodes specified
in TORDER tuples, and in paths generated for multicasting.
The fifth element in TWEIGHT is hop_wt which controls the relative importance
of the lengths (number of hops) and weights (usage of links) of optimal paths.
Default nearest-neighbor and TORDER link weights can be overridden by
weights specified in the TLINK and TORDER tuples, respectively.
In the following discussion, the link weight should be assumed to be the
appropriate weight for the given tuple.
.PP
In generating the first routing tables,
.I map
starts by filling the information available from the TNODE, TLINK and TROUTE
tuples.
It then begins to generate optimal (lightest and/or shortest) paths for all
pairs of nodes not yet connected, following a certain order.
An optimal path between a pair of nodes is the path with the minimum cost,
where the cost of a path is determined by the sum of two quantities.
One is the path weight which is the sum of weights of the links forming
the path.
The second is the path length which is the number of links along the path
multiplied by the value of hop_wt.
Setting hop_wt to zero makes path length not relevant.
Setting hop_wt to a large value effectively forces shortest paths because it
makes it relatively expensive to take a hop.
.PP
The first step is to generate paths from every node in the system to the
origin host node, and back.
The weights of the links along these paths are incremented
by the appropriate weight.
The
.I map
tool then generates the paths specified in TORDER tuples, in the order in which
they appear in the boot schema, and increments the appropriate link weights.
The final phase is to ensure full connectivity of the topology.
In this phase, the
.I map
tool generates optimal paths for the remaining pairs of nodes for which
routes do not yet exist.
Link weights again are incremented.
The order in which the routes are generated in this phase is random.
.PP
In general, the earlier a path between two nodes is generated, the higher
its quality.
This is due to the fact that, as the route tables get filled, there are less
choices available for routing through the system.
This gradual reduction in flexibility ultimately affects the quality (number
of hops and total weight) of the paths being generated.
The user has the flexibility of modifying the order in which paths are
generated through judicious use of the TORDER tuples, and the power to
bias the shape of the paths by manipulating the various weights available.
Ultimate control can be achieved by setting some or all of the route table
entries by using the TROUTE tuples.
.PP
In generating the second routing tables,
.I map
creates for each node in the system a minimum-hop spanning tree rooted at
that node.
The size of each hop is determined by the weight given to the link in the
boot schema.
If the
.I \-W
option is used, the link weights are ignored and each link becomes
effectively a size-one hop.
This allows a faster algorithm to be used in generating the second routing
tables but the quality of the routes suffers a slight degradation.
This is due to the fact that all links are treated as equal whereas in reality
there usually is a difference in performance between links connecting two ITB
nodes and links that connect ITB nodes to OTB nodes.
The route tables are filled such that any message originating from that node
and destined to a multicast group of nodes, will traverse the system following
the paths formed by the branches of that spanning tree.
Note that the user has less control over the generation of the second routing
tables as they are not affected neither by the TROUTE tuples nor by the
TORDER tuples.
The TROUTE entries affecting the second routing tables are disregarded by the
.I map
tool.
These entries are only useful when the
.I rasm
tool is used.
.SH EXAMPLES
.TP 4
.I "map -v"
Generate routing tables from the default boot schema file.
The routing tables are stored in the default route file.
The verbose option reports important stages in the generation process.
.TP
.I "map -s -c bnail.mytree"
Generate the routing tables using the system described in the custom
boot schema file `bnail.mytree' and store them in `route.mytree'.
Work silently but report routing statistics when done.
.TP
.I "map -sw -W -c bnail.hmmm"
Generate the routing tables using the system described in the custom
boot schema file `bnail.hmmm' but do not store the resulting routes.
Work silently but report routing statistics when done.
Use the faster algorithm to generate the second routing tables by
ignoring the link weights.
This is typically used when evaluating several boot schemata, comparing
the qualities of the routes generated with different settings of the
various weights and TORDER tuples.
.SH FILES
.TP 38
$TROLLIUSHOME/boot/bnail.def
default boot schema file
.TP
$TROLLIUSHOME/boot/route.def
default route file
.SH "SEE ALSO"
nail(l), rasm(l), spread(l), trout(l)
