.TH NAIL l  "April 7, 1992" "Release 2.2" "TROLLIUS TOPOLOGY TOOLS"
.SH NAME
NaIL \- Trollius Node and Interconnect Language
.SH DESCRIPTION
.PP
The
.I Node and Interconnect Language (NaIL)
is a grammar which describes an arbitrary interconnected network of nodes
and links constituting a multicomputer system.
In Trollius, it is used to precisely specify the configuration
of host and compute nodes that will be booted with the operating system.
This specification is called the
.I boot schema.
By driving link reconfiguration, booting, and message routing from a
boot schema, complete topology independence is achieved.
Eventually, NaIL will be used to build other types of schemata,
ranging from the complete description of all installed hardware
(as opposed to the subset of nodes and links used in a particular
user's incarnation of Trollius) to a higher level description of
the computation and communication of an application.
.PP
The NaIL grammar used for the boot schema is very powerful and
allows for easy description of any random network topology.
The boot schema is used by a group of NaIL tools to handle
the following topology related functions: generation of a near-optimal
routing table based on minimum weight paths (see map(l)),
software configuration of the communication links (see solder(l)),
and initialization (booting) of Trollius (see spread(l)).
.PP
Common boot schemata for a particular multicomputer installation are
supplied by the system administrator.
They typically start with the prefix
.I bnail.
Each user is also free to customize an application specific machine.
The NaIL tools expect a user to specify in the boot schema minimum
information about nodes and links.
The map(l) tool will generate near-optimal routing paths between
non-neighbouring nodes starting with no further information.
At the other extreme NaIL permits a user to completely control the
routing tables by giving further information in the boot schema.
In general, the boot schema is first parsed by the NaIL tools
to trap syntactic errors.
Any logical errors are also identified and reported.
.SH NaIL GRAMMAR
.PP
The NaIL grammar for the boot schema is described here.
All information is specified in the form of tuples which are listed below.
.PP
.TP 12
TWEIGHT
= <[link_wt], [host_wt], [order_wt], [mcast_wt], [hop_wt]>;
.TP
TNODE
= <nodeid, node_type>;
.TP
TLINK
= <from_node, outlink, [oname], to_node, inlink, [iname], [this_link_wt]>;
.TP
TROUTE
= <any_node, dest_node, outlink, [outlink2]>;
.TP
TORDER
= <src_node, dest_node, [this_order_wt]>;
.TP
TSOLDER
= <node, [hostname], [sysnode]>;
.PP
A tuple consists of fields separated by commas and bounded by
angular parentheses.
A tuple-line consists of a tuple name, "=", one or more
tuples of the corresponding type, and a terminating semi-colon.
White space (blanks, tabs and newlines) is ignored.
Comments are bounded by "/*" and "*/".
Fields bounded by square parentheses are optional, but their commas
are mandatory.
Tuple-lines of a particular type may be used repeatedly.
In the case of duplication or conflict of information between various
tuples, the information in the first tuple to appear in the boot schema
is used.
Subsequent tuples do not over-write this information.
.SH TUPLE DESCRIPTION
.PP
.TP 12
TWEIGHT
= <[link_wt], [host_wt], [order_wt], [mcast_wt], [hop_wt]>;
.TP
Usage:
These weights are used by map(l) to guide the generation of the
routing tables for each node.
.I Link_wt
is used for assigning weights to direct links defined in TLINK tuples;
.I host_wt
is used for paths to and from the origin host node;
.I order_wt
is used for paths specified in TORDER tuples;
.I mcast_wt
is used for the remaining paths while creating a fully connected graph;
.I hop_wt
is used to control the relative importance between lengths (number of hops)
and weights (usage of links) while generating shortest paths.
Each weight can be set to any arbitrary integer, it is the relative
value that is important.
The weights have a default value of 100, except for hop_wt which defaults
to the value 0.
.TP
Ex:
<1, 1, 1, 1, 0>
.TP 12
TNODE
= <nodeid, node_type>;
.TP
Usage:
.I Nodeid
is the Trollius node identifier.
.I Node_type
is a bit-field describing the characteristics of the node in question.
For a detailed description see getnet(l).
The identifiers established in TNODE tuples are used to identify nodes in
Trollius user interface commands, and in C and FORTRAN programming under
Trollius.
Since the TNODE tuples define the nodeids used by other tuples to describe
the topology of the multicomputer, they all have to be specified before any
other tuple is used.
The TWEIGHT tuple is the only exception to this rule since it does not
rely on the definition of nodeids.
.TP
Ex:
<100, 0> <0x10000, 1>
.TP 12
TLINK
= <from_node, outlink, [oname], to_node, inlink, [iname], [this_link_wt]>;
.TP
Usage:
List all unidirectional physical links of the multicomputer.
The node
.I from_node
is attached to
.I to_node
by the link
.I outlink,
which is one of the links leaving from
.I from_node.
.I To_node
is attached to
.I from_node
by the link
.I inlink,
which is
one of the links arriving at
.I to_node.
.I This_link_wt
overrides the default
.I link_wt
for this link only.
In general, the optional link names
.I oname
and
.I iname
are only used on host nodes, whose links are identified by devices.
.TP
Ex:
<100, 0, "/dev/nap0", 0x10000, 1,, 20>
<0x10000, 1,, 100, 0, "/dev/nap0",>
.TP 12
TROUTE
= <any_node, dest_node, outlink, [outlink2]>;
.TP
Usage:
A message at
.I any_node
that is destined for
.I dest_node
must be sent on
.I outlink
of
.I any_node.
.I Outlink2
is the link used by the broadcast daemon at
.I any_node
when sending messages to
.I dest_node.
If
.I outlink2
is not specified,
.I outlink
is used instead.
This tuple effectively specifies one entry in one node's routing table.
All or part of a complete path between two nodes may be specified with
multiple TROUTE tuples.
.TP
Ex:
<100, 105, 0, 1> <101, 105, 2, >
.TP 12
TORDER
= <src_node, dest_node, [this_order_wt]>;
.TP
Usage:
Give priority to generating the optimal path from
.I src_node
to
.I dest_node.
Add
.I this_order_wt
to each intermediate link along the path to produce updated link weights.
.I This_order_wt
overrides the default
.I order_wt
for this route only.
.TP
Ex:
<101, 200, 5> <101, 201,,>
.TP
TSOLDER
= <node, [hostname], [sysnode]>;
.TP
Usage:
Place Trollius
.I node
on a physical node denoted by the vendor specific name
.I sysnode,
that is controled by the host node,
.I hostname.
.TP
Ex:
<0x10000,, "Acme 1.0"> <0x20003,, "Acme 1.3">
.SH NaIL TOOLS
The first step after creating a new valid boot schema is to
use map(l) to create a file of routing tables, one for each node.
Generally, this file is named
.I route,
suffixed by the boot schema file's suffix.
Solder(l) is then used to configure the nodes and links and, finally,
spread(l) boots Trollius on the complete multicomputer.
All Trollius commands and programs must then use the node identifiers
given in the TNODE tuples.
.SH FILES
.TP 38
$TROLLIUSHOME/boot/bnail.def
default boot schema file
.TP
$TROLLIUSHOME/boot/route.def
default route file
.SH "SEE ALSO"
The Trollius User's Reference Manual, Appendix A:  Writing a boot schema.
.br
getnet(l), map(1), rasm(l), solder(l), spread(l)
