.TH NCOMBINE l "April 26, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
ncombine \- Trollius user-defined message combining
.SH SYNOPSIS
.nf
int	ncombine (int castid, int event, void (*cmbfunc)())
.fi
.SH DESCRIPTION
.PP
The
.I ncombine()
function handles user-defined message combining during multireeling operations.
It is used when the programmer requires a combining function that is not
provided in the set of built-in functions (see rcast(l)).
.PP
.I Ncombine()
receives the packets of messages multireeled on the spanning tree defined by
.I castid
and combines them two by two using the user defined combining function
.I cmbfunc.
.I Event
is the positive number given in lieu of a combining function during the
cast registration using either rcast(l) or rcasttype(l).
It is used for communication between the cast daemon and the
.I ncombine()
function.
The task of
.I ncombine()
is coordinated by the cast daemon which provides the message packets as
they arrive, making sure packets are combined in their proper order within
the message.
The cast identifier
.I castid
has to be registered before
.I ncombine()
is called.
To guarantee that the cast registration is completed, the user may invoke
rrnodewait(l) in order to block till
.I castid
is defined.
.PP
The user defined combining function
.I cmbfunc
has the following calling sequence:
.PP
.nf
.ta 1i 2i 3.5i
void	cmbfunc (int length, char *msg1, char *msg2,
			int *data1, int *data2)
.fi
.PP
The function accepts two message packets to be combined, each packet having
a message buffer and a data pouch.
.I Msg1
and
.I msg2
are the two message buffer having an equal size of
.I length
bytes.
.I Data1
and
.I data2
are the two data pouches of size NHDSIZE as defined in <net.h>.
After combining, the resulting message buffer and data pouch have to
be put in
.I msg1
and
.I data1
since their contents are returned to the cast daemon.
Notice that the user cannot modify the length of the message buffers because
the packetization strategy is strictly controlled by Trollius.
.SH "EXAMPLE"
The following code waits till the cast registration of castid 400 (with
a combining event of 7) is completed then invokes
.I ncombine()
to combine the multireeled messages.
The user defined combining function adds all the integers in the packet
modulo 1000.
To keep the example code simple, no checking of error conditions is done.
.nf

#include <net.h>

static void	mycombine();

main()
{
	kinit(0);
	rrnodewait(getnodeid(), 400);
	ncombine(400, 7, mycombine);
	kexit(0);
}

static void
mycombine(length, msg1, msg2, data1, data2)

int		length, *data1, *data2;
char		*msg1, msg2;

{
	int	i, *p1, *p2;

	p1 = (int *) msg1;
	p2 = (int *) msg2;

	for (i = 0; i < length/sizeof(int); i++) {
		p1[i] = (p1[i] + p2[i]) % 1000;
	}
	for (i = 0; i < NHDSIZE; i++) {
		data1[i] = (data1[i] + data2[i]) % 1000;
	}
}
.fi
.SH "RETURN VALUE"
.PP
Upon successful completion, 0 is returned.
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH "SEE ALSO"
rcast(l), rrnodewait(l)
