.TH NSEND l "April 24, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
nsend, ntry_send, nrecv, ntry_recv  \- Trollius network layer message passing
.SH SYNOPSIS
.nf
#include <net.h>

int	nsend (struct nmsg *header);
int	ntry_send (struct nmsg *header);
int	nrecv (struct nmsg *header);
int	ntry_recv (struct nmsg *header);
.fi
.SH DESCRIPTION
.PP
There are four sublayers of message passing within the Trollius network
(inter-node) layer.
They are, in order of increasing functionality,
the physical, datalink, network, and transport sublayers.
The network sublayer functions are the most commonly used, most generic
communication functions in Trollius.
They provide routing services and invoke datalink layer functions.
.PP
.I Nsend()
and
.I nrecv()
send messages to, and receive messages from, any process on any node.
If no receiving, sending, or forwarding process is ready, the process calling
.I nsend()
or
.I nrecv()
will block (more on this below).
.I Ntry_send()
and
.I ntry_recv()
differ from
.I nsend()
and
.I nrecv(),
in that they will never cause the calling process to block.
The message is either immediately transferred, or an error is immediately
returned, indicating that the process would have blocked.
Both the sending process and the synchronizing receiving process
may use the non-blocking functions (unlike ksend(l)).
.SH "NETWORK MESSAGE DESCRIPTOR"
.PP
All of the functions accept a pointer to a message descriptor which is an
extension of the local layer message structure used by ksend(l) and krecv(l).
The network message structure is defined in <net.h> as:
.PP
.nf
.RS
.ta 1i 2i
struct nmsg {
	int	nh_dl_event,
		nh_node,
		nh_event,
		nh_type,
		nh_length,
		nh_flags,
		nh_data[NHDSIZE];
	char	*nh_msg;
};
.RE
.fi
.PP
The
.I nh_dl_event
field is not used by
.I nsend()
or
.I nrecv().
It is unchanged after a call to
.I nrecv(),
but is set to the event of the forwarding process after a call to
.I nsend()
(see dsend(l)).
.PP
The
.I nh_node
field is used by
.I nsend()
to identify the node running the intended receiver, local or remote.
The
.I nh_node
field is not used by
.I nrecv().
A receiver thus cannot directly specify its synchronizing sender.
Instead, a receiver is "matched" to a sender by one or both of the
.I nh_event
and
.I nh_type
fields.
Two special node identifiers are defined in <net.h>.
LOCAL refers to the local node and causes
.I nsend()
to bypass its routing step.
ORIGIN refers to the host node from which the Trollius multicomputer was booted.
The
.I nh_node
field remains unchanged after calling
.I nsend()
or
.I nrecv().
.PP
A message can be sent to, or received from, a group of nodes.
The nodes must first be registered as a multicast destination with rcast(l).
The cast identifier can then be used in the
.I nh_node
field.
See rcast(l) for more information.
.PP
.I Nh_event
is an arbitrary positive integer, used as a second level of
synchronization after
.I nh_node.
The values assigned by the sender and the receiver must be equal.
This field remains unchanged after calling
.I nsend()
or
.I nrecv().
.PP
The
.I nh_type
field further filters messages that match on the
.I nh_event
field.
A message will be passed only if the
.I nh_type
fields of the sender and the receiver have at least one bit set in
an identical position.
In other words, the bitwise logical AND of the type fields specified by
the two parties must not equal zero.
A zero value in the
.I nh_type
field is a special case \-
that process will match any other process using the same
.I nh_event.
The
.I nh_type
field remains unchanged after calling
.I nsend(),
but is set to the sender's
.I nh_type
field after calling
.I nrecv().
.PP
The
.I nh_length
field holds the length (in bytes) of the message to be sent.
If the sender and the receiver specify different lengths,
the lesser amount will be transferred.
The
.I nh_length
field remains unchanged after calling
.I nsend(),
but is set to the minimum of the sender's and receiver's lengths after calling
.I nrecv().
The datalink functions called by
.I nsend()
have a maximum message length of MAXNMSGLEN (defined in <net.h> and
currently set to 4096 bytes).
Messages exceeding this length will be implicitly broken down into
a series of smaller messages, which will be recombined at the receiving end.
Because of this packetization, one call to
.I nsend()
can introduce many messages into the network, a possible source of confusion
during debugging.
.PP
The
.I nh_flags
field is normally set to 0, and remains unchanged after calling
.I nsend()
or
.I nrecv().
When the NOBUF bit is set in
.I nh_flags,
buffers will not be used.
The KHOLD bit is set to create a virtual circuit (see "Virtual Circuits").
When the DFORWARD bit is set in the receiver's
.I nh_flags
field, all message structure fields will be changed to those of the
sending process except for
.I nh_msg.
DFORWARD has no effect when used with
.I nsend().
It is not typically used by user processes.
The NREEL bit of the sending process is set for multireeling (see rcast(l)).
During multireeling, the
.I nh_data[6]
field of the network message descriptor is not available to the program.
Flags used to assure that the data representation is correct for the receiving
node are discussed under "Data Representation".
NOBUF, KHOLD, DFORWARD, NREEL, and the data representation flags
are defined in <net.h>.
.PP
The
.I nh_data
field is a convenient data pouch within the network message descriptor.
Its array size is NHDSIZE, which is defined in <net.h> and is currently
set to 8 (words).
It can be used for sending short messages (in which case
.I nh_length
is set to 0)
or for appending control information to the message body.
After the message is passed, the information placed in
.I nh_data
by the sender will appear (possibly with conversion, see "Data Representation")
in the same field of the receiver.
The sender's
.I nh_data
field will not change (except in the cases of multireeling and tracing).
.PP
The
.I nh_msg
field holds the address of the first byte of data to be sent.
The data must be stored contiguously in memory.
This field remains unchanged after calling
.I nsend()
or
.I nrecv().
.SH "DATA REPRESENTATION"
.PP
On nodes of different types, data may have different representations.
For example, integers may be stored with the most significant byte
first in memory (big-endian) or with the most significant byte last in
memory (little-endian).
All fields in the network message structure, except
.I nh_msg,
are automatically converted if passed to a node with different data
representation.
In converting the
.I nh_data
field, it is assumed to hold all integers.
.PP
The
.I nh_flags
field of the sender's message structure can be set to the following data
representation flags (they will have no effect when used with
.I nrecv()).
Each flag assumes a data type, and will make the appropriate change
in the data representation of the given field.
They will have no effect if data conversion is not needed.
.TP 14
DINTDATA
.I nh_data
holds 8 32-bit integers.
.TP
DFLTDATA
.I nh_data
holds 8 single precision real numbers.
.TP
DDBLDATA
.I nh_data
holds 4 double precision real numbers.
.TP
DRAWDATA
.I nh_data
representation will not be changed.
.TP
DINTMSG
.I nh_msg
points to integers.
.TP
DFLTMSG
.I nh_msg
points to single precision real numbers.
.TP
DDBLMSG
.I nh_msg
points to double precision real numbers.
.TP
DRAWMSG
.I nh_msg
representation will not be changed.
.PP
If
.I nh_data
or
.I nh_msg
contains a mixture of data types, the user will have to change the
representation using the functions ltot(l), ttol(l), etc.
.SH "EXAMPLE USAGE"
.PP
The following is an example of a typical usage of the network message
structure for a message passed between two nodes with similar data
representations, utilizing a minimum level of message filtering.
This is intended only as a summary of a simple case.
Many variations can be constructed using the detailed information
given in the above section.
.PP
The sender sets
.I nh_node
to the identifier of the receiving node,
.I nh_event
to an arbitrary integer,
.I nh_type
and
.I nh_flags
to 0,
.I nh_msg
to a pointer to a data structure containing relevant information, and
.I nh_length
to the size of this data structure.
.I Nh_data
and
.I nh_dl_event
are left unassigned.
The receiver leaves
.I nh_node, nh_data,
and
.I nh_dl_event
unassigned.
It sets
.I nh_event
to the same integer specified by the sender,
.I nh_type
and
.I nh_flags
to 0,
.I nh_msg
to a pointer to a data structure, and
.I nh_length
to the size of this data structure.
After the message is passed, the receiver's data structure holds information
passed by the sender,  the receiver's
.I nh_data
field holds whatever (meaningless) information happened
to be at the address of the sender's
.I nh_data
field, and the receiver's
.I nh_length
field is set to the minimum of the sender's and receiver's
.I nh_length
fields (unchanged if the lengths were equal).
The only change to the sender's message structure is in the
.I nh_dl_event
field, which is set to the event of the forwarding process and can
be ignored by most programs.
.SH BLOCKING
.PP
A process calling
.I nrecv()
blocks until the message sent by the process calling
.I nsend()
entirely arrives.
A process calling
.I nsend()
blocks only until its message is picked up by a) a local process calling
.I nrecv(),
b) a buffer process on the same node, or c) a link process on the same node.
The only thing that is guaranteed by a successful return from
.I nsend()
is that the message has entirely left the calling process.
.I Nsend()
thus differs from ksend(l), which blocks unless a receiver is already waiting.
.PP
The loose blocking behavior of
.I nsend()
introduces a fundamental danger of Trollius message passing:  a sender
can transmit a message that may never be received due to programming
error or deadlock.
This message will never be dropped or timed out.
Some Trollius process will always be stuck with it, waiting for a synchronizing
.I nrecv()
that will never happen.
If that unfortunate process is a buffer, it can
be located by the user and swept clean (see sweep(l)).
However, if the process is a link proprietor, the link is henceforth
plugged and useless.
An intelligent buffering system does try to prevent this from
happening (see buffer(l)), but it remains theoretically possible.
.PP
It is also important to note that a transputer datalink input process
can be a buffer for one message.
.SH "VIRTUAL CIRCUITS"
.PP
The virtual circuit capability described in ksend(l) is also available
for network messages.
A virtual circuit eliminates the need for interaction with the kernel
and other server processes on subsequent
.I nsend()
and
.I nrecv()
calls, and thus increases the speed of message passing.
A network virtual circuit is a chain of local virtual circuits beginning
with the sender, following through any number of link proprietor processes,
and ending with the destination receiver.
For the life of the circuit, the intervening links cannot be used by
other processes, including system processes.
Thus it is recommended that virtual circuits be used only for nearest
neighbor message passing.
.PP
A virtual circuit is created by setting the KHOLD bit in the
.I nh_flags
field of both the sender and receiver's message structures.
Unlike the kernel layer, there is protection against length mismatch for
network layer virtual circuits.
The lesser amount specified in the sender or receiver's
.I nh_length
field will always be transmitted.
Resetting the KHOLD flag of both message structures to 0 sends or receives
the message along the virtual circuit, but terminates the circuit after
transmission.
Communicating parties must set and reset the KHOLD bit on the same message,
or both parties will block and the node may fail.
Use virtual circuits with care and only after debugging an application
without them.
Buffers are automatically disabled (NOBUF) when KHOLD is set.
.SH TRACING
.PP
The network sublayer functions are capable of generating traces that
record the time and nature of their activities.
Tracing is enabled for a process by the -t option of loadgo(l) and
toggled on or off with ktrstart(l) and ktrstop(l).
Tracing is provided at the cost of one word of storage in the data pouch.
If trace generation is enabled, the
.I nh_data[7]
field of the network message descriptor is not available to the program.
.SH "RETURN VALUE"
.PP
Upon successful completion, 0 is returned.
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 23
[EWOULDBLOCK]
.I ntry_send()
or
.I ntry_recv()
failed because the message could not be forwarded or received, respectively.
A call to
.I nsend()
or
.I nrecv()
would have blocked.
.TP
[ENOEVDESCRIPTORS]
The Trollius kernel ran out of message descriptor space.
.TP
[EBADNODE]
The specified destination node does not exist.
.SH "SEE ALSO"
dsend(l), ksend(l), psend(l), tsend(l)
