.TH PARSE_NODES l  "April 7, 1992" "Release 2.2" "TROLLIUS INTERNALS"
.SH NAME
.B parse_nodes
\- Parse command line mnemonic node identifiers.
.SH SYNOPSIS
.nf
.ft B
int	parse_nodes (int *argc, char **argv);
void	getnode (int *index, int *nodeid, int *flags);
char	*mnemonic (int node);
void	mnusage ();
.ft R
.fi
.SH DESCRIPTION
.I Parse_nodes()
is a package of functions that process mnemonic node ID's given on the
command line.
The format of the mnemonic ID's is system dependent, and a different
version of the file
.I parse_nodes.c
must be written for each system on which custom node ID mnemonics are desired.
The generic version is described here as an example.
.PP
In the generic case, nodes ID's appear on the command line as single
nodes (for example,
.I n0),
as comma-separated lists of ranges, or in lists separated by spaces.
Ranges of nodes have consecutive numerical ID's and are given as two numbers
separated by a dash.
For example, the mnemonic argument
.I "n0-3,12-15"
specifies nodes 0, 1, 2, 3, 12, 13, 14 and 15.
.PP
Command-line arguments are processed as follows.
Option letters that are prefixed with a dash in the command line,
and parameters that follow selected option letters,
are processed by the do_args package (see do_args(l)).
Then
.I parse_nodes()
is called with the remains of the
.I argv
structure.
What remains in
.I argv
at this point are all those command arguments that are neither
options nor node identifiers.
.PP
After calling
.I parse_nodes(),
node ID information is stored in an internal circular list.
All the nodes referred to on the command line will appear in the list, in
the order in which they were specified on the command line.
Node information is retrieved by calling
.I getnode().
Each call to
.I getnode()
advances one position in the circular list, and supplies values
for the three integer variables pointed to by the
.I getnode()
parameters.
These are the index of the position in the circular list (0 and up),
the absolute (non-mnemonic) node ID, and the
.I flags
variable.
Each bit in the flags variable represents a possible flag associated
with the mnemonic node ID.
If no mnemonic node ID's were given in the command line,
.I getnode()
returns -1 as the index, and NOTNODEID (0x80000000) as the node address.
.PP
To facilitate the response of the main program to errors and other
conditions, two additional utility functions are provided in the file
.I parse_nodes.c.
The function
.I mnemonic()
returns a pointer to a static character string that is the mnemonic
representation of the node address given as its argument.
The function
.I mnusage()
prints a short description of the valid mnemonics to standard output.
.SH DIAGNOSTICS
Upon successful completion,
.I parse_nodes()
returns 0.
Otherwise, \-1 is returned and the global variable
.I errno
is set to one of the following integer values, symbolically defined
in the include file
.I <terror.h>.
.TP 22
EBADNODESPEC
A mnemonic node specification is invalid.
.TP
EBADASCIINUMB
A character string that is supposed to represent a number
has a format that is invalid.
.SH "SEE ALSO"
do_args(l)
