.TH RBFSTATE l "April 10, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rbfstate
\- Report status of remote Trollius buffers.
.SH SYNOPSIS
.nf
#include <bufreq.h>

int rbfstate (int node, struct bfstate *table, int maxsize,
	bfparms *parms);
.fi
.SH DESCRIPTION
.PP
The remote function
.I rbfstate()
accesses a remote Trollius buffer daemon and reports information on each
buffered message on the given node, and on the buffer daemon itself.
.PP
The arguments to
.I rbfstate()
are:
.I node,
the target node identifier;
.I table,
a pointer to an array of structures where state information will be stored;
.I maxsize,
the number of elements in that array; and
.I parms,
a pointer to a structure which will hold information about the buffer daemon.
.PP
Returned information about buffered messages
is stored in an array of structures of the type
.I bfstate,
defined in <bufreq.h> as:
.PP
.nf
	struct bfstate {
		int	bfs_node;
		int	bfs_event;
		int	bfs_type;
		int	bfs_length;
	};
.fi
.PP
Upon returning from
.I rbfstate(),
each
.I bfstate
structure holds information about one buffered message.
The fields in the structure hold:
the destination node of the buffered message,
the destination event and type (see nsend(l)),
and the length in bytes of the message.
One of the returned
.I bfstate
structures will describe a message
bound for the event NOTNODEID (defined in <net.h>).
This is a dummy entry that serves to separate "blocked" buffered messages
from those which are "waiting".
Blocked messages are held by a buffer worker process, waiting messages
are held in the buffer daemon's internal buffer pool (see buffer(l)).
.PP
Returned information about the buffer daemon is stored in a
structure of the type
.I bfparms,
defined in <bufreq.h> as:
.PP
.nf
	struct bfparms {
		int	bfp_maxworkers;
		int	bfp_maxspace;
		int	bfp_nworkers;
		int	bfp_nspace;
		int	bfp_nfull;
		int	bfp_maxusr;
	};
.fi
.PP
The fields in this structure hold the maximum number
of buffer workers, the maximum size in bytes of the internal buffer pool,
the current number of workers, the current amount of message
body space (in bytes) in the internal buffer pool,
the current number of workers that are blocked with messages,
and the space size threshold for user messages.
.SH "RETURN VALUE"
.PP
Upon successful completion of
.I rbfstate(),
the number of buffered messages held by the remote buffer daemon is returned,
.I table
points to an array of filled
.I bfstate
structures, and
.I parms
points to information about the buffer daemon.
The return value will be larger than the number of
.I bfstate
structures, if
.I maxsize
was insufficient to hold state information for all buffered messages.
If an error occurred, \-1 is returned and the global variable
.I errno
is set to one of the error codes returned by nsend(l) or nrecv(l).
.SH BUGS
.PP
If the buffer daemon's workers are all occupied, the order of network
messages handled by the target node could be jumbled by the
service request initiated by \fIrbfstate()\fR.
.SH "SEE ALSO"
bfstate(l), buffer(l), rbuffer(l)
