.TH RBOOT l "April 7, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rboot \- Boot remote ITB nodes.
.SH SYNOPSIS
.nf
.ft B
#include <bootreq.h>
.PP
.ft B
int	rboot (int node, struct bootreq *request);
.fi
.SH DESCRIPTION
The function
.I rboot()
enables a Trollius process to boot a remote node, via a boot server on a
previously booted node adjacent to the desired node.
The parameters passed to
.I rboot()
are:
.I node,
the identifier of the staging node on which the boot server resides, and
.I request,
a pointer to a filled boot request structure.
.PP
The booting process consists of preparing the staging node and booting the
target node.
When the target node is booted, it is also prepared to serve as a staging
node in subsequent boots.
Thus, the staging node will already have been prepared if it was booted with
.I rboot()
without the DONT_CONT boot request flag (see below), or
with vboot(l) without the -s option.
In these cases, one call to
.I rboot()
will boot the target node.
If the staging node was booted with sboot(l), if DONT_CONT or -s were used,
or if you wish to reboot a previously booted node, a separate call to
.I rboot()
must first be made for each step needed to prepare the staging node.
These steps are:  flushing the table of modules to be loaded onto the target
node, starting to rebuild the table with a pointer to a bootstrap loader,
and adding pointers to module images to be loaded.
If a frozen boot is to be performed you will need to identify the frozen boot
file instead of adding the pointers to module images.
.SS THE BOOT REQUEST STRUCTURE
The boot request structure is defined in <bootreq.h> as follows:
.nf
.PP
.RS
.DT
.if n .ta .5i 1.0i 2.0i 2.5i
struct bootreq {
	int		btq_src_node;
	int		btq_src_event;
	int		btq_request;
	int		btq_subnode;
	BIMAGE	btq_image;
	int		btq_ilink;
	int		btq_frztag;
	int		btq_frzsize;
	int		btq_memstart;
	int		btq_flags;
	struct exec	*btq_ptchtout;
	char		*btq_ptchdata;
};
.RE
.PP
.fi
The first two fields,
.I btq_src_node
and
.I btq_src_event,
are filled in by
.I rboot().
They are the node and event the boot server replies to after completing
the request.
.PP
The
.I btq_request
field identifies the type of request.
The possible values
are:
.TP 24
BOOT_FLUSH
Flush the table of module images on the staging node.
.TP
BOOT_SETBOOT
Set the bootstrap pointer and size on the staging node.
.TP
BOOT_ADDIMAGE
Add an image to the table on the staging node.
.TP
BOOT_DOIT
Boot the target node.
.TP
BOOT_FROZEN
Boot the target node using a frozen boot file.
.TP
BOOT_SETFROZEN
Set the frozen boot image pointer and size on the staging node.
.TP
BOOT_PATCH
Relocate a load module.
.TP
BOOT_SETROUTE
Set the route file pointer and size on the staging node.
.PP
Which of the remaining fields need to be filled depends on the request.
.PP
The
.I btq_subnode
field holds the target node identifier (see the Boot Schema).
You will need to fill this field for BOOT_DOIT and BOOT_FROZEN requests
unless the recursive boot feature is used (see below).
.PP
The
.I btq_image
field, used with BOOT_ADDIMAGE requests, is a structure of type BIMAGE
(see below).
It supplies information on a module that will be loaded onto the
target node during booting.
You will need to make a separate request for each program that you wish to load.
Consult the process schemata (see conf(l)) for the programs that constitute
the Trollius operating system under the default configuration.
.PP
The
.I btq_ilink
field is the transputer link that communicates between the staging and target
nodes (see the Boot Schema).
You will need to fill this field for BOOT_DOIT and BOOT_FROZEN requests.
If this field is set to the special value ALLLINK, it inidicates that
all children of the staging node in the boot spanning tree are to be
booted and loaded with route tables.
A recursive boot request must be preceded by a BOOT_SETROUTE request.
.PP
The
.I btq_frztag
field holds a tag identifying the block of memory into which the frozen boot
file was loaded.
The
.I btq_frzsize
specifies the size of the frozen boot file.
These fields are used with the BOOT_SETFROZEN request.
These fields are also known as
.I btq_rtag
and
.I btq_rsize
and are used in similar roles for the BOOT_SETROUTE request.
.PP
The
.I btq_memstart
field is not currently used.
.PP
The
.I btq_flags
field specifies boot request flags used with BOOT_DOIT and BOOT_FROZEN requests.
The boot request flags are (for further information see vboot(l)):
.TP 17
DONT_FISH
Skip the reset step.
This step is irrelevant on systems that cannot reset ITB neighbors.
.TP
DONT_EXPORT
Don't export the bootstrap program to the target node.
.TP
DONT_BOOT
Export bootstrap program, but don't boot image modules or frozen image.
.TP
DONT_CONT
Skip staging steps to the target node.
.TP
RUNP_ONLY
Maintain bootstrap control after booting (leave moses(l) running) - useful
in debugging.
.PP
The
.I btq_ptchtout
and
.I btq_ptchdata
fields are used with BOOT_PATCH requests, and specify
the addresses where the text and data segments of the load module will reside.
.SS THE BOOT IMAGE STRUCTURE
The boot module image structure is defined in <bootreq.h> as follows:
.nf
.PP
.RS
.DT
struct bootimage {
	int	bimg_tag;
	char	*bimg_datapt;
	int	bimg_avtag;
	int	bimg_argc;
	int	bimg_stk;
	int	bimg_flags;
};
typedef struct bootimage BIMAGE;
.RE
.PP
.fi
The
.I bimg_tag
field is the tag identifying the memory block holding the module (given to
rload(l) when loading the module from a file).
The
.I bimg_datapt
field holds the actual address of the data/bss segment, and usually is not
filled before the call to
.I rboot().
The
.I bimg_avtag
field holds the tag associated with the argv structure.
The
.I bimg_argc
field tells how many strings are in the argv structure.
The
.I bimg_stk
field holds the desired size of the stack for that module upon execution.
The
.I bimg_flags
field is a bitmask.
The bits correspond to the following options:
.TP 16
SHR_TEXT
Share text if two of the same programs are loaded.
.TP
CALLIT
call/runp flag:
This flag must be set for the last process to load, and for no others.
This module will be called rather than set up as a separate process,
thus taking over from the bootstrap loader.
.TP
PRIORITY
Set priority to low for this module, otherwise the priority is high.
.TP
STRIPIT
Strip the module on download.
Don't load the symbol or relocation table.
The process cannot be spawned.
.SH "RETURN VALUE"
Upon successful completion, 0 is returned.
Otherwise \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 25
[EBADLINK]
The link specified in the
.I btq_ilink
field is invalid.
.TP
[EBADNODE]
The node identifier argument is invalid.
.TP
[EBADTAG]
A boot image tag (or its argv tag) is invalid.
.TP
[ENOBTDESCRIPTORS]
There are too many modules and the boot process is out of
module descriptor space.
.SH "SEE ALSO"
vboot(l)
