.TH RBTADD l "March 16, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rbtadd, rbtaddstage, rbtaddfrozen \- Load Trollius boot object on a remote
node.
.SH SYNOPSIS
.nf
#include <bootreq.h>
.PP
int rbtadd (int node, int object, int tag);
int rbtaddstage (int node, int tag, int flags, int stksize, int priority);
int rbtaddfrozen (int node, int tag, int size);
.fi
.SH DESCRIPTION
These functions register (or `add') boot objects loaded onto a remote node.
The
.I rbtadd()
function adds the route or bootstrap object, depending on the argument
.I object,
which is one of BTO_STRAP or BTO_ROUTE, as defined in <bootreq.h>.
The stage and frozen objects each have their own dedicated add function,
.I rbtaddstage()
and
.I rbtaddfrozen().
All the functions assume that the object has been previously loaded
onto the remote node with either rload(l) or rflat(l).
The
.I tag
returned by these loading functions serves as one parameter to
the add functions.
As with all functions of the remote layer, the first parameter,
.I node,
is the remote node identifier.
.PP
The stage and frozen objects require additional information to
be properly registered on the remote node.
Since a stage object may eventually become a process on a neighbour
of the remote node, process attributes like the stack size
and execution priority must be given.
The following boot flags may also be passed in the
.I flags
argument:
.TP 15
BTSHARE
Share the text segment with a previously booted program of the same name,
if one exists.
.TP
BTONCHIPSTK
Load the stack segment into the on-chip RAM of a transputer.
.PP
The frozen object only requires its total size to be passed to
.I rbtaddfrozen().
.SH "RETURN VALUE"
.PP
Upon successful completion, 0 is returned.
Otherwise \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 20
[ENOBTDESCRIPTORS]
The remote node is fresh out of stage object descriptors.
.TP
[EINVAL]
The
.I object
argument is invalid.
.PP
These functions may also return any of the error codes applicable
to nsend(l) or nrecv(l).
.SH "SEE ALSO"
rflat(l), rload(l), vboot(l)
