.TH RBTBOOT l "March 16, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rbtboot \- Boot Trollius on a remote ITB node.
.SH SYNOPSIS
.nf
#include <bootreq.h>
.PP
int rbtboot (int node, int object, int ilink);
.fi
.SH DESCRIPTION
.PP
The function
.I rbtboot()
boots a remote ITB node, the
.I target
node, that is a direct neighbour of another remote ITB node, the
.I staging
node.
The objects are booted by transmitting them across the staging
node's communication link to the target node.
Links are logically numbered in the range of 0 to N - 1, where
N is the total number of links on a node.
The staging node identifier is given in the first parameter to
.I rbtboot().
.SH OBJECTS
.PP
Several objects may be booted as a result of one call to
.I rbtboot()
but all objects specified must already reside on the staging node.
See rbtadd(l) for a description on how to prepare for booting.
The
.I object
parameter is a bit field with the following possible object
masks, defined in <bootreq.h>:
.TP 12
BTO_STRAP
The strap object is the moses(l) bootstrap program.
The strap boot procedure starts with reseting the target node,
then booting the bootstrap and finally synchronizing the
target node's timers (a feature of moses(l)) with those of
the staging node.
This object must be booted before any other object.
.TP
BTO_STAGE
The stage object is the process schema, a table of programs
that will constitute the operating system on the target node.
For booting purposes, it is mutually exclusive with BTO_FROZEN
and BTO_SPAWN.
.TP
BTO_FROZEN
The frozen object is the frozen process schema.
.TP
BTO_SPAWN
The spawn object is the table of programs that are running
on the staging node.
These programs must all be re-entrant if the spawn object is
chosen for booting.
.TP
BTO_ROUTE
The route object is the combination of route tables for every
node in the multicomputer.
Booting this object means extracting the route table for the
target node and sending it to the target node.
The target node must be already be running Trollius as a result
of previously booting the stage, frozen or spawn object.
.SH "RETURN VALUE"
.PP
Upon successful completion, 0 is returned.
Otherwise \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 20
[ENOBOOTSTRAP]
The strap object has been selected for booting but it has not be loaded
on the staging node.
.TP
[EHIGHWATER]
The combined booted system of programs is exceeds the internal memory limit
placed on it.
.TP
[EINVAL]
The
.I object
argument is invalid or the link index is out of range.
.PP
This function may also return any of the error codes applicable
to nsend(l) or nrecv(l).
.SH "SEE ALSO"
vboot(l)
