.TH RBTSTATE l "April 7, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rbtstate
\- Report status of boot objects on a remote ITB node.
.SH SYNOPSIS
.nf
#include <bootreq.h>
.PP
int rbtstate (int node, struct bootstate *table, int maxsize, int object);
.fi
.SH DESCRIPTION
.PP
The function
.I rbtstate()
fills a table of program descriptors with information from selected
boot objects on remote nodes.
Some boot objects are made of up one or more programs which may
be booted on neighbouring ITB nodes.
The last argument is the object selector and the object masks are
defined in <bootreq.h>.
One object may be selected from the following list:
.TP 12
BTO_STAGE
The stage object is the process schema.
Each program in the process schema is described by one entry in the
descriptor table.
If
.I rbtstate()
returns 0, then the stage object has not been loaded on the remote
node (see rbtadd(l)).
.TP
BTO_SPAWN
The spawn object is the table of running programs on the remote node.
Any ITB node that is running should have a non-zero spawn table.
.PP
The first argument to
.I rbtstate()
is the remote node identifier.
The
.I table
argument points to the descriptor table to be filled with the selected object.
The
.I maxsize
argument defines the number of entries in the descriptor table.
.I Rbtstate()
will not overflow the table.
.PP
The descriptor table entry is defined by the following structure,
found in <bootreq.h>:
.PP
.ta .5i 1.0i 2.0i 4.0i
.nf
	struct bootstate {
		void	*bts_load;	/* load module/text addr */
		void	*bts_data;	/* data segment address */
		int4	bts_stksize;	/* stack size */
		int4	bts_flags;	/* boot flags */
		int4	bts_priority;	/* process priority */
		char	bts_name[BTMAXNAME];
	};
.fi
.PP
The last field of the structure holds the name of the program.
The boot flags are listed below:
.TP 15
BTSHARE
Share the text segment with a previously booted program of the same name,
if one exists.
.TP
BTONCHIPSTK
Load the stack segment into the on-chip RAM of a transputer.
.SH "RETURN VALUE"
.PP
Upon successful completion of
.I rbtstate(),
the number of program descriptors in the object is returned.
If this number is greater than the
.I maxsize
argument, the remote node object was bigger than the table provided.
In all cases of error, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 12
[EINVAL]
The
.I object
argument is invalid.
.PP
This function may also return any of the error codes applicable
to nsend(l) or nrecv(l).
.SH "SEE ALSO"
vboot(l)
