.TH RCAST l "April 28, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rcast, rcasttype, rmelt, rcask \- Register or remove a multicast destination;
report number of multireel messages.
.SH SYNOPSIS
.nf
.ft B
#include <castreq.h>
#include <net.h>

int	rcast (int node, int castid, int reelfunc, int *pcast, int ncast);
int	rcasttype (int node, int castid, int reelfunc, int type, int mask);
int	rmelt (int node, int castid);
int	rcask (int node, int castid);
.fi
.SH DESCRIPTION
Multicasting and multireeling allow messages to be sent to, or received
from, a group of nodes, with one call to nsend(l)/nrecv(l).
This is achieved by defining the group of nodes under a multicast identifier,
which can then be used in the destination node field of the message
structure (see nsend(l)).
Only one process per node can participate in the multicast or multireel on
a specific message event for a given multicast identifier.
The root node, i.e. the node that defines the multicast identifier, may be 
included in the multicast identifier.
.PP
Multicast identifiers can also be used as the destination node argument to
many remote functions, allowing the action to be taken on a group of nodes.
.PP
.I Rcast()
or
.I rcasttype()
must be used before multicasting or multireeling to register the
multicast identifier with all the cast daemons needed to participate in
a multicast or multireel.
This registration is specific to the root node.
Therefore, each Trollius node wishing to transmit multicast
messages must make a separate call to
.I rcast()/rcasttype().
.I Rmelt()
is the inverse of 
.I rcast()/rcasttype().
It removes the information and frees up space in the cast daemon's data
structures.
Since the cast daemon has a static amount of space for cast registration,
multicast identifiers should be removed when no longer needed. 
.I Rcask()
reports the number of messages that are expected to arrive at a
given cast member on the way to the root node.
The usefulness of
.I rcask()
in multireeling is diminished with the introduction of ncombine(l)
to handle user-defined combining of multi-packet multireeled messages.
.PP
The arguments to
.I rcast()
are:  the root
.I node;
the multicast identifier
.I castid,
which can be any integer that is not a node identifier;
.I reelfunc, 
which identifies a combining function for multireeling;
.I pcast,
a pointer to an array holding the identifiers of the nodes
to be included in the multicast; and 
.I ncast,
the number of nodes in
.I pcast.
Except in rare circumstances,
.I node
should be set to LOCAL for a multicast from a compute node, or to
ORIGIN for a multicast from the host.
LOCAL and ORIGIN are defined in <net.h>.
.I Reelfunc 
does not need to be set for multicasting.
The first three arguments of
.I rcasttype()
are identical to those of
.I rcast().
The last two arguments offer a different way of specifying the identifiers
of the nodes to included in the multicast.
Instead of a full enumeration, the user gives the characteristics of
these nodes by using the two bit-fields
.I type
and
.I mask
to specify the node types required.
For a full explanation on how the
.I type
and
.I mask
bit-fields are used, see getntype(l).
.I Rcasttype()
uses these arguments to get the list of node identifiers that fit the
required characteristics and then invokes
.I rcast()
to do the multicast registration.
.I Rmelt()
and
.I rcask()
require only the root 
.I node
and the 
.I castid.
.PP
.I Rcast()
operates as follows.
A registration message is sent to the cast daemon on the root node.
The cast daemon consults the route daemon to learn which outgoing
multicomputer links are used to send messages to each node specified in
the multicast.
The cast members are then grouped according to outgoing link.
This information is stored in the cast daemon's data structures.
The registration message is then propagated across the links for which there
is a non-zero list of cast members, to neighboring nodes.
The cast daemons on these nodes proceed through the steps performed by the
original daemon, subdividing their sub-group of cast members, storing 
information, and propagating the registration message.
In this way, 
.I rcast()
proceeds through several recursive cast daemon requests.
.SH MULTICASTING
To multicast a message, a process starts by registering the multicast
destination with
.I rcast().
Then the network message structure is filled,
with the destination node field set to the multicast identifier
.I castid,
instead of the usual single destination node.
The process then calls nsend(l).
The nodes included in the multicast receive the message with nrecv(l).
The messages are matched on event and type, as usual.
.PP
Multicast messages are naturally more efficient than multiple messages
due to their multiple forwarding characteristic.
However, multicast messages incur an added overhead because they
must pass through an additional process, the cast daemon, on each node
of the journey.
.SH MULTIREELING
In a multireel, the member nodes of a cast reverse the operation of a 
multicast, i.e., they all send a message to the root node.
For a multireel, the root node registers the multicast destination and
calls nrecv(l).
Each of the member nodes sets the 
.I nh_flags
field of its network message structure to NREEL (defined in <net.h>)
and its destination node field to
.I castid. 
The member nodes then call nsend(l).
There is no restriction on the length of the multireeled messages provided
all messages have an equal number of packets.
Message packetization is explained in nsend(l).
During a multireeling operation, the
.I nh_data[6]
entry of the message header is used by the cast daemons and any user data
residing in it is lost.
.PP
Note that although the root node registers 
.I castid, 
the multicast nodes use
.I castid.  
This makes it imperative that the call to 
.I rcast()
be completed before the multicast nodes call nsend(l).
The recommended procedure is for the root node to register the 
.I castid,
while each of the multicast nodes blocks waiting for the
.I castid
to be registered with its local route daemon.
This can be done using the rrnodewait(l) function.
This way the multicast nodes cannot initiate the multireel until the
multicast is completed, which occurs after the cast registration.
.PP
During a multireel, the cast daemons operate as follows.
When the daemon receives two messages from different nodes of the
multireel, and each time it receives a subsequent message, it sends
the messages to a local combining process, which returns a single message.
When all information has been received and combined, the processed
message is forwarded to the next node.
Eventually, a single message is forwarded to a process at the root node.
.PP
The function used to combine message information is specified in the
.I reelfunc
argument to
.I rcast().
Several built-in combining functions that operate on integers (or, more
specifically, on word-size items interpreted as integers) are defined 
in <castreq.h>, and are listed below.
These functions combine both the network message structure's 
.I nh_data
and 
.I nh_msg
fields (see nsend(l)).
All functions combine the entire fields, except for CBN_ERR, which operates
on the first word only.
For these functions to work properly the user must set the DINTDATA and
DINTMSG flags in the messages sent in order to specify that the messages
contain integers.
.PP
.TP 14
CBN_ERR
Combine error codes in first integer only.
Retain non-zero value.
If both values are non-zero, either one may be retained.
.TP 
CBN_ADD
Combine by adding all corresponding integers.
.TP
CBN_MIN
Combine by taking the minimum.
.TP
CBN_MAX
Combine by taking the maximum.
.TP
CBN_AND
Combine by applying a bitwise AND.
.TP
CBN_OR
Combine by applying a bitwise OR.
.TP
CBN_XOR
Combine by applying a bitwise XOR.
.PP
If a multireel is to be used purely for synchronization, use CBN_ERR, and set
.I nh_length
to 0.
.PP
The user can choose to supply their own combining process.
When registering the cast identifier with 
.I rcast()
or
.I rcasttype(),
the 
.I reelfunc
argument should be set to a positive event used to identify messages
passed to and from the combining process.
The combining process itself should use rrnodewait(l) and ncombine(l)
to fulfill its task (see ncombine(l) for a sample code).
In order to insure that the combining process runs on every node on the
multireel route that needs to do combining, the combining process should
run on every ITB node.
The combining process can either be an independent process or part of
the user program.
.SH "RETURN VALUE"
.I Rcast(), rcasttype()
and
.I rmelt()
return 0 if the message is successfully registered or de-registered
at the first (local) node.
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.I Rcask()
returns the expected number of multireel messages.
As well as the errors given below,
.I rcast()
and
.I rcasttype()
can incur any of the errors associated with nsend(l) or nrecv(l).
.PP
.SH ERRORS
.TP 15
[ECOLLIDE]
multicast registration conflicts with a defined multicast identifier
.TP
[EINPROGRESS]
multicast registration conflicts with a multicast identifier still being
registered
.TP
[EFULL]
out of cast descriptors
.TP
[EBADNODE]
non-existent destination node
.SH "BUGS"
Simultaneous reels on the same group of nodes using the same event, will
not work.
To be safe, synchronize multireeling with the source node before sending
another multireeled message on the same castid/event pair.
This can be done by having the root node send a multicast acknowledgment
message.
Another way is to guarantee by careful timing that a multireel finishes
before a second one is started although this technique will not be scalable.
.SH "SEE ALSO"
nsend(l), ncombine(l), rrnodewait(l), getntype(l)
