.TH RCREATE l "April 7, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rcreate \- Create a Trollius process on remote nodes.
.SH SYNOPSIS
.ft B
#include <kio.h>
.PP
.ft B
int	rcreate (int node, int loadpt_tag, int flags, int wpsize, int *pid,
int argc, int argv_tag);
.SH DESCRIPTION
The remote function
.I rcreate()
enables a Trollius process to create another process on a specified node
or group of nodes.
The load module and the
.I argv
structure must already be
in memory on the target node (see rload(l) and rflat(l)).
.PP
The parameters passed to
.I rcreate()
are:
.I node,
the target node identifier;
.I loadpt_tag,
the integer through which the loaded module is referenced;
.I flags,
the runtime flags;
.I wpsize,
the size in bytes of the stack to be allocated;
.I pid,
a pointer to an integer where a new process ID will be returned;
.I argc,
the number of command-line arguments; and
.I argv_tag,
the integer through which the
.I argv
structure is referenced.
.PP
.I Loadpt_tag
and
.I argv_tag
are the same parameters that were passed to rload(l) and rflat(l), respectively.
.I Flags
.B must
be set to RTF_NOWAIT, to specify that
.I rcreate()
should return immediately, rather than upon exit of the new process.
.I Flags
may additionally be set to RTF_ONCHIP in order to allocate the stack from fast
on-chip RAM.
If this flag is used,
.I wpsize
should not exceed 3 Kbytes.
RTF_ONCHIP and RTF_NOWAIT are defined in <kio.h>.
.PP
The
.I argv
structure starts with
.I (argc
+ 1)
.I uninitialized
pointers to strings followed by the null-terminated strings.
The strings must be contiguous in memory.
The pointers are initialized by the remote Trollius server
that creates the process.
.PP
.I Loadpt_tag
and
.I argv_tag
are erased from the flat daemon tables during the call to
.I rcreate().
They cannot be used again, unless rload(l) or rflat(l)
are called again with the same tag.
.PP
An
.I rcreate()
request can be sent to a group of nodes.
The nodes must first be registered as a multicast destination with rcast(l).
The cast identifier can then be used as the
.I node
argument to
.I rcreate().
See rcast(l).
.SH "RETURN VALUE"
Upon successful completion of an
.I rcreate()
request to a single destination node, the index (in the remote process table)
of the created process is returned, and the
.I pid
argument points to the new process ID.
If an error occurred,
\-1 is returned and the global variable
.I errno
is set to indicate the error.
An
.I rcreate()
request sent to a multicast destination returns the same values.
There is no way, however, for the user to know which process index
or process ID was returned, or for which process an error was generated.
.SH ERRORS
.TP 22
[EBADNODE]
The node address argument is invalid.
.TP
[ENOMEM]
There isn't enough memory to allocate the stack space.
.TP
[EBADTAG]
The load point tag or the
.I argv
tag is invalid.
.TP
[ENOPDESCRIPTORS]
The
.I kenya
process descriptor tables are full.
.SH "SEE ALSO"
rcast(l), rflat(l), rload(l), rloadgo(l), rspawn(l)
