.TH RCSTATE l "April 29, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rcstate, rcstate2
\- Report status of multicast daemon on a remote node.
.SH SYNOPSIS
.nf
.ft B
#include <castreq.h>

int	rcstate (int node, struct cstate *table, int maxsize);
int	rcstate2 (int node, int castid, struct cstate2 *table, int maxsize);
.fi
.SH DESCRIPTION
The remote functions
.I rcstate()
and
.I rcstate2()
access a Trollius multicast daemon (castd) on a remote node,
and report information on each cast identifier being used and on any
message event being used in a multireeling operation on that node.
.I Rcstate()
is used to get information about all cast identifiers used on that node
whereas
.I rcstate2()
is used to get details about the message events for a particular identifier.
.PP
The arguments to
.I rcstate()
are:
.I node,
the target node identifier,
.I table,
a pointer to an array of structures where multicast information will
be stored, and
.I maxsize,
the number of elements in that array.
The arguments to
.I rcstate2()
are:
.I node,
the target node identifier,
.I castid,
the specific cast identifier desired,
.I table,
a pointer to an array of structures where message event information will
be stored, and
.I maxsize,
the number of elements in that array.
.PP
The information returned by
.I rcstate()
is stored in an array of structures of the type
.I cstate,
defined in <castreq.h>:
.PP
.ta .5i 1.0i 2.5i 3.0i
.nf
	struct cstate {
		int		cs_castid;
		int		cs_root_node;
		int		cs_parent_node;
		int		cs_combine;
		int		cs_nbranch;
		int		cs_nreels;
	};
.fi
.PP
The information returned by
.I rcstate2()
is stored in an array of structures of the type
.I cstate2,
defined in <castreq.h>:
.PP
.ta .5i 1.0i 2.5i 3.0i
.nf
	struct cstate2 {
		int		cs2_castid;
		int		cs2_event;
		int		cs2_type;
		int		cs2_npkt;
	};
.fi
.PP
Upon returning from
.I rcstate(),
each
.I cstate
structure holds information about one cast identifier.
The fields in the structure hold:
the cast identifier, the root node of the multicast spanning tree,
the parent node in the tree, the combining function, the number of
leaf nodes, and the number of multireels currently taking place on different
message events at that node.
A positive combining function indicates a user defined combining process
whereas a negative function indicates a system builtin function.
Builtin functions are defined in <castreq.h>.
The sign of the
.I cs_nbranch
field indicates whether the registration of that cast identifier has been
completed: a positive number signals a completed registration, a negative
number signals a registration still in progress.
Three special cast identifiers, HOST2ITB, HOST2OTB and HOST2ALL are defined
in <net.h> and indicate multicast spanning trees rooted at the host node
and branching to all ITB nodes, all OTB nodes, and all nodes in the system
respectively.
These three identifiers might get defined by some Trollius commands.
.PP
Upon returning from
.I rcstate2(),
each
.I cstate2
structure holds information about one multireeled message event.
The fields in the structure hold:
the cast identifier, the message event used, the message type, and
the number of packets currently stored and/or combined.
In Trollius, long messages are packetized.
Packetization is explained in nsend(l).
.SH "RETURN VALUE"
.PP
Upon successful completion of
.I rcstate(),
the number of cast identifier entries on the remote node is returned, and
.I table
points to an array of filled
.I cstate
structures.
Upon successful completion of
.I rcstate2(),
the number of message event entries for the given cast identifier is
returned, and
.I table
points to an array of filled
.I cstate2
structures.
In both cases, the return values will be larger than the number of filled
.I cstate/cstate2
structures if
.I maxsize
was less than the number of entries on the remote node.
In all cases of error, \-1 is returned and the global variable
.I errno
is set to one of the error code.
.SH "SEE ALSO"
cstate(l)
