.TH RDOOM l "April 7, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rdoom \- Doom processes in remote ITB (compute) nodes.
.SH SYNOPSIS
.nf
.ft B
#include <preq.h>
#include <ksignal.h>
.PP
.ft B
int	rdoom(int node, int sflags, int svalue, int signal);
.fi
.SH DESCRIPTION
The remote function
.I rdoom()
enables a Trollius process to send a signal to selected processes on a
specified ITB (compute) node or nodes.
The signal will be delivered the next time the target process attempts any
kernel request, or immediately if the target process is blocked on an event.
.I Rdoom()
cannot be used to send a signal to a host process.
.PP
The parameters passed to
.I rdoom()
are:
.I node,
the node to receive the signal;
.I sflags,
the process selection flags;
.I svalue,
the process selection value; and
.I signal,
the signal number.
The process selection flags
.I sflags
are used to determine on what basis processes are selected to receive
the signal.
The process selection flags, defined in <preq.h>, are:
.TP 18
SELECT_PID
Select by process ID.
.TP
SELECT_INDEX
Select by process table index.
.TP
SELECT_BLKD
Select by the event the process is blocked on.
.TP
SELECT_LOADPT
Select by the load point of the process.
.TP
SELECT_APPL
Select all application (user) processes.
.TP
SELECT_SYS
Select all system processes.
.TP
SELECT_ALL
Select all application and all system processes.
.PP
The process selection value
.I svalue
is set to the process ID, the process table index, the event, or the
load point, whichever corresponds with the process selection flag.
If the process selection flag is SELECT_APPL, SELECT _SYS, or SELECT_ALL,
the process selection value does not need to be set.
.PP
The signal number
.I signal
can be set to the following codes, defined in <ksignal.h>:
.TP 15
SIGROUTE
Invalidate route cache.
.TP
SIGBACK
This is used to implement controlled processes.
.TP
SIGNOOP
This is used to implement controlled processes.
.TP
SIGUDIE
Terminate.
.TP
SIGARREST
Suspend execution.
.TP
SIGRELEASE
Continue execution.
.TP
SIGA
This is user defined.
.TP
SIGB
This is user defined.
.PP
An
.I rdoom()
request can be sent to a group of nodes.
The nodes must first be registered as a multicast destination with rcast(l).
The cast identifier can then be used as the
.I node
argument to
.I rdoom().
.SH "RETURN VALUE"
Upon successful completion, 0 is returned.
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 20
[EBADNODE]
The node address argument is invalid.
.TP
[ENOTPROCESS]
No such process exists.
.SH "SEE ALSO"
doom(l), kdoom(l)
