.TH RLOADGO l "April 8, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rloadgo \- Load and execute a module in remote ITB nodes.
.SH SYNOPSIS
.nf
.ft B
#include <kio.h>

.if n int	rloadgo (int src_node, int dest_node, int wpsize, int flags,
.if n	char *filename, int argc, char **argv, int argvsize, int *pid);
.if t int	rloadgo (int src_node, int dest_node, int wpsize, int flags, char *filename, int argc,
.if t	char **argv, int argvsize, int *pid);
.fi
.SH DESCRIPTION
The remote function
.I rloadgo()
loads an executable program onto an ITB (compute) node or group of ITB nodes,
and creates a new running process or processes with the given runtime arguments.
.PP
The arguments to
.I rloadgo()
are:
.I src_node,
the node on which the executable file resides;
.I dest_node,
the node which the program is to be loaded onto;
.I wpsize,
the size in bytes of the workspace (stack) to be allocated;
.I flags,
the runtime flags;
.I filename,
the name of the file to load;
.I argc,
the number of command-line arguments;
.I argv,
a pointer to the command-line argument structure;
.I argvsize,
the size of the
.I argv
structure; and
.I pid,
a pointer to an integer variable where the ID of the new process will be
returned.
.PP
.I Flags
.B must
be set to RTF_NOWAIT.
RTF_NOWAIT specifies that
.I rloadgo()
should return immediately, rather than upon the exit of the new processes.
.I Flags
may additionally be set to RTF_ONCHIP in order to allocate the stack
from the fast on-chip RAM.
If this flag is used,
.I wpsize
should not exceed 3 Kbytes.
RTF_ONCHIP and RTF_NOWAIT are defined in <kio.h>.
.PP
The
.I argv
structure starts with
.I (argc
+ 1) uninitialized
pointers to strings, followed by the null-terminated strings.
The pointers are initialized by the remote Trollius server that creates
the process.
The strings themselves should all be contiguous in memory.
.PP
An
.I rloadgo()
request can be sent to a group of nodes.
The nodes must first be registered as a multicast destination with rcast(l).
The cast identifier can then be used as the
.I dest_node
argument to
.I rloadgo().
.SH "RETURN VALUE"
Upon successful completion of an
.I rloadgo()
request to a single destination node, the index (in the remote process table)
of the created process is returned, and the
.I pid
argument points to the process ID.
If an error occurred, \-1 is returned and the global variable
.I errno
is set to indicate the error.
An
.I rloadgo()
request sent to a multicast destination returns the same values.
There is no way, however, for the user to know which process index or process
ID was returned, or for which process an error was generated.
.SH ERRORS
.TP 24
[EBADNODE]
A node address argument is invalid.
.TP
[ENOMEM]
There isn't enough memory to hold the load module and/or allocate the
stack space.
.TP
[ENOPDESCRIPTORS]
The
.I kenya
process descriptor tables are full.
.TP
[ENOFLDESCRIPTORS]
The table of memory blocks (kept by the
.I flat
process) is full.
.SH "SEE ALSO"
rload(l), loadgo(l)
