.TH RMALLOC l "April 8, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rmalloc, rfree, rfreeall
\- low level Trollius memory (de)allocation functions
.SH SYNOPSIS
.nf
.ft B
#include <mareq.h>
.PP
.ft B
char	*rmalloc (int node, unsigned int amount, int region, int align);
int	rfree (int node, char *address);
int	rfreeall (int node, int pid);
.fi
.SH DESCRIPTION
These functions access Trollius low level memory allocation services
on ITB (compute) nodes.
The high level memory allocation functions malloc(l), free(l), calloc(l),
and realloc(l) make use of
.I rmalloc().
.PP
.I Rmalloc()
attempts to allocate, to the given
.I node,
the specified
.I amount
(bytes) inside the memory region specified by
.I region,
aligned to an address that is a multiple of
.I align
bytes.
Word alignment is always guaranteed.
If you do not care about alignment, set
.I align
to 0.
The starting address of the allocated block is returned,
or NULL if such a block is not available.
.PP
The possible regions are system dependent and are defined in <maconfig.h>.
There is always a memory region labelled MAIN.
All types of nodes allow the caller to specify the region ANYBANK (0),
meaning the caller does not care which region is used.
If ANYBANK is given, the first NMAIN regions are searched in
ascending order for a suitable block of free memory.
See <maconfig.h> for the order of regions and the search limit imposed by NMAIN.
If a specific region is given, only that region is searched.
The allocated memory block is guaranteed to fall entirely within one region.
.SH TRANSPUTERS
.PP
Transputer nodes always have the special region ONCHIP.
If ONCHIP is specified, up to 1 Kbyte (T414) or 3 Kbytes (T800)
of the fast on-chip RAM can be allocated to the caller.
.SH "MALLOC & FREE"
.PP
If you wish to allocate, free and reallocate many small blocks of memory from
ANYBANK on the local node, and have no special address alignment requirements,
use malloc(l) and free(l), rather than
.I rmalloc()
and
.I rfree().
See malloc(l) for more information on these faster memory allocation functions.
.SH DEALLOCATION
.PP
Memory allocated by
.I rmalloc()
must be deallocated by
.I rfree()
or
.I rfreeall().
.I Rfree()
requests that the block of memory formerly allocated via
.I rmalloc()
on
.I node,
starting at
.I address,
be returned to the system.
Locally allocated blocks are automatically released upon termination of the
process.
Remotely allocated blocks, however, must be explicitly returned by
.I rfree().
Otherwise, they are lost until the node is rebooted.
.PP
The function
.I rfreeall()
returns to the system all the memory blocks allocated via
.I rmalloc()
to the process with the given process ID
.I pid
on the specified
.I node.
Use with extreme caution.
.PP
One must be careful to pass pointers returned by
.I rmalloc()
only to
.I rfree(),
and not to free(l) or realloc(l).
Similarly, pointers returned by malloc(l) can be passed to
free(l) or realloc(l), but not to
.I rfree().
.SH "RETURN VALUE"
.I "Rmalloc()"
returns the address of the allocated block.
If the request cannot be satisfied, NULL (0) is returned.
Upon successful completion of
.I rfree()
or
.I rfreeall(),
0 is returned.
Otherwise \-1 is returned.
In all cases of error, the global variable
.I errno
is set to one of the following error codes or one of the error codes
returned by nsend(l) and nrecv(l).
.SH ERRORS
.TP 20
[ENOMEM]
An allocation request cannot be satisfied because the amount requested
is too big, too small (<= 0 bytes), or memory is exhausted or fragmented.
.TP
[ENOTMALLOCED]
The memory block was rejected by
.I rfree()
since it was never obtained by
.I rmalloc(),
or because of an intervening call to
.I rfreeall().
.TP
[EINVAL]
The
.I region
argument is out of range.
.SH "SEE ALSO"
malloc(l), rflat(l), mstate(l)
