.TH RMSTATE l "April 8, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rmstate
\- Report status of memory allocation on a remote node.
.SH SYNOPSIS
.nf
.ft B
#include <mareq.h>

int	rmstate (int node, struct memblock *table, int maxsize);
.fi
.SH DESCRIPTION
The remote function
.I rmstate()
accesses a Trollius memory management daemon (mallocd) on a remote node,
and reports information on each allocated and unallocated memory block
on that node.
.PP
The arguments to
.I rmstate()
are:
.I node,
the target node identifier,
.I table,
a pointer to an array of structures where memory allocation information will
be stored, and
.I maxsize,
the number of elements in that array.
.PP
The information returned by
.I rmstate()
is stored in an array of structures of the type
.I memblock,
defined in <mareq.h>:
.PP
.ta .5i 1.0i 2.5i 3.0i
.nf
	struct memblock {
		int		ma_node;
		int		ma_pid;
		char		*ma_start;
		int		ma_size;
		struct memblock	*ma_next;
	};
.fi
.PP
Upon returning from
.I rmstate(),
each
.I memblock
structure holds information about one memory block.
The fields in the structure hold:
the node ID and process ID of the client process for which the memory
block was allocated, the starting address and size in bytes of the memory block,
and a link field which is unused in this context.
If the node ID is given as NOTPROCESS (defined in <kernel.h>),
the memory block is unallocated.
.SH "MEMORY OVERHEAD"
.PP
The memory management daemon also uses the
.I memblock
structure as descriptors for memory blocks.
These descriptors are not embedded into each separate memory block.
Instead, they are allocated in contiguous sets, a distinct memory
block themselves, from any of the regions covered by ANYBANK (see
rmalloc(l)).
There are MAMAX (defined in <mareq.h>) descriptors in each descriptor set.
The daemon starts with one set (within system space, not in any of
its regions) but may allocate and deallocate more sets as necessary
if the user processes make heavy use of rmalloc(l) and rfree(l).
Information on these descriptor sets are returned as with any
other memory block by
.I rmstate().
They can be identified with the daemon's process ID.
.SH "RETURN VALUE"
.PP
Upon successful completion of
.I rmstate(),
the number of memory block descriptors on the remote node is returned, and
.I table
points to an array of filled
.I memblock
structures.
The return value will be larger than the number of filled
.I memblock
structures if
.I maxsize
was less than the number of memory blocks on the remote node.
In all cases of error, \-1 is returned and the global variable
.I errno
is set to one of the error codes returned by nsend(l) or nrecv(l).
.SH BUGS
.PP
The memory management daemon returns empty descriptors to
.I rmstate()
along with descriptors for allocated and unallocated memory blocks.
An unused or empty descriptor has a starting address,
.I ma_start,
of NULL (0).
Thus, the selection of
.I maxsize
must exceed the total number of descriptors in the daemon, not
just the total number of memory blocks, if all the memory blocks
(the important information) are to be returned.
.SH "SEE ALSO"
.PP
mstate(l), rmalloc(l)
