.TH RPATTACH l "April 8, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rpattach, rpdetach \- Attach, detach a Trollius process
from kenya deamon (process control server) on local node.
.SH SYNOPSIS
.ft B
#include <kio.h>
.PP
.nf
.ft B
int	rpattach (int node);
int	rpdetach (int node, int pid, int status);
.fi
.SH DESCRIPTION
The remote function
.I rpattach()
enables a Trollius process on an ITB node to attach itself to the kenya
deamon on the local node.
Such a kenya-attached Trollius process can access kenya
remote services like creating, dooming, and spawning Trollius
processes, and checking their status (see rcreate(l), rdoom(l),
rspawn(l), rstate(l)) on ITB nodes.
Actually, kenyad can affect processes not attached to it if the client
specifies the pid (as on the OTB node).
With
.I rpattach(),
the client can locate the process without knowing the pid.
.PP
The
.I node
argument to
.I rpattach()
or
.I rpdetach()
is always the local node identifier (also specified by LOCAL) to whose
kenya deamon the caller desires to attach or detach from.
The reason they don't work on non-local ITB nodes is that kenyad
is like the remote agent of the kernel.
Pids should be valid locally, not remotely.
Also,
.I rpattach()
is used to get kenyad to clean up your memory when you invoke
.I rpdetach().
Memory addresses are only valid locally.
.PP
With
.I rpdetach(),
a process that had attached to a local kenya deamon
can detach itself or another attached process on its own node
from the kenya deamon.
This results in discontinuation of the kenya remote services to
the detached process.
Not only this, the kenya deamon assumes that the detached process
is proceding to exit and waits for it to terminate after which
kenyad frees its memory.
.PP
The parameters passed to
.I rpdetach()
are:
.I node,
the local nodeid;
.I pid,
the process ID of the process being detached; and
.I status,
the error code which is passed to the parent, if any, of the detached process.
A pid of 0 is interpreted as a request to detach the calling process, and
is the normal usage of
.I rpdetach().
The status argument is 0 for a normal exit of the detached process,
or is some nonzero error code.
.PP
.I Rpdetach()
is a dangerous request which should be used by processes
to detach themselves immediately prior to detaching from the
kernel or by experienced Trollius programmers who need to clean
up dead kenya descriptors.
.PP
The
.I rpattach()
and
.I rpdetach()
functions return error if called by an OTB (host) Trollius process since
the kenya deamon is absent on OTB node.
.SH "RETURN VALUE"
.I Rpattach()
and
.I rpdetach()
return 0 upon successful completion.
Otherwise, -1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 22
[EBADNODE]
The node address argument is invalid.
.TP
[ENOPDESCRIPTORS]
The
.I kenya
process descriptor tables are full.
.TP
[ENOTPROCESS]
The process whose process identifier was passed to
.I rpdetach()
is not attached to the kenya deamon.
.TP
[ESERVICE]
The remote function service is not available.
.SH "SEE ALSO"
rcreate(l), rdoom(l), rstate(l), rspawn(l)
