.TH RSPAWN l "April 8, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rspawn \- Spawn processes onto a remote ITB (compute) node.
.SH SYNOPSIS
.ft B
#include <preq.h>
.PP
.ft B
int	rspawn(int src_node, int sflags, int svalue, int dest_node, int argc, int argv_tag, int *pid);
.SH DESCRIPTION
The remote function
.I rspawn()
duplicates existing remote node processes and starts them running on
a different remote node.
This is a request to the
.I kenya
process-control server on the source node.
When a process is spawned an exact copy of the code segment and the current
data segment are made, but a fresh stack is set up and the new process
is started from the original initial entry point.
.PP
The first four parameters passed to
.I rspawn()
are:
.I src_node,
the node on which the original processes are running;
.I sflags,
the process selection flags;
.I svalue,
the process selection value; and
.I dest_node,
the node on which the new processes will be running.
The next argument,
.I argc,
is set to zero if the command line arguments of the original process are to
be used.
Otherwise it is set to
the number of command line arguments to be passed to the new processes.
.I Argv_tag
is an arbitrary integer.
If new command line arguments are to be used, rflat(l) must be called before
.I rspawn().
Rflat(l) will associate
.I argv_tag
with the address of the loaded module.
The last argument,
.I pid,
is a pointer to the address where a new process ID will be returned.
.PP
The process selection flags
.I sflags
are used to determine on what basis processes will be selected to be duplicated.
The process selection flags, defined in <preq.h>, are:
.TP 18
SELECT_PID
Select by process ID.
.TP
SELECT_INDEX
Select by process table index.
.TP
SELECT_BLKD
Select by the event the process is blocked on.
.TP
SELECT_LOADPT
Select by the load point of the process.
.TP
SELECT_APPL
Select all application (user) processes.
.PP
The process selection value
.I svalue
is set to the process ID, the process table index, the event, or the load
point, whichever corresponds with the process selection flag.
If the process selection flag is SELECT_APPL the process
selection value does not need to be set.
.PP
An
.I rspawn()
request can be sent to a group of nodes.
The nodes must first be registered as a multicast destination with rcast(l).
The cast identifier can then be used as the
.I dest_node
argument to
.I rspawn().
.PP
If you do not wish to use the command line arguments of the original processes,
you will need to set up an
.I argv
structure.
The
.I argv
structure starts with
.I (argc
+ 1)
.I uninitialized
pointers to strings, followed by the null-terminated strings.
The strings themselves should all be contiguous in memory.
.SH "RETURN VALUE"
Upon successful completion 0 is returned and the
.I pid
argument points to a new process ID.
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 22
[EBADTAG]
The
.I argv
tag is invalid.
.TP
[EBADNODE]
A node address argument is invalid.
.TP
[ENOFLDESCRIPTORS]
The
.I flatd
descriptor table is full.
.TP
[ENOMEM]
A memory allocation request cannot be satisfied.
.TP
[ENOPDESCRIPTORS]
The
.I kenya
process-descriptor tables are full.
.TP
[ENOTPROCESS]
No process exists satisfying the selection parameters.
.SH "SEE ALSO"
rcreate(l), rflat(l), rloadgo(l), spawn(l)
