.TH RSTATE l "April 8, 1992" "Release 2.2" "TROLLIUS REMOTE LIBRARY"
.SH NAME
rstate \- Find state of processes on a remote ITB (compute) node.
.SH SYNOPSIS
.ft B
.nf
#include <preq.h>
#include <kreq.h>
.fi
.PP
.ft B
int	rstate (int node, int sflags, int svalue, struct pstate *reply_state, int maxsize);
.SH DESCRIPTION
The remote function
.I rstate()
enables a Trollius process to inquire about the state of
specified processes on a specified compute node.
.PP
The parameters passed to
.I rstate()
are:
.I node,
the target node identifier;
.I sflags,
the process selection flags;
.I svalue,
the process selection value;
.I reply_state,
a pointer to an array of structures where state information will be stored; and
.I maxsize,
the number of elements in the array.
.PP
The process selection flags
.I sflags
are used to determine on what basis processes are selected to be stated.
The process selection flags, defined in <preq.h>, are:
.TP 18
SELECT_PID
Select by process ID.
.TP
SELECT_INDEX
Select by process table index.
.TP
SELECT_BLKD
Select by the event the process is blocked on.
.TP
SELECT_LOADPT
Select by the load point of the process.
.TP
SELECT_APPL
Select all application (user) processes.
.TP
SELECT_SYS
Select all system processes.
.TP
SELECT_ALL
Select all application and all system processes.
.PP
The process selection value
.I svalue
is set to the process ID, the process table index, the event, or the
load point, whichever corresponds with the process selection flag.
If the process selection flag is SELECT_APPL, SELECT_SYS, or SELECT_ALL, the
process selection value does not need to be set.
.PP
The information returned by
.I rstate()
is stored in an array of structures of the type
.I pstate,
defined in <preq.h> as follows:
.PP
.nf
.RS
.DT
.if n .ta .5i 1.0i 2.25i 2.75i
struct pstate {
	int		ps_reply;
	int		ps_index;
	struct pprocess	ps_kenya;
	KS_STATE	ps_kernel;
	char		ps_name[PSMAXNAME];
};
.RE
.fi
.PP
PSMAXNAME is defined in <preq.h> and is currently set to 64.
.PP
Each
.I pstate
structure holds information about one process.
The fields hold: the reply code from the kenya daemon (always 0),
the process-table index, a structure holding information obtained from the
kenya daemon, a structure holding information obtained from the Trollius
kernel, and the process name (derived from the file name used when loading).
.PP
The
.I ps_kenya
field holds process-control information kept by the Trollius kenya daemon.
It is a
.I pprocess
structure, defined in <preq.h> as follows:
.PP
.nf
.RS
.DT
.if n .ta .5i 1.0i 2.0i 2.5i
struct pprocess {
	int		p_pid;
	int		p_stksize;
	char		*p_stack;
	char		*p_static;
	struct exec	*p_loadpt;
	int		p_argc;
	char		**p_argv;
	int		p_flags;
	int		p_nodeid;
	int		p_event;
};
.RE
.fi
.PP
The fields contain: the process ID, its stack size in bytes, the starting
addresses of the stack and static data areas, the load point address,
the argc and argv arguments, process flags (see <kio.h>),
the parent process' node identifier, and the parent process' reply event
(meaningful for processes created from within another Trollius process).
.PP
The
.I ps_kernel
field holds information about the process that is kept in the Trollius kernel.
It is a
.I kstate
structure, defined in <kreq.h> as follows:
.PP
.nf
.RS
.DT
struct kstate {
	int	ks_state;
	int	ks_priority;
	int	ks_event;
	int	ks_type;
	int	ks_signal;
};
typedef struct kstate KS_STATE;
.RE
.fi
.PP
The fields contain: the process state, the priority, the event and type the
process is blocked on (if any), and the signal waiting
for delivery to the process (if any).
The process state, defined in <kreq.h>, can be:
.TP 15
SSYNC
ready, having called ksend(l) or krecv(l), often seen as SREADY
.TP
SSTATE
ready, having called kstate(l)
.TP
SERROR
ready, having encountered an error on any kernel request
.TP
SATTACH
ready, having called kattach(l)
.TP
SREADY
ready, having called kdetach(l), kdoom(l), ksend(l) or krecv(l)
.TP
SSEND_BLKD
blocked sending a message to an event
.TP
SRECV_BLKD
blocked receiving a message from an event
.TP
SACTIVE
running - not under Trollius kernel control
.SH "RETURN VALUE"
Upon successful completion of
.I rstate(),
the number of processes matching the selection criteria is returned, and
.I reply_state
points to an array of filled
.I pstate
structures.
The return value will be larger than the number of filled
.I pstate
structures if
.I maxsize
was less than the number of qualifying processes.
If an error occurred,
\-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 20
[EBADNODE]
The node address argument is invalid.
.SH "SEE ALSO"
kstate(l), state(l)
