.TH SENDB l "April 8, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
sendb, recvb \- Basic Trollius message passing.
.SH SYNOPSIS
.nf
.ft B
#include <net.h>

int	sendb (int destnode, void *msg, int length, int datatype);
int	recvb (int srcnode, void *msg, int length, int datatype);
.fi
.SH DESCRIPTION
.I Sendb()
and
.I recvb()
provide very simple message passing at the Trollius network level.
They allow the user to make point-to-point data transfers between two nodes.
These functions are meant to be used in pairs, each message sent by
.I sendb()
being received by
.I recvb().
.PP
.I Destnode
is used by
.I sendb()
to identify the destination node running the intended receiving process, local
or remote.
.I Srcnode
is used by
.I recvb()
to identify the source node running the process that is sending the message.
A receiving process thus can directly specify its synchronizing sending process.
Two special node identifiers are defined in <net.h>.
LOCAL refers to the local node and ORIGIN refers to the host node from which
the Trollius multicomputer was booted.
Since the sender and receiver processes are only identified by the nodes on
which they run (source node and destination node), this limits the applicability
of these functions to the case of a single user process per node.
.PP
The message being transmitted takes the form of an array of data elements.
.I Msg
holds the address of that array, i.e. the address of the first byte of data
to be sent.
The data must be stored contiguously in memory.
.PP
.I Length
holds the length of the message as a number of data elements.
For example, if the mesage to be sent is an array of 20 floating point numbers,
.I length
should be set to 20.
If the sender and the receiver specify different lengths, the lesser amount
will be transferred.
.PP
The type of the data elements is given by
.I datatype.
The functions transmit the message between any two Trollius nodes.
On nodes of different types, data may have different representations.
For example, integers may be stored with the most significant byte
first in memory (big-endian) or with the most significant byte last in
memory (little-endian).
.I Sendb()
and
.I recvb()
automatically modify the byte ordering to suit the internal data
representation of each node.
The data types supported, defined in <net.h>, are:
.TP 14
DINTMSG
The message is an array of 32-bit integers.
.TP
DFLTMSG
The message is an array of single precision real numbers.
.TP
DDBLMSG
The message is an array of double precision real numbers.
.TP
DRAWMSG
The message representation will not be changed.
.PP
DRAWMSG is typically used when sending an array of bytes (e.g. an ASCII
text string) or when the message array contains a mixture of data types.
In this later case, the user will have to specify the size of the array
in bytes and perform any necessary changes in the data representation by
using the several byte-order translation functions available in Trollius
(see ttol(l)).
.SH EXAMPLE USAGE
The following is an example of a typical usage of
.sendb()
and
.recvb()
to transmit an array of single precision floating point numbers
between two nodes (n3 and n7).
.PP
The sending process on node 3 sends the contents of the data array of length
5 to the receiving process on node 7 by executing:
.nf
.ta 1i
.ft B

float	data[5];

sendb(7, data, 5, DFLTMSG);
.fi
.PP
The receiving process on node 7, not knowing how many floating point numbers
are going to be sent, sets a maximum limit of 20.
The value returned by
.I recvb()
indicates how many numbers were actually received.
The code executed by the receiver is:
.nf
.ta 1i
.ft B

float	indata[20];
int	numdata;

numdata = recvb(3, indata, 20, DFLTMSG);
.fi
.SH BLOCKING
A process calling
.I recvb()
blocks until the message sent by the process calling
.I sendb()
arrives.
A process calling
.I sendb()
blocks until its message is picked by the Trollius message routing system.
The only thing that is guaranteed by a successful return from
.I sendb()
is that the message has entirely left the calling process, thus the sender
is free to reuse the space of the message array.
In the example above, the user can modify the values in the
.I data
array after
.I sendb()
returns.
.SH "RETURN VALUE
Upon successful completion,
.I sendb()
returns the number of data elements sent (i.e. the length of the array).
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
Upon successful completion,
.I recvb()
returns the number of data elements received (i.e. the lesser amount between
the length of the array sent and the length of the array expected).
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 23
[EINVAL]
The given data type is invalid.
.TP
[ENOEVDESCRIPTORS]
The Trollius kernel ran out of message descriptor space.
.TP
[EBADNODE]
The specified destination node does not exist.
.SH "SEE ALSO"
ttol(l)
