.TH TPRINT l "April 8, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
tprint, tprintf, tputc, tputs, tflush \- print messages
.SH SYNOPSIS
.nf
.ft B
int	tprint (char *str);
int	tprintf (char *format [, arg] ...);
int	tputc (int c);
int	tputs (char *str);
int	tflush();
.fi
.SH DESCRIPTION
These functions provide for easy output of text from a Trollius
program, without using the "stdio" library.
They are intended for the printing of debugging and diagnostic messages
from programs that otherwise do not need to output text.
.I Tprint()
is portable to all nodes; the other functions are limited to compute nodes.
.PP
.I Tprint()
is a simple wrapper around twrite(l), and simply outputs the
given null-terminated string to standard output.
.I Tprint()
does not do buffering of any sort: each call actually
sends the string out to the file server.
.PP
.I Tprintf()
provides a small subset of the formatting features of the
standard library function printf(3s), still without linking in the
full-blown stream I/O and formatting library.
The formatting options available are:
.TP 6
%c
character
.TP
%s
null-terminated string
.TP
%d
decimal integer
.TP
%u
unsigned decimal integer
.TP
%o
octal integer
.TP
%x
hexadecimal integer
.PP
The formatted output from
.I tprintf()
is actually written out (flushed from a buffer) after each call.
.PP
The functions
.I tputc(), tputs(),
and
.I tflush()
implement local buffering of the text, similar to the buffering that is
done in stream I/O.
Their purposes are analogous to those of fputc(3s), fputs(3s), and fflush(3s),
to output a character, output a string, and flush the buffer, respectively.
The buffer is flushed to stdout upon a newline, a call to
.I tflush(),
or when the buffer is full.
These functions may be used directly, but their
main purpose is for the implementation of
.I tprintf().
.SH RETURN VALUE
All of these functions return -1 upon error.
Upon successful completion,
.I tprint()
returns the number of bytes written,
.I tprintf(), tputs(),
and
.I tflush()
return 0, and
.I tputc()
returns the argument
.I c.
.SH BUGS
The d, u, o and x format letters in
.I tprintf()
may be preceded by 'l',
but that has no effect since integers are assumed to be the same as longs.
The method of implementation of the variable number of arguments in
.I tprintf()
is not portable.
