.TH TSEND l "April 24, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
tsend, trecv  \- Trollius transport level message passing
.SH SYNOPSIS
.nf
.ft B
#include <net.h>

int	tsend (struct nmsg *header);
int	trecv (struct nmsg *header);
.fi
.SH DESCRIPTION
These functions access the transport layer message passing services of
the Trollius network level.
.I Tsend()
and
.I trecv()
add flow control to the network layer services (see nsend(l))
and enable multiple processes to send multi-packet messages to a
receiver using the same event while maintaining the atomicity of
the messages.
The sending process sends a "ready-to-send" message to the receiver
and blocks until the receiving process signals that it is ready.
The receiver returns to the sender a new event, its process id negate, to
be used in the transfer of the body of the message.
The introduction of the second event allows multi-packet messages, using the
same event and destined to the same node, to be received with no mixing up
of packets.
The transport layer functions are also typically used during the development
and debugging stages.
Naturally, they have higher overhead than their network layer counterparts.
.PP
The purpose of the transport layer is to overcome the basic danger
of Trollius network message passing, which is the lack of strong
synchronization at the network and datalink layers.
A sending process can transmit a message that may never be received
due to programming error or deadlock.
This message will never be dropped or timed out.
Some Trollius process will always be stuck with it, waiting for a synchronizing
.I drecv()
or
.I nrecv()
that will never happen.
If that unfortunate process is a buffer, it can be located by the user
and swept clean (see sweep(l)).
However, if the process is a link proprietor, the link is henceforth plugged
and useless.
An intelligent buffering system does try to prevent this from
happening (see buffer(l)), but it remains theoretically possible.
.PP
The transport layer functions solve this problem by causing the sending
process to block until it receives a "ready-to-receive" protocol
message from the receiving process.
If due to a programming error there is no receiving process at the other
end, the "ready-to-send" message is guaranteed not to plug a link proprietor.
The second advantage in this case is that the sender is stopped from
flooding the system with messages that no receiver will consume.
After the pre-synchronizing exchange, the network layer functions
are invoked and the message is transmitted.
.PP
Use of the network message structure passed to
.I tsend()
and
.I trecv()
is the same as described under nsend(l).
.SH "RETURN VALUE
Upon successful completion, 0 is returned.
Otherwise, \-1 is returned and the global variable
.I errno
is set to indicate the error.
.SH ERRORS
.TP 23
[ENOEVDESCRIPTORS]
The Trollius kernel ran out of message descriptor space.
.TP
[EBADNODE]
The specified source or destination node does not exist.
.SH "SEE ALSO"
ksend(l), psend(l), dsend(l), nsend(l)
