.TH TTIME l "April 8, 1992" "Release 2.2" "TROLLIUS LOCAL LIBRARY"
.SH NAME
ttime
\- Trollius timing function
.SH SYNOPSIS
.nf
.ft B
.ta 0.75i
double	ttime ();
.fi
.SH DESCRIPTION
The
.I ttime()
function enables a program running on any Trollius node (ITB or OTB)
to read the current value of the system timer in a portable fashion.
This is useful for measurements of execution times of program sections.
.I Ttime()
returns the timer value in seconds.
.PP
On the host (OTB) the function returns the time value using the
gettimeofday(2) system call.
On a transputer (ITB), it returns the current value of the transputer timer.
There are actually two timers.
Processes running at low priority have an associated timer that is incremented
once every 64 microseconds.
Processes running at high priority read a timer that is incremented once
every microsecond.
.I Ttime()
handles both cases, adjusting the timer value depending on the
priority of the calling process, returning a reading in seconds.
.SH "EXAMPLE"
.nf
.ta 1.5i
#include <stdio.h>
extern double	ttime();
double	time, start, end;
start = ttime();
my_sub();
end = ttime();
time = end - start;
printf ("my_sub took: %f seconds to execute\\n", time);
.fi
.SH "BUGS"
Note that the transputer timers wrap around (from the most positive
to the most negative values).
This happens infrequently (less than once an hour for the high priority
timer), but does happen.
The resolution of the measurements is bound by that of the underlying
system (gettimeofday(2) on the host and the hardware timers on the transputer).
.SH "SEE ALSO"
gettimeofday(2)
