.TH TTOL l "April 8, 1992" "Release 2.2" "TROLLIUS NETWORK LIBRARY"
.SH NAME
ttol, ltot, ttols, ltots, ttolf, ltotf, ttold, ltotd, ttolcpy, ltotcpy
\- Trollius network byte-order translation
.SH SYNOPSIS
.nf
.ft B
#include <t_types.h>

.ft B
.ta 0.75i
int	ttol (int iword);
int	ltot (int iword);
short	ttols (short sword);
short	ltots (short sword);
float	ttolf (float fword);
float	ltotf (float fword);
double	ttold (double dword);
double	ltotd (double dword);
void	ttolcpy (int *dest, int *src, int nword);
void	ltotcpy (int *dest, int *src, int nword);
.fi
.SH DESCRIPTION
The order of bytes in a word may not be uniform across the nodes in a
given Trollius network.
Often, the host machine uses a different byte order than the compute nodes.
As a result, message and file data must frequently be converted to the
appropriate byte order.
Conversion is not necessary if the processes involved are on the same machine.
.PP
The functions convert the order of bytes between the local representation,
regardless of what that is, and a standard (Trollius) representation.
If no conversion is necessary, the functions will have no effect.
In reality, these functions are implemented as macros and thus should not
be accessed through pointers to functions.
.PP
When sending messages between nodes with different byte ordering,
only the message data, i.e., what is in the
.I nh_msg
field of the network message structure, must be converted.
All other fields, including
.I nh_data,
are automatically converted when the message is passed.
Data conversion functions should be used by both the sending and receiving
processes.
Data should be converted from local to Trollius representation before sending,
and from Trollius to local representation after receiving.
.PP
As an alternative to the byte-order translation functions, the programmer can
set various bits in the
.I nh_flags
field to convert the
.I nh_msg
data, or to convert
.I nh_data
in other than the default manner (see nsend(l)).
The default assumption is that
.I nh_data
is an array of 8 integers.
.PP
File data should be converted from local to Trollius representation before
writing, and from Trollius to local representation after reading.
.PP
The conversion functions are:
.TP 10
.I ttol()
Converts an integer from Trollius to local representation.
.TP
.I ltot()
Converts an integer from local to Trollius representation.
.TP
.I ttols()
Converts a short from Trollius to local representation.
.TP
.I ltots()
Converts a short from local to Trollius representation.
.TP
.I ttolf()
Converts a float from Trollius to local representation.
.TP
.I ltotf()
Converts a float from local to Trollius representation.
.TP
.I ttold()
Converts a double from Trollius to local representation.
.TP
.I ltotd()
Converts a double from local to Trollius representation.
.TP
.I ttolcpy()
Converts an array of integers from Trollius to local representation.
.TP
.I ltotcpy()
Converts an array of integers from local to Trollius representation.
.SH "RETURN VALUE"
Upon completion, the byte-order translation functions return the argument
with the desired value; the machine representation has been changed to
achieve this.
The array conversion function
.I ttolcpy()
and
.I ltotcpy()
do not return any value.
They place the converted array of dimension
.I nword
in
.I dest
leaving
.I src
unmodified.
.I Dest
and
.I src
may point to the same array.
.SH "SEE ALSO"
nsend(l)
