.TH VBOOT l  "January 5, 1992" "Release 2.2" "TROLLIUS COMMANDS"
.SH NAME
vboot \- Boot Trollius on a remote ITB node.
.SH SYNTAX
.B vboot
[-dhoqvN] [-{abefs}{xlptz}] [-c <config>]
.br
[-T <topology>] [-L <link>] [-n <nodeid>] [-i <frozen>]
.SH OPTIONS
.TP 15
.B \-d
Turn on debugging.
.TP
.B \-h
Print useful information on this tool.
.TP
.B \-o
Follow the old boot procedure.
.TP
.B \-q
Follow the quick boot procedure (fast boot).
.TP
.B \-v
Be verbose.
.TP
.B \-N
Pretend - go through the motions but take no effective action.
.TP
.B \-c <conf>
Use <conf> as the process schema file.
.TP
.B \-i <frozen>
Use <frozen> as the frozen process schema file.
.TP
.B \-a
Add the given object to the remote node.
.TP
.B \-b
Boot the neighbour ITB node of the remote node using the given object.
This option must be accompanied by the \fI-L\fR option.
.TP
.B \-f
Flush (remove) the given object from the remote node.
.TP
.B \-e
Recursively boot starting at the remote node using the given object.
.TP
.B \-s
Stage the given object from the remote node to the remote node
given by the required \fI-n\fR option.
.TP
.B \-L <link>
Boot the ITB neighbour node across this link number.
.TP
.B \-n <nodeid>
Stage this destination node.
.TP
.B \-T <topology>
Redefine the topology variable in the process schema (see conf(l)).
.SH DESCRIPTION
.PP
.I Vboot
boots multiple programs on an ITB node which is directly connected
to another ITB node running Trollius.
The strategy is to load the programs on the Trollius node, also called the
.I staging
node, and then boot the programs across the communication link connecting the
.I target
node.
The booting phase is similar to the function of sboot(l); a full boot
will require the functionality of tfish(l), export(l) and cboot(l).
The difference is that sboot(l) coordinates utilities operating over
a local communication link and
.I vboot
directs the same procedure remotely, over the Trollius network,
with a boot daemon process operating the remote communication link.
.I Vboot
is part of the booting sequence directed by spread(l) to boot Trollius
but it can be used directly to boot any system of processes.
.PP
A process schema is a description of the programs
which constitute the operating system on any given node.
Naturally, the process schema used by
.I vboot
should be the one that describes Trollius on an ITB node.
The grammar of the process schema is described in conf(l).
.SH OBJECTS
.PP
The remote booting system is roughly object oriented.
There is a set of objects and a set of operators.
One operator appended with one object forms one option to
.I vboot.
The boot objects are listed below with their key letters in
parentheses:
.sp
.TP 12
strap (x)
The strap object is the moses(l) bootstrap program.
The boot operator resets the target ITB node,
boots moses(l), and synchronizes the target node's timers with
its own.
.TP
stage (l)
The stage object is the process schema.
The add operator parses the process schema and loads the
programs onto the staging node.
The boot operator boots these programs into the target node,
from the staging node.
.TP
frozen (z)
The frozen object is the frozen process schema.
The add operator loads this file onto the staging node
and the boot operator boots it into the target node,
from the staging node.
.TP
spawn (p)
The spawn object is the spawn table, which is created on
each target node as it is booted.
Each entry describes one program that was booted.
The idea is that a target node may eventually become a
staging node and the running programs on the staging node
will be spawned to the target node.
Staging of the process schema will not be required.
Booting the spawn table assumes that the target's operating
system will be identical to the staging node's.
Also, all programs described in the staging table must
be re-entrant if spawn table booting is invoked.
.TP
route (t)
The route object is the route file, which contains a route
table for every node in the Trollius multicomputer.
The add operator loads the route file onto the staging node.
The boot operator extracts the route table for the target
node and sends it to the router process on the target node.
.SH OPERATORS
.PP
An action is specified to
.I vboot
by combining an operator and an object into a command line option.
For example,
.I -ax
means add (load) the bootstrap onto the staging node.
Not all operators work on all objects or not all combinations
are available at the command layer via
.I vboot.
The operators and their valid objects are listed below:
.sp
.TP 17
flush (f){xltz}
After an object has been used for booting or spreading, flushing
it releases memory back to the staging node's operating system.
.TP
add (a){xlz}
Objects can be loaded from local host node files.
Obviously an object must be loaded on the staging node before
is can be used by the boot or spread operators.
.TP
boot (b){xlpz}
The stage, spawn or frozen objects must not be booted before
the bootstrap object.
.TP
stage (s){xltz}
Objects can also be loaded from a staging node.
The target node must be previously booted with Trollius.
.TP
spread (e){lpz}
This operator executes a complicated recursive boot of the
boot spanning tree beginning with the staging node.
Choosing the spread operator with one of the stage, spawn
or frozen objects automatically means choosing the
strap and route objects as well.
The first staging node will boot all of its neighbours
with node type NT_BOOT (see rroute(l)).
It will then stage all of the spread objects (x, t and
one of l, p, or z) to each of the booted target nodes.
Finally, it invokes the spread operation on each of the
target nodes.
This procedure will recurse in parallel until it reaches
nodes with no type NT_BOOT neighbours.
The map(l) topology layer tool establishes the NT_BOOT
nodes for each node in the multicomputer.
.PP
When multiple operators are specified,
.I vboot
executes them in a fixed order, the same order as the operators
are listed above.
If an operator is given multiple times for multiple objects,
the order of execution is not important for all except the
boot operator.
The order of booting the target node is very important:
strap, stage or frozen or spawn (mutually exclusive), and
finally route.
Route object loading and booting can also be done by the
trout(l) command.
.PP
Multiple staging nodes can be specified on the command line.
The notation is n<list>, where <list> is a comma separated
list of node identifiers or identifier ranges, such as 0-3.
Node identifiers are established in the boot schema (see nail(l)),
and may be written in decimal or hexadecimal notation.
.SH "SPECIAL MENU"
.PP
Two short-hand options represent common combinations of operators
and objects.
The
.I -o
option stands for -fx -fl -ax -al -bx -bl -sx -sl, which is the
longest, most laborious, assume nothing, boot one can imagine.
The
.I -q
option stands for -bx -bl -sx -sl, which assumes that the objects
have been loaded on the staging node.
Along with trout(l), these two options could be use to sequentially
boot an arbitrary size multicomputer, in a fairly efficient manner.
.SH EXAMPLES
.TP 4
.I "vboot -v -ax -al n0,4"
Add the bootstrap and the process schema to staging nodes 0 and 4.
Report about each step as it is done.
.TP
.I "vboot -bx -bz -sx -sz -L 2 -n 1 n0"
On staging node 0, boot then stage the bootstrap and frozen process schema
on target node 1 connected via staging node's link 2.
.TP
.I "vboot -ep n0"
Recursively boot, starting with the spawn table of staging node 0,
all the children of node 0 in the boot spanning tree.
.SH FILES
.TP 4
$TROLLIUSHOME/boot/conf.ivb
default process schema for remote ITB node, where $TROLLIUSHOME is
the installation directory
.TP
$TROLLIUSHOME/boot/conf.ivbz
default frozen process schema for remote ITB node
.SH DIAGNOSTICS
If a particular process schema, frozen or not, contains programs which
either do not work or do not work together,
.I vboot
may fail since moses(l) may be clobbered.
The only recourse is to use tkill(l) to reset the host node, followed by
tfish(l) to reset all of the ITB nodes.
.SH SEE ALSO
conf(l), spread(l), trout(l)
