/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	btstate.c,v $
 * Revision 2.2  92/04/30  14:09:14  trillium
 * Ohio Release
 * 
 *	Function:	- prints status information gathered from bootd
 *			- nodes are selected in a system dependent fashion
 *	Accepts:	- see the "usage" variable
 */

#include <bootreq.h>
#include <stdio.h>
#include <terror.h>			/* error codes */

/*
 * external variables
 */
extern int 		errno;		/* global error variable */

/*
 * external functions
 */
extern void		terror();
extern void		validopts();	/* set the valid options */
extern int		do_args();	/* parse the valid options */
extern int		opt_taken();	/* Was a particular option selected? */
extern int		parse_nodes();	/* parse the node identifiers */
extern int		rbtstate();	/* remote bootd status */
extern int4		getnode();	/* get next node entry */
extern char		*malloc();
extern char		*mnemonic();	/* ascii representation of nodeid */

/*
 * local functions
 */
static void		btstate1();
static void		help();

/*
 * local variables
 */
static char *_copyright = "Copyright 1991 The Ohio State University $Revision: 2.2 $";

static struct bootstate	*btstbl;	/* memory block descriptors */

/*
 * syntax for btstate command
 */
static char *usage = "btstate [-hp] [-B <#progs>] <nodes>";

main(argc, argv)

int   			argc;    
char   			*argv[];

{
	int4		n_index;		/* index of node entry */
	int4		nodeid;			/* value of node entry */
	int4		n_flags;		/* extra node entry info */
	int4		nprog;			/* # descs prog state table */
	int4		object;		/* selected bootd objects */
/*
 * Parse the command line.
 */
	validopts("hp");
	followed("B");

	if ((do_args(&argc, argv) < 0) ||
		(parse_nodes(&argc, argv) < 0) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n", usage);
		} else {
			terror("btstate (arguments)");
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Allocate status descriptor table.
 */
	if (opt_taken('B')) {
		intparam('B', &nprog);
	} else {
		nprog = BTSMAX;
	}

	btstbl = (struct bootstate *) malloc((unsigned) nprog *
		sizeof(struct bootstate));

	if (! btstbl) {
		terror("btstate (malloc)");
		exit(errno);
	}

	object = opt_taken('p') ? BTO_SPAWN : BTO_STAGE;
/*
 * Attach to kernel.
 */
	if (kinit(0) < 0) {
		terror("btstate (kinit)");
		exit(errno);
	}
/*
 * Loop through all specified nodes.
 */
	getnode(&n_index, &nodeid, &n_flags);

	if (n_index >= 0) {
/*
 * Print status header.
 */
		printf("NODE      LOAD     DATA     STKSIZE  PRI NAME\n");

		do {
			if (n_flags) {
				fprintf(stderr, "usage: %s\n", usage);
				kexit(errno);
			}

			btstate1(nodeid, nprog, object);
			getnode(&n_index, &nodeid, &n_flags);
		} while (n_index);
	}

	else {
		fprintf(stderr, "usage: %s\n", usage);
		kexit(errno);
	}

	kexit(0);
}

/*
 *	btstate1
 *
 *	Function:	- gets and prints the boot program state of 1 node
 *
 *	Accepts:	- node identifier
 *			- maximum # of program descriptors to print
 *			- bootd object selector
 */

static void
btstate1(nodeid, nprog, object)

int4			nodeid;
int4			nprog;
int4			object;

{
	int4		ndaemon;	/* # programs in daemon */
	int4		i;		/* favourite index variable */

	if ((ndaemon = rbtstate(nodeid, btstbl, nprog, object)) < 0) {
		terror("btstate (rbtstate)");
		kexit(errno);
	}

	if (nprog > ndaemon) nprog = ndaemon;

	for (i = 0; i < nprog; ++i) {
		printf("%-9.9s", mnemonic(nodeid));
		printf(" %-8X %-8X %-8X %-3d",
			(int4) btstbl[i].bts_load,
			(int4) btstbl[i].bts_data,
			btstbl[i].bts_stksize,
			btstbl[i].bts_priority);

		printf(" %-20.20s\n", btstbl[i].bts_name);
	}

	if (ndaemon > nprog) {
		printf("%-9.9s ...more...\n", mnemonic(nodeid));
	}
}

/*
 *	  help
 *
 *	  Function:	- prints helpful information on the btstate command
 */

static void
help()
{
	printf("\nSynopsis:	btstate [options] <nodes> [processes]\n");
	printf("\nDescription:	Print information on boot programs.\n");
	printf("\nOptions:	-h	Print this help message.\n");
	printf("\t\t-p	Display programs in spawn table.\n");
	printf("\t\t-B #	Print # program descriptors.\n");
	mnusage();
	printf("\nRestriction:\tUse btstate only with ITB nodes.\n");
	printf("\nExample:	btstate n3\n");
	printf("\t\t\t\"Display staged program information for node 3.\"\n\n");
}
