/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	buffer.c,v $
 * Revision 2.2  92/04/30  14:10:04  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  11:31:17  gdburns
 * Control user space threshold.
 * 
 * Revision 2.1  91/03/20  02:52:16  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:26:29  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:53:55  trillium
 * Homo erectus
 *
 *	Function:	- command line access to the buffer server
 */

#include <bufreq.h>
#undef NULL
#include <stdio.h>
#include <terror.h>
#include <typical.h>

/*
 * local variables
 */
static char		*usage =
	"buffer [-hS] [-b <#workers>] [-s <#space>] [-u <#usrspace>] <nodes>";

/*
 * local functions
 */
static void		help();

/*
 * external variables
 */
extern int 		errno;

/*
 * external functions
 */
extern void		validopts();
extern void		followed();
extern void		terror();
extern int		doargs();
extern int		opt_taken();
extern int		parse_nodes();
extern int		predef();
extern int		intparam();
extern int		rbuffer();


main(argc, argv)

int			argc;    
char			*argv[];

{
	int		n_index;		/* for getnode() */
	int		n_flags;		/* for getnode() */
	int		nodeid;			/* target node ID */
	int		parm;			/* # workers or # space */
	int		taken;			/* at least one opt taken? */

/*
 * Parse command-line arguments.
 */
	validopts("bhsuS");
	followed("bsu");

	if ((do_args(&argc, argv) < 0) ||
		(parse_nodes(&argc, argv) < 0) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n", usage);
		} else {
			terror("buffer (arguments)");
		}

		exit(errno);
	}

	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Initialize and attach to Trollius kernel.
 */
	if (kinit(0) < 0) {
		terror("buffer (kinit)");
		exit(errno);
	}
/*
 * Loop through all specified nodes.
 */
	getnode(&n_index, &nodeid, &n_flags);

	if (n_index < 0) {
		fprintf(stderr, "usage: %s\n", usage);
		kexit(EUSAGE);
	}

	do {
		taken = FALSE;
/*
 * Register any host castids (predefs) required.
 */
		if (predef(nodeid)) {
			terror("buffer (predef)");
			kexit(errno);
		}

		if (opt_taken('b')) {
			taken = TRUE;

			if (intparam('b', &parm)) {
				fprintf(stderr, "usage: %s\n", usage);
				kexit(EUSAGE);
			}
		
			if (rbuffer(nodeid, BFWORKER, parm)) {
				terror("buffer (rbuffer)");
				kexit(errno);
			}
		}

		if (opt_taken('s')) {
			taken = TRUE;

			if (intparam('s', &parm)) {
				fprintf(stderr, "usage: %s\n", usage);
				kexit(EUSAGE);
			}
		
			if (rbuffer(nodeid, BFSPACE, parm)) {
				terror("buffer (rbuffer)");
				kexit(errno);
			}
		}

		if (opt_taken('u')) {
			taken = TRUE;

			if (intparam('u', &parm)) {
				fprintf(stderr, "usage: %s\n", usage);
				kexit(EUSAGE);
			}
		
			if (rbuffer(nodeid, BFUSRSPACE, parm)) {
				terror("buffer (rbuffer)");
				kexit(errno);
			}
		}

		if (opt_taken('S')) {
			taken = TRUE;
		
			if (rbuffer(nodeid, BFSWEEP, 0)) {
				terror("buffer (rbuffer)");
				kexit(errno);
			}
		}

		if (! taken) {
			fprintf(stderr,
				"buffer: Use bfstate for buffer status.\n");
			kexit(EUSAGE);
		}

		getnode(&n_index, &nodeid, &n_flags);
	} while (n_index);

	kexit(0);
}


/*
 *	  help
 *
 *	  Function:	- prints helpful information on the buffer command
 */

static void
help()
{
	printf("\nSynopsis:	buffer [options] <nodes>\n");
	printf("\nDescription:	Control Trollius message buffers.\n");
	printf("\nOptions:	-h	this help message\n");
	printf("\t\t-b #	set max # of buffer workers\n");
	printf("\t\t-s #	set max total amount buffer space\n");
	printf("\t\t-u #	set max amount user space\n");
	printf("\t\t-S	sweep out all buffers (equiv: sweep)\n");
	mnusage();
	printf("\t\th	(local host node)\n");
	printf("\t\tC	(all ITB nodes)\n");
	printf("\t\tH	(all host nodes)\n");
	printf("\t\tN	(all nodes)\n");
	printf("\nExample:	buffer C -s 0x40000\n");
	printf("\t\t\t\"Set maximum buffer space to 256K bytes on all\n");
	printf("\t\t\t ITB nodes.\"\n");
}
