/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	cast.c,v $
 * Revision 2.2  92/04/30  14:11:09  trillium
 * Ohio Release
 * 
 *	Function:	- command to control the cast daemon
 *
 *	Accepts:	- see the "usage" variable
 *
 *	Examples:	- cast -c 400 n10
 *			- melt the tree of castid 400 starting from node 10
 */

#include <stdio.h>
#include <string.h>
#include <portable.h>
#include <net.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>
#include <castreq.h>

/*
 * external variables
 */
extern int 		errno;		/* global error variable */

/*
 * external functions
 */
extern void		exit();
extern void		followed();
extern void		intparam();
extern void		terror();
extern void		validopts();
extern char		*getparam();
extern char		*mnemonic();
extern int		do_args();
extern int		getnode();
extern int		opt_taken();
extern int		parse_nodes();
extern int		rmelt();
	
/*
 * local functions
 */
static void		help();

/*
 * local variables
 */
static int4		castid;		/* requested castid */
static char		*me;		/* tool name */

static char *_copyright =
	"Copyright 1992 The Ohio State University $Revision: 2.2 $";

/*
 * syntax for cstate command
 */
static char *usage = "usage: %s [-h] [-c castid <nodes>]\n";

/*
 * Tool starts here.
 */
main(argc, argv)

int   			argc;    
char   			*argv[];

{
	int4		nodeid;		/* value of node entry */
	int		n_flags;	/* extra node entry info */
	int		n_index;	/* index of node entry */
	char		*c_param;	/* 'c' option parameter */
/*
 * Set the tool name.
 */
	if ((me = strrchr(argv[0], STRDIR)) == NULL) {
		me = argv[0];
	} else {
		me++;
	}
/*
 * Parse the command line.
 */
	validopts("h");
	followed("c");

	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, usage, me);
		} else {
			terror(me);
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Get the castid.
 */
	if (opt_taken('c')) {
		c_param = getparam('c');
		if (strcmp(c_param, "HOST2ITB") == 0) {
			castid = HOST2ITB;
		} else if (strcmp(c_param, "HOST2OTB") == 0) {
			castid = HOST2OTB;
		} else if (strcmp(c_param, "HOST2ALL") == 0) {
			castid = HOST2ALL;
		} else {
			(void) intparam('c', &castid);
		}
	} else {
		fprintf(stderr, usage, me);
		exit(EUSAGE);
	}
/*
 * Attach to kernel.
 */
	if (kinit(0)) {
		terror(me);
		exit(errno);
	}
/*
 * Loop through all specified nodes.
 */
	getnode(&n_index, &nodeid, &n_flags);

	if (n_index != ERROR) {

		do {
			if (n_flags) {
				fprintf(stderr, usage, me);
				kexit(errno);
			}

			if (rmelt(nodeid, castid)) {
				terror(me);
				kexit(errno);
			}

			getnode(&n_index, &nodeid, &n_flags);
		} while (n_index);
	}
	else {
		fprintf(stderr, usage, me);
		kexit(errno);
	}

	kexit(0);
}

/*
 *	  help
 *
 *	  Function:	- prints helpful information on the cstate command
 */
static void
help()

{
	printf("\nSynopsis:\t%s [options] <nodes>\n", me);
	printf("\nDescription:\tControl Trollius multicasting/multireeling.\n");
	printf("\nOptions:\t-h\t\tthis help message\n");
	printf("\t\t-c <castid>\tmelt the tree of this castid\n");
	mnusage();
	printf("\t\th\t(local host node)\n");
	printf("\nCastid:\t\tHOST2ITB, HOST2OTB, HOST2ALL, or <constant>\n");
	printf("\nExample:\tcast -c 400 n10\n");
	printf("\t\t\t\"Melt the spanning tree of castid 400 starting\n");
	printf("\t\t\t from node 10.\"\n");
}
