/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	cstate.c,v $
 * Revision 2.2  92/04/30  14:11:41  trillium
 * Ohio Release
 * 
 *	Function:	- command to print cast daemon status information
 *
 *	Accepts:	- see the "usage" variable
 *
 *	Examples:	- cstate n10
 *			- describe all cast identifiers declared on node 10
 */

#include <stdio.h>
#include <string.h>
#include <portable.h>
#include <net.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>
#include <castreq.h>

/*
 * external variables
 */
extern int 		errno;		/* global error variable */

/*
 * external functions
 */
extern void		exit();
extern void		followed();
extern void		intparam();
extern void		terror();
extern void		validopts();
extern char		*getparam();
extern char		*malloc();
extern char		*mnemonic();
extern int		do_args();
extern int		getnode();
extern int		opt_taken();
extern int		parse_nodes();
extern int4		rcstate();	/* remote castd status */
extern int4		rcstate2();	/* remote castd secondary status */
	
/*
 * local functions
 */
static void		do_cstate();
static void		do_cstate2();
static void		header();
static void		help();
static void		print_castid();
static void		print_combine();

/*
 * local variables
 */
static int4		castid;		/* requested castid */
static int4		cmax;		/* max entries */
static char		*me;		/* tool name */

static char *_copyright =
	"Copyright 1992 The Ohio State University $Revision: 2.2 $";

/*
 * syntax for cstate command
 */
static char *usage = "usage: %s [-h] [-c castid] [-C <maxcasts>] <nodes>\n";

/*
 * Tool starts here.
 */
main(argc, argv)

int   			argc;    
char   			*argv[];

{
	int4		nodeid;		/* value of node entry */
	int		n_flags;	/* extra node entry info */
	int		n_index;	/* index of node entry */
	char		*c_param;	/* 'c' option parameter */
/*
 * Set the tool name.
 */
	if ((me = strrchr(argv[0], STRDIR)) == NULL) {
		me = argv[0];
	} else {
		me++;
	}
/*
 * Parse the command line.
 */
	validopts("h");
	followed("cC");

	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, usage, me);
		} else {
			terror(me);
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Get the limit on the number of entries allowed.
 */
	if (opt_taken('C')) {
		(void) intparam('C', &cmax);
	} else {
		cmax = MAXCAST;
	}
/*
 * Check if primary or secondary information requested.
 */
	if (opt_taken('c')) {
		c_param = getparam('c');
		if (strcmp(c_param, "HOST2ITB") == 0) {
			castid = HOST2ITB;
		} else if (strcmp(c_param, "HOST2OTB") == 0) {
			castid = HOST2OTB;
		} else if (strcmp(c_param, "HOST2ALL") == 0) {
			castid = HOST2ALL;
		} else {
			(void) intparam('c', &castid);
		}
	} else {
		castid = NOTNODEID;
	}
/*
 * Attach to kernel.
 */
	if (kinit(0)) {
		terror(me);
		exit(errno);
	}
/*
 * Loop through all specified nodes.
 */
	getnode(&n_index, &nodeid, &n_flags);

	if (n_index != ERROR) {
/*
 * Print status header.
 */
		header();

		do {
			if (n_flags) {
				fprintf(stderr, usage, me);
				kexit(errno);
			}

			if (castid == NOTNODEID) {
				do_cstate(nodeid);
			} else {
				do_cstate2(nodeid, castid);
			}

			getnode(&n_index, &nodeid, &n_flags);
		} while (n_index);
	}
	else {
		fprintf(stderr, usage, me);
		kexit(errno);
	}

	kexit(0);
}

/*
 *	do_cstate
 *
 *	Function:	- find and print the castd state of a remote node
 *			- primary state information is printed
 *
 *	Accepts:	- node identifier
 */
static void
do_cstate(nodeid)

int4			nodeid;

{
	int4		total;		/* total # castd entries */
	int4		size;		/* size of state table */
	int4		i;		/* favorite index */
	int		save_errno;	/* copy of error code */
	struct cstate	*ctable;	/* state table */
/*
 * Allocate the state table.
 */
	ctable = (struct cstate *) malloc((uint4) cmax * sizeof(struct cstate));
	if (ctable == CSNULL) {
		terror(me);
		kexit(errno);
	}

	if ((total = rcstate(nodeid, ctable, cmax)) == ERROR) {
		terror(me);
		save_errno = errno;
		(void) free((char *) ctable);
		kexit(save_errno);
	}

	size = min(total, cmax);
	for (i = 0; i < size; i++) {
		printf("%-9.9s", mnemonic(nodeid));
		print_castid(ctable[i].cs_castid);
		printf(" %-9.9s", mnemonic(ctable[i].cs_root_node));
		printf(" %-9.9s", mnemonic(ctable[i].cs_parent_node));
		print_combine(ctable[i].cs_combine);
		printf(" %-3d%-4.4s", abs(ctable[i].cs_nbranch),
			(ctable[i].cs_nbranch < 0) ? "*" : " ");
		printf(" %-8d\n", ctable[i].cs_nreels);
	}

	if (total > cmax) {
		printf("...and more registered castids...\n");
	}

	(void) free((char *) ctable);
}

/*
 *	do_cstate2
 *
 *	Function:	- find and print the castd state of message reeling
 *			  for the specified cast identifier
 *			- secondary state information is printed
 *
 *	Accepts:	- node identifier
 *			- cast identifier
 */
static void
do_cstate2(nodeid, cast_id)

int4			nodeid;
int4			cast_id;

{
	int4		total;		/* total # castd entries */
	int4		size;		/* state table size */
	int4		i;		/* favorite index */
	int		save_errno;	/* copy of error code */
	struct cstate2	*ctable;	/* state table */
/*
 * Allocate the state table.
 */
	if ((ctable = (struct cstate2 *) malloc((uint4) cmax *
		sizeof(struct cstate2))) == CS2NULL) {

		terror(me);
		kexit(errno);
	}

	if ((total = rcstate2(nodeid, cast_id, ctable, cmax)) == ERROR) {
		terror(me);
		save_errno = errno;
		(void) free((char *) ctable);
		kexit(save_errno);
	}

	size = min(total, cmax);
	for (i = 0; i < size; i++) {
		printf("%-9.9s", mnemonic(nodeid));
		print_castid(ctable[i].cs2_castid);
		printf(" %-11d", ctable[i].cs2_event);
		printf(" %-9d", ctable[i].cs2_type);
		printf(" %-9d\n", ctable[i].cs2_npkt);
	}

	if (total > cmax) {
		printf("...and more registered events...\n");
	}

	(void) free((char *) ctable);
}

/*
 *	  header
 *
 *	  Function:	- print the cstate command primary
 *			  or secondary table header
 */
static void
header()

{
	if (opt_taken('c')) {
/*
 * Print secondary header.  Castid reeling information.
 */
		printf("%-9.9s %-11.11s %-11.11s %-9.9s %-9.9s\n",
			"NODE", "CASTID", "EVENT", "TYPE", "PACKETS");
	}
 	else {
/*
 * Print primary header.  Main castid information.
 */
		printf("%-9.9s %-11.11s %-9.9s %-9.9s %-11.11s %-7.7s %-8.8s\n",
			"NODE", "CASTID", "ROOT", "PARENT", "COMBINE",
			"LEAVES", "REELS");
	}
}

/*
 *	print_castid
 *
 *	Function:	- print the given castid
 *			- handle the predefined castids
 *
 *	Accepts:	- cast identifier
 */
static void
print_castid(castid)

int4			castid;

{
	switch(castid) {
		case HOST2ITB:
			printf(" %-11.11s", "HOST2ITB"); break;
		case HOST2OTB:
			printf(" %-11.11s", "HOST2OTB"); break;
		case HOST2ALL:
			printf(" %-11.11s", "HOST2ALL"); break;
		default:
			printf(" %-11d", castid); break;
	}
}

/*
 *	print_combine
 *
 *	Function:	- print name of internal combining function
 *			  or event number if it is an external one
 *
 *	Accepts:	- combining function
 */
static void
print_combine(cmbfunc)

int4			cmbfunc;

{
	switch(cmbfunc) {
		case CBN_ERR:
			printf(" %-11.11s", "ERR"); break;
		case CBN_MIN:
			printf(" %-11.11s", "MIN"); break;
		case CBN_MAX:
			printf(" %-11.11s", "MAX"); break;
		case CBN_ADD:
			printf(" %-11.11s", "ADD"); break;
		case CBN_AND:
			printf(" %-11.11s", "AND"); break;
		case CBN_OR:
			printf(" %-11.11s", "OR"); break;
		case CBN_XOR:
			printf(" %-11.11s", "XOR"); break;
		default:
			printf(" %-11d", cmbfunc); break;
	}
}

/*
 *	  help
 *
 *	  Function:	- prints helpful information on the cstate command
 */
static void
help()

{
	printf("\nSynopsis:\t%s [options] <nodes>\n", me);
	printf("\nDescription:\tPrint information on cast registrations.\n");
	printf("\nOptions:\t-h\t\tthis help message\n");
	printf("\t\t-c <castid>\tsecondary information on this castid\n");
	printf("\t\t-C <#>\t\tset maximum # of castids to print\n");
	mnusage();
	printf("\t\th\t(local host node)\n");
	printf("\nCastid:\t\tHOST2ITB, HOST2OTB, HOST2ALL, or <constant>\n");
	printf("\nExample:\tcstate h n3-9,11\n");
	printf("\t\t\t\"Get primary cast information on the local host\n");
	printf("\t\t\t node and nodes 3 thru 9 and 11.\"\n");
}
