/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB/DLF
 *
 *	$Log:	drain.c,v $
 * Revision 2.2  92/04/30  14:12:30  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  02:52:20  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:26:28  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:53:54  trillium
 * Homo erectus
 * 
 *
 *	Function : - drain memory into a file
 *
 *	Accepts  : - various flags [-h -v]
 *		   - source node of file
 *		   - start address of memory to drain
 *		   - number of bytes to drain
 *		   - filename to drain data into
 *
 *
 *	Example : 
 *
 *		   drain n2 0x106A3 1000 file1	- drain 1000 bytes from node
 *						  2 starting at address 0x106A3
 *						  and place them in file1
 */


#include <stdio.h>
#include <terror.h>	/* error values */
#include <typical.h>	/* commonly used constants */


/*
 * external variables
 */
extern int	errno;

/*
 * external functions
 */
extern int	do_args();
extern int	kinit();
extern int	rdrain();
extern int	opt_taken();
extern void	validopts();

/*
 * local functions
 */
static void	help();

/*
 * local variables
 */
static char *usage = "drain [-hv] <node> <address> <length> [<filename>]";



main(argc, argv)

int		argc;    
char		*argv[];

{
	int	address;		/* remote address */
	int	nodeid;			/* remote node ID */
	int	index;			/* parsed node index */
	int	flags;			/* parsed node flags */
	int	length;			/* #bytes to drain */
	char	*filename;		/* local node destination file */


        validopts("hv");
 
        if ((do_args(&argc, argv) == ERROR) ||
                (parse_nodes(&argc, argv) == ERROR) ||
                (errno = (argc <= 4) ? 0 : EUSAGE)) {
 
                if (errno == EUSAGE) {
                        fprintf(stderr, "usage: %s\n", usage);
                }
 
                else {
                        terror("drain (arguments)");
                }
 
                exit(errno);
        }
/*
 * Check for help request.
 */
        if (opt_taken('h')) {
                help();
                exit(0);
        }

	getnode(&index, &nodeid, &flags);

	if ((index == ERROR) || flags || (argc < 3)) {
		fprintf(stderr, "usage: %s\n", usage);
		exit(EUSAGE);
	}

	address = stoi(argv[1]);
	length = stoi(argv[2]);
	filename = (argc > 3) ? argv[3] : "drain.out";
/*
 * Attach to kernel.
 */
	if (kinit(0)) {
		terror("drain (kinit)");
		exit(errno);
	}
/*
 * Dump to file.
 */
	if (rdrain(nodeid, address, length, filename) == ERROR) {
		terror("drain (rdrain)");
		kexit(errno);
	}

	if (opt_taken('v')) {
		printf("drained OK %d bytes at %X on node %X into %s\n",
			length, address, nodeid, filename);
	}

	kexit(0);
}

 
/*
 *        help
 *
 *        Function:     - prints helpful information on the drain command
 */
 
static void
help()
{
        printf("\nSynopsis:\tdrain [options] <node> <addr> <size> [<file>]\n");
        printf("\nDescription:    Dump memory from a Trollius compute node.\n");
        printf("\nOptions:	-h	this help message\n");
        printf("\t\t-v      turn on verbose mode\n");
        printf("\t\t<addr>  compute node starting address\n");
        printf("\t\t<size>  drain size in bytes\n");
        printf("\t\t<file>  local destination file (default: `drain.out')\n");
        printf("\nNode:		n<constant>\n");
        printf("\t\t\t(any single compute node)\n");
        printf("\t\t\twhere <constant> = decimal or hexadecimal (0x)\n");
        printf("\nExample:        drain n10 100 1024 image\n");
        printf("\t\t\t\"Dump 1K bytes from address 100 on node 10\n");
        printf("\t\t\t into the file `image'.\"\n");
}
