/*
 *	OSU Research Computing, architects and contractors	Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	fstate.c,v $
 * Revision 2.2  92/04/30  14:12:45  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/06  10:59:40  gdburns
 * Add formatting room for the internal file descriptor number.
 * 
 * Revision 2.1.1.1  91/12/18  15:59:39  gdburns
 * Use int4 cast on constants in comparison operation.
 * 
 * Revision 2.1  91/02/27  02:43:33  gdburns
 * Ohio Release
 *
 *	Function:	- command to print Trillium filed status info
 *
 *	Accepts:	- see the "usage" variable
 */

#include <fcntl.h>
#include <freq.h>
#include <net.h>
#include <stdio.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>

#define FSMAX	256
#define FQCXMAX	9

/*
 * external variables
 */
extern int 		errno;		/* global error variable */

/*
 * external functions
 */
extern void		validopts();	/* set the valid options */
extern void		terror();
extern void		intparam();
extern void		followed();
extern char		*malloc();
extern int		do_args();	/* parse the valid options */
extern int		getnode();	/* get next node entry */
extern int		opt_taken();	/* Was a particular option selected? */
extern int		parse_nodes();	/* parse the node identifiers */
extern int		rfstate();	/* remote filed status */

/*
 * local functions
 */
static void		header();
static void		help();
static void		do_fstate();

/*
 * local variables
 */
static char *_copyright
	= "Copyright 1990 The Ohio State University $Revision: 2.2 $";

static struct fstate	*ftable;		/* file descriptors */
static int		fsmax;			/* max entries */
static int		bcubix;			/* cubix display? */

static char		*reqname[FQCXMAX] = {
				"open",
				"close",
				"read",
				"write",
				"lseek",
				"cxwrite",
				"chcx",
				"order",
				"idle"
			};

/*
 * syntax for fstate command
 */
static char *usage = "fstate [-ach] [-F <maxdesc>] <nodes>";


main(argc, argv)

int   			argc;    
char   			*argv[];

{
	int		n_index;		/* index of node entry */
	int		nodeid;			/* value of node entry */
	int		n_flags;		/* extra node entry info */
/*
 * Parse the command line.
 */
	validopts("Cach");
	followed("F");

	if ((do_args(&argc, argv) < 0) ||
		(parse_nodes(&argc, argv) < 0) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n", usage);
		} else {
			terror("fstate (arguments)");
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Allocate filed state table.
 */
	if (opt_taken('F')) {
		 intparam('F', &fsmax);
	} else {
		fsmax = FSMAX;
	}

	ftable = (struct fstate *) 
		malloc((unsigned) (fsmax * sizeof(struct fstate)));

	if (! ftable) {
		terror("fstate (malloc)");
		exit(errno);
	}

	bcubix = opt_taken('c') || opt_taken('C');
/*
 * Attach to kernel.
 */
	if (kinit(0) < 0) {
		terror("fstate (kinit)");
		exit(errno);
	}
/*
 * Loop through all specified nodes.
 */
	getnode(&n_index, &nodeid, &n_flags);

	if (n_index >= 0) {
/*
 * Print status header.
 */
		header();

		do {
			if (n_flags) {
				fprintf(stderr, "usage: %s\n", usage);
				kexit(errno);
			}

			do_fstate(nodeid);
			getnode(&n_index, &nodeid, &n_flags);
		} while (n_index);
	}

	else {
		fprintf(stderr, "usage: %s\n", usage);
		kexit(errno);
	}

	kexit(0);
}


/*
 *	do_fstate
 *
 *	Function:	- find and print the filed state of a remote node
 *	Accepts:	- node identifier
 */

static void
do_fstate(nodeid)

int			nodeid;

{
	int		count;		/* # of return blocks */
	int		limit;		/* # of blocks to print */
	int		i;		/* favourite index */
	char		buf[24];	/* field buffer */
	char		*p;		/* ptr within buf */


	if ((count = rfstate(nodeid, ftable, fsmax)) < 0) {
		terror("fstate");
		kexit(errno);
	}

	limit = (count < fsmax) ? count : fsmax;

	for (i = 0; i < limit; ++i) {
/*
 * Forgive me for this.
 */
		if (bcubix && !(ftable[i].fs_tflags & FCUBIX)) {
			continue;
		}

		if (opt_taken('a') && !(ftable[i].fs_tflags & FACTIVE)) {
			continue;
		}

		printf("%-6d ", ftable[i].fs_tfd);
		p = buf;

		if (ftable[i].fs_tflags & O_WRONLY) {
			*p++ = 'W';
		}

		else if (ftable[i].fs_tflags & O_RDWR) {
			*p++ = 'R';
			*p++ = '|';
			*p++ = 'W';
		} else {
			*p++ = 'R';
		}

		if (ftable[i].fs_tflags & O_SINGL) {
			*p++ = '|';
			*p++ = 'S';
		} else if (ftable[i].fs_tflags & O_MULTI) {
			*p++ = '|';
			*p++ = 'M';
		}

		if (ftable[i].fs_tflags & FLOCK) {
			*p++ = '|';
			*p++ = 'L';
		} else if (ftable[i].fs_tflags & FACTIVE) {
			*p++ = '|';
			*p++ = 'A';
		} else {
			*p++ = '|';
			*p++ = 'I';
		}

		*p = '\0';
		printf("%-9.9s", buf);

		if (bcubix) {
			sprintf(buf, "%d/%d", ftable[i].fs_nheard,
				ftable[i].fs_ncubix);
			printf("%-8.8s ", buf);

			printf("%-14.14s ", reqname[ftable[i].fs_req]);
		}

		else {
			printf("%-8d ", ftable[i].fs_flow);

			if (ftable[i].fs_src_node == (int4) NOTNODEID) {
				printf("none           ");
			} else if (ftable[i].fs_src_node == (int4) LOCAL) {
				sprintf(buf, "local/%d",
					ftable[i].fs_src_event);
				printf("%-14.14s ", buf);
			} else {
				sprintf(buf, "n%d/%d", ftable[i].fs_src_node,
					ftable[i].fs_src_event);
				printf("%-14.14s ", buf);
			}
		}

		printf("%-39.39s\n", ftable[i].fs_name);
	}

	if (count > fsmax) {
		printf("...and more file descriptors...\n");
	}
}

/*
 *	  header
 *
 *	  Function:	- print the fstate command table header
 */

static void
header()
{
	if (opt_taken('c') || opt_taken('C')) {
		printf("FD     FLAGS    SYNC'D   REQ            NAME\n");
	} else {
		printf("FD     FLAGS    FLOW     CLIENT         NAME\n");
	}
}

/*
 *	  help
 *
 *	  Function:	- prints helpful information on the fstate command
 */

static void
help()
{
	printf("\nSynopsis:	fstate [options] <nodes>\n");
	printf("\nDescription:	Print information on file descriptors.\n");
	printf("\nOptions:	-h	Print this help message.\n");
	printf("\t\t-a	Display active descriptors only.\n");
	printf("\t\t-c	Display cubix descriptors only.\n");
	printf("\t\t-F <#>	Set maximum # of descriptors to print.\n");
	mnusage();
	printf("\t\th\t(local host node)\n");
	printf("\nExample:	fstate h\n");
	printf("\t\t\t\"Get file status information on the local host.\"\n");
}
