/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB
 *
 *	$Log:	load.c,v $
 * Revision 2.2  92/04/30  14:13:13  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  11:44:37  raja
 * new multicast procedure
 * 
 * Revision 2.1  91/03/20  02:52:24  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:26:10  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:53:40  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  00:49:45  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:10:25  trillium
 * Initial revision
 * 
 *
 *	Function:	- load files into nodes using the loadd server
 *
 *	Example:
 *
 *		load n0-7 data1		- Load the file "data1" onto nodes
 *					  0 through 7.
 */


#include <terror.h>		/* error values */
#include <typical.h>		/* commonly used constants */
#include <loadreq.h>		/* loadd request */
#include <net.h>		/* network IPC */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <stdio.h>


/*
 * local variables
 */
static char *_copyright = "Copyright 1989 The Ohio State University $Revision: 2.2 $";
static char *usage = "load [-hv] <nodes> <files>";

/*
 * external variables
 */
extern int	errno;		/* error variable */

/*
 * external functions
 */
extern char	*getwd();
extern char	*strcat();
extern char	*strcpy();
extern char	*rindex();
extern int	predef();
extern int	stat();

/*
 * local functions
 */
static void	help();


main(argc, argv)

int		argc;    
char		*argv[];

{
	char		fullpath[FNAMELEN];	/* absolute pathname buffer */
	char		*shortname;		/* short file name */
	char		*addr;			/* target address */
	int		i;			/* favourite index variable */
	int		nodeid;			/* destination node */
	int		index;			/* parsed node index */
	int		flags;			/* parsed node flags */
	struct stat	status;			/* file status */

/*
 * Parse the command line.
 */
        validopts("hv");
 
        if ((do_args(&argc, argv) == ERROR) ||
                (parse_nodes(&argc, argv) == ERROR)) {
 
		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n", usage);
		}
 
		else {
			terror("load (arguments)");
		}
 
		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}

	getnode(&index, &nodeid, &flags);
	++argv;
	--argc;

	if ((index == ERROR) || (argc == 0)) {
		fprintf(stderr, "usage: %s\n", usage);
		exit(EUSAGE);
	}
/*
 * Make sure every filename exists.
 */
	for (i = 0; i < argc; ++i) {

		if (stat(argv[i], &status)) {
			perror(argv[i]);
			exit(errno);
		}
	}
/*
 * Attach to kernel.
 */
	if (kinit(0)) {
		terror("load (kinit)");
		exit(errno);
	}
/*
 * Load every file onto every node.
 */
	for (i = 0; i < argc; ++i) {
/*
 * Register HOST2ITB castid (predef) if required.
 * HOST2OTB and HOST2ALL are not allowed here.
 */
		if ((nodeid == HOST2OTB) || (nodeid == HOST2ALL)) {
			errno = EBADNODE;
			terror("load");
			kexit(errno);
		}
		else if (predef(nodeid)) {
			terror("load (predef)");
			kexit(errno);
		}
/*
 * Make filename absolute.
 */
		if (argv[i][0] != '/') {
			(void) getwd(fullpath);
			(void) strcat(fullpath, "/");
			(void) strcat(fullpath, argv[i]);
		}

		else {
			(void) strcpy(fullpath, argv[i]);
		}

		shortname = rindex(fullpath, '/') + 1;

		do {
			if (rload(LOCAL, nodeid, getpid(), fullpath) == ERROR) {
				terror("load (rload)");
				kexit(errno);
			}

			if (opt_taken('v')) {
				printf("%s loaded OK on ", shortname);

				if (nodeid == HOST2ITB) {
				    printf("all compute nodes with tag %d\n",
						getpid());
				}

				else {
				    printf("node %X with tag %d\n",
						nodeid, getpid());
				}
			}

			getnode(&index, &nodeid, &flags);
		} while (index);
	}

	kexit(0);
}

 
/*
 *        help
 *
 *        Function:     - prints helpful information on the load command
 */
 
static void
help()
{
        printf("\nSynopsis:       load [options] <nodes> <files>\n");
        printf("\nDescription:    Load files onto Trollius compute nodes.\n");
        printf("\nOptions:        -h	this help message\n");
        printf("\t\t-v      turn on verbose mode\n");
        mnusage();
        printf("\t\tC\t(all compute nodes)\n");
        printf("\nRestriction:    Only compute nodes may be loaded.\n");
        printf("\nExample:        load n0-7 data\n");
        printf("\t\t\t\"Load the file `data' on nodes 0 thru 7.\"\n\n");
}
