/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	loadgo.c,v $
 * Revision 2.2  92/04/30  14:13:36  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/27  11:45:29  raja
 * new multicast procedure
 * 
 * Revision 2.1.1.1  92/02/03  20:46:40  gdburns
 * Add trace support.
 * 
 * Revision 2.1  91/03/20  02:52:26  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/03/19  23:49:51  gdburns
 * Upgraded for compatibility with the new filed.
 * 
 * Revision 2.0.1.2  90/07/13  19:26:14  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/20  12:34:21  gdburns
 * Gave proper permission bits to file created by -o.
 * 
 * Revision 2.0  90/02/20  11:53:42  trillium
 * Homo erectus
 * 
 *	Function:	- load files into nodes using loadd
 *			- load runtime arguments into nodes using flatd
 *			- creates loaded files as processes
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 */

#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/param.h>

#include <boot.h>
#include <terror.h>		/* error values */
#include <typical.h>		/* commonly used constants */
#include <kio.h>
#include <loadreq.h>		/* loadd request */
#include <preq.h>		/* kenyad request */
#include <net.h>		/* network IPC */
#include <freq.h>

/*
 * local variables
 */
static char		cwd[MAXPATHLEN];
static char		*usage =
	"loadgo [-fhtvwy] [-s <stksize>] <nodes> <files>";

/*
 * external variables
 */
extern int		errno;		/* error variable */
extern int		argtailc;	/* # of arguments to processes */
extern char		**argtailv;	/* ptrs to arguments */
extern struct kio_t	_kio;
extern struct fclient	_ufd[FUMAX];	/* user file desc. */

/*
 * external functions
 */
extern char		*getwd();
extern char		*strcat();
extern char		*strcpy();
extern char		*rindex();
extern char		**pack_argv();
extern char		*ttyname();
extern char		*getparam();
extern char		*findpath();
extern void		getnode();
extern int		predef();
extern int		rloadgo();
extern int		rpwait();
extern int		rfopen();
extern int		rfincr();

/*
 * local functions
 */
static void		help();
static int		pwait();

/*
 * local variables
 */
static int		fd[3] = {-1, -1, -1};

main(argc, argv)

int			argc;    
char			*argv[];

{
	char	       *shortname;		/* short file name */
	char	      **argpackv;		/* pack argument structure */
	char	      **dirv;			/* search directories */
	char	       *temp;			/* brief temporary */
	int		argpackc;		/* number of packed arguments */
	int		argpackn;		/* size of packed arguments */
	int		i;			/* favourite index variable */
	int		nodeid;			/* destination node */
	int		index;			/* parsed node index */
	int		nflags;			/* parsed node flags */
	int		pflags;			/* process flags */
	int		stksize;		/* stack size */
	int		pid;			/* created process ID */
	int		px;			/* created process index */
	int		nproc;			/* # processes to create */
	int		ncreate;		/* # procs created */
	int		ndecr;			/* # to decrement file count */
	int		nitb;			/* # ITB nodes */
	int		nloadgo;		/* # distinct loadgos */
	int		dirc;			/* # search directories */

/*
 * Parse the command line.
 */
        validopts("fhtvwy");
        followed("sio");
 
        if ((do_args(&argc, argv) < 0) ||
                (parse_nodes(&argc, argv) < 0)) {
 
		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s\n", usage);
		} else {
			terror("loadgo (arguments)");
		}
 
		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Establish stack size.
 */
	if (opt_taken('s')) {
		intparam('s', &stksize);
	} else {
		stksize = PSTKSIZE;
	}

	pflags = opt_taken('w') ? 0 : RTF_NOWAIT;

	if (opt_taken('y')) {
		pflags |= RTF_ONCHIP;
	}

	if (opt_taken('t')) {
		pflags |= RTF_TRACE;
	}

	getnode(&index, &nodeid, &nflags);
	++argv;
	--argc;

	if ((index < 0) || (argc == 0)) {
		fprintf(stderr, "usage: %s\n", usage);
		exit(EUSAGE);
	}
/*
 * Find every file.
 */
	dirc = 0;
	dirv = (char **) 0;
	(void) addarg(&dirc, &dirv, getwd(cwd));
	(void) addarg(&dirc, &dirv, "$TROLLIUSHOME/itb/bin");
	(void) addarg(&dirc, &dirv, DEFPPROGS);

	for (i = 0; i < argc; ++i) {

		temp = findpath(argv[i], dirv, R_OK | X_OK);

		if (! temp) {
			perror(argv[i]);
			exit(errno);
		} else {
			argv[i] = temp;
		}
	}
/*
 * Attach to kernel.
 */
	if (kinit(0)) {
		terror("loadgo (kinit)");
		exit(errno);
	}
/*
 * Open stdin, stdout, and stderr for the Trollius remote filesystem.
 * Rfincr may be necessary so that descriptors do not close when loadgo exits.
 */
	if (opt_taken('i')) {
		fd[0] = rfopen(getparam('i'), O_RDONLY, 0);

		if (fd[0] < 0) {
			terror("loadgo (rfopen)");
			kexit(errno);
		}

		_kio.ki_stdin = _ufd[fd[0]].fu_tfd;
	} else if (isatty(0) && !opt_taken('f')) {
		fd[0] = rfopen(ttyname(0), O_RDONLY | O_REUSE | O_NOCBX, 0);

		if (fd[0] < 0) {
			terror("loadgo (rfopen)");
			kexit(errno);
		}

		_kio.ki_stdin = _ufd[fd[0]].fu_tfd;
	}

	if (opt_taken('o')) {
		fd[1] = rfopen(getparam('o'),
			O_WRONLY | O_CREAT | O_TRUNC, 0644);

		if (fd[1] < 0) {
			terror("loadgo (rfopen)");
			kexit(errno);
		}

		_kio.ki_stdout = _ufd[fd[1]].fu_tfd;
	} else if (isatty(1) && !opt_taken('f')) {
		fd[1] = rfopen(ttyname(1), O_WRONLY | O_REUSE | O_NOCBX, 0);

		if (fd[1] < 0) {
			terror("loadgo (rfopen)");
			kexit(errno);
		}

		_kio.ki_stdout = _ufd[fd[1]].fu_tfd;
	}

	if (isatty(2) && !opt_taken('f')) {
		fd[2] = rfopen(ttyname(2), O_RDWR | O_REUSE | O_NOCBX, 0);

		if (fd[2] < 0) {
			terror("loadgo (rfopen)");
			kexit(errno);
		}

		_kio.ki_stderr = _ufd[fd[2]].fu_tfd;
	}
/*
 * Count the total number of new processes.
 */
	nitb = getnitb();
	nproc = 0;

	do {
		if ((nodeid == HOST2ALL) || (nodeid == HOST2OTB)) {
			errno = EBADNODE;
			terror("loadgo");
			kexit(errno);
		} else if (nodeid == HOST2ITB) {
			nproc += nitb;
/*
 * Register the HOST2ITB castid (predef) if required.
 */
			if (predef(nodeid)) {
				terror("loadgo (predef)");
				kexit(errno);
			}
		} else {
			nproc += 1;
		}

		getnode(&index, &nodeid, &nflags);
	} while (index);

	nproc *= argc;
/*
 * Set file descriptor count.
 */
	if (nproc > 1) {

		if (_kio.ki_stdin > 2) {
			(void) rfincr(fd[0], nproc - 1);
		}

		if (_kio.ki_stdout > 2) {
			(void) rfincr(fd[1], nproc - 1);
		}

		if (_kio.ki_stderr > 2) {
			(void) rfincr(fd[2], nproc - 1);
		}
	}
/*
 * Load every file onto every node.
 */
	ncreate = 0;
	nloadgo = 0;

	for (i = 0; i < argc; ++i) {
/*
 * Every new process gets the same arguments.
 */
		shortname = rindex(argv[i], '/');
		shortname = (shortname) ? shortname + 1 : argv[i];

		if (argtailc) {
			*(argtailv - 1) = shortname;
			argpackc = argtailc + 1;
			argpackv = pack_argv(argpackc, argtailv - 1, &argpackn);
		} else {
			argpackc = 1;
			argpackv = pack_argv(1, &shortname, &argpackn);
		}

		do {
/*
 * Remotely load and execute the current program on the current node.
 */
			if ((px = rloadgo(LOCAL, nodeid, stksize, pflags,
				argv[i], argpackc, argpackv, argpackn,
				&pid)) < 0) {

				terror("loadgo (rloadgo)");
/*
 * Uh, what about those open file descriptors?
 */
				ndecr = nproc - ncreate;

				if (ndecr > 1) {
					ndecr -= 1;

					if (_kio.ki_stdin > 2) {
						(void) rfincr(fd[0], -ndecr);
					}

					if (_kio.ki_stdout > 2) {
						(void) rfincr(fd[1], -ndecr);
					}

					if (_kio.ki_stderr > 2) {
						(void) rfincr(fd[2], -ndecr);
					}
				}

				if (pwait(nloadgo) < 0) {
					terror("loadgo (rpwait)");
				}

				kexit(errno);
			} else {
				ncreate += (nodeid == HOST2ITB) ? nitb : 1;
				nloadgo += 1;
			}

			if (opt_taken('v')) {
				printf("[%d] %d %s running on ",
					px, pid, shortname);

				if (nodeid == HOST2ITB) {
					printf("all compute nodes\n");
				} else {
					printf("node %x\n", nodeid);
				}
			}

			getnode(&index, &nodeid, &nflags);

		} while (index);
	}
/*
 * Wait for the process(es) to terminate.
 */
	if (pwait(nloadgo) < 0) {
		terror("loadgo (rpwait)");
	}

	_kexit(0);
}

/*
 *        help
 *
 *        Function:     - prints helpful information on the loadgo command
 */
 
static void
help()
{
    printf("\nSynopsis:\tloadgo [options] <nodes> <files> [-- <arguments>]\n");
    printf("\nDescription:\tExecute programs on Trollius ITB nodes.\n");
    printf("\nOptions:        -h\t\tthis help message\n");
    printf("\t\t-v\t\tturn on verbose mode\n");
    printf("\t\t-f\t\tdo not open file descriptors\n");
    printf("\t\t-t\t\tenable trace collection\n");
    printf("\t\t-w\t\twait for process completion\n");
    printf("\t\t-y\t\tallocate stack from on-chip RAM\n");
    printf("\t\t-s <size>\tset stack size\n");
    printf("\t\t-i <filename>\tredirect standard input\n");
    printf("\t\t-o <filename>\tredirect standard output\n");
    mnusage();
    printf("\t\tC\t(all compute nodes)\n");
    printf("\nRestriction:\tUse loadgo only with ITB nodes.\n");
    printf("\nExample:        loadgo n0-7 prog1\n");
    printf("\t\t\t\"Execute `prog1' on nodes 0 thru 7.\"\n\n");
}

/*
 *	pwait
 *
 *	Function:	- if requested, waits for processes to complete
 *	Accepts:	- number of processes to wait
 *	Returns:	- 0 or ERROR
 */

static int
pwait(nwait)

int			nwait;

{
	int		nodeid;		/* child's node ID */
	int		pid;		/* child's process ID */
	int		pstatus;	/* child's exit status */

	if (opt_taken('w')) {

		while (nwait > 0) {

			if (rpwait(&nodeid, &pid, &pstatus) < 0) {
				return(ERROR);
			}

			if (pstatus) {
				printf("%d exited with status %d\n",
					pid, pstatus);
			}

			--nwait;
		}
	}

	return(0);
}
