/*
 *	OSU Research Computing, architects and engineers	Trillium
 *
 *	Copyright 1989 The Ohio State University
 *
 *	GDB
 *
 *	$Log:	fcubix.c,v $
 * Revision 2.2  92/04/30  14:18:25  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/02/24  20:59:33  gdburns
 * Ohio Release
 * 
 *
 *	Function:	- support routines for cubix access method
 */

#include <errno.h>
#include "filed.h"
#include <freq.h>
#include <net.h>
#include <portable.h>
#include <stdio.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;

/*
 * external functions
 */
extern char		*malloc();
extern void		bcopy();
extern struct fcast	*fcfind();	/* find a cast entry */
extern struct fcast	*fcdef();	/* check for a predefined cast entry */

/*
 * global functions
 */
int			cxhear();	/* hear from a cubix cast member */
void			cxreset();	/* reset cubix cast to "not heard" */
struct cxnode		*cxnew();	/* allocate and init a cubix cast */
struct cxnode		*cxfind();	/* find a cubix member */

/*
 *	cxnew
 *
 *	Function:	- establish a cubix structure from a known file cast
 *			- allocates and initializes the cubix structure
 *			  using the members of the file cast
 *
 *	Accepts:	- file cast identifier
 *			- ptr to number of cast members (value returned)
 *
 *	Returns:	- ptr to cubix array structure or NULL ptr
 */

struct cxnode *
cxnew(id, num)

int4			id;
int4			*num;

{
	struct fcast	*pfc;		/* ptr cast descriptor */
	struct cxnode	*pcx;		/* new cubix array */
	int4		i;		/* favourite counter */

/*
 * Find the file cast entry.
 */
	pfc = fcfind(id);

	if (pfc == FCNULL) {
/*
 * Check for predefined file casts.
 */
		pfc = fcdef(id);

		if (pfc == FCNULL) {
			errno = EINVAL;
			return(CXNULL);
		}
	}
/*
 * Allocate and initialize new cubix array structure.
 */
	pcx = (struct cxnode *) malloc((unsigned int) (pfc->fc_num *
		sizeof(struct cxnode)));

	if (pcx == CXNULL) {
		fprintf(stderr, "filed: can't allocate cubix array\n");
		perror("filed");
		kexit(errno);
	}

	for (i = 0; i < pfc->fc_num; ++i) {
		pcx[i].cx_src_node = pfc->fc_members[i];
		pcx[i].cx_heard = FALSE;
	}

	*num = pfc->fc_num;
	return(pcx);
}

/*
 *	cxhear
 *
 *	Function:	- "hear" from a cubix cast member
 *			- If there are n cast members, the first n - 1
 *			  requests on a cubix file desc just record pertinent
 *			  information.
 *			- The nth request causes the file operation to be taken.
 *
 *	Accepts:	- ptr to filed descriptor
 *			- ptr to filed request
 *
 *	Returns:	- ERROR
 *			- 0 for requests 1 thru n - 1
 *			- 1 for request n
 */

int
cxhear(pfd, request)

struct fdesc		*pfd;
struct freq		*request;

{
	int4		i;		/* favourite index */
	int4		ncx;		/* favourite count */
	struct cxnode	*pcx;		/* favourite ptr */

/*
 * Check for proper request code synchronization.
 */
	if (((pfd->f_req != request->fq_req) && (pfd->f_req != FQIDLE)) ||
		((request->fq_req == FQOPEN) && (pfd->f_req == FQIDLE))) {
		errno = ECXREQ;
		return(ERROR);
	}
/*
 * Note the request and find the cubix cast member.
 */
	pfd->f_req = request->fq_req;
	ncx = pfd->f_ncubix;
	pcx = cxfind(request->fq_src_node, pfd);

	if ((pcx == CXNULL) || (pcx->cx_heard == TRUE)) {
		errno = ECXMEMBER;
		return(ERROR);
	}
/*
 * Note that the member has heard the request.
 */
	pcx->cx_heard = TRUE;
	pcx->cx_src_event = request->fq_src_event;
	pcx->cx_param = request->fq_param1;
	pfd->f_nheard += 1;
/*
 * If all members have heard the request, reset the cubix portion
 * of the file descriptor.
 */
	if (pfd->f_nheard == pfd->f_ncubix) {
		pfd->f_req = FQIDLE;
		pfd->f_nheard = 0;

		for (i = 0; i < ncx; ++i) {
			pfd->f_pcubix[i].cx_heard = FALSE;
		}

		return(1);
	} else {
		return(0);
	}
}

/*
 *	cxfind
 *
 *	Function:	- find a cubix member
 *
 *	Accepts:	- node ID of cubix member
 *			- ptr to filed descriptor
 *
 *	Returns:	- ptr to cubix member descriptor
 */

struct cxnode *
cxfind(node, pfd)

int4			node;
struct fdesc		*pfd;

{
	int4		i;		/* favourite index */

	for (i = 0; (i < pfd->f_ncubix) &&
		(pfd->f_pcubix[i].cx_src_node != node); ++i) {
	}

	if (i < pfd->f_ncubix) {
		return(pfd->f_pcubix + i);
	} else {
		return(CXNULL);
	}
}
