/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	fface.c,v $
 * Revision 2.2.1.1  92/06/05  15:59:40  raja
 * Use uint4 instead of "unsigned int4".
 *
 * Revision 2.2  92/04/30  14:18:30  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/02/28  10:45:09  gdburns
 * Give our best effort to stat() and fstat().
 * 
 * Revision 2.1.1.2  91/12/18  16:27:43  gdburns
 * Use NOCTTY on opens.
 * Use setsid() to get away from process groups and
 * controlling terminals.
 * 
 * Revision 2.1.1.1  91/06/05  15:11:33  gdburns
 * Pass a pointer not the whole stat structure in stat() and fstat().
 * 
 * Revision 2.1  91/02/24  20:59:37  gdburns
 * Ohio Release
 * 
 *	Function:	- abstracts native filesystem interface
 *			- Trollius remote file access follows the Unix I/O
 *			  standard assuming a 32 bit interface (int4).
 *			- In case the standard Unix calls are not available,
 *			  they may be emulated in this module.
 *			- Unix version
 */

#include <fcntl.h>
#include "filed.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <typical.h>

/*
 * external functions
 */
extern int		open();
extern int		close();
extern int		read();
extern int		write();
extern int		chdir();
extern int		access();
extern int		stat();
extern int		fstat();
extern int		truncate();
extern int		ftruncate();
extern int		mkdir();
extern int		rmdir();
extern int		unlink();
extern int		system();
extern long		lseek();
extern char		*getwd();
extern pid_t		setsid();

/*
 * local functions
 */
static void		conv_stat();

/*
 * Unix version - boring, so just blurt them out.
 * Here we know that int4 is int.
 */

int4
myopen(path, flags, mode)

char			*path;
int4			flags;
int4			mode;

{
	int4		fd;

	fd = open(path, flags | O_NOCTTY, mode);
	return(fd);
}

int4 myclose(fd) int4 fd;
{ return(close(fd)); }

int4 myread(fd, buf, nbytes) int4 fd; char *buf; int4 nbytes;
{ return(read(fd, buf, nbytes)); }

int4 mywrite(fd, buf, nbytes) int4 fd; char *buf; int4 nbytes;
{ return(write(fd, buf, nbytes)); }

int4 mylseek(fd, offset, whence) int4 fd; int4 offset; int4 whence;
{ return(lseek(fd, (long) offset, whence)); }

int4 mychdir(newdir) char *newdir;
{ return(chdir(newdir)); }

char * mygetwd(dir) char *dir;
{ return(getwd(dir)); }

int4 myunlink(name) char *name;
{ return(unlink(name)); }

int4 myrmdir(dir) char *dir;
{ return(rmdir(dir)); }

int4 mysystem(cmd) char *cmd;
{ return(system(cmd)); }

int4 mymkdir(path, mode) char *path; int4 mode;
{ return(mkdir(path, mode)); }

int4 myaccess(path, mode) char *path; int4 mode;
{ return(access(path, mode)); }

int4 mytrunc(path, len) char *path; int4 len;
{ return(truncate(path, (off_t) len)); }

int4 myftrunc(fd, len) int4 fd; int4 len;
{ return(ftruncate(fd, (off_t) len)); }


/*
 *	myinit
 *
 *	Function:	- machine dependent initialization
 *			- in Unix case, take care of process group/
 *			  controlling terminal headaches
 *
 *	Returns:	- 0 or ERROR
 */

int
myinit()

{
/*
 * Create a new session.  This detaches our controlling terminal
 * so we don't get TTIN and TTOU signals when reading/writing the tty.
 * We also get out of the tty's process group so that we don't
 * catch special character signals from it.
 *
 * In the future, we will have worker processes to deal with ttys
 * (and all weird and/or slow devices) so we won't have this hassle.
 */
	if (setsid() < 0) {
		return(ERROR);
	}

	return(0);
}

/*
 *	mystat
 */
int4
mystat(path, buf)

char			*path;
struct stat		*buf;

{
	struct stat	status;
	int4		ret;

	ret = stat(path, &status);
	conv_stat(&status, (unsigned int *) buf);
	return(ret);
}

/*
 *	myfstat
 */
int4
myfstat(fd, buf)

int4			fd;
struct stat		*buf;

{
	struct stat	status;
	int4		ret;

	ret = fstat(fd, &status);
	conv_stat(&status, (unsigned int *) buf);
	return(ret);
}

/*
 *	conv_stat
 *
 *	Function:	- stores status structure in an int4 array
 *	Accepts:	- ptr to stat structure
 *			- ptr to int4 array
 */

static void
conv_stat(status, buf)

struct stat		*status;
uint4			*buf;

{
	buf[0] = (uint4) status->st_dev;
	buf[1] = (uint4) status->st_ino;
	buf[2] = (uint4) status->st_mode;
	buf[3] = (uint4) status->st_nlink;
	buf[4] = (uint4) status->st_uid;
	buf[5] = (uint4) status->st_gid;
	buf[6] = (uint4) status->st_rdev;
	buf[7] = (uint4) status->st_size;
	buf[8] = (uint4) status->st_atime;
	buf[9] = (uint4) status->st_spare1;
	buf[10] = (uint4) status->st_mtime;
	buf[11] = (uint4) status->st_spare2;
	buf[12] = (uint4) status->st_ctime;
	buf[13] = (uint4) status->st_spare3;
	buf[14] = (uint4) status->st_blksize;
	buf[15] = (uint4) status->st_blocks;
	buf[16] = (uint4) status->st_spare4[0];
	buf[17] = (uint4) status->st_spare4[1];
}
