/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	fface.unicos.c,v $
 * Revision 2.2  92/04/30  14:18:31  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/12/18  16:28:42  gdburns
 * initial version
 * 
 *	Function:	- abstracts native filesystem interface
 *			- Trollius remote file access follows the Unix I/O
 *			  standard assuming a 32 bit interface (int4).
 *			- In case the standard Unix calls are not available,
 *			  they may be emulated in this module.
 *			- Unicos version
 */

#include <fcntl.h>
#include "filed.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <typical.h>

/*
 * external functions
 */
/*
 * extern int		open();
 * extern int		close();
 * extern int		read();
 * extern int		write();
 * extern int		chdir();
 * extern int		access();
 * extern int		stat();
 * extern int		fstat();
 * extern int		truncate();
 * extern int		ftruncate();
 * extern int		mkdir();
 * extern int		rmdir();
 * extern int		unlink();
 * extern int		system();
 */
extern long		lseek();
extern char		*getwd();

/*
 * Here we know that int4 is short.
 */

int4
myopen(path, flags, mode)

char			*path;
int4			flags;
int4			mode;

{
	int		fd;

	fd = open(path, (int) flags, (int) mode);
	return((int4) fd);
}

int4 myclose(fd) int4 fd;
{ return((int4) close((int) fd)); }

int4 myread(fd, buf, nbytes) int4 fd; char *buf; int4 nbytes;
{ return((int4) read((int) fd, buf, (unsigned int) nbytes)); }

int4 mywrite(fd, buf, nbytes) int4 fd; char *buf; int4 nbytes;
{ return((int4) write((int) fd, buf, (unsigned int) nbytes)); }

int4 mylseek(fd, offset, whence) int4 fd; int4 offset; int4 whence;
{ return((int4) lseek((int) fd, (long) offset, (int) whence)); }

int4 mychdir(newdir) char *newdir;
{ return((int4) chdir(newdir)); }

char *mygetwd(dir) char *dir;
{ return(getwd(dir)); }

int4 myunlink(name) char *name;
{ return((int4) unlink(name)); }

int4 myrmdir(dir) char *dir;
{ return((int4) rmdir(dir)); }

int4 mysystem(cmd) char *cmd;
{ return((int4) system(cmd)); }

int4 mymkdir(path, mode) char *path; int4 mode;
{ return((int4) mkdir(path, (int) mode)); }

int4 myaccess(path, mode) char *path; int4 mode;
{ return((int4) access(path, (int) mode)); }

int4 mystat(path, buf) char *path; struct stat *buf;
{ return((int4) stat(path, buf)); }

int4 myfstat(fd, buf) int4 fd; struct stat *buf;
{ return((int4) fstat((int) fd, buf)); }

int4 myftrunc(fd, len) int4 fd; int4 len;
{ return((int4) ftruncate(fd, (long) len)); }

/*
 *	mytrunc
 *
 *	Function:	- Simulate truncate() since it does not seem
 *			  to exist in Unicos.
 *
 *	Accepts:	- filename
 *			- desired length
 *
 *	Returns:	- 0 or ERROR
 */

int4
mytrunc(path, len)

char			*path;
int4			len;

{
	int		fd;

	fd = open(path, O_WRONLY, 0);

	if (fd < 0) return(ERROR);

	if (ftruncate(fd, (long) len) < 0) {
		(void) close(fd);
		return(ERROR);
	}
	
	if (close(fd) < 0) return(ERROR);

	return(0);
}

/*
 *	myinit
 *
 *	Function:	- machine dependent initialization
 *			- in Unix case, take care of process group/
 *			  controlling terminal headaches
 *
 *	Returns:	- 0 or ERROR
 */

int
myinit()

{
/*
 * Create a new session.  This detaches our controlling terminal
 * so we don't get TTIN and TTOU signals when reading/writing the tty.
 * We also get out of the tty's process group so that we don't
 * catch special character signals from it.
 *
 * In the future, we will have worker processes to deal with ttys
 * (and all weird and/or slow devices) so we won't have this hassle.
 */
	if (setsid() < 0) {
		return(ERROR);
	}

	return(0);
}
