/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	filed.c,v $
 * Revision 2.2  92/04/30  14:18:33  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/06  10:56:15  gdburns
 * Add flags capability to sendreply().
 * Remove time stamp from internal file descriptor.
 * 
 * Revision 2.1.1.1  91/10/09  15:43:26  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1  91/02/24  20:59:39  gdburns
 * Ohio Release
 * 
 *	Function:	- filed mainline 
 *			- remote daemon
 *			- a lean, mean fighting machine
 *			- You, too, can maintain this program.
 *
 *			- Filed is organized into four parts:
 *			  1) filed - the request/reply mainline
 *			  2) fq* - the guts of the requests
 *			  3) f* - data structure modules
 *			  4) fface - the native file interface
 *
 *			- Filed provides the standard Unix I/O
 *			  operations as a remote service to Trollius
 *			  clients.  Additional requests, specific to
 *			  Trollius, control and monitor filed functions.
 *			  Filed also provides a special parallel access
 *			  method, called Cubix, in addition to the typical
 *			  single client Unix behaviour.  Only the five
 *			  "biggie" functions: open, close, read, write and
 *			  lseek, work on cubix method descriptors.
 *
 *			- There are actually three types of file descriptors
 *			  in the filed world.  First, there is the user file
 *			  descriptor, ufd, which resides wholly in the client
 *			  and has the look and feel of a Unix file descriptor.
 *			  The ufd has a field which contains the Trollius
 *			  file descriptor handle.  It is passed to filed
 *			  from the client, so that filed can identify the
 *			  target Trollius file descritor, tfd.  Before an
 *			  operation is carried out on a file, the tfd must
 *			  have an associated native OS (eg, Unix) file
 *			  descriptor open, and this is referred to as the fd.
 *			  Thus, an active file has an associated open fd,
 *			  while an inactive file does not.
 *
 *			- The word handle means the identifying number of
 *			  the descriptor while the word descriptors refers
 *			  to the whole data structure associated with an open
 *			  file.
 *
 *			- Filed does not ensure that the client that called for
 *			  a file open is the same client that calls for later
 *			  operations.  Since the client does not manage the
 *			  tfd handle (he only sees the ufd), this should not
 *			  be a problem.
 */

#include <events.h>
#include "filed.h"
#include <freq.h>
#include <net.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */

/*
 * external functions
 */
extern void		do_open();
extern void		do_close();
extern void		do_read();
extern void		do_write();
extern void		do_lseek();
extern void		do_cxwrite();
extern void		do_state();
extern void		do_order();
extern void		do_chcx();
extern void		do_chdir();
extern void		do_getwd();
extern void		do_rmfd();
extern void		do_access();
extern void		do_stat();
extern void		do_fstat();
extern void		do_trunc();
extern void		do_ftrunc();
extern void		do_dup();
extern void		do_mkdir();
extern void		do_system();
extern void		do_rmdir();
extern void		do_unlink();
extern void		do_incr();
extern void		do_f77read();
extern void		do_f77back();
extern void		sendreply();	/* to client */
extern void		terror();	/* report on errno */
extern void		fdstdio();	/* set up 0, 1 and 2 */
extern void		kexit();	/* detach and exit */
extern int		kinit();	/* initialize and attach to kernel */
extern int		myinit();	/* machine dependent initialization */

/*
 * local variables
 */
static void		(*(do_freq[FQMAX]))() = {
			do_open,
			do_close,
			do_read,
			do_write,
			do_lseek,
			do_cxwrite,
			do_chcx,
			do_order,
			do_state,
			do_rmfd,
			do_incr,
			do_access,
			do_mkdir,
			do_unlink,
			do_system,
			do_chdir,
			do_rmdir,
			do_getwd,
			do_stat,
			do_fstat,
			do_dup,
			do_trunc,
			do_ftrunc,
			do_f77read,
			do_f77back,
			};
/*
 * global variables
 */
char			buf[MAXNMSGLEN];/* data for certain reqs */

main()

{
	struct freq	*request;	/* filed request details */
	struct nmsg	incoming;	/* incoming filed request */

	if (myinit() < 0) {
		terror("filed (myinit)");
		exit(errno);
	}
/*
 * Get acquainted with the Trollius kernel.
 */
	if (kinit(0) < 0) {
		terror("filed (kinit)");
		exit(errno);
	}
/*
 * Set up standard I/O descriptors.
 */
	fdstdio();
/*
 * Set up request message descriptor.
 */
	incoming.nh_event = EVFILED;
	incoming.nh_flags = 0;
	incoming.nh_msg = buf;
	request = (struct freq *) incoming.nh_data;

	for (;;) {
/*
 * Receive a request.
 */
		incoming.nh_type = 0;
		incoming.nh_length = sizeof(buf);

		if (nrecv(&incoming) < 0) {
			terror("filed (nrecv)");
			kexit(errno);
		}
/*
 * Process the request.
 * Individual request functions handle replies, if any.
 */
		if ((request->fq_req >= FQMAX) ||
			(request->fq_req < 0)) {

			sendreply(request->fq_src_node,
				request->fq_src_event,
				EBADREQUEST, (int4) ERROR, INT4_NIL, INT4_NIL);
		}

		else {
			(*(do_freq[request->fq_req]))(request);
		}
	}
}

/*
 *	sendreply
 *
 *	Function:	- send reply back to client
 *			- copies parameters into the outgoing (reply) message
 *			  descriptor and calls nsend()
 *
 *	Accepts:	- client node
 *			- client event
 *			- reply error number (will be set as client errno)
 *			- reply return value (will be return to client caller)
 *			- length of buffer to return
 *			- message flags
 */

void
sendreply(node, event, err, ret, length, flags)

int4			node;
int4			event;
int			err;
int4			ret;
int4			length;
int4			flags;

{
	struct nmsg	outgoing;	/* outgoing filed reply */
	struct freply	*reply;		/* filed reply details */

	outgoing.nh_node = node;
	outgoing.nh_event = event;
	outgoing.nh_type = 0;
	outgoing.nh_flags = flags;
	outgoing.nh_length = length;
	outgoing.nh_msg = buf;

	reply = (struct freply *) outgoing.nh_data;
	reply->fr_errno = (int4) err;
	reply->fr_ret = ret;

	if (nsend(&outgoing) < 0) {
		terror("filed (sendreply)");
		kexit(errno);
	}
}
