/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	fqclose.c,v $
 * Revision 2.2  92/04/30  14:18:36  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/06  10:56:59  gdburns
 * Add flags capability to sendreply().
 * Remove time stamp from internal file descriptor.
 * 
 * Revision 2.1.1.1  91/10/09  15:44:04  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1  91/02/24  20:59:40  gdburns
 * Ohio Release
 * 
 *	Function:	- FQCLOSE request
 *			- three possibilites are:
 *			  1) Cubix file - use cxhear() and reply to everybody
 *			  2) reused file - decrement the count and reply
 *			  3) simple file - close it and reply
 */

#include "filed.h"
#include <freq.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */

/*
 * external functions
 */
extern struct fdesc	*fdfind();	/* find a filed descriptor */
extern void		fdcut();	/* free filed descriptor */
extern void		sendreply();	/* general purpose file reply */
extern int		cxhear();	/* note sync of one Cubix node */
extern int4		myclose();	/* portable close interface */

/*
 *	do_close
 *
 *	Function:	- handles simple client close
 *			- handles synchronized Cubix close
 *
 *	Accepts:	- ptr to filed request
 *			- ptr to filed reply
 */

void
do_close(request)

struct freq		*request;

{
	struct fdesc	*f;		/* ptr to target descriptor */
	int		status;		/* Cubix status */
	int4		ret;		/* result of close() */
	int4		i;		/* favourite index */
	int		err;		/* errno */

	f = fdfind(request->fq_tfd);

	if (f == FDNULL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, (int4) ERROR, INT4_NIL, INT4_NIL);
	}
/*
 * Cubix file
 */
	else if (f->f_tflags & FCUBIX) {
		status = cxhear(f, request);

		if (status == ERROR) {
			sendreply(request->fq_src_node, request->fq_src_event,
				errno, (int4) ERROR, INT4_NIL, INT4_NIL);
		}

		else if (status == 0) {
		}

		else {
			ret = (f->f_tflags & FACTIVE) ? myclose(f->f_fd) : 0;
			err = (ret == 0) ? 0 : errno;

			for (i = 0; i < f->f_ncubix; ++i) {
				sendreply(f->f_pcubix[i].cx_src_node,
					f->f_pcubix[i].cx_src_event,
					err, ret, INT4_NIL, INT4_NIL);
			}

			fdcut(f);
		}
	}
/*
 * reused file
 */
	else if (f->f_count > 1) {
		f->f_count -= 1;
		f->f_src_node = request->fq_src_node;
		f->f_src_event = request->fq_src_event;
		sendreply(request->fq_src_node, request->fq_src_event,
			0, INT4_NIL, INT4_NIL, INT4_NIL);
	}
/*
 * plain file
 */
	else {
		if (f->f_tfd <= 2) {
			ret = 0;
		} else {
			ret = (f->f_tflags & FACTIVE) ? myclose(f->f_fd) : 0;
		}

		if (ret == 0) errno = 0;
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, ret, INT4_NIL, INT4_NIL);
		fdcut(f);
	}
}
