/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	fqcubix.c,v $
 * Revision 2.2  92/04/30  14:18:37  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/02/06  10:57:02  gdburns
 * Add flags capability to sendreply().
 * Remove time stamp from internal file descriptor.
 * 
 * Revision 2.1.1.2  91/10/09  15:44:12  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1.1.1  91/06/11  17:35:54  gdburns
 * Keep a better record of clients in chcx.
 * 
 * Revision 2.1  91/02/24  20:59:42  gdburns
 * Ohio Release
 * 
 *	Function:	- Cubix specific requests
 */

#include <errno.h>
#include <events.h>
#include "filed.h"
#include <freq.h>
#include <net.h>
#include <portable.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */
extern char		buf[];		/* input data from read */

/*
 * external functions
 */
struct fdesc		*fdfind();	/* find a filed descriptor by tfd */
struct fdesc		*fdfindname();	/* find a filed descriptor by name */
struct fdesc		*fdinit();	/* initialize new file descriptor */
struct cxnode		*cxnew();	/* initialize new Cubix descriptor */
void			sendreply();	/* general purpose file reply */
void			fdcut();	/* clean and free file descriptor */
int			cxhear();	/* note sync of one Cubix node */
int4			myclose();	/* close() */
char			*strcpy();

/*
 *	do_order
 *
 *	Function:	- reorder sequence of Cubix cast members
 *			- determines the order in which data is read
 *			  or written from among the members
 *			- only relevant for multi mode
 *
 *	Accepts:	- ptr to filed request
 */

void
do_order(request)

struct freq		*request;

{
	struct fdesc	*f;		/* ptr to target file descriptor */
	struct cxnode	*pcx;		/* ptr to Cubix descriptor */
	int4		i;		/* favourite index */
	int		status;		/* Cubix status */

	f = fdfind(request->fq_tfd);

	if (f == FDNULL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, (int4) ERROR, INT4_NIL, INT4_NIL);
		return;
	}

	f->f_src_node = request->fq_src_node;
	f->f_src_event = request->fq_src_event;
/*
 * This has to be a Cubix method descriptor.
 */
	if (! (f->f_tflags & FCUBIX)) {
		sendreply(request->fq_src_node, request->fq_src_event,
			EINVAL, (int4) ERROR, INT4_NIL, INT4_NIL);
		return;
	}

	status = cxhear(f, request);

	if (status == ERROR) {
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, (int4) ERROR, INT4_NIL, INT4_NIL);
	}

	else if (status == 0) {
	}
/*
 * Juggle the order.
 */
	else {
		pcx = f->f_pcubix;

		for (i = 0; i < f->f_ncubix; ++i) {
			pcx[pcx[i].cx_seq].cx_new_node = pcx[i].cx_src_node;
			pcx[pcx[i].cx_seq].cx_new_event = pcx[i].cx_src_event;
		}

		for (i = 0; i < f->f_ncubix; ++i) {
			pcx[i].cx_src_node = pcx[i].cx_new_node;
			pcx[i].cx_src_event = pcx[i].cx_new_event;

			sendreply(pcx[i].cx_src_node, pcx[i].cx_src_event,
					0, INT4_NIL, INT4_NIL, INT4_NIL);
		}
	}
}

/*
 *	do_chcx
 *
 *	Function:	- change access method to/from/within Cubix
 *			- most applicable for switching been singl
 *			  and multi modes, eg, when ordered, distributed
 *			  data is followed by global, singular values, or
 *			  vice versa
 *
 *	Accepts:	- ptr to filed request
 */

void
do_chcx(request)

struct freq		*request;

{
	struct fdesc	*f;		/* ptr to old target descriptor */
	struct fdesc	*x;		/* ptr to Cubix target descriptor */
	int4		i;		/* favourite index */
	int		status;		/* Cubix status */
	int4		oldflags;	/* previous access method */
	int4		newflags;	/* subsequent access method */

	f = fdfind(request->fq_tfd);

	if (f == FDNULL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, (int4) ERROR, INT4_NIL, INT4_NIL);
		return;
	}

	f->f_src_node = request->fq_src_node;
	f->f_src_event = request->fq_src_event;
	oldflags = f->f_tflags;
	newflags = (oldflags & ~FMETHOD) | (request->fq_newflags & FMETHOD);
/*
 * Establish a ptr to a Cubix method file descriptor, either by finding
 * the old one or creating a new one.  Except for the trivial case
 * of async -> async, this is where the synchronization must occur.
 */
	if (oldflags & FCUBIX) {
		x = f;
	} else if (newflags & FCUBIX) {
		request->fq_param1 = (int4) f;
		x = fdfindname(f->f_name, newflags & ~O_LOCK);

		if (x == FDNULL) {
			newflags &= ~FACTIVE;
			strcpy(buf, f->f_name);
			x = fdinit(newflags);
			x->f_pcubix = cxnew(HOST2ITB, &(x->f_ncubix));
			x->f_req = FQCHCX;
			x->f_nheard = 0;
			x->f_ptr = f->f_ptr;
			x->f_flow = f->f_flow;
		}

		x->f_src_node = request->fq_src_node;
		x->f_src_event = request->fq_src_event;
	} else {
/*
 * File in async mode gets changed to async mode - waste of my time.
 */
		sendreply(request->fq_src_node, request->fq_src_event,
			0, f->f_tfd, INT4_NIL, INT4_NIL);
		return;
	}
		
	status = cxhear(x, request);

	if (status == ERROR) {
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, (int4) ERROR, INT4_NIL, INT4_NIL);
	}

	else if (status == 0) {
	}
/*
 * Switch access methods. 
 */
	else {

		if (oldflags & FCUBIX) {
/*
 * Simply change from one Cubix method to another.
 */
			if (newflags & FCUBIX) {
				x->f_tflags = newflags;

				for (i = 0; i < x->f_ncubix; ++i) {
					sendreply(x->f_pcubix[i].cx_src_node,
						x->f_pcubix[i].cx_src_event,
						0, x->f_tfd, INT4_NIL,
						INT4_NIL);
				}
			}
/*
 * Convert Cubix method descriptor to async method descriptor.
 * Set f_count to reflect the number of Cubix members.
 */
			else {
				x->f_tflags = newflags;
				x->f_count = x->f_ncubix;

				for (i = 0; i < x->f_ncubix; ++i) {
					sendreply(x->f_pcubix[i].cx_src_node,
						x->f_pcubix[i].cx_src_event,
						0, f->f_tfd, INT4_NIL,
						INT4_NIL);
				}
			}
		}
/*
 * Collapse n async method descriptors into 1 Cubix method descriptor.
 * Clean up and release old async method descriptors.
 */
		else {
			for (i = 0; i < x->f_ncubix; ++i) {
				f = (struct fdesc *) x->f_pcubix[i].cx_pfd;

				sendreply(x->f_pcubix[i].cx_src_node,
					x->f_pcubix[i].cx_src_event,
					0, x->f_tfd, INT4_NIL, INT4_NIL);

				if (f->f_tfd > 2) {
					f->f_count -= 1;

					if (f->f_count <= 0) {

						if (f->f_tflags & FACTIVE) {
							(void) myclose(f->f_fd);
						}

						fdcut(f);
					}
				}
			}
		}
	}
}
