/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	fqf77.c,v $
 * Revision 2.2  92/04/30  14:18:39  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.4  92/02/06  10:57:05  gdburns
 * Add flags capability to sendreply().
 * Remove time stamp from internal file descriptor.
 * 
 * Revision 2.1.1.3  91/11/19  17:48:46  gdburns
 * Malloc()'s argument is unsigned int, not int4.
 * 
 * Revision 2.1.1.2  91/10/27  20:53:58  gdburns
 * Change lseek to Sys V.
 * 
 * Revision 2.1.1.1  91/10/09  15:44:16  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1  91/02/24  20:59:44  gdburns
 * Ohio Release
 * 
 *	Function:	- F77 specific requests
 *			- works on files assumed to be in an F77 record format
 *			- generally does not detect bad record format, but
 *			  will try not to die solely due to bad record format
 *			- much more efficient to do these operations here
 *			  than on a remote client
 *			- does not handle direct access files, sequential
 *			  access only
 *			- unformatted files end records with newline ('\n')
 *			- formatted files head and trail record with length
 */

#include "filed.h"
#include <freq.h>
#include <net.h>
#include <portable.h>
#include <sys/types.h>
#include <terror.h>
#include <typical.h>
#include <unistd.h>

#undef	NULL
#define NULL		((char *) 0)
#define NCARD		(80)

/*
 * external variables
 */
extern int		errno;		/* global error variable */
extern char		buf[];		/* input data from read */

/*
 * external functions
 */
struct fdesc		*fdfind();	/* find a filed descriptor by tfd */
void			fdactivate();	/* ensure fd is open */
void			sendreply();	/* general purpose file reply */
int4			myread();	/* read() */
int4			mylseek();	/* lseek() */
char			*malloc();

/*
 *	do_f77back
 *
 *	Function:	- moves file ptr to start of previous record
 *			- record format based on file format, either
 *			  "formatted" or "unformatted"
 *			- assumes we have a sequential F77 file
 *
 *	Accepts:	- ptr to filed request
 */

void
do_f77back(request)

struct freq		*request;

{
	struct fdesc	*f;		/* ptr to target file descriptor */
	int4		nback;		/* # bytes to jump back */
	int4		nscan;		/* # bytes scanned */
	int4		ret;		/* return value of f77back */
	char		*p;		/* scanning byte ptr */
	int		nfound;		/* # newlines found */
	int		err;		/* error code for f77back */

	f = fdfind(request->fq_tfd);

	if (f == FDNULL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, (int4) ERROR, INT4_NIL, INT4_NIL);
		return;
	}

	f->f_src_node = request->fq_src_node;
	f->f_src_event = request->fq_src_event;
	fdactivate(f);
/*
 * If file is formatted then seek backwards for two newlines ('\n').
 */
	if (request->fq_fmt == FMT) {
		nfound = 0;
		err = 0;
		ret = 0;
/*
 * Continue until we find two newlines or we hit the start of the file.
 */
		while ((nfound < 2) && (f->f_ptr > 0)) {

			nback = (f->f_ptr < NCARD) ? f->f_ptr : NCARD;
			nscan = 0;
/*
 * Back up a reasonable distance.
 */
			if ((f->f_ptr = mylseek(f->f_fd, -nback,
					SEEK_CUR)) < 0) {
				perror("filed");
				kexit(errno);
			}
/*
 * Read that amount into our multi-purpose buffer.
 */
			if (myread(f->f_fd, buf, nback) < nback) {
				perror("filed");
				kexit(errno);
			}

			f->f_ptr += nback;
/*
 * Scan it backwards for newlines.
 */
			p = buf + nback;

			while ((nback > 0) && (nfound < 2)) {
				p--; nback--; nscan++;

				if (*p == '\n') {
					nfound++;
				}
			}
/*
 * Bump forward past the second newline, if found.
 */
			if (nfound >= 2) {
				nback++;
				nscan--;
			}
/*
 * Back up past the scanned portion.
 */
			if ((f->f_ptr = mylseek(f->f_fd, -nscan, SEEK_CUR))
					< 0) {
				perror("filed");
				kexit(errno);
			}
		}
	}
/*
 * Otherwise, read the length of the previous record and then
 * seek backwards that amount.
 */
	else if ((f->f_ptr > 0) && (f->f_ptr < 4)) {
		err = EBADREC;
		ret = ERROR;
	}
/*
 * We're already at the head of the file - essentially a NO-OP.
 */
	else if (f->f_ptr == 0) {
		err = 0;
		ret = 0;
	}
/*
 * It's an unformatted file.
 */
	else {
/*
 * Back up in position to read the trailing byte count of the last record.
 */
		if ((f->f_ptr = mylseek(f->f_fd, -4, SEEK_CUR)) < 0) {
			perror("filed");
			kexit(errno);
		}
/*
 * Read the byte count of the last record.
 */
		if (myread(f->f_fd, (char *) &nback, 4) < 4) {
			perror("filed");
			kexit(errno);
		}

		f->f_ptr += 4;
		nback += 2 * 4;
/*
 * Back up to the beginning of the last record.
 */
		if ((ret = mylseek(f->f_fd, -nback, SEEK_CUR)) < 0) {
			perror("filed");
			kexit(errno);
		}

		if (ret < 0) {

			if (errno == EINVAL) {
				err = EBADREC;
			} else {
				perror("filed");
				kexit(errno);
			}
		} else {
			err = 0;
			f->f_ptr = ret;
			ret = 0;
		}
	}

	sendreply(request->fq_src_node, request->fq_src_event, err, ret,
		INT4_NIL, INT4_NIL);
}

/*
 *	do_f77read
 *
 *	Function:	- read a record from a F77 file
 *			- reads up to a newline if file is formatted
 *			- reads the length and then the body of the record
 *			  if file is unformatted
 *			- assumes we have a sequential F77 file
 *
 *	Accepts:	- ptr to filed request
 */

void
do_f77read(request)

struct freq		*request;

{
	struct fdesc	*f;		/* ptr to old target descriptor */
	int4		nread;		/* #bytes actually read */
	int4		nret;		/* #bytes replied to client */
	int4		nrec;		/* #bytes in full record */
	int4		npkt;		/* #bytes in one reply packet */
	int4		r;		/* result of mylseek */
	int4		n;		/* countdown of nret */
	char		*prec;		/* maximum record ptr */
	char		*p;		/* moving buffer ptr */
	char		*phole;		/* safe place to read & trash bytes */
	int		bfound;		/* newline found? */

	f = fdfind(request->fq_tfd);

	if (f == FDNULL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, (int4) ERROR, INT4_NIL, INT4_NIL);
		return;
	}

	if (request->fq_length <= INT4_NIL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			EINVAL, (int4) ERROR, INT4_NIL, INT4_NIL);
		return;
	}

	f->f_src_node = request->fq_src_node;
	f->f_src_event = request->fq_src_event;
	fdactivate(f);
/*
 * Allocate a buffer large enough to hold the maximum record
 * size indicated by the client.
 */
	if (request->fq_length < MAXNMSGLEN) {
		prec = buf;
	} else {
		prec = malloc((unsigned) request->fq_length);

		if (prec == NULL) {
			perror("filed");
			kexit(errno);
		}
	}

	nread = myread(f->f_fd, prec, request->fq_length);

	if (nread < 0) {
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, (int4) ERROR, INT4_NIL, INT4_NIL);
		return;
	}

	f->f_ptr += nread;
/*
 * If file is formatted then read up to next newline ('\n').
 */
	if (request->fq_fmt == FMT) {
		bfound = FALSE;
		nrec = 0;
		nret = 0;
		p = prec;
/*
 * Scan for a newline.
 */
		while ((nread > 0) && !bfound) {
			nrec++;
			nret++;
			nread--;
			bfound = (*p == '\n');
			p++;
		}
/*
 * Usually, read less than the maximum record and will
 * have to backup over the unused portion.
 */
		if (nread > 0) {
			f->f_ptr = mylseek(f->f_fd, -nread, SEEK_CUR);

			if (f->f_ptr < 0) {
				perror("filed");
				kexit(errno);
			}
/*
 * But if the maximum record is scanned without a newline,
 * move forward to the next record and determine the
 * size of the full current record.
 */
		} else if (!bfound) {
			phole = malloc((unsigned) NCARD);

			if (phole == NULL) {
				perror("filed");
				kexit(errno);
			}
/*
 * Read reasonable sized chunks from the file and continue scanning
 * for the newline until it is found or the end of the file is reached.
 */
			do {
				nread = myread(f->f_fd, phole, NCARD);

				if (nread < 0) {
					perror("filed");
					kexit(errno);
				} else if (nread == 0) {
					bfound = TRUE;
				} else {
/*
 * Scan current chunk for a newline.
 */
					p = phole;

					while ((nread > 0) && !bfound) {
						nrec++;
						nread--;
						bfound = (*p == '\n');
						p++;
					}
/*
 * Backup over unused portion of current chunk.
 */
					if (nread > 0) {
						f->f_ptr = mylseek(f->f_fd,
							-nread, SEEK_CUR);

						if (f->f_ptr < 0) {
							perror("filed");
							kexit(errno);
						}
					}
				}
			} while (!bfound);

			(void) free(phole);
		}
	}
/*
 * The file is unformatted and length is found at the head and
 * tail of the record.
 */
	else {
		if (nread < 4) {
			sendreply(request->fq_src_node, request->fq_src_event,
				EBADREC, (int4) ERROR, INT4_NIL, INT4_NIL);
			return;
		}
/*
 * Total length of record accounts for head and tail record lengths.
 */
		nrec = *((int4 *) prec) + (2 * 4);
		nret = (nrec > nread) ? nread : nrec;
/*
 * Move file ptr to the end of the record, forward or backward.
 */
		r = mylseek(f->f_fd, nrec - nread, SEEK_CUR);

		if (r < INT4_NIL) {

			if (errno == EINVAL) {
				sendreply(request->fq_src_node,
					request->fq_src_event,
					EBADREC, (int4) ERROR, INT4_NIL,
					INT4_NIL);
				return;
			} else {
				perror("filed");
				kexit(errno);
			}
		} else {
			f->f_ptr = r;
		}
	}

	n = nret;
	p = prec;

	do {
		npkt = (n > MAXNMSGLEN) ? MAXNMSGLEN : n;
		bcopy(p, buf, npkt);
		sendreply(request->fq_src_node, request->fq_src_event,
				0, nret, npkt, f->f_tflags & O_DMSG);
		n -= npkt;
		p += npkt;
	} while (n > 0);

	if (prec != buf) {
		(void) free(prec);
	}
}
