/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	fqlseek.c,v $
 * Revision 2.2  92/04/30  14:18:40  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/06  10:57:08  gdburns
 * Add flags capability to sendreply().
 * Remove time stamp from internal file descriptor.
 * 
 * Revision 2.1.1.1  91/10/09  15:44:20  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1  91/02/24  20:59:45  gdburns
 * Ohio Release
 * 
 *	Function:	- FQLSEEK request
 *			- supports Cubix
 */

#include "filed.h"
#include <freq.h>
#include <portable.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */

/*
 * external functions
 */
struct fdesc		*fdfind();	/* find a filed descriptor */
void			sendreply();	/* general purpose file reply */
void			fdactivate();	/* make sure tfd is active */
int4			mylseek();	/* portable int4 lseek() */
int			cxhear();	/* note sync of one Cubix node */

/*
 *	do_lseek
 *
 *	Function:	- handles simple client lseek
 *			- handles synchronized Cubix lseek
 *
 *	Accepts:	- ptr to filed request
 */

void
do_lseek(request)

struct freq		*request;

{
	struct fdesc	*f;		/* ptr to target descriptor */
	int4		i;		/* favourite index */
	int4		ret;		/* return value of lseek() */
	int		status;		/* Cubix status */
	int		err;		/* errno */

	f = fdfind(request->fq_tfd);

	if (f == FDNULL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, (int4) ERROR, INT4_NIL, INT4_NIL);
	}
/*
 * Cubix method
 */
	else if (f->f_tflags & FCUBIX) {
		status = cxhear(f, request);

		if (status == ERROR) {
			sendreply(request->fq_src_node, request->fq_src_event,
				errno, (int4) ERROR, INT4_NIL, INT4_NIL);
		}

		else if (status == 0) {
		}

		else {
			fdactivate(f);
/*
 * Do a single lseek.
 */
			if (f->f_tflags & FSINGL) {
				ret = mylseek(f->f_fd, request->fq_offset,
					request->fq_whence);

				if (ret == ERROR) {
					err = errno;
				} else {
					err = 0;
					f->f_ptr = ret;
				}
			}
/*
 * Reply to everybody in order, perhaps doing each member's
 * seek beforehand if in multi mode.
 */
			for (i = 0; i < f->f_ncubix; ++i) {

				if (f->f_tflags & FMULTI) {
					ret = mylseek(f->f_fd,
						f->f_pcubix[i].cx_offset,
						request->fq_whence);

					if (ret == ERROR) {
						err = errno;
					} else {
						err = 0;
						f->f_ptr = ret;
					}
				}

				sendreply(f->f_pcubix[i].cx_src_node,
					f->f_pcubix[i].cx_src_event,
					err, ret, INT4_NIL, INT4_NIL);
			}
		}
	}
/*
 * async method, one lseek and one reply
 */
	else {
		fdactivate(f);
		f->f_src_node = request->fq_src_node;
		f->f_src_event = request->fq_src_event;
		ret = mylseek(f->f_fd, request->fq_offset, request->fq_whence);

		if (ret != ERROR) {
			errno = 0;
			f->f_ptr = ret;
		}

		sendreply(request->fq_src_node, request->fq_src_event,
			errno, ret, INT4_NIL, INT4_NIL);
	}
}
