/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	fqread.c,v $
 * Revision 2.2  92/04/30  14:18:43  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.4  92/02/06  10:57:12  gdburns
 * Add flags capability to sendreply().
 * Remove time stamp from internal file descriptor.
 * 
 * Revision 2.1.1.3  91/10/09  15:44:26  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1.1.2  91/08/07  12:45:48  raja
 * Another place where errno is substituted by err.
 * 
 * Revision 2.1.1.1  91/05/01  22:23:47  gdburns
 * Substitute local variable err for volatile errno.
 * 
 * Revision 2.1  91/02/24  20:59:49  gdburns
 * Ohio Release
 * 
 *	Function:	- read request
 *			- supports Cubix
 *			- exits daemon in case of non-user error
 *			- read operations longer than a network message
 *			  are done in packets
 */

#include <errno.h>
#include "filed.h"
#include <freq.h>
#include <net.h>
#include <portable.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */

/*
 * external functions
 */
struct fdesc		*fdfind();	/* find a filed descriptor */
void			sendreply();	/* general purpose file reply */
void			fdactivate();	/* make sure tfd is active */
int4			myread();	/* portable read() */
int			cxhear();	/* record a Cubix operation */

/*
 * local functions
 */
static void		readcubix();	/* Cubix read hand waving */
static void		readsend();	/* read and ship to client */

/*
 * local variables
 */
extern char		buf[];		/* input data from read */

void
do_read(request)

struct freq		*request;

{
	struct fdesc	*f;		/* target filed descriptor */

	
	f = fdfind(request->fq_tfd);

	if (f == FDNULL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			EBADF, (int4) ERROR, INT4_NIL, INT4_NIL);
	}

	else {
		f->f_src_node = request->fq_src_node;
		f->f_src_event = request->fq_src_event;

		if (f->f_tflags & FCUBIX) {
			readcubix(f, request);
		} else {
			readsend(f, request);
		}
	}
}

/*
 *	readsend
 *
 *	Function:	- read bytes from a file and send them to a client
 *
 *	Accepts:	- ptr to fdesc
 *			- ptr to filed request
 */

static void
readsend(f, request)

struct fdesc		*f;
struct freq		*request;

{
	int4		ret;		/* length returned by myread */
	int4		len;		/* read packet length */
	struct cxnode	*p;		/* favourite ptr */
	int4		i;		/* favourite index */
	int4		err;		/* consistent errno */

	fdactivate(f);
/*
 * We want to perform at least one read operation even if the
 * length is 0.  Multiple reads are performed if the length requested
 * exceeds the network message size.  The client will be able to
 * detect, based on the reply->fr_ret, which is the last packet.
 */
	do {
		len = (request->fq_length > MAXNMSGLEN) ? MAXNMSGLEN :
				request->fq_length;
		ret = myread(f->f_fd, buf, len);

		if (ret == ERROR) {
			request->fq_length = 0;
			err = errno;
			len = 0;
		} else {
			err = 0;
			f->f_ptr += ret;
			f->f_flow += ret;

			if (ret < len) {
				request->fq_length = 0;
				len = ret;
			} else {
				request->fq_length -= len;
			}
		}
/*
 * This may not be pretty here, but if SINGL mode we need to send the
 * same data to everybody and this is the easiest place to do it.
 */
		if (f->f_tflags & FSINGL) {
			i = 0;
			p = f->f_pcubix;

			while (i < f->f_ncubix) {
				sendreply(p->cx_src_node, p->cx_src_event,
					err, ret, len, f->f_tflags & O_DMSG);
				++i;
				++p;
			}
		}
/*
 * Send to one client.  This takes care of async and multi methods.
 */
		else {
			sendreply(request->fq_src_node, request->fq_src_event,
				err, ret, len, f->f_tflags & O_DMSG);
		}
		
	} while (request->fq_length > 0);
}

/*
 *	readcubix
 *
 *	Function:	- Cubix read operation
 *			- Unless all other cast members have made the same
 *			  request, readcubix() simply notes the event and
 *			  takes no other action.
 *			- If all other members have been "heard" all members'
 *			  requests are executed in order (multi) or
 *			  a single read operation is performed and the result
 *			  sent to each (singl).
 *
 *	Accepts:	- ptr to fdesc
 *			- ptr to filed request
 */

static void
readcubix(f, request)

struct fdesc			*f;
struct freq			*request;

{
	int			status;		/* Cubix status */
	int4			i;		/* favourite index */
	struct cxnode		*p;		/* favourite ptr */

	status = cxhear(f, request);

	if (status < 0) {
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, (int4) ERROR, INT4_NIL, INT4_NIL);
	}

	else if (status == 0) {
/*
 * Not all cast members have been heard from.
 * Nothing to do at this point.
 */
	}
/*
 * singl method
 */
	else if (f->f_tflags & FSINGL) {
		readsend(f, request);
	}
/*
 * multi method
 */
	else {
		p = f->f_pcubix;
		i = 0;

		while (i < f->f_ncubix) {
			request->fq_src_node = p->cx_src_node;
			request->fq_src_event = p->cx_src_event;
			request->fq_length = p->cx_length;
			readsend(f, request);
			++i;
			++p;
		}
	}
}
