/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	fqsimplex.c,v $
 * Revision 2.2  92/04/30  14:18:45  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.4  92/02/28  10:45:56  gdburns
 * Give our best effort to stat() and fstat().
 * 
 * Revision 2.1.1.3  92/02/06  10:57:14  gdburns
 * Add flags capability to sendreply().
 * Remove time stamp from internal file descriptor.
 * 
 * Revision 2.1.1.2  91/10/09  15:44:31  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1.1.1  91/05/22  16:09:20  raja
 * Incremented the string length in do_getwd() to account for the NULL.
 * 
 * Revision 2.1  91/02/24  20:59:51  gdburns
 * Ohio Release
 * 
 *	Function:	- simple filed requests
 *			- none support Cubix
 *
 *	Accepts:	- ptr to filed request structure
 */

#include <errno.h>
#include "filed.h"
#include <freq.h>
#include <net.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <typical.h>
#include <t_types.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */
extern char		buf[];		/* multi-purpose I/O buffer */

/*
 * global functions
 */
void			do_access();
void			do_chdir();
void			do_dup();
void			do_fstat();
void			do_ftrunc();
void			do_getwd();
void			do_incr();
void			do_mkdir();
void			do_rmdir();
void			do_rmfd();
void			do_stat();
void			do_trunc();
void			do_unlink();

/*
 * external functions
 */
extern int		strlen();
extern int4		mychdir();
extern int4		myaccess();
extern int4		mymkdir();
extern int4		myrmdir();
extern int4		myunlink();
extern int4		myclose();
extern int4		myfstat();
extern int4		mystat();
extern int4		myftrunc();
extern int4		mytrunc();
extern char		*mygetwd();
extern char		*strcpy();
extern void		sendreply();		/* to client */
extern void		fdcut();		/* clean & remove descriptor */
extern void		fdactivate();		/* make descriptor active */
extern struct fdesc	*fdfind();		/* find descriptor by tfd */
extern struct fdesc	*fdinit();		/* create & initialize desc. */

/*
 *	do_ftrunc
 *
 *	Function:	- set an open file to a specified length
 */

void
do_ftrunc(request)

struct freq		*request;

{
	int4		ret;			/* result of myftrunc() */
	struct fdesc	*f;			/* target descriptor */
	int4		*plen;			/* ptr desired length */

	f = fdfind(request->fq_tfd);

	if (f == FDNULL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			EBADF, (int4) ERROR, INT4_NIL, INT4_NIL);
	} else {
		plen = (int4 *) buf;
		fdactivate(f);
		ret = myftrunc(f->f_fd, ttol(*plen));
		sendreply(request->fq_src_node, request->fq_src_event,
			(ret == ERROR) ? errno : 0, ret, INT4_NIL, INT4_NIL);
	}
}

/*
 *	do_trunc
 *
 *	Function:	- set a named file to a specified length
 */

void
do_trunc(request)

struct freq		*request;

{
	int4		ret;			/* result of mytrunc() */

	ret = mytrunc(buf, request->fq_trunclen);
	sendreply(request->fq_src_node, request->fq_src_event,
		(ret == ERROR) ? errno : 0, ret, INT4_NIL, INT4_NIL);
}

/*
 *	do_fstat
 *
 *	Function:	- get status of a file descriptor
 */

void
do_fstat(request)

struct freq		*request;

{
	int4		ret;			/* result of stat() */
	struct fdesc	*f;			/* target descriptor */

	f = fdfind(request->fq_tfd);

	if (f == FDNULL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			EBADF, (int4) ERROR, INT4_NIL, INT4_NIL);
	} else {
		fdactivate(f);
		ret = myfstat(f->f_fd, buf);
		sendreply(request->fq_src_node, request->fq_src_event,
			(ret == ERROR) ? errno : 0, ret, sizeof(struct stat),
			DINTMSG);
	}
}

/*
 *	do_stat
 *
 *	Function:	- get status of a named file
 */

void
do_stat(request)

struct freq		*request;

{
	int4		ret;			/* result of stat() */

	ret = mystat(buf, buf);
	sendreply(request->fq_src_node, request->fq_src_event,
		(ret == ERROR) ? errno : 0, ret, sizeof(struct stat), DINTMSG);
}

/*
 *	do_unlink
 *
 *	Function:	- remove a file
 */

void
do_unlink(request)

struct freq		*request;

{
	int4		ret;			/* result of unlink() */

	ret = myunlink(buf);
	sendreply(request->fq_src_node, request->fq_src_event,
		(ret == ERROR) ? errno : 0, ret, INT4_NIL, INT4_NIL);
}

/*
 *	do_rmdir
 *
 *	Function:	- remove a directory
 */

void
do_rmdir(request)

struct freq		*request;

{
	int4		ret;			/* result of rmdir() */

	ret = myrmdir(buf);
	sendreply(request->fq_src_node, request->fq_src_event,
		(ret == ERROR) ? errno : 0, ret, INT4_NIL, INT4_NIL);
}

/*
 *	do_system
 *
 *	Function:	- issue a system command
 */

void
do_system(request)

struct freq		*request;

{
	int4		ret;			/* result of system() */

	ret = mysystem(buf);
	sendreply(request->fq_src_node, request->fq_src_event,
		(ret == ERROR) ? errno : 0, ret, INT4_NIL, INT4_NIL);
}

/*
 *	do_mkdir
 *
 *	Function:	- make a new directory
 */

void
do_mkdir(request)

struct freq		*request;

{
	int4		ret;			/* result of mkdir() */

	ret = mymkdir(buf, request->fq_accmode);
	sendreply(request->fq_src_node, request->fq_src_event,
		(ret == ERROR) ? errno : 0, ret, INT4_NIL, INT4_NIL);
}

/*
 *	do_access
 *
 *	Function:	- check accessibility of a file
 */

void
do_access(request)

struct freq		*request;

{
	int4		ret;			/* result of access() */

	ret = myaccess(buf, request->fq_accmode);
	sendreply(request->fq_src_node, request->fq_src_event,
		(ret == ERROR) ? errno : 0, ret, INT4_NIL, INT4_NIL);
}

/*
 *	do_chdir
 *
 *	Function:	- change working directory
 */

void
do_chdir(request)

struct freq		*request;

{
	int4		ret;			/* result of chdir() */

	ret = mychdir(buf);
	sendreply(request->fq_src_node, request->fq_src_event,
		(ret == ERROR) ? errno : 0, ret, INT4_NIL, INT4_NIL);
}

/*
 *	do_getwd
 *
 *	Function:	- get working directory
 */

void
do_getwd(request)

struct freq		*request;

{
	(void) mygetwd(buf);
	sendreply(request->fq_src_node, request->fq_src_event,
		0, INT4_NIL, (int4) strlen(buf) + 1, INT4_NIL);
}

/*
 *	do_rmfd
 *
 *	Function:	- blow away file descriptor
 *			- NOTFD indicates that all descriptors should
 *			  be blown away
 */

void
do_rmfd(request)

struct freq		*request;

{
	struct fdesc	*f;		/* ptr to old target descriptor */

	if (request->fq_tfd == NOTFD) {
		f = fdfind(2);

		while (f->f_next != FDNULL) {
			f = f->f_next;

			if (f->f_tflags & FACTIVE) {
				(void) myclose(f->f_fd);
			}

			fdcut(f);
		}
	}

	else {
		f = fdfind(request->fq_tfd);

		if (f == FDNULL) {
			sendreply(request->fq_src_node, request->fq_src_event,
				errno, (int4) ERROR, INT4_NIL, INT4_NIL);
			return;
		}

		else {

			if (f->f_tflags & FACTIVE) {
				(void) myclose(f->f_fd);
			}

			fdcut(f);
		}
	}

	sendreply(request->fq_src_node, request->fq_src_event,
		0, INT4_NIL, INT4_NIL, INT4_NIL);
}

/*
 *	do_dup
 *
 *	Function:	- duplicate file descriptor
 *			- inherits old descriptors file ptr but
 *			  flow is reset to 0
 *			- new descriptor starts inactive
 *			- does not support Cubix
 */

void
do_dup(request)

struct freq		*request;

{
	struct fdesc	*of;		/* ptr to old target descriptor */
	struct fdesc	*nf;		/* ptr to new target descriptor */

	of = fdfind(request->fq_tfd);

	if (of == FDNULL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, (int4) ERROR, INT4_NIL, INT4_NIL);
		return;
	}
/*
 * We don't allow Cubix duplications.
 */
	if (of->f_tflags & FCUBIX) {
		sendreply(request->fq_src_node, request->fq_src_event,
			EBADF, (int4) ERROR, INT4_NIL, INT4_NIL);
		return;
	}

	(void) strcpy(buf, of->f_name);
	nf = fdinit(of->f_tflags);
	nf->f_ptr = of->f_ptr;
	nf->f_count = of->f_count;
	nf->f_src_node = of->f_src_node = request->fq_src_node;
	nf->f_src_event = of->f_src_event = request->fq_src_event;

	sendreply(request->fq_src_node, request->fq_src_event,
		0, nf->f_tfd, INT4_NIL, INT4_NIL);
}

/*
 *	do_incr
 *
 *	Function:	- increment the usage count of a file descriptor
 *			- used by loadgo to set the usage count of
 *			  redirected stdio files to match the number of
 *			  created processes - when all processes die,
 *			  usage count falls to 0 and file is removed
 *			- should be used whenever a tfd is given to another
 *			  process without a reuse open operation
 */

void
do_incr(request)

struct freq		*request;

{
	struct fdesc	*f;		/* ptr to target descriptor */
	int4		*incr;		/* increment amount */

	f = fdfind(request->fq_tfd);

	if (f == FDNULL) {
		sendreply(request->fq_src_node, request->fq_src_event,
			errno, (int4) ERROR, INT4_NIL, INT4_NIL);
		return;
	}

	incr = (int4 *) buf;
	f->f_count += ttol(*incr);
	sendreply(request->fq_src_node, request->fq_src_event, 0, INT4_NIL,
		INT4_NIL, INT4_NIL);
}
