/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	fqstate.c,v $
 * Revision 2.2  92/04/30  14:18:46  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/06  10:57:17  gdburns
 * Add flags capability to sendreply().
 * Remove time stamp from internal file descriptor.
 * 
 * Revision 2.1.1.1  91/10/09  15:44:39  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1  91/02/24  20:59:52  gdburns
 * Ohio Release
 * 
 *	Function:	- FQSTATE request
 *			- a monitoring request with no associated file operation
 */

#include "filed.h"
#include <freq.h>
#include <net.h>
#include <t_types.h>
#include <terror.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */
extern char		buf[];		/* general purpose message buffer */
extern struct fdesc	*pfdtop;	/* top of descriptor list */

/*
 * external functions
 */
extern char		*strcpy();
void			sendreply();	/* general purpose file reply */

/*
 *	do_state
 *
 *	Function:	- send back every file descriptor
 *
 *	Accepts:	- ptr to filed request
 *			- assumes MAXNMSGLEN > sizeof(struct fdesc)
 */

void
do_state(request)

struct freq		*request;

{
	struct fdesc	*f;		/* ptr to target descriptor */
	struct fstate	*p;		/* ptr to desc state */
	int4		n;		/* total space used in buf */

	p = (struct fstate *) buf;
	n = 0;

	for (f = pfdtop; f != FDNULL; f = f->f_next) {
/*
 * If buffer is almost full, send it back to the client and reset.
 */
		if ((n + sizeof(struct fstate)) > MAXNMSGLEN) {
			sendreply(request->fq_src_node,
				request->fq_src_event,
				0, INT4_NIL, n, INT4_NIL);
			p = (struct fstate *) buf;
			n = 0;
		}
/*
 * Add to the buffer.
 */
		else {
			p->fs_tfd = ltot(f->f_tfd);
			p->fs_tflags = ltot(f->f_tflags);
			p->fs_flow = ltot(f->f_flow);
			p->fs_ncubix = ltot(f->f_ncubix);
			p->fs_nheard = ltot(f->f_nheard);
			p->fs_req = ltot(f->f_req);
			p->fs_src_node = ltot(f->f_src_node);
			p->fs_src_event = ltot(f->f_src_event);
			(void) strcpy(p->fs_name, f->f_name);
			p += 1;
			n += sizeof(struct fstate);
		}
	}

	sendreply(request->fq_src_node, request->fq_src_event,
		EEOF, INT4_NIL, n, NOBUF);
}
