/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 Cornell University
 *	GDB
 *
 *	$Log:	kernelio.udp.c,v $
 * Revision 2.2  92/04/30  14:19:26  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/27  11:57:26  gdburns
 * enhanced robustness
 * new tkill procedure
 * 
 * Revision 2.1.1.1  91/11/01  22:14:44  gdburns
 * Signal accepts pointer to a function returning void.
 * 
 * Revision 2.1  91/03/20  02:58:36  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:23:37  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:50:57  trillium
 * Homo erectus
 *
 *	Function:	- receive kernel requests and send replys
 *			- OTB specific code
 *			- uses INET domain datagram sockets
 *			- binds a fixed, globally published port number
 *			  defined in kernel.h
 *			- should only receive requests from within
 *			  the native machine
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#include <stdio.h>
#include <kernel.h>
#include <kreq.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <signal.h>

/*
 * external variables
 */
extern int		errno;			/* global error variable */
extern struct kreq	request;		/* kernel request */
extern struct kreply	reply;			/* kernel reply */

/*
 * local variables
 */
static struct sockaddr_in	kernel;		/* kernel's full socket addr */
static struct sockaddr_in	client;		/* client's full socket addr */
static u_long			ip_addr;	/* native machine's INET addr */
static int			sock;		/* kernel I/O socket */

/*
 * external functions
 */
extern void		kkillall();		/* kill all clients */
extern char		*kkillopen();		/* open kill file */
extern char		*strcpy();

/*
 *	init_request
 *
 *	Function:	- initializes communication structures for
 *			  sending and receiving kernel requests
 */

void
init_request()

{
	char		*f_kill;		/* kill filename */
	char		myhostname[MAXHNAME];	/* space for hostname */
	struct hostent	*myhostent;		/* my host entry */

	(void) strcpy(myhostname, "localhost");

	if ((myhostent = gethostbyname(myhostname)) == (struct hostent *) 0) {
		fprintf(stderr, "kernel: cannot get host entry for %s\n",
			myhostname);
		perror("kernel (gethostbyname)");
		exit(1);   /* should be h_errno */
	}

	bcopy((caddr_t) myhostent->h_addr, (caddr_t) &ip_addr,
		myhostent->h_length);
/*
 * Establish kernel socket.
 */
/*
 * Set INET domain address.
 */
	bzero((caddr_t) &kernel, sizeof(kernel));
	bcopy((caddr_t) &ip_addr, (caddr_t) &kernel.sin_addr, sizeof(ip_addr));
	kernel.sin_family = AF_INET;
	bcopy((caddr_t) &kernel, (caddr_t) &client, sizeof(kernel));
/*
 * Set kernel port number.
 */
	kernel.sin_port = htons((u_short) KERNEL_PORT + geteuid());

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		fprintf(stderr, "kernel: cannot open kernel request socket\n");
		perror("kernel (socket)");
		exit(errno);
	}

	if (bind(sock, (struct sockaddr *) &kernel, sizeof(kernel)) < 0) {
		fprintf(stderr, "kernel: cannot bind internet port %d\n",
			kernel.sin_port);
		perror("kernel (bind)");
		exit(errno);
	}
/*
 * Catch SIGTERM and kill all attached processes.
 */
	(void) signal(SIGTERM, kkillall);
	(void) signal(SIGTSTP, SIG_IGN);
/*
 * Open the kill record file.
 */
	if ((f_kill = kkillopen()) != (char *) 0) {
		fprintf(stderr, "kernel: %s\n", f_kill);
		perror("kernel (kkillopen)");
		exit(errno);
	}
}

/*
 *	recv_req
 *
 *	Function:	- receives the next kernel request
 */

void
recv_req()

{
	int		lsize;		/* size of client INET address */

	if (recvfrom(sock, (caddr_t) &request, sizeof(request), 
			0, (struct sockaddr *) &client, &lsize) < 0) {
		perror("kernel (recv_req)");
		exit(errno);
	}
}

/*
 *	send_reply
 *
 *	Function:	- send reply back to the client
 *			- message contains result of kernel request
 *	Accepts:	- client's kernel communication port
 */

void
send_reply(client_port)

int			client_port;		/* INET port number of client */

{
	client.sin_port = (u_short) client_port;

	if (sendto(sock, (caddr_t) &reply, sizeof(reply), 0, 
			(struct sockaddr *) &client, sizeof(client)) < 0) {

		perror("kernel (send_reply)");
		exit(errno);
	}
}
