/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 Cornell University
 *	GDB
 *
 *	$Log:	kkill.c,v $
 * Revision 2.2  92/04/30  14:19:27  trillium
 * Ohio Release
 * 
 *	Function:	- supports client cleanup by the kernel and by tkill(l)
 *			- OTB specific code
 */

#include <stdio.h>
#include <boot.h>
#include <fcntl.h>
#include <kernel.h>
#include <kreq.h>
#include <portable.h>
#include <signal.h>
#include <typical.h>
#include <unistd.h>

/*
 * global functions
 */
void			kkillall();
char			*kkillopen();
int			kkillclose();
int			kkillrewrite();
int			kkillwrite();

/*
 * external functions
 */
extern int		addarg();
extern char		*findpath();
extern char		*getenv();
extern char		*malloc();
extern char		*strcpy();
extern char		*strcat();

/*
 * external variables
 */
extern struct kprocess	ptable[];	/* process table */

/*
 * local variables
 */
static FILE		*fp_kill;	/* kill file ptr */
static int		fd_kill;	/* kill file desc. */

/*
 *	kkillall
 *
 *	Function:	- catches SIGTERM and SIGHUP
 *			- kills all attached processes
 *			- exits
 */

void
kkillall()

{
	int		i;			/* favourite index variable */

	for (i = 0; i < PMAX; ++i) {

		if (ptable[i].k_state != SFREE) {
			(void) kill(ptable[i].k_pid, SIGCONT);
			(void) kill(ptable[i].k_pid, SIGHUP);
		}
	}

	exit(0);
}

/*
 *	kkillopen
 *
 *	Function:	- opens the kill file
 *			- locates the kill directory and constructs
 *			  the full kill pathname by appending the
 *			  userid to the kill directory
 *			- writes the kernel's own pid in the first
 *			  entry of the kill file
 *	Returns:	- null ptr or ptr to kill file if error
 */

char *
kkillopen()

{
	char		*f_kill;	/* kill file */
	char		*d_kill;	/* kill directory */
	char		**p_kill;	/* kill pathname prefixes */
	char		*full;		/* full pathname */
	int		n_kill;		/* # of p_kill prefixes */
/*
 * Initialize search paths.
 */
	n_kill = 0;
	p_kill = (char **) 0;
	addarg(&n_kill, &p_kill, "");
	addarg(&n_kill, &p_kill, "$TROLLIUSHOME");
	addarg(&n_kill, &p_kill, DEFPLOCK);
/*
 * Initialize file names.
 */
	d_kill = "locks";
	f_kill = getenv("LOGNAME");

	if (! f_kill) {
		return("LOGNAME");
	}
/*
 * Try to locate the reset directory.
 */
	full = findpath(d_kill, p_kill, W_OK | X_OK);

	if (! full) {
		return(d_kill);
	}
/*
 * Build a reset filename from the directory name and the userid.
 */
	d_kill = full;
	full = malloc((unsigned) strlen(d_kill) + strlen(f_kill) + 2);

	if (! full) {
		return("malloc");
	}

	(void) strcpy(full, d_kill);
	(void) strcat(full, "/");
	(void) strcat(full, f_kill);
	f_kill = full;
/*
 * Open the reset file.
 */
	fd_kill = open(f_kill, O_WRONLY | O_CREAT | O_EXCL, 0644);

	if (fd_kill < 0) {
		return(f_kill);
	}

	fp_kill = fdopen(fd_kill, "w");

	if (! fp_kill) {
		return(f_kill);
	}

	setlinebuf(fp_kill);

	if (kkillwrite(getpid()) < 0) {
		return(f_kill);
	}

	return((char *) 0);
}

/*
 *	kkillwrite
 *
 *	Function:	- writes pid into lock file
 *	Accepts:	- pid
 *	Returns:	- 0 or ERROR
 */

int
kkillwrite(pid)

int4			pid;

{
	if (fprintf(fp_kill, "%d\n", pid) == EOF) {
		return(ERROR);
	} else {
		return(0);
	}
}

/*
 *	kkillrewrite
 *
 *	Function:	- rewrites entire kill file from current process list
 *	Returns:	- 0 or ERROR
 */

int
kkillrewrite()

{
	int		i;			/* favourite index variable */

/*
 * Rewind the stdio descriptor.
 */
	rewind(fp_kill);
/*
 * Rewind and truncate the real thing.
 */
	if (ftruncate(fd_kill, (long) 0) < 0) {
		return(ERROR);
	}
/*
 * Write the kernel pid first.
 */
	if (fprintf(fp_kill, "%d\n", getpid()) == EOF) {
		return(ERROR);
	}
/*
 * Write every current client pid.
 */
	for (i = 0; i < PMAX; ++i) {

		if (ptable[i].k_state != SFREE) {

			if (fprintf(fp_kill, "%d\n", ptable[i].k_pid) == EOF) {
				return(ERROR);
			}
		}
	}

	return(0);
}

/*
 *	kkillclose
 *
 *	Function:	- Flush and close the kill file.
 *	Returns:	- 0 or ERROR
 */

int
kkillclose()

{
	if (fclose(fp_kill) == EOF) {
		return(ERROR);
	} else {
		return(0);
	}
}
