/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	DLF/GDB
 *
 *	$Log:	loadd.c,v $
 * Revision 2.2  92/04/30  14:19:57  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/01/10  17:31:46  gdburns
 * Use new tout module and toutrev() function.
 * 
 * Revision 2.1  91/03/20  02:59:29  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:23:42  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:52:21  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  00:53:38  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0.1.1  88/05/24  13:23:36  fielding
 * Fixed file descriptor to close on an error
 * 
 * Revision 1.0  87/12/19  15:13:17  trillium
 * Initial revision
 * 
 *
 *
 *	Function : - loader process (server)
 *		   - loads file onto node that the user specifies
 */

#include <loadreq.h>		/* loadd structures */
#include <events.h>		/* loadd event */
#include <debug.h>		/* debugging mode */
#include <t_types.h>		/* bit  order conversion routines */
#include <net.h>		/* network layer structs */
#include <typical.h>		/* commonly used constants */
#include <t.out.h>		/* executable header constants */

#include <sys/types.h>		/* system type definitions */
#include <sys/stat.h>		/* must be able to stat the file */
#include <sys/file.h>		/* unix file system constants */
#include <stdio.h>


/*
 * external variables
 */
extern int		errno;		/* error number */

/*
 * external functions
 */
extern void		toutrev();	/* reverse executable byte ordering */
extern char		*malloc();	/* local memory allocation */
extern int		rflat();	/* flatd interface */
extern int		(*(set_order()))();
					/* determines byte ordering */
/*
 * local variables
 */
static struct nmsg	nhead;		/* network message header */
static struct lreq	*request;	/* loadd request structure */

static char *_copyright = "Copyright 1987 Cornell University $Revision: 2.2 $";

/*
 * local functions
 */
static void		l_sendreply();


main()
{
	struct	stat		status;	/* store file status	*/
	struct 	exec		*thead,	/* header for executable file */
				*ohead, /* ptr to ordered header */
				oexec;	/* ordered header */
	struct to_relative	*reloc; /* pointer to relocation table */
	struct nlist		*symbol;/* pointer to relocation table */

	char	filename[FNAMELEN];	/* file to be loaded */
	int	fdesc;			/* local file descriptor	*/
	int	msize;			/* total size to malloc		*/
	int	fsize;			/* total size to load (T4EX)	*/
	int 	offset1;		/* temporary pointer		*/
	int 	offset2;		/* temporary pointer		*/
	int	(*order)();		/* byte ordering routine	*/
	char	*frem;			/* temporary allocated memory   */
	char	*drem;			/* temporary allocated memory   */


	ohead = &oexec;
/*
 * Attach to kernel.
 */
	if (kinit(0)) {
		terror("loadd (kinit)");
		exit(errno);
	}

	request = (struct lreq *) nhead.nh_data;
/*
 * Run forever.
 */
	for (;;) {
/*
 * Set up header to receive a loadd request.
 */
		nhead.nh_event	= EVLOADD;
		nhead.nh_type	= 0;
		nhead.nh_length	= FNAMELEN;
		nhead.nh_flags	= 0;
		nhead.nh_msg	= filename;
/*
 * Receive request to load file.
 */
		if (nrecv(&nhead)) {
			terror("loadd (nrecv)");
			kexit(errno);
		}


		db(("loadd: dest_node: %X src_node: %X file: %s\n",
			request->ld_dest_node, request->ld_src_node, filename));

/*
 *		1) file can be RAW or TRILLIUM executable
 *
 *			a) trillium 	- verify trillium executable
 *					- find uninitialized data size (udata)
 *					- move data segment
 *					- convert byte ordering
 *					- rflat file to destination node
 *
 *			b) raw		- rflat file to destination node
 */

/*
 * Check if file actually exists.
 */
		if (stat(filename, &status)) {
			l_sendreply(errno);
			/* recover and resume */
			continue;
		}
/*
 * Open file to be loaded.
 */
		if ((fdesc = open(filename, O_RDONLY)) == ERROR) {
			/* return error */
			l_sendreply(errno);
			/* recover and resume */
			continue;
		}
/*
 * Malloc space for the file, OK with limited node memory.
 */
		if ((frem = malloc((unsigned) status.st_size)) == NULL) {
			/* return error */
			l_sendreply(errno);
			/* recover and resume */
			close(fdesc);
			continue;
		}
/*
 * Read in the whole file.
 */
		if (read(fdesc, frem, (int) status.st_size) == ERROR) {
			/* return error */
			l_sendreply(errno);
			/* recover and resume */
			close(fdesc);
			continue;
		}
/*
 * Close up the file descriptor.
 */
		if (close(fdesc) == ERROR) {
			/* return error */
			l_sendreply(errno);
			/* recover and resume */
			continue;
		}			
/*
 * Is it a Trillium executable?
 */
		thead = (struct exec *) frem;

		if (order = set_order(thead)) {
/*
 * Yes, set up a header in local byte ordering.
 */
			bcopy(frem, (char *) ohead, sizeof(struct exec));
			ohead->a_text = (*order)(thead->a_text);
			ohead->a_data = (*order)(thead->a_data);
			ohead->a_bss = (*order)(thead->a_bss);
			ohead->a_syms = (*order)(thead->a_syms);
			ohead->a_drsize = (*order)(thead->a_drsize);
			ohead->a_entry = (*order)(thead->a_entry);

			/* calculate space for load module */
			msize = ohead->a_text
				+ ohead->a_data
				+ ohead->a_bss
				+ ohead->a_syms
				+ ohead->a_drsize 
				+ sizeof(struct exec);

			/* calculate size to load */
			fsize = msize - ohead->a_bss;
/*
 * Swap segment ordering.
 */
			/* malloc space to save data segment */
			if ((drem = malloc((unsigned) ohead->a_data)) == NULL) {

				/* return error */
				l_sendreply(errno);
				/* recover and resume */
				continue;
			}

			/* set pointer to move data segment */
			offset1 = ohead->a_text + sizeof(struct exec);

			/* move data segment */
			bcopy(frem + offset1, drem, (int) ohead->a_data);

			/* set pointers to reorganize array */
			offset2 = offset1 + ohead->a_data;

			/* shift tables */
			bcopy(frem + offset2, frem + offset1,
				(int) (ohead->a_syms + ohead->a_drsize));

			offset2 = offset1 + ohead->a_drsize + ohead->a_syms;

			/* tack on data segment */
			bcopy(drem, frem + offset2, (int) ohead->a_data);
/*
 * Ensure that the entire executable is in Trillium byte ordering.
 */
			if (ltot(TO_MAGIC_T8) != (*order)(TO_MAGIC_T8)) {
				reloc = (struct to_relative *)
					(((char *) thead) +
					ohead->a_text +
					sizeof(struct exec));
				symbol = (struct nlist *)
					(((char *) reloc) +
					ohead->a_drsize);

				toutrev(ohead, thead, reloc, symbol);
			}
/*
 * Load Trillium executable file on destination node.
 */
			if (rflat(request->ld_dest_node, frem, fsize, msize,
					request->ld_tag) == ERROR) {
				/* return error */
				l_sendreply(errno);
				/* recover and resume */
				continue;
			}

			/* free memory for data segment */
			free(drem);

		} else {
/*
 * No, load raw file on destination node.
 */
			if (rflat(request->ld_dest_node, frem, status.st_size,
					status.st_size, request->ld_tag)
					== ERROR) {
				/* return error */
				l_sendreply(errno);
				/* recover and resume */
				continue;
			}
		}
/*
 * Free memory allocated for the file.
 */
		free(frem);
/*
 * Reply to client.
 */
		l_sendreply(0);
	}
}

/*
 *	l_sendreply
 *
 *	Function:	- send a reply back to the client
 *
 *	Accepts:	- error code
 */

static void
l_sendreply(code)

int 			code;

{
	struct lreply	*reply;


	nhead.nh_node	= request->ld_src_node;
	nhead.nh_event	= request->ld_src_event;
	nhead.nh_type	= 0;
	nhead.nh_flags	= 0;
    	nhead.nh_length	= 0;
    	nhead.nh_msg	= (char *) 0;
 
	reply = (struct lreply *) nhead.nh_data;
	reply->ld_reply = code;

    	if (nsend(&nhead)) {
		terror("loadd (nsend)");
		kexit(errno);
    	}
}
