/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University
 *
 *	GDB
 *
 *	$Log:	tfile.c,v $
 * Revision 2.2  92/04/30  14:14:50  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/12/02  16:15:12  gdburns
 * Make sure MAXPATHLEN is defined.
 * This should be converted to POSIX.
 * 
 * Revision 2.1  91/02/27  02:47:38  gdburns
 * Ohio Release
 * 
 *
 *	Function:	- command to control certain filed functions
 *
 *	Accepts:	- see the "usage" variable
 */

#include <freq.h>
#include <stdio.h>
#include <sys/param.h>
#include <terror.h>
#include <typical.h>

#ifndef MAXPATHLEN
#define MAXPATHLEN	1024
#endif

/*
 * external variables
 */
extern int 		errno;		/* global error variable */

/*
 * external functions
 */
extern void		validopts();	/* set the valid options */
extern void		terror();
extern void		intparam();
extern void		followed();
extern void		exclusive();
extern int		do_args();	/* parse the valid options */
extern int		getnode();	/* get next node entry */
extern int		opt_taken();	/* Was a particular option selected? */
extern int		parse_nodes();	/* parse the node identifiers */
extern int		rfchdir();
extern int		rfrmfd();
extern char		*rfgetwd();

/*
 * local functions
 */
static void		help();

/*
 * local variables
 */
static char *_copyright
	= "Copyright 1990 The Ohio State University $Revision: 2.2 $";
static char *usage = " [-hS] [-s <descno>] [<directory>] <nodes>";

static char		pwd[MAXPATHLEN];


main(argc, argv)

int   			argc;    
char   			*argv[];

{
	int		n_index;		/* index of node entry */
	int		nodeid;			/* value of node entry */
	int		n_flags;		/* extra node entry info */
	int		tfd;			/* filed descriptor # */
	char		*me;			/* invocation name */

/*
 * Parse the command line.
 */
	validopts("Ssh");
	followed("s");
	exclusive("Ss");
	me = argv[0];

	if ((do_args(&argc, argv) == ERROR) ||
		(parse_nodes(&argc, argv) == ERROR) ||
		(errno = (argc <= 2) ? 0 : EUSAGE)) {

		if (errno == EUSAGE) {
			fprintf(stderr, "usage: %s%s\n", me, usage);
		}

		else {
			terror(me);
		}

		exit(errno);
	}
/*
 * Check for help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Attach to kernel.
 */
	if (kinit(0) == ERROR) {
		terror(me);
		exit(errno);
	}
/*
 * Loop through all specified nodes.
 */
	getnode(&n_index, &nodeid, &n_flags);

	if (n_index != ERROR) {

		do {
			if (n_flags) {
				fprintf(stderr, "usage: %s\n", usage);
				kexit(errno);
			}
/*
 * Clean up file descriptors in filed.
 */
			if (opt_taken('s') || opt_taken('S')) {

				if (opt_taken('s')) {
		 			intparam('s', &tfd);
				} else {
					tfd = NOTFD;
				}

				if (rfrmfd(nodeid, tfd) == ERROR) {
					terror(me);
					kexit(errno);
				}
			}
/*
 * Print the current working directory.
 */
			else if (argc == 1) {

				if (rfgetwd(nodeid, pwd) == (char *) 0) {
					terror(me);
					kexit(errno);
				}

				printf("%s\n", pwd);
			}
/*
 * Change current working directory.
 */
			if (argc == 2) {

				if (rfchdir(nodeid, argv[1]) == ERROR) {
					terror(me);
					kexit(errno);
				}
			}

			getnode(&n_index, &nodeid, &n_flags);
		} while (n_index);
	}

	else {
		fprintf(stderr, "usage: %s\n", usage);
		kexit(errno);
	}

	kexit(0);
}

/*
 *	  help
 *
 *	  Function:	- prints helpful information on the tfile command
 */

static void
help()
{
	printf("\nSynopsis:	tfile [options] <nodes> [<dir>]\n");
	printf("\nDescription:	Control Trollius file daemon operations.\n");
	printf("\nOptions:	-h	this help message\n");
	printf("\t\t-S	blow away all open descriptors\n");
	printf("\t\t-s <#>	blow away one specific descriptor\n");
	printf("\t\t<dir>	change to new working directory\n");
	printf("\t\tnone	print current working directory\n");
	mnusage();
	printf("\t\th\t(local host node)\n");
	printf("\nExample:	tfile h -S\n");
	printf("\t\t\t\"Clean up (reset) the file daemon.\"\n");
}
