/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	vboot.c,v $
 * Revision 2.2  92/04/30  14:16:37  trillium
 * Ohio Release
 * 
 *	Function:	- boots remote ITB operating system
 */

#include <boot.h>
#include <bootreq.h>
#include <conf.h>
#include <debug.h>
#include <machine/machine.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <terror.h>
#include <typical.h>
#include <unistd.h>

/*
 * external variables
 */
extern int		errno;
extern char		**exttopv;	/* -T will override conf topology */
extern int		exttopc;

/*
 * external functions
 */
extern void		split();
extern void		cnfparse();
extern void		mnusage();
extern void		followed();
extern void		twoletter();
extern void		validopts();
extern int		addarg();
extern int		do_args();
extern int		rload();
extern int		rflat();
extern int		rbtadd();
extern int		rbtaddstage();
extern int		rbtboot();
extern int		rbtflush();
extern int		rbtstage();
extern int		strcmp();
extern int4		cnfshare();
extern int4		intparam();
extern char		*findpath();
extern char		*getparam();
extern char		*malloc();
extern char		*rindex();
extern char		**pack_argv();
extern CONF		*cnffind();

/*
 * local functions
 */
static void		help();
static void		vaddfrozen();
static void		vaddstage();
static void		vaddstrap();
static void		setdefaults();

/*
 * local variables
 */
static char		*f_conf;	/* process schema file */
static char		*f_frozen;	/* frozen process schema file */
static char		*f_strap;	/* bootstrap file */
static char		**p_prog;	/* program pathname prefixes */
static char		**p_conf;	/* conf pathname prefixes */
static char		*me;		/* who am I? */

static int		fl_debug;	/* debugging option */
static int		fl_verbose;	/* verbose option */
static int		n_prog;		/* # of p_prog prefixes */
static int		n_conf;		/* # of p_conf prefixes */

static char *usage = " [-dhoqvN] [-{abefs}{xlptz}] [-c <config>] [-T <topology>]\n\t[-L <link>] [-n <neighbour>] [-i <frozen>]";

main(argc, argv)

int			argc;
char			*argv[];

{
	int		bto;		/* object option let */
	int4		nodeid;		/* target node ID */
	int4		flags;		/* for getnode() */
	int4		i;		/* favourite index */
	int4		ilink;		/* objects to stage */
	int4		index;		/* index of multiple nodeids */
	int4		neighbour;	/* nodeid for staging */
	int4		nnodes;		/* # of node args */
	int4		o_add;		/* objects to add/load */
	int4		o_boot;		/* objects to boot */
	int4		o_spread;	/* objects to spread */
	int4		o_flush;	/* objects to flush */
	int4		o_stage;	/* objects to stage */
	int4		*pnodes;	/* array of node args */

	me = argv[0];
/*
 * Initialize option parser.
 */
	validopts("dhoqvN");
	followed("cinLT");
	twoletter("a:xlz");
	twoletter("b:xlpz");
	twoletter("e:lpz");
	twoletter("f:xltz");
	twoletter("s:xltz");

	if ((do_args(&argc, argv) < 0) ||
		(parse_nodes(&argc, argv) < 0) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		fprintf(stderr, "usage: %s%s\n", me, usage);
		exit(errno);
	}

	if (opt_taken('h')) {
		help();
		exit(0);
	}

	setdefaults();
/*
 * Define, from the options, what actions are to be taken.
 */
	o_add = 0;
	o_boot = 0;
	o_spread = 0;
	o_flush = 0;
	o_stage = 0;
/*
 * Look for the special menu options.
 */
	if (opt_taken('q')) {
		o_boot = BTO_STRAP | BTO_STAGE;
		o_stage = BTO_STRAP | BTO_STAGE;
	}

	if (opt_taken('o')) {
		o_flush = BTO_STRAP | BTO_STAGE;
		o_add = BTO_STRAP | BTO_STAGE;
		o_boot = BTO_STRAP | BTO_STAGE;
		o_stage = BTO_STRAP | BTO_STAGE;
	}
/*
 * Look for the a la carte options.
 */
	if (opt_taken('a')) {
		bto = get_2nd_let('a');

		while (bto >= 0) {
			if (bto == 'x') o_add |= BTO_STRAP;
			if (bto == 'l') o_add |= BTO_STAGE;
			if (bto == 'z') o_add |= BTO_FROZEN;
			bto = get_2nd_let('a');
		}
	}

	if (opt_taken('b')) {
		bto = get_2nd_let('b');

		while (bto >= 0) {
			if (bto == 'x') o_boot |= BTO_STRAP;
			if (bto == 'l') o_boot |= BTO_STAGE;
			if (bto == 'p') o_boot |= BTO_SPAWN;
			if (bto == 'z') o_boot |= BTO_FROZEN;
			bto = get_2nd_let('b');
		}
	}

	if (opt_taken('e')) {
		o_spread = (BTO_STRAP | BTO_ROUTE);
		bto = get_2nd_let('e');

		while (bto >= 0) {
			if (bto == 'l') o_spread |= BTO_STAGE;
			if (bto == 'p') o_spread |= BTO_SPAWN;
			if (bto == 'z') o_spread |= BTO_FROZEN;
			bto = get_2nd_let('e');
		}
	}

	if (opt_taken('f')) {
		bto = get_2nd_let('f');

		while (bto >= 0) {
			if (bto == 'x') o_flush |= BTO_STRAP;
			if (bto == 'l') o_flush |= BTO_STAGE;
			if (bto == 't') o_flush |= BTO_ROUTE;
			if (bto == 'z') o_flush |= BTO_FROZEN;
			bto = get_2nd_let('f');
		}
	}

	if (opt_taken('s')) {
		bto = get_2nd_let('s');

		while (bto >= 0) {
			if (bto == 'x') o_stage |= BTO_STRAP;
			if (bto == 'l') o_stage |= BTO_STAGE;
			if (bto == 't') o_stage |= BTO_ROUTE;
			if (bto == 'z') o_stage |= BTO_FROZEN;
			bto = get_2nd_let('s');
		}
	}

	if (o_boot) {

		if (opt_taken('L')) {
			intparam('L', &ilink);
		} else {
			fprintf(stderr, "%s: must have -L to boot\n", me);
			exit(1);
		}
	}

	if (o_stage) {

		if (opt_taken('n')) {
			intparam('n', &neighbour);
		} else {
			fprintf(stderr, "%s: must have -n to stage\n", me);
			exit(1);
		}
	}

	getnode(&index, &nodeid, &flags);

	if (index < 0) {
		fprintf(stderr, "usage: %s%s\n", me, usage);
		exit(EUSAGE);
	}
/*
 * Become a Trollius process.
 */
	if (kinit(0) < 0) {
		terror(me);
		exit(errno);
	}
/*
 * Take all actions except spread for all nodes in sequence.
 */
	nnodes = 0;

	do {
		nnodes++;
/*
 * Take all flushing actions.
 */
		if (o_flush) {
			VERBOSE("flush ...\n");

			if (! opt_taken('N')) {
				if (rbtflush(nodeid, o_flush) < 0) {
					terror(me);
					kexit(errno);
				}
			}
		}
/*
 * Take all adding actions.
 */
		if (o_add & BTO_STRAP) {
			VERBOSE("load strap ...\n");
			vaddstrap(nodeid);
		}

		if (o_add & BTO_STAGE) {
			VERBOSE("load ...\n");
			vaddstage(nodeid);
		}

		if (o_add & BTO_FROZEN) {
			VERBOSE("load frozen ...\n");
			vaddfrozen(nodeid);
		}
/*
 * Take all booting actions.
 */
		if (o_boot) {
			VERBOSE("boot ...\n");

			if (! opt_taken('N')) {
				if (rbtboot(nodeid, o_boot, ilink) < 0) {
					terror(me);
					kexit(errno);
				}
			}
		}
/*
 * Take all staging actions.
 */
		if (o_stage) {
			VERBOSE("stage ...\n");

			if (! opt_taken('N')) {
				if (rbtstage(nodeid, o_stage, neighbour) < 0) {
					terror(me);
					kexit(errno);
				}
			}
		}

		getnode(&index, &nodeid, &flags);
	} while (index > 0);
/*
 * Take spread action for all nodes at once.
 */
	if (o_spread) {
		VERBOSE("spread ...\n");
		pnodes = (int4 *) malloc((unsigned) nnodes * sizeof(int4));

		if (pnodes == (int4 *) 0) {
			perror(me);
			kexit(errno);
		}

		i = 0;

		do {
			pnodes[i] = nodeid;
			getnode(&index, &nodeid, &flags);
			i++;
		} while (index > 0);

		if (! opt_taken('N')) {
			if (rbtspread(pnodes, nnodes, o_spread) < 0) {
				terror(me);
				kexit(errno);
			}
		}
	}

	kexit(0);
	return(0);
}

/*
 *	vaddstrap
 *
 *	Functions:	- adds/loads bootstrap load module to remote node
 *			- notifies bootd
 *
 *	Accepts:	- remote node ID
 */

static void
vaddstrap(nodeid)

int4			nodeid;

{
	char		*full;		/* full pathname */
/*
 * Locate the bootstrap load module.
 */
	full = findpath(f_strap, p_prog, R_OK | X_OK);

	if (! full) {
		fprintf(stderr, "%s: can't find strap file - %s\n",
				me, f_strap);
		kexit(1);
	} else {
		f_strap = full;
		DBUG("%s: f_strap = \"%s\"\n", me, f_strap);
	}

	if (! opt_taken('N')) {
		if (rload(LOCAL, nodeid, getpid(), f_strap) < 0) {
			terror(me);
			kexit(errno);
		}

		DBUG("%s: %s loaded OK with tag %d\n", me, f_strap, getpid());

		if (rbtadd(nodeid, BTO_STRAP, getpid())) {
			terror(me);
			kexit(errno);
		}
	}
}

/*
 *	vaddfrozen
 *
 *	Functions:	- adds/loads frozen process schema to remote node
 *			- notifies bootd
 *
 *	Accepts:	- remote node ID
 */

static void
vaddfrozen(nodeid)

int4			nodeid;

{
	char		*full;		/* full pathname */
	struct stat	frozen;		/* frozen file status */
/*
 * Locate the frozen file.
 */
	full = findpath(f_frozen, p_conf, R_OK);

	if (! full) {
		fprintf(stderr, "%s: can't find frozen file - %s\n",
				me, f_frozen);
		kexit(1);
	} else {
		f_frozen = full;
		DBUG("%s: f_frozen = \"%s\"\n", me, f_frozen);
	}

	if (! opt_taken('N')) {
/*
 * Get the size of the frozen file.
 */
		if (stat(f_frozen, &frozen) < 0) {
			perror(me);
			kexit(errno);
		}

		if (rload(LOCAL, nodeid, getpid(), f_frozen) < 0) {
			terror(me);
			kexit(errno);
		}

		DBUG("%s: %s loaded OK with tag %d\n", me, f_frozen, getpid());

		if (rbtaddfrozen(nodeid, getpid(), (int4) frozen.st_size)) {
			terror(me);
			kexit(errno);
		}
	}
}

/*
 *	vaddstage
 *
 *	Functions:	- adds/loads staged program to remote node and
 *			  notifies bootd
 *
 *	Accepts:	- remote node ID
 */

static void
vaddstage(nodeid)

int4			nodeid;

{
	CONF		*p;		/* favourite ptr */
	CONF		*conf;		/* program list */
	char		*tail;		/* tail end of program pathname */
	char		*full;		/* full pathname */
	char		**largv;	/* ptr to pack arg vectors */
	int4		i;		/* favourite index */
	int4		tag;		/* flatd tag */
	int		n;		/* favourite counter */
	int		lsize;		/* total arg vector size */
	FILE		*fp_conf;	/* process schema file ptr */

	conf = CNFNULL;
/*
 * Locate the conf file.
 */
	full = findpath(f_conf, p_conf, R_OK);

	if (! full) {
		fprintf(stderr, "%s: can't find conf file - %s\n", me, f_conf);
		kexit(1);
	} else {
		f_conf = full;
	}
/*
 * Dump debugging information.
 */
	DBUG("%s: f_conf = \"%s\"\n", me, f_conf);
/*
 * Open the conf file.
 */
	fp_conf = fopen(f_conf, "r");

	if (! fp_conf) {
		fprintf(stderr, "%s: can't open conf file - %s\n", me, f_conf);
		perror(me);
		kexit(errno);
	}
/*
 * Allow -T to override conf file topology variable.
 */
	if (opt_taken('T')) {
		exttopv = (char **) 0;
		exttopc = 0;
		split(getparam('T'), &exttopv, &exttopc);
	}
/*
 * Parse the conf file.
 */
	cnfparse(fp_conf, &conf);
/*
 * Find programs with sufficient permissions.
 */
	if ((p = cnffind(p_prog, conf)) != CNFNULL) {
		fprintf(stderr, "%s: can't find %s with permissions\n",
			me, p->cnf_argv[0]);
		kexit(1);
	}

	if (fl_debug) {

		for (p = conf; p; p = p->cnf_next) {
			printf("%s: found %s\n", me, p->cnf_argv[0]);
		}
	}

	for (p = conf, n = 1; p; p = p->cnf_next, ++n) {
/*
 * Look for possible text segment sharing and if found, use the
 * shared module's tag, otherwise use the local program pointer
 * as a tag.
 */
		tag = 0;

		if (p->cnf_flags & BTSHARE) {
			tag = cnfshare(p, conf);
		}
/*
 * Pack and load argument vector.
 */
		largv = pack_argv(p->cnf_argc, p->cnf_argv, &lsize);

		if (! opt_taken('N')) {
			if (rflat(nodeid, (char *) largv, (int4) lsize,
				(int4) lsize, tag ? tag : (int4) p) < 0) {

				terror(me);
				kexit(errno);
			}
		}

		(void) free((char *) largv);
/*
 * Load the load module.
 */
		if (!tag) {
	    		tag = (int4) p;

			if (! opt_taken('N')) {
				if (rload(LOCAL, nodeid, tag, p->cnf_argv[0])
						< 0) {
					terror(me);
					kexit(errno);
				}
			}
		}

		p->cnf_tag = tag;
/*
 * Notify remote boot daemon.
 */
		if (! opt_taken('N')) {
			if (rbtaddstage(nodeid, tag, p->cnf_flags,
					p->cnf_stksize, p->cnf_priority) < 0) {
				terror(me);
				kexit(errno);
			}
		}

		if (fl_verbose) {
			tail = rindex(p->cnf_argv[0], STRDIR);
			tail = (tail) ? tail + 1 : p->cnf_argv[0];
			printf("[%2d] %7d %s", n, tag, tail);

			for (i = 1; i < p->cnf_argc; i++) {
	    			printf(" %s", p->cnf_argv[i]);
			}

			printf("\n");
		}
	}
}

/*
 *	setdefaults
 *
 *	Function:	- sets default files and paths
 */

static void
setdefaults()

{
/*
 * prefix paths
 */
	n_conf = 0;
	p_conf = (char **) 0;
	addarg(&n_conf, &p_conf, "");
	addarg(&n_conf, &p_conf, "$TROLLIUSHOME/boot");
	addarg(&n_conf, &p_conf, DEFPCONFIG);

	n_prog = 0;
	p_prog = (char **) 0;
	addarg(&n_prog, &p_prog, "");
	addarg(&n_prog, &p_prog, "$TROLLIUSHOME/itb/bin");
	addarg(&n_prog, &p_prog, DEFPPROGV);
/*
 * flags
 */
	fl_debug = opt_taken('d');;
	fl_verbose = opt_taken('v');
/*
 * files
 */
	if (opt_taken('c')) {
		f_conf = getparam('c');
	} else {
		f_conf = DEFFCONFIGV;
	}

	if (opt_taken('i')) {
		f_frozen = getparam('i');
	} else {
		f_frozen = DEFFFROZENV;
	}

	f_strap = DEFFSTRAP;
}

/*
 *	help
 *
 *	Function:	- prints helpful information on the vboot command
 */

static void
help()

{
static char *usage = " [-dhoqvN] [-{abefs}{xlptz}] [-c <config>] [-T <topology>]\n\t[-L <link>] [-n <neighbour>] [-i <frozen>]";
	printf("\nSynopsis:	hboot [options]\n");
	printf("\nDescription:	Boot an ITB node.\n");
	printf("\nOptions:	-h		Print this message.\n");
	printf("\t\t-d		Print debug information.\n");
	printf("\t\t-v		Be verbose.\n");
	printf("\t\t-c <conf>	Use <conf> as the config file.\n");
	printf("\t\t-i <frozen>	Use <frozen> as the frozen file.\n");
	printf("\t\t-N		Pretend to take actions.\n");
	printf("\t\t-T <topo>	Set $topology variable.\n");
	printf("\t\t-a[xlz]		Add/load object.\n");
	printf("\t\t-b[xlptz]	Boot object.\n");
	printf("\t\t-e[lpz]		Recursively boot.\n");
	printf("\t\t-f[xltz]	Flush object.\n");
	printf("\t\t-s[xltz]	Stage object.\n");
	printf("\t\t-q		Execute a quick boot.\n");
	printf("\t\t-o		Execute an old boot.\n");
	printf("\t\t-L <ilink>	Use link <ilink> to neighbour.\n");
	printf("\t\t-n <node>	Use neighbour node <node>.\n");
	mnusage();
}
