/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1988 Cornell University
 *
 *	$Log:	do_args.c,v $
 * Revision 2.2  92/04/30  14:20:51  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:28:30  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:31:20  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:56:53  trillium
 * Homo erectus
 * 
 *
 *	DO_ARGS  --  Process argc/argv in a general way.
 *
 *	MB, 	881219
 */


#include <terror.h>
#include <typical.h>

#define DO_TWOLETTER
#define PARAMS	8		/* hardwired limits */
#define N_SET	8
#define N_SCND	8

/*** public variables ***/

int	argtailc = 0;		/* count of unparsed args (after "--") */
char	**argtailv;		/* unparsed tail of argv */

/*** external variables ***/
extern int	errno;

/*** typedefines ***/

/* bitmap structure used for compact tables below */

typedef struct {
	long	lc;
	long	uc;
} abcmap;

/* struct of option-following parameter entries */

typedef struct {
	int	let;
	char	*ptr;
} param;

/* struct of option-following parameter entries */

typedef struct {
	int	let;			/* first letter */
	abcmap	menu;			/* valid completions */
	abcmap	let2;			/* actual completion(s) */
} scndst;

/*** private variables ***/

static abcmap bm_validopts = {0, 0};	/* allowed options */
static abcmap bm_followed  = {0, 0};	/* options followed by a parameter */
static abcmap bm_options   = {0, 0};	/* taken options */

#ifdef DO_TWOLETTER
static abcmap bm_twoletter = {0, 0};	/* these are two-letter options */
#endif

static int	n_set = 0;		/* count of exclusive sets */
static abcmap	set_array[N_SET];	/* array of exclusive sets */

static int	paramsc = 0;		/* count of following params */
static param	paramsv[PARAMS];	/* pointers to following params */

#ifdef DO_TWOLETTER
static int	scndc = 0;
static scndst	scndlet[N_SCND];	/* pointers to second letter */
#endif

/*** private subroutines ***/

/* pick the right bit */

static long
mapbit (let)
	int let;
{
	if (let == '0')			/* special "-<number>" option */
		return (2 << ('z'-'a'+1));
	if (let >= 'a' && let <= 'z')
		return (2 << (let - 'a'));
	if (let >= 'A' && let <= 'Z')
		return (1 | (2 << (let - 'A')));	/* lsb flags uc */
	errno = EUSAGE;
	return (ERROR);
}

/* set a bit in an abcmap */

static void
setbit (let, map)
	int	let;
	abcmap	*map;
{
	long	bit;

	bit = mapbit (let);
	if (bit < 0)  return;
	if (bit & 1)
		map->uc |= (bit & 0x0FFFFFFE);
	else
		map->lc |= bit;
}

/* clear a bit in an abcmap */

static void
clearbit (let, map)
	int	let;
	abcmap	*map;
{
	long	bit;

	bit = mapbit (let);
	if (bit < 0)  return;
	if (bit & 1)
		map->uc &= ~(bit & 0x0FFFFFFE);
	else
		map->lc &= ~bit;
}

/* test a bit in an abcmap */

static int
testbit (let, map)
	int	let;
	abcmap	*map;
{
	long	bit;

	bit = mapbit (let);
	if (bit < 0)  return (0);
	if (bit & 1)
		return ((map->uc & bit) != 0);
	return ((map->lc & bit) != 0);
}

/* count the number of bits set in a long */

static int
bitcount (word)
	long	word;
{
	int	count;

	count = 0;
	word &= 0x7FFFFFFF;	/* some machine might sign-extend... */
	while (word) {
		if (word & 1)  ++count;
		word >>= 1;
	}
	return (count);
}

/* check for the mutual exclusivity */

static int
excl_test()
{
	int	i;

	for (i=0; i<n_set; i++) {
		if (bitcount (bm_options.lc & set_array[i].lc)
		  + bitcount (bm_options.uc & set_array[i].uc) > 1) {
			errno = EUSAGE;
			return (ERROR);
		}
	}
	return (0);
}

#ifdef DO_TWOLETTER
/* set the second letter taken */
static int
pair (c, c2)
	int c, c2;
{
	int i;

	for (i=0; i<scndc; i++) {
		if (scndlet[i].let == c) {
			if (! testbit (c2, &scndlet[i].menu))
				return (0);	/* not a valid 2nd letter */
			setbit (c2, &scndlet[i].let2);
			return (1);
		}
	}
	return (0);				/* not found */
}
#endif

/*** public subroutines ***/

/* set option letters as exclusive, followed, or just plain valid */

/*
 *	If called as validopts("ab:c"); means 'b' is "followed".
 */
void
validopts (str)
	char	*str;
{
	int c;

	while ((c = *str++) != '\0') {
		setbit (c, &bm_validopts);
		if (c == '0' || *str == ':')
			setbit (c, &bm_followed);
		if (*str == ':')
			++str;
	}
}

void
exclusive (str)
	char	*str;
{
	int	c;
	abcmap	*set;

	if (n_set == N_SET)			/* no slots left */
		return;
	validopts (str);
	set = &set_array[n_set++];		/* new set */
	while ((c = *str++) != '\0') {
		if (c == ':')
			continue;
		setbit (c, set);
	}
}

#ifdef DO_TWOLETTER
/*
 *	Called as twoletter("xabc"); or twoletter("x:abc");
 *	meaning 'x' can be completed with 'a' or 'b' or 'c'.
 */
void
twoletter (str)
	char	*str;
{
	int c, d;

	if (scndc == N_SCND)		/* out of slots */
		return;
	d = *str++;
	setbit (d, &bm_validopts);
	setbit (d, &bm_twoletter);
	scndlet[scndc].let = d;
	if (*str == ':')
		++str;
	while ((c = *str++) != '\0') {
		if (c == ':' && *str == '\0')
			break;
		setbit (c, &scndlet[scndc].menu);
	}
	if (c == ':')
		setbit (d, &bm_followed);
	++scndc;
}
#endif

void
followed (str)
	char	*str;
{
	int c;

	while ((c = *str++) != '\0') {
		setbit (c, &bm_validopts);
		setbit (c, &bm_followed);
	}
}

/* test whether an option was taken */

int
opt_taken (let)
	int	let;
{
	return (testbit (let, &bm_options));
}

/* find which, if any, of the exclusive set including 'o' was chosen */

int
opt_chosen (o)
	int o;
{
	int  c, i;

	for (i=0; i<n_set; i++) {
		/* find a set that includes 'o': */
		if (! testbit (o, &set_array[i]))
			continue;
		/* for each 'c' in the set, see if opted: */
		for (c='a';  c!=0;  c = (c=='Z'? 0 : (c=='z'? 'A' : c+1))) {
			if (! testbit (c, &set_array[i]))
				continue;
			if (testbit (c, &bm_options))
				return (c);
		}
	}
	return (ERROR);		/* none opted */
}

/* get the pointer to a following parameter */

#define NOPARAM ((char *) 0)

char *
getparam (let)
	int	let;
{
	int	ii = 0;
	int	i;

	if (! opt_taken (let))
		return (NOPARAM);		/* option not taken */
	if (! testbit (let, &bm_followed))
		return (NOPARAM);		/* option not followed */
	for (i=ii; i<paramsc; i++) {
		if (paramsv[i].let == let) {
			ii = i+1;
			return (paramsv[i].ptr);
		}
	}
	return (NOPARAM);			/* not found */
}

#ifdef DO_TWOLETTER

/* get the second letter taken */

#define NOLET2	(-1)
int
get_2nd_let (let)
	int	let;
{
	int	i, c;

	if (! opt_taken (let))
		return (NOLET2);		/* option not taken */
	if (! testbit (let, &bm_twoletter))
		return (NOLET2);		/* option not two-letter */
	for (i=0; i<scndc; i++) {
		if (scndlet[i].let != let)	/* not in this set */
			continue;
		for (c='a';  c!=0;  c = (c=='Z'? 0 : (c=='z'? 'A' : c+1))) {
			if (testbit (c, &scndlet[i].let2)) {
				clearbit (c, &scndlet[i].let2);
				return (c);
			}
		}
	}
	return (NOLET2);			/* not found */
}
#endif  /* DO_TWOLETTER */

/* convert a string into a number, using either base 10 or 16 */

int
xtoi (cpp, ip)
	char **cpp;
	int  *ip;
{
	int  c, base, sign;
	char *cp;

	cp = *cpp;

	if (cp == 0) {
		errno = EUSAGE;
		return (ERROR);
	}
	sign = 1;
	if (*cp == '-') {
		sign = (ERROR);
		++cp;
	}
	base = 10;
	if (cp[0] == '0' && (cp[1] == 'x' || cp[1] == 'X')) {
		base = 16;
		cp += 2;
	}
	if (*cp == 0) {
		errno = EBADASCIINUMB;
		return (ERROR);
	}
	*ip = 0;
	while (((c=(*cp)) != '\0') && (c != '-') && (c != ',')) {
		if (base == 10 && ! (c>='0' && c<='9')) {
			errno = EBADASCIINUMB;
			return (ERROR);
		}
		if (base == 16) {
			if (c >= 'a' && c <= 'f')
				c = (c-'a') + '9' + 1;
			else if (c >= 'A' && c <= 'F')
				c = (c-'A') + '9' + 1;
			else if (! (c>='0' && c<='9')) {
				errno = EBADASCIINUMB;
				return (ERROR);
			}
		}
		*ip = (*ip) * base  +  c - '0';
		cp++;
	}
	*cpp = cp;	/* advance caller's pointer */
	*ip *= sign;
	return (0);
}

/* get a following numerical parameter if applicable */

int
intparam (c, p)
	int c, *p;
{
	char	*cp;

	cp = getparam(c);
	return (xtoi (&cp, p));
}

/*
 * Main argc/argv processor
 *
 *	The abcmaps must be set up before calling this function!
 */

int
do_args (argcp, argv)
	int	*argcp;
	char	*argv[];
{
	int	c, c2;
	char	*cp;
	int	argc;
	int	othersc;
	char	**othersv;

	othersc = 1;			/* pass argv[0] through */
	othersv = argv;
	argc = *argcp;
	--argc;				/* skip argv[0] in processing */
	++argv;

	while (argc > 0) {
		cp = argv[0];
		if (*cp++ == '-') {			/* an option */
			c = *cp++;
			if (c=='-' && *cp=='\0') {	/* "--" */
				argtailc = argc - 1;
				argtailv = argv + 1;
				break;			/* rest unparsed */
			}
			if (c >= '0' && c <= '9'	/* "-<number>"   */
					|| c == '-') {	/*   negative    */
				c = '0';		/* simulate "-0" */
				--cp;			/* point to number */
			}
			while (c != '\0') {
				if (! testbit (c, &bm_validopts)) {
					errno = EUSAGE;
					return (ERROR);
				}
				setbit (c, &bm_options);
#ifdef DO_TWOLETTER
				if (testbit (c, &bm_twoletter)) {
					if ((c2 = *cp) != '\0')
						++cp;
					if (! pair (c, c2)) {
						errno = EUSAGE;
						return (ERROR);
					}
				}
#endif
				if (testbit (c, &bm_followed))
					break;
				c = *cp++;		/* allow '-abc' */
			}
			if (c != '\0' && paramsc < PARAMS) {
				if (*cp == '\0') {	/* followed w/space */
					if (argc == 1) {	/* last arg */
						errno = EUSAGE;
						return (ERROR);
					}
					--argc;
					++argv;
					paramsv[paramsc].ptr = argv[0];
				} else {
					paramsv[paramsc].ptr = cp;
				}
				paramsv[paramsc].let = c;
				++paramsc;
			}
		} else {
			/* current argv[] element overwrites earlier one: */
			othersv[othersc] = argv[0];
			++othersc;
		}
		--argc;
		++argv;
	}

	*argcp = othersc;
	return (excl_test());
}

