/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB/MB
 *
 *	Function:	- parse node specifiers from an argument list
 *			- generic machine independent version
 *			- based on the model created by MB
 */


#include <errno.h>
#include <stdio.h>
#include <typical.h>
#include <net.h>
#include "range.h"

/*
 * external variables
 */
extern int		errno;

/*
 * external functions
 */
extern char		*malloc();
extern int		addentry();
extern int		getnodeid();
extern int		getrange();
extern void		getentry();


/*
 * local variables
 */
static struct list_el	*node_head = 0,	/* head of node list */
			*node_tail;	/* tail of node list */



/*
 *	parse_nodes
 *
 *	Functions:	- search for and parse node identifier mneumonics
 *			- basic strategy is to bump a secondary pointer
 *			  and counter whenever a non-identifier is found
 *			- arguments are value returned
 *			- argc, argv are undefined in case of ERROR
 *
 *	Accepts:	- argument string pointer and counter
 *
 *	Returns:	- 0 or ERROR
 */

int
parse_nodes (argc, argv)

int		*argc;
char		**argv;

{
	int	othersc;		/* left over count */
	char	**othersv;		/* left over arguments */


/*
 * Pass argument 0 through.
 */
	othersc = 1;
	othersv = argv++;
	++othersv;
	--*argc;
/*
 * Loop through all the arguments.
 */
	while (*argc) {
/*
 * Is it a generic node identifier?
 */
		if (**argv == 'n') {
/*
 * Parsing node identifiers may bomb due to lack of memory or simply syntax.
 */
			if (getrange(*argv + 1, 0, &node_head, &node_tail)
				== ERROR) {

				if (errno == ENOMEM) {
					return(errno);
				}
/*
 * Anything that starts as a nodeid but fails due to syntax is assumed to
 * be a valid non-nodeid argument.
 */
				else {
					*othersv++ = *argv;
					othersc++;
				}
			}
		}
/*
 * Is it the local host identifier (h)?
 */
		else if ((**argv == 'h') && (*(*argv + 1) == '\0')) {

			if (addentry(LOCAL, 0, &node_head, &node_tail)
				== ERROR) {

				return(errno);
			}
		}
/*
 * Is it a multicast destination?
 */
		else if ((**argv == 'H') && (*(*argv + 1) == '\0')) {

			if (addentry(HOST2OTB, NT_CAST, &node_head, &node_tail)
				== ERROR) {

				return(errno);
			}
		}

		else if ((**argv == 'C') && (*(*argv + 1) == '\0')) {

			if (addentry(HOST2ITB, NT_CAST, &node_head, &node_tail)
				== ERROR) {

				return(errno);
			}
		}

		else if ((**argv == 'N') && (*(*argv + 1) == '\0')) {

			if (addentry(HOST2ALL, NT_CAST, &node_head, &node_tail)
				== ERROR) {

				return(errno);
			}
		}
/*
 * Bump the non-nodeid pointer and counter.
 */
		else {
			*(othersv++) = *argv;
			++othersc;
		}

		--*argc;
		++argv;
	}

	*argc = othersc;
	return (0);
}

/*
 *	getnode
 *
 *	Function:	- get one node specifier from top of list
 *			- rotates it to bottom of list
 *			- If none in list, returns NOTNODE as the nodeid and
 *			  ERROR as the index.
 *
 *	Accepts:	- value returned index pointer
 *			- value returned node ID pointer
 *			- value returned flags pointer
 */

void
getnode (ixp, idp, flagsp)
	int *ixp, *idp, *flagsp;
{
	getentry(ixp, idp, flagsp, &node_head, &node_tail);
}


/*
 *	mnemonic (GDB)
 *
 *	Function:	- translate a node ID into a mnemonic string
 *
 *	Accepts:	- node identifier
 *
 *	Returns:	- mnemonic string pointer
 */

char *
mnemonic (node)

int			node;

{
	static char str[16];		/* more than 16 long ain't mnemonic! */


	if (node == NOTNODE) {
		str[0] = '?';
		str[1] = '\0';
	}

	else if (node == LOCAL) {
		strcpy(str, "local");
	}

	else if (node == HOST2OTB) {
		strcpy(str, "H");
	}

	else if (node == HOST2ITB) {
		strcpy(str, "C");
	}

	else if (node == HOST2ALL) {
		strcpy(str, "N");
	}

	else if (node == getorigin()) {
		str[0] = 'h';
		str[1] = '\0';
	}

	else {
		sprintf(str, "n%d", node);
	}

	return (str);
}

/* call this for a nodespec usage message */

void
mnusage()
{
	printf ("\nNodes:		n<list> (any node)\n");
	printf ("\t\t\twhere <list>  = <range>[,<range>]\n");
	printf ("\t\t\tand   <range> = <constant>[-<constant>]\n");
	printf ("\t\t\tand   <constant> = decimal or hexadecimal (0x)\n");
}
