/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB/MB
 *
 *	Function:	- parse process specifiers from an argument list
 *			- generic machine independent version
 *			- based on the node model created by MB
 */


#include <errno.h>
#include <stdio.h>
#include <preq.h>
#include <typical.h>
#include "range.h"

/*
 * external variables
 */
extern int		errno;

/*
 * external functions
 */
extern char		*malloc();
extern int		addentry();
extern int		getrange();
extern void		getentry();


/*
 * local variables
 */
static struct list_el	*proc_head = 0,	/* head of process list */
			*proc_tail;	/* tail of process list */



/*
 *	parse_procs
 *
 *	Functions:	- search for and parse process identifier mneumonics
 *			- basic strategy is to bump a secondary pointer
 *			  and counter whenever a non-identifier is found
 *			- arguments are value returned
 *			- argc, argv are undefined in case of ERROR
 *
 *	Accepts:	- argument string pointer and counter
 *
 *	Returns:	- 0 or ERROR
 */

int
parse_procs (argc, argv)

int		*argc;
char		**argv;

{
	int	proc_select;		/* process selection method */
	int	othersc;		/* left over count */
	char	**othersv;		/* left over arguments */

/*
 * Pass argument 0 through.
 */
	othersc = 1;
	othersv = argv++;
	++othersv;
	--*argc;
/*
 * Loop through all the arguments.
 */
	while (*argc) {
		proc_select = 0;
/*
 * Is it a process ID specification?
 */
		switch (**argv) {
		case 'p':
			proc_select = SELECT_PID;
			break;
		case '%':
			proc_select = SELECT_INDEX;
			break;
		case '@':
			proc_select = SELECT_LOADPT;
			break;
		case '#':
			proc_select = SELECT_BLKD;
			break;
		}

		if (proc_select) {
/*
 * Parsing process specifiers may bomb due to lack of memory or simply syntax.
 */
			if (getrange(*argv + 1, proc_select, &proc_head,
				&proc_tail) == ERROR) {

				if (errno == ENOMEM) {
					return(errno);
				}
/*
 * Anything that starts as a process but fails due to syntax is assumed to
 * be a valid non-process argument.
 */
				else {
					*(othersv++) = *argv;
					++othersc;
				}
			}
		}
/*
 * Bump the non-process pointer and counter.
 */
		else {
			*(othersv++) = *argv;
			++othersc;
		}

		--*argc;
		++argv;
	}

	*argc = othersc;
	return (0);
}

/*
 *	getproc
 *
 *	Function:	- get one process specifier from top of list
 *			- rotates it to bottom of list
 *			- If none in list, returns NOTPROCESS as the process
 *			  value and ERROR as the index.
 *
 *	Accepts:	- value returned index pointer
 *			- value returned process value pointer
 *			- value returned flags pointer
 */

void
getproc (ixp, idp, flagsp)
	int *ixp, *idp, *flagsp;
{
	getentry(ixp, idp, flagsp, &proc_head, &proc_tail);
}
