/*
 *	Cornell Theory Center				Trillium
 *
 *	Copyright 1988 Cornell University
 *
 *	MB/GDB
 *
 *	$Log:	range.c,v $
 * Revision 2.2  92/04/30  14:20:59  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:29:02  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:31:25  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:56:57  trillium
 * Homo erectus
 * 
 *
 *	Function:	- portable range parser and list manager
 */


#include <terror.h>
#include <typical.h>
#include "range.h"

/*
 * external variables
 */
extern int		errno;

/*
 * external functions
 */
extern char		*malloc();
extern int		xtoi();


/*
 *	addentry
 *
 *	Function:	- add a node to the list
 *
 *	Accepts:	- node identifier
 *			- node description flags
 *
 *	Returns:	- 0 or ERROR
 */

int
addentry (nodeid, flags, head, tail)
	int 		nodeid, flags;
	struct list_el	**head, **tail;
{
	struct list_el *lep;

	lep = (struct list_el *) malloc (sizeof (struct list_el));
	if (lep == 0)
		return (ERROR);
	if (*head == 0) {		/* first in list */
		*head = lep;
		lep->ns_index = 0;
	} else {
		(*tail)->ns_next = lep;
		lep->ns_index = (*tail)->ns_index + 1;
	}
	*tail		= lep;
	lep->ns_next	= 0;
	lep->ns_value	= nodeid;
	lep->ns_flags	= flags;
	return (0);
}


/*
 *	getrange
 *
 *	Function:	- parse the range specification format
 *			- example: 1,3-5,7
 *
 *	Accepts:	- pointer to specification string
 *			- flags to use with every value found
 *			- ptr to head of list ptr
 *			- ptr to tail of list ptr
 *
 *	Returns:	- 0 or ERROR
 */

int
getrange (cp, flags, head, tail)

char			*cp;
int			flags;
struct list_el		**head, **tail;

{
	int	c, n, m, oldn;

	m = NEW;
	while ((c=(*cp)) != '\0') {
		if (c == ',') {
			if (m == NEW) {
				errno = EBADNODESPEC;
				return (ERROR);
			}
			++cp;
			m = NEW;
			continue;
		}
		if (c == '-') {
			if (m != ONE) {
				errno = EBADNODESPEC;
				return (ERROR);
			}
			++cp;
			m = MANY;
			continue;
		}
		if (xtoi (&cp, &n) < 0)
			return (ERROR);		   /* bad number */
		if (m == NEW) {
			if (addentry (n, flags, head, tail) < 0)
				return (ERROR);
			oldn = n;
			m = ONE;
		} else if (m == MANY) {
			if (n <= oldn) {
				errno = EBADNODESPEC;
				return (ERROR);
			}
			while (++oldn <= n) {
				if (addentry (oldn, flags, head, tail) < 0)
					return (ERROR);
			}
		} else {
			errno = EBADNODESPEC;
			return (ERROR);
		}
	}					/* end of while() */
	return (0);
}


/*
 *	getentry
 *
 *	Function:	- get one entry from top of list
 *			- rotates it to bottom of list
 *			- If none in list, returns 0 as the value and
 *			  ERROR as the index.
 *
 *	Accepts:	- value returned index pointer
 *			- value returned list entry value pointer
 *			- value returned flags pointer
 *			- ptr to head of list ptr
 *			- ptr to tail of list ptr
 */

void
getentry (ixp, idp, flagsp, head, tail)
	int 		*ixp, *idp, *flagsp;
	struct list_el	**head, **tail;
{
	struct list_el *lep;

	lep = *head;
	if (lep == 0) {			/* empty list */
		*ixp = ERROR;
		*idp = 0;
		return;
	}

	if (lep->ns_next) {		/* not only one in list */
		*head = lep->ns_next;
		(*tail)->ns_next = lep;
		*tail = (*tail)->ns_next;
		lep->ns_next = 0;
	}

	*ixp    = lep->ns_index;	/* copy data to caller */
	*idp    = lep->ns_value;
	*flagsp = lep->ns_flags;
	return;
}
