/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	Function:	- common routines to all process schema commands
 */

#include <bootreq.h>
#include <conf.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <typical.h>
#include <unistd.h>

/*
 * globally defined functions
 */
int			cnfexec();
CONF			*cnffind();
void			cnfprint();
int4			cnfshare();

/*
 * external variables
 */
extern int		errno;

/*
 * external functions
 */
extern char		*findpath();

/*
 *	cnfexec
 *
 *	Function:	- forks, execs and waits for a subordinate program
 *
 *	Accepts:	- argument vector, argv[0] is program
 *
 *	Returns:	- status code or ERROR
 */

int
cnfexec(argv)

char			*argv[];

{
	int		status;		/* exit status */
	int		pid;		/* whoami? */

	if ((pid = fork()) < 0) {
		return(ERROR);
    	}

	else if (pid == 0) {		/* child */
		execvp(argv[0], argv);
		return(ERROR);
    	}

	else {				/* parent */

		while ((wait(&status) != 0) && (! WIFEXITED(status)));
	}

	return(WEXITSTATUS(status));
}

/*
 *	cnffind
 *
 *	Function:	- expands each argv[0] to full pathname
 *
 *	Accepts:	- argv style structure of program directory pathnames
 *			- ptr to process schema program list
 *
 *	Returns:	- CNFNULL or ptr to program that can't be located
 */

CONF *
cnffind(p_prog, p)

char			**p_prog;
CONF			*p;

{
	char		*full;		/* full pathname */
/*
 * Loop through all the programs in the parsed process schema.
 */
	while (p) {

		if (p->cnf_argv != (char **) 0) {
			full = findpath(p->cnf_argv[0], p_prog, R_OK | X_OK);

			if (! full) {
				return(p);
			} else {
				p->cnf_argv[0] = full;
			}
		}

		p = p->cnf_next;
	}

	return(CNFNULL);
}

/*
 *	cnfprint
 *
 *	Function:	- print the process schema
 *
 *	Accepts:	- ptr to process schema program list
 */

void
cnfprint(p)

CONF			*p;

{
	int		i;		/* favourite index */

	while (p) {

		for (i = 0; i < p->cnf_argc; ++i) {
			printf("%s ", p->cnf_argv[i]);
		}

		printf("{\n");
		printf("\tstack(%x);\n", p->cnf_stksize);
		printf("\tdelay(%x);\n", p->cnf_delay);
		printf("\tpriority(%x);\n", p->cnf_priority);
		printf("\tshare(%x);\n", (p->cnf_flags & BTSHARE) ? 1 : 0);
		printf("\tonchip(%x);\n", (p->cnf_flags & BTONCHIPSTK) ? 1 : 0);
		printf("}\n");

		p = p->cnf_next;
	}
}

/*
 *	cnfshare
 *
 *	Function:	- traverses the program list to find any possible
 *			  shared texts
 *			- Two programs are considered to have the same text
 *			  if they have identical argv[0]'s.
 *	
 *	Accepts:	- ptr to program for which a match is sought
 *			- top of program list ptr
 *
 *	Returns:	- tag of shared text or 0
 */

int4
cnfshare(key, conf)

CONF			*key;
CONF			*conf;

{
	CONF		*p;		/* favourite potr */

	for (p = conf; p != key; p = p->cnf_next) {

		if (strcmp(p->cnf_argv[0], key->cnf_argv[0]) == 0) {
			return(p->cnf_tag);
		}
	}

	return(0);
}
