%{
/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB/AKP
 *
 *	$Log:	confl.l,v $
 * Revision 2.2  92/04/30  14:23:14  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/24  17:30:17  gdburns
 * Disallow C-style comments.
 * 
 * Revision 2.1.1.1  92/02/23  22:24:15  gdburns
 * toolkit layer overhaul
 * 
 * Revision 2.1  91/03/20  11:39:33  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:30:25  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/06/01  16:55:15  gdburns
 * Homo erectus
 *
 *	Function:	- tokenizes input stream for process schema parser
 *
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#include <portable.h>
#include <stdio.h>
#include <y.conf.h>

/*
 * external functions
 */
extern int4		stoi();			/* string to integer */

/*
 * keyword table
 */
struct keyword {
	char		*key_str;		/* identifier */
	int		key_val;		/* token value */
};

struct keyword		keytable[] = {
	{"topology",	T_TOPOLOGY},
	{"stack",	T_STACK},
	{"priority",	T_PRIORITY},
	{"delay",	T_DELAY},
	{"share",	T_SHARE},
	{"onchip",	T_ONCHIP},
	{ (char *) 0,	T_IDENTIFIER}
};

YYSTYPE			yylval;
%}

DIGIT			[0-9]
XDIGIT			[0-9A-Fa-f]
ALPHA			[A-Za-z]
WHITE			[ \t\n]
NOTSPECIAL		[^ \t\n;(){}"#$]

%%
";"			return(T_SEMI);
"("			return(T_LPAREN);
")"			return(T_RPAREN);
"{"			return(T_LBRACE);
"}"			return(T_RBRACE);
"$"			return(T_DOLLAR);

\"[^"\n]*\"	{	/* strings */
			yytext[yyleng - 1] = '\0';
			yylval.sval = yytext + 1;
			return(T_STRING);
		}

{ALPHA}+	{	/* alphabetic keywords and identifiers */
			yylval.sval = yytext;
			return(keylookup(yytext));
		}

{DIGIT}+/[ \t]*\)	{	/* decimal numbers */
			yylval.ival = stoi(yytext);
			return(T_NUMBER);
		}

0[xX]{XDIGIT}+/[ \t]*\) {	/* hexadecimal numbers */
			yylval.ival = stoi(yytext);
			return(T_NUMBER);
		}

{NOTSPECIAL}+	{	/* stranger identifiers */
			yylval.sval = yytext;
			return(T_IDENTIFIER);
		}

{WHITE}+	;	/* swallow white space */

"#"[^\n]*\n	;	/* swallow pound comments */

%%

/*
 *	keylookup
 *
 *	Function:	- looks up identifier in keyword table
 *	Accepts:	- identifier string
 */

int
keylookup(s)

char			*s;

{
	int		i = 0;		/* favourite index */

	while (keytable[i].key_str && strcmp(s, keytable[i].key_str)) {
		i++;
	}

	return(keytable[i].key_val);
}

/*
 *	yyerror
 *
 *	Function:	- prints error message and bails out
 *	Accepts:	- error message
 */

void
yyerror(s)

char			*s;

{
	fprintf(stderr, "%s on line %d: \"%s\"\n", s, yylineno, yytext);
	exit(1);
}
