%{
/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	confy.y,v $
 * Revision 2.2  92/04/30  14:23:15  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/02/23  22:24:17  gdburns
 * toolkit layer overhaul
 * 
 *	Function:	- parses process schema
 */

#include <bootreq.h>
#include <conf.h>
#include <stdio.h>

/*
 * external functions
 */
extern char		*malloc();
extern void		yyerror();

/*
 * local variables
 */
static int		defdelay;	/* default delay after startup */
static int		defpriority;	/* default priority */
static int		defstksize;	/* default stack size */
static int		defflags;	/* default booting flags */

static int		argc;		/* length of argument vector */
static char		**argv;		/* current program's argument vector */
static int		topc;		/* length of topology vector */
static char		**topv;		/* conf defined topology vector */

static CONF		*new;		/* new program */
static CONF		*cnflast;	/* bottom of process schema list */
static CONF		**cnffirst;	/* top of process schema list */

/*
 * globally defined variables
 */
int			exttopc = 0;
char			**exttopv = (char **) 0;
					/* value of $topology, possibly set
					   in another module */
/*
 * external variables
 */
extern FILE		*yyin;		/* input stream */

/*
 *	cnfparse
 *
 *	Function:	- parses process schema
 *	Accepts:	- input stream ptr
 *			- ptr to ptr to parsed process schema list
 *			  (value returned)
 */

void
cnfparse(f, first)

FILE			*f;
CONF			**first;

{
	yyin = f;
	cnffirst = first;
	cnflast = CNFNULL;

	defpriority = CNFPRIORITY;
	defdelay = CNFDELAY;
	defstksize = CNFSTKSIZE;
	defflags = CNFFLAGS;
	argc =  0;
	argv = (char **) 0;

	yyparse();
}

%}

/*
 * special character tokens
 */
%token <ival>	T_SEMI
%token <ival>	T_LPAREN
%token <ival>	T_RPAREN
%token <ival>	T_LBRACE
%token <ival>	T_RBRACE
%token <ival>	T_DOLLAR

/*
 * global variable tokens
 */
%token <ival>	T_TOPOLOGY

/*
 * option tokens
 */
%token <ival>	T_STACK
%token <ival>	T_PRIORITY
%token <ival>	T_DELAY
%token <ival>	T_SHARE
%token <ival>	T_ONCHIP

/*
 * miscellaneous tokens
 */
%token <sval>	T_IDENTIFIER
%token <sval>	T_STRING
%token <ival>	T_NUMBER

/*
 * non-terminal types
 */
%type <ival>	setpriority
%type <ival>	setdelay
%type <ival>	setshare
%type <ival>	setonchip
%type <ival>	setstksize

/*
 * token value template
 */
%union {
	int4	ival;
	char	*sval;
}

/*
 * start symbol
 */
%start conf

%%

conf:		/* empty */
	|	conf line
	;

line:		global_var T_SEMI
	|	default T_SEMI
	|	program
        ;

global_var:	settopology
	;

settopology:	T_TOPOLOGY T_LPAREN T_STRING
		{
			topc =  0;
			topv = (char **) 0;
			split($3, &topv, &topc);
		}
		T_RPAREN
	;

default:	setpriority
		{ defpriority = $1; }
	|	setdelay
		{ defdelay = $1; }
	|	setshare
		{
			if ($1) defflags |= BTSHARE;
			else defflags &= ~BTSHARE;
		}
	|	setonchip
		{
			if ($1) defflags |= BTONCHIPSTK;
			else defflags &= ~BTONCHIPSTK;
		}
	|	setstksize
		{ defstksize = $1; }
	;

program:	args
	{
	/*
	 * Allocate a new program desc.
	 */
		new = (CONF *) malloc((unsigned) sizeof(CONF));

		if (!new) {
		    yyerror("malloc: no space for pathname");
		}

		new->cnf_priority = defpriority;
		new->cnf_flags = defflags;
		new->cnf_delay = defdelay;
		new->cnf_stksize = defstksize;
	/*
	 * Copy argv.
	 */
		new->cnf_argc = argc;
		new->cnf_argv = argv;
		argc = 0;
		argv = (char **) 0;
	}
		optionlist
	{
	/*
	 * Append new program to end of the pschema list.
	 */
		new->cnf_next = CNFNULL;

		if (cnflast) {
			cnflast->cnf_next = new;
		} else {
			*cnffirst = new;
		}

		cnflast = new;
	}
	;

args:		arg
	|	args arg

arg:		T_IDENTIFIER
		{
			addarg(&argc, &argv, $1);
		}
	|	T_DOLLAR T_TOPOLOGY
		{
			int		i;
			int		ac;
			char		**av;

			if (exttopc > 0) {
				ac = exttopc;
				av = exttopv;
			} else {
				ac = topc;
				av = topv;
			}

			for (i = 0; i < ac; ++i) {
				addarg(&argc, &argv, av[i]);
			}
		}
	;

optionlist:	T_SEMI
	|	T_LBRACE options T_RBRACE
	;

options:	/* empty */
	|	options option
	;

option:		setstksize T_SEMI
		{ new->cnf_stksize = $1; }
	|	setpriority T_SEMI
		{ new->cnf_priority = $1; }
	|	setdelay T_SEMI
		{ new->cnf_delay = $1; }
	|	setshare T_SEMI
		{
			if ($1) {
				new->cnf_flags |= BTSHARE;
			} else {
				new->cnf_flags &= ~BTSHARE;
			}
		}
	|	setonchip T_SEMI
		{
			if ($1) {
				new->cnf_flags |= BTONCHIPSTK;
			} else {
				new->cnf_flags &= ~BTONCHIPSTK;
			}
		}
	;

setstksize:	T_STACK T_LPAREN T_NUMBER T_RPAREN
		{ $$ = $3; }
	;

setdelay:	T_DELAY T_LPAREN T_NUMBER T_RPAREN
		{ $$ = $3; }
	;

setpriority:	T_PRIORITY T_LPAREN T_NUMBER T_RPAREN
		{ $$ = $3; }
	;

setshare:	T_SHARE T_LPAREN T_NUMBER T_RPAREN
		{ $$ = $3; }
	;

setonchip:	T_ONCHIP T_LPAREN T_NUMBER T_RPAREN
		{ $$ = $3; }
	;
