/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD
 *
 *	$Log:	nail.c,v $
 * Revision 2.2  92/04/30  14:23:18  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  91/11/26  14:00:15  raja
 * Modified to use libtools.
 * 
 * Revision 2.1.1.1  91/06/05  21:39:18  raja
 * Enhanced portability: use an (int4) cast on the TNODEINCR constant.
 * 
 * Revision 2.1  91/03/20  11:30:06  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/11/16  15:23:02  raja
 * Overhaul of the topology layer tools.
 * 
 * Revision 2.0.1.1  90/07/13  19:30:41  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:56:29  trillium
 * Homo erectus
 *
 *	Function:	- functions shared by the NaIL tools
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 */

#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <portable.h>
#include <typical.h>
#include <net.h>
#include <events.h>
#include <nail.h>
#include <all_queue.h>

/*
 * external variables
 */
extern char		*me;			/* program name */

/*
 * static functions
 */
static void		nail_error();
static int		add_splink();


/*
 *	print_bootorder
 *
 *	Function:	- print the sequential booting order of nodes
 *
 *	Accepts:	- ptr to the boot schema
 */
void
print_bootorder(nail)

struct bschema		*nail;

{
	struct t_node	*tnode;		/* ptr to node */
/*
 * Check the trivial cases.
 */
	if ((tnode = nail->bs_hostnode) == (struct t_node *) 0) {
		nail_error("print_bootorder", "no hostnode declared");
		return;
	}
/*
 * Loop through the linked list of nodes to boot.
 */
	for ( ; tnode; tnode = (tnode->tn_specs)->ns_bootnext) {
		printf("node booted: %d  with %d link(s) out\n",
			tnode->tn_nodeid, (tnode->tn_specs)->ns_nspanout);
	}
}

/*
 *	print_spantree
 *
 *	Function:	- print the booting spanning tree of the system
 *
 *	Accepts:	- ptr to the boot schema
 */
void 
print_spantree(nail)

struct bschema		*nail;

{
	struct t_node	*tnode;		/* ptr to tnode entries */
	struct t_node	**ptnode;	/* ptr to tnode pointer */
	struct nspecs	*specs;		/* ptr to node specifications */
	struct link	*link;		/* ptr to link element */
	QUEUE		*qd;		/* ptr to queue descriptor */
	int		i;		/* favorite counter */
/*
 * Set up a buffering queue used for breadth-first graph navigation.
 */
	qd = aq_init((int4) TNODEINCR, (int4) sizeof(tnode));
	if (qd == (QUEUE *) 0) {
		nail_error("print_spantree", "cannot create a queue");
		return;
	}
	if ((tnode = nail->bs_hostnode) == (struct t_node *) 0) {
		return;
	}
	if (aq_insert(qd, &tnode)) {
		nail_error("print_spantree", "cannot enqueue first element");
		aq_free(qd);
		return;
	}
/*
 * Print out the spanning tree from each node in the queue.
 * Keep looping until the queue is empty.
 */
	while (aq_count(qd) > INT4_NIL) {
/*
 * Get the next node in the queue.
 */
		if ((ptnode = (struct t_node **) aq_find(qd))
			== (struct t_node **) 0) {

			nail_error("print_spantree", "no queue element found");
			aq_free(qd);
			return;
		}

		tnode = *ptnode;
		if (aq_delete(qd)) {
			nail_error("print_spantree", "cannot dequeue");
			aq_free(qd);
			return;
		}

		specs = tnode->tn_specs;
		printf("node booted: %d, children nodes: {", tnode->tn_nodeid);
/*
 * Put all its spanning tree children in the queue.
 */
		link = specs->ns_lkspan;
		for (i = 0; i < specs->ns_nspanout; i++) {
			tnode = (link->lk_link)->tl_destnode;
			if (i == 0) {
				printf(" %d", tnode->tn_nodeid);
			}
			else {
				printf(", %d", tnode->tn_nodeid);
			}
			if (aq_shove(qd, &tnode)) {
				nail_error("print_spantree",
					"cannot shove a node child");
				aq_free(qd);
				return;
			}
			link = link->lk_next;
		}
		printf(" }\n");
	}
}

/*
 *	pretty_bnail
 *
 *	Function:	- pretty print the boot schema
 *
 *	Accepts:	- ptr to the boot schema
 */
void 
pretty_bnail(nail)

struct bschema		*nail;

{
	int4		i;		/* index */
	int		cnt;		/* elements per line counter */
	int		flag;		/* TRUE = print a ';' at the end */
	struct t_node	*tnode;
	struct t_link	*tlink;
	struct t_route	*troute;
	struct t_order	*torder;
	struct nspecs	*specs;
	HASH		*tnhd;

	printf("\n/* boot schema */\n\n");
/*
 * Print the TWEIGHT tuple.
 */
	printf("TWEIGHT = <%d, %d, %d, %d, %d>;\n\n",
		(nail->bs_tweight).tw_link, (nail->bs_tweight).tw_host,
		(nail->bs_tweight).tw_order, (nail->bs_tweight).tw_mcast,
		(nail->bs_tweight).tw_hop);
/* 
 * Print the TNODE tuples.
 */
	printf("/* node enumeration */\n");
	cnt = 0;
	flag = FALSE;
	tnhd = nail->bs_tnhd;
	tnode = (struct t_node *) ah_top(tnhd);
	while (tnode) {
		if (cnt == 0) {
			printf("TNODE =");
			flag = TRUE;   
		}
		printf(" <%d, %d>", tnode->tn_nodeid,
			(tnode->tn_specs)->ns_type);
		if (cnt == (TNODEPERLINE-1)) {
			printf(";\n");
			flag = FALSE;
		}
		cnt = ((cnt + 1) % TNODEPERLINE);
		tnode = (struct t_node *) ah_next(tnhd, (void *) tnode);
	}
	if (flag) printf(";\n");
/* 
 * Print the TLINK tuples.
 */
	printf("\n/* link connections */\n");
	cnt = 0;
	flag = FALSE;
	for (tlink=nail->bs_tlink; tlink; tlink=tlink->tl_next) {
		if (cnt == 0) {
			printf("TLINK =");
			flag = TRUE;
		}
		printf(" <%d, %d,", (tlink->tl_srcnode)->tn_nodeid,
			tlink->tl_srclink);
		if (tlink->tl_srcname == NULL) {
			printf(", ");
		}
		else {
			printf(" \"%s\", ", tlink->tl_srcname);
		}
		printf("%d, %d,", (tlink->tl_destnode)->tn_nodeid,
			tlink->tl_destlink);
		if (tlink->tl_destname == NULL) {
			printf(", ");
		}
		else {
			printf(" \"%s\", ", tlink->tl_destname);
		}
		if (tlink->tl_weight == NONUM) {
			printf(">");
		}
		else {
			printf(" %d>", tlink->tl_weight);
		}
		if (cnt == (TLINKPERLINE-1)) {
			printf(";\n");
			flag = FALSE;
		}
		cnt = ((cnt + 1) % TLINKPERLINE);
	}
	if (flag) printf(";\n");
/* 
 * Print the TORDER tuples.
 */
	printf("\n/* non-direct communication paths */\n");
	cnt = 0;
	flag = FALSE;
	for (torder=nail->bs_torder; torder; torder=torder->to_next) {
		if (cnt == 0) {
			printf("TORDER =");
			flag = TRUE;
		}
		printf(" <%d, %d,", torder->to_srcnode, torder->to_destnode);
		if (torder->to_weight == NONUM) {
			printf(">");
		}
		else {
			printf(" %d>", torder->to_weight);
		}
		if (cnt == (TORDERPERLINE-1)) {
			printf(";\n");
			flag = FALSE;
		}
		cnt = ((cnt + 1) % TORDERPERLINE);
	}
	if (flag) printf(";\n");
/* 
 * Print the TROUTE tuples.
 */
	printf("\n/* route table entries */\n");
	cnt = 0;
	flag = FALSE;
	tnode = (struct t_node *) ah_top(tnhd);
	while (tnode) {
		troute = (tnode->tn_specs)->ns_troute;
		for ( ; troute; troute=troute->tr_next) {
			if (cnt == 0) {
				printf("TROUTE =");
				flag = TRUE;
			}
			printf(" <%d, %d, %d>", tnode->tn_nodeid,
				troute->tr_destnode, troute->tr_destlink);
			if (cnt == (TROUTEPERLINE-1)) {
				printf(";\n");
				flag = FALSE;
			}
			cnt = ((cnt + 1) % TROUTEPERLINE);
		}
		tnode = (struct t_node *) ah_next(tnhd, (void *) tnode);
	}
	if (flag) printf(";\n");
/* 
 * Print the TSOLDER tuples.
 */
	printf("\n/* soldering info */\n");
	cnt = 0;
	flag = FALSE;
	tnode = (struct t_node *) ah_top(tnhd);
	while (tnode) {
		if (cnt == 0) {
			printf("TSOLDER =");
			flag = TRUE;
		}
		printf(" <%d,", tnode->tn_nodeid);
		specs = tnode->tn_specs;
		if (specs->ns_host) {
			printf(" \"%s\",", specs->ns_host);
		}
		else {
			printf(",");
		}
		if (specs->ns_sysnode) {
			printf(" \"%s\">", specs->ns_sysnode);
		}
		else {
			printf(">");
		}
		if (cnt == (TSOLDERPERLINE-1)) {
			printf(";\n");
			flag = FALSE;
		}
		cnt = ((cnt + 1) % TSOLDERPERLINE);
		tnode = (struct t_node *) ah_next(tnhd, (void *) tnode);
	}
	if (flag) printf(";\n\n");
}

/*
 *	find_host
 *
 *	Function:	- determine the host node in the system
 *
 *	Accepts:	- ptr to the boot schema
 *
 *	Returns:	- 0 or ERROR
 */
int
find_host(nail)

struct bschema		*nail;

{
	int4		i;		/* favorite index */
	struct t_node	*tnode;		/* an okay pointer */
	HASH		*tnhd;		/* TNODE hash table desc. */

	tnhd = nail->bs_tnhd;
	tnode = (struct t_node *) ah_top(tnhd);
	while (tnode) {
		if (tisotb((tnode->tn_specs)->ns_type)) {
			nail->bs_hostnode = tnode;
			return(0);
		}
		tnode = (struct t_node *) ah_next(tnhd, (void *) tnode);
	}

	return(ERROR);
}

/*
 *	printfv
 *
 *	Function:	- print an array of strings to stdio
 *
 *	Accepts:	- the number of strings
 *			- ptr to the list of strings
 */
void 
printfv(narg, arg)

int			narg;
char			*arg[];

{
	for (; narg > 0; narg--, arg++) {
		printf("%s ", *arg);
	}
	printf("\n");
}

/*
 *	make_spantree
 *
 *	Function:	- create a network spanning tree from the host
 *
 *	Accepts:	- ptr to the network description
 *
 *	Returns:	- 0 or ERROR
 */
int 
make_spantree(nail)

struct bschema		*nail;

{
	int4		remain;			/* nodes not visited */
	int		i;
	struct t_node	*tnode;			/* ptr to node */
	struct t_node	**ptnode;		/* ptr to node pointer */
	struct t_node	*prev_tnode;		/* ptr to previous node */
	struct nspecs	*specs;			/* ptr to node specs */
	struct nspecs	*dest_specs;		/* destination node specs */
	struct link	*link;
	struct t_link	*tlink;			/* ptr to tlink */
	QUEUE		*waitr;			/* nodes to visit */
/* 
 * Handle the trivial cases.
 */
	if ((remain = ah_count(nail->bs_tnhd)) == 0) {
		return(0);
	}
	if ((tnode = nail->bs_hostnode) == (struct t_node *) 0) {
		nail_error("make_spantree", "no hostnode defined");
		return(ERROR);
	}
/*
 * Initialize the waiting room.  It is a queue of nodes to visit next.
 */
	waitr = aq_init((int4) TNODEINCR, (int4) sizeof(tnode));
	if (waitr == (QUEUE *) 0) {
		nail_error("make_spantree", "cannot create waiting room queue");
		return(ERROR);
	}
	if (aq_insert(waitr, &tnode)) {
		nail_error("make_spantree", "cannot insert in waiting room");
		aq_free(waitr);
		return(ERROR);
	}
	(tnode->tn_specs)->ns_nspanout = 0;
/*
 *  Start navigating the network.
 */
	prev_tnode = (struct t_node *) 0;
	while (aq_count(waitr) > INT4_NIL) {
/*
 * Get the node to visit.
 */
		if ((ptnode = (struct t_node **) aq_find(waitr)) ==
			(struct t_node **) 0) {

			nail_error("make_spantree", "no tnode in waiting room");
			aq_free(waitr);
			return(ERROR);
		}
		tnode = *ptnode;
		specs = tnode->tn_specs;

		if (aq_delete(waitr)) {
			nail_error("make_spantree", "cannot dequeue");
			aq_free(waitr);
			return(ERROR);
		}
		remain--;
/*
 * Link it to the previous node visited.
 */
		if (prev_tnode) {
			(prev_tnode->tn_specs)->ns_bootnext = tnode;
		}
		prev_tnode = tnode;
/*
 * Loop over all links out from the current node.
 */
		link = specs->ns_lkout;
		for (i = 0; i < specs->ns_nout; i++, link = link->lk_next) {
			tlink = link->lk_link;
			dest_specs = (tlink->tl_destnode)->tn_specs;

			if (dest_specs->ns_nspanout == NOTVISITED) {
/*
 * The destination node is still free.  Link it to the spanning tree.
 */
				if (add_splink(tlink, tnode)) {
					nail_error("make_spantree",
						"cannot add spanning link");
					aq_free(waitr);
					return(ERROR);
				}
/*
 * Declare the destination node "visited" and put it in the waiting room queue.
 */
				dest_specs->ns_nspanout = 0;
				if (aq_shove(waitr, &(tlink->tl_destnode))) {
					nail_error("make_spantree",
						"cannot shove node in waitQ");
					aq_free(waitr);
					return(ERROR);
				}
			}
		}		/* end for */
	}		/* end while */
/*
 * Cleanup and check if all nodes were visited.
 */
	aq_free(waitr);
	if (remain) {
		nail_error("make_spantree", "network is not fully connected");
		return(ERROR);
	}
	else {
		return(0);
	}
}

/*
 *	on_span
 *
 *	Function:	- check if the given tlink is part of the 
 *			  boot spanning tree
 *
 *	Accepts:	- ptr to tlink
 *
 *	Returns:	- TRUE or FALSE
 */
int
on_span(p_tlink)

struct t_link		*p_tlink;

{
	int		i;			/* favorite counter */
	struct link	*p;			/* favorite pointer */
	struct nspecs	*spec;			/* ptr to node specs */

	spec = (p_tlink->tl_srcnode)->tn_specs;
	p = spec->ns_lkspan;
	for (i = 0; i < spec->ns_nspanout; i++, p = p->lk_next) {
		if (p->lk_link == p_tlink) {
			return(TRUE);
		}
	}

	return(FALSE);
}

/*
 *	add_splink
 *
 *	Function:	- add a spanning tree link to a node
 *
 *	Accepts:	- ptr to the link
 *			- ptr to the node
 *
 *	Returns:	- 0 or ERROR
 */
static int
add_splink(tlink, tnode)

struct t_link		*tlink;
struct t_node		*tnode;

{
	struct link	*link;
	struct link	**p;		/* favorite pointer */
	struct nspecs	*specs;
/*
 * Allocate a linked list element and fill it.
 */
	if ((link = (struct link *) malloc(sizeof(struct link)))
		== (struct link *) 0) {
		return(ERROR);
	}
	link->lk_link = tlink;
/*
 * Insert the element at the end of the linked list.
 * Increment the number of children for this node.
 */
	specs = tnode->tn_specs;
	for (p = &(specs->ns_lkspan) ; *p; p = &((*p)->lk_next)) { }
	link->lk_next = *p;
	*p = link;
	specs->ns_nspanout++;

	return(0);
}

/*
 *	nail_error
 *
 *	Function:	- report error message in nail code
 *
 *	Accepts:	- module name
 *			- error message
 */
static void
nail_error(module, message)

char			*module,
			*message;

{
	fprintf(stderr, "%s", me);
	if (module) {
		fprintf(stderr, " (%s)", module);
	}
	if (message) {
		fprintf(stderr, ": %s", message);
	}
	fprintf(stderr, "\n");
}
