/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD
 *
 *	$Log:	nail.h,v $
 * Revision 2.2  92/04/30  14:23:20  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/11/26  14:04:23  raja
 * Modified to use libtools.
 * 
 * Revision 2.1  91/03/20  11:30:08  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/11/15  16:52:01  raja
 * Full overhaul of the data structures of the NaIL tools.
 * 
 * Revision 2.0.1.1  90/07/13  19:30:24  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:56:23  trillium
 * Homo erectus
 *
 *	Function:	- Network and Interconnect Language templates
 *			  and constants
 *			- handles the boot schema
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 */


#ifndef _NAIL
#define _NAIL

#include <portable.h>
#include <all_hash.h>

/*
 * function prototypes (nail.c)
 */
extern int4		next_prime();
extern int		find_host();
extern int		is_prime();
extern int		make_spantree();
extern void		pretty_bnail();
extern void		printfv();
extern void		print_bootorder();
extern void		print_spantree();

/*
 * misc. constants and macros
 */
#define NOTVISITED	(int)  -1	/* node not visited */
#define NONUM		(int4) -2	/* no number specified */

#define VERBOSE		if (fl_verbose) printf
#define BOOTNODE(x)	(( ! fl_thisnode) || (this_node == (x)))

/*
 * network constants
 */
#define LINKPERNODE		4	/* maximum nbr. of links/node */
#define TNODEINCR	      317	/* tnode array size increment */
#define DEFTWEIGHT	      100	/* default weight value */

/*
 * number of elements per line when pretty-printing schemas
 */
#define TNODEPERLINE		4
#define TLINKPERLINE		2
#define TROUTEPERLINE		2
#define TORDERPERLINE		4
#define TSOLDERPERLINE		4

/*
 * default filenames
 */
#define DEFFBNAIL	"bnail.def"

/*
 * default search paths
 */
#ifndef DEFPBNAIL
#define DEFPBNAIL	"/usr/local/trillium/boot"
#endif

/*
 * default tools
 */
#define DEFTHBOOT	"hboot"
#define DEFTSBOOT	"sboot"
#define DEFTVBOOT	"vboot"
#define DEFTTROUT	"trout"
#define DEFTTKILL	"tkill"
#define DEFTTSERV	"tserver"

/* 
 * structure definitions (boot schema)
 */
struct t_weight {				/* tweight information */
	int4			tw_link;	/* link weight */
	int4			tw_host;	/* host paths weight */
	int4			tw_order;	/* torder paths weight */
	int4			tw_mcast;	/* multicast weight */
	int4			tw_hop;		/* hop weight */
};

struct bschema {				/* boot schema */
	struct	t_weight	bs_tweight;	/* tweight tuple */
	HASH			*bs_tnhd;	/* ptr tnode hash desc. */
	struct	t_node		*bs_hostnode;	/* boot spanning tree root */
	int4			bs_ntlink;	/* number of tlinks */
	struct	t_link		*bs_tlink;	/* prt to tlink list */
	int4			bs_ntorder;	/* number of torder paths */
	struct	t_order		*bs_torder;	/* ptr to torder list */
	int4			bs_ntroute;	/* number of troute entries */
};

struct nspecs {					/* node specifications */
	int4			ns_type;	/* node type */
	int			ns_nout;	/* nbr. output links */
	int			ns_nin;		/* nbr. input links */
	struct	link		*ns_lkout;	/* ptr to output links */
	struct	link		*ns_lkin;	/* ptr to input links */
	int4			ns_ntroute;	/* number of troute entries */
	struct	t_route		*ns_troute;	/* ptr to troute entries */
	char			*ns_host;	/* host machine name */
	char			*ns_sysnode;	/* vendor node name */
	int			ns_nspanout;	/* nbr. children nodes */
	struct	link		*ns_lkspan;	/* ptr to span. tree links */
	struct	t_node		*ns_bootnext;	/* ptr to next node to boot */
};

struct t_node {					/* tnode entry */
	int4			tn_nodeid;	/* node identifier */
	struct	nspecs		*tn_specs;	/* ptr to node specs */
};
	
struct t_link {					/* link description */
	struct t_node		*tl_srcnode;	/* ptr to source node */
	int4			tl_srclink;	/* source link number */
	char			*tl_srcname;	/* source link name */
	struct t_node		*tl_destnode;	/* ptr to destination node */
	int4			tl_destlink;	/* destination link number */
	char			*tl_destname;	/* destination link name */
	int4			tl_weight;	/* link weight */
	struct	t_link		*tl_next;	/* ptr to next entry */
};

struct link {					/* list of t_link pointers */
	struct	t_link		*lk_link;	/* ptr to t_link */
	struct	link		*lk_next;	/* ptr to next entry */
};

struct t_route {				/* troute entry */
	int4			tr_destnode;	/* destination node */
	int4			tr_destlink;	/* destination link number */
	int4			tr_destlink2;	/* min-route link number */
	struct	t_route		*tr_next;	/* ptr to next entry */
};

struct t_order {				/* torder entry */
	int4			to_srcnode;	/* source node */
	int4			to_destnode;	/* destination node */
	int4			to_weight;	/* link weight override */
	struct	t_order		*to_next;	/* ptr to next entry */
};

#endif
