%{

/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University
 *
 *	RBD
 *
 *	$Log:	naill.l,v $
 * Revision 2.2  92/04/30  14:23:21  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/24  17:35:12  raja
 * Fixed clash between C-style comments and strings starting with '/'.
 * 
 * Revision 2.1.1.1  91/05/26  21:09:57  raja
 * Added a check in case the EOF is reached while a comment is still open.
 * 
 * Revision 2.1  91/03/20  11:39:36  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/11/16  15:21:59  raja
 * Overhaul of the topology layer tools.
 * 
 * Revision 2.0.1.1  90/07/13  19:30:35  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/06/01  16:57:18  gdburns
 * Homo erectus
 * 
 *
 *	Function:	- NaIL lexical analyzer
 *			- handles the boot schema
 */

#include <stdio.h>
#include <portable.h>
#include <typical.h>
#include <y.nail.h>
#include <nail.h>

/*
 * external variables
 */
extern int			errno;
extern int			lineno;		/* file line number */
extern char			*me;		/* program name */

/* 
 * external functions
 */
extern char			*malloc();

/*
 * static functions
 */
static void			rm_comment();	/* remove comments */
static void			rm_quotes();	/* remove quotes */
static char			*put_string();	/* store strings */
static int			yywrap();

%}

%%

"/*"				{ rm_comment(); }
\n				{ lineno++; }
[ \t]+				{ }

TWEIGHT				{ return(TWEIGHT); }
TNODE				{ return(TNODE); }
TLINK				{ return(TLINK); }
TORDER				{ return(TORDER); }
TROUTE				{ return(TROUTE); }
TSOLDER				{ return(TSOLDER); }
"="				{ return('='); }
"<"				{ return('<'); }
">"				{ return('>'); }
","				{ return(','); }
";"				{ return(';'); }

[\-+]?[0-9]+			{ sscanf(yytext, "%d", &yylval.num);
					return(NUMBER); }
[0-9a-fA-F]+			{ sscanf(yytext, "%x", &yylval.num);
					return(NUMBER); }
0[xX][0-9a-fA-F]+		{ sscanf(yytext+2,"%x", &yylval.num);
					return(NUMBER); }

\"[^\"\n\t]*\"			{ rm_quotes(); yylval.string = put_string();
					return(STRING); }
[a-zA-Z\.][^ =<>,;\n\t]*	{ yylval.string = put_string(); 
					return(STRING); }
[/][^\*][^ =<>,;\n\t]*		{ yylval.string = put_string(); 
					return(STRING); }

%%

/*
 *	rm_comment
 *
 *	Function:	- remove simple C-style comments
 *			- nested comments are not allowed
 */
static void 
rm_comment()    
{
	int		c1, c2;		/* two-character shift register */

	c1 = input();
	c2 = input();

	if (c1 == '\n') {
		lineno++;
	}
/*
 * Suck the input stream till an end of comment is found.
 * Update the line counter for every '\n'.
 */
	while ((c1 != '*') || (c2 != '/')) {
		if (c1 == 0) {
			fprintf(stderr,
				"%s: EOF reached with open comment\n", me);
			exit(ERROR);
		}
		if (c2 == '\n') {
			lineno++;
		}
		c1 = c2;
		c2 = input();
	}
}

/*
 *	put_string
 *
 *	Function:	- allocate storage for a string
 *			- store the string
 *			- abort is no storage space is available
 *
 *	Returns:	- ptr to the string
 */
static char *
put_string()  
{
	char		*string;
	int		i;

	if ((string = (char *)malloc(yyleng+1)) == (char *)NULL) {
		fprintf(stderr, "%s (put_string): cannot malloc\n", me);
		exit(errno);
	}

	for (i=0; i < yyleng; i++) {
		string[i] = yytext[i];
	}
	string[yyleng] = '\0';

	return(string);
}

/*
 *	rm_quotes
 *
 *	Function:	- remove first and last quotes of the string
 *			- work directly on the yytext[] array
 *			- update the string length yyleng
 *			- abort in case of error
 */
static void 
rm_quotes()
{
	int		i;

	if (yytext[0] != '\"') {
		fprintf(stderr, "%s (rm_quotes): no beginning quote\n", me);
		exit(ERROR);
	}
	if (yytext[yyleng-1] != '\"') {
		fprintf(stderr, "%s (rm_quotes): no ending quote\n", me);
		exit(ERROR);
	}
/*
 * Compact by shifting the string one character to the beginning.
 * Decrement the string length (yyleng) by two.
 */
	for (i=1; i < yyleng-1; i++) {
		yytext[i-1] = yytext[i];
	}
	yyleng -= 2;
}

/*
 *	yywrap
 *
 *	Function:	- stub to satisfy the lex generated code
 */
static int
yywrap()
{
	return(1);
}
