%{

/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD
 *
 *	$Log:	naily.y,v $
 * Revision 2.2  92/04/30  14:23:22  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  91/11/26  14:08:17  raja
 * Modified to use libtools and support a second route table.
 * 
 * Revision 2.1.1.1  91/06/05  21:42:36  raja
 * Enhanced portability: force default "int"s to "int4"s.
 * 
 * Revision 2.1  91/03/20  11:39:38  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  90/11/19  01:45:50  raja
 * Made the optional weights of TLINK and TORDER take their default values
 * from the TWEIGHT tuple if defined, otherwise take DEFTWEIGHT.
 * 
 * Revision 2.0.1.1  90/07/13  19:30:39  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/06/01  16:57:27  gdburns
 * Homo erectus
 *
 *	Function:	- NaIL parser
 *			- handles the boot schema
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 */

#include <stdio.h>
#include <portable.h>
#include <typical.h>
#include <net.h>
#include <nail.h>
#include <all_hash.h>

/*
 * global variables
 */
int				lineno;		/* file line number */

/*
 * external variables
 */
extern int			errno;
extern char			*me;		/* program name */

/*
 * static variables
 */
static int			errsyntax;	/* syntax error flag */
static struct bschema		*net;		/* ptr to the boot schema */

/*
 * external functions
 */
extern char			*malloc();

/*
 * static functions
 */
static void			do_tlink();
static void			do_tnode();
static void			do_torder();
static void			do_troute();
static void			do_tsolder();
static void			do_tweight();
static void			warning();
static int			expand_tnode();
static int			yyerror();
%}

%union {
	int4	num;
	char	*string;
}

%token	<num>		NUMBER
%token	<string>	STRING 
%token	<num>		TWEIGHT TNODE TLINK TORDER TROUTE TSOLDER 
%token	<num>		'=' '<' '>' ',' ';'

%type	<num>		optnum
%type	<string>	optstr

%start bschema

%%

bschema		: /* nothing */
		| bschema expression
		;

expression	: tweight ';'
		| tnode ';'
		| tlink ';'
		| torder ';'
		| troute ';'
		| tsolder ';'
		| error	';'		{ yyerrok; errsyntax = TRUE; }
		;
/*
 * TWEIGHT expression
 */
tweight		: TWEIGHT '=' '<' optnum ',' optnum ','
			optnum ',' optnum ',' optnum '>'
{
	do_tweight($4, $6, $8, $10, $12);
}
		| TWEIGHT '=' '<' optnum ',' optnum ',' optnum ',' optnum '>'
{
	do_tweight($4, $6, $8, $10, (int4) 0);
}
		;
/*
 * TNODE expression
 */
tnode		: TNODE '=' tnode_list
		;

tnode_list	: tnode_tuple
		| tnode_list tnode_tuple
		;

tnode_tuple	: '<' NUMBER ',' NUMBER '>'
{
	do_tnode($2,$4);
}
		;
/*
 * TLINK expression
 */
tlink		: TLINK '=' tlink_list
		;

tlink_list	: tlink_tuple
		| tlink_list tlink_tuple
		;

tlink_tuple	: '<' NUMBER ',' NUMBER ',' optstr ',' 
			NUMBER ',' NUMBER ',' optstr ',' optnum '>'
{
	do_tlink($2,$4,$6,$8,$10,$12,$14);
}
		;
/*
 * TORDER expression
 */
torder		: TORDER '=' torder_list
		;

torder_list	: torder_tuple
		| torder_list torder_tuple
		;

torder_tuple	: '<' NUMBER ',' NUMBER ',' optnum '>'
{
	do_torder($2,$4,$6);
}
		;
/*
 * TROUTE expression
 */
troute		: TROUTE '=' troute_list
		;

troute_list	: troute_tuple
		| troute_list troute_tuple
		;

troute_tuple	: '<' NUMBER ',' NUMBER ',' NUMBER ',' optnum '>'
{
	do_troute($2,$4,$6,$8);
}
		;
/*
 * TSOLDER expression
 */
tsolder		: TSOLDER '=' tsolder_list
		;

tsolder_list	: tsolder_tuple
		| tsolder_list tsolder_tuple
		;

tsolder_tuple	: '<' NUMBER ',' optstr ',' optstr '>'  
{
	do_tsolder($2,$4,$6);
}
		;
/*
 * optional number
 */
optnum		: /* no number */	{ $$ = NONUM; }
		| NUMBER		{ $$ = $1; }
		;
/*
 * optional string
 */
optstr		: /* no string */	{ $$ = NULL; }
		| STRING		{ $$ = $1; }
		;

%%

/*
 *	do_tweight
 *
 *	Function:	- store the tweight tuple
 *
 *	Accepts:	- links weight (optional)
 *			- host paths weight (optional)
 *			- torder paths weight (optional)
 *			- multicast paths weight (optional)
 *			- hop weight (optional)
 */
static void
do_tweight(tw_link, tw_host, tw_order, tw_mcast, tw_hop)

int4			tw_link,
			tw_host,
			tw_order,
			tw_mcast,
			tw_hop;

{
	if (tw_link != NONUM) {
		net->bs_tweight.tw_link = tw_link;
	}
	if (tw_host != NONUM) {
		net->bs_tweight.tw_host = tw_host;
	}
	if (tw_order != NONUM) {
		net->bs_tweight.tw_order = tw_order;
	}
	if (tw_mcast != NONUM) {
		net->bs_tweight.tw_mcast = tw_mcast;
	}
	if (tw_hop != NONUM) {
		net->bs_tweight.tw_hop = tw_hop;
	}
}

/*
 *	do_tnode
 *
 *	Function:	- store the tnode tuple
 *			- expand the TNODE array if needed
 *
 *	Accepts:	- the nodeid
 *			- the node type
 */
static void 
do_tnode(node, type)

int4			node,
			type;

{
	int4		new_size;	/* new hash table size */
	struct nspecs	*p;		/* favorite pointer */
	struct t_node	tn_elem;	/* tnode element */
/*
 * Allocate a node specification structure and initialize it.
 */
	p = (struct nspecs *) malloc(sizeof(struct nspecs));
	if (p == (struct nspecs *) 0) {
		warning("do_tnode", "cannot malloc specs structure");
		return;
	}

	p->ns_type	= type;
	p->ns_nout	= 0;
	p->ns_nin	= 0;
	p->ns_lkout	= (struct link *) 0;
	p->ns_lkin	= (struct link *) 0;
	p->ns_ntroute	= INT4_NIL;
	p->ns_troute	= (struct t_route *) 0;
	p->ns_host	= (char *) 0;
	p->ns_sysnode	= (char *) 0;
	p->ns_nspanout	= NOTVISITED;
	p->ns_lkspan	= (struct link *) 0;
	p->ns_bootnext	= (struct t_node *) 0;
/*
 * Prepare the TNODE hash table element and insert it.
 */
	tn_elem.tn_nodeid = node;
	tn_elem.tn_specs = p;
	if (ah_insert(net->bs_tnhd, &tn_elem)) {
/*
 * Not enough space.  Expand the TNODE hash table.
 */
		new_size = next_prime(ah_size(net->bs_tnhd) + TNODEINCR);
		if (ah_expand(net->bs_tnhd, new_size)) {
			warning("do_tnode", "cannot expand the TNODE table");
			free((char *) p);
			return;
		}
/*
 * Try to insert the element again.
 */
		if (ah_insert(net->bs_tnhd, &tn_elem)) {
			warning("do_tnode", "cannot store post-expansion !!!");
			free((char *) p);
			return;
		}
	}

	return;
}

/*
 *	do_tsolder
 *
 *	Function:	- store the tsolder tuple
 *			- the node must have been previously declared
 *
 *	Accepts:	- the nodeid
 *			- the hostname
 *			- the system dependent node name
 */
static void 
do_tsolder(node, hostname, sysnode)

int4			node;
char			*hostname,
			*sysnode;

{
	struct nspecs	*p;		/* favorite pointer */
	struct t_node	*tnode;		/* tnode pointer */

	if ((tnode = (struct t_node *) ah_find(net->bs_tnhd, node))
		== (struct t_node *) 0) {

		warning("do_tsolder", "undefined nodeid");
		return;
	}
	p = tnode->tn_specs;
	p->ns_host = hostname;
	p->ns_sysnode = sysnode;
	return;
}

/*
 *	do_troute
 *
 *	Function:	- store the troute tuple
 *			- the node must have been previously declared
 *
 *	Accepts:	- the source nodeid
 *			- the destination nodeid
 *			- the output link number
 *			- min-route output link number (optional)
 */
static void 
do_troute(src_node, dest_node, outlink, outlink2)

int4			src_node,
			dest_node,
			outlink,
			outlink2;

{
	struct t_route	*p;		/* troute pointer */
	struct t_route	**p1;		/* another favorite pointer */
	struct nspecs	*specs;		/* node specifications */
	struct t_node	*tnode;		/* tnode pointer */

	if ((tnode = (struct t_node *) ah_find(net->bs_tnhd, src_node))
		== (struct t_node *) 0) {

		warning("do_troute", "undefined nodeid");
		return;
	}
	specs = tnode->tn_specs;
/*
 * Find the last troute entry for the node.
 */
	for (p1 = &(specs->ns_troute); *p1; p1 = &((*p1)->tr_next));
/*
 * Allocate a new troute entry.
 */
	if ((p = (struct t_route *) malloc(sizeof(struct t_route))) == NULL) {
		warning("do_troute", "cannot malloc troute structure");
		return;
	}
	p->tr_next = *p1;
	*p1 = p;
/*
 * Fill the troute entry.
 */
	p->tr_destnode = dest_node;
	p->tr_destlink = outlink;
	p->tr_destlink2 = (outlink2 == NONUM) ? outlink : outlink2;
	(specs->ns_ntroute)++;
	(net->bs_ntroute)++;

	return;
}

/*
 *	do_tlink
 *
 *	Function:	- store the tlink tuple
 *			- both nodes must have been declared
 *
 *	Accepts:	- the source nodeid
 *			- the output link
 *			- the output link name
 *			- the destination nodeid
 *			- the input link
 *			- the input link name
 *			- the link weight (optional)
 */
static void 
do_tlink(src_node, outlink, outname, dest_node, inlink, inname, weight)

int4			src_node,
			outlink;
char			*outname;
int4			dest_node,
			inlink;
char			*inname;
int4			weight;

{
	struct t_node	*tn1, *tn2;		/* tnode pointers */
	struct nspecs	*p1, *p2;		/* node ptrs */
	struct t_link	*p;			/* tlink ptrs */
	struct t_link	**pp;			/* search prt */
	struct link	*l1, *l2;		/* node-link ptrs */
	struct link	**lp;			/* node-link search ptr */
/*
 * Check if both nodes have been declared.
 */
	tn1 = (struct t_node *) ah_find(net->bs_tnhd, src_node);
	tn2 = (struct t_node *) ah_find(net->bs_tnhd, dest_node);
	if ((tn1 == (struct t_node *) 0) || (tn2 == (struct t_node *) 0)) {
		warning("do_tlink", "undefined node(s)");
		return;
	}
/*
 * Get the node specification structures.  Check if the number of 
 * output/input links in the source/destination nodes has reached its maximum.
 */
	p1 = tn1->tn_specs;
	p2 = tn2->tn_specs;
	if ((p1->ns_nout >= LINKPERNODE) || (p2->ns_nin >= LINKPERNODE)) {
		warning("do_tlink", "links per node exceeded");
		return;
	}
/*
 * Allocate entries in the link output/input lists for the nodes.
 */
	l1 = (struct link *) malloc(sizeof(struct link));
	l2 = (struct link *) malloc(sizeof(struct link));
	if ((l1 == (struct link *) 0) || (l2 == (struct link *) 0)) {
		warning("do_tlink", "cannot malloc node link-entries");
		if (l1) {
			free((char *) l1);
		}
		if (l2) {
			free((char *) l2);
		}
		return;
	}
/*
 * Allocate a tlink entry and fill it.
 */
	p = (struct t_link *) malloc(sizeof(struct t_link));
	if (p == (struct t_link *) 0) {
		warning("do_tlink", "cannot malloc tlink structure");
		free((char *) l1);
		free((char *) l2);
		return;
	}

	p->tl_srcnode	= tn1;
	p->tl_srclink	= outlink;
	p->tl_srcname	= outname;
	p->tl_destnode	= tn2;
	p->tl_destlink	= inlink;
	p->tl_destname	= inname;
	p->tl_weight	= (weight == NONUM) ? net->bs_tweight.tw_link : weight;
/*
 * Link it to the end of the tlink linked list. 
 * Increment the number of tlinks in the network structure.
 */
	for (pp = &(net->bs_tlink); *pp; pp = &((*pp)->tl_next));
	p->tl_next = *pp;
	*pp = p;
	(net->bs_ntlink)++;
/*
 * Add the link to the end of the list of output links of the source node.
 */
	l1->lk_link = p;
	(p1->ns_nout)++;
	for (lp = &(p1->ns_lkout); *lp; lp = &((*lp)->lk_next));
	l1->lk_next = *lp;
	*lp = l1;
/*
 * Add the link to the end of the list of input links of the destination node.
 */
	l2->lk_link = p;
	(p2->ns_nin)++;
	for (lp = &(p2->ns_lkin); *lp; lp = &((*lp)->lk_next));
	l2->lk_next = *lp;
	*lp = l2;

	return;
}

/*
 *	do_torder
 *
 *	Function:	- store the torder tuple
 *
 *	Accepts:	- the source node
 *			- the destination node
 *			- the path weight (optional)
 */
static void 
do_torder(src_node, dest_node, weight)

int4			src_node,
			dest_node,
			weight;

{
	struct t_order	*p;		/* torder ptr */
	struct t_order	**pp;		/* search ptr */

/*
 * Allocate a torder structure and fill it.
 */
	p = (struct t_order *) malloc(sizeof(struct t_order));
	if (p == (struct t_order *) 0) {
		warning("do_torder", "cannot malloc torder structure");
		return;
	}
	p->to_srcnode	= src_node;
	p->to_destnode	= dest_node;
	p->to_weight	= (weight == NONUM) ? net->bs_tweight.tw_order : weight;
	(net->bs_ntorder)++;
/*
 * Link it at the end of the torder linked list.
 */
	for (pp = &(net->bs_torder); *pp; pp = &((*pp)->to_next));
	p->to_next = *pp;
	*pp = p;

	return;
}

/*
 *	yyerror
 *
 *	Function:	- report yacc syntax errors
 *
 *	Accepts:	- the error message
 *
 *	Returns:	- 0
 */
static int
yyerror(message)

char			*message;

{
	warning((char *) 0, message);
	return(0);
}

/*
 *	warning
 *
 *	Function:	- print the warning message
 *
 *	Accepts:	- the module name
 *			- the warning message
 */
static void
warning(mname, message)

char			*mname,
			*message;

{
	errsyntax = TRUE;
	fprintf(stderr, "%s", me);
	if (mname) {
		fprintf(stderr, " (%s)", mname);
	}
	if (message) {
		fprintf(stderr, ": %s", message);
	}
	fprintf(stderr, ": near line %d\n", lineno);
}

/*
 *	parse_nail
 *
 *	Function:	- entry point of the NaIL parser
 *			- initializes the data structures
 *			- invokes the actual parser (yyparse)
 *
 *	Accepts:	- ptr to the boot schema structure
 *
 *	Returns:	- 0 or ERROR
 */
int
parse_nail(bnail)

struct bschema		*bnail;

{
/*
 * Make the pointer statically accessible.
 * Initialize the network structure.
 */
	net = bnail;
	net->bs_tweight.tw_link = DEFTWEIGHT;
	net->bs_tweight.tw_host = DEFTWEIGHT;
	net->bs_tweight.tw_order = DEFTWEIGHT;
	net->bs_tweight.tw_mcast = DEFTWEIGHT;
	net->bs_tweight.tw_hop = DEFTWEIGHT;

	if ((net->bs_tnhd = ah_init((int4) 1, (int4) sizeof(struct t_node),
		NOTNODEID, AHFIXED)) == (HASH *) 0) {

		warning("parse_node", "cannot create TNODE hash table");
		return(ERROR);
	}

	net->bs_hostnode = (struct t_node *) 0;;
	net->bs_ntlink = (int4) 0;
	net->bs_tlink = (struct t_link *) 0;
	net->bs_ntorder = (int4) 0;
	net->bs_torder = (struct t_order *) 0;
	net->bs_ntroute = (int4) 0;

	lineno = 1;
	errsyntax = FALSE;
/*
 * Call the real parser.
 */
	if (yyparse() || errsyntax) {
		return(ERROR);
	}

	return(0);
}
