/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	bflist.c,v $
 * Revision 2.2  92/04/30  14:26:43  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  12:13:57  gdburns
 * Add buffer flow control of user messages.
 * 
 * Revision 2.1  91/03/20  11:31:34  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:31:12  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:56:48  trillium
 * Homo erectus
 * 
 *	Function:	- buffer daemon
 *			- list management routines
 *			- ITB/OTB portable module
 */

#include <bufreq.h>
#include <dlreq.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;

/*
 * external functions
 */
extern char		*malloc();

/*
 * globally defined functions
 */
struct buffer		*find();
struct buffer		*init();
void			insert();
void			delete();

/*
 *	insert
 *
 *	Function:	- inserts a buffer descriptor into a list
 *			- SORT by event and type - Why? - not anymore
 *			  now just stick at the end of the list
 *
 *	Accepts:	- ptr to top of list ptr
 *			- ptr to new buffer descriptor
 */

void
insert(top, new)

struct buffer		**top;
struct buffer		*new;

{
	struct buffer	*curr;		/* search pointer */
	struct buffer	*prev;

/*SORT
 *	curr = find(*top, new);
 */
/*
 * If a similar buffer descriptor is not in the list, insert the
 * new descriptor at the bottom of the list.
 */
/*SORT
 *	if (!curr) {
 */
		for (prev = curr = *top; curr;
			prev = curr, curr = curr->bf_next) {
		}

		if (prev) {
			prev->bf_next = new;
		} else {
			*top = new;
		}

		new->bf_next = NULL;
/*SORT
 *	 }
 */
/*
 * Otherwise, insert the new descriptor immediately after all others
 * of the same event and type.
 */
/*SORT
 *	else {
 *
 *		for (prev = curr; curr && (curr->bf_head.nh_dl_event ==
 *			new->bf_head.nh_dl_event) &&
 *			(curr->bf_head.nh_type == new->bf_head.nh_type);
 *			prev = curr, curr = curr->bf_next) {
 *		}
 *
 *		if (curr == *top) {
 *			new->bf_next = *top;
 *			*top = new;
 *		}
 *
 *		else {
 *			new->bf_next = prev->bf_next;
 *			prev->bf_next = new;
 *		}
 *	}
 */
}

/*
 *	delete
 *
 *	Function:	- deletes a buffer descriptor from a list
 *	Accepts:	- ptr to top of list ptr
 *			- ptr to existing buffer descriptor in the list
 */

void
delete(top, old)

struct buffer		**top;
struct buffer		*old;

{
	struct buffer	*curr;		/* search pointer */
	struct buffer	*prev;
/*
 * Search the list for a matching pointer.
 */
	for (prev = curr = *top; curr && (curr != old);
		prev = curr, curr = curr->bf_next) {
	}

	if (! curr) {
		errno = EIMPOSSIBLE;
		panic(errno);
	}
/*
 * Delete descriptor from the list.
 */
	if (curr == *top) {
		*top = curr->bf_next;
	}

	else {
		prev->bf_next = curr->bf_next;
	}
}

/*
 *	find
 *
 *	Function:	- find a descriptor of matching event and type
 *			  in a list
 *			- type matching based on the calculated datalink
 *			  type, using the same procedure as dsend()
 *
 *	Accepts:	- top of list ptr
 *			- ptr to key descriptor
 *
 *	Returns:	- ptr to matching descriptor or NULL
 */

struct buffer *
find(top, key)

struct buffer		*top;
struct buffer		*key;

{
	struct buffer	*p;		/* favourite ptr */
	int		key_type;	/* key datalink type */
	int		p_type;		/* datalink type */
	int		found;		/* found matching buffer? */

	key_type = (key->bf_head.nh_dl_event == key->bf_head.nh_event)
		? key->bf_head.nh_type : (key->bf_head.nh_event < 0)
		? DLSYS : DLUSR;
	found = 0;
	p = top;

	while (p && !found) {
		p_type = (p->bf_head.nh_dl_event == p->bf_head.nh_event)
			? p->bf_head.nh_type : (p->bf_head.nh_event < 0)
			? DLSYS : DLUSR;

		found = (p->bf_head.nh_dl_event == key->bf_head.nh_dl_event)
			&& (key_type == p_type);

		if (!found) p = p->bf_next;
	}

	return(p);
}

/*
 *	init
 *
 *	Function:	- allocate and initialize a new batch of free
 *			  buffer descriptors
 *	Returns:	- ptr to top of newly initialized free list or NULL
 */

struct buffer *
init()

{
	struct buffer	*p;		/* favourite ptr */
	int		i;		/* favourite index */

	p = (struct buffer *) malloc((unsigned)(BFMAX * sizeof(struct buffer)));
/*
 * Initialize the descriptors if successfully allocated.
 */
	if (p) {

		for (i = 0; i < (BFMAX - 1); ++i) {
			p[i].bf_next = &p[i + 1];
		}

		p[BFMAX - 1].bf_next = NULL;
	}

	return(p);
}
