/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	bfworker.c,v $
 * Revision 2.2  92/04/30  14:26:45  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  12:14:39  gdburns
 * Add buffer flow control of user messages.
 * 
 * Revision 2.1  91/03/20  11:31:35  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:31:13  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:56:49  trillium
 * Homo erectus
 * 
 *	Function:	- buffer worker
 *			- OTB/ITB portable module
 */

#include <bufreq.h>
#include <events.h>
#include <net.h>
#include <priority.h>
#include <shared.h>
#include <terror.h>

/*
 * external variables
 */
extern int		errno;
extern char		msg[MAXNMSGLEN];	/* message body */

/*
 * local functions
 */
static void		getmsg();

void
bfworker()

{
	struct bufreq	*req;			/* bufferd request */
	struct nmsg	nready;			/* ready message desc. */
	struct nmsg	nbuf;			/* buffered message desc. */

/*
 * Initialize.
 */
	if (kinit(PRBUFFER)) {
		panic(errno);
	}

	req = (struct bufreq *) nready.nh_data;
	nready.nh_node = LOCAL;
	nready.nh_event = nready.nh_dl_event = EVBUFFERD;
	nready.nh_type = BFREQ;
	nready.nh_length = 0;
	nready.nh_flags = NOBUF;
	req->bfq_req = BFREADY;
/*
 * Receive and forward network messages until we receive a message
 * destined for ourselves, then terminate.
 */
	getmsg(&nbuf);
/*
 * Initialize with kenyad.  I think that this might lead to deadlock
 * problems.
 *
 *	if (rpattach(LOCAL) && (errno != ESERVICE)) {
 *		panic(errno);
 *	}
 */

	while (nbuf.nh_event != EVBUFFER) {
		nbuf.nh_flags |= BYBUF;
		nbuf.nh_msg = msg;

		if (dsend(&nbuf)) {
			panic(errno);
		}
/*
 * Send a ready message to bufferd.
 */
		req->bfq_dl_event = nbuf.nh_dl_event;
		req->bfq_type = nbuf.nh_type;
		req->bfq_event = nbuf.nh_event;

		if (dsend(&nready)) {
			panic(errno);
		}

		getmsg(&nbuf);
	}
/*
 * Use kenyad to clean up our data/stack segments.
 */
	if (rpattach(LOCAL) && (errno != ESERVICE)) {
		panic(errno);
	}

	kexit(0);
}

/*
 *	getmsg
 *
 *	Function:	- get next buffered message from the
 *			  buffer daemon
 *	Accepts:	- ptr to network message descriptor
 */

static void
getmsg(nbuf)

struct nmsg		*nbuf;

{
	struct kmsg	kforward;		/* forwarded from daemon */

	kforward.k_event = EVBUFFER;
	kforward.k_type = 0;
	kforward.k_flags = KHOLD;
	kforward.k_length = sizeof(struct nmsg);
	kforward.k_msg = (char *) nbuf;

	if (krecv(&kforward)) {
		panic(errno);
	}

	kforward.k_flags = 0;
	kforward.k_length = nbuf->nh_length;
	kforward.k_msg = msg;

	if (krecv(&kforward)) {
		panic(errno);
	}
}
