/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	combine.c,v $
 * Revision 2.2  92/04/30  14:36:55  trillium
 * Ohio Release
 * 
 *	Function:	- functions to reel long messages
 */

#include <portable.h>
#include <castreq.h>
#include <events.h>
#include <net.h>
#include <shared.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>
#include <all_list.h>

/*
 * external variables
 */
extern int		errno;
extern struct nmsg	inhead;			/* network message desc. */
extern int4		filter;
extern int4		maxspace;

/*
 * external functions
 */
extern char		*malloc();
extern char		*memcpy();
extern int4		getroute2();

/*
 * local functions
 */
static void		cbn_int();
static void		lcombine();
static void		sendup();
static int		cmp_pkts();
static int		cmp_reels();

/*
 *	setspace
 *
 *	Function:	- update memory space usage and message filtering
 *			- user messages accepted if maxspace >= MAXNMSGLEN
 *
 *	Accepts:	- memory size increment
 */
void
setspace(incr)

int4			incr;

{
	maxspace += incr;
	if (maxspace >= MAXNMSGLEN) {
		filter = CFALL;
	} else {
		filter = CFSYS;
	}
}

/*
 *	reelit
 *
 *	Function:	- reel in multipacket messages
 *			- system panics in few locations if small
 *			  amounts of memory cannot be allocated
 *
 *	Accepts:	- ptr to cast table entry
 */
void
reelit(pcast)

struct cast		*pcast;

{
	struct castreel	crelem, *preel;
	struct castpkt	cpelem, *ppkt;
/*
 * Start by locating/creating the storage space for the new packet.
 * The location in the list of reels is dictated by the message event.
 * The location in the list of packets is dictated by the packet number.
 */
	if (pcast->c_reels == (LIST *) 0) {
/*
 * First reeling ever, create a list of reels.
 */
		if ((pcast->c_reels = al_init((int4) sizeof(struct castreel),
			cmp_reels)) == (LIST *) 0) {

			panic(errno);
		}
	}
/*
 * Locate the entry in the list of reels that matches the message event.
 */
	crelem.cr_event = inhead.nh_event;
	preel = (struct castreel *) al_find(pcast->c_reels, (char *) &crelem);
	if (preel == CRNULL) {
/*
 * This is the first message using this event, create a new entry.
 */
		crelem.cr_type = inhead.nh_type;
		crelem.cr_pkt = (LIST *) 0;
		if ((preel = (struct castreel *) al_insert(pcast->c_reels,
			(char *) &crelem)) == CRNULL) {

			panic(errno);
		}
	}
/*
 * Create a packet list if needed.
 */
	if (preel->cr_pkt == (LIST *) 0) {
		if ((preel->cr_pkt = al_init((int4) sizeof(struct castpkt),
			cmp_pkts)) == (LIST *) 0) {

			panic(errno);
		}
	}
/*
 * Locate the entry of the current packet number.
 */
	cpelem.cp_pktnum = inhead.nh_data[6];
	if ((ppkt = (struct castpkt *) al_find(preel->cr_pkt,
		(char *) &cpelem)) == CPNULL) {
/*
 * Packet entry does not exist, create a new one.
 */
		cpelem.cp_nreel = pcast->c_nbranch;
		cpelem.cp_length = 0;
		cpelem.cp_msg = (char *) 0;
		if ((ppkt = (struct castpkt *) al_insert(preel->cr_pkt,
			(char *) &cpelem)) == CPNULL) {

			panic(errno);
		}
	}
/*
 * Now that we have pin-pointed the location, process the message.
 * If the message is new, just store it, otherwise do the combining.
 */
	ppkt->cp_nreel--;
	if (ppkt->cp_nreel == pcast->c_nbranch - 1) {
/*
 * New message.  Allocate space and store it.
 */
		ppkt->cp_nh_flags = inhead.nh_flags;
		(void) memcpy((char *) ppkt->cp_nh_data,
			(char *) inhead.nh_data, NHDSIZE * sizeof(int4));
		if ((ppkt->cp_length = inhead.nh_length) > 0) {
			ppkt->cp_msg = malloc((uint4) ppkt->cp_length);
			if (ppkt->cp_msg == (char *) 0) {
				panic(errno);
			} else {
				(void) memcpy(ppkt->cp_msg,
					inhead.nh_msg, ppkt->cp_length);
				setspace( - ppkt->cp_length);
			}
		}
	}
	else {
/*
 * Subsequent message, merge it and adjust the message length if needed.
 */
		ppkt->cp_length = min(ppkt->cp_length, inhead.nh_length);
		if (pcast->c_combine >= 0) {
/*
 * Use a local combining process.
 */
			lcombine(ppkt, &inhead, pcast->c_combine);
		}
		else {
/*
 * Use the builtin combining functions.
 */
			cbn_int(NHDSIZE * sizeof(int4), ppkt->cp_nh_data,
				(int4 *) inhead.nh_data, pcast->c_combine);
			cbn_int(ppkt->cp_length, (int4 *) ppkt->cp_msg,
				(int4 *) inhead.nh_msg, pcast->c_combine);
		}
	}

	if (ppkt->cp_nreel == INT4_NIL) {
/*
 * This packet is ready to be shipped up the tree and deleted.
 */
		sendup(pcast, preel, ppkt);
		if (ppkt->cp_msg) {
			(void) free(ppkt->cp_msg);
			setspace(ppkt->cp_length);
		}
		(void) al_delete(preel->cr_pkt, (char *) ppkt);

		if (al_count(preel->cr_pkt) == INT4_NIL) {
/*
 * That was the last packet in the message, delete the whole list of
 * packets and remove the event entry in the list of reels.  We don't
 * delete the list of reels even if it is empty because we assume the castid
 * will be used again.  Only melting the castid removes the list of reels.
 */
			(void) al_free(preel->cr_pkt);
			(void) al_delete(pcast->c_reels, (char *) preel);
		}
	}
}

/*
 *	sendup
 *
 *	Function:	- send the reeled packet upstream
 *
 *	Accepts:	- ptr cast table entry
 *			- ptr reel list entry
 *			- ptr packet list entry
 */
static void
sendup(p_cast, p_reel, p_pkt)

struct cast		*p_cast;
struct castreel		*p_reel;
struct castpkt		*p_pkt;

{
	struct nmsg	outhead;		/* message header */

	(void) memcpy((char *) outhead.nh_data, (char *) p_pkt->cp_nh_data,
		NHDSIZE * sizeof(int4));

	outhead.nh_event = p_reel->cr_event;
	outhead.nh_type = p_reel->cr_type;
	outhead.nh_flags = p_pkt->cp_nh_flags;
	outhead.nh_msg = p_pkt->cp_msg;
	outhead.nh_length = p_pkt->cp_length;

	if (p_cast->c_root_node == getnodeid()) {
/*
 * Send the message to the local client.  Use the buffers if needed.
 */
		outhead.nh_dl_event = outhead.nh_event;
		outhead.nh_node = LOCAL;
		outhead.nh_flags &= ~BYBUF;
	}
	else {
/*
 * Send the message to the castd upstream.  Bypass buffers.
 * Use the message reeling protocol: nh_data[6] has the packet number.
 */
		outhead.nh_dl_event =
				getroute2(p_cast->c_parent_node, NOTEVENT);
		outhead.nh_node = p_cast->c_castid;
		outhead.nh_flags |= BYBUF;
		outhead.nh_data[6] = p_pkt->cp_pktnum;
	}

	if (dsend(&outhead)) {
		panic(errno);
	}
}

/*
 *	cbn_int
 *
 *	Function:	- combine by integer
 *			- selects on specific combining function
 *
 *	Accepts:	- length of messages (in bytes)
 *			- first operand ptr
 *			- second operand ptr
 *			- combining code
 */	
static void
cbn_int(length, msg1, msg2, cbn_code)

int4			length;
int4			*msg1;
int4			*msg2;
int4			cbn_code;

{
	int4		i;			/* favourite index */

	if (length == 0) {
		return;
	}

	if (cbn_code == CBN_ERR) {
		*msg1 = *msg1 ? *msg1 : *msg2;
		return;
	}

	length /= sizeof(int4);

	switch(cbn_code) {

		case CBN_ADD:
			for (i = 0; i < length; i++, msg1++, msg2++) {
				*msg1 += *msg2;
			}
			break;

		case CBN_MIN:
			for (i = 0; i < length; i++, msg1++, msg2++) {
				*msg1 = min(*msg1, *msg2);
			}
			break;

		case CBN_MAX:
			for (i = 0; i < length; i++, msg1++, msg2++) {
				*msg1 = max(*msg1, *msg2);
			}
			break;

		case CBN_XOR:
			for (i = 0; i < length; i++, msg1++, msg2++) {
				*msg1 ^= *msg2;
			}
			break;

		case CBN_OR:
			for (i = 0; i < length; i++, msg1++, msg2++) {
				*msg1 |= *msg2;
			}
			break;

		case CBN_AND:
			for (i = 0; i < length; i++, msg1++, msg2++) {
				*msg1 &= *msg2;
			}
			break;
	}
}

/*
 *	lcombine
 *
 *	Function:	- combine by sending two messages to a local process
 *			  on a given event and collecting one message back
 *			- because of packetization, the message length
 *			  cannot be changed by the user
 *
 *	Accepts:	- ptr message packet
 *			- ptr incoming message
 *			- local event of combining process
 */	
static void
lcombine(pkt, incoming, cbn_code)

struct castpkt		*pkt;
struct nmsg		*incoming;
int4			cbn_code;

{
	struct nmsg	nhead;

	nhead.nh_node = LOCAL;
	nhead.nh_event = nhead.nh_dl_event = cbn_code;
	nhead.nh_type = 0;
	nhead.nh_flags = pkt->cp_nh_flags | NOBUF;
	nhead.nh_length = pkt->cp_length;

	nhead.nh_msg = pkt->cp_msg;
	(void) memcpy((char *) nhead.nh_data,
			(char *) pkt->cp_nh_data, NHDSIZE * sizeof(int4));

	if (dsend(&nhead)) {
		return;
	}

	nhead.nh_msg = incoming->nh_msg;
	(void) memcpy((char *) nhead.nh_data,
			(char *) incoming->nh_data, NHDSIZE * sizeof(int4));

	if (dsend(&nhead)) {
		return;
	}

	nhead.nh_msg = pkt->cp_msg;

	if (drecv(&nhead)) {
		return;
	}

	(void) memcpy((char *) pkt->cp_nh_data,
			(char *) nhead.nh_data, NHDSIZE * sizeof(int4));
}

/*
 *	cmp_reels
 *
 *	Function:	- compare two reel entries
 *
 *	Accepts:	- ptr first reel entry
 *			- ptr second reel entry
 *
 *	Returns:	- neg. if first < second
 *			   0   if first == second
 *			  pos. if first > second
 */
static int
cmp_reels(preel1, preel2)

struct castreel		*preel1,
			*preel2;

{
	return(preel1->cr_event - preel2->cr_event);
}

/*
 *	cmp_pkts
 *
 *	Function:	- compare two packet entries
 *
 *	Accepts:	- ptr first packet entry
 *			- ptr second packet entry
 *
 *	Returns:	- neg. if first < second
 *			   0   if first == second
 *			  pos. if first > second
 */
static int
cmp_pkts(ppkt1, ppkt2)

struct castpkt		*ppkt1,
			*ppkt2;

{
	return(ppkt1->cp_pktnum - ppkt2->cp_pktnum);
}
