/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	dld.c,v $
 * Revision 2.2  92/04/30  14:39:24  trillium
 * Ohio Release
 * 
 *	Function:	- dld, datalink service daemon
 *			- yet another resgistration daemon
 *			- operates the output flow control valves
 */

#include <dlreq.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <terror.h>

#define NOTSIG		-1

/*
 * external variables
 */
extern int		errno;			/* global error variable */

/*
 * global variables
 */
int4			ppid[MAXNLINK];		/* child dlo processes */

void
dld()

{
	int4		event;			/* route for client node */
	int4		sig;			/* signal for dlo child */
	struct dlreq	*req;			/* datalink request */
	struct nmsg	nhead;			/* network message */
	struct kstate	sigstat;		/* signal delivery status */
	int4		nextsig[MAXNLINK];	/* next signal to deliver */
	int4		pendsig[MAXNLINK];	/* pending signal delivery */
	int4		i;			/* favourite index */
	int4		block;			/* poll or block? */

	for (i = 0; i < MAXNLINK; ++i) {
	    nextsig[i] = NOTSIG;
	    pendsig[i] = NOTSIG;
	}
/*
 * Initialize kernel interface and become a kernel process.
 */
	if (kinit(0) < 0) {
	    panic(errno);
	}
/*
 * Attach to kenyad (ITB only).
 */
	if (rpattach(LOCAL) && (errno != ESERVICE)) {
	    panic(errno);
	}

	req = (struct dlreq *) nhead.nh_data;
/*
 * server main loop
 */
	nhead.nh_event = EVDLMOM;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	for (;;) {
	    block = NRECV;

	    for (i = 0; i < MAXNLINK; ++i) {
		if (pendsig[i] != NOTSIG) {
		    block = NTRY_RECV;
		}
	    }

	    if (do_nrecv(&nhead, block) < 0) {

		if (errno == EWOULDBLOCK) {
		    req->dlq_req = DLQPOLL;
		} else {
		    panic(errno);
		}
	    }

	    switch (req->dlq_req) {

		case DLQVALVE:
		event = getroute(req->dlq_node, NOTEVENT);
		i = dl_link(event);

		if (event != NOTEVENT) {
		    sig = (req->dlq_valve == DLON) ? SIGA : SIGB;

		    if (nextsig[i] != NOTSIG) {

			if (sig != nextsig[i]) {
			    nextsig[i] = NOTSIG;
			}
		    } else if (pendsig[i] != NOTSIG) {

			if (sig != pendsig[i]) {
			    nextsig[i] = sig;
			}
		    } else {
			pendsig[i] = sig;

			if (kdoom(ppid[i], sig) < 0) {
			    panic(errno);
			}
		    }
		}

		break;

		case DLQATTACH:
		ppid[dl_link(req->dlq_event)] = req->dlq_pid;
		break;
	    }


	    for (i = 0; i < MAXNLINK; ++i) {

		if (pendsig[i] != NOTSIG) {

		    if (kstate(ppid[i], &sigstat) < 0) {
			panic(errno);
		    }

		    if (! (sigstat.ks_signal & sigmask(pendsig[i]))) {

			if (nextsig[i] != NOTSIG) {

			    if (kdoom(ppid[i], nextsig[i]) < 0) {
				panic(errno);
			    }

			    pendsig[i] = nextsig[i];
			    nextsig[i] = NOTSIG;
			} else {
			    pendsig[i] = NOTSIG;
			}
		    }
		}
	    }
	}
}
