/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD
 *
 *	$Log:	prime.c,v $
 * Revision 2.2  92/04/30  14:40:43  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/11/11  14:24:22  raja
 * initial revision
 * 
 *	Function:	- test and generate prime numbers
 *				- is_prime(): test is number is prime
 *				- next_prime(): return the next prime number
 */

#include <math.h>
#include <portable.h>
#include <typical.h>

/* 
 *	is_prime
 *
 *	Function:	- check if the number is prime
 *
 *	Accepts:	- a positive integer
 *
 *	Returns:	- TRUE or FALSE
 */
int
is_prime(n)

int4			n;

{
	int4		i;		/* favorite counter */
	int4		maxdiv;		/* largest divisor */

	if (n <= (int4) 0) {
		return(FALSE);
	}
	else if ((n == (int4) 1) || (n == (int4) 2)) {
		return(TRUE);
	}
	else if ((n % 2) == (int4) 0) {
		return(FALSE);
	}
/*
 * Loop through all odd number smaller than the largest divisor
 * checking if any divides evenly the given number.
 */
	else {
		maxdiv = (int4) (sqrt((double) n) + 1);
		for (i = 3; i <= maxdiv; i += 2) {
			if ((n % i) == (int4) 0) {
				return(FALSE);
			}
		}
	}

	return(TRUE);
}

/*
 *	next_prime
 *
 *	Function:	- returns the first prime number larger than
 *			  or equal to the given number
 *
 *	Accepts:	- a positive number
 *
 *	Returns:	- a prime number or ERROR
 */
int4
next_prime(n)

int4			n;

{
	if (n < (int4) 0) {
		return((int4) ERROR);
	}
	else if (n < (int4) 1) {
		return((int4) 1);
	}
	else if (n < (int4) 2) {
		return((int4) 2);
	}
	else {
/*
 * Find the next prime number in the general case.
 */
		if ((n % 2) == (int4) 0) {
			n++;
		}
		while (! is_prime(n)) {
			n += 2;
		}
		return(n);
	}
}
