/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB, RBD
 *
 *	$Log:	rw.c,v $
 * Revision 2.2  92/04/30  14:40:47  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/27  17:23:51  raja
 * Add rwcpy() to change the byte ordering of an array of integers.
 * Various cosmetic changes and enhanced portability.
 * 
 * Revision 2.1.1.1  91/07/15  18:03:45  raja
 * Words to be reversed are now unsigned, as they should be.
 * This fixes a bug that shows up under the ULS compiler.
 * 
 * Revision 2.1  91/03/20  11:32:47  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:31:05  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:56:45  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:12:13  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:47:03  trillium
 * Initial revision
 * 
 *	Function:	- reverse bytes
 *			- handles words, short words and arrays of words
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 */

#include <portable.h>

/*
 *	rw
 *
 *	Function:	- reverse byte ordering in a word
 *
 *	Accepts:	- any word
 *
 *	Returns:	- the word with the byte ordering reversed
 */
uint4
rw(rword)

uint4			rword;		/* word to be reversed */

{
	uint4		new_rword;	/* reversed word */
	int		i;		/* loop counter */
	
	for (i = 0, new_rword = 0; i < sizeof(int4); ++i) {
		new_rword = (new_rword << 8) + (rword & INT4_LSB);
		rword >>= 8;
	}

	return(new_rword);
}

/*
 *	rs
 *
 *	Function:	- reverse bytes in a short word
 *
 *	Accepts:	- any short word
 *
 *	Returns:	- the short word with the byte ordering reversed
 */
unsigned short
rs(rword)

unsigned short		rword;

{
	unsigned short	new_rword;	/* reversed word */
	int		i;		/* loop counter */

	
	for (i = 0, new_rword = 0; i < sizeof(short); ++i) {
		new_rword = (new_rword << 8) + (rword & INT4_LSB);
		rword >>= 8;
	}

	return(new_rword);
}

/*
 *	rwcpy
 *
 *	Function:	- reverse byte ordering of words in an array
 *
 *	Accepts:	- ptr to destination array (reversed words)
 *			- ptr to source array (words to be reversed)
 *			- number of words in array
 */
void
rwcpy(dest, src, nwords)

uint4			*dest;
uint4			*src;
int4			nwords;

{
	int4		i;			/* favorite counter */

	for (i = 0; i < nwords; i++) {
		dest[i] = rw(src[i]);
	}
}
