/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	search.c,v $
 * Revision 2.2  92/04/30  14:40:48  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/01/20  19:10:12  gdburns
 * Overhaul to use access().
 * 
 * Revision 2.1.1.1  91/05/06  21:13:51  gdburns
 * Watch out for null prefix.
 * 
 * Revision 2.1  91/03/10  23:01:06  gdburns
 * Ohio Release
 * 
 *	Function:	- file searching
 */

#include <errno.h>
#include <portable.h>
#include <typical.h>
#include <unistd.h>

/*
 * globally defined functions
 */
char			*findpath();

/*
 * external functions
 */
extern int		access();
extern int		strlen();
extern char		*getenv();
extern char		*malloc();
extern char		*strcpy();
extern char		*strcat();
extern char		*strchr();

/*
 * local functions
 */
static char		*trypath();

/*
 * external variables
 */
extern int		errno;

/*
 *	findpath
 *
 *	Function:	- locates a file with certain permissions
 *			- environment variable can appear in the form
 *			  $variable at the start of a prefix path
 *			  and will be replaced by the environment value
 *			  if it is defined otherwise the whole prefix
 *			  is ignored
 *			- environment variable must be following by
 *			  a path delimiter or end-of-string
 *
 *	Accepts:	- file name
 *			- array of search directories
 *			- target permissions which must be satisfied
 *
 *	Returns:	- full pathname of located file or NULL
 */

char *
findpath(base, pfixv, mode)

char			*base;
char			**pfixv;
int			mode;

{
	char		*fullpath;	/* full pathname of search file */
	char		*delimit;	/* ptr to first delimiter in prefix */
	char		*env;		/* ptr to environment var */
	char		*pfix;		/* prefix directory */
	int		i;		/* favourite index */
/*
 * Check stupid errors.
 */
	if (! base || ! pfixv) {
		errno = EINVAL;
		return(NULL);
	}
/*
 * initialization
 */
	fullpath = NULL;
	i = 0;
/*
 * Consider each prefix until the file is found.
 * Thus, the order of prefixes is important.
 */
	while ((pfixv[i]) && (! fullpath)) {
/*
 * Replace environment variable at the head of the string.
 */
		if (*pfixv[i] == '$') {
			delimit = strchr(pfixv[i], STRDIR);

			if (delimit != NULL) { *delimit = '\0'; }

			env = getenv(pfixv[i] + 1);

			if (delimit != NULL) { *delimit = STRDIR; }

			if (env != NULL) {

				if (delimit == NULL) {
					fullpath = trypath(base, env, mode);
				} else {
					pfix = malloc((unsigned) strlen(env) +
						strlen(delimit) + 1);

					if (pfix == NULL) {
						return(NULL);
					}

					(void) strcpy(pfix, env);
					(void) strcat(pfix, delimit);
					fullpath = trypath(base, pfix, mode);
					(void) free(pfix);
				}
			}
		}

		else {
			fullpath = trypath(base, pfixv[i], mode);
		}

		i++;
	}

	return(fullpath);
}

/*
 *	trypath
 *
 *	Function:	- forms a complete pathname and checks it for
 *			  existance and permissions
 *
 *	Accepts:	- filename
 *			- path prefix
 *			- permissions which must be satisfied
 *
 *	Returns:	- full pathname or NULL
 */

static char *
trypath(base, pfix, mode)

char			*base;
char			*pfix;
int4			mode;

{
	char		*fullpath;	/* full pathname of search file */
/*
 * Check stupid errors.
 */
	if (! base || ! pfix) {
		errno = EINVAL;
		return(NULL);
	}
/*
 * Allocate space for the full pathname.
 */
	fullpath = malloc((unsigned) strlen(pfix) + strlen(base) + 2);

	if (! fullpath) { return(NULL); }

	if (strlen(pfix) > 0) {
		(void) strcpy(fullpath, pfix);
		(void) strcat(fullpath, STRSDIR);
		(void) strcat(fullpath, base);
	} else {
		(void) strcpy(fullpath, base);
	}
/*
 * Get status on the full path name to check for existance.
 * Then check the permissions.
 */
	if (access(fullpath, mode) < 0) {
		fullpath = NULL;
	}

	return(fullpath);
}
