/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	terror.c,v $
 * Revision 2.2  92/04/30  14:40:52  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/27  12:00:58  gdburns
 * Add EI860PROC and protect against string overflows.
 * 
 * Revision 2.1.1.1  92/01/27  17:55:30  gdburns
 * Remove ENOPORT.
 * 
 * Revision 2.1  91/03/20  11:32:50  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.5  91/01/08  12:40:01  gdburns
 * Added EBADREC.
 * 
 * Revision 2.0.1.4  90/11/27  17:55:46  gdburns
 * replaced file daemon errors
 * 
 * Revision 2.0.1.3  90/10/15  16:14:31  gdburns
 * Added EBLOWN.
 * 
 * Revision 2.0.1.2  90/07/13  19:31:07  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:56:46  trillium
 * Homo erectus
 *
 *	Function:	- prints error messages for Unix and Trillium
 *			- checks if Unix error and uses perror
 *			- checks if Trillium error and uses table of
 *			  error messages
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University
 *
 *	Accepts:	- user's error message
 */

#include <stdio.h>
#include <terror.h>
#include <typical.h>

char			*emsg[] = {
		"out of event descriptors",	/* ENOEVDESCRIPTORS */
		"virtual circuit in use",	/* ECIRCUITINUSE */
		"out of exit descriptors",	/* ENOEXDESCRIPTORS */
		"out of descriptors",		/* EFULL */
		"not a process",		/* ENOTPROCESS */
		"invalid request",		/* EBADREQUEST */
		"block not allocated",		/* ENOTMALLOCED */
		"unused error code",		/* unused */
		"service unavailable",		/* ESERVICE */
		"bad magic number",		/* EMAGIC */
		"blown stack",			/* EBLOWN */
		"impossible, ergo a bug",	/* EIMPOSSIBLE */
		"limit one i860 process",	/* EI860PROC */
		"garbage character",		/* EGARBAGE */
		"no running kernel",		/* ENOKERNEL */
		"not attached to kernel",	/* ENOTATTACHED */
		"lost system integrity",	/* EPANIC */
		"improper message size",	/* EMSGLEN */
		"end of file",			/* EEOF */
		"cast spanning tree collision",	/* ECOLLIDE */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"can't find shared segment",	/* ESHARE */
		"invalid link index",		/* EBADLINK */
		"bad acknowledgement",		/* EACK */
		"improper command invocation",	/* EUSAGE */
		"bad/unknown option",		/* EOPTION */
		"bad node range specification", /* EBADRANGE */
		"invalid node",			/* EBADNODE */
		"invalid node(s) specification",/* EBADNODESPEC */
		"unused error code",		/* unused */
		"out of route descriptors",	/* ENORDESCRIPTORS */
		"out of buffer descriptors",	/* ENOBFDESCRIPTORS */
		"out of boot descriptors",	/* ENOBTDESCRIPTORS */
		"out of flat descriptors",	/* ENOFLDESCRIPTORS */
		"out of kenya descriptors",	/* ENOPDESCRIPTORS */
		"out of process descriptors",	/* ENOKDESCRIPTORS */
		"bad moses request",		/* EBADMOSESREQ */
		"can't kill datalink",		/* EMAXBULLETS */
		"invalid address tag",		/* EBADTAG */
		"delete from list failed",	/* EDELETE */
		"no bootstrap present",		/* ENOBOOTSTRAP */
		"invalid record format",	/* EBADREC */
		"bad ascii number",		/* EBADASCIINUM */
		"booted system is too big",	/* EHIGHWATER */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"unused error code",		/* unused */
		"badly synchronized cubix req",	/* ECXREQ */
		"invalid cubix member",		/* ECXMEMBER */
		"OK, cubix singl method",	/* ECXSINGL */
			};

char			*baderror = 
		"?  Get out of town!\n";	/* unknown errno */


extern int		errno;			/* global error variable */

char			*strcpy(),
			*strcat();

int
terror(umsg)

char			*umsg;			/* user supplied message */

{
	char		buffer[64];		/* complete message */
	char		anum[12];		/* ascii error code */
	int		len;			/* string length */

	if ((errno >= ELOW) && (errno <= EHIGH)) {
/*
 * Trillium error
 */
		strncpy(buffer, umsg, 32);
		buffer[32] = '\0';
		strcat(buffer, ": ");
		strcat(buffer, emsg[errno - ELOW]);
		strcat(buffer, "\n");
		len = strlen(buffer);
		if (write(STDERR, buffer, len) != len) return(ERROR);
	}

	else if ((errno > EHIGH) || (errno < 0)) {
/*
 * Unknown error
 */
		strcpy(buffer, umsg);
		strcat(buffer, ": ");
		my_itoa(errno, anum);
		strcat(buffer, anum);
		strcat(buffer, baderror);
		len = strlen(buffer);
		if (write(STDERR, buffer, len) != len) return(ERROR);
	}

	else {
		perror(umsg);
	}

	return(0);
}

/*
 *	my_itoa
 *
 *	Function:	- convert n to characters in s
 *			- lifted from the itb libc (unidot)
 *
 *	Accepts:	- the number to convert
 *			- ptr to the ascii buffer
 */

my_itoa(n, s)  register char *s; register int n;{

	char		sbuf[16];
	register char	*p;

	p = sbuf;
	if( n < 0 ){/* record sign */
		n = -n; /* make n positive */
		*p = '-';
	}
	do	/* generate digits in reverse order */
		*p++ = ((unsigned)n) % 10 + '0'; /* get next digit */
	while( n = ((unsigned)n) / 10 );
	while( p > sbuf ) *s++ = *--p;
	*s = 0;
}
