/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	tout.c,v $
 * Revision 2.2  92/04/30  14:40:54  trillium
 * Ohio Release
 * 
 * Revision 2.1  92/01/20  19:19:08  gdburns
 * initial revision
 * 
 *	Function:	- functions that manipulate the t.out load
 *			  module format
 */

#include <portable.h>
#include <t.out.h>
#include <t_types.h>

/*
 * global functions
 */
void			*toutdataptr();	/* ptr to data segment */
void			toutpatch();	/* relocate data segment */
void			*toutrelptr();	/* ptr to reloc/sym tables */
void			toutrev();	/* reverse words in load module */
void			*touttextptr();	/* ptr to text segment */
int4			toutentry();	/* entry point offset */
int4			toutbsssize();	/* bss segment size */
int4			toutdatasize();	/* data segment size */
int4			touthtxtsize();	/* text segment + header size */
int4			toutloadsize();	/* total load module size */
int4			touttextsize();	/* text segment size */

/*
 *	toutpatch
 *
 *	Function:	- relocates 32 bit addresses in the data segment
 *
 *	Accepts:	- ptr to load module
 *			- ptr to relocation and symbol tables
 *			- ptr to data segment to be patched
 *			- eventual text segment location
 *			- eventual data segment location
 */

void
toutpatch(pload, reloc, ldatapt, textpt, datapt)

void			*pload;		/* ptr to load module */
void			*ldatapt;	/* address of local data segment */
void			*textpt;	/* text address */
void			*datapt;	/* data address */
struct to_relative	*reloc;		/* pointer to relocate table */

{
	struct exec	*tout;		/* ptr to load module */
	int4		relocsize;	/* size of relocation table */
	int4		i;		/* favourite index variable */
	int4		s1_value;	/* symbol 1 value */
	int4		s2_value;	/* symbol 2 value */
	int4		patch_value;	/* relocation patch value */
	int4		*patch_addr;	/* local patch address */
	int4		stub();		/* byte ordering NOP */
	struct nlist	*symbol;	/* pointer to symbol table */

	tout = (struct exec *) pload;
/*
 * Determine local symbol table address.
 */
	symbol = (struct nlist *) (((char *) reloc) + tout->a_drsize);
/*
 * Determine number of relocation table entries.
 */
	relocsize = tout->a_drsize / sizeof(struct to_relative);
/*
 * Step through the relocate table and make the patches.
 */
	for (i = 0; i < relocsize; ++i) {
/*
 * All patches are of the form (s1 - s2) + constant + base.
 */
		s1_value = (reloc[i].s1.index == -1)
			? 0
			: symbol[reloc[i].s1.index].n_value;

		s2_value = (reloc[i].s2.index == -1)
			? 0
			: symbol[reloc[i].s2.index].n_value;
/*
 * local patch address
 */
		patch_addr = (int *) ((char *) ldatapt +
			(reloc[i].pc * sizeof(int)));
/*
 * Add in appropriate base to patch value.
 */
		if (reloc[i].s1_text) {
			patch_value = (s1_value - s2_value) + (int4) textpt;
		} else {
			patch_value = (((s1_value * sizeof(int))
					- (s2_value * sizeof(int)))
					+ (int4) datapt);
		}
/*
 * Add in constant.
 */
		patch_value += reloc[i].cnst;
/*
 * Patch that word!
 */
		*patch_addr = ltot(patch_value);
	}
}

/*
 *	touttextsize
 *	touthtxtsize
 *	toutdatasize
 *	toutbsssize
 *	toutloadsize
 *	touttextptr
 *	toutdataptr
 *	toutrelptr
 *	toutentry
 *
 *	Function:	- extracts information about a t.out load module
 */

int4 touttextsize(t) void *t;
{
	struct exec *e;
	e = (struct exec *) t;
	return((int4) e->a_text);
}

int4 touthtxtsize(t) void *t;
{
	struct exec *e;
	e = (struct exec *) t;
	return((int4) e->a_text + sizeof(struct exec));
}

int4 toutdatasize(t) void *t;
{
	struct exec *e;
	e = (struct exec *) t;
	return((int4) e->a_data);
}

int4 toutbsssize(t) void *t;
{
	struct exec *e;
	e = (struct exec *) t;
	return((int4) e->a_bss);
}

int4 toutloadsize(t) void *t;
{
	struct exec *e;
	e = (struct exec *) t;
	return((int4) (sizeof(struct exec) + e->a_text + e->a_drsize +
		e->a_syms + e->a_data));
}

void *touttextptr(t) void *t;
{
	char *p;
	p = (char *) t;
	return((void *) (p + sizeof(struct exec)));
}

void *toutdataptr(t) void *t;
{
	char *p;
	struct exec *e;
	p = (char *) t;
	e = (struct exec *) t;
	return((void *) (p + sizeof(struct exec) + e->a_text +
		e->a_drsize + e->a_syms));
}

void *toutrelptr(t) void *t;
{
	char *p;
	struct exec *e;
	p = (char *) t;
	e = (struct exec *) t;
	return((void *) (p + sizeof(struct exec) + e->a_text));
}

int4 toutentry(t) void *t;
{
	struct exec *e;
	e = (struct exec *) t;
	return((int4) e->a_entry + sizeof(struct exec));
}

/*
 *	toutrev
 *
 *	Function:	- reorder the bytes of an executable's header,
 *			  reloc and symbol tables
 *
 *	Accepts:	- pointer to locally ordered executable header
 *			- pointer to executable header (to be reversed)
 *			- pointer to relocation table
 *			- pointer to symbol table
 */

/*
 * external functions
 */
extern int		rw();		/* reverse word */

void
toutrev(ohead, thead, reloc, symbol)

struct exec		*ohead;		/* locally ordered executable header */
struct exec		*thead;		/* executable header */
struct to_relative	*reloc;		/* relocation table */
struct nlist		*symbol;	/* symbol table */

{
	int		i;		/* favourite index variable */
	int		relocsize;	/* #entries in relocation table */
	int		symbolsize;	/* #entries in symbol table */

/*
 * Reorder the relocation table.
 */
	relocsize = ohead->a_drsize / sizeof(struct to_relative);

	for (i = 0; i < relocsize; ++i) {
		(*(int *) &reloc[i]) = rw((int) *(int *) &reloc[i]);
		reloc[i].s1.index = rw((int) reloc[i].s1.index);
		reloc[i].s2.index = rw((int) reloc[i].s2.index);
		reloc[i].cnst = rw((int) reloc[i].cnst);
	}
/*
 * Reorder the symbol table.
 */
	symbolsize = ohead->a_syms / sizeof(struct nlist);

	for (i = 0; i < symbolsize; ++i) {
		symbol[i].n_value = rw((int) symbol[i].n_value);
	}
/*
 * Reorder the header.
 */
	thead->a_magic = rw((int) thead->a_magic);
	thead->a_text = rw((int) thead->a_text);
	thead->a_data = rw((int) thead->a_data);
	thead->a_bss = rw((int) thead->a_bss);
	thead->a_syms = rw((int) thead->a_syms);
	thead->a_drsize = rw((int) thead->a_drsize);
	thead->a_entry = rw((int) thead->a_entry);
}
