/*
 *	OSU Research Computing, architects and contractors	Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	rfcubix.c,v $
 * Revision 2.2  92/04/30  14:41:36  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/10/27  20:50:12  gdburns
 * Change O_ASYNC to O_NOCBX.
 * 
 * Revision 2.1  91/02/24  20:57:46  gdburns
 * Ohio Release
 * 
 *	Function:	- cubix specific filed requests and runtime
 *			  functions
 */

#include <events.h>
#include <freq.h>
#include <ksignal.h>
#include <net.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* error number */
extern struct fclient	_ufd[FUMAX];	/* client file descriptor table */

/*
 * external functions
 */
extern int		_fdvalid();	/* check validity of file desc */

/*
 * backwards compatible
 */
int tmulti(fd) int fd; { return(multi(fd)); }
int tsingl(fd) int fd; { return(singl(fd)); }
int tasync(fd) int fd; { return(async(fd)); }
int torder(fd, seq) int fd; int4 seq; { return(order(fd, seq)); }

int ismulti(fd) int fd;
{ return(_fdvalid(fd) && (_ufd[fd].fu_tflags & FMULTI)); }

int issingl(fd) int fd;
{ return(_fdvalid(fd) && (_ufd[fd].fu_tflags & FSINGL)); }

int isasync(fd) int fd;
{ return(_fdvalid(fd) && (_ufd[fd].fu_tflags & O_NOCBX)); }

int multi(fd) int fd; { return(rfchcx(fd, FMULTI)); }

int singl(fd) int fd; { return(rfchcx(fd, FSINGL)); }

int async(fd) int fd; { return(rfchcx(fd, O_NOCBX)); }

int order(fd, seq) int fd; int4 seq; { return(rforder(fd, seq)); }

/*
 *	rfchcx
 *
 *	Function:	- change the cubix access method in
 *			  use for an open file
 *
 *	Accepts:	- open file descriptor
 *			- new cubix access method
 *
 *	Returns:	- 0 or ERROR
 */

int
rfchcx(fd, flags)

int			fd;
int4			flags;

{
	struct nmsg	nhead;		/* network message desc. */
	struct freq  	*request;	/* filed request */
	struct freply	*reply;		/* filed reply */
	int		mask;		/* signal mask */

	if (_fdvalid(fd) == ERROR) {
		return(ERROR);
	}

	request = (struct freq *) nhead.nh_data;
	reply = (struct freply *) nhead.nh_data;

	request->fq_src_node = getnodeid();
	request->fq_src_event = -getpid();
	request->fq_req = FQCHCX;
	request->fq_tfd = _ufd[fd].fu_tfd;
	request->fq_newflags = flags;

	if ((fd <= 2) && (_ufd[fd].fu_node == NOTNODEID)) {
		_ufd[fd].fu_node = getorigin();
	}

	nhead.nh_node = _ufd[fd].fu_node;
	nhead.nh_event = EVFILED;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	if (! ((_ufd[fd].fu_node & FCUBIX) || (flags & FCUBIX))) {
		mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	}

	nhead.nh_event = -getpid();

	if (nrecv(&nhead)) {
		return(ERROR);
	}

	if (! ((_ufd[fd].fu_node & FCUBIX) || (flags & FCUBIX))) {
		(void) ksigsetmask(mask);
	}

	if (reply->fr_errno != 0) {
		errno = reply->fr_errno;
		return(ERROR);
	}

	else {
		_ufd[fd].fu_tflags &= ~FMETHOD;
		_ufd[fd].fu_tflags |= flags & FMETHOD;
		_ufd[fd].fu_tfd = reply->fr_ret;
		return(0);
	}
}

/*
 *	rforder
 *
 *	Function:	- change the cubix member ordering for MULTI mode
 *			- sequence number range from 0 to #members - 1
 *
 *	Accepts:	- open file descriptor
 *			- new sequence number
 *
 *	Returns:	- 0 or ERROR
 */

int
rforder(fd, seq)

int			fd;
int4			seq;

{
	struct nmsg	nhead;		/* network message desc. */
	struct freq  	*request;	/* filed request */
	struct freply	*reply;		/* filed reply */
	int		mask;		/* signal mask */

	if (_fdvalid(fd) == ERROR) {
		return(ERROR);
	}

	request = (struct freq *) nhead.nh_data;
	reply = (struct freply *) nhead.nh_data;

	request->fq_src_node = getnodeid();
	request->fq_src_event = -getpid();
	request->fq_req = FQORDER;
	request->fq_tfd = _ufd[fd].fu_tfd;
	request->fq_seq = seq;

	if ((fd <= 2) && (_ufd[fd].fu_node == NOTNODEID)) {
		_ufd[fd].fu_node = getorigin();
	}

	nhead.nh_node = _ufd[fd].fu_node;
	nhead.nh_event = EVFILED;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();

	if (nrecv(&nhead)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);

	if (reply->fr_errno != 0) {
		errno = reply->fr_errno;
		return(ERROR);
	}

	else {
		return(0);
	}
}
