/*
 *	OSU Research Computing, architects and contractors	Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	rflseek.c,v $
 * Revision 2.2  92/04/30  14:41:37  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/10/24  11:04:39  gdburns
 * Declare global functions for tempermental compilers.
 * 
 * Revision 2.1  91/02/24  20:57:48  gdburns
 * Ohio Release
 *
 *	Function:	- seek on a file using remote filed
 *
 *	Accepts:	- runtime file descriptor
 *			- seek offset
 *			- type of seek operation
 *
 *	Returns:	- 0 or ERROR
 */

#include <events.h>
#include <freq.h>
#include <ksignal.h>
#include <net.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* error number */
extern struct fclient	_ufd[FUMAX];	/* client file descriptor table */

/*
 * external functions
 */
extern int		_fdvalid();	/* check validity of file desc */

/*
 * global functions
 */
int4			rflseek();

/*
 * backwards compatible
 */
int4 tlseek(fd, offset, whence) int fd; int4 offset; int whence;
{ return(rflseek(fd, offset, whence)); }

int4
rflseek(fd, offset, whence)

int			fd;
int4			offset;
int			whence;

{
	struct nmsg	nhead;		/* network message desc. */
	struct freq  	*request;	/* filed request */
	struct freply	*reply;		/* filed reply */
	int		mask;		/* signal mask */

/*
 * Do simple error checking.
 */
	if (_fdvalid(fd) == ERROR) {
		return((int4) ERROR);
	}

	request = (struct freq *) nhead.nh_data;
	reply = (struct freply *) nhead.nh_data;

	request->fq_src_node = getnodeid();
	request->fq_src_event = -getpid();
	request->fq_req = FQLSEEK;
	request->fq_tfd = _ufd[fd].fu_tfd;
	request->fq_offset = offset;
	request->fq_whence = (int4) whence;

	if ((fd <= 2) && (_ufd[fd].fu_node == NOTNODEID)) {
		_ufd[fd].fu_node = getorigin();
	}

	nhead.nh_node = _ufd[fd].fu_node;
	nhead.nh_event = EVFILED;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return((int4) ERROR);
	}

	if ( !(_ufd[fd].fu_tflags & FCUBIX)) {
		mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	}

	nhead.nh_event = -getpid();

	if (nrecv(&nhead)) {
		return((int4) ERROR);
	}

	if (reply->fr_errno != 0) {
		errno = reply->fr_errno;
		return((int4) ERROR);
	}

	if ( !(_ufd[fd].fu_tflags & FCUBIX)) {
		(void) ksigsetmask(mask);
	}

	return(reply->fr_ret);
}
