/*
 *	OSU Research Computing, architects and contractors	Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	rfopen.c,v $
 * Revision 2.2  92/04/30  14:41:39  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/10/27  20:50:30  gdburns
 * Change O_ASYNC to O_NOCBX.
 * 
 * Revision 2.1  91/02/24  20:57:50  gdburns
 * Ohio Release
 *
 *	Function:	- open a file using remote filed
 *			- server nodeid may be specified by prefixing
 *			  filename with <nodeid>: as in 100:temp
 *
 *	Accepts:	- ptr to file name
 *			- open flags
 *			- mode for newly created files
 *
 *	Returns:	- remote file descriptor
 */

#include <events.h>
#include <errno.h>
#include <freq.h>
#include <ksignal.h>
#include <net.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* error number */
extern struct fclient	_ufd[FUMAX];	/* client file descriptor table */

/*
 * external functions
 */
extern char		*_fnparse();	/* parse nodeid from file name */

/*
 * backwards compatible
 */
int topen(name, flags, mode) char *name; int4 flags; int4 mode;
{ return(rfopen(name, flags, mode)); }

int
rfopen(name, flags, mode)

char			*name;
int4			flags;
int4			mode;

{
	struct nmsg	nhead;		/* network message desc. */
	struct freq  	*request;	/* filed request */
	struct freply	*reply;		/* filed reply */
	int		mask;		/* signal mask */
	int		fd;		/* new file descriptor (ufd) */

/*
 * Find slot in client fd table.
 */
	for (fd = 3; (fd < FUMAX) && (_ufd[fd].fu_tflags != 0); ++fd) {
	}

	if (fd >= FUMAX) {
		errno = ENFILE;
		return(ERROR);
	}
/*
 * Determine server node ID.
 */
	name = _fnparse(name, &nhead.nh_node);
		
	request = (struct freq *) nhead.nh_data;
	reply = (struct freply *) nhead.nh_data;

	request->fq_src_node = getnodeid();
	request->fq_src_event = -getpid();
	request->fq_req = FQOPEN;
	request->fq_mode = mode;

	if (! (flags & FCUBIX)) {
		flags |= O_NOCBX;
	}

	request->fq_flags = flags;

	nhead.nh_event = EVFILED;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = strlen(name) + 1;

	if (nhead.nh_length > MAXNMSGLEN) {
		errno = ENAMETOOLONG;
		return(ERROR);
	}

	nhead.nh_msg = name;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	if (! (_ufd[fd].fu_tflags & FCUBIX)) {
		mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	}

	nhead.nh_event = -getpid();
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nrecv(&nhead)) {
		return(ERROR);
	}

	if (! (_ufd[fd].fu_tflags & FCUBIX)) {
		(void) ksigsetmask(mask);
	}

	if (reply->fr_errno != 0) {
		errno = reply->fr_errno;
		return(ERROR);
	}

	else {
		_ufd[fd].fu_tflags = flags;
		_ufd[fd].fu_node = nhead.nh_node;
		_ufd[fd].fu_tfd = reply->fr_ret;
		return(fd);
	}
}
