/*
 *	OSU Research Computing, architects and contractors	Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	rfread.c,v $
 * Revision 2.2  92/04/30  14:41:41  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/10/24  11:05:14  gdburns
 * Declare global functions for tempermental compilers.
 * 
 * Revision 2.1  91/02/24  20:57:51  gdburns
 * Ohio Release
 *
 *	Function:	- read from a file using remote filed
 *
 *	Accepts:	- runtime file descriptor
 *			- buffer ptr
 *			- buffer length
 *
 *	Returns:	- 0 or ERROR
 */

#include <events.h>
#include <freq.h>
#include <ksignal.h>
#include <net.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* error number */
extern struct fclient	_ufd[FUMAX];	/* client file descriptor table */

/*
 * external functions
 */
extern int		_fdvalid();	/* check validity of file desc */

/*
 * global functions
 */
int4			rfread();

/*
 * backwards compatible
 */
int4 tread(fd, buf, len) int fd; char *buf; int4 len;
{ return(rfread(fd, buf, len)); }

int4
rfread(fd, buf, len)

int			fd;
char			*buf;
int4			len;

{
	struct nmsg	nhead;		/* network message desc. */
	struct freq  	*request;	/* filed request */
	struct freply	*reply;		/* filed reply */
	int		mask;		/* signal mask */
	int4		tot_len = 0;	/* total length written */
	int4		pkt_len;	/* one packet length */

/*
 * Do simple error checking.
 */
	if (_fdvalid(fd) == ERROR) {
		return((int4) ERROR);
	}

	request = (struct freq *) nhead.nh_data;
	reply = (struct freply *) nhead.nh_data;

	request->fq_src_node = getnodeid();
	request->fq_src_event = -getpid();
	request->fq_req = FQREAD;
	request->fq_tfd = _ufd[fd].fu_tfd;
	request->fq_length = len;

	if ((fd <= 2) && (_ufd[fd].fu_node == NOTNODEID)) {
		_ufd[fd].fu_node = getorigin();
	}

	nhead.nh_node = _ufd[fd].fu_node;
	nhead.nh_event = EVFILED;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return((int4) ERROR);
	}

	if (! (_ufd[fd].fu_tflags & FCUBIX)) {
		mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	}
/*
 * Receive input buffer packets.
 */
	nhead.nh_msg = buf;
	nhead.nh_event = -getpid();

	do {
		nhead.nh_type = 0;
		nhead.nh_length = (len > MAXNMSGLEN) ? MAXNMSGLEN : len;
		pkt_len = nhead.nh_length;

		if (nrecv(&nhead)) {
			return((int4) ERROR);
		}

		if (reply->fr_errno != 0) {
			errno = reply->fr_errno;
			return((int4) ERROR);
		}

		else if (reply->fr_ret != pkt_len) {
			len = 0;
		}

		else {
			len -= pkt_len;
			nhead.nh_msg += pkt_len;
		}

		tot_len += reply->fr_ret;
	} while (len > 0);

	if (! (_ufd[fd].fu_tflags & FCUBIX)) {
		(void) ksigsetmask(mask);
	}

	return(tot_len);
}
