/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	rfsimplex.c,v $
 * Revision 2.2  92/04/30  14:41:42  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.4  92/02/28  10:47:30  gdburns
 * Pull out rfstat() and rffstat().
 * 
 * Revision 2.1.1.3  91/10/24  13:03:30  gdburns
 * Make sure MAXPATHLEN is defined.
 * This should probably be converted to POSIX.
 * 
 * Revision 2.1.1.2  91/10/09  14:08:51  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1.1.1  91/05/22  16:18:21  raja
 * Enhanced portability by casting few constants for int4.
 * 
 * Revision 2.1  91/02/24  20:57:53  gdburns
 * Ohio Release
 *
 *	Function:	- simple, basic, filed requests
 */

#include <errno.h>
#include <freq.h>
#include <net.h>
#include <typical.h>
#include <t_types.h>

#undef NULL
#define NULL		((char *) 0)

#ifndef MAXPATHLEN
#define MAXPATHLEN	1024
#endif

/*
 * global functions
 */
int			rfaccess();
int			rfchdir();
int			rfdup();
int			rfdup2();
int			rfincr();
int			rfftruncate();
int			rfmkdir();
int			rfrmdir();
int			rfrmfd();
int			rfsystem();
int			rftruncate();
int			rfunlink();
char			*rfgetwd();

/*
 * external variables
 */
extern int		errno;		/* error number */
extern struct fclient	_ufd[FUMAX];	/* client file descriptor table */

/*
 * external functions
 */
extern int		_fdvalid();	/* check validity of file desc */
extern int4		_makefreq();	/* make simple file request */
extern char		*_fnparse();	/* parse nodeid from file name */

/*
 * backwards compatibility
 */
int tmkdir(fname, mode) char *fname; int4 mode;
{ return(rfmkdir(fname, mode)); }
int taccess(fname, mode) char *fname; int4 mode;
{ return(rfaccess(fname, mode)); }
int tchdir(node, newdir) int4 node; char *newdir;
{ return(rfchdir(node, newdir)); }
char *tgetwd(node, currdir) int4 node; char *currdir;
{ return(rfgetwd(node, currdir)); }
int tsystem(node, cmd) int4 node; char *cmd;
{ return(rfsystem(node, cmd)); }
int tunlink(fname) char *fname;
{ return(rfunlink(fname)); }
int trmdir(fname) char *fname;
{ return(rfrmdir(fname)); }
int tincr(fd, amount) int fd; int4 amount;
{ return(rfincr(fd, amount)); }
int tdup(ofd) int ofd;
{ return(rfdup(ofd)); }
int tdup2(ofd, nfd) int ofd; int nfd;
{ return(rfdup2(ofd, nfd)); }
int ttruncate(fname, len) char *fname; int4 len;
{ return(rftruncate(fname, len)); }
int tftruncate(fd, len) int fd; int4 len;
{ return(rfftruncate(fd, len)); }

/*
 *	rfmkdir
 *
 *	Function:	- make a new directory
 *	Accepts:	- ptr to extended file name
 *			- access mode for directory
 *	Returns:	- 0 or ERROR
 */

int
rfmkdir(fname, mode)

char			*fname;
int4			mode;

{
	int4		node;			/* server node ID */
	int4		len;			/* length of file name */

	fname = _fnparse(fname, &node);
	len = (int4) (strlen(fname) + 1);
	if (len > MAXNMSGLEN) len = MAXNMSGLEN;
	return((int) _makefreq(node, FQMKDIR, mode, fname, len, NULL,
		INT4_NIL));
}

/*
 *	rfaccess
 *
 *	Function:	- check accessibility of a file
 *	Accepts:	- ptr to extended file name
 *			- access mode sought
 *	Returns:	- 0 or ERROR
 */

int
rfaccess(fname, mode)

char			*fname;
int4			mode;

{
	int4		node;			/* server node ID */
	int4		len;			/* length of file name */

	fname = _fnparse(fname, &node);
	len = (int4) (strlen(fname) + 1);
	if (len > MAXNMSGLEN) len = MAXNMSGLEN;
	return((int) _makefreq(node, FQACCESS, mode, fname, len, NULL,
		INT4_NIL));
}

/*
 *	rfchdir
 *
 *	Function:	- change the current working directory
 *	Accepts:	- nodeid of filed
 *			- new working directory
 *	Returns:	- 0 or ERROR
 */

int
rfchdir(node, newdir)

int4			node;
char			*newdir;

{
	int4		len;

	len = (int4) (strlen(newdir) + 1);
	if (len > MAXNMSGLEN) len = MAXNMSGLEN;
	return((int) _makefreq(node, FQCHDIR, INT4_NIL, newdir, len, NULL,
		INT4_NIL));
}

/*
 *	rfgetwd
 *
 *	Function:	- get the current working directory
 *	Accepts:	- nodeid of filed
 *			- caller's buffer for working directory
 *	Returns:	- 0 or ERROR
 */

char *
rfgetwd(node, currdir)

int4			node;
char			*currdir;

{
	if (_makefreq(node, FQGETWD, INT4_NIL, NULL, INT4_NIL,
		currdir, (int4) MAXPATHLEN) == ERROR) {

		return(NULL);
	} else {
		return(currdir);
	}
}

/*
 *	rfrmfd
 *
 *	Function:	- remove a filed descriptor
 *	Accepts:	- nodeid of filed
 *			- file descriptor handle
 *	Returns:	- 0 or ERROR
 */

int
rfrmfd(node, tfd)

int4			node;
int4			tfd;

{
	return((int) _makefreq(node, FQRMFD, tfd, NULL, INT4_NIL, NULL,
		INT4_NIL));
}

/*
 *	rfsystem
 *
 *	Function:	- issue a shell command
 *	Accepts:	- node ID of filed
 *			- shell command
 *	Returns:	- exit status
 */

int
rfsystem(node, cmd)

int4			node;
char			*cmd;

{
	int4		len;

	len = (int4) (strlen(cmd) + 1);
	if (len > MAXNMSGLEN) len = MAXNMSGLEN;
	return((int) _makefreq(node, FQSYSTEM, INT4_NIL, cmd, len, NULL,
		INT4_NIL));
}

/*
 *	rfunlink
 *
 *	Function:	- remove a file
 *	Accepts:	- extended file name
 *	Returns:	- 0 or ERROR
 */

int
rfunlink(fname)

char			*fname;

{
	int4		node;			/* server node ID */
	int4		len;

	fname = _fnparse(fname, &node);
	len = (int4) (strlen(fname) + 1);
	if (len > MAXNMSGLEN) len = MAXNMSGLEN;
	return((int) _makefreq(node, FQUNLINK, INT4_NIL, fname, len, NULL,
		INT4_NIL));
}

/*
 *	rfrmdir
 *
 *	Function:	- remove a directory
 *	Accepts:	- pathname of directory
 *	Returns:	- 0 or ERROR
 */

int
rfrmdir(fname)

char			*fname;

{
	int4		node;			/* server node ID */
	int4		len;

	fname = _fnparse(fname, &node);
	len = (int4) (strlen(fname) + 1);
	if (len > MAXNMSGLEN) len = MAXNMSGLEN;
	return((int) _makefreq(node, FQRMDIR, INT4_NIL, fname, len, NULL,
		INT4_NIL));
}

/*
 *	rfftruncate
 *
 *	Function:	- set an open file to a specified length
 *	Accepts:	- file descriptor handle
 *			- new file length
 *	Returns:	- 0 or ERROR
 */

int
rfftruncate(fd, len)

int			fd;
int4			len;

{
	if (_fdvalid(fd) == ERROR) {
		return(ERROR);
	}

	if ((fd <= 2) && (_ufd[fd].fu_node == NOTNODEID)) {
		_ufd[fd].fu_node = getorigin();
	}

	len = ltot(len);
	return((int) _makefreq(_ufd[fd].fu_node, FQFTRUNC, _ufd[fd].fu_tfd,
		(char *) &len, (int4) sizeof(int4), NULL, INT4_NIL));
}

/*
 *	rftruncate
 *
 *	Function:	- set a named file to a specified length
 *	Accepts:	- file name
 *			- new file length
 *	Returns:	- 0 or ERROR
 */

int
rftruncate(fname, len)

char			*fname;
int4			len;

{
	int4		node;			/* server node ID */
	int4		lfname;			/* length of fname */

	fname = _fnparse(fname, &node);
	lfname = (int4) (strlen(fname) + 1);
	if (lfname > MAXNMSGLEN) lfname = MAXNMSGLEN;
	return((int) _makefreq(node, FQTRUNC, len, fname, lfname, NULL,
		INT4_NIL));
}

/*
 *	rfincr
 *
 *	Function:	- increment usage count of a file descriptor
 *	Accepts:	- file descriptor handle
 *			- increment amount
 *	Returns:	- 0 or ERROR
 */

int
rfincr(fd, amount)

int			fd;
int4			amount;

{
	if (_fdvalid(fd) == ERROR) {
		return(ERROR);
	}

	if ((fd <= 2) && (_ufd[fd].fu_node == NOTNODEID)) {
		_ufd[fd].fu_node = getorigin();
	}

	amount = ltot(amount);
	return((int) _makefreq(_ufd[fd].fu_node, FQINCR, _ufd[fd].fu_tfd,
		(char *) &amount, (int4) sizeof(int4), NULL, INT4_NIL));
}

/*
 *	rfdup
 *
 *	Function:	- duplicate file descriptor
 *	Accepts:	- existing open file descriptor
 *	Returns:	- 0 or ERROR
 */

int
rfdup(ofd)

int			ofd;

{
	int		nfd;			/* new file descriptor */

	if (_fdvalid(ofd) == ERROR) {
		return(ERROR);
	}
/*
 * Find slot in client fd table.
 */
	for (nfd = 3; (nfd < FUMAX) && (_ufd[nfd].fu_tflags != 0); ++nfd) {
	}

	if (nfd >= FUMAX) {
		errno = ENFILE;
		return(ERROR);
	}

	if ((ofd <= 2) && (_ufd[ofd].fu_node == NOTNODEID)) {
		_ufd[ofd].fu_node = getorigin();
	}

	_ufd[nfd].fu_tfd = _makefreq(_ufd[ofd].fu_node, FQDUP,
		_ufd[ofd].fu_tfd, NULL, INT4_NIL, NULL, INT4_NIL);

	if (_ufd[nfd].fu_tfd == ERROR) {
		return(ERROR);
	} else {
		_ufd[nfd].fu_tflags = _ufd[ofd].fu_tflags;
		_ufd[nfd].fu_node = _ufd[ofd].fu_node;
		return(nfd);
	}
}

/*
 *	rfdup2
 *
 *	Function:	- duplicate file descriptor to specific handle
 *	Accepts:	- existing open file descriptor
 *			- new file descriptor
 *	Returns:	- 0 or ERROR
 */

int
rfdup2(ofd, nfd)

int			ofd;
int			nfd;

{
	if (_fdvalid(ofd) == ERROR) {
		return(ERROR);
	}

	if ((nfd < 0) || (nfd >= FUMAX)) {
		errno = EBADF;
		return(ERROR);
	}

	if (_ufd[nfd].fu_tflags != 0) {

		if (rfclose(nfd) == ERROR) {
			return(ERROR);
		}
	}

	if ((ofd <= 2) && (_ufd[ofd].fu_node == NOTNODEID)) {
		_ufd[ofd].fu_node = getorigin();
	}

	_ufd[nfd].fu_tfd = _makefreq(_ufd[ofd].fu_node, FQDUP,
		_ufd[ofd].fu_tfd, NULL, INT4_NIL, NULL, INT4_NIL);

	if (_ufd[nfd].fu_tfd == ERROR) {
		return(ERROR);
	} else {
		_ufd[nfd].fu_tflags = _ufd[ofd].fu_tflags;
		_ufd[nfd].fu_node = _ufd[ofd].fu_node;
		return(nfd);
	}
}
