/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	rfstat.c,v $
 * Revision 2.2  92/04/30  14:41:44  trillium
 * Ohio Release
 * 
 * Revision 2.1  92/02/28  10:46:32  gdburns
 * initial revision
 * 
 *	Function:	- rffstat() and rfstat()
 *			- get file status
 *			- SunOS version
 */

#include <freq.h>
#include <net.h>
#include <sys/stat.h>
#include <typical.h>

#undef NULL
#define NULL		((char *) 0)

/*
 * global functions
 */
int			rffstat();
int			rfstat();

/*
 * external variables
 */
extern int		errno;		/* error number */
extern struct fclient	_ufd[FUMAX];	/* client file descriptor table */

/*
 * external functions
 */
extern int		_fdvalid();	/* check validity of file desc */
extern int4		_makefreq();	/* make simple file request */
extern char		*_fnparse();	/* parse nodeid from file name */

/*
 * local functions
 */
static void		conv_stat();	/* convert to status structure */

/*
 * backwards compatibility
 */
int tstat(fname, buf) char *fname; char *buf;
{ return(rfstat(fname, buf)); }
int tfstat(fd, buf) int fd; char *buf;
{ return(rffstat(fd, buf)); }

/*
 *	rfstat
 *
 *	Function:	- get status on a named file
 *	Accepts:	- file name
 *			- ptr to status buffer
 *	Returns:	- 0 or ERROR
 */

int
rfstat(fname, status)

char			*fname;
char			*status;

{
	int4		node;			/* server node ID */
	int4		len;
	uint4		buf[17];
	int		ret;

	fname = _fnparse(fname, &node);
	len = (int4) (strlen(fname) + 1);
	if (len > MAXNMSGLEN) len = MAXNMSGLEN;

	ret = _makefreq(node, FQSTAT, INT4_NIL, fname, len, (char *) buf,
		(int4) sizeof(buf));

	if (ret == 0) {
		conv_stat((struct stat *) status, buf);
	}

	return(ret);
}

/*
 *	rffstat
 *
 *	Function:	- get status on a file descriptor
 *	Accepts:	- file descriptor handle
 *			- ptr to status buffer
 *	Returns:	- 0 or ERROR
 */

int
rffstat(fd, status)

int			fd;
char			*status;

{
	uint4		buf[17];
	int		ret;

	if (_fdvalid(fd) == ERROR) {
		return(ERROR);
	}

	if ((fd <= 2) && (_ufd[fd].fu_node == NOTNODEID)) {
		_ufd[fd].fu_node = getorigin();
	}

	ret = _makefreq(_ufd[fd].fu_node, FQFSTAT, _ufd[fd].fu_tfd,
		NULL, INT4_NIL, (char *) buf, (int4) sizeof(buf));

	if (ret == 0) {
		conv_stat((struct stat *) status, buf);
	}

	return(ret);
}

/*
 *	conv_stat
 *
 *	Function:	- stores an int4 array into a status structure
 *	Accepts:	- ptr to stat structure
 *			- ptr to int4 array
 */

static void
conv_stat(status, buf)

struct stat		*status;
uint4			*buf;

{
	status->st_dev = buf[0];
	status->st_ino = buf[1];
	status->st_mode = buf[2];
	status->st_nlink = buf[3];
	status->st_uid = buf[4];
	status->st_gid = buf[5];
	status->st_rdev = buf[6];
	status->st_size = buf[7];
	status->st_atime = buf[8];
	status->st_spare1 = buf[9];
	status->st_mtime = buf[10];
	status->st_spare2 = buf[11];
	status->st_ctime = buf[12];
	status->st_spare3 = buf[13];
	status->st_blksize = buf[14];
	status->st_blocks = buf[15];
	status->st_spare4[0] = buf[16];
	status->st_spare4[1] = buf[17];
}
