/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	rfstate.c,v $
 * Revision 2.2  92/04/30  14:41:49  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/03/02  18:16:23  gdburns
 * Replace bcopy() with memcpy().
 * 
 * Revision 2.1.1.2  92/03/02  18:07:03  gdburns
 * Replace bcopy() with memcpy().
 * 
 * Revision 2.1.1.1  92/02/23  22:32:16  gdburns
 * Remove NOBUF.
 * 
 * Revision 2.1  91/02/24  20:57:55  gdburns
 * Ohio Release
 * 
 *	Function:	- gather state from a remote filed
 *			- receives file descriptors from
 *			  filed until reply code is EEOF
 *
 *	Accepts:	- target node ID
 *			- ptr to file descriptor table
 *			- # of entries in table
 *
 *	Returns:	- number of replied descriptors or ERROR
 */

#include <freq.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>

/*
 * external variables
 */
extern int		errno;		/* error number */

/*
 * local variables
 */
static char		hole[MAXNMSGLEN];

int
rfstate(nodeid, ftbl, maxcount)

int			nodeid;		/* target node ID */
int			maxcount;	/* count of descriptor table */
struct fstate		ftbl[];		/* ptr to descriptor table */

{
	struct nmsg	nhead;		/* network message desc. */
	struct freq  	*request;	/* filed request */
	struct freply	*reply;		/* filed reply */
	struct fstate	*fhole;		/* ptr incoming file descs */
	int		count;		/* # descs in this msg */
	int		totcount;	/* total # replied descs */
	int		i;		/* favourite index */
	int		mask;		/* signal mask */


	totcount = 0;
	fhole = (struct fstate *) hole;
	request = (struct freq *) nhead.nh_data;
	reply = (struct freply *) nhead.nh_data;

	request->fq_src_node = getnodeid();
	request->fq_src_event = -getpid();
	request->fq_req = FQSTATE;

	nhead.nh_node = nodeid;
	nhead.nh_event = EVFILED;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();
	nhead.nh_msg = hole;

	do {
		nhead.nh_length = MAXNMSGLEN;

		if (nrecv(&nhead)) {
			return(ERROR);
		}

		count = nhead.nh_length / sizeof(struct fstate);
		totcount += count;

		if (maxcount > 0) {
/*
 * Convert byte ordering (sigh).
 */
			for (i = 0; i < count; ++i) {
				fhole[i].fs_tfd = ttol(fhole[i].fs_tfd);
				fhole[i].fs_tflags = ttol(fhole[i].fs_tflags);
				fhole[i].fs_flow = ttol(fhole[i].fs_flow);
				fhole[i].fs_ncubix = ttol(fhole[i].fs_ncubix);
				fhole[i].fs_nheard = ttol(fhole[i].fs_nheard);
				fhole[i].fs_req = ttol(fhole[i].fs_req);
				fhole[i].fs_src_node =
					ttol(fhole[i].fs_src_node);
				fhole[i].fs_src_event =
					ttol(fhole[i].fs_src_event);
			}

			if (count > maxcount) count = maxcount;

			(void) memcpy((char *) ftbl, hole,
				count * sizeof(struct fstate));
			maxcount -= count;
			ftbl += count;
		}

	} while (! reply->fr_errno);

	(void) ksigsetmask(mask);

	if (reply->fr_errno == EEOF) {
		return(totcount);
	}

	else {
		errno = reply->fr_errno;
		return(ERROR);
	}
}
