/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	rfwrite.c,v $
 * Revision 2.2  92/04/30  14:41:50  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/23  22:33:03  gdburns
 * Use data conversion flags.
 * 
 * Revision 2.1.1.1  91/10/24  11:05:19  gdburns
 * Declare global functions for tempermental compilers.
 * 
 * Revision 2.1  91/02/24  20:57:57  gdburns
 * Ohio Release
 *
 *	Function:	- writes to a file using remote filed
 *	Accepts:	- runtime file descriptor
 *			- buffer ptr
 *			- buffer length
 *	Returns:	- 0 or ERROR
 */

#include <events.h>
#include <freq.h>
#include <ksignal.h>
#include <net.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;		/* error number */
extern struct fclient	_ufd[FUMAX];	/* client file descriptor table */

/*
 * external functions
 */
extern int		_fdvalid();	/* check validity of file desc */
extern int		strlen();

/*
 * global functions
 */
int4			rfwrite();

/*
 * backwards compatible
 */
int4 twrite(fd, buf, len) int fd; char *buf; int4 len;
{ return(rfwrite(fd, buf, len)); }
int4 tprint(buf) char *buf;
{ return(rfwrite(1, buf, strlen(buf))); }

int4
rfwrite(fd, buf, len)

int			fd;
char			*buf;
int4			len;

{
	struct nmsg	nhead;		/* network message desc. */
	struct freq  	*request;	/* filed request */
	struct freply	*reply;		/* filed reply */
	int		mask;		/* signal mask */
	int4		tot_len = 0;	/* total length written */
	int4		pkt_len;	/* one packet length */

/*
 * Do simple error checking.
 */
	if (_fdvalid(fd) < 0) {
		return((int4) ERROR);
	}

	request = (struct freq *) nhead.nh_data;
	reply = (struct freply *) nhead.nh_data;

	if ((fd <= 2) && (_ufd[fd].fu_node == NOTNODEID)) {
		_ufd[fd].fu_node = getorigin();
	}

	nhead.nh_node = _ufd[fd].fu_node;
	nhead.nh_flags = _ufd[fd].fu_tflags & O_DMSG;
/*
 * We need to synchronize if this is a cubix file.
 */
	if (_ufd[fd].fu_tflags & FCUBIX) {
		request->fq_src_node = getnodeid();
		request->fq_src_event = -getpid();
		request->fq_req = FQCXWRITE;
		request->fq_tfd = _ufd[fd].fu_tfd;
		request->fq_length = len;

		nhead.nh_event = EVFILED;
		nhead.nh_type = 0;
		nhead.nh_length = 0;
		nhead.nh_msg = (char *) 0;

		if (nsend(&nhead)) {
			return((int4) ERROR);
		}

		if (! (_ufd[fd].fu_tflags & FCUBIX)) {
			mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
		}

		nhead.nh_event = -getpid();

		if (nrecv(&nhead)) {
			return((int4) ERROR);
		}

		if (reply->fr_errno == ECXSINGL) {
			return(len - reply->fr_ret);
		}

		else if (reply->fr_errno != 0) {
			errno = reply->fr_errno;
			return((int4) ERROR);
		}
	}

	nhead.nh_msg = buf;
/*
 * Packetize output buffer.
 */
	do {
		request->fq_src_node = getnodeid();
		request->fq_src_event = -getpid();
		request->fq_req = FQWRITE;
		request->fq_tfd = _ufd[fd].fu_tfd;

		nhead.nh_event = EVFILED;
		nhead.nh_type = 0;
		nhead.nh_length = (len > MAXNMSGLEN) ? MAXNMSGLEN : len;
		pkt_len = nhead.nh_length;
		request->fq_length = pkt_len;

		if (nsend(&nhead)) {
			return((int4) ERROR);
		}

		if ((nhead.nh_msg == buf) && !(_ufd[fd].fu_tflags & FCUBIX)) {
			mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
		}

		if (! (_ufd[fd].fu_tflags & F1WAY)) {
			nhead.nh_event = -getpid();
			nhead.nh_type = 0;
			nhead.nh_length = 0;

			if (nrecv(&nhead)) {
				return((int4) ERROR);
			}

			if (reply->fr_errno != 0) {
				errno = reply->fr_errno;
				return((int4) ERROR);
			}

			else if (reply->fr_ret != pkt_len) {
				len = 0;
			}

			else {
				len -= pkt_len;
				nhead.nh_msg += pkt_len;
			}
		} else {
			reply->fr_ret = pkt_len;
			len -= pkt_len;
			nhead.nh_msg += pkt_len;
		}

		tot_len += reply->fr_ret;
	} while (len > 0);

	if ( !(_ufd[fd].fu_tflags & FCUBIX)) {
		(void) ksigsetmask(mask);
	}

	return(tot_len);
}
