/*
 *	OSU Research Computing, architects and contractors	Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	freq.h,v $
 * Revision 2.2  92/04/30  14:45:22  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/02/06  10:58:22  gdburns
 * Add data representation conversion flags.
 * 
 * Revision 2.1.1.2  91/10/27  20:50:50  gdburns
 * Change O_ASYNC to O_NOCBX.
 * 
 * Revision 2.1.1.1  91/05/22  16:24:27  raja
 * Enhanced portability by casting a few constants to int4.
 * 
 * Revision 2.1  91/02/27  02:53:49  gdburns
 * Ohio Release
 * 
 *	Function:	- filed request/reply templates and constants
 *
 */

#ifndef _FREQ
#define _FREQ

#include <portable.h>

/*
 * the bread and butter request codes
 */
#define FQOPEN		(0) 
#define FQCLOSE		(1) 
#define FQREAD		(2)
#define FQWRITE		(3) 
#define FQLSEEK		(4) 
#define FQCXWRITE	(5) 		/* cubix sync before write */

/*
 * Trollius specific requests
 */
#define FQCHCX		(6)		/* change access method */
#define FQORDER		(7)		/* change cubix node order */
#define FQIDLE		(8) 		/* for cubix - nothing in progress */
#define FQSTATE		(8)		/* filed status */
#define FQRMFD		(9)		/* blow away one or all file descs */
#define FQINCR		(10)		/* increment usage counter */

/*
 * request for the enthusiasts
 */
#define FQACCESS	(11)
#define FQMKDIR		(12) 
#define FQUNLINK	(13) 
#define FQSYSTEM	(14) 
#define FQCHDIR		(15) 
#define FQRMDIR		(16) 
#define FQGETWD		(17) 
#define FQSTAT		(18) 
#define FQFSTAT		(19) 
#define FQDUP		(20)
#define FQTRUNC		(21) 
#define FQFTRUNC	(22)

/*
 * requests for F77
 */
#define FQF77READ	(23)
#define FQF77BACK	(24)
#define FQMAX		(25)		/* # of valid requests */

/*
 * Trollius specific open flags
 */
#define	O_LOCK		((int4) 1 << 16)
#define	O_REUSE		((int4) 1 << 17)
#define	O_SINGL		((int4) 1 << 18)
#define	O_MULTI		((int4) 1 << 19)
#define	O_NOCBX		((int4) 1 << 20)
#define	O_DRAW		((int4) 0)
#define	O_DINT		((int4) 1 << 21)
#define	O_DFLT		((int4) 1 << 22)
#define	O_DDBL		((int4) 1 << 23)
#define	O_1WAY		((int4) 1 << 24)
#define O_FILED	(O_LOCK | O_REUSE | O_SINGL | O_MULTI | O_NOCBX | O_1WAY | O_DINT | O_DFLT | O_DDBL)
#define O_DMSG		(O_DINT | O_DFLT | O_DDBL)

/*
 * Trollius status flags
 */
#define FLOCK		O_LOCK
#define FSINGL		O_SINGL
#define FMULTI		O_MULTI
#define FNOCBX		O_NOCBX
#define FCUBIX		(FSINGL | FMULTI)
#define FMETHOD		(FSINGL | FMULTI | FNOCBX)
#define F1WAY		O_1WAY
#define FACTIVE		((int4) 1 << 25)

/*
 * limits
 */
#define FUMAX		32		/* per process fd limit, min 3 */

/*
 * special values
 */
#define NOTFD		-1		/* not a file descriptor handle */

/*
 * client (user) file descriptor
 */
struct fclient {
	int4		fu_node;	/* server node ID */
	int4		fu_tfd;		/* filed descriptor */
	int4		fu_tflags;	/* open flags, 0 - unused */
};

/*
 * filed reply
 */
struct	freply {
        int4		fr_errno;
        int4		fr_ret;
};

/*
 * filed request
 */
struct 	freq	{
	int4		fq_req;
	int4		fq_src_node;
	int4		fq_src_event;
	int4		fq_param0;
#define	fq_tfd		fq_param0
#define fq_flags	fq_param0
#define fq_accmode	fq_param0
#define fq_trunclen	fq_param0
	int4		fq_param1;
#define	fq_mode		fq_param1
#define	fq_length	fq_param1
#define	fq_offset	fq_param1
#define	fq_newflags	fq_param1
#define	fq_seq		fq_param1
	int4		fq_param2;
#define	fq_whence	fq_param2
#define	fq_fmt		fq_param2
};

struct fstate {
	int4		fs_tfd;		/* Trollius file descriptor */
	int4		fs_tflags;	/* Trollius peculiar flags */
	int4		fs_flow;	/* total data flow */
	int4		fs_src_node;	/* last client's node */
	int4		fs_src_event;	/* last client's node */
	int4		fs_ncubix;	/* # of cubix cast members */
	int4		fs_nheard;	/* # of heard members */
	int4		fs_req;		/* current cubix request */
	char		fs_name[24];	/* file name */
};

#endif
