/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	kernel.h,v $
 * Revision 2.2  92/04/30  14:45:30  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/01/27  17:50:38  gdburns
 * Expand process table.
 * 
 * Revision 2.1  91/03/20  11:34:49  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/11/16  15:08:29  raja
 * Change in the configuration of on-chip RAM.
 * 
 * Revision 2.0.1.1  90/07/13  19:28:20  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:55:06  trillium
 * Homo erectus
 *
 *	Function:	- kernel constants and data structure templates
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#ifndef _KERNEL
#define _KERNEL

/*
 * compile time tunable parameters
 */
#ifdef itb
#define PMAX		97		/* maximum number of kernel processes */
#define EMAX		101		/* maximum number of message events */
#else
#define PMAX		23		/* maximum number of kernel processes */
#define EMAX		29		/* maximum number of message events */
#define MAXHNAME	64		/* maximum size hostname */
#endif

/*
 * reserved values
 */
#define NOTPROCESS	0x80000000	/* not a process identifier */
#ifndef NOTEVENT
#define NOTEVENT	0x80000001	/* not an event number */
#endif
#define MININT		0x80000000	/* most negative number */
#undef  NULL
#define NULL		((struct kprocess *) 0)
					/* null process pointer */
/*
 * gateway to the kernel (OTB)
 */
#ifndef itb
#define KERNEL_PORT	2401		/* fixed kernel port number */
#define KERNEL_REPLY	1		/* reply from kernel type - msgop(2) */
#define KERNEL_DATA	2		/* data transfer - msgop(2) */
#define KERNEL_ID       'K'		/* token id for msg queue creation */
#endif

/*
 * process descriptor
 */
struct kprocess {
	struct kprocess	*k_pnext[2];	/* linked list fields */
#define	PPROCESS	0		/* process list */
#define	PBLKD		1		/* event blocked list */
	struct kprocess	*k_ptarget;	/* ptr to target process */
	int		k_pid;		/* process identifier */
	int		k_priority;	/* process priority */
	int		k_state;	/* process run state */
	int		k_signal;	/* pending signals */
	int		k_port;		/* communication port */
	int		k_aux;		/* auxiliary communication port */
	int		k_event;	/* blocked message event */
	int		k_type;		/* message type */
	int		k_client;	/* client's reply port */
	int		k_hold;		/* multi-purpose holding variable */
#define	k_size		k_hold		/* message length */
#define	k_errno		k_hold		/* reply error number */
#define	k_kqs_pid	k_hold		/* state request pid */
#define	k_index		k_hold		/* index of this descriptor */
};

/*
 * event descriptor
 */
struct kevent {
	int		ev_event;	/* message event */
	struct kprocess	*ev_head;	/* head of blocked list */
	struct kprocess	*ev_tail;	/* tail of blocked list */
};

#define rem(dividend, divisor) (((dividend) % (divisor)))
#define hash(hnum, hcons) ((int) rem(((unsigned) hnum), ((unsigned) hcons)))

#endif
