/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB
 *
 *	$Log:	kreq.h,v $
 * Revision 2.2  92/04/30  14:45:44  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:34:52  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:28:13  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:55:02  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:09:59  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:46:12  trillium
 * Initial revision
 * 
 *
 *	Function:	- declare kernel interface data structures
 *			- define kernel request constants
 */


#ifndef _KREQ

#define _KREQ

/*
 * kernel request codes
 */
#define KATTACH		1		/* attach to the kernel as a process */
#define KDETACH		2		/* detach a process from the kernel */
#define KSTATE		3		/* get the state of a process */
#define KDOOM		4		/* doom an existing process */
#define KTRY_SEND	5		/* try to send a message */
#define KSEND		6		/* send a message to a process */
#define KTRY_RECV	7		/* try to receive a message */
#define KRECV		8		/* receive a message from a process */

#define NORMAL		1		/* normalize request to zero */
#define NREQS		8		/* number of request types */

#define stor(x)		((x) += 2)	/* convert send to receive */
#define rtos(x)		((x) -= 2)	/* convert receive to send */

/*
 * process states
 */
#define NSTATES		9		/* number of distinct kernel states */
#define SSYNC		1		/* synchronized, ready to run */
#define SSTATE		2		/* needs state info, ready to run */
#define SERROR		3		/* unsuccessful request, ready to run */
#define SATTACH		4		/* reply process index, ready to run */
#define SREADY		5		/* ready to run */
#define SSEND_BLKD	6		/* send blocked on an event */
#define SRECV_BLKD	7		/* recv blocked on an event */
#define SACTIVE		8		/* running */
#define SFREE		9		/* unused */

/*
 * One function handles both blocking and non-blocking requests.
 */
#define krecv(x)	(do_krecv(x, KRECV))
#define ktry_recv(x)	(do_krecv(x, KTRY_RECV))
#define ksend(x)	(do_ksend(x, KSEND))
#define ktry_send(x)	(do_ksend(x, KTRY_SEND))

/*
 * kernel request template
 */
struct kreq {				/* kernel request format */
	int		kq_index;	/* requester's ptable index */
	int		kq_request;	/* request code */
	int		kq_parm1;	/* parameters */
#define	kqa_port	kq_parm1
#define	kqu_pid		kq_parm1
#define	kqd_pid		kq_parm1
#define	kqs_pid		kq_parm1
#define	kqm_event	kq_parm1
	int		kq_parm2;
#define	kqa_client	kq_parm2
#define kqd_signal	kq_parm2
#define kqm_type	kq_parm2
	int		kq_parm3;
#define kqa_priority	kq_parm3
#define kqm_length	kq_parm3
	int		kq_parm4;
#define kqa_pid		kq_parm4
#define kqm_aux		kq_parm4
};

/*
 * kernel reply template
 */
struct kreply {				/* kernel reply format */
	int		kr_reply;	/* kernel reply code */
	int		kr_signal;	/* delivered signals */
	int		kr_result1;	/* results */
#define	kra_index	kr_result1
#define	krs_state	kr_result1
#define	krm_port	kr_result1
	int		kr_result2;
#define	krs_priority	kr_result2
#define	krm_aux		kr_result2
	int		kr_result3;
#define	krs_event	kr_result3
#define	krm_length	kr_result3
	int		kr_result4;
#define	krs_signal	kr_result4
#define	krm_type	kr_result4
	int		kr_result5;
#define	krs_type	kr_result5
};

/*
 * kernel process status template - for client use only
 */
struct kstate {
	int		ks_state;
	int		ks_priority;
	int		ks_event;
	int		ks_type;
	int		ks_signal;
};

typedef struct kstate KS_STATE;


#endif
