/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	mareq.h,v $
 * Revision 2.2  92/04/30  14:45:54  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/11/10  23:24:32  gdburns
 * Minor overhaul fixes several problems.
 * 
 * Revision 2.1  91/03/20  11:34:58  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.4  90/11/16  15:05:03  raja
 * Change in the configuration of on-chip RAM.
 * 
 * Revision 2.0.1.3  90/09/13  11:20:52  gdburns
 * Removed machine dependencies into separate header file.
 * 
 * Revision 2.0.1.2  90/07/13  19:27:53  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/04  15:59:38  gdburns
 * Added more information to onchip RAM layout.
 * 
 * Revision 2.0  90/02/20  11:55:02  trillium
 * Homo erectus
 * 
 *	Function:	- define malloc request structures and constants
 *			- prefix is ma
 *			- based on Trollius 2.0, Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

/*
 * limitations
 */
#define MAMAX		64		/* #descriptors per table */
#define MASMAX		(4 * MAMAX)	/* max #descriptors for mstate */

/*
 * generic constants
 */
#define MANULL		((struct memblock *) 0)

/*
 * mallocd request codes
 */
#define MAQALLOC	0		/* malloc request */
#define MAQFREE		1		/* free request */
#define MAQSTATE	2		/* report state info */
#define MAQFREEALL	3		/* free all request */

/*
 * search criteria
 */
#define BYPTR		1
#define BYSIZE		2
#define BYSTART		3
#define BYPID		4
#define BYLAST		5

/*
 * block compaction
 */
#define COMPACT		1
#define NOCOMPACT	0

/*
 * mallocd request template
 */
struct mareq {
	int			maq_src_node;	/* source node */
	int			maq_src_event;	/* source event */
	int			maq_request;	/* malloc request type */
	int			maq_pid;	/* pid to be freed or mallocd */
	int			maq_node;	/* node used */
	int			maq_region;	/* memory region requested */
#define maq_address 		maq_region	/* address of memory to free */
#define maq_maxblks 		maq_region	/* max # descs client accepts */
	int			maq_amount;	/* amount of memory requested */
	int			maq_align;	/* address alignment */
};

/*
 * mallocd reply template
 */
struct mareply {
	int			mar_reply;	/* result of malloc */
	char		       *mar_address;	/* ptr to requested amount 
						   of memory */
	int			mar_nblks;	/* total # of mallocd descs */
};

/*
 * memory block descriptor
 */
struct memblock {
        int   	          	ma_node;	/* allocated to this node */
        int   	          	ma_pid;		/* ...and this process ID */
        char   	          	*ma_start;	/* block starting address */
        int             	ma_size;	/* block size */
#define ma_iregion		ma_size		/* region index of origin */
        struct memblock        	*ma_next;	/* next descriptor */
}; 
