/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB/RBD
 *
 *	$Log:	net.h,v $
 * Revision 2.2  92/04/30  14:46:06  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/27  21:21:16  gdburns
 * Add more flags, of course.
 * 
 * Revision 2.1.1.1  92/01/19  20:46:27  gdburns
 * Move dl_event() macro from events.h to here.
 * 
 * Revision 2.1  91/03/20  11:35:02  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.5  91/02/05  17:10:05  raja
 * Added NT_BOOT for parallel booting.  Fixed NT_ERROR definition.
 * 
 * Revision 2.0.1.4  90/11/20  16:59:11  dixit-v
 * Added NT_JONES for routing tables.
 * 
 * Revision 2.0.1.3  90/07/13  19:28:09  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.2  90/06/20  12:22:23  gdburns
 * Changed NOTTYPE to NOTNODETYPE.
 * 
 * Revision 2.0.1.1  90/04/26  21:32:33  gdburns
 * Added NT_CAST to getnotb.
 * Added NOTTYPE, NT_ERROR and tisbad() support for gettype().
 * Cleaned up Raja's RCS mess :-).
 * 
 * Revision 2.0  90/02/20  11:54:56  trillium
 * Homo erectus
 *
 *	Function:	- constants and structures associated with the
 *			  network subsystem
 *
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#ifndef _NET
#define _NET

#include "kmsg.h"

/*
 * network constants
 */
#define	MAXNMSGLEN	4096		/* maximum network frame size */
#define	NOTNODEID	0x80000000	/* not a valid node ID */
#define NOTLINKID       0x80000000      /* not a valid logical link ID */
#define	NOTNODETYPE	0x80000000	/* not a valid node type */
#define	LOCAL		0x80000001	/* on the local node */
#define	ORIGIN		(getorigin())	/* should be user's node */

/*
 * node types and masks
 */
#define NT_OTB		0x00		/* OTB, not-ITB */
#define NT_ITB		0x01		/* ITB, not-OTB */
#define NT_CAST		0x02		/* a multicast of nodes */
#define NT_WASTE	0x04		/* not fully used/wasted */
#define NT_DISK		0x08		/* node has a disk */
#define NT_TUBE		0x10		/* node has video unit */
#define NT_ALL		0x1F		/* match all bit-flags */
#define NT_JONES	0x20		/* node is my neighbour */
#define NT_BOOT		0x40		/* node to boot */
#define NT_ERROR	NOTNODETYPE	/* invalid node type */

/*
 * node type inquiry macros
 */
#define tisitb(x)	((x) & NT_ITB)		/* node is itb */
#define tisotb(x)	(!tisitb(x))		/* node is otb */
#define tiscast(x)	((x) & NT_CAST)		/* is a cast */
#define tiswaste(x)	((x) & NT_WASTE)	/* node is wasted */
#define tisdisk(x)	((x) & NT_DISK)		/* node has disk */
#define tistube(x)	((x) & NT_TUBE)		/* node has video */
#define tisbad(x)	((x) & NT_ERROR)	/* node type is invalid */

#define getnitb()	getntype(NT_ITB, NT_ALL)	/* pure itbs */
#define getnotb()	getntype(NT_OTB, NT_ITB|NT_CAST)/* all otbs */
#define getnall()	getntype(0, NT_CAST)		/* all nodes */
#define getncast()	getntype(NT_CAST, NT_CAST)	/* all casts */

/*
 * node casts (implies an alias and a source)
 */
#define HOST2ALL	0x80000002	/* Host to all nodes */
#define HOST2ITB	0x80000003	/* Host to all ITB nodes */
#define HOST2OTB	0x80000005	/* Host to all OTB nodes */

/*
 * network information
 */
struct net_info {
	int		ni_nodeid;	/* my node ID */
	int		ni_nodetype;	/* type of node */
	int		ni_origin;	/* where the user sits */
};

/*
 * network calls
 */
#define DSEND		KSEND
#define DTRY_SEND	KTRY_SEND
#define DRECV		KRECV
#define DTRY_RECV	KTRY_RECV
#define NSEND		KSEND
#define NTRY_SEND	KTRY_SEND
#define NRECV		KRECV
#define NTRY_RECV	KTRY_RECV

#define dsend(x)	(do_dsend(x, DSEND))
#define dtry_send(x)	(do_dsend(x, DTRY_SEND))
#define drecv(x)	(do_drecv(x, DRECV))
#define dtry_recv(x)	(do_drecv(x, DTRY_RECV))
#define nsend(x)	(do_nsend(x, NSEND))
#define ntry_send(x)	(do_nsend(x, NTRY_SEND))
#define nrecv(x)	(do_nrecv(x, NRECV))
#define ntry_recv(x)	(do_nrecv(x, NTRY_RECV))

/*
 * network message header
 */
#define NHDSIZE		8		/* # of data words in header */

struct nmsg {
	int	 nh_dl_event;		/* datalink output event from router */
	int	 nh_node;		/* destination process's node */
	int	 nh_event;		/* destination process's event */
	int	 nh_type;		/* destination process's type */
	int	 nh_length;		/* how much data to transmit */
	int	 nh_flags;		/* network flags */
#define	DHOLD		0x02		/* not used in 2.0 */
#define	NHOLD		0x04		/* use KHOLD in multi-packet msg */
#define HOLD	(KHOLD|DHOLD|NHOLD)	/* virtual circuit at all layers */
#define NOBUF		0x10		/* don't do buffering at all */
#define DNODATA		0x80		/* don't ship header data */
#define DFORWARD	0x100		/* datalink store and forward */
#define NREEL		0x400		/* multireel */
#define BYBUF		0x800		/* skip buffer on this node */
#define NMORE		0x1000		/* more packets coming */
#define DRAWDATA	0x10000		/* nh_data contains raw data */
#define DINTDATA	0x20000		/* nh_data contains integer data */
#define DFLTDATA	0x40000		/* nh_data contains float data */
#define DDBLDATA	0x80000		/* nh_data contains double data */
#define DMSKDATA	0xf0000		/* mask nh_data flags */
#define DRAWMSG		0x100000	/* nh_msg contains raw data */
#define DINTMSG		0x200000	/* nh_msg contains integer data */
#define DFLTMSG		0x400000	/* nh_msg contains float data */
#define DDBLMSG		0x800000	/* nh_msg contains double data */
#define DMSKMSG		0xf00000	/* mask nh_msg flags */
#define DDLI		0x1000000	/* datalink input */
#define NPKT		0x2000000	/* multipacket message */
	int	 nh_data[NHDSIZE];	/* builtin data field */
	char	*nh_msg;		/* points to contiguous data */
};

/*
 * network frame
 */
struct netframe {
	struct nmsg	nf_header;
	char		nf_data[MAXNMSGLEN];
};

/*
 * function declarations
 */
extern int		do_drecv();
extern int		do_dsend();
extern int		do_nrecv();
extern int		do_nsend();
extern int		do_precv();
extern int		do_psend();
extern int		getnodeid();
extern int		getorigin();

/*
 * backward compatibility
 */
#define ator(x) (x)
#define HOST ORIGIN

/*
 * Convert a link number to a datalink event.
 */
#define dl_event(x)     (0x80000000 | (((x) & 3) << 2))

/*
 * Convert a datalink event to a link number.
 */
#define dl_link(x)	(((x) & 0xF) >> 2)

#endif
