/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB
 *
 *	$Log:	pcb.h,v $
 * Revision 2.2  92/04/30  14:46:08  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:35:03  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:28:25  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:55:11  trillium
 * Homo erectus
 * 
 *
 *	Function:	- define templates and constants for the
 *			  Process Control Block
 */


#ifndef _PCB

#define _PCB

#include "t.out.h"

/*
 * This is the Process Control Block, containing information on a
 * compute node process available at runtime.
 */
struct pcb_t {
	int		pc_pid;		/* process identifier */
	struct exec	*pc_header;	/* load module header */
	int		pc_nodeid;	/* optional method of setting node ID */
	int		pc_argc;	/* # of runtime arguments */
	char		**pc_argv;	/* runtime arguments */
	int		pc_rtf;		/* initial runtime flags */
	char		*pc_stack;	/* stack or workspace */
	int		pc_stksize;	/* stack size */
	int		pc_stdin;	/* standard input descriptor */
	int		pc_stdout;	/* standard output descriptor */
	int		pc_stderr;	/* standard error descriptor */
};

/*
 * This is the extended workspace start-off template, used only
 * to initialize new processes.
 */
struct crt0_t {
	char		*t0_ip;		/* instruction pointer */
	char		*t0_ret;	/* return address */
	char		*t0_slink;	/* data segment */
	struct pcb_t	*t0_ppcb;	/* ptr to PCB */
	struct pcb_t	t0_pcb;		/* process control block */
};

/*
 * This structure contains the information necessary to create a process
 * from a load module excepting the runtime arguments.
 * It is redundant with crt0_t, but exists so as not to confuse the user
 * with extraneous fields.
 */

struct kproc_t {
	struct exec	*kp_header;	/* load module header */
	int		kp_rtf;		/* initial runtime flags */
	int		kp_stksize;	/* stack size */
	char		*kp_slink;	/* data segment */
	char		*kp_stack;	/* stack ptr (value returned) */
};

#endif
