/*
 *	OSU Research Computing					Trillium
 *	Copyright 1990 The Ohio State University & Cornell University
 *	GDB,VSD,RBD
 *
 *	$Log:	rreq.h,v $
 * Revision 2.2  92/04/30  14:46:17  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/11/27  14:59:17  raja
 * Support second route table and RNODEWAIT request.
 * 
 * Revision 2.1  91/03/20  11:35:08  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:28:07  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/05/10  16:26:19  raja
 * Modified the rr_event field in the rreply structure from rr_result1
 * to rr_result2 in order to accomodate the new RGETROUTE service which
 * needs rr_nodeid and rr_event to coexist.
 * 
 * Revision 2.0  90/02/20  11:54:55  trillium
 * Homo erectus
 * 
 *	Function:	- constants and templates associated with the router
 */


#ifndef _RREQ
#define _RREQ

#include <portable.h>

/*
 * router requests
 */
#define NREQUESTS	13	/* number of router requests */
#define RNORMAL		1	/* normalize request value */
#define RGETROUTE	1	/* request for route information */
#define RGETNET		2	/* get network information */
#define RSETNET		3	/* set network information */
#define RATTACH		4	/* attach and register stand alone process */
#define RDETACH		5	/* detach and forget stand alone process */
#define RLOADROUTES	6	/* load a brand new route table */
#define RSETROUTE	7	/* change or add one specific route */
#define RHOLD		8	/* forces sa_rtr to wait for next message
				   from the sender of RHOLD message */
#define RNODES		9	/* get list of nodes */
#define RREMROUTE	10	/* remove one specific route */
#define RGETSMART	11	/* wait for a smart local router */
#define RGETNTYPE	12	/* get number of nodes of certain type */
#define RNODEWAIT	13	/* wait till nodeid is defined */

/*
 * router request types
 */
#define RDUMB		1			/* dumb router without table */
#define RSMART		2			/* smart router with table */

/*
 * router request template
 */
struct rreq {
	int4		rq_src_node;		/* client's node */
	int4		rq_src_event;		/* client's event */
	int4		rq_request;		/* specific router request */
	int4		rq_parm1;		/* first parameter */
#define rq_nodeid	rq_parm1		/* node ID */
#define rq_length	rq_parm1		/* length of route table */
	int4		rq_parm2;		/* second parameter */
#define rq_event	rq_parm2		/* forwarding event */
#define rq_pid		rq_parm2		/* process ID */
	int4		rq_parm3;		/* fourth parameter */
#define rq_nodetype	rq_parm3		/* node type */
	int4		rq_parm4;		/* fifth parameter */
#define rq_origin	rq_parm4		/* origin node ID */
#define rq_typemask	rq_parm4		/* type mask */
#define rq_event2	rq_parm4		/* min-route event */
};

/*
 * router reply template
 */
struct rreply {
	int4		rr_reply;		/* return code */
	int4		rr_result1;		/* first result */
#define rr_nodeid	rr_result1		/* node ID */
	int4		rr_result2;		/* second result */
#define rr_event	rr_result2		/* forwarding event */
#define rr_ntype	rr_result2		/* # of nodes of given type */
	int4		rr_result3;		/* fourth result */
#define rr_nodetype	rr_result3		/* node type */
	int4		rr_result4;		/* fifth result */
#define rr_origin	rr_result4		/* orgin node ID */
#define rr_event2	rr_result4		/* min-route event */
};

/*
 * route table entry
 */
struct route {
	int4		r_nodeid;		/* destination node */
	int4		r_event;		/* forward to this event */
	int4		r_event2;		/* min-route event */
	int4		r_nodetype;		/* ITB or OTB */
};

/*
 * waiting list entry
 */
struct rwait {
	int4		rw_nodeid;		/* nodeid awaited for */
	int4		rw_src_node;		/* client's nodeid */
	int4		rw_src_event;		/* client's event */
};

#endif
