/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	kexit.c,v $
 * Revision 2.2  92/04/30  14:47:34  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/02/23  22:43:23  gdburns
 * _kcleanup() now returns void.
 * 
 * Revision 2.1  91/03/20  11:36:59  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:27:43  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/14  16:41:28  gdburns
 * To terminate process, call _kbail() in system dependent clientio module.
 * 
 * Revision 2.0  90/02/20  11:54:41  trillium
 * Homo erectus
 * 
 *	Function:	- exit package
 *			- performs all tasks required when a process exits
 */

#include <kio.h>
#include <terror.h>
#include <typical.h>
#include <net.h>

/*
 * external variables
 */
extern struct kio_t	_kio;			/* Kernel Input/Output block */
extern int		errno;			/* global error variable */

/*
 * external functions
 */
extern int		kdetach();
extern int		rpdetach();
extern void		_kcleanup();	 	/* destroy mesasge queues */
extern void		_kbail();		/* terminate self */

/*
 * global functions
 */
void			_kexit();
void			kexit();
int			atkexit();

/*
 *	atkexit
 *
 *	Function:	- add function to cleanup table
 *	Accepts:	- pointer to exit function
 *			- argument to exit function
 *			- does not store a combination of exit function
 *			  and argument more than once
 *	Returns:	- 0 or ERROR
 */

int
atkexit(exitfunc, arg)

int			(*exitfunc)();
char			*arg;

{
	int		i;		/* favourite index variable */

	for (i = 0; (i < TNEXIT) && (_kio.ki_exit[i]) &&
			!((_kio.ki_exit[i] == exitfunc) &&
			(_kio.ki_arg[i] == arg)); ++i);

	if (i < TNEXIT) {

		if (!_kio.ki_exit[i]) {
			_kio.ki_exit[i] = exitfunc;
			_kio.ki_arg[i] = arg;
		}

		return(0);
	}

	else {
		errno = ENOEXDESCRIPTORS;
		return(ERROR);
	}
}

/*
 *	kexit
 *
 *	Function:	- exit from Trollius
 *			- calls cleanup routines
 *			- does not return
 *	Accepts:	- exit status
 */

void
kexit(status)

int			status;

{
	int		i;		/* favourite index variable */

	for (i = TNEXIT - 1; i >= 0; --i) {

		if (_kio.ki_exit[i]) {

			(*(_kio.ki_exit[i]))(_kio.ki_arg[i], status);
		}
	}

	_kexit(status);
}

/*
 *	_kexit
 *
 *	Function:	- exit from Trollius
 *			- does not call cleanup routines
 *			- does not return
 *	Accepts:	- exit status
 */

void
_kexit(status)

int			status;

{
	if (_kio.ki_rtf & RTF_KERNEL) {

		if (_kio.ki_rtf & RTF_KENYA) {
			(void) rpdetach(LOCAL, 0, status);
		}

		(void) kdetach(0);
	}
/*
 * Destroy underlying communication structures and free related resources.
 */
	_kcleanup();
/*
 * _kbail does not return.
 */
	_kbail(status);
}
