/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	kinit.c,v $
 * Revision 2.2  92/04/30  14:47:35  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/02/23  22:42:00  gdburns
 * _init_kio() now returns void.
 * 
 * Revision 2.1.1.2  91/10/27  20:58:50  gdburns
 * Use fcntl.h for open() flags.
 * 
 * Revision 2.1.1.1  91/10/03  16:35:49  gdburns
 * Add runtime support for trace collection and transport.
 * 
 * Revision 2.1  91/03/20  11:37:00  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  91/03/12  16:20:37  gdburns
 * compatible with new file daemon
 * 
 * Revision 2.0.1.1  90/07/13  19:27:38  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:38  trillium
 * Homo erectus
 * 
 *	Function:	- initialize data structures for kernel interface
 *			  by calling `init_kio'
 *			- calls `kattach'
 *			- defines global space for kernel I/O block
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 *
 *	Accepts:	- a priority to be used with kattach
 *
 *	Returns:	- 0 or ERROR, kernel replied errors will
 *			  show up in errno
 */

#include <typical.h>
#include <kernel.h>
#include <kio.h>
#include <net.h>
#include <freq.h>
#include <fcntl.h>

/*
 * external data structures
 */
extern struct kio_t		_kio;		/* Kernel I/O block */
extern struct fclient		_ufd[FUMAX];	/* user file desc. */

/*
 * external functions
 */
extern int			_createport();	/* create comm. port */
extern int			atkexit();	/* register cleanup */
extern int			kattach();	/* attach to the kernel */
extern int			rfatexit();	/* filed cleanup */
extern void			_init_kio();	/* initialize kernel I/O */
extern void			_siginit();	/* initialize signals */
extern void			ktrinit();	/* initialize tracing */

int
kinit(priority)

int				priority;	/* if necessary, attach at this
						   priority */
{
	int			i;		/* favourite index variable */

	if (!(_kio.ki_rtf & RTF_KERNEL)) {
/*
 * Initialize virtual circuit table.
 */
		for (i = 0; i < MAXKHOLD; ++i) {
			_kio.ki_khold[i].kh_event = NOTEVENT;
			_kio.ki_khold[i].kh_aux = 0;
		}
/*
 * Initialize signal handling module.
 */
		_siginit();
/*
 * Initialize the exit table.
 */
		for (i = 0; i < TNEXIT; ++i) {
			_kio.ki_exit[i] = (int (*)()) 0;
		}
/*
 * Initialize route cache.
 */
		for (i = 0; i < MAXRCACHE; ++i) {
			_kio.ki_route[i].r_nodeid = NOTNODEID;
		}

		_kio.ki_index = -1;
/*
 * Initialize kernel request interface.
 */
		_init_kio();
/*
 * Initialize user file descriptor table
 */
		for (i = 0; i < FUMAX; ++i) {
			_ufd[i].fu_tflags = 0;
		}

		_ufd[0].fu_tflags = O_NOCBX | O_RDONLY;
		_ufd[0].fu_node = NOTNODEID;
		_ufd[0].fu_tfd = _kio.ki_stdin;
		_ufd[1].fu_tflags = O_NOCBX | O_WRONLY;
		_ufd[1].fu_node = NOTNODEID;
		_ufd[1].fu_tfd = _kio.ki_stdout;
		_ufd[2].fu_tflags = O_NOCBX | O_WRONLY;
		_ufd[2].fu_node = NOTNODEID;
		_ufd[2].fu_tfd = _kio.ki_stderr;

		(void) atkexit(rfatexit, 0);
/*
 * Attach process to the kernel.
 */
		if (kattach(priority) < 0) {
			return(ERROR);
		}
/*
 * Initialize the trace subsystem.
 */
		if (_kio.ki_rtf & RTF_TRACE) {
			ktrinit();
		}

		return(0);
	}

	else {
/*
 * Attach process to the kernel.
 */
		return(kattach(priority));
	}
}
