/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB
 *
 *	$Log:	krecv.c,v $
 * Revision 2.2  92/04/30  14:47:36  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/10/09  14:09:34  gdburns
 * Use upgraded portable.h.
 * 
 * Revision 2.1  91/03/20  11:37:03  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.4  91/03/12  16:21:22  gdburns
 * extra checking for ENOTATTACHED
 * 
 * Revision 2.0.1.3  91/01/14  15:32:15  gdburns
 * Made routine reentrant to signals by localizing pick_*.
 * 
 * Revision 2.0.1.2  90/09/27  18:02:16  gdburns
 * Converted to modern portable.h.
 * 
 * Revision 2.0.1.1  90/07/13  19:27:39  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:38  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:11:51  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:46:51  trillium
 * Initial revision
 * 
 *
 *	Function:	- receive a message from another process
 *
 *	Accepts:	- a pointer to a kernel message header
 *
 *	Returns:	- an error indicator
 */
			  

#include <events.h>
#include <kio.h>
#include <kmsg.h>
#include <kreq.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>


/*
 * external variables
 */
extern struct kio_t	_kio;			/* kernel I/O block */
extern int		errno;

/*
 * external functions
 */
extern int		_make_kreq();		/* make request to kernel */
extern int		_createport();		/* create comm. port */
extern void		_follow_signal();	/* check signals */


int
do_krecv(msgptr, try)

struct kmsg		*msgptr;		/* pointer to message header */
int			try;			/* krecv request type */

{
	struct kreq	req;			/* kernel request */
	struct kreply	reply;			/* kernel reply */
	int		fhold;			/* KHOLD bit */
	struct khold_t	*phold;			/* virtual circuit descriptor */
	int		ev;			/* caller's event */
	int		ret;			/* return code */
	int		pick_aux;		/* my aux port */
	int		pick_auxd;		/* my aux port desc. */

/*
 * It is at this point that we must decide if we are using
 * a virtual circuit or we have to make a kernel request.  Either way,
 * the goal is to come up with values for _kio.ki_peer_port and
 * _kio.ki_peer_aux, which are the data exchange ports for local
 * message passing.
 */
/*
 * Calculate the pointer to the correct virtual circuit descriptor.
 */
	ev = msgptr->k_event;
	phold = _kio.ki_khold + ((ev & HOLDBITS) +
		((ev & INT4_SIGN) ? (HOLDBITS + 1) : 0));
/*
 * Is a circuit already defined for the given event?
 */
	if (phold->kh_event == ev) {
		_kio.ki_peer_port = phold->kh_peer_port;
		_kio.ki_peer_aux = phold->kh_peer_aux;
		pick_auxd = phold->kh_auxd;
		pick_aux = phold->kh_aux;
/*
 * If the caller has defined KREVPORT, we need to save this fact in
 * the virtual circuit descriptor.  If the caller has not defined
 * KREVPORT, then we need to define it in k_flags if it has been defined 
 * in the virtual ciruit descriptor.
 */
		phold->kh_flags |= msgptr->k_flags & KREVPORT;
		msgptr->k_flags |= phold->kh_flags & KREVPORT;
/*
 * Do we need to break the circuit?
 */
		if (!(msgptr->k_flags & KHOLD)) {
			phold->kh_event = NOTEVENT;
			fhold = FALSE;
		}

		else {
			fhold = TRUE;
		}
	}

	else {
/*
 * The circuit is not now defined, but WILL be defined after a kernel
 * request IF the KHOLD bit is set.  In this case, we want to be sure
 * we don't have a collision in the circuit table BEFORE we make a
 * kernel request.
 */
		if (msgptr->k_flags & KHOLD) {
			fhold = TRUE;

			if (phold->kh_event != NOTEVENT) {

				errno = (_kio.ki_rtf & RTF_KERNEL) ?
					ECIRCUITINUSE : ENOTATTACHED;
				return(ERROR);
			}

			if (!phold->kh_aux) {
				phold->kh_aux = _createport(&phold->kh_auxd);

				if (phold->kh_aux == ERROR) {
					return(ERROR);
				}
			}
/*
 * Save the KREVPORT flag.
 */
			phold->kh_flags = msgptr->k_flags & KREVPORT;

			req.kqm_aux = phold->kh_aux;
			pick_aux = phold->kh_aux;
			pick_auxd = phold->kh_auxd;
		}

		else {
			fhold = FALSE;
			req.kqm_aux = _kio.ki_aux;
			pick_auxd = _kio.ki_auxd;
			pick_aux = _kio.ki_aux;
		}
/*
 * Formulate the KRECV request.
 */
		req.kq_request = try;
		req.kq_index = _kio.ki_index;
		req.kqm_event = ev;
		req.kqm_type = msgptr->k_type;
		req.kqm_length = msgptr->k_length;
/*
 * Send the request to the kernel.
 */
		if (!_make_kreq(&req, &reply)) {
			msgptr->k_type = reply.krm_type;
			msgptr->k_length = reply.krm_length;
			_kio.ki_peer_port = reply.krm_port;
			_kio.ki_peer_aux = reply.krm_aux;

			if (fhold) {
				phold->kh_event = ev;
				phold->kh_peer_port = reply.krm_port;
				phold->kh_peer_aux = reply.krm_aux;
			}
		}

		else {
			return(ERROR);
		}
	}   /* else */

/*
 * Receive the data from the receiving process.
 */
	ret = _recv_kmsg(msgptr, pick_aux, pick_auxd);
/*
 * Check signals.
 * We don't want to check signals if we are
 * passing messages on a virtual circuit.
 */
	if (_kio.ki_signal && !fhold) {
		_follow_signal();
	}

	return(ret);
}
