/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB
 *
 *	$Log:	kstate.c,v $
 * Revision 2.2  92/04/30  14:47:41  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:37:07  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:27:42  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:40  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:11:59  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:46:53  trillium
 * Initial revision
 * 
 *
 *	Function:	- find state of a process
 *
 *	Accepts:	- a Trillium process identifier (0 means yourself)
 *			- a pointer to a state structure where information
 *			  from the kernel will be stored
 *
 *	Returns:	- an error indicator
 */
			  

#include <kreq.h>
#include <kio.h>
#include <typical.h>


/*
 * external variables
 */
extern struct kio_t	_kio;			/* kernel I/O block */

/*
 * external functions
 */
extern int		_make_kreq();		/* make request of kernel */
extern void		_follow_signal();	/* check signals */


int
kstate(pid, pstate)

int			pid;			/* Trillium process ID */
struct kstate		*pstate;		/* pointer to state info */

{
	int		ret_code;		/* return code */
	struct kreq	req;			/* kernel request */
	struct kreply	reply;			/* kernel reply */

/*
 * Formulate the KSTATE kernel request.
 */
	req.kq_request = KSTATE;
	req.kq_index = _kio.ki_index;
	req.kqs_pid = pid;

	if (!(ret_code = _make_kreq(&req, &reply))) {
		pstate->ks_state = reply.krs_state;
		pstate->ks_priority = reply.krs_priority; 
		pstate->ks_event = reply.krs_event;
		pstate->ks_type = reply.krs_type;
		pstate->ks_signal = reply.krs_signal;
	}
/*
 * Check signals.
 */
	if (_kio.ki_signal) {
		_follow_signal();
	}

	return(ret_code);
}
