/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	drecv.c,v $
 * Revision 2.2  92/04/30  14:48:28  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/02/28  16:11:20  gdburns
 * Replace move() with memcpy().
 * 
 * Revision 2.1.1.1  92/02/23  22:45:32  gdburns
 * Change tools/c.h to c.h.
 * 
 * Revision 2.1  91/03/20  11:37:21  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:27:30  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/03/06  14:10:29  gdburns
 * Placed #ifdef around tools/c.h for compatibility
 * with stand-alone system.
 * 
 * Revision 2.0  90/02/20  11:54:33  trillium
 * Homo erectus
 *
 *	Function:	- receives a message at the datalink sublayer
 *			- contains most of the meat of the network layer
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#include <typical.h>
#include <kmsg.h>
#include <kio.h>
#include <net.h>
#include <terror.h>

/*
 * external variables
 */
extern struct kio_t	_kio;		/* Kernel Input/Output */
extern int		errno;

int
do_drecv(nhead, dr_req)

struct nmsg		*nhead;
int			dr_req;

{
	struct kmsg	khead;		/* kernel message descriptor */
	struct nmsg	incoming;	/* received network message header */


	if (((unsigned) nhead->nh_length) > MAXNMSGLEN) {
		errno = EMSGLEN;
		return(ERROR);
	}
/*
 * Set up kernel message descriptor to receive network frame header.
 */     
	khead.k_event = nhead->nh_event;
	khead.k_type = nhead->nh_type;
	khead.k_length = sizeof(struct nmsg);
	khead.k_flags = KHOLD;
	khead.k_msg = (char *) &incoming;
/*
 * Receive network frame header.
 */
	if (do_krecv(&khead, dr_req)) {
		return(ERROR);
	}
/*
 * Continue virtual circuit if specified by network header.
 * Otherwise notify krecv() that a datalink virtual circuit is now broken.
 * Set up kernel message descriptor to receive data portion of network frame.
 * Pass along KREVPORT flag to krecv().
 */
	khead.k_flags = ((nhead->nh_flags & KHOLD) ? KHOLD : KBREAK) |
			(incoming.nh_flags & KREVPORT);
	khead.k_msg = nhead->nh_msg;
/*
 * Determine minimum length.  Warn krecv about sender length overrun.
 */
	if (nhead->nh_length < incoming.nh_length) {
		khead.k_flags |= KCLOBBER;
		_kio.ki_slush = incoming.nh_length - nhead->nh_length;
		khead.k_length = nhead->nh_length;
	}

	else {
		khead.k_length = incoming.nh_length;
	}

	if (nhead->nh_flags & DFORWARD) {
/*
 * We intend to store and forward this frame.  Copy everything except
 * the nh_msg field which is designed to appear last in the network
 * frame header.
 */
		memcpy((char *) nhead, (char *) &incoming,
			(sizeof(struct nmsg) - sizeof(int)));
/*
 * It's dangerous to leave this flag on.
 */
		nhead->nh_flags &= ~KREVPORT;
	}

	else {
/*
 * We want to reuse the network frame header and don't want important
 * fields trashed by dsend().  Just copy the data and the nh_type field.
 * Set the nh_length field to the minimum transferred.
 */
/*
 * If the user has set the NMORE bit, then clear it if the
 * incoming NMORE bit clear.  If the user has NMORE clear, don't touch it.
 */
		nhead->nh_flags &= ((incoming.nh_flags & NMORE) | ~NMORE);
		nhead->nh_type = incoming.nh_type;
		nhead->nh_length = khead.k_length;

		memcpy((char *) nhead->nh_data,
			(char *) incoming.nh_data, NHDSIZE * sizeof(int));
	}
/*
 * Receive network frame data.
 */
	return(krecv(&khead));
}
