/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	dsend.c,v $
 * Revision 2.2  92/04/30  14:48:30  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  12:25:23  gdburns
 * Categorize user and system messages.
 * 
 * Revision 2.1  91/03/20  11:37:22  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:27:31  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:33  trillium
 * Homo erectus
 *
 *	Function:	- sends a message at the datalink sublayer
 *			- contains most of the meat of the network layer
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#include <bufreq.h>
#include <dlreq.h>
#include <events.h>
#include <kmsg.h>
#include <net.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int		errno;

int
do_dsend(nhead, ds_req)

struct nmsg		*nhead;
int			ds_req;

{
	struct kmsg	khead;		/* kernel message descriptor */	
	int		bf_flag;	/* to buffer or not to buffer */
	int		bf_req;		/* blocking or non-blocking */
	int		sv_flags;	/* save nh_flags */

	if (((unsigned) nhead->nh_length) > MAXNMSGLEN) {
		errno = EMSGLEN;
		return(ERROR);
	}
/*
 * Turn off buffering if the caller says so, or if the caller wants KHOLD mode.
 */
	bf_flag = !(nhead->nh_flags & (NOBUF | BYBUF | KHOLD));
	bf_req = bf_flag ? KTRY_SEND : ds_req;
	sv_flags = nhead->nh_flags;
	nhead->nh_flags &= ~(DDLI | BYBUF);
/*
 * Set up kernel message descriptor to send the network frame header.
 */
	khead.k_event = nhead->nh_dl_event;
	khead.k_length = sizeof(struct nmsg);
	khead.k_type = (nhead->nh_dl_event == nhead->nh_event) ? nhead->nh_type
		: (nhead->nh_event >= 0) ? DLUSR : DLSYS;
	khead.k_flags = KHOLD;
	khead.k_msg = (char *) nhead;

	if (do_ksend(&khead, bf_req)) {
/*
 * Try for a buffer.
 */
		if (bf_flag) {
			khead.k_event = EVBUFFERD;
			khead.k_type = (sv_flags & DDLI) ? BFDLI :
				(nhead->nh_event >= 0) ? BFUSR : BFSYS;

			if (do_ksend(&khead, ds_req)) {
				return(ERROR);
			}
		} else {
			return(ERROR);
		}
	}
/*
 * Set up kernel message descriptor to send the network frame data.
 * Set the kernel KHOLD flag equal to the network KHOLD flag.
 */
	khead.k_length = nhead->nh_length; 
	khead.k_flags = (nhead->nh_flags & KHOLD) ? KHOLD : KBREAK;
	khead.k_msg = nhead->nh_msg;	

	(void) ksend(&khead);
	nhead->nh_flags = sv_flags;
	return(0);
}
