/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	ncombine.c,v $
 * Revision 2.2  92/04/30  14:48:31  trillium
 * Ohio Release
 * 
 *	Function:	- user defined local combining of messages
 *			- handles multipacket reeled messages
 *
 *	Accepts:	- cast identifier
 *			- message event
 *			- ptr combining function
 *
 *	Returns:	- 0 or ERROR 
 */

#include <portable.h>
#include <net.h>
#include <typical.h>

#undef	NULL
#define	NULL	((char *) 0)

/*
 * external variables
 */
extern int		errno;			/* global error code */

/*
 * external functions
 */
extern char		*malloc();
extern int4		rcask();

int
ncombine(castid, event, cmbfunc)

int4			castid,
			event;
void			(*cmbfunc)();

{
	struct nmsg	nhead1, nhead2;		/* message headers */
	int4		ncomb;			/* # of combines */
	int		save_errno;		/* saved error code */
/*
 * Get the number of messages to reel.
 */
	if ((ncomb = rcask(LOCAL, castid)) == ERROR) {
		return(ERROR);
	}
	else if (--ncomb == 0) {
		return(0);
	}
/*
 * Allocate two packet buffers.
 */
	if ((nhead1.nh_msg = malloc((uint4) MAXNMSGLEN)) == NULL) {
		return(ERROR);
	}
	if ((nhead2.nh_msg = malloc((uint4) MAXNMSGLEN)) == NULL) {
		save_errno = errno;
		(void) free(nhead1.nh_msg);
		errno = save_errno;
		return(ERROR);
	}
 
	nhead1.nh_event = nhead2.nh_event = event;

	while (ncomb > 0) {
		nhead1.nh_type = nhead2.nh_type = 0;
		nhead1.nh_flags = nhead2.nh_flags = NMORE;
		nhead1.nh_length = nhead2.nh_length = MAXNMSGLEN;
/*
 * Get the two packets to combine.
 */
		if (drecv(&nhead1)) {
			save_errno = errno;
			(void) free(nhead1.nh_msg);
			(void) free(nhead2.nh_msg);
			errno = save_errno;
			return(ERROR);
		}

		if (drecv(&nhead2)) {
			save_errno = errno;
			(void) free(nhead1.nh_msg);
			(void) free(nhead2.nh_msg);
			errno = save_errno;
			return(ERROR);
		}
/*
 * Decrement the combining count if any packet has NMORE turned off.
 */
		if ((nhead1.nh_flags & nhead2.nh_flags & NMORE) == FALSE) {
			ncomb--;
			nhead1.nh_flags &= ~NMORE;
		}
/*
 * Invoke the user's combining function.
 */
		nhead1.nh_length = min(nhead1.nh_length, nhead2.nh_length);

		cmbfunc(nhead1.nh_length, nhead1.nh_msg, nhead2.nh_msg,
			nhead1.nh_data, nhead2.nh_data);
/*
 * Return the combined message to castd.
 */
		nhead1.nh_dl_event = nhead1.nh_event;
		nhead1.nh_node = LOCAL;
		nhead1.nh_flags |= NOBUF;
		if (dsend(&nhead1)) {
			save_errno = errno;
			(void) free(nhead1.nh_msg);
			(void) free(nhead2.nh_msg);
			errno = save_errno;
			return(ERROR);
		}
	}

	(void) free(nhead1.nh_msg);
	(void) free(nhead2.nh_msg);
	return(0);
}
