/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	nrecv.c,v $
 * Revision 2.2  92/04/30  14:48:49  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/01/10  17:40:31  gdburns
 * Add TID, the process's trace identifier.
 * 
 * Revision 2.1.1.1  91/10/03  17:14:54  gdburns
 * Add runtime support for tracing.
 * 
 * Revision 2.1  91/03/20  11:37:23  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:27:23  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:54:31  trillium
 * Homo erectus
 * 
 *	Function:	- receive a network message from an event
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 */

#include <kio.h>
#include <ktrace.h>
#include <net.h>
#include <typical.h>

/*
 * external variables
 */
extern struct kio_t	_kio;

int
do_nrecv(nhead, nr_req)

struct nmsg		*nhead;		/* network message header */
int			nr_req;		/* blocking/non-blocking */

{
	int		remlen;		/* remaining length */
	int		totlen;		/* total length */
	int		stamp;		/* trace time stamp */
	char		*save_msg;	/* saved msg */
	struct trrecv	*ptrace;	/* recv trace ptr */
/*
 * Save time stamp, if tracing.
 */
	if (((_kio.ki_rtf & RTF_TRON) == RTF_TRON) &&
		(nhead->nh_event >= 0)) {
		stamp = ldtimer();
	}

/*
 * Receive the network message using dsend.
 */
	remlen = nhead->nh_length;
	totlen = 0;
	save_msg = nhead->nh_msg;
	nhead->nh_flags |= NMORE;
	nhead->nh_length = MAXNMSGLEN;

	while ((remlen > MAXNMSGLEN) && (nhead->nh_flags & NMORE)) {

		if (do_drecv(nhead, nr_req)) {
			return(ERROR);
		}

		nr_req = NRECV;
		remlen -= nhead->nh_length;
		totlen += nhead->nh_length;
		nhead->nh_msg += nhead->nh_length;
	}

	if (nhead->nh_flags & NMORE) {
/*
 * The next packet should be the last.
 */
		nhead->nh_length = remlen;

		if (do_drecv(nhead, nr_req)) {
			return(ERROR);
		}

		totlen += nhead->nh_length;
/*
 * Suck excess packets into the vacuum of space.
 */
		nhead->nh_length = 0;

		while (nhead->nh_flags & NMORE) {

			if (do_drecv(nhead, nr_req)) {
				return(ERROR);
			}
		}
	}
/*
 *	else if (remlen > MAXNMSGLEN)
 *
 * We just ran out of packets.
 */
	nhead->nh_flags &= ~NMORE;
	nhead->nh_length = totlen;
	nhead->nh_msg = save_msg;

/*
 * Store recv trace, if tracing is enabled.
 */
	if (((_kio.ki_rtf & RTF_TRON) == RTF_TRON) &&
		(nhead->nh_event >= 0)) {

		ptrace = (struct trrecv *) ktrpoint(sizeof(struct trrecv));
		ptrace->tr6_head.trt_type = TRRECV;
		ptrace->tr6_head.trt_stamp = stamp;
		ptrace->tr6_head.trt_tid = _kio.ki_tid;
		ptrace->tr6_frtid = nhead->nh_data[7];
		ptrace->tr6_event = nhead->nh_event;
		ptrace->tr6_length = nhead->nh_length;
		ptrace->tr6_total = ldtimer() - stamp;
		ptrace->tr6_blocked = 0;
		ptrace->tr6_transmit = 0;
	}

	return(0);
}
