/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB/RBD
 *
 *	$Log:	nsend.c,v $
 * Revision 2.2  92/04/30  14:48:50  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/04/27  17:03:42  raja
 * Handle multireeling packet numbering.
 * 
 * Revision 2.1.1.2  92/01/10  17:40:47  gdburns
 * Add TID, the process's trace identifier.
 * 
 * Revision 2.1.1.1  91/10/03  17:15:10  gdburns
 * Add runtime support for tracing.
 * 
 * Revision 2.1  91/03/20  11:37:25  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  90/07/13  19:27:33  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.2  90/05/13  22:17:26  raja
 * Nsend does not check if the destination nodeid is the sending nodeid.
 * It always calls getroute() which tests for this condition.
 * 
 * Revision 2.0.1.1  90/05/10  20:18:27  raja
 * Removed the caching code, since getrentc does the caching now.
 * 
 * Revision 2.0  90/02/20  11:54:34  trillium
 * Homo erectus
 *
 *	Function:	- sends a message to any event on any node
 *			- caching is done in getroute
 *			- packetizes long messages
 *			- can collect trace data
 *			- for NREEL nh_data[6] is the packet number
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 */

#include <events.h>
#include <kio.h>
#include <ktrace.h>
#include <net.h>
#include <typical.h>

/*
 * external functions
 */
extern int		getroute();

/*
 * external variables
 */
extern struct kio_t	_kio;

int
do_nsend(nhead, ns_req)

struct nmsg		*nhead;		/* network message header */
int			ns_req;		/* blocking/non-blocking */

{
	int		f_nreel;	/* NREEL status flag */
	int		length;		/* remaining length */
	int		stamp;		/* time stamp for tracing */
	int		save_length;	/* saved length */
	char		*save_msg;	/* saved message ptr */
	struct trsend	*ptrace;	/* send trace ptr */
/*
 * Record trace information, if tracing is turned on.
 * This means that word 7 of the data pouch is reserved for us.
 */
	if (((_kio.ki_rtf & RTF_TRON) == RTF_TRON) && (nhead->nh_event >= 0)) {
		stamp = ldtimer();
		nhead->nh_data[7] = _kio.ki_tid;
	}
/*
 * Find the correct value for nh_dl_event.  Get the routing event.
 */
	if ((nhead->nh_dl_event = getroute(nhead->nh_node,
			nhead->nh_event)) == NOTEVENT) {
		return(ERROR);
	}
/*
 * Forward the network message using dsend.
 */
	length = nhead->nh_length;
/*
 * If this send is for a multireel, nh_data[6] contains the packet number.
 * Do this only if packetization (NPKT) was not done previously.
 */
	f_nreel = FALSE;
	if (((nhead->nh_flags & NREEL) == NREEL) &&
			((nhead->nh_flags & NPKT) == 0)) {
		f_nreel = TRUE;
		nhead->nh_flags |= NPKT;
		nhead->nh_data[6] = 0;
	}
			
	if (nhead->nh_length > MAXNMSGLEN) {
		save_msg = nhead->nh_msg;
		save_length = nhead->nh_length;
		nhead->nh_length = MAXNMSGLEN;
		nhead->nh_flags |= NMORE;
/*
* Send first packet.
*/
		if (do_dsend(nhead, ns_req)) {
			return(ERROR);
		}

		ns_req = NSEND;
		length -= MAXNMSGLEN;
		nhead->nh_msg += MAXNMSGLEN;

		if (f_nreel) {
			nhead->nh_data[6]++;
		}
/*
* Send the middle packets.
*/
		while (length > MAXNMSGLEN) {		

			if (do_dsend(nhead, ns_req)) {
				return(ERROR);
			}

			length -= MAXNMSGLEN;
			nhead->nh_msg += MAXNMSGLEN;
			if (f_nreel) {
				nhead->nh_data[6]++;
			}
		}

		nhead->nh_length = length;
		nhead->nh_flags &= ~NMORE;
/*
* Send the last packet.
*/
		if (do_dsend(nhead, ns_req)) {
			return(ERROR);
		}

		nhead->nh_msg = save_msg;
		nhead->nh_length = save_length;
	}

	else {
		if (do_dsend(nhead, ns_req)) {
			return(ERROR);
		}
	}
/*
 * Turn the NPKT flag off.
 */
	if (f_nreel) {
		nhead->nh_flags &= ~NPKT;
	}
/*
 * Store send trace if tracing is turned on.
 */
	if (((_kio.ki_rtf & RTF_TRON) == RTF_TRON) &&
		(nhead->nh_event >= 0)) {

		ptrace = (struct trsend *) ktrpoint(sizeof(struct trsend));
		ptrace->tr5_head.trt_type = TRSEND;
		ptrace->tr5_head.trt_stamp = stamp;
		ptrace->tr5_head.trt_tid = _kio.ki_tid;
		ptrace->tr5_tonode = nhead->nh_node;
		ptrace->tr5_event = nhead->nh_event;
		ptrace->tr5_length = nhead->nh_length;
		ptrace->tr5_total = ldtimer() - stamp;
		ptrace->tr5_blocked = 0;
		ptrace->tr5_transmit = 0;
	}

	return(0);
}
