/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB,RBD,VSD
 *
 *	$Log:	router.c,v $
 * Revision 2.2  92/04/30  14:49:17  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.4  92/04/27  12:26:10  gdburns
 * I am now a controlled process so I don't mess up
 * the buffer system.
 * 
 * Revision 2.1.1.3  92/03/12  13:10:56  gdburns
 * Don't use rmalloc().
 * 
 * Revision 2.1.1.2  92/01/27  17:55:46  gdburns
 * Return ENOTNODE if node not found in route table.
 * 
 * Revision 2.1.1.1  91/11/27  15:36:28  raja
 * Support the second route table and use hashing functions in libtools.
 * Support the new RNODEWAIT request.
 * 
 * Revision 2.1  91/03/20  11:37:40  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.5  91/02/20  15:56:58  gdburns
 * checks for invalid table entries in getnodes()
 * 
 * Revision 2.0.1.4  90/07/13  19:30:20  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.3  90/06/20  12:25:31  gdburns
 * Changed NOTTYPE to NOTNODETYPE.
 * 
 * Revision 2.0.1.2  90/05/13  22:13:17  raja
 * The RGETROUTE request does not accept a destination event anymore.
 * Ir returns the route table entry as is, even for the local nodeid.
 * 
 * Revision 2.0.1.1  90/05/10  16:10:37  raja
 * Modified the RGETROUTE request to return a full route table entry.
 * This service is now called by getrent().
 * 
 * Revision 2.0  90/02/20  11:56:22  trillium
 * Homo erectus
 * 
 *	Function:	- the router daemon: network accessible, topology
 *			  omnipotent, centrally located, totally reconfigurable
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 */

#include <portable.h>
#include <events.h>
#include <net.h>
#include <priority.h>
#include <rreq.h>
#include <shared.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>
#include <all_hash.h>
#include <all_list.h>

#undef NULL
#define NULL (char *) 0

/*
 * external variables
 */
extern int	errno;			/* global error variable */

/*
 * external functions
 */
extern int4	stoi4();

/*
 * local functions
 */
static void		do_getroute();
static void		do_getnet();
static void		do_setnet();
static void		do_noservice();
static void		do_loadroutes();
static void		do_setroute();
static void		do_remroute();
static void		do_nodes();
static void		do_getsmart();
static void		do_getntype();
static void		do_nodewait();
static void		recv_req();
static void		send_reply();
static void		wait_unblock();
static int4		count_ntype();

/*
 * local variables
 */
static struct net_info	net;		/* node and network information */
static struct nmsg	inbound;	/* network message header (requests) */
static struct nmsg	outbound;	/* network message header (replies) */
static struct rreq	*request;	/* router request */
static struct rreply	*reply;		/* router reply */
static int4		reqtype;	/* request type (smart or dumb) */

static void	(*do_req[NREQUESTS])() = {
			do_getroute,
			do_getnet,
			do_setnet,
			do_noservice,	
			do_noservice,
			do_loadroutes,
			do_setroute,
			do_noservice,
			do_nodes,
			do_remroute,
			do_getsmart,
			do_getntype,
			do_nodewait
		};
/*
 * The do_noservice calls are for RATTACH, RDETACH, and RHOLD reply.
 * which are meaningful for stand-alone router.
 */
static HASH		*rtd;		/* route table hash descriptor */
static LIST		*wld;		/* waiting list descriptor */
static int4		dumb_route;	/* route to use before getting a
					   proper route table */
				
main(argc, argv)

int			argc;
char			**argv;

{
	dumb_route = EVDL0;
	reqtype = RDUMB;
	rtd = (HASH *) 0;
	wld = (LIST *) 0;
/*
 * If network information is supplied on the command line, use it.
 */
	if (argc > 1) {
		net.ni_nodeid = stoi4(argv[1]);
		net.ni_nodetype = stoi4(argv[2]);
		net.ni_origin = stoi4(argv[3]);
	} else {
		net.ni_nodeid = NOTNODEID;
		net.ni_nodetype = NOTNODETYPE;
		net.ni_origin = NOTNODEID;
	}
/*
 * Initialize and attach to the local Trillium kernel.
 */
	request = (struct rreq *) inbound.nh_data;
	reply = (struct rreply *) outbound.nh_data;

	if (kinit(0) < 0) {
		panic(errno);
	}

	if (kattach(PRROUTER) < 0) {
		panic(errno);
	}

	if (rpattach(LOCAL) && (errno != ESERVICE)) {
		panic(errno);
	}
/*
 * Continuously receive and process router requests.
 */
	for(;;) {
		recv_req();

		if (request->rq_request > NREQUESTS) {
			reply->rr_reply = EBADREQUEST;
			panic(errno);
		}
		else {
			request->rq_request -= RNORMAL;
			(*(do_req[request->rq_request]))();
		}
	}
}

/*
 *	do_getroute
 *
 *	Function:	- perform RGETROUTE request
 */
static void
do_getroute()

{
	int4		nodeid;			/* client's node ID */
	struct route	*p;			/* favorite pointer */

	reply->rr_reply = 0;
	nodeid = request->rq_nodeid;
/*
 * Get the node entry in the route table.
 */
	if (rtd) {
		p = (struct route *) ah_find(rtd, nodeid);

		if (p == (struct route *) 0) {
			reply->rr_reply = EBADNODE;
		} else {
			reply->rr_nodetype = p->r_nodetype;
			reply->rr_event = p->r_event;
			reply->rr_event2 = p->r_event2;
		}
	} else {
/*
 * I don't have a proper route table, so use the dumb route.
 */
		reply->rr_nodetype = NOTNODETYPE;
		reply->rr_event = dumb_route;
		reply->rr_event2 = dumb_route;
	}

	send_reply(NULL, INT4_NIL, NOBUF);
}

/*
 *	do_getnet
 *
 *	Function;	- perform RGETNET request
 *			- replies with all information from the
 *			  "net" structure
 */
static void
do_getnet()

{
	reply->rr_reply = 0;
	reply->rr_nodeid = net.ni_nodeid;
	reply->rr_nodetype = net.ni_nodetype;
	reply->rr_origin = net.ni_origin;
	send_reply(NULL, INT4_NIL, NOBUF);
}

/*
 *	do_setnet
 *
 *	Function:	- perform RSETNET request
 *			- does not reply
 *			- fills in information found in the "net" structure
 */
static void
do_setnet()

{
	net.ni_nodeid = request->rq_nodeid;
	net.ni_nodetype = request->rq_nodetype;
	net.ni_origin = request->rq_origin;
}

/*
 *	do_noserivce
 *
 *	Function:	- not supported by the generic router
 *			- supported by the stand-alone router
 */
static void
do_noservice()

{
	reply->rr_reply = ESERVICE;
	send_reply(NULL, INT4_NIL, NOBUF);
}

/*
 *	do_loadroutes
 *
 *	Function:	- perform RLOADROUTES request
 *			- load a new route table
 *			- allocate space and if successful, receive route
 *			  table on client's event
 *			- not exactly a steel box of robustness if clients
 *			  are completely uncoordinated
 */
static void
do_loadroutes()

{
	HASH		*newrtd;		/* new route table desc. */
	struct nmsg	chunk;			/* network message header */
	int4		length;			/* #bytes in route table */
	int4		nroutes;		/* number of route entries */
	struct rwait	*p, *pp;		/* two favorite pointers */

	length = request->rq_length;
	nroutes = length / sizeof(struct route);
/*
 * Allocate the new route table.
 */
	if ((newrtd = ah_init(nroutes, (int4) sizeof(struct route),
		NOTNODEID, AHFIXED)) != (HASH *) 0) {
/*
 * Receive messages for route table.
 * Nothing fancy in the message passing for now.
 * Don't check every possible error since we are communicating with
 * another system routine.
 */
		chunk.nh_event = request->rq_src_event;
		chunk.nh_flags = INT4_NIL;
		chunk.nh_msg = (char *) newrtd->ah_table;

		while (length) {
			chunk.nh_type = INT4_NIL;
			chunk.nh_length = (length > MAXNMSGLEN) ?
				MAXNMSGLEN : length;

			if (nrecv(&chunk)) {
				panic(errno);
			}

			chunk.nh_msg += chunk.nh_length;
			length -= chunk.nh_length;
		}

		if (rtd) {
			ah_free(rtd);
		}
/*
 * Convert table to local byte ordering.  The call to ttolcpy
 * works because all elements of a route table entry are int4.
 */
		ttolcpy((int4 *) newrtd->ah_table, (int4 *) newrtd->ah_table,
			nroutes * (sizeof(struct route) / sizeof(int4)));

		rtd = newrtd;
	}
	else {
		panic(errno);
	}

	reqtype |= RSMART;

	if (rtd && wld) {
/*
 * Loop through the waiting list releasing any client
 * blocked on a nodeid that is now in the new route table
 * and delete that entry in the waiting list.
 */
		p = (struct rwait *) al_top(wld);
		while (p) {
			pp = (struct rwait *) al_next(wld, (char *) p);
			if (ah_find(rtd, p->rw_nodeid)) {
				wait_unblock(p);
			}
			p = pp;
		}
	}
}

/*
 *	do_setroute
 *
 *	Function:	- perform RSETROUTE request
 */
static void
do_setroute()

{
	struct route	rentry;			/* route table entry */
	struct rwait	*p, *pp;		/* two favorite pointers */
	int		f_wlcheck;		/* check waiting list flag */

	reply->rr_reply = 0;
	f_wlcheck = FALSE;

	if (! rtd) {
/*
 * Set the dumb route.
 */
		dumb_route = request->rq_event;
	}
	else {
/*
 * Insert or change the entry in the route table.
 */
		if (request->rq_nodeid != NOTNODEID) {
			rentry.r_nodeid = request->rq_nodeid;
			rentry.r_nodetype = request->rq_nodetype;
			rentry.r_event = request->rq_event;
			rentry.r_event2 = request->rq_event2;

			if (ah_insert(rtd, (void *) &rentry)) {
				if (errno == EFULL) {
					reply->rr_reply = ENORDESCRIPTORS;
				}
				else {
					reply->rr_reply = errno;
				}
			}
			else {
				f_wlcheck = TRUE;
			}
		}
		else {
			reply->rr_reply = EBADNODE;
		}
	}

	send_reply(NULL, INT4_NIL, NOBUF);

	if (f_wlcheck && wld) {
/*
 * Loop through the waiting list releasing any client
 * blocked on the nodeid just defined and delete that list entry.
 */
		p = (struct rwait *) al_top(wld);
		while (p) {
			pp = (struct rwait *) al_next(wld, (char *) p);
			if (p->rw_nodeid == request->rq_nodeid) {
				wait_unblock(p);
			}
			p = pp;
		}
	}
}

/*
 *	do_nodes
 *
 *	Function:	- deliver a list of nodes to the client
 *
 */
static void
do_nodes()

{
	int4		*pnodes;		/* node list */
	int4		*p;			/* favourite pointer */
	int4		nnodes;			/* length of node list */
	int4		nodetype;		/* requested node type */
	int4		typemask;		/* nodetype bit mask */
	struct route	*p_route;		/* ptr to route entry */

	reply->rr_reply = 0;
/*
 * Allocate the array of nodeids needed.
 */
	typemask = request->rq_typemask;
	nodetype = request->rq_nodetype;
	nnodes = count_ntype(nodetype, typemask) * sizeof(int4);
	nodetype &= typemask;
	p = pnodes = (int4 *) malloc(nnodes);

	if (!p) {
		reply->rr_reply = errno;
		send_reply(NULL, INT4_NIL, NOBUF);
		return;
	}
/*
 * Loop through the route table looking for the matching nodeids.
 */
	p_route = (struct route *) ah_top(rtd);
	while (p_route) {
		if ((p_route->r_nodetype & typemask) == nodetype) {
			*p++ = p_route->r_nodeid;
		}
		p_route = (struct route *) ah_next(rtd, (void *) p_route);
	}

	send_reply((char *) pnodes, nnodes, NOBUF);

	(void) free((char *) pnodes);
}

/*
 *	do_remroute
 *
 *	Function:	- perform RREMROUTE request
 */
static void
do_remroute()

{
	reply->rr_reply = 0;

	if (! rtd) {
/*
 * Set the dumb route to return an error.
 */
		dumb_route = NOTEVENT;
	}
	else {
		if (ah_delete(rtd, request->rq_nodeid)) {
			reply->rr_reply = errno;
		}
	}

	send_reply(NULL, INT4_NIL, NOBUF);
}

/*
 *	do_getsmart
 *
 *	Function:	- perform RGETSMART request
 */
static void
do_getsmart()

{
/*
 * Since we received this request, we must be smart and the client is
 * unblocked and happy.
 */
}

/*
 *	do_getntype
 *
 *	Function:	- perform RGETNTYPE request.
 */
static void
do_getntype()

{
	reply->rr_reply = 0;
	reply->rr_ntype =
		count_ntype(request->rq_nodetype, request->rq_typemask);
	send_reply(NULL, INT4_NIL, NOBUF);
}

/*
 *	do_nodewait
 *
 *	Function:	- perform RNODEWAIT request.
 */
static void
do_nodewait()

{
	struct rwait	wl_elem;		/* waiting list element */

	if ( rtd && (ah_find(rtd, request->rq_nodeid) != (void *) 0) ) {
/*
 * The nodeid is defined in the route table, reply unblocking the client.
 */
			reply->rr_reply = 0;
			send_reply(NULL, INT4_NIL, INT4_NIL);
			return;
	}
	else {
/*
 * Either the route table is not known or the nodeid is not defined.
 * Link the request on the waiting list.
 */
		if (wld == (LIST *) 0) {
			wld = al_init(sizeof(struct rwait), NULL);
			if (wld == (LIST *) 0) {
				panic(errno);
			}
		}

		wl_elem.rw_nodeid = request->rq_nodeid;
		wl_elem.rw_src_node = request->rq_src_node;
		wl_elem.rw_src_event = request->rq_src_event;
		if (al_append(wld, (char *) &wl_elem) == NULL) {
			panic(errno);
		}
	}
}

/*
 *	count_ntype
 *
 *	Function:	- Returns the number of nodes in the
 *			  route table that satisfy the given
 *			  nodetype and typemask.  
 *			- If no route table is available, returns zero.
 *
 *	Accepts:	- The nodetype bit pattern required.
 *			- The mask pattern for the nodetype.
 */
static int4
count_ntype(nodetype, typemask)

int4			nodetype,
			typemask;

{
	int4		numnodes;		/* # nodes that fit */
	struct route	*p;			/* favorite pointer   */

	numnodes        = 0;

	if (rtd) {
/*
 * Loop through the route table counting the matching nodeids.
 */
		nodetype &= typemask;
		p = (struct route *) ah_top(rtd);
		while (p) {
			if ((p->r_nodetype & typemask) == nodetype) {
				numnodes++;
			}
			p = (struct route *) ah_next(rtd, (void*) p);
		}
	}

	return(numnodes);
}

/*
 *	recv_req
 *
 *	Function:	- receive a request from a client
 */
static void
recv_req()

{
	inbound.nh_event = EVROUTER;
	inbound.nh_type = reqtype;
	inbound.nh_flags = 0;
	inbound.nh_length = 0;

	if (nrecv(&inbound)) {
		panic(errno);
	}
}

/*
 *	send_reply
 *
 *	Function:	- send a reply back to the client
 *			- The trick here is that router is a remote server
 *			  (wasn't the case in 1.1) and therefore would like
 *			  to use nsend.  Nsend() would naturally call the
 *			  router.  Since we are the router, we have to look
 *			  in our route table, fill in nh_dl_event and use
 *			  dsend().
 *
 *	Accepts:	- reply message pointer
 *			- reply message length
 */
static void
send_reply(msg, length, flags)

char			*msg;
int4			length;
int4			flags;

{
	struct route	*p;		/* favorite pointer */

	outbound.nh_event = request->rq_src_event;
	outbound.nh_type = 0;
	outbound.nh_flags = flags;
	outbound.nh_length = length;
	outbound.nh_msg = msg;

	if ((request->rq_src_node == net.ni_nodeid) ||
		(request->rq_src_node == LOCAL)) {
/*
 * The client is on the local node.  No sweat here.
 */
		outbound.nh_dl_event = outbound.nh_event;
	}
	else {
		outbound.nh_node = request->rq_src_node;
/*
 * The client is on a remote node.  Look up the route in our own tables.
 */
		if (rtd) {
			p = (struct route *) ah_find(rtd, request->rq_src_node);

			if (p == (struct route *) 0) {
				panic(errno);
			} else {
				outbound.nh_dl_event = p->r_event;
			}
		} else {
			outbound.nh_dl_event = dumb_route;
		}
	}

	if (dsend(&outbound)) {
		panic(errno);
	}
}

/*
 *	wait_unblock
 *
 *	Function:	- unblock the waiting client
 *			- delete the waiting list element
 *
 *	Accepts:	- ptr to waiting list element
 */
static void
wait_unblock(p)

struct rwait		*p;

{
	if (p) {
		reply->rr_reply = 0;
		request->rq_src_node = p->rw_src_node;
		request->rq_src_event = p->rw_src_event;
		send_reply(NULL, INT4_NIL, INT4_NIL);

		(void) al_delete(wld, (char *) p);
	}
}
