/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB/RBD
 *
 *	$Log:	getnet.c,v $
 * Revision 2.2  92/04/30  14:49:40  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/27  15:47:49  gdburns
 * Dsend() requires that nmsg be completely filled.
 * 
 * Revision 2.1.1.1  91/06/03  15:08:04  raja
 * Corrected signal masking in ksigblock().
 * 
 * Revision 2.1  91/03/20  11:37:52  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:30:47  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/20  12:17:37  gdburns
 * Changed NOTTYPE to NOTNODETYPE.
 * Always set errno in functions with vague return values.
 * 
 * Revision 2.0  90/02/20  11:56:32  trillium
 * Homo erectus
 * 
 *	Function:	- makes the RGETNET request to the local router
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 *	Accepts: 	- ptr to node information structure
 *	Returns:	- 0 or ERROR
 */

#include <kio.h>
#include <net.h>
#include <events.h>
#include <ksignal.h>
#include <typical.h>
#include <rreq.h>

/*
 * external variables
 */
extern int		errno;		/* global error variable */
extern struct kio_t	_kio;		/* kernel interface */

/*
 * global functions
 */
int			getnet();
int			getnodeid();
int			getnodetype();
int			getorigin();
int			getntype();

int
getnet(nodes)

struct net_info		*nodes;

{
	struct rreq	*request;	/* router request */
	struct rreply	*reply;		/* router reply */
	struct nmsg	nhead;		/* network header */
	int		mask;

/*
 * Set up the router request.
 */
	request = (struct rreq *) nhead.nh_data;
	request->rq_src_node = LOCAL;
	request->rq_src_event = -getpid();
	request->rq_request = RGETNET;
	request->rq_pid = getpid();

	nhead.nh_dl_event = EVROUTER;
	nhead.nh_node = LOCAL;
	nhead.nh_event = EVROUTER;
	nhead.nh_type = 0;
	nhead.nh_flags = NOBUF;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (dsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();

	if (drecv(&nhead)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);
	reply = (struct rreply *) nhead.nh_data;

	if (reply->rr_reply) {
		errno = reply->rr_reply;
		return(ERROR);
	} else {
		nodes->ni_nodeid = reply->rr_nodeid;
		nodes->ni_nodetype = reply->rr_nodetype;
		nodes->ni_origin = reply->rr_origin;
		return(0);
	}
}

/*
 *	getnodeid
 *
 *	Function:	- gets local node identifier
 *			- also sets origin node identifier
 *	Returns:	- local node identifier
 */

int
getnodeid()

{
	struct net_info	nodes;		/* complete node information */

	if (! (_kio.ki_rtf & RTF_NODEID)) {

		if (getnet(&nodes)) {
			_kio.ki_nodeid = NOTNODEID;
			_kio.ki_origin = NOTNODEID;
		} else {
			_kio.ki_nodeid = nodes.ni_nodeid;
			_kio.ki_origin = nodes.ni_origin;
			errno = 0;

			if (nodes.ni_nodeid != NOTNODEID) {
				_kio.ki_rtf |= RTF_NODEID;
			}
		}
	}

	return(_kio.ki_nodeid);
}

/*
 *	getnodetype
 *
 *	Returns:	- local node type
 */

int
getnodetype()

{
	struct net_info	nodes;		/* complete node info */

	if (getnet(&nodes)) {
		return(NOTNODETYPE);
	} else {
		errno = 0;
		return(nodes.ni_nodetype);
	}
}

/*
 *	getorigin
 *
 *	Returns:	- origin node identifier
 */

int
getorigin()

{
	(void) getnodeid();
	return(_kio.ki_origin);
}

/*
 *	getntype
 *
 *	Returns:	- # of nodes with types satisfying the
 *			  given type in all bits that are set
 *			  in the mask field.  0 may be an error.
 *
 *	Accepts:	- nodetype bit pattern required
 *			- mask pattern for the nodetype
 */

int
getntype(node_type, type_mask)

int 			node_type;
int			type_mask;

{
	struct rreq	*request;	/* router request */
	struct rreply	*reply;		/* router reply   */
	struct nmsg	nhead;		/* network header */
	int		mask;

/*
 * Set up the router request.
 */
	request = (struct rreq *)nhead.nh_data;
	request->rq_src_node = LOCAL;
	request->rq_src_event = -getpid();
	request->rq_request = RGETNTYPE;
	request->rq_pid = getpid();
	request->rq_nodetype = node_type;
	request->rq_typemask = type_mask;

	nhead.nh_dl_event = EVROUTER;
	nhead.nh_node = LOCAL;
	nhead.nh_event = EVROUTER;
	nhead.nh_type = 0;
	nhead.nh_flags = NOBUF;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (dsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();

	if (drecv(&nhead)) {
		return(0);
	}

	(void) ksigsetmask(mask);
	reply = (struct rreply *)nhead.nh_data;

	if (reply->rr_reply) {
		errno = reply->rr_reply;
		return(0);
	} else {
		errno = 0;
		return(reply->rr_ntype);
	}
}
