/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB,RBD
 *
 *	$Log:	getnodes.c,v $
 * Revision 2.2  92/04/30  14:49:41  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/06/03  15:13:42  raja
 * Corrected signal masking in ksigblock().
 * 
 * Revision 2.1  91/03/20  11:37:53  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:30:54  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:56:39  trillium
 * Homo erectus
 * 
 *
 *	Function:	- make the RNODES request to the local router
 *
 *	Accepts: 	- ptr to node array
 *			- number of elements in node array
 *			- type of nodes to be returned
 *			- bit mask for the type field
 *
 *	Returns:	- 0 or ERROR
 */


#include <net.h>
#include <events.h>
#include <ksignal.h>
#include <typical.h>
#include <rreq.h>


/*
 * external variables
 */
extern int			errno;		/* global error variable */

/*
 * locally defined global functions
 */
int				getnodes();


int
getnodes(pnodes, nnodes, nodetype, typemask)

int				*pnodes;
int				nnodes;
int				nodetype;
int				typemask;

{
	struct rreq		*request;	/* router request */
	struct rreply		*reply;		/* router reply */
	struct nmsg		nhead;		/* network header */
	int			mask;

/*
 * Set up the router request.
 */
	request = (struct rreq *) nhead.nh_data;
	request->rq_src_node = LOCAL;
	request->rq_src_event = -getpid();
	request->rq_request = RNODES;
	request->rq_nodetype = nodetype;
	request->rq_typemask = typemask;

	nhead.nh_dl_event = EVROUTER;
	nhead.nh_type = 0;
	nhead.nh_flags = NOBUF;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (dsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));

	nhead.nh_event = -getpid();
	nhead.nh_length = nnodes * sizeof(int);
	nhead.nh_msg = (char *) pnodes;

	if (drecv(&nhead)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);
	reply = (struct rreply *) nhead.nh_data;

	if (reply->rr_reply) {
		errno = reply->rr_reply;
		return(ERROR);
	}

	else {
		return(0);
	}
}

