/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	getrent.c,v $
 * Revision 2.2.1.2  93/05/14  16:34:44  raja
 * Do not re-initialize the route cache in getrentc().
 *
 * Revision 2.2.1.2  93/05/14  16:34:44  raja
 * Do not re-initialize the route cache in getrentc().
 *
 * Revision 2.2  92/04/30  14:49:42  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.4  92/04/27  15:48:06  gdburns
 * Dsend() requires that nmsg be completely filled.
 * 
 * Revision 2.1.1.3  92/02/28  16:30:29  gdburns
 * Replace bcopy() with memcpy().
 * 
 * Revision 2.1.1.2  91/11/27  14:33:49  raja
 * Added support for the second route table.
 * 
 * Revision 2.1.1.1  91/06/03  15:15:40  raja
 * Corrected signal masking in ksigblock().
 * 
 * Revision 2.1  91/03/20  11:37:54  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  90/07/13  19:30:45  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.2  90/06/20  12:18:58  gdburns
 * Changed NOTTYPE to NOTNODETYPE.
 * 
 * Revision 2.0.1.1  90/05/13  22:23:24  raja
 * Modified getroute() to check if the destination nodeid is its nodeid
 * in which case it returns the destination event it was given.  Also
 * modified getrent() to send to the router only the nodeid field.
 * 
 * Revision 2.0  90/05/10  16:17:19  raja
 * Homo erectus
 *
 *	Function:	- make the RGETROUTE request to the local router
 *			- gets a full route table entry from the router
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 *
 *	Accepts: 	- ptr to a route table structure
 *
 *	Returns:	- 0 or ERROR
 */


#include <portable.h>
#include <net.h>
#include <events.h>
#include <ksignal.h>
#include <kio.h>
#include <typical.h>
#include <rreq.h>

/*
 * external functions
 */
extern int4		rd_hash();
extern int4		wr_kick();

/*
 * external variables
 */
extern int		errno;
extern struct kio_t	_kio;


int
getrent(rent)

struct route		*rent;

{
	struct rreq	*request;	/* router request */
	struct rreply	*reply;		/* router reply */
	struct nmsg	nhead;		/* network header */
	int4		mask;
/*
 * Set up the router request.
 */
	request = (struct rreq *) nhead.nh_data;
	request->rq_src_node = LOCAL;
	request->rq_src_event = -getpid();
	request->rq_request = RGETROUTE;
	request->rq_nodeid = rent->r_nodeid;

	nhead.nh_dl_event = EVROUTER;
	nhead.nh_node = LOCAL;
	nhead.nh_event = EVROUTER;
	nhead.nh_type = 0;
	nhead.nh_flags = NOBUF;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (dsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();

	if (drecv(&nhead)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);

	reply = (struct rreply *) nhead.nh_data;

	if (reply->rr_reply) {
		errno = reply->rr_reply;
		return(ERROR);
	}

	else {
		rent->r_nodetype = reply->rr_nodetype;
		rent->r_event = reply->rr_event;
		rent->r_event2 = reply->rr_event2;
		return(0);
	}
}

/*
 *	getrentc
 *
 *	Function:	- get and cache route table entries
 *			- uses getrent to consult the router
 *
 *	Returns:	- 0 or ERROR
 */
int
getrentc(rent)

struct route		*rent;

{
	int		i;
/*
 * Check the route cache.
 */
	if ((i = rd_hash(rent->r_nodeid, _kio.ki_route, MAXRCACHE,
		sizeof(struct route))) != ERROR) {

		memcpy((char *) rent, (char *) &_kio.ki_route[i],
			sizeof(struct route));
	}
/*
 * Get the route table entry from the router.
 */
	else {
		if (getrent(rent)) {
			return(ERROR);
		} else if (rent->r_nodetype != NOTNODETYPE) {
/*
 * It is not the dumb route so insert the entry in the route cache.
 */
			i = wr_kick(rent->r_nodeid, _kio.ki_route, MAXRCACHE,
				sizeof(struct route));
			if (i != ERROR) {
				memcpy((char *) &_kio.ki_route[i],
					(char *) rent, sizeof(struct route));
			}
		}
	}

	return(0);
}

/*
 *	getroute
 *
 *	Function:	- get routing information for a destination
 *			  node and a destination event
 *
 *	Returns:	- routing event or NOTEVENT
 */
int4
getroute(dest_node, dest_event)

int4			dest_node,
			dest_event;

{
	struct route	rent;			/* route table entry */

	if (dest_node == LOCAL) {
		return(dest_event);
	} else {
/*
 * Get a route table entry.
 */
		rent.r_nodeid = dest_node;
		if (getrentc(&rent)) {
			return(NOTEVENT);
		} else if (rent.r_event == NOTEVENT) {
/*
 * This must be my own node.  Return dest_event.
 */
			return(dest_event);
		} else {
/*
 * Return the event from the route table entry.
 */
			return(rent.r_event);
		}
	}
}

/*
 *	getrtype
 *
 *	Function:	- get the type of a given node identifier
 *
 *	Returns:	- the nodetype or NOTNODETYPE
 */
int4
getrtype(nodeid)

int4			nodeid;

{
	struct route	rent;			/* route table entry */

	rent.r_nodeid = nodeid;

	if (getrentc(&rent)) {
		return(NOTNODETYPE);
	} else {
/*
 * Note: Even if getrentc() is successful, the nodetype returned 
 * in rent might still be NOTNODETYPE when a dumb route is returned.
 */
		errno = 0;
		return(rent.r_nodetype);
	}
}

/*
 *	getroute2
 *
 *	Function:	- get information from the second route table
 *			- used for minimum-hop route paths
 *
 *	Returns:	- routing event or NOTEVENT
 */
int4
getroute2(dest_node, dest_event)

int4			dest_node,
			dest_event;

{
	struct route	rent;			/* route table entry */

	if (dest_node == LOCAL) {
		return(dest_event);
	} else {
/*
 * Get a route table entry.
 */
		rent.r_nodeid = dest_node;
		if (getrentc(&rent)) {
			return(NOTEVENT);
		} else if (rent.r_event2 == NOTEVENT) {
/*
 * This must be my own node.  Return dest_event.
 */
			return(dest_event);
		} else {
/*
 * Return the second event from the route table entry.
 */
			return(rent.r_event2);
		}
	}
}
