/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	VSD
 *
 *	$Log:	rrattach.c,v $
 * Revision 2.2  92/04/30  14:49:46  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/06/03  15:18:15  raja
 * Corrected signal masking in ksigblock().
 * 
 * Revision 2.1  91/03/20  11:37:57  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:30:53  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:56:39  trillium
 * Homo erectus
 * 
 *
 *	Function:	- make the RATTACH request to the local router
 *
 *	Accepts: 	- pid and nodeid from launch process
 *
 *	Returns:	- 0 in case of success,
 *			  and ERROR in case of failure.
 */


#include <net.h>
#include <events.h>
#include <ksignal.h>
#include <typical.h>
#include <rreq.h>


/*
 * external variables
 */
extern int			errno;		/* global error variable */


int
rrattach( n_pid, n_nodeid )

int				n_pid;
int				n_nodeid;

{
	struct rreq		*request;	/* router request */
	struct rreply		*reply;		/* router reply */
	struct nmsg		nhead;		/* network header */
	int			mask;

/*
 * Set up the router request.
 */
	request = (struct rreq *) nhead.nh_data;
	request->rq_src_node = LOCAL;
	request->rq_src_event = -getpid();
	request->rq_request = RATTACH;
	request->rq_pid = n_pid;
	request->rq_nodeid = n_nodeid;

	nhead.nh_dl_event = -getpid();
	nhead.nh_type = 0;
	nhead.nh_flags = NOBUF;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (dsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();

	if (drecv(&nhead)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);
	reply = (struct rreply *) nhead.nh_data;

	if (reply->rr_reply) {
		errno = reply->rr_reply;
		return(ERROR);
	}

	else {
		return(0);
	}
}


/*
 *	rrdetach
 *
 *	Function:	- make the RDETACH request to the local router
 *
 *	Accepts: 	- pid from launch process
 *
 *	Returns:	- 0 in case of success,
 *			  and ERROR in case of failure.
 */

int
rrdetach( n_pid )

int				n_pid;

{
	struct rreq		*request;	/* router request */
	struct rreply		*reply;		/* router reply */
	struct nmsg		nhead;		/* network header */
	int			mask;

/*
 * Set up the router request.
 */
	request = (struct rreq *) nhead.nh_data;
	request->rq_src_node = LOCAL;
	request->rq_src_event = -getpid();
	request->rq_request = RDETACH;
	request->rq_pid = n_pid;

	nhead.nh_dl_event = EVROUTER;
	nhead.nh_type = 0;
	nhead.nh_flags = NOBUF;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (dsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();

	if (drecv(&nhead)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);
	reply = (struct rreply *) nhead.nh_data;

	if (reply->rr_reply) {
		errno = reply->rr_reply;
		return(ERROR);
	}

	else {
		return(0);
	}
}
