/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	rroute.c,v $
 * Revision 2.2  92/04/30  14:49:52  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/01/20  19:24:14  gdburns
 * Use the shared library.
 * 
 * Revision 2.1.1.2  91/11/27  14:38:20  raja
 * Support second route table and faster byte ordering.
 * 
 * Revision 2.1.1.1  91/06/03  15:19:23  raja
 * Corrected signal masking in ksigblock().
 * 
 * Revision 2.1  91/03/20  11:37:59  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  91/03/20  00:20:56  gdburns
 * Allow the use of buffering.
 * 
 * Revision 2.0.1.1  90/07/13  19:30:48  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:56:33  trillium
 * Homo erectus
 *
 *	Function:	- make the RLOADTABLE request to the local router
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 *			- It is a bad idea to use your own pid as an
 *			  event on another node.  We pull the same kludge
 *			  as in rflat() and turn off bit 29.
 *
 *	Accepts: 	- remote node ID
 *			- table pointer
 *			- number of table entries
 *
 *	Returns:	- 0 or ERROR
 */

#include <portable.h>
#include <net.h>
#include <events.h>
#include <ksignal.h>
#include <typical.h>
#include <rreq.h>
#include <t_types.h>
#include <shared.h>

/*
 * external variables
 */
extern int			errno;		/* global error variable */

int
rroute(node, ptable, ntable)

int				node;
struct route			*ptable;
int				ntable;

{
	struct rreq		*request;	/* router request */
	struct nmsg		nhead;		/* network header */
	int			length;		/* table length in bytes */
	int			i;		/* favourite counter */
	int			mask;
/*
 * Set up the router request.
 */
	request = (struct rreq *) nhead.nh_data;
	request->rq_src_node = (node == LOCAL) ? node : getnodeid();
	request->rq_src_event = (-getpid()) & 0xDFFFFFFF;
	request->rq_request = RLOADROUTES;
	request->rq_length = length = ntable * sizeof(struct route);

	nhead.nh_node = node;
	nhead.nh_event = EVROUTER;
	nhead.nh_type = 0;
	nhead.nh_flags = (node == LOCAL) ? NOBUF : 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
/*
 * Convert table to transputer byte ordering.
 */
	ltotcpy((int4 *) ptable, (int4 *) ptable,
		ntable * (sizeof(struct route) / sizeof(int4)));
/*
 * Ship the table.
 */
	nhead.nh_event = (-getpid()) & 0xDFFFFFFF;
	nhead.nh_type = 0;
	nhead.nh_msg = (char *) ptable;

	while (length) {
		nhead.nh_length = (length > MAXNMSGLEN) ? MAXNMSGLEN : length;

		if (nsend(&nhead)) {
			return(ERROR);
		}

		nhead.nh_msg += nhead.nh_length;
		length -= nhead.nh_length;
	}

	(void) ksigsetmask(mask);
	return(0);
}
