/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	rsetroute.c,v $
 * Revision 2.2  92/04/30  14:49:55  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.3  92/01/20  19:24:29  gdburns
 * Use the shared library.
 * 
 * Revision 2.1.1.2  91/11/27  14:40:54  raja
 * Support the second route table.
 * 
 * Revision 2.1.1.1  91/06/03  15:50:11  raja
 * Corrected signal masking in ksigblock().
 * 
 * Revision 2.1  91/03/20  11:38:01  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:30:51  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:56:37  trillium
 * Homo erectus
 *
 *	Function:	- make the RSETROUTE request to a remote router
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 *
 *	Accepts: 	- node ID of affected router process
 *			- destination node of routed message
 *			- destination event of routed message
 *			- destination event of multicasted message
 *			- type of destination node
 */

#include <portable.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <rreq.h>
#include <typical.h>
#include <shared.h>

/*
 * external variables
 */
extern int		errno;

int
rsetroute(node, dest_node, dest_event, dest_event2, dest_type)

int4				node;
int4				dest_node;
int4				dest_event;
int4				dest_event2;
int4				dest_type;

{
	struct rreq		*request;	/* router request */
	struct rreply		*reply;		/* router reply */
	struct nmsg		nhead;		/* network header */
	int4			mask;
/*
 * Set up the router request.
 */
	request = (struct rreq *) nhead.nh_data;

	request->rq_src_node = (node == LOCAL) ? node : getnodeid();
	request->rq_src_event = -getpid();
	request->rq_request = RSETROUTE;
	request->rq_nodeid = dest_node;
	request->rq_event = dest_event;
	request->rq_event2 = dest_event2;
	request->rq_nodetype = dest_type;

	nhead.nh_node = node;
	nhead.nh_event = EVROUTER;
	nhead.nh_flags = NOBUF;
	nhead.nh_length = INT4_NIL;
	nhead.nh_msg = (char *) 0;

	nhead.nh_type = (request->rq_nodetype == NT_CAST) ? RSMART : INT4_NIL;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();
	nhead.nh_type = INT4_NIL;

	if (nrecv(&nhead)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);
	reply = (struct rreply *) nhead.nh_data;

	if (reply->rr_reply) {
		errno = reply->rr_reply;
		return(ERROR);
	} else {
		return(0);
	}
}

/*
 *	rremroute
 *
 *	Function:	- remove a route from a route table on a remote node
 *
 *	Accepts:	- target node ID
 *			- destination node ID
 *
 *	Returns:	- 0 or ERROR
 */

int
rremroute(node, dest_node)

int4				node;
int4				dest_node;

{
	struct rreq		*request;	/* router request */
	struct rreply		*reply;		/* router reply */
	struct nmsg		nhead;		/* network header */
	int4			mask;
/*
 * Set up the router request.
 */
	request = (struct rreq *) nhead.nh_data;

	request->rq_src_node = getnodeid();
	request->rq_src_event = -getpid();
	request->rq_request = RREMROUTE;
	request->rq_nodeid = dest_node;

	nhead.nh_node = node;
	nhead.nh_event = EVROUTER;
	nhead.nh_type = INT4_NIL;
	nhead.nh_flags = NOBUF;
	nhead.nh_length = INT4_NIL;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();

	if (nrecv(&nhead)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);
	reply = (struct rreply *) nhead.nh_data;

	if (reply->rr_reply) {
		errno = reply->rr_reply;
		return(ERROR);
	} else {
		return(0);
	}
}
