/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	RBD
 *
 *	$Log:	bfprobe.c,v $
 * Revision 2.2  92/04/30  14:51:55  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:38:24  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:30:12  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/29  16:10:04  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0  90/05/06  16:26:29  raja
 * Homo erectus
 * 
 * Revision 2.0  90/05/06  15:57:22  raja
 * Homo erectus
 * 
 * 
 *
 *	Function:	- checks for the existance of messages with
 *			  specific events/types on a remote bufferd
 *			- receives information on whether such a 
 *			  message exists and its characteristics
 *
 *	Accepts:	- ptr to message header
 *
 *	Returns:	- 1 if message exists, 0 if not
 *			- ERROR in case of error (-1)
 *			- entries in nhead are updated: 
 *			  type, length, flags.
 */


#include <bufreq.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>


/*
 * External variables.
 */
extern int		errno;		/* error number */


int
bfprobe(header)

struct	nmsg		*header;	/* ptr to network message desc. */

{
	struct nmsg	nhead;		/* bufferd message */
	struct bufreq  	*request;	/* bufferd request */
	struct bufreply *reply;		/* bufferd reply */
	int		mask;


	request = (struct bufreq *) nhead.nh_data;
	reply = (struct bufreply *) nhead.nh_data;

	request->bfq_src_node = LOCAL;
	request->bfq_src_event = -getpid();
	request->bfq_req = BFPROBE;
	request->bfq_dl_event = header->nh_event;
	request->bfq_type = header->nh_type;

	nhead.nh_node = LOCAL;
	nhead.nh_event = EVBUFFERD;
	nhead.nh_type = BFREQ;
	nhead.nh_flags = NOBUF;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));

	nhead.nh_event = -getpid();
	nhead.nh_type = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nrecv(&nhead)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);

	if (reply->bfr_reply < 0) {
		errno = reply->bfr_reply;
		return(ERROR);
	}
	else if (reply->bfr_reply > 0) {
/*
 * Pass the information back to the user.
 */
		header->nh_type = reply->bfr_type;
		header->nh_length = reply->bfr_length;
		header->nh_flags = reply->bfr_flags;
	}

	return(reply->bfr_reply);
}
