/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	rbfstate.c,v $
 * Revision 2.2  92/04/30  14:52:21  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  92/04/27  15:49:00  gdburns
 * new information - user space threshold
 * 
 * Revision 2.1.1.1  92/03/02  18:16:52  gdburns
 * Replace bcopy() with memcpy().
 * 
 * Revision 2.1  91/03/20  11:38:25  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:30:16  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/29  16:10:19  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0  90/02/20  11:56:18  trillium
 * Homo erectus
 *
 *	Function:	- gather state from a remote bufferd
 *			- receives buffer descriptor tables from
 *			  mallocd until reply code is EEOF
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 *
 *	Accepts:	- target node ID
 *			- ptr to buffer descriptor table
 *			- # of entries in table
 *
 *	Returns:	- number of replied descriptors or ERROR
 */

#include <bufreq.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>

/*
 * external variables
 */
extern int		errno;		/* error number */

/*
 * local variables
 */
static char		hole[MAXNMSGLEN];

int
rbfstate(nodeid, tbldesc, maxsize, parms)

int			nodeid;		/* target node ID */
int			maxsize;	/* size of descriptor table */
struct bfstate		tbldesc[];	/* ptr to descriptr table */
struct bfparms		*parms;		/* buffer parameters */

{
	struct nmsg	nhead;		/* network message desc. */
	struct bufreq  	*request;	/* bufferd request */
	struct bufreply *reply;		/* bufferd reply */
	struct bfstate	*bfhole;	/* ptr incoming buffer descs */
	int		size;		/* # descs in this msg */
	int		totsize;	/* total # replied descs */
	int		i;		/* favourite index */
	int		mask;		/* favourite index */

	totsize = 0;
	bfhole = (struct bfstate *) hole;
	request = (struct bufreq *) nhead.nh_data;
	reply = (struct bufreply *) nhead.nh_data;

	request->bfq_src_node = (nodeid == LOCAL) ? nodeid : getnodeid();
	request->bfq_src_event = -getpid();
	request->bfq_req = BFSTATE;

	nhead.nh_node = nodeid;
	nhead.nh_event = EVBUFFERD;
	nhead.nh_type = BFREQ;
	nhead.nh_flags = NOBUF;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();
	nhead.nh_type = 0;
	nhead.nh_length = MAXNMSGLEN;
	nhead.nh_msg = hole;

	if (nrecv(&nhead)) {
		return(ERROR);
	}

	while (! reply->bfr_reply) {
		size = nhead.nh_length / sizeof(struct bfstate);
		totsize += size;

		if (maxsize) {
/*
 * Convert byte ordering (sigh).
 */
			for (i = 0; i < size; ++i) {
				bfhole[i].bfs_node = ttol(bfhole[i].bfs_node);
				bfhole[i].bfs_event = ttol(bfhole[i].bfs_event);
				bfhole[i].bfs_type = ttol(bfhole[i].bfs_type);
				bfhole[i].bfs_length =
					ttol(bfhole[i].bfs_length);
			}

			if (size > maxsize) size = maxsize;

			(void) memcpy((char *) tbldesc, hole,
				size * sizeof(struct bfstate));
			maxsize -= size;
			tbldesc += size;
		}

		nhead.nh_length = MAXNMSGLEN;

		if (nrecv(&nhead)) {
			return(ERROR);
		}
	}

	(void) ksigsetmask(mask);

	if (reply->bfr_reply == EEOF) {
		parms->bfp_maxworkers = reply->bfr_maxworkers;
		parms->bfp_maxspace = reply->bfr_maxspace;
		parms->bfp_nworkers = reply->bfr_nworkers;
		parms->bfp_nspace = reply->bfr_nspace;
		parms->bfp_nfull = reply->bfr_nfull;
		parms->bfp_maxusr = reply->bfr_maxusr;
		return(totsize);
	} else {
		errno = reply->bfr_reply;
		return(ERROR);
	}
}
