/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	rbtadd.c,v $
 * Revision 2.2  92/04/30  14:52:22  trillium
 * Ohio Release
 * 
 * Revision 2.1  92/01/20  19:29:55  gdburns
 * initial revision
 * 
 *	Function:	- adds object to remote boot daemon
 *
 *	Accepts:	- remote nodeid
 *			- object bit filed (one object only)
 *			- tag for flatd
 *
 *	Returns:	- 0 or ERROR
 */

#include <bootreq.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <portable.h>
#include <shared.h>
#include <typical.h>

/*
 * external variables
 */
extern int4			errno;		/* global error */

/*
 * local functions
 */
static int			rbtaddobj();

int
rbtadd(nodeid, object, tag)

int4				nodeid;		/* target node ID */
int4				object;		/* bit field of boot objects */
int4				tag;		/* flatd tag */

{
	struct nmsg		nhead;		/* network message */
	struct bootreq		*request;	/* bootd request template */

	request = (struct bootreq *) nhead.nh_data;
	request->btq_object = object;
	return(rbtaddobj(nodeid, tag, &nhead));
}

/*
 *	rbtaddstage
 *
 *	Function:	- adds staged program to remote boot daemon
 *	Accepts:	- remote nodeid
 *			- tag for flatd
 *			- option flags
 *			- stack size
 *			- process priority
 *	Returns:	- 0 or ERROR
 */

int
rbtaddstage(nodeid, tag, flags, stksize, priority)

int4				nodeid;
int4				tag;
int4				flags;
int4				stksize;
int4				priority;

{
	struct nmsg		nhead;		/* network message */
	struct bootreq		*request;	/* bootd request template */

	request = (struct bootreq *) nhead.nh_data;
	request->btq_object = BTO_STAGE;
	request->btq_flags = flags;
	request->btq_stksize = stksize;
	request->btq_priority = priority;

	return(rbtaddobj(nodeid, tag, &nhead));
}

/*
 *	rbtaddfrozen
 *
 *	Function:	- adds frozen boot image to remote boot daemon
 *	Accepts:	- remote nodeid
 *			- tag for flatd
 *			- frozen size
 *	Returns:	- 0 or ERROR
 */

int
rbtaddfrozen(nodeid, tag, size)

int4				nodeid;
int4				tag;
int4				size;

{
	struct nmsg		nhead;		/* network message */
	struct bootreq		*request;	/* bootd request template */

	request = (struct bootreq *) nhead.nh_data;
	request->btq_object = BTO_FROZEN;
	request->btq_size = size;

	return(rbtaddobj(nodeid, tag, &nhead));
}

/*
 *	rbtaddobj
 *
 *	Function:	- adds object to remote boot daemon
 *
 *	Accepts:	- remote nodeid
 *			- tag for flatd
 *			- ptr to network msg descriptor
 *
 *	Returns:	- 0 or ERROR
 */

static int
rbtaddobj(nodeid, tag, nhead)

int4				nodeid;
int4				tag;
struct nmsg			*nhead;

{
	struct bootreq		*request;	/* bootd request template */
	struct bootreply	*reply;		/* bootd reply template */
	int4			mask;		/* signal mask */

	request = (struct bootreq *) nhead->nh_data;
	reply = (struct bootreply *) nhead->nh_data;

	request->btq_src_node = (nodeid == LOCAL) ? nodeid : getnodeid();
	request->btq_src_event = -getpid();
	request->btq_request = BTQADD;
	request->btq_tag = tag;

	nhead->nh_node = nodeid;
	nhead->nh_event = EVBOOTD;
	nhead->nh_type = 0;
	nhead->nh_flags = 0;
	nhead->nh_length = 0;
	nhead->nh_msg = (char *) 0;

	if (nsend(nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead->nh_event = -getpid();

	if (nrecv(nhead)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);

	if (reply->btr_reply) {
		errno = reply->btr_reply;
		return(ERROR);
	} else {
		return(0);
	}
}
