/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	rbtspread.c,v $
 * Revision 2.2  92/04/30  14:52:34  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/02/23  19:03:49  gdburns
 * Initialize the error variable.
 * 
 * Revision 2.1  92/01/20  19:29:57  gdburns
 * initial revision
 * 
 *	Function:	- recursively boots starting at the remote node
 *	Accepts:	- remote nodeid array
 *			- length of array
 *			- object selector
 *	Returns:	- 0 or ERROR
 */

#include <bootreq.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <portable.h>
#include <shared.h>
#include <typical.h>

/*
 * external variables
 */
extern int4		errno;		/* global error */

int
rbtspread(pnode, nnode, object)

int4			pnode[];
int4			nnode;
int4			object;

{
	struct nmsg	nhead;		/* network message */
	struct bootreq	*request;	/* bootd request template */
	struct bootreply
			*reply;		/* bootd reply template */
	int4		mask;		/* signal mask */
	int4		i;		/* favourite index */
	int		err = 0;	/* error code */

	request = (struct bootreq *) nhead.nh_data;
	reply = (struct bootreply *) nhead.nh_data;

	request->btq_src_event = -getpid();
	request->btq_request = BTQSPREAD;
	request->btq_object = object;

	nhead.nh_event = EVBOOTD;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	for (i = 0; i < nnode; ++i) {
		nhead.nh_node = pnode[i];
		request->btq_src_node = (pnode[i] == LOCAL) ?
			pnode[i] : getnodeid();

		if (nsend(&nhead)) {
			return(ERROR);
		}

		if (i == 0) {
			mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
		}
	}

	nhead.nh_event = -getpid();

	for (i = 0; i < nnode; ++i) {

		if (nrecv(&nhead)) {
			return(ERROR);
		}

		if (reply->btr_reply) {
			err = (int) reply->btr_reply;
		}
	}

	(void) ksigsetmask(mask);

	if (err) {
		errno = err;
		return(ERROR);
	} else {
		return(0);
	}
}
