/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	rbtstate.c,v $
 * Revision 2.2  92/04/30  14:52:38  trillium
 * Ohio Release
 * 
 * Revision 2.1  92/01/20  19:29:58  gdburns
 * initial revision
 * 
 *	Function:	- gathers state from a remote bootd
 *
 *	Accepts:	- target node ID
 *			- ptr to program descriptor table
 *			- # of entries in caller's table
 *			- bootd object selector
 *
 *	Returns:	- # of descriptors in bootd
 */

#include <bootreq.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <portable.h>
#include <t_types.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int			errno;		/* error number */

int4
rbtstate(nodeid, tbldesc, maxprog, object)

int4				nodeid;		/* target node ID */
int4				maxprog;	/* size of descriptor table */
int4				object;		/* which objects? */
struct bootstate		tbldesc[];	/* ptr to descriptr table */

{
	struct nmsg		nhead;		/* network message desc. */
	struct bootreq  	*request;	/* bootd request */
	struct bootreply   	*reply;		/* bootd reply */
	int4			mask;		/* signal mask */

	if (maxprog == 0) {
		return(0);
	}

	request = (struct bootreq *) nhead.nh_data;
	reply = (struct bootreply *) nhead.nh_data;

	request->btq_src_node = (nodeid == LOCAL) ? nodeid : getnodeid();
	request->btq_src_event = -getpid();
	request->btq_request = BTQSTATE;
	request->btq_maxprog = maxprog;
	request->btq_object = object;

	nhead.nh_node = nodeid;
	nhead.nh_event = EVBOOTD;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();
	nhead.nh_type = 0;
	nhead.nh_length = sizeof(struct bootstate);
	nhead.nh_msg = (char *) tbldesc;

	if (nrecv(&nhead)) {
		return(ERROR);
	}

	if (reply->btr_nprog < maxprog) {
		maxprog = reply->btr_nprog;
	}

	while (!reply->btr_reply && (maxprog > 0)) {

		tbldesc->bts_load = (void *) ttol((int4) tbldesc->bts_load);
		tbldesc->bts_data = (void *) ttol((int4) tbldesc->bts_data);
		tbldesc->bts_stksize = ttol(tbldesc->bts_stksize);
		tbldesc->bts_flags = ttol(tbldesc->bts_flags);
		tbldesc->bts_priority = ttol(tbldesc->bts_priority);

		maxprog--;
		tbldesc++;
		nhead.nh_msg = (char *) tbldesc;

		if (maxprog > 0) {

			if (nrecv(&nhead)) {
				return(ERROR);
			}
		}
	}

	(void) ksigsetmask(mask);

	if (reply->btr_reply) {
		errno = reply->btr_reply;
		return(ERROR);
	} else {
		return(reply->btr_nprog);
	}
}
