/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB
 *
 *	$Log:	rcask.c,v $
 * Revision 2.2  92/04/30  14:52:41  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  17:15:47  raja
 * Use CFSYS message type in the request.
 * 
 * Revision 2.1  91/03/20  11:38:28  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:30:15  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/29  16:10:28  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0  90/02/20  11:56:20  trillium
 * Homo erectus
 * 
 *
 *	Function:	- get the number of branches of a cast ID at
 *			  a given node
 *
 *	Accepts:	- target node
 *			- multicast identifier
 */


#include <castreq.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>


/*
 * external variables
 */
extern int			errno;


int
rcask(node, castid)

int				node;
int				castid;

{
	struct nmsg		nhead;
	struct castreq		*req;
	struct castreply	*reply;
	int			mask;


	req = (struct castreq *) nhead.nh_data;
	req->cq_req = CASK;
	req->cq_src_event = -getpid();
	req->cq_castid = castid;
	req->cq_src_node = getnodeid();

	nhead.nh_node = node;
	nhead.nh_event = EVCASTD;
	nhead.nh_type = CFSYS;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	reply = (struct castreply *) nhead.nh_data;
	nhead.nh_event = -getpid();
	nhead.nh_type = 0;

	if (nrecv(&nhead)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);

	if (reply->cr_reply) {
		errno = reply->cr_reply;
		return(ERROR);
	}

	else {
		return(reply->cr_nbranch);
	}
}
