/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB, RBD
 *
 *	$Log:	rcast.c,v $
 * Revision 2.2  92/04/30  14:52:42  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  17:19:55  raja
 * Use CFSYS message type in the request.  Add rcasttype() request.
 * Various cosmetic changes and enhanced portability.
 * 
 * Revision 2.1  91/03/20  11:38:29  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  90/07/13  19:29:48  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.2  90/07/04  14:35:15  gdburns
 * Revised for better robustness in castd registration.
 * 
 * Revision 2.0.1.1  90/06/29  16:11:15  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0  90/02/20  11:55:58  trillium
 * Homo erectus
 *
 *	Function:	- register a multicast with bcastd
 *
 *	Accepts:	- target node
 *			- multicast identifier
 *			- combine function selector
 *			- array of multicast members
 *			- number of multicast members
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 */

#include <portable.h>
#include <castreq.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>

/*
 * external variables
 */
extern int			errno;

/*
 * external functions
 */
extern char			*malloc();
extern int4			getntype();
extern int4			getrtype();

int
rcast(node, castid, combine, pcast, ncast)

int4				node;
int4				castid;
int4				combine;
int4				*pcast;
int4				ncast;

{
	struct nmsg		nhead;		/* network msg header */
	struct castreq		*req;		/* CADD request */
	struct castreply	*reply;		/* CADD reply */
	int4			mask;		/* signal mask */

	if (ncast == 0) {
		errno = EINVAL;
		return(ERROR);
	}

	req = (struct castreq *) nhead.nh_data;
	req->cq_req = CADD;
	req->cq_src_event = -getpid();
	req->cq_castid = castid;
	req->cq_combine = combine;
	req->cq_src_node = getnodeid();
	req->cq_root_node = (node == LOCAL) ? getnodeid() : node;
/*
 * Convert node list to Trillium byte order.
 */
	ltotcpy((uint4 *) pcast, (uint4 *) pcast, ncast);

	nhead.nh_node = node;
	nhead.nh_event = EVCASTD;
	nhead.nh_type = CFSYS;
	nhead.nh_flags = 0;
	nhead.nh_length = ncast * sizeof(int4);
	nhead.nh_msg = (char *) pcast;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	reply = (struct castreply *) nhead.nh_data;
	nhead.nh_event = -getpid();
	nhead.nh_type = 0;

	if (nrecv(&nhead)) {
		(void) ksigsetmask(mask);
		return(ERROR);
	}

	(void) ksigsetmask(mask);
/*
 * Convert node list to local byte order.
 */
	ttolcpy((uint4 *) pcast, (uint4 *) pcast, ncast);

	if (reply->cr_reply) {
		errno = reply->cr_reply;
		return(ERROR);
	}
	else {
		return(0);
	}
}

/*
 *	rmelt
 *
 *	Function:	- remove a multicast registration with castd
 *
 *	Accepts:	- target node
 *			- multicast identifier
 *
 *	Returns:	- 0 or ERROR
 */
int
rmelt(node, castid)

int4			node;
int4			castid;

{
	struct nmsg	nhead;		/* network message header */
	struct castreq	*req;		/* CREM request */

	req = (struct castreq *) nhead.nh_data;
	req->cq_req = CREM;
	req->cq_castid = castid;

	if ( (getrtype(castid) == NOTNODETYPE) ||
			( ! tiscast(getrtype(castid))) ) {
		errno = EINVAL;
		return(ERROR);
	}

	nhead.nh_node = node;
	nhead.nh_event = EVCASTD;
	nhead.nh_type = CFSYS;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	return(nsend(&nhead));
}

/*
 *	rcasttype
 *
 *	Function:	- register a multicast with bcastd
 *			  for certain node types
 *
 *	Accepts:	- target node
 *			- multicast identifier
 *			- combine function selector
 *			- node type desired
 *			- node type bitmask
 *
 *	Returns:	- 0 or ERROR
 */
int
rcasttype(node, castid, combine, nodetype, typemask)

int4			node;
int4			castid;
int4			combine;
int4			nodetype;
int4			typemask;

{
	int4		nnodes;			/* number of nodes */
	int4		*pnodes;		/* array of nodeids */
	int		ret;			/* returned value */
	int		save_errno;		/* saved errno */
/*
 * Find the specified nodes.
 */
	if ((nnodes = getntype(nodetype, typemask)) <= 0) {
		errno = EINVAL;
		return(ERROR);
	}

	pnodes = (int4 *) malloc((uint4) nnodes * sizeof(int4));
	if (pnodes == (int4 *) 0) {
		return(ERROR);
	}

	if (getnodes(pnodes, nnodes, nodetype, typemask)) {
		save_errno = errno;
		(void) free((char *) pnodes);
		errno = save_errno;
		return(ERROR);
	}
/*
 * Do the actual broadcast registration.
 */
	ret = rcast(node, castid, combine, pnodes, nnodes);

	save_errno = errno;
	(void) free((char *) pnodes);
	errno = save_errno;
	return(ret);
}
