/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	rcreate.c,v $
 * Revision 2.2  92/04/30  14:52:43  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:38:31  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  90/07/13  19:30:02  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.2  90/06/29  16:11:18  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0.1.1  90/06/22  15:40:17  gdburns
 * Checked destination node for all possible castids and
 * set source node accordingly.
 * 
 * Revision 2.0  90/02/20  11:56:12  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:13:19  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:48:25  trillium
 * Initial revision
 * 
 *
 * 	Function:	- create a process on a given node
 * 			- assumes load module already in memory
 * 			- assumes argv structure already in memory
 * 
 * 	Accepts:	- target node ID
 * 			- address of load module
 * 			- process flags
 * 			- size of workspace
 * 			- ptr to pid (value returned)
 * 			- argc
 * 			- argv tag
 * 
 * 	Returns:	- index or ERROR
 */


#include <net.h>
#include <kio.h>
#include <events.h>
#include <ksignal.h>
#include <typical.h>
#include <t_types.h>
#include <preq.h>


/*
 * external functions
 */
extern int			getrtype();	/* returns node type */

/*
 * external variables
 */
extern int			errno;		/* global error variable */
extern struct kio_t		_kio;


int
rcreate(nodeid, loadpt_tag, flags, wpsize, pid, argc, argv_tag)

int				wpsize,		/* workspace size */
				loadpt_tag,	/* tag of load module */
				argv_tag,	/* tag of argument strings */
				flags,		/* process flags */
				nodeid,		/* target node address */
				*pid,		/* ptr to returned pid */
				argc;		/* number of runtime args */

{
	struct nmsg		header;		/* network header */
	struct preq  	 	request;	/* kenya request  */
	struct preply		reply;		/* kenya reply    */
	int			mask;


	request.pq_loadpt = ltot(loadpt_tag);
	request.pq_flags = ltot(flags);
	request.pq_stksize = ltot(wpsize);
	request.pq_argc = ltot(argc);
	request.pq_argv = ltot(argv_tag);
	request.pq_stdin = ltot(_kio.ki_stdin);
	request.pq_stdout = ltot(_kio.ki_stdout);
	request.pq_stderr = ltot(_kio.ki_stderr);

	request.pq_src_node = ((nodeid == LOCAL) ||
		(getrtype(nodeid) & NT_CAST)) ?
		ltot(nodeid) : ltot(getnodeid());
	request.pq_src_event = ltot(-getpid());
	request.pq_req = ltot(PCREATE);

	header.nh_node = nodeid;
	header.nh_event = EVKENYAD;
	header.nh_flags = 0;
	header.nh_type = 0;
	header.nh_length = sizeof(request);
	header.nh_msg = (char *) &request;

	if (nsend(&header)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	header.nh_event = -getpid();
	header.nh_length = sizeof(reply);
	header.nh_msg = (char *) &reply;

	if (nrecv(&header)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);

	if (reply.pr_reply) {
		errno = ttol(reply.pr_reply);
		return(ERROR);
	}

	else {
		*pid = ttol(reply.pr_pid);
		return(ttol(reply.pr_index));
	}
}
