/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	rcstate.c,v $
 * Revision 2.2  92/04/30  14:52:45  trillium
 * Ohio Release
 * 
 *	Function:	- return state information from castd
 *			- primary information about castids registered
 *			- secondary information about reeling in progress
 */

#include <portable.h>
#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <terror.h>
#include <typical.h>
#include <t_types.h>
#include <castreq.h>

/*
 * external variables
 */
extern int		errno;			/* error number */

/*
 * local functions
 */
static int4		do_rcstate();

/*
 *	rcstate
 *
 *	Function:	- get primary state information from a remote castd
 *
 *	Accepts:	- target node ID
 *			- ptr to cast state table
 *			- maximum # of entries in table
 *
 *	Returns:	- total # of entries in castd or ERROR
 */
int4
rcstate(nodeid, table, maxsize)

int4			nodeid;
struct cstate		*table;
int4			maxsize;

{
	return(do_rcstate(nodeid, NOTNODEID, (void *) table, maxsize,
		(int4) sizeof(struct cstate), CSTAT));
}

/*
 *	rcstate2
 *
 *	Function:	- get secondary state information from a remote
 *			  castd for a specific cast identifier
 *
 *	Accepts:	- target node ID
 *			- castid requested
 *			- ptr to cast state table
 *			- maximum # of entries in table
 *
 *	Returns:	- total # of entries in castd or ERROR
 */
int4
rcstate2(nodeid, castid, table, maxsize)

int4			nodeid;
int4			castid;
struct cstate2		*table;
int4			maxsize;

{
	return(do_rcstate(nodeid, castid, (void *) table, maxsize,
		(int4) sizeof(struct cstate2), CSTAT2));
}

/*
 *	do_rcstate
 *
 *	Function:	- gather state from a remote castd
 *			- handles both primary and secondary information
 *
 *	Accepts:	- target node ID
 *			- castid requested (or NOTNODEID)
 *			- ptr to state table
 *			- maximum # of entries in table
 *			- size of table element
 *			- request code (CSTAT or CSTAT2)
 *
 *	Returns:	- total # of entries in castd or ERROR
 */
static int4
do_rcstate(nodeid, castid, table, maxsize, elemsize, reqcode)

int4			nodeid;
int4			castid;
void			*table;
int4			maxsize;
int4			elemsize;
int4			reqcode;

{
	struct nmsg	nhead;		/* network message desc. */
	struct castreq 	*request;	/* castd request */
	struct castreply
			*reply;		/* bufferd reply */
	int4		mask;		/* signal bit mask */
	int4		rem;		/* remaining # of table entries */

	request = (struct castreq *) nhead.nh_data;
	reply = (struct castreply *) nhead.nh_data;

	request->cq_src_node = (nodeid == LOCAL) ? nodeid : getnodeid();
	request->cq_src_event = -getpid();
	request->cq_castid = castid;
	request->cq_maxstat = maxsize;
	request->cq_req = reqcode;

	nhead.nh_node = nodeid;
	nhead.nh_event = EVCASTD;
	nhead.nh_type = CFSYS;
	nhead.nh_flags = DINTMSG;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return((int4) ERROR);
	}
	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));

	nhead.nh_event = -getpid();
	nhead.nh_type = 0;
	nhead.nh_flags = DINTMSG;
	nhead.nh_length = maxsize * elemsize;
	nhead.nh_msg = (char *) table;

	if (nrecv(&nhead)) {
		(void) ksigsetmask(mask);
		return((int4) ERROR);
	}

	if (reply->cr_reply) {
		(void) ksigsetmask(mask);
		errno = (int) reply->cr_reply;
		return((int4) ERROR);
	}
/*
 * Get all the returning packets.
 */
	else {
		rem = min(reply->cr_result, maxsize) -
				(nhead.nh_length / elemsize);
		maxsize = reply->cr_result;

		while (rem > 0) {
			nhead.nh_msg += nhead.nh_length;
			nhead.nh_length = rem * elemsize;

			if (nrecv(&nhead)) {
				(void) ksigsetmask(mask);
				return((int4) ERROR);
			}

			rem -= nhead.nh_length / elemsize;
		}

		(void) ksigsetmask(mask);
		return(maxsize);
	}
}
