/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	rdoom.c,v $
 * Revision 2.2  92/04/30  14:53:04  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:38:32  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  90/07/13  19:29:58  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.2  90/06/29  16:11:20  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0.1.1  90/06/22  15:41:09  gdburns
 * Checked destination node for all possible castids and
 * set source node accordingly.
 * 
 * Revision 2.0  90/02/20  11:56:07  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:13:24  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:48:27  trillium
 * Initial revision
 * 
 *
 *	Function:	- doom a process on a given node
 *
 *	Accepts:	- process ID
 *			- process index
 *			- node address
 *			- signal number
 *
 *	Returns:	- 0 or ERROR
 */


#include <net.h>
#include <events.h>
#include <ksignal.h>
#include <typical.h>
#include <t_types.h>
#include <preq.h>

/*
 * external variables
 */
extern int			errno;		/* global error variable */

/*
 * external functions
 */
extern int			getrtype();	/* returns node type */

int
rdoom(nodeid, sflags, svalue, signal)

int				sflags;		/* process select flags */
int				svalue;		/* process select value */
int				nodeid;		/* target node address */
int				signal;		/* signal number */

{
	struct nmsg		header;		/* network header */
	struct preq  	 	request;	/* kenya request  */
	struct preply		reply;		/* kenya reply    */
	int			mask;


	request.pq_sel_method = ltot(sflags);
	request.pq_sel_value = ltot(svalue);
	request.pq_src_node =
		((nodeid == LOCAL) || (getrtype(nodeid) & NT_CAST)) ?
		ltot(nodeid) : ltot(getnodeid());
	request.pq_src_event = ltot(-getpid());
	request.pq_signal = ltot(signal);
	request.pq_req = ltot(PDOOM);

	header.nh_node = nodeid;
	header.nh_event = EVKENYAD;
	header.nh_type = 0;
	header.nh_flags = 0;
	header.nh_length = sizeof(request);
	header.nh_msg = (char *) &request;

	if (nsend(&header)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	header.nh_event = -getpid();
	header.nh_length = sizeof(reply);
	header.nh_msg = (char *) &reply;

	if (nrecv(&header)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);

	if (reply.pr_reply) {
		errno = ttol(reply.pr_reply);
		return(ERROR);
	}

	else {
		return(0);
	}
}
