/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	recho.c,v $
 * Revision 2.2  92/04/30  14:53:08  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:38:35  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  90/07/13  19:29:52  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.2  90/06/29  16:11:24  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0.1.1  90/06/22  15:41:12  gdburns
 * Checked destination node for all possible castids and
 * set source node accordingly.
 * 
 * Revision 2.0  90/02/20  11:56:01  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:13:30  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:48:29  trillium
 * Initial revision
 * 
 *
 *	Function:	- runtime call to the echo daemon
 *
 *	Accepts:	- a destination node
 *			- outgoing message pointer
 *			- imcoming message pointer
 *			- length of the outgoing message
 *
 *	Returns:	- length of the incoming message
 */

#include <events.h>
#include <ksignal.h>
#include <net.h>
#include <ereq.h>
#include <typical.h>
#include <terror.h>

/*
 * external variables
 */
extern int			errno;

/*
 * external functions
 */
extern int			getrtype();	/* returns node type */

int
recho(node, src_msg, dest_msg, length)

int				node,		/* destination node */
				length;		/* length of echo message */
char				*src_msg,	/* ptr to source message */
				*dest_msg;	/* echo message here */

{
	struct nmsg		nhead;		/* message descriptor */
	struct ereq		*request;	/* echo request */
	int			mask;

/*
 * Check sanity on the length field.
 */
	if ((length < 0) || (length > MAXNMSGLEN)) {
		errno = EMSGLEN;
		return(ERROR);
	}
/*
 * Fill in echod request.
 */
	request = (struct ereq *) nhead.nh_data;
	request->eq_event = -getpid();
	request->eq_node =
		((node == LOCAL) || (getrtype(node) & NT_CAST)) ?
		node : getnodeid();
	
	nhead.nh_node = node;
	nhead.nh_length = length;
	nhead.nh_event = EVECHOD;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_msg = src_msg;
/*
 * Send request to the echo daemon.
 */
	if (nsend(&nhead) == ERROR) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();
	nhead.nh_msg = dest_msg;
	nhead.nh_length = length;
/*
 * Received echoed message.
 */
	if (nrecv(&nhead) == ERROR) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);

	return(nhead.nh_length);
}
