/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	rflat.c,v $
 * Revision 2.2  92/04/30  14:53:11  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:38:36  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.5  91/03/19  23:58:46  gdburns
 * Use a modified reply event.
 * 
 * Revision 2.0.1.4  90/07/13  19:29:56  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.3  90/06/29  16:11:27  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0.1.2  90/06/22  15:41:15  gdburns
 * Checked destination node for all possible castids and
 * set source node accordingly.
 * 
 * Revision 2.0.1.1  90/05/16  15:42:34  gdburns
 * Added extra reply message from flatd to ensure that the data
 * is fully loaded before the client can start announcing the tag value.
 * 
 * Revision 2.0  90/02/20  11:56:06  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:13:33  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0.1.3  88/06/01  17:10:24  gdburns
 * Whoops.  Forgot to include the correct reference to flatreq.h
 * 
 * Revision 1.0.1.2  88/06/01  16:34:55  gdburns
 * Updated to use new stateless protocol with flatd.
 * 
 * Revision 1.0.1.1  88/03/04  14:13:40  gdburns
 * The type field, used to ensure proper sequencing, was not correctly set.
 * 
 * Revision 1.0  87/12/19  15:48:30  trillium
 * Initial revision
 * 
 *
 *	Function:	- runtime access to the flatd server
 *			- loads a buffer into a node and assigns it a tag
 */


#include <events.h>
#include <flatreq.h>
#include <ksignal.h>
#include <net.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int			errno;

/*
 * external functions
 */
extern int			getrtype();	/* returns node type */

int
rflat(nodeid, buffer, ld_length, ma_length, tag)

int				nodeid,		/* destination node ID */
				ld_length,	/* length of buffer to load */
				ma_length,	/* length of buffer to malloc */
				tag;		/* tag to mark loaded buffer */
char				*buffer;	/* buffer pointer */

{
	struct flat_request	*request;	/* flatd request structure */
	struct flat_reply	*reply;		/* flatd reply structure */
	struct nmsg		nhead;		/* message descriptor */
	int			mask;


	if (!ma_length) return(0);

	request = (struct flat_request *) nhead.nh_data;
	reply = (struct flat_reply *) nhead.nh_data;

	request->fl_src_node =
		((nodeid == LOCAL) || (getrtype(nodeid) & NT_CAST)) ?
		nodeid : getnodeid();
/*
 * It is dangerous to use our pid as an event on a remote node.
 * We turn off bit 30.
 */
	request->fl_src_event = (-getpid()) & 0xBFFFFFFF;
	request->fl_req = FL_LOAD;
	request->fl_ld_length = ld_length;
	request->fl_ma_length = ma_length;
	request->fl_tag = tag;

	nhead.nh_node = nodeid;
	nhead.nh_event = EVFLATD;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_msg = (char *) 0;
	nhead.nh_length = 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = (-getpid()) & 0xBFFFFFFF;
	
	if (nrecv(&nhead)) {
		return(ERROR);
	}

	if (reply->fl_reply) {
		errno = reply->fl_reply;
		return(ERROR);
	}

	else {
		if (!ld_length) return(0);
/*
 * Load buffer to destination node.
 */
		nhead.nh_node = nodeid;
		nhead.nh_msg = buffer;
		nhead.nh_type = 0;

		while (ld_length > 0) {

			nhead.nh_length = (ld_length > MAXNMSGLEN)
						? MAXNMSGLEN : ld_length;

			if (nsend(&nhead)) {
				return(ERROR);
			}

			ld_length -= nhead.nh_length;
			nhead.nh_msg += nhead.nh_length;
			nhead.nh_type += nhead.nh_length;
		}
	}
/*
 * Receive a reply that ensures that all data has been fully loaded.
 */
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nrecv(&nhead)) {
		return(ERROR);
	}

	if (reply->fl_reply) {
		errno = reply->fl_reply;
		return(ERROR);
	}

	(void) ksigsetmask(mask);

	return(0);
}
