/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	rget.c,v $
 * Revision 2.2  92/04/30  14:53:14  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:38:38  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:30:05  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/29  16:11:29  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0  90/02/20  11:56:14  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:13:36  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:48:31  trillium
 * Initial revision
 * 
 *
 *	Function:	- runtime access to the flatd server
 *			- inquires about a given tag
 */


#include <net.h>
#include <typical.h>
#include <events.h>
#include <flatreq.h>
#include <ksignal.h>


/*
 * external variables
 */
extern int			errno;

/*
 * local functions
 */
static char			*do_get();


/*
 *	rget
 *
 *	Function:	- make the FL_GET request
 *			- simply calls do_get
 *
 *	Accepts:	- destination node ID
 *			- tag value
 *
 *	Returns:	- address of given tag or ERROR
 */

char
*rget(nodeid, tag)

int			nodeid,		/* destination node */
			tag;		/* memory tag value */

{
	return(do_get(nodeid, tag, FL_GET));
}


/*
 *	rforget
 *
 *	Function:	- make the FL_FORGET request
 *			- simply calls do_get
 *
 *	Accepts:	- destination node ID
 *			- tag value
 *
 *	Returns:	- address of given tag or ERROR
 */

char
*rforget(nodeid, tag)

int			nodeid,		/* destination node */
			tag;		/* memory tag value */

{
	return(do_get(nodeid, tag, FL_FORGET));
}


/*
 *	do_get
 *
 *	Function:	- gets an address for a given tag by calling
 *			  the flatd server
 *			- request type is either FL_GET or FL_FORGET
 *
 *	Accepts:	- destination node ID
 *			- tag value
 *			- request type
 *
 *	Returns:	- address of given tag or ERROR
 */

static char
*do_get(nodeid, tag, req)

int				nodeid,		/* destination node ID */
				tag,		/* tag to mark loaded buffer */
				req;		/* request type */

{
	struct flat_request	*request;	/* flatd request structure */
	struct flat_reply	*reply;		/* flatd reply structure */
	struct nmsg		nhead;		/* message descriptor */
	int			mask;


	request = (struct flat_request *) nhead.nh_data;
	reply = (struct flat_reply *) nhead.nh_data;

	request->fl_src_node = (nodeid == LOCAL) ? nodeid : getnodeid();
	request->fl_src_event = -getpid();
	request->fl_req = req;
	request->fl_tag = tag;

	nhead.nh_node = nodeid;
	nhead.nh_event = EVFLATD;
	nhead.nh_type = 0;
	nhead.nh_flags = (nodeid == LOCAL) ? NOBUF : 0;
	nhead.nh_msg = (char *) 0;
	nhead.nh_length = 0;

	if (nsend(&nhead)) {
		return((char *) 0);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();
	
	if (nrecv(&nhead)) {
		return((char *) 0);
	}

	(void) ksigsetmask(mask);

	if (reply->fl_reply) {
		errno = reply->fl_reply;
		return((char *) 0);
	}

	else {
		return(reply->fl_address);
	}
}
