/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	DLF/GDB
 *
 *	$Log:	rload.c,v $
 * Revision 2.2  92/04/30  14:53:16  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:38:40  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:30:06  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/29  16:11:35  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0  90/02/20  11:56:15  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:13:42  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:48:33  trillium
 * Initial revision
 * 
 *
 *	Function:	- run-time call for loading files into any node
 *
 *	Accepts:	- source node of file 
 *			- destination node 
 *			- flags
 *			- filename to load
 */


#include <errno.h>
#include <events.h>		/* loader events */
#include <ksignal.h>
#include <net.h>		/* network layer structures */
#include <typical.h>		/* commonly used constants */
#include <loadreq.h>		/* loadd structures */
#include <terror.h>


/*
 * external variables
 */
extern int	errno;		/* global errors */

/*
 * external functions
 */
extern int	strlen();


int
rload(src_node, dest_node, tag, filename)

int	src_node,	/* node where original copy of file reside */
	dest_node,	/* node where new copy will be placed      */
	tag;		/* tag to associate with loaded file */
char	*filename;

{
	struct lreply		*reply;		/* load reply     */
	struct lreq		*req;		/* load request   */
	struct nmsg		nhead;		/* network header */
	int			mask;
	
/*
 * Check filename length.
 */
	nhead.nh_length = strlen(filename) + 1;

	if (nhead.nh_length > FNAMELEN) {
		errno = ENAMETOOLONG;
		return(ERROR);
	}

	req = (struct lreq *) nhead.nh_data;
	reply = (struct lreply *) nhead.nh_data;
/*
 * Complete request information structure.
 */
	req->ld_req = LD_LOAD;
	req->ld_src_node = (src_node == LOCAL) ? src_node : getnodeid();
	req->ld_src_event = -getpid();
	req->ld_dest_node = dest_node;
	req->ld_tag = tag;
/*
 * Set up network header for sending request.
 */
	if ((src_node >= HOST2ALL) && (src_node <= HOST2OTB)) {
		errno = EINVAL;
		return(ERROR);
	}

	nhead.nh_node = src_node;
	nhead.nh_event = EVLOADD;
	nhead.nh_type = 0;	  
	nhead.nh_flags = 0;
	nhead.nh_msg = filename;
/*
 * Send loader request.
 */
	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();
/*
 * Receive reply.
 */
	if (nrecv(&nhead)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);
/*
 * Set local errno with returned errno value.
 */
	if (reply->ld_reply) {
		errno  = reply->ld_reply;
		return(ERROR);
	}

	else {
		return(0);
	}
}
