/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	rmem.c,v $
 * Revision 2.2  92/04/30  14:53:21  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:38:43  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:29:47  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/29  16:11:43  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0  90/02/20  11:55:58  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:13:52  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:48:36  trillium
 * Initial revision
 * 
 *
 *	Function:	- itb memory read/write runtime interface
 *			- reads/writes a given buffer from/to memory
 *			- assumes that target memory is allocated
 *			  when writing
 *	
 *	Accepts:	- pointer to a buffer
 *			- length of source buffer
 *			- destination address
 *			- destination node
 *			- a READ/WRITE switch
 */


#include <net.h>
#include <memreq.h>
#include <events.h>
#include <ksignal.h>
#include <terror.h>
#include <typical.h>


/*
 * external variables
 */
extern int			errno;


int
rmem(node, src_address, dest_address, length, rw_switch)

int				node;		/* destination node */
char				*src_address,	/* buffer to be written */
				*dest_address;	/* target address on node */
int				length,		/* length of buffer */
				rw_switch;	/* read/write switch */

{
	struct nmsg		nhead;		/* network message desc. */
	struct memreq		*request;	/* memd request */
	char			*laddr,		/* local address */
				*raddr;		/* remote address */
	int			mask;


	if (rw_switch == MWRITE) {
		laddr = src_address;
		raddr = dest_address;
	}

	else if (rw_switch == MREAD) {
		laddr = dest_address;
		raddr = src_address;
	}

	else {
		errno = EBADREQUEST;
		return(ERROR);
	}

	request = (struct memreq *) nhead.nh_data;

	request->mem_address = raddr;
	request->mem_length = length;
	request->mem_event = -getpid();
	request->mem_node = getnodeid();

	nhead.nh_node = node;
	nhead.nh_event = EVMEMD;
	nhead.nh_type = rw_switch;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();
	nhead.nh_type = 0;
	nhead.nh_flags = NOBUF;
	nhead.nh_msg = laddr;

	while (length) {

		nhead.nh_length = (length > MAXNMSGLEN) ? MAXNMSGLEN : length;

		if (rw_switch == MWRITE) {

			if (nsend(&nhead)) {
				return(ERROR);
			}
		}

		else {
			if (nrecv(&nhead)) {
				return(ERROR);
			}
		}

		length -= nhead.nh_length;
		nhead.nh_msg += nhead.nh_length;
	}

	(void) ksigsetmask(mask);
	return(0);
}
