/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	rmstate.c,v $
 * Revision 2.2  92/04/30  14:53:25  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.2  91/11/10  23:35:30  gdburns
 * Minor overhaul fixes several problems.
 * 
 * Revision 2.1.1.1  91/10/09  18:22:10  gdburns
 * Substantially rewrite to remove faulty design.
 * 
 * Revision 2.1  91/03/20  11:38:46  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:30:08  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/29  16:11:46  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0  90/02/20  11:56:16  trillium
 * Homo erectus
 * 
 *	Function:	- gathers state from a remote mallocd
 *
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell
 *			  Research Foundation
 *
 *	Accepts:	- target node ID
 *			- ptr to memory block descriptor table
 *			- # of entries in caller's table
 *
 *	Returns:	- # of descriptors in mallocd
 */

#include <events.h>
#include <ksignal.h>
#include <mareq.h>
#include <net.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>

/*
 * external variables
 */
extern int			errno;		/* error number */

int4
rmstate(nodeid, tbldesc, maxblks)

int4				nodeid;		/* target node ID */
int4				maxblks;	/* size of descriptor table */
struct memblock			tbldesc[];	/* ptr to descriptr table */

{
	struct nmsg		nhead;		/* network message desc. */
	struct mareq  	 	*request;	/* mallocd request */
	struct mareply   	*reply;		/* mallocd reply */
	int4			mask;		/* signal mask */
	int4			nblks;		/* #entries in this message */

	request = (struct mareq *) nhead.nh_data;
	reply = (struct mareply *) nhead.nh_data;

	request->maq_src_node = (nodeid == LOCAL) ? nodeid : getnodeid();
	request->maq_src_event = -getpid();
	request->maq_request = MAQSTATE;
	request->maq_maxblks = maxblks;

	nhead.nh_node = nodeid;
	nhead.nh_event = EVMALLOCD;
	nhead.nh_type = 0;
	nhead.nh_flags = 0;
	nhead.nh_length = 0;
	nhead.nh_msg = (char *) 0;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();
	nhead.nh_msg = (char *) tbldesc;
	nhead.nh_type = 0;
	nblks = (MAMAX > maxblks) ? maxblks : MAMAX;
	nhead.nh_length = nblks * sizeof(struct memblock);

	if (nrecv(&nhead)) {
		return(ERROR);
	}

	if (reply->mar_reply) {
		errno = reply->mar_reply;
		return(ERROR);
	}

	if (maxblks > reply->mar_nblks) {
		maxblks = reply->mar_nblks;

		if (nblks > reply->mar_nblks) {
			nblks = reply->mar_nblks;
		}
	}

	maxblks -= nblks;
	nhead.nh_msg += nhead.nh_length;

	while ((maxblks > 0) && ! reply->mar_reply) {
		nblks = (MAMAX > maxblks) ? maxblks : MAMAX;
		nhead.nh_length = nblks * sizeof(struct memblock);

		if (nrecv(&nhead)) {
			return(ERROR);
		}

		maxblks -= nblks;
		nhead.nh_msg += nhead.nh_length;
	}

	(void) ksigsetmask(mask);

	if (reply->mar_reply) {
		errno = reply->mar_reply;
		return(ERROR);
	} else {
		return(reply->mar_nblks);
	}
}
