/*
 *	OSU Research Computing, architects and contractors	Trillium
 *
 *	Copyright 1990 The Ohio State University & Cornell University
 *
 *	GDB
 *
 *	$Log:	rpdetach.c,v $
 * Revision 2.2  92/04/30  14:53:26  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:38:48  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:30:14  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/29  16:11:52  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0  90/02/20  11:56:19  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  02:00:45  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  16:21:21  trillium
 * Initial revision
 * 
 *
 *	Function:	- detach caller from the local kenya server
 *
 *	Accepts:	- destination node
 *			- target process ID
 *			- exit status
 *
 *	Returns:	- 0 or ERROR
 */


#include <net.h>
#include <events.h>
#include <ksignal.h>
#include <typical.h>
#include <preq.h>
#include <kio.h>
#include <t_types.h>

/*
 * external variables
 */
extern int			errno;		/* global error variable */
extern struct kio_t		_kio;		/* Kernel Input/Output */


int
rpdetach(node, pid, status)

int				node;
int				pid;
int				status;

{
	struct nmsg		header;		/* network header */
	struct preq  	 	request;	/* kenya request  */
	struct preply		reply;		/* kenya reply    */
	int			mask;


	request.pq_src_node = ltot((node == LOCAL) ? LOCAL : getnodeid());
	request.pq_src_event = ltot(-getpid());
	request.pq_pid = ltot(pid ? pid : getpid());
	request.pq_status = ltot(status);
	request.pq_req = ltot(PDETACH);

	header.nh_node = node;
	header.nh_event = EVKENYAD;
	header.nh_type = 0;
	header.nh_flags = NOBUF;
	header.nh_length = sizeof(request);
	header.nh_msg = (char *) &request;

	if (nsend(&header)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	header.nh_event = -getpid();
	header.nh_length = sizeof(reply);
	header.nh_msg = (char *) &reply;

	if (nrecv(&header)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);
	_kio.ki_rtf &= ~RTF_KENYA;

	if (reply.pr_reply) {
		errno = ttol(reply.pr_reply);
		return(ERROR);
	}

	else {
		return(0);
	}
}
