/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	rspawn.c,v $
 * Revision 2.2  92/04/30  14:53:30  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:38:50  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:30:04  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/29  16:11:57  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0  90/02/20  11:56:13  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:13:58  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:48:38  trillium
 * Initial revision
 * 
 *
 *	Function:	- spawn processes on a given node to a given node
 *			- spawns an existing kenya process on the source node
 *			- assumes argv structure already in memory on
 *			  destination node
 *
 *	Accepts:	- source node ID
 *			- process select flags
 *			- process select value
 *			- destination node ID
 *			- argc
 *			- argv
 *			- ptr to process ID (value returned)
 *
 *	Returns:	- 0 or ERROR
 */


#include <net.h>
#include <events.h>
#include <ksignal.h>
#include <typical.h>
#include <t_types.h>
#include <preq.h>
#include <kio.h>


/*
 * external variables
 */
extern int			errno;		/* global error variable */
extern struct kio_t		_kio;


int
rspawn(src_nodeid, sflags, svalue, dest_nodeid, argc, argv_tag, pid)

int				src_nodeid,	/* source node ID */
				sflags,		/* process select flags */
				svalue,		/* process select value */
				dest_nodeid,	/* destination node ID */
				argc,		/* number of runtime args */
				argv_tag,	/* runtime args tag */
				*pid;		/* newly created pid */

{
	struct nmsg		nhead;		/* network header */
	struct preq  	 	request;	/* kenya request */
	struct preply		reply;		/* kenya reply */
	int			mask;


	request.pq_sel_method = ltot(sflags);
	request.pq_sel_value = ltot(svalue);
	request.pq_argc = ltot(argc);
	request.pq_argv = ltot(argv_tag);
	request.pq_nodeid = ltot(dest_nodeid);
	request.pq_stdin = ltot(_kio.ki_stdin);
	request.pq_stdout = ltot(_kio.ki_stdout);
	request.pq_stderr = ltot(_kio.ki_stderr);
	request.pq_src_node = ltot(getnodeid());
	request.pq_src_event = ltot(-getpid());
	request.pq_req = ltot(PSPAWN);

	nhead.nh_node = src_nodeid;
	nhead.nh_event = EVKENYAD;
	nhead.nh_flags = 0;
	nhead.nh_type = 0;
	nhead.nh_length = sizeof(request);
	nhead.nh_msg = (char *) &request;

	if (nsend(&nhead)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	nhead.nh_event = -getpid();
	nhead.nh_length = sizeof(reply);
	nhead.nh_msg = (char *) &reply;

	if (nrecv(&nhead)) {
		return(ERROR);
	}

	(void) ksigsetmask(mask);

	if (reply.pr_reply) {
		errno = ttol(reply.pr_reply);
		return(ERROR);
	}

	else {
		*pid = ttol(reply.pr_pid);
		return(ttol(reply.pr_index));
	}
}
