/*
 *	Cornell Theory Center					Trillium
 *
 *	Copyright 1987 Cornell University
 *
 *	GDB
 *
 *	$Log:	rstate.c,v $
 * Revision 2.2  92/04/30  14:53:32  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  11:38:51  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.2  90/07/13  19:29:59  gdburns
 * Homo sapien
 * 
 * Revision 2.0.1.1  90/06/29  16:11:59  gdburns
 * Corrected brain damaged ksigblock().
 * 
 * Revision 2.0  90/02/20  11:56:10  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:14:03  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  15:48:40  trillium
 * Initial revision
 * 
 *
 *	Function:	- find the state of a process on a given node
 *
 *	Accepts:	- target node ID
 *			- process select flags
 *			- process select value
 *			- ptr to process state array
 *			- size of process state array
 *
 *	Returns:	- number of replied states or ERROR
 */


#include <net.h>
#include <events.h>
#include <ksignal.h>
#include <typical.h>
#include <t_types.h>
#include <terror.h>
#include <preq.h>


/*
 * external variables
 */
extern int			errno;		/* global error variable */


int
rstate(nodeid, sflags, svalue, reply_state, maxsize)

int				nodeid,		/* target node address */
				sflags,		/* process select flags */
				svalue,		/* process select value */
				maxsize;	/* size of state array */
struct pstate			reply_state[];	/* process state */

{
	struct nmsg		header;		/* network header */
	struct preq  	 	request;	/* kenya request  */
	struct pstate		black_hole;	/* space to lose replies */
	int			reply_code,	/* reply error code */
				num_replies,	/* number of replied states */
				i,		/* favourite index variable */
				*pint;		/* integer pointer for
						   reversing byte ordering */
	int			mask;


	num_replies = reply_code = 0;

	request.pq_sel_method = ltot(sflags);
	request.pq_sel_value = ltot(svalue);
	request.pq_req = ltot(PSTATE);
	request.pq_src_node = ltot(getnodeid());
	request.pq_src_event = ltot(-getpid());

	header.nh_node = nodeid;
	header.nh_event = EVKENYAD;
	header.nh_type = 0;
	header.nh_flags = 0;
	header.nh_length = sizeof(request);
	header.nh_msg = (char *) &request;

	if (nsend(&header)) {
		return(ERROR);
	}

	mask = ksigblock(sigmask(SIGUDIE) | sigmask(SIGARREST));
	header.nh_event = -getpid();
	header.nh_length = sizeof(struct pstate);

	while (maxsize && !reply_code) {

		header.nh_msg = (char *) reply_state;

		if (nrecv(&header)) {
			return(ERROR);
		}
/*
 * Convert byte ordering (sigh).
 */
		pint = (int *) reply_state;

		for (i = 0; i < (PSTATESIZE / sizeof (int)); ++i) {
			*pint = ltot(*pint);
			++pint;
		}

		reply_code = reply_state->ps_reply;
		--maxsize;
		++num_replies;
		++reply_state;
	}

	header.nh_msg = (char *) &black_hole;

	while (!reply_code) {
/*
 * We ran out of process state entries so suck back
 * any outstanding replies from kenyad.
 */
		if (nrecv(&header)) {
			return(ERROR);
		}

		reply_code = ttol(black_hole.ps_reply);
		++num_replies;
	}

	(void) ksigsetmask(mask);

	if (!reply_code || (reply_code == EEOF)) {
		return(--num_replies);
	}

	else {
		errno = reply_code;
		return(ERROR);
	}
}
