/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	export.c,v $
 * Revision 2.2  92/04/30  14:54:59  trillium
 * Ohio Release
 * 
 *	Function:	- otb to itb bootstrap downloader
 *			- Transputer boot-from-link version
 *			- loads and boots up to 255 bytes of code
 */

#include <boot.h>
#include <fcntl.h>
#include <stdio.h>
#include <t.out.h>
#include <unistd.h>

/*
 * external functions
 */
extern char		*findpath();
extern char		*getparam();
extern char		*malloc();
extern int		do_args();
extern int		read();
extern int		CloseLink();	/* close device */
extern int		OpenLink();	/* open device */
extern int		WriteLink();	/* write to device */
extern void		addarg();	/* add arg to arg list */
extern void		followed();
extern void		validopts();

/*
 * local functions
 */
static void		setpaths();	/* initialize search paths */
static void		help();

/*
 * local variables
 */
static char		*me;		/* name I was exec'd by */
static char		**p_prog;	/* bootstrap search paths */
static int		n_prog;		/* count of p_prog */

static char		*usage = "usage: %s <bootstrap> <device>\n";

/*
 * external variables
 */
extern int		errno;


main(argc, argv)

int			argc;
char			*argv[];

{
	struct exec	tout;		/* executable info header */
	char		len;		/* text length */
	char		*ptext;		/* ptr to text */
	char		*ifname;	/* input filename */
	char		*ofname;	/* output filename */
	char		*ifullname;	/* input full pathname */
	char		*p;		/* favourite ptr */
	int		ifd;		/* input file descriptor */
	int		ofd;		/* output file descriptor */

	me = argv[0];
/*
 * Initialize option parser.
 */
	validopts("hiov");
	followed("io");

	if (do_args(&argc, argv) < 0) {
		(void) fprintf(stderr, usage, me);
		exit(errno);
	}

	if (opt_taken('h')) {
		help();
		exit(0);
	}

	--argc;
	++argv;
	setpaths();
	ifname = NULL;
	ofname = NULL;
/*
 * Check for a single file argument with no decidability between input
 * and output.
 */
	if (!opt_taken('o') && !opt_taken('i') && (argc == 1)) {
		(void) fprintf(stderr, usage, me);
		exit(1);
	}
/*
 * Establish input file name.
 */
	if (opt_taken('i')) {
		ifname = getparam('i');
	} else if (argc > 0) {
		ifname = argv[0];
		--argc;
		++argv;
	}
/*
 * Establish output file name.
 */
	if (opt_taken('o')) {
		ofname = getparam('o');
	} else if (argc > 0) {
		ofname = argv[0];
		--argc;
		++argv;
	}

	if (argc > 0) {
		(void) fprintf(stderr, usage, me);
		exit(1);
	}
/*
 * Establish input file descriptor.
 */
	if (ifname == NULL) {
		ifd = 0;
		ifname = "stdin";
	} else {
	    	if ((ifullname = findpath(ifname, p_prog, R_OK)) == NULL) {
			fprintf(stderr, "%s: can't find %s\n", me, ifname);
			exit(1);
		}

		ifname = ifullname;
		ifd = open(ifname, O_RDONLY, 0);

		if (ifd < 0) {
			perror(ifname);
			exit(errno);
		}
	}

	if (opt_taken('v')) fprintf(stderr, "%s: input is %s\n", me, ifname);
/*
 * Read in the executable info structure.
 */
	if (read(ifd, (char *) &tout, sizeof(tout)) < sizeof(tout)) {
		perror(me);
		exit(errno);
	}
/*
 * Verify magic number of a valid executable file.
 */
	if ((tout.a_magic != TO_MAGIC_TN_X) &&
		(tout.a_magic != TO_MAGIC_T4_X) &&
		(tout.a_magic != TO_MAGIC_T8_X)) {

		(void) fprintf(stderr, "%s: %s: bad magic number\n", me,
			ifname);
		exit(1);
	}
/*
 * Check length of text segment.
 */
	if (tout.a_text > 255) {
		(void) fprintf(stderr,
			"%s: warning: text segment too long (%d > 255 bytes)\n",
			me, tout.a_text);
		tout.a_text = 255;
	}

	ptext = malloc((unsigned) tout.a_text);

	if (! ptext) {
		perror(me);
		exit(errno);
	}
/*
 * Read in the text (code) only.
 */
	if (read(ifd, ptext, (int) tout.a_text) < (int) tout.a_text) {
		perror(me);
		exit(errno);
	}

	len = (char) (tout.a_text & 0xff);
	if (opt_taken('v')) fprintf(stderr, "%s: %d bytes of text\n", me,
		tout.a_text);
/*
 * Establish output file descriptor.
 */
	if (ofname == NULL) {
		ofd = 1;
		ofname = "stdout";
	} else {

#ifdef PARSY
/*
 * Replace '/' with ' ' in Parsytec link name.
 * Thanks to the person who decided to embed blanks in a single identifier.
 */
		for (p = ofname; *p; ++p) if (*p == '/') *p = ' ';
#endif
/*
 * Open the device.
 */
		ofd = OpenLink(ofname);

		if (ofd < 0) {
			(void) fprintf(stderr, "%s: cannot open %s\n",
				me, ofname);
		}
	}

	if (opt_taken('v')) fprintf(stderr, "%s: output is %s\n", me, ofname);
/*
 * Load and boot text into reset boot-from-link transputer.
 */
	if (WriteLink(ofd, &len, 1, 0) < 0) {
		(void) fprintf(stderr, "%s: cannot write device\n", me);
		exit(1);
	}

	if (WriteLink(ofd, ptext, (int) tout.a_text, 0) < (int) tout.a_text) {
		(void) fprintf(stderr, "%s: cannot write device\n", me);
		exit(1);
	}

	(void) close(ifd);
	(void) CloseLink(ofd);

	return(0);
}

/*
 *	setpaths
 *
 *	Function:	- initializes search paths for input and output files
 */

static void
setpaths()

{
/*
 * input file
 */
	n_prog = 0;
	p_prog = (char **) 0;
	addarg(&n_prog, &p_prog, "");
	addarg(&n_prog, &p_prog, "$TROLLIUSHOME");
	addarg(&n_prog, &p_prog, DEFPPROGS);
}
 
/*
 *        help
 *
 *        Function:     - prints helpful information on the export command
 */
 
static void
help()
{
    printf("\nSynopsis:\texport [options] <bootstrap> <device>\n");
    printf("\nDescription:\tLoad and execute bootstrap on neighbour node.\n");
    printf("\nOptions:        -h\t\tPrint this message.\n");
    printf("\t\t-v\t\tBe verbose.\n");
    printf("\t\t-i <bootstrap>\tSpecify bootstrap input file.\n");
    printf("\t\t-o <device>\tSpecify device output.\n");
    printf("\nExample:        export /dev/link0 moses.t4\n");
    printf("\t\t\t\"Bootstrap `moses.t4' over `/dev/link0'.\"\n\n");
}
