/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	GDB
 *
 *	$Log:	hboot.c,v $
 * Revision 2.2  92/04/30  14:55:29  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  92/04/27  15:56:00  gdburns
 * major rewrite
 * 
 *	Function:	- boots OTB operating system
 */

#include <boot.h>
#include <conf.h>
#include <debug.h>
#include <stdio.h>
#include <terror.h>
#include <unistd.h>

/*
 * external variables
 */
extern int		errno;
extern char		**exttopv;	/* -T will override conf topology */
extern int		exttopc;

/*
 * external functions
 */
extern void		split();
extern void		cnfparse();
extern void		followed();
extern void		validopts();
extern int		addarg();
extern int		do_args();
extern char		*findpath();
extern char		*getparam();
extern char		*rindex();
extern CONF		*cnffind();

/*
 * local functions
 */
static void		help();
static void		setdefaults();

/*
 * local variables
 */
static char		*f_conf;		/* process schema file */
static char		**p_prog;		/* program pathname prefixes */
static char		**p_conf;		/* conf pathname prefixes */
static char		*me;			/* who am I? */

static int		fl_debug;		/* debugging option */
static int		fl_verbose;		/* verbose option */
static int		n_prog;			/* # of p_prog prefixes */
static int		n_conf;			/* # of p_conf prefixes */

static char
	*usage = " [-hdvNcV] [-c <conf>] [-T <topology>]";

main(argc, argv)

int			argc;
char			*argv[];

{
	FILE		*fp_conf;	/* process schema file ptr */
	CONF		*conf;		/* parsed process schema list */
	CONF		*p;		/* favourite ptr */
	char		*full;		/* full pathname */
	char		*tail;		/* tail of full pathname */
	int4		i;		/* favourite index */
	int		n;		/* favourite counter */
	int		pid;		/* child PID */

	me = argv[0];
	conf = CNFNULL;
/*
 * Initialize option parser.
 */
	validopts("hdvNV");
	followed("cT");

	if ((do_args(&argc, argv) < 0) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		fprintf(stderr, "usage: %s%s\n", me, usage);
		exit(errno);
	}

	if (opt_taken('h')) {
		help();
		exit(0);
	}

	setdefaults();
/*
 * Locate the conf file.
 */
	full = findpath(f_conf, p_conf, R_OK);

	if (! full) {
		fprintf(stderr, "%s: can't find conf file - %s\n", me, f_conf);
		exit(1);
	} else {
		f_conf = full;
	}
/*
 * Dump debugging information.
 */
	DBUG("%s: f_conf = \"%s\"\n", me, f_conf);
/*
 * Open the conf file.
 */
	fp_conf = fopen(f_conf, "r");

	if (! fp_conf) {
		fprintf(stderr, "%s: can't open conf file - %s\n", me, f_conf);
		perror(me);
		exit(errno);
	}
/*
 * Allow -T to override conf file topology variable.
 */
	if (opt_taken('T')) {
		exttopv = (char **) 0;
		exttopc = 0;
		split(getparam('T'), &exttopv, &exttopc);
	}
/*
 * Parse the conf file.
 */
	cnfparse(fp_conf, &conf);

	if (opt_taken('V')) {
		cnfprint(conf);
		exit(0);
	}

	if (! conf) {
		fprintf(stderr, "%s: nothing to do in %s\n", me, f_conf);
		exit(0);
	}
/*
 * Find programs with sufficient permissions.
 */
	if ((p = cnffind(p_prog, conf)) != CNFNULL) {
		fprintf(stderr, "%s: can't find %s with permissions\n",
			me, p->cnf_argv[0]);
		exit(1);
	}

	if (fl_debug) {

		for (p = conf; p; p = p->cnf_next) {
			printf("%s: found %s\n", me, p->cnf_argv[0]);
		}
	}
/*
 * Bail out here, if pretending.
 */
	if (opt_taken('N')) {
		exit(0);
	}

	VERBOSE("boot ...\n");

	n = 0;
	setpgrp(0, 0);
/*
 * Loop through all the programs in the parsed config file.
 */
	while (conf) {
		DBUG("%s: fork %s\n", me, conf->cnf_argv[0]);

		if ((pid = fork()) < 0) {
			perror(me);
			exit(errno);
		}

		else if (pid == 0) {		/* child */
	    		execvp(conf->cnf_argv[0], conf->cnf_argv);
	    		exit(1);
		}

		else {				/* parent */
	    		n++;

			if (fl_verbose) {
				tail = rindex(conf->cnf_argv[0], '/');
				tail = (tail) ? tail + 1 : conf->cnf_argv[0];
				printf("[%d] %5d %s", n, pid, tail);

				for (i = 1; i < conf->cnf_argc; i++) {
		    			printf(" %s", conf->cnf_argv[i]);
				}

				printf("\n");
			}
		}

		if (conf->cnf_delay > 0) {
			sleep((unsigned int) conf->cnf_delay);
		}

		conf = conf->cnf_next;
	}

	return(0);
}

/*
 *	setdefaults
 *
 *	Function:	- sets default files and paths
 */

static void
setdefaults()

{
/*
 * prefix paths
 */
	n_conf = 0;
	p_conf = (char **) 0;
	addarg(&n_conf, &p_conf, "");
	addarg(&n_conf, &p_conf, "$TROLLIUSHOME/boot");
	addarg(&n_conf, &p_conf, DEFPCONFIG);

	n_prog = 0;
	p_prog = (char **) 0;
	addarg(&n_prog, &p_prog, "");
	addarg(&n_prog, &p_prog, "$TROLLIUSHOME/bin");
	addarg(&n_prog, &p_prog, DEFPPROGH);
/*
 * flags
 */
	fl_debug = opt_taken('d');;
	fl_verbose = opt_taken('v');
/*
 * files
 */
	if (opt_taken('c')) {
		f_conf = getparam('c');
	} else {
		f_conf = DEFFCONFIGH;
	}
}

/*
 *	help
 *
 *	Function:	- prints helpful information on the hboot command
 */

static void
help()

{
    printf("\nSynopsis:	hboot [options]\n");
    printf("\nDescription:	Boot an OTB node.\n");
    printf("\nOptions:	-h		Print this message.\n");
    printf("\t\t-d		Print debug information.\n");
    printf("\t\t-v		Be verbose.\n");
    printf("\t\t-N		Pretend to hboot.\n");
    printf("\t\t-b		Skip the boot step.\n");
    printf("\t\t-V		Pretty print the conf file.\n");
    printf("\t\t-c <conf>	Use <conf> as the process schema.\n");
    printf("\t\t-T <topo>	Set $topology variable.\n\n");
}
