/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD
 *
 *	$Log:	map.h,v $
 * Revision 2.2  92/04/30  14:56:36  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/12/04  17:16:50  raja
 * Added link weight option to route2 algorithm.
 * 
 * Revision 2.1  91/11/26  18:09:39  raja
 * initial revision
 * 
 *	Function:	- templates and constants used by the map tool
 */

#ifndef _MAP
#define _MAP

#include <portable.h>
#include <all_hash.h>
#include <all_list.h>
#include <rreq.h>
#include <nail.h>

/* 
 * structure definitions
 */
struct cmstats {				/* routing graph statistics */
	int4			cs_nlink;	/* number of links */
	int4			cs_r1minlw;	/* minimum link weight */
	int4			cs_r1maxlw;	/* maximum link weight */
	int4			cs_r1minlu;	/* minimum link usage */
	int4			cs_r1maxlu;	/* maximum link usage */
	float			cs_r1avelw;	/* average link weight */
	float			cs_r1avelu;	/* average link usage */
	double			cs_r1stdlw;	/* std. dev. link weight */
	double			cs_r1stdlu;	/* std. dev. link usage */
	int4			cs_r2minbh;	/* minimum broadcast hops */
	int4			cs_r2maxbh;	/* maximum broadcast hops */
	float			cs_r2avebh;	/* average broadcast hops */
	double			cs_r2stdbh;	/* std. dev. broadcast hops */
};

struct cmdesc {					/* connectivity matrix desc. */
	int4			cd_nnodes;	/* number of nodes */
	HASH			*cd_index;	/* hash table index */
	struct cmelem		**cd_table;	/* 2-D array of elements */
	LIST			*cd_cleanup;	/* lists to cleanup */
	struct cmstats		cd_stats;	/* routing statistics */
};

struct cmindex {				/* hash table index */
	int4			ci_nodeid;	/* node identifier, hash key */
	int4			ci_offset;	/* route array offset */
};

struct cmelem {					/* matrix element */
	int4			ce_done;	/* path has been done */
	int4			ce_nodetype;	/* destination nodetype */
	int4			ce_event;	/* route 1 event */
	int4			ce_event2;	/* route 2 event */
	LIST			*ce_cost;	/* link/cost list */
};

struct cmlink {					/* link/cost element */
	int4			cl_linkno;	/* link number */
	int4			cl_cost;	/* accumulated cost */
	int4			cl_weight;	/* link weight */
	int4			cl_use;		/* number of times used */
};

struct navelem {				/* navigation table element */
	int4			ne_nodeid;	/* node identifier, hash key */
	struct t_node		*ne_tnode;	/* ptr to node's TNODE tuple */
	int4			ne_mincost;	/* minimum path cost */
	struct navelem		*ne_prevnode;	/* previous node */
	struct cmlink		*ne_prevlink;	/* link from previous node */
};

struct stelem {					/* spanning tree element */
	int4			st_nodeid;	/* node identifier, hash key */
	int4			st_mindist;	/* minimum path distance */
	int4			st_children;	/* number of children nodes */
	struct t_node		*st_tnode;	/* ptr to node's TNODE tuple */
	struct stelem		*st_upnode;	/* ptr parent node */
	struct t_link		*st_uplink;	/* ptr parent node's TLINK */
};

/*
 * function declarations
 */
extern struct cmdesc	*cm_init();
extern struct cmelem	*cm_find();
extern void		cm_free();

#endif
