/*
 *	OSU Research Computing					Trillium
 *	Copyright 1991 The Ohio State University
 *	RBD
 *
 *	$Log:	map_funct.c,v $
 * Revision 2.2  92/04/30  14:56:38  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.1  91/11/26  18:03:49  raja
 * Big overhaul.
 * 
 *	Function :	- miscellaneous functions used by the map tool
 */

#include <stdio.h>
#include <malloc.h>
#include <portable.h>
#include <typical.h>
#include <net.h>
#include <events.h>
#include <rreq.h>
#include <nail.h>
#include <routeio.h>
#include <debug.h>
#include <all_hash.h>
#include <all_list.h>
#include <map.h>

/*
 * external variables
 */
extern struct bschema	nail;		/* network description */
extern char		*me;		/* tool name */
extern int		fl_debug;	/* debug mode flag */
extern int		fl_verbose; 	/* verbose mode flag */

/*
 * external functions
 */
extern void		exit();


/*
 *	cm_init
 *
 *	Function:	- allocate and initialize the connectivity matrix
 *
 *	Accepts:	- ptr to the network information
 *			- extra number of route table entries
 *
 *	Returns:	- ptr to the matrix descriptor or NULL
 */
struct cmdesc *
cm_init(net, extra)

struct bschema		*net;
int4			extra;

{
	int4		prime;			/* size of hash tables */
	int4		nnodes;			/* number of nodes */
	int4		i, j;			/* favorite counters */
	struct t_node	*tnode;			/* ptr to TNODE entry */
	struct cmelem	**p;			/* favorite pointer */
	struct cmelem	*pp;			/* another pointer */
	struct cmdesc	*cm;			/* matrix descriptor */
	struct cmindex	index_elem;		/* matrix index element */

	nnodes = ah_count(net->bs_tnhd);
	if (nnodes < (int4) 1) {
		fprintf(stderr, "%s (cm_init): network has %d nodes\n", nnodes);
		return((struct cmdesc *) 0);
	}
/*
 * Allocate the descriptor.
 */
	cm = (struct cmdesc *) malloc(sizeof(struct cmdesc));
	if (cm == (struct cmdesc *) 0) {
		fprintf(stderr, "%s (cm_init): cannot malloc descriptor\n", me);
		return((struct cmdesc *) 0);
	}

	cm->cd_nnodes = nnodes;
	cm->cd_index = (HASH *) 0;
	cm->cd_table = (struct cmelem **) 0;
	cm->cd_cleanup = (LIST *) 0;
/*
 * Determine the size of the hash tables.  In order to minimize hashing
 * collisions, select a prime number to be the size.
 */
	prime = next_prime( (((nnodes + extra) * 12) / 10) + 1 );
	DBUG("%s: nnodes %d\textra %d\tprime %d\n", me, nnodes, extra, prime);
/*
 * Allocate the index hash table that is used in accessing the
 * connectivity matrix.  It returns the offset in the 2-D array
 * to be used for a given nodeid.
 */ 
	if ((cm->cd_index = ah_init(prime, (int4) sizeof(struct cmindex),
		NOTNODEID, AHFIXED)) == (HASH *) 0) {

		fprintf(stderr, "%s (cm_init): cannot malloc index\n", me);
		cm_free(cm);
		return((struct cmdesc *) 0);
	}
/*
 * Allocate the 2-D array of route table entries.  It is constructed
 * as N separately allocated arrays of N entries each.  These arrays are
 * accessed using an array of N pointers, each pointing to one of these
 * arrays.  This setup resembles the C argv[] structure and has the advantage
 * of breaking the array into smaller chunks to be malloc'ed.
 */
	if ((cm->cd_table = (struct cmelem **) malloc((uint4) nnodes *
		sizeof(struct cmelem *))) == (struct cmelem **) 0) {

		fprintf(stderr, "%s (cm_init): cannot malloc head table\n", me);
		cm_free(cm);
		return((struct cmdesc *) 0);
	}

	for (i = 0, p = cm->cd_table; i < nnodes; i++, p++) {
		*p = (struct cmelem *) 0;
	}

	for (i = 0, p = cm->cd_table; i < nnodes; i++, p++) {
		if ((*p = (struct cmelem *) malloc((uint4) nnodes *
			sizeof(struct cmelem))) == (struct cmelem *) 0) {

			fprintf(stderr,
				"%s (cm_init): cannot malloc matrix row %d\n",
				me, i);
			cm_free(cm);
			return((struct cmdesc *) 0);
		} 
		else {
/*
 * Initialize the array elements.
 */
			for (j = 0, pp = *p; j < nnodes; j++, pp++) {
				pp->ce_done = FALSE;
				pp->ce_nodetype = NOTNODETYPE;
				pp->ce_event = NOTEVENT;
				pp->ce_event2 = NOTEVENT;
				pp->ce_cost = (LIST *) 0;
			}
		}
	}
/*
 * Read all the nodes in the network and hash them into the index.
 */
	index_elem.ci_offset = 0;
	tnode = (struct t_node *) ah_top(net->bs_tnhd);
	while (tnode) {
		index_elem.ci_nodeid = tnode->tn_nodeid;
		if (ah_insert(cm->cd_index, &index_elem)) {
			fprintf(stderr,
				"%s (cm_init): cannot hash node %d in index\n",
				me, tnode->tn_nodeid);
			cm_free(cm);
			return((struct cmdesc *) 0);
		}

		(index_elem.ci_offset)++;
		tnode = (struct t_node *) ah_next(net->bs_tnhd, (void *) tnode);
	}
/*
 * Allocate the cleanup list.  It is a linked list of all linked lists
 * in the 2-D array.  This allows for fast cleanup at the end.
 */
	cm->cd_cleanup = (LIST *) al_init(sizeof(LIST *), NULL);
	if (cm->cd_cleanup == (LIST *) 0) {
		fprintf(stderr, "%s (cm_init): cannot malloc cleanup list\n");
		cm_free(cm);
		return((struct cmdesc *) 0);
	} 

	return(cm);
}

/*
 *	cm_free
 *
 *	Function:	- free the given connectivity matrix
 *
 *	Accepts:	- ptr to matrix descriptor
 */
void
cm_free(cm)

struct cmdesc		*cm;

{
	int4		i;		/* favorite counter */
	struct cmelem	**p;		/* favorite pointer */
	LIST		**pp;		/* another pointer */

	if (cm == (struct cmdesc *) 0) {
		return;
	}
/*
 * Free the index hash table.
 */
	if (cm->cd_index) {
		ah_free(cm->cd_index);
	}
/*
 * Free the 2-D array of route table entries.
 */
	if (p = cm->cd_table) {
		for (i = 0; i < cm->cd_nnodes; i++, p++) {
			if (*p) {
				free((char *) *p);
			}
		}

		free((char *) cm->cd_table);
	}
/*
 * Free all the lists linked in the cleanup list.
 */
	if (cm->cd_cleanup) {
		pp = (LIST **) al_top(cm->cd_cleanup);
		while (pp) {
			(void) al_free(*pp);
			pp = (LIST **) al_next(cm->cd_cleanup, (char *) pp);
		}
	}

	(void) al_free(cm->cd_cleanup);

	free((char *) cm);
}

/*
 *	cm_find
 *
 *	Function:	- locate an entry in the connectivity matrix
 *			- no bound checking is done
 *
 *	Accepts:	- ptr to matrix descriptor
 *			- source nodeid
 *			- destination nodeid
 *
 *	Returns:	- ptr to matrix entry
 */
struct cmelem *
cm_find(cm, src, dest)

struct cmdesc		*cm;
int4			src;
int4			dest;

{
	struct cmindex	*p_src;		/* ptr to source index */
	struct cmindex	*p_dest;	/* ptr to destination index */
/*
 * Hash the source and destination nodeids and get the indices.
 */
	p_src = (struct cmindex *) ah_find(cm->cd_index, src);
	p_dest = (struct cmindex *) ah_find(cm->cd_index, dest);
	if ((p_src == (struct cmindex *) 0) ||
		(p_dest == (struct cmindex *) 0)) {

		return((struct cmelem *) 0);
	}

	return( (cm->cd_table)[p_src->ci_offset] + p_dest->ci_offset );
}

/*
 *	cm_linkcomp
 *
 *	Function:	- compare two link/cost list entries
 *			- used to navigate the link/cost list
 *
 *	Accepts:	- ptr to first link entry
 *			- ptr to second link entry
 *
 *	Returns:	- negative if first < second
 *			  0        if first == second
 *			  positive if first > second
 */
int4
cm_linkcomp(link1, link2)

struct cmlink		*link1,
			*link2;

{
	return(link1->cl_linkno - link2->cl_linkno);
}

/*
 *	is_link
 *
 *	Function:	- check if the given link number is defined
 *			  between source and destination nodes
 *			  (i.e. for the given connectivity matrix entry)
 *
 *	Accepts:	- ptr to matrix entry
 *			- link number
 *
 *	Returns:	- TRUE or FALSE
 */
int
is_link(cmentry, link)

struct cmelem		*cmentry;
int4			link;

{
	struct cmlink	link_entry;

	link_entry.cl_linkno = link;
	if (al_find(cmentry->ce_cost, &link_entry)) {
		return(TRUE);
	}
	else {
		return(FALSE);
	}
}

/*
 *	boot_cmp
 *
 *	Function:	- compare two nodes in the booting order
 *			  to determine which one is booted first
 *
 *	Accepts:	- first nodeid
 *			- second nodeid
 *
 *	Returns:	- negative if first node is booted first
 *			  0 if both nodes are the same
 *			  positive if second node is booted first
 */
int
boot_cmp(node1, node2)

int4			node1,
			node2;

{
	struct t_node	*p;		/* favorite pointer */

	if (node1 == node2) {
		return(0);
	}
/*
 * Loop over the nodes following the boot order.
 * If node1 is encountered first return -1.
 * If node2 is encountered first return +1.
 */
	for (p = nail.bs_hostnode; p; p = (p->tn_specs)->ns_bootnext) {
		if (p->tn_nodeid == node1) {
			return(-1);
		}
		else if (p->tn_nodeid == node2) {
			return(1);
		}
	}
/*
 * Neither node is in the system, declare this a fatal error and exit.
 */
	fprintf(stderr, "%s (boot_cmp): nodes %d and %d are not valid\n",
		me, node1, node2);
	exit(1);
}
