/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	moses.c,v $
 * Revision 2.2  92/04/30  14:59:16  trillium
 * Ohio Release
 * 
 *	Function:	- command interface to moses running on a
 *			  device connected ITB node
 *
 */

#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <portable.h>
#include <typical.h>
#include <machine/machine.h>

/*
 * global variables
 */
int			fl_debug;		/* debug flag */
int			fl_pretend;		/* pretend flag */
int			linkdesc;		/* device file descriptor */
char			*linkname;		/* device name */
char			*me;			/* program name */

/*
 * external variables
 */
extern int		errno;			/* global error variable */

/*
 * static variables
 */
static char *_copyright = "Copyright 1992 The Ohio State University $Revision: 2.2 $";
static char *usage = "usage: %s [-dhN] <device>\n";

/*
 * external functions
 */
extern void		validopts();		/* set the valid options */
extern int		OpenLink();
extern int		do_args();		/* parse the valid options */
extern int		opt_taken();		/* was option selected ? */
extern char		*strrchr();

/*
 * static functions
 */
static void		help();			/* print help menu */
static int		set_defaults();		/* set the option defaults */

int
main(argc, argv)

int			argc;
char			*argv[];

{
	char		*p;

	if ((me = strrchr(argv[0], '/')) == NULL) {
		me = argv[0];
	} else {
		me++;
	}
/*
 * Parse the command line.
 */
	validopts("dhN");

	if (do_args(&argc, argv) < 0) {
		fprintf(stderr, usage, me);
		exit(errno);
	}
/*
 * Check for the help request.
 */
	if (opt_taken('h')) {
		help();
		exit(0);
	}
/*
 * Set the default options.
 */
	if (set_defaults(argc, argv)) {
		fprintf(stderr, usage, me);
		exit(errno);
	}
#ifdef PARSY
/*
 * Replace '/' with ' ' in Parsytec link name.
 * Thanks to the person who decided to embed blanks in a single identifier.
 */
		for (p = linkname; *p; ++p) if (*p == '/') *p = ' ';
#endif
/*
 * Open the device.
 */
	linkdesc = OpenLink(linkname);

	if (linkdesc < 0) {
		fprintf("%s: can't open %s\n", me, linkname);
		exit(errno);
	}
/*
 * Call the moses command parser.  That's the heart of the tool.
 */
	printf("ms> ");
	exit(yyparse());
}

/*
 *	help
 *
 *	Function:	- print the help menu for the tool
 */
static void
help()
{
	printf("\nSynopsis:\t%s [options] [<device>]\n", me);
	printf("\nDescription:\tCommand interface to moses.\n");
	printf("\nOptions:\t-h                      this help message\n");
	printf("\t\t-d                      turn on debugging\n");
	printf("\t\t-N                      \"pretend\" flag\n");
}

/*
 *	set_defaults
 *
 *	Function:	- set the default values of key variables
 *			- checks most of the command line arguments
 *
 *	Accepts:	- the number of remaining arguments
 *			- ptr to the list of arguments
 *
 *	Returns:	- 0 or ERROR
 */

static int
set_defaults(argc, argv)

int			argc;
char			*argv[];

{
/*
 * Set the flags.
 */
	fl_debug = opt_taken('d');
	fl_pretend = opt_taken('N');
/*
 * Check the remaining arguments and set the device name.
 */
	if (argc > 2) {
		return(ERROR);
	} else if (argc > 1) {
		linkname = argv[1];
	} else {
		linkname = DEFFDEVICE;
	}

	return(0);
}
