%{
/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD/AKP
 *
 *	$Log:	mosesl.l,v $
 * Revision 2.2  92/04/30  14:59:26  trillium
 * Ohio Release
 * 
 *	Function:	- moses lexical analyzer
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and
 *			  Cornell Research Foundation
 */

#include <portable.h>
#include <moses.h>
#include <typical.h>
#include <y.moses.h>

/*
 * external variables
 */
extern int		errno;			/* global error code */
extern char		*me;			/* program name */

/*
 * static functions
 */
static void		rm_quotes();
static int		yywrap();
static char		*put_str();

%}

%%

writ(e)?|WRIT(E)?	{ yylval.num = M_WRIT; return(MS_WRITE); }
read|READ		{ yylval.num = M_READ; return(MS_READ); }
load|LOAD		{ yylval.num = M_LOAD; return(MS_LOAD); }
f(e)?tch|F(E)?TCH 	{ yylval.num = M_FTCH; return(MS_FETCH); }
call|CALL		{ yylval.num = M_CALL; return(MS_CALL); }
runp|RUNP		{ yylval.num = M_RUNP; return(MS_RUNP); }
sttimer|STTIMER		{ yylval.num = M_STTIMER; return(MS_STTIMER); }
warm|WARM		{ yylval.num = M_WARM; return(MS_WARM); }
out|OUT			{ yylval.num = M_OUT;  return(MS_OUT); }
in|IN			{ yylval.num = M_IN;   return(MS_IN); }

q(uit)?|Q(UIT)?		{ return(QUIT); }
e(xit)?|E(XIT)?		{ return(QUIT); }
help|HELP		{ return(HELP); }
\n			{ return(NEWLINE); }
[ \t]+			{ }
"#"[^\n]*\n		{ }

[0-9a-fA-F]+		{ sscanf(yytext, "%x", &yylval.num); return(NUMBER); }
0(x|X)[0-9a-fA-F]+	{ sscanf(yytext+2,"%x", &yylval.num); return(NUMBER); }
[0-9]+			{ sscanf(yytext, "%d", &yylval.num); return(NUMBER); }

\"[^\n\t]*\"		{ rm_quotes();yylval.string=put_str();return(STRING); }
[a-zA-Z/\.][^ \n\t]*	{ yylval.string=put_str(); return(STRING); }

%%

/*
 *	rm_quotes
 *
 *	Function:	- remove first and last quotes of the string
 *			- work directly on the yytext[] array
 *			- update the string length yyleng
 *			- abort in case of error
 */
static void 
rm_quotes()
{
	int	i;

	if (yytext[0] != '\"') {
		fprintf(stderr, "%s (rm_quotes): no beginning quote\n", me);
		exit(ERROR);
	}
	if (yytext[yyleng-1] != '\"') {
		fprintf(stderr, "%s (rm_quotes): no ending quote\n", me);
		exit(ERROR);
	}
/*
 * Compact by shifting the string one character to the beginning.
 * Decrement the string length (yyleng) by two.
 */
	for (i=1; i < yyleng-1; i++) {
		yytext[i-1] = yytext[i];
	}
	yyleng -= 2;
}

/*
 *	put_str
 *
 *	Function:	- make a copy of the parsed string
 *			- memory is dynamically allocated
 *			- abort in case of error
 *
 *	Returns:	- ptr to the string
 */
static char *
put_str()
{
	char		*string;
	int		i;

	if ((string = (char *)malloc(yyleng+1)) == NULL) {
		fprintf(stderr, "%s (put_str): cannot malloc\n", me);
		exit(errno);
	}

	for (i = 0; i < yyleng; i++) {
		string[i] = yytext[i];
	}
	string[yyleng] = '\0';

	return(string);
}

/*
 *	yywrap
 *
 *	Function:	- stub to satisfy the lex generated code.
 */
static int
yywrap()

{
	return(1);
}
