%{
/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD/GDB
 *
 *	$Log:	mosesy.y,v $
 * Revision 2.2  92/04/30  14:59:33  trillium
 * Ohio Release
 * 
 *	Function:	- moses command filter
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and
 *			  Cornell Research Foundation
 */

#include <portable.h>
#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <t_types.h>
#include <moses.h>

#define MYBUFSIZE	512

/*
 * external functions
 */
extern char		*malloc();
extern int		WriteLink();
extern int		ReadLink();

/*
 * external variables
 */
extern int		errno;
extern int		fl_debug;
extern int		fl_pretend;
extern int		linkdesc;
extern char		*me;

/*
 * static functions
 */
static void		do_call();
static void		do_fetch();
static void		do_in();
static void		do_load();
static void		do_out();
static void		do_read();
static void		do_runp();
static void		do_sttimer();
static void		do_warm();
static void		do_write();
static void		help_moses();
static void		ms_arg();
static void		ms_cmnd();
static void		ms_read();
static void		ms_readw();
static void		ms_write();
static void		warning();
static int		yyerror();

%}

%union {
	int4	num;
	char	*string;
}

%token <num>	NUMBER
%token <string>	STRING

%token <num>	'=' HELP QUIT NEWLINE

%token <num>	MS_WRITE MS_READ MS_LOAD MS_FETCH MS_CALL
%token <num>	MS_RUNP MS_STTIMER MS_WARM MS_OUT MS_IN

%start commands

%%

commands	: /* nothing */
		| commands command		{ printf("ms> "); }
		;

command		: no_args NEWLINE
		| one_args NEWLINE
		| two_args NEWLINE
		| three_args NEWLINE
		| help NEWLINE
		| quit NEWLINE
		| NEWLINE
		| error NEWLINE			{ yyerrok; }
		;

help		: HELP				{ help_moses(); }
		;

quit		: QUIT				{ return(0); }
		;

no_args		: MS_WARM			{ do_warm(); }
		;

one_args	: MS_READ NUMBER		{ do_read($2); }
		| MS_STTIMER NUMBER		{ do_sttimer($2); }
		;

two_args	: MS_WRITE NUMBER NUMBER	{ do_write($2, $3); }
		| MS_LOAD STRING NUMBER		{ do_load($3, $2); }
		| MS_LOAD NUMBER NUMBER		{ printf("not supported\n"); }
		| MS_FETCH NUMBER NUMBER	{ do_fetch($2, $3, (char *)0); }
		| MS_CALL NUMBER NUMBER		{ do_call($2, $3); }
		| MS_RUNP NUMBER NUMBER		{ do_runp($2, $3); }
		;

three_args	: MS_FETCH STRING NUMBER NUMBER	{ do_fetch($3, $4, $2); }
		| MS_OUT NUMBER NUMBER NUMBER	{ do_out($2, $3, $4); }
		| MS_IN NUMBER NUMBER NUMBER	{ do_in($2, $3, $4); }
		;

%%

/*
 *	do_write
 *
 *	Function:	- execute the moses write command
 *			- write a word at memory location
 *
 *	Accepts:	- the address of the memory location
 *			- the word to write
 */
static void
do_write(address, word)

int4			address,
			word;

{
	if (fl_debug) printf("write 0x%x 0x%x\n", address, word);
	if (! fl_pretend) {
		ms_cmnd(M_WRIT);
		ms_arg(address);
		ms_arg(word);
	}
}

/*
 *	do_read
 *
 *	Function:	- execute the moses read command
 *			- read and print the contents of a memory location
 *
 *	Accepts:	- the address of the memory location
 */
static void
do_read(address)

int4			address;

{
	int4		word;

	if (fl_debug) printf("read 0x%x\n", address);
	if (! fl_pretend) {
		ms_cmnd(M_READ);
		ms_arg(address);
		ms_readw(&word);
		printf("0x%08X\n", word);
	}
}

/*
 *	do_load
 *
 *	Function:	- load a file onto the memory of the node
 *
 *	Accepts:	- the memory address
 *			- the file name
 */
static void
do_load(address, filename)

int4			address;
char			*filename;

{
	int		fd;			/* file descriptor */
	int4		blksize;		/* block size */
	int4		length;			/* file length */
	char		*buffer;
	struct stat	filestat;

	if (fl_debug) printf("load %s 0x%x\n", filename, address);
	if (fl_pretend) {
		if (filename) free(filename);
		return;
	}
/*
 * Open the file.
 */
	fd = open(filename, O_RDONLY, 0644);
	if (fd < 0) {
		warning("do_load", "cannot open file");
		if (filename) free(filename);
		return;
	}
/*
 * Get the file size.
 */
	if (fstat(fd, &filestat)) {
		warning("do_load", "cannot get file stat");
		close(fd);
		if (filename) free(filename);
		return;
	}
	if ((length = (int4) filestat.st_size) == 0) {
		close(fd);
		if (filename) free(filename);
		return;
	}
	printf("%d>", length);
	fflush(stdout);
/*
 * Allocate the buffer.
 */
	if ((buffer = malloc(MYBUFSIZE)) == NULL) {
		warning("do_load", "cannot malloc buffer");
		close(fd);
		if (filename) free(filename);
		return;
	}
/*
 * Do the loading.
 */
	ms_cmnd(M_LOAD);
	ms_arg(length);
	ms_arg(address);
	while (length > 0) {
		blksize = (length > MYBUFSIZE) ? MYBUFSIZE : length;
		read(fd, buffer, blksize);
		ms_write(buffer, blksize);
		length -= blksize;
		printf(".");
		fflush(stdout);
	}
/*
 * Clean up.
 */
	printf("\n");
	close(fd);
	if (filename) free(filename);
	free(buffer);
}

/*
 *	do_fetch
 *
 *	Function:	- execute the moses fetch command
 *			- fetch a block of memory and optionally
 *			  store it in a file
 *
 *	Accepts:	- the length of the block (in bytes)
 *			- the starting address of the memory block
 *			- the file name
 */
static void
do_fetch(length, address, filename)

int4			length,
			address;
char			*filename;

{
	int		fd;			/* file descriptor */
	int4		addr;			/* tmp address */
	int4		bufsize;		/* buffer size (bytes) */
	int4		blksize;		/* size of a fetched block */
	int4		i, j, remain;
	char		*buffer;
	char		ascii[17];

	if (fl_debug) printf("fetch %s 0x%x 0x%x\n", filename,length,address);
	if (fl_pretend) {
		if (filename) free(filename);
		return;
	}
/*
 * Take care of the trivial case.
 */
	if (length <= 0) {
		if (filename) free(filename);
		return;
	}

	if (filename) {
/*
 * A filename is given.  Open it and allocate a buffer of a fixed size.
 */
		fd = open(filename, (O_WRONLY|O_CREAT|O_TRUNC), 0644);
		if (fd < 0) {
			warning("do_fetch", "cannot open file");
			if (filename) free(filename);
			return;
		}
		buffer = malloc(MYBUFSIZE);
		bufsize = MYBUFSIZE;
	}
	else {
/*
 * No filename.  Allocate a buffer of the given length.
 * Make sure the length is a multiple of 4 (word = 4 bytes).
 */
		if ((length % 4) != 0) {
			length = ((length / 4) + 1) * 4;
		}
		buffer = malloc(length);
		bufsize = length;
	}
	if (buffer == NULL) {
		warning("do_fetch", "cannot malloc buffer");
		if (filename) free(filename);
		return;
	}
/*
 * Fetch the bytes from the node.  Note that in case of a fetch
 * to a file, several fetches of a fixed size will be made.
 */
	addr = address;
	remain = length;
	while (remain > 0) {
/*
 * Fetch a chunk from memory.
 */
		blksize = (remain < bufsize) ? remain : bufsize;
		ms_cmnd(M_FTCH);
		ms_arg(blksize);
		ms_arg(addr);
		ms_read(buffer, blksize);

		if (filename) {
/*
 * Write it to the file.
 */
			(void) write(fd, buffer, blksize);
		}
		else {
/*
 * Print it out in a nice format.
 */
			printf("%08X ", addr);
			i = j = 0;
			do {
				printf(" %02X", (buffer[i] & 0xFF));
				if ((buffer[i] >= ' ') && (buffer[i] < 0x7F)) {
					ascii[j] = buffer[i];
				}
				else {
					ascii[j] = '.';
				}
				ascii[++j] = (char) 0;
				i++;
				if ((i % 16) == 0) {
					printf("  /%s/\n", ascii);
					j = 0;
					if (i < blksize) printf("%08X ",addr+i);
				}
			} while (i < blksize);
			if ((i % 16) != 0) {
				for ( ; (i % 16) != 0; i++) { printf("   "); }
				printf("  /%s/\n", ascii);
			}
		}
/*
 * Update the loop parameters.
 */
		remain -= blksize;
		addr += blksize;
	}
/*
 * Clean up.
 */
	if (filename) {
		close(fd);
		free(filename);
	}
	free(buffer);
}

/*
 *	do_call
 *
 *	Function:	- execute the moses call command
 *			- call a process with a given instruction pointer
 *			  and stack pointer
 *	Accepts:	- the instruction pointer value
 *			- the stack pointer value
 */
static void
do_call(ip, sp)

int4			ip,
			sp;

{
	if (fl_debug) printf("call 0x%x 0x%x\n", ip, sp);
	if (! fl_pretend) {
		ms_cmnd(M_CALL);
		ms_arg(ip);
		ms_arg(sp);
	}
}

/*
 *	do_runp
 *
 *	Function:	- execute the moses runp command
 *			- create a parallel process with a given
 *			  instruction pointer and stack pointer
 *	Accepts:	- the instruction pointer value
 *			- the stack pointer value
 */
static void
do_runp(ip, sp)

int4			ip,
			sp;

{
	if (fl_debug) printf("runp 0x%x 0x%x\n", ip, sp);
	if (! fl_pretend) {
		ms_cmnd(M_RUNP);
		ms_arg(ip);
		ms_arg(sp);
	}
}

/*
 *	do_sttimer
 *
 *	Function:	- executes the moses sttimer command
 *			- sets the timers
 */
static void
do_sttimer(time)

int4			time;

{
	int4		word;

	if (fl_debug) printf("sttimer 0x%x\n", time);
	if (! fl_pretend) {
		ms_cmnd(M_STTIMER);
		ms_arg(time);
	}
}

/*
 *	do_warm
 *
 *	Function:	- execute the moses warm command
 *			- warm restart the processor
 */
static void
do_warm()
{
	if (fl_debug) printf("warm\n");
	if (! fl_pretend) {
		ms_cmnd(M_WARM);
	}
}

/*
 *	do_out
 *
 *	Function:	- execute the moses out command
 *			- do an out on a channel
 *	Accepts:	- the buffer address
 *			- the channel address
 *			- the buffer size (bytes)
 */
static void
do_out(buf, chan, size)

int4			buf;
int4			chan;
int4			size;

{
	if (fl_debug) printf("out 0x%x 0x%x 0x%x\n", buf, chan, size);
	if (! fl_pretend) {
		ms_cmnd(M_OUT);
		ms_arg(buf);
		ms_arg(chan);
		ms_arg(size);
	}
}

/*
 *	do_in
 *
 *	Function:	- execute the moses in command
 *			- do an in on a channel
 *	Accepts:	- the buffer address
 *			- the channel address
 *			- the buffer size (bytes)
 */
static void
do_in(buf, chan, size)

int4			buf;
int4			chan;
int4			size;

{
	if (fl_debug) printf("in 0x%x 0x%x 0x%x\n", buf, chan, size);
	if (! fl_pretend) {
		ms_cmnd(M_IN);
		ms_arg(buf);
		ms_arg(chan);
		ms_arg(size);
	}
}

/*
 *	ms_cmnd
 *
 *	Function:	- send the moses command to the compute node
 *	Accepts:	- the command
 */
static void
ms_cmnd(com)

int			com;

{
	char		c_com;
	int		n;

	c_com = (char) com;
	n = WriteLink(linkdesc, &c_com, sizeof(c_com), 0);

	if (n < 0) {
		fprintf(stderr, "moses: can't write to device\n");
		exit(1);
	}
}

/*
 *	ms_arg
 *
 *	Function:	- send the moses argument to the compute node
 *			- the argument is a 4-byte integer
 *	Accepts:	- the argument
 */
static void
ms_arg(arg)

int4			arg;

{
	int		n;

	arg = ltot(arg);
	n = WriteLink(linkdesc, &arg, sizeof(arg), 0);

	if (n < 0) {
		fprintf(stderr, "moses: can't write to device\n");
		exit(1);
	}
}

/*
 *	ms_readw
 *
 *	Function:	- read a word from the compute node
 *	Accepts:	- ptr to the word
 */
static void
ms_readw(word)

int4			*word;

{
	int4		nret;

	nret = ReadLink(linkdesc, (char *) word, sizeof(int4), 0);

	if (nret < 0) {
		fprintf(stderr, "moses: can't read from device\n");
		exit(1);
	}

	else if (nret < 4) {
		warning("ms_read", "got less than expected");
		fprintf(stderr, "got %d expected 4\n", nret);
	}

	*word = ttol(*word);
}

/*
 *	ms_read
 *
 *	Function:	- read a specified number of bytes from the node
 *	Accepts:	- ptr to the buffer
 *			- the maximum number of bytes to receive
 */
static void
ms_read(buf, nbytes)

char			*buf;
int4			nbytes;

{
	int4		nret;

	nret = ReadLink(linkdesc, buf, nbytes, 0);

	if (nret < 0) {
		fprintf(stderr, "moses: can't read from device\n");
		exit(1);
	}

	else if (nret < nbytes) {
		warning("ms_readw", "got less than expected");
		fprintf(stderr, "got %d expected %d\n", nret, nbytes);
	}
}

/*
 *	ms_write
 *
 *	Function:	- write a specified number of bytes to the node
 *	Accepts:	- ptr to the buffer
 *			- the number of bytes
 */
static void
ms_write(buf, nbytes)

char			*buf;
int4			nbytes;

{
	int		n;

	n = WriteLink(linkdesc, buf, nbytes, 0);

	if (n < 0) {
		fprintf(stderr, "moses: can't write to device\n");
		exit(1);
	}
}

/*
 *	help_moses
 *
 *	Function:	- print help menu for moses commands
 */
static void
help_moses()
{
    printf("\n\t%s commands:\n\n", me);
    printf("\thelp                      this message\n");
    printf("\tquit                      exit %s\n", me);
    printf("\texit                      exit %s\n", me);
    printf("\twrite addr word           write word at location addr\n");
    printf("\tread addr                 read the contents of addr\n");
    printf("\tload file addr            load file starting from addr\n");
    printf("\tfetch length addr         fetch length bytes from addr\n");
    printf("\tfetch file length addr    fetch and store in the file\n");
    printf("\tcall ip stack             call new process using ip and stack\n");
    printf("\trunp ip stack             create process using ip and stack\n");
    printf("\tsttimer time              set the time\n");
    printf("\twarm                      warm restart the processor\n");
    printf("\tout buf chan cnt          out on a channel\n");
    printf("\tin buf chan cnt           in on a channel\n");
}

/*
 *	yyerror
 *
 *	Function:	- report yacc syntax errors
 *	Accepts:	- the error message
 *	Returns:	- 0
 */
static int
yyerror(message)

char			*message;

{
	warning((char *) 0, message);
	return(0);
}

/*
 *	warning
 *
 *	Function:	- print a warning message
 *	Accepts:	- the module name
 *			- the warning message
 */
static void
warning(mname, message)

char			*mname,
			*message;

{
	fprintf(stderr, "%s", me);

	if (mname) {
		fprintf(stderr, " (%s)", mname);
	}

	if (message) {
		fprintf(stderr, ": %s\n", message);
	} else {
		fprintf(stderr, "\n");
	}
}
