/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	RBD
 *
 *	$Log:	spread.c,v $
 * Revision 2.2  92/04/30  15:15:46  trillium
 * Ohio Release
 * 
 * Revision 2.1.1.5  92/02/23  19:02:23  gdburns
 * Make cosmetic change to help menu.
 * 
 * Revision 2.1.1.4  92/01/31  18:21:28  gdburns
 * Upgrade to suit the overhauled boot subsystem.
 * 
 * Revision 2.1.1.3  91/11/26  17:28:15  raja
 * Modified to use libtools and did a face lift.
 * 
 * Revision 2.1.1.2  91/07/18  14:57:43  raja
 * Removed the size limitation on the hboot topology string by making it
 * dynamically allocated.
 * 
 * Revision 2.1.1.1  91/06/05  21:45:12  raja
 * Enhanced portability: cast the NT_ITB and NT_ALL constants to int4.
 * 
 * Revision 2.1  91/03/20  13:03:17  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.3  91/03/11  01:48:44  gdburns
 * Added recursive boot option.
 * 
 * Revision 2.0.1.2  90/11/16  16:25:14  raja
 * Overhaul of the topology layer tools.
 * 
 *	Function:	- topology booting tool
 *			- boots Trollius on a network of processors
 *			- topology is specified using NaIL
 *			- uses hboot, sboot, vboot and trout
 *			- based on Trollius 2.0 Copyright 1990
 *			  The Ohio State University and Cornell Research
 *			  Foundation
 *
 *      Accepts:        - options described in man page and help menu
 *
 *      Returns:        - 0 or ERROR
 */

#include <stdio.h>
#include <unistd.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>
#include <net.h>
#include <events.h>
#include <debug.h>
#include <nail.h>
#include <routeio.h>

/*
 * global variables
 */
char			*me;			/* tool name */
char		 	*t_hboot;		/* hboot tool */
char			*t_sboot;		/* sboot tool */
char		 	*t_tkill;		/* tkill tool */
char			*t_trout;		/* trout tool */
char			*t_tserv;		/* tserver tool (PC Trollius) */
char		 	*t_vboot;		/* vboot tool */
int			fl_debug;		/* debug mode */
int			fl_frozen;		/* use frozen files */
int			fl_pretend;		/* pretend mode */
int			fl_recurse;		/* recursive booting */
int			fl_verbose;		/* verbose mode */

/*
 * external variables
 */
extern int		errno;
extern FILE		*yyin;			/* lex input file */

/*
 * static variables
 */
static struct bschema	nail;			/* network description */

static int		fl_header;		/* print the header */
static int		fl_troutes;		/* send route tables */
static int		topc;			/* nbr. of topology arguments */
static char		rtfname[256];		/* route filename */
static char		*f_bnail;		/* boot schema filename */
static char		**p_bnail;		/* boot schema pathname */
static char		**topv;			/* topology arguments */

static char *usage = "%s [-dhtvEHNVZ] [-c <bnail>] [-p <route>] [-T <topology>]\n";
static char *header= "Ohio Trollius 2.2 - The Ohio State University";

/* 
 * external functions
 */
extern void		exit();
extern void		exclusive();
extern void		followed();
extern void		free();
extern void		validopts();
extern char		*findpath();
extern char		*getenv();
extern char		*getparam();
extern char		*malloc();
extern char		*sprintf();
extern char		*strcat();
extern char		*strchr();
extern char		*strcpy();
extern char		*strrchr();
extern int		addarg();
extern int		do_args();
extern int		exec_hboot();
extern int		exec_sboot();
extern int		exec_stage();
extern int		exec_tkill();
extern int		exec_tool();
extern int		exec_tserv();
extern int		exec_vboot();
extern int		opt_taken();
extern int		parse_nail();
extern int4		intparam();
extern int4		stoi4();

/* 
 * static functions
 */
static int		boot_system();
static int		find_path();
static int		find_ntype();
static int		send_rtable();
static int		set_defaults();
static void		help_spread();
static void		prt_defaults();

/*
 * Main program.
 */
int
main(argc, argv)  

int			argc;
char			*argv[];

{
	char		*full;		/* temporary for full pathname */
/*
 * Assign the tool name.
 */
	if ((me = strrchr(argv[0], STRDIR)) == (char *) 0) {
		me = argv[0];
	}
	else {
		me++;
	}
/*
 * Parse the command line.
 */
	validopts("dhtvHNV");
	followed("cpT");
	exclusive("eE");
	exclusive("zZ");

	if ((do_args(&argc, argv) == ERROR) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {
		fprintf(stderr, usage, me);
		exit(1);
	}
 
	if (opt_taken('h')) {
		help_spread();
		exit(0);
	}
/*
 * Set the mode of operation.
 */
	if (set_defaults()) {
		exit(1);
	}
	if (fl_header) {
		printf("\n%s\n\n", header);
	}
	if (fl_debug) {
		prt_defaults();
	}
/* 
 * Locate the tools and files.  The route file is not located here, but in
 * the trout tool instead.  For PC Trollius, tserver is located instead
 * of hboot.
 */
	if ((full = findpath(f_bnail, p_bnail, R_OK)) == (char *) 0) {
		fprintf(stderr, "%s: cannot locate \"%s\"\n", me, f_bnail);
		exit(errno);
	} else {
		f_bnail = full;
	}
/*
 * Open the boot schema file.
 */
	DBUG("%s: opening the boot schema file: %s ...\n", me, f_bnail);
	if ((yyin = fopen(f_bnail, "r")) == NULL) {
		perror(me);
		exit(errno);
	}
/*
 * Parse the boot schema.
 */
	if (parse_nail(&nail)) {
		fprintf(stderr, "%s: cannot parse the NaIL syntax\n", me);
		exit(1);
	}
/*
 * Close the boot schema.
 */
	if (fclose(yyin)) {
		perror(me);
		exit(errno);
	}
/*
 * Pretty print the boot schema out if needed.
 */
	if (opt_taken('V')) { 
		pretty_bnail(&nail);
		exit(0);
	}
/*
 * Determine the host node of the network.  It will be
 * the root of the spanning tree.
 */
	if (find_host(&nail)) {
		fprintf(stderr, "%s: cannot find the network host\n", me);
		exit(1);
	}
	DBUG("%s: found hostnode %d\n", me, nail.bs_hostnode->tn_nodeid);
/*
 * Generate the network spanning tree.
 */
	DBUG("%s: generating the booting order ...\n", me);
	if (make_spantree(&nail)) {
		fprintf(stderr, "%s: cannot generate spanning tree\n", me);
		exit(1);
	}
	if (fl_debug) {  
		print_spantree(&nail);
	}
/*
 * Boot the system.
 */
	DBUG("%s: booting the system ...\n", me);
	if (boot_system()) {
		fprintf(stderr, "%s: unable to boot the system\n", me);
		exit(1);
	} else {
		DBUG("%s: booted OK\n", me);
	}

	exit(0);
}

/*
 *	help_spread
 *
 *	Function:	- print the help menu
 *
 */
static void 
help_spread()
{
	printf("\nSynopsis:\t%s [options]\n", me);
	printf("\nDescription:\tBoot a Trollius multicomputer.\n");
	printf("\nOptions:\t-h                      Print this message.\n");
	printf("\t\t-d                      Turn on debugging.\n");
	printf("\t\t-t                      Do not ship the route tables.\n");
	printf("\t\t-v                      Be verbosee\n");
	printf("\t\t-E                      Do not boot recursively.\n");
	printf("\t\t-H                      Don't print the header.\n");
	printf("\t\t-N                      Pretend to take action.\n");
	printf("\t\t-V                      Pretty print the bnail file.\n");
	printf("\t\t-Z                      Do not use frozen files.\n");
	printf("\t\t-c bnail                Use \"bnail\" as boot schema.\n");
	printf("\t\t-p route                Use \"route\" as the route file\n");
	printf("\t\t-T \"device link\"        Use \"device\" as \"link\".\n");
}

/*
 *	set_defaults
 *
 *	Function:	- set the values of key variables controlling
 *			  the operation of the tool
 *
 *	Returns:	- 0 or ERROR
 */
static int 
set_defaults()
{
	int		n_bnail;
	char		*p;			/* favorite pointer */
/*
 * Set the pathnames for the tool and the boot schema.
 */
	n_bnail = 0;
	p_bnail = (char **) 0;
	addarg(&n_bnail, &p_bnail, "");				/* absolute */
	addarg(&n_bnail, &p_bnail, "$TROLLIUSHOME/boot");	/* env */
	addarg(&n_bnail, &p_bnail, DEFPBNAIL);			/* default */
/*
 * Set the tool names.
 */
	t_hboot = DEFTHBOOT;
	t_sboot = DEFTSBOOT;
	t_tkill = DEFTTKILL;
	t_trout = DEFTTROUT;
	t_tserv = DEFTTSERV;
	t_vboot = DEFTVBOOT;
/*
 * Set the file names.
 */
	if (opt_taken('c')) {
		f_bnail = getparam('c');
	} else if (f_bnail = getenv("TROLLIUSBNAIL")) {
	} else {
		f_bnail = DEFFBNAIL;
	}
  
	if (opt_taken('p')) {
		(void)strcpy(rtfname, getparam('p'));
	}
	else {
		(void)strcpy(rtfname, DEFFROUTE);
		if (p = strchr(rtfname, '.')) {
			*p = '\0';
		}
		p = strrchr(f_bnail, STRDIR);
		if (p = strchr((p) ? p + 1 : f_bnail, '.')) {
			(void)strcat(rtfname, ".");
			(void)strcat(rtfname, p + 1);
		}
	}
/*
 * Set the flags.
 */
	fl_debug = opt_taken('d');
	fl_recurse = ( ! opt_taken('E'));
	fl_verbose = opt_taken('v') || fl_debug;
	fl_pretend = opt_taken('N');
	fl_troutes = ( ! opt_taken('t'));
	fl_frozen = ( ! opt_taken('Z'));
	fl_header = ( ! opt_taken('H'));
/*
 * Split up the topology string, if it exists.
 */
	topc = 0;
	topv = (char **) 0;
	if (opt_taken('T')) {
		split(getparam('T'), &topv, &topc);
	}
    
	return(0);
}

/*
 *	prt_defaults
 *
 *	Function:	- print the values of key variables
 *			- mostly used for debugging
 */
static void 
prt_defaults()
{
	printf("\nDefault settings\n\n");
	printf("The boot file: %s\n", f_bnail);
	printf("The route file: %s\n", rtfname);
	printf("debug flag: %d\n", fl_debug);
	printf("recurse flag: %d\n", fl_recurse);
	printf("verbose flag: %d\n", fl_verbose);
	printf("pretend flag: %d\n", fl_pretend);
	printf("header flag: %d\n", fl_header);
	printf("route table shipping flag: %d\n", fl_troutes);
	printf("use frozen flag: %d\n", fl_frozen);
	printf("f_bnail = \"%s\"\n", f_bnail);
	printf("t_hboot = \"%s\"\n", t_hboot);
	printf("t_sboot = \"%s\"\n", t_sboot);
	printf("t_vboot = \"%s\"\n", t_vboot);
	printf("t_trout = \"%s\"\n", t_trout);
	printf("t_tserv = \"%s\"\n", t_tserv);
}

/*
 *	boot_system
 *
 *	Function:	- boot the Trollius network as follows:
 *			  1) tkill any previous system
 *			  2) sboot the device connected itb nodes
 *			  3) hboot the host (skip in PC Trollius)
 *			  4) trout the host
 *			  5) trout the device connected itb nodes
 *			  6) stage the device connected itb nodes (vboot)
 *			  7) vboot and trout all other itb nodes
 *			  8) exec tserver (only in PC Trollius)
 *
 *	Returns:	- 0 or ERROR
 */
static int 
boot_system()

{
	int		i, j;			/* favorite counters */
	int		n_topo;			/* # topology strings */
	int4		nodeid, hostid;		/* node identifiers */
	int4		nnodes;			/* # of nodeids in array */
	int4		*nodes;			/* array of nodeids */
	int4		*snodes;		/* array of sboot nodes */
	char		tmp_string[256];	/* used by sprintf */
	char		*device;		/* host device name */
	char		*topology;		/* topology string */
	char		**v_topo;		/* topology strings for hboot */
	struct link	*links;			/* ptr to list of links */
	struct nspecs	*specs;			/* ptr to node specification */
	struct t_link	*tlink;			/* ptr to link */
	struct t_node	*tnode, *dc_tnode;	/* node pointers */
	HASH		*tnhd;			/* TNODE hash table desc. */
/*
 * Check to see if there is anything to boot at all.
 */
	tnhd = nail.bs_tnhd;
	if (ah_count(tnhd) < (int4) 1) {
		fprintf(stderr, "%s : no nodes to boot\n", me);
		return(ERROR);
	}
/*
 * Issue a tkill command.  This allows users to spread several
 * times without having to remember to tkill the system.
 */
	if (exec_tkill()) {
		return(ERROR);
	}
	specs = nail.bs_hostnode->tn_specs;

#ifdef DOS
/*
 * For the current version of PC Trollius only one device
 * is supported.  Check if more than one is being used.
 */
	if (specs->ns_nspanout > 1) {
		fprintf(stderr, "%s : multiple devices not allowed\n", me);
		return(ERROR);
	}
#endif

/*
 * Sboot the device connected nodes.  These are the nodes directly
 * attached to the host.
 */
	n_topo = 0;
	v_topo = (char **) 0;
	links = specs->ns_lkspan;
	for (i = 0; i < specs->ns_nspanout; i++, links = links->lk_next) {
		tlink = links->lk_link;
/*
 * Check for command line override and determine the device name.
 */
		for (j = 1; (j < topc) && (stoi4(topv[j]) !=
			tlink->tl_srclink); j += 2) {
		}

		if (j < topc) {
			device = topv[j - 1];
		} else {
			device = tlink->tl_srcname;
		}

		if (device == (char *) 0) {
			fprintf(stderr, "%s : no device name from n%d to n%d\n",
				me, (tlink->tl_srcnode)->tn_nodeid,
				(tlink->tl_destnode)->tn_nodeid);
			return(ERROR);
		}
/*
 * Build the topology string for hboot.
 */
		sprintf(tmp_string, "0x%X", dl_event(tlink->tl_srclink));
		addarg(&n_topo, &v_topo, device);
		addarg(&n_topo, &v_topo, tmp_string);
/*
 * Boot the device connected node.
 */
		nodeid = (tlink->tl_destnode)->tn_nodeid;
		if (exec_sboot(nodeid, device)) {
			return(ERROR);
		}
	}
/*
 * Boot the host node. Host booting is a stub in PC-Trollius.
 * Build the single topology information string.
 * Reserve space for one space character after each string
 * and one space for the null character at the end.
 */
	for (i = 0, j = 1; i < n_topo; i++) {
		j += (strlen(v_topo[i]) + 1);
	}
	if ((topology = malloc(j)) == (char *) 0) {
		perror(me);
		return(ERROR);
	}
	for (i = 0, *topology = '\0'; i < n_topo; i++) {
		(void) strcat(topology, v_topo[i]);
		(void) strcat(topology, " ");
	}

	hostid = nail.bs_hostnode->tn_nodeid;
	if (exec_hboot(hostid, topology)) {
		return(ERROR);
	}
	free(topology);
/*
 * Send the route tables to the host and all device connected nodes.
 * In PC Trollius, do not send the route table to the host.
 * Allocate and fill the array of nodeids to be trouted.
 */
	nnodes = specs->ns_nspanout;
	nodes = (int4 *) malloc((nnodes + 1) * sizeof(int4));
	if (nodes == (int4 *) 0) {
		fprintf(stderr, "%s: cannot malloc node list to trout\n", me);
		return(ERROR);
	}
	snodes = nodes;

#ifndef DOS
/* 
 * Include the hostid in the list of nodeids.
 */
	nnodes++;
	*snodes++ = hostid;
#endif

/*
 * Include all device connected nodes in the list of nodeids.
 */
	links = specs->ns_lkspan;
	for (i = 0; i < specs->ns_nspanout; i++, links = links->lk_next) {
		snodes[i] = ((links->lk_link)->tl_destnode)->tn_nodeid;
	}
/*
 * Send all the route tables.
 */
	if (send_rtable(nodes, nnodes, device, FALSE)) {
		return(ERROR);
	}
/*
 * Get the list of device connected nodes ignoring all nodes that
 * have no spanning children nodes.  If no such nodes exists, we are done.
 */
	nnodes = 0;
	links = specs->ns_lkspan;
	for (i = 0; i < specs->ns_nspanout; i++, links = links->lk_next) {
		dc_tnode = (links->lk_link)->tl_destnode;
		if ((dc_tnode->tn_specs)->ns_nspanout > 0) {
			nodes[i] = dc_tnode->tn_nodeid;
			nnodes++;
		}
	}

	if (nnodes == 0) {
		return(0);
	}

	if (fl_recurse) {
/*
 * Recursive booting is specified.  Send the file of routing tables
 * of all nodes to the device connected nodes that have no children.
 */
		if (send_rtable(nodes, nnodes, device, TRUE)) {
			return(ERROR);
		}
	}
/* 
 * Stage all the device connected nodes that have children nodes.
 * These nodes will be used to boot the remaining nodes in the system.
 * If recursive booting is chosen, this stage is all we need to do.
 */
	if (exec_stage(nodes, nnodes, hostid, device)) {
		return(ERROR);
	}

	if ( ! fl_recurse) {
/* 
 * No recursive booting required, just loop sequentially from the host.
 * Boot the rest of the itb nodes in the booting order specified
 * in the network description.  For each node booted, send its 
 * route table.
 */
		tnode = specs->ns_bootnext;
		for ( ; tnode; tnode = specs->ns_bootnext) {
			specs = tnode->tn_specs;
/* 
 * Boot all nodes spanning from the current node.
 */
			links = specs->ns_lkspan;
			for (i = 0; i < specs->ns_nspanout; i++) {
				tlink = links->lk_link;
				nodeid = (tlink->tl_destnode)->tn_nodeid;
				if (exec_vboot(tnode->tn_nodeid, nodeid, hostid,
					tlink->tl_srclink, device)) {
					return(ERROR);
				}
				if (send_rtable(&nodeid, (int4) 1, 
					device, FALSE)) {
					return(ERROR);
				}
				links = links->lk_next;
			}
		}
	}
/*
 * Execute the host server.  This is only used in PC Trollius,
 * and is only a stub in the Unix (tm) version.
 */
	if (exec_tserv(hostid, device,
		find_ntype((int4) NT_ITB, (int4) NT_ALL)))  {
		return(ERROR);
	}

	(void) free((char *) nodes);
	return(0);
}

/*
 *	send_rtable
 *
 *	Function:	- send route tables to a list of nodeids
 *			- format and issue a trout command
 *
 *	Accepts:	- array of nodeids
 *			- number of nodeids
 *			- device name
 *			- recursive booting flag
 *
 *	Returns:	- 0 or ERROR
 */
static int
send_rtable(nodeids, nnodes, device, recursive)

int4			*nodeids;
int4			nnodes;
char			*device;
int			recursive;

{
	int		n_cmd;
	char		**v_cmd;
	char		tmp_string[256];	/* used by sprintf */
/*
 * Build the command line arguments for the tool.
 */
	n_cmd = 0;
	v_cmd = (char **) 0;
	addarg(&n_cmd, &v_cmd, t_trout);
	if (fl_verbose) {
		addarg(&n_cmd, &v_cmd, "-v");
	}
	if (! fl_troutes) {
		addarg(&n_cmd, &v_cmd, "-t");
	}
	if (rtfname) {
		addarg(&n_cmd, &v_cmd, "-p");
		addarg(&n_cmd, &v_cmd, rtfname);
    	}
	if (recursive) {
		addarg(&n_cmd, &v_cmd, "-e");
	}

	VERBOSE("trout ");
	for ( ; nnodes > 0; nnodes--, nodeids++) {
		sprintf(tmp_string, "n%#x", *nodeids);
		addarg(&n_cmd, &v_cmd, tmp_string);
		VERBOSE("n%d ", *nodeids);
	}
	VERBOSE("...\n");
/*
 * Make lint happy because this variable is only used in PC-T.
 */
	device = device;

	if (fl_debug) {
		printfv(n_cmd, v_cmd);
	}
/* 
 * Execute the command.
 */
	if (! fl_pretend) {
		return(exec_tool(t_trout, v_cmd));
	} else {
		return(0);
	}
}

/*
 *	find_ntype
 *
 *	Function:	- get the number of nodes having given
 *			  node type attributes.
 *
 *	Accepts:	- node type (bit values desired)
 *			- type mask (bit locations desired)
 *
 *	Returns:	- a positive integer
 */
static int4
find_ntype(type, mask)

int4			type,
			mask;

{
	int4		nnode;
	struct t_node	*tnode;
	HASH		*tnhd;

	type &= mask;
	nnode = INT4_NIL;
	tnhd = nail.bs_tnhd;
	tnode = (struct t_node *) ah_top(tnhd);
	while (tnode) {
		if (((tnode->tn_specs)->ns_type & mask) == type) {
			nnode++;
		}
		tnode = (struct t_node *) ah_next(tnhd, (void *) tnode);
	}

	return(nnode);
}
