/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	thaw.c,v $
 * Revision 2.2  92/04/30  15:16:37  trillium
 * Ohio Release
 * 
 *	Function:	- OTB to ITB frozen loader
 *			- loads multi-program system frozen with moses cmds
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <boot.h>
#include <t.out.h>
#include <typical.h>
#include <machine/machine.h>

#define BUFSIZE		8192

/*
 * external functions
 */
extern char		*findpath();
extern char		*getparam();
extern char		*malloc();
extern int		do_args();
extern int		read();
extern int		OpenLink();
extern int		CloseLink();
extern int		WriteLink();
extern void		addarg();	/* add arg to arg list */
extern void		followed();
extern void		validopts();

/*
 * local functions
 */
static void		setpaths();	/* initialize search paths */
static void		help();

/*
 * local variables
 */
static int		n_config;	/* count of p_config */
static char		*me;		/* name I was exec'd by */
static char		**p_config;	/* frozen search paths */
static char		buffer[BUFSIZE];
static char		*usage = "usage: %s <frozen> <device>\n";

/*
 * external variables
 */
extern int		errno;

main(argc, argv)

int			argc;
char			*argv[];

{
	char		*ifname;	/* input filename */
	char		*ofname;	/* output filename */
	char		*ifullname;	/* input full pathname */
	char		*p;		/* favourite ptr */
	int		len;		/* input buffer length */
	int		ifd;		/* input file descriptor */
	int		ofd;		/* output file descriptor */

	me = argv[0];
/*
 * Initialize option parser.
 */
	validopts("hiov");
	followed("io");

	if (do_args(&argc, argv) < 0) {
		(void) fprintf(stderr, usage, me);
		exit(errno);
	}

	if (opt_taken('h')) {
		help();
		exit(0);
	}

	--argc;
	++argv;
	setpaths();
	ifname = NULL;
	ofname = NULL;
/*
 * Check for a single file argument with no decidability between input
 * and output.
 */
	if (!opt_taken('o') && !opt_taken('i') && (argc == 1)) {
		(void) fprintf(stderr, usage, me);
		exit(1);
	}
/*
 * Establish input file name.
 */
	if (opt_taken('i')) {
		ifname = getparam('i');
	} else if (argc > 0) {
		ifname = argv[0];
		--argc;
		++argv;
	}
/*
 * Establish output file name.
 */
	if (opt_taken('o')) {
		ofname = getparam('o');
	} else if (argc > 0) {
		ofname = argv[0];
		--argc;
		++argv;
	}

	if (argc > 0) {
		(void) fprintf(stderr, usage, me);
		exit(1);
	}
/*
 * Establish input file descriptor.
 */
	if (ifname == NULL) {
		ifd = 0;
		ifname = "stdin";
	} else {
	    	if ((ifullname = findpath(ifname, p_config, R_OK)) == NULL) {
			fprintf(stderr, "%s: can't find %s\n", me, ifname);
			exit(1);
		}

		ifname = ifullname;
		ifd = open(ifname, O_RDONLY, 0);

		if (ifd < 0) {
			perror(ifname);
			exit(errno);
		}
	}

	if (opt_taken('v')) fprintf(stderr, "%s: input is %s\n", me, ifname);
/*
 * Establish output file descriptor.
 */
	if (ofname == NULL) {
		ofd = 1;
		ofname = "stdout";
	} else {

#ifdef PARSY
/*
 * Replace '/' with ' ' in Parsytec link name.
 * Thanks to the person who decided to embed blanks in a single identifier.
 */
		for (p = ofname; *p; ++p) if (*p == '/') *p = ' ';
#endif
/*
 * Open the device.
 */
		ofd = OpenLink(ofname);

		if (ofd < 0) {
			fprintf(stderr, "%s: cannot open link\n", me);
			exit(1);
		}
	}

	if (opt_taken('v')) fprintf(stderr, "%s: output is %s\n", me, ofname);
/*
 * Copy input file to output file.
 */
	if ((len = read(ifd, buffer, BUFSIZE)) < 0) {
		perror(me);
		exit(errno);
	}

	while (len > 0) {

		if (WriteLink(ofd, buffer, len, 0) < 0) {
			perror(me);
			exit(errno);
		}

		if ((len = read(ifd, buffer, BUFSIZE)) < 0) {
			perror(me);
			exit(errno);
		}
	}

	(void) close(ifd);
	(void) CloseLink(ofd);

	return(0);
}

/*
 *	setpaths
 *
 *	Function:	- initialize search paths for input and output files
 */

static void
setpaths()

{
/*
 * input files
 */
	n_config = 0;
	p_config = (char **) 0;
	addarg(&n_config, &p_config, "");
	addarg(&n_config, &p_config, "$TROLLIUSHOME");
	addarg(&n_config, &p_config, DEFPCONFIG);
}
 
/*
 *        help
 *
 *        Function:     - prints helpful information on the thaw command
 */
 
static void
help()
{
    printf("\nSynopsis:\tthaw [options] <frozen> <device>\n");
    printf("\nDescription:\tLoad frozen system on neighbour node.\n");
    printf("\nOptions:        -h\t\tPrint this message.\n");
    printf("\t\t-v\t\tBe verbose.\n");
    printf("\t\t-i <frozen>\tSpecify frozen input.\n");
    printf("\t\t-o <device>\tSpecify device output.\n");
    printf("\nExample:        thaw conf.isbz /dev/link0\n");
    printf("\t\t\t\"Load `conf.isbz' over `/dev/link0'.\"\n\n");
}
