/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	kio.msg.c,v $
 * Revision 2.2  92/04/30  15:19:17  trillium
 * Ohio Release
 * 
 *	Function:	- cleans up OTB system IPC resources
 */

#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#include <debug.h>
#include <kernel.h>

/*
 * external variables
 */
extern int		fl_debug;
extern char		*me;

/*
 * external functions
 */
extern char		*getenv();
extern key_t		ftok();

/*
 * local functions
 */
static void		kiokill();

/*
 *	kiokillk
 *
 *	Function:	- cleans up kernel process IPC
 *			- searches and removes message queues
 */

void
kiokillk()

{
	char		f_token[32];		/* pathname to create token */

	DBUG("%s: kernel msgq...", me);
/*
 * Generate the token file used by the process to create its message queue.
 */
	sprintf(f_token, "/tmp/%s%dk", getenv("USER"), getuid());
	kiokill(f_token);
	DBUG("\n");
}

/*
 *	kiokillc
 *
 *	Function:	- cleans up client process IPC
 *			- searches and removes message queues
 *	Accepts:	- process ID
 */

void
kiokillc(pid)

int			pid;

{
	char		f_token[32];		/* pathname to create token */

	DBUG(", msgq ...");
/*
 * Generate the token file used by the process to create its message queue.
 */
	sprintf(f_token, "/tmp/%s%dc", getenv("USER"), pid);
	kiokill(f_token);
}

/*
 *	kiokill
 *
 *	Function:	- cleans up IPC
 *			- searches and removes message queues
 *	Accepts:	- token creation filename
 */

static void
kiokill(f_token)

char			*f_token;

{
	key_t 		token;			/* unique token */
	int		qid;			/* msg queue ID */
/*
 * Create the token (key) used to specify the message queue.
 */
	if ((token = ftok(f_token, KERNEL_ID)) < 0) {

		if (errno == ENOENT) {
			DBUG(" already removed");
		} else {
			DBUG(" (ftok) errno %d", errno);
		}

		return;
	}		
/*
 * Learn process qid.
 */
	if ((qid = msgget(token, 0600)) < 0) {

		if (errno == ENOENT) {
			DBUG(" already removed");
		} else {
			DBUG(" (msgget) errno %d", errno);
		}

		return;
	}
/*
 * Remove the message queue.
 */
 	if (msgctl(qid, IPC_RMID, (struct msqid_ds *) 0) < 0) {
		DBUG(" (msgctl) errno %d", errno);
	} else {
		DBUG(" removed");
	}
/*
 * Remove the token creation file.
 */
	(void) unlink(f_token);
}
