/*
 *	OSU Research Computing					Trillium
 *	Copyright 1992 The Ohio State University
 *	GDB
 *
 *	$Log:	tkill.c,v $
 * Revision 2.2  92/04/30  15:19:30  trillium
 * Ohio Release
 * 
 *	Function:	- kills OTB operating system
 *			- reads process IDs from the kill file, aka
 *			  the lock file for historical reasons
 *			- makes a very good attempt at ensuring the
 *			  death of each pid in the file
 */

#include <boot.h>
#include <debug.h>
#include <portable.h>
#include <signal.h>
#include <stdio.h>
#include <sys/errno.h>
#include <terror.h>
#include <typical.h>
#include <unistd.h>

#define TIMEOUT		10000000	/* timeout on process death */

/*
 * external variables
 */
extern int		errno;

/*
 * external functions
 */
extern void		followed();
extern void		kiokillc();
extern void		kiokillk();
extern void		validopts();
extern int		addarg();
extern int		do_args();
extern int		putenv();
extern int		usleep();
extern char		*findpath();
extern char		*getenv();
extern char		*getparam();
extern char		*strrchr();

/*
 * local functions
 */
static void		help();
static void		setdefaults();

/*
 * global variables
 */
char			*me;		/* who am I? */
int			fl_debug;	/* debugging option */
int			fl_verbose;	/* verbose option */

/*
 * local variables
 */
static char		*f_kill;	/* kill file */
static char		**p_device;	/* kill pathname prefixes */

static int		fl_killer;	/* killing others */
static int		n_device;	/* # of p_device prefixes */

static char		*user;		/* who am i? */
static char		*usage = " [-hdvN] [-f <killfile>]";

main(argc, argv)

int			argc;
char			*argv[];

{
	FILE		*fp_kill;	/* kill file ptr */
	FILE		*fp_email;	/* email file ptr */
	char		*full;		/* full pathname */
	int		n;		/* favourite counter */
	int		pid;		/* lots of space to hold a PID */
	unsigned int	usec;		/* microseconds to sleep */
	char		*f_tail;	/* tail of pathname */
	static char	f_email[24];	/* temp file for email to victim */
	static char	syscmd[80];	/* shell command for email */

	me = argv[0];
/*
 * Initialize option parser.
 */
	validopts("hdvKN");
	followed("f");

	if ((do_args(&argc, argv) < 0) ||
		(errno = (argc == 1) ? 0 : EUSAGE)) {

		fprintf(stderr, "usage: %s%s\n", me, usage);
		exit(errno);
	}

	if (opt_taken('h')) {
		help();
		exit(0);
	}

	setdefaults();
/*
 * Try to locate the kill file.
 */
	full = findpath(f_kill, p_device, F_OK);

	if (! full) {

		if ((errno != ENOENT) || fl_killer) {
			fprintf(stderr, "%s: can't access kill file - %s\n",
				me, f_kill);
			exit(1);
		} else {
			exit(0);
		}
	} else {
		f_kill = full;
	}
/*
 * Dump debugging information.
 */
	DBUG("%s: f_kill = \"%s\"\n", me, f_kill);
/*
 * Open the kill file.
 */
	fp_kill = fopen(f_kill, "r+");

	if (!fp_kill) {
		fprintf(stderr, "%s: can't open kill file - %s\n", me, f_kill);
		perror(me);
		exit(errno);
	}

	VERBOSE("reset ...\n");

	n = fscanf(fp_kill, "%d", &pid);

	while (n > 0) {
		DBUG("%s: %d ...", me, pid);
/*
 * Send SIGTERM to the process.
 */
		if (opt_taken('N')) {
			DBUG("\n");
		} else {

			if (kill(pid, SIGTERM) < 0) {

				if (errno == ESRCH) {
					DBUG(" already dead");
				} else {
					perror(me);
					exit(errno);
				}
			} else {
				usec = 2;

				while ((kill(pid, SIGTERM) >= 0) &&
						(usec <= TIMEOUT)) {
					(void) usleep(usec);
					usec *= 2;
				}
	
				if (usec > TIMEOUT) {
					DBUG(" cannot kill");
				} else {
					DBUG(" killed");
				}
			}
/*
 * Clean up any IPC resources.
 */
			kiokillc(pid);
			DBUG("\n");
		}

		n = fscanf(fp_kill, "%d", &pid);
	}

	if (opt_taken('N')) {
		exit(0);
	}

	kiokillk();
/*
 * Send email to victim if not suicide.
 */
	if (fl_killer) {
		sprintf(f_email, "/tmp/tkill%d", getpid());
		fp_email = fopen(f_email, "w");

		if (! fp_email) {
			fprintf(stderr, "%s: cannot open %s for email msg\n",
				me, f_email);
			perror(me);
			exit(errno);
		}

		fprintf(fp_email,
			"\nYour Trollius session was killed by %s.\n", user);
		fprintf(fp_email,
			"\nThis message generated automatically...\n\n");

		if (fclose(fp_email) == EOF) {
			fprintf(stderr, "%s: cannot close %s for email msg\n",
				me, f_email);
			perror(me);
			exit(errno);
		}
/*
 * Don't hassle me about this.
 */
		(void) putenv("MAILRC=/dev/null");
		f_tail = strrchr(f_kill, STRDIR);
		f_tail = f_tail ? f_tail + 1 : f_kill;

		sprintf(syscmd, "/usr/ucb/mail -s \"tkill by %s\" %s < %s",
			user, f_tail, f_email);
		system(syscmd);

		if (unlink(f_email) < 0) {
			perror(me);
			exit(errno);
		}
	}
/*
 * Remove the kill file.
 */
	if (unlink(f_kill) < 0) {
		fprintf(stderr, "%s: cannot close %s\n", me, f_kill);
		perror(me);
		exit(errno);
	}

	return(0);
}

/*
 *	setdefaults
 *
 *	Function:	- sets default files and paths
 *			- may exit
 */

static void
setdefaults()

{
/*
 * prefix paths
 */
	n_device = 0;
	p_device = (char **) 0;
	addarg(&n_device, &p_device, "");
	addarg(&n_device, &p_device, "$TROLLIUSHOME/locks");
	addarg(&n_device, &p_device, DEFPRESET);
/*
 * flags
 */
	fl_debug = opt_taken('d');;
	fl_verbose = opt_taken('v');
/*
 * files
 */
	user = getenv("USER");

	if (opt_taken('f')) {
		f_kill = getparam('f');
		fl_killer = strcmp(f_kill, user);
	} else {

		if (! user) {
			fprintf(stderr, "%s: environment USER not defined\n",
				me);
			exit(1);
		}

		f_kill = user;
		fl_killer = FALSE;
	}
}

/*
 *	help
 *
 *	Function:	- prints helpful information on the tkill command
 */

static void
help()

{
    printf("\nSynopsis:	tkill [options]\n");
    printf("\nDescription:	Kill Trollius on an OTB node.\n");
    printf("\nOptions:	-h		Print this message.\n");
    printf("\t\t-d		Print debug information.\n");
    printf("\t\t-v		Be verbose.\n");
    printf("\t\t-N		Pretend to tkill.\n");
    printf("\t\t-f <kill>	Use <kill> for the kill file.\n\n");
}
