/*
 *	Cornell Theory Center					Trillium
 *
 *	(third party software provided under license)
 *
 *	$Log:	tvw.c,v $
 * Revision 2.2  92/04/30  15:24:14  trillium
 * Ohio Release
 * 
 * Revision 2.1  91/03/20  13:03:45  gdburns
 * Ohio Release
 * 
 * Revision 2.0.1.1  90/07/13  19:32:21  gdburns
 * Homo sapien
 * 
 * Revision 2.0  90/02/20  11:57:33  trillium
 * Homo erectus
 * 
 * Revision 1.1  88/06/05  01:45:18  trillium
 * long overdue after 127 fixes and upgrades
 * 
 * Revision 1.0  87/12/19  16:15:27  trillium
 * Initial revision
 * 
 *
 *	We're just putting our stamp on this for housekeeping reasons.
 */

/* tvw   
 */

#ifdef dsi
#include <t.out.h>
#else
#include <t.out.h>
#endif

#include <stdio.h>

FILE		*fp;
struct exec	h;
struct nlist *sym;
struct to_relative *tr, *dr;
char *string;

main(ac, av)
	char	**av;
{
	int	i;

	for(i=1; i<ac; i++)
	    dofile(av[i]);
}

dofile(cp)
	char	*cp;
{
	if ((fp = fopen(cp, "r")) == NULL) {
	    fprintf(stderr, "tvw: cannot open %s\n",cp);
	    return;
	}

	to_header();
	to_symbol();
	to_tr();
	to_dr();
}

to_header()
{
	char	*x_type;

	if (fread(&h, sizeof(struct exec), 1, fp) != 1) {
	    fprintf(stderr, "tvw: cannot read header\n");
	    return;
	}

	if (h.a_magic == TO_MAGIC_TN_X) {
		x_type = "unspecified transputer";
	}
	else if (h.a_magic == TO_MAGIC_T2_X) {
		x_type = "T2 compatible";
	}
	else if (h.a_magic == TO_MAGIC_T4_X) {
		x_type = "T4 compatible";
	}
	else if (h.a_magic == TO_MAGIC_T8_X) {
		x_type = "T8 compatible";
	}
	else {
	    fprintf(stderr, "tvw: bad magic number\n");
	    return;
	}

	printf("file:              %s\n", x_type);
	printf("text size:         %d\n", h.a_text);
	printf("data size:         %d\n", h.a_data);
	printf("bss size:          %d\n", h.a_bss);
	printf("symbols:           %d\n", h.a_syms/sizeof(struct nlist));
	printf("entry point:     0x%08x\n", h.a_entry);
	printf("text relocations:  %d\n",h.a_trsize/sizeof(struct to_relative));
	printf("data relocations:  %d\n",h.a_drsize/sizeof(struct to_relative));
}

to_symbol()
{
	int	i, c;
	int	slen;

	sym = (struct nlist *) malloc (h.a_syms);
	fseek(fp, N_SYMOFF(h), 0);
	c = fread(sym, 1, h.a_syms, fp);
	if (c != h.a_syms) {
	    fprintf(stderr, "tvw: unable to read %d symbols\n", h.a_syms);
	    exit(1);
	}
	fseek(fp, N_STROFF(h), 0);
	fread(&slen, 4, 1, fp);
	string = (char *) malloc(slen);
	fread(string, 1, slen, fp);

	printf("\nSymbol table:\n");
	for(i=0; i<(h.a_syms/sizeof(struct nlist)); i++) {
	    printf("    symbol: %s, ", sym[i].n_un.n_strx + string - 4);
	    if (sym[i].n_other & N_REG)
		printf("internal, ");
	    if (sym[i].n_type == (N_UNDF | N_EXT))
		if (sym[i].n_value)
		    printf("common of size %d\n", sym[i].n_value);
		else
		    printf("undefined external symbol\n");
	    else {
		if (sym[i].n_type & N_EXT)
		    printf("external, ");
		switch(sym[i].n_type & N_TYPE) {
		    case N_UNDF:	printf("Undefined local\n"); break;
		    case N_ABS: 	if (sym[i].n_type & N_EXT)
					    printf("global absolute");
					else
					    printf("local absolute");
					printf(", value:%d\n",sym[i].n_value);
					break;
		    case N_TEXT:	if (sym[i].n_type & N_EXT)
					    printf("global text");
					else
					    printf("local text");
					printf(", value:%d\n",sym[i].n_value);
					break;
		    case N_DATA:	if (sym[i].n_type & N_EXT)
					    printf("global data");
					else
					    printf("local data");
					printf(", value:%d\n",sym[i].n_value);
					break;
		    case N_BSS: 	if (sym[i].n_type & N_EXT)
					    printf("global bss");
					else
					    printf("local bss");
					printf(", value:%d\n",sym[i].n_value);
					break;
		    case N_COMM:	printf("weird comm, "); break;
		    case N_FN:		printf("file name\n"); break;
		    default:		printf("weird symbol type\n");
					break;
		}
	    }
	}
}

to_tr()
{
	int	i;

	tr = (struct to_relative *) malloc(h.a_trsize);
	fseek(fp, N_TXTOFF(h) + h.a_text + h.a_data, 0);
	fread(tr, 1, h.a_trsize, fp);

	printf("\nText relocation\n");
	for(i=0; i< h.a_trsize/sizeof(struct to_relative); i++) {
	    printf("    at %d", tr[i].pc);
	    if (tr[i].type)
		printf(", word");
	    else
		printf(", prefix");
	    if (tr[i].s1.index != -1)
		if (tr[i].s1.index > h.a_syms/sizeof(struct nlist) ||
		    tr[i].s1.index < 0)
		    printf(", s1.index: %d (out of range)", tr[i].s1.index);
		else
		    printf(", s1:%s",
			sym[tr[i].s1.index].n_un.n_strx + string - 4);
	    if (tr[i].s1_external)
		printf(", s1 external");
	    if (tr[i].s2.index != -1)
		if (tr[i].s2.index > h.a_syms/sizeof(struct nlist) ||
		    tr[i].s2.index < 0)
		    printf(", s2.index: %d (out of range)", tr[i].s2.index);
		else
		    printf(", s2:%s",
			sym[tr[i].s2.index].n_un.n_strx + string - 4);
	    if (tr[i].s2_external)
		printf(", s2 external");
	    printf(", const:%d\n", tr[i].cnst);
	}
}

to_dr()
{
	int	i;

	dr = (struct to_relative *) malloc(h.a_drsize);
	fseek(fp, N_TXTOFF(h) + h.a_text + h.a_data + h.a_trsize, 0);
	fread(dr, 1, h.a_drsize, fp);

	printf("\nData relocation\n");
	for(i=0; i< h.a_drsize/sizeof(struct to_relative); i++) {
	    printf("    at %d", dr[i].pc);
	    if (dr[i].s1_text)
		printf(", text relative");
	    else
		printf(", data relative");
	    if (dr[i].type)
		printf(", word");
	    else
		printf(", prefix");
	    if (dr[i].s1.index != -1)
		printf(", s1:%s", sym[dr[i].s1.index].n_un.n_strx + string - 4);
	    if (dr[i].s1_external)
		printf(", s1 external");
	    if (dr[i].s2.index != -1)
		printf(", s2:%s", sym[dr[i].s2.index].n_un.n_strx + string - 4);
	    if (dr[i].s2_external)
		printf(", s2 external");
	    printf(", const:%d\n", dr[i].cnst);
	}
}
