/*
 *	Cornell Theory Center				tti
 *
 *	Copyright 1988 Cornell University
 *
 *	AKP
 *
 *	$Log:	compare.ts,v $
Revision 1.1  88/08/19  15:35:44  gdburns
Initial revision

 */

/*
** Date: Fri, 19 Dec 86 14:36:36 EST
** From: gdburns@tcgould.tn.cornell.edu (Greg Burns)
**
**	Thanks, Greg.
**
*/

/*
**    Compare(dict, search)
**
**	Compare a dictionary name entry to a null terminated string.
**	Dictionary entry is a DICTMAX byte field followed by a full
**	length stored in the next byte.  The dictionary name entry is
**	declared equivalent to the search string if the two match in
**	length and the leading bytes are equivalent up to a maximum
**	of DICTMAX bytes.
**
**	CALL:
**		ld?  search		;pointer to search string
**		ld?  dict		;pointer to dictionary name entry
**		call Compare
**
**	True or false is returned in the A register and the length of
**	the null terminated string is always stored in the B register
**
*/


#define AREG		1
#define	BREG		2
#define SEARCH		0
#define DICT		1
#define DICTBOUND	2
#define SEARCHLEN	3
#define DICTLEN		4

#define LOCAL		5
#define DICTMAX		7


	EXPORT	Compare

	CSEG

Compare:

	ajw	-LOCAL			;allocate space for locals

	ldl	LOCAL + AREG		;load A parameter (dict entry)
	stl	DICT			;save pointer to dict entry
	ldl	LOCAL + BREG;		;load B parameter (search string)
	stl	SEARCH			;save pointer to search string

	ldc	0
	stl	SEARCHLEN		;initialize search string length

	ldc	DICTMAX
	stl	DICTBOUND		;initialize dict boundary length

	ldl	DICT
	adc	DICTMAX
	lb				;extract full dict entry length
	ldc	0x7f			;mask off flag bit
	and				;
	stl	DICTLEN			;save full dict entry length

cont:
	ldl	DICTLEN			;check length of dict entry
	cj	testnull		;end reached, check search null term.
	
	ldl	SEARCH			;check search null terminator
	lb				;load current byte in search string
	cj	false			;lengths don't match, goto false

	ldl	DICTBOUND		;check dict entry boundary length
	cj	spin			;dict entry exhausted, spin to end
	j	nextbyte		;not exhausted, compare next byte
	
testnull:
	ldl	SEARCH			;check search null terminator
	lb				;load current byte in search string
	cj	true			;lengths match, bytes equal, goto true
	j	false			;lengths don't match, goto false

nextbyte:
	inc(SEARCHLEN)

;
;Compare next two bytes.
;
	ldl	SEARCH			;current search address
	lb				;current search byte
	ldl	DICT			;current dict address
	lb				;current dict byte
	sub				;compare bytes
	eqc	0			;reverse logic
	cj	false			;bytes not equal, goto false

	inc(SEARCH)

	inc(DICT)

	ldl	DICTBOUND		;dec dict entry boundary
	adc	-1
	stl	DICTBOUND

	ldl	DICTLEN			;dec dict entry length
	adc	-1
	stl	DICTLEN

	j	cont			;continue comparing bytes

spin:					;bytes are equal, check lengths
	inc(SEARCHLEN)

	inc(SEARCH)

	ldl	DICTLEN			;dec dict entry length
	adc	-1
	stl	DICTLEN

	ldl	SEARCH			;check for null terminator
	lb
	cj	testlen			;null found, check dict length
	j	spin			;null not found, continue spin

testlen:
	ldl	DICTLEN			;check dictionary entry length
	cj	true			;lengths match, got true

false:
	ldl	SEARCHLEN		;load search string length
	ldc	0			;load false
	ajw	LOCAL			;deallocate space for locals
	ret				;return to caller

true:
	ldl	SEARCHLEN		;load search string length
	ldc	1			;load true
	ajw	LOCAL			;deallocate space for locals
	ret				;return to caller
