/*
 *	Cornell Theory Center				tti
 *
 *	Copyright 1988 Cornell University
 *
 *	AKP
 *
 *	Function:	- some conversions
 *
 *	$Log:	cvt.ts,v $
Revision 1.1.1.1  88/09/29  21:34:07  gdburns
Use tti.h like everybody else.

Revision 1.1  88/08/19  15:35:45  gdburns
Initial revision

 */


#include "tti.h"


	CSEG
	NAME	CVT
	EXPORT	CVT_Str2Int
	EXPORT	CVT_Int2Str

;=============================================================
;
;    CVT_digit(n)
;
;	Map byte-value "n" into a character in [0-9A-Z]
;	and return it in aReg.
;
;	CALL:
;		ld?  n
;		call CVT_digit
;
;	RETURN:
;	    ASCII character in aReg.
;
;=============================================================
CVT_digit:
	ldl aReg	; the byte to be mapped
	ldc 0x09	;
	gt		; is the byte > 9?
	cj  L_add48	; jump if byte <= 9...
	ldl aReg	; its > 9 so we need to bump it up
	adc 0x07	; into A..Z range by adding 0x07
	stl aReg	; and saving it.
L_add48:
	ldl aReg	; the byte (may already have an offset)
	adc 0x30	; ASCII '0'

	ret		; return n = n + 48 + (n > 9 ? 7 : 0)

;=============================================================
;
;    CVT_UnMap(ch)
;
;	Convert a character in the set [0-9A-Za-z] into
;	a value between 0 and 35, inclusive.  The value
;	is returned in the aReg.  The registers are
;	preserved (except aReg).
;
;	CALL:
;		ld?  ch
;		call CVT_UnMap
;
;=============================================================
#define ch	aReg
CVT_UnMap:
	ldc  'a'		; convert to upper case...
	ldl  ch			; (ch >= 'a')
	gt			;
	eqc  0			;
	cj   UM_norm		; jump if (ch < 'a')
	ldl  ch			; (ch <= 'z')
	ldc  'z'		;
	gt			;
	eqc  0			;
	cj   UM_norm		; jump if (ch > 'z')
	ldl  ch			; ah ha: ch is in [a-z]
	ldc  32			; upper-case it by subracting 32
	sub			;
	stl  ch			;

UM_norm:
	ldc  '0'		; skip the normalization if its a digit...
	ldl  ch			; (ch >= '0')
	gt			;
	eqc  0			;
	cj   UM_bad		; less than '0' is bad...jump if so...
	ldl  ch			; (ch <= '9')
	ldc  '9'		;
	gt			;
	cj   UM_sub0		; its a digit -- go subtract off '0'

	ldc  'A'		; normalize into [0-9:-@A-S] (48..83)
	ldl  ch			; (ch >= 'A')
	gt			;
	eqc  0			;
	cj   UM_bad		; jump if (ch < 'A') (not allowed)
	ldl  ch			; (ch <= 'Z')
	ldc  'z'		;
	gt			;
	eqc  0			;
	cj   UM_bad		; jump if (ch > 'Z') (not allowed)
	ldl  ch			; ah ha: ch is in [A-Z]
	ldc  7			; normalize it by subtracting 7
	sub			;
	stl  ch			;

UM_sub0:			; finally, subtract off '0'
	ldl  ch			;
	ldc  '0'		;
	sub			;
	stl  ch			;

UM_ret:
	ldl  cReg		; restore the registers
	ldl  bReg		;
	ldl  ch			;

	ret

UM_bad:
	ldc -1			; an out of range value
	stl ch			;
	j   UM_ret		;

#undef ch

;=============================================================
;
;    CVT_IsValid(ch, radix)
;
;	Return TRUE in aReg if ch is a valid character in
;	radix "radix" FALSE, otherwise.  radix is in bReg,
;	ch is in aReg.
;
;	CALL:
;		ld?  radix
;		ld?  ch
;		call CVT_IsValid
;
;=============================================================
#define ch    aReg
#define radix bReg
#define tmp   cReg

CVT_IsValid:
	ldl  ch			; first need to UnMap it
	call CVT_UnMap		;
	stl  tmp		;

	ldc  0			; ((tmp >= 0) && (tmp < radix)
	ldl  tmp		;
	gt			;
	eqc  0			;
	ldl  radix		;
	ldl  tmp		;
	gt			;
	and			; I think I like stack machines...

	ret

#undef ch
#undef radix
#undef tmp

;=============================================================
;
;    CVT_Str2Int(radix, str)
;
;	Convert "str" (a null-terminated ASCII string) to
;	a 32 bit word using "radix" as the conversion base.
;	"str" is the absolute address of a string and is
;	in bReg, "radix" (in range 2 to 36 inclusive) is in
;	aReg.  The number will be returned in bReg and a
;	success/failure (1 == successful conversion) is
;	returned in aReg.
;
;	CALL:
;		ld?  str
;		ld?  radix
;		call CVT_Str2Int
;
;=============================================================
#define LOCAL	4
#define p	0
#define pos	1
#define scale	2
#define val	3
#define str	(LOCAL + bReg)
#define radix   (LOCAL + aReg)

CVT_Str2Int:
	ajw -LOCAL		; make room for our local variables
	ldc  0			; set val to a sensible value
	stl  val		;
	ldc  1			; assume that the number is positive
	stl  pos		;
	ldc  1			; set scale to 1
	stl  scale		;

	ldl  str		; check for optional sign character
	lb			;
	eqc  '-'		;
	cj   S2I_pos		; jump if str[0] <> '-'
	ldc  0			; ah ha: they say its a negative number
	stl  pos		;
	inc(str)		; skip the minus sign
	j    S2I_n		; go skip leading zeros

S2I_pos:
	ldl  str		; check for '+'
	lb			;
	eqc  '+'		;
	cj   S2I_n		; jump if str[0] <> '+'
	inc(str)		; ah ha: its '+' be we already assumed that
S2I_n:
	ldl  str		; have to check that there is at least one...
	lb			; digit...
	cj   S2I_NotValid	; jump if it is a null...

S2I_skip0:
	ldl  str		; skip all leading '0's
	lb			;
	cj   S2I_all0		; if its a '\000', then it was all '0's...
	ldl  str		;
	lb			;
	eqc  '0'		;
	cj   S2I_save		; jump if str[0] <> '0'
	inc(str)		; skip over the '0'
	j    S2I_skip0		;

S2I_all0:
	ldc  0			; the string was all zeros...
	stl  val		;
	j    S2I_ret		;

S2I_save:
	ldl  str		; save a pointer to the first good...
	stl  p			; character

S2I_check:
	ldl  str		; check each character for validity and
	lb			; advance to the end of the string
	cj   S2I_cvt		; jump if the byte is '\000'
	ldl  radix		; ah ha: non-null character that needs...
	ldl  str		; to be checked for validity
	lb			;
	call CVT_IsValid	;
	cj   S2I_NotValid	; jump if the "IsValid" returned FALSE
	inc(str)		; ah ha: its a good character...
	j    S2I_check		;

S2I_NotValid:
	ldc  0			; the ill-converted number
	ldc  0			; it didn't work...
	ajw  LOCAL		;
	ret			;

S2I_cvt:
	ldl  str		; convert each character from right...
	adc -1			; to left
	stl  str		;
	ldl  str		; unmap the character
	lb			;
	call CVT_UnMap		;
	ldl  scale		; val = val + (scale * UnMap(str[0]))
	mul			;
	ldl  val		;
	add			;
	stl  val		;
	ldl  scale		; scale = scale * radix
	ldl  radix		;
	mul			;
	stl  scale		;
	ldl  str		; do {} while (str != p)
	ldl  p			;
	diff			;
	eqc  0			;
	cj   S2I_cvt		;

	ldl  pos		; do we need to negate it
	eqc  0			;
	cj   S2I_ret		;
	ldl  val		; val = ~val + 1
	not			;
	adc  1			;
	stl  val		;

S2I_ret:
	ldl  val		; the converted value
	ldc  1			; it worked
	ajw  LOCAL		; unallocate our space

	ret

#undef LOCAL
#undef p
#undef pos
#undef scale
#undef val
#undef buffer
#undef radix

;=============================================================
;
;    CVT_Int2Str(n, radix, buffer)
;
;	Convert integer "n" to a null-terminated ASCII string
;	using "radix" for conversion (ie, 2, 8, 10, 16).  If
;	radix is < 0 (ie, -10) a signed conversion will be
;	done.  "dst" should be the absolute address of a
;	sufficiently large space.
;
;	CALL:
;		ldp? buffer 
;		ld?  radix
;		ld?  n
;		call CVT_Int2Str
;
;=============================================================
#define LOCAL	10
#define buffer	0
#define p	9
#define val	(LOCAL + aReg)
#define radix	(LOCAL + bReg)
#define dst	(LOCAL + cReg)

CVT_Int2Str:
	ajw -LOCAL		; we need 10 words (33 chars + 1 word)

	ldc  0			; is 0 > radix?
	ldl  radix		;
	gt			;
	cj   L_unsigned		; jump to L_unsigned if radix >= 0

	ldc  0			; Okay, we're doing a signed conversion
	ldl  val		; Now, is 0 > val?
	gt			;
	cj   L_signed		; jump to L_signed if val >= 0

	ldc  '-'		; val is < 0, throw in a minus sign
	ldl  dst		;
	sb			;

	ldl  dst		; dst = dst + 1
	adc  1			; bump the index into the dst buffer
	stl  dst		;

	ldl  val		; val = -val
	not			; now pretend to do an unsigned...
	adc  1			; conversion and flip the sign of...
	stl  val		; the number to be converted

L_signed:
	ldl  radix		; radix = -radix
	not			; doing a signed conversion and have...
	adc  1			; already taken care of the minus sign...
	stl  radix		; so now pretend to do an unsigned conversion

L_unsigned:
	ldlp buffer		; get the address of the last byte...
	adc  32 		; in our buffer
	stl  p			;

	ldc  0			; make sure that there is a terminating...
	ldl  p			; null byte
	sb			;

	/*  Convert the integer by building our local buffer backwards, */
	/*  digit by digit.  It will be later copied forwards.          */
L_int2chr:
	ldl  p			; p = p - 1
	adc -1			;
	stl  p			;

	ldc  0			; digit = (unsigned) val % radix
	ldl  val		;
	ldl  radix		;
	ldiv			; ...modulus part is in bReg so...
	rev			; bring it to the top of the stack
	call CVT_digit		; map into [0-9A-Z] (returned in aReg)
	ldl  p			; the index into our buffer...
	sb			; and stash the ASCII char away 

	ldc  0			; val = (unsigned) val / radix
	ldl  val		;
	ldl  radix		;
	ldiv			; ...div part is in aReg
	stl  val		;

	ldl  val		; do { ...above... } while (val != 0)
	eqc  0			;
	cj   L_int2chr		;

	/* nothing more to convert, copy our buffer to theirs */
L_copy:
	ldl  p			; while (*dst++ = *p++);  /* ugh! */
	lb			;
	ldl  dst		;
	sb			; *dst = *p

	ldl  p			; get the byte we just copied...
	lb			; and leave it on the stack

	ldl  dst		;
	adc  1			; increment each pointer
	stl  dst		;

	ldl  p			;
	adc  1			;
	stl  p			;

	eqc  0			; is this the null byte?
	cj   L_copy		; from the "ldl p; lb" above

L_done:
	ajw  LOCAL		; unallocate the space
	ret			; and return

#undef dst
#undef radix
#undef val
#undef p
#undef buffer
#undef LOCAL
