/*
 *	Cornell Theory Center				tti
 *
 *	Copyright 1988 Cornell University
 *
 *	AKP
 *
 *	Function:	- THE Dictionary
 *
 *	$Log:	dict.ts,v $
Revision 1.1  88/08/19  15:35:46  gdburns
Initial revision

 */


#include "ascii.h"
#include "tti.h"

	IMPORT	Boot, Start, Abort, Colon, Next, Semi, Run
	IMPORT	Constant, storeFP, FPplus
	IMPORT	Compare
	IMPORT	SetCompile, ClrCompile
	IMPORT	prompt, huh, resetmsg
	IMPORT	If, Goto
	IMPORT	Konstant, String, Variable
	IMPORT	strlen
	IMPORT	OutChar, OutString, OutByt, OutWrd
	IMPORT	CVT_Str2Int, CVT_Int2Str
	IMPORT	sss
	EXPORT	Dictionary

	ISEG

Dictionary:

/*
**
**  "Boot", "START", and "ABORT"
**	These routines have their code as part of the inner-interpreter
**	but have headers defined here for user access.
**
*/
	header("Boot",   _Boot, _Start, Boot);

	header("START", _Start, _Abort, Start);

	header("ABORT", _Abort, Execute, Abort);

	header("EXECUTE", Execute, DefWord, _exec);

_exec:
	ldl  FLAG		; in execute or compile/immediate mode?
	ldc  F_MODE		;
	and			;
	cj   _exec_it		; execute the word if the bit isn't set
	ldl  FLAG		; compiling: execute if an immediate word,
	ldc  F_IMMED		; otherwise enclose it in the dictionary.
	and			;
	cj   _comma		;
_exec_it:
	pop(DS, DINC)		; get the address
	stl WA			; save it in the word address register
	j Run			; execute it

/*
**
**  ":"
**	Define a new word.
**
*/
	header(":", DefWord, DefEnd, Colon)

	ISEG
	dw codep(Create)
	dw storeFP
	dw Colon
	dw SetCompile
	dw Semi

/*
**
**  ";"
**	Terminate the definition of a word.
**
*/	header(";", DefEnd, Repeat, Colon, IMMEDIATE)

	ISEG
	dw storeFP
	dw Semi
	dw ClrCompile
	dw Semi

/*
**
**  "repeat"
**	Mark the beginning of a "repeat...until" loop when
**	defining a word.
*/
	header("repeat", Repeat, Until, _repeat, IMMEDIATE)
_repeat:
	ldl FP			; the address to jump to...
	push(DS, DINC)		; save it on the stack...
	j Next			; I hope it doesn't get trashed.

/*
**
**  "until"
**	Produce code to jump to the address found at the
**	top of the stack if at runtime the top of stack is
**	false.
*/
	header("until", Until, Iff, _until, IMMEDIATE)
_until:
	ldl SL			; Plop down the "If"
	ldnlp	If		;
	ldl FP			;
	stnl 0			;
	incw(FP)		;
	pop(DS, DINC)		; where to jump to...
	ldl FP			; current address...
	sub			; now have relative offset
	ldl FP			; plop that down
	stnl 0			;
	incw(FP)		;
	j Next

/*
**
**  "if"
**	Compile in the first part of an "if ... [else...end]|[end]"
**	construct.
*/
	header("if", Iff, End, _iff, IMMEDIATE)
_iff:
	ldl SL			; Plop down the "If"
	ldnlp	If		;
	ldl FP			;
	stnl 0			;
	incw(FP)		;
	ldl FP			; save the backpatch address on the...
	push(DS, DINC)		; data stack
	incw(FP)		; skip over the space for the jump...
	j Next			; offset

/*
**
**  "end"
**	Backpatch the end of an "if...[else...]end" statement.
*/
	header("end", End, Else, _end, IMMEDIATE)
_end:
	pop(DS, DINC)		; get the backpatch address
	stl R0			;
_bpatch:
	ldl FP			; compute the needed relative offset
	ldl R0			;
	sub			;
	ldl R0			; patch what R0 points at
	stnl 0			;
	j Next			;

/*
**
**  "else"
**	Plop down a "goto" (to skip over the "else" part) and leave
**	the address of the offset on the stack for "end".  Also,
**	backpatch the relative offset of the "if".
*/
	header("else", Else, Plus, _else, IMMEDIATE)
_else:
	pop(DS, DINC)		;
	stl R0			; save the backpatch address
	ldl SL			; Plop down the "Goto"
	ldnlp	Goto		;
	ldl FP			;
	stnl 0			;
	incw(FP)		;
	ldl FP			; save the backpatch address on the...
	push(DS, DINC)		; data stack
	incw(FP)		; skip over the space for the jump...
	j _bpatch		; perform the backpatch for "if"

/*
**
**  "+"
**	Add top two numbers (32 bit) on data stack.  Leaves the sum on
**	the data stack.  (A = B + A)
**
*/
	header("+", Plus, Minus, _plus)
_plus:
	pop(DS, DINC)
	pop(DS, DINC)
	add
	push(DS, DINC)
	j Next


/*
**
**  "-"
**	Subtract top two numbers (32 bit) on data stack.  Leaves the result
**	on the data stack.  (A = B - A)
**
*/
	header("-", Minus, Times, _minus)
_minus:
	pop(DS, DINC)
	pop(DS, DINC)
	rev
	sub
	push(DS, DINC)
	j Next

/*
**
**  "*"
**	Multiply top two numbers (32 bit) on data stack.  Leaves the result
**	on the data stack.  (A = B * A)
**
*/
	header("*", Times, Divide, _mult)
_mult:
	pop(DS, DINC)
	pop(DS, DINC)
	mul
	push(DS, DINC)
	j Next

/*
**
**  "/"
**	Divide top two numbers (32 bit) on data stack.  Leaves the result
**	on the data stack.  (A = B / A)
**
*/
	header("/", Divide, Modulus, _div)
_div:
	pop(DS, DINC)
	pop(DS, DINC)
	rev
	div
	push(DS, DINC)
	j Next

/*
**
**  "MOD"
**	Modulus top two numbers (32 bit) on data stack.  Leaves the result
**	on the data stack.  (A = B MOD A)
**
*/
	header("MOD", Modulus, Plus1, _mod)
_mod:
	pop(DS, DINC)
	pop(DS, DINC)
	rev
	rem
	push(DS, DINC)
	j Next

/*
**
**  "1+"
**	Add 1 to the number on top of the stack. (A = A + 1)
**
*/
	header("1+", Plus1, Minus1, _add1)
_add1:
	pop(DS, DINC)
	adc 1
	push(DS, DINC)
	j Next

/*
**
**  "1-"
**	Subtract 1 from the number on top of the stack. (A = A - 1)
**
*/
	header("1-", Minus1, Plus2, _sub1)
_sub1:
	pop(DS, DINC)
	adc -1
	push(DS, DINC)
	j Next

/*
**
**  "2+"
**	Add 2 to the number on top of the stack. (A = A + 2)
**
*/
	header("2+", Plus2, Minus2, _add2)
_add2:
	pop(DS, DINC)
	adc 2
	push(DS, DINC)
	j Next

/*
**
**  "2-"
**	Subtract 2 from the number on top of the stack. (A = A - 2)
**
*/
	header("2-", Minus2, PlusBPW, _sub2)
_sub2:
	pop(DS, DINC)
	adc -2
	push(DS, DINC)
	j Next

/*
**
**  "BPW+"
**	Add BPW to the number on top of the stack. (A = A + BPW)
**
*/
	header("BPW+", PlusBPW, MinusBPW, _addBPW)
_addBPW:
	pop(DS, DINC)
	adc BPW
	push(DS, DINC)
	j Next

/*
**
**  "BPW-"
**	Subtract BPW from the number on top of the stack. (A = A - BPW)
**
*/
	header("BPW-", MinusBPW, Times2, _subBPW)
_subBPW:
	pop(DS, DINC)
	adc -BPW
	push(DS, DINC)
	j Next

/*
**
**  "2*"
**	Fast multiply (bit shift) the number on top of the stack by 2.
**	(A = 2 * A)
**
*/
	header("2*", Times2, DivBy2, _mul2)
_mul2:
	pop(DS, DINC)
	ldc 2
	shl
	push(DS, DINC)
	j Next

/*
**
**  "2/"
**	Fast divide (bit shift) the number on top of the stack by 2.
**	(A = A / 2)
**
*/
	header("2/", DivBy2, Negate, _div2)
_div2:
	pop(DS, DINC)
	ldc 1
	shr
	push(DS, DINC)
	j Next

/*
**
**  "NEG"
**	Unary minus of the number on top of the stack.
**	(A = -A)
**
*/
	header("NEG", Negate, ShiftL, _negate)
_negate:
	pop(DS, DINC)
	not
	adc 1
	push(DS, DINC)
	j Next

/*
**
**  "<<"
**	Shift next-on-stack left by top-of-stack bits.  (A = B << A)
**
*/
	header("<<", ShiftL, ShiftR, _shl)
_shl:
	pop(DS, DINC)
	pop(DS, DINC)
	rev
	shl
	push(DS, DINC)
	j Next

/*
**
**  ">>"
**	Shift next-on-stack right by top-of-stack bits.  (A = B >> A)
**
*/
	header(">>", ShiftR, Band, _shr)
_shr:
	pop(DS, DINC)
	pop(DS, DINC)
	rev
	shr
	push(DS, DINC)
	j Next

/*
**
**  "AND"
**	Bitwise AND of the top two numbers on the stack.
**	(A = A & B)
**
*/
	header("AND", Band, Bor, _and)
_and:
	pop(DS, DINC)
	pop(DS, DINC)
	and
	push(DS, DINC)
	j Next

/*
**
**  "OR"
**	Bitwise OR of the top two numbers on the stack.
**	(A = A | B)
**
*/
	header("OR", Bor, Bxor, _or)
_or:
	pop(DS, DINC)
	pop(DS, DINC)
	or
	push(DS, DINC)
	j Next

/*
**
**  "XOR"
**	Bitwise exclusive-OR of the top two numbers on the stack.
**	(A = A ^ B)
**
*/
	header("XOR", Bxor, Bnot, _xor)
_xor:
	pop(DS, DINC)
	pop(DS, DINC)
	xor
	push(DS, DINC)
	j Next

/*
**
**  "NOT"
**	Bitwise NOT (complement) of the top number on the stack.
**	(A = ~A)
**
*/
	header("NOT", Bnot, DotQ, _not)
_not:
	pop(DS, DINC)
	not
	push(DS, DINC)
	j Next

/*
**
**  ".""
**    If compiling:
**	Compile in a literal handler for strings, followed by the string.
**    If executing:
**	Print the null-terminated ASCII string whose address is at
**	the top of the stack.  Pops address from the stack.
**
*/
	/*headr("."", DotQ, Dot, Colon)*/
	ALIGN 4
	EXPORT DotQ
	ISEG
DotQ:
	db ".", "\"", 0, 0, 0, 0, 0, 2 + 128
	dw Dot
	dw Colon
	dw codep(Compile)
	dw If
	dw 3 * BPW
	dw compstr
	dw Semi
	dw printstr
	dw Semi
	
	/* this is run if not compiling */
printstr:
	dw DotQuote

	CSEG
DotQuote:
	pop(DS, DINC)		; get the string's starting address...
	ldl  CO			; and our output channel is the console...
	call OutString		; output it...
	j Next			; and return.

	/* this is run iff compiling */
	ISEG
compstr:
	dw Colon
	dw storeFP		/* store the handler pointer */
	dw String
	dw Constant		/* get the string */
	dw DQUOTE
	dw codep(Token)
	dw storestr		/* skip around it */
	dw Semi

storestr:
	dw _storestr

	CSEG
_storestr:
	/* not TOO rough...the string is already nicely in the right
	** spot, just have to bump up the FP to the next word after
	** the null put in by TOKEN
	*/
	ldl FP
	lb
	cj  _strdone
	inc(FP)
	j _storestr

_strdone:
	/* move FP to the nearest word boundary */
	ldl FP
	ldc -BPW
	and
	adc BPW
	stl FP
	j Next

/*
**
**  "."
**	Print (and pop) the number on the top of the stack...
*/
	header(".", Dot, DotCX, _dot)
_dot:
	pop(DS, DINC)		; get the number
	stl  R0			;
	ldl  FP			; build the string in free space above free
	adc  PAD		; space to skip over any good things left.
	ldl  BASE		; the conversion base
	ldl  R0			; the number
	call CVT_Int2Str	; convert it
	ldl  FP			; 
	adc  PAD		; string's address...
	ldl  CO			; output channel...
	call OutString		; print it...
	j    Next		; and return.


/*
**
**  ".cx"
**	Print a byte in semi-formatted hex (always 2 chars: ie 01 or
**	F3)
*/
	header(".cx", DotCX, DotX, _dotcx)
_dotcx:
	pop(DS, DINC)		; get the value to print
	ldl CO			;
	call OutByt		; print it
	j Next			; and return

/*
**
**  ".x"
**	Print a 32-bit number in semi-formatted hexidecimal (always
**	8 characters)  This could easily become a secondary...
**
*/
	header(".x", DotX, GetCH, _dotx)
_dotx:
	pop(DS, DINC)		; the value to print
	ldl CO			; output channel...
	call OutWrd		;
	j Next			;

/*
**
**  "getc"
**	Get a character from the console and leave it in the least
**	significant byte of the word on the stack
**
*/
	header("getc", GetCH, PutCH, getc)
getc:
	ldlp R0			; a place to put it
	ldl  CI			; console input
	ldc   1			; get 1 byte
	in			;
	ldl  R0			;
	push(DS, DINC)		; throw it on the stack
	j Next

/*
**
**  "putc"
**	Put a character from the top of the stack out to the
**	console.
**
*/
	header("putc", PutCH, NL, putc)
putc:
	pop(DS, DINC)		; get the byte
	stl  R0			; stash it away
	ldlp R0			; the address of the byte
	ldl  CO			; console output
	ldc   1			;
	out			; send it out
	j Next			;

/*
**
**  "<nl>"
**	Issue a newline to the console.
**
*/
	header("<nl>", NL, Space, Colon)

	ISEG
	dw Constant
	dw NEWLINE
	dw codep(PutCH)
	dw Constant
	dw CR
	dw codep(PutCH)
	dw Semi

/*
**
**  "<sp>"
**	Issue a space to the console.
**
*/
	header("<sp>", Space, Token, Colon)

	ISEG
	dw Constant
	dw BLANK
	dw codep(PutCH)
	dw Semi

/*
**
**  "TOKEN"
**	Place a pointer to the next token in the input stream based
**	on the separator found on top of the stack.  The token will
**	be placed in free memory and null terminated.
**
*/
	header("TOKEN", Token, Number, _token)
_token:
	pop(DS, DINC)		; get the token separator
	stl  R0			; and save it
	ldl  FP			; get the address of where to place the...
	stl  R1			; next character
	ldc   0			; clean out high bytes of R2 so that it...
	stl  R2			; works as a one character buffer.
	ldl  FLAG		; clear the F_EOL token-ended bit
	ldc  ~(F_EOL)		;
	and			;
	stl  FLAG		;

	/* skip leading separators */
tkn_lead:
	ldlp R2			; a place to put it
	ldl  CI			; where to get it
	ldc   1			; how many to get
	in			;
	ldl  R0			; is it a separator?
	ldl  R2			;
	diff			;
	cj   tkn_lead		; jump if they are equal...

	ldl  R2			; is it a tab?
	ldc  TAB		;
	diff			;
	cj   tkn_lead		; always skip tabs!

	ldl  R2			; is it a newline?
	eqc  NEWLINE		;
	cj   tkn_strt		; found token start if not newline

	ldl  FLAG		; A newline in execute mode is treated...
	ldc  F_MODE		; as a null-token, in compile mode it...
	and			; is ignored.
	cj   tkn_eol		; jump iff not compiling
	j    tkn_lead		; compiling, so ignore the newline

tkn_strt:
	ldl  R2			; the first character of the token...
	ldl  R1			;
	sb			;
	inc(R1)			;

	/* copy characters until reaching a separator */
tkn_bld:
	ldlp R2			; get another character
	ldl  CI			;
	ldc   1			;
	in			;

	ldl  R0			; is it a separator?
	ldl  R2			;
	diff			;
	cj   tkn_xit		; jump if it is a separator...

	ldl  R2			; if its a newline AND the separator is...
	eqc  NEWLINE		; a space, then end-of-token.
	ldl  R0			;
	eqc  BLANK		;
	and			;
	eqc  0			;
	cj   tkn_eol		;

	ldl  R2			; otherwise, store the character.
	ldl  R1			;
	sb			;
	inc(R1)			;
	j    tkn_bld		;

tkn_eol:
	ldl  FLAG		; set the Token-Ended-With-A-Newline bit
	ldc  F_EOL		;
	or			;
	stl  FLAG		;

	/* null-terminate the token and return */
tkn_xit:
	ldc 0			;
	ldl R1			;
	sb			;
	j   Next		;

/*
**
**  "NUMBER"
**	Try to convert the null-terminated ASCII string pointed to by
**	FP into a 32-bit number using the current number base.  If
**	successful, the number is pushed to the stack followed by
**	True, otherwise only False is pushed.
**
*/
	header("NUMBER", Number, Search, _number)
_number:
	ldl BASE		;
	stl R0			;
	ldc 0			;
	ldl R0			;
	gt			;
	cj  _Ncvt		; jump if base >= 0
	ldl R0			;
	not			;
	adc 1			;
	stl R0			;
_Ncvt:
	ldl  FP			; the string's address
	ldl  R0			; absolute value of conversion base
	call CVT_Str2Int	; try the conversion (Str2Int(base, buf))
	cj  _NaN		; jump if it couldn't be converted
	push(DS, DINC)		; it worked. Push the number
	ldc  1			; success
	push(DS, DINC)		;
	j Next			;
_NaN:
	ldc  0			; failure
	push(DS, DINC)		;
	j Next			;

/*
**
**  "SEARCH"
**	Search the dictionary for a word having a header which
**	"matches" the token pointed to by FP.  No paramaters are
**	needed on the stack on entry.  On exit, False is pushed
**	if the token could not be found.  If it was found, the
**	code address is pushed to the stack followed by True.
**
*/
	header("SEARCH", Search, Question, _srch)
_srch:
	ldc 0			; assume failure
	stl R0			;
	ldl DP			; our copy of the dictionary pointer
	stl R1			;
	ldc ~(F_IMMED)		; clear the F_IMMED bit
	ldl FLAG		;
	and			;
	stl FLAG		;
srch_eod:
	ldl R1			; have we reached the end of the list?
	cj  srch_xit		; jump if so
	ldl FP			; not done yet...try the token...
	ldl R1			; with the current entry
	call Compare		; courtesy of Greg
	stl R0			; save the flag...
	ldl R0			;
	cj  srch_no		; jump if not this entry...
	ldl R1			; we found it!
	adc 12			; offset to code address...
	push(DS, DINC)		; put it on the stack...
	ldl R1			; if it is an immediate word, set the bit
	adc  7			; in the FLAG register.
	lb			;
	ldc 0x80		;
	and			;
	cj  srch_xit		; jump if it isn't an immediate word
	ldc  F_IMMED		; it's an immediate word, set F_IMMED
	ldl  FLAG		;
	or			;
	stl  FLAG		;
	j srch_xit		; return (and put the flag on the stack)

srch_no:
	ldl R1			; didn't match...follow the link
	adc  8			; offset to link pointer
	ldnl 0			;
	stl R1			;
	j srch_eod		; try this entry...

srch_xit:
	ldl R0			; the success/failure flag
	push(DS, DINC)		;
	j Next			;

/*
**
**  "QUESTION"
**	If the byte pointed to by FP is a null, then we have reached the
**	end of the input line.
**	If it is anything else, then it is an unknown token so the address
**	of the "huh" string is put in WA and we exit to "Abort".
*/
	header("QUESTION", Question, Stack, _quest)
_quest:
	ldl  FP			; check for "null-ness"
	lb			;
	cj   Next		; if null, all is well.
	ldl  FP			; uh oh...we didn't handle that one...
	ldl  CO			; so echo print the offending token...
	call OutString		;
	ldl  SL			; and abort with a "huh?"
	ldnlp	huh		;
	stl  WA			;
	j    Abort		; restart

/*
**
**  "STACK"
**	Check the data stack for underflow.  Exits to "Abort" if
**	there are problems.
*/
	header("STACK", Stack, DotS, _stack)
_stack:
	ldl	SL		; check for data stack underflow...
	ldnlp	sss		;
	ldl	DS		;
	gt			; is top of stack > DS
	cj   Next		; no stack troubles.  everything is ducky.
	ldl  SL			; uh oh...stack underflow...
	ldnlp	resetmsg	;
	stl  WA			;
	j    Abort		;

/*
**
**  ".s"
**	Non-destructively print the contents of the data stack in the
**	current number radix.
*/
	header(".s", DotS, Prompt, _dots)
_dots:
	ldl  DS			; initialize a pointer to the stack
	adc -DINC		;
	stl  R0			;
_dots_l:
	ldl	SL		;
	ldnlp	sss		;
	ldl   R0		; is R0 < start of stack space?
	gt			;
	eqc   0			;
	cj    Next		; on to next word if reached bottom of stack
	ldl  FP			; build the string in free space above free
	adc  PAD		; space to skip over any good things left.
	ldl  BASE		; the conversion base
	ldl  R0			; the number
	ldnl  0			;
	call CVT_Int2Str	; convert it
	ldl  FP			; 
	adc  PAD		; string's address...
	ldl  CO			; output channel...
	call OutString		; print it...
	ldc  BLANK		; spacer...
	ldl  CO			;
	call OutChar		;
	ldl  R0			;
	adc -DINC		;
	stl  R0			;
	j    _dots_l		;

/*
**  "PROMPT"
**	Issue the prompt string if the F_EOL bit is set and
**	not in compile mode.
*/
	header("PROMPT", Prompt, Fetch, _prompt)
_prompt:
	ldl FLAG		; compile mode?
	ldc F_MODE		;
	and			;
	eqc 0			;
	cj  Next		; jump if in compile mode.
	ldl FLAG		; is the F_EOL bit on?
	ldc F_EOL		;
	and			;
	cj  Next		; jump if not set
	ldl SL			; its set, prompt
	ldnlp	prompt		;
	ldl  CO			;
	call OutString		;
	j    Next		;


/*
**
**  "@"
**	Replace the top stack entry by the contents of that memory
**	location.
*/
	header("@", Fetch, Store, _fetch)
_fetch:
	pop(DS, DINC)		; get the address
	ldnl 0			; get what it points to
	push(DS, DINC)		; put it on the stack
	j Next			;

/*
**
**  "!"
**	Store the next-on-stack entry at the memory address indicated
**	by the top of the stack.
*/
	header("!", Store, FplusS, _store)
_store:
	pop(DS, DINC)		; the place to put it
	pop(DS, DINC)		; get the word to store
	rev			;
	stnl 0			; store it
	j Next			;

/*
**
**  "+!"
**	Increment the contents of top-of-stack by next-on-stack.
*/
	header("+!", FplusS, FetchB, _FplusS)
_FplusS:
	pop(DS, DINC)		; target address
	stl R0			;
	ldl R0			; fetch it...
	ldnl 0			;
	pop(DS, DINC)		; increment value
	add			;
	ldl R0			; put it back
	stnl 0			;
	j Next

/*
**  "c@"
**	Fetch the 8 bit quantity (the "c" is historically for "character")
**	pointed to by the top of stack.
*/
	header("c@", FetchB, StoreB, _fetchb)
_fetchb:
	pop(DS, DINC)		; the address
	lb			; get the byte
	push(DS, DINC)		; put it on the stack
	j Next			;

/*
**
**  "c!"
**	Store the byte value at next-on-stack in the location pointed
**	to by the top-of-stack.
*/
	header("c!", StoreB, Fish, _storeb)
_storeb:
	pop(DS, DINC)		; the address to store it
	pop(DS, DINC)		; the byte value
	rev
	sb			; put it there
	j Next			;

/*
**
**  "FISH"
**	This calls the ROM and will (hopefully) completely
**	reset the machine.  It will not return.
*/
	header("FISH", Fish, Dup, _fish)
_fish:
	mint			;
	ldc 0x48		;
	add			;
	gajw			;
	ldc 0x7ffffffe		; address of FPS ROM reset
	gcall			; bye bye

/*
**  "dup"
**	Duplicate the top stack entry.
**
*/
	header("dup", Dup, Drop, _dup)
_dup:
	pop(DS, DINC)		; remember to fix this so its faster...
	stl R0			;
	ldl R0			;
	push(DS, DINC)		;
	ldl R0			;
	push(DS, DINC)		;
	j Next			;

/*
**  "drop"
**	Drop the top stack entry.
**
*/
	header("drop", Drop, Swap, _drop)
_drop:
	pop(DS, DINC)		; pretty simple, ay?
	j Next			;

/*
**
**  "swap"
**	Swap the two entries on the top of the stack.
**
*/
	header("swap", Swap, Over, _swap)
_swap:
	pop(DS, DINC)		;
	pop(DS, DINC)		;
	rev			;
	push(DS, DINC)		;
	push(DS, DINC)		;
	j Next			;

/*
**
**  "over"
**	Duplicates the second stack entry and pushes it to the top
**	of the stack.
*/
	header("over", Over, rLP, _over)
_over:
	pop(DS, DINC)		; top stack entry
	stl R0			;
	pop(DS, DINC)		; 2nd entry
	stl R1			;
	ldl R1			; restore 2nd
	push(DS, DINC)		;
	ldl R0			; restore top
	push(DS, DINC)		;
	ldl R1			; new copy
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$LP"
**	Task register accesss word.  Push the address of 'LP' to the stack.
**
*/
	header("$LP", rLP, rIP, _rLP)
_rLP:
	ldlp LP			; the address of LP
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$IP"
**	Task register accesss word.  Push the address of 'IP' to the stack.
**
*/
	header("$IP", rIP, rWA, _rIP)
_rIP:
	ldlp IP			; the address of IP
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$WA"
**	Task register accesss word.  Push the address of 'WA' to the stack.
**
*/
	header("$WA", rWA, rDS, _rWA)
_rWA:
	ldlp WA			; the address of WA
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$DS"
**	Task register accesss word.  Push the address of 'DS' to the stack.
**
*/
	header("$DS", rDS, rRS, _rDS)
_rDS:
	ldlp DS			; the address of DS
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$RS"
**	Task register accesss word.  Push the address of 'RS' to the stack.
**
*/
	header("$RS", rRS, rDP, _rRS)
_rRS:
	ldlp RS			; the address of RS
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$DP"
**	Task register accesss word.  Push the address of 'DP' to the stack.
**
*/
	header("$DP", rDP, rFP, _rDP)
_rDP:
	ldlp DP			; the address of DP
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$FP"
**	Task register accesss word.  Push the address of 'FP' to the stack.
**
*/
	header("$FP", rFP, rSSS, _rFP)
_rFP:
	ldlp FP			; the address of FP
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$SSS"
**	Task register accesss word.  Push the address of 'SSS' to the stack.
**
*/
	header("$SSS", rSSS, rSL, _rSSS)
_rSSS:
	ldlp SSS		; the address of SSS
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$SL"
**	Task register accesss word.  Push the address of 'SL' to the stack.
**
*/
	header("$SL", rSL, rFLAG, _rSL)
_rSL:
	ldlp SL			; the address of SL
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$FLAG"
**	Task register accesss word.  Push the address of 'FLAG' to the stack.
**
*/
	header("$FLAG", rFLAG, rBASE, _rFLAG)
_rFLAG:
	ldlp FLAG		; the address of FLAG
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$BASE"
**	Task register accesss word.  Push the address of 'BASE' to the stack.
**
*/
	header("$BASE", rBASE, rCI, _rBASE)
_rBASE:
	ldlp BASE		; the address of BASE
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$CI"
**	Task register accesss word.  Push the address of 'CI' to the stack.
**
*/
	header("$CI", rCI, rCO, _rCI)
_rCI:
	ldlp CI			; the address of CI
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$CO"
**	Task register accesss word.  Push the address of 'CO' to the stack.
**
*/
	header("$CO", rCO, Vlist, _rCO)
_rCO:
	ldlp CO			; the address of CO
	push(DS, DINC)		;
	j Next			;

/*
**
**  "vlist"
**	Print the contents of the current dictionary.  Output is of the
**	form "address name" where the number is the address
**	of the entry, and name are the ASCII characters identifying that
**	word.
**
*/
	header("vlist", Vlist, StrLen, _vlist)
_vlist:
	ldl DP			; initialize our private pointer
	stl R0			;
	ldc  0			; a cheap words-per-line counter
	ldl FP			;
	stnl 0			;
_vlst:
	ldl R0			; end-of-dictionary?
	cj  _vdone		; jump if so...
	ldl R0			; print the address of the entry...
	ldl CO			;
	call OutWrd		;
	ldc BLANK		;
	ldl CO			;
	call OutChar		;
	ldl R0			; print the name field
	adc  7			; offset to length byte...
	lb			;
	ldc 0x7f		; strip off the "IMMEDIATE" bit
	and			;
	stl R1			; R1 holds the length of the name...
	ldc  0			;
	stl R2			;
_vnam:
	ldl R2			; jump if printed all of name field...
	ldc  6			; only print 7 characters if there were
	gt			; more (that's all we keep around)
	eqc 0			;
	cj  _vnxt		;
	ldl R1			;
	ldl R2			;
	diff			;
	cj  _vpad		; fill in with spaces if less than 7 printed...
	ldl R0			; base address of entry...
	ldl R2			; byte offset...
	add			;
	lb			; retrieve the character
	ldl CO			;
	call OutChar		; and print it
	inc(R2)			;
	j _vnam			;
_vpad:
	ldc BLANK		;
	ldl CO			;
	call OutChar		;
	inc(R2)			;
	ldl R2			;
	ldc  6
	gt
	cj  _vpad

_vnxt:
	ldl FP			; increment the sleazy counter
	ldnl 0			;
	adc  1			;
	ldl FP			;
	stnl 0			;

	ldl FP			; need to start on a fresh line?
	ldnl 0			;
	ldc  4			;
	rem			;
	cj  _vnl		;
	ldc BLANK		; put some space inbetween each word
	ldl CO			;
	call OutChar		;
	ldc BLANK		;
	ldl CO			;
	call OutChar		;
_vlnk:
	ldl R0			; follow the link...
	adc  8			;
	ldnl 0			;
	stl R0			;
	j _vlst			;

_vnl:
	ldc NEWLINE
	ldl CO
	call OutChar
	ldc 0
	ldl FP
	stnl 0
	j _vlnk

_vdone:
	ldc NEWLINE		;
	ldl CO			;
	call OutChar		;
	j Next

/*
**
**  "strlen"
**	Compute the length of the null-terminated string pointed to
**	by the top of stack.
**
*/
	header("strlen", StrLen, Create, _strlen)
_strlen:
	pop(DS, DINC)		; get the address..
	call strlen		;
	push(DS, DINC)		;
	j Next			;

/*
**
**  "CREATE"
**	Make a partial dictionary entry composed of the next token
**	in the input stream.
**
**	COMPRESS THIS SOMEHOW -- FIX ME!
*/
	header("CREATE", Create, Const, Colon)

	ISEG
	dw Constant
	dw BLANK
	dw SetCompile
	dw codep(Token)
	dw ClrCompile
	dw codep(rFP)
	dw codep(Fetch)
	dw codep(StrLen)
	dw FPplus
	dw 7
	dw codep(StoreB)
	dw codep(rDP)
	dw codep(Fetch)
	dw FPplus
	dw 8
	dw codep(Store)
	dw codep(rFP)
	dw codep(Fetch)
	dw codep(rDP)
	dw codep(Store)
	dw FPplus
	dw 12
	dw codep(rFP)
	dw codep(Store)
	dw Semi

/*
**
**  "const"
**	Create a constant with the value found on the top of the stack.
**
*/
	header("const", Const, Var, Colon, IMMEDIATE)

	ISEG
	dw codep(Create)
	dw storeFP
	dw Konstant
	dw codep(Comma)
	dw Semi

/*
**
**  "var"
**	Create a variable initialized to the value found on the
**	top of the stack.
**
*/
	header("var", Var, Chan, Colon, IMMEDIATE)

	ISEG
	dw codep(Create)
	dw storeFP
	dw Variable
	dw codep(Comma)
	dw Semi

/*
**
**  "chan"
**	Create a channel and initialize it to NotProcess.p
**	(Its really only a variable with an initializer of
**	 MININT)
*/
	header("chan", Chan, MinInt, Colon)

	ISEG
	dw codep(MinInt)
	dw codep(Var)
	dw Semi

/*
**
**  "MinInt"
**	Put the minimum integer (NotProcess.p) on the stack.
**
*/
	header("MinInt", MinInt, Comma, _minint)
_minint:
	mint			; geese, this is too simple
	push(DS, DINC)		;
	j Next			;

/*
**
**  ","
**	Store the value found at the top of the stack in the word
**	pointed to by FP and increment FP.
**
*/
	/* header(",", Comma, Greater, _comma) */
	ISEG
	ALIGN 4
	EXPORT Comma
Comma:
	db ",", 0, 0, 0, 0, 0, 0, 1
	dw Greater
	dw _comma

	CSEG
_comma:
	pop(DS, DINC)		; get the value...
	ldl FP			;
	stnl 0			; and store it away
	incw(FP)		; bump up FP
	j Next			;

/*
**
**  ">"
**	Return true if next-on-stack is greater than top-of-stack.
**
*/
	header(">", Greater, Less, _greater)
_greater:
	pop(DS, DINC)		; top of stack
	pop(DS, DINC)		; next on stack
	rev			; get them in the right order
	gt			;
	push(DS, DINC)		;
	j Next			;

/*
**
**  "<"
**	Return true if next-on-stack is less than
**	top-of-stack.
**
*/
	header("<", Less, Equal, _less)
_less:
	pop(DS, DINC)		; top...
	pop(DS, DINC)		; next...
	gt			; they have implicitly reversed
	push(DS, DINC)		;
	j Next			;

/*
**
**  "="
**	Return true if the top two values on the stack are
**	equal.
**
*/
	header("=", Equal, ZEqual, _equal)
_equal:
	pop(DS, DINC)		;
	pop(DS, DINC)		;
	diff			;
	eqc 0			;
	push(DS, DINC)		;
	j Next			;

/*
**
**  "0="
**	Return true if the top value on the stack is equal to 0.
*/
	header("0=", ZEqual, LEqual, _zequal)
_zequal:
	pop(DS, DINC)		;
	eqc 0			;
	push(DS, DINC)		;
	j Next			;

/*
**
**  "<="
**	Return true if next-on-stack is less-than or equal-to
**	the top of the stack.
*/
	header("<=", LEqual, GEqual, _lequal)
_lequal:
	pop(DS, DINC)		;
	pop(DS, DINC)		;
	rev			;
	gt			;
	eqc 0			;
	push(DS, DINC)		;
	j Next			;

/*
**
**  ">="
**	Return true if next-on-stack is greater-than or equal-to
**	the top of the stack.
*/
	header(">=", GEqual, NEqual, _gequal)
_gequal:
	pop(DS, DINC)		;
	pop(DS, DINC)		;
	gt			;
	eqc 0			;
	push(DS, DINC)		;
	j Next			;

/*
**
**  "<>"
**	Return true if the top two stack elements are not equal.
*/
	header("<>", NEqual, Display, _nequal)
_nequal:
	pop(DS, DINC)		;
	pop(DS, DINC)		;
	diff			;
	eqc 0			;
	eqc 0			;
	push(DS, DINC)		;
	j Next			;
	
/*
**
**  "du"
**	Display bytes in the address space.  Top value on stack
**	is the starting address.  Next value on stack is the number
**	of bytes to display.  Output from (5 f5f5f5f5 d) might look like:
**
**	f5f5f5f5  41 42 43 44 45                     /ABCDE/
**
**	Leaves nothing on the data stack.
**
*/
	header("du", Display, HighQ, _display)
_display:
	pop(DS, DINC)		;get starting address
	stl	R0		;save it
	pop(DS, DINC)		;get byte count
	stl	R1		;save it

lines:
	ldc	0		;initialize byte count for new line
	stl	R2

	ldl	R1		;check if count has expired
	ldl	R2
	gt
	cj	quit

	ldl	R0		;print out the starting address
	ldl	CO
	call	OutWrd

	ldc	BLANK		;insert spacer
	ldl	CO
	call	OutChar

bytes:				;print data as hexadecimal bytes
	ldl	R1		;check if count has expired
	ldl	R2
	gt
	cj	filler

	ldc	BLANK		;insert spacer
	ldl	CO
	call	OutChar

	ldl	R0		;calculate address of next byte
	ldl	R2
	sum
	lb			;get and print next byte
	ldl	CO
	call	OutByt

	ldl	R2		;bump the bytes-this-line counter
	adc	1
	stl	R2

	ldl	R0		;Check for end of line
	ldl	R2
	sum
	ldc	0xf
	and
	cj	filler
	j	bytes

filler:				;print filler blanks
	ldl	R2		;save R2 on data stack
	push(DS, DINC)

	ldc	0x10		;calculate number of filler blanks to print
	ldl	R2
	diff
	ldc	3		;each byte occupies three cursor positions
	mul
	stl	R2

blanks:				;print blanks
	ldl	R2
	cj	ascii1
	ldl	R2
	adc	-1
	stl	R2
	ldc	BLANK
	ldl	CO
	call	OutChar
	j	blanks
	
ascii1:
	pop(DS,DINC)		;retrieve bytes-this-line counter
	stl	R2		;save it

	ldc	BLANK		;print spacers
	ldl	CO
	call	OutChar
	ldc	BLANK
	ldl	CO
	call	OutChar
	ldc	DELIMIT		;print delimiter
	ldl	CO
	call	OutChar

	ldl	R1		;decrement overall byte count
	ldl	R2
	diff
	stl	R1

chars:
	ldl	R2		;check bytes-this-line
	cj	ascii2

	ldl	R0		;check printable boundaries
	lb
	ldc	0x1f
	gt
	cj	period
	ldc	0x7f
	ldl	R0
	lb
	gt
	cj	period

	ldl	R0		;printable character
	lb
	ldl	CO
	call	OutChar
	j	bump

period:
	ldc	UNPRINTABLE	;print "unprintable" token
	ldl	CO
	call	OutChar

bump:				;bump pointer and counter
	ldl	R0
	adc	1
	stl	R0
	ldl	R2
	adc	-1
	stl	R2
	j	chars

ascii2:				;print second delimiter
	ldc	DELIMIT
	ldl	CO
	call	OutChar
	ldc	NEWLINE
	ldl	CO
	call	OutChar
	j	lines

quit:
	j	Next

/*
**
**  "$qp0"
**	Push the high priority queue pointers onto the stack. The
**	front pointer will be on the top of the stack.
**
*/
	header("$qp0", HighQ, LowQ, _highq)
_highq:
	ldlp R0			; a pointer to some space...
	saveh			; put the pointers into R0 and R1
	ldl  R1			; the back pointer
	push(DS, DINC)		;
	ldl  R0			; the front pointer
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$qp1"
**	Push the low priority queue pointers onto the stack. The
**	front pointer will be on the top of the stack.
**
*/
	header("$qp1", LowQ, Priority, _lowq)
_lowq:
	ldlp R0			; a pointer to some space...
	savel			; put the pointers into R0 and R1
	ldl  R1			; the back pointer
	push(DS, DINC)		;
	ldl  R0			; the front pointer
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$pri"
**	Push the current priority to the stack...
**
*/
	header("$pri", Priority, Timer, _pri)
_pri:
	ldpri			; load the current priority
	push(DS, DINC)		;
	j Next			;

/*
**
**  "$time"
**	Get the current value of Clock0 or Clock1, depending upon
**	the priority.
**
*/
	header("$time", Timer, Load, _timer)
_timer:
	ldtimer			; load the value of the timer
	push(DS, DINC)		;
	j Next			;

/*
**
**  "load"
**	Use the top of the stack as the address at which to load
**	bytes read from the console input channel.  The first 4
**	bytes of the load are taken to be the number of bytes to
**	read.
**
*/
	header("load", Load, dword, _load)
_load:
;	ldc  tty1data		; FIX ME! (save the current config)
;	ldnl ttyCTL		;
;	stl  R2			;
;	ldc  RAW		;
;	ldc  tty1data		;
;	stnl ttyCTL		;

	pop(DS, DINC)		; get the starting address
	stl  R0			;
	ldlp R1			; read the data length into R1
	ldl  CI			;
	ldc   1			;
	in			;
	ldlp R1			; R1[1]
	adc   1			;
	ldl  CI			;
	ldc   1			;
	in			;
	ldlp R1			; R1[2]
	adc   2			;
	ldl  CI			;
	ldc   1			;
	in			;
	ldlp R1			; R1[3]
	adc   3			;
	ldl  CI			;
	ldc   1			;
	in			;
_loadl:
	ldl R1			; check to see if have read all bytes
	cj  _loadxit		;
	ldl R0			; address to put the byte...
	ldl CI			; where to get it from...
	ldc  1			; get 1 byte...
	in			; go get it...
	inc(R0)			; bump the address pointer...
	ldl R1			; bump the byte counter...
	adc -1			;
	stl R1			;
	ldc 0			;
	cj _loadl		;
_loadxit:
;	ldl R2			; restore the tty settings
;	ldc tty1data		;
;	stnl ttyCTL		;
	j Next			;

/*
**
**  "dw"
**
**  Displays memory in word format.  Top word on stack is the starting
**  address.  Next thing on stack is the count of words to be printed.
**  Formatting rules are similar to "du" without the ascii field.
**
*/
	header("dw", dword, Runit, Colon)

	ISEG
	dw Constant			;truncate address to word boundary
	dw 0xfffffffc
	dw codep(Band)

line_loop:				;print a new line
	dw codep(Swap)
	dw codep(Dup)
	dw If

test_quit:				;check if counter has expired
	/*dw dw_quit - test_quit*/
	dw 28 * BPW

	dw codep(Swap)			;print line address
	dw codep(Dup)
	dw codep(DotX)

	dw codep(Space)
dw_loop:
	dw codep(Space)

	dw codep(Dup)			;print next data word
	dw codep(Fetch)
	dw codep(DotX)

	dw codep(PlusBPW)		;increment word pointer

	dw codep(Swap)			;decrement word count
	dw codep(Minus1)

	dw codep(Dup)
	dw If

test_counter:
	/*dw newline1 - test_counter*/	;counter expired, print newline
	dw 10 * BPW

	dw codep(Swap)			;check for line boundary
	dw codep(Dup)
	dw Constant
	dw 0xf
	dw codep(Band)
	dw If

test_bound:				;boundary reached, print newline
	/*dw newline2 - test_bound*/
	dw 4 * BPW

	dw Goto				;continue word loop

dw_continue:
	dw dw_loop - dw_continue

newline1:				;swap to get to known stack state
	dw codep(Swap)

newline2:				;print newline character
	dw codep(NL)

	dw Goto				;continue line loop

line_continue:
	dw line_loop - line_continue

dw_quit:
	dw codep(Drop)
	dw codep(Drop)
	dw Semi

/*
**
**  "run"
**	Top of stack is treated as a workspace pointer, next-on-stack
**	is assumed to be the instruction pointer.  The process described
**	by that is scheduled for execution.
**
*/
	header("run", Runit, _In, _run)
_run:
	pop(DS, DINC)		; workspace pointer
	stl R0			;
	pop(DS, DINC)		; instruction pointer
	ldl R0			;
	stnl -1			; store the iReg at wReg - 1
	ldl R0			;
	runp			; schedule it
	j Next			;


/*
**
**  "in"
**	Takes three parameters on the stack.  The topmost entry is
**	the number of bytes to transfer, the next entry is the
**	address of the channel to transfer through, the bottom
**	entry is address of the bytes to transfer.
**
*/
	header("in", _In, _Out, _in)
_in:
	pop(DS, DINC)		; number of bytes to transfer
	stl R0			;
	pop(DS, DINC)		; channel address
	stl R1			;
	pop(DS, DINC)		; data pointer
	ldl R1			;
	ldl R0			;
	in			;
	j Next			; pretty simple, no?

/*
**
**  "out"
**	Takes three parameters on the stack.  The topmost entry is
**	the number of bytes to transfer, the next entry is the
**	address of the channel to transfer through, the bottom
**	entry is address of the bytes to transfer.
**
*/
	header("out", _Out, _Move, _out)
_out:
	pop(DS, DINC)		; number of bytes to transfer
	stl R0			;
	pop(DS, DINC)		; channel address
	stl R1			;
	pop(DS, DINC)		; data pointer
	ldl R1			;
	ldl R0			;
	out			;
	j Next			; pretty simple, no?

/*
**
**  "move"
**	Copy a region of memory.  The third stack argument is assumed to
**	be the source address, 2nd stack entry is the destination, and the
**	top stack entry is the number of bytes to copy.
**
**	NB: destination must not overlap source (T414 restriction)
*/
	header("move", _Move, Sleep, _move)
_move:
	pop(DS, DINC)		; byte count
	stl R0			;
	pop(DS, DINC)		; destination address
	pop(DS, DINC)		; source address
	rev			; ...put operands in the right order
	ldl R0			;
	move			; do it
	j Next			;

/*
**
**  "sleep"
**	Reschedule the calling process after top-of-stack clock ticks.
*/
	header("sleep", Sleep, Comment, _sleep)
_sleep:
	pop(DS, DINC)		; get the time delay
	ldtimer			; current time
	sum			; "AFTER t"
	tin			;
	j Next			;

/*
**  "("
**
**	Begin of comment.  Skip everything until we find a ")".
*/
	/*header(left-parenthesis, Comment, Compile, Colon, IMMEDIATE)*/

	ISEG
	ALIGN 4
	EXPORT Comment
Comment:
	db 0x28, 0, 0, 0, 0, 0, 0, 1 + 128
	dw Compile
	dw Colon
	dw Constant
	dw 0x29
	dw codep(Token)
	dw Semi

/*
**  "?COMPILE"
**
**	Return TRUE if in compile mode.
*/
	header("?COMPILE", Compile, ResetC, _compilep)
_compilep:
	ldl FLAG
	ldc F_MODE
	and
	eqc F_MODE
	push(DS, DINC)
	j Next

/*
**  "resetch"
**	Reset the channel pointed to by the top-of-stack
**	and return the value found there.
*/
	header("resetch", ResetC, _Disc, _resetc)
_resetc:
	pop(DS, DINC)
	resetch
	push(DS, DINC)
	j Next

/*
**  "disc"
**	Report the results of a disable channel instruction,
**	using the top-of-stack as the channel address.
*/
	header("disc", _Disc, 0, _disc)
_disc:
	pop(DS, DINC)		; channel address
	ldc 1			; boolean TRUE
	ldl 0			; we don't care...
	ajw -1			; don't clobber anything...
	disc			;
	ajw  1			;
	push(DS, DINC)		;
	j Next
