/*
 *	Cornell Theory Center				tti
 *
 *	Copyright 1988 Cornell University
 *
 *	AKP
 *
 *	Function:	- execution support for numeric and string constants
 *
 *	$Log:	literal.ts,v $
Revision 1.1  88/08/19  15:35:47  gdburns
Initial revision

 */


#include "tti.h"


	IMPORT	Next, Semi, _semi, Boot
	EXPORT	storeFP, FPplus
	EXPORT  Constant, Literal, Konstant, Variable
	EXPORT  SetCompile, ClrCompile
	EXPORT	String

/*
**  In the interests of space, these two were coded this way
**  rather than as lengthy secondaries.  They are used by
**  ":" and ";" to respectively set and clear the compile bit.
**
*/

_comp1:
	ldc  F_MODE	; turn on compilation...
	ldl  FLAG	;
	or		;
	stl  FLAG	;
	j    Next	;

_comp0:
	ldc  ~(F_MODE)	; turn off compilation...
	ldl  FLAG	;
	and		;
	stl  FLAG	;
	j    Next	;

/*
**
**  For numeric constants, the value is stored one memory cell after
**  the address of this routine.  All we have to do is load what the IP
**  points at and increment IP by one word.
**
**  NB: Does not have a header as the run-time support is not needed
**      (or reasonable) to have available to a user.
*/

_cons:
	ldl  IP		; the address of the constants value 
	ldnl 0		; the constant's value
	push(DS, DINC)	; put it on the data stack
	incw(IP)	; bump IP over the constant
	j Next		; on to the next word...

/*
**
**  For variables, the value of the variable is stored one memory
**  cell after the address of this routine.  We don't care about
**  what's stored there, all we have to do is push the value of
**  the current WA (ie the address of the variable) to the stack.
*/
Variable:
	ldl WA		; the address of the variable
	push(DS, DINC)	; throw it on the stack...
	j Next		; pretty simple!

/*
**
**  Runtime support of user defined constants is very similar to that
**  of variables.
*/
Konstant:
	ldl WA		; the address of the constant
	ldnl 0		; its value
	push(DS, DINC)	;
	j Next		;

/*
**  If compiling a constant into a word, and it isn't already defined
**  as a constant (it happens, you know), then the address of "Constant"
**  has to be put there first, followed by the number.  Its kind of
**  ugly.
*/

_literal:
	ldl  FLAG	; are we compiling?
	ldc  F_MODE	;
	and		;
	cj   Next	; jump if the bit isn't set
	ldl  SL		; compiling...put in the pointer to "Constant"
	ldnlp	Constant;
	ldl  FP		;
	stnl 0		;
	incw(FP)	;
	pop(DS, DINC)	; the actual constant
	ldl  FP		;
	stnl 0		;
	incw(FP)	;
	j    Next	; dats all there is!

/*
**
**  It turned out that I needed runtime support for storing an
**  assemble time constant where FP points (ie Colon, Semi, etc). FP is
**  adjusted accordingly.  This is actually a hack to keep down
**  the size of predefined secondaries...
*/

_storeFP:
	ldl IP		; get the address of the constant
	ldnl 0		; get the value
	ldl FP		;
	stnl 0		; stash it away
	incw(FP)	; bump up FP
	incw(IP)	;
	j Next		;

/*
**  It also is convenient to be able to add an assemble time constant
**  to FP without going through the whole FP, fetch, constant, x,
**  business.
*/

_FPplus:
	ldl IP		; the constant's address
	ldnl 0		; the value
	ldl FP		;
	add		;
	push(DS, DINC)	;
	incw(IP)	;
	j Next		;

/*
**  For string constants, the string is stored starting at the word
**  following the pointer to this routine.  Print until reaching a
**  null byte, then bump IP to the next word boundary.
*/

	IMPORT OutChar
_prtit:
	ldl  IP		; check for nullness...or print it.
	lb		;
	stl  R0
	ldl  R0
	cj   _strnd
	ldl  R0
	ldl  CO
	call OutChar
	inc(IP)
	j    _prtit
_strnd:
	ldl  IP
	ldc -BPW
	and
	adc  BPW
	stl  IP
	j    Next


	ISEG

SetCompile:
	dw _comp1
ClrCompile:
	dw _comp0
Constant:
	dw _cons
Literal:
	dw _literal
storeFP:
	dw _storeFP
FPplus:
	dw _FPplus
String:
	dw _prtit
