/*
 *	Cornell Theory Center				tti
 *
 *	Copyright 1988 Cornell University
 *
 *	AKP
 *
 *	Function:	- helper subroutines...
 *
 *	$Log:	subrs.ts,v $
Revision 1.1  88/08/19  15:35:49  gdburns
Initial revision

 */


#include "tti.h"


/*
**
**  strlen
**
**	Count the non-null characters in the ASCII string pointed
**	to by aReg and return the count in aReg.  Registers are
**	not preserved.
**
*/
	EXPORT strlen
strlen:
	ldc 0			; initialize the counter
	stl bReg		;
strlup:
	ldl aReg		; have we reached the end-of-string?
	lb			;
	cj  strret		; jump if it is a nul
	inc(aReg)		; bump the pointer
	inc(bReg)		; bump the counter
	j strlup		;
strret:
	ldl bReg		; load up the conter's value
	ret			;

/*
**  OutChar
**
**	Use the value in aReg as the channel on which to output
**	the character in bReg.
**
*/
	EXPORT OutChar
OutChar:
	ldlp bReg		; pointer...
	ldl  aReg		; channel...
	ldc  1			; send 1 byte...
	out			;
	ret			;

/*
**
**  OutString
**
**	Use the value in aReg as the channel on which to output the
**	null-terminated ASCII string pointed to by aReg.  Registers
**	are not preserved.
**
*/
	EXPORT OutString
OutString:
	ldl bReg		; are we done yet?
	lb			;
	cj  OutS_ret		; return if the byte is a null
	ldl bReg		; pointer...
	ldl aReg		; channel...
	ldc 1			; send 1 byte...
	out			;
	inc(bReg)		; bump up the pointer...
	j OutString		; send the next byte
OutS_ret:
	ret			;

/*
**
**  OutNybble
**
**	Use the value in aReg as the address of the channel on which
**	to output the nybble (ie a 4-bit value) in bReg.  The nybble
**	is converted to the corresponding ASCII hexidecimal character.
**	Registers are not preserved.
**
*/
	EXPORT OutNybble
OutNybble:
	ldl bReg		; will it simply be a digit (0-9)
	ldc 0x9			;
	gt			;
	cj  OtN_add0		; jump if it's <= 9
	ldl bReg		; not a digit...bump it into A-F
	adc 0x7			;
	stl bReg		;
OtN_add0:
	ldl  bReg		; add '0' to it
	adc  '0'		;
	stl  bReg		;
	ldlp bReg		; output it...address of the byte
	ldl  aReg		; channel...
	ldc  1			; send 1 byte...
	out			;
	ret			;


/*
**
**  OutByt
**
**	Use the value in aReg as the channel on which to output the
**	byte value in bReg.  The value in bReg is converted to two
**	ASCII characters.
**
*/
	EXPORT OutByt
OutByt:
	ldl  bReg		; the value
	ldc  0xff		;
	and			;
	stl  bReg		; only do the least significant byte
	ldl  bReg		;
	ldc  4			;
	shr			;
	ldl  aReg		;
	call OutNybble		;
	ldl  bReg		; now do the least significant nybble
	ldc  0xf		;
	and			;
	ldl  aReg		;
	call OutNybble		;
	ret			;

/*
**
**  OutWrd
**
**	Use the value in aReg as the channel address on which to output
**	the 32-bit word in bReg.  The value in bReg is converted into
**	8 ASCII hexidecimal characters.  Registers are not preserved.
**
*/
	EXPORT OutWrd
OutWrd:
	ldc  28			; number of bits to shift...
	stl  cReg		;
_owordl:
	ldl  bReg		; do each nybble of the value
	ldl  cReg		;
	shr			;
	ldc  0xf		;
	and			;
	ldl  aReg		;
	call OutNybble		;
	ldl  cReg		; if there are no more nybbles then...
	cj   _owxit		;
	ldl  cReg		;
	adc  -4			;
	stl  cReg		;
	ldc  0			; don't deschedule...
	cj   _owordl		;
_owxit:
	ret
