/*
 *	Cornell Theory Center				tti
 *
 *	Copyright 1988 Cornell University
 *
 *	AKP
 *
 *	Function:	- task register table
 *			- offsets from local workspace
 *
 *	$Log:	tti.h,v $
 * Revision 1.1.1.1  88/09/29  21:34:27  gdburns
 * Use a private copy of T414.h.
 * This makes tti a stand-alone tool independent of Trillium.
 * 
 * Revision 1.1  88/08/19  15:35:52  gdburns
 * Initial revision
 * 
 */

#ifndef T414
#include "T414.h"
#endif

/*
** Word offsets from the workspace pointer for each interpreter
** task.  aka "The Task Register Table"
**
**	LP	Load Point		Absolute address of first byte
**	IP	Instruction Pointer	Interpreter's instruction pointer
**	WA	Word Address		Holding register for word address
**	DS	Data Stack		Data Stack pointer
**	RS	Return Stack		Return Stack pointer
**	DP	Dictionary Pointer	pointer to first word in dictionary
**	FP	Free Space Pointer	pointer to next free byte
**	SSS	Stack Space Size	no. of bytes (holds both stacks)
**	SL	Static Link		pointer to data segment
**	FLAG	Flag bits		state of interpreter
**	BASE	I/O number base		current system radix
**	CI	Console IN		console input channel
**	CO	Console OUT		console output channel
**	R0	temporary		not preserved from word to word
**	R1	temporary
**	R2	temporary
**	PAD	constant		area above the current FP
**
*/
#define LP   0x0
#define IP   0x1
#define WA   0x2
#define DS   0x3
#define RS   0x4
#define DP   0x5
#define FP   0x6
#define SSS  0x7
#define SL   0x8
#define FLAG 0x9
#define BASE 0xa
#define CI   0xb
#define CO   0xc
#define R0   0xd
#define R1   0xe
#define R2   0xf
#define PAD  128

#define TRTSIZE	0x10

/*
**  FLAG bits
*/
#define F_MODE		0x01
#define	F_ERROR		0x02
#define F_STACK		0x04
#define F_RESET		0x08
#define F_IMMED		0x10
#define F_EOL		0x20
#define F_RSVD1		0x40
#define F_RSVD2		0x80

/*
**  Stack pointer increments
**
**	DINC	Data Stack Increment	grow toward hi memory
**	RINC	Return Stack Increment	grow toward lo memory
*/
#define DINC		BPW
#define RINC	       -BPW

/*
** Space between Task Register Tables to provide room for
** subroutine calls and Transputer housekeeping.
*/
#define TRTGAP		0x400

/*
 * Size of free space (ISEG).
 */

#define FSSize		0x1000

/*
 * Size of data/return stack (ISEG).
 */

#define SSSize		0x100
