/*
 *	Cornell Theory Center				tti
 *
 *	Copyright 1988 Cornell University
 *
 *	AKP
 *
 *	Function:	- startup code
 *			- the inner interpreter
 *
 *			- converted to conform with the current assembler
 *			  "tas" by GDB
 *
 *			- Breathing new life into tti revolved around three
 *			  main issues:  1) syntactic differences with tas
 *			  2) lack of some convenient features in tas (like
 *			  replicated initialization "5#0")  3) tti's assumption
 *			  that all data/code was addressable from the load
 *			  point of the program.
 *			
 *			  The LP register is no longer used and the DSS register
 *			  was never used.  We now have a new register which
 *			  is the familiar static link of the process.  The
 *			  entire dictionary is in the data segment (ISEG).
 *			  We now link in crt0 and rely on other tools to
 *			  properly initialize our workspace/instruction ptr/
 *			  static link.   (GDB)
 *
 *	$Log:	tti.ts,v $
Revision 1.1  88/08/19  15:35:50  gdburns
Initial revision

 */


#include "tti.h"


	NAME	TTI

	IMPORT	Dictionary
	IMPORT	Outer, startmsg, resetmsg
	IMPORT	TTIEND
	IMPORT	fss, sss, ssstop
	EXPORT  Boot, _main
	EXPORT	Colon, Semi, _semi, Next, Run
	EXPORT	Start, Abort

	CSEG

_main:
	ldl	1		; address of data segment
	ajw	-TRTSIZE	; make room for TRT
	stl	SL		; save as "static link"

/*
 * Build the initial Task Register Table.
 */
Boot:
	ldc	Link1In		; set console input from channel tty0In
	stl	CI		;
	ldc	Link1Out	; set console output to channel tty0Out
	stl	CO		;
	ldc	F_RESET		; make sure that we do a cold start...
	stl	FLAG		; ...a 0 means do a warm start.
	ldc	SSSize		; initial stack space in bytes
	stl	SSS		;
	ldl	SL		; the start of the dictionary
	ldnlp	Dictionary	;
	stl	DP		;
Start:
	ldl	SL		; (we assume a warm start...)
	ldnlp	resetmsg	;
	stl	WA		; ...and put it in WA

	ldl	FLAG		; check the reset flag bit
	ldc	F_RESET		;
	and			; 1 = cold start / 0 = warm start
	cj	Abort		; jump if a warm start

	ldc	0x10		; Okay, cold start. Hex is default base.
	stl	BASE		;

	ldl	SL		; initialize the free space pointer
	ldnlp	fss		;
	stl	FP		; points to just after the input buffer

	ldl	SL		; Cold start message
	ldnlp	startmsg	;
	stl	WA		;

	ldc	0		; clear all flag bits to insure warm...
	stl	FLAG		; start (unless the bit is flipped)

	/* warm start entry point */

Abort:
	ldl	SL		; initialize the data stack pointer
	ldnlp 	sss		;
	stl	DS		;

	ldl	SL		; initialize the return stack pointer
	ldnlp	ssstop		;
	adc	-BPW		; just below top of combined stack space
	stl	RS		;

	ldc	(~(F_MODE|F_ERROR|F_STACK|F_RESET|F_IMMED)) | F_EOL
	stl	FLAG		;

	ldl	SL		; set the IP to the outer interpreter
	ldnlp	Outer		;
	stl	IP		;
	ldl	WA		; save the error message address on DS
	push(DS, DINC)		;
	j	Next		; ...and away we go!


	/* The inner sanctum... */
Colon:
        ldl	IP		;
        push(RS, RINC)		; PUSH IP
        ldl	WA		;
        stl	IP		; IP := WA
        j	Next		; jump to Next

_semi:
        pop(RS, RINC)		;
        stl	IP		; POP IP

Next:
        ldl	IP		;
	ldnl	0		;
	stl	WA		;
        incw(IP)                ; WA := (IP+)

Run:
        ldl	WA		; its a register stack machine, right?
	ldnl	0		;
        incw(WA)		;
        gcall			; jump (WA+)

	ISEG
Semi:
	dw	_semi		;
